#!/usr/bin/perl -W
#--------------------------------------------------------------
# This is a set of functions to deal with beam attenuation
#
#     Authors: Satish Devarapally and Sergey Stepanov
#
# Version-1.0:  2006/03/17 using Pezca and the simplex recursive algorithm
#                          for matching requested attenulation (Satish).
# Version-2.0:  2008/02/11 added reading foils from mySQL
# Version-2.1:  2008/02/11 removed exceptions from the calls (simplified)
# Version-2.2:  2009/05/20 added Au
# Version-2.3:  2010/02/24 changed the order filters are commanded IN/OUT:
#                          first all needed IN, then all needed OUT
# Version-2.4:  2010/06/07 added taking out the foils that are marked 'unused';
#                          for that purposed added fixAttenuationExceptions
# Version-2.5:  2010/06/08 changed to reading rb PVs instead of bo for the
#                          filters status.
# Version-2.6:  2011/01/25 added CA-monitors function
#--------------------------------------------------------------
# 1a. ($status,$attenuation_actual) = &setAttenuation1($attenuation);
#                                                        # automatically finds energy.
#                                                        # By default attenuation is %,
#                                                        # but that can be changed to x
#                                                        # with global $ATTENUATION_USE_FACTOR=1
#
# 1b. ($status,$attenuation_actual) = &setAttenuation2($attenuation,$energyKeV,[$report]);
#                                                        # By default attenuation is %,
#                                                        # but that can be changed to x
#                                                        # with global $ATTENUATION_USE_FACTOR=1
#
# 2a. ($status,$attenuation_actual) = &getAttenuation1();
#                                                        # automatically finds energy
#                                                        # By default attenuation is %,
#                                                        # but that can be changed to x
#                                                        # with global $ATTENUATION_USE_FACTOR=1
#
# 2b. ($status,$attenuation_actual) = &getAttenuation2($energyKeV);
#                                                        # By default attenuation is %,
#                                                        # but that can be changed to x
#                                                        # with global $ATTENUATION_USE_FACTOR=1
#
# 3a.              ($status,[$msg]) = &makeAttenuatorDB_DESC([\@material,\@thinkness,\@exceptions]);	#optionally fills local arrays as copies of global
#                                                        # fills global arrays:
#                                                        # @Attenuators_material,
#                                                        # @Attenuators_thickness,
#                                                        # @attenuatorExceptions
# 3b.              ($status,[$msg]) = &makeAttenuatorDB_mySQL([\@material,\@thinkness,\@exceptions]);	#optionally fills local arrays as copies of global
#                                                        # fills global arrays:
#                                                        # @Attenuators_material,
#                                                        # @Attenuators_thickness,
#                                                        # @attenuatorExceptions
# 4.                                  &printAttenuatorDB();
# 5.                        $status = &makeF2DB($material,\@EnergyTable,\@F2Table,\$Density,\$AtomicWeight);
#                                                        # fills global arrays for material=Al, Ag, or Au :
#                                                        # @xxEnergyTable,
#                                                        # @xxF2Table,
#                                                        # $xxMaterialDensity,
#                                                        # $xxAtomicWeight
# 6.                        $status = &readF2DB($material,\@EnergyTable,\@F2Table);
# 7.               ($status,$state) = &getAttenuatorState($channel);			#1=IN, 0=OUT
# 8.                        $status = &setAttenuatorState($channel,$state);             #1=IN, 0=OUT
# 9.                  ($status,$Mu) = &calculateMu($material,$energykeV);
#10.                 ($status,$MuT) = &calculateFilterMuT($material,$thickness,energykeV);
#11.                   $requiredMuT = &calculateRequiredMuT($requestedAttenuation);
#12.                   $attenuation = &calculateAttenuation(\@combination,\@attenuatorsMuT);
#13.                @sortedPosition = &sortArrayDescending(\@elements);
#14.   ($deltaMuT,@originalPattern) = &sortAndSearchAttenuation($targetMuT,@MuTList);
#15.     ($deltaMuT,@sortedPattern) = &searchAttenuation($targetMuT,@sortedMuTList);
#16.                        $status = &fixAttenuationExceptions();
#17.                        $status = &setAttenuationMonitors();
#--------------------------------------------------------------
# use warnings::unused;
  use strict;
# use Pezca;
# use DBI;

### Required global import:
  our ($beamline, $attenuatorPV, $attenuatrbPV, $nAttenuators);

### If the following equals '1', the attenuation is calculated/reported
### as a factor; otherwise it is reported as a percent.
  our $ATTENUATION_USE_FACTOR;
  our $MAX_FACTOR_ATTENUATION = 1.E9;

  our (@Attenuators_material, @Attenuators_thickness);
  our (@alEnergyTable, @alF2Table, $alMaterialDensity, $alAtomicWeight);
  our (@agEnergyTable, @agF2Table, $agMaterialDensity, $agAtomicWeight);
  our (@auEnergyTable, @auF2Table, $auMaterialDensity, $auAtomicWeight);
  our (@attenuatorExceptions, $nAttenuatorExceptions);

#------------------------------------------------------------------------------------------
# our $attenuatorPV   ='bo0:ch';	# this may be beamline specific! Moved to beamline.h
# our $attenuatrbPV   ='rb0:ch';	# this may be beamline specific! Moved to beamline.h
# our $attenuatorPV   ='23d:bo0:ch';	# this may be beamline specific! Moved to beamline.h
# our $nAttenuators   = 16;		# this may be beamline specific! Moved to beamline.h
# our $Attenuators_IN = 0;		# this may be beamline specific! Moved to beamline.h
# @alEnergyTable     = qw( 3010.00 3024.63 3069.00 3073.55 3123.26 3129.00 3173.78 3190.00 3225.11 3253.00 3277.27 3316.00 3330.28 3381.00 3384.15 3438.88 3447.00 3494.50 3515.00 3551.02 3583.00 3608.46 3653.00 3666.82 3725.00 3726.13 3786.40 3847.64 3872.00 3909.87 3948.00 3973.11 4025.00 4037.38 4102.68 4169.03 4236.46 4266.00 4304.98 4349.00 4374.62 4434.00 4445.37 4517.27 4590.33 4664.58 4740.03 4791.00 4816.69 4885.00 4894.60 4973.77 4980.00 5054.21 5135.96 5219.03 5278.00 5303.44 5381.00 5389.22 5476.39 5564.97 5594.00 5654.98 5746.44 5815.00 5839.39 5928.00 5933.83 6029.81 6127.33 6226.44 6327.15 6406.00 6429.48 6531.00 6533.48 6639.15 6746.54 6855.65 6922.00 6966.54 7057.00 7079.22 7193.72 7310.07 7428.31 7548.45 7625.00 7670.54 7775.00 7794.61 7920.68 8048.79 8178.98 8311.26 8401.00 8445.69 8565.00 8582.29 8721.11 8862.16 9005.50 9077.00 9151.16 9255.00 9299.17 9436.00 9449.58 9602.42 9757.73 9915.55 10000.00 10075.90 10200.00 10238.90 10390.00 10404.50 10572.80 10743.80 10917.60 11020.00 11094.20 11230.00 11273.60 11450.00 11455.90 11641.20 11829.50 11900.00 12020.80 12140.00 12215.30 12370.00 12412.80 12613.60 12817.60 13025.00 13110.00 13235.60 13370.00 13449.70 13630.00 13667.20 13888.30 14112.90 14341.20 14450.00 14573.10 14730.00 14808.90 15020.00 15048.40 15291.80 15310.00 15539.10 15610.00 15790.40 15920.00 16045.80 16230.00 16305.40 16540.00 16569.10 16837.10 16870.00 17109.40 17200.00 17386.10 17530.00 17667.40 17880.00 17953.10 18230.00 18243.50 18538.60 18580.00 18838.40 18950.00 19143.10 19320.00 19452.70 19690.00 19767.40 20080.00 20087.10 20412.00 20470.00 20742.10 20870.00 21077.60 21280.00 21418.50 21700.00 21765.00 22117.00 22474.70 22550.00 22838.20 22990.00 23207.60 23440.00 23583.00 23900.00 23964.40 24352.00 24745.90 24850.00 25146.20 25330.00 25552.90 25830.00 25966.20 26330.00 26386.10 26812.90 26850.00 27246.60 27370.00 27687.30 27910.00 28135.10 28450.00 28590.20 29010.00 29052.60 29522.50 29580.00 30000.00 30150.00 30740.00 31340.00 31960.00 32580.00 33220.00 33870.00 34530.00 35210.00 35890.00 36600.00 37310.00 38040.00 38790.00 39540.00 40320.00 );
# @alF2Table         = qw( 1.489000 1.480330 1.441000 1.439610 1.399890 1.394000 1.361160 1.348000 1.323390 1.304000 1.286570 1.261000 1.250640 1.220000 1.215640 1.181540 1.180000 1.148310 1.141000 1.115910 1.103000 1.084360 1.066000 1.053630 1.030000 1.023690 0.994542 0.966167 0.962200 0.938535 0.929700 0.911645 0.898300 0.885448 0.859966 0.835163 0.811033 0.809600 0.787559 0.781900 0.764708 0.755100 0.742491 0.720878 0.699839 0.679399 0.659516 0.655900 0.640176 0.633000 0.621389 0.603106 0.610900 0.585342 0.568080 0.551279 0.548700 0.534968 0.529300 0.519116 0.503715 0.488726 0.492400 0.474187 0.460062 0.457900 0.446336 0.441500 0.433003 0.420026 0.407459 0.395224 0.383348 0.381100 0.371815 0.367200 0.360616 0.349739 0.339182 0.328927 0.328500 0.318973 0.316600 0.309309 0.299926 0.290817 0.281974 0.273391 0.272800 0.265058 0.262700 0.256972 0.249121 0.241506 0.234114 0.226938 0.226000 0.219978 0.217600 0.213218 0.206663 0.200304 0.194131 0.194100 0.188140 0.186900 0.182331 0.179900 0.176693 0.171229 0.165921 0.160780 0.160400 0.155788 0.154400 0.150945 0.148600 0.146248 0.141693 0.137274 0.132988 0.132300 0.128831 0.127300 0.124800 0.122500 0.120890 0.117099 0.113422 0.113300 0.109856 0.109000 0.106398 0.104800 0.103045 0.997941E-01 0.966416E-01 0.935848E-01 0.932300E-01 0.906209E-01 0.896400E-01 0.877473E-01 0.861800E-01 0.849611E-01 0.822598E-01 0.796408E-01 0.771018E-01 0.765700E-01 0.746402E-01 0.735900E-01 0.722539E-01 0.707300E-01 0.699405E-01 0.676979E-01 0.679800E-01 0.655240E-01 0.653300E-01 0.634165E-01 0.627800E-01 0.613737E-01 0.603200E-01 0.593935E-01 0.579600E-01 0.574741E-01 0.556136E-01 0.556800E-01 0.538102E-01 0.535000E-01 0.520623E-01 0.513900E-01 0.503681E-01 0.493600E-01 0.487260E-01 0.474100E-01 0.471345E-01 0.455920E-01 0.455400E-01 0.440970E-01 0.437300E-01 0.426482E-01 0.420000E-01 0.412440E-01 0.403300E-01 0.398831E-01 0.387200E-01 0.385533E-01 0.372771E-01 0.371800E-01 0.360423E-01 0.356900E-01 0.348475E-01 0.342700E-01 0.336915E-01 0.328900E-01 0.325731E-01 0.314909E-01 0.304440E-01 0.303000E-01 0.294311E-01 0.290800E-01 0.284511E-01 0.279100E-01 0.275030E-01 0.267800E-01 0.265858E-01 0.256985E-01 0.248401E-01 0.246600E-01 0.240097E-01 0.236600E-01 0.232064E-01 0.227000E-01 0.224293E-01 0.217800E-01 0.216776E-01 0.209505E-01 0.208900E-01 0.202471E-01 0.200500E-01 0.195668E-01 0.192400E-01 0.189087E-01 0.184600E-01 0.182722E-01 0.177200E-01 0.176565E-01 0.170611E-01 0.170000E-01 0.164852E-01 0.163100E-01 0.156500E-01 0.150100E-01 0.144000E-01 0.138100E-01 0.132500E-01 0.127100E-01 0.121900E-01 0.116900E-01 0.112100E-01 0.107500E-01 0.103100E-01 0.989000E-02 0.948400E-02 0.909400E-02 0.872000E-02 );
# $alMaterialDensity = 2.7;
# $alAtomicWeight    = 26.98;
# @agEnergyTable     = qw( 3010.00 3024.63 3069.00 3073.55 3123.26 3129.00 3173.78 3190.00 3225.11 3253.00 3277.27 3316.00 3330.28 3338.00 3339.00 3340.00 3341.00 3342.00 3343.00 3344.00 3345.00 3346.00 3347.00 3348.00 3349.00 3350.00 3351.00 3351.10 3351.20 3352.00 3353.00 3354.00 3355.00 3356.00 3357.00 3358.00 3359.00 3360.00 3361.00 3362.00 3363.00 3364.00 3381.00 3384.15 3438.88 3447.00 3494.50 3512.00 3513.00 3514.00 3515.00 3516.00 3517.00 3518.00 3519.00 3520.00 3521.00 3522.00 3523.00 3523.60 3523.80 3524.00 3525.00 3526.00 3527.00 3528.00 3529.00 3530.00 3531.00 3532.00 3533.00 3534.00 3535.00 3536.00 3537.00 3538.00 3539.00 3540.00 3551.02 3583.00 3608.46 3653.00 3666.82 3725.00 3726.13 3786.40 3792.00 3793.00 3794.00 3795.00 3796.00 3797.00 3798.00 3799.00 3800.00 3801.00 3802.00 3803.00 3804.00 3805.00 3805.70 3805.90 3806.00 3806.50 3807.00 3808.00 3809.00 3810.00 3811.00 3812.00 3813.00 3814.00 3815.00 3816.00 3817.00 3818.00 3819.00 3820.00 3821.00 3822.00 3823.00 3824.00 3847.64 3872.00 3909.87 3948.00 3973.11 4025.00 4037.38 4102.68 4104.00 4169.03 4184.00 4236.46 4266.00 4304.98 4349.00 4374.62 4434.00 4445.37 4517.27 4521.00 4590.33 4609.00 4664.58 4699.00 4740.03 4791.00 4816.69 4885.00 4894.60 4973.77 4980.00 5054.21 5078.00 5135.96 5177.00 5219.03 5278.00 5303.44 5381.00 5389.22 5476.39 5487.00 5564.97 5594.00 5654.98 5703.00 5746.44 5815.00 5839.39 5928.00 5933.83 6029.81 6044.00 6127.33 6163.00 6226.44 6283.00 6327.15 6406.00 6429.48 6531.00 6533.48 6639.15 6659.00 6746.54 6789.00 6855.65 6922.00 6966.54 7057.00 7079.22 7193.72 7195.00 7310.07 7336.00 7428.31 7479.00 7548.45 7625.00 7670.54 7775.00 7794.61 7920.68 7927.00 8048.79 8082.00 8178.98 8240.00 8311.26 8401.00 8445.69 8565.00 8582.29 8721.11 8732.00 8862.16 8903.00 9005.50 9077.00 9151.16 9255.00 9299.17 9436.00 9449.58 9602.42 9620.00 9757.73 9808.00 9915.55 10000.00 10075.90 10200.00 10238.90 10390.00 10404.50 10572.80 10600.00 10743.80 10810.00 10917.60 11020.00 11094.20 11230.00 11273.60 11450.00 11455.90 11641.20 11680.00 11829.50 11900.00 12020.80 12140.00 12215.30 12370.00 12412.80 12613.60 12620.00 12817.60 12860.00 13025.00 13110.00 13235.60 13370.00 13449.70 13630.00 13667.20 13888.30 13900.00 14112.90 14170.00 14341.20 14450.00 14573.10 14730.00 14808.90 15020.00 15048.40 15291.80 15310.00 15539.10 15610.00 15790.40 15920.00 16045.80 16230.00 16305.40 16540.00 16569.10 16837.10 16870.00 17109.40 17200.00 17386.10 17530.00 17667.40 17880.00 17953.10 18230.00 18243.50 18538.60 18580.00 18838.40 18950.00 19143.10 19320.00 19452.70 19690.00 19767.40 20080.00 20087.10 20412.00 20470.00 20742.10 20870.00 21077.60 21280.00 21418.50 21700.00 21765.00 22117.00 22120.00 22474.70 22550.00 22838.20 22990.00 23207.60 23440.00 23583.00 23900.00 23964.40 24352.00 24370.00 24745.90 24800.00 24810.00 24820.00 24830.00 24840.00 24850.00 24860.00 24870.00 24880.00 24890.00 24900.00 24910.00 24920.00 24930.00 24940.00 24950.00 24960.00 24970.00 24980.00 24990.00 25000.00 25010.00 25020.00 25030.00 25040.00 25050.00 25060.00 25070.00 25080.00 25090.00 25100.00 25110.00 25120.00 25130.00 25140.00 25146.20 25150.00 25160.00 25170.00 25180.00 25190.00 25200.00 25210.00 25220.00 25230.00 25240.00 25250.00 25260.00 25270.00 25280.00 25290.00 25300.00 25310.00 25320.00 25330.00 25340.00 25350.00 25360.00 25370.00 25380.00 25390.00 25400.00 25410.00 25420.00 25430.00 25440.00 25450.00 25460.00 25470.00 25480.00 25490.00 25500.00 25510.00 25513.90 25514.10 25520.00 25530.00 25540.00 25550.00 25552.90 25560.00 25570.00 25580.00 25590.00 25600.00 25610.00 25620.00 25630.00 25640.00 25650.00 25660.00 25670.00 25680.00 25690.00 25700.00 25710.00 25720.00 25730.00 25740.00 25750.00 25760.00 25770.00 25780.00 25790.00 25800.00 25810.00 25820.00 25830.00 25840.00 25850.00 25860.00 25870.00 25880.00 25890.00 25900.00 25910.00 25920.00 25930.00 25940.00 25950.00 25960.00 25966.20 25970.00 25980.00 25990.00 26000.00 26010.00 26020.00 26030.00 26040.00 26050.00 26060.00 26070.00 26080.00 26090.00 26100.00 26110.00 26120.00 26130.00 26140.00 26150.00 26160.00 26170.00 26180.00 26190.00 26200.00 26210.00 26220.00 26230.00 26240.00 26250.00 26260.00 26270.00 26330.00 26386.10 26812.90 26850.00 27246.60 27370.00 27687.30 27910.00 28135.10 28450.00 28590.20 29010.00 29052.60 29522.50 29580.00 30000.00 30150.00 30740.00 31340.00 31960.00 32580.00 33220.00 33870.00 34530.00 35210.00 35890.00 36600.00 37310.00 38040.00 38790.00 39540.00 40320.00 );
# @agF2Table         = qw( 3.75600 3.85353 3.64700 3.76361 3.67596 3.54000 3.59056 3.43600 3.50738 3.33500 3.42637 3.23700 3.34743 3.20400 3.20400 3.20200 3.20000 3.19900 3.19800 3.19600 3.19500 3.19300 3.19200 3.19000 3.18900 3.18700 3.18600 3.31742 10.8260 10.5200 10.5200 10.5200 10.5200 10.5200 10.5100 10.5100 10.5100 10.5100 10.5000 10.5000 10.5000 10.5000 10.4700 10.6678 10.4135 10.3300 10.1657 10.1700 10.1700 10.1600 10.1600 10.1600 10.1600 10.1500 10.1500 10.1500 10.1500 10.1400 10.1400 10.0399 13.7502 13.8100 13.8000 13.8000 13.8000 13.7900 13.7900 13.7900 13.7800 13.7800 13.7800 13.7800 13.7700 13.7700 13.7700 13.7600 13.7600 13.7600 13.5923 13.6200 13.2689 13.3800 12.9532 13.1000 12.6444 12.3435 12.8200 12.8100 12.8100 12.8000 12.8000 12.8000 12.7900 12.7900 12.7800 12.7800 12.7800 12.7700 12.7600 12.7600 12.2496 14.0388 12.7600 14.6200 14.6100 14.6100 14.6000 14.6000 14.5900 14.5900 14.5800 14.5800 14.5700 14.5700 14.5700 14.5600 14.5600 14.5500 14.5500 14.5400 14.5400 14.5300 13.8258 14.3000 13.5182 13.9100 13.2155 13.5100 12.9155 12.6208 13.0800 12.3301 12.6400 12.0439 12.1800 11.7626 11.7100 11.4863 11.3300 11.2136 10.9460 10.9700 10.6833 10.6200 10.4251 10.3000 10.1717 10.00000 9.92324 9.70400 9.67948 9.44039 9.41600 9.20601 9.13300 8.97641 8.85800 8.75149 8.59000 8.53119 8.32900 8.31557 8.10442 8.07500 7.89794 7.82800 7.69578 7.58700 7.49817 7.35300 7.30492 7.12600 7.11596 6.93126 6.90400 6.75079 6.68900 6.57444 6.48000 6.40224 6.27600 6.23409 6.07900 6.06975 5.90945 5.88600 5.75290 5.70000 5.60018 5.51900 5.45094 5.34400 5.30552 5.16361 5.17300 5.02523 5.00800 4.89031 4.84800 4.75869 4.69200 4.63026 4.54100 4.50510 4.38327 4.39500 4.26427 4.25400 4.14847 4.11600 4.03554 3.98300 3.92563 3.85300 3.81844 3.71407 3.72800 3.61232 3.60600 3.51332 3.48900 3.41702 3.37500 3.32309 3.26500 3.23157 3.14262 3.15800 3.05603 3.05400 2.97157 2.95400 2.88946 2.85700 2.80955 2.76300 2.73178 2.65609 2.67200 2.58245 2.58300 2.51078 2.49800 2.44106 2.41500 2.37322 2.33500 2.30723 2.24304 2.25700 2.18059 2.18200 2.11985 2.10900 2.06077 2.03900 2.00331 1.94743 1.97000 1.89309 1.90400 1.84024 1.84000 1.78885 1.77800 1.73889 1.71800 1.69030 1.64307 1.66000 1.59714 1.60400 1.55249 1.55000 1.50908 1.49700 1.46688 1.44600 1.42586 1.38599 1.39700 1.34723 1.35000 1.30955 1.30400 1.27293 1.26000 1.23733 1.21700 1.20274 1.16911 1.17500 1.13643 1.13500 1.10467 1.09600 1.07380 1.05900 1.04381 1.02300 1.01465 0.986325 0.987500 0.958794 0.953600 0.932041 0.920800 0.906044 0.889100 0.880780 0.858500 0.856231 0.832377 0.828900 0.809197 0.800300 0.786673 0.772600 0.764788 0.745800 0.743521 0.722858 0.720000 0.702780 0.695000 0.683270 0.670800 0.664314 0.647400 0.645895 0.624900 0.627998 0.610608 0.603000 0.593712 0.583900 0.583600 0.583200 0.582800 0.582400 0.581900 0.581500 0.581100 0.580600 0.580200 0.579800 0.579400 0.578900 0.578500 0.578100 0.577600 0.577200 0.576800 0.576400 0.575900 0.575500 0.575100 0.574700 0.574200 0.573800 0.573400 0.572900 0.572500 0.572100 0.571700 0.571200 0.570800 0.570400 0.570000 0.569500 0.577294 0.569100 0.568700 0.568300 0.567800 0.567400 0.567000 0.566800 0.566400 0.565900 0.565500 0.565100 0.564700 0.564200 0.563800 0.563400 0.563000 0.562300 0.562100 0.561600 0.561300 0.560900 0.560400 0.560200 0.559800 0.559400 0.559000 0.558500 0.558100 0.557700 0.557300 0.556900 0.556400 0.556000 0.555600 0.555400 0.555000 0.554500 0.562829 3.56169 3.61900 3.61500 3.61200 3.60900 3.55330 3.60600 3.60300 3.60000 3.59700 3.59500 3.59200 3.58900 3.58600 3.58300 3.58000 3.57700 3.57400 3.57200 3.57000 3.56700 3.56400 3.56100 3.55800 3.55500 3.55200 3.55000 3.54700 3.54500 3.54200 3.53900 3.53600 3.53300 3.53100 3.52900 3.52600 3.52300 3.52000 3.51800 3.51500 3.51200 3.51100 3.50800 3.50500 3.50200 3.50000 3.49700 3.46583 3.49500 3.49300 3.49000 3.48700 3.48500 3.48200 3.48100 3.47800 3.47500 3.47300 3.47000 3.46800 3.46600 3.46400 3.46100 3.45800 3.45600 3.45500 3.45200 3.45000 3.44700 3.44500 3.44200 3.44100 3.43800 3.43600 3.43300 3.43100 3.43000 3.42700 3.42500 3.41100 3.38000 3.29582 3.31300 3.21329 3.22300 3.13240 3.13200 3.05314 3.04100 2.97550 2.95100 2.89948 2.82506 2.86100 2.75222 2.77100 2.68200 2.59400 2.50700 2.42100 2.33600 2.26200 2.19000 2.11900 2.05100 1.98500 1.92100 1.85900 1.79800 1.74000 1.68300 );
# $agMaterialDensity = 10.5;
# $agAtomicWeight    = 107.868;

#=================================================================
### Usage:
### ($status,$attenuation_actual) = &setAttenuation1($attenuation);
###
### Purpose:
###   sets requested attenuation by calculating best-match attenuator positions.
###
###   This subroutine takes current monochromator energy for calculations.
###
sub setAttenuation1 ($;$@) {

  our $beamline;
  my ($script, $attenuation, $energyPV, $energyKeV, $attenuationActual);
  my ($status, $statusDescr);

  $script      = 'setAttenuation1';
  $attenuation = shift(@_);

### Retrieve energy value from EPICS:
  $energyPV = $beamline.':MO:E:ActPos';
  ($status, $energyKeV) = Pezca::GetDouble($energyPV);
  if ($status) {
     $statusDescr = &ezcaStatusCode($status);
     print STDOUT '*** '.$script.': EZCA error='.$status
                                .' ['.$statusDescr.'] on PV='.$energyPV."\n";
     return $status;
  }
# print STDOUT $script.': Energy='.$energyKeV.'keV'."\n";

  ($status,$attenuationActual) = &setAttenuation2($attenuation,$energyKeV);
  return $status,$attenuationActual;
}

#=================================================================
### Usage:
###   ($status,$attenuation_actual) = &setAttenuation2($attenuation,$energyKeV);
###
### Purpose:
###   sets requested attenuation by calculating best-match attenuator positions
###
sub setAttenuation2 ($$;$) {

  our $ATTENUATION_USE_FACTOR;
  our (@Attenuators_material, @Attenuators_thickness);
  our (@attenuatorExceptions, $nAttenuatorExceptions);
  my  ($script, $attenuation, $energyKeV, $report, $status, $IN, $OUT);
  my  (@Working_attenuators_material, @Working_attenuators_thickness);
  my  ($nWorking_attenuators, @Working_attenuators_channel, $i, $ex);
  my  ($attenuator, $attenuation_min, $attenuation_max);
  my  ($filterThickness, $filterMaterial, $filterMuT, @attenuatorsMuT);
  my  ($requiredMuT, $MuT_error, @combination);
  my  ($attenuationActual, @elementList, $element);

  $script        = 'setAttenuation2';
  $attenuation   = shift(@_);
  $energyKeV     = shift(@_);
  $report        = shift(@_);	if (!defined $report) {$report=1;}

  if (!defined $ATTENUATION_USE_FACTOR) {$ATTENUATION_USE_FACTOR = 0;}

  if ($energyKeV <= 0.0) {
     print STDOUT '*** '.$script.': incorrect energy='.$energyKeV."\n";
     return -1;
  }
# if ($report) {print STDOUT $script.': Energy='.$energyKeV.'keV'."\n";}

  @elementList = ('Al', 'Ag');
  if ($beamline =~ /^21/) {push(@elementList,'Au');}

  $status = 0;
  $IN  = 1;
  $OUT = 1 - $IN;

### makeAttenuationDB has to be invoked at least once before
### calling setAttenuation2. It fills in global arrays
### @Attenuators_material, @Attenuators_thickness, and @attenuatorExceptions
### Initialize attenuation DB
  if (@Attenuators_material < 2) {
#    if ($report) {print STDOUT $script.': Initializing AttenuationDB'."\n";}
### This call of makeAttenuatorDB initializes
### the following global arrays:
### our (@Attenuators_material, @Attenuators_thickness);
### our (@attenuatorExceptions, $nAttenuatorExceptions);
#    $status = &makeAttenuatorDB_DESC();
     $status = &makeAttenuatorDB_mySQL();
     if ($status) {
        print STDOUT '*** '.$script.': makeAttenuationDB returned status='.$status."\n";
        return $status;
     }
     foreach $element (@elementList) {
        if ($element eq 'Al') {
           $status = &makeF2DB($element,\@alEnergyTable,\@alF2Table,\$alMaterialDensity,\$alAtomicWeight);
        }
        elsif ($element eq 'Ag') {
           $status = &makeF2DB($element,\@agEnergyTable,\@agF2Table,\$agMaterialDensity,\$agAtomicWeight);
        }
        elsif ($element eq 'Au') {
           $status = &makeF2DB($element,\@auEnergyTable,\@auF2Table,\$auMaterialDensity,\$auAtomicWeight);
        }
        else {$status = -1;}
        if ($status) {
           print STDOUT '*** '.$script.': makeF2DB for material='.$element
                                      .' returned status='.$status."\n";
           return $status;
        }
     }
  }

### Create attenuators_thickness and attenuators_material
### excluding the elements in exceptions array:
###
### step 1: Copy all attenuators from Attenuators_material/thickness array
###         to Working_attenuators_thickness/material array
### step 2: For each element in the exceptions array set
###         Working_attenuators_thickness element to -1
### step 3: Exclude those elements from Working_attenuators arrays
###
### Working_attenuators_thickness/material: Array containing
###'working' attenuators thickness/material

  @Working_attenuators_material  = @Attenuators_material;
  @Working_attenuators_thickness = @Attenuators_thickness;
  $nWorking_attenuators = @Working_attenuators_thickness;
# if ($report) {print STDOUT $script.': nWorking_attenuators = '.$nWorking_attenuators."\n";}
  for ($i=0; $i<$nWorking_attenuators; $i++) {$Working_attenuators_channel[$i]=$i;}

### Mark exceptions:
  if ($nAttenuatorExceptions > 0) {
     foreach $ex (@attenuatorExceptions) {
        if ($ex >= 0 && $ex < $nWorking_attenuators) {
           $Working_attenuators_thickness[$ex] = -1;		#mark as not present
        } else {
           print STDOUT '*** '.$script.': requested excluded attenuator number='
                                      .$ex.' is outside of range'."\n";
           return -1;
        }
     }
### Exclude not present attenuators:
     for ($i=$nWorking_attenuators-1; $i>=0; $i--) {
        if ($Working_attenuators_thickness[$i] == -1) {
           splice(@Working_attenuators_material,  $i, 1);
           splice(@Working_attenuators_thickness, $i, 1);
           splice(@Working_attenuators_channel,   $i, 1);
        }
     }
     $nWorking_attenuators = @Working_attenuators_thickness; 	#actual number of attenuators
#    if ($report) {print STDOUT $script.': nWorking_attenuators = '.$nWorking_attenuators."\n";}
  }

###----------------------------------------------------
### Calculate attenuators to open and close
  if ($ATTENUATION_USE_FACTOR) {		#attenuation is expressed in times
     $attenuation_min = 1.;
     $attenuation_max = 1.E6;
  } else {
     $attenuation_min = 1.;			#attenuation is expressed in percent
     $attenuation_max = 100.;
  }

  if  ($attenuation >= $attenuation_max) {
     if ($report) {print STDOUT 'Closing all attenuators'."\n";}
### Put in every attenuator without any calculation,
     foreach $attenuator (@Working_attenuators_channel) {
        $status = &setAttenuatorState($attenuator, $IN);
        if ($status) {
           print STDOUT '*** '.$script.': setAttenuatorState returned status='.$status."\n";
           return $status;
        }
     }
  }
  elsif ($attenuation <= $attenuation_min) {
     if ($report) {print STDOUT 'Opening all attenuators'."\n";}
     foreach $attenuator (@Working_attenuators_channel) {
        $status = &setAttenuatorState($attenuator, $OUT);
        if ($status) {
           print STDOUT '*** '.$script.': setAttenuatorState returned status='.$status."\n";
           return $status;
        }
     }
  }

### SATISH: Loop over all of the filters and update
### the amount of attenuation each filter provides:
  for ($i=0; $i<$nWorking_attenuators; $i++) {
     $filterThickness      = $Working_attenuators_thickness[$i];
     $filterMaterial       = $Working_attenuators_material[$i];
     ($status, $filterMuT) = &calculateFilterMuT($filterMaterial,$filterThickness,$energyKeV);
     if ($status) {
        print STDOUT '*** '.$script.': calculateFilterMuT returned status='.$status."\n";
        return $status;
     }
#    push(@attenuatorsMuT,$filterMuT);
     $attenuatorsMuT[$i] = $filterMuT;
#    print 'filtermut: '.$filterMaterial.', '.$filterThickness.', '.$energyKeV.', '.$filterMuT."\n"
  }

### Required muT:
  $requiredMuT = &calculateRequiredMuT($attenuation);
# if ($report) {print STDOUT 'required mut: '.$requiredMuT.' for attn '.$attenuation."\n";}

### Search for the combination of filters that will
### give us the closest value to the desired attenuation:
  ($MuT_error, @combination) = &sortAndSearchAttenuation($requiredMuT,@attenuatorsMuT);
# if ($report) {print STDOUT $script.": @{result}\n";}
  $attenuationActual = &calculateAttenuation(\@combination,\@attenuatorsMuT);

### Open and close the filters based on the search result.
### Changed 2010/02/24 by SS: in order to prevent a possibility that the
### attenuation may be temporary lower than requested, the loop over
### attenuators was split into two. Namely:
### ... In the first loop we close everything that needs to be closed
  for ($i=0; $i<@combination; $i++) {
     if($combination[$i] == 1) {
        $status = &setAttenuatorState($Working_attenuators_channel[$i], $IN);
        if ($status) {
           print STDOUT '*** '.$script.': setAttenuatorState IN returned status='.$status."\n";
           return $status;
        }
     }
  }
### ... and only then in the second loop we open everything that needs
### to be opened.
  for ($i=0; $i<@combination; $i++) {
     if($combination[$i] != 1) {
        $status = &setAttenuatorState($Working_attenuators_channel[$i], $OUT);
        if ($status) {
           print STDOUT '*** '.$script.': setAttenuatorState OUT returned status='.$status."\n";
           return $status;
        }
     }
  }
# if ($report) {print STDOUT 'Attenuation is set: Rqs='.$attenuation.'  Act='.$attenuationActual."\n";}

###Finally take out all filters marked as disconnected:
  $status = &fixAttenuationExceptions();
  if ($status) {
     print STDOUT '*** '.$script.': fixAttenuationExceptions returned status='.$status."\n";
  }

  return 0,$attenuationActual;
}

#=================================================================
### Usage:
### ($status,$attenuation_actual) = &getAttenuation1();
###
### Purpose:
###   Returns current attenuation.
###   This subroutine takes current monochromator energy for calculations.
###
sub getAttenuation1 () {

  our $beamline;
  my ($script, $energyPV, $energyKeV, $status, $attenuation, $statusDescr);

  $script = 'getAttenuation1';

### Retrieve energy value from EPICS:
  $energyPV = $beamline.':MO:E:ActPos';
  ($status, $energyKeV) = Pezca::GetDouble($energyPV);
  if ($status) {
     $statusDescr = &ezcaStatusCode($status);
     print STDOUT '*** '.$script.': EZCA error='.$status
                                .' ['.$statusDescr.'] on PV='.$energyPV."\n";
     return $status;
  }
# print STDOUT $script.': Energy='.$energyKeV.'keV'."\n";

  ($status,$attenuation) = &getAttenuation2($energyKeV);
  return $status, $attenuation;
}

#=================================================================
### Usage:
###   ($status,$attenuation_actual) = &getAttenuation2($energyKeV);
###
### Purpose:
###   Returns current attenuation.
###
sub getAttenuation2 ($$;\@) {

  our $ATTENUATION_USE_FACTOR;
  our (@Attenuators_material, @Attenuators_thickness);
  our (@attenuatorExceptions, $nAttenuatorExceptions);
  our (@alEnergyTable, @alF2Table, $alMaterialDensity, $alAtomicWeight);
  our (@agEnergyTable, @agF2Table, $agMaterialDensity, $agAtomicWeight);
  our (@auEnergyTable, @auF2Table, $auMaterialDensity, $auAtomicWeight);
  my  ($script, $energyKeV, $status, $state, $i);
  my  ($MuAl, $MuAg, $MuAu);
  my  ($nAttenuat, @Attenuators_channel, $ex);
  my  ($transmission, $attenuation, @elementList, $element);

  $script    = 'getAttenuation2';
  $energyKeV = shift(@_);

  if (!defined $ATTENUATION_USE_FACTOR) {$ATTENUATION_USE_FACTOR = 0;}

  if ($energyKeV <= 0.0) {
     print STDOUT '*** '.$script.': incorrect energy='.$energyKeV.'KeV'."\n";
     return -1;
  }
# print STDOUT $script.': Energy='.$energyKeV.'keV'."\n";

  @elementList = ('Al', 'Ag');
  if ($beamline =~ /^21/) {push(@elementList,'Au');}

### makeAttenuationDB has to be invoked at least once before
### calling getAttenuation2. It fills in global arrays
### @Attenuators_material, @Attenuators_thickness, and @attenuatorExceptions
### Initialize attenuation DB:
  if (@Attenuators_material < 2) {
#    print STDOUT $script.': Initializing AttenuationDB'."\n";
### This call of makeAttenuatorDB initializes
### the following global arrays:
### our (@Attenuators_material, @Attenuators_thickness);
### our (@attenuatorExceptions, $nAttenuatorExceptions);
#    $status = &makeAttenuatorDB_DESC();
     $status = &makeAttenuatorDB_mySQL();
     if ($status) {
        print STDOUT '*** '.$script.': makeAttenuationDB returned status='.$status."\n";
        return $status;
     }
     foreach $element (@elementList) {
        if ($element eq 'Al') {
           $status = &makeF2DB($element,\@alEnergyTable,\@alF2Table,\$alMaterialDensity,\$alAtomicWeight);
        }
        elsif ($element eq 'Ag') {
           $status = &makeF2DB($element,\@agEnergyTable,\@agF2Table,\$agMaterialDensity,\$agAtomicWeight);
        }
        elsif ($element eq 'Au') {
           $status = &makeF2DB($element,\@auEnergyTable,\@auF2Table,\$auMaterialDensity,\$auAtomicWeight);
        }
        else {$status = -1;}
        if ($status) {
           print STDOUT '*** '.$script.': makeF2DB for material='
                                      .$element.' returned status='.$status."\n";
           return $status;
        }
     }
  }
### Calculate Mu for Al, Ag and Au:
  foreach $element (@elementList) {
     if    ($element eq 'Al') {($status,$MuAl) = &calculateMu($element,$energyKeV);}
     elsif ($element eq 'Ag') {($status,$MuAg) = &calculateMu($element,$energyKeV);}
     elsif ($element eq 'Au') {($status,$MuAu) = &calculateMu($element,$energyKeV);}
     else {$status = -1;}
     if ($status) {
        print STDOUT '*** '.$script.': calculateMu('.$element
                                   .') failed with status='.$status."\n";
        return $status;
     }
  }
### Mark exceptions:
  $nAttenuat = @Attenuators_material;
  for ($i=0; $i<$nAttenuat; $i++) {$Attenuators_channel[$i]=$i;}
  if ($nAttenuatorExceptions > 0) {
     foreach $ex (@attenuatorExceptions) {
        if ($ex >= 0 && $ex < $nAttenuat) {
           $Attenuators_thickness[$ex] = -1;			# mark as not present
        } else {
           print STDOUT '*** '.$script.': requested excluded attenuator number='
                                      .$ex.' out of range.'."\n";
           return -1;
        }
     }
  }
### Calculate MuT of the attenuators array:
  my $muTSum = 0.0;
  for ($i=0; $i<$nAttenuat; $i++) {
     if ($Attenuators_thickness[$i] != -1) {
        ($status,$state) = &getAttenuatorState($i);
        if ($status) {
           print STDOUT '*** '.$script.': getAttenuatorState('.$i.') failed with status='.$status."\n";
           return $status;
        }
        if ($state == 1) {
           if    ($Attenuators_material[$i] eq 'Al') {$muTSum = $muTSum + $MuAl * $Attenuators_thickness[$i];}
           elsif ($Attenuators_material[$i] eq 'Ag') {$muTSum = $muTSum + $MuAg * $Attenuators_thickness[$i];}
           elsif ($Attenuators_material[$i] eq 'Au') {$muTSum = $muTSum + $MuAu * $Attenuators_thickness[$i];}
           else {
              printf STDOUT '*** '.$script.': unknown material %s for channel='.$i."\n",$Attenuators_material[$i];
              return -1;
           }
        }
     }
  }
### transmission = e^(-mu*t)
  $transmission = exp(-$muTSum);
  if ($ATTENUATION_USE_FACTOR) {			#express in times (factor)
     if ($transmission > 1./$MAX_FACTOR_ATTENUATION) {
        $attenuation  = 1./$transmission;
     } else {
        $attenuation  = $MAX_FACTOR_ATTENUATION;}
  } else {						#express in percent
     $attenuation  = (1. - $transmission) * 100.;
  }
  return 0, $attenuation;
}

#=================================================================
### Usage:
###   ($status,[$msg]) = &makeAttenuatorDB_DESC([\@material,\@thinkness,\@exceptions]);
### Purpose:
###    This routine fills in global arrays:
###    @Attenuators_material and @Attenuators_thickness.
###
###    Data is taken from Acromag-9440 Digital I/O channel descriptions.
###
### If the \@material,\@thinkness pointers are passed, they are filled along
### with the global arrays.

sub makeAttenuatorDB_DESC(;\@\@\@) {

  our ($beamline, $attenuatorPV, $attenuatrbPV, $nAttenuators);
  our (@Attenuators_material, @Attenuators_thickness);
  our (@attenuatorExceptions, $nAttenuatorExceptions);
  my  ($script, @result, $pv, $i, $description);
  my  (@elementList, $status, $statusDescr);
  my  ($materialRef, $thicknessRef, $exceptionsRef, $msg, $nr);

  $script = 'makeAttenuatorDB_DESC';
  $materialRef   = shift(@_);
  $thicknessRef  = shift(@_);
  $exceptionsRef = shift(@_);

  @elementList = ('Al', 'Ag');
  if ($beamline =~ /^21/) {push(@elementList,'Au');}

  if ($attenuatorPV !~ /^${beamline}/) {$attenuatorPV = $beamline.':'.$attenuatorPV;}
  if (!defined $attenuatrbPV)          {$attenuatrbPV = $attenuatorPV;}				#then readback=command
  if ($attenuatrbPV !~ /^${beamline}/) {$attenuatrbPV = $beamline.':'.$attenuatrbPV;}

  @Attenuators_material  = ();
  @Attenuators_thickness = ();
  @attenuatorExceptions  = ();
  $nAttenuatorExceptions = 0;
  for ($i=0; $i<$nAttenuators; $i++) {
     $pv = $attenuatorPV.$i.'.DESC';
     ($status,$description) = Pezca::GetString($pv);
#    &report_exit($script, $status, $pv);
     if ($status) {
        $statusDescr = &ezcaStatusCode($status);
        $msg = '*** '.$script.': EZCA error='.$status.' ['.$statusDescr.'] on PV='.$pv."\n";
        print STDOUT $msg;
        return $status,$msg;
     }
### If no filter is present, the description must read 'excluded', 'removed',
### or 'unused'; then we mark it as 'DontCare':
     if ($description =~ /excluded/i || $description =~ /removed/i || $description =~ /unused/i) {
        $Attenuators_thickness[$i] = 0.0;
        $Attenuators_material[$i]  = 'DontCare';
        $attenuatorExceptions[$nAttenuatorExceptions] = $i;
        $nAttenuatorExceptions++;
     }
     else {
        $description =~ s/um//g;			# remove 'microns' label
        @result = split(/\s+/,$description);
        $nr = @result;
        if ($nr != 2) {
           $msg = '*** '.$script.': invalid description ['.$description.'] in PV='.$pv."\n"
                . '*** '.$script.': the number of words is '.$nr.' while expecting 2'."\n"
                . '*** '.$script.': example valid description is "40 Al"'."\n"
                . '*** '.$script.': where 40 is attenuator thickness in um.'."\n";
           print STDOUT $msg;
           return 1,$msg;
        }
        if    (&is_decimal($result[0]) && $result[1] =~ /^[a-zA-Z]*$/) {
           ($Attenuators_thickness[$i],$Attenuators_material[$i]) = @result;
        }
        elsif (&is_decimal($result[1]) && $result[0] =~ /^[a-zA-Z]*$/) {
           ($Attenuators_material[$i],$Attenuators_thickness[$i]) = @result;
        }
        else {
           ($Attenuators_material[$i],$Attenuators_thickness[$i]) = ('','');
        }
        if ($Attenuators_thickness[$i] < 1.0) {
           $msg = '*** '.$script.': invalid description ['.$description.'] in PV='.$pv."\n"
                . '*** '.$script.': too small thickness='.$Attenuators_thickness[$i].' < 1um'."\n"
                . '*** '.$script.': example valid description is [40 Al] where 40 is thickness in um.'."\n";
           print STDOUT $msg;
           return 1,$msg;
        }
        if (! grep(/^$Attenuators_material[$i]$/, @elementList)) {
           $msg = '*** '.$script.': invalid description ['.$description.'] in PV='.$pv."\n"
                . '*** '.$script.': unrecognized chemical element='.$Attenuators_material[$i]."\n"
                . '*** '.$script.': example valid description is [40 Al] where 40 is thickness in um.'."\n";
           print STDOUT '*** '.$script.': Supported materials are '."(@{elementList})\n";
           return 1,$msg;
        }
     }
  }
  if (defined $materialRef)   {@$materialRef   = @Attenuators_material;}
  if (defined $thicknessRef)  {@$thicknessRef  = @Attenuators_thickness;}
  if (defined $exceptionsRef) {@$exceptionsRef = @attenuatorExceptions;}
  return 0,'';
}

#=================================================================
### Usage:
###   ($status,$msg) = &makeAttenuatorDB_mySQL([\@material,\@thinkness,\@exceptions]);
### Purpose:
###    This routine fills in global arrays:
###    @Attenuators_material and @Attenuators_thickness.
###
###    Data is taken from BluIce mySQL database.
###
sub makeAttenuatorDB_mySQL(;\@\@\@) {

  our ($nAttenuators, @Attenuators_material, @Attenuators_thickness);
  our (@attenuatorExceptions, $nAttenuatorExceptions);
  our ($mysqlHost, $mysqlDb, $mysqlUser, $mysqlPass);
  my  ($script, $mysqlTbl);
  my  ($dsn, $dbh, $sth, $nrecords, @val, $nval, $enable, $i);
  my  ($materialRef, $thicknessRef, $exceptionsRef, $msg);

  $script = 'makeAttenuatorDB_mySQL';
  $materialRef   = shift(@_);
  $thicknessRef  = shift(@_);
  $exceptionsRef = shift(@_);

  $mysqlTbl = 'EpicsShutter';
  $dsn = 'DBI:mysql:host='.$mysqlHost.';database='.$mysqlDb;
  $dbh = DBI->connect($dsn,$mysqlUser,$mysqlPass,{PrintError=>0,});
  if (!defined $dbh) {
     $msg = '*** '.$script.': Cannot connect to mySQL server '.$mysqlHost."\n";
     print STDOUT $msg;
     return -1,$msg;
  }

  @Attenuators_material  = ();
  @Attenuators_thickness = ();
  @attenuatorExceptions  = ();
  $nAttenuatorExceptions = 0;
  for ($i=0; $i<$nAttenuators; $i++) {
     $sth = $dbh->prepare('SELECT * FROM '.$mysqlTbl.' WHERE channel="ch'.$i.'"');
     $nrecords = $sth->execute();			#nrecords works for MySQL only. For SQLite result is def/undef only.
     if (!defined $nrecords) {
        $sth->finish();
        $dbh->disconnect();
        $msg = '*** '.$script.': Failed reading channel=ch'.$i."\n"
             . '*** '.$script.': in table='.$mysqlDb.'.'.$mysqlTbl."\n"
             . '*** '.$script.': at mySQL server='.$mysqlHost."\n"
             . '*** '.$script.': '.$sth->errstr."\n";
        print STDOUT $msg;
        return -1,$msg;
     }
     elsif ($nrecords == 0) {
        $sth->finish();
        $dbh->disconnect();
        $msg = '*** '.$script.': No records found for channel=ch'.$i."\n"
             . '*** '.$script.': in table='.$mysqlDb.'.'.$mysqlTbl."\n"
             . '*** '.$script.': at mySQL server='.$mysqlHost."\n";
        print STDOUT $msg;
        return -1,$msg;
     }
     elsif ($nrecords > 1) {
        $sth->finish();
        $dbh->disconnect();
        $msg = '*** '.$script.': Multiple records ('.$nrecords.') found for channel=ch'.$i."\n"
             . '*** '.$script.': in table='.$mysqlDb.'.'.$mysqlTbl."\n"
             . '*** '.$script.': at mySQL server='.$mysqlHost."\n";
        print STDOUT $msg;
        return -1,$msg;
     }
### id,name,channel,beamline,recordID,readback,enable,thickness,material,controllerGID
### 0   1     2       3        4        5       6        7         8         9
     @val = $sth->fetchrow_array();
     $nval = @val;
     if ($nval != 10) {
        $sth->finish();
        $dbh->disconnect();
        $msg = '*** '.$script.': Incorrect record ['."@{val}".'] for channel=ch'.$i."\n"
             . '*** '.$script.': in table='.$mysqlDb.'.'.$mysqlTbl."\n"
             . '*** '.$script.': at mySQL server='.$mysqlHost."\n";
        print STDOUT $msg;
        return -1,$msg;
     }
     $enable = $val[6];
     if ($enable) {
        $Attenuators_material[$i]  = $val[8];
        $Attenuators_thickness[$i] = $val[7];
     } else {
### If channel is not enabled (same as 'excluded', 'removed', or 'unused'
### in makeAttenuatorDB_DESC), we mark it as 'DontCare':
        $Attenuators_material[$i]  = 'DontCare';
        $Attenuators_thickness[$i] = 0.0;
        $attenuatorExceptions[$nAttenuatorExceptions] = $i;
        $nAttenuatorExceptions++;
     }
  }
  $sth->finish();
  $dbh->disconnect();
  if (defined $materialRef)   {@$materialRef   = @Attenuators_material;}
  if (defined $thicknessRef)  {@$thicknessRef  = @Attenuators_thickness;}
  if (defined $exceptionsRef) {@$exceptionsRef = @attenuatorExceptions;}
  return 0,'';
}

#=================================================================
### Usage:
###  &printAttenuatorDB();
###
sub printAttenuatorDB() {

  our @Attenuators_thickness;
  my  ($script, $i);

  $script = 'printAttenuatorDB';
  for ($i=0; $i<@Attenuators_thickness; $i++) {
    print STDOUT $i.'  '.$Attenuators_thickness[$i].'  '.$Attenuators_material[$i]."\n";
  }
}

#=================================================================
### Usage:
###   $status = &makeF2DB($material,\@EnergyTable,\@F2Table,\$Density,\$AtomicWeight);
###                                                        # fills global arrays for material=Al, Ag or Au:
### Purpose:
###   Fills global arrays for material=Al/Ag/Au: @xxEnergyTable,@xxF2Table,$xxMaterialDensity,$xxAtomicWeight
###
sub makeF2DB($\@\@\$\$) {

### This routine fills tables of filter material properties:
### $F2($Energy), $Density, and $AtomicWeight

  my ($script, $material, $EnRef, $F2Ref, $RoRef, $AWRef, $status);

  $script   = 'makeF2DB';
  $material = shift(@_);
  $EnRef    = shift(@_);
  $F2Ref    = shift(@_);
  $RoRef    = shift(@_);
  $AWRef    = shift(@_);
  $status   = 0;

# no strict 'refs';
  if ($material eq 'Al') {
     $$RoRef = 2.7;
     $$AWRef = 26.98;
  }
  elsif ($material eq 'Ag') {
     $$RoRef = 10.5;
     $$AWRef = 107.868;
  }
  elsif ($material eq 'Au') {
     $$RoRef = 19.3;
     $$AWRef = 196.96;
  }
  else {
     print STDOUT '*** '.$script.': no data for material='.$material."\n";
     $status = -1;
  }
# use strict 'refs';
  $status = &readF2DB($material,$EnRef,$F2Ref);
  return $status;
}

#=================================================================
### Usage:
###    $status = &readF2DB($material,\@EnergyTable,\@F2Table);
### Purpose:
###    this routine fills tables of filter material properties:
###    @EnergyTable and @F2Table by reading respective files
###    from FASTSCANS database.
###
sub readF2DB($\@\@) {

  my ($script, $material, $EnRef, $F2Ref, $file, $filehandle, $status);
  my ($string, $nwords, @words, $line, $FASTSCANS);

  $script   = 'readF2DB';
  $material = shift(@_);
  $EnRef    = shift(@_);
  $F2Ref    = shift(@_);
  $status   = 0;

  if (!defined $ENV{FASTSCANS}) {
     print STDOUT '*** '.$script.': Environment variable FASTSCANS is undefined'."\n";
     return -1;
  }
  $FASTSCANS = $ENV{FASTSCANS};
  $FASTSCANS =~ s/\\/\//g;					#change \ to /
  $FASTSCANS =~ s/\/$//;					#remove ending /
  $FASTSCANS =~ s/\/bin$//;					#remove ending /bin if exits
  $file = $FASTSCANS.'/data/'.$material.'.dat';

  if (!(-f $file)) {
     print STDOUT '*** '.$script.': Material DB file="'.$file.'" does not exist'."\n";
     return -1;
  }
# print STDOUT 'Reading material DB file ['.$file.']'."\n";

  $filehandle = open (MATERIAL_DATA, '< '.$file);
  if (!defined $filehandle) {
     print STDOUT '*** '.$script.': Failed to open material DB file "'.$file.'"'."\n";
     return -1;
  }
  select MATERIAL_DATA; $|=1;

# Retrieve lines from file:
  $line = 0;
  while (<MATERIAL_DATA>) {
     $line++;
     $string = $_;
     chomp $string;		# trail CR
     $string =~ s/;.*$//g;	# trail comments started by ';' ("." matches any character, '*' means match 0 or more times, '$' matches end of string)
     $string =~ s/#.*$//g;	# trail comments started by '#' ("." matches any character, '*' means match 0 or more times, '$' matches end of string)
     $string =~ s/\t/ /g;	# replace tabs by spaces
     $string =~ s/ +/ /g;	# compress spaces ('+' means match 1 or more times)
     $string =~ s/^ *//g;	# remove beginning-of-the-line spaces
     $string =~ s/ *$//g;	# remove end-of-the-line spaces
     if (length($string) == 0) {next;}
     @words = split ' ', $string;
     $nwords = @words;
#    print STDOUT 'line='.$line.'  string='.$string."\n";
#    print STDOUT 'nwords='.$nwords.'  words='."@{words}\n";
     if ($nwords != 2) {
        print STDOUT '*** '.$script.': Number of words at line='.$line
                                   .' is not 2: nwords='.$nwords."\n";
        close(MATERIAL_DATA);
        return -1;
     }
     no strict 'refs';
     push (@$EnRef,$words[0]);
     push (@$F2Ref,$words[1]);
     use strict 'refs';
  }
  close(MATERIAL_DATA);
  return 0;
}

#=================================================================
### Usage:
###   ($status,$state) = &getAttenuatorState($channel);
### Purpose:
###   returns state=1 for IN and state=0 for OUT.
###
sub getAttenuatorState($) {

  our ($beamline, $attenuatorPV, $attenuatrbPV, $Attenuators_IN);
  my  ($script, $channel, $pv, $status, $statusDescr, $VAL, $Attenuators_OUT);

  $script  = 'getAttenuatorState';
  $channel = shift(@_);

  if ($attenuatorPV !~ /^${beamline}/) {$attenuatorPV = $beamline.':'.$attenuatorPV;}
  if (!defined $attenuatrbPV)          {$attenuatrbPV = $attenuatorPV;}	#then it is the same
  if ($attenuatrbPV !~ /^${beamline}/) {$attenuatrbPV = $beamline.':'.$attenuatrbPV;}

  $pv = $attenuatrbPV.$channel;

  ($status, $VAL) = Pezca::GetDouble($pv);
  if ($status) {
     $statusDescr = &ezcaStatusCode($status);
     print STDOUT '*** '.$script.': EZCA error='.$status
                                .' ['.$statusDescr.'] on PV='.$pv."\n";
     return $status;
  }
  $Attenuators_OUT = 1 - $Attenuators_IN;
  if    ($VAL == $Attenuators_IN)  {return 0, 1;}		#IN
  elsif ($VAL == $Attenuators_OUT) {return 0, 0;}		#OUT
  else  {
     print STDOUT '*** '.$script.': Unknown state='.$VAL.' on PV='.$pv."\n";
     return 1;
  }
}

#=================================================================
### Usage:
###   $status = &setAttenuatorState($channel,$state);
### Purpose:
###   dtives IN or OUT requested attenuator
###
sub setAttenuatorState($$) {

  our ($beamline, $attenuatorPV, $attenuatrbPV, $Attenuators_IN);
  my  ($script, $channel, $pv, $state, $status, $statusDescr);
  my  ($IN, $OUT, $VAL, $Attenuators_OUT);

  $script  = 'setAttenuatorState';
  $channel = shift(@_);
  $state   = shift(@_);
  if ($attenuatorPV !~ /^${beamline}/) {$attenuatorPV = $beamline.':'.$attenuatorPV;}
  if (!defined $attenuatrbPV)          {$attenuatrbPV = $attenuatorPV;}		#then readback=command
  if ($attenuatrbPV !~ /^${beamline}/) {$attenuatrbPV = $beamline.':'.$attenuatrbPV;}

  $pv = $attenuatorPV.$channel;

  $IN  = 1;
  $OUT = 1 - $IN;
  $Attenuators_OUT = 1 - $Attenuators_IN;

  if    ($state == $IN)  {$VAL = $Attenuators_IN;}
  elsif ($state == $OUT) {$VAL = $Attenuators_OUT;}

  $status = Pezca::PutDouble($pv,$VAL);
# &report_exit($script, $status, $pv);
  if ($status) {
     $statusDescr = &ezcaStatusCode($status);
     print STDOUT '*** '.$script.': EZCA error='.$status
                                .' ['.$statusDescr.'] on PV='.$pv."\n";
  }
  return $status;
}

#=================================================================
### Usage:
###    ($status,$Mu) = &calculateMu($material,$energykeV);
###        where material is Al, Ag or Au; e.g.: &calculateMu(Al, 5);
### returns: status, absorptionVal
###         where status 1- error; 0- success
###
### Purpose:
###    This routine calculates Mu for given material and energy.
###    It is used by calculateFilterMuT subroutine.
###
sub calculateMu($$) {

  our (@alEnergyTable, @alF2Table, $alMaterialDensity, $alAtomicWeight);
  our (@agEnergyTable, @agF2Table, $agMaterialDensity, $agAtomicWeight);
  our (@auEnergyTable, @auF2Table, $auMaterialDensity, $auAtomicWeight);
  my  ($script, $material, $energykeV);
  my  (@energyTable, @f2Table);
  my  ($materialDensity, $atomicWeight, $status);
  my  ($totalNumEnergies, $energy, $indx);
  my  ($E_1, $F2_1, $E_2, $F2_2, $f2, $coeff, $mu);

  $script    = 'calculateMu';
  $material  = shift(@_);
  $energykeV = shift(@_);

  if ($material eq 'Al') {
     if (!defined $alAtomicWeight) {
        $status = &makeF2DB('Al',\@alEnergyTable,\@alF2Table,\$alMaterialDensity,\$alAtomicWeight);
        if ($status) {return 1, 0;}
     }
     @energyTable     = @alEnergyTable;
     @f2Table         = @alF2Table;
     $materialDensity = $alMaterialDensity;
     $atomicWeight    = $alAtomicWeight;
  }
  elsif ($material eq 'Ag') {
     if (!defined $agAtomicWeight) {
        $status = &makeF2DB('Ag',\@agEnergyTable,\@agF2Table,\$agMaterialDensity,\$agAtomicWeight);
        if ($status) {return 1, 0;}
     }
     @energyTable     = @agEnergyTable;
     @f2Table         = @agF2Table;
     $materialDensity = $agMaterialDensity;
     $atomicWeight    = $agAtomicWeight;
  }
  elsif ($material eq 'Au') {
     if (!defined $auAtomicWeight) {
        $status = &makeF2DB('Au',\@auEnergyTable,\@auF2Table,\$auMaterialDensity,\$auAtomicWeight);
        if ($status) {return 1, 0;}
     }
     @energyTable     = @auEnergyTable;
     @f2Table         = @auF2Table;
     $materialDensity = $auMaterialDensity;
     $atomicWeight    = $auAtomicWeight;
  }
  else {
     print STDOUT '*** '.$script.': input parameter material="'.$material.'" is invalid'."\n";
     return 1, 0;
  }

### Number of elements in energyTable array
  $totalNumEnergies = @energyTable;
  $energy = $energykeV * 1000;

  $indx=0;
  while ($energy >= $energyTable[$indx] && $indx < $totalNumEnergies) {$indx++;}

###  SATISH:
### If the input energy is between E_1 and E_2 use linear approximation to calculate f2
### linear approximation:               E - E_1
###                      f2 = F2_1 + ---------- (F2_2 - F2_1)
###                                   E_2 - E_1
###
### If input energy < the first energy in the table (Eg: 3009)


  if ($indx == 0) {
     $E_1  = 0;
     $F2_1 = 0;
     $E_2  = $energyTable[0];
     $F2_2 = $f2Table[0];
  } else {
     $E_1  = $energyTable[$indx-1];
     $F2_1 = $f2Table [$indx-1];
     $E_2  = $energyTable[$indx];
     $F2_2 = $f2Table[$indx];
  }
### Liner intetpolation from the table:
  $f2 = $F2_1 + ( ($energy - $E_1) / ($E_2 - $E_1) ) * ($F2_2 - $F2_1);
### Calculation of $mu:
# my $w2e      = 12.3984186;				#keV*Angstrom
# my $pi       = 3.14159265359;
# my $e_radius = 2.81794e-05;				#Angstrom
# my $Avogadro = 6.0221367e+23;				#1/gram (?)
# my $wave     = $w2e / $energyKeV;			#(1)
# my $td       = 2. * $wave * $e_radius * $f2;		#(2)
# my $atomGram = $atomicWeight / $Avogadro;		#from a.e.m to gram
# my $vol      = (1.E24) * $atomGram / $materialDensity;$A^3
# my $xi0      = $td * $wave / (2*$pi*$vol);
# my $mu       = (2.E4) * $pi * $xi0 / $wave;		#1/um
### As a result, we get:
# my $coeff    = 2. * $w2e * $e_radius * $Avogadro / (1.E20);
# print STDOUT $coeff."\n";
  $coeff    = 4.2080282;
  $mu = $coeff * $materialDensity * $f2 / ($energykeV * $atomicWeight);

  return 0, $mu;
 }

#=================================================================
### Usage:
###   ($status,$MuT) = &calculateFilterMuT($material,$thickness,energykeV);
###
sub calculateFilterMuT ($$$) {

  my ($script, $material, $thickness, $energykeV);
  my ($status, $mu, $muT);

  $script    = 'calculateFilterMuT';
  $material  = shift(@_);
  $thickness = shift(@_);
  $energykeV = shift(@_);

  ($status, $mu)  = &calculateMu($material,$energykeV);

  if ($status) {
    print STDOUT '*** '.$script.': calculateMu returned status='.$status."\n";
    return 1,0;
  }

  $muT = $mu * $thickness;
  return 0, $muT;
}

#=================================================================
### Usage:
###   $requiredMuT = &calculateRequiredMuT($requestedAttenuation);
### Purpose:
###    Converts attenuation into MuT
###
sub calculateRequiredMuT ($) {

  our $ATTENUATION_USE_FACTOR;
  my  ($script, $requestedAttenuation, $transmission);
  my  ($min_transmission, $requiredMuT);

  $script = 'calculateRequiredMuT';
  $requestedAttenuation = shift(@_);

  if (!defined $ATTENUATION_USE_FACTOR) {$ATTENUATION_USE_FACTOR = 0;}

  $min_transmission = 1./$MAX_FACTOR_ATTENUATION;
### transmission = e^(-mu*t)
### mu * t= - log(transmission)
  if ($ATTENUATION_USE_FACTOR) {
     if ($requestedAttenuation > $MAX_FACTOR_ATTENUATION) {
        $transmission = $min_transmission;
     }
     elsif ($requestedAttenuation < 1.) {
        $transmission = 1.;
     } else {
        $transmission = 1./$requestedAttenuation;
     }
  } else {
     if ($requestedAttenuation > 100.*(1.-$min_transmission)) {
        $transmission = $min_transmission;
     }
     elsif ($requestedAttenuation < 0.) {
        $transmission = 1.;
     } else {
        $transmission = 1. - $requestedAttenuation/100.0;
     }
  }
  $requiredMuT = - log($transmission);
  return $requiredMuT;
}

#=================================================================
### Usage:
###   $attenuation = &calculateAttenuation(\@combination,\@attenuatorsMuT);
### Purpose:
###   Calculate attenuation for given combination of IN filters.
###   The MuT values for each filter are pre-calculated elswhere.
###
sub calculateAttenuation(\@\@) {

  our $ATTENUATION_USE_FACTOR;
  my  ($script, $combinationRef, $MuTRef, $MuTSum, $nMuT, $i);
  my  ($transmission, $attenuation);

  $script = 'calculateAttenuation';
  $combinationRef = shift(@_);
  $MuTRef         = shift(@_);

  if (!defined $ATTENUATION_USE_FACTOR) {$ATTENUATION_USE_FACTOR = 0;}

  no strict 'refs';
  $MuTSum = 0.;
  $nMuT  = @$MuTRef;
  for ($i=0; $i<$nMuT; $i++) {
     if (@$combinationRef[$i] == 1) {
        $MuTSum = $MuTSum + @$MuTRef[$i];
     }
  }
  use strict 'refs';
### transmission = e^(-mu*t)
  $transmission = exp(-$MuTSum);
  if ($ATTENUATION_USE_FACTOR) {
     if ($transmission > 1.E-6) {$attenuation  = 1./$transmission;}	#express in times (factor)
     else                       {$attenuation  = 1.E6;}
  } else {
     $attenuation = (1. - $transmission) * 100.;			#express in percent
  }
  return $attenuation;
}

#=================================================================
### Usage:
###   @sortedPosition = &sortArrayDescending(\@elements);
### Return:
###   sorted_array + array_element_position_in_sorted_array
###
### Purpose:
###   The array is sorted in descending order
###
### Please see the comments in the code for more details
###
sub sortArrayDescending(\@) {

  my ($script, $elements, @position, $numElements);
  my ($i, $j, $j_plus_1, $a, $b, $temp_pos, $temp_element);
  my (@unsortedPosition, @sortedPosition);

  $script   = 'sortArrayDescending';
  $elements = shift(@_);

  no strict 'refs';
  $numElements = @$elements;
  if ($numElements <= 0) {
     print STDOUT '!!! '.$script.': incorrect number of elements='.$numElements."\n";
  }

### Position: position of elements in unsorted array
  for ($i=0; $i<$numElements; $i++) {
#   push(@position, $i);
    $position[$i] = $i;
  }

  for ($i=0; $i < ($numElements-1); $i++) {
     for ($j=0; $j < ($numElements-1-$i); $j++) {
        $j_plus_1 = $j + 1;

        $a = @$elements[$j];
        $b = @$elements[$j_plus_1];

        if ($a < $b) {
           $temp_element         = @$elements[$j];
           @$elements[$j]        = @$elements[$j_plus_1];
           @$elements[$j_plus_1] = $temp_element;

           $temp_pos            = $position[$j];
           $position[$j]        = $position[$j_plus_1];
           $position[$j_plus_1] = $temp_pos;
        }
     }
  }
  use strict 'refs';

### SATISH:
### Example:          {20 2 4 1 25}
### sorted array:     {25 20 4 2 1}
### unsortedPosition: { 4  0 2 1 3}
###                     |  | | | |
###                     |  | | | +---old position of  1 (of element-4)
###                     |  | | +-----old position of  2 (of element-3)
###                     |  | +-------old position of  4 (of element-2)
###                     |  +---------old position of 20 (of element-1)
###                     +------------old position of 25 (of element-0)
###                     ^ 25 is at position 4 in unsorted array
### position of element in the unsorted/input array

  for ($i=0; $i < $numElements; $i++) {
#    push(@unsortedPosition,$position[$i]);
     $unsortedPosition[$i] = $position[$i];
  }

### sortedPosition: position of element in sorted array
### Example:          {20 2 4 1 25}
### sorted array:     {25 20 4 2 1}
### sortedPosition:     1  3 2 4 0
###                     |  | | | |
###                     |  | | | +--new position of 25 (of element-4)
###                     |  | | +----new position of  1 (of element-3)
###                     |  | +------new position of  4 (of element-2)
###                     |  +--------new position of  2 (of element-1)
###                     +-----------new position of 20 (of element-0)
###                     ^ 20 is at indx 1 in sorted array

  for ($i=0; $i < $numElements; $i++) {
     for ($j=0; $j < $numElements; $j++) {
        if ($unsortedPosition[$j] == $i) {
           push(@sortedPosition,$j);
        }
     }
  }
### This actually returns one combined array:
  return @sortedPosition;
}

#=================================================================
### Usage:
###   ($deltaMuT,@originalPattern) = &sortAndSearchAttenuation($targetMuT,@MuTList);
### Purpose:
###    sorts @MuTList in descending order and then calls &searchAttenuation
###    to find best match to $targetMuT. Returns $deltaMuT (the difference
###    between $targetMuT best found MuT) and respective array of IN-OUTs
###    for filters.
###
sub sortAndSearchAttenuation($@) {

  my ($script, $targetMuT, @MuTList, $delta, $position);
  my (@sortedMuTList, @sortedPosition, @sortedPattern, @originalPattern);

  $script    = 'sortAndSearchAttenuation';
  $targetMuT = shift(@_);
  @MuTList   = @_;
# print STDOUT 'target = '.$targetMuT."\n";
# print STDOUT 'MuT = '."@{MuTList}\n";

###sort MuTList in descending order:
  @sortedMuTList = @MuTList;
  @sortedPosition = &sortArrayDescending(\@sortedMuTList);

# print STDOUT 'Sorted Positions: '."@{sortedPosition}\n\n";
# print STDOUT 'targetMuT: '.$targetMuT."\n\n";
# print STDOUT 'Sorted MuT: '."@{sortedMuTList}\n\n";

  ($delta,@sortedPattern) = &searchAttenuation($targetMuT,@sortedMuTList);

  foreach $position (@sortedPosition) {
     push(@originalPattern,$sortedPattern[$position]);
  }

  return  $delta,@originalPattern;

}


#=================================================================
### SATISH:
###  ported from bluice04/dcss/scripts/devices
###  NB: some bugs were fixed as commented below
###
### Usage:
###   ($deltaMuT,@sortedPattern) = &searchAttenuation($targetMuT,@sortedMuTList);
###  Inputs:
###   targetMuT -> This it the requested MuT
###   MuTList   -> This is the list of MuT for available filters.
###                      The MuT list must be ordered by decreasing MuT
###                      for this function to work.
###  Outputs:
###            delta  -> The difference between the target MuT and the best
###                      achievable MuT from the component filters.
###            pattern-> This will be a list of binary numbers indicating whether
###                      the component filter should be used or not.
### Purpose:
###     This function searches for a combination of filters that is closest
###     to, but not exceeding the desired MuT.
###
###  The function will recursively call itself while searching for the answer.
###
sub searchAttenuation ($@) {

  my ($script, $targetMuT, @MuTList, $totalChannels, $bestDelta);
  my (@bestPattern, @thisPattern, @possiblePattern, $mut, $MuTSum);
  my ($count, $thisMuT, @thisResult, @temp_MuTList, $delta, @pattern);

  $script    = 'searchAttenuation';
  $targetMuT = shift(@_);
  @MuTList   = @_;

  $totalChannels = @MuTList;
  $bestDelta     = $targetMuT;
  @bestPattern   = qw();
  @thisPattern   = qw();

### SATISH: set best pattern to all 0's
  foreach $mut (@MuTList) {push(@bestPattern,0);}

### SATISH: shortcircuit the recursive calling if the target MuT
### is less than or equals to zero
  if ($targetMuT <= 0) {return  $bestDelta,@bestPattern;}

### Shortcircuit the recursive calling if the sum of the MuT
###  is less than the target MuT.
  $MuTSum = 0;
  foreach $mut (@MuTList) {
     $MuTSum = $MuTSum + $mut;
     push (@possiblePattern, 1);
  }

  if ($MuTSum <= $targetMuT) {return $targetMuT-$MuTSum, @possiblePattern;}

### Loop over all the filters, except the last filter in the list.
  $count = 0;
  while ($count+1 < $totalChannels) {

     $thisMuT = $MuTList[$count];

### Search for best resulting pattern for sub string
     @temp_MuTList = @MuTList;
     splice(@temp_MuTList,0,$count+1);
     @thisResult= &searchAttenuation($targetMuT-$thisMuT,@temp_MuTList);

     $delta   = shift(@thisResult);
     @pattern = @thisResult;

### SATISH:
     $delta = abs($delta);

     if ($delta >= $bestDelta) {
### That one didn't work
        push(@thisPattern,0);
### SATISH:
### Bug: Assumes if delta > bestDelta bestPattern has been found.
### It's possible that best pattern is not found in some cases
### Eg: searchAttenuation 25 {23 22 20 4 1} would give wrong value
###
### ORIGINAL: break
### MODIFIED: commented
#       break
     } else {
### SATISH:
###  Bug: $delta == $bestDelta case not considered
###  Eg: searchAttenuation 9 {8 6 5 4 2} would give wrong value
###
###  ORIGINAL: if { $delta < $bestDelta}
###  MODIFIED: else part is added
        @bestPattern = @thisPattern;
        push(@bestPattern,1);
        push(@bestPattern,@pattern);
        $bestDelta = $delta;
        push(@thisPattern,0);
     }
     $count++;
  }

### Look at the last filter in the list.
  $delta = $targetMuT - $MuTList[$#MuTList];

### SATISH:
  $delta = abs($delta);

  if ($delta < $bestDelta) {
     @bestPattern = @thisPattern;
     push(@bestPattern,1);
     $bestDelta = $delta;
  }

# print STDOUT ' '.$bestDelta.', '."@{bestPattern}".', '."@{MuTList}\n";
  return  $bestDelta,@bestPattern;
}

#=================================================================
### Usage:
###   $status = &fixAttenuationExceptions();
### Purpose:
###    find if the filters marked as "Unused" are actually "in" and if
###    so, fix it with a warning.
###
sub fixAttenuationExceptions() {

  our (@attenuatorExceptions, $nAttenuatorExceptions);
  my ($script, $state, $status, $i, $channel);

  $script = 'fixAttenuationExceptions';
  if (! @attenuatorExceptions || !defined $nAttenuatorExceptions) {
     print STDOUT '*** '.$script.': filters exceptions are undefined.'
                                .' Please call makeAttenuatorDB first.'."\n";
     return -1;
  }
  for ($i=0; $i<$nAttenuatorExceptions; $i++) {
     $channel = $attenuatorExceptions[$i];
     ($status,$state) = &getAttenuatorState($channel);			#1=IN, 0=OUT
     if ($status) {
        print STDOUT '*** '.$script.': error='.$status
                           .' from getAttenuatorState on channel='.$channel."\n";
        next;
     } else {
        if ($state) {							#1=IN, 0=OUT
           print STDOUT '!!! '.$script.': filter='.$channel
                              .' marked as unused is IN. Fixing...'."\n";
           $state = 0;
           $status = &setAttenuatorState($channel,$state);             #1=IN, 0=OUT
           if ($status) {
              print STDOUT '*** '.$script.': error='.$status
                                 .' from setAttenuatorState on channel='.$channel."\n";
              print STDOUT '*** '.$script.': could not fix channel='.$channel."\n";
           }
        }
     }
  }
}

#=================================================================
### Usage:
###   $status = &setAttenuationMonitors();
### Purpose: sets CA monitors on the attenuators PVs (for repeated use)
###   returns status=1 in case of failure and status=0 in case of success.
###
sub setAttenuationMonitors() {

  our ($beamline, $attenuatorPV, $attenuatrbPV, $nAttenuators);
  my  ($script, $channel, $pv, $status, $statusTotal);

  $script  = 'setAttenuationMonitors';

  if (!defined $attenuatorPV ||
      !defined $attenuatrbPV ||
      !defined $nAttenuators) {
     print STDOUT '*** '.$script.': missing attenuators information in beamline.pl'."\n";
     return -1;
  }
  if (!defined $attenuatrbPV)          {$attenuatrbPV = $attenuatorPV;}	#then it is the same
  if ($attenuatorPV !~ /^${beamline}/) {$attenuatorPV = $beamline.':'.$attenuatorPV;}
  if ($attenuatrbPV !~ /^${beamline}/) {$attenuatrbPV = $beamline.':'.$attenuatrbPV;}

  $statusTotal = 0;
  for ($channel=0; $channel<$nAttenuators; $channel++) {
     $pv = $attenuatrbPV.$channel;
     $status = addCaMonitor($pv);
     if ($status) {
        print STDOUT '*** '.$script.': cannot set CAmonitor on PV='.$pv."\n";
        $statusTotal++;
     }
     ### Do we need these monitors? We only write those PVs:
     $pv = $attenuatorPV.$channel;
     $status = addCaMonitor($pv);
     if ($status) {
        print STDOUT '*** '.$script.': cannot set CAmonitor on PV='.$pv."\n";
        $statusTotal++;
     }
  }
  return $statusTotal;
}
