#!/usr/bin/perl -W
#--------------------------------------------------------------
# These are common subs based on PDL that may be used by other scripts
#
#     	             Author: Sergey Stepanov
#
# Version-1.00:  2010/06/22
# Version-1.01:  2018/11/16 changed internal calls in peakFit from findNoisyPeakRef to findPeakRef 
#--------------------------------------------------------------
# List of functions:
# ^^^^^^^^^^^^^^^^^
# 1. ($status,$y0,$A,$xc,$w) = &peakFit     (\@x_array,\@y_array,['gaussian|loretzian');        #uses &gaussianFunc (default) or &lorentzianFunc 
# 2. ($status,$y0,$A,$xc,$w) = &peakFit_full(\@x_array,\@y_array,['gaussian|loretzian');        #uses &gaussianFunc_full (default) or &lorentzianFunc_full
# 3. $y = &gaussianFunc       ($xarg,$par,$ym,$dyda);                                           #calculates $y for ($xc,$w) while ($y0,$A) are taken as ($y_min,$y_max-$y_min)
# 4. $y = &gaussianFunc_full  ($xarg,$par,$ym,$dyda);                                           #calculates $y for ($y0,$A,$xc,$w)
# 5. $y = &lorentzianFunc     ($xarg,$par,$ym,$dyda);                                           #calculates $y for ($xc,$w) while ($y0,$A) are taken as ($y_min,$y_max-$y_min)
# 6. $y = &lorentzianFunc_full($xarg,$par,$ym,$dyda);                                           #calculates $y for ($y0,$A,$xc,$w)
# 7. ($status,$a,$b,$c)     = &quadFit(\@x_array, \@y_array, [$eps]);                           #uses &quadFunc to find the fit
# 8. $y = &quadFunc            ($xarg,$par,$ym,$dyda);                                          #calculates $y for ($a,$b,$c)
# 9. ($status,$a,$b)        = &linearFit(\@x_array, \@y_array, [$eps]);                         #uses &linearFunc to find the fit
#10. $y = &linearFunc          ($xarg,$par,$ym,$dyda);                                          #calculates $y for ($a,$b)
#--------------------------------------------------------------
  use strict;
  use PDL;                                      #Perl data language
  use PDL::Fit::LM;                             #Levenberg-Marquardt fitting
# use PDL::Math;				#extended math functions (atan, erf,...)
# use warnings::unused;

#==========================================================================

# ($status,$y0,$A,$xc,$w) = &peakFit(\@x_array, \@y_array, ['gaussian|loretzian');
sub peakFit (\@\@;$) {

### Fit the curve given by @x_array and @y_array to a Gaussian using the
### Marquardt-Levenberg non-linear least squares fitting (simplified 
### version where the amplitude and minimum are simply taken from data,
### which should not be noisy).

  our $report;
  our ($peakFit_y0, $peakFit_A);
  my ($script, $x_ref, $y_ref, $fitfunc, $y0, $A, $xc, $w, $FWHM);
  my ($npts, $y_max, $y_min, $pt_max, $pt_min, $pt_FWHM, $pt_left, $pt_right);
  my ($xpdl, $ypdl, $inipar, $weight);
  my ($yfit, $parfit, $covarianceMatrix, $iterations);

  $script = 'peakFit';
  $x_ref   = shift(@_);
  $y_ref   = shift(@_);
  $fitfunc = shift(@_);		if (!defined $fitfunc) {$fitfunc = 'gaussian';}
  if ($fitfunc !~ /^gaussian$/i && $fitfunc !~ /^lorentzian$/i) {
    print STDOUT  '*** '.$script.': unknown fit function "'.$fitfunc.'"'."\n";
    return 3;
  }
  if (!defined $report) {$report=0;}

  $npts = @$x_ref;
  ($y_max,$y_min,$pt_max,$pt_min,$pt_FWHM,$pt_left,$pt_right) = &findPeakRef($npts,$y_ref,'auto','noisy');
  if ($pt_FWHM <= 0) {						        #if peak not found
     print STDOUT '*** '.$script.': no peak found on data!'."\n";
     return 1;
  }
### Initial data for fitting:
  $y0   = $y_min;
  $A    = $y_max-$y_min;
  $xc   = @$x_ref[$pt_max];
  $FWHM = abs(@$x_ref[$pt_right] - @$x_ref[$pt_left]);
### See: http://en.wikipedia.org/wiki/Gaussian_function
### See: http://en.wikipedia.org/wiki/Lorentzian_function
  if    ($fitfunc =~ /^gaussian$/i)   {$w = $FWHM/2.35482;}	        #$w=$FWHM/(2*sqrt(2*ln(2)))
  elsif ($fitfunc =~ /^lorentzian$/i) {$w = $FWHM/2.;}
  if ($report) {
     printf STDOUT "\n".$script.': initial fitting parameters for LMFIT fitting:'."\n"
			  .'  y0 = %.3f'."\n"
			  .'   A = %.3f'."\n"
			  .'  xc = %.3f'."\n"
			  .'FWHM = %.3f'."\n", $y0, $A, $xc, $FWHM;
  }
  $xpdl = pdl ($x_ref);
  $ypdl = pdl ($y_ref);

# $inipar = pdl [$y0, $A, $xc, $w];				        #this is for full (all 4 pars) fit
  $inipar = pdl [$xc, $w]; $peakFit_y0=$y0; $peakFit_A=$A;
  $weight = 1;							        #weights of different points in fitting
  if ($fitfunc =~ /^gaussian$/i) {
     ($yfit, $parfit, $covarianceMatrix, $iterations) =
                lmfit ($xpdl, $ypdl, $weight, \&gaussianFunc,           #see: PDL::Fit::LM
                       $inipar, {Maxiter => 300, Eps => 1e-4});
  }
  elsif  ($fitfunc =~ /^lorentzian$/i) {
     ($yfit, $parfit, $covarianceMatrix, $iterations) =
                lmfit ($xpdl, $ypdl, $weight, \&lorentzianFunc,         #see PDL::Fit::LM
                       $inipar, {Maxiter => 300, Eps => 1e-4});
  }
  if (defined $parfit) {
#    ($y0, $A, $xc, $w) = list $parfit;
     ($xc, $w) = list $parfit;
     $w = abs($w);
     if    ($fitfunc =~ /^gaussian$/i)   {$FWHM = $w*2.35482;}	        #$FWHM=2*$w*sqrt(2*ln(2))
     elsif ($fitfunc =~ /^lorentzian$/i) {$FWHM = $w*2.;}	        #$FWHM=2*$w
     if ($report) {
        printf STDOUT $script.': '.$fitfunc.' fitting parameters found by LMFIT:'."\n"
			  .'  y0 = %.3f'."\n"
			  .'   A = %.3f'."\n"
			  .'  xc = %.3f'."\n"
			  .'FWHM = %.3f'."\n\n", $y0, $A, $xc, $FWHM;
     }
     return 0, $y0, $A, $xc, $w;
  } else {
     print STDOUT '*** '.$script.' '.$fitfunc.' fitting with LMFIT failed'."\n"
                 .'*** '.$script.' LMFIT number of iterations='.$iterations."\n\n";
     return 2;
  }
}

#==========================================================================

# ($status,$y0,$A,$xc,$w) = &peakFit_full(\@x_array, \@y_array, ['gaussian|loretzian');
sub peakFit_full (\@\@;$) {

### Fit the curve given by @x_array and @y_array to a Gaussian using the
### Marquardt-Levenberg non-linear least squares fitting:

  our $report;
  my ($script, $x_ref, $y_ref, $fitfunc, $y0, $A, $xc, $w, $FWHM);
  my ($npts, $y_max, $y_min, $pt_max, $pt_min, $pt_FWHM, $pt_left, $pt_right);
  my ($xpdl, $ypdl, $inipar, $weight);
  my ($yfit, $parfit, $covarianceMatrix, $iterations);

  $script = 'peakFit_full';
  $x_ref   = shift(@_);
  $y_ref   = shift(@_);
  $fitfunc = shift(@_);		if (!defined $fitfunc) {$fitfunc = 'gaussian';}
  if ($fitfunc !~ /^gaussian$/i && $fitfunc !~ /^lorentzian$/i) {
    print STDOUT  '*** '.$script.': unknown fit function "'.$fitfunc.'"'."\n";
    return 3;
  }
  if (!defined $report) {$report=0;}

  $npts = @$x_ref;
  ($y_max,$y_min,$pt_max,$pt_min,$pt_FWHM,$pt_left,$pt_right) = &findPeakRef($npts,$y_ref,'auto','noisy');
  if ($pt_FWHM <= 0) {						        #if peak not found
     print STDOUT '*** '.$script.': no peak found on data!'."\n";
     return 1;
  }
### Initial data for fitting:
  $y0   = $y_min;
  $A    = $y_max-$y_min;
  $xc   = @$x_ref[$pt_max];
  $FWHM = abs(@$x_ref[$pt_right] - @$x_ref[$pt_left]);
### See: http://en.wikipedia.org/wiki/Gaussian_function
### See: http://en.wikipedia.org/wiki/Lorentzian_function
  if    ($fitfunc =~ /^gaussian$/i)   {$w = $FWHM/2.35482;}	        #$w=$FWHM/(2*sqrt(2*ln(2)))
  elsif ($fitfunc =~ /^lorentzian$/i) {$w = $FWHM/2.;}
  if ($report) {
     printf STDOUT "\n".$script.': initial fitting parameters for LMFIT fitting:'."\n"
			  .'  y0 = %.3f'."\n"
			  .'   A = %.3f'."\n"
			  .'  xc = %.3f'."\n"
			  .'FWHM = %.3f'."\n", $y0, $A, $xc, $FWHM;
  }
  $xpdl = pdl ($x_ref);
  $ypdl = pdl ($y_ref);
  $inipar = pdl [$y0, $A, $xc, $w];
  $weight = 1;							        #weights of different points in fitting
  if ($fitfunc =~ /^gaussian$/i) {
     ($yfit, $parfit, $covarianceMatrix, $iterations) =
                lmfit ($xpdl, $ypdl, $weight, \&gaussianFunc_full,      #see PDL::Fit::LM
                       $inipar, {Maxiter => 300, Eps => 1e-4});
  }
  elsif  ($fitfunc =~ /^lorentzian$/i) {
     ($yfit, $parfit, $covarianceMatrix, $iterations) =
                lmfit ($xpdl, $ypdl, $weight, \&lorentzianFunc_full,    #see PDL::Fit::LM
                        $inipar, {Maxiter => 300, Eps => 1e-4});
  }
  if (defined $parfit) {
     ($y0, $A, $xc, $w) = list $parfit;
     $w = abs($w);
     if    ($fitfunc =~ /^gaussian$/i)   {$FWHM = $w*2.35482;}	        #$FWHM=$w*(2*sqrt(2*ln(2)))
     elsif ($fitfunc =~ /^lorentzian$/i) {$FWHM = $w*2.;}
     if ($report) {
        printf STDOUT $script.': '.$fitfunc.' fitting parameters found by LMFIT:'."\n"
			  .'  y0 = %.3f'."\n"
			  .'   A = %.3f'."\n"
			  .'  xc = %.3f'."\n"
			  .'FWHM = %.3f'."\n\n", $y0, $A, $xc, $FWHM;
     }
     return 0, $y0, $A, $xc, $w;
  } else {
     print STDOUT '*** '.$script.' '.$fitfunc.' fitting with LMFIT failed'."\n"
                 .'*** '.$script.' LMFIT number of iterations='.$iterations."\n\n";
     return 2;
  }
}

#==========================================================================

#sub gaussianFunc ($xarg,$par,$ym,$dyda) {              #y = y0+A*exp[-2(x-xc)^2/w^2]
sub gaussianFunc ($$$$) {
#
### See: http://en.wikipedia.org/wiki/Gaussian_function
#
# We use: y=y0+A*exp[-(x-xc)^2/2w^2]; then at x-xc=w*sqrt(2*ln(2)) => y=y0+A/2 and FWHM=2*w*sqrt(2*ln(2))=2.35482*w
#
### Input parameters:
# $xarg - a vector (piddle) of independent values $x
# $par  - a vector (piddle) of fitting parameters
# $ym   - a vector (piddle) of dependent variables that will be assigned upon return
# $dyda - a matrix of partial derivatives with respect to the fitting parameters that will be assigned upon return

  our ($peakFit_y0, $peakFit_A);
  my ($xarg, $par, $ym, $dyda);
  my ($y0, $A, $xc, $w, @dy);
  my ($arg, $expo);
  ($xarg, $par, $ym, $dyda) = @_;
  ($xc, $w) = map {$par->slice("($_)")} (0..1);                         #only 2 parameters are varied
  ($y0, $A) = ($peakFit_y0, $peakFit_A);                                #the rest 2 are fixed & passed as "our"
  (@dy)     = map {$dyda->slice(",($_)")} (0..1);
  $arg = ($xarg-$xc)/$w;
  $expo = exp(-0.5*$arg*$arg);
  $ym    .= $y0 + $A*$expo;                                             #function itself
  $dy[0] .= $A*$expo*$arg/$w;                                           #derivative over par[0], i.e. over $xc
  $dy[1] .= $A*$expo*$arg*$arg/$w;                                      #derivative over par[1], i.e. over $w
# $dy[2] .= 0;
# $dy[3] .= 0;
}

#==========================================================================

#sub gaussianFunc_full ($xarg,$par,$ym,$dyda) {         #y = y0+A*exp[-2(x-xc)^2/w^2]
sub gaussianFunc_full ($$$$) {
#
# We use: y = y0+A*exp[-2(x-xc)^2/w^2]; then at x-xc=w/2 => y=y0+A*exp[-1/2]
#
### Input parameters:
# $xarg - a vector (piddle) of independent values $x
# $par  - a vector (piddle) of fitting parameters
# $ym   - a vector (piddle) of dependent variables that will be assigned upon return
# $dyda - a matrix of partial derivatives with respect to the fitting parameters that will be assigned upon return

  my ($xarg, $par, $ym, $dyda);
  my ($y0, $A, $xc, $w, @dy);
  my ($arg, $expo);
  ($xarg, $par, $ym, $dyda) = @_;
  ($y0, $A, $xc, $w) = map {$par->slice("($_)")} (0..3);                #all 4 parameters are varied
  (@dy)              = map {$dyda->slice(",($_)")} (0..3);
  $arg = ($xarg-$xc)/$w;
  $expo = exp(-0.5*$arg*$arg);                                          
  $ym    .= $y0 + $A*$expo;                                             #function itself
  $dy[0] .= 1.;                                                         #derivative over par[0], i.e. over $y0
  $dy[1] .= $expo;                                                      #derivative over par[1], i.e. over $A
  $dy[2] .= $A*$expo*$arg/$w;                                           #derivative over par[2], i.e. over $xc
  $dy[3] .= $A*$expo*$arg*$arg/$w;                                      #derivative over par[3], i.e. over $w
}

#==========================================================================

#sub lorentzianFunc ($xarg,$par,$ym,$dyda) {            #y=y0+A/[(x-xc)^2/w^2+1]
sub lorentzianFunc ($$$$) {
#
### See: http://en.wikipedia.org/wiki/Lorentzian_function
#
# We use: y=y0+A/[(x-xc)^2/w^2+1]; then at x-xc=w => y=y0+A/2 and FWHM=2*w
#
### Input parameters:
# $xarg - a vector (piddle) of independent values $x
# $par  - a vector (piddle) of fitting parameters
# $ym   - a vector (piddle) of dependent variables that will be assigned upon return
# $dyda - a matrix of partial derivatives with respect to the fitting parameters that will be assigned upon return

  our ($peakFit_y0, $peakFit_A);
  my ($xarg, $par, $ym, $dyda);
  my ($y0, $A, $xc, $w, @dy);
  my ($arg, $lrnz);
  ($xarg, $par, $ym, $dyda) = @_;
  ($xc, $w) = map {$par->slice("($_)")} (0..1);                         #only 2 parameters are varied
  ($y0, $A) = ($peakFit_y0, $peakFit_A);                                #the rest 2 are fixed & passed as "our"
  (@dy)     = map {$dyda->slice(",($_)")} (0..1);
  $arg  = ($xarg-$xc)/$w;
  $lrnz = 1./($arg*$arg+1.);
  $ym    .= $y0 + $A*$lrnz;                                             #function itself
  $dy[0] .= (2.*$A*$arg/$w)*$lrnz*$lrnz;                                #derivative over par[0], i.e. over $xc
  $dy[1] .= (2.*$A/$w)*$lrnz*(1.-$lrnz);                                #derivative over par[1], i.e. over $w
# $dy[2] .= 0;
# $dy[3] .= 0;
}

#==========================================================================

#sub lorentzianFunc_full ($xarg,$par,$ym,$dyda) {       #y=y0+A*w^2/[4(x-xc)^2+w^2]
sub lorentzianFunc_full ($$$$) {
#
# We use: y = y0+A*w^2/[4(x-xc)^2+w^2]; then at x-xc=w/2 => y=y0+A/2
#
### Input parameters:
# $xarg - a vector (piddle) of independent values $x
# $par  - a vector (piddle) of fitting parameters
# $ym   - a vector of dependent variables that will be assigned upon return
# $dyda - a matrix of partial derivatives with respect to the fitting parameters that will be assigned upon return

  my ($xarg, $par, $ym, $dyda);
  my ($y0, $A, $xc, $w, @dy);
  my ($arg, $lrnz);
  ($xarg, $par, $ym, $dyda) = @_;
  ($y0, $A, $xc, $w)        = map {$par->slice("($_)")} (0..3);         #all 4 parameters are varied
  (@dy)                     = map {$dyda->slice(",($_)")} (0..3);
  $arg  = ($xarg-$xc)/$w;
  $lrnz = 1./($arg*$arg+1.);
  $ym    .= $y0 + $A*$lrnz;                                             #function itself
  $dy[0] .= 1.;                                                         #derivative over par[0], i.e. over $y0
  $dy[1] .= $lrnz;                                                      #derivative over par[1], i.e. over $A
  $dy[2] .= (2.*$A*$arg/$w)*$lrnz*$lrnz;                                #derivative over par[2], i.e. over $xc
  $dy[3] .= (2.*$A/$w)*$lrnz*(1.-$lrnz);                                #derivative over par[3], i.e. over $w
}

#==========================================================================

# ($status,$a,$b,$c) = &quadFit(\@x_array, \@y_array, [$eps]);
sub quadFit (\@\@;$) {                                   #y = a*x^2 + b*x + c

### Fit the curve given by @x_array and @y_array to a quadratic polinomial
### using the Marquardt-Levenberg non-linear least squares fitting:

  use if ! defined &PDL,          'PDL';
  use if ! defined &PDL::Fit::LM, 'PDL::Fit::LM';

  my ($script, $x_ref, $y_ref, $eps, $a, $b, $c, $npts);
  my ($xpdl, $ypdl, $inipar, $weight);
  my ($yfit, $parfit, $covarianceMatrix, $iterations);
  my ($x1, $y1, $x2, $y2);

  $script = 'quadFit';
  $x_ref  = shift(@_);
  $y_ref  = shift(@_);
  $eps    = shift(@_);   
  if (!defined $eps) {$eps = 1.e-4;}
  if ($eps < 1.e-4)  {$eps = 1.e-4;}

  $xpdl = pdl ($x_ref);
  $ypdl = pdl ($y_ref);

### Initial data for fitting (using linear fit as a start):
  $npts = @$x_ref;
  if ($npts < 3) {return -1;}
  $a = 0;
  $x1 = $$x_ref[0];  $x2 = $$x_ref[$npts-1];
  $y1 = $$y_ref[0];  $y2 = $$y_ref[$npts-1];
# $y1 = $b*$x1 + $c;
# $y2 = $b*$x2 + $c;
  if ($x1 == $x2) {return -2;}
  $b = ($y2-$y1)/($x2-$x1);
  $c = $y2 - $b*$x2;
  $inipar = pdl [$a, $b, $c];
  $weight = 1;							        #weights of different points in fitting
  ($yfit, $parfit, $covarianceMatrix, $iterations) =
                lmfit ($xpdl, $ypdl, $weight, \&quadFunc,               #see: PDL::Fit::LM
                       $inipar, {Maxiter=>300,Eps=>$eps});
  if (defined $parfit) {
     ($a, $b, $c) = list $parfit;
     return 0, $a, $b, $c;
  } else {
     print STDOUT '*** '.$script.' LMFIT quadratic fitting failed (interations='.$iterations.').'."\n\n";
     return -3;
  }
}

#==========================================================================

#sub quadFunc ($x,$par,$ym,$dyda) {                                     #y = a*x^2 + b*x + c
sub quadFunc ($$$$) {
#
# We use: y = a*x^2 + b*x + c
#
### Input parameters:
# $x    - a vector (piddle) of independent values $x
# $par  - a vector (piddle) of fitting parameters
# $ym   - a vector (piddle) of dependent variables that will be assigned upon return
# $dyda - a matrix of partial derivatives with respect to the fitting parameters that will be assigned upon return

  my ($x, $par, $ym, $dyda);
  my ($a, $b, $c, @dy);
  my ($arg, $expo);
  ($x, $par, $ym, $dyda) = @_;
  ($a, $b, $c) = map {$par->slice("($_)")} (0..2);
  (@dy)        = map {$dyda->slice(",($_)")} (0..2);
  $ym    .= $a*$x**2 + $b*$x + $c;                                      #function itself
  $dy[0] .= $x**2;                                                      #derivative over par[0], i.e. over $a
  $dy[1] .= $x;                                                         #derivative over par[1], i.e. over $b
  $dy[2] .= 1.;                                                         #derivative over par[2], i.e. over $c
}

#==========================================================================

# ($status,$a,$b,$c) = &linearFit(\@x_array, \@y_array, [$eps]);
sub linearFit (\@\@;$) {                                   #y = a*x + b

### Fit the curve given by @x_array and @y_array to a quadratic polinomial
### using the Marquardt-Levenberg non-linear least squares fitting:

  use if ! defined &PDL,          'PDL';
  use if ! defined &PDL::Fit::LM, 'PDL::Fit::LM';

  my ($script, $x_ref, $y_ref, $eps, $a, $b, $npts);
  my ($xpdl, $ypdl, $inipar, $weight);
  my ($yfit, $parfit, $covarianceMatrix, $iterations);
  my ($x1, $y1, $x2, $y2);

  $script = 'linearFit';
  $x_ref  = shift(@_);
  $y_ref  = shift(@_);
  $eps    = shift(@_);   
  if (!defined $eps) {$eps = 1.e-4;}
  if ($eps < 1.e-4)  {$eps = 1.e-4;}

  $xpdl = pdl ($x_ref);
  $ypdl = pdl ($y_ref);

### Initial data for fitting (using linear fit as a start):
  $npts = @$x_ref;
  if ($npts < 2) {return -1;}
  $x1 = $$x_ref[0];  $x2 = $$x_ref[$npts-1];
  $y1 = $$y_ref[0];  $y2 = $$y_ref[$npts-1];
# $y1 = $a*$x1 + $b;
# $y2 = $a*$x2 + $b;
  if ($x1 == $x2) {return -2;}
  $a = ($y2-$y1)/($x2-$x1);
  $b = $y2 - $a*$x2;
  $inipar = pdl [$a, $b];
  $weight = 1;							        #weights of different points in fitting
  ($yfit, $parfit, $covarianceMatrix, $iterations) =
                lmfit ($xpdl, $ypdl, $weight, \&linearFunc,             #see: PDL::Fit::LM
                       $inipar, {Maxiter=>300,Eps=>$eps});
  if (defined $parfit) {
     ($a, $b) = list $parfit;
     return 0, $a, $b;
  } else {
     print STDOUT '*** '.$script.' LMFIT linear fitting failed (interations='.$iterations.').'."\n\n";
     return -3;
  }
}

#==========================================================================

#sub linearFunc ($x,$par,$ym,$dyda) {                                     #y = a*x + b
sub linearFunc ($$$$) {
#
# We use: y = a*x + b
#
### Input parameters:
# $x    - a vector (piddle) of independent values $x
# $par  - a vector (piddle) of fitting parameters
# $ym   - a vector (piddle) of dependent variables that will be assigned upon return
# $dyda - a matrix of partial derivatives with respect to the fitting parameters that will be assigned upon return

  my ($x, $par, $ym, $dyda);
  my ($a, $b, $c, @dy);
  my ($arg, $expo);
  ($x, $par, $ym, $dyda) = @_;
  ($a, $b) = map {$par->slice("($_)")} (0..1);
  (@dy)    = map {$dyda->slice(",($_)")} (0..1);
  $ym    .= $a*$x + $b;                                                 #function itself
  $dy[0] .= $x;                                                         #derivative over par[0], i.e. over $a
  $dy[1] .= 1;                                                          #derivative over par[1], i.e. over $b
}
