#!/usr/bin/perl -W
#==============================================================
# These are CRL-related subs for beamline controls Perl scripts
#
#     			Author: Sergey Stepanov
#
# Version-1.0:  2019/02/08 implemented calcCRLfocus() for fixed lens set (no In/Out)
# Version-1.2:  2020/09/01 added readCRLconfigSQL() and for clarity renamed 
#                          &readCRLboxSavedPosSQL and &interpolateCRLboxPos.
# Version-1.3:  2020/09/20 added calculating and setting transfocator focus.
#                          The calucation, which includes &matchTransfocatorFocus
#                          &setTransfocatorFocus uses the same simplex recursive
#                          algorithm as in the attenuation subs.
# Version-1.4:  2020/11/20 added calculating weighted center (&calcTransfocatorCenter)
# Version-2.0:  2021/02/11 added &matchTransfocatorFocus2 with another way to find
#                          the best matching focus (storing all lens combinations 
#                          in SQL and then sorting them by best match to focusRqs).
# Version-3.0:  2021/05/15 added a set of routines based on Brul's mathod:
#                          &getTransfocatorFocus3, &calcTransfocatorFocus3,
#                          &matchTransfocatorFocus3, &setTransfocatorFocus3;
#                          also corrected &getTransfocatorFocus and &matchTransfocatorFocus2
#                          to include L/6 (Mamyrbayev);
#                          also corrected attenuation optimization: from the number
#                          of blocks IN to the number of lenses IN (blocks may contain
#                          different number of lenses)
# Version-3.1: 2021/07/19 added &setTransfocatorInOut used by all four &setTransfocatorFocus routines
# Version-3.2: 2021/07/19 added added $prefilter and $filerpm for &matchTransfocatorFocus4 
# Version-3.3: 2021/08/19 simplified &saveTFstate() and &getTFserverState() -- excuded shiftIncluded and centerIncluded
# Version-3.4: 2021/09/15 eliminated center from &readTFstate
# Version-3.5: 2021/09/26 removed obsolete subs (versions 1,2,3) and added &excludeCRLradius()
# Version-3.6: 2021/10/27 removed prefilter - did not work for high energy when there is no good combination
# Version-4.0: 2021/12/13 added &calcTransfocatorFocus5, &getTransfocatorFocus5, &matchTransfocatorFocus5,
#                         &setTransfocatorFocus5, &calcTransfocatorFocusComb and auxilary routines for tetris-like focus matching
#==============================================================
# List of functions:
# ^^^^^^^^^^^^^^^^^
# --------------
#01.        $nrecords = &readCRLboxSavedPosSQL(\@hashBoxPos);					#reads saved CRL-box positions: date,energy,pitch,yaw,height,outboard,downstream
#02.          $status = &interpolateCRLboxPos($E,$poskey,\$pos,[\@hashBoxPos,$dE,$lastNmonths]);#interpolates $pos for $key from SQL data on CRL positions per energy
#03.          $status = &saveTFstate();								#save Transfocator positions, energy, focus, timestamp to PV
#04.          $status = &readTFstate(\$timestamp,\$X,\$H,\$V,\$energy,\$focus);			#read Transfocator positions, energy, focus, timestamp from PV
#
#05.        $nrecords = &readCRLconfigSQL($device,\@hashLensData);				#reads CRL data: device,name,radius,N-lenses,material,PV-In/Out,downstream-offset
#06.        $nrecords = &readCRLconfigSQLpopulated($device,\@hashLensData);			#calls readCRLconfigSQL and then excludes unpopulated groups
#07.        $nrecords = &excludeCRLradius($minRadius,[$device,\$msg]);				#excludes all lenses with radius <= $minRadius from @hashLensData and takes them out
#08.          $status = &readDeltaDB($material,[\$message]);					#if not yet cached, reads Material f' from respective file in $PEZCA/Data & stores in global array @DeltaDB
#09.           $delta = &calcDelta($E,$material,[\$msg]);					#calculates (interpolates) $delta for given x-ray energy and material (e.g. Be)
#10.           $focus = &calcCRLfocus($E,$R,$n,[$material,\$msg]);				#returns focus of CRL block in mm. $R-radius is um, $n=number of lenses. Default material=Be (Be.dat), can also be $delta($E)
#11.           $power = &calcCRLpower($E,$R,$n,[$material,\$msg]);				#returns focus of CRL block in mm. $R-radius is um, $n=number of lenses. Default material=Be (Be.dat), can also be $delta($E)
#12.      @oneByFocus = &calcAllOneByFocus([$E,$device,\$msg]);					#returns array of power=1/focus for all Transfocator lens groups. Defaults: E=current, device=CRL-2020
#13.          $status = &getTransfocatorInOut(\@InOut,[$device,\$msg]);				#returns current transfocator lenses In/Out state
#14.          $status = &setTransfocatorInOut(\@InOut,[$device,\$msg]);				#sets requested transfocator lenses In/Out states
#15. ($status,$ncorr) = &setTransfocatorRB2BO([$device,\$msg]);					#sets all transfocator BO from RB. Useful after reboot when BO is lost, but RB is stored in the digital I/O device.
#16.            $n_in = &InOut_str2arr($str,\@InOut);						#converts a string of In/Out states (e.g. '11010001111000') into respective array
#17.     ($str,$n_in) = &InOut_arr2str(\@InOut);						#converts an array of In/Out states (In=1,Out=0) into respective string (e.g. '11010001111000')
#
#------begin Bruls method with principal plane & checking all combinations ---v
#18.             $focus = &getTransfocatorFocus4 ([$E,$device,\$msg,\@InOut]);			#returns overall Transfocator focus (Bruls method) in mm for current lenses in/out configuration (can return it too). Defaults: E=current, device=CRL-2020
#19.             $focus = &calcTransfocatorFocus4 (\@InOut,[$E,$device,\$msg]);			#returns overall Transfocator focus (Bruls method) in mm for specified lenses in/out configuration. Defaults: E=current, device=CRL-2020
#20. ($focusAct,@InOut) = &matchTransfocatorFocus4 ($focusRqs,[$E,$device,\$msg,		#match requested Transfocator focus (Bruls method, check all combinations) - calc, but do not set.
#                                                   $minimizeAttenuation,
#                                                   $focusRange,$TF_axis_PV]);
#21. ($focusAct,@InOut) = &setTransfocatorFocus4  ($focusRqs,[$E,$device,\$msg,			#match requested Transfocator focus  (Bruls method, check all combinations) - calc & set.
#                                                  $minimizeAttenuation,
#                                                  $focusRange,$TF_axis_PV]);
#------end Bluls method with principal plane & checking all combinations -----^
#
#------begin Bruls method with principal plane & tetris-like search for best match ---v
#22.            $status = &fillCRLcombPD(\@CRLcomb,[$E,$device,\$msg]);				#Fills power and distances for the @CRLcomb array used by &calcTransfocatorFocus5
#23.            $status = &fillCRLcombIO(\@CRLcomb,\@IO);					#Fills In/Out for the @CRLcomb array used by &calcTransfocatorFocus5
#24.             $focus = &getTransfocatorFocus5 ($E,$device,[\$msg,\@InOut]);			#returns overall Transfocator focus (Bruls method) in mm for current lenses in/out (same as getTransfocatorFocus4)
#25.             $focus = &calcTransfocatorFocus5 (\@InOut,[$E,$device,\$msg]);			#returns overall Transfocator focus (Bruls method) in mm for specified lenses in/out (same as calcTransfocatorFocus4). Defaults: E=current, device=CRL-2020
#26.             $focus = &calcTransfocatorFocusComb (\@CRLcombination);			#returns overall Transfocator focus (Bruls method) for given combination of {'Power'}, {'Offset'}, {'InOut'}.
#27. ($focusAct,@InOut) = &matchTransfocatorFocus5 ($focusRqs,[$E,$device,\$msg]);		#match requested Transfocator focus (Bruls method, "tetris" match) - calc, but do not set.
#28. ($focusAct,@InOut) = &setTransfocatorFocus5 ($focusRqs,[$E,$device,\$msg]);		#match requested Transfocator focus (Bruls method, "tetris" match) - calc & set.
#------end Bluls method with principal plane & tetris-like search for best match -----^
#
#29.       ($serverState,$focusRef,
#            $focusRqs,$focusAct,$xPos,   
#            $searchMode,$rangeMax) = &getTFserverState();					#read all parameters of Transfocator Server from PVs
#30.              ($focusAct,$xPos) = &setTFserverFocus($focusRqs,[$wait_max]);			#command the Transfocator Server to set requested focus with whatever method is used
#
#==============================================================
# use Pezca;

# use warnings::unused;
  use strict;
# no strict 'vars';
# no strict 'subs';
# no strict 'refs';

#### 1 or commented out = check, 0=no check ####
# our $transfocatorCheckInOut = 1;
#### 1 or commented out = check, 0=no check ####

#=====================================================================

# $nrecords = &readCRLboxSavedPosSQL(\@hashBoxPos);
sub readCRLboxSavedPosSQL(\@) {
###
### Reads saved CRL-box positions: date,energy,pitch,yaw,height,outboard,downstream
###
  our ($mysqlHost, $mysqlUser, $mysqlPass, $mysqlDb);			#this is from beamline.pl
  my ($mysqlTable, $dsn, $dbh, $sth);
  my ($script, $refBoxPos, $nrecords, $i);

  $script    = 'readCRLboxSavedPosSQL';
  $refBoxPos = shift(@_);

  $mysqlTable = 'CRLpositions';
  $dsn = 'DBI:mysql:host='.$mysqlHost.';database='.$mysqlDb;
# print STDOUT $script.': Connecting to mySQL database='.$mysqlDb.' on host='.$mysqlHost."\n";
  $dbh = DBI->connect($dsn,$mysqlUser,$mysqlPass,{PrintError=>0,});
  if (!defined $dbh) {
     print STDOUT '*** '.$script.': Failed connecting to mySQL DB='.$mysqlDb.' on '.$mysqlHost."\n";
     return -1;
  }
  $sth = $dbh->prepare('SELECT * FROM '.$mysqlTable);
  if (!defined $sth) {
     print STDOUT '*** '.$script.': Error on mySQL table='.$mysqlTable.'. -- '.$dbh->errstr."\n";
     $dbh->disconnect(); 
     return -1;
  }
  $nrecords = $sth->execute();						#nrecords works for MySQL only. For SQLite result is def/undef only.
  if (!defined $nrecords) {
     print STDOUT '*** '.$script.': Failed executing SELECT for table='.$mysqlTable.'. -- '.$sth->errstr."\n";
     $sth->finish(); 
     $dbh->disconnect(); 
     return -1;
  }
  if ($nrecords < 1) {
     print STDOUT '!!! '.$script.': No saved CRL positions yet in table='.$mysqlTable."\n";
     $sth->finish(); 
     $dbh->disconnect(); 
     return $nrecords;
  }
  for ($i=0; $i<$nrecords; $i++) {
     $$refBoxPos[$i] = $sth->fetchrow_hashref();
#    $energy     = $$refBoxPos[$i]->{'energy'};
#    $epoch      = $$refBoxPos[$i]->{'epoch'};
#    $date       = $$refBoxPos[$i]->{'date'};
#    $pitch      = $$refBoxPos[$i]->{'pitch'};
#    $yaw        = $$refBoxPos[$i]->{'yaw'};
#    $height     = $$refBoxPos[$i]->{'height'};
#    $outboard   = $$refBoxPos[$i]->{'outboard'};
#    $downstream = $$refBoxPos[$i]->{'downstream'};
#    $bigheight  = $$refBoxPos[$i]->{'bigheight'};
  }
  $sth->finish(); 
  $dbh->disconnect();
  return $nrecords;
}

#=====================================================================

#sub interpolateCRLboxPos($E,$poskey,\$pos,[\@hashBoxPos,$dE,$lastNmonths]) {
sub interpolateCRLboxPos($$\$;\@$$) {
###
### Interpolates $pos for $poskey from SQL data on CRL positions vs energy
### where $poskey is one of (pitch yaw height outboard downstream bigheight)
###
  my ($script, $E, $poskey, $refBoxPos, $posref, $dE, $age);
  my ($nrec, $epoch, $secnds, $estart, $i, @keys);
  my ($E_hash, $ep_hash, @energy, @positn, @indx, $nfit);
  my ($status, $aa, $bb, $cc, $base, $tolerance, $diff);
  
  $script    = 'interpolateCRLboxPos';
  $E         = shift(@_);						#energy of interest in keV
  $poskey    = shift(@_);						#pitch,yaw,height,outboard,downstream,bigheight
  $posref    = shift(@_);						#output: position to calculate
  $refBoxPos = shift(@_);						#array of hashes with SQL data (if not defined, will be filled in the sub)
  $dE        = shift(@_); if (!defined $dE)  {$dE=1.E9;}		#SQL energy data filter: +-range in keV
  $age       = shift(@_); if (!defined $age) {$age=12*10;}		#SQL data age filter in months

# $energy     = $$refBoxPos[$i]->{'energy'};
# $epoch      = $$refBoxPos[$i]->{'epoch'};
# $date       = $$refBoxPos[$i]->{'date'};
# $pitch      = $$refBoxPos[$i]->{'pitch'};
# $yaw        = $$refBoxPos[$i]->{'yaw'};
# $height     = $$refBoxPos[$i]->{'height'};
# $outboard   = $$refBoxPos[$i]->{'outboard'};
# $downstream = $$refBoxPos[$i]->{'downstream'};
# $bigheight  = $$refBoxPos[$i]->{'bigheight'};

  @keys = qw(pitch yaw height outboard downstream bigheight);
  if (! grep /^$poskey$/, @keys) {
     print STDOUT '*** '.$script.': Incorrect key='.$poskey.'. Must be one of '."(@keys)\n";
     return -1;
  }

  if (ref(@$refBoxPos) eq 'ARRAY') {$nrec = @$refBoxPos;}
  else                             {$nrec = &readCRLboxSavedPosSQL(\@$refBoxPos);}
  $epoch  = time();
  $secnds = int($age * 30 *24 * 3600);					#from months to seconds
  $estart = $epoch - $secnds;  if ($estart < 0) {$estart = 0;}
  @energy = ();
  @positn = ();
  $nfit = 0;
  if ($poskey =~ /height/) {
     if    ($poskey eq 'height')    {$base = $$refBoxPos[0]->{'bigheight'};}
     elsif ($poskey eq 'bigheight') {$base = $$refBoxPos[0]->{'height'};}
  }
  $tolerance = 0.005;
  for ($i=0; $i<$nrec; $i++) {
     $E_hash  = $$refBoxPos[$i]->{'energy'};
     $ep_hash = $$refBoxPos[$i]->{'epoch'};
     if (abs($E-$E_hash) < $dE && $ep_hash > $estart) {
        $energy[$nfit] = $E_hash;
        $positn[$nfit] = $$refBoxPos[$i]->{$poskey};
        if ($poskey =~ /height/) {
           if    ($poskey eq 'height')    {$diff = abs($base-$$refBoxPos[$i]->{'bigheight'});}
           elsif ($poskey eq 'bigheight') {$diff = abs($base = $$refBoxPos[$i]->{'height'});}
           if ($diff > $tolerance) {
              printf STDOUT '!!! '.$script.': on %s second CRL height deviates from base by %.3fmm'."\n",$$refBoxPos[$i]->{'date'},$diff;
           }
        }
        $nfit++;
     }
  }
  if ($nfit < 3) {
     printf STDOUT '*** '.$script.': Insufficient data ('.$nfit.' pts) for interpolating CRL positions at Energy=%.3f keV.'."\n", $E;
     return -1;
  }
### Sort over energy:
  @indx = sort { $energy[$a] <=> $energy[$b] } 0..$#energy;
  @energy = @energy[@indx];
  @positn = @positn[@indx];
  
### Do Marquardt-Levenberg non-linear least squares fitting (see common_PDL.pl):
  if ($poskey eq 'downstream') {($status,$aa,$bb,$cc) = &quadFit(\@energy,\@positn);}
  else                         {($status,$aa,$bb)   = &linearFit(\@energy,\@positn);}
  if ($status) {
     printf STDOUT '*** '.$script.': Fitting of CRL [%s] SQL data at E=%.3f keV failed.'."\n", $poskey, $E;
     return $status;
  }
  if ($poskey eq 'downstream') {$$posref = $aa*$E**2 + $bb*$E + $cc;}
  else                         {$$posref = $aa*$E + $bb;}
  return 0;
}

#=====================================================================

# sub saveTFstate() {							#returns $status
sub saveTFstate() {							#returns $status
###
### Saves Transfocator positions, energy, focus, timestamp to PV
###
  our ($beamline);
  my ($script, $status);
  my ($X_PV, $X);
  my ($H_PV, $H);
  my ($V_PV, $V);
  my ($E_PV, $E);
  my ($F_PV, $F); 
# my ($C_PV, $C);							# *** deprecated *** 
  my ($str, $PV);
  my (@ct, $year, $mon, $mday, $hour, $minute);
  my ($server_state_PV, $server_state);
# my ($server_shift_PV, $server_shift);					# *** deprecated ***
  $script = 'saveTFstate';
  if (! defined $beamline) {
     print STDOUT '*** '.$script.': undefined $beamline variable.'."\n";
     return 1;
  }
  $X_PV            = $beamline.':CRL:X:ActPos';				#upstream/downstream position
  $H_PV            = $beamline.':CRL:Y:ActPos';				#horizontal position
  $V_PV            = $beamline.':CRL:Z:ActPos';				#vertical position
  $E_PV            = $beamline.':MO:E:ActPos';				#current energy (for reference
  $F_PV            = $beamline.':focusAct';				#current focus (for reference)
# $C_PV            = $beamline.':focusDeviceCenter';			#current weigthened center of TF with respect to its geometrical center (*** deprecated ***)
  $PV              = $beamline.':focusPositionHistory';			#waveform PV to save the information
  $server_state_PV = $beamline.':focusSts';				#transfocator server status PV
# $server_shift_PV = $beamline.':focusShiftIncluded';			#transfocator server shift included PV

### If transfocator server is running and shift is included, read
### current focus and current weigthened center:
  ($status,$server_state) = Pezca::GetDouble($server_state_PV);			#0=stopped,1=processing,2=ready
  if ($status) {&report_only($script,$status,$server_state_PV); $server_state=0;}
# ($status,$server_shift) = Pezca::GetDouble($server_shift_PV);			#0=not included,1=included
# if ($status) {&report_only($script,$status,$server_shift_PV); $server_shift=0;}
  if ($server_state != 2) {
     $str = '*** '.$script.': Cannot save TF state when the server is not running ***';
     print STDOUT $str."\n";
     $status = &PezcaPutStringWF($PV,$str,'oldca');
     &report_only($script,$status,$PV); 
     return -2;
  }
 
# $str = 'D=2016.06.07-13:00 X=12.123 H=0.012 V=0.000 E=13500 F=12345.678 C=12.345';
  @ct = localtime(time);
  ($minute,$hour,$mday,$mon,$year) = ($ct[1],$ct[2],$ct[3],$ct[4]+1,$ct[5]+1900);
  $str = sprintf 'D=%04d-%02d-%02d_%02d:%02d', $year, $mon, $mday, $hour, $minute;

  ($status, $X) = Pezca::GetDouble($X_PV);
  if ($status) {
     &report_only($script,$status,$X_PV); 
     print STDOUT '*** '.$script.': error reading TF downstream position from '.$X_PV."\n";
     return -1;
  }
  $str .= sprintf ' X=%.3f',$X;

  ($status, $H) = Pezca::GetDouble($H_PV);
  if ($status) {
     &report_only($script,$status,$H_PV); 
     print STDOUT '*** '.$script.': error reading TF horizontal position from '.$H_PV."\n";
     return -1;
  }
  $str .= sprintf ' H=%.3f',$H;

  ($status, $V) = Pezca::GetDouble($V_PV);
  if ($status) {
     &report_only($script,$status,$V_PV); 
     print STDOUT '*** '.$script.': error reading TF vertical position from '.$V_PV."\n";
     return -1;
  }
  $str .= sprintf ' V=%.3f',$V;

  ($status, $E) = Pezca::GetDouble($E_PV);
  if ($status) {
     &report_only($script,$status,$E_PV); 
     print STDOUT '*** '.$script.': error reading current energy from '.$E_PV."\n";
     return -1;
  }
  $str .= sprintf ' E=%.3f',$E;

  ($status, $F) = Pezca::GetDouble($F_PV);
  if ($status) {
     &report_only($script,$status,$F_PV); 
     print STDOUT '*** '.$script.': error reading current focus from '.$F_PV."\n";
     return -1;
  }
  $str .= sprintf ' F=%.3f',$F;

# ($status, $C) = Pezca::GetDouble($C_PV);				# *** deprecated ***
# if ($status) {
#    &report_only($script,$status,$C_PV); 
#    print STDOUT '*** '.$script.': error reading current TF weighted center from '.$C_PV."\n";
#    return -1;
# }
# $str .= sprintf ' C=%.3f',$C;

### Post all data into waveform:
  $status = &PezcaPutStringWF($PV,$str,'oldca');
  if ($status) {
     &report_only($script,$status,$PV); 
     print STDOUT '*** '.$script.': error saving TF status into '.$PV."\n";
  }
  return $status;
}

#=====================================================================

# sub readTFstate(\$timestamp,\$X,\$H,\$V,\$energy,\$focus) {		#returns status 
sub readTFstate(\$\$\$\$\$\$) {
###
### Reads Transfocator positions, energy, focus, timestamp from PV
###
  our ($beamline);
  my ($script, $D_ref, $X_ref, $H_ref, $V_ref, $E_ref, $F_ref);
  my ($PV, $str, %data, @kwords, $nw, @vals, $key, $i, $status);
  $script = 'readTFstate';
  $D_ref = shift(@_);
  $X_ref = shift(@_);
  $H_ref = shift(@_);
  $V_ref = shift(@_);
  $E_ref = shift(@_);
  $F_ref = shift(@_);
  
  if (! defined $beamline) {
     print STDOUT '*** '.$script.': undefined $beamline variable.'."\n";
     return 1;
  }
  $PV   = $beamline.':focusPositionHistory';				#waveform PV to save the information
  ($status,$str) = &PezcaGetStringWF($PV);
  if ($status) {
     &report_only($script,$status,$PV); 
     print STDOUT '*** '.$script.': error reading TF status from '.$PV."\n";
     return $status;
  }
  if (length($str) == 0) {
     print STDOUT '!!! '.$script.': no TF status in '.$PV."\n";
     return $status;
  }
  %data = map {split /=/, $_, 2} map {split /\s+/, $_} $str;
  @kwords = qw(D X H V E F);
  $nw = @kwords;
  @vals = ();
  $status = 0;
  for ($i=0; $i<$nw; $i++) {
     $key = $kwords[$i];
     if (defined $data{$key}) {
        push(@vals,$data{$key});
     } else {
        print STDOUT '!!! '.$script.': no TF data for key='.$key.' in '.$PV."\n";
        push(@vals,undef);
        $status++;
    }
  }
  for ($i=1; $i<$nw; $i++) {						#skip date
     if (defined $vals[$i] && ! &is_decimal($vals[$i])) {
        $key = $kwords[$i];
        print STDOUT '!!! '.$script.': TF  data='.$vals[$i].' for key='.$key.' in '.$PV.' is not decimal'."\n";
        undef $vals[$i];
        $status++;
     }
  }
  $$D_ref = $vals[0];  $$D_ref =~ s/_/ /;
  $$X_ref = $vals[1]; 
  $$H_ref = $vals[2];
  $$V_ref = $vals[3];
  $$E_ref = $vals[4];
  $$F_ref = $vals[5];
  return $status;
}

#=====================================================================

# $status = &readCRLconfigSQL($device,\@hashLensData);
sub readCRLconfigSQL($\@) {
###
### Reads CRL data from SQL: device,name,radius,N-lenses,material,PV-In/Out,downstream-offset
###
  our ($beamline);							#defined in beamline.pl
  our ($mysqlHost, $mysqlUser, $mysqlPass, $mysqlDb);                   #this is from beamline.pl
  my ($mysqlTable, $dsn, $dbh, $sth);
  my ($script, $device, $refLensData, $nrecords, $i);

  $script      = 'readCRLconfigSQL';
  $device      = shift(@_);
  $refLensData = shift(@_);						#reference to hash array with lens data
  if (!defined $beamline) {die $script.': no beamline defined.'."\n";}
  if (!defined $device) {
     $device = 'CRL-2020';
     print STDOUT $script.': transfocator device is set to "'.$device.'"'."\n";
  }

  $mysqlTable = 'CRLconfig';
  $dsn = 'DBI:mysql:host='.$mysqlHost.';database='.$mysqlDb;
  print STDOUT $script.': Connecting to mySQL database='.$mysqlDb.' on host='.$mysqlHost."\n";
  $dbh = DBI->connect($dsn,$mysqlUser,$mysqlPass,{PrintError=>0,});
  if (!defined $dbh) {
     print STDOUT '*** '.$script.': Failed connecting to mySQL DB='.$mysqlDb.' on '.$mysqlHost."\n";
     return -1;
  }
  if ($device ne '*') {$sth = $dbh->prepare('SELECT * FROM '.$mysqlTable.' WHERE Device="'.$device.'"');}
  else                {$sth = $dbh->prepare('SELECT * FROM '.$mysqlTable.' WHERE Device like "%"');}
  if (!defined $sth) {
     print STDOUT '*** '.$script.': Error on mySQL table='.$mysqlTable.'. -- '.$dbh->errstr."\n";
     $dbh->disconnect(); 
     return -1;
  }
  $nrecords = $sth->execute();						#nrecords works for MySQL only. For SQLite result is def/undef only.
  if (!defined $nrecords) {
     print STDOUT '*** '.$script.': Failed executing SELECT for table='.$mysqlTable.' and Device="'.$device.'". -- '.$sth->errstr."\n";
     $sth->finish(); 
     $dbh->disconnect(); 
     return -1;
  }
  if ($nrecords < 1) {
     print STDOUT '!!! '.$script.': No data for CRL Device="'.$device.'" in table='.$mysqlTable."\n";
     $sth->finish(); 
     $dbh->disconnect(); 
     return $nrecords;
  }
  for ($i=0; $i<$nrecords; $i++) {
     $$refLensData[$i] = $sth->fetchrow_hashref();

     if (!defined $$refLensData[$i]->{'PV_ctrl'}) {
        die '*** '.$script.': no PV_ctrl PV for device='.$$refLensData[$i]->{'Device'}.', lensNr='.$i."\n";
     }
     if (!defined $$refLensData[$i]->{'PV_rb_in'}) {
        die '*** '.$script.': no PV_rb_in PV for device='.$$refLensData[$i]->{'Device'}.', lensNr='.$i."\n";
     }
     if (!defined $$refLensData[$i]->{'PV_rb_out'}) {
        die '*** '.$script.': no PV_rb_out PV for device='.$$refLensData[$i]->{'Device'}.', lensNr='.$i."\n";
     }

     if (length($$refLensData[$i]->{'PV_ctrl'}) > 0 && $$refLensData[$i]->{'PV_ctrl'} !~ /^$beamline/) {
        $$refLensData[$i]->{'PV_ctrl'}   = $beamline.':'.$$refLensData[$i]->{'PV_ctrl'};
     }
     if (length($$refLensData[$i]->{'PV_rb_in'}) > 0 && $$refLensData[$i]->{'PV_rb_in'} !~ /^$beamline/) {
        $$refLensData[$i]->{'PV_rb_in'}  = $beamline.':'.$$refLensData[$i]->{'PV_rb_in'};
     }
     if (length($$refLensData[$i]->{'PV_rb_out'}) > 0 && $$refLensData[$i]->{'PV_rb_out'} !~ /^$beamline/) {
        $$refLensData[$i]->{'PV_rb_out'} = $beamline.':'.$$refLensData[$i]->{'PV_rb_out'};
     }
#    $device    = $$refLensData[$i]->{'Device'};			#CRL-box name
#    $name      = $$refLensData[$i]->{'Name'};				#Name of lenses subset driven by $PV_ctrl
#    $PV_ctrl   = $$refLensData[$i]->{'PV_ctrl'};			#Digital output for lenses In/Out
#    $PV_rb_in  = $$refLensData[$i]->{'PV_rb_in'};			#Digital input for lenses in the beam
#    $PV_rb_out = $$refLensData[$i]->{'PV_rb_out'};			#Digital input for lenses out of the beam
#    $offset    = $$refLensData[$i]->{'Offset'};			#Offset (mm) of lenses subset from the centre of the whole box (may not be used)
#    $radius    = $$refLensData[$i]->{'Radius'};			#Lenses radius (um). If $radius=0 no lenses are installed in the subset
#    $number    = $$refLensData[$i]->{'Number'};			#Number of lenses in the box. In $number=0, no lenses are installed in the subset
#    $material  = $$refLensData[$i]->{'Material'};			#Lenses material, normally Be. If $material is blank, no lenses are installed in the subset
  }
  $sth->finish(); 
  $dbh->disconnect();
  @$refLensData = sort{ $a->{'Name'} cmp $b->{'Name'} } @$refLensData;
  return $nrecords;
}

#=====================================================================

# $status = &readCRLconfigSQLpopulated($device,\@hashLensData);
sub readCRLconfigSQLpopulated($\@) {
###
### Calls &readCRLconfigSQL and then excludes unpopulated groups of lenses
### (those were  Radius, Number of lenses or Material are missing)
###
  my ($script, $device, $refLensData, $nrecords, $i, $excluded, $material);

  $script      = 'readCRLconfigSQLpopulated';
  $device      = shift(@_);
  $refLensData = shift(@_);

  $nrecords = &readCRLconfigSQL($device,$refLensData);
  if ($nrecords < 1) {return -1;}

  if (!defined $device) {
     $device = $$refLensData[0]->{'Device'};
     print STDOUT $script.': transfocator device is set to "'.$device.'"'."\n";
  }

  $excluded = 0;
  for ($i=0; $i<$nrecords; $i++) {
     if ($$refLensData[$i]->{'Radius'} <= 0 ||
         $$refLensData[$i]->{'Number'} == 0 ||
         $$refLensData[$i]->{'Material'} eq '') {
        $material = $$refLensData[$i]->{'Material'};
        if (length($material) == 0) {$material = '<not specified>';}
        printf STDOUT $script.': excluding lens block [%s] due to conditions: radius=%s nlenses=%s material=%s'."\n", 
                                           $$refLensData[$i]->{'Name'}, 
                                           $$refLensData[$i]->{'Radius'}, 
                                           $$refLensData[$i]->{'Number'},
                                           $material;

        splice @$refLensData,$i,1;
        $i--;
        $nrecords--;
        $excluded++;
     }
  }
  @$refLensData = sort{ $a->{'Name'} cmp $b->{'Name'} } @$refLensData;

# if ($excluded > 0) {print STDOUT $script.': '.$excluded.' groups excluded as unpopulated from the CRL Device="'.$device.'"'."\n";}
  if ($nrecords < 1) {print STDOUT '!!! '.$script.': No data for CRL Device="'.$device.'": all CRL groups are disabled'."\n";}
# for ($i=0; $i<$nrecords; $i++) {
#    $device    = $$refLensData[$i]->{'Device'};			#CRL-box name
#    $name      = $$refLensData[$i]->{'Name'};				#Name of lenses subset driven by $PV_ctrl
#    $PV_ctrl   = $$refLensData[$i]->{'PV_ctrl'};			#Digital output for lenses In/Out
#    $PV_rb_in  = $$refLensData[$i]->{'PV_rb_in'};			#Digital input for lenses in the beam
#    $PV_rb_out = $$refLensData[$i]->{'PV_rb_out'};			#Digital input for lenses out of the beam
#    $offset    = $$refLensData[$i]->{'Offset'};			#Offset (mm) of lenses subset from the centre of the whole box (may not be used)
#    $radius    = $$refLensData[$i]->{'Radius'};			#Lenses radius (um). If $radius=0 no lenses are installed in the subset
#    $number    = $$refLensData[$i]->{'Number'};			#Number of lenses in the box. In $number=0, no lenses are installed in the subset
#    $material  = $$refLensData[$i]->{'Material'};			#Lenses material, normally Be. If $material is blank, no lenses are installed in the subset
# }
  return $nrecords;
}

#=====================================================================

# $nrecords = &excludeCRLradius($minRadius,[$device,\$msg]);
sub excludeCRLradius($;$\$) {
###
### Excludes all lenses with radius <= $minRadius from @hashLensData 
### and takes them out of the beam (used for low energies)
###
  our (@CRLconfig);							#cached array of the latest CRL config (what is normally read from SQL)
  our ($beamline);							#defined in beamline.pl
  my ($script, $minRadius, $device, $msgref, $nrecords, $i);
  my (@InOut, $radius, $exclude, $takeout, $status, $msg);

  $script    = 'excludeCRLradius';
  $minRadius = shift(@_);
  $device    = shift(@_);
  $msgref    = shift(@_);

  if (defined $msgref) {$$msgref = '';}
  
### Read SQL if no cache or cached for different device:
  $nrecords = @CRLconfig;
  if ($nrecords == 0 || (defined $device && $CRLconfig[0]{'Device'} ne $device)) {
     @CRLconfig = ();
     $nrecords = &readCRLconfigSQLpopulated($device,\@CRLconfig);
  }
  if ($nrecords < 1) {return -1;}

### Set device if not specified in input:
  if (!defined $device) {
     $device = $CRLconfig[0]{'Device'};
     print STDOUT $script.': transfocator device is set to "'.$device.'"'."\n";
  }

  $status = &getTransfocatorInOut(\@InOut,$device,$msgref);
  if ($status) {return -1;}

  $exclude = 0;
  $takeout = 0;
  for ($i=0; $i<$nrecords; $i++) {
     $radius = $CRLconfig[$i]{'Radius'};   
     if ($radius <= $minRadius) {
        $exclude++;
        if ($InOut[$i] == 1) {
           $InOut[$i] = 0;
           $takeout++;
        }
     }
  }
  if ($takeout) {
     $msg = 'Taking out of the beam '.$takeout.' lenses with radius <= '.$minRadius.' um'."\n";
     print STDOUT $msg;
     if (defined $msgref) {$$msgref .= $msg;}
     $status = &setTransfocatorInOut(\@InOut,$device,$msgref);
     if ($status) {return -1;}
  }

  if ($exclude) {
     $msg = 'Excluding from the CRL config '.$exclude.' lenses with radius <= '.$minRadius.' um'."\n";
     print STDOUT $msg;
     if (defined $msgref) {$$msgref .= $msg;}
     for ($i=0; $i<$nrecords; $i++) {
        $radius = $CRLconfig[$i]{'Radius'};   
        if ($radius <= $minRadius) {
           splice @CRLconfig,$i,1;
           $i--;
           $nrecords--;
        }
     }
  }
  if ($nrecords < 1) {print STDOUT '!!! '.$script.': No data for CRL Device="'.$device.'": all CRL groups are disabled'."\n";}
# for ($i=0; $i<$nrecords; $i++) {
#    $device    = $CRLconfig[$i]{'Device'};				#CRL-box name
#    $name      = $CRLconfig[$i]{'Name'};				#Name of lenses subset driven by $PV_ctrl
#    $PV_ctrl   = $CRLconfig[$i]{'PV_ctrl'};				#Digital output for lenses In/Out
#    $PV_rb_in  = $CRLconfig[$i]{'PV_rb_in'};				#Digital input for lenses in the beam
#    $PV_rb_out = $CRLconfig[$i]{'PV_rb_out'};				#Digital input for lenses out of the beam
#    $offset    = $CRLconfig[$i]{'Offset'};				#Offset (mm) of lenses subset from the centre of the whole box (may not be used)
#    $radius    = $CRLconfig[$i]{'Radius'};				#Lenses radius (um). If $radius=0 no lenses are installed in the subset
#    $number    = $CRLconfig[$i]{'Number'};				#Number of lenses in the box. In $number=0, no lenses are installed in the subset
#    $material  = $CRLconfig[$i]{'Material'};				#Lenses material, normally Be. If $material is blank, no lenses are installed in the subset
# }
  return $nrecords;
}

#=====================================================================

# $status = &readDeltaDB($material,[\$message]);
sub readDeltaDB($) {
###
### If not yet cached, reads Material f' from respective file in
### $PEZCA/Data & stores in global array @DeltaDB
### 
  our (@DeltaDB);							#array of hashes containing f' for materials per energy
  my ($script, $material, $msgref, $msg, $status);
  my ($PEZCA, $file, @lines, $str, $m);
  my (@words, $nwords, @energy, @delta);
  
  $script   = 'readDeltaDB';
  $material = shift(@_);						#material if not Be (will look for $material.dat)
  $msgref   = shift(@_);

  if (defined $msgref) {$$msgref = '';}

  foreach $m (@DeltaDB) {						#if aready cached, nothing to do
     if ($m->{material} eq $material) {return 0;}
  }

  if (!defined $material) {
     $msg = sprintf '*** '.$script.': undefined material.';
     print STDOUT $msg,"\n";
     if (defined $msgref) {$$msgref = $msg;}
     return -1;
  }
  $file = $material.'.dat';
  if (!defined $ENV{PEZCA}) {
     $msg = '*** '.$script.': Undefined PEZCA environment.';
     print STDOUT $msg,"\n";
     if (defined $msgref) {$$msgref = $msg;}
     return -1;
  } else {
     $PEZCA = $ENV{PEZCA};
     $PEZCA =~ s/\\/\//g;
     if ($PEZCA !~ /\/$/) {$PEZCA .='/';}
  }
  $file = $PEZCA.'/Data/'.$file;
  if (! -e $file) {
     $msg = '*** '.$script.': Data file ['.$file.'] not found.';
     print STDOUT $msg,"\n";
     if (defined $msgref) {$$msgref = $msg;}
     return -1;
  }
# print STDOUT $script.': Opening material file: '.$file."\n";
  $status = &read_entire_file($file,\@lines);
  if ($status) {
     $msg = '*** '.$script.': Error reading data file ['.$file.'].';
     print STDOUT $msg,"\n";
     if (defined $msgref) {$$msgref = $msg;}
     return -1;
  }
  @energy = ();
  @delta  = ();
  foreach $str (@lines) {
     if ($str =~ /^\#/) {next;}
     $str =~ s/\s+$//;							#remove end-of-line spaces
     if (length($str) == 0) {next;}
     @words = split(/,\s+/,$str);
     $nwords = @words;
     if ($nwords != 2) {
        $msg = '*** '.$script.': incorrect record ['.$str.'] in data file ['.$file.'] -- nWords='.$nwords;
        print STDOUT $msg,"\n";
        if (defined $msgref) {$$msgref = $msg;}
        return -1;
     }
     $words[0] =~ s/^\s+//;
     if (! &is_float($words[0]) || ! &is_float($words[1])) {
        $msg = '*** '.$script.': incorrect record ['.$str.'] in data file ['.$file.'] -- non-float value.';
        print STDOUT $msg,"\n";
        if (defined $msgref) {$$msgref = $msg;}
        return -1;
     }
     push(@energy,$words[0]);
     push(@delta, $words[1]);
  }
  push @DeltaDB, {material=>$material, energy=>\@energy, delta=>\@delta};
  return 0;
}

#=====================================================================

# $delta = &calcDelta($E,$material,[\$msg]);
sub calcDelta($$;\$) {
###
### Calculates (interpolates) $delta for given x-ray energy and material (e.g. Be)
### 
  our (@DeltaDB);							#array of hashes containing f' for materials per energy: ${DeltaDB[$i]->{material}}, @{DeltaDB[$i]->{energy}}, @{DeltaDB[$i]->{delta}}
  my ($script, $E, $material, $found);
  my ($msgref, $msg, $i, $m, $status);
  my (@energy, @deltas, $nE, $delta);

  $script   = 'calcDelta';
  $E        = shift(@_);						#x-ray energy in keV
  $material = shift(@_);						#material if not Be (will look for $material.dat)
  $msgref   = shift(@_);

  if (defined $msgref) {$$msgref = '';}

  if ($E < 1 || $E> 60) {
     $msg = sprintf '*** '.$script.': E=%.3f is not in range. X-ray energy must be in the range [1-60] keV.',$E;
     print STDOUT $msg,"\n";
     if (defined $msgref) {$$msgref = $msg;}
     return 0;
  }

  $status = &readDeltaDB($material,$msgref);
  if ($status) {return 0;}

  @energy = ();
  @deltas = ();
  $found  = 0;
  foreach $m (@DeltaDB) {
     if ($m->{material} eq $material) {
        @energy = @{$m->{energy}};
        @deltas = @{$m->{delta}};
        $found  = 1;
        last;
     }
  }
  $nE = @energy;
  if (! $found || $nE == 0) {
     $msg = '*** '.$script.': no refraction coefficient "delta" data for material='.$material;
     print STDOUT $msg,"\n";
     if (defined $msgref) {$$msgref = $msg;}
     return 0;
  }
  if ($E < $energy[0] || $E > $energy[$nE-1]) {
     $msg = sprintf '*** '.$script.': E=%.3f is outside of delta data for material='.$material.', which is within [%.3f-%.3f] keV.',$E,$energy[0],$energy[$nE-1];
     print STDOUT $msg,"\n";
     if (defined $msgref) {$$msgref = $msg;}
     return 0;
  }
  for ($i=0; $i<$nE; $i++) {if ($energy[$i] > $E) {last;}}
  $delta = $deltas[$i-1] + ($deltas[$i]-$deltas[$i-1])*($E-$energy[$i-1])/($energy[$i]-$energy[$i-1]);
  return $delta;
}

#=====================================================================

# $focus = &calcCRLfocus($E,$R,$N,[$material,\$msg]);			#
sub calcCRLfocus($$$;$\$) {
###
### Returns focus in mm. $R-radius is um, $n=number of lenses. 
### Default material=Be (Be.dat), or can be $delta($E).
### This sub is reciprocal with &calcCRLpower: focus=1/power.
###
  my ($script, $E, $R, $N, $material, $msgref);
  my ($msg, $delta, $focus);

### For calculating focal distance with multiple lenses when they are spaced, see:
### https://underthemath.wordpress.com/2020/05/31/equivalent-lens-system/
### Also: https://physics.stackexchange.com/questions/247617/how-is-focal-length-defined-for-a-two-lens-system-separated-by-a-distance-d

  $script   = 'calcCRLfocus';
  $E        = shift(@_);						#x-ray energy in keV
  $R        = shift(@_);						#CRL lenses radius in um
  $N        = shift(@_);						#number of CRL lenses
  $material = shift(@_);						#material if not Be (will look for $material.dat); can also be $delta for given energy
  $msgref   = shift(@_);

# $power = &calcCRLpower($E,$R,$N,$material,$msgref);			#alternative way
# if ($power > 0) {return 1./$power;} else {return -1;}
 
  if (defined $msgref) {$$msgref = '';}
  if (!defined $material) {$material = 'Be';}

  if ($E < 1 || $E> 60) {
     $msg = sprintf '*** '.$script.': E=%.3f is not in range. X-ray energy must be in the range [1-60] keV.',$E;
     print STDOUT $msg,"\n";
     if (defined $msgref) {$$msgref = $msg;}
     return -1;
  }
  if ($R < 1 || $R> 1000) {
     $msg = sprintf '*** '.$script.': R=%.3f is not in range. Lens radius must be in the range [1-1000] um.',$R;
     print STDOUT $msg,"\n";
     if (defined $msgref) {$$msgref = $msg;}
     return -1;
  }
  if ($N < 1 || $N> 100) {
     $msg = sprintf '*** '.$script.': N=%d is not in range. Number of lenses must be in the range [1-100].',$N;
     print STDOUT $msg,"\n";
     if (defined $msgref) {$$msgref = $msg;}
     return -1;
  }

  if (&is_float($material)) {$delta = $material;}			#if $delta($E) is directly specified instead of material
  else                      {$delta = &calcDelta($E,$material,$msgref);}
  if ($delta == 0) {return -1;}
  
  $focus = (0.001*$R)/(2*$delta*$N);					#this is in mm
  return $focus;
}

#=====================================================================

# $power = &calcCRLpower($E,$R,$N,[$material,\$msg]);
sub calcCRLpower($$$;$\$) {
###
### Returns power=1/focus in 1/mm. $R-radius is um, $n=number of lenses. 
### Default material=Be (Be.dat), or can be $delta($E).
### This sub is reciprocal with &calcCRLfocus: focus=1/power.
###
  my ($script, $E, $R, $N, $material);
  my ($msgref, $msg, $delta, $P);

### For calculating focal distance with multiple lenses when they are spaced, see:
### https://underthemath.wordpress.com/2020/05/31/equivalent-lens-system/
### Also: https://physics.stackexchange.com/questions/247617/how-is-focal-length-defined-for-a-two-lens-system-separated-by-a-distance-d

  $script   = 'calcCRLpower';
  $E        = shift(@_);						#x-ray energy in keV
  $R        = shift(@_);						#CRL lenses radius in um
  $N        = shift(@_);						#number of CRL lenses
  $material = shift(@_);						#material if not Be (will look for $material.dat); can also be $delta for given energy
  $msgref   = shift(@_);
  
  if (defined $msgref) {$$msgref = '';}
  if (!defined $material) {$material = 'Be';}

  if ($E < 1 || $E> 60) {
     $msg = sprintf '*** '.$script.': E=%.3f is not in range. X-ray energy must be in the range [1-60] keV.',$E;
     print STDOUT $msg,"\n";
     if (defined $msgref) {$$msgref = $msg;}
     return -1;
  }
  if ($R < 1 || $R> 1000) {
     $msg = sprintf '*** '.$script.': R=%.3f is not in range. Lens radius must be in the range [1-1000] um.',$R;
     print STDOUT $msg,"\n";
     if (defined $msgref) {$$msgref = $msg;}
     return -1;
  }
  if ($N < 1 || $N> 100) {
     $msg = sprintf '*** '.$script.': N=%d is not in range. Number of lenses must be in the range [1-100].',$N;
     print STDOUT $msg,"\n";
     if (defined $msgref) {$$msgref = $msg;}
     return -1;
  }

  if (&is_float($material)) {$delta = $material;}			#if $delta($E) is directly specified instead of material
  else                      {$delta = &calcDelta($E,$material,$msgref);}
  if ($delta == 0) {return -1;}
  
  $P = (2*$delta*$N)/(0.001*$R);					#this is in 1/mm
  return $P;
}

#=====================================================================

# @oneByFocus = &calcAllOneByFocus($E,$device,[\$msg]);
sub calcAllOneByFocus(;$$\$) {
###
### Returns array of power=1/focus for all transfocator lens groups. 
### Defaults: E=current, device=CRL-2020
### 
  our (@CRLconfig);							#cached array of the latest CRL config (what is normally read from SQL)
  our ($beamline);							#defined in beamline.pl
  my ($script, $E, $device, $msgref, $nrecords, $i);
  my ($r, $n, $m, $m0, $delta, @power, $msg);
  my ($energyPV, $status, $statusDescr);

  $script = 'calcAllOneByFocus';
  $E      = shift(@_);							#if not specified, taken from EPICS
  $device = shift(@_);							#if not specified, taken from currently stored SQL cache, or if not available, then default of &readCRLconfigSQL
  $msgref = shift(@_);							#if not specified, messages are not returned

  if (defined $msgref) {$$msgref = '';}

  @power = ();
  
### Read SQL if no cache or cached for different device:
  $nrecords = @CRLconfig;
  if ($nrecords == 0 || (defined $device && $CRLconfig[0]{'Device'} ne $device)) {
     @CRLconfig = ();
     $nrecords = &readCRLconfigSQLpopulated($device,\@CRLconfig);
  }
  if ($nrecords < 1) {return -1;}

### Set device if not specified in input:
  if (!defined $device) {
     $device = $CRLconfig[0]{'Device'};
     print STDOUT $script.': transfocator device is set to "'.$device.'"'."\n";
  }
### Get current energy if not specified in input:
  if (!defined $E) {
     if (!defined $beamline) {die $script.': no beamline defined.'."\n";}
     $energyPV = $beamline.':MO:E:ActPos';
     ($status, $E) = Pezca::GetDouble($energyPV);
     if ($status) {
        $statusDescr = &ezcaStatusCode($status);
        $msg = '*** '.$script.': EZCA error='.$status.' ['.$statusDescr.'] on PV='.$energyPV;
        print STDOUT $msg,"\n";
        if (defined $msgref) {$$msgref = $msg;}
        return -1;
     }
  }

  for ($i=0; $i<$nrecords; $i++) {
     $r = $CRLconfig[$i]{'Radius'};					#Lenses radius (um). If $radius=0 no lenses are installed in the subset
     $n = $CRLconfig[$i]{'Number'};					#Number of lenses in the box. In $number=0, no lenses are installed in the subset
     $m = $CRLconfig[$i]{'Material'};					#Lenses material, normally Be. If $material is blank, no lenses are installed in the subset
     if (!defined $m0 || $m ne $m0) {
        $delta = &calcDelta($E,$m,$msgref);
        if ($delta == 0) {return -1;}
        $m0 = $m;
     }
     if ($r <= 0) {
        $msg = '*** '.$script.': lens group='.$CRLconfig[$i]{'Name'}.' R='.$r;
        print STDOUT $msg,"\n";
        if (defined $msgref) {$$msgref = $msg;}
        return -1;
     }
     $power[$i] = (2*$delta*$n)/(0.001*$r)
  }
  return @power;
}

#=====================================================================

# $status = &getTransfocatorInOut(\@InOut,[$device,\$msg]);
sub getTransfocatorInOut(\@$\$) {
###
### Returns current In/Out of transfocator lens groups
### 
  our (@CRLconfig);							#cached array of the latest CRL config (what is normally read from SQL)
  our ($beamline);							#defined in beamline.pl
  our ($transfocatorCheckInOut);
  my ($script, $device, $msgref, $IOref, $status);
  my (@IO, $ON, $OFF, $stateCTRL, $stateIN, $stateOUT);
  my ($pv, $msg, $nrecords, $i, @offon);
  my ($wait_max, $wait_step, $wait);

  $script = 'getTransfocatorInOut';
  $IOref  = shift(@_);							#returns current In/Out
  $device = shift(@_);							#if not specified, taken from currently stored SQL cache, or if not available, then default of &readCRLconfigSQL
  $msgref = shift(@_);							#if specified, messages are returned

  if (defined $msgref) {$$msgref = '';}
  if (!defined $beamline) {die $script.': no beamline defined.'."\n";}
  if (!defined $transfocatorCheckInOut) {$transfocatorCheckInOut = 1;}
  
  @offon = qw(Off On);
  $wait_max  = 5;
  $wait_step = 0.05;

### Read SQL if no cache or cached for different device:
  $nrecords = @CRLconfig;
  if ($nrecords == 0 || (defined $device && $CRLconfig[0]{'Device'} ne $device)) {
     @CRLconfig = ();
     $nrecords = &readCRLconfigSQLpopulated($device,\@CRLconfig);
  }
  if ($nrecords < 1) {return -1;}

### Set device if not specified in input:
  if (!defined $device) {
     $device = $CRLconfig[0]{'Device'};
     print STDOUT $script.': transfocator device is set to "'.$device.'"'."\n";
  }

  $ON  = 1;								#what digital-In reading corresponds to actuator triggered?
  $OFF = 1 - $ON;
  @IO = ();
  for ($i=0; $i<$nrecords; $i++) {
     $pv = $CRLconfig[$i]{'PV_ctrl'}; if ($pv !~ /^$beamline/) {$pv = $beamline.':'.$pv;}
     $status = &addCaMonitor($pv); if ($status) {return -1;}
     ($status,$stateCTRL) = Pezca::GetDouble($pv);
     if ($status) {&report_only($script,$status,$pv); $stateCTRL=$OFF;}
#
### Check for consistency:
     if ($transfocatorCheckInOut) {  #----------------------------------#begin consistency check
        $pv = $CRLconfig[$i]{'PV_rb_in'}; if ($pv !~ /^$beamline/) {$pv = $beamline.':'.$pv;}
        $status = &addCaMonitor($pv); if ($status) {return -1;}
        ($status,$stateIN) = Pezca::GetDouble($pv);
        if ($status) {&report_only($script,$status,$pv); $stateIN=$OFF;}
#
        $pv = $CRLconfig[$i]{'PV_rb_out'}; if ($pv !~ /^$beamline/) {$pv = $beamline.':'.$pv;}
        $status = &addCaMonitor($pv); if ($status) {return -1;}
        ($status,$stateOUT) = Pezca::GetDouble($pv);
        if ($status) {&report_only($script,$status,$pv); $stateOUT=$ON;}
#
### Extra check for consistency:
        if ($stateIN != $stateCTRL || $stateOUT != 1-$stateCTRL) { 
           $wait = 0;
           do {
              select(undef,undef,undef,$wait_step);			# sleep(..);
              $wait += $wait_step; 

              $pv = $CRLconfig[$i]{'PV_ctrl'}; if ($pv !~ /^$beamline/) {$pv = $beamline.':'.$pv;}
              ($status,$stateCTRL) = Pezca::GetDouble($pv);
              if ($status) {&report_only($script,$status,$pv); $stateCTRL=$OFF;}

              $pv = $CRLconfig[$i]{'PV_rb_in'}; if ($pv !~ /^$beamline/) {$pv = $beamline.':'.$pv;}
              ($status,$stateIN) = Pezca::GetDouble($pv);
              if ($status) {&report_only($script,$status,$pv); $stateIN=$OFF;}

              $pv = $CRLconfig[$i]{'PV_rb_out'}; if ($pv !~ /^$beamline/) {$pv = $beamline.':'.$pv;}
              ($status,$stateOUT) = Pezca::GetDouble($pv);
              if ($status) {&report_only($script,$status,$pv); $stateOUT=$ON;}
           } while (($stateIN != $stateCTRL || $stateOUT != 1-$stateCTRL) && $wait < $wait_max);
        }
#
        if ($stateIN != $stateCTRL || $stateOUT != 1-$stateCTRL) { 
           $msg = '*** '.$script.': lens group='.$CRLconfig[$i]{'Name'}.' is in inconsistent state after '.$wait_max.'s:'."\n"
                        .         ' CTRL('.$CRLconfig[$i]{'PV_ctrl'}  .')='.$stateCTRL.' ('.$offon[$stateCTRL].')'."\n"
                        .  ' readback-IN('.$CRLconfig[$i]{'PV_rb_in'} .')='.$stateIN."\n"
                        . ' readback-OUT('.$CRLconfig[$i]{'PV_rb_out'}.')='.$stateOUT."\n";
           print STDOUT $msg;
           if (defined $msgref) {$$msgref .= $msg;}
           return -1;
        }
     } else { #---------------------------------------------------------#end condistency check
        $stateIN  = $stateCTRL;
        $stateOUT = 1-$stateCTRL;
     }
     if ($stateIN == $ON) {$IO[$i] = 1;}				#lenses block is IN 
     else                 {$IO[$i] = 0;}				#lenses block is OUT
  }
  @$IOref = @IO;
  return 0;
}


#=====================================================================

# $status = &setTransfocatorInOut(\@InOut,[$device,\$msg]);
sub setTransfocatorInOut(\@$\$) {
###
### Sets requested In/Out for all Transfocator lens groups & returns status
### 
  our (@CRLconfig);							#cached array of the latest CRL config (what is normally read from SQL)
  our ($beamline);							#defined in beamline.pl
  our ($transfocatorCheckInOut);
  my ($script, $device, $msgref, $IOref, $status);
  my ($ON, $OFF, $stateCTRL, $stateIN, $stateOUT);
  my ($pv, $msg, $nrecords, $i, @offon, @IO);
  my ($wait_max, $wait_step, $wait, $statusChk);

  $script = 'setTransfocatorInOut';
  $IOref  = shift(@_);							#returns current In/Out
  $device = shift(@_);							#if not specified, taken from currently stored SQL cache, or if not available, then default of &readCRLconfigSQL
  $msgref = shift(@_);							#if specified, messages are returned

  if (defined $msgref) {$$msgref = '';}
  if (!defined $beamline) {die $script.': no beamline defined.'."\n";}
  if (!defined $transfocatorCheckInOut) {$transfocatorCheckInOut = 1;}

  @offon = qw(Off On);
  $wait_step = 0.01;
  $wait_max  = 7;
  
### Read SQL if no cache or cached for different device:
  $nrecords = @CRLconfig;
  if ($nrecords == 0 || (defined $device && $CRLconfig[0]{'Device'} ne $device)) {
     @CRLconfig = ();
     $nrecords = &readCRLconfigSQLpopulated($device,\@CRLconfig);
  }
  if ($nrecords < 1) {return -1;}

### Set device if not specified in input:
  if (!defined $device) {
     $device = $CRLconfig[0]{'Device'};
     print STDOUT $script.': transfocator device is set to "'.$device.'"'."\n";
  }

  $ON  = 1;								#what digital-In reading corresponds to actuator triggered?
  $OFF = 1 - $ON;

  $statusChk = 0;
### Set lenses IN and OUT:
  for ($i=0; $i<$nrecords; $i++) {
     $pv = $CRLconfig[$i]{'PV_ctrl'}; if ($pv !~ /^$beamline/) {$pv = $beamline.':'.$pv;}
     if (@$IOref[$i]) {$stateCTRL=$ON;} 
     else             {$stateCTRL=$OFF;}
     $status = Pezca::PutDoubleOldCa($pv,$stateCTRL);
     if ($status) {&report_only($script,$status,$pv); $statusChk++;}
#
     $pv = $CRLconfig[$i]{'PV_rb_in'}; if ($pv !~ /^$beamline/) {$pv = $beamline.':'.$pv;}
     $status = &addCaMonitor($pv); if ($status) {return -1;}
#
     $pv = $CRLconfig[$i]{'PV_rb_out'}; if ($pv !~ /^$beamline/) {$pv = $beamline.':'.$pv;}
     $status = &addCaMonitor($pv); if ($status) {return -1;}
#
     select(undef,undef,undef,0.1);					#sleep(..);
  }

### Verify that all is set:
  if ($transfocatorCheckInOut) { #--------------------------------------#begin consistency check
     for ($i=0; $i<$nrecords; $i++) {
        if (@$IOref[$i]) {$stateCTRL=$ON;} 
        else             {$stateCTRL=$OFF;}
#
        $pv = $CRLconfig[$i]{'PV_rb_in'}; if ($pv !~ /^$beamline/) {$pv = $beamline.':'.$pv;}
        ($status,$stateIN) = Pezca::GetDouble($pv);
        if ($status) {&report_only($script,$status,$pv); $stateIN=$OFF;}
#
        $pv = $CRLconfig[$i]{'PV_rb_out'}; if ($pv !~ /^$beamline/) {$pv = $beamline.':'.$pv;}
        ($status,$stateOUT) = Pezca::GetDouble($pv);
        if ($status) {&report_only($script,$status,$pv); $stateOUT=$ON;}
#
        if ($stateIN != $stateCTRL || $stateOUT != 1-$stateCTRL) {
### Retry:
           $wait = 0;
           do {
              select(undef,undef,undef,$wait_step);			#sleep(..);
              $wait += $wait_step;
#
              $pv = $CRLconfig[$i]{'PV_ctrl'}; if ($pv !~ /^$beamline/) {$pv = $beamline.':'.$pv;}
              $status = Pezca::PutDoubleOldCa($pv,$stateCTRL);
              &report_only($script,$status,$pv);
#
              $pv = $CRLconfig[$i]{'PV_rb_in'}; if ($pv !~ /^$beamline/) {$pv = $beamline.':'.$pv;}
              ($status,$stateIN) = Pezca::GetDouble($pv);
              if ($status) {&report_only($script,$status,$pv); $stateIN=$OFF;}
#
              $pv = $CRLconfig[$i]{'PV_rb_out'}; if ($pv !~ /^$beamline/) {$pv = $beamline.':'.$pv;}
              ($status,$stateOUT) = Pezca::GetDouble($pv);
              if ($status) {&report_only($script,$status,$pv); $stateOUT=$ON;}
#
           } while (($stateIN != $stateCTRL || $stateOUT != 1-$stateCTRL) && $wait < $wait_max);
        }
###  Extra check for consistency:
        if ($stateIN != $stateCTRL || $stateOUT != 1-$stateCTRL) { 
           $msg = '*** '.$script.': lens group='.$CRLconfig[$i]{'Name'}.' is in inconsistent state after '.$wait_max.'s:'."\n"
                        .         ' CTRL('.$CRLconfig[$i]{'PV_ctrl'}  .')='.$stateCTRL.' ('.$offon[$stateCTRL].')'."\n"
                        .  ' readback-IN('.$CRLconfig[$i]{'PV_rb_in'} .')='.$stateIN."\n"
                        . ' readback-OUT('.$CRLconfig[$i]{'PV_rb_out'}.')='.$stateOUT."\n"
                        .  ' - retrying (no second check)...'."\n";
           print STDOUT $msg;
           if (defined $msgref) {$$msgref .= $msg;}
           $pv = $CRLconfig[$i]{'PV_ctrl'}; if ($pv !~ /^$beamline/) {$pv = $beamline.':'.$pv;}
           $status = Pezca::PutDoubleOldCa($pv,$stateCTRL);
           &report_only($script,$status,$pv);
           select(undef,undef,undef,0.1);					#sleep(..);
           $statusChk++;
#          return -1;
        }
     }
  }  #------------------------------------------------------------------#end consistency check
  if ($statusChk) {							#if there were inconsistency errors
     $msg = '!!! '.$script.': '.$statusChk.' inconsistent states detected, rechecking...'."\n";
     print STDOUT $msg;
     if (defined $msgref) {$$msgref .= $msg;}
     @IO  = ();
     $msg = '';
     $status = &getTransfocatorInOut(\@IO,$device,\$msg);               #we do not want to pass $IOref as it will be overwritten
     if ($status) {
        if (defined $msgref) {$$msgref .= $msg;}
        $msg = '*** '.$script.': rechecking did not help!'."\n";
     } else {
        $msg = $script.': all lens groups are consistent after rechecking!'."\n";
     }
     print STDOUT $msg;
     if (defined $msgref) {$$msgref .= $msg;}
  } else {								#if no inconsitency errors
     $status = 0;
  }
  return $status;
}


#=====================================================================

# $status = &setTransfocatorRB2BO([$device,\$msg]);
sub setTransfocatorRB2BO($\$) {
###
### Sets all BO PVs from the values of RB PVs & returns status
### 
  our (@CRLconfig);							#cached array of the latest CRL config (what is normally read from SQL)
  our ($beamline);							#defined in beamline.pl
  my ($script, $device, $msgref, $status);
  my ($pv, $msg, $nrecords, $i, @IO, $state, $ncorrected);

  $script = 'setTransfocatorRB2BO';
  $device = shift(@_);							#if not specified, taken from currently stored SQL cache, or if not available, then default of &readCRLconfigSQL
  $msgref = shift(@_);							#if specified, messages are returned

  if (defined $msgref) {$$msgref = '';}
  if (!defined $beamline) {die $script.': no beamline defined.'."\n";}

  $ncorrected = 0;

### Read SQL if no cache or cached for different device:
  $nrecords = @CRLconfig;
  if ($nrecords == 0 || (defined $device && $CRLconfig[0]{'Device'} ne $device)) {
     @CRLconfig = ();
     $nrecords = &readCRLconfigSQLpopulated($device,\@CRLconfig);
  }
  if ($nrecords < 1) {return -1,$ncorrected;}

### Set device if not specified in input:
  if (!defined $device) {
     $device = $CRLconfig[0]{'Device'};
     print STDOUT $script.': transfocator device is set to "'.$device.'"'."\n";
  }

### Read current readbacks:
  @IO = ();
  for ($i=0; $i<$nrecords; $i++) {
     $pv = $CRLconfig[$i]{'PV_ctrl'}; if ($pv !~ /^$beamline/) {$pv = $beamline.':'.$pv;}
     if ($pv !~ /:bo\d{1,2}$/) {
        $msg = '*** '.$script.': lens group='.$CRLconfig[$i]{'Name'}.' unexpected control PV='.$pv."\n";
        print STDOUT $msg;
        if (defined $msgref) {$$msgref .= $msg;}
        return -1,$ncorrected;
     }
     $pv =~ s/:bo/:rb/; 
     ($status,$IO[$i]) = Pezca::GetDouble($pv);
     if ($status) {
        &report_only($script,$status,$pv); 
        $msg = '*** '.$script.': lens group='.$CRLconfig[$i]{'Name'}.' error reading PV='.$pv."\n";
        print STDOUT $msg;
        if (defined $msgref) {$$msgref .= $msg;}
        return -1,$ncorrected;
     }
  }

### Copy readbacks to controls if do not coincide:
  for ($i=0; $i<$nrecords; $i++) {
     $pv = $CRLconfig[$i]{'PV_ctrl'}; if ($pv !~ /^$beamline/) {$pv = $beamline.':'.$pv;}
     ($status,$state) = Pezca::GetDouble($pv);
     if ($status) {
        &report_only($script,$status,$pv); 
        $msg = '*** '.$script.': lens group='.$CRLconfig[$i]{'Name'}.' error reading PV='.$pv."\n";
        print STDOUT $msg;
        if (defined $msgref) {$$msgref .= $msg;}
        return -1,$ncorrected;
     }
     if ($state ne $IO[$i]) {
        $ncorrected++;
        $status = Pezca::PutDoubleOldCa($pv,$IO[$i]);
        if ($status) {
           &report_only($script,$status,$pv);
           $msg = '*** '.$script.': lens group='.$CRLconfig[$i]{'Name'}.' error writing PV='.$pv."\n";
           print STDOUT $msg;
           if (defined $msgref) {$$msgref .= $msg;}
           return -1,$ncorrected;
        }
        select(undef,undef,undef,0.15);					#sleep(..);
     }
  }
  return 0,$ncorrected;
}

#=====================================================================

#$n_in = &InOut_str2arr($str,\@InOut);
sub InOut_str2arr($\@) {
###
### Converts a string of In/Out states (e.g. '11010001111000') into respective array
###
  my ($script, $str, $IOref);
  my ($nlens, $i, $n_in);
  
  $script = 'InOut_str2arr';
  $str    = shift(@_);
  $IOref  = shift(@_);

  if (!defined $str) {
     print STDOUT '*** '.$script.': undefined str'."\n"; 
     &call_trace();
     return 0;
  }
  $nlens  = length($str);
  @$IOref = ();
  $n_in   = 0;
  for ($i=0; $i<$nlens; $i++) {
     if (substr($str,$i,1) eq '1') {@$IOref[$i]=1; $n_in++;}
     else                          {@$IOref[$i]=0;}
  }
  return $n_in;  
}  

#=====================================================================

#($str,$n_in) = &InOut_arr2str(\@InOut);
sub InOut_arr2str(\@) {
###
### Converts an array of In/Out states (In=1,Out=0) into respective string (e.g. '11010001111000')
###
  my ($script, $str, $IOref);
  my ($nlens, $n_in, $i);
  
  $script = 'InOut_arr2str';
  $IOref  = shift(@_);

  $nlens  = @$IOref;
  $str    = '';
  $n_in   = 0;
  for ($i=0; $i<$nlens; $i++) {
     $str .= @$IOref[$i];
     if (@$IOref[$i]) {$n_in++;}
  }
  return ($str,$n_in);  
}  

#=====================================================================

# $focus = &getTransfocatorFocus4($E,$device,[\$msg,\@InOut]);		#
sub getTransfocatorFocus4(;$$\$\@) {
###
### Returns overall Transfocator focus (Bruls method) in mm for current 
### lenses in/out configuration. First it calls &getTransfocatorInOut
### and then &calcTransfocatorFocus4 to calculate the focus
###
  our (@CRLconfig);							#cached array of the latest CRL config (what is normally read from SQL)
  our ($beamline);							#defined in beamline.pl
  my ($script, $E, $device, $msgref, $IOref);
  my (@IO, $focus, $msg, $nrecords);
  my ($energyPV, $status, $statusDescr);

  $script         = 'getTransfocatorFocus4';
  $E              = shift(@_);						#if not specified, taken from EPICS
  $device         = shift(@_);						#if not specified, taken from currently stored SQL cache, or if not available, then default of &readCRLconfigSQL
  $msgref         = shift(@_);						#if specified, messages are returned
  $IOref          = shift(@_);						#if specified, returns current In/Out

  if (defined $msgref)          {$$msgref = '';}
  if (!defined $beamline)       {die $script.': no beamline defined.'."\n";}
  
### Read SQL if no cache or cached for different device:
  $nrecords = @CRLconfig;
  if ($nrecords == 0 || (defined $device && $CRLconfig[0]{'Device'} ne $device)) {
     @CRLconfig = ();
     $nrecords = &readCRLconfigSQLpopulated($device,\@CRLconfig);
  }
  if ($nrecords < 1) {return -1;}

### Set device if not specified in input:
  if (!defined $device) {
     $device = $CRLconfig[0]{'Device'};
     print STDOUT $script.': transfocator device is set to "'.$device.'"'."\n";
  }
### Get current energy if not specified in input:
  if (!defined $E) {
     $energyPV = $beamline.':MO:E:ActPos';
     ($status, $E) = Pezca::GetDouble($energyPV);
     if ($status) {
        $statusDescr = &ezcaStatusCode($status);
        $msg = '*** '.$script.': EZCA error='.$status.' ['.$statusDescr.'] on PV='.$energyPV;
        print STDOUT $msg,"\n";
        if (defined $msgref) {$$msgref = $msg;}
        return -1;
     }
  }

  @IO = ();
  $status = &getTransfocatorInOut(\@IO,$device,$msgref);
  if ($status) {return -1;}
  if (defined $IOref) {@$IOref = @IO;}

  $focus = &calcTransfocatorFocus4(\@IO,$E,$device,$msgref);
  return $focus;
}

#=====================================================================

#$focus = &calcTransfocatorFocus4(\@InOut,[$E,$device,\$msg]);
sub calcTransfocatorFocus4(\@;$$\$) {
###
### For given @InOut and $E, calculates exact Transfocator focus by the Bruls method:
### G.J.C.L. Bruls "Exact formulas for a thin-lens system with an arbitrary
### number of lenses", Optik, Volume 126, Issue 6, March 2015, Pages 659-662.
### (https://dx.doi.org/10.1016/j.ijleo.2015.01.035)
###
  our (@CRLconfig);							#cached array of the latest CRL config (what is normally read from SQL)
  our ($beamline);							#defined in beamline.pl
  my ($script, $IOref, $E, $device, $msgref);
  my ($nblocks, $r, $n, $m, $d, $f, $nIO);
  my ($energyPV, $status, $statusDescr);
  my ($m0, $delta, $msg, $i, $j, $nBlocksIn);
  my (@P, @D, @Ptotal, $sum, $Pr2Ln, $fadd);

  $script         = 'calcTransfocatorFocus4';
  $IOref          = shift(@_);						#reference to array of IN=1 & OUT=0 lense blocks
  $E              = shift(@_);						#if not specified, taken from EPICS
  $device         = shift(@_);						#if not specified, taken from currently stored SQL cache, or if not available, then default of &readCRLconfigSQL
  $msgref         = shift(@_);						#if not specified, messages are not returned

  if (defined $msgref)          {$$msgref = '';}
  if (!defined $beamline)       {die $script.': no beamline defined.'."\n";}
  
### Read SQL if no cache or cached for different device:
  $nblocks = @CRLconfig;						#has the config has not been read?
  if ($nblocks == 0 || (defined $device && $CRLconfig[0]{'Device'} ne $device)) {
     @CRLconfig = ();
     $nblocks = &readCRLconfigSQLpopulated($device,\@CRLconfig);
  }
  if ($nblocks < 1) {return -1;}

### Set device if not specified in input:
  if (!defined $device) {
     $device = $CRLconfig[0]{'Device'};
     print STDOUT $script.': transfocator device is set to "'.$device.'"'."\n";
  }
  $nIO = @$IOref;
  if ($nIO != $nblocks) {
     $msg = '*** '.$script.': In/Out array length='.$nIO.' does not match used lenses blocks number='.$nblocks.' in device='.$device;
     print STDOUT $msg,"\n";
     if (defined $msgref) {$$msgref = $msg;}
     return -1;
  }
### Get current energy if not specified in input:
  if (!defined $E) {
     $energyPV = $beamline.':MO:E:ActPos';
     ($status, $E) = Pezca::GetDouble($energyPV);
     if ($status) {
        $statusDescr = &ezcaStatusCode($status);
        $msg = '*** '.$script.': EZCA error='.$status.' ['.$statusDescr.'] on PV='.$energyPV;
        print STDOUT $msg,"\n";
        if (defined $msgref) {$$msgref = $msg;}
        return -1;
     }
  }
### Create an array of powers (p=1/f) and distances for those lenses that are In:
  @D = ();
  @P = ();
  @Ptotal = ();
  $j = 0;
  for ($i=0; $i<$nblocks; $i++) {
     if ($$IOref[$i]) {							#lenses block is IN
        $r = $CRLconfig[$i]{'Radius'};					#Lenses radius (um). If $radius=0 no lenses are installed in the subset
        $n = $CRLconfig[$i]{'Number'};					#Number of lenses in the box. In $number=0, no lenses are installed in the subset
        $m = $CRLconfig[$i]{'Material'};				#Lenses material, normally Be. If $material is blank, no lenses are installed in the subset
        $d = $CRLconfig[$i]{'Offset'};					#the distance of lenses block from the transfocator center
        if (!defined $m0 || $m ne $m0) {
           $delta = &calcDelta($E,$m,$msgref);
           if ($delta == 0) {return -1;}
           $m0 = $m;
        }
        $f = &calcCRLfocus($E,$r,$n,$delta,$msgref);
        if ($f > 0) {
           $P[$j] = 1./$f;
           $D[$j] = $d;
           $j++;
        }
        elsif ($f == 0) {
           $msg = '*** '.$script.': zero focus for lenses group='.$CRLconfig[$i]{'Name'};
           print STDOUT $msg,"\n";
           if (defined $msgref) {$$msgref = $msg;}
           return -1;
        }
        else {
           $msg =  '!!! '.$script.': calcCRLfocus error='.$f.' for lenses group='.$CRLconfig[$i]{'Name'}; 
           print STDOUT $msg,"\n";
           if (defined $msgref) {$$msgref = $msg;}
           return -1;
        }
     }
  }
  $nBlocksIn = $j;
  if    ($nBlocksIn == 0) {
     $f     = 1.E10;
     $Pr2Ln = 0;
     $fadd  = 0;
  }
  elsif ($nBlocksIn == 1) {
     $f     = 1./$P[0];
     $Pr2Ln = 0;
     $fadd  = $D[0] + $Pr2Ln;
  }
  else {
     $Ptotal[0] = $P[0];
     for ($i=1; $i<$nBlocksIn; $i++) {
        $sum = 0;
        for ($j=0; $j<$i; $j++) {
           $d = abs($D[$j+1]-$D[$j]);
           $sum += $Ptotal[$j]*$d;
        }
        $Ptotal[$i] = $Ptotal[$i-1] + $P[$i]*(1-$sum);
     }
     if ($Ptotal[$nBlocksIn-1] <= 0) {return -1;}
     $f = 1./$Ptotal[$nBlocksIn-1];
### Now once we have the focus calculated, it is measured from the rear
### principal optical plane. The distance of that plane from the last
### lens is given by Brul's equation (17a):
###                  n-1
### P_r - L_n = -f * sum (P_i * D_(i,i+1)
###                  i=1
     $sum = 0;
     for ($j=0; $j<$nBlocksIn-1; $j++) {
        $d = abs($D[$j+1]-$D[$j]);
        $sum += $Ptotal[$j]*$d;
     }
     $Pr2Ln = -$f * $sum;
     $fadd  = $D[$nBlocksIn-1] + $Pr2Ln;
  }
  $f += $fadd;
  return $f;
}

#=====================================================================

# ($focusAct,@InOut) = &matchTransfocatorFocus4($focusRqs,[$E,$device,\$msg,$minimizeAttenuation,$focusRange,$TF_axis_PV])
# ($focusAct,@InOut) = &matchTransfocatorFocus4($focusRqs,
#                                               [$E,			#Default: $E=current
#                                                $device,		#Default: CRL-2020
#                                                \$msg,			#Default: none (this is output)
#                                                $minimizeAttenuation,	#Default: 0=minimize |Act-Rqs|
#                                                $focusRange,	 	#Default: $focusRange1=1/4 of TF travel range
#                                                $TF_axis_PV]);		#Default: 'CRL:X'
sub matchTransfocatorFocus4($;$$\$$$$) {
###
### Finds closest match to requested transfocator focus in mm, but does not set.
### Users Bruls exact formula for multi-lens focus & checks all possible lens combinations.
### Defaults: E=current, device=CRL-2020.
###
  our (@CRLconfig);							#cached array of the latest CRL config (what is normally read from SQL)
  our ($beamline);							#defined in beamline.pl
  my ($script, $focusRqs, $E, $device, $msgref);
  my ($minimizeFlag, $range, $axis_PV);
  my (@InOut, @Power, $nblocks, $nBlocksIn, $binstr);
  my ($energyPV, $status, $statusDescr, $msg, $nLensIn);
  my ($drvl, $drvh, $i, $j, $k, $n, $debug, $dump);
  my (@Ptotal, @P, @D, $sum, $d, $Pr2Ln);
  my (@CRL_Bruls, $ncomb);							#array of sum{1/f} for all in/out combinations
  my ($Bruls_inout, $Bruls_in, $Bruls_focus, $Bruls_fadd);

  $script = 'matchTransfocatorFocus4';
# $debug  = 10;								#set to the number of top choices to print
  $debug  = 0;								#set to the number of top choices to print
  $dump   = 0;								#$dump=1 requests printing all (A LOT!) combinations

  $focusRqs     = shift(@_);						#reference to array of IN=1 & OUT=0 lense blocks
  $E            = shift(@_);						#if not specified, taken from EPICS
  $device       = shift(@_);						#if not specified, taken from currently stored SQL cache, or if not available, then default of &readCRLconfigSQL
  $msgref       = shift(@_);						#if not specified, messages are not returned
  $minimizeFlag = shift(@_);						#if not specified, taken 0=minimize |Act-Rqs|; can be 0 or 1=minimize attenuation
  $range        = shift(@_);						#if not specified, taken as 1/4 of TF travel range
  $axis_PV      = shift(@_);						#if not specified, taken as 'CRL:X' (Transfocator upstream/downstream position axis)

  if (defined $msgref)        {$$msgref = '';}
  if (!defined $beamline)     {die $script.': no beamline defined.'."\n";}
  $beamline =~ s/:$//;
  if (!defined $axis_PV)      {$axis_PV = $beamline.':CRL:X';}
  if ($axis_PV !~ /^$beamline/) {
     if ($axis_PV =~ /^:/)    {$axis_PV = $beamline.$axis_PV;}
     else                     {$axis_PV = $beamline.$axis_PV;}
  }
  if (!defined $minimizeFlag) {$minimizeFlag = 0;}
  if ($minimizeFlag ne '1')   {$minimizeFlag = 0;}
  if (!defined $debug)        {$debug = 0;}
  
### Read SQL if no cache or if cached for different device (device name is needed!):
  $nblocks = @CRLconfig;
  if ($nblocks == 0 || (defined $device && $CRLconfig[0]{'Device'} ne $device)) {
     @CRLconfig = ();
     $nblocks = &readCRLconfigSQLpopulated($device,\@CRLconfig);
     if ($nblocks < 1) {return -1;}
  }

### Set device if not specified in input, but we have @CRLconfig cached:
  if (!defined $device) {
     $device = $CRLconfig[0]{'Device'};
     if ($debug) {print STDOUT $script.': transfocator device is set to "'.$device.'"'."\n";}
  }

### Get current energy if not specified in input:
  if (!defined $E) {
     $energyPV = $beamline.':MO:E:ActPos';
     ($status, $E) = Pezca::GetDouble($energyPV);
     if ($status) {
        $statusDescr = &ezcaStatusCode($status);
        $msg = '*** '.$script.': EZCA error='.$status.' ['.$statusDescr.'] on PV='.$energyPV;
        print STDOUT $msg,"\n";
        if (defined $msgref) {$$msgref = $msg;}
        return -1;
     }
  }

  if ($minimizeFlag) {
     if (!defined $range) {						#TF upstream/downstream motion range
        if (!defined &function_exists('getAxisDRVLH')) {
           print STDOUT '*** '.$script.': function "getAxisDRVLH" is undefined.'."\n"
                              .'Please source "motor_subs.pl"'."\n";
           return -1;
        }
        ($drvl,$drvh) = &getAxisDRVLH ($axis_PV,\$status);
        if ($status) {return -1;}
        if (!defined $drvl || !defined $drvh || abs($drvh-$drvl)<0.01) {return -1;}
        $range = 0.25*abs($drvh-$drvl);
     } else {
        $range = abs($range);
     }
  } else {
     $range = 0;
  }
  if ($debug) {
     print  STDOUT $script.': focus search mode = ';
     if ($minimizeFlag) {
        print  STDOUT 'minimum attenuation'."\n";
        printf STDOUT $script.': focus range of deviations for DB search are [%.2f : %.2f]'."\n",-$range,$range;
     } else {
        print  STDOUT 'best match'."\n";
     }
  }
  if ($minimizeFlag && $range == 0) {
     print STDOUT '*** '.$script.': focus range of deviations cannot be zero in the attenuation mode.'."\n";
     return -1;
  }     

  @Power = &calcAllOneByFocus($E,$device,$msgref);
  $status = 0;
  for ($i=0; $i<$nblocks; $i++) {
     if ($Power[$i] <= 0) {
        $msg = '*** '.$script.': 1/f=0 for lense block='.$i."\n";
        print STDOUT $msg;
        if (defined $msgref) {$$msgref .= $msg;}
        $status++;
     }
  }
  if ($status) {return -1;}

### Fill @CRL_Bruls (array of sum{1/f} for all in/out combinations):
  @CRL_Bruls = ();
  $ncomb = (2**$nblocks) - 1;
  if ($debug) {print STDOUT $script.': nblocks='.$nblocks.', ncomb='.$ncomb."\n";}
  $n = 0;
  for ($k=0; $k<$ncomb; $k++) { #---------------------------------------begin loop over all possible combinations
     $binstr = sprintf '%0'.$nblocks.'b', $k+1;				#this string should have length $nblocks
#    if (!defined $binstr || length($binstr) == 0) {
#       print STDOUT '*** '.$script.': undefined $binstr at nblocks='.$nblocks.', ncomb='.$ncomb.', k+1=,'.($k+1)."\n";
#    }
     $nBlocksIn = &InOut_str2arr($binstr,\@InOut);
     $nLensIn   = 0;
     @D = ();
     @P = ();
     @Ptotal = ();
     $j = 0;
     for ($i=0; $i<$nblocks; $i++) { #----------------------------------begin of loop over all blocks for given combination
        if ($InOut[$i]) {						#lenses block is IN
           $P[$j] = $Power[$i];
           $D[$j] = $CRLconfig[$i]{'Offset'};
           $nLensIn += $CRLconfig[$i]{'Number'};
           $j++;
        }
     } #----------------------------------------------------------------end of loop over all blocks for given combination
     $nBlocksIn = $j;
     $Bruls_inout = $binstr;
     ### For optimizing attenuation we can count either number
     ### of lens blocks in or number of lenses in:
#    $Bruls_in = $nBlocksIn;
     $Bruls_in = $nLensIn;
     if    ($nBlocksIn == 0) {
        $Bruls_focus = 1.E10;
        $Bruls_fadd  = 0.;
     }
     elsif ($nBlocksIn == 1) {
        $Bruls_focus = 1./$P[0];
        $Bruls_fadd  = $D[0];
     } else {
        $Ptotal[0] = $P[0];
        for ($j=1; $j<$nBlocksIn; $j++) {
           $sum = 0;
           for ($i=0; $i<$j; $i++) {
              $d = abs($D[$i+1]-$D[$i]);
              $sum += $Ptotal[$i]*$d;
           }
           $Ptotal[$j] = $Ptotal[$j-1] + $P[$j]*(1-$sum);
        }
        ### In principle, $Ptotal can be negative (it is a focus with respect to back principal plane):
        if ($debug && $Ptotal[$nBlocksIn-1] <= 0) {printf STDOUT 'k='.$k.'/'.$ncomb.' nBlocksIn=%-2d Ptotal=%.4f'."\n", $nBlocksIn, $Ptotal[$nBlocksIn-1];}
#       if ($Ptotal[$nBlocksIn-1] <= 0) {die '*** '.$script.': k='.$k.' inout=('.$binstr.') Ptotal='.$Ptotal[$nBlocksIn-1]."\n";}
        if ($Ptotal[$nBlocksIn-1] == 0) {$Ptotal[$nBlocksIn-1]=1.E-10;}
        $Bruls_focus = 1./$Ptotal[$nBlocksIn-1];
        ###
        ### Now, once we have the focus calculated, it is measured from the rear
        ### principal optical plane. The distance of that plane from the last
        ### lens is given by Brul's equation (17a):
        ###                  n-1
        ### P_r - L_n = -f * sum (P_i * D_(i,i+1)
        ###                  i=1
        $sum = 0;
        for ($j=0; $j<$nBlocksIn-1; $j++) {
           $d = abs($D[$j+1]-$D[$j]);
           $sum += $Ptotal[$j]*$d;
        }
        $Pr2Ln = -$Bruls_focus * $sum;
        $Bruls_fadd = $D[$nBlocksIn-1] + $Pr2Ln;
     }
     $Bruls_focus += $Bruls_fadd;
#    print STDOUT 'In/Out='.$Bruls_inout.' f='.$Bruls_focus.', f_add='.$Bruls_fadd."\n";
     $CRL_Bruls[$n]{'inout'} = $Bruls_inout;
     $CRL_Bruls[$n]{'in'}    = $Bruls_in;
     $CRL_Bruls[$n]{'focus'} = $Bruls_focus;
     $CRL_Bruls[$n]{'fadd'}  = $Bruls_fadd;
     $n++;
  } #-------------------------------------------------------------------end of loop over all possible combinations

### Sort array over the deviation from requested focus

  if ($n > 1) {
     @CRL_Bruls = sort{ abs($focusRqs - $a->{'focus'}) <=> abs($focusRqs - $b->{'focus'}) } @CRL_Bruls;
  }
  if ($debug) {
     printf STDOUT 'Best match to focus(rqs)=%.2f at E=%.3f keV:'."\n",$focusRqs,$E;
     for ($j=0; $j<&minval($debug,$ncomb); $j++) {
        printf STDOUT '%3d.  %s, in=%d, focus=%.2f, diff=%.2f'."\n", 
            $j,
            $CRL_Bruls[$j]{'inout'},
            $CRL_Bruls[$j]{'in'},
            $CRL_Bruls[$j]{'focus'},
            abs($focusRqs-$CRL_Bruls[$j]{'focus'});
     }
  }
  if ($minimizeFlag && $n > 1) {					# if minimize attenuation
     for ($j=0; $j<$ncomb; $j++) {
        if (abs($focusRqs - $CRL_Bruls[$j]{'focus'}) > $range) {
           $#CRL_Bruls = $j;
           $ncomb      = $j+1;
           last;
        }
     }
     if ($debug) {
        printf STDOUT 'Number of Best Attenuation candidates for range=+-%.2f is n=%d'."\n",$range,$ncomb;
     }
     @CRL_Bruls = sort{$a->{'in'} <=> $b->{'in'}} @CRL_Bruls;
     if ($debug) {
        printf STDOUT 'Best attenuation match to focus(rqs)=%.2f at E=%.3f keV:'."\n",$focusRqs,$E;
        for ($j=0; $j<$ncomb; $j++) {
           printf STDOUT '%3d/%d.  %s, in=%d, focus=%.2f, diff=%.2f'."\n", 
               $j, $ncomb,
               $CRL_Bruls[$j]{'inout'}, 
               $CRL_Bruls[$j]{'in'},
               $CRL_Bruls[$j]{'focus'},
               abs($focusRqs-$CRL_Bruls[$j]{'focus'});
        }
     }
  }
  if ($dump || !defined $CRL_Bruls[0]{'inout'}) {
     print STDOUT 'All found data:'."\n";
     for ($j=0; $j<$ncomb; $j++) {
        printf STDOUT '%d/%d   %s, in=%d, focus=%.2f, diff=%.2f'."\n", 
            $j, $ncomb, 
            $CRL_Bruls[$j]{'inout'},
            $CRL_Bruls[$j]{'in'},
            $CRL_Bruls[$j]{'focus'},
            abs($focusRqs-$CRL_Bruls[$j]{'focus'});
     }
  }
  $i = &InOut_str2arr($CRL_Bruls[0]{'inout'},\@InOut);
  return ($CRL_Bruls[0]{'focus'}, @InOut);
}

#=====================================================================

# ($focusAct,@InOut) = &setTransfocatorFocus4($focusRqs,[$E,$device,\$msg,$minimizeAttenuation,$focusRange,$TF_axis_PV]);
# ($focusAct,@InOut) = &setTransfocatorFocus4($focusRqs,
#                                                       [$E,			#Default: $E=current
#                                                        $device,		#Default: CRL-2020
#                                                        \$msg,			#Default: none (this is output)
#                                                        $minimizeAttenuation,	#Default: 0=minimize |Act-Rqs|
#                                                        $focusRange,		#Default: $focusRange1=1/4 of TF travel range
#                                                        $TF_axis_PV]);		#Default: 'CRL:X'
sub setTransfocatorFocus4($;$$\$$$$) {
###
### Finds closest match to requested transfocator focus in mm and then sets it. 
### Uses Bruls exact formula for multi-lens focus & checks all combinations.
### First calls &matchTransfocatorFocus4 and then &setTransfocatorInOut
###
  our (@CRLconfig);							#cached array of the latest CRL config (what is normally read from SQL)
  our ($beamline);							#defined in beamline.pl
  my ($script, $E, $device, $msgref, $minimizeFlag);
  my ($range, $axis_PV, $nblocks);
  my ($focusRqs, $focusAct, @InOut);
  my ($energyPV, $status, $statusDescr, $msg);

  $script = 'setTransfocatorFocus4';

  $focusRqs     = shift(@_);						#reference to array of IN=1 & OUT=0 lense blocks
  $E            = shift(@_);						#if not specified, taken from EPICS
  $device       = shift(@_);						#if not specified, taken from currently stored SQL cache, or if not available, then default of &readCRLconfigSQL
  $msgref       = shift(@_);						#if not specified, messages are not returned
  $minimizeFlag = shift(@_);						#if not specified, taken 0=minimize |Act-Rqs|; can be 0 or 1=minimize attenuation
  $range        = shift(@_);						#if not specified, taken as 1/4 of TF travel range
  $axis_PV      = shift(@_);						#if not specified, taken as 'CRL:X'

  if (defined $msgref) {$$msgref = '';}
  if (!defined $beamline) {die $script.': no beamline defined.'."\n";}

### Read SQL if no cache or if cached for different device (device name is needed!):
  $nblocks = @CRLconfig;
  if ($nblocks == 0 || (defined $device && $CRLconfig[0]{'Device'} ne $device)) {
     @CRLconfig = ();
     $nblocks = &readCRLconfigSQLpopulated($device,\@CRLconfig);
  }
  if ($nblocks < 1) {return -1;}

### Set device if not specified in input, but we have @CRLconfig cached:
  if (!defined $device) {
     $device = $CRLconfig[0]{'Device'};
     print STDOUT $script.': transfocator device is set to "'.$device.'"'."\n";
  }
### Get current energy if not specified in input:
  if (!defined $E) {
     $energyPV = $beamline.':MO:E:ActPos';
     ($status, $E) = Pezca::GetDouble($energyPV);
     if ($status) {
        $statusDescr = &ezcaStatusCode($status);
        $msg = '*** '.$script.': EZCA error='.$status.' ['.$statusDescr.'] on PV='.$energyPV;
        print STDOUT $msg,"\n";
        if (defined $msgref) {$$msgref = $msg;}
        return -1;
     }
  }

  ($focusAct,@InOut) = &matchTransfocatorFocus4($focusRqs,$E,$device,$msgref,$minimizeFlag,$range,$axis_PV);
  if (!defined $focusAct) {
     print STDOUT '*** '.$script.': undefined $focusAct from &matchTransfocatorFocus4'."\n"; 
     &call_trace();
     return -1;
  }
  elsif ($focusAct < 0) {return -1;}

### Set lenses IN and OUT:
  $status = &setTransfocatorInOut(\@InOut,$device,\$msg);
  if ($status) {
     $msg .= "\n".'*** '.$script.': error='.$status.' from setTransfocatorInOut';
     print STDOUT $msg,"\n";
     if (defined $msgref) {$$msgref = $msg;}
  }

  return ($focusAct,@InOut);
}

#=====================================================================

# $status = &fillCRLcombPD(\@CRLcomb,[$E,$device,\$msg]);
sub fillCRLcombPD(\@;$$\$) {
###
### Fills power and distances for the @CRLcomb array used by &calcTransfocatorFocus5
### 
  our (@CRLconfig);							#cached array of the latest CRL config (what is normally read from SQL)
  our ($beamline);							#defined in beamline.pl
  my ($script, $CRLcomb_ref, $E, $device, $msgref);
  my ($msg, $nrecords, @CRLcomb);
  my ($energyPV, $status, $statusDescr);
  my ($i, $r, $n, $m, $m0, $delta);

  $script      = 'fillCRLcombPD';
  $CRLcomb_ref = shift(@_);						#associate array of power and distances to fill
  $E           = shift(@_);						#if not specified, taken from EPICS
  $device      = shift(@_);						#if not specified, taken from currently stored SQL cache, or if not available, then default of &readCRLconfigSQL
  $msgref      = shift(@_);						#if specified, messages are returned

  if (defined $msgref)          {$$msgref = '';}
  if (!defined $beamline)       {die $script.': no beamline defined.'."\n";}
  
  @$CRLcomb_ref = ();

### Read SQL if no cache or cached for different device:
  $nrecords = @CRLconfig;
  if ($nrecords == 0 || (defined $device && $CRLconfig[0]{'Device'} ne $device)) {
     @CRLconfig = ();
     $nrecords = &readCRLconfigSQLpopulated($device,\@CRLconfig);
  }
  if ($nrecords < 1) {return -1;}

### Set device if not specified in input:
  if (!defined $device) {
     $device = $CRLconfig[0]{'Device'};
     print STDOUT $script.': transfocator device is set to "'.$device.'"'."\n";
  }
### Get current energy if not specified in input:
  if (!defined $E) {
     $energyPV = $beamline.':MO:E:ActPos';
     ($status, $E) = Pezca::GetDouble($energyPV);
     if ($status) {
        $statusDescr = &ezcaStatusCode($status);
        $msg = '*** '.$script.': EZCA error='.$status.' ['.$statusDescr.'] on PV='.$energyPV;
        print STDOUT $msg,"\n";
        if (defined $msgref) {$$msgref = $msg;}
        return -1;
     }
  }

  @CRLcomb = ();
  for ($i=0; $i<$nrecords; $i++) {
     $r = $CRLconfig[$i]{'Radius'};					#Lenses radius (um). If $radius=0 no lenses are installed in the subset
     $n = $CRLconfig[$i]{'Number'};					#Number of lenses in the box. In $number=0, no lenses are installed in the subset
     $m = $CRLconfig[$i]{'Material'};					#Lenses material, normally Be. If $material is blank, no lenses are installed in the subset
     if (!defined $m0 || $m ne $m0) {					#if block material differs, calculate delta for new material
        $delta = &calcDelta($E,$m,$msgref);
        if ($delta == 0) {return -1;}
        $m0 = $m;
     }
     $CRLcomb[$i]{'Power'}  = &calcCRLpower($E,$r,$n,$delta,$msgref);	#the power of lens block (1/focus)
     $CRLcomb[$i]{'Offset'} = $CRLconfig[$i]{'Offset'};			#the distance of lens block from the transfocator center
     $CRLcomb[$i]{'InOut'}  = 0;					#the In/Out state of lens block
  }
  @$CRLcomb_ref = @CRLcomb;
  return 0;
}

#=====================================================================

# $status = &fillCRLcombIO(\@CRLcomb,\@IO);
sub fillCRLcombIO(\@\@) {
###
### Fills In/Out for the @CRLcomb array used by &calcTransfocatorFocus5
### 
  my ($script, $CRLcomb_ref, $InOut_ref);
  my ($nCRL, $nIO, $i);

  $script      = 'fillCRLcombIO';
  $CRLcomb_ref = shift(@_);						#associate array of power and distances to fill
  $InOut_ref   = shift(@_);						#array of In/Out

### Read SQL if no cache or cached for different device:
  $nCRL = @$CRLcomb_ref;
  $nIO  = @$InOut_ref;
  if ($nCRL != $nIO) {
     print STDOUT '*** '.$script.': numbers not match: nCRL='.$nCRL.' nIO='.$nIO."\n";
     return -1;
  }

  for ($i=0; $i<$nCRL; $i++) {
     @$CRLcomb_ref[$i]->{'InOut'} = @$InOut_ref[$i];			#the distance of lenses block from the transfocator center
  }
  return 0;
}

#=====================================================================

# $focus = &getTransfocatorFocus5($E,$device,[\$msg,\@InOut]);
sub getTransfocatorFocus5(;$$\$\@) {
### 
### Returns overall Transfocator focus (Bruls method) in mm for current
### lenses in/out configuration. Calls &getTransfocatorInOut and then
### &calcTransfocatorFocus5.
### This routine does exactly same as &getTransfocatorFocus4 except it
### calls &calcTransfocatorFocus5 instead of &calcTransfocatorFocus4.
### The results coincide and either one can be called.
### 
  our (@CRLconfig);							#cached array of the latest CRL config (what is normally read from SQL)
  our ($beamline);							#defined in beamline.pl
  my ($script, $E, $device, $msgref, $IOref);
  my (@IO, $focus, $msg, $nrecords);
  my ($energyPV, $status, $statusDescr);

  $script         = 'getTransfocatorFocus5';
  $E              = shift(@_);						#if not specified, taken from EPICS
  $device         = shift(@_);						#if not specified, taken from currently stored SQL cache, or if not available, then default of &readCRLconfigSQL
  $msgref         = shift(@_);						#if specified, messages are returned
  $IOref          = shift(@_);						#if specified, returns current In/Out

  if (defined $msgref)          {$$msgref = '';}
  if (!defined $beamline)       {die $script.': no beamline defined.'."\n";}
  
### Read SQL if no cache or cached for different device:
  $nrecords = @CRLconfig;
  if ($nrecords == 0 || (defined $device && $CRLconfig[0]{'Device'} ne $device)) {
     @CRLconfig = ();
     $nrecords = &readCRLconfigSQLpopulated($device,\@CRLconfig);
  }
  if ($nrecords < 1) {return -1;}

### Set device if not specified in input:
  if (!defined $device) {
     $device = $CRLconfig[0]{'Device'};
     print STDOUT $script.': transfocator device is set to "'.$device.'"'."\n";
  }
### Get current energy if not specified in input:
  if (!defined $E) {
     $energyPV = $beamline.':MO:E:ActPos';
     ($status, $E) = Pezca::GetDouble($energyPV);
     if ($status) {
        $statusDescr = &ezcaStatusCode($status);
        $msg = '*** '.$script.': EZCA error='.$status.' ['.$statusDescr.'] on PV='.$energyPV;
        print STDOUT $msg,"\n";
        if (defined $msgref) {$$msgref = $msg;}
        return -1;
     }
  }

  @IO = ();
  $status = &getTransfocatorInOut(\@IO,$device,$msgref);
  if ($status) {return -1;}
  if (defined $IOref) {@$IOref = @IO;}

  $focus = &calcTransfocatorFocus5(\@IO,$E,$device,$msgref);
  return $focus;
}

#=====================================================================

#$focus = &calcTransfocatorFocus5(\@InOut,[$E,$device,\$msg]);
sub calcTransfocatorFocus5(\@;$$\$) {
###
### For given @InOut and $E, calculates exact Transfocator focus by the Bruls method:
### G.J.C.L. Bruls "Exact formulas for a thin-lens system with an arbitrary
### number of lenses", Optik, Volume 126, Issue 6, March 2015, Pages 659-662.
### (https://dx.doi.org/10.1016/j.ijleo.2015.01.035)
### This routine uses the same algorithm as &calcTransfocatorFocus4 and
### the results fully coincide. The only difference is in the arrangement
### of calls: this one uses &calcTransfocatorFocusComb(\@CRLcombination);
###
  our (@CRLconfig);							#cached array of the latest CRL config (what is normally read from SQL)
  our ($beamline);							#defined in beamline.pl
  my ($script, $IOref, $E, $device, $msgref);
  my (@CRLcomb, $nblocks, $focus, $nIO);
  my ($energyPV, $status, $statusDescr);
  my ($msg);

  $script = 'calcTransfocatorFocus5';
  $IOref  = shift(@_);							#reference to array of IN=1 & OUT=0 lense blocks
  $E      = shift(@_);							#if not specified, taken from EPICS
  $device = shift(@_);							#if not specified, taken from currently stored SQL cache, or if not available, then default of &readCRLconfigSQL
  $msgref = shift(@_);							#if not specified, messages are not returned

  if (defined $msgref)          {$$msgref = '';}
  if (!defined $beamline)       {die $script.': no beamline defined.'."\n";}
  
### Read SQL if no cache or cached for different device:
  $nblocks = @CRLconfig;						#has the config has not been read?
  if ($nblocks == 0 || (defined $device && $CRLconfig[0]{'Device'} ne $device)) {
     @CRLconfig = ();
     $nblocks = &readCRLconfigSQLpopulated($device,\@CRLconfig);
  }
  if ($nblocks < 1) {return -1;}

### Set device if not specified in input:
  if (!defined $device) {
     $device = $CRLconfig[0]{'Device'};
     print STDOUT $script.': transfocator device is set to "'.$device.'"'."\n";
  }
  $nIO = @$IOref;
  if ($nIO != $nblocks) {
     $msg = '*** '.$script.': In/Out array length='.$nIO.' does not match used lenses blocks number='.$nblocks.' in device='.$device;
     print STDOUT $msg,"\n";
     if (defined $msgref) {$$msgref = $msg;}
     return -1;
  }
### Get current energy if not specified in input:
  if (!defined $E) {
     $energyPV = $beamline.':MO:E:ActPos';
     ($status, $E) = Pezca::GetDouble($energyPV);
     if ($status) {
        $statusDescr = &ezcaStatusCode($status);
        $msg = '*** '.$script.': EZCA error='.$status.' ['.$statusDescr.'] on PV='.$energyPV;
        print STDOUT $msg,"\n";
        if (defined $msgref) {$$msgref = $msg;}
        return -1;
     }
  }

  @CRLcomb = ();
  $status = &fillCRLcombPD(\@CRLcomb,$E,$device,$msgref);
  if ($status) {return -1;}

  $status = &fillCRLcombIO(\@CRLcomb,$IOref);
  if ($status) {return -1;}

  $focus = &calcTransfocatorFocusComb(\@CRLcomb);
  return $focus;
}

#=====================================================================

#$focus = &calcTransfocatorFocusComb(\@CRLcombination);
sub calcTransfocatorFocusComb(\@) {
###
### For given @CRLcombination calculates exact Transfocator focus by the Bruls method:
### G.J.C.L. Bruls "Exact formulas for a thin-lens system with an arbitrary
### number of lenses", Optik, Volume 126, Issue 6, March 2015, Pages 659-662.
### (https://dx.doi.org/10.1016/j.ijleo.2015.01.035)
### Here @CRLcombination is an array of references to {'Power'}, {'Offset'}, {'InOut'}
### This routine is used by &calcTransfocatorFocus5 and &matchTransfocatorFocus5.
###
  my ($script, $CRLref, @CRLcomb);
  my ($nblocks, $d, $f);
  my ($i, $j, $nBlocksIn);
  my (@P, @D, @Ptotal, $sum, $Pr2Ln, $fadd);

  $script = 'calcTransfocatorFocusComb';
  $CRLref = shift(@_);							#reference to array of IN=1 & OUT=0 lense blocks
  
  $nblocks = @$CRLref;							#have any lens blocks been passed?
  if ($nblocks < 1) {return -1;}

### Lens blocks may be in any order. I am not sure that
### it is important for the Bruls method, but just in 
### case we want then in the upstream-to-downsteam order:
  @CRLcomb = sort{ $a->{'Offset'} <=> $b->{'Offset'} } @$CRLref;	#sort in the offset increasing order

### Create an array of powers (p=1/f) and distances for those lenses that are In:
  @D = ();
  @P = ();
  @Ptotal = ();
  $j = 0;
  for ($i=0; $i<$nblocks; $i++) {
     if ($CRLcomb[$i]{'InOut'}) {					#lenses block is IN
        $P[$j] = $CRLcomb[$i]{'Power'};
        $D[$j] = $CRLcomb[$i]{'Offset'};
        $j++;
     }
  }
  $nBlocksIn = $j;
  if    ($nBlocksIn == 0) {
     $f     = 1.E10;
     $Pr2Ln = 0;
     $fadd  = 0;
  }
  elsif ($nBlocksIn == 1) {
     $f     = 1./$P[0];
     $Pr2Ln = 0;
     $fadd  = $D[0] + $Pr2Ln;
  }
  else {
     $Ptotal[0] = $P[0];
     for ($i=1; $i<$nBlocksIn; $i++) {
        $sum = 0;
        for ($j=0; $j<$i; $j++) {
           $d = abs($D[$j+1]-$D[$j]);
           $sum += $Ptotal[$j]*$d;
        }
        $Ptotal[$i] = $Ptotal[$i-1] + $P[$i]*(1-$sum);
     }
     if ($Ptotal[$nBlocksIn-1] <= 0) {return -1;}
     $f = 1./$Ptotal[$nBlocksIn-1];
### Now once we have the focus calculated, it is measured from the rear
### principal optical plane. The distance of that plane from the last
### lens is given by Brul's equation (17a):
###                  n-1
### P_r - L_n = -f * sum (P_i * D_(i,i+1)
###                  i=1
     $sum = 0;
     for ($i=0; $i<$nBlocksIn-1; $i++) {
        $d = abs($D[$i+1]-$D[$i]);
        $sum += $Ptotal[$i]*$d;
     }
     $Pr2Ln = -$f * $sum;
     $fadd  = $D[$nBlocksIn-1] + $Pr2Ln;
  }
  $f += $fadd;
  return $f;
}

#=====================================================================

# ($focusAct,@InOut) = &matchTransfocatorFocus5($focusRqs,[$E,$device,\$msg]);
# ($focusAct,@InOut) = &matchTransfocatorFocus5($focusRqs,
#                                               [$E,			#Default: $E=current
#                                                $device,		#Default: CRL-2020
#                                                \$msg]);		#Default: none (this is output)
sub matchTransfocatorFocus5($;$$\$) {
###
### Finds closest match to requested transfocator focus in mm, but does not set.
### (Bruls exact formula for multi-lens focus)
### Defaults: E=current, device=CRL-2020.
### Unlike matchTransfocatorFocus4, this routine does not check all 
### combinations, but uses a "tetris" method: first tries to match
### requested power with most powerful blocks and then keeps adding
### smaller and smaller blocks.
###
  our (@CRLconfig);							#cached array of the latest CRL config (what is normally read from SQL)
  our ($beamline);							#defined in beamline.pl
  my ($script, $focusRqs, $E, $device, $msgref);
  my (@InOut, @Power, $nblocks, $nBlocksIn, $binstr);
  my ($energyPV, $status, $statusDescr, $msg);
  my ($focus, @CRLcomb, @indx, @indb, $i, $j, $debug);

  $script = 'matchTransfocatorFocus5';
  $debug  = 0;								#set to 1 for debug messages

  $focusRqs = shift(@_);						#reference to array of IN=1 & OUT=0 lense blocks
  $E        = shift(@_);						#if not specified, taken from EPICS
  $device   = shift(@_);						#if not specified, taken from currently stored SQL cache, or if not available, then default of &readCRLconfigSQL
  $msgref   = shift(@_);						#if not specified, messages are not returned

  if (defined $msgref)        {$$msgref = '';}
  if (!defined $beamline)     {die $script.': no beamline defined.'."\n";}
  if (!defined $debug)        {$debug = 0;}
  
### Read SQL if no cache or if cached for different device (device name is needed!):
  $nblocks = @CRLconfig;
  if ($nblocks == 0 || (defined $device && $CRLconfig[0]{'Device'} ne $device)) {
     @CRLconfig = ();
     $nblocks = &readCRLconfigSQLpopulated($device,\@CRLconfig);
     if ($nblocks < 1) {return -1;}
  }
  if ($debug) {print STDOUT $script.': nblocks='.$nblocks."\n";}

### Set device if not specified in input, but we have @CRLconfig cached:
  if (!defined $device) {
     $device = $CRLconfig[0]{'Device'};
     if ($debug) {print STDOUT $script.': transfocator device is set to "'.$device.'"'."\n";}
  }

### Get current energy if not specified in input:
  if (!defined $E) {
     $beamline =~ s/:$//;
     $energyPV = $beamline.':MO:E:ActPos';
     ($status, $E) = Pezca::GetDouble($energyPV);
     if ($status) {
        $statusDescr = &ezcaStatusCode($status);
        $msg = '*** '.$script.': EZCA error='.$status.' ['.$statusDescr.'] on PV='.$energyPV;
        print STDOUT $msg,"\n";
        if (defined $msgref) {$$msgref = $msg;}
        return -1;
     }
  }

  @Power = &calcAllOneByFocus($E,$device,$msgref);
### Sort over decreasing power (we want to consider the most powerful blocks first):
  @indx = sort {$Power[$b] <=> $Power[$a]} 0..$#Power;			#new indices
  @indb = sort {$indx[$a]  <=> $indx[$b]}  0..$#indx;			#old places of new indices

  $status = 0;
  for ($i=0; $i<$nblocks; $i++) {
     if ($Power[$i] <= 0) {
        $msg = '*** '.$script.': 1/f=0 for lense block='.$i."\n";
        print STDOUT $msg;
        if (defined $msgref) {$$msgref .= $msg;}
        $status++;
     }
  }
  if ($status) {return -1;}

### Fill @CRL_Bruls (array of sum{1/f} for all in/out combinations):
  @CRLcomb = ();
  for ($i=0; $i<$nblocks; $i++) { #-------------------------------------#begin loop over all lens blocks
     $j = $indx[$i];
     $CRLcomb[$i]{'Power'}  = $Power[$j];
     $CRLcomb[$i]{'Offset'} = $CRLconfig[$j]{'Offset'},
     $CRLcomb[$i]{'InOut'}  = 1;
     $focus = &calcTransfocatorFocusComb(\@CRLcomb);
     if ($focus < $focusRqs) {$CRLcomb[$i]{'InOut'} = 0;}		#if too much power with the last addition, consider Out
  }
  if ($debug) {
     print STDOUT 'Block   Power   Offset   InOut'."\n";
     for ($i=0; $i<$nblocks; $i++) { #-------------------------------------#begin loop over all lens blocks
        printf STDOUT '%3d %10.3g %7.2f %5d'."\n",
                      ($i+1),
                      $CRLcomb[$i]{'Power'},
                      $CRLcomb[$i]{'Offset'},
                      $CRLcomb[$i]{'InOut'};
     }
  }

  $focus = &calcTransfocatorFocusComb(\@CRLcomb);
  @InOut = ();
  $binstr = '';
  $nBlocksIn = 0;
  for ($i=0; $i<$nblocks; $i++) { #-------------------------------------#begin loop over all lens blocks
     $j = $indb[$i];
     $InOut[$i] = $CRLcomb[$j]{'InOut'};
     $binstr .= $InOut[$i];
     if ($InOut[$i]) {$nBlocksIn++;}
  }
  if ($debug) {
     printf STDOUT 'Best match to focus(rqs)=%.2f at E=%.3f keV:'."\n",$focusRqs,$E;
     printf STDOUT '(%s), in=%d, focus=%.2f, diff=%.2f'."\n", 
            $binstr,
            $nBlocksIn,
            $focus,
            abs($focusRqs-$focus);
  }
  return ($focus, @InOut);
}

#=====================================================================

# ($focusAct,@InOut) = &setTransfocatorFocus5($focusRqs,[$E,$device,\$msg]);
# ($focusAct,@InOut) = &setTransfocatorFocus5($focusRqs,
#                                                       [$E,			#Default: $E=current
#                                                        $device,		#Default: CRL-2020
#                                                        \$msg]);		#Default: none (this is output)
sub setTransfocatorFocus5($;$$\$) {
###
### Finds closest match to requested transfocator focus in mm and then sets it. 
### Uses Bruls exact formula for multi-lens focus & a "tetris" method: first
### tries to match requested power with most powerful blocks and then keeps
### adding smaller and smaller blocks.
### First calls &matchTransfocatorFocus5 and then &setTransfocatorInOut
###
  our (@CRLconfig);							#cached array of the latest CRL config (what is normally read from SQL)
  our ($beamline);							#defined in beamline.pl
  my ($script, $E, $device, $msgref);
  my ($focusRqs, $focusAct, @InOut, $nblocks);
  my ($energyPV, $status, $statusDescr, $msg);

  $script = 'setTransfocatorFocus5';

  $focusRqs     = shift(@_);						#reference to array of IN=1 & OUT=0 lense blocks
  $E            = shift(@_);						#if not specified, taken from EPICS
  $device       = shift(@_);						#if not specified, taken from currently stored SQL cache, or if not available, then default of &readCRLconfigSQL
  $msgref       = shift(@_);						#if not specified, messages are not returned

  if (defined $msgref) {$$msgref = '';}
  if (!defined $beamline) {die $script.': no beamline defined.'."\n";}

### Read SQL if no cache or if cached for different device (device name is needed!):
  $nblocks = @CRLconfig;
  if ($nblocks == 0 || (defined $device && $CRLconfig[0]{'Device'} ne $device)) {
     @CRLconfig = ();
     $nblocks = &readCRLconfigSQLpopulated($device,\@CRLconfig);
  }
  if ($nblocks < 1) {return -1;}

### Set device if not specified in input, but we have @CRLconfig cached:
  if (!defined $device) {
     $device = $CRLconfig[0]{'Device'};
     print STDOUT $script.': transfocator device is set to "'.$device.'"'."\n";
  }
### Get current energy if not specified in input:
  if (!defined $E) {
     $energyPV = $beamline.':MO:E:ActPos';
     ($status, $E) = Pezca::GetDouble($energyPV);
     if ($status) {
        $statusDescr = &ezcaStatusCode($status);
        $msg = '*** '.$script.': EZCA error='.$status.' ['.$statusDescr.'] on PV='.$energyPV;
        print STDOUT $msg,"\n";
        if (defined $msgref) {$$msgref = $msg;}
        return -1;
     }
  }

  ($focusAct,@InOut) = &matchTransfocatorFocus5($focusRqs,$E,$device,$msgref);
  if (!defined $focusAct) {
     print STDOUT '*** '.$script.': undefined $focusAct from &matchTransfocatorFocus5'."\n"; 
     &call_trace();
     return -1;
  }
  elsif ($focusAct < 0) {return -1;}

### Set lenses IN and OUT:
  $status = &setTransfocatorInOut(\@InOut,$device,\$msg);
  if ($status) {
     $msg .= "\n".'*** '.$script.': error='.$status.' from setTransfocatorInOut';
     print STDOUT $msg,"\n";
     if (defined $msgref) {$$msgref = $msg;}
  }

  return ($focusAct,@InOut);
}

#=====================================================================

#($serverState,$focusRef,$focusRqs,$focusAct,
#  $xPos,$searchMode,$maxRange) = &getTFserverState();
sub getTFserverState() {
###
### Reads all parameters of Transfocator Server from PVs
###
  our ($beamline);							#defined in beamline.pl
  my ($script, $status);
  my ($serverSts,  $serverSts_PV);
  my ($focusRef,   $focusRef_PV);
  my ($focusRqs,   $focusRqs_PV);
  my ($focusAct,   $focusAct_PV);
  my ($deviceXpos, $deviceXpos_PV);
  my ($searchMode, $searchMode_PV);
  my ($rangeMax,   $rangeMax_PV);

  $script = 'getTFserverState';
  if (!defined $beamline) {die $script.': no beamline defined.'."\n";}

  $serverSts_PV  = $beamline.':focusSts';				#0=stopped,1=processing,2=ready
  $focusRef_PV   = $beamline.':focusReference';
  $focusRqs_PV   = $beamline.':focusRqs';
  $focusAct_PV   = $beamline.':focusAct';
  $deviceXpos_PV = $beamline.':CRL:X:ActPos';
  $searchMode_PV = $beamline.':focusSearchMode';
  $rangeMax_PV   = $beamline.':focusMaxRange';

  ($status,$serverSts) = Pezca::GetDouble($serverSts_PV);		#0=stopped,1=processing,2=ready
  if ($status) {
     &report_only($script,$status,$serverSts_PV);
     $serverSts = 0;
  }
  ($status,$focusRef) = Pezca::GetDouble($focusRef_PV);			#reference focus (setpoint)
  if ($status) {
     &report_only($script,$status,$focusRef_PV);
     $focusRef = 0;
  }
  ($status,$focusRqs) = Pezca::GetDouble($focusRqs_PV);			#requested focus
  if ($status) {
     &report_only($script,$status,$focusRqs_PV);
     $focusRqs = 0;
  }
  ($status,$focusAct) = Pezca::GetDouble($focusAct_PV);			#actual focus
  if ($status) {
     &report_only($script,$status,$focusAct_PV);
     $focusAct = 0;
  }
  ($status,$deviceXpos) = Pezca::GetDouble($deviceXpos_PV);		#device upstream/downstream position
  if ($status) {
     &report_only($script,$status,$deviceXpos_PV);
     $deviceXpos = 0;
  }
  ($status,$searchMode) = Pezca::GetDouble($searchMode_PV);		#0=find best focus match, 1=minimize attenuation within range
  if ($status) {
     &report_only($script,$status,$searchMode_PV);
     $searchMode = 0;
  }
  ($status,$rangeMax) = Pezca::GetDouble($rangeMax_PV);			#max range of |focusAct-FocusRqs| for minimizing attenuation
  if ($status) {
     &report_only($script,$status,$rangeMax_PV);
     $rangeMax = 0;
  }
  return ($serverSts,$focusRef,$focusRqs,$focusAct,$deviceXpos,$searchMode,$rangeMax);
}

#=====================================================================

# ($focusAct,$xPos) = &setTFserverFocus($focusRqs,[$wait_max]);
sub setTFserverFocus($;$) {
###
### Commands the Transfocator Server to set requested focus with whatever
### method the Server is using.
### 
  our ($beamline);							#defined in beamline.pl
  my ($script, $status, $wait_max, $wait_step, $wait, $msg);
  my ($serverSts,  $serverSts_PV);
  my ($focusSet,   $focusSet_PV);
  my ($focusRqs,   $focusRqs_PV);
  my ($focusAct,   $focusAct_PV);
  my ($deviceXpos, $deviceXpos_PV);

  $script = 'setTFserverFocus';

  $focusRqs = shift(@_);
  $wait_max = shift(@_);	if (!defined $wait_max) {$wait_max = 60;}  
  
  if (!defined $beamline) {die $script.': no beamline defined.'."\n";}
  $serverSts_PV  = $beamline.':focusSts';				#0=stopped,1=processing,2=ready
  $focusSet_PV   = $beamline.':focusSet';
  $focusRqs_PV   = $beamline.':focusRqs';
  $focusAct_PV   = $beamline.':focusAct';
  $deviceXpos_PV = $beamline.':CRL:X:ActPos';

  $msg = '';
  $status = &addCaMonitor($serverSts_PV); 
  if ($status) {$msg .= '*** '.$script.': failed to set CA monitor on '.$serverSts_PV."\n"; print STDOUT $msg; return (-1,0,$msg);}
  $status = &addCaMonitor($focusSet_PV); 
  if ($status) {$msg .= '*** '.$script.': failed to set CA monitor on '.$focusSet_PV."\n";  print STDOUT $msg; return (-1,0,$msg);}
  ($status,$serverSts) = Pezca::GetDouble($serverSts_PV);		#0=stopped,1=processing,2=ready
  if ($status) {&report_only($script,$status,$serverSts_PV); $serverSts=0;}

  if ($serverSts == 0) {$msg .= '*** '.$script.': Transfocator server is stopped'."\n"; print STDOUT $msg; return (-1,0,$msg);}
  $status = Pezca::PutDoubleOldCa($focusRqs_PV,$focusRqs);
  if ($status) {&report_only($script,$status,$focusRqs_PV); return -1;}
  $wait = 2;
  select(undef,undef,undef,$wait);					#sleep(..);
  $wait_step = 0.1;
  do {
     select(undef,undef,undef,$wait_step);				#sleep(..);
     $wait += $wait_step;
     ($status,$serverSts) = Pezca::GetDouble($serverSts_PV);		#0=stopped,1=processing,2=ready
     if ($status) {&report_only($script,$status,$serverSts_PV); $serverSts=0;}
     ($status,$focusSet) = Pezca::GetDouble($focusSet_PV);		#0=idle 1=setting
     if ($status) {&report_only($script,$status,$focusSet_PV); $focusSet=1;}
  } while (($serverSts != 2 || $focusSet != 0) && $wait < $wait_max);

  if ($wait >= $wait_max) {$msg .= '*** '.$script.': timeout on waiting for focus to set.'."\n";}

  ($status,$focusAct) = Pezca::GetDouble($focusAct_PV);			#actual focus
  if ($status) {&report_only($script,$status,$focusAct_PV);   $focusAct = 0;}

  ($status,$deviceXpos) = Pezca::GetDouble($deviceXpos_PV);		#device upstream/downstream position
  if ($status) {&report_only($script,$status,$deviceXpos_PV); $deviceXpos = 0;}

  return ($focusAct,$deviceXpos,$msg);
}
