#!/usr/bin/perl -W
#--------------------------------------------------------------
# These are DAC related subs for beamline controls Perl scripts
#
#     		Author: Sergey Stepanov
#
# Version-1.0:  2004/07/23 using Pezca
# Version-2.0:  2011/01/10 scan overload flag added
# Version-3.0:  2011/11/09 added filess plot to scanDAC
# Version-3.1:  2013/08/31 added optional DAC check (ADC "encoder") to scanDAC
# Version-3.2:  2013/10/03 added shutter to scanDAC
# Version-3.3:  2017/10/25 added DACslopeSupported and driveDAC_slope
# Version-3.4:  2018/11/16 added Gaussian/Lorentzian peak fitting to stepScanDAC
#                          (needs PDL, PDL::Fit::LM and peakFit from common_PDL.pl).
#                          Also moved all after scan operations to DAC_afterScan
# Version-3.5:  2023/02/06 added PVlist to stepScanDAC (NOTE new parameter in calling!)
#--------------------------------------------------------------
# List of functions:
# ^^^^^^^^^^^^^^^^^
#  1.       $status = &setupDAC ($dacPV);
#  2.    $dac_value = &readDAC ($dacPV);
#  3. ($drvL,$drvH) = &limitsDAC ($dacPV);
#  4.       $status = &driveDAC ($dacPV, $rqsPos, $drvL, $drvH, [$delay,$inlimits]);     #drvL/drvH: 0/0=no check, undef=auto
#  5.       $status = &shiftDAC ($dacPV, $rqsRel, $drvL, $drvH, [$delay,$inlimits]);     #drvL/drvH: 0/0=no check, undef=auto
#  6.       $status = &stepScanDAC($dacPV, $dacBegin, $dacEnd, $dacStep,
#                                  $scaler, $countTime, 
#                                  \@inputs, [\@pvlist, $fileBkg,
#                                  $gnuplot, $logscale, \@gnuopts,
#                                  $afterScan, $checkOverload, $gnufile,
#                                  $postarrayPV,$post_interval,
#                                  $adc_encdrPV, $dac_tolerance]);
#                    ----> This is a *STEP* scan
#                    ----> Also returns global arrays @x1DAC, @y1DAC, ... @ynDAC
#                    ----> See also 'SRstepScanDAC' in scanrec_subs.pl
#
#  7.                 &DAC_afterScan($afterScan,\@xarr,\@yarr,$begin,$end,$dac_ini,$dacPV);
#  8. ($status,$dac,$i0) = &optimizeDAC($dacPV, $dacStep, $scaler, $countTime, $input,
#                                       [$fileBkg, $delay]);
#-------------------
# Functions for Hytec DAC8402
# (advanced DAC with ramping):
#  9. $dacPV_common = &DAC2common ($dacPV);
# 10.       $status = &setDACslope ($dacPV, $slope);
# 11.        $slope = &getDACslope ($dacPV);
# 12.        $yesno = &DACslopeSupported ($dacPV);
# 13.    $dac_value = &getDACactpos ($dacPV);
# 14.       $status = &waitDAC ($dacPV, [$timeout, $noabort]);
# 15.       $status = &abortDAC ($dacPV);
# 16.       $status = &driveDAC_slope ($dacPV, $rqsPos, $drvL, $drvH, $slope);     #drvL/drvH: 0/0=no check, undef=auto
#                    ----> See also 'SRflyScanDAC' (the *ON-THE-FLY* scan) in scanrec_subs.pl
#--------------------------------------------------------------
# use Pezca;
# do 'dac_subs.pl';

# use warnings::unused;
  use strict;
# no strict 'vars';
# no strict 'subs';
# no strict 'refs';

#=================================================================

# sub setupDAC ($dacPV) {
sub setupDAC ($) {

### This routine sets DAC mode to 'Normal'

  my ($script, $dacPV, $status, $state);

  $script = 'setupDAC';
  $dacPV  = shift(@_);
  $state  = 0;					# 0=Normal, 1=Pulse

# Set the mode:
  $status = Pezca::PutDouble($dacPV.'_pulse', $state);
  &report_exit($script, $status, $dacPV.'_pulse');
  return $status;
}

#=================================================================

# sub readDAC ($dacPV) {
sub readDAC ($) {

### This routine drives DAC to a specified value

  my ($script, $dacPV, $status, $dac_value);

  $script = 'readDAC';
  $dacPV  = shift(@_);

  ($status,$dac_value) = Pezca::GetDouble($dacPV);
  if ($status) {&report_exit($script,$status,$dacPV); $dac_value=0;}

  return $dac_value;
}

#=================================================================

# sub limitsDAC ($dacPV) {
sub limitsDAC ($) {

  our $report;
  my ($script, $dacPV, $status, $drvL, $drvH);

  $script = 'limitsDAC';
  $dacPV  = shift(@_);
  if (!defined $report)   {$report=0;};

  if ($report) {print STDOUT 'Getting DAC Limits [DAC='.$dacPV.']'."\n";}

  ($status,$drvL) = Pezca::GetDouble($dacPV.'.DRVL');
  if ($status) {&report_exit($script,$status,$dacPV.'.DRVL'); $drvL=0;}

  ($status,$drvH) = Pezca::GetDouble($dacPV.'.DRVH');
  if ($status) {&report_exit($script,$status,$dacPV.'.DRVH'); $drvH=0;}

  if ($report) {print STDOUT 'DAC limits: DRVL='.$drvL.' DRVH='.$drvH."\n";}

  return $drvL,$drvH;
}

#=================================================================

# sub driveDAC ($dacPV, $rqsPos, $drvL, $drvH, [$delay, $inlimits) {
sub driveDAC ($$$$;$) {

### This routine drives DAC to a specified value

  my ($script, $dacPV, $rqsPos, $drvL, $drvH, $delay, $inlimits);
  my ($status, $current_time);

  $script   = 'driveDAC';
  $dacPV    = shift(@_);
  $rqsPos   = shift(@_);
  $drvL     = shift(@_);
  $drvH     = shift(@_);
  $delay    = shift(@_);
  $inlimits = shift(@_);
  if (!defined $delay)    {$delay=0.1;}
  if (!defined $inlimits) {$inlimits=0;}

  if (!defined $drvL || !defined $drvH) {
     print STDOUT '!!! '.$script.': no Hi/Lo passed for DAC='.$dacPV.' -- getting auto'."\n";
     ($drvL,$drvH) = &limitsDAC($dacPV);
  }

  if ($drvL != 0 || $drvH != 0) {		#if limits specified
     if ($rqsPos < $drvL || $rqsPos > $drvH) {
        if ($inlimits) {
           if    ($rqsPos < $drvL) {$rqsPos = $drvL;}
           elsif ($rqsPos > $drvH) {$rqsPos = $drvH;}
        } else {
           $current_time = scalar localtime;
           print  STDOUT '*** '.$script.': '.$current_time."\n";
           printf STDOUT '*** '.$script.': '.$dacPV.': rqsPos=%5.3f not in range [%3.2f - %3.2f]'."\n",$rqsPos,$drvL, $drvH;
           return 1;
        }
     }
  }

  $status = Pezca::PutDouble($dacPV, $rqsPos);
  &report_exit($script, $status, $dacPV);
### Wait to settle:
  select(undef,undef,undef,$delay);        # sleep in sec
  return $status;
}

#=================================================================

# sub shiftDAC ($dacPV, $rqsRel, $drvL, $drvH, [$delay, $inlimits]) {
sub shiftDAC ($$$$;$) {

### This routine shifts DAC by a specified value

  my ($script, $dacPV, $rqsRel, $drvL, $drvH, $delay, $inlimits);
  my ($actPos, $rqsPos, $status);

  $script   = 'shiftDAC';
  $dacPV    = shift(@_);
  $rqsRel   = shift(@_);
  $drvL     = shift(@_);
  $drvH     = shift(@_);
  $delay    = shift(@_);
  $inlimits = shift(@_);

  $actPos = &readDAC($dacPV);
  $rqsPos = $actPos + $rqsRel;
  $status = &driveDAC ($dacPV, $rqsPos, $drvL, $drvH, $delay, $inlimits);
  return $status;
}


#=================================================================

# sub stepScanDAC($dacPV, $dacBegin, $dacEnd, $dacStep,
#             $scaler, $countTime, 
#             \@inputs, [\@pvlist, $fileBkg,
#             $gnuplot, $logscale, \@gnuopts,
#             $afterScan, $checkOverload, $gnufile,
#             $postarrayPV, $post_interval,
#             $adc_encdrPV, $dac_tolerance]) {
sub stepScanDAC($$$$$$\@;\@$$$\@$$$$$$$) {
###
### This subroutine implements *STEP* SCAN with DAC.
### Usage: $status = &stepScanDAC(...);
### The sub also returns global arrays @x1DAC, @y1DAC, ... @ynDAC
### +-------------------------------------------------------------+
### | ATTENTION: file scanrec_subs.pl also contains a scan-record |
### |            version (&SRstepScanDAC) of the DAC *STEP SCAN*, |
### |            but it is not very reliable since in few test    |
### |            scans the record was skipping scan points        |
### +-------------------------------------------------------------+
  our ($beamline, $shutter, $shutter_readback, @Offsets);
  our (@x1DAC, $gnuopen);
  our ($DACscan2ndTime, $IGNORE_ERRORS);
  my ($script, $dacPV, $dacBegin, $dacEnd, $dacStep, $scaler, $countTime);
  my ($inputsRef, $pvlistRef, $fileBkg);
  my ($gnuplot, $logscale, $gnuoptRef, $afterScan);
  my ($checkOverload, $threshold, $current, $npts, $npts_last, $status);
  my ($dac_ini, $dacLo, $dacHi, $dac_dest, $dacAct);
  my (@Intensity, $point, $name, $filehandle, $i);
  my ($nx, $xcol, $ycol, $ninputs, $npv, $pvval, $gnufile);
  my ($I_lo, $I_hi, $I_range, $i_plot, $plot_all, $label, $msg);
  my ($postarrayPV, $post_interval, $scan_index, $narr);
  my (@post_array, $time_epoch, $time_passed, $name_plot);
  my ($OVERLOAD, $overload_detected, $adc_offset, $adc_ini);
  my ($FOLLERR, $follerr_detected);
  my ($adc_encdrPV, $dac_tolerance, $tolerance_min);

  $script        = 'stepScanDAC';
  $dacPV         = shift(@_);		#DAC PV to scan
  $dacBegin      = shift(@_);		#begin DAC value
  $dacEnd        = shift(@_);		#end DAC value
  $dacStep       = shift(@_);		#DAC step
  $scaler        = shift(@_);		#scaler PV
  $countTime     = shift(@_);		#scaler count time
  $inputsRef     = shift(@_);		#pointer to the array of scaler inputs
  $pvlistRef     = shift(@_);		#pointer to the array of PVs to record
  $fileBkg       = shift(@_);		#joerger background file ('auto' for automatic, '' for none)
  $gnuplot       = shift(@_);		#1=plot with gnuplot, 0=no, 2=plot and save to image
  $logscale      = shift(@_);		#1=logscale, 0=no
  $gnuoptRef     = shift(@_);		#reference to array of gnu options
  $afterScan     = shift(@_);		#0=stay, 1=startpos, 2=priorpos 3=maxpos 4,14=peakpos 5=minpos 6,16=valleypos 7=gaussian 8=lorentzian
  $checkOverload = shift(@_);		#1=check, 0=ignore
  $gnufile       = shift(@_);		#TMP file for gnuplot data
  $postarrayPV   = shift(@_);		#PV of array into which scan may post progress/final data
  $post_interval = shift(@_);		#post_interval (seconds)
  $adc_encdrPV   = shift(@_);		#ADC PV to test DAC output (optional)
  $dac_tolerance = shift(@_);		#tolerance when testing DAC output (optional; default=step/2)

  if (!defined $fileBkg)        {$fileBkg       = 'auto';}
  if (!defined $gnuplot)        {$gnuplot       = 0;}
  if (!defined $logscale)       {$logscale      = 0;}
  if (!defined $afterScan)      {$afterScan     = 0;}
  if (!defined $checkOverload)  {$checkOverload = 0;}
  if (!defined $gnufile)        {$gnufile       = 'gnu.tmp';}
  if (length($gnufile) == 0)    {$gnufile       = '-';}		#filess plot
  if (!defined $post_interval)  {$post_interval = 2;}		#post interval
  if (!defined $adc_encdrPV)    {$adc_encdrPV   = $dacPV;}	#ADC PV to test DAC output (optional)
  if (!defined $dac_tolerance)  {$dac_tolerance = 0.} 		#ADC channel to test DAC output (optional)

  if (!defined $gnuopen)        {$gnuopen       = 0;}		#global flag
  if (!defined $DACscan2ndTime) {$DACscan2ndTime= 0;}		#global flag
  if (!defined $IGNORE_ERRORS)  {$IGNORE_ERRORS = 0;}		#global flag
  if (defined $shutter) {
     if ($shutter !~ /^${beamline}:/) {$shutter = $beamline.':'.$shutter;}
  }
  if (defined $shutter_readback) {
     if ($shutter_readback !~ /^${beamline}:/) {
                              $shutter_readback = $beamline.':'.$shutter_readback;}
  }

  $ninputs   = @$inputsRef;
  if (defined $pvlistRef) {$npv = @$pvlistRef;} 
  else                    {$npv = 0;}
# $threshold = 0.95E+5;			#threshold intensity for detecting overload (before 2013)
  $threshold = 0.95E+6;			#threshold intensity for detecting overload (2013)
  $OVERLOAD  = 100000;			#scan failure flag when overload detected
  $overload_detected = 0;
  $FOLLERR   = 200000;			#scan failure flag when following error detected
  $tolerance_min = 0.2;			#the minimum sensible DAC error because of different DAC & ADC calibrations
  $scan_index = 0;		        #scan index to save with the post (e.g. 11-vertical, 21-horizontal)
  $follerr_detected = 0;
  $dac_tolerance = abs($dac_tolerance);
  if ($dac_tolerance == 0.) {$dac_tolerance = abs(0.5*$dacStep);}
  if ($dac_tolerance < $tolerance_min) {$dac_tolerance = $tolerance_min;}

  $nx   = 1;
  $xcol = 1;

  if (defined $shutter) {
     print STDOUT 'Closing shutter ['.$shutter.']'."\n";
     &closeShutter ($shutter,$shutter_readback);
  }

  if ($dacStep != 0) {
     if (($dacEnd - $dacBegin)*$dacStep < 0) {$dacStep = - $dacStep;}
     $npts = int(($dacEnd - $dacBegin)/$dacStep + 1.5);
  } else {
     $npts = 1;
  }

  print  STDOUT "\n".
                '# Starting step scan with DAC='.$dacPV;
  if ($adc_encdrPV ne $dacPV) {
     printf STDOUT ',  control ADC='.$adc_encdrPV.', tolerance=%.3f',$dac_tolerance;
  }
  print  STDOUT "\n";

  printf STDOUT '# Begin=%.3f, End=%.3f, Step=%.3f, npts=%d'."\n",$dacBegin,$dacEnd,$dacStep,$npts;
  print  STDOUT '# Scaler='.$scaler.', countTime='.$countTime.'s, inputs='."(@${inputsRef})\n";

  if ($adc_encdrPV ne $dacPV) {
     $status = &addCaMonitor($adc_encdrPV);
     if ($status) {
        print STDOUT '!!! '.$script.': Incorrect ADC PV='.$adc_encdrPV
                           .' Disabling ADC control during DAC scanning'."\n";
        $adc_encdrPV = $dacPV;
     }
  }

### Read the scaler background offsets:
  if (!$DACscan2ndTime) {@Offsets = &getJoergerOffsets ($fileBkg, $scaler);}

### set scaler:
  if (!$DACscan2ndTime) {&setJoerger ($scaler, $countTime);}

### setup DAC (prepare to scan):
  if (!$DACscan2ndTime) {$status=&setupDAC($dacPV); if($status) {return 1;}}

### get current DAC:
  $dac_ini = &readDAC($dacPV);

### Measure DAC-ADC offset:
  $adc_offset = 0.;
  if ($adc_encdrPV ne $dacPV) {
     ($status,$adc_ini) = Pezca::GetDouble($adc_encdrPV);
     if ($status) {&report_only($script, $status, $adc_encdrPV);}
     else {
        $adc_offset=$adc_ini-$dac_ini;
        $msg = sprintf "\n".$script.': DAC=%.2f, ADC(rb)=%.2f, Offset=%.2f'."\n\n",
                                       $dac_ini,$adc_ini,$adc_offset;
        print STDOUT $msg;
        if (abs($adc_offset) > $dac_tolerance) {&tellBluIce ($msg,'err'); $adc_offset = 0.;}
     }
  }

### get DAC limits:
  ($dacLo,$dacHi) = &limitsDAC($dacPV);

### -----------------------------------------------------------------
  if ($gnuplot && !$gnuopen) {
     $gnuopen = &open_gnuplot ($dacBegin, $dacEnd, $logscale, $gnuoptRef);
     if ($gnuopen) {
        if ($gnufile ne '-') {					#if not filess plot
           $filehandle  = open (TMP,'> '.$gnufile);
           if (defined $filehandle) {select TMP; $|=1;}		#set unbuffered output
           else {
              print STDOUT '!!! '.$script.': Failed to open GnuPlot data file "'
                          .$gnufile.'" for writing. Trying filess plot.'."\n";
              $gnufile = '-';
              $plot_all = 0;					#cannot plot multiple columns filessly
           }
        }
        print GNUPLOT 'set xlabel "DAC (Volts)"'."\n";
        print GNUPLOT 'set ylabel "Intensity (cps)"'."\n";
     }
  }

no strict 'refs';
  @x1DAC = ();
  for ($i=1; $i<=$ninputs+$npv; $i++) {		#reset arrays
     $name  = 'y'.$i.'DAC';
     @$name = ();				#here we define global @y1DAC, @y2DAC, etc.
  }
use strict 'refs';
### = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
### Do a generic step scan:
  $I_lo       = +1.E+37;
  $I_hi       = -1.E+37;
  $I_range    = 0.0;
  $i_plot     = 1;				#the main plot = 1st input ($inputs[0])
  $plot_all   = 1;				#plot other inputs using main normalization
  $name_plot  = 'y'.$i_plot.'DAC';
  $label      = 'Intensity';
  $npts_last  = 0;
  $time_epoch = time();				#seconds since epoch

  if (defined $shutter) {
     print STDOUT 'Opening shutter ['.$shutter.']'."\n";
     $status = &openShutter ($shutter,$shutter_readback);
     if ($status) {return $status;}
  }
  no strict 'refs';
  for ($point=0; $point<$npts; $point++) {	#----begin-scan-over-data-points------
     $dac_dest = $dacBegin + $dacStep * $point;
     if ($dacHi != 0 || $dacLo != 0) {
        if ($dac_dest > $dacHi) {$dac_dest = $dacHi;}
        if ($dac_dest < $dacLo) {$dac_dest = $dacLo;}
     }

     &driveDAC ($dacPV, $dac_dest, $dacLo, $dacHi);
     $dacAct = &readDAC ($adc_encdrPV);
     if (abs($dacAct-$dac_dest-$adc_offset) > $dac_tolerance) {
        printf STDOUT '!!! '.$script.': DAC mismatch problem at rqsDAC=%.3f:'
        .' readback=%.3f (rbk-offset=%.3f, err=%.3f, tolerance=%.3f)'."\n",
        $dac_dest,$dacAct,($dacAct-$adc_offset),
        abs($dacAct-$dac_dest-$adc_offset),$dac_tolerance;
        $follerr_detected++;
     }

     printf STDOUT '%4d/%d. DAC=%-8.3f ', $point+1, $npts, $dac_dest;

### Run measurements:
     @Intensity = &runJoerger ($scaler, $countTime, $ninputs, @$inputsRef);
     for ($i=0; $i<$npv; $i++) {
        ($status,$pvval) = Pezca::GetDouble(@$pvlistRef[$i]);
        if ($status) {&report_only($script,$status,@$pvlistRef[$i]); $pvval=0;}
        push @Intensity,$pvval;
     }  
     $current = $Intensity[$i_plot-1];
     if ($current >= $threshold) {
        if ($checkOverload) {
### Aborting measurements and closing plot if overload detected
           print STDOUT '!!! '.$script.': Intensity overload detected. Restoring DAC'."\n\n";
           &driveDAC ($dacPV, $dac_ini, $dacLo, $dacHi);
           if ($gnuopen) {
              &close_gnuplot();
              $gnuopen = 0;
              if ($gnufile ne '-') {close(TMP);}
           }
           use strict 'refs';
           return $OVERLOAD;
        } else {
           $overload_detected++;
        }
     }
     &subtractJoergerOffsets(\@Intensity,$inputsRef,\@Offsets);
     $x1DAC[$point] = $dac_dest;
     for ($i=1; $i<=$ninputs; $i++) {
        $name     = 'y'.$i.'DAC';
        @$name[$point] = $Intensity[$i-1];
        printf STDOUT '  %12.5g', @$name[$point];
        if ($i == $i_plot) {
           if (@$name[$point] < $I_lo) {$I_lo = @$name[$point];}
           if (@$name[$point] > $I_hi) {$I_hi = @$name[$point];}
           $I_range = $I_hi - $I_lo;
        }
     }
     for ($i=$ninputs+1; $i<=$ninputs+$npv; $i++) {
        $name     = 'y'.$i.'DAC';
        @$name[$point] = $Intensity[$i-1];
        printf STDOUT '  %12.5g', @$name[$point];
     }
     print STDOUT "\n";

     if ($gnuopen) {	#- - - - - Begin GnuPlot Business- - - - - - - - - -

        if ($I_hi <= $I_lo) {printf GNUPLOT 'set yrange [%.2g:%.2g]'."\n",$I_hi-1,$I_lo+1;}
        else                {print  GNUPLOT 'set autoscale y'."\n";}

        if ($gnufile ne '-') {
           printf TMP ' %12.5g', $x1DAC[$point];
           for ($i=1; $i<=$ninputs; $i++) {
              $name = 'y'.$i.'DAC';
              printf TMP '  %12.5g', @$name[$point];
           }
           print TMP "\n";
        }
        if ($point > 5) {
### Plot with GnuPlot:
### Syntax:  plot "file" using 1:2, "file" using 1:3, .....
           $ycol = $nx + $i_plot;
           print GNUPLOT 'plot "'.$gnufile.'" using '.$xcol.':'.$ycol
                        .' title "'.$label
                        .' (scaler-'.@$inputsRef[$i_plot-1].')"';
           if ($gnufile eq '-') {
              print GNUPLOT "\n";
              for ($i=0; $i<=$point; $i++) {
                 printf  GNUPLOT '%.5g, %.5g'."\n", $x1DAC[$i], @$name_plot[$point];
                 if ($i == 10*int($i/10)) {&gnuplot_filess_delay();}
              }
              print GNUPLOT 'e'."\n";					#end of data
           }
           elsif ($plot_all) {
              for ($i=1; $i<=$ninputs; $i++) {
                 if ($i != $i_plot) {
                    $ycol = $nx + $i;
                    print GNUPLOT ', "" using '.$xcol.':'.$ycol		#"" means same file
                                  .' title "'.$label
                                  .' (scaler-'.@$inputsRef[$i-1].')"';
                 }
              }
              print GNUPLOT "\n";
           }
           else {
              print GNUPLOT "\n";
           }
        }
     }              	#- - - - - - End GnuPlot Business- - - - - - - - - -

     if (defined $postarrayPV) {  #- - Begin post array bisiness - - - - - -
        $time_passed = time() - $time_epoch;
        if ($time_passed >= $post_interval && $point > $npts_last+1) {
           $time_epoch = time();			#reset seconds since epoch
           $npts_last  = $point+1;
           @post_array = ($scan_index,$npts_last,$dacBegin,$dacEnd,@x1DAC,@$name_plot);
           $narr = @post_array;
#          if ($narr < 1024) {$status = Pezca::PutList($postarrayPV, 'DOUBLE', @post_array);}
           if ($narr < 2048) {$status = Pezca::PutList($postarrayPV, 'FLOAT', @post_array);}
           else {
              print STDOUT '!!! '.$script.': cannot post '.$narr.'-element array into PV='
                          .$postarrayPV.'. Array is too long. Further posting will be disabled.'."\n\n";
              $status = 0;
              undef $postarrayPV;
           }
           if ($status) {
              &report_only($script, $status, $postarrayPV);
              print STDOUT '!!! '.$script.': error posting '.$narr.'-element array into PV='
                          .$postarrayPV.'. Further posting will be disabled.'."\n\n";
              undef $postarrayPV;
           }
        }
     }                            #- - - End post array bisiness - - - - - -

  }                                             #------end-scan-over-data-points------
  if (defined $shutter) {
     print STDOUT 'Closing shutter ['.$shutter.']'."\n";
     &closeShutter ($shutter,$shutter_readback);
  }
### = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
  if (defined $postarrayPV && $npts_last < $npts) {
     @post_array = ($scan_index,$npts,$dacBegin,$dacEnd,@x1DAC,@$name_plot);
     $narr = @post_array;
#    if ($narr < 1024) {$status = Pezca::PutList($postarrayPV, 'DOUBLE', @post_array);}
     if ($narr < 2048) {$status = Pezca::PutList($postarrayPV, 'FLOAT', @post_array);}
     else {
        print STDOUT '!!! '.$script.': cannot post '.$narr.'-element array into PV='
                    .$postarrayPV.'. Array is too long. Further posting will be disabled.'."\n\n";
        $status = 0;
        undef $postarrayPV;
     }
     if ($status) {
        &report_only($script, $status, $postarrayPV);
        print STDOUT '!!! '.$script.': error posting '.$narr
                    .'-element array into PV='.$postarrayPV."\n\n";
        undef $postarrayPV;
     }
  }
### = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
  if ($gnuopen) {
     if ($gnufile ne '-') {close(TMP);}
#    if ($gnufile ne '-') {&print_gnuplot();} else {&close_gnuplot();}
     if ($gnuplot == 1) {&close_gnuplot(); $gnuopen=0;}
  }

  if ($follerr_detected > 0) {
     $msg = sprintf "\n".'!!! '.$script.': DAC following error detected %d times of %d.'."\n"
                        .'DAC=%s ADC(readback)=%s Tolerance=%.2fV Offset=%.2fV'."\n\n",
                        $follerr_detected,$npts,$dacPV,$adc_encdrPV,$dac_tolerance,$adc_offset;
     print STDOUT $msg;
     &tellBluIce ($msg,'warn');
     if ($follerr_detected > 0.1*$npts) {
        $msg = '!!! '.$script.':  Rescan needed. Restoring DAC.'."\n\n";
        print STDOUT $msg;
        &tellBluIce ($msg,'warn');
        $dac_dest = $dac_ini;
        &driveDAC ($dacPV, $dac_dest, $dacLo, $dacHi);
        return $FOLLERR;
     } else {
        $msg = '!!! '.$script.':  ignoring following errors as their rate is under 10%'."\n\n";
        print STDOUT $msg;
        &tellBluIce ($msg,'warn');
     }
  }
### = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
### afterScan:
  &DAC_afterScan($afterScan,\@x1DAC,\@$name_plot,$dacBegin,$dacEnd,$dac_ini,$dacPV);

  print STDOUT '..DAC scan finished!'."\n";
  use strict 'refs';
  if ($overload_detected > 0) {
     print STDOUT '!!! '.$script.': Intensity overload detected '.$overload_detected
                                                                  .' times.'."\n\n";
     return $OVERLOAD;
  }
  return 0;
}

#=================================================================

#sub DAC_afterScan($afterScan,\@xarr,\@yarr,$begin,$end,$dac_ini,$dacPV) {
sub DAC_afterScan($\@\@$$$$) {
### -----------------------------------------------------
### afterScan operations for &SRflyScanDAC (scanrec_subs.pl) and &stepScanDAC (dac_subs.pl)
### -----------------------------------------------------

  my ($script, $afterScan, $xname, $yname, $begin, $end, $dac_ini, $dacPV);
  my ($npts, $dac_dest, $dacLo, $dacHi, $I_range, $I_noise);
  my ($yesno, $msg, $status, $searchType, $discriminator);
  my ($y_max, $y_min, $pt_max, $pt_min, $pt_FWHM);
  my ($pt_left, $pt_right, $x_left, $x_right, $fit);
  my ($ybase, $yscale, $xwidth, $wfactor, $x_peak, $x_valley);

  $afterScan = shift(@_);  if (!defined $afterScan) {$afterScan=0;}     #valid options 0-11,14,15,17,18,110,111
  $xname     = shift(@_);
  $yname     = shift(@_);
  $begin     = shift(@_);
  $end       = shift(@_);
  $dac_ini   = shift(@_);
  $dacPV     = shift(@_);

  $script         = 'DAC_afterScan';
  $I_noise        = 4.;                         		        #noise intensity variation threshold
  $npts           = @$yname;
  ($dacLo,$dacHi) = &limitsDAC ($dacPV);

  ($pt_min,$y_min,$pt_max,$y_max) = &arrayStatsRef(\@$yname);

  $I_range = $y_max-$y_min;

  if ($I_range <= $I_noise && $afterScan >= 3) {
     printf STDOUT '!!! '.$script.': noise-level ('.$I_range.'cts) intensity variation during the scan.'."\n"
                  .'!!! Driving DAC to initial coordinate=%.3f'."\n",$dac_ini;
     $afterScan = 2;
  }
  
# $afterScan: 0=stay, 1=startpos, 2=priorpos 
#             3=maxpos 4,14=peakpos(noisy), 5,15=peakpos(smooth)
#             6=minpos 7,17=valleypos(noisy), 8,18=valleypos(smooth)
#             9=centroid, 10,110=gaussian 11,111=lorentzian
  $yesno = 1;
  if    ($afterScan == 14)  {$afterScan=4;  undef $yesno;}
  elsif ($afterScan == 15)  {$afterScan=5;  undef $yesno;}
  elsif ($afterScan == 17)  {$afterScan=7;  undef $yesno;}
  elsif ($afterScan == 18)  {$afterScan=8;  undef $yesno;}
  elsif ($afterScan == 110) {$afterScan=10; undef $yesno;}
  elsif ($afterScan == 111) {$afterScan=11; undef $yesno;}

  if ($afterScan == 10 || $afterScan == 11) {                           #Gaussian or Lerentzian fit
     if (!defined &peakFit) {
        print STDOUT '!!! '.$script.': no peakFit function available.'."\n"
                    .'!!! Perhaps common_PDL.pl is not imported.'."\n"
                    .'!!! Reverting to afterScan="Noisy Peak".'."\n";
        $afterScan = 4;
     }
  }

  if    ($afterScan == 0) {				                #0=stay,
     $dac_dest = $end;
     printf STDOUT $script.': leaving DAC at the end of scan coordinate=%.3f'."\n",$dac_dest;
     sleep (2);
  }
  elsif ($afterScan == 1) {				                #1=startpos (begin of scan interval)
     $dac_dest = $begin;
     printf STDOUT $script.': driving DAC to the beginning of scan coordinate=%.3f'."\n",$dac_dest;
     &driveDAC ($dacPV,$dac_dest,$dacLo,$dacHi);
  }
  elsif ($afterScan == 2) {				                #2=priorpos (initial DAC value before scan)
     $dac_dest = $dac_ini;
     printf STDOUT $script.': driving DAC to the initial coordinate=%.3f'."\n",$dac_dest;
     &driveDAC ($dacPV,$dac_dest,$dacLo,$dacHi);
  }
  elsif ($afterScan == 3) {				                #3=maxpos
     $dac_dest = @$xname[$pt_max];
     printf STDOUT $script.': driving DAC to the I0-max coordinate=%.3f'."\n",$dac_dest;
     &driveDAC ($dacPV,$dac_dest,$dacLo,$dacHi);
  }
  elsif ($afterScan == 4 || $afterScan == 5) {		                #4=peakpos(noisy),5=peakpos(smooth)
     if    ($afterScan == 4) {$searchType='noisy';}                     #search half-peak from tails
     elsif ($afterScan == 5) {$searchType='smooth';}                    #search half-peak from max.intensity
     ($y_max,$y_min,$pt_max,$pt_min,$pt_FWHM,
      $pt_left,$pt_right,$x_left,$x_right,$x_peak) = &findPeakRef($npts,\@$yname,'auto',$searchType);
     if ($pt_FWHM > 0) {
        $dac_dest = 0.5*(&pt2coord($x_left,\@$xname) + &pt2coord($x_right,\@$xname));
        printf STDOUT $script.': driving DAC to '.$searchType.' I0-peak coordinate=%.3f'."\n",$dac_dest;
     } else {
        if (!defined $yesno) {
           $msg = sprintf 'No '.$searchType.' I0-peak found during DAC scan. Do you want'."\n"
                         .'to drive DAC to the maximum I0 coordinate=%.3f ?',@$xname[$pt_max];
           $yesno = &zenity_question($msg);
        }
        print STDOUT '!!! '.$script.': no '.$searchType.' I0-peak found. '."\n";
        if ($yesno) {
           $dac_dest = @$xname[$pt_max];
           printf STDOUT '!!! '.$script.': Driving DAC to the I0-max coordinate=%.3f'."\n",$dac_dest;
        } else {
           $dac_dest = $dac_ini;
           printf STDOUT '!!! '.$script.': Driving DAC to the initial coordinate=%.3f'."\n",$dac_dest;
        }
     }
     &driveDAC ($dacPV,$dac_dest,$dacLo,$dacHi);
  }
  elsif ($afterScan == 6) {				                #6=minpos
     $dac_dest = @$xname[$pt_min];
     printf STDOUT $script.': driving DAC to the I0-min coordinate=%.3f'."\n",$dac_dest;
     &driveDAC ($dacPV,$dac_dest,$dacLo,$dacHi);
  }
  elsif ($afterScan == 7 || $afterScan == 8) {		                #7=valleypos(noisy),8=valleypos(smooth)
     if    ($afterScan == 7) {$searchType='noisy';}                     #search half-valey from tails
     elsif ($afterScan == 8) {$searchType='smooth';}                    #search half-valey from min.intensity
     ($y_max,$y_min,$pt_max,$pt_min,$pt_FWHM,
      $pt_left,$pt_right,$x_left,$x_right,$x_valley) = &findValleyRef($npts,\@$yname,'auto',$searchType);
     if ($pt_FWHM > 0) {
        $dac_dest = 0.5*(&pt2coord($x_left,\@$xname) + &pt2coord($x_right,\@$xname));
        printf STDOUT $script.': driving DAC to '.$searchType.' I0-valley coordinate=%.3f'."\n",$dac_dest;
     } else {
        if (!defined $yesno) {
           $msg = sprintf 'No '.$searchType.' I0-valley found during DAC scan. Do you want'."\n"
                         .'to drive DAC to the minimum I0 coordinate=%.3f ?',@$xname[$pt_min];
           $yesno = &zenity_question($msg);
        }
        print STDOUT '!!! '.$script.': no '.$searchType.' I0-valley found. '."\n";
        if ($yesno) {
           $dac_dest = @$xname[$pt_min];
           printf STDOUT '!!! '.$script.': Driving DAC to the I0-min coordinate=%.3f'."\n",$dac_dest;
        } else {
           $dac_dest = $dac_ini;
           printf STDOUT '!!! '.$script.': Driving DAC to the initial coordinate=%.3f'."\n",$dac_dest;
        }
     }
     &driveDAC ($dacPV,$dac_dest,$dacLo,$dacHi);
  }
  elsif ($afterScan == 9) {             		                #9=centroid
     $discriminator = 0.1;                                              #only account points with intensity > 10% of max
     $dac_dest = &findCentroidRef($npts,\@$xname,\@$yname,$discriminator);
     printf STDOUT $script.': driving DAC to the I0-centroid coordinate=%.3f'."\n",$dac_dest;
     &driveDAC ($dacPV,$dac_dest,$dacLo,$dacHi);
  }
  elsif ($afterScan == 10 || $afterScan == 11) {             		#10=Gaussian 11=Lorentzian ask
     ### Gaussian fit is:   $y=$ybase+$yscale*exp(-0.5*(($x-$x_peak)/$xwidth)**2)
     ### Lorentzian fit is: $y=$ybase+$yscale/(1+((x-$x_peak)/$x_width)**2) 
     if    ($afterScan == 10) {$fit = 'Gaussian';   $wfactor=2.35482;}  #$FWHM=$xwidth*(2*sqrt(2*ln(2)))
     elsif ($afterScan == 11) {$fit = 'Lorentzian'; $wfactor=2.0;}      #$FWHM=$xwidth*2
     ($status,$ybase,$yscale,$x_peak,$xwidth) = &peakFit(\@$xname, \@$yname, $fit);
     if ($status == 0) {
        printf STDOUT $script.': '.$fit.' fit xcenter=%.3f  xwidth=%.3f  FWHM=%.3f'."\n",$x_peak,$xwidth,$xwidth*$wfactor;
        $dac_dest = $x_peak;
        printf STDOUT $script.': driving DAC to the '.$fit.' fit of I0-peak coordinate=%.3f'."\n",$dac_dest;
     } else {
        print STDOUT '*** '.$script.': '.$fit.' fitting failed with status='.$status."\n";
        if (!defined $yesno) {
           $msg = sprintf 'No '.$fit.' I0-peak found during DAC scan. Do you want'."\n"
                         .'to drive DAC to the maximum I0 coordinate=%.3f ?',@$xname[$pt_max];
           $yesno = &zenity_question($msg);
        }
        print STDOUT '!!! '.$script.': no '.$fit.' I0-peak found. '."\n";
        if ($yesno) {
           $dac_dest = @$xname[$pt_max];
           printf STDOUT '!!! '.$script.': Driving DAC to the I0-max coordinate=%.3f'."\n",$dac_dest;
        } else {
           $dac_dest = $dac_ini;
           printf STDOUT '!!! '.$script.': Driving DAC to the initial coordinate=%.3f'."\n",$dac_dest;
        }
     }
     &driveDAC ($dacPV,$dac_dest,$dacLo,$dacHi);
  }
  else {
     print STDOUT '!!! '.$script.': invalid option value afterScan='.$afterScan."\n";
  }
  return;
}

#=================================================================

#  ($status,$dac,$i0) = &optimizeDAC($dacPV, $dacStep, $scaler, $countTime, $input,
#                                    [$fileBkg, $delay]);
sub optimizeDAC($$$$$;$$) {
###
### This subroutine tweaks DAC from the current position to find the maximum $i0.
  our ($beamline, $shutter, $shutter_readback, @Offsets);
  our ($DACscan2ndTime, $IGNORE_ERRORS);
  my ($script, $dacPV, $dacStep, $scaler, $countTime, $input, $fileBkg, $delay);
  my ($status, $dac, $i0, $dac_ini, $dacLo, $dacHi, @Intensity, $off);
  my ($i0_ini, $i0_1, $i0_2, $direction, $n);

  $script     = 'optimizeDAC';
  $dacPV      = shift(@_);		#DAC PV to scan
  $dacStep    = shift(@_);		#DAC step
  $scaler     = shift(@_);		#scaler PV
  $countTime  = shift(@_);		#scaler count time
  $input      = shift(@_);		#pointer to the array of scaler inputs
  $fileBkg    = shift(@_);		#joerger background file ('auto' for automatic, '' for none)
  $delay      = shift(@_);		#delay after changing DAC and before running scaler

  if (!defined $fileBkg)        {$fileBkg       = 'auto';}
  if (!defined $delay)          {$delay         = 0;}

  if (!defined $DACscan2ndTime) {$DACscan2ndTime= 0;}		#global flag
  if (!defined $IGNORE_ERRORS)  {$IGNORE_ERRORS = 0;}		#global flag
  if (defined $shutter) {
     if ($shutter !~ /^${beamline}:/) {$shutter = $beamline.':'.$shutter;}
  }
  if (defined $shutter_readback) {
     if ($shutter_readback !~ /^${beamline}:/) {
                              $shutter_readback = $beamline.':'.$shutter_readback;}
  }

  if (defined $shutter) {&closeShutter ($shutter,$shutter_readback);}

  print  STDOUT "\n".$script.': Tweaking DAC='.$dacPV
         .' to max intensity on scaler='.$scaler.' input='.$input."\n";
  printf STDOUT $script.': DAC step=%.3f'."\n",$dacStep;
  print  STDOUT $script.': Scaler='.$scaler.', countTime='
         .$countTime.'s, input='.$input."\n";

  if (!$DACscan2ndTime) {
     ### Read the scaler background offsets:
     @Offsets = &getJoergerOffsets ($fileBkg, $scaler);
     ### set scaler:
     &setJoerger ($scaler, $countTime);
     ### setup DAC (prepare to scan):
     $status=&setupDAC($dacPV); if($status) {return -1;}
  }
  $off = $Offsets[$input-1];

### get current DAC:
  $dac_ini = &readDAC($dacPV);

### get DAC limits:
  ($dacLo,$dacHi) = &limitsDAC($dacPV);

  if (defined $shutter) {
     $status = &openShutter ($shutter,$shutter_readback);
     if ($status) {return $status;}
  }

### get current intensity:
  @Intensity = &runJoerger ($scaler, $countTime, 1, ($input));
  $i0_ini = $Intensity[0] - $off;
  printf STDOUT $script.': DAC='.$dacPV.' I0[0]=%.2f'."\n",$i0_ini;

### = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
### Find the direction to go:

  $dacStep = abs($dacStep);
  $dac = $dac_ini - $dacStep;
  if ($dacHi != 0 || $dacLo != 0) {
     if ($dac < $dacLo) {undef $dac;}
  }
  if (defined $dac) {
     &driveDAC ($dacPV, $dac, $dacLo, $dacHi);
     if ($delay > 0) {select(undef,undef,undef,$delay);}			#sleep in sec
     @Intensity = &runJoerger ($scaler, $countTime, 1, ($input));
     $i0_1 = $Intensity[0] - $off;
     printf STDOUT $script.': DAC='.$dacPV.' I0[-1]=%.2f'."\n",$i0_1;
  } else {
     $i0_1 = -100000000;
  }

  $dac = $dac_ini + $dacStep;
  if ($dacHi != 0 || $dacLo != 0) {
     if ($dac > $dacHi) {undef $dac;}
  }
  if (defined $dac) {
     &driveDAC ($dacPV, $dac, $dacLo, $dacHi);
     if ($delay > 0) {select(undef,undef,undef,$delay);}			#sleep in sec
     @Intensity = &runJoerger ($scaler, $countTime, 1, ($input));
     $i0_2 = $Intensity[0] - $off;
     printf STDOUT $script.': DAC='.$dacPV.' I0[+1]=%.2f'."\n",$i0_2;
  } else {
     $i0_2 = -100000000;
  }

  if    ($i0_2 > $i0_ini && $i0_ini > $i0_1) {$direction = 1; $i0=$i0_2}	#positive
  elsif ($i0_1 > $i0_ini && $i0_ini > $i0_2) {$direction =-1; $i0=$i0_1}	#negative
  else {
     if (defined $shutter) {&closeShutter ($shutter,$shutter_readback);}
     printf STDOUT $script.': No optimization required for DAC='.$dacPV."\n";
     &driveDAC ($dacPV, $dac_ini, $dacLo, $dacHi);
     return 0,$dac_ini,$i0_ini;
  }

### Run measurements:
  $n = 1;
  do {
     $n++;
     $dac = $dac_ini + $n*$dacStep*$direction;
     if ($dacHi != 0 || $dacLo != 0) {
        if ($dac > $dacHi) {undef $dac;}
        if ($dac < $dacLo) {undef $dac;}
     }
     if (defined $dac) {
        &driveDAC ($dacPV, $dac, $dacLo, $dacHi);
        if ($delay > 0) {select(undef,undef,undef,$delay);}			#sleep in sec
        @Intensity = &runJoerger ($scaler, $countTime, 1, ($input));
        $i0_1 = $i0;
        $i0 = $Intensity[0] - $off;
        printf STDOUT $script.': DAC='.$dacPV.' I0[%d]=%.2f'."\n",$n*$direction,$i0;
     } else {
        if (defined $shutter) {&closeShutter ($shutter,$shutter_readback);}
        printf STDOUT '!!! '.$script.': optimization for DAC='.$dacPV
                     .' brought it to limit.'."\n";
        $dac = $dac_ini + ($n-1)*$dacStep*$direction;
        return 0,$dac,$i0;
     }
  } while ($i0 > $i0_1);							#while intensity is growing
  if (defined $shutter) {&closeShutter ($shutter,$shutter_readback);}
  $n--;
  $dac = $dac_ini + $n*$dacStep*$direction;
  &driveDAC ($dacPV, $dac, $dacLo, $dacHi);
  printf STDOUT $script.': DAC='.$dacPV.' corrected from %.4f to %.4f (by %d steps)'."\n", $dac_ini,$dac,$n;
  return 0,$dac,$i0_1;
}

#=================================================================

# sub DAC2common ($dacPV) {
sub DAC2common ($) {

### This routine parses common DAC PV based on given DAC channel:

  my ($script, $dacPV, $dacPV_common);

  $script = 'DAC2common';
  $dacPV  = shift(@_);

  $dacPV_common = $dacPV;
  $dacPV_common =~ s/_[0-9]*$//;		#remove channel number: 23i:DAC2_14 -> 23i:DAC2
  return $dacPV_common;
}


#=================================================================

# sub setDACslope ($dacPV, $slope) {
sub setDACslope ($$) {

### This routine sets DAC scanning slope (V/s)

  my ($script, $dacPV, $status, $slope, $delay);

  $script = 'setDACslope';
  $dacPV  = shift(@_);
  $slope  = shift(@_);
  $delay  = 0.1;

  if ($slope < 0) {return -1;}

# $dacPV =~ s/_[0-9]*$//;					#remove channel number: 23i:DAC2_14 -> 23i:DAC2
### Set the ramp time:
  $status = Pezca::PutDouble($dacPV.'_wfSlope', $slope);
  &report_exit($script, $status, $dacPV.'_wfSlope');
  select(undef,undef,undef,$delay);			# sleep in sec
  return $status;
}

#=================================================================

# sub getDACslope ($dacPV) {
sub getDACslope ($) {

### This routine reads DAC scanning slope (V/s)

  my ($script, $dacPV, $status, $slope);

  $script = 'getDACslope';
  $dacPV  = shift(@_);

# $dacPV =~ s/_[0-9]*$//;			#remove channel number: 23i:DAC2_14 -> 23i:DAC2
### Get the ramp time:
  ($status,$slope) = Pezca::GetDouble($dacPV.'_wfSlope');
  if ($status) {&report_only($script,$status,$dacPV.'_wfSlope'); $slope=0;}

  return $slope;
}

#=================================================================

# sub DACslopeSupported ($dacPV) {
sub DACslopeSupported ($) {

### This routine checks if DAC supports scanning slope (V/s)

  my ($script, $dacPV, $status, $slope);

  $script = 'DACslopeSuppoered';
  $dacPV  = shift(@_);

# $dacPV =~ s/_[0-9]*$//;				#remove channel number: 23i:DAC2_14 -> 23i:DAC2
  &Pezca::AutoErrorMessageOff();			#start suppressing error messages
  ($status,$slope) = Pezca::GetDouble($dacPV.'_wfSlope');
  &Pezca::AutoErrorMessageOn();				#finish suppressing error messages
  if (! $status) {
     print STDOUT $script.': DAC='.$dacPV.' supports sloping.'."\n";
     return 1;						#PV responds, which means slope supported
  } else {
     print STDOUT $script.': DAC='.$dacPV.' does not support sloping.'."\n";
     return 0;						#PV access error, which means slope is not supported
  }
}

#=================================================================

# sub getDACactpos ($dacPV) {
sub getDACactpos ($) {

### This routine reads DAC actual value:

  my ($script, $dacPV, $status, $dac_value);

  $script = 'getDACactpos';
  $dacPV  = shift(@_);

### Get the actual value:
  ($status,$dac_value) = Pezca::GetDouble($dacPV.'rb.VAL');
  if ($status) {&report_exit($script,$status,$dacPV.'rb.VAL'); $dac_value=0;}

  return $dac_value;
}

#=================================================================

# sub waitDAC ($dacPV, [$timeout, $noabort]) {
sub waitDAC ($;$$) {

### This routine  DAC ramping:

  my ($script, $dacPV, $dacPV_common, $status, $timeout, $noabort);
  my ($wait_, $wait_step, $scanning, $current_time);

  $script  = 'waitDAC';
  $dacPV   = shift(@_);
  $timeout = shift(@_);
  $noabort = shift(@_);
  if    (!defined $timeout) {$timeout = 30;}
  elsif ($timeout <= 0.0)   {$timeout = 30;}

  $wait_step = 0.1;
  $wait_     = 0.0;
  $scanning  = 1;
  $status    = 0;

  $dacPV_common = &DAC2common($dacPV);

  $status = &addCaMonitor($dacPV_common.'_wfScan');

  while ($scanning && $wait_ < $timeout) {
    select(undef,undef,undef,$wait_step);	# sleep($wait_step);
    $wait_ += $wait_step;
    ($status,$scanning) = Pezca::GetDouble($dacPV_common.'_wfScan');
  }

  if ($wait_ >= $timeout) {
    $current_time = scalar localtime;
    print  STDOUT "\n";
    print  STDOUT '*** '.$current_time."\n";
    printf STDOUT '*** '.$script.': DAC='.$dacPV.' scanning did not complete after %2.0fs'."\n",$timeout;
    $status++;
    ### $noabort must be defined when called from $abortDAC because
    ### otherwise we arrange infinite abortDAT-waitDAC loop:
    if (!defined $noabort) {
       $status += &abortDAC ($dacPV);
    }
  }

  return $status;
}

#=================================================================

# sub abortDAC ($dacPV) {
sub abortDAC ($) {

### This routine abort DAC scanning:

  my ($script, $dacPV, $dacPV_common, $status, $delay);
  my ($timeout, $noabort);

  $script = 'abortDAC';
  $dacPV  = shift(@_);
  $delay  = 0.1;
  $status = 0;

  $dacPV_common = &DAC2common($dacPV);

### Abort DAC scanning:
  $status += Pezca::PutDoubleOldCa($dacPV_common.'_wfAbort', 1);
  &report_exit($script, $status, $dacPV_common.'_wfAbort');
  select(undef,undef,undef,$delay);        # sleep in sec

### Wait DAC to finish:
### (tell waitDAC that it should not send its own abort
### because it would make infinite abortDAT-waitDAC loop): 
  ($timeout, $noabort) = (5, 1);
  $status += &waitDAC($dacPV_common, $timeout, $noabort);

### Set scan slope to 0:
  $status += &setDACslope ($dacPV, 0.0);

### Read actual DAC and set requested DAC equal to actual:
# $dac_value = &getDACactpos ($dacPV);
# $status += &driveDAC ($dacPV, $dac_value, 0.0, 0.0, $delay);

  return $status;
}

#=================================================================

# sub driveDAC_slope ($dacPV, $rqsPos, $drvL, $drvH, $slope) {
sub driveDAC_slope ($$$$$) {

### This routine drives DAC to a specified value with given speed

  my ($script, $dacPV, $rqsPos, $drvL, $drvH, $slope);
  my ($status, $current_time, $delay, $timeout, $startPos);

  $script = 'driveDAC_slope';
  $dacPV  = shift(@_);
  $rqsPos = shift(@_);
  $drvL   = shift(@_);
  $drvH   = shift(@_);
  $slope  = shift(@_);

  if (!defined $drvL || !defined $drvH) {
    print STDOUT '!!! '.$script.': no Hi/Lo passed for DAC='.$dacPV.' -- getting auto'."\n";
    ($drvL,$drvH) = &limitsDAC($dacPV);
  }

  if ($drvL != 0 || $drvH != 0) {		#if limits specified
    if ($rqsPos < $drvL || $rqsPos > $drvH) {
      $current_time = scalar localtime;
      print  STDOUT '*** '.$script.': '.$current_time."\n";
      printf STDOUT '*** '.$script.': '.$dacPV.': rqsPos=%5.3f not in range [%3.2f - %3.2f]'."\n",$rqsPos,$drvL, $drvH;
      return 1;
    }
  }

  if ($slope != 0) {
     if (! &DACslopeSupported($dacPV)) {$slope = 0;}
  }
  if ($slope == 0) {
     $delay = 1;
     $status = &driveDAC ($dacPV, $rqsPos, $drvL, $drvH, $delay); 
  } else {
     $startPos = &readDAC ($dacPV);
     $timeout = abs($rqsPos - $startPos)/$slope + 3.;
     $status = Pezca::PutDouble($dacPV, $rqsPos);
     if ($status) {
        &report_only($script, $status, $dacPV);
        print STDOUT '*** '.$script.': starting DAC='.$dacPV.' slopping failed with status='.$status.' Aborting...'."\n";
        &abortDAC ($dacPV);
     } else {
### Wait to finish:
        $status = &waitDAC ($dacPV, $timeout);
        if ($status) {
           print STDOUT '*** '.$script.': waiting for DAC='.$dacPV.' slopping failed with status='.$status.' Aborting...'."\n";
           &abortDAC ($dacPV);
        }
     }
  }
  return $status;
}
