#!/usr/bin/perl -W
#-----------------------------------------------------------------------
# These are Fluorescent Detector (MCA + ADC) related subs for beamline
# controls Perl scripts. These functions were originally developed for
# Canberra MCA/ADC/TCA (NIM modules, EPICS driver & DB by Mark Rivers).
#
#     		    Author: Sergey Stepanov
#
# Version-1.0:  2001/06/22 using Pezca for Canberra MCA/ADC/TCA
# Version-2.0:  2006/12/08 changed MCA calibration to data from EPICS
# Version-3.0:  2008/12/05 added functions to read edge, flu peak and scan
#                          range data from mySQL (shared with BluIce)
# Version-4.0:  2019/10/29 modified most of functions to also work with
#                          Falcon DXP (EPICS driver & DB by Mark Rivers)
#-----------------------------------------------------------------------
# List of functions:
# ^^^^^^^^^^^^^^^^^
#  1.                               $mcaNMAX = &MCAnmax($mca);
#  2.                                          &setMCA($mca,$count_time);
#  3.                                          &setMCA_fastest($mca,$count_time);
#  4.                     ($roi_lo, $roi_hi) = &checkMCA_ROI($mca,[$roi_Num]);
#  5.                                          &setMCA_ROI($mca,$roi_lo,$roi_hi,[$roi_Num]);
#  6.                                          &startMCA($mca);						#uses global $EraseMCA
#  7.                                          &abortMCA($mca);
#  8.                                          &eraseMCA($mca);
#  9.                                          &waitMCA($mca);
# 10.                                $status = &wait_readingMCA($mca);					#stop MCA, initiate reading from device into record and wait it to complete
# 11.                  ($num_pts,@Intensity) = &readMCA($mca);
# 12.                                $status = &readMCA_ref($mca,\@Intensity);				#fills-in intensity
# 13.                      ($total,$num_pts) = &readMCA_total($mca);
# 14.                  ($sum_data,$net_data) = &readMCA_ROI($mca,[$roi_Num]);
# 15.                             $dead_time = &readMCA_deadTime($mca);
# 16.                             $dead_time = &readMCA_deadTime_from_live($mca);
#------------ADC and TCA are for Canberra only:
# 17.                                          &setupMCA_ADC($adc,[$gain,$offset,$low]);
# 18.                                          &setMCA_ADC($adc,$gain,$offset);
# 19.                                          &setMCA_ADC_discr($adc,$low,$high);
# 20.                        ($gain,$offset) = &getMCA_ADC($adc);
# 21.                           ($low,$high) = &getMCA_ADC_discr($adc,$low,$high);
# 22.                                          &setupMCA_TCA($tca,$low);
# 23.                                          &setMCA_TCA($tca,$low,$high,[$scaNo]);
# 24.                           ($low,$high) = &getMCA_TCA($tca,[$scaNo]);
# 25.                                          &getDerekMCAch2en($adc); 				#defines global ($derekSlope, $derekOffset) [channel -> eV]
# 26.                   ($E_scale,$E_offset) = &getMCAcalib($adc,[$gain,$offset]); 			#in eV; if ($gain,$offset) not passed, they are read from ADC
# 27.                               $E_scale = &MCAcalibDB2006($gain);					#in eV
# 28.                                 $scale = &scaleMCA($mcaNMAX,$gain);				#in eV ('$scale' is E-range over all MCA channels)
# 29.					  $E = &mcaRoi2energy($roi,$adc,[$E_scale,$E_offset]);		#recalculates MCA channel to eV
# 30.					$roi = &mcaEnergy2roi($E,$adc,[$E_scale,$E_offset]);		#recalculates energy in eV to MCA channel
# 31.                             ($N1, $N2) = &optimizeMCA_ADC($E1,$E2,$mcaNMAX,$adc,[$tooWideFlag]);	#$E1,$E2 are in eV; toowide: 0=error,1=adjust_E1,2=$adjust_E2
# 32.                                          &printMCA($LUN,$num_pts,@Intensity);
# -----------2008, fast scans bypassing MCA (Canberra & Falcon; getDerekMCAroi2dis used with Canberra only):
# 33.                                          &getDerekMCAroi2dis($adc);				#defines global ($derekR2D_slope, $derekR2D_offlo, $derekR2D_offhi);
# 34.                           ($low,$high) = &calcDiscrFromROI($mca,$adc);				#for fast scans with Struck
# 35.                                          &setDiscrFromROI($mca,$adc,$tca);			#for fast scans with Struck
# -----------2008, mySQL:
# 36.                                $status = &getEdgeDataSQL($element,$edge,
#                                                    \$EdgeEnergy,\@EnStep,\@EnPM,			#all in eV
#                                                    \$FluoEnergy,\$FluoPMwidth);			#all in eV
# 37.                                $nEdges = &getEdgeListSQL(\@ElementEdge,\@EdgeEnergy);		#in eV
# 38.                                $nLines = &getFluoListSQL(\@ElementLine,\@FluoEnergy);		#in eV
# 39.                                $status = &checkFalcon2BluiceConsistency();
# -----------2020, Falcon-specific:
# 40.                   ($E_scale,$E_offset) = &getFalconCalib($mca);					#in eV
# 41.                             ($N1, $N2) = &setFalconROI($mca,$E1,$E2,[$N,$E_scale,$E_offset]);	#in eV
# 42.                                $status = &setFalconSCA($mca,$sca_lo,$sca_hi,[$N]);
# 43.                                $status = &setFalconROI2SCA($mca);
# 44.                                     $E = &FalconRoi2energy($roi,$mca,[$E_scale,$E_offset]);	#in eV
# -----------2020, MEDM:
# 45.                                          &openMedmCanberraMCA ($mca);
# 46.                                          &openMedmCanberraADC ($adc);
# 47.                                          &openMedmCanberraTCA ($tca);
#--------------------------------------------------------------
# use Pezca;
# use DBI; 			# mySQL interface (for getting edge & scanrange data)
# do 'fluo_subs.pl';

# use warnings::unused;
  use strict;
# no strict 'vars';
# no strict 'subs';
# no strict 'refs';

#=================================================================

# sub MCAnmax ($mca) {
sub MCAnmax ($) {
### This function returns maximum number of channels in MCA:

  my ($script, $mca, $status, $mcaNMAX, $pv,$mcapv);
  $script = 'MCAnmax';
  $mca    = shift(@_);

  if ($mca !~ /falcon/) {
     $mca =~ s/:$//;					#mca should not be ended by ":"
     $mcapv = $mca;
  } else {
     if ($mca !~ /:$/) {$mca.=':';}			#falcon must be ended by ":"
     $mcapv = $mca.'mca1';
  }
  
### Read the maximum possible number of channels:
  $pv = $mcapv.'.NMAX';
  ($status, $mcaNMAX) = Pezca::GetDouble($pv);
  if ($status) {&report_exit($script, $status, $pv); $mcaNMAX=-1;}

  return $mcaNMAX;
}

#=================================================================

# sub setMCA ($mca, $count_time) {
sub setMCA ($$) {

  my ($script, $mca, $count_time, $status, $mcaNMAX);
  my ($pv, $mcapv, $state);
  $script     = 'setMCA';
  $mca        = shift(@_);
  $count_time = shift(@_);
  $status     = 0;

  ### Read the maximum possible number of channels:
  $mcaNMAX = &MCAnmax($mca);

  if ($mca !~ /falcon/) {
     $mca =~ s/:$//;					#mca should not be ended by ":"
     $mcapv = $mca;
  } else {
     if ($mca !~ /:$/) {$mca.=':';}			#falcon must be ended by ":"
     $mcapv = $mca.'mca1';
  }

  ### Set number of channels:
  $pv = $mcapv.'.NUSE';
  $status = Pezca::PutDoubleOldCa($pv,$mcaNMAX);
  &report_exit($script, $status, $pv);

  ### Set Preset Counts
  $pv = $mcapv.'.PCT';
  $status = Pezca::PutDoubleOldCa($pv,0);
  &report_exit($script, $status, $pv);

  ### Set Preset Start Channel:
  $pv = $mcapv.'.PCTL';
  $status = Pezca::PutDoubleOldCa($pv,0);
  &report_exit($script, $status, $pv);

  ### Set Preset End Channel:
  $pv = $mcapv.'.PCTH';
  $status = Pezca::PutDoubleOldCa($pv,$mcaNMAX-1);
  &report_exit($script, $status, $pv);

  ### Set Mode to PHA:
  $pv = $mcapv.'.MODE';
  $status = Pezca::PutDoubleOldCa($pv,0);
  &report_exit($script, $status, $pv);

  ### Set Channel Advance to Internal:
  $pv = $mcapv.'.CHAS';
  $status = Pezca::PutDoubleOldCa($pv,0);
  &report_exit($script, $status, $pv);

  ### Set MCS internal Dwell time=0:
  $pv = $mcapv.'.DWEL';
  $status = Pezca::PutDoubleOldCa($pv,0);
  &report_exit($script, $status, $pv);

  ### Set MCS Ext. Prescale=1:
  $pv = $mcapv.'.PSCL';
  $status = Pezca::PutDoubleOldCa($pv,1);
  &report_exit($script, $status, $pv);

  ### Set Preset Live Time:
  $pv = $mcapv.'.PLTM';
  $status = Pezca::PutDoubleOldCa($pv,$count_time);
  &report_exit($script, $status, $pv);

  ### Set Preset Real Time:
  $pv = $mcapv.'.PRTM';
  $status = Pezca::PutDoubleOldCa($pv,$count_time);
  &report_exit($script, $status, $pv);

  if ($mca !~ /falcon/) { #---------------------------------------------Canberra

     ### Disable wait for client:
     $pv = $mcapv.'EnableWait';
     $status = Pezca::PutDoubleOldCa($pv,0);
     &report_exit($script, $status, $pv);

  } else { #------------------------------------------------------------Falcon 

     ### Set Collection mode to 0="MCA spectra":
     $pv = $mca.'CollectMode';
     $status = Pezca::PutDoubleOldCa($pv,0);
     &report_exit($script, $status, $pv);

     ### Set Preset Mode to 1="Real time":
     $pv = $mca.'PresetMode';
     $status = Pezca::PutDoubleOldCa($pv,1);
     &report_exit($script, $status, $pv);

     ### Set Preset Real Time (this also sets ".PRTM"):
     $pv = $mca.'PresetReal';
     $status = Pezca::PutDoubleOldCa($pv,$count_time);
     &report_exit($script, $status, $pv);

     ### Set SCA Trigger Mode to 3="Always":
     $pv = $mca.'SCATriggerMode';
     $status = Pezca::PutDoubleOldCa($pv,3);
     &report_exit($script, $status, $pv);

     ### Set Wait for Client to 0="Disable":
     $pv = $mca.'EnableClientWait';
     $status = Pezca::PutDoubleOldCa($pv,0);
     &report_exit($script, $status, $pv);

     ### Set Detector polarity to 1="Pos":
     $pv = $mca.'dxp1:DetectorPolarity';
     $status = Pezca::PutDoubleOldCa($pv,1);
     &report_exit($script, $status, $pv);

     ### Set Detection filter to 2="MidRate":
     $pv = $mca.'dxp1:DetectionFilter';
     $status = Pezca::PutDoubleOldCa($pv,2);
     &report_exit($script, $status, $pv);

     ### Set number of SCA to 1:
     $pv = $mca.'NumSCAs';
     $status = Pezca::PutDoubleOldCa($pv,1);
     &report_exit($script, $status, $pv);

     ### Check SNL status (expecting 1="Connected"):
     $pv = $mca.'SNL_Connected';
     ($status,$state) = Pezca::GetDouble($pv);
     if ($status) {
        &report_exit($script, $status, $pv); 
        $state=0;
     }
     if ($state != 1) {&press_key_exit('*** '.$script.': SNL not connected for MCA='.$mca,1);}

  }  #------------------------------------------------------------------end of Canberra/Falcon fork
}

#=================================================================

# sub setMCA_fastest ($mca, $count_time) {
sub setMCA_fastest ($$) {

  my ($script, $mca, $count_time, $status, $pv);
  $script      = 'setMCA_fastest';
  $mca         = shift(@_);
  $count_time  = shift(@_);
  $status      = 0;

  &setMCA($mca,$count_time);

  if ($mca !~ /falcon/) {
     $mca =~ s/:$//;					#mca should not be ended by ":"
  } else {
     if ($mca !~ /:$/) {$mca.=':';}			#falcon must be ended by ":"
  }

### Set 'Read Spectrum' to 'Passive':
### In this case the spectrum is only read after the aquisition is finished.
  if ($mca !~ /falcon/) {$pv = $mca.'Read.SCAN';}
  else                  {$pv = $mca.'ReadAll.SCAN';}
  $status = Pezca::PutDoubleOldCa($pv,0);
  &report_exit($script, $status, $pv);

### Set 'Check Done' to '0.1s':
### In this case the mca record 'learns' that acquisition is finished in shortest time.
  if ($mca !~ /falcon/) {$pv = $mca.'Status.SCAN';}
  else                  {$pv = $mca.'StatusAll.SCAN';}
  $status = Pezca::PutDoubleOldCa($pv,9);
  &report_exit($script, $status, $pv);
}

#=================================================================

# sub checkMCA_ROI ($mca,[$N]) {
sub checkMCA_ROI ($;$) {			# Returns ($roi_lo, $roi_hi)

  my ($script, $mca, $N, $status, $roi_lo, $roi_hi, $mcaNMAX, $pv, $mcapv);
  $script = 'checkMCA_ROI';
  $mca    = shift(@_);
  $N      = shift(@_);
  if (!defined $N) {$N = 0;}				# default ROI number

### Read the maximum possible number of channels:
  $mcaNMAX = &MCAnmax($mca);

  if ($mca !~ /falcon/) {
     $mca =~ s/:$//;					#mca should not be ended by ":"
     $mcapv = $mca;
  } else {
     if ($mca !~ /:$/) {$mca.=':';}			#falcon must be ended by ":"
     $mcapv = $mca.'mca1';
  }

### Read LO:
  $pv = $mcapv.'.R'.$N.'LO';				#e.g. ".R1LO"
  ($status,$roi_lo) = Pezca::GetDouble($pv);
  if ($status) {&report_exit($script, $status, $pv); $roi_lo = -1;}

### Read HI:
  $pv = $mcapv.'.R'.$N.'HI';				#e.g. ".R1HI"
  ($status,$roi_hi) = Pezca::GetDouble($pv);
  if ($status) {&report_exit($script, $status, $pv); $roi_hi = -1;}

  if ($roi_lo != -1 || $roi_hi != -1) {             	#(-1,-1) corresponds to ROI disabled
    if ($roi_lo < 0 || $roi_lo > $mcaNMAX-1) {
       &press_key_exit('*** '.$script.': Low ROI-'.$N.' = '.$roi_lo
                             .' is not set or not in valid range');
    }
    if ($roi_hi < 0 || $roi_hi > $mcaNMAX-1) {
       &press_key_exit('*** '.$script.': High ROI-'.$N.' = '.$roi_hi
                             .' is not set or not in valid range');
    }
    if ($roi_lo >= $roi_hi) {
       &press_key_exit('*** '.$script.': Incorrect data for ROI-'.$N
                             .': LO='.$roi_lo.' exceeds or equals to HI='.$roi_hi);
    }
  }

### Disable ROI preset (0="N"):
  $pv = $mcapv.'.R'.$N.'IP';				#e.g. ".R1IP"
  $status = Pezca::PutDoubleOldCa($pv,0);
  &report_exit($script, $status, $pv);

### Set nAvg=0:
  $pv = $mcapv.'.R'.$N.'BG';				##e.g. ".R1BG"
  $status = Pezca::PutDoubleOldCa($pv,0);
  &report_exit($script, $status, $pv);

  return $roi_lo, $roi_hi;
}

#=================================================================

# sub setMCA_ROI ($mca, $roi_lo, $roi_hi, [$N]) {
sub setMCA_ROI ($$$;$) {

  my ($script, $mca, $roi_lo, $roi_hi, $N, $status, $mcaNMAX, $pv, $mcapv, $dxppv);
  $script = 'setMCA_ROI';
  $mca    = shift(@_);
  $roi_lo = shift(@_);
  $roi_hi = shift(@_);
  $N      = shift(@_);
  if (!defined $N) {$N = 0;}				# default ROI number

### Read the maximum possible number of channels:
  $mcaNMAX = &MCAnmax($mca);

  if ($roi_lo != -1 || $roi_hi != -1) {             	#(-1,-1) corresponds to ROI disabled
    if ($roi_lo < 0 || $roi_lo > $mcaNMAX-1) {
       &press_key_exit('*** '.$script.': Low ROI-'.$N.' = '.$roi_lo
                                     .' is not in valid range');
    }
    if ($roi_hi < 0 || $roi_hi > $mcaNMAX-1) {
       &press_key_exit('*** '.$script.': High ROI-'.$N.' = '.$roi_hi
                                     .' is not in valid range');
    }
    if ($roi_lo >= $roi_hi) {
       &press_key_exit('*** '.$script.': Incorrect data for ROI-'.$N
                             .': LO='.$roi_lo.' exceeds or equals to HI='.$roi_hi);
    }
  }

  if ($mca !~ /falcon/) {
     $mca =~ s/:$//;					#mca should not be ended by ":"
     $mcapv = $mca;
  } else {
     if ($mca !~ /:$/) {$mca.=':';}			#falcon must be ended by ":"
     $mcapv = $mca.'mca1';
     $dxppv = $mca.'dxp1';
  }

### re-set ROI-LO to "0":
  $pv = $mcapv.'.R'.$N.'LO';				#e.g. ".R1LO"
  $status = Pezca::PutDoubleOldCa($pv,0);
  &report_exit($script, $status, $pv);
  select(undef,undef,undef,0.1);			#sleep($wait_step);

### set ROI-HI:
  $pv = $mcapv.'.R'.$N.'HI';				#e.g. ".R1HI"
  $status = Pezca::PutDoubleOldCa($pv,$roi_hi);
  &report_exit($script, $status, $pv);

### set ROI-LO:
  $pv = $mcapv.'.R'.$N.'LO';				#e.g. ".R1LO"
  $status = Pezca::PutDoubleOldCa($pv,$roi_lo);
  &report_exit($script, $status, $pv);

### Disable ROI preset (0="N"):
  $pv = $mcapv.'.R'.$N.'IP';				#e.g. ".R1IP"
  $status = Pezca::PutDoubleOldCa($pv,0);
  &report_exit($script, $status, $pv);

### Set nAvg=0:
  $pv = $mcapv.'.R'.$N.'BG';				#e.g. ".R1BG"
  $status = Pezca::PutDoubleOldCa($pv,0);
  &report_exit($script, $status, $pv);

  if ($mca =~ /falcon/) {
     select(undef,undef,undef,0.1);			#sleep($wait_step);
### copy ROI-LO and ROI-HI to SCA-Low and SCA-High:
     $pv = $mca.'CopyROI_SCA';
     $status = Pezca::PutDoubleOldCa($pv,1);
     &report_exit($script, $status, $pv);
### set SCA-Low
#    $pv = $dxppv.':SCA'.$N.'Low';			#e.g. ":SCA0Low"
#    $status = Pezca::PutDoubleOldCa($pv,$roi_lo);
#    &report_exit($script, $status, $pv);
### set SCA-High:
#    $pv = $dxppv.':SCA'.$N.'High';			#e.g. ".SCA0High"
#    $status = Pezca::PutDoubleOldCa($pv,$roi_hi);
#    &report_exit($script, $status, $pv);
  }
}

#=================================================================

# sub startMCA ($mca) {			# uses global $EraseMCA
sub startMCA ($) {			# uses global $EraseMCA

  our $EraseMCA;
  my ($script, $mca, $status, $pv);
  $script = 'startMCA';
  $mca    = shift(@_);
  $status = 0;
  if (!defined $EraseMCA) {$EraseMCA=0;}

  if ($mca !~ /falcon/) {
     $mca =~ s/:$//;					#mca should not be ended by ":"
  } else {
     if ($mca !~ /:$/) {$mca.=':';}			#falcon must be ended by ":"
  }

  if ($EraseMCA) {
    if ($mca !~ /falcon/) {$pv = $mca.'Start';}
    else                  {$pv = $mca.'StartAll';}
    $status = Pezca::PutDoubleOldCA($pv,1);
    &report_exit($script, $status, $pv);
  }
  else  {
    $pv = $mca.'EraseStart';
    $status = Pezca::PutDoubleOldCa($pv,1);
    &report_exit($script, $status, $pv);
  }
}

#=================================================================

# sub abortMCA ($mca) {
sub abortMCA ($) {

  my ($script, $mca, $status, $pv);
  $script = 'abortMCA';
  $mca    = shift(@_);
  $status = 0;

  if ($mca !~ /falcon/) {
     $mca =~ s/:$//;					#mca should not be ended by ":"
  } else {
     if ($mca !~ /:$/) {$mca.=':';}			#falcon must be ended by ":"
  }

  if ($mca !~ /falcon/) {$pv = $mca.'Stop';}
  else                  {$pv = $mca.'StopAll';}
  $status = Pezca::PutDouble($pv,1);
  &report_exit($script, $status, $pv);
}

#=================================================================

# sub eraseMCA ($mca) {
sub eraseMCA ($) {

  my ($script, $mca, $status, $pv);
  $script = 'eraseMCA';
  $mca    = shift(@_);
  $status = 0;

  if ($mca !~ /falcon/) {
     $mca =~ s/:$//;					#mca should not be ended by ":"
  } else {
     if ($mca !~ /:$/) {$mca.=':';}			#falcon must be ended by ":"
  }

  if ($mca !~ /falcon/) {$pv = $mca.'Erase';}
  else                  {$pv = $mca.'EraseAll';}
  $status = Pezca::PutDouble($pv,1);
  &report_exit($script, $status, $pv);
}

#=================================================================

# sub waitMCA ($mca) {
sub waitMCA ($) {

  my ($script, $mca, $status, $count_time, $acquiring, $pv);
  my ($wait_step, $wait_max, $wait_, $current_time, $msg);
  $script    = 'waitMCA';
  $mca       = shift(@_);
  $wait_step = 0.1;

  if ($mca !~ /falcon/) {
     $mca =~ s/:$//;					#mca should not be ended by ":"
  } else {
     if ($mca !~ /:$/) {$mca.=':';}			#falcon must be ended by ":"
  }

### Read Preset Real Time:
  if ($mca !~ /falcon/) {$pv = $mca.'.PRTM';}
  else                  {$pv = $mca.'PresetReal';}
  ($status,$count_time) = Pezca::GetDouble($pv);
  if ($status) {&report_exit($script, $status, $pv); $count_time=0;}
  if ($count_time <= 0) {
     &press_key_exit('*** '.$script.': zero or negative Preset Real Time');
  }
  $wait_max = $count_time + 10;

  if ($mca !~ /falcon/) {$pv = $mca.'.ACQG';}
  else                  {$pv = $mca.'Acquiring';}
  $status = &addCaMonitor($pv);

  $acquiring = 1;
  $wait_     = 0.5;
  select(undef,undef,undef,$wait_);		#wait a little to ensure that MCA started
  while ($acquiring == 1 && $wait_ < $wait_max) {
    select(undef,undef,undef,$wait_step);	# sleep($wait_step);
    $wait_ += $wait_step;
    ($status,$acquiring) = Pezca::GetDouble($pv);
    if ($status) {&report_exit($script, $status, $pv); $acquiring=0;}
  }

  if ($wait_ >= $wait_max) {
     &abortMCA($mca);
     $current_time = scalar localtime;
     $msg = sprintf '*** '.$current_time."\n".
                    '*** '.$script.': MCA did not finish counting after %4.0fs'
                                 ."\n", $wait_;
     &press_key_exit($msg);
  }
}

#=================================================================

# sub wait_readingMCA ($mca) {
sub wait_readingMCA ($) {			# - returns ($num_pts,@Intensity)

  my ($script, $mca, $status, $msg);
  my ($wait_step, $wait_, $wait_max, $current_time, $readFlag, $rdngFlag);
  my ($pv, $mcapv);
  $script = 'wait_readingMCA';
  $mca    = shift(@_);

  if ($mca !~ /falcon/) {
     $mca =~ s/:$//;						#mca should not be ended by ":"
     $mcapv = $mca;
  } else {
     if ($mca !~ /:$/) {$mca.=':';}				#falcon must be ended by ":"
     $mcapv = $mca.'mca1';
  }

### This is absolutely necessary precaution because the attempt
### to read running MCA results in its never stopping until
### the IOC is rebooted:
  &abortMCA($mca);

  $status = Pezca::PutDouble($mcapv.'.READ',1);
  &report_exit($script, $status, $mcapv.'.READ');
# select(undef,undef,undef,0.02);				# msleep(20) -- repalced by wait

### Wait reading to complete (change of 2008/10/17):
  $status = &addCaMonitor($mcapv.'.READ');
  $status = &addCaMonitor($mcapv.'.RDNG');
  $wait_step  = 0.1;
  $wait_max   = 5.0;
  $wait_      = 0.0;
  $readFlag   = 1;
  $rdngFlag   = 1;
  while (($readFlag || $rdngFlag) && $wait_ < $wait_max) {
     select(undef,undef,undef,$wait_step);			# sleep($wait_step);
     $wait_ += $wait_step;
     $pv = $mcapv.'.READ';
     ($status,$readFlag) = Pezca::GetDouble($pv);
     if ($status) {&report_exit($script, $status, $pv); $readFlag=0;}
     $pv = $mcapv.'.RDNG';
     ($status,$rdngFlag) = Pezca::GetDouble($pv);
     if ($status) {&report_exit($script, $status, $pv); $rdngFlag=0;}
  }
  if ($wait_ >= $wait_max) {
     &abortMCA($mca);
     $current_time = scalar localtime;
     $msg = sprintf '*** '.$current_time."\n".
                    '*** '.$script.': MCA [%s] did not finish reading scaler after %4.0fs'
                                  ."\n", $mcapv, $wait_;
     print STDOUT $msg."\n";
#    &press_key_exit($msg);
     return 1;
  } else {
     return 0;
  }
}

#=================================================================

# sub readMCA ($mca) {
sub readMCA ($) {			# - returns ($num_pts,@Intensity)

  our $EZCALONG;
  my ($script, $mca, $status, $num_pts, $arr_len, @Intensity, $msg);
  my ($pv, $mcapv, $current_time, $maxBytes, $dataType);
  
  if (defined $ENV{EPICS_CA_MAX_ARRAY_BYTES}) {$maxBytes = $ENV{EPICS_CA_MAX_ARRAY_BYTES};}
  else                                        {$maxBytes = 16384;}
### This is how PEZCA is compiled (see #define LISTBUFSIZE in Pezca.xs):
  if ($maxBytes > 65536)                      {$maxBytes = 65536;}

  $script = 'readMCA';
  $mca    = shift(@_);

  if (!defined $EZCALONG) {&setEZCA32or64();}

  if ($mca !~ /falcon/) {
     $mca =~ s/:$//;						#mca should not be ended by ":"
     $mcapv = $mca;
  } else {
     if ($mca !~ /:$/) {$mca.=':';}				#falcon must be ended by ":"
     $mcapv = $mca.'mca1';
  }

  $status = &wait_readingMCA($mca);				#stop MCA, initiate reading from device into record and wait it to complete
  if ($status) {return 0,();}

### Read the number of points:
  $pv = $mcapv.'.NUSE';
  ($status,$num_pts) = Pezca::GetDouble($pv);
  if ($status) {&report_exit($script, $status, $pv); $num_pts=0;}

### Read MCA:
  if ($mca !~ /falcon/) {$dataType = $EZCALONG;}                #do not care, these arrays are < 4000 elements 
  else {
### Type of variable  LONG      FLOAT     DOUBLE
### Length 64 bit       -         8         16
### Length 32-bit       4         4          8
     if ($EZCALONG =~ /FLOAT/) {                                #64-bit system
        if ( 16*$num_pts < $maxBytes) {$dataType='DOUBLE';}else{$dataType='FLOAT';}
     } else {                                                   #32-bit systems
        if ( 8*$num_pts < $maxBytes)  {$dataType='DOUBLE';}{$dataType = 'LONG';}
     }
  }
# print STDOUT $script.': num_pts='.$num_pts.'  dataType='.$dataType.'  EPICS_CA_MAX_ARRAY_BYTES='.$maxBytes."\n";
  $pv = $mcapv.'.VAL';
# ($status,@Intensity) = Pezca::GetList($pv,$EZCALONG,$num_pts);
  ($status,@Intensity) = Pezca::GetList($pv,$dataType,$num_pts);
  if ($status) {&report_exit($script, $status, $pv); @Intensity=0;}
  $arr_len = @Intensity;
  if ($arr_len < $num_pts) {
     $current_time = scalar localtime;
     $msg = sprintf '!!! '.$current_time."\n".
                    '!!! '.$script.': MCA [%s] returned data array shorter than requested'
                       .' (%d < %d)'."\n", $mcapv, $arr_len, $num_pts;
     print STDOUT $msg."\n";
  }
  return $num_pts,@Intensity;
}

#=================================================================

# sub readMCA_ref ($mca,\@Intensity) {
sub readMCA_ref ($\@) {					# - returns $status

  our $EZCALONG;
  my ($script, $mca, $status, $status_final, $intensityRef, $num_pts);
  my ($pv, $mcapv, $current_time, $arr_len, $msg, $maxBytes, $dataType);
  
  if (defined $ENV{EPICS_CA_MAX_ARRAY_BYTES}) {$maxBytes = $ENV{EPICS_CA_MAX_ARRAY_BYTES};}
  else                                        {$maxBytes = 16384;}
### This is how PEZCA is compiled (see #define LISTBUFSIZE in Pezca.xs):
  if ($maxBytes > 65536)                      {$maxBytes = 65536;}

  $script       = 'readMCA_ref';
  $mca          = shift(@_);
  $intensityRef = shift(@_);
  $status_final = 0;

  if (!defined $EZCALONG) {&setEZCA32or64();}

  if ($mca !~ /falcon/) {
     $mca =~ s/:$//;						#mca should not be ended by ":"
     $mcapv = $mca;
  } else {
     if ($mca !~ /:$/) {$mca.=':';}				#falcon must be ended by ":"
     $mcapv = $mca.'mca1';
  }

  $status = &wait_readingMCA($mca);				#stop MCA, initiate reading from device into record and wait it to complete
# if ($status) {return 0,();}
  if ($status) {$status_final++;}

### Read the number of points:
  $pv = $mcapv.'.NUSE';
  ($status,$num_pts) = Pezca::GetDouble($pv);
  if ($status) {&report_exit($script, $status, $pv); $num_pts=0; $status_final++;}

### Read MCA:
  if ($mca !~ /falcon/) {$dataType = $EZCALONG;}                #do not care, these arrays are < 4000 elements 
  else {
### Type of variable  LONG      FLOAT     DOUBLE
### Length 64 bit       -         8         16
### Length 32-bit       4         4          8
     if ($EZCALONG =~ /FLOAT/) {                                #64-bit system
        if ( 16*$num_pts < $maxBytes) {$dataType='DOUBLE';}else{$dataType='FLOAT';}
     } else {                                                   #32-bit systems
        if ( 8*$num_pts < $maxBytes)  {$dataType='DOUBLE';}{$dataType = 'LONG';}
     }
  }
# print STDOUT $script.': num_pts='.$num_pts.'  dataType='.$dataType.'  EPICS_CA_MAX_ARRAY_BYTES='.$maxBytes."\n";
  $pv = $mcapv.'.VAL';
# ($status,@$intensityRef) = Pezca::GetList($pv,$EZCALONG,$num_pts);
  ($status,@$intensityRef) = Pezca::GetList($pv,$dataType,$num_pts);
  if ($status) {&report_exit($script, $status, $pv); @$intensityRef=0; $status_final++;}
  $arr_len = @$intensityRef;
  if ($arr_len < $num_pts) {
     $msg = sprintf '!!! '.$current_time."\n".
                    '!!! '.$script.': MCA [%s] returned data array shorter than requested'
                       .' (%d < %d)'."\n", $mcapv, $arr_len, $num_pts;
     print STDOUT $msg."\n";
  }
  return $status_final;
}

#=================================================================

# sub readMCA_total ($mca) {
sub readMCA_total ($) {					# - returns ($total,$num_pts)

  my ($script, $mca, $status, $num_pts, $total, @Intensity, $i);
  $script  = 'readMCA_total';
  $mca     = shift(@_);
  $num_pts = 0;
  $total   = 0;

# ($num_pts,@Intensity) = &readMCA($mca);
  $status = &readMCA_ref($mca,\@Intensity);
  if ($status == 0) {
     $num_pts = @Intensity;
     for ($i=0; $i<$num_pts; $i++) {$total += $Intensity[$i];}
  }
  return ($total, $num_pts);
}

#=================================================================

# sub readMCA_ROI($mca,[$N]) {
sub readMCA_ROI($;$) {			# - returns ($sum_data,$net_data)

  my ($script, $mca, $status, $N, $sum_data, $net_data, $pv, $mcapv);
  $script = 'readMCA_ROI';
  $mca    = shift(@_);
  $N      = shift(@_);
  if (!defined $N) {$N = 0;}		# default ROI number

  if ($mca !~ /falcon/) {
     $mca =~ s/:$//;						#mca should not be ended by ":"
     $mcapv = $mca;
  } else {
     if ($mca !~ /:$/) {$mca.=':';}				#falcon must be ended by ":"
     $mcapv = $mca.'mca1';
  }

  $status = &wait_readingMCA($mca);				#stop MCA, initiate reading from device into record and wait it to complete
# if ($status) {return 0,0;}

### Read MCA ROI:
  $pv = $mcapv.'.R'.$N;						#e.g. ".R1"
  ($status,$sum_data) = Pezca::GetDouble($pv);
  if ($status) {&report_exit($script, $status, $pv); $sum_data=0;}
  $pv = $mcapv.'.R'.$N.'N';					#e.g. ".R1N"
  ($status,$net_data) = Pezca::GetDouble($pv);
  if ($status) {&report_exit($script, $status, $pv); $net_data=0;}

  return ($sum_data, $net_data);
}

#=================================================================

# sub readMCA_deadTime($mca) {
sub readMCA_deadTime($) {			# - returns $dead_time

  my ($script, $mca, $status, $dead_time, $mcapv, $pv);
  $script = 'readMCA_deadTime';
  $mca    = shift(@_);

  if ($mca !~ /falcon/) {
     $mca =~ s/:$//;						#mca should not be ended by ":"
     $mcapv = $mca;
  } else {
     if ($mca !~ /:$/) {$mca.=':';}				#falcon must be ended by ":"
     $mcapv = $mca.'mca1';
  }

### Read MCA dead time:
  $pv = $mcapv.'.IDTIM';
  ($status,$dead_time) = Pezca::GetDouble($pv);
  if ($status) {&report_exit($script, $status, $pv); $dead_time=0;}

  return $dead_time;
}

#=================================================================

# sub readMCA_deadTime_from_live($mca) {
sub readMCA_deadTime_from_live($) {			# - returns $dead_time

  my ($script, $mca, $status, $dead_time, $live_time, $real_time, $mcapv, $pv);
  $script = 'readMCA_deadTime_from_live';
  $mca    = shift(@_);

  if ($mca !~ /falcon/) {
     $mca =~ s/:$//;						#mca should not be ended by ":"
     $mcapv = $mca;
  } else {
     if ($mca !~ /:$/) {$mca.=':';}				#falcon must be ended by ":"
     $mcapv = $mca.'mca1';
  }

### Read MCA live time:
  $pv = $mcapv.'.ELTM';
  ($status,$live_time) = Pezca::GetDouble($pv);
  if ($status) {&report_exit($script, $status, $pv); $live_time=0;}

### Read MCA real time:
  $pv = $mcapv.'.ERTM';
  ($status,$real_time) = Pezca::GetDouble($pv);
  if ($status) {&report_exit($script, $status, $pv); $real_time=0;}

  if ($real_time > 0) {$dead_time = 100.*($real_time-$live_time)/$real_time;}
  else {
    printf STDOUT '*** '.$script.': RealTime=%3.2fs is invalid. '
                        .'Cannot calculate dead time'."\n",
                         $real_time;
    $dead_time = 0;
  }

  return $dead_time;
}

#=================================================================

# sub setupMCA_ADC($adc, [$gain, $offset, $low]) {			#Canberra only
sub setupMCA_ADC($;$$$) {

### This routine sets initial parameters for the MCA ADC
### -- if gain and offset are not specified, they are skipped
### -- if low discrim. is not specified, it is set to 5.
### -- high discrim. is always set to 100.
  my ($script, $status, $adc, $gain, $offset, $low, $high);
  my ($pv, $gainEPICS);
  $script = 'setupMCA_ADC';
  $adc    = shift(@_);
  $gain   = shift(@_);
  $offset = shift(@_);
  $low    = shift(@_);

  if (defined $adc && $adc !~ /falcon/) {
     $adc =~ s/:$//;						#mca should not be ended by ":"

     $status = 0;

     if (defined $gain) {
        $gainEPICS = -1;
        ### The input parameter gain can be specified as a
        ### band number in range [0:6] or as a band label that
        ### users read on the MEDM screen  (256,512,1024,...).
        ### In the later case the label needs to be converted
        ### into respective PV value:
        if    ($gain == 256)   {$gainEPICS = 0;}
        elsif ($gain == 512)   {$gainEPICS = 1;}
        elsif ($gain == 1024)  {$gainEPICS = 2;}
        elsif ($gain == 2048)  {$gainEPICS = 3;}
        elsif ($gain == 4096)  {$gainEPICS = 4;}
        elsif ($gain == 8192)  {$gainEPICS = 5;}
        elsif ($gain == 16384) {$gainEPICS = 6;}
        if ($gainEPICS < 0 || $gainEPICS > 6) {
           &press_key_exit('*** '.$script.': gain='.$gain.' is invalid');
        }
        ### Set gain:
        $pv = $adc.'GAIN';
        $status = Pezca::PutDoubleOldCa($pv,$gainEPICS);
        &report_exit($script, $status, $pv);
     }

     if (defined $offset) {
        if ($offset < 0 || $offset > 16128) {			#16128 = 128*126 = 256*63
           &press_key_exit('*** '.$script.': offset='.$offset
                                 .' is not in range [0-16128]');
        }
        ### Set offset:
        $pv = $adc.'OFFSET';
        $status = Pezca::PutDoubleOldCa($pv,$offset);
        &report_exit($script, $status, $pv);
     }

     if (!defined $low) {$low=5.;}
     $high = 100;
     &setMCA_ADC_discr($adc, $low, $high)
  } else {
     print STDOUT '!!! '.$script.': nothing to do for Falcon.'."\n";
  }
}

#=================================================================

# sub setMCA_ADC($adc, $gain, $offset) {			#Canberra only
sub setMCA_ADC($$$) {

  my ($script, $status, $adc, $gain, $offset, $gainEPICS, $pv);
  $script = 'setMCA_ADC';
  $adc    = shift(@_);
  $gain   = shift(@_);
  $offset = shift(@_);

  if (defined $adc && $adc !~ /falcon/) {
     $adc =~ s/:$//;						#mca should not be ended by ":"

     $status = 0;

     $gainEPICS = -1;
     ### The input parameter gain can be specified as a
     ### band number in range [0:6] or as a band label that
     ### users read on the MEDM screen  (256,512,1024,...).
     ### In the later case the label needs to be converted
     ### into respective PV value:
     if    ($gain == 256)   {$gainEPICS = 0;}
     elsif ($gain == 512)   {$gainEPICS = 1;}
     elsif ($gain == 1024)  {$gainEPICS = 2;}
     elsif ($gain == 2048)  {$gainEPICS = 3;}
     elsif ($gain == 4096)  {$gainEPICS = 4;}
     elsif ($gain == 8192)  {$gainEPICS = 5;}
     elsif ($gain == 16384) {$gainEPICS = 6;}
     if ($gainEPICS < 0 || $gainEPICS > 6) {
        &press_key_exit('*** '.$script.': gain='.$gain.' is invalid');
     }
     ### Set gain:
     $pv = $adc.'GAIN';
     $status = Pezca::PutDoubleOldCa($pv,$gainEPICS);
     &report_exit($script, $status, $pv);

     if ($offset < 0 || $offset > 16128) {			#16128 = 128*126 = 256*63
        &press_key_exit('*** '.$script.': offset='.$offset
                                   .' is not in range [0-16128]');
     }
     ### Set offset:
     $pv = $adc.'OFFSET';
     $status = Pezca::PutDoubleOldCa($pv,$offset);
     &report_exit($script, $status, $pv);
  } else {
     print STDOUT '!!! '.$script.': nothing to do for Falcon.'."\n";
  }
}

#=================================================================

# sub setMCA_ADC_discr($adc, $low, $high) {			#Canberra only
sub setMCA_ADC_discr($$$) {

  my ($script, $status, $adc, $low, $high, $pv);
  $script = 'setMCA_ADC_discr';
  $adc    = shift(@_);
  $low    = shift(@_);
  $high   = shift(@_);

  if (defined $adc && $adc !~ /falcon/) {
     $adc =~ s/:$//;						#mca should not be ended by ":"

     $status = 0;

     if ($low < 0 || $low > 100) {
        &press_key_exit('*** '.$script.': Lower_discr='.$low
                                   .' is not in range [0-100]');
     }

     if ($high < 0 || $high > 100) {
        &press_key_exit('*** '.$script.': Higher_discr='.$high
                                   .' is not in range [0-100]');
     }

     if ($high <= $low) {
        &press_key_exit('*** '.$script.': Higher_discr='.$high
                                   .' is inconsistent with Lower_discr='.$low);
     }

     ### Set low:
     $pv = $adc.'LLD';
     $status = Pezca::PutDoubleOldCa($pv,$low);
     &report_exit($script, $status, $pv);

     ### Set high:
     $pv = $adc.'ULD';
     $status = Pezca::PutDoubleOldCa($pv,$high);
     &report_exit($script, $status, $pv);

     ### Set zero:
     $pv = $adc.'ZERO';
     $status = Pezca::PutDoubleOldCa($pv,0);
     &report_exit($script, $status, $pv);
  } else {
     print STDOUT '!!! '.$script.': nothing to do for Falcon.'."\n";
  }
}

#=================================================================

# sub getMCA_ADC($adc) {			#Canberra only
sub getMCA_ADC($) {				#returns ($gain, $offset)

  my ($script, $status, $adc, $gain, $gainEPICS, $offset, $pv);
  $script = 'getMCA_ADC';
  $adc    = shift(@_);

  if (defined $adc && $adc !~ /falcon/) {
     $adc =~ s/:$//;						#mca should not be ended by ":"

     ### Get gain:
     $pv = $adc.'GAIN';
     ($status,$gainEPICS) = Pezca::GetDouble($pv);
     if ($status) {&report_exit($script, $status, $pv); $gainEPICS=-1;}

     ### Get offset:
     $pv = $adc.'OFFSET';
     ($status,$offset) = Pezca::GetDouble($pv);
     if ($status) {&report_exit($script, $status, $pv); $offset=0;}

     ### Convert the gain from the PV value (band number in range [0:6])
     ### into respective band label (256,512,1024,...) that users read
     ### on the MEDM screen:
     if    ($gainEPICS == 0) {$gain = 256;}
     elsif ($gainEPICS == 1) {$gain = 512;}
     elsif ($gainEPICS == 2) {$gain = 1024;}
     elsif ($gainEPICS == 3) {$gain = 2048;}
     elsif ($gainEPICS == 4) {$gain = 4096;}
     elsif ($gainEPICS == 5) {$gain = 8192;}
     elsif ($gainEPICS == 6) {$gain = 16384;}
     else  {
        &press_key_exit('*** '.$script.': gain='.$gainEPICS.' is invalid');
     }
  } else {
     print STDOUT '!!! '.$script.': nothing to do for Falcon.'."\n";
     $gain = 1;
     $offset = 0;
  }
  return ($gain, $offset);
}

#=================================================================

# sub getMCA_ADC_discr($adc) {			#Canberra only
sub getMCA_ADC_discr($) {			#returns ($low, $high)

  my ($script, $status, $adc, $low, $high, $pv);
  $script = 'getMCA_ADC_discr';
  $adc    = shift(@_);

  if (defined $adc && $adc !~ /falcon/) {
     $adc =~ s/:$//;						#mca should not be ended by ":"

     $status = 0;

     ### Get low:
     $pv = $adc.'LLD';
     ($status,$low) = Pezca::GetDouble($pv);
     &report_exit($script, $status, $pv);

     ### Get high:
     $pv = $adc.'ULD';
     ($status,$high) = Pezca::GetDouble($pv);
     &report_exit($script, $status, $pv);
  } else {
     print STDOUT '!!! '.$script.': nothing to do for Falcon.'."\n";
     $low  = 0;
     $high = 1;
  }
  return ($low,$high);
}

#=================================================================

# sub setupMCA_TCA ($tca,$low) {			#Canberra only
sub setupMCA_TCA ($$) {

### Setup 3-channel analyzer:

  my ($script, $tca, $low, $status, $gateEnable, $scaLo, $scaHi);
  my ($pv, $i);
  $script = 'setupMCA_TCA';
  $tca    = shift(@_);
  $low    = shift(@_);

  if (defined $tca && $tca !~ /falcon/) {
     $tca =~ s/:$//;						#mca should not be ended by ":"

     if ($low < 0 || $low > 99) {
        &press_key_exit('*** '.$script.': low='.$low.' is invalid');
     }

     ### Set Output Polarity = Normal:
     $pv = $tca.'POLARITY';
     $status = Pezca::PutDoubleOldCa($pv,0);
     &report_exit($script, $status, $pv);

     ### Set ICR Threshold = Auto
     $pv = $tca.'THRESHOLD';
     $status = Pezca::PutDoubleOldCa($pv,0);
     &report_exit($script, $status, $pv);

     ### Set SCA Enable = Yes
     $pv = $tca.'SCA_ENABLE';
     $status = Pezca::PutDoubleOldCa($pv,1);
     &report_exit($script, $status, $pv);

     ### Set PUR Enable = No
     $pv = $tca.'PUR_ENABLE';
     $status = Pezca::PutDoubleOldCa($pv,0);
     &report_exit($script, $status, $pv);

     ### Set Select TCA = Yes
     $pv = $tca.'TCA_SELECT';
     $status = Pezca::PutDoubleOldCa($pv,1);
     &report_exit($script, $status, $pv);

     ### Set ROI to SCA = Disable
     $pv = $tca.'ROI_SCA_ENABLE';
     $status = Pezca::PutDoubleOldCa($pv,0);
     &report_exit($script, $status, $pv);

     ### Loop over 3 SCA
     for ($i=1; $i<=3; $i++) {
        ### SCA-1 is enabled:
        ### SCA-2 and SCA-3 are disabled:
        if ($i == 1) {$gateEnable = 1;  $scaLo = $low;  $scaHi = 100.;}
        else         {$gateEnable = 0;  $scaLo = 0.;    $scaHi = 0.;}
        ### Set Gate = Enable/Disable
        $pv = $tca.'SCA'.$i.'_GATE';
        $status = Pezca::PutDoubleOldCa($pv,$gateEnable);
        &report_exit($script, $status, $pv);
        ### Set PUR = No
        $pv = $tca.'SCA'.$i.'_PUR';
        $status = Pezca::PutDoubleOldCa($pv,0);
        &report_exit($script, $status, $pv);
        ### Set Low = ...
        $pv = $tca.'SCA'.$i.'_LOW';
        $status = Pezca::PutDoubleOldCa($pv,$scaLo);
        &report_exit($script, $status, $pv);
        ### Set High = ...
        $pv = $tca.'SCA'.$i.'_HI';
        $status = Pezca::PutDoubleOldCa($pv,$scaHi);
        &report_exit($script, $status, $pv);
     }
  } else {
     print STDOUT '!!! '.$script.': nothing to do for Falcon.'."\n";
  }
}

#=================================================================

# sub setMCA_TCA ($tca,$low,$high,[$scaNo]) {
sub setMCA_TCA ($$$;$) {

# Setup Low and High on a given scaler of 3-channel analyzer:

  my ($script, $tca, $status, $low, $high, $scaNo, $pv);
  $script = 'setMCA_TCA';
  $tca    = shift(@_);
  $low    = shift(@_);
  $high   = shift(@_);
  $scaNo  = shift(@_);  if (!defined $scaNo) {$scaNo=1;}

  if (defined $tca && $tca !~ /falcon/) {
     $tca =~ s/:$//;						#mca should not be ended by ":"

     $status = 0;

     if ($low < 0 || $low > $high || $high > 100) {
        &press_key_exit('*** '.$script.': low='.$low.', high='.$high
                                   .' is invalid data');
     }

     if ($scaNo < 1 || $scaNo > 3) {
        &press_key_exit('*** '.$script.': SCA='.$scaNo
                                   .' is not in range [1--3]');
     }

     ### Set Low = ...
     $pv = $tca.'SCA'.$scaNo.'_LOW';
     $status = Pezca::PutDoubleOldCa($pv,$low);
     &report_exit($script, $status, $pv);
     ### Set High = ...
     $pv = $tca.'SCA'.$scaNo.'_HI';
     $status = Pezca::PutDoubleOldCa($pv,$high);
     &report_exit($script, $status, $pv);
  } else {
     print STDOUT '!!! '.$script.': nothing to do for Falcon.'."\n";
  }
}

#=================================================================

# sub getMCA_TCA($tca,[$scaNo]) {
sub getMCA_TCA($;$) {					# - returns ($low,$high)

  my ($script, $tca, $status, $scaNo, $low, $high, $pv);
  $script = 'getMCA_TCA';
  $tca    = shift(@_);
  $scaNo  = shift(@_);  if (!defined $scaNo) {$scaNo=1;}

  if (defined $tca && $tca !~ /falcon/) {
     $tca =~ s/:$//;						#mca should not be ended by ":"

     if ($scaNo < 1 || $scaNo > 3) {
        &press_key_exit('*** '.$script.': SCA='.$scaNo
                                   .' is not in range [1--3]');
     }

     ### Get low for a given SCA:
     $pv = $tca.'SCA'.$scaNo.'_LOW';
     ($status,$low) = Pezca::GetDouble($pv);
     if ($status) {&report_exit($script, $status, $pv); $low=0;}

     ### Get high for a given SCA:
     $pv = $tca.'SCA'.$scaNo.'_HI';
     ($status,$high) = Pezca::GetDouble($pv);
     if ($status) {&report_exit($script, $status, $pv); $high=0;}

  } else {
     print STDOUT '!!! '.$script.': nothing to do for Falcon.'."\n";
     $low  = 0;
     $high = 1;
  }
  return ($low, $high);
}

#=================================================================

# sub getDerekMCAch2en($adc) {                          #called from getMCAcalib and optimizeMCA_ADC (both Canberra only)
sub getDerekMCAch2en($) {
#----------------------------------------------------
# This function defines global vars ($derekSlope, $derekOffset)
# needed for recalculating MCA channels into energy and fills
# global variables $derekSlope, $derekOffset with these data.
#----------------------------------------------------
  our ($derekSlope, $derekOffset);	# GLOBAL - must be on the MCA calibration form (channel -> eV)
  our $beamline;                        # GLOBAL (usually from beamline.pl)
  my ($script, $adc, $status, $adcpv, $pv);
  $script = 'getDerekMCAch2en';
  $adc    = shift(@_);

  if (defined $adc && $adc !~ /falcon/) {
     $adc =~ s/:$//;					#mca should not be ended by ":"
     $adcpv = $adc;
#######################################################################
### Derek's calibration ($s, $o):                                   ###
### ==> for Canberra                                                ###
### $E(eV) = $s*(2048/gain) * ($channel + $offsetReal) - $o;        ###
### ==> for Falcon                                                  ###
### $E(eV) = $s*$channel - $o;                                      ###
#######################################################################
     if (!defined $derekSlope) {
        $pv = $adcpv.'.CALS';
        ($status, $derekSlope) = Pezca::GetDouble($pv);
        if ($status) {&report_exit($script, $status, $pv); $derekSlope=0;}
     }
     if (!defined $derekOffset) {
        $pv = $adcpv.'.CALO';
        ($status, $derekOffset) = Pezca::GetDouble($pv);
        if ($status) {&report_exit($script, $status, $pv); $derekOffset=0;}
     }

### These are not supposed to be used:
     if ($derekSlope == 0 && $derekOffset == 0) {
        &press_key_exit("\n".'*** '.$script.': no channel-to-energy calibration for'."\n"
                       .'MCA_ADC='.$adcpv.' in '.$adcpv.'.CALS and/or '.$adcpv.'.CALO EPICS PVs');
     }
  } else {
     print STDOUT '!!! '.$script.': nothing to do for Falcon.'."\n";
     $derekSlope  = 0;
     $derekOffset = 0;
  }
}

#=================================================================

# sub getMCAcalib($adc,[$gain,$offset]) {       #if gain,offset are not passed, current device data is read
sub getMCAcalib($;$$) {
#----------------------------------------------------
# This function returns ($Escale,$Eoffset) for
# recalculating MCA channels into energy in [eV]:
#
#          $E(eV) = $E_scale * $channel + $E_offset;
# or:
#          $channel = ($E - $E_offset) / $E_scale;
#
# where 
# ==> for Canberra:
#           $E_scale(eV) = $derekSlope * (2048/$gain);
#	   $E_offset(eV) = $E_scale * $offsetReal - $derekOffset;
#        $offsetReal(eV) = $offsetStep*[int($offset/$offsetStep)]
# and $channel is 0-2047 and $offsetStep=128
#
# ==> for Falcon: -- see "&getFalconCalib"
#
#----------------------------------------------------
  our ($derekSlope, $derekOffset);	#GLOBAL - must be on the MCA calibration form (channel -> eV)
  my ($script, $adc, $gain, $offset);
  my ($gain_ADC, $offset_ADC);
  my ($offsetStep, $offsetReal);
  my ($E_scale, $E_offset);		#returned parmaters (output)
  $script     = 'getMCAcalib';
  $adc        = shift(@_);
  $gain       = shift(@_);
  $offset     = shift(@_);
  $offsetStep = 128;	 		#CONSTANT: offset step really affecting Canberra (was 256).

  if (defined $adc && $adc !~ /falcon/) { #---------------------------------------------Canberra
     ############################################################
     ### Derek's calibration ($s, $o):                        ###
     ### $E = $s*(2048/gain) * ($channel + $offsetReal) - $o; ###
     ############################################################

     if (!defined $derekSlope || !defined $derekOffset) {&getDerekMCAch2en($adc);}

     ### Get current gain & offset from the device:
     if (!defined $gain || !defined $offset) {
        ($gain_ADC, $offset_ADC) = &getMCA_ADC($adc);
        if (!defined $gain)   {$gain = $gain_ADC;}
        if (!defined $offset) {$offset = $offset_ADC;}
     }

     if ($gain < 256 || $gain > 16384) {
        &press_key_exit('*** '.$script.': gain='.$gain
                                  .' is not in range [256--16384]');
     }

     if ($offset < 0 || $offset > 16128) {				#16128 = 128*126 = 256*63
        &press_key_exit('*** '.$script.': offset='.$offset
                                   .' is not in range [0--16128]');
     }
     ### Round $offset to $offsetStep:
     $offsetReal = $offsetStep * int($offset/$offsetStep);		#128*(0--126) or 256*(0--63)

     $E_scale  = $derekSlope * (2048/$gain);
     $E_offset = $E_scale * $offsetReal - $derekOffset;
     
  } else { #------------------------------------------------------------Falcon

     print STDOUT '!!! '.$script.': nothing to do for Falcon.'."\n";
     $E_scale  = 0;
     $E_offset = 0;

  }     
  return ($E_scale, $E_offset);
}

#=================================================================

# sub MCAcalibDB2006($gain) {                   #called from scaleMCA and optimizeMCA_ADC
sub MCAcalibDB2006($) {							#Canberra only
#----------------------------------------------------
# This function returns $E_scale for given gain that can be
# used for recalculating MCA channels into energy in [eV]:
#
#                 $E(eV) = $E_scale * $channel + $E_offset;
# where:
# ==> for Canberra:
#           $E_scale(eV) = $derekSlope * (2048/$gain);
#	   $E_offset(eV) = $E_scale * $offsetReal - $derekOffset;
#        $offsetReal(eV) = $offsetStep*[int($offset/$offsetStep)]
# and $channel is 0-2047 and $offsetStep=128
#----------------------------------------------------
  our ($derekSlope, $derekOffset);	#GLOBAL - must be predefined in getMCAcalib (channel -> eV)
  my ($script, $gain);
  my ($E_scale);			#returned parmater (output)
  $script = 'MCAcalibDB2006';
  $gain   = shift(@_);

  if (defined $gain) {							#Canberra

     ############################################################
     ### Derek's calibration ($s, $o):                        ###
     ### $E = $s*(2048/gain) * ($channel + $offsetReal) - $o; ###
     ############################################################
     if (!defined $derekSlope || !defined $derekOffset) {
        &press_key_exit('*** '.$script.': calibrations are undefined. '
                                      .'Call getDerekMCAch2en first.');
     }
     ### For Canberra gain=256,512,1024,2048.4096,8192,16384
     if ($gain < 256 || $gain > 16384) {
        &press_key_exit('*** '.$script.': gain='.$gain
                              .' is not in range [256--16384]');
     }
     $E_scale = $derekSlope * (2048/$gain);

  } else {								#Falcon

     print STDOUT '!!! '.$script.': nothing to do for Falcon.'."\n";
     $E_scale  = 0;

  }
# print STDOUT $script.' data for gain='.$gain.': E_scale='.$E_scale."\n";
  return $E_scale;
}

#=================================================================

# sub scaleMCA($mcaNMAX,$gain) {
sub scaleMCA($$) {
#----------------------------------------------------
# This function returns the MCA energy scale ($dE) at given $gain.
#----------------------------------------------------
  my ($script, $mcaNMAX, $gain, $E_scale, $scale);
  $script  = 'scaleMCA';
  $mcaNMAX = shift(@_);
  $gain    = shift(@_);
  
  if (defined $gain) {							#Canberra

     $E_scale = &MCAcalibDB2006($gain);
     $scale = ($mcaNMAX - 1) * $E_scale;				#in eV

  } else {								#Falcon

     print STDOUT '!!! '.$script.': nothing to do for Falcon.'."\n";
     $scale = 0;

  }
  return $scale;							#in eV
}

#=================================================================

# sub mcaRoi2energy($roi,$adc,[$gain,$offset]) {
sub mcaRoi2energy($$;$$) {
#----------------------------------------------------
# This function recalculates MCA channel or ROI border into Energ(eV)
#----------------------------------------------------
  my ($script, $E, $roi, $adc, $E_scale, $E_offset);
  $script  = 'mcaRoi2energy';
  $roi      = shift(@_);
  $adc      = shift(@_);
  $E_scale  = shift(@_);
  $E_offset = shift(@_);

  if (defined $adc && $adc !~ /falcon/) {				#Canberra

     if (!defined $E_scale || !defined $E_offset) {($E_scale,$E_offset) = &getMCAcalib($adc);}

     if ($E_scale != 0) {$E = $E_scale * $roi + $E_offset;}
     else               {&press_key_exit('*** '.$script.': Zero E_scale'); $E=0;}

  } else {								#Falcon

     print STDOUT '!!! '.$script.': nothing to do for Falcon.'."\n";
     $E = 0;

  }

  return $E;								#in eV
}

#=================================================================

#sub mcaEnergy2roi($E,$adc,$E_scale,$E_offset) {
sub mcaEnergy2roi($$;$$) {
#----------------------------------------------------
# This function recalculates Energy(eV) into MCA channel or ROI border
#----------------------------------------------------
  my ($script, $E, $roi, $adc, $E_scale, $E_offset);

  $script  = 'mcaEnergy2roi';
  $E        = shift(@_);						#in eV
  $adc      = shift(@_);
  $E_scale  = shift(@_);
  $E_offset = shift(@_);

  if (defined $adc && $adc !~ /falcon/) {				#Canberra

     if (!defined $E_scale || !defined $E_offset) {($E_scale,$E_offset) = &getMCAcalib($adc);}

     if ($E_scale != 0) {$roi = ($E - $E_offset) / $E_scale;}
     else               {&press_key_exit('*** '.$script.': Zero E_scale'); $roi=0;}
     
  } else {								#Falcon

     print STDOUT '!!! '.$script.': nothing to do for Falcon.'."\n";
     $E = 0;

  }

  return $roi;
}

#=================================================================

# sub optimizeMCA_ADC($E1,$E2,$mcaNMAX,$adc,[$tooWideFlag) {	#$E1,$E2 are in eV; toowide: 0=error,1=adjust_E1,2=$adjust_E2
sub optimizeMCA_ADC($$$$) {						#Canberra only
#----------------------------------------------------
# This function optimizes ADC gain and offset so that energies
# $E1 and $E2 are displayed with the maximum resolution and at
# the center of MCA plot. It returns channel numbers ($N1,$N2)
# where $E1 and $E2 will be displayed.
#----------------------------------------------------
  our ($derekSlope, $derekOffset);	#GLOBAL - must be predefined in getMCAcalib (channel -> eV)
  our $report;
  my ($script, $adc);
  my ($E1, $E2, $E_, $tooWide, $N1, $N2, $mcaNMAX, $gain);
  my ($E_center, $dE, $E_scale, $E_offset, $E_range);
  my ($offsetStep, $offset, $offsetReal, $range_changed);
  $script     = 'optimizeMCA_ADC';
  $E1         = shift(@_);		#in eV
  $E2         = shift(@_);		#in eV
  $mcaNMAX    = shift(@_);
  $adc        = shift(@_);
  $tooWide    = shift(@_);		#what to do when range is too wide: 0=error,1=adjust_E1,2=$adjust_E2
  if (!defined $tooWide) {$tooWide=0;}
  $offsetStep = 128; 			#CONSTANT: offset step affecting Canberra (before it was 256).
  if (!defined $report) {$report=0;}

  if (defined $adc && $adc !~ /falcon/) {				#Canberra
     if ($E2 < $E1) {			#we expect $E2 > $E1; swap them if it is not the case
        $E_ = $E2;
        $E2 = $E1;
        $E1 = $E_;
        undef $E_;
     }
     $E_center = ($E2 + $E1) / 2.0;	#in eV
     $dE       = ($E2 - $E1);		#in eV
     if ($report) {
        printf STDOUT $script.': Need to optimize to: E_center=%.2feV, dE=%.2feV'."\n",$E_center,$dE;
     }

     if (!defined $derekSlope || !defined $derekOffset) {&getDerekMCAch2en($adc);}
     if ($report) {
        print STDOUT $script.': MCA/ADC calibrations by Derek: slope='
                            .$derekSlope.', offset='.$derekOffset."\n";
     }

     ### Find the maxiimum gain at which the MCA scale is wider that $dE:
     $gain = 16384;
     while (&scaleMCA($mcaNMAX,$gain) <= $dE) {
        $gain = int($gain/2);
        if ($gain < 2048) {
           if ($tooWide == 0) {
              &press_key_exit('*** '.$script.': Requested energy range dE='.$dE
                                         .'eV is wider than MCA could analyze.');
           } else {
              last;
           }
        }
     }
     if ($gain < 2048) {
        $gain = 2048;
        $E_range = &scaleMCA($mcaNMAX,$gain);
        printf STDOUT '!!! '.$script.': Requested energy range dE=%.2feV is wider'
                            .' than MCA could analyze=%.2feV.'."\n",$dE,$E_range;
        if    ($tooWide == 1) {
           $E_ = $E1;
           $E1 = $E2 - $E_range;
           printf STDOUT '!!! '.$script.': adjusting lower energy threshold from %.2feV to %.2feV'."\n",$E_,$E1;
        }
        elsif ($tooWide == 2) {
           $E_ = $E2;
           $E2 = $E1 + $E_range;
           printf STDOUT '!!! '.$script.': adjusting upper energy threshold from %.2feV to %.2feV'."\n",$E_,$E2;
        }
        else  {&press_key_exit('*** '.$script.': incorrect tooWide flag='.$tooWide);}
        $E_center = ($E2 + $E1) / 2.0;	#in eV (new after shrinking the range)
        $dE       = ($E2 - $E1);		#in eV (new after shrinking the range)
        if ($report) {
           printf STDOUT $script.' (adjusted): Need to optimize to E_center=%.2feV, dE=%.2feV'."\n",$E_center,$dE;
        }
        $range_changed = 1;
     } else {
        $E_range = &scaleMCA($mcaNMAX,$gain);
     }
     if ($report) {
        print  STDOUT $script.': found ADC gain='.$gain.' =>'."\n";
        printf STDOUT $script.': MCA_range=%.2feV (vs required dE=%.2feV)'
                             ."\n", $E_range, $dE;
     }

     ### Get scaling parameters for the new gain:
     $E_scale = &MCAcalibDB2006($gain);	#in eV/channel
     if ($report) {
        printf STDOUT $script.': for this gain E_scale=%.2feV/channel'
                             ."\n",$E_scale;
     }

     ### Find $offset at which $E_center appears at the center ($channel=1023)
     ### of the MCA (follows from: $E=$E_scale*$channel+$E_offset;)
     $E_offset = $E_center - $E_scale * ($mcaNMAX/2);	#in eV
     if ($report) {
        printf STDOUT $script.': E_offset=%.2feV (the ideal value needed '
                             .'for the peak to be exactly at center)'
                             ."\n",$E_offset;
     }

     ### This follows from: $E_offset=$E_scale*$offsetReal-$derekOffset;
     $offset     = ($E_offset + $derekOffset) / $E_scale;	  #in channels
     $offsetReal = $offsetStep * int ($offset/$offsetStep);  #in channels

     if ($offsetReal < 0 || $offsetReal > 16128) {  	#16128 = 128*126 or 256*63
#       &press_key_exit('*** '.$script.': calculated offset='
#                             .$offsetReal.' is not in range [0--16128]');
        print STDOUT '!!! '.$script.': calculated offset='.$offsetReal
                                    .' is not in range [0--16128]'."\n";
        print STDOUT '!!! '.$script.': trying to fix by placing requested '
                                    .'energy range somewhat off the '
                                    .'MCA scale center.'."\n";
        if    ($offsetReal < 0)     {$offsetReal=0;}
        elsif ($offsetReal > 16128) {$offsetReal=16128;}
     }

     ### Calculate E_offset with the account that MCA offset is discrete:
     $E_offset = $E_scale*$offsetReal-$derekOffset;
     if ($report) {
        printf STDOUT $script.': E_offset=%.2feV (the real offset '
                             .'accounting for MCA offset discreteness)'
                             ."\n", $E_offset;
     }

     ### Calculate channels for $E1 and $E2:  $E = $E_scale * $channel + $E_offset;
     $N1 = int (($E1 - $E_offset) / $E_scale);
     $N2 = int (($E2 - $E_offset) / $E_scale);
     if ($N1 < 0 || $N1 > $mcaNMAX-1) {
        &press_key_exit('*** '.$script.': calculated N1='.$N1
                              .' is not in range [0--'.$mcaNMAX.']');
        if    ($N1 < 0)          {$N1 = 0;}
        elsif ($N1 > $mcaNMAX-1) {$N1 = $mcaNMAX-1;}
     }
     if ($N2 < 0 || $N2 > $mcaNMAX-1) {
        &press_key_exit('*** '.$script.': calculated N2='.$N2
                              .' is not in range [0--'.$mcaNMAX.']');
        if    ($N2 < 0)          {$N2 = 0;}
        elsif ($N2 > $mcaNMAX-1) {$N2 = $mcaNMAX-1;}
     }
     if ($report) {print STDOUT $script.': N1='.$N1.' N2='.$N2."\n";}
 
     ### It is safe to overwrite $E1,$E2 because we are not using them anymore:
     $E1 = $E_scale *      0       + $E_offset;
     $E2 = $E_scale * ($mcaNMAX-1) + $E_offset;
     printf STDOUT $script.': Setting MCA ADC to the energy range'
                          .' [%.2f--%.2f]eV:'."\n", $E1, $E2;
     print  STDOUT $script.': Setting MCA ADC: gain='.$gain
                          .' offset='.$offsetReal."\n";
     &setMCA_ADC($adc,$gain,$offsetReal);
     if (!defined $range_changed) {return ($N1, $N2);}
     else                         {return ($N1, $N2, $E1, $E2);}		#return corrected enrgies too

  } else {								#Falcon

     print STDOUT '!!! '.$script.': nothing to do for Falcon.'."\n";
     return (0,0);

  }
}

#=================================================================

# sub  printMCA ($LUN, $num_pts, @Intensity) {
sub  printMCA ($$@) {

  my ($script, $LUN, $num_pts, $i, $intensity);
  $script  = 'printMCA';
  $LUN     = shift(@_);
  $num_pts = shift(@_);
  for ($i=0; $i<$num_pts; $i++) {
     $intensity = shift(@_);
     printf $LUN '%5.0f, %12.5g'."\n",$i,$intensity;
  }
}

#=================================================================

# sub getDerekMCAroi2dis($adc) {				#canberra only
sub getDerekMCAroi2dis($) {
#----------------------------------------------------
# This function defines global vars ($derekR2D_slope, $derekR2D_offlo, $derekR2D_offhi)
# needed for recalculating ROI channels into ADC and TCA discrimination:
#
#  $TCA_discr_low  = 100% * ($derekR2D_slope * (($MCA_roi_lo + $ADC_offset) / $ADC_gain) - $derekR2D_offlo);
#  $TCA_discr_high = 100% * ($derekR2D_slope * (($MCA_roi_hi + $ADC_offset) / $ADC_gain) - $derekR2D_offhi);
#----------------------------------------------------
### GLOBAL - must be on the GMCA supplement to MCA calibration form:
  our ($derekR2D_slope, $derekR2D_offlo, $derekR2D_offhi);
  my ($script, $adc, $status, $pv);
  $script = 'getDerekMCAroi2dis';
  $adc    = shift(@_);

#######################################################################
### Derek's calibration ROI-to-Discrimination ($s, $lo, $hi):       ###
###  $discr_lo = 100 * ($s * (($roi_lo + $offset) / $gain) - $lo);  ###
###  $discr_hi = 100 * ($s * (($roi_hi + $offset) / $gain) - $hi);  ###
#######################################################################

### These calibrations are for Canberra only !!!

  if (defined $adc && $adc !~ /falcon/) {				#Canberra

     $adc =~ s/:$//;					#adc should not be ended by ":"

     if (!defined $derekR2D_slope) {
        $pv = $adc.':R2D:SLOPE';
        ($status, $derekR2D_slope) = Pezca::GetDouble($pv);
        if ($status) {&report_exit($script, $status, $pv); $derekR2D_slope=0;}
     }
     if (!defined $derekR2D_offlo) {
        $pv = $adc.':R2D:OFFLO';
        ($status, $derekR2D_offlo) = Pezca::GetDouble($pv);
        if ($status) {&report_exit($script, $status, $pv); undef $derekR2D_offlo;}
     }
     if (!defined $derekR2D_offhi) {
        $pv = $adc.':R2D:OFFHI';
        ($status, $derekR2D_offhi) = Pezca::GetDouble($pv);
        if ($status) {&report_exit($script, $status, $pv); undef $derekR2D_offhi;}
     }

     ### We should not be using this:
     if ($derekR2D_slope == 0 || !defined $derekR2D_offlo || !defined $derekR2D_offhi) {
#       $derekR2D_slope = 1.0811;               # data by Derek 01/2006, ID-out?
#       $derekR2D_offlo = 0.0004;               # data by Derek 01/2006, ID-out?
#       $derekR2D_offhi = 0.0012;               # data by Derek 01/2006, ID-out?
        $derekR2D_slope = 1.0611254;            # data by Derek 01/2008, what ID?
        $derekR2D_offlo = 0.0;                  # data by Derek 01/2008, what ID?
        $derekR2D_offhi = 0.0;                  # data by Derek 01/2008, what ID?
        print STDOUT "\n".'!!! '.$script
                    .': no ROI-to-discrimination calibration for'."\n",
                     'MCA_ADC='.$adc.' in '.$adc
                    .':R2D:SLOPE, ..:OFFLO, ..:OFFHI EPICS PVs.'."\n",
                     'Using potentially obsolete embedded data:'."\n",
                     'MCA_R2D_slope     = '.$derekR2D_slope."\n",
                     'MCA_R2D_offset_lo = '.$derekR2D_offlo."\n",
                     'MCA_R2D_offset_hi = '.$derekR2D_offhi."\n\n";
     }
  } else {								#Falcon

     ### No Derek's corrections in case of Falcon
     print STDOUT '!!! '.$script.': nothing to do for Falcon.'."\n";
     $derekR2D_slope = 0;
     $derekR2D_offlo = 0;
     $derekR2D_offhi = 0;

  }
  return $derekR2D_slope,$derekR2D_offlo,$derekR2D_offhi;
}

#=================================================================

# sub calcDiscrFromROI($mca,$adc) {		#for fast scans with Struck
sub calcDiscrFromROI($$) {						#Canberra only

### This program returns ($low,$high) TCA & ADC discrimination levels
###				      ^^^^^^^^^
### This program is used when we are feeding the fluorescent detector
### output into Struck, In this case we are bypassing the MCA and
### the only way to filter required fluorescent radiation is to set
### low/high bandwidth filters on the ADC and TCA.
### The task of this program is to calculate the TCA low/high bandwidth
### filters values based on current MCA ROI and ADC gain and offset.

### GLOBAL - must be on the GMCA supplement to MCA calibration form:
  our ($derekR2D_slope, $derekR2D_offlo, $derekR2D_offhi);
  my ($script, $mca, $adc);
  my ($roi_lo, $roi_hi, $gain, $offset, $low, $high);
  $script = 'calcDiscrFromROI';
  $mca    = shift(@_);
  $adc    = shift(@_);

  ($roi_lo, $roi_hi) = &checkMCA_ROI($mca);

  if (defined $adc && $mca !~ /falcon/) {				#Canberra

     if (!defined $derekR2D_slope ||
         !defined $derekR2D_offlo ||
         !defined $derekR2D_offhi)   {&getDerekMCAroi2dis($adc);}

     ($gain, $offset)   = &getMCA_ADC($adc);

     if ($gain < 256 || $gain > 16384) {
       &press_key_exit('*** '.$script.': gain='.$gain
                             .' is not in range [256--16384]');
     }

     $low  = 100 * ($derekR2D_slope * (($roi_lo + $offset) / $gain) - $derekR2D_offlo);
     $high = 100 * ($derekR2D_slope * (($roi_hi + $offset) / $gain) - $derekR2D_offhi);

     if ($low < 0. || $low >= $high || $high > 100.) {
        print STDOUT '!!! '.$script.': calculated invalid discr.pair: low='
                                  .$low.' high='.$high."\n";
        if ($low  <  0.  ) {$low  = 0.;      }
        if ($high <= $low) {$high = $low+1.; }
        if ($high >  100.) {$high = 100.;    }
        if ($high <= $low) {$low  = $high-1.;}
        print STDOUT '!!! '.$script.': the corrected discr.pair: low='
                                   .$low.' high='.$high."\n";
     }
     
  } else {								#Falcon

#    print STDOUT '!!! '.$script.': nothing to do for Falcon.'."\n";
     ($low, $high) = ($roi_lo, $roi_hi);

  }
  return ($low, $high);
}

# 31.                                          &convertROI2discr($mca,$adc,$tca);

#=================================================================

# sub setDiscrFromROI($mca,$adc,$tca) {		#for fast scans with Struck
sub setDiscrFromROI($$$) {

### This program calculates and SETs ($low,$high) discrimination on TCA & ADC
### that should be equivalent to given ($ROI_lo,$ROI_hi,$gain,$offset) on MCA.
###				      ^^^^^^^^^
### This program is used when we are feeding the fluorescent detector
### output into Struck, In this case we are bypassing the MCA and
### the only way to filter required fluorescent radiation is to set
### low/high bandwidth filters on the ADC and TCA.

  my ($script, $mca, $adc, $tca);
  my ($low, $high);
  $script = 'setDiscrFromROI';
  $mca    = shift(@_);
  $adc    = shift(@_);
  $tca    = shift(@_);

  if (defined $adc && defined $tca && $mca !~ /falcon/) {		#Canberra

     ($low,$high) = &calcDiscrFromROI($mca,$adc);

     ### Commented out 2010/06/03 per Derek's request: should
     ### only set discrimination in TCA:
     # printf STDOUT $script.': setting ADC discrimination: low='
     #                      .'%3.2f, high=%3.2f'."\n", $low, $high;
     # &setMCA_ADC_discr($adc,$low,$high);
     printf STDOUT $script.': setting TCA discrimination: low='
                          .'%3.2f, high=%3.2f'."\n", $low, $high;
     &setMCA_TCA($tca,$low,$high);

  } else {								#Falcon

#    print STDOUT '!!! '.$script.': nothing to do for Falcon.'."\n";
     &setFalconROI2SCA($mca);

  }
}

#=================================================================

# $status = &getEdgeDataSQL($element,$edge,\$EdgeEnergy,\@EnStep,\@EnPM,\$FluoEnergy,\$FluoPMwidth);
# sub getEdgeDataSQL ($element,$edge,\$EdgeEnergy,\@EnStep,\@EnPM,\$FluoEnergy,\$FluoPMwidth) {
sub getEdgeDataSQL ($$\$\@\@\$\$) {

### This routine takes ($element,$edge) as keywords and then searches mySQL
### database for respective data of:
### ---  $EdgeEnergy,
### ---  arrays of @EnStep and @EnPM (typically three intervals at either
###      side of the edge, the first two have symmetric ranges and farthermost
###      is asymmetric, so we read $EnStep1,$EnStep2,$EnStep3, $EnPM1,$EnPM2,
###      $EnPM3,$EnPM4).
### ---  $FluoEnergy,$FluoPMwidth, which are the fluorescent peak position
###      and width on MCA, respectively.

  our ($mysqlHost, $mysqlDb, $mysqlUser, $mysqlPass);
  my ($script, $status, $element, $edge, $mysqlTbl, $nrecords);
  my ($EdgeEnRef, $EnStepRef, $EnPMRef, $FluoEnRef, $FluoPMRef);
  my ($dsn, $dbh, $sth, @val, @valMaster, $nval, $needMaster, $i);
  my (@names, $name, $pointer);

  $script    = 'getEdgeDataSQL';
  $element   = shift(@_);
  $edge      = shift(@_);
  $EdgeEnRef = shift(@_); 			#in eV
  $EnStepRef = shift(@_);			#in eV
  $EnPMRef   = shift(@_);			#in eV
  $FluoEnRef = shift(@_);			#in eV
  $FluoPMRef = shift(@_);			#in eV
  $mysqlTbl  = 'EdgeData';
### Record structure:
  @names = ('Element','Edge','EdgeEnergy','FluoEnergy',
            'E_range1','E_step1','E_range2','E_step2','E_range3','E_step3','E_range4',
            'FluoRange','graphDisplay');

  $status = 0;
  if (!defined $element)  {
     print STDOUT '*** '.$script.': no Element name passed'."\n";
     $status++;
  }
  if (!defined $edge)  {
     print STDOUT '*** '.$script.': no Edge name passed'."\n";
     $status++;
  }
  if ($status) {return $status;}

  $dsn = 'DBI:mysql:host='.$mysqlHost.';database='.$mysqlDb;
  $dbh = DBI->connect($dsn,$mysqlUser,$mysqlPass,{PrintError=>0,});
  if (!defined $dbh) {
     print STDOUT '*** '.$script.': Cannot connect to mySQL server '
                                .$mysqlHost."\n";
     return -1;
  }
  $sth = $dbh->prepare('SELECT * FROM '.$mysqlTbl
                      .' WHERE Element="'.$element.'" '
                      .' AND Edge="'.$edge.'" ');
  $nrecords = $sth->execute();				#nrecords works for MySQL only. For SQLite result is def/undef only.
  if (!defined $nrecords) {
     print STDOUT '*** '.$script.': Failed reading data for'."\n",
                  '***       Element='.$element.' and Edge='.$edge."\n",
                  '***       in table '.$mysqlDb.'.'.$mysqlTbl."\n",
                  '***       at mySQL server '.$mysqlHost."\n";
     $sth->finish();
     $dbh->disconnect();
     return -1;
  }
  elsif ($nrecords == 0) {
     print STDOUT '*** '.$script.': No records found for'."\n",
                  '***       Element='.$element.' and Edge='.$edge."\n",
                  '***       in table '.$mysqlDb.'.'.$mysqlTbl."\n",
                  '***       at mySQL server '.$mysqlHost."\n";
     $sth->finish();
     $dbh->disconnect();
     return -1;
  }
  elsif ($nrecords > 1) {
     print STDOUT '*** '.$script.': Multiple records ('.$nrecords.') found for'."\n",
                  '***       Element='.$element.' and Edge='.$edge."\n",
                  '***       in table '.$mysqlDb.'.'.$mysqlTbl."\n",
                  '***       at mySQL server '.$mysqlHost."\n";
     $sth->finish();
     $dbh->disconnect();
     return -1;
  }
# @val = $sth->fetchrow_array();
  $pointer = $sth->fetchrow_hashref();
  $sth->finish();
  @val = ();
  foreach $name (@names) {
     if (defined $pointer->{$name}) {push (@val, $pointer->{$name});}
     else  {
        print STDOUT '*** '.$script.': Incorrect record for'."\n",
                     '***       Element='.$element.' and Edge='.$edge."\n",
                     '***       in table '.$mysqlDb.'.'.$mysqlTbl."\n",
                     '***       at mySQL server '.$mysqlHost.' .'."\n",
                     '***       No data for column '.$name."\n";
        $sth->finish();
        $dbh->disconnect();
        return -1;
     }
  }
### The mySQL record structure is:
### Element Edge EdgeEnergy FluoEnergy E_range1 E_step1 E_range2 E_step2 E_range3 E_step3 E_range4 FluoRange graphDisplay
###   0      1      2          3          4       5        6       7        8        9      10       11       12
  if ($val[10] == 0) {$val[10] = $val[8]}		#E_range4 = E_range3

### Check if we need to read master record:
  $needMaster = 0;
  $nval = @val;
  for ($i=2; $i<$nval; $i++) {
     if ($val[$i] == 0) {$needMaster = 1; last;}
  }

### Read master data:
  if ($needMaster) {
     $sth = $dbh->prepare('SELECT * FROM '.$mysqlTbl.' WHERE Element="master" AND Edge="master" ');
     $nrecords = $sth->execute();				#nrecords works for MySQL only. For SQLite result is def/undef only.
     if (!defined $nrecords) {
        print STDOUT '*** '.$script.': Failed reading data for'."\n",
                     '***       Element=master and Edge=master'."\n",
                     '***       in table '.$mysqlDb.'.'.$mysqlTbl."\n",
                     '***       at mySQL server '.$mysqlHost."\n";
        $sth->finish();
        $dbh->disconnect();
        return -1;
     }
     elsif ($nrecords == 0) {
        print STDOUT '*** '.$script.': No records found for'."\n",
                     '***       Element=master and Edge=master'."\n",
                     '***       in table '.$mysqlDb.'.'.$mysqlTbl."\n",
                     '***       at mySQL server '.$mysqlHost."\n";
        $sth->finish();
        $dbh->disconnect();
        return -1;
     }
     elsif ($nrecords > 1) {
        print STDOUT '*** '.$script.': Multiple records ('.$nrecords.') found for'."\n",
                     '***       Element=master and Edge=master'."\n",
                     '***       in table '.$mysqlDb.'.'.$mysqlTbl."\n",
                     '***       at mySQL server '.$mysqlHost."\n";
        $sth->finish();
        $dbh->disconnect();
        return -1;
     }
#    @valMaster = $sth->fetchrow_array();
#    $nval = @valMaster;
     $pointer = $sth->fetchrow_hashref();
     $sth->finish();
     @valMaster = ();
     foreach $name (@names) {
        if (defined $pointer->{$name}) {push (@valMaster, $pointer->{$name});}
        else  {
           print STDOUT '*** '.$script.': Incorrect record for'."\n",
                        '***       Element=master and Edge=master'."\n",
                        '***       in table '.$mysqlDb.'.'.$mysqlTbl."\n",
                        '***       at mySQL server '.$mysqlHost.' .'."\n",
                        '***       No data for column '.$name."\n";
           $sth->finish();
           $dbh->disconnect();
           return -1;
        }
     }
     if ($valMaster[10] == 0) {$valMaster[10] = $valMaster[8]}	#E_range4 = E_range3
     for ($i=2; $i<$nval; $i++) {
        if ($val[$i] == 0) {$val[$i] = $valMaster[$i];}
     }
  }
  $dbh->disconnect();

  $$EdgeEnRef = $val[2];
  $$FluoEnRef = $val[3];
  @$EnPMRef   = ($val[4],$val[6],$val[8],$val[10]);
  @$EnStepRef = ($val[5],$val[7],$val[9]);
  $$FluoPMRef = $val[11];

  return 0;
}

#=================================================================

# $nEdges = &getEdgeListSQL(\@ElementEdge,\@EdgeEnergy);
# sub getEdgeListSQL (\@ElementEdge,\@EdgeEnergy) {
sub getEdgeListSQL (\@\@) {

### This routine fills Element-Edge and EdgeEnergy arrays with data from BluIce mySQL DB

  our ($mysqlHost, $mysqlDb, $mysqlUser, $mysqlPass);
  my ($script, $mysqlTbl, $nColumns);
  my ($ElementEdgeRef, $EdgeEnergyRef, $nRecords);
  my ($dsn, $dbh, $sth, @val, $nval, $i);

  $script         = 'getEdgeListSQL';
  $ElementEdgeRef = shift(@_);
  $EdgeEnergyRef  = shift(@_);			#in eV
  $mysqlTbl  = 'EdgeData';
### Record structure:
### (Element,Edge),(EdgeEnergy,FluoEnergy),(E_range1,E_step1,E_range2,E_step2,E_range3,E_step3,E_range4),FluoRange,graphDisplay
  $nColumns  = 2+2+7+1+1;

  $dsn = 'DBI:mysql:host='.$mysqlHost.';database='.$mysqlDb;
  $dbh = DBI->connect($dsn,$mysqlUser,$mysqlPass,{PrintError=>0,});
  if (!defined $dbh) {
     print STDOUT '*** '.$script.': Cannot connect to mySQL server '.$mysqlHost."\n";
     return -1;
  }
  $sth = $dbh->prepare('SELECT * FROM '.$mysqlTbl.' WHERE NOT Element="master" AND NOT Edge="master" ');
# $sth = $dbh->prepare('SELECT * FROM '.$mysqlTbl.' WHERE NOT Element="master" ');
# $sth = $dbh->prepare('SELECT * FROM '.$mysqlTbl);
  $nRecords = $sth->execute();				#nrecords works for MySQL only. For SQLite result is def/undef only.
  if (!defined $nRecords) {
     print STDOUT '*** '.$script.': Failed reading data list from'."\n",
                  '***       table '.$mysqlDb.'.'.$mysqlTbl."\n",
                  '***       at mySQL server '.$mysqlHost."\n";
     $sth->finish();
     $dbh->disconnect();
     return -1;
  }
  elsif ($nRecords == 0) {
     print STDOUT '!!! '.$script.': No records found in'."\n",
                  '!!!       table '.$mysqlDb.'.'.$mysqlTbl."\n",
                  '!!!       at mySQL server '.$mysqlHost."\n";
     $sth->finish();
     $dbh->disconnect();
     return $nRecords;
  }
  for ($i=0; $i<$nRecords; $i++) {
     @val = $sth->fetchrow_array();
     $nval = @val;
     if ($nval != $nColumns) {
        print STDOUT '*** '.$script.': Incorrect record ['."@{val}".'] in'."\n",
                     '***       table '.$mysqlDb.'.'.$mysqlTbl."\n",
                     '***       at mySQL server '.$mysqlHost.' .'."\n",
                     '***       Expected '.$nColumns.' columns, but found '.$nval.'.'."\n";
        $sth->finish();
        $dbh->disconnect();
        return -1;
     }
### The mySQL record structure is:
### Element Edge EdgeEnergy FluoEnergy E_range1 E_step1 E_range2 E_step2 E_range3 E_step3 E_range4 FluoRange graphDisplay
###   0      1      2          3          4       5        6       7        8        9      10       11       12

     @$ElementEdgeRef[$i] = $val[0].'-'.$val[1];
     @$EdgeEnergyRef[$i]  = $val[2];
  }
  $sth->finish();
  $dbh->disconnect();
  return $nRecords;
}

#=================================================================

# $nLines = &getFluoListSQL(\@ElementLine,\@FluoEnergy);
# sub getFluoListSQL (\@ElementLine,\@FluoEnergy) {
sub getFluoListSQL (\@\@) {

### This routine fills Element-Edge and FluoEnergy arrays with data from BluIce mySQL DB

  our ($mysqlHost, $mysqlDb, $mysqlUser, $mysqlPass);
  my ($script, $mysqlTbl, $nColumns);
  my ($ElementLineRef, $FluoEnergyRef, $nRecords);
  my ($dsn, $dbh, $sth, @val, $nval, $i);

  $script         = 'getFluoListSQL';
  $ElementLineRef = shift(@_);
  $FluoEnergyRef  = shift(@_);			#in eV
  $mysqlTbl  = 'EdgeData';
### Record structure:
### (Element,Edge),(EdgeEnergy,FluoEnergy),(E_range1,E_step1,E_range2,E_step2,E_range3,E_step3,E_range4),FluoRange,graphDisplay
  $nColumns  = 2+2+7+1+1;

  $dsn = 'DBI:mysql:host='.$mysqlHost.';database='.$mysqlDb;
  $dbh = DBI->connect($dsn,$mysqlUser,$mysqlPass,{PrintError=>0,});
  if (!defined $dbh) {
     print STDOUT '*** '.$script.': Cannot connect to mySQL server '.$mysqlHost."\n";
     return -1;
  }
  $sth = $dbh->prepare('SELECT * FROM '.$mysqlTbl.' WHERE NOT Element="master" AND NOT Edge="master" ');
  $nRecords = $sth->execute();				#nrecords works for MySQL only. For SQLite result is def/undef only.
  if (!defined $nRecords) {
     print STDOUT '*** '.$script.': Failed reading data list from'."\n",
                  '***       table '.$mysqlDb.'.'.$mysqlTbl."\n",
                  '***       at mySQL server '.$mysqlHost."\n";
     $sth->finish();
     $dbh->disconnect();
     return -1;
  }
  elsif ($nRecords == 0) {
     print STDOUT '!!! '.$script.': No records found in'."\n",
                  '!!!       table '.$mysqlDb.'.'.$mysqlTbl."\n",
                  '!!!       at mySQL server '.$mysqlHost."\n";
     $sth->finish();
     $dbh->disconnect();
     return $nRecords;
  }
  for ($i=0; $i<$nRecords; $i++) {
     @val = $sth->fetchrow_array();
     $nval = @val;
     if ($nval != $nColumns) {
        print STDOUT '*** '.$script.': Incorrect record ['."@{val}".'] in'."\n",
                     '***       table '.$mysqlDb.'.'.$mysqlTbl."\n",
                     '***       at mySQL server '.$mysqlHost.' .'."\n",
                     '***       Expected '.$nColumns.' columns, but found '.$nval.'.'."\n";
        $sth->finish();
        $dbh->disconnect();
        return -1;
     }
### The mySQL record structure is:
### Element Edge EdgeEnergy FluoEnergy E_range1 E_step1 E_range2 E_step2 E_range3 E_step3 E_range4 FluoRange graphDisplay
###   0      1      2          3          4       5        6       7        8        9      10       11       12

     @$ElementLineRef[$i] = $val[0].'-'.$val[1];
     @$FluoEnergyRef[$i]  = $val[3];
  }
  $sth->finish();
  $dbh->disconnect();
  return $nRecords;
}

#=================================================================

# $status = &checkFalcon2BluiceConsistency();
sub checkFalcon2BluiceConsistency() {

### This routine checks that Pezca and BluIce mySQL DB about using Falcon coinside

  our ($USE_FALCON, $mysqlHost, $mysqlDb, $mysqlUser, $mysqlPass);
  my ($script, $mysqlTbl, $dsn, $dbh, $sth, $pointer, $nRecords);
  my ($key, $mca, $JBL_FALCON);

  $script         = 'checkFalcon2BluiceConsistency';
  if (! defined $USE_FALCON) {$USE_FALCON = 0;}

  $mysqlTbl  = 'PVPrefix';
  $key = 'mca';

  print STDOUT 'Checking consistency of MCA choice between PEZCA and JBluIce...'."\n";

  $dsn = 'DBI:mysql:host='.$mysqlHost.';database='.$mysqlDb;
  $dbh = DBI->connect($dsn,$mysqlUser,$mysqlPass,{PrintError=>0,});
  if (!defined $dbh) {
     print STDOUT '*** '.$script.': Cannot connect to mySQL server '.$mysqlHost."\n";
     return -1;
  }
  $sth = $dbh->prepare('SELECT * FROM '.$mysqlTbl.' WHERE Label="'.$key.'"');
  $nRecords = $sth->execute();				#nrecords works for MySQL only. For SQLite result is def/undef only.
  if (!defined $nRecords) {
     print STDOUT '*** '.$script.': Failed reading data for Label="'.$key.'"'."\n",
                  '***       for table '.$mysqlDb.'.'.$mysqlTbl."\n",
                  '***       at mySQL server '.$mysqlHost."\n";
     $sth->finish();
     $dbh->disconnect();
     return -1;
  }
  elsif ($nRecords == 0) {
     print STDOUT '*** '.$script.': No records found for Label="'.$key.'"'."\n",
                  '***       in table '.$mysqlDb.'.'.$mysqlTbl."\n",
                  '***       at mySQL server '.$mysqlHost."\n";
     $sth->finish();
     $dbh->disconnect();
     return -1;
  }
  elsif ($nRecords > 1) {
     print STDOUT '*** '.$script.': Multiple ('.$nRecords.') records found for Label="'.$key.'"'."\n",
                  '***       in table '.$mysqlDb.'.'.$mysqlTbl."\n",
                  '***       at mySQL server '.$mysqlHost."\n";
     $sth->finish();
     $dbh->disconnect();
     return -1;
  }
  $pointer = $sth->fetchrow_hashref();
  $sth->finish();
  $dbh->disconnect();
  if (! defined $pointer->{Prefix}) {
     print STDOUT '*** '.$script.': Incorrect record for Label="'.$key.'"'."\n",
                  '***       in table '.$mysqlDb.'.'.$mysqlTbl."\n",
                  '***       at mySQL server '.$mysqlHost.' .'."\n",
                  '***       No data for key=Prefix'."\n";
     return -1;
  }
  $mca = $pointer->{Prefix};
  if ($USE_FALCON)      {print STDOUT 'MCA choice for PEZCA  = Falcon'."\n";}
  else                  {print STDOUT 'MCA choice for PEZCA  = Canberra'."\n";}
  if ($mca =~ /falcon/) {print STDOUT 'MCA choice for BluIce = Falcon'."\n";   $JBL_FALCON=1;}
  else                  {print STDOUT 'MCA choice for BluIce = Canberra'."\n"; $JBL_FALCON=0;}
  if ($USE_FALCON != $JBL_FALCON) {print STDOUT 'PEZCA and BluIce choices about MCA mismatch'."\n"; return 1;}
  else                            {return 0;}
}

#=================================================================

# sub getFalconCalib($mca) {                   #return slope and offset
sub getFalconCalib($) {
#----------------------------------------------------
# This function returns ($E_scale,$E_offset) for
# recalculating MCA channels into energy in [eV]:
#
#          $E(eV) = $E_scale * $channel + $E_offset;
# or:
#          $channel = ($E - $E_offset) / $E_scale;
#
# where 
# ==> for Falcon ONLY!!!:
#
# ==> for Canberra: see getMCAcalib
#
#----------------------------------------------------
  my ($script, $mca, $mcapv, $pv, $status);
  my ($E_scale, $E_offset);		#returned parmaters (output)
  $script = 'getFalconCalib';
  $mca    = shift(@_);

  if (defined $mca && $mca =~ /falcon/) { #---------------------------------------------Canberra
     if ($mca !~ /:$/) {$mca .= ':';}
     $mcapv = $mca.'mca1';
     
     $pv = $mcapv.'.CALS';
     ($status, $E_scale) = Pezca::GetDouble($pv);
     if ($status) {&report_exit($script, $status, $pv); $E_scale=0;}

     $pv = $mcapv.'.CALO';
     ($status, $E_offset) = Pezca::GetDouble($pv);
     if ($status) {&report_exit($script, $status, $pv); $E_offset=0;}

  } else { #------------------------------------------------------------Falcon

     print STDOUT '!!! '.$script.': nothing to do for Canberra.'."\n";
     $E_scale  = 0;
     $E_offset = 0;

  }     
  return ($E_scale, $E_offset);
}

#=================================================================

# ($N1,$N2) = &setFalconROI($mca,$E1,$E2,[$N,$E_scale,$E_offset]);      #in eV
  sub setFalconROI($$$;$$$) {                                            #returns ($N1,$N2) for ROU-$N
#----------------------------------------------------
# This function returns $E_scale for given gain that can be
# used for recalculating MCA channels into energy in [eV]:
#
#                 $E(eV) = $E_scale * $channel + $E_offset;
# where:
# ==> for Canberra:
#           $E_scale(eV) = $derekSlope * (2048/$gain);
#	   $E_offset(eV) = $E_scale * $offsetReal - $derekOffset;
#        $offsetReal(eV) = $offsetStep*[int($offset/$offsetStep)]
# and $channel is 0-2047 and $offsetStep=128
#----------------------------------------------------
  my ($script, $mca, $E1, $E2, $ROI_num, $E_scale, $E_offset);
  my ($N1, $N2, $E_);
  $script   = 'setFalconROI';
  $mca      = shift(@_);
  $E1       = shift(@_);
  $E2       = shift(@_);
  $ROI_num  = shift(@_);   if (!defined $ROI_num) {$ROI_num = 0;}       #default ROI number
  $E_scale  = shift(@_);
  $E_offset = shift(@_);

  if ($mca =~ /falcon/) {						#Falcon

     if (!defined $E_scale || !defined $E_offset) {
        ($E_scale, $E_offset) = &getFalconCalib($mca);
     }
     if (!defined $E_scale || $E_scale == 0) {
        &press_key_exit('*** '.$script.': zero or undefined E_scale');
     }

     if ($E2 < $E1) {$E_=$E2; $E2=$E1; $E1=$E_;}

     $N1 = int(($E1 - $E_offset) / $E_scale + 0.5);
     $N2 = int(($E2 - $E_offset) / $E_scale + 0.5);

     &setMCA_ROI ($mca, $N1, $N2, $ROI_num);

  } else {								#Falcon

     print STDOUT '!!! '.$script.': nothing to do for Canberra'."\n";
     $N1 = 0;
     $N2 = 0;

  }
  return ($N1, $N2);
}

#=================================================================

# $status = &setFalconSCA($mca,$sca_lo,$sca_hi,[$N);			#in channel numbers
  sub setFalconSCA($$$;$) {
  my ($script, $mca, $sca_lo, $sca_hi, $N);
  my ($dxppv, $pv, $mcaNMAX, $status), 
  $script   = 'setFalconSCA';
  $mca      = shift(@_);
  $sca_lo   = shift(@_);
  $sca_hi   = shift(@_);
  $N        = shift(@_);  
  if (!defined $N) {$N=0;}
  if ($N < 0 || $N > 2) {print STDOUT '*** '.$script.': incorrect N='.$N."\n"; return -1;}

  if ($mca =~ /falcon/) {						#Falcon

     ### Read the maximum possible number of channels:
     $mcaNMAX = &MCAnmax($mca);

     if ($mca !~ /:$/) {$mca.=':';}					#falcon must be ended by ":"
     $dxppv = $mca.'dxp1';

     if ($sca_lo < 0 || $sca_lo > $mcaNMAX-1) {
        print STDOUT '*** '.$script.': Low SCA-'.$N.' = '.$sca_lo.' is not in valid range'."\n";
        return -1;
     }
     if ($sca_hi < 0 || $sca_hi > $mcaNMAX-1) {
        print STDOUT '*** '.$script.': High SCA-'.$N.' = '.$sca_hi.' is not in valid range'."\n";
        return -1;
     }
     if ($sca_lo >= $sca_hi) {
        print STDOUT '*** '.$script.': Incorrect SCA-'.$N.': LO='.$sca_lo.' >= HI='.$sca_hi."\n";
        return -1;
     }
     ### re-set SCA-Low to "0"
     $pv = $dxppv.':SCA'.$N.'Low';			#e.g. ":SCA0Low"
     $status = Pezca::PutDoubleOldCa($pv,0);
     if ($status) {&report_only($script, $status, $pv); return -1;}
     ### set SCA-High:
     $pv = $dxppv.':SCA'.$N.'High';			#e.g. ".SCA0High"
     $status = Pezca::PutDoubleOldCa($pv,$sca_hi);
     if ($status) {&report_only($script, $status, $pv); return -1;}
     ### set SCA-Low
     $pv = $dxppv.':SCA'.$N.'Low';			#e.g. ":SCA0Low"
     $status = Pezca::PutDoubleOldCa($pv,$sca_lo);
     if ($status) {&report_only($script, $status, $pv); return -1;}
     return 0;

  } else {								#Falcon

     print STDOUT '!!! '.$script.': nothing to do for Canberra'."\n";
     return -1;

  }
}

#=================================================================

# $status = &setFalconROI2SCA($mca);
  sub setFalconROI2SCA($) {
  my ($script, $mca);
  my ($pv, $status), 
  $script   = 'setFalconROI2SCA';
  $mca      = shift(@_);

  if ($mca =~ /falcon/) {						#Falcon

     if ($mca !~ /:$/) {$mca.=':';}					#falcon must be ended by ":"
### copy ROI-LO and ROI-HI to SCA-Low and SCA-High:
     $pv = $mca.'CopyROI_SCA';
     $status = Pezca::PutDoubleOldCa($pv,1);
     &report_exit($script, $status, $pv);
     return $status;

  } else {								#Falcon

     print STDOUT '!!! '.$script.': nothing to do for Canberra'."\n";
     return -1;

  }
}

#=================================================================

# sub FalconRoi2energy($roi,$mca,[$E_scale,$E_offset]) {
sub FalconRoi2energy($$;$$) {
#----------------------------------------------------
# This function recalculates MCA channel or ROI border into Energ(eV)
#----------------------------------------------------
  my ($script, $E, $roi, $mca, $E_scale, $E_offset);
  $script  = 'FalconRoi2energy';
  $roi      = shift(@_);
  $mca      = shift(@_);
  $E_scale  = shift(@_);
  $E_offset = shift(@_);

  if (defined $mca && $mca =~ /falcon/) {				#Falcon

     if (!defined $E_scale || !defined $E_offset) {($E_scale,$E_offset) = &getFalconCalib($mca);}

     if ($E_scale != 0) {$E = $E_scale * $roi + $E_offset;}
     else               {&press_key_exit('*** '.$script.': Zero E_scale'); $E=0;}

  } else {								#Canberra

     print STDOUT '!!! '.$script.': nothing to do for Canberra.'."\n";
     $E = 0;

  }

  return $E;								#in eV
}

#=================================================================

# sub openMedmCanberraMCA ($mca) {
sub openMedmCanberraMCA ($) {

  our ($beamline);
  my ($script, $mca);
  my ($medmPar, $adl, $path, $status);

  $script = 'openMedmCanberraMCA';
  $mca    = shift(@_);

  if ($^O !~ /Win/i) {			 				#if UNIX
     if    (defined $ENV{PEZCA})     {$path=$ENV{PEZCA};}
     elsif (defined $ENV{FASTSCANS}) {$path=$ENV{FASTSCANS};}
     elsif (-e '/home/gmca/epics')   {$path='/home/gmca/epics';}
     else                            {$path='.';}
     $path =~ s/\\/\//g;						#change \ to /
     $path =~ s/\/$//;							#remove end /
     $path =~ s/\/bin$//;						#remove end bin/
     $path =~ s/\/fastscans$//;						#remove end fastscans/
     $path =~ s/\/pezca$//;						#remove end pezca/
     $adl = $path.'/mcaApp/adl/mca.adl';
     $mca =~ s/^${beamline}://;
     $medmPar = 'P='.$beamline.':,M='.$mca;
     $status = &open_medm($adl,$medmPar);
  }
}

#=================================================================

# sub openMedmCanberraADC ($adc) {
sub openMedmCanberraADC ($) {

  our ($beamline);
  my ($script, $adc);
  my ($medmPar, $adl, $path, $status);

  $script = 'openMedmCanberraADC';
  $adc    = shift(@_);

  if ($^O !~ /Win/i) {			 				#if UNIX
     if    (defined $ENV{PEZCA})     {$path=$ENV{PEZCA};}
     elsif (defined $ENV{FASTSCANS}) {$path=$ENV{FASTSCANS};}
     elsif (-e '/home/gmca/epics')   {$path='/home/gmca/epics';}
     else                            {$path='.';}
     $path =~ s/\\/\//g;						#change \ to /
     $path =~ s/\/$//;							#remove end /
     $path =~ s/\/bin$//;						#remove end bin/
     $path =~ s/\/fastscans$//;						#remove end fastscans/
     $path =~ s/\/pezca$//;						#remove end pezca/
     $adl = $path.'/mcaApp/adl/xxIcbAdc.adl';
     $adc =~ s/^${beamline}://;
     $medmPar = 'P='.$beamline.':,ADC='.$adc;
     $status = &open_medm($adl,$medmPar);
  }
}

#=================================================================

# sub openMedmCanberraTCA ($tca) {
sub openMedmCanberraTCA ($) {

  our ($beamline);
  my ($script, $tca);
  my ($medmPar, $adl, $path, $status);

  $script = 'openMedmCanberraTCA';
  $tca    = shift(@_);

  if ($^O !~ /Win/i) {			 				#if UNIX
     if    (defined $ENV{PEZCA})     {$path=$ENV{PEZCA};}
     elsif (defined $ENV{FASTSCANS}) {$path=$ENV{FASTSCANS};}
     elsif (-e '/home/gmca/epics')   {$path='/home/gmca/epics';}
     else                            {$path='.';}
     $path =~ s/\\/\//g;						#change \ to /
     $path =~ s/\/$//;							#remove end /
     $path =~ s/\/bin$//;						#remove end bin/
     $path =~ s/\/fastscans$//;						#remove end fastscans/
     $path =~ s/\/pezca$//;						#remove end pezca/
     $adl = $path.'/mcaApp/adl/xxIcbTCA.adl';
     $tca =~ s/^${beamline}://;
     $medmPar = 'P='.$beamline.':,TCA='.$tca;
     $status = &open_medm($adl,$medmPar);
  }
}
