#!/usr/bin/perl -W
#--------------------------------------------------------------
#        These are common subs for dealing with gnuplot
#
#     	             Author: Sergey Stepanov
#
# Version-1.0:  2004/09/09
# Version-2.0:  2007/07/31 added possibility to redefine GNUPLOT with global $GNUPLOT_FH
# Version-3.0:  2008/02/05 added support for different gnuplot versions (global $GNUPLOT_VERSION).
#                          The default version is 4.0.
# Version-3.1   2019/12/12 added &gnuplot_viewer and &image_gnuplot
#--------------------------------------------------------------
# List of functions:
# ^^^^^^^^^^^^^^^^^
# 1. $gnuplot = &gnuplot_executable();
# 2. $viewer  = &gnuplot_viewer();
# 3. $success = &open_gnuplot([$scan_start],[$scan_end],[$logscale],[\@gnuargs]);      -- uses LUN=GNUPLOT
# 4. $success = &open_gnuplot_fh($LUN,[$scan_start],[$scan_end],[$logscale],[\@gnuargs]);
# 5. $printed = &print_gnuplot([$autoplot,$GUIprompt]);                                -- uses LUN=GNUPLOT, autoplot=1|0|-1
# 6. $printed = &print_gnuplot_fh($LUN,[$autoplot,$GUIprompt]);                        -- autoplot=1|0|-1
# 7. $status  = &image_gnuplot($file,[$width,$height,$view,$GUIprompt]);               -- view=1 to call image viewer
# 8. $status  = &view_gnuplot_image($file);
# 9. $status  = &image_gnuplot_fh($LUN,$file,[$width,$height,$view,$GUIprompt]);       -- view=1 to call image viewer
#10.            &close_gnuplot();                                                      -- uses LUN=GNUPLOT
#11.            &close_gnuplot_fh($LUN);
#12.            &gnuplot_filess_delay([$milliseconds])				       -- default is 1ms
#--------------------------------------------------------------

# use warnings::unused;
  use strict;
# no strict 'vars';
# no strict 'subs';
# no strict 'refs';

#=================================================================

# sub gnuplot_executable() {		#returns the name of gnuplot pipe executable
sub gnuplot_executable() {
  my ($script, $gnuplot);
  $script = 'gnuplot_executable';
  if ($^O !~ /Win/i) {
     $gnuplot = &Which('gnuplot');	       				# UNIX
  } else {
     $gnuplot = &Which('pgnuplot.exe');					# Windows -- use before GnuPlot5
     if (!defined $gnuplot) {$gnuplot = &Which('gnuplot.exe');}		# Windows -- use with GnuPlot5 (no pgnuplot)
  }
  if ( !defined $gnuplot ) {
     print STDERR '!!! '.$script.': cannot find gnuplot executable'."\n";
  }
  return $gnuplot;
}

#=================================================================

# sub gnuplot_viewer() {		#returns the name of gnuplot pipe executable
sub gnuplot_viewer() {
  my ($script, $viewer);
  $script = 'gnuplot_viewer';
  if ($^O !~ /Win/i) {
     $viewer = &Which('feh');						# feh
     if (!defined $viewer) {$viewer = &Which('eog');}			# eye of Gnome
     if (!defined $viewer) {$viewer = &Which('eom');}			# eye of MATE
     if (!defined $viewer) {$viewer = &Which('gthumb');}		# Gthumb
     if (!defined $viewer) {$viewer = &Which('okular');}		# Gthumb
  } else {
     if    (-f 'C:/Progra~1/IrfanView/i_view32.exe') {$viewer = 'C:/Progra~1/IrfanView/i_view32.exe';}
     elsif (-f 'C:/Progra~2/IrfanView/i_view32.exe') {$viewer = 'C:/Progra~2/IrfanView/i_view32.exe';}
     elsif (-f 'C:/Progra~1/IrfanView/i_view64.exe') {$viewer = 'C:/Progra~1/IrfanView/i_view64.exe';}
     else  {$viewer = 'rundll32.exe C:/WINDOWS/system32/shimgvw.dll,ImageView_Fullscreen';}
  }
  if (!defined $viewer) {
     print STDERR '!!! '.$script.': cannot find graphic viewer for GnuPlot'."\n";
  }
  return $viewer;
}
#=================================================================

# sub open_gnuplot ([$scan_start, $scan_end], [$logscale], [\@gnuargs]) {
sub open_gnuplot (;$$$\@) {

  my ($scan_start, $scan_end, $logscale, $gnuargs, $success);

  $scan_start = shift(@_);
  $scan_end   = shift(@_);
  $logscale   = shift(@_);
  $gnuargs    = shift(@_);

  $success = &open_gnuplot_fh(\*GNUPLOT,$scan_start,$scan_end,$logscale,$gnuargs);
  return $success;
}

#=================================================================

# sub open_gnuplot_fh ($GNUPLOT_FH,[$scan_start, $scan_end], [$logscale], [\@gnuargs]) {
sub open_gnuplot_fh ($;$$$\@) {

  our ($GNUPLOT_VERSION);
  my ($GNUPLOT_FH, $scan_start, $scan_end, $logscale, $gnuargs);
  my ($script, $gnuplot, $terminal, $pid);

  $script = 'open_gnuplot_fh';
  $GNUPLOT_FH = shift(@_);  if (!defined $GNUPLOT_FH) {$GNUPLOT_FH=\*GNUPLOT;}
  $scan_start = shift(@_);  if (!defined $scan_start) {$scan_start=0;}
  $scan_end   = shift(@_);  if (!defined $scan_end)   {$scan_end=0;}
  $logscale   = shift(@_);  if (!defined $logscale)   {$logscale=0;}
  $gnuargs    = shift(@_);

  if (!defined $GNUPLOT_VERSION) {$GNUPLOT_VERSION = 4.0;} 	#default
  $gnuplot = &gnuplot_executable();
  if (!defined $gnuplot) {return 0;}
  if ($^O !~ /Win/i) {$terminal = 'x11';}			#Linux
  else               {$terminal = 'windows';}			#Windows
# else               {$terminal = 'wxt';}			#cross-platform wxWindows

### Possible GnuPlot args (may not work on older gnuplot with pgnuplot for Windows):
###  `-mono`     forces monochrome rendering on color displays.
###  `-gray`     requests grayscale rendering on grayscale or color displays.
###              (Grayscale displays receive monochrome rendering by default.)
###  `-clear`    requests that the window be cleared momentarily before a
###              new plot is displayed.
###  `-tvtwm`    requests that geometry specifications for position of the
###              window be made relative to the currently displayed portion
###              of the virtual root.
###  `-raise`    raises plot window after each plot
###  `-noraise`  does not raise plot window after each plot
###  `-noevents` does not process mouse and key events
###  `-persist`  plot windows survive after main gnuplot program exit
# $$gnuargs = ('-tvtwn', '-noraise', '-persist', '2>/dev/null');

# print STDOUT 'Opening pipe to '.$gnuplot.'-'.$gnuinstance."\n";
### GNUplot arguments are not supported on Windows:
  if (defined $gnuargs && $gnuplot !~ /pgnuplot/) {
#   print STDOUT  $script.': opening pipe to "'.$gnuplot.'" with option(s): '."@${gnuargs}\n";
    if ($^O !~ /Win/i) {
       $pid = open($GNUPLOT_FH, '|-', $gnuplot, @$gnuargs);		#this is calling directly
    } else {
       $pid = open($GNUPLOT_FH, '| '.$gnuplot." @{$gnuargs}");		#this is using shell (Windows does not support list form of pipe)
    }
  } else {
#   print STDOUT  $script.': opening pipe to "'.$gnuplot.'" without options.'."\n";
#   $pid = open($GNUPLOT_FH, '| '.$gnuplot);				#this is using shell
    $pid = open($GNUPLOT_FH, '|-', $gnuplot);				#this is calling directly
  }
  if (defined $pid) {
     $SIG{PIPE} = sub {die $script.': gnuplot pipe crashed'};
     select $GNUPLOT_FH; $|=1;						# set unbuffered output
# Set mouse before using terminal (by default mouse is disabled on pipes)
     if ($GNUPLOT_VERSION >= 4) {print $GNUPLOT_FH 'set mouse'."\n";}
     print $GNUPLOT_FH 'set terminal '.$terminal."\n";
     print $GNUPLOT_FH 'set tics out'  ."\n";
     print $GNUPLOT_FH 'set mxtics 10' ."\n";
     print $GNUPLOT_FH 'set mytics 10' ."\n";
     print $GNUPLOT_FH 'set my2tics 10'."\n";
     print $GNUPLOT_FH 'set grid'."\n";
     if ($GNUPLOT_VERSION >= 4) {print $GNUPLOT_FH 'set style data linespoints'."\n";}
     else                       {print $GNUPLOT_FH 'set data style linespoints'."\n";}
     print $GNUPLOT_FH 'set format y "%8.2e"'."\n";
     print $GNUPLOT_FH 'set xlabel "Scan axis"'."\n";
     print $GNUPLOT_FH 'set ylabel "Intensity"'."\n";
     if ($scan_start || $scan_end) {
        printf $GNUPLOT_FH 'set xrange [%.4g:%.4g]'."\n",$scan_start,$scan_end;
     }
     if ($logscale == 1) {
        print $GNUPLOT_FH 'set logscale y'."\n";
        print $GNUPLOT_FH 'set logscale y2'."\n";
     } else {
        if ($GNUPLOT_VERSION >= 4) {print $GNUPLOT_FH 'unset logscale'."\n";}
	else                       {print $GNUPLOT_FH 'set nologscale'."\n";}
     }
     return 1;

  } else {

     print STDOUT  '*** '.$script.': cannot open pipe to '.$gnuplot.": $!";
     return 0;

  }
}

#=================================================================

# sub print_gnuplot ([$autoplot,$GUIprompt]) {
sub print_gnuplot (;$$) {

  my ($autoplot, $GUIprompt, $printed);

  $autoplot  = shift(@_);
  $GUIprompt = shift(@_);

  $printed   = &print_gnuplot_fh(\*GNUPLOT,$autoplot,$GUIprompt);
  return $printed;
}

#=================================================================

# sub print_gnuplot_fh ($GNUPLOT_FH,[$autoplot,$GUIprompt]) {
sub print_gnuplot_fh ($;$$) {

  my ($GNUPLOT_FH, $autoplot, $GUIprompt);
  my ($sleep_time, $status, $printed, $printPRG);

  $GNUPLOT_FH = shift(@_);  if (!defined $GNUPLOT_FH) {$GNUPLOT_FH = \*GNUPLOT;}
  $autoplot   = shift(@_);  if (!defined $autoplot)   {$autoplot   = -1;}	#1=autoprint;  0=skip -1=ask
  $GUIprompt  = shift(@_);  if (!defined $GUIprompt)  {$GUIprompt   = 0;}

  $sleep_time = 2;
  $status     = 0;
  $printed    = 0;
  if ($^O !~ /Win/i) {$printPRG = 'lpr';    }    			#UNIX
# else               {$printPRG = 'gsprint';}    			#Windows
  else               {$printPRG = 'gsview32';}   			#Windows

  if ($autoplot == -1) {						#ask
     if ($GUIprompt) {
        $_ = &zenity_question('Print plot?','Yes','No');
     } else {
        print STDOUT 'Print plot [n]?_';
        ### This works on Linux, but not on Windows (input freezes):
        if ($^O !~ /Win/i) {$_=uc <STDIN>; chomp $_; print STDOUT "\n";}
        ### This works on Windows:
        else               {$_=getc STDIN; print STDOUT "$_\n";}
     }
  }
  elsif ($autoplot == 1) {$_ = 'Y'; print STDOUT 'Printing plot...'."\n";}
  else                   {$_ = 'N'; print STDOUT 'Skipping printer copy...'."\n";}

  if ($_ eq 'Y' || $_ eq 'y' || $_ eq '1') {
     $printed = 1;
     print STDOUT 'Creating temp file "gnu.ps"'."\n";
#    print $GNUPLOT_FH 'set terminal postscript'."\n";
     print $GNUPLOT_FH 'set terminal postscript color colortext'."\n";
     print $GNUPLOT_FH 'set output "gnu.ps"'."\n";
     print $GNUPLOT_FH 'replot'."\n";
     select (undef, undef, undef, $sleep_time);	# sleep($sleep_time);
  }

  &close_gnuplot_fh($GNUPLOT_FH);

  if ($_ eq 'Y' || $_ eq 'y' || $_ eq '1') {
     select (undef, undef, undef, $sleep_time);	# sleep($sleep_time);
     print STDOUT 'Sending plot to printer with the "'.$printPRG.'" command...'."\n";
     $status = `${printPRG} gnu.ps`;
     if ($status eq '1' || $status eq '-1') {
	print STDOUT 'Plot printing failed with status = '.$status."\n";
     }
  }
  return $printed;
}

#=================================================================

# sub image_gnuplot ($file,[$width,$height,$view,$GUIprompt]) {
sub image_gnuplot ($;$$$$) {

  my ($file, $width, $height, $view, $GUIprompt, $status);

  $file      = shift(@_);
  $width     = shift(@_);
  $height    = shift(@_);
  $view      = shift(@_);
  $GUIprompt = shift(@_);

  $status = &image_gnuplot_fh(\*GNUPLOT,$file,$width,$height,$view,$GUIprompt);
  return $status;
}

#=================================================================

# sub image_gnuplot_fh ($GNUPLOT_FH,$file,[$width,$height,$view,$GUIprompt]) {
sub image_gnuplot_fh ($$;$$$$) {

  my ($GNUPLOT_FH, $file, $width, $height, $view, $GUIprompt);
  my ($script, $ext, $terminal, $fontsize, $enhanced);
  my ($filesize, $status);

  $GNUPLOT_FH = shift(@_);  if (!defined $GNUPLOT_FH) {$GNUPLOT_FH = \*GNUPLOT;}
  $file       = shift(@_);
  $width      = shift(@_);  if (!defined $width)      {$width      = 800;}
  $height     = shift(@_);  if (!defined $height)     {$height     = 600;}
  $view       = shift(@_);  if (!defined $view)       {$view       = 0;}
  $GUIprompt  = shift(@_);  if (!defined $GUIprompt)  {$GUIprompt  = 0;}

  $script = 'image_gnuplot_fh';
  if (!defined $file) {
     print STDERR '!!! '.$script.': no image filename specified.'."\n";
     return -1;
  }
  $ext = $file;
  $ext =~ s/^.*\.//g;                                                   #distil extenstion
  if    ($ext =~ /^jpg$/i || $ext =~ /^jpeg$/i) {$terminal='jpeg';}
  elsif ($ext =~ /^gif$/i)                      {$terminal='gif';}
  elsif ($ext =~ /^png$/i)                      {$terminal='png';}
  else {
     print STDERR '!!! '.$script.': unsupported gnuplot image type "'.$ext.'".'."\n";
     return -1;
  }
  if (-e $file) {
     if (! -w $file) {  
        print STDERR '!!! '.$script.': image file "'.$file.'" exists and cannot be overwritten.'."\n";
        return -1;
     }
     if (! -f $file) {  
        print STDERR '!!! '.$script.': image file "'.$file.'" exists and not a regular file.'."\n";
        return -1;
     }
     if ($GUIprompt) {
        $_ = &zenity_question('File ['.$file.'] exists. Overwrite?','Yes','No');
     } else {
        print STDOUT 'File ['.$file.'] exists. Overwrite [n]?_';
        ### This works on Linux, but not on Windows (input freezes):
        if ($^O !~ /Win/i) {$_=uc <STDIN>; chomp $_; print STDOUT "\n";}
        ### This works on Windows:
        else               {$_=getc STDIN; print STDOUT "$_\n";}
     }
     if ($_ !~ /^[y1]$/i) {return -1;}                                  #if not Y, y, or 1
     unlink($file);                                                     #delete previous file
  }

  $fontsize = 'small';                                                  #options are tiny | small | medium | large | giant
### "enhanced" enables the enhanced text processing features, (subscripts, 
### superscripts and mixed fonts). It is supported by PNG/JPEG/GIF.
  $enhanced = 'enhanced';
  
  print STDOUT 'Saving plot as "'.$file.'"...'."\n";

  print $GNUPLOT_FH 'set terminal '.$terminal.' '.$fontsize.' size '.$width.','.$height.' '.$enhanced."\n";
  print $GNUPLOT_FH 'set output "'.$file.'"'."\n";
  print $GNUPLOT_FH 'replot'."\n";

  &close_gnuplot_fh($GNUPLOT_FH);

  if (! -e $file) {
     print STDERR '!!! '.$script.': gnuplot failed to create image file "'.$file.'".'."\n";
     return -1;
  }
  $filesize = -s $file;
  if ($filesize < 100) {
     print STDERR '!!! '.$script.': gnuplot failed plotting to "'.$file.'": filesize is only '.$filesize.' bytes.'."\n";
     return -1;
  }

  if ($view) {
     $status = &view_gnuplot_image ($file);
     return $status;
  } else {
     return 0;
  }
}

#=================================================================

# sub view_gnuplot_image ($file) {
sub view_gnuplot_image ($) {

  my ($file, $script, $filesize, $viewer, @pars);

  $file       = shift(@_);
  $script = 'view_gnuplot_image';

  if (!defined $file) {
     print STDERR '!!! '.$script.': no image filename specified.'."\n";
     return -1;
  }
  if (! -e $file) {
     print STDERR '!!! '.$script.': no image file "'.$file.'". found (did gnuplot create it?)'."\n";
     return -1;
  }
  $filesize = -s $file;
  if ($filesize < 100) {
     print STDERR '!!! '.$script.': gnuplot likely failed plotting to "'.$file.'": filesize is only '.$filesize.' bytes.'."\n";
     return -1;
  }

  $viewer = &gnuplot_viewer();
  if (defined $viewer) {
     print STDOUT 'Viewing "'.$file.'" with "'.$viewer.'".'."\n";
     if ($^O !~ /Win/i) {
### We want to completely separate the viewer process from the script 
### so that it does not hold stdout and stderr and allows the script
### to exit because there is no attached process. See:
### https://unix.stackexchange.com/questions/3886/difference-between-nohup-disown-and
### To suppress creation of nohup.out, redirect both stdout & stderr to /dev/null, see:
### https://stackoverflow.com/questions/10408816/how-do-i-use-the-nohup-command-without-getting-nohup-out
#       @pars = ($viewer,$file,'>/dev/null','2>&1','&');			#does not work (blocks because "&" needs bash to be interpreted)
#       @pars = ('nohup',$viewer,$file,'>/dev/null','2>&1','&');		#does not work (blocks because "&" needs bash to be interpreted)
#       system(@pars);
#       system($viewer.' '.$file.' >/dev/null 2>&1 &');				#calling script can continue, but kills viewer on exit or gets stuck when using tee
#       system($viewer.' '.$file.' >/dev/null 2>&1 &; disown -a');		#does not work
#       system('xterm -e nohup '.$viewer.' '.$file.' >/dev/null 2>&1 &');	#works, but ugly
        system('nohup '.$viewer.' '.$file.' >/dev/null 2>&1 &');		#works if both stdout & stderr redirected (no nohup.out is created)
#       system('nohup '.$viewer.' '.$file.' &>/dev/null &');			#works if both stdout & stderr redirected (no nohup.out is created), may not work with "sh", but OK woth "bash"
#       system('screen -d -m '.$viewer.' '.$file.' >/dev/null 2>&1 &');		#works (both forms are the same)
#       system('screen -d -m '.$viewer.' '.$file.' &>/dev/null &');		#works (both forms are the same)
     } else {
        if ($viewer =~ /^rundll32/i) {
           ($viewer,@pars) = split(/\s/,$viewer);
           system(1, $viewer, @pars, $file);
        } else {
           system(1, $viewer, $file);
        }
     }
  }
  return 0;
}

#=================================================================

sub close_gnuplot () {

### This routine closes GnuPlot pipe:
  &close_gnuplot_fh(\*GNUPLOT);
}

#=================================================================

# sub close_gnuplot_fh ($GNUPLOT_FH) {
sub close_gnuplot_fh ($) {

### This routine closes GnuPlot pipe:

  my ($GNUPLOT_FH, $status);

  $GNUPLOT_FH = shift(@_);  if (!defined $GNUPLOT_FH) {$GNUPLOT_FH=\*GNUPLOT;}

### This should close gnuplot:
  print $GNUPLOT_FH 'exit'."\n";
### close $GNUPLOT_FH or die "bad pipe: $! $?";
  $status = close($GNUPLOT_FH);
# if ($status) {print STDOUT "Bad pipe at closing gnuplot: $! $?\n";}
}

#=================================================================

# sub gnuplot_filess_delay([$milliseconds]) {
sub gnuplot_filess_delay (;$) {

  my $milliseconds = shift(@_);

  if (!defined $milliseconds) {$milliseconds=1;}
### Use one of the two:
  if (exists &usleep) {				#this is imported as "use Time::HiRes qw(usleep);"
     usleep(1000*$milliseconds);
  } else {
     select (undef, undef, undef, 0.001*$milliseconds);
  }
}

