#!/usr/bin/perl -W
#--------------------------------------------------------------
#   These are subs for working with ALS-style automounter
#       (State-notations robot server by Oleg Makarov)
#
#     	             Author: Sergey Stepanov
#
# Version-1:  2007/11/01 initial implementation
# Version-2:  2017/10/19 added Hall (pin) sensor
#--------------------------------------------------------------
# List of functions:
# ^^^^^^^^^^^^^^^^^
#  1.          $status = &pauseRobot($robot);
#  2.          $status = &unpauseRobot($robot);
#  3.           $state = &getRobotState($robot);
#  4.    $statusString = &readRobotStatus($robot,[-1|-2]);
#  5.          $status = &waitRobotReady ($robot,[$report_interval]);
#  6.          $status = &loadRobotPin ($robot,$puck,$style,$pin,[$waitCompletion,$report_interval]);
#  7.          $status = &fillRobotDACs($robot,\%rdbkHash,\%ctrlHash);
#                        %rdbkHash = {'longHorzOutPV', 'longHorzInPV', 'longVertUpPV', 'longVertDnPV', 'shortVertUpPV', 'shortVertDnPV'}
#                        %ctrlHash = {'shortVertPV','lognVertPV','rotateHorzPV'}
#  8.           $state = &checkGripperSafe($robot,[$robotType]); 			#state: [1=OK, 0=unsafe] robotType: BAM1/BAM2
#  9.          $status = &setGripperSafe($robot,[$robotType]);				#robotType: BAM1/BAM2
# 10.           $state = &getDewarLidState();                                      	#1=open 0=closed -1=unknown
# 11.          $status = &showRobotVideo();
# 12.     $pinsensorPV = &getPinSensorPV([$beamline]);
# 13. ($status,$yesno) = &pinMounted([$pinsensorPV]);
#--------------------------------------------------------------
# use Pezca;

# use warnings::unused;
  use strict;
# no strict 'vars';
# no strict 'subs';
# no strict 'refs';

#=================================================================

# sub pauseRobot ($robot) {
sub pauseRobot ($) {
    my ($script, $robot, $pausePV, $status, $state);
    $script = 'pauseRobot';
    $robot  = shift(@_);
    if (!defined $robot) {return -1;}
    if ($robot !~ /:$/) {$robot .= ':';}
    $pausePV = $robot.'pause';
### 0=ready 1=paused
    $state = &getRobotState($robot);
    if ($state == 0) {
      print STDOUT "Pausing sample mounting robot\n";
      $status = Pezca::PutDoubleOldCa($pausePV, 1);
      &report_only($script, $status, $pausePV);
    }
    elsif ($state == 1) {$status = 0;}
    else                {$status = -1;}
    return $status;
}

#=================================================================

# sub unpauseRobot ($robot) {
sub unpauseRobot ($) {
    my ($script, $robot, $pausePV, $status, $state);
    $script = 'unpauseRobot';
    $robot  = shift(@_);
    if (!defined $robot) {return -1;}
    if ($robot !~ /:$/) {$robot .= ':';}
    $pausePV = $robot.'pause';
### 0=ready 1=paused
    $state = &getRobotState($robot);
    if ($state == 1) {
      print STDOUT "Un-pausing sample mounting robot\n";
      $status = Pezca::PutDoubleOldCa($pausePV, 0);
      &report_only($script, $status, $pausePV);
    }
    elsif ($state == 0) {$status = 0;}
    else                {$status = -1;}
    return $status;
}

#=================================================================

# sub getRobotState ($robot) {
sub getRobotState ($) {
    my ($script, $robot, $pausePV, $status, $state);
    $script = 'getRobotState';
    $robot = shift(@_);
    if (!defined $robot) {return -1;}
    if ($robot !~ /:$/) {$robot .= ':';}
    $pausePV = $robot.'pause';
### 0=ready 1=paused
    ($status,$state) = Pezca::GetDouble($pausePV);
    if ($status) {
      &report_only($script, $status, $pausePV);
      $state = -1;
    }
    return $state;
}

#=================================================================

# sub readRobotStatus ($robot,[queue]) {
sub readRobotStatus ($;$) {
    my ($script, $robot, $queue, $statusPV, $status, $statusString);
    $script = 'readRobotStatus';
    $robot = shift(@_);
    $queue = shift(@_);
    if (!defined $robot) {return -1;}
    if (!defined $queue) {$queue = '';}
    else {
      if ($queue ne '-1' && $queue ne '-2') {
	 print STDOUT '*** '.$script.': incorrect queue argument='.$queue.'. Expecting "-1" or "-2"'."\n";
	 return -1;
      }
    }
    if ($robot !~ /:$/) {$robot .= ':';}
    $statusPV = $robot.'sta'.$queue;
    ($status, $statusString) = Pezca::GetString($statusPV);
    if ($status) {
      &report_only($script, $status, $statusPV);
      $statusString = '';
    }
    return $statusString;
}

#=================================================================

# sub waitRobotReady ($robot,[$report_interval]) {
sub waitRobotReady ($;$) {
### Wait for Automounter 'ready' status:
    my ($script, $robot, $print_interval, $start_sleep, $sleep_interval);
    my ($elapsed_time, $print_time, $statusString, $ROBOTREADY, $status);
    $script = 'waitRobotReady';
    $robot          = shift(@_);
    $print_interval = shift(@_);	#delay between printouts in sec,
    if (!defined $robot)          {return -1;}
    if (!defined $print_interval) {$print_interval = 5.0;}	#default
    if ($robot !~ /:$/)           {$robot .= ':';}
    $start_sleep    = 1;		#delay before begin monitoring in sec.
    $sleep_interval = 0.05;		#delay between successive queries in sec.
    $elapsed_time   = 0.0;		#counter
    $print_time     = 0.0;		#counter
    $statusString   = '';
    $ROBOTREADY     = 'SMR: ready';

### Wait for 'ready' status:
    print STDOUT "Waiting for robot ready status \"${ROBOTREADY}\"...\n";
    select(undef,undef,undef,$start_sleep);	 	# start sleep
    $elapsed_time += $start_sleep;
    $print_time   += $start_sleep;

### This is for 'readRobotStatus':
    $status = &addCaMonitor ($robot.'sta', 'string');
    &report_only($script, $status, $robot.'sta');

    do {
       select(undef,undef,undef,$sleep_interval);	 # short sleep
       $elapsed_time += $sleep_interval;
       $print_time   += $sleep_interval;
       $statusString = &readRobotStatus($robot);
### Only print messages if interval is non-zero:
       if ($print_time > $print_interval && $print_interval > 0) {
          printf STDOUT "%5.0fs waiting time...\n", $elapsed_time;
          $print_time = 0;
       }
    } while ($statusString !~ /${ROBOTREADY}$/);
    print STDOUT "...OK!\n";
    return 0;
    }

#=================================================================

# sub &loadRobotPin ($robot,$cassette,$style,$pin,[$waitCompletion,$report_interval]) {
sub loadRobotPin ($$$$;$$) {
    my ($script, $robot, $puck, $style, $pin, $loadwait, $ds_print);
    my (@pucks, @styles, $maxpins, $cmdPV, $cmd, $status);
    $script = 'loadRobotPin';
    $robot = shift(@_);
    $puck  = shift(@_);
    $style = shift(@_);
    $pin   = shift(@_);
    $loadwait = shift(@_);
    $ds_print = shift(@_);
    if (!defined $robot) {return -1;}
    if (!defined $puck)  {return -1;}
    if (!defined $style) {return -1;}
    if (!defined $pin)   {return -1;}
    if (!defined $loadwait) {$loadwait = 0;}
    if ($robot !~ /:$/)  {$robot .= ':';}
    @pucks  = ('A', 'B', 'C', 'D', 'E', 'F');
    @styles = ('A', 'R');
### Validate input parameters:
    $puck  = uc($puck);
    if (!grep(/^${puck}$/, @pucks)) {
      print STDOUT "*** ${script}: incorrect puck=${puck}\n";
      return -1;
    }
    $style = uc($style);
    if (!grep(/^${style}$/, @styles)) {
      print STDOUT "*** ${script}: incorrect style=${style}\n";
      return -1;
    }
    if    ($style eq 'A') {$maxpins = 16;}
    elsif ($style eq 'R') {$maxpins = 12;}
    if ($pin < 1 || $pin > $maxpins) {
      print STDOUT "*** ${script}: pin=${pin} not in range [1-${maxpins}]\n";
      return -1;
    }
### Send 'load pin' command:
    $cmdPV = $robot.'cmd';
    $cmd   = 'load '.$puck.' '.$style.' '.$pin;
    print STDOUT "Sending robot command:  ${cmd}\n";
    $status = Pezca::PutString($cmdPV, $cmd);
    if ($status) {
      &report_exit($script, $status, $cmdPV);
      return $status;
    }
### Optionally wait for completion:
    if ($loadwait) {$status = &waitRobotReady($robot,$ds_print);}
    return $status;
}

#=================================================================

# sub fillRobotDACs($robot, %rdbkHash, \%ctrlHash) {		#returns status
#                    %rdbkHash = {'longHorzOutPV', 'longHorzInPV', 'longVertUpPV',
#                                 'longVertDnPV', 'shortVertUpPV', 'shortVertDnPV', ...}
#                    %ctrlHash = {'shortVertPV','lognVertPV','rotateHorzPV', ...}
sub fillRobotDACs ($\%\%) {
    my ($script, $robot, $ctrlHashRef, $rdbkHashRef, $bl, $DAC_ctrl, $DAC_rdbk);
    $script = 'fillRobotDACs';
    $robot       = shift(@_);
    $rdbkHashRef = shift(@_);
    $ctrlHashRef = shift(@_);

    if (defined $ctrlHashRef) {%$ctrlHashRef = ();}
    if (defined $rdbkHashRef) {%$rdbkHashRef = ();}

    if (length($robot) < 3) {
       print STDOUT '*** '.$script.': too short robot PV=['.$robot.']. Expecting beamline prefix.'."\n";
       return 1;
    }
    $bl = substr($robot,0,3);
    if ($bl !~ /^23[iobd]$/) {
       print STDOUT '*** '.$script.': incorrect robot PV=['.$robot.']. Expecting beamline prefix.'."\n";
       return 2;
    }
    $DAC_ctrl = $bl.':pmac21:acc65e:bo';
    $DAC_rdbk = $bl.':pmac21:acc65e:bi';
    if (defined $rdbkHashRef) {
	$$rdbkHashRef{'longHorzOutPV'} = $DAC_rdbk.'0';
	$$rdbkHashRef{'longHorzInPV'}  = $DAC_rdbk.'1';
	$$rdbkHashRef{'longVertUpPV'}  = $DAC_rdbk.'4';
	$$rdbkHashRef{'longVertDnPV'}  = $DAC_rdbk.'5';
	$$rdbkHashRef{'shortVertUpPV'} = $DAC_rdbk.'6';
	$$rdbkHashRef{'shortVertDnPV'} = $DAC_rdbk.'7';
	$$rdbkHashRef{'rotateHorzPV'}  = $DAC_rdbk.'8';
	$$rdbkHashRef{'rotateVertPV'}  = $DAC_rdbk.'9';
    }
    if (defined $ctrlHashRef) {
	$$ctrlHashRef{'longHorzPV'}    = $DAC_ctrl.'0';
	$$ctrlHashRef{'longVertPV'}    = $DAC_ctrl.'2';
	$$ctrlHashRef{'shortVertPV'}   = $DAC_ctrl.'3';
	$$ctrlHashRef{'rotateHorzPV'}  = $DAC_ctrl.'4';
    }
    return 0;
}

#=================================================================

# sub checkGripperSafe($robot,[$robotType]) { 			#returns state: 1=OK, 0=unsafe
sub  checkGripperSafe($;$) {
    our ($beamline);
    my ($script, $robot, %rdbkHash, %valHash, $status, $state);
    my ($longHorzOutPV, $longHorzInPV,  $longVertUpPV, $longVertDnPV);
    my ($shortVertUpPV, $shortVertDnPV, $rotateHorzPV, $rotateVertPV);
    my ($robotType, $gripperAssy, $gripperVAxis, $gripperHomed);
    my ($gripperVAct, $gripperSafe, $val);
    $script    = 'checkGripperSafe';
    $robot     = shift(@_);
    $robotType = shift(@_);	if (!defined $robotType) {$robotType='BAM1';}
    $state = 0;						#unsafe
    if ($robotType =~ /^BAM1$/i) {
       %rdbkHash = ();
       $status = &fillRobotDACs($robot,\%rdbkHash);
       if ($status) {return $state;}
       $longHorzOutPV = $rdbkHash{'longHorzOutPV'};
       $longHorzInPV  = $rdbkHash{'longHorzInPV'};
       $longVertUpPV  = $rdbkHash{'longVertUpPV'};
       $longVertDnPV  = $rdbkHash{'longVertDnPV'};
       $shortVertUpPV = $rdbkHash{'shortVertUpPV'};
       $shortVertDnPV = $rdbkHash{'shortVertDnPV'};
       $rotateHorzPV  = $rdbkHash{'rotateHorzPV'};
       $rotateVertPV  = $rdbkHash{'rotateVertPV'};

       ($status,$valHash{'longHorzOut'}) = Pezca::GetDouble($longHorzOutPV);
       if ($status) {&report_only($script,$status,$longHorzOutPV); return $state;}

       ($status,$valHash{'longHorzIn'}) = Pezca::GetDouble($longHorzInPV);
       if ($status) {&report_only($script,$status,$longHorzInPV); return $state;}

       ($status,$valHash{'longVertUp'}) = Pezca::GetDouble($longVertUpPV);
       if ($status) {&report_only($script,$status,$longVertUpPV); return $state;}

       ($status,$valHash{'longVertDn'}) = Pezca::GetDouble($longVertDnPV);
       if ($status) {&report_only($script,$status,$longVertDnPV); return $state;}

       ($status,$valHash{'shortVertUp'}) = Pezca::GetDouble($shortVertUpPV);
       if ($status) {&report_only($script,$status,$shortVertUpPV); return $state;}

       ($status,$valHash{'shortVertDn'}) = Pezca::GetDouble($shortVertDnPV);
       if ($status) {&report_only($script,$status,$shortVertDnPV); return $state;}

       ($status,$valHash{'rotateHorz'}) = Pezca::GetDouble($rotateHorzPV);
       if ($status) {&report_only($script,$status,$rotateHorzPV); return $state;}

       ($status,$valHash{'rotateVert'}) = Pezca::GetDouble($rotateVertPV);
       if ($status) {&report_only($script,$status,$rotateVertPV); return $state;}

       if ($valHash{'longHorzOut'} == 1 && $valHash{'longHorzIn'} == 0) {
          if ($valHash{'longVertUp'} == 1 && $valHash{'longVertDn'} == 0 &&
             $valHash{'shortVertUp'} == 1 && $valHash{'shortVertDn'} == 0) {
             $state = 1;
          }
       }
       elsif ($valHash{'longHorzOut'} == 0 && $valHash{'longHorzIn'} == 1) {
          $state = 1;
       }
       elsif ($valHash{'rotateHorz'} == 1 && $valHash{'rotateVert'} == 0) {
          $state = 1;
       }
       if ($state == 0) {
          print STDOUT "\n".'!!! '.$script.': unsafe gripper state:'."\n\n";
	  foreach $val (keys  %valHash) {printf STDOUT '   %s = %d'."\n", $val, $valHash{$val};}
          print STDOUT "\n";
       }
       return $state;
    }
    elsif ($robotType =~ /^BAM2$/i) {
       $gripperAssy  = $beamline.':RB:Gr';
       $gripperVAxis = $beamline.':RB:GV';
       $status = 0;
       $gripperHomed = &getHomed($gripperAssy,\$status);
       if ($status) {
          print STDOUT '*** '.$script.': failed to get homed status of '.$gripperAssy."\n";
          return 0;						#unsafe
       }
       if (! $gripperHomed) {
          print STDOUT '*** '.$script.': gripper '.$gripperAssy.' is not homed. Status is unknown'."\n";
          return 0;						#unsafe
       }
       $status = 0;
       $gripperVAct = &getAxisAct ($gripperVAxis,\$status);
       if ($status) {
          print STDOUT '*** '.$script.': failed to get vertical pos. of gripper vertical axis '.$gripperVAxis."\n";
          return 0;						#unsafe
       }
       $gripperSafe = -10;					#according to Oleg, 0mm position is 5mm above Dewar door
       if ($gripperVAct < $gripperSafe) {
          printf STDOUT '*** '.$script.': gripper vertical pos=%.1f is'
                       .' below the save position of %.1f'."\n",
                                      $gripperVAct, $gripperSafe;
          return 0;						#unsafe
       } else {
          printf STDOUT $script.': gripper vertical pos=%.1f is safe.'."\n",$gripperVAct;
          return 1;						#safe
       }
    }
    else {
       print STDOUT "\n".'*** '.$script.': unrecognized robot type: '.$robotType."\n\n";
       return 0;						#unsafe
    }
}

#=================================================================

# sub setGripperSafe($robot,[$robotType]) { 			#returns status
sub  setGripperSafe($) {
    our ($beamline);
    my ($script, $robot, $robotType, %ctrlHash, $status, $state);
    my ($longVertPV, $shortVertPV, $msg);
    my ($gripperAssy, $gripperVAxis, $gripperHomed, $gripperSafe);
    $script    = 'setGripperSafe';
    $robot     = shift(@_);
    $robotType = shift(@_);	if (!defined $robotType) {$robotType='BAM1';}
    ### Check current robot state:
    $state = &checkGripperSafe($robot,$robotType);				#1=safe/0=unsafe
    if ($state) {
       print STDOUT $script.': robot ['.$robot.'] gripper is in safe position. No action needed.'."\n";
       return 0;
    }
    print STDOUT $script.': moving robot ['.$robot.'] gripper to safe position.'."\n";
    if ($robotType =~ /^BAM1$/i) {
       %ctrlHash = ();
       $status = &fillRobotDACs($robot,undef,\%ctrlHash);
       if ($status) {return $status;}
       $longVertPV	 = $ctrlHash{'longVertPV'};
       $shortVertPV = $ctrlHash{'shortVertPV'};

       $status = Pezca::PutDouble($longVertPV, 0);
       if ($status) {&report_only($script,$status,$longVertPV); return $status;}

       $status = Pezca::PutDouble($shortVertPV, 0);
       if ($status) {&report_only($script,$status,$shortVertPV); return $status;}
    }
    elsif ($robotType =~ /^BAM2$/i) {
       $gripperAssy  = $beamline.':RB:Gr';
       $gripperVAxis = $beamline.':RB:GV';
       $status = 0;
       $gripperHomed = &getHomed($gripperAssy,\$status);
       if ($status) {
          print STDOUT '*** '.$script.': failed to get homed status of '.$gripperAssy."\n";
          return 1;						#failed
       }
       if (! $gripperHomed) {
          print STDOUT '*** '.$script.': gripper '.$gripperAssy.' is not homed. Status is unknown'."\n";
          return 1;						#failed
       }
       $gripperSafe = -10;					#according to Oleg, 0mm position is 5mm above Dewar door
       $msg = '';
       $status = &runMotor ($gripperAssy, $gripperVAxis, $gripperSafe, 'checkAuto',\$msg);
       if ($status) {
          printf STDOUT '*** '.$script.': failed to move gripper '
                        .$gripperVAxis.' to %.1f'."\n".$msg, $gripperSafe;
          return 1;						#failed
       }
    }
    else {
       print STDOUT "\n".'*** '.$script.': unrecognized robot type: '.$robotType."\n\n";
       return 1;						#failed
    }

    ### Check robot state again:
    $state = &checkGripperSafe($robot,$robotType);				#1=safe/0=unsafe
    if ($state) {
       print STDOUT $script.': robot ['.$robot.'] gripper has been moved to safe position.'."\n";
       return 0;
    } else {
       print STDOUT $script.': failed to move robot ['.$robot.'] gripper to safe position.'."\n";
       return 1;
    }
}

#=================================================================

# 10.        $state = &getDewarLidState();                      #1=open 0=closed -1=unknown
# sub getDewarLidState() {
sub  getDewarLidState() {
    our ($beamline);
    my ($lidOpenPV, $lidClosedPV);
    my ($status, $lidOpen, $lidClosed, $script);

    $script = 'getDewarLidState';
    if ($beamline eq '23i') {
       $lidOpenPV   = $beamline.':pmac21:acc65e:bi10';
       $lidClosedPV = $beamline.':pmac21:acc65e:bi11';
    }
    elsif ($beamline eq '23o') {
       $lidOpenPV   = $beamline.':pmac21:acc65e:bi2';
       $lidClosedPV = $beamline.':pmac21:acc65e:bi3';
    }
    elsif ($beamline eq '23b') {
       $lidOpenPV   = $beamline.':pmac21:acc65e:bi10';
       $lidClosedPV = $beamline.':pmac21:acc65e:bi11';
    }
    else {
       return -1;
    }
    ($status,$lidOpen) = Pezca::GetDouble($lidOpenPV);
    if ($status) {
       &report_only($script, $status, $lidOpenPV);
       return -1;
    }
    ($status,$lidClosed) = Pezca::GetDouble($lidClosedPV);
    if ($status) {
       &report_only($script, $status, $lidClosedPV);
       return -1;
    }
    if    (  $lidOpen && ! $lidClosed) {return 1;}
    elsif (! $lidOpen &&   $lidClosed) {return 0;}
    else {
       print STDOUT '*** '.$script.' Dewar lid Open ('.$lidOpenPV.') and Closed ('.$lidClosedPV.') in the same state.'."\n";
       return -1;
    }
}

#=================================================================

# 11.       $status = &showRobotVideo();
sub showRobotVideo() {
    our ($axisCameraURL2);
    my ($script, $camera, $preset, $status);

    $script = 'showRobotVideo';
    $preset = 'robot';
    $camera = 1;
    $status = &gotoPZTcameraPreset($preset,$axisCameraURL2,$camera);
    if ($status) {
       print STDOUT "\n".'*** '.$script.': error='.$status
      .' from gotoPZTcameraPreset when requesting camera URL='
      .$axisCameraURL2.' to show preset='.$preset."\n\n";
       return $status;
    }
    $status = &openCameraHomepage($axisCameraURL2);
    if ($status) {
       print STDOUT "\n".'*** '.$script.': error='.$status
      .' from openCameraVideoStream when requesting Firefox'
      .' to show camera URL='.$axisCameraURL2."\n\n";
       return $status;
    } else {
       sleep 3;
       return 0;
    }
}

#=================================================================

# 12.  $pinsensorPV = &getPinSensorPV([$beamline]);
sub getPinSensorPV(;$) {
    our ($beamline);
    my ($script, $bl, $nopinPV, $PV, $status, $string, @words);
    $script = 'getPinSensorPV';
    $bl = shift(@_);
    if (!defined $bl && defined $beamline) {$bl = $beamline;}
    if (!defined $bl) {
       print STDOUT '*** '.$script.': unknown beamline!'."\n";
       return;						#returns undefined
    }
    $nopinPV = $bl.':bi:nopin.INPA';
    ($status, $string) = Pezca::GetString($nopinPV);
    if ($status) {
       &report_only($script, $status, $nopinPV);
       return;						#returns undefined
    }
    @words = split(/\s+/,$string);
    $PV = $words[0];
    $PV =~ s/\.VAL$//i;
    return $PV;
}

#=================================================================

# 13. ($status,$yesno) = &pinMounted([$pinsensorPV]);
sub pinMounted(;$) {
    my ($script, $pinsensorPV, $status, $readout, $mounted);
    $script = 'pinMounted';
    $pinsensorPV = shift(@_);
    if (!defined $pinsensorPV) {$pinsensorPV = &getPinSensorPV();}
    if (!defined $pinsensorPV) {
       print STDOUT '*** '.$script.': cannot determine pin sensor PV. Presuming pin is mounted.'."\n";
       return 1,1;				#presuming pin is mounted
    }
    ($status, $readout) = Pezca::GetDouble($pinsensorPV);
    if ($status) {
       &report_only($script, $status, $pinsensorPV);
       return 1,1;				#presuming pin is mounted
    } else {
       $mounted = 1 - $readout;
       return 0,$mounted;
    }
}
