#!/usr/bin/perl -W
# _CALLED_ON_DIRECTNET_
#--------------------------------------------------------------
# This script monitors the DirectNet IOC (the GMCA Equipment Protection
# system) and sends e-mais to the list of interested paties in case of a
# problem. This script normally runs all the time along with the directNet
# IOC on the directnet computer.
#
#               Initial implementation: Derek Yoder
#                Version-2.0 and after: Sergey Stepanov
#
# Version-2.0 2012/12/03 Speed/stability cleanup
# Version-2.1 2013/12/04 Change to 2D-arrays
# Version-2.2 2014/02/04 Added processing of pressure fluctuations;
#                        Introduced readable arrays indices
# Version-2.3 2015/06/25  added unlisted (unexpected) states for @EPS_PVs, @FEV_PVs, @BIV_PVs
# Version-2.4 2015/10/26 added analysis of individual bits when unexpected @EPS_PVs state happens
# Version-2.5 2015/11/20 made unexpected @EPS_PVs state=3 hidden
# Version-3.0 2016/02/09 added history array to show all data preceeding a problem
# Version-3.1 2016/02/12 added sending email with after-program data
# Version-3.2 2016/02/15 reduced the poll rate from 0.1s to 1s
# Version-3.3 2016/02/15 added flag thatFEV and BIV trips are the same;
#                        disabled sending history on restoring normal state
# Version-3.4 2016/10/11 separated after-dumps for trips (level=1) and warnings (level=2)
# Version-3.5 2017/10/03 changed to 8 states of EPS status instead of 4.
# Version-3.6 2018/02/01 added $MAX_EMAILS_PER_DAY
# Version-3.7 2019/03/25 added checking Directnet status
# Version-3.8 2019/10/01 added checking liquid nitrogen pressure and valves
# Version-4.0 2019/10/22 added %CA_Failure array for time stamps of CA errors and delayed CA err reports
# Version-4.1 2020/04/02 combined EPS-TRIPPED and EPS-OFF since they may alternate in case of failure and thus cause flood of emails.
# Version-4.2 2020/11/19 added reporting energy
# Version-4.3 2022/09/09 added monitoring water flow through WBSL with new APS BLEPS
# Version-5.0 2023/02/01 adapted to new APS BLEPS instead of ACCEL EPS
#-------------------------------------------------------------
  use strict;
  use Pezca;
  use MIME::Lite;
  use Time::HiRes qw(gettimeofday tv_interval time);

  my (@EPS_PVs, $n_EPS_PVs);
  my (@FEV_PVs, $n_FEV_PVs);
  my (@BIV_PVs, $n_BIV_PVs);
  my (@P_PVs, $n_P_PVs, $P_warn, $P_watch);
  my (@T_PVs, $n_T_PVs, $T_warn_1, $T_warn_2);
  my (@C_PVs, $n_C_PVs);
  my (@V_PVs, $n_V_PVs);
  my (@F_PVs, $n_F_PVs);
  my (@LN2_PVs, $n_LN2_PVs, $LN2_min, $LN2_max);
  my (@LN2_V_PVs, $n_LN2_V_PVs, $LN2_V_OPEN);
  my (@IP_PVs, $n_IP_PVs);
  my ($PEZCA, $status, $statui, $datum, $i, $msg_flags, $debug);
  my ($sleeptime, $sleepstep, $sleeping, $message);
  my ($script, $email_list, $day_changed, $trips, $warnings, $recovers);
  my (@processes, $nprocesses, $filehandle, @filenames, $filename, $logdir);
  my ($epoch, $epochPV, $timestampPV, $condition, $descr);
  my ($fluct_percent, $fluct_maxtimes, $timestamp);
  my ($PVname, $PVdescr, $statusPrev, $valuePrev, $P_nfluct, $T_setpnt);
# my ($EPS_TRIPPED, $EPS_OPERATOR, $EPS_OK, $EPS_PRETRIPPED);
# my ($EPS_OFF, $EPS_TRIPPED, $EPS_OPERATOR, $EPS_OPERTRIP, $EPS_OK, $EPS_TAG);
  my ($EPS_TRIPPED, $EPS_OK, $EPS_TAG);
  my ($FEV_MOVING,  $FEV_CLOSED,   $FEV_OPEN, $FEV_ERROR);
  my ($BIV_MOVING,  $BIV_CLOSED,   $BIV_OPEN, $BIV_ERROR);
  my ($C_OK, $IP_OK);
  my ($F_OK, $F_shtr2cls);
  my ($V_OPEN);
  my ($last_output_time_ref, @history, $maxhistory, $nhistory);
  my (@headers, $nvar, $shortstamp, $now_usec, $last_now_day);
  my ($now_ss, $now_mm, $now_hh, $now_day, $now_mon, $now_year);
  my ($send_history, $send_afterdump, $afterdump_interval, $afterdump_list);
  my ($last_email_time_ref, $last_email_stamp, $need_after_mail, $fev_biv_same);
  my ($MAX_EMAILS_PER_DAY, $n_email_sent);
  my ($arg, $monitorEPS, $monitorPval, $monitorT, $monitorFlow);
  my ($monitorValves, $monitorLN2, $monitorIonPumps);
  my (%CA_Failure, $CAerr_timeout, $age);
  my (@energyPVs, @energy, $en, $j);

  select STDERR; $|=1;                          # Set unbuffered output
  select STDOUT; $|=1;                          # Set unbuffered output

# $debug = 1;
  $MAX_EMAILS_PER_DAY = 25;
  $n_email_sent = 0;				#how many emails have been sent today
  $CAerr_timeout = 600;                         #10 minutes
    
### Do common subs -------------------------------------------------------------
  if (!defined $ENV{PEZCA}) {
     print STDERR '*** Undefined PEZCA environment.'."\n\n";
     if ($^O !~ /Win/i) {$PEZCA = '/home/gmca/epics/pezca';}
     else               {$PEZCA = 'c:/gmca/epics/pezca';}
  } else {
     $PEZCA = $ENV{PEZCA};
     $PEZCA =~ s/\\/\//g;
  }
  if ($PEZCA !~ /\/$/) {$PEZCA .='/';}

  @filenames = qw(common_subs.pl shutter_subs.pl beamline_23i.pl);	#beamline_23i contains all station shutters
  foreach $filename (@filenames) {
     $filename = $PEZCA.'Common/'.$filename;
     if (-e $filename) {do $filename;
        if ($@) {die 'Failed parsing '.$filename.': '.$@;}              # '$@' is error message
     } else     {die 'Did not find '.$filename;}
  }

  $script = &makeScriptName('noext');

  our $nRetry=5;
  &setEpicsRetry();
  Pezca::AutoErrorMessageOff();

### Signal handler -------------------------------------------------------------
  $SIG{INT}  = 'signal_exit';
  $SIG{QUIT} = 'signal_exit';

  die '!!! '.$script.': EPS monitoring is disabled for the APS Dark Period. Contact Sergey if you need to resume.'."\n";

### Command line ---------------------------------------------------------------
  if (@ARGV == 0) {
     $monitorEPS      = 1;
     $monitorPval     = 1;
     $monitorIonPumps = 1;
     $monitorT        = 1;
     $monitorValves   = 1;
     $monitorLN2      = 1;
     $monitorFlow     = 1;
  } else {
     $monitorEPS      = 0;
     $monitorPval     = 0;
     $monitorIonPumps = 0;
     $monitorT        = 0;
     $monitorValves   = 0;
     $monitorLN2      = 0;
     $monitorFlow     = 0;
     foreach $arg (@ARGV) {
        if    ($arg =~ /^EPS$/i)    {$monitorEPS      = 1;}
        elsif ($arg =~ /^P$/i)      {$monitorPval     = 1;}
        elsif ($arg =~ /^IP$/i)     {$monitorIonPumps = 1;}
        elsif ($arg =~ /^T$/i)      {$monitorT        = 1;}
        elsif ($arg =~ /^Valves$/i) {$monitorValves   = 1;}
        elsif ($arg =~ /^LN2$/i)    {$monitorLN2      = 1;}
        elsif ($arg =~ /^Flow$/i)   {$monitorFlow     = 1;}
        else {die '*** '.$script.': incorrect command line agrgument "'.$arg.'"'."\n"
                .'Syntax: '.$script.' [EPS] [P] [T] [Valves] [LN2] [Flow]'."\n";
        }        
     }     
  }       

### Watch PVs from all beamlines ----------------------------------------------
  if (defined $ENV{EPICS_CA_ADDR_LIST_FULL}) {$ENV{EPICS_CA_ADDR_LIST}=$ENV{EPICS_CA_ADDR_LIST_FULL};}
  else                                       {die $script.': EPICS_CA_ADDR_LIST_FULL not defined.'."\n";}

  if ($^O !~ /Win/i) {
     if (defined $ENV{HOME}) {$logdir = $ENV{HOME}.'/logfiles/';}
     else                    {die $script.': Environment variable HOME is not defined.'."\n";}
  } else {
     $logdir = 'c:/gmca/logfiles';
  }
  if (! -d $logdir) {
     print STDOUT $script.': Creating LOG directory: '.$logdir."\n";
     $status = mkdir $logdir;
     if ($status < 1) {
        print STDOUT $script.': Could not create LOG directory: '.$logdir."\n";
        exit;
     }
  } else {
     print STDOUT $script.': LOG directory is: '.$logdir."\n";
  }


### Check for pre-existing script ----------------------------------------------

  if ($^O !~ /Win/i) {
     $nprocesses = &count_Nrunning_lsof(\@processes);
     print STDOUT 'nprocesses='.$nprocesses."\n";
     $nprocesses--;								#ignore this script
     if ($nprocesses > 0) {
        $filename = $logdir.$script.'_error.log.';
        $filehandle = open (LOG,'>> '.$filename);       			#open LOG file for writing
        if (!defined $filehandle) {
           print STDOUT $script.': Could not open LOG file: '.$filename."\n";
           exit;
        }
        select LOG; $|=1;                               			#set unbuffered output
        $timestamp = scalar localtime;                  			#write logfile header;
        print LOG 'Attempted to start EPS monitor: '.$timestamp."\n"
                . 'Found '.$nprocesses.' process(es) already running.'."\n"
                . 'Process list including the current one:'."\n\n";
        foreach $i (@processes) {print LOG $i,"\n";}
        close LOG;
        print STDOUT $script.': Attempted to start EPS monitor: '.$timestamp."\n"
                   . $script.': Found '.$nprocesses.' process(es) already running.'."\n"
                   . $script.': Process list including the current one:'."\n\n";
        foreach $i (@processes) {print STDOUT $i,"\n";}
        exit;
     }
  }

### Define email addresses here ------------------------------------------------
  $email_list = 'rfischetti@anl.gov'
              . ',sstepanov@anl.gov'
              . ',makarov@anl.gov'
              . ',mhilgart@anl.gov'
              . ',qxu@anl.gov'
              . ',scorcoran@anl.gov'
              . ',fergusond@anl.gov'
              . ',ogata@anl.gov'
              . ',vnagas@anl.gov'
              . ',mbecker@anl.gov'
              . ',dkissick@anl.gov'
               ;
$afterdump_list = 'rfischetti@anl.gov'
              . ',sstepanov@anl.gov'
              . ',makarov@anl.gov'
              . ',scorcoran@anl.gov'
              . ',fergusond@anl.gov'
               ;

# $email_list     = 'sstepanov@anl.gov';					#temp for testing
# $afterdump_list = 'sstepanov@anl.gov';					#temp for testing

### Open logfile ---------------------------------------------------------------
  &open_file();

### Set sleeptime --------------------------------------------------------------
  $sleeptime = 1;
  $sleepstep = 0.1;

### Set history length ---------------------------------------------------------
  @history            = ();
  $send_history       = 0;							#0=disable emailing history at problems, 1=email history at problems
  $send_afterdump     = 0;							#0=disable after-emailing, 1=email data after dumps, 2=email data after warnings
  $afterdump_interval = 120;							#seconds
  $maxhistory         = int(2*$afterdump_interval/$sleeptime+1.5);		#the length of history array
  $nhistory           = 0;							#current index of history array (no data yet)
  $last_email_time_ref= [0,0];
  $last_email_stamp   = 'YYYY-MM-DD_hh-mm-ss';
  $need_after_mail    = 0;
  %CA_Failure         = ();

### List of PVs of interest ----------------------------------------------------

  $epochPV     = '23o:epsmon_epoch';
  $timestampPV = '23o:epsmon_timestamp';
### Trip levels ----------------------------------------------------------------
  $P_watch  = 1.000e-7;
  $P_warn   = 4.690e-6;
# $T_warn_1 = 102;                              				#2013/04/25 changed from 100 to 102 per Bob's request.
  $T_warn_1 = 105;                              				#2021/09/21 changed from 102 to 105 per Bob's request.
  $T_warn_2 = 310;
  $fluct_percent  = 25;                         				#value exceeding watch, but within fluctuation (watch*(1+0.01*fluctuation_percent))
  $fluct_maxtimes = 6;                          				#are only reported after $fluct_maxtimes*$sleeptime
  $fev_biv_same = 1;								#set to "1" if FEV/BIV trips are the same

### Random values for testing purposes
# $P_watch  = 1.000e-8;
# $P_warn   = 4.690e-8;
# $T_warn_1 = 100;
# $T_warn_2 = 100;

  $PVname     = 0;
  $PVdescr    = 1;
  $statusPrev = 2;
  $valuePrev  = 3;
  $P_nfluct   = 4;
  $T_setpnt   = 4;
  $LN2_min    = 4;
  $LN2_max    = 5;
  $F_shtr2cls = 4;
# $EPS_TRIPPED=0; $EPS_OPERATOR=1;  $EPS_OK=2;       $EPS_PRETRIPPED=3;
# $EPS_OFF=0;     $EPS_TRIPPED=1;   $EPS_OPERATOR=1; $EPS_OPERTRIP=3;   $EPS_OK=4;
  $EPS_TRIPPED=1; $EPS_OK=0;
  $FEV_MOVING=0;  $FEV_CLOSED=1;    $FEV_OPEN=2;     $FEV_ERROR=3;
  $BIV_MOVING=0;  $BIV_CLOSED=1;    $BIV_OPEN=2;     $BIV_ERROR=3;
  $C_OK=1;
  $F_OK=0;
  $IP_OK=1;
  $V_OPEN=1;
#                      0               1            2          3
### General EPS PV = PVname,         name,    CA_status_prev, Prev
  @EPS_PVs = (
     ['23id:BLEPS:A_FAULT_EXISTS', 'BL EPS',       0,    $EPS_OK],
     ['23id:BLEPS:A_TRIP_EXISTS',  'BL EPS',       0,    $EPS_OK]
  );
  @FEV_PVs = (
     ['EPS:23:ID:FEV:POSITION',    'FEV POSITION', 0,    $FEV_OPEN]
  );
  @BIV_PVs = (
     ['EPS:23:ID:BIV:POSITION',    'BIV POSITION', 0,    $BIV_OPEN]
  );
  $n_EPS_PVs = @EPS_PVs; 
  $n_FEV_PVs = @FEV_PVs; 
  $n_BIV_PVs = @BIV_PVs; 

#                  0                1                    2         3        4
# Pressure PVs = PVname,          Name,          CA_status_prev, Prev, nFluctuation
  @P_PVs   = (
#    ['23o:V3001_mbar', 'Pressure_mbar_WBSL_IDout',      0,        0,       0],
#    ['23o:V3003_mbar', 'Pressure_mbar_DCM_IDout',       0,        0,       0],
#    ['23o:V3005_mbar', 'Pressure_mbar_IBST_IDout',      0,        0,       0],
#    ['23o:V3007_mbar', 'Pressure_mbar_HDM_IDout',       0,        0,       0],
#    ['23o:V3123_mbar', 'Pressure_mbar_MBSH_IDout',      0,        0,       0],
#    ['23o:V3125_mbar', 'Pressure_mbar_KBM_IDout',       0,        0,       0],
#    ['23i:V3011_mbar', 'Pressure_mbar_IBSH_IDin',       0,        0,       0],
#    ['23i:V3127_mbar', 'Pressure_mbar_WBCO_IDin',       0,        0,       0],
#    ['23i:V3267_mbar', 'Pressure_mbar_WBSL_IDin',       0,        0,       0],
#    ['23i:V3271_mbar', 'Pressure_mbar_DCM_IDin',        0,        0,       0],
#    ['23i:V3273_mbar', 'Pressure_mbar_MBSL_IDin',       0,        0,       0],
#    ['23i:V3275_mbar', 'Pressure_mbar_KBM_IDin',        0,        0,       0]
  );
  $n_P_PVs = @P_PVs; 
  if ($n_P_PVs < 1) {$monitorPval = 0;}
#
#               0                      1                   2         3
# Ion Pump PVs = PVname,              Name,        CA_status_prev,  Prev
  @IP_PVs = (
     ['23ID:BLEPS:IP1_STATUS',  'Ion_pump_WBSL_IDout',      0,     $IP_OK],	#0=tripped, 1=OK
     ['23ID:BLEPS:IP2_STATUS',  'Ion_pump_DCM_IDout',       0,     $IP_OK],	#0=tripped, 1=OK
     ['23ID:BLEPS:IP3_STATUS',  'Ion_pump_IBST_IDout',      0,     $IP_OK],	#0=tripped, 1=OK
     ['23ID:BLEPS:IP4_STATUS',  'Ion_pump_HDM_IDout',       0,     $IP_OK],	#0=tripped, 1=OK
     ['23ID:BLEPS:IP5_STATUS',  'Ion_pump_MBSH_IDout',      0,     $IP_OK],	#0=tripped, 1=OK
     ['23ID:BLEPS:IP6_STATUS',  'Ion_pump_KBM_IDout',       0,     $IP_OK],	#0=tripped, 1=OK
     ['23ID:BLEPS:IP7_STATUS',  'Ion_pump_IBSH_IDin',       0,     $IP_OK],	#0=tripped, 1=OK
     ['23ID:BLEPS:IP8_STATUS',  'Ion_pump_WBCO_IDin',       0,     $IP_OK],	#0=tripped, 1=OK
     ['23ID:BLEPS:IP9_STATUS',  'Ion_pump_WBSL_IDin',       0,     $IP_OK],	#0=tripped, 1=OK
     ['23ID:BLEPS:IP10_STATUS', 'Ion_pump_DCM_IDin',        0,     $IP_OK],	#0=tripped, 1=OK
     ['23ID:BLEPS:IP11_STATUS', 'Ion_pump_MBSL_IDin',       0,     $IP_OK],	#0=tripped, 1=OK
     ['23ID:BLEPS:IP12_STATUS', 'Ion_pump_KBM_IDin',        0,     $IP_OK] 	#0=tripped, 1=OK
  );
  $n_IP_PVs = @IP_PVs;
  if ($n_IP_PVs < 1) {$monitorIonPumps = 0;}
#
#                     0                1                         2       3      4
# Temperature PVs = PVname,            Name,          CA_status_prev, Prev, TempSetPoint
  @T_PVs = (
     ['23i:BLEPS:TEMP6_CURRENT',  'Temperature_Kelvin_DCMin_XTAL1',   0, 0, $T_warn_1],
     ['23i:BLEPS:TEMP7_CURRENT',  'Temperature_Kelvin_DCMin_XTAL2',   0, 0, $T_warn_1],
     ['23i:BLEPS:TEMP8_CURRENT',  'Temperature_Kelvin_DCMin_SHIELD',  0, 0, $T_warn_2],
     ['23i:BLEPS:TEMP9_CURRENT',  'Temperature_Kelvin_DCMin_BPM_IB',  0, 0, $T_warn_2],
     ['23i:BLEPS:TEMP10_CURRENT', 'Temperature_Kelvin_DCMin_BPM_OB',  0, 0, $T_warn_2],
#
     ['23o:BLEPS:TEMP1_CURRENT', 'Temperature_Kelvin_DCMout_XTAL1',   0, 0, $T_warn_1],
     ['23o:BLEPS:TEMP2_CURRENT',  'Temperature_Kelvin_DCMout_XTAL2',  0, 0, $T_warn_1],
     ['23o:BLEPS:TEMP3_CURRENT',  'Temperature_Kelvin_DCMout_SHIELD', 0, 0, $T_warn_2],
     ['23o:BLEPS:TEMP4_CURRENT', 'Temperature_Kelvin_DCMout_BPM_IB',  0, 0, $T_warn_2],
     ['23o:BLEPS:TEMP5_CURRENT', 'Temperature_Kelvin_DCMout_BPM_OB',  0, 0, $T_warn_2]
  );
  $n_T_PVs = @T_PVs;
  if ($n_T_PVs < 1) {$monitorT = 0;}
#
#               0                       1               2         3
# Valve PVs = PVname,                 Name,      CA_status_prev,  Prev
  @V_PVs = (
     ['23ID:BLEPS:GV1_OPENED_LS',  'Valve_ComA1',       0,     $V_OPEN],	#0=open,1=closed,2=bothLimits,3=noneLimits
     ['23ID:BLEPS:GV2_OPENED_LS',  'Valve_ComA2',       0,     $V_OPEN],	#0=open,1=closed,2=bothLimits,3=noneLimits
     ['23ID:BLEPS:GV3_OPENED_LS',  'Valve_ComA3',       0,     $V_OPEN],	#0=open,1=closed,2=bothLimits,3=noneLimits
     ['23ID:BLEPS:GV4_OPENED_LS',  'Valve_ComA4',       0,     $V_OPEN],	#0=open,1=closed,2=bothLimits,3=noneLimits
     ['23ID:BLEPS:GV5_OPENED_LS',  'Valve_IDin_B1',     0,     $V_OPEN],	#0=open,1=closed,2=bothLimits,3=noneLimits
     ['23ID:BLEPS:GV6_OPENED_LS',  'Valve_IDin_C1',     0,     $V_OPEN],	#0=open,1=closed,2=bothLimits,3=noneLimits
     ['23ID:BLEPS:GV7_OPENED_LS',  'Valve_IDin_C2',     0,     $V_OPEN],	#0=open,1=closed,2=bothLimits,3=noneLimits
     ['23ID:BLEPS:GV8_OPENED_LS',  'Valve_IDin_C3',     0,     $V_OPEN],	#0=open,1=closed,2=bothLimits,3=noneLimits
     ['23ID:BLEPS:GV9_OPENED_LS',  'Valve_IDin_D1',     0,     $V_OPEN],	#0=open,1=closed,2=bothLimits,3=noneLimits
     ['23ID:BLEPS:GV10_OPENED_LS', 'Valve_IDout_B1',    0,     $V_OPEN],	#0=open,1=closed,2=bothLimits,3=noneLimits
     ['23ID:BLEPS:GV11_OPENED_LS', 'Valve_IDout_B2',    0,     $V_OPEN] 	#0=open,1=closed,2=bothLimits,3=noneLimits
  );
  $n_V_PVs = @V_PVs;
  if ($n_V_PVs < 1) {$monitorValves = 0;}
#
#               0             1                       2          3
# Cryo PVs = PVname,        Name,             CA_status_prev,  Prev
  @C_PVs = (
     ['23ID:BLEPS:Cryo1_Ready', 'Cryo_IDout_ready',    0,    $C_OK],		#1=OK,0=tripped
     ['23ID:BLEPS:Cryo2_Ready', 'Cryo_IDin_ready',     0,    $C_OK],		#1=OK,0=tripped
  );
  $n_C_PVs = @C_PVs;
#
#               0             1                          2         3    4    5
# LN2 PVs = PVname,       Name,               CA_status_prev,    Prev  Min  Max
  @LN2_PVs = (
     ['23i:CC:PV_LT19',  '23ID-D LN2 Level',             0,       50,  10., 94.],
     ['23o:CC:PV_LT19',  '23ID-B LN2 Level',             0,       50,  10., 94.]
  );
  $n_LN2_PVs = @LN2_PVs;
#               0             1                          2         3
  $LN2_V_OPEN = 'Opened';
# LN2_V PVs = PVname,       Name,               CA_status_prev,  Prev
  @LN2_V_PVs = (
     ['LN:S23ValveStatus', 'APS LN2 Valve ',             0,     $LN2_V_OPEN]          #0=unknown,1=closed,2=open,3=SwitchFlt,4-unknown,5=closed,6=open,7=SwitchFlt
  );
  $n_LN2_V_PVs = @LN2_V_PVs;
#
#               0             1                         2          3        4
# Flow PVs = PVname,        Name,               CA_status_prev,  Prev   Shutter2close
  @F_PVs = (
     ['23id:BLEPS:FLOW1_TRIP', 'WBSL New/Out flow',      0,      $F_OK,   'FE'],	#1=OK,0=tripped
     ['23id:BLEPS:FLOW2_TRIP', 'WBSL Ole/In flow',       0,      $F_OK,    'C'],	#1=OK,0=tripped
  );
  $n_F_PVs = @F_PVs;

  @energyPVs = ('23i:MO:E:ActPos','23o:MO:E:ActPos');

### Set monitors ---------------------------------------------------------------

  print STDOUT "\n".'Setting PV monitors ...'."\n";
  do {
     $status = 0;
     for ($i=0; $i<$n_EPS_PVs; $i++)   {$status += &addCaMonitor($EPS_PVs[$i][$PVname]);}       #EPS status
     for ($i=0; $i<$n_FEV_PVs; $i++)   {$status += &addCaMonitor($FEV_PVs[$i][$PVname]);}       #FEV status
     for ($i=0; $i<$n_BIV_PVs; $i++)   {$status += &addCaMonitor($BIV_PVs[$i][$PVname]);}       #BIV status
     for ($i=0; $i<$n_P_PVs; $i++)     {$status += &addCaMonitor($P_PVs[$i][$PVname]);}         #Pressures
     for ($i=0; $i<$n_T_PVs; $i++)     {$status += &addCaMonitor($T_PVs[$i][$PVname]);}         #Temperatures
     for ($i=0; $i<$n_V_PVs; $i++)     {$status += &addCaMonitor($V_PVs[$i][$PVname]);}         #Valves
     for ($i=0; $i<$n_C_PVs; $i++)     {$status += &addCaMonitor($C_PVs[$i][$PVname]);}         #Cryo
     for ($i=0; $i<$n_LN2_PVs; $i++)   {$status += &addCaMonitor($LN2_PVs[$i][$PVname]);}       #Cryo
     for ($i=0; $i<$n_LN2_V_PVs; $i++) {$status += &addCaMonitor($LN2_V_PVs[$i][$PVname]);}     #Cryo
     for ($i=0; $i<$n_F_PVs; $i++)     {$status += &addCaMonitor($F_PVs[$i][$PVname]);}         #Water flow
     $status += &addCaMonitor($energyPVs[0]);
     $status += &addCaMonitor($energyPVs[1]);
     if ($status) {print STDOUT '  ... failed to set '.$status.' PV monitors. Retrying ...'."\n";}
  } while ($status != 0);
  print STDOUT '  ... all PV monitors are set.'."\n";


### Add PV descriptions from the DESC field in case they changed
### compared to what is hardcoded into this script:
  print STDOUT "\n".'Reading PV descriptions ...'."\n";
  for ($i=0; $i<$n_EPS_PVs; $i++) {&addDescr($EPS_PVs[$i][$PVname],\$EPS_PVs[$i][$PVdescr]);}
  for ($i=0; $i<$n_FEV_PVs; $i++) {&addDescr($FEV_PVs[$i][$PVname],\$FEV_PVs[$i][$PVdescr]);}
  for ($i=0; $i<$n_BIV_PVs; $i++) {&addDescr($BIV_PVs[$i][$PVname],\$BIV_PVs[$i][$PVdescr]);}
  for ($i=0; $i<$n_P_PVs;   $i++) {&addDescr($P_PVs[$i][$PVname],  \$P_PVs[$i][$PVdescr]);}
  for ($i=0; $i<$n_T_PVs;   $i++) {&addDescr($T_PVs[$i][$PVname],  \$T_PVs[$i][$PVdescr]);}
  for ($i=0; $i<$n_C_PVs;   $i++) {&addDescr($C_PVs[$i][$PVname],  \$C_PVs[$i][$PVdescr]);}
  for ($i=0; $i<$n_V_PVs;   $i++) {&addDescr($V_PVs[$i][$PVname],  \$V_PVs[$i][$PVdescr]);}
  for ($i=0; $i<$n_F_PVs;   $i++) {&addDescr($F_PVs[$i][$PVname],  \$F_PVs[$i][$PVdescr]);}

### Headers for history:
  $nvar = 0;
  $headers[$nvar] = 'timestamp';         $nvar++;
  for ($i=0; $i<$n_P_PVs;   $i++) {$headers[$nvar] = $P_PVs[$i][$PVdescr];   $nvar++;}
  for ($i=0; $i<$n_T_PVs;   $i++) {$headers[$nvar] = $T_PVs[$i][$PVdescr];   $nvar++;}
  for ($i=0; $i<$n_C_PVs;   $i++) {$headers[$nvar] = $C_PVs[$i][$PVdescr];   $nvar++;}
  for ($i=0; $i<$n_V_PVs;   $i++) {$headers[$nvar] = $V_PVs[$i][$PVdescr];   $nvar++;}
  for ($i=0; $i<$n_F_PVs;   $i++) {$headers[$nvar] = $F_PVs[$i][$PVdescr];   $nvar++;}
  for ($i=0; $i<$n_EPS_PVs; $i++) {$headers[$nvar] = $EPS_PVs[$i][$PVdescr]; $nvar++;}
  for ($i=0; $i<$n_FEV_PVs; $i++) {$headers[$nvar] = $FEV_PVs[$i][$PVdescr]; $nvar++;}
  for ($i=0; $i<$n_BIV_PVs; $i++) {$headers[$nvar] = $BIV_PVs[$i][$PVdescr]; $nvar++;}

###----------------------------------------------------------------------------
### Get data, repeat -----------------------------------------------------------
  $last_now_day = -1;

  print STDOUT "\n".'Entering the monitoring loop ...'."\n";
  $last_output_time_ref = [gettimeofday()];						#this is reference to ($epoch,$now_usec)
  while (1) { #================================================ MAIN LOOP ============================================================
     ($epoch,$now_usec) = gettimeofday();						#Time::HiRes: (seconds, mircroseconds) since epoch;
     $timestamp  = scalar localtime($epoch);     					#This is timestamp
     ($now_ss,$now_mm,$now_hh,$now_day,$now_mon,$now_year) = localtime($epoch);		#(sec,min,hour,mday,mon,year,wday,yday,isdst)
     $shortstamp = sprintf '%02d:%02d:%02d.%03d',
                           $now_hh,$now_mm,$now_ss,int($now_usec/1000);

     $status = Pezca::PutDoubleOldCa($epochPV,$epoch);
#    if ($status) {&report_only($script, $status, $epochPV);}
     $status = Pezca::PutStringOldCa($timestampPV,$timestamp);
#    if ($status) {&report_only($script, $status, $timestampPV);}

     if ($now_day != $last_now_day) {
        $day_changed = 1;
        $n_email_sent = 0;									#reset email cap
        print STDOUT "\n\n",$timestamp,"\n\n";
        print LOG    "\n\n",$timestamp,"\n\n";
     } else {
        $day_changed = 0;
     }
     $last_now_day = $now_day;
     $trips    = 0;
     $warnings = 0;
     $recovers = 0;
     $message = '';
     
     ### Record beamline energies
     for ($i=0; $i<2; $i++) {
        ($status, $energy[$i]) = Pezca::GetDouble($energyPVs[$i]);
        if ($status) {
#          &report_only($script, $status, $energyPVs[$i]);
           $energy[$i] = 0;
        }
     }

     ### EPS Status ---------------------------------------------------------------
     if ($monitorEPS) {
        for ($i=0; $i<$n_EPS_PVs; $i++) {
           ($status, $datum) = Pezca::GetDouble($EPS_PVs[$i][$PVname]);
           if ($status) {
#             &report_only($script, $status, $EPS_PVs[$i][$PVname]);
              $datum = $EPS_PVs[$i][$valuePrev];
              $EPS_TAG = 'CA_FAILURE';
              if ($EPS_PVs[$i][$statusPrev] == 0) {
                 $warnings += 1;
                 $message .= $timestamp.' [CA err] PV='.$EPS_PVs[$i][$PVname].' ('.$EPS_PVs[$i][$PVdescr].')'."\n";
              }
              elsif ($day_changed) {
                 $warnings += 1;
                 $message .= $timestamp.' [CA err] PV='.$EPS_PVs[$i][$PVname].' ('.$EPS_PVs[$i][$PVdescr].') -- ongoing'."\n";
              }
           }
           else {										#PV is accessible
              if ($status != $EPS_PVs[$i][$statusPrev]) {
                 $recovers += 1;
                 $message .= $timestamp.' [OK] PV='.$EPS_PVs[$i][$PVname].' ('.$EPS_PVs[$i][$PVdescr].') CA_ACCESS restored !'."\n";
              }
              ($statui, $EPS_TAG) = Pezca::GetString($EPS_PVs[$i][$PVname]);
              if (!defined $EPS_TAG) {$EPS_TAG = 'CA_FAILURE';}
              $datum = sprintf '%.0f',$datum;                         			#round
              ### $EPS_TRIPPED=1; $EPS_OK=0;
              if ($datum == $EPS_TRIPPED  || 
                  $datum > 1) {
                 if ($EPS_PVs[$i][$valuePrev] != $EPS_TRIPPED  && 
                     $EPS_PVs[$i][$valuePrev] <= 1) {
                    $trips += 1;
                    $message .= $timestamp.' [WARN] '.$EPS_PVs[$i][$PVdescr].' TRIPPED';
                    if ($datum == $EPS_TRIPPED) {$message .= ' !!!'."\n";}
                    else                        {$message .= ' (STATE='.$datum.' TAG='.$EPS_TAG.') !!!'."\n".$msg_flags;}
                 }
                 elsif ($day_changed) {
                    $warnings += 1;
                    $message .= $timestamp.' [WARN] '.$EPS_PVs[$i][$PVdescr].' TRIPPED, ongoing';
                    if ($datum == $EPS_TRIPPED) {$message .= '.'."\n";}
                    else                        {$message .= ' (STATE='.$datum.' TAG='.$EPS_TAG.').'."\n".$msg_flags;}
                 }
              }
              elsif ($datum == $EPS_OK) {
                 if ($EPS_PVs[$i][$valuePrev] != $datum) {
                    $recovers += 1;
                    $message .= $timestamp.' [OK] '.$EPS_PVs[$i][$PVdescr].' status is back to normal !'."\n";
                 }
              }
              else  {
                 if ($datum != $EPS_PVs[$i][$valuePrev]) {
                    $warnings += 1;
                    $message .= $timestamp.' [WARN] UNEXPECTED '.$EPS_PVs[$i][$PVdescr].' STATE='.$datum.' (TAG='.$EPS_TAG.') !!!'."\n".$msg_flags;
                 }
                 elsif ($day_changed) {
                    $warnings += 1;
                    $message .= $timestamp.' [WARN] UNLISTED '.$EPS_PVs[$i][$PVdescr].' STATE='.$datum.' (ongoing)'."\n".$msg_flags;
                 }
              }
           }
#          print STDOUT '---> datum=',$datum,'  valuePrev=',$valuePrev,'  EPS[prev]=',$EPS_PVs[$i][$valuePrev],"\n";
           $EPS_PVs[$i][$valuePrev]  = $datum;
           $EPS_PVs[$i][$statusPrev] = $status;
        }
        ### FEV Status --------------------------------------------------------------
        for ($i=0; $i<$n_FEV_PVs; $i++) {
           ($status, $datum) = Pezca::GetDouble($FEV_PVs[$i][$PVname]);
           if ($status) {
#             &report_only($script, $status, $FEV_PVs[$i][$PVname]);
              $datum = $FEV_PVs[$i][$valuePrev];
              if ($FEV_PVs[$i][$statusPrev] == 0) {
                 $warnings += 1;
                 $message .= $timestamp.' [CA err] PV='.$FEV_PVs[$i][$PVname].' ('.$FEV_PVs[$i][$PVdescr].')'."\n";
              }
              elsif ($day_changed) {
                 $warnings += 1;
                 $message .= $timestamp.' [CA err] PV='.$FEV_PVs[$i][$PVname].' ('.$FEV_PVs[$i][$PVdescr].') -- ongoing'."\n";
              }
           }
           else  {
              if ($status != $FEV_PVs[$i][$statusPrev]) {
                 $recovers += 1;
                 $message .= $timestamp.' [OK] PV='.$FEV_PVs[$i][$PVname].' ('.$FEV_PVs[$i][$PVdescr].') CA_ACCESS restored !'."\n";
              }
              $datum = sprintf '%.0f',$datum;                         			#round
              if ($datum != $FEV_OPEN) {
                 if ($FEV_PVs[$i][$valuePrev] != $datum) {
                    if ($fev_biv_same == 0) {
                       $trips += 1;
                       $message .= $timestamp.' [WARN] '.$FEV_PVs[$i][$PVdescr].' TRIPPED !!!'."\n";
                    }
                    elsif ($fev_biv_same == 1 && $BIV_PVs[$i][$valuePrev] == $BIV_OPEN) {
                       $trips += 1;
                       $message .= $timestamp.' [WARN] '.$FEV_PVs[$i][$PVdescr].' TRIPPED !!!'."\n";
                    }
                 }
                 elsif ($day_changed) {
                    $warnings += 1;
                    $message .= $timestamp.' [WARN] '.$FEV_PVs[$i][$PVdescr].' tripped (ongoing)'."\n";
                 }
              }
              elsif ($datum == $FEV_OPEN) {
                 if ($FEV_PVs[$i][$valuePrev] != $datum) {
                    if ($fev_biv_same == 0) {
                       $recovers += 1;
                       $message .= $timestamp.' [OK] '.$FEV_PVs[$i][$PVdescr].' is back to open !'."\n";
                    }
                    elsif ($fev_biv_same == 1 && $BIV_PVs[$i][$valuePrev] != $BIV_OPEN) {
                       $recovers += 1;
                       $message .= $timestamp.' [OK] '.$FEV_PVs[$i][$PVdescr].' is back to open !'."\n";
                    }
                 }
              }
              else  {
                 if ($datum != $FEV_PVs[$i][$valuePrev]) {
                    $warnings += 1;
                    $message .= $timestamp.' [WARN] UNLISTED FEV STATE='.$datum.' !!!'."\n";
                 }
                 elsif ($day_changed) {
                    $warnings += 1;
                    $message .= $timestamp.' [WARN] UNLISTED FEV STATE='.$datum.' (ongoing)'."\n";
                 }
              }
           }
           $FEV_PVs[$i][$valuePrev]  = $datum;
           $FEV_PVs[$i][$statusPrev] = $status;
        }
        ### BIV Status --------------------------------------------------------------
        for ($i=0; $i<$n_BIV_PVs; $i++) {
           ($status, $datum) = Pezca::GetDouble($BIV_PVs[$i][$PVname]);
           if ($status) {
#             &report_only($script, $status, $BIV_PVs[$i][$PVname]);
              $datum = $BIV_PVs[$i][$valuePrev];
              if ($BIV_PVs[$i][$statusPrev] == 0) {
                 $warnings += 1;
                 $message .= $timestamp.' [CA err] PV='.$BIV_PVs[$i][$PVname].' ('.$BIV_PVs[$i][$PVdescr].')'."\n";
              }
              elsif ($day_changed) {
                 $warnings += 1;
                 $message .= $timestamp.' [CA err] PV='.$BIV_PVs[$i][$PVname].' ('.$BIV_PVs[$i][$PVdescr].') -- ongoing'."\n";
              }
           }
           else  {
              if ($status != $BIV_PVs[$i][$statusPrev]) {
                 $recovers += 1;
                 $message .= $timestamp.' [OK] PV='.$BIV_PVs[$i][$PVname].' ('.$BIV_PVs[$i][$PVdescr].') CA_ACCESS restored !'."\n";
              }
              $datum = sprintf '%.0f',$datum;                         			#round
              if ($datum != $BIV_OPEN) {
                 if ($BIV_PVs[$i][$valuePrev] != $datum) {
                    if ($fev_biv_same == 0) {
                       $trips += 1;
                       $message .= $timestamp.' [WARN] BIV TRIPPED !!!'."\n";
                    }
                    elsif ($fev_biv_same == 1 && $FEV_PVs[$i][$valuePrev] == $FEV_OPEN) {
                       $trips += 1;
                       $message .= $timestamp.' [WARN] BIV/FEV TRIPPED !!!'."\n";
                    }
                 }
                 elsif ($day_changed) {
                    $warnings += 1;
                    $message .= $timestamp.' [WARN] BIV tripped (ongoing)'."\n";
                 }
              }
              elsif ($datum == $BIV_OPEN) {
                 if ($BIV_PVs[$i][$valuePrev] != $datum) {
                    if ($fev_biv_same == 0) {
                       $recovers += 1;
                       $message .= $timestamp.' [OK] BIV is back to open !'."\n";
                    }
                    elsif ($fev_biv_same == 1 && $FEV_PVs[$i][$valuePrev] != $FEV_OPEN) {
                       $recovers += 1;
                       $message .= $timestamp.' [OK] BIV/FEV is back to open !'."\n";
                    }
                 }
              }
              else  {
                 if ($datum != $BIV_PVs[$i][$valuePrev]) {
                    $warnings += 1;
                    $message .= $timestamp.' [WARN] UNLISTED BIV STATE='.$datum.' !!!'."\n";
                 }
                 elsif ($day_changed) {
                    $warnings += 1;
                    $message .= $timestamp.' [WARN] UNLISTED BIV STATE='.$datum.' (ongoing)'."\n";
                 }
              }
           }
           $BIV_PVs[$i][$valuePrev]  = $datum;
           $BIV_PVs[$i][$statusPrev] = $status;
        }
     }


     ### Pressures ---------------------------------------------------------------
     if ($monitorPval) {
        for ($i=0; $i<$n_P_PVs; $i++) {
           ($status,$datum) = Pezca::GetDouble($P_PVs[$i][$PVname]);
           if ($status) {
#             &report_only($script, $status, $P_PVs[$i][$PVname]);
              $datum = $P_PVs[$i][$valuePrev];
              if ($P_PVs[$i][$statusPrev] == 0) {                                          #CA_status_prev
                 $warnings += 1;
                 $message .= $timestamp.' [CA err] PV='.$P_PVs[$i][$PVname].' ('.$P_PVs[$i][$PVdescr].')'."\n";
              }
              elsif ($day_changed) {
                 $warnings += 1;
                 $message .= $timestamp.' [CA err] PV='.$P_PVs[$i][$PVname].' ('.$P_PVs[$i][$PVdescr].') -- ongoing'."\n";
              }
           }
           else {
              if ($status != $P_PVs[$i][$statusPrev]) {
                 $recovers += 1;
                 $message .= $timestamp.' [OK] PV='.$P_PVs[$i][$PVname].' ('.$P_PVs[$i][$PVdescr].') CA_ACCESS restored !'."\n";
              }
              if ($datum > $P_watch) {$P_PVs[$i][$P_nfluct]++;}
              else                   {$P_PVs[$i][$P_nfluct]=0;}
              if    (substr($energyPVs[0],0,3) eq substr($P_PVs[$i][$PVname],0,3)) {$en = sprintf '%.3f',$energy[0];}
              elsif (substr($energyPVs[1],0,3) eq substr($P_PVs[$i][$PVname],0,3)) {$en = sprintf '%.3f',$energy[1];}
              else                                                                 {$en = 'unknown';}
              if ($datum > $P_warn) {                                                      #severe excess pressure
                 if ($P_PVs[$i][$valuePrev] <= $P_warn) {
                    $warnings += 1;
#                   $message .= sprintf $timestamp.' [WARN] %s P=%.3e > %.3e !!! (at E='.$en.')'."\n",
#                                                  $P_PVs[$i][$PVdescr],$datum,$P_warn;
                    $message .= sprintf $timestamp.' [WARN] %s P=%.3e > %.3e !!! (at E='.$en.'), Prev=%.3e'."\n",
                                                   $P_PVs[$i][$PVdescr],$datum,$P_warn,$P_PVs[$i][$valuePrev];
                 }
                 elsif ($day_changed) {
                    $warnings += 1;
#                   $message .= sprintf $timestamp.' [WARN] %s P=%.3e > %.3e (ongoing, at E='.$en.')'."\n",
#                                                  $P_PVs[$i][$PVdescr],$datum,$P_warn;
                    $message .= sprintf $timestamp.' [WARN] %s P=%.3e > %.3e (ongoing, at E='.$en.'), Prev=%.3e'."\n",
                                                   $P_PVs[$i][$PVdescr],$datum,$P_warn,$P_PVs[$i][$valuePrev];
                 }
              }
              elsif ($datum > $P_watch*(1+0.01*$fluct_percent) ||
                    ($datum > $P_watch && $P_PVs[$i][$P_nfluct] > $fluct_maxtimes)) {      #less severe pressure
                 if ($P_PVs[$i][$valuePrev] <= $P_watch) {
                    $warnings += 1;
#                   $message .= sprintf $timestamp.' [watch] %s P=%.3e > %.3e !!! (at E='.$en.')'."\n",
#                                                   $P_PVs[$i][$PVdescr],$datum,$P_watch;
                    $message .= sprintf $timestamp.' [watch] %s P=%.3e > %.3e !!! (at E='.$en.'), Prev=%.3e'."\n",
                                                    $P_PVs[$i][$PVdescr],$datum,$P_watch,$P_PVs[$i][$valuePrev];
                 }
                 elsif ($day_changed) {
                    $warnings += 1;
                    $message .= sprintf $timestamp.' [watch] %s P=%.3e > %.3e (ongoing, at E='.$en.')'."\n",
                                                    $P_PVs[$i][$PVdescr],$datum,$P_watch;
                 }
              }
              elsif ($P_PVs[$i][$valuePrev] > $P_warn) {
                 $recovers += 1;
#                $message .= sprintf $timestamp.' [OK] %s restored: P=%.3e < %.3e (at E='.$en.')'."\n",
#                                                  $P_PVs[$i][$PVdescr],$datum,$P_warn;
                 $message .= sprintf $timestamp.' [OK] %s restored: P=%.3e < %.3e (at E='.$en.'), Prev=%.3e'."\n",
                                                   $P_PVs[$i][$PVdescr],$datum,$P_warn,$P_PVs[$i][$valuePrev];
              }
           }
           $P_PVs[$i][$valuePrev]  = $datum;
           $P_PVs[$i][$statusPrev] = $status;
        }
     }


     ### Temperatures ------------------------------------------------------------
     if ($monitorT) {
        for ($i=0; $i<$n_T_PVs; $i++) {
           ($status,$datum) = Pezca::GetDouble($T_PVs[$i][$PVname]);
           if ($status) {
#             &report_only($script, $status, $T_PVs[$i][$PVname]);
              $datum = $T_PVs[$i][$valuePrev];
              if ($T_PVs[$i][$statusPrev] == 0) {
                 $warnings += 1;
                 $message .= $timestamp.' [CA err] PV='.$T_PVs[$i][$PVname].' ('.$T_PVs[$i][$PVdescr].')'."\n";
              }
              elsif ($day_changed) {
                 $warnings += 1;
                 $message .= $timestamp.' [CA err] PV='.$T_PVs[$i][$PVname].' ('.$T_PVs[$i][$PVdescr].') -- ongoing'."\n";
              }
           }
           else  {
              if ($status != $T_PVs[$i][$statusPrev]) {
                 $recovers += 1;
                 $message .= $timestamp.' [OK] PV='.$T_PVs[$i][$PVname].' ('.$T_PVs[$i][$PVdescr].') CA_ACCESS restored !'."\n";
              }
              if    (substr($energyPVs[0],0,3) eq substr($T_PVs[$i][$PVname],0,3)) {$en = sprintf '%.3f',$energy[0];}
              elsif (substr($energyPVs[1],0,3) eq substr($T_PVs[$i][$PVname],0,3)) {$en = sprintf '%.3f',$energy[1];}
              else                                                                 {$en = 'unknown';}
              if ($datum > $T_PVs[$i][$T_setpnt]) {
                 if ($T_PVs[$i][$valuePrev] <= $T_PVs[$i][$T_setpnt]) {
                    $warnings += 1;
                    $message .= sprintf $timestamp.' [WARN] %s T=%.3f > %.3f !!! (at E='.$en.')'."\n",
                                     $T_PVs[$i][$PVdescr],$datum,$T_PVs[$i][$T_setpnt];
                 }
                 elsif ($day_changed) {
                    $warnings += 1;
                    $message .= sprintf $timestamp.' [WARN] %s T=%.3f > %.3f (ongoing, at E='.$en.')'."\n",
                                      $T_PVs[$i][$PVdescr],$datum,$T_PVs[$i][$T_setpnt];
                 }
              }
              elsif ($T_PVs[$i][$valuePrev] > $T_PVs[$i][$T_setpnt]) {
                 $recovers += 1;
                 $message .= sprintf $timestamp.' [OK] %s restored: T=%.3f < %.3f (at E='.$en.')'."\n",
                                     $T_PVs[$i][$PVdescr],$datum,$T_PVs[$i][$T_setpnt];
              }
           }
           $T_PVs[$i][$valuePrev]  = $datum;
           $T_PVs[$i][$statusPrev] = $status;
        }
     }


     ### Ion Pumps ------------------------------------------------------------------
     if ($monitorIonPumps) {
        for ($i=0; $i<$n_IP_PVs; $i++) {
           ($status,$datum) = Pezca::GetDouble($IP_PVs[$i][$PVname]);
           if ($status) {
#             &report_only($script, $status, $IP_PVs[$i][$PVname]);
              $datum = $IP_PVs[$i][$valuePrev];
              if ($IP_PVs[$i][$statusPrev] == 0) {
                 $warnings += 1;
                 $message .= $timestamp.' [CA err] PV='.$IP_PVs[$i][$PVname].' ('.$IP_PVs[$i][$PVdescr].')'."\n";
              }
              elsif ($day_changed) {
                 $warnings += 1;
                 $message .= $timestamp.' [CA err] PV='.$IP_PVs[$i][$PVname].' ('.$IP_PVs[$i][$PVdescr].' -- ongoing'."\n";
              }
           }
           else {
              if ($status != $IP_PVs[$i][$statusPrev]) {
                 $recovers += 1;
                 $message .= $timestamp.' [OK] PV='.$IP_PVs[$i][$PVname].' ('.$IP_PVs[$i][$PVdescr].') CA_ACCESS restored !'."\n";
              }
              $datum = sprintf '%.0f',$datum;                              		#round
              if ($datum != $IP_OK) {
                 if ($IP_PVs[$i][$valuePrev] != $datum) {
                    $trips += 1;
                    $message .= $timestamp.' [WARN] '.$IP_PVs[$i][$PVdescr].' TRIPPED !!!'."\n";
                 }
                 elsif ($day_changed) {
                    $warnings += 1;
                    $message .= $timestamp.' [WARN] '.$IP_PVs[$i][$PVdescr].' TRIPPED (ongoing)'."\n";
                 }
              }
              elsif ($IP_PVs[$i][$valuePrev] != $datum) {
                 $recovers += 1;
                 $message .= $timestamp.' [OK] '.$IP_PVs[$i][$PVdescr].' TRIP ended'."\n";
              }
           }
           $IP_PVs[$i][$valuePrev]  = $datum;
           $IP_PVs[$i][$statusPrev] = $status;
        }
     }


     ### Valves ------------------------------------------------------------------
     if ($monitorValves) {
        for ($i=0; $i<$n_V_PVs; $i++) {
           ($status,$datum) = Pezca::GetDouble($V_PVs[$i][$PVname]);
           if ($status) {
#             &report_only($script, $status, $V_PVs[$i][$PVname]);
              $datum = $V_PVs[$i][$valuePrev];
              if ($V_PVs[$i][$statusPrev] == 0) {
                 $warnings += 1;
                 $message .= $timestamp.' [CA err] PV='.$V_PVs[$i][$PVname].' ('.$V_PVs[$i][$PVdescr].')'."\n";
              }
              elsif ($day_changed) {
                 $warnings += 1;
                 $message .= $timestamp.' [CA err] PV='.$V_PVs[$i][$PVname].' ('.$V_PVs[$i][$PVdescr].' -- ongoing'."\n";
              }
           }
           else {
              if ($status != $V_PVs[$i][$statusPrev]) {
                 $recovers += 1;
                 $message .= $timestamp.' [OK] PV='.$V_PVs[$i][$PVname].' ('.$V_PVs[$i][$PVdescr].') CA_ACCESS restored !'."\n";
              }
              $datum = sprintf '%.0f',$datum;                              		#round
              if ($datum != $V_OPEN) {
                 if ($V_PVs[$i][$valuePrev] != $datum) {
                    $trips += 1;
                    $message .= $timestamp.' [WARN] '.$V_PVs[$i][$PVdescr].' TRIPPED !!!'."\n";
                 }
                 elsif ($day_changed) {
                    $warnings += 1;
                    $message .= $timestamp.' [WARN] '.$V_PVs[$i][$PVdescr].' TRIPPED (ongoing)'."\n";
                 }
              }
              elsif ($V_PVs[$i][$valuePrev] != $datum) {
                 $recovers += 1;
                 $message .= $timestamp.' [OK] '.$V_PVs[$i][$PVdescr].' TRIP ended'."\n";
              }
           }
           $V_PVs[$i][$valuePrev]  = $datum;
           $V_PVs[$i][$statusPrev] = $status;
        }
     }


     ### LN2 ---------------------------------------------------------------------
     if ($monitorLN2) {
        for ($i=0; $i<$n_C_PVs; $i++) {
           ($status,$datum) = Pezca::GetDouble($C_PVs[$i][$PVname]);
           if ($status) {
#             &report_only($script, $status, $C_PVs[$i][$PVname]);
              $datum = $C_PVs[$i][$valuePrev];
              if ($C_PVs[$i][$statusPrev] == 0) {
#                $warnings += 1;
#                $message .= $timestamp.' [CA err] PV='.$C_PVs[$i][$PVname].' ('.$C_PVs[$i][$PVdescr].')'."\n";
                 &CAerr_age($C_PVs[$i][$PVname]);                                       #record epoch of error start
              }
              elsif (&CAerr_age($C_PVs[$i][$PVname]) > $CAerr_timeout) {
                 $warnings += 1;
                 $age = &CAerr_age($C_PVs[$i][$PVname],-1);                             #mark "already reported"
                 $message .= $timestamp.' [CA err] PV='.$C_PVs[$i][$PVname].' ('.$C_PVs[$i][$PVdescr].') -- longer than '.$age.'s'."\n";
              }
              elsif ($day_changed) {
                 &CAerr_age($C_PVs[$i][$PVname],-1);                                    #mark "already reported"
                 $warnings += 1;
                 $message .= $timestamp.' [CA err] PV='.$C_PVs[$i][$PVname].' ('.$C_PVs[$i][$PVdescr].') -- ongoing'."\n";
              }
           }
           else {
              if ($status != $C_PVs[$i][$statusPrev]) {
                 if (&CAerr_age($C_PVs[$i][$PVname]) == -1) {
                    $recovers += 1;
                    $message .= $timestamp.' [OK] PV='.$C_PVs[$i][$PVname].' ('.$C_PVs[$i][$PVdescr].') CA_ACCESS restored !'."\n";
                 }
                 &clear_CAerr($C_PVs[$i][$PVname]);                                     #erase record for the PV
              }
              $datum = sprintf '%.0f',$datum;                                           #round
              if ($datum != $C_OK) {
                 if ($C_PVs[$i][$valuePrev] != $datum) {
                    $trips += 1;
                    $message .= sprintf $timestamp.' [WARN] %s TRIPPED !!!'."\n", $C_PVs[$i][$PVdescr];
                 }
                 elsif ($day_changed) {
                    $warnings += 1;
                    $message .= sprintf $timestamp.' [WARN] %s TRIPPED (ongoing)'."\n", $C_PVs[$i][$PVdescr];
                 }
              }
              elsif ($C_PVs[$i][$valuePrev] != $datum) {
                 $recovers += 1;
                 $message .= sprintf $timestamp.' [OK] %s TRIP ended'."\n", $C_PVs[$i][$PVdescr];
              }
           }
           $C_PVs[$i][$valuePrev]  = $datum;
           $C_PVs[$i][$statusPrev] = $status;
        }

        for ($i=0; $i<$n_LN2_PVs; $i++) {
           ($status,$datum) = Pezca::GetDouble($LN2_PVs[$i][$PVname]);
           if ($status) {
#             &report_only($script, $status, $LN2_PVs[$i][$PVname]);
              $datum = $LN2_PVs[$i][$valuePrev];
              if ($LN2_PVs[$i][$statusPrev] == 0) {                                     #CA_status_prev
#                $warnings += 1;
#                $message .= $timestamp.' [CA err] PV='.$LN2_PVs[$i][$PVname].' ('.$LN2_PVs[$i][$PVdescr].')'."\n";
                 &CAerr_age($LN2_PVs[$i][$PVname]);                                     #record epoch of error start
              }
              elsif (&CAerr_age($LN2_PVs[$i][$PVname]) > $CAerr_timeout) {
                 $warnings += 1;
                 $age = &CAerr_age($LN2_PVs[$i][$PVname],-1);                           #mark "already reported"
                 $message .= $timestamp.' [CA err] PV='.$LN2_PVs[$i][$PVname].' ('.$LN2_PVs[$i][$PVdescr].') -- longer than '.$age.'s'."\n";
              }
              elsif ($day_changed) {
                 &CAerr_age($LN2_PVs[$i][$PVname],-1);                                  #mark "already reported"
                 $warnings += 1;
                 $message .= $timestamp.' [CA err] PV='.$LN2_PVs[$i][$PVname].' ('.$LN2_PVs[$i][$PVdescr].') -- ongoing'."\n";
              }
           }
           else {
              if ($status != $LN2_PVs[$i][$statusPrev]) {
                 if (&CAerr_age($LN2_PVs[$i][$PVname]) == -1) {
                    $recovers += 1;
                    $message .= $timestamp.' [OK] PV='.$LN2_PVs[$i][$PVname].' ('.$LN2_PVs[$i][$PVdescr].') CA_ACCESS restored !'."\n";
                 }
                 &clear_CAerr($LN2_PVs[$i][$PVname]);                                   #erase record for the PV
              }
              if ($datum > $LN2_PVs[$i][$LN2_max]) {
                 if ($LN2_PVs[$i][$valuePrev] <= $LN2_PVs[$i][$LN2_max]) {
                    $warnings += 1;
                    $message .= sprintf $timestamp.' [WARN] %s LN2 Level=%.0f%s > %.0f%s !!!'."\n",
                                     $LN2_PVs[$i][$PVdescr],$datum,'%',$LN2_PVs[$i][$LN2_max],'%';
                 }
                 elsif ($day_changed) {
                    $warnings += 1;
                    $message .= sprintf $timestamp.' [WARN] %s LN2 Level=%.0f%s > %.0f%s (ongoing)'."\n",
                                      $LN2_PVs[$i][$PVdescr],$datum,'%',$LN2_PVs[$i][$LN2_max];
                 }
              }
              elsif ($datum < $LN2_PVs[$i][$LN2_min]) {
                 if ($LN2_PVs[$i][$valuePrev] > $LN2_PVs[$i][$LN2_min]) {
                    $warnings += 1;
                    $message .= sprintf $timestamp.' [WARN] %s LN2 Level=%.0f%s < %.0f%s !!!'."\n",
                                     $LN2_PVs[$i][$PVdescr],$datum,'%',$LN2_PVs[$i][$LN2_min],'%';
                 }
                 elsif ($day_changed) {
                    $warnings += 1;
                    $message .= sprintf $timestamp.' [WARN] %s LN2 Level=%.0f%s < %.0f%s (ongoing)'."\n",
                                      $LN2_PVs[$i][$PVdescr],$datum,'%',$LN2_PVs[$i][$LN2_min];
                 }
              }
              elsif ($LN2_PVs[$i][$valuePrev] <= $LN2_PVs[$i][$LN2_min] ||
                     $LN2_PVs[$i][$valuePrev] >= $LN2_PVs[$i][$LN2_max]) {
                 $recovers += 1;
                 $message .= sprintf $timestamp.' [OK] %s restored: LN2 Level=%.0f%s is in limits [%.0f:%.0f]'."\n",
                                     $LN2_PVs[$i][$PVdescr],$datum,'%',$LN2_PVs[$i][$LN2_min],$LN2_PVs[$i][$LN2_max];
              }
           }
           $LN2_PVs[$i][$valuePrev]  = $datum;
           $LN2_PVs[$i][$statusPrev] = $status;
        }
        for ($i=0; $i<$n_LN2_V_PVs; $i++) {
           ($status,$datum) = Pezca::GetString($LN2_V_PVs[$i][$PVname]);
           if ($status) {
#             &report_only($script, $status, $LN2_V_PVs[$i][$PVname]);
              $datum = $LN2_V_PVs[$i][$valuePrev];
              if ($LN2_V_PVs[$i][$statusPrev] == 0) {
#                $warnings += 1;
#                $message .= $timestamp.' [CA err] PV='.$LN2_V_PVs[$i][$PVname].' ('.$LN2_V_PVs[$i][$PVdescr].')'."\n";
                 &CAerr_age($LN2_V_PVs[$i][$PVname]);                                   #record epoch of error start
              }
              elsif (&CAerr_age($LN2_V_PVs[$i][$PVname]) > $CAerr_timeout) {
                 $warnings += 1;
                 $age = &CAerr_age($LN2_V_PVs[$i][$PVname],-1);                         #mark "already reported"
                 $message .= $timestamp.' [CA err] PV='.$LN2_V_PVs[$i][$PVname].' ('.$LN2_V_PVs[$i][$PVdescr].') -- longer than '.$age.'s'."\n";
              }
              elsif ($day_changed) {
                 &CAerr_age($LN2_V_PVs[$i][$PVname],-1);                                #mark "already reported"
                 $warnings += 1;
                 $message .= $timestamp.' [CA err] PV='.$LN2_V_PVs[$i][$PVname].' ('.$LN2_V_PVs[$i][$PVdescr].') -- ongoing'."\n";
              }
           }
           else {
              if ($status != $LN2_V_PVs[$i][$statusPrev]) {
                 if (&CAerr_age($LN2_V_PVs[$i][$PVname]) == -1) {
                    $recovers += 1;
                    $message .= $timestamp.' [OK] PV='.$LN2_V_PVs[$i][$PVname].' ('.$LN2_V_PVs[$i][$PVdescr].') CA_ACCESS restored !'."\n";
                 }
                 &clear_CAerr($LN2_V_PVs[$i][$PVname]);                                 #erase record for the PV
              }
              if ($datum ne $LN2_V_OPEN) {
                 if ($LN2_V_PVs[$i][$valuePrev] ne $datum) {
                    $trips += 1;
                    $message .= $timestamp.' [WARN] '.$LN2_V_PVs[$i][$PVdescr].' NOT OPEN !!!'."\n";
                 }
                 elsif ($day_changed) {
                    $warnings += 1;
                    $message .= $timestamp.' [WARN] '.$LN2_V_PVs[$i][$PVdescr].' NOT OPEN (ongoing)'."\n";
                 }
              }
              elsif ($LN2_V_PVs[$i][$valuePrev] ne $datum) {
                 $recovers += 1;
                 $message .= $timestamp.' [OK] '.$LN2_V_PVs[$i][$PVdescr].' RE-OPENED'."\n";
              }
           }
           $LN2_V_PVs[$i][$valuePrev]  = $datum;
           $LN2_V_PVs[$i][$statusPrev] = $status;
        }
    }

     ### Water Flow Trips ---------------------------------------------------------------
     if ($monitorFlow) {
        for ($i=0; $i<$n_F_PVs; $i++) {
           ($status,$datum) = Pezca::GetDouble($F_PVs[$i][$PVname]);
           if ($status) {
#             &report_only($script, $status, $F_PVs[$i][$PVname]);
              $datum = $F_PVs[$i][$valuePrev];
              if ($F_PVs[$i][$statusPrev] == 0) {
                 $warnings += 1;
                 $message .= $timestamp.' [CA err] PV='.$F_PVs[$i][$PVname].' ('.$F_PVs[$i][$PVdescr].')'."\n";
              }
              elsif ($day_changed) {
                 $warnings += 1;
                 $message .= $timestamp.' [CA err] PV='.$F_PVs[$i][$PVname].' ('.$F_PVs[$i][$PVdescr].' -- ongoing'."\n";
              }
           }
           else {
              if ($status != $F_PVs[$i][$statusPrev]) {
                 if (&CAerr_age($F_PVs[$i][$PVname]) == -1) {
                    $recovers += 1;
                    $message .= $timestamp.' [OK] PV='.$F_PVs[$i][$PVname].' ('.$F_PVs[$i][$PVdescr].') CA_ACCESS restored !'."\n";
                 }
                 &clear_CAerr($F_PVs[$i][$PVname]);                                     #erase record for the PV
              }
              $datum = sprintf '%.0f',$datum;                                           #round
              if ($datum != $F_OK) {
                 if ($F_PVs[$i][$valuePrev] != $datum) {
                    $trips += 1;
                    $message .= sprintf $timestamp.' [WARN] %s TRIPPED !!!'."\n", $F_PVs[$i][$PVdescr];
                    if ($F_PVs[$i][$F_shtr2cls] eq 'FE') {
                       $message .= $timestamp.' [WARN] Closing FE shutter !!!'."\n";
                       $status = &closeFrontEndShutter('23o');
                    }
                    elsif ($F_PVs[$i][$F_shtr2cls] eq 'C') {
                       $message .= $timestamp.' [WARN] Closing C shutter !!!'."\n";
                       $status = &closeCshutter('23i'); 
                    }
                    else {
                       $message .= sprintf $timestamp.' [WARN] cannot close unrecognized shutter [%s] !!!'."\n", $F_PVs[$i][$F_shtr2cls];
                       $status = -1;
                    }
                    if ($status) {
                       $message .= $timestamp.' [WARN] failed to close shutter to protect WBSL at tripped water flow, status='.$status.' !!!'."\n";
                    }
                 }
                 elsif ($day_changed) {
                    $warnings += 1;
                    $message .= sprintf $timestamp.' [WARN] %s TRIPPED (ongoing)'."\n", $F_PVs[$i][$PVdescr];
                 }
              }
              elsif ($F_PVs[$i][$valuePrev] != $datum) {
                 $recovers += 1;
                 $message .= sprintf $timestamp.' [OK] %s TRIP ended'."\n", $F_PVs[$i][$PVdescr];
              }
           }
           $F_PVs[$i][$valuePrev]  = $datum;
           $F_PVs[$i][$statusPrev] = $status;
        }
     }


     $message .="\n";

     #print STDOUT "\n\n";
     #print STDOUT "@EPS_PVs\n";
     #print STDOUT "@FEV_PVs\n";
     #print STDOUT "@BIV_PVs\n";
     #print STDOUT "@P_PVs\n";
     #print STDOUT "@T_PVs\n";
     #print STDOUT "@C_PVs\n";
     #print STDOUT "@IP_PVs\n";
     #print STDOUT "@V_PVs\n";
     #print STDOUT "\n\n";
     #print STDOUT 'message = '.$message;

### Add last data to history:
     if ($nhistory > 0) {unshift @history, undef;}
     $nvar = 0;
     $history[0][$nvar] = $shortstamp;           $nvar++;
     for ($i=0; $i<$n_P_PVs;  $i++)    {$history[0][$nvar] = sprintf '%.3e', $P_PVs[$i][$valuePrev];     $nvar++;}
     for ($i=0; $i<$n_T_PVs;  $i++)    {$history[0][$nvar] = sprintf '%.3f', $T_PVs[$i][$valuePrev];     $nvar++;}
     for ($i=0; $i<$n_C_PVs;  $i++)    {$history[0][$nvar] = sprintf '%d',   $C_PVs[$i][$valuePrev];     $nvar++;}
     for ($i=0; $i<$n_IP_PVs; $i++)    {$history[0][$nvar] = sprintf '%d',   $IP_PVs[$i][$valuePrev];    $nvar++;}
     for ($i=0; $i<$n_V_PVs;  $i++)    {$history[0][$nvar] = sprintf '%d',   $V_PVs[$i][$valuePrev];     $nvar++;}
#    for ($i=0; $i<$n_LN2_PVs; $i++)   {$history[0][$nvar] = sprintf '%.1f', $LN2_PVs[$i][$valuePrev];   $nvar++;}
#    for ($i=0; $i<$n_LN2_V_PVs; $i++) {$history[0][$nvar] = sprintf '%.1f', $LN2_V_PVs[$i][$valuePrev]; $nvar++;}
     for ($i=0; $i<$n_EPS_PVs; $i++)   {$history[0][$nvar] = $EPS_PVs[$i][$valuePrev];                   $nvar++;}
     for ($i=0; $i<$n_FEV_PVs; $i++)   {$history[0][$nvar] = $FEV_PVs[$i][$valuePrev];                   $nvar++;}
     for ($i=0; $i<$n_BIV_PVs; $i++)   {$history[0][$nvar] = $BIV_PVs[$i][$valuePrev];                   $nvar++;}
     $nhistory++;
     if  ($nhistory > $maxhistory) {pop @history;  $nhistory--;}

     if ($trips || $warnings || $recovers) {
        print LOG    'EPS message @ '.$timestamp.'!'."\n";
        print LOG    $message."\n";
        print STDOUT "\n".'EPS message @ '.$timestamp.'!'."\n";
        print STDOUT $message."\n";
        &email();
        $last_output_time_ref = [$epoch,$now_usec];
	if (($trips    && $send_afterdump >= 1) ||
            ($warnings && $send_afterdump >= 2)) {					#we do not count recover messages as events
           $last_email_time_ref = [$epoch,$now_usec];
           $need_after_mail = 1;							#this is flag for after-dump that it needs to send another email few minutes after
	}
     } else {
	if (tv_interval($last_output_time_ref) > $sleeptime) {
           print STDOUT '.';
	   $last_output_time_ref = [$epoch,$now_usec];
	}
     }
     if ($send_afterdump && $need_after_mail && tv_interval($last_email_time_ref) > $afterdump_interval) {
        &email_afterdump();
    	$need_after_mail = 0;								#reset the flag indicating that error message was sent (the one we are following up)
     }

     $sleeping = 0;
     do {
        select(undef,undef,undef,$sleepstep);
	$sleeping += $sleepstep;
     } while ($sleeping < $sleeptime);
  } #=================================================== END OF MAIN LOOP ============================================================

  exit;

#---------------------------------------------------------------

  sub open_file {

     my ($filename, $filehandle, $outfile);
     my (@timearray, $day, $month, $year);
     @timearray = localtime(time);                                      		#this is (sec,min,hour,mday,mon,year,wday,yday,isdst)
     $day       = $timearray[3];
     $month     = $timearray[4] + 1;
     $year      = $timearray[5] + 1900;
     $outfile   = 'EPS_monitor';
     $filename  = sprintf $logdir.$outfile.'_%04d_%02d_%02d.log',$year,$month,$day;
     $filename  =~ s/ /0/g;
     if (-f $filename) {$filehandle = open (LOG,'>> '.$filename);}      		#open LOG file for writing
     else              {$filehandle = open (LOG,'>  '.$filename);}      		#open LOG file for writing
     if (!defined $filehandle) {
        print STDOUT 'Could not open LOG file: '.$filename."\n";
        &signal_exit();
     }
     ### Set unbuffered output
     select LOG; $|=1;
     ### Write logfile header;
     $timestamp = scalar localtime;
     print LOG 'EPS monitor started at: '.$timestamp."\n\n";
     print STDOUT $script.': Now logging data to '.$filename."\n".'Press CTRL+C to quit...'."\n\n";
  }

#---------------------------------------------------------------

  sub email {

    my ($msg, @lines, $nlines, $subject, $buf, $data, $i, $j);

    if (!defined $debug) {$debug = 0;}

    @lines  = split (/\n/, $message);
    $nlines = @lines;
    if    ($nlines > 2) {$subject = 'EPS ALERT!';}
    elsif ($nlines < 1) {$subject = 'EPS ALERT test';}
    else {
       $lines[0] =~ s/^.*\[/\[/;
       $subject = 'EPS ALERT: '.$lines[0];
    }

    $last_email_stamp = sprintf '%04d-%02d-%02d_%02d-%02d-%02d',
                              $now_year+1900,$now_mon+1,$now_day,$now_hh,$now_mm,$now_ss;

    $buf = '';
    if (($trips || $warnings) && $send_history && $nhistory > 0) {		     #we do not sent history if this is a recover message
       for ($j=0; $j<$nvar; $j++) {
          if ($j < $nvar-1) {$buf .='"'.$headers[$j].'", ';}
          else              {$buf .='"'.$headers[$j].'"'."\n";}
       }
       for ($i=0; $i<$nhistory; $i++) {
          for ($j=0; $j<$nvar; $j++) {
             $data = $history[$i][$j];
             if ($j < $nvar-1) {$buf .= $data.',  ';}
             else              {$buf .= $data."\n";}
          }
       }
    }

    if ($^O !~ /Win/i && $debug == 0) {
       if ($n_email_sent == $MAX_EMAILS_PER_DAY) {$message .= "\n".'>>> Daily quota of '.$MAX_EMAILS_PER_DAY.' emails reached.'."\n";}
       $msg = MIME::Lite->new(
          From     => 'sstepanov@anl.gov',
          To       => $email_list,
          Subject  => $subject,
          Type     => 'multipart/mixed',
       );
       $msg->attach(
          Type     => 'TEXT',                   					#case-sensitive TEXT/BINARY/AUTO
          Encoding => 'quoted-printable',
          Data     => $message."\n",
       );
       if (length($buf) > 0) {
          $msg->attach(
             Type     => 'BINARY',                   					#case-sensitive TEXT/BINARY/AUTO
             Encoding => 'base64',
             Data     => $buf."\n",
             Filename => 'EPS_history_'.$last_email_stamp.'.csv',
          );
       }
       if ($n_email_sent <= $MAX_EMAILS_PER_DAY) {
          $msg->send();
          $n_email_sent++;
          print STDOUT 'Mail sent to '.$email_list.' .'."\n";
       } else {
          print STDOUT 'Daily quota of '.$MAX_EMAILS_PER_DAY.' emails exceeded.'."\n";
          print LOG 'Mail was not sent because daily quota of '.$MAX_EMAILS_PER_DAY.' messages exceeded.'."\n";
          print LOG 'SUBJECT: '.$subject."\n";
          print LOG 'MESSAGE: '.$message."\n";
          if (length($buf) > 0) {print LOG 'HISTORY: '.$buf."\n";}
       }
    } else  {
       if ($^O =~ /Win/i) {print STDOUT 'Mail not sent: no sendmail client on Windows OS.'."\n";}
       print STDOUT 'SUBJECT: '.$subject."\n";
       print STDOUT 'MESSAGE: '.$message."\n";
       if (length($buf) > 0) {
          print STDOUT 'HISTORY: '.$buf."\n";
       }
    }
  }

#---------------------------------------------------------------

  sub email_afterdump {

    my ($msg, $subject, $buf, $data, $i, $j);

    if (!defined $debug) {$debug = 0;}

    $subject = 'After-dump for EPS ALERT of '.$last_email_stamp;

    $buf = '';
    if ($nhistory > 0) {
       for ($j=0; $j<$nvar; $j++) {
          if ($j < $nvar-1) {$buf .='"'.$headers[$j].'", ';}
          else              {$buf .='"'.$headers[$j].'"'."\n";}
       }
       for ($i=0; $i<$nhistory; $i++) {
          for ($j=0; $j<$nvar; $j++) {
             $data = $history[$i][$j];
             if ($j < $nvar-1) {$buf .= $data.',  ';}
             else              {$buf .= $data."\n";}
          }
       }
    }

    if ($^O !~ /Win/i && $debug == 0) {
       if ($n_email_sent == $MAX_EMAILS_PER_DAY) {$buf .= "\n".'>>> Daily quota of '.$MAX_EMAILS_PER_DAY.' emails reached.'."\n";}
       $msg = MIME::Lite->new(
          From     => 'sstepanov@anl.gov',
#         To       => $email_list,
          To       => $afterdump_list,
          Subject  => $subject,
          Type     => 'multipart/mixed',
       );
       if (length($buf) > 0) {
          $msg->attach(
             Type     => 'BINARY',                   					#case-sensitive TEXT/BINARY/AUTO
             Encoding => 'base64',
             Data     => $buf."\n",
             Filename => 'EPS_after-dump_'.$last_email_stamp.'.csv',
          );
       }
       if ($n_email_sent <= $MAX_EMAILS_PER_DAY) {
          $msg->send();
          $n_email_sent++;
          print STDOUT 'After-dump mail sent to '.$email_list.' .'."\n";
       } else {
          print STDOUT 'Daily quota of '.$MAX_EMAILS_PER_DAY.' emails exceeded.'."\n";
          print LOG 'Mail was not sent because daily quota of '.$MAX_EMAILS_PER_DAY.' messages exceeded.'."\n";
          print LOG 'SUBJECT: '.$subject."\n";
          print LOG 'MESSAGE: '.$buf."\n";
       }
    } else  {
       if ($^O =~ /Win/i) {print STDOUT 'After-dump mail not sent: no sendmail client on Windows OS.'."\n";}
       print STDOUT 'SUBJECT: '.$subject."\n";
       print STDOUT 'MESSAGE: '.$buf."\n";
    }
  }

#---------------------------------------------------------------

# sub addDescr($pv,\$comment) {
sub addDescr($\$) {

  my ($pv, $commentRef, $descr);
  $pv         = shift(@_);
  $commentRef = shift(@_);

  $descr = &getPVdescription($pv);
  if (length($descr) > 3) {$$commentRef .= ' (DESC: '.$descr.')';}
}

#---------------------------------------------------------------

# sub CAerr_age($pv,[-1]) {
sub CAerr_age($;$) {

  my ($pv, $clr, $age);
  $pv  = shift(@_);
  $clr = shift(@_);

  if (defined $CA_Failure{$pv}) {
     if ($CA_Failure{$pv} > 0) {$age = $epoch - $CA_Failure{$pv};}      # $epoch is current time defined in main program
     else                      {$age = -1;}                             # event reported
     if (defined $clr) {$CA_Failure{$pv} = -1;}
     return $age;
  }
  else {
     $CA_Failure{$pv} = $epoch;
     return 0;
  }
}

#---------------------------------------------------------------

# sub clear_CAerr($pv) {
sub clear_CAerr($) {

  my ($pv);
  $pv = shift(@_);

  if (defined $CA_Failure{$pv}) {undef $CA_Failure{$pv};}
}

#---------------------------------------------------------------

  sub signal_exit {

     $timestamp = scalar localtime;
     if (defined fileno(LOG)) {
        print LOG "\n".'Beamline monitor stopped at: '.$timestamp."\n\n";
        close LOG;
     }
     print STDOUT "\n\n"
                 .$script.': Exiting!'."\n";
     print STDOUT $script.': Clearing monitors... '."\n";
     &clearAllCaMonitors();
     print STDOUT $script.': Done.'."\n";
#    &press_key_exit();
     exit;
  }
