#!/usr/bin/perl -W
#--------------------------------------------------------------
#  This script saves Lock-In Amplifier (LIA) parameters to a data file
#  that can be later loaded back into LIA with restoreLIA.pl
#
#     	Authors: Sergey Stepanov & Sioan Zohar
#
# Version-1.0:  2014/06/10 using Pezca
#--------------------------------------------------------------
  use strict;
  use Pezca;

  our ($beamline, $NONINTERACTIVE, $IGNORE_ERRORS, $USE_ZENITY_FOR_EXIT);
  my ($PEZCA, @beamlinelist, $script, $syntax, $narg, $snap, $i, $arg);
  my ($lockin_asyn_PV, $fc, $v, $handle, $user, $currentTime, $portState);

  our ($phase, $amplitude, $frequency, $gain,$timeConstant, $display);
  our ($lineFilter, $filterRollOff, $frontOutput);

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;

  do $PEZCA.'/Common/common_subs.pl';
  do $PEZCA.'/Common/beamline.pl';

  select STDERR; $|=1;			# set unbuffered output
  select STDOUT; $|=1;			# set unbuffered output

  $NONINTERACTIVE      = 1;		# set as "1" to skip all dialogs
  $IGNORE_ERRORS       = 1;		# ignore EZCA errors
  $USE_ZENITY_FOR_EXIT = 1;		# use graphical popups

  @beamlinelist = ('23i', '23o', '23b', '23d');
  $script       = &makeScriptName();
  $syntax       = "\n\n".'Syntax: '.$script.' snapFile [beamline]'."\n";

  $narg = @ARGV;

  if ($NONINTERACTIVE == 0) {undef $NONINTERACTIVE;}

### Parse command line arguments:
  for ($i=0; $i<$narg; $i++ ) {
     $arg = $ARGV[$i];
     if (grep(/^$arg$/i, @beamlinelist)) {
        if ($arg ne $beamline) {
           print STDOUT "\n".'!!! '.$script.': Cross-beamline operation: controlling '.$arg.' from '.$beamline."\n\n";
           $beamline = $arg;
           do $PEZCA.'/Common/beamline_'.$beamline.'.pl';
	}
     }
     else {
        $snap = $ARGV[$i];
	print STDOUT $script.': snap file = '.$snap."\n";
     }
  }

  if (!defined $snap) {&press_key_exit('*** '.$script.': please specify a snap file to write!'.$syntax);}

  if (-e $snap && &ask_to_overwrite($snap) == 0) {exit;}

  print STDOUT "\n";
  print STDOUT 'This script saves LIA settings to a snap file'."\n";
  print STDOUT '     beamline = '.$beamline."\n";
  print STDOUT '     snapFile = '.$snap."\n";
# &ask_to_continue();

  $lockin_asyn_PV   = $beamline.':feedback:serialLIA';
### Different command for each beam line because different model lock-ins
  $fc = 'DDEF?';    if ($beamline eq '23i') {$fc .= '1';}

### Open SNAP file for writing:
  print STDOUT $script.': Opening snap file: '.$snap."\n\n";
  $handle = open (SNAP, '> '.$snap);
  if (!defined $handle) {&press_key_exit('*** '.$script.': Error opening '.$snap."\n");}
  select SNAP; $|=1;
  $currentTime = scalar localtime;
  $user        = $ENV{USER};
  if (!defined $user) {$user = $ENV{USERNAME};}
  if (!defined $user) {$user = 'unknown';}
  print SNAP '# LIA backup file for beamline='.$beamline."\n";
  print SNAP '# This file was created on '.$currentTime."\n";
  print SNAP '# Producer='.$script."\n";
  print SNAP '# User='.$user."\n";
  print SNAP '# '."\n";

  $portState = &getPV($lockin_asyn_PV.'.TMOD');
  if ($portState ne 'Write/Read') {&putPV($lockin_asyn_PV.'.TMOD', 'Write/Read');}

  #our ($phase, $amplitude,$frequency,$gain,$timeConst,$lineFilter,$filterRollOff,$frontOutput);

  $v = &getLIA($lockin_asyn_PV, 'PHAS?');	if (defined $v) {print SNAP '  $phase         = '.$v.';'."\n";}
  $v = &getLIA($lockin_asyn_PV, 'SLVL?');	if (defined $v) {print SNAP '  $amplitude     = '.$v.';'."\n";}
  $v = &getLIA($lockin_asyn_PV, 'FREQ?');	if (defined $v) {print SNAP '  $frequency     = '.$v.';'."\n";}
  $v = &getLIA($lockin_asyn_PV, 'SENS?');	if (defined $v) {print SNAP '  $gain          = '.$v.';'."\n";}
  $v = &getLIA($lockin_asyn_PV, 'OFLT?');	if (defined $v) {print SNAP '  $timeConstant  = '.$v.';'."\n";}
  $v = &getLIA($lockin_asyn_PV, 'OFSL?');	if (defined $v) {print SNAP '  $filterRollOff = '.$v.';'."\n";}
  $v = &getLIA($lockin_asyn_PV, 'ILIN?');	if (defined $v) {print SNAP '  $lineFilter    = '.$v.';'."\n";}
  $v = &getLIA($lockin_asyn_PV,   $fc  );	if (defined $v) {print SNAP '  $frontOutput   = '."'${v}';\n";}

  if ($portState ne 'Write/Read') {&putPV($lockin_asyn_PV.'.TMOD',$portState);}
  close(SNAP);

  &press_key_exit($script.': Finished!',0,'info');


#==========================================================================
sub putPV ($$) {

  my ($pv2change, $new_pv_value, $ca_error, $maxStrLen);

  $pv2change    = shift (@_);
  $new_pv_value = shift (@_);

  $maxStrLen = 39;		  	# maximum length of EPICS string PV

  $ca_error = Pezca::PutString ($pv2change, substr($new_pv_value,0,$maxStrLen) );
  &report_only($script, $ca_error, $pv2change);
  select(undef,undef,undef,0.1);
  return $ca_error;
}

#==========================================================================
sub getPV ($) {

  my ($pv2get, $ca_error, $pv_value);

  $pv2get  = shift(@_);

  ($ca_error,$pv_value) = Pezca::GetString($pv2get);
  &report_only($script, $ca_error, $pv2get);
  select(undef,undef,undef,0.1);
  return $pv_value;
}

#==========================================================================
sub getLIA ($$) {

  my ($cmd, $pv, $val, $ca_error);

  $pv  = shift (@_);
  $cmd = shift (@_);

  if (&putPV($pv.'.AOUT', $cmd)) {return undef;}
  else                           {return &getPV($pv.'.TINP');}
}
