#!/usr/bin/perl -w
#------------------------------------------------------------------------
# This script watches for stopped CAEN QuadEM when it reports "Acquiring"
# and then restarts it and send an email. The script is supposed to run
# in screen, optimally on the NUC computer controlling CAEN.
#
# Command for running from SCREEN:
# screen -d -m -h 5120 -S caen_restarter bash -c "source /etc/bashrc; caen_restarter.pl; read -rsp $'Press any key...\n' -n1 key"
# screen -d -m bash -c "source /etc/bashrc; caen_restarter.pl; read -rsp $'Press any key...\n' -n1 key"
# screen -d -m bash -c "source /etc/bashrc; caen_restarter.pl"
#
#                                Author: Sergey Stepanov
#
# Version-1.0 2021/11/11: initial impelentation
#--------------------------------------------------------------
  use strict;
# use Pezca;
# require MIME::Lite;			# loaded below
  use warnings::unused;

  our ($beamline, $NONINTERACTIVE, $IGNORE_ERRORS);
  my ($script, $PEZCA, $date, $status);
  my ($caen_acquire_pv, $caen_acquire);
  my ($caen_num_acq_pv, $caen_num_acq, $caen_num_prev);
  my ($sleeptime, $n_coincide, $n_coincide_max);

# &get_login_env();						#tried for running from SCREEN, did not work

  select(STDERR);  $|=1;					#flush output after each print
  select(STDOUT);  $|=1;					#flush output after each print

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;

  do $PEZCA.'/Common/common_subs.pl';
  do $PEZCA.'/Common/beamline.pl';

  $script = &makeScriptName();

  $SIG{'HUP'}  = 'signal_handler';			#(1) hangup (see signum.h)
  $SIG{'INT'}  = 'signal_handler';			#(2) interrupt (see signum.h)
  $SIG{'QUIT'} = 'signal_handler';			#(3) quit (see signum.h)
# $SIG{'KILL'} = 'signal_handler';			#(9) this is what is sent by kill -9 -- unblockable
  $SIG{'TERM'} = 'signal_handler';			#(15) terminate -- sent by kill without paramaters
# $SIG{'STOP'} = 'signal_handler';			#(19) stop -- unblockable

  if (&try_load_module('Pezca'))	{die '*** '.$script.': Cannot load MIME::Lite PERL mddule';}
  if (&try_load_module('MIME::Lite'))	{die '*** '.$script.': Cannot load Pezca PERL module';}

  $NONINTERACTIVE = 1;
  $IGNORE_ERRORS = 1;

  $caen_acquire_pv = $beamline.':AH501:Acquire';
  $caen_num_acq_pv = $beamline.':AH501:NumAcquired';

  $status = &addCaMonitor($caen_acquire_pv);
  &report_only($script,$status,$caen_acquire_pv);

  $status = &addCaMonitor($caen_num_acq_pv);
  &report_only($script,$status,$caen_num_acq_pv);

### -------------------------------------------------------------------------------------------
  $n_coincide = 0;
  $n_coincide_max = 10;
  $caen_num_acq = 0;
  print STDOUT $script.': starting monitoring CAEN QuadEM at beamline '.$beamline."\n"
                      .'Press Ctrl+C to abort...'."\n";
  while (1) {
     $sleeptime = 0.1 + rand(0.9);				#sleep from 0.1 to 1s.
#    print STDOUT $sleeptime,"\n";
     select(undef,undef,undef,$sleeptime);
     ($status,$caen_acquire) = Pezca::GetDouble($caen_acquire_pv);
     &report_only($script,$status,$caen_acquire_pv);
     if ($status || ! &is_int($caen_acquire) || $caen_acquire == 0) {next;}
     if (defined $caen_num_acq) {$caen_num_prev = $caen_num_acq;}
     else                       {$caen_num_prev = 0;}
     ($status,$caen_num_acq) = Pezca::GetDouble($caen_num_acq_pv);
     &report_only($script,$status,$caen_num_acq_pv);
     if ($status || ! &is_int($caen_num_acq)) {next;}
     if ($caen_num_acq != $caen_num_prev) {$n_coincide = 0; next;}
     else {
        $n_coincide++;
        if ($n_coincide > $n_coincide_max) {
#          $status = Pezca::PutDoubleOldCa($caen_acquire_pv,0);
#          &report_only($script,$status,$caen_acquire_pv);
#          select(undef,undef,undef,0.15);
           $status = Pezca::PutDoubleOldCa($caen_acquire_pv,1);
           &report_only($script,$status,$caen_acquire_pv);
           $date = scalar localtime;
           &sendmail('*** '.$script.': CAEN QuadEM at beamline '.$beamline.' is dead. Restarted with status='.$status.' at '.$date.'...'."\n");
           $n_coincide = 0;
        }
     }
  }
  exit 0;

#==========================================================================================
sub get_login_env {                                             #Sergey: my version 2019.02 tested with crontab
  my ($shell, $env, @lines, $line, $key, $val);
  if ($^O =~ /^Win/i) {return;}					#this function does not work on Windows (no getpwuid)
  $shell = shift || (getpwuid($<))[8];		               	#getpwuid does not exist on Windows
  $env = `$shell -lc env`;
  @lines = split(/\n/,$env);
  foreach $line (@lines) {
     if ($line =~ /=/) {
        ($key,$val) = split("=",$line,2);
        if ($key !~ /BASH_FUNC_module/ && length($key)>0) {$ENV{$key} = $val;}
     }
  }
}

#=================================================================

sub sendmail($) {
  my ($msg, $mail);
  $msg  = shift(@_);
  print STDOUT $msg."\n\n";
  $mail = MIME::Lite->new(
    From     => 'sstepanov@anl.gov',
    To       => 'sstepanov@anl.gov,makarov@anl.gov,vnagas@anl.gov',
#   Cc       => 'sstepanov@anl.gov',
    Subject  => 'CAEN QuadEM restarted on beamline '.$beamline ,
    Type     => 'TEXT',
#   Encoding => 'quoted-printable',
    Encoding => 'base64',
    Data     => $msg
  );
  if ($^O =~ /Win/i) {MIME::Lite->send('smtp', 'px1.gmca.aps.anl.gov', Timeout=>60);}
  $mail->send;
}

#=================================================================

sub signal_handler {
  my $signame = shift;
# local($signame) = @_;
  print STDOUT "\n".$script.': *** Caught a SIG'.$signame."\n";
  &clearAllCaMonitors();
  print STDOUT $script.': exiting!'."\n\n";
  exit 1;
}
