#!/usr/bin/perl -W
#--------------------------------------------------------------
# This script sets new beamline user and clears up all JBluIce data
# of the previous user.
#
#     			Author: Sergey Stepanov
#
# Version-1.0: 2009/06/04 using Pezca
# 2010/01/28 (MH): Raster runs, new collect run PVs, and SQL are cleared
# 2010/02/02 (MH): Raster directory was written with doubles, not strings
# 2010/02/05 (MH): Enable processing for raster tabs defaults to true now
# 2010/05/17 (SS): Added default scanning mode
# 2010/05/19 (MH): Changed raster defaults, added new scan PVs
# 2010/09/23 (MH): Add scanCenter & enableVectorCollect, fix spreadsheetDir256
# 2011/01/27 (MH): Added contour PVs, auto raster runs, auto raster PVs and numberOfSites PV
# 2011/02/08 (SP): cleared webice session id and SIL ID and colorcoding screened PVs
# 2011/05/24 (MH): Two new tables are dropped, 17 new PVs are reset
# 2011/09/28 (MH): Two more tables dropped, 15 new PVs reset
# 2012/02/07 (MH): 14 new PVs reset
# 2012/02/09 (MH): Default ROI setting for fl. raster is now opt MCA
# 2012/05/29 (SP): Clearing dataproc PVs and defaults for screening action PVs
# 2012/06/02 (MH): 25 new PVs cleared
# 2012/08/27 (MH): Bin mode PVs cleared
# 2013/01/16 (MH): PVs: helicalMode, request, response
# 2013/05/21 (MH): data dir changes
# 2013/10/22 (MH): clear FPE filename PVs
# 2014/01/21 (MH): added first approximation for Pilatus
# 2014/01/25 (SS): added precautions for safe restarting of FURKA (checking previous user)
# 2014/05/02 (SS): added option to clear JBluIce settings without changing user
# 2014/07/29 (MH): changed Pilatus defaults: raster delta=0, image viewer=albula
# 2014/12/18 (MH): added four new PVs
# 2015/02/11 (SS): added '--tab0only" flag, changed default to clearing all tabs and added Pezca::AutoErrorMessageOff
# 2015/02/16 (SS): added check when creating working directory as root
# 2016/03/02 (MH): added PVs for import collect mode
# 2016/06/01 (SS): added support for Eiger
# 2016/06/23 (MH): updating defaults for the Eiger
# 2017/01/17 (MH): adding PVs for repeat mode and new Eiger options
# 2017/04/06 (SS)  added Pezca::StartGroup() and Pezca::EndGroup(); made autodetectNruns optional
# 2017/04/10 (SS)  splitted Pezca::StartGroup() into multiple subgroups because it looks like IOC cannot handle
#                  ~3000 requests at once (some buffer overflow? - no error messages);
#                  added "slow" command line argument for the no-grouping behavior;
#                  changed from PutOldCa to Put because OLdCa was randomly loosing some puts
#                  (on the server side?)
# 2017/12/07 (SS)  added added check if FURKA is anabled. FURKA may be disabled when filewriter
#                  is not in use and images are saved by StreamWriter. Then no need for FURKA.
# 2018/04/09 (SS)  added reverting JBluIce staff-controlled settings stored in SQL to their defaults.
#                  For now only "raster_single_arm" is reverted, others can be added later.
# 2019/01/25 (MH): Adding 4 PVs and a new SQL table
# 2019/06/04 (MH): Redis additions
# 2020/01/15 (SS): added sync from $bluiceUser to $ccdUser for Pilatus & Eiger. Disabled making users
#                  with local home directories (e.g. gmca, root) curent users
# 2021/05/14 (MH): Call pybluice2/bin/redis_change_user.py for Redis clearing
# 2021/07/13 (SS): added &getCCDfpehost()
# 2021/09/02 (SS): bug fixes
# 2021/10/05 (MH): Apply BluIce Limits PV for raster delta
# 2021/11/03 (SS): removed fpeuser PV
# 2022/06/23 (SS): added support for mar on bl3
# 2022/09/23 (SS): added handling of backticks errors
#--------------------------------------------------------------
  use strict;
  use Pezca;
# use DBI;				#this is loaded below -- see &try_load_module
# use Date::Manip;                      #defined on Linux only; loaded below -- see &try_load_module
# use warnings::unused;

  our ($beamline, $mysqlHost, $IGNORE_ERRORS, $NONINTERACTIVE, $USE_ZENITY_FOR_EXIT);
  our ($EZCA_GROUP);
  my ($PEZCA, $script, $narg, $systemUser, $bluiceUser, $newUser, $forceChange, $i, $nRuns);
  my ($user_pw, $passwd, $uid, $gid, $quota, $comment, $gcos, $homedir, $shell);
  my (@tablesToDelete, $beamlineString, $datadir, $testdir, @prefixList, $nRasterRuns);
  my ($dsn, $dbh, $sth, $status, $yesno, $host, $bln, $syntax, @beamlinelist);
  my ($detector, $furkaUser, $sshcmd, $furkacmd, $prompt, $clearOnly, $bl, $val);
  my ($defaultExposureTime, $defaultCollectDelta, $defaultRasterDelta, $tab0only);
  my ($import_xyz);
  my ($CCD_D, $BS_D, $MO_E, $nerr, $msg, $slow, $autodetectNruns, $server, $redis);
  my ($RASTER_DELTA_MIN, $RASTER_DELTA_MAX);
# my ($count);

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;

  do ${PEZCA}.'/Common/ccd_subs.pl';
  do ${PEZCA}.'/Common/common_subs.pl';
  do ${PEZCA}.'/Common/beamline.pl';

  select STDERR; $|=1;					#set unbuffered output
  select STDOUT; $|=1;					#set unbuffered output

  $script = &makeScriptName();
# $script =~ s/\.pl$//i;

  if ($^O =~ /Win/i) {&press_key_exit('*** '.$script.': must run on Linux');}

  if (&try_load_module('DBI'))         {exit;}
  if (&try_load_module('Date::Manip')) {exit;}

  $IGNORE_ERRORS       = 1;
  $NONINTERACTIVE      = 1;
  $USE_ZENITY_FOR_EXIT = 1;
  $nRuns       = 16;					#(16) this will quit the loop when PVs end
  $nRasterRuns = 16;					#(16) this will quit the loop when PVs end
  &Pezca::AutoErrorMessageOff();			#suppress error messages

### Determine who is running this:
  if    (defined $ENV{USER})     {$systemUser = $ENV{USER};}
  elsif (defined $ENV{USERNAME}) {$systemUser = $ENV{USERNAME};}
  elsif ($^O !~ /Win/i)          {$systemUser = getpwuid($<);}
  else                           {&press_key_exit('*** '.$script.': cannot detect current user');}

### Determine who is current JBluIce user:
  $bluiceUser = &pvget('bi:runs:currentUser', 'string');

  $forceChange     = 0;					#default if no cmdline is give
  $newUser         = $systemUser;			#default if no cmdline is given
  $prompt          = 0;        				#default if no cmdline is given
  $clearOnly       = 0;        				#default if no cmdline is given
  $tab0only        = 0;					#default if no cmdline is given
  $import_xyz      = 0;					#default if no cmdline is given
  $autodetectNruns = 0;					#default if no cmdline is given
  $slow            = 0;					#0=use ezcaGroup, 1=individual ezcaPuts
  @beamlinelist = ('23i:', '23o:', '23b:', '23d:');

### Parse cmdline
  $narg = @ARGV;
  $syntax = "\n\n".'Syntax:'."\n".$script.'.pl [newJBluiceUser] [--force]'."\n"
                                 .'[--clearRunsOnly|clearOnly] [--prompt]'."\n"
				 .'[--tab0only] [--autodetectNruns]'."\n"
                                 .'[--slow|--fast] [beamline] [--?|--help]'."\n";
  for ($i=0; $i<$narg; $i++) {
     if    ($ARGV[$i] =~ /^-{0,2}\?$/i ||
            $ARGV[$i] =~ /^-{0,2}help$/i )           {&press_key_exit($syntax);}
     elsif ($ARGV[$i] =~ /^-{0,2}force$/i)           {$forceChange = 1;}
     elsif ($ARGV[$i] =~ /^-{0,2}prompt$/i)          {$prompt      = 1;}
     elsif ($ARGV[$i] =~ /^-{0,2}tab0only$/i)        {$tab0only    = 1;}
     elsif ($ARGV[$i] =~ /^-{0,2}clearOnly$/i ||
            $ARGV[$i] =~ /^-{0,2}clearRunsOnly$/i)   {$clearOnly   = 1;}
     elsif ($ARGV[$i] =~ /^-{0,2}autodetectNruns$/i) {$autodetectNruns = 1;}	#auto-detect $nRuns & $nRasterRuns (you can set them high)
     elsif ($ARGV[$i] =~ /^-{0,2}slow$/i)            {$slow        = 1;}	#do not group ezcaPuts
     elsif ($ARGV[$i] =~ /^-{0,2}fast$/i)            {$slow        = 0;}	#group ezcaPuts
     elsif (grep(/^$ARGV[$i]/, @beamlinelist)) {
        $bl = $ARGV[$i];
        $bl =~ s/:$//;
        if ($bl ne $beamline) {&press_key_exit("\n".'*** '.$script.'.pl: cannot run across beamlines!');}
     }
     else {
        if ($ARGV[$i] =~ /^-/) {&press_key_exit("\n".'*** '.$script.'.pl: incorrect command line argument "'.$ARGV[$i].'".'.$syntax);}
        else                   {$newUser     = $ARGV[$i];}
     }
  }

  if ($tab0only) {$nRuns=1; $nRasterRuns=1;}
  if ($beamline eq '23b') {$import_xyz=1;}

###-------------------------------------------------------------------------
   if    ($beamline eq '23i') {$bln = 'bl1';}
   elsif ($beamline eq '23o') {$bln = 'bl2';}
   elsif ($beamline eq '23b') {$bln = 'bl3';}
   else                       {$bln = '';}

  ($detector,$status) = &getCCDname();
  if ($status || $detector =~ /unknown/i) {
     &press_key_exit("\n".'*** '.$script.'.pl: error='.$status.', detector='.$detector.' from getCCDname!');
  }
  print STDOUT 'Beamline detector = '.$detector."\n";

  if ($detector =~ /pilatus/i || $detector =~ /eiger/i || $detector =~ /marccd/i ) {
     ($server,$status) = &getCCDfpehost();
     if ($status || $server eq 'unknown') {
        if    ($detector =~ /pilatus/i) {$server = $bln.'ppu';}
        elsif ($detector =~ /eiger/i)   {$server = $bln.'epu';}
        elsif ($detector =~ /marccd/i)  {$server = $bln.'mar';}
     }
     $server =~ s/(\.gmca\.aps\.anl\.gov)*$//;
     $server =~ s/(-40g)*$//;
     $server =~ s/(-10g)*$//;
     print STDOUT 'Detector server   = '.$server."\n";
  } else {
     &press_key_exit("\n".'*** '.$script.'.pl: cannot determine server computer for detector='.$detector);
  }

###-----------------------------------------------------------------
  if ($clearOnly == 0) {
     ($user_pw, $passwd, $uid, $gid, $quota, $comment, $gcos, $homedir, $shell) = getpwnam ($newUser);
     if    (!defined $homedir)    {&press_key_exit('*** '.$script.': requested user='.$newUser.' does not exist!');}
     elsif ($homedir !~ /^\/mnt/) {&press_key_exit('*** '.$script.': requested user='.$newUser.' has local homedir='.$homedir.' No change because JBLuIce cannot run on local dirs!');}
     elsif ($newUser eq 'root' || 
            $newUser eq 'gmca')   {&press_key_exit('*** '.$script.': requested user='.$newUser.' is not allowed to be JBluIce master!');}

     if ($newUser ne $systemUser && $systemUser ne 'root') {
        &press_key_continue('!!! '.$script.': user='.$systemUser.' trying to set JBluIce for *another* user='.$newUser);
     }

     ###############################################################################
#    if    ($detector =~ /pilatus/i || $detector =~ /eiger/i) {&restart_furka(0);} #
     if    ($detector =~ /pilatus/i) {&restart_furka(0);}                          #
     elsif ($detector =~ /marccd/i) {&restart_marccd(0);}                          #
     ###############################################################################

     if (($newUser eq $bluiceUser) && ($forceChange != 1)) {
        &press_key_exit('*** '.$script.': new user same as before -- will not cleanup unless you type:'
                       ."\n".$script.' --force '.$newUser);
     }

     &press_key_continue('!!! '.$script.': JBluIce Runs will be DELETED'."\n"
                        .'!!! '.$script.': JBluIce username will be set to "'.$newUser.'"');
  } else {
     print STDOUT $script.': called with no JBluIce user changing option.'."\n\n";
     if ($prompt) {
        $yesno = &zenity_question ($script.': Do you want to clear all JBluIce Runs without changing user?');
        if ($yesno eq '0') {exit;}
     }
  }

  ### Clear Redis databases
  print STDOUT 'Sending flushall to JBluIce Redis...'."\n";
  $redis = '/mnt/software/bluice/tools/scripts/jbluice_redis_defaults.py';
  if (! -e $redis) {&press_key_exit('*** '.$script.': cannot find '.$redis);}
  if (! -x $redis) {&press_key_exit('*** '.$script.': cannot exec '.$redis);}
  print STDOUT `$redis flushall`;
  if (${^CHILD_ERROR_NATIVE} || $?) {print STDERR '*** '.$script.' command:'."\n".$redis.' flushall'."\n".'failed with code='.${^CHILD_ERROR_NATIVE}.' ('.$?.').'."\n\n";}

  if ($beamline eq '23i' || $beamline eq '23o') {
     print STDOUT 'Sending flushall to PyBluIce Redis...'."\n";
     $redis = '/mnt/software/pybluice/bin/redis_change_user.py';
     if (! -e $redis) {&press_key_exit('*** '.$script.': cannot find '.$redis);}
     if (! -x $redis) {&press_key_exit('*** '.$script.': cannot exec '.$redis);}
     print STDOUT `PYBLUICE_SQL=$mysqlHost $redis`;
     if (${^CHILD_ERROR_NATIVE} || $?) {print STDERR '*** '.$script.' command:'."\n".'PYBLUICE_SQL='.$mysqlHost.' '.$redis."\n".'failed with code='.${^CHILD_ERROR_NATIVE}.' ('.$?.').'."\n\n";}
  }

  ### clear SQL before setting PVs, so when the raster run is updated there is no data
  print STDOUT 'Dropping SQL tables...'."\n";
  $dsn = 'DBI:mysql:host='.$mysqlHost.';database=user_data';
  $dbh = DBI->connect($dsn,'dhs','',{PrintError=>0,});
  if (defined $dbh) {
     # MH: two more tables added 9/28/11
     @tablesToDelete = ();
     push (@tablesToDelete,('RasterResults', 'RasterSites', 'RasterHits'));
     push (@tablesToDelete,('AutoindexResults', 'DataProcResults', 'HKLDatasets', 'HKLDatasetsList'));
     push (@tablesToDelete,('StrategyResults', 'MultiStrategyResults'));
     push (@tablesToDelete,('MultiStrategyReferenceData', 'MultiStrategyMergeDatasets', 'MultiStrategyCombineDatasets'));
     push (@tablesToDelete,('Macros', 'SpotFinderResults3', 'XDSSettings'));
     if ($import_xyz) {push (@tablesToDelete,('ImportSites'));}
     foreach (@tablesToDelete) {
        $sth = $dbh->prepare('DROP TABLE IF EXISTS '.$_);
        $status = $sth->execute();
        if (!defined $status) {
           print STDOUT '*** '.$script.': Failed to remove table '.$_.' from SQL.  host='.$mysqlHost."\n";
        }
        $sth->finish();
     }
     $dbh->disconnect();
  }

  print STDOUT 'Reverting staff settings in SQL their defaults...'."\n";
  &write_jbluice_config('raster_single_arm',1);

  ### Generate beamline string
  $beamlineString = '23';
  if ($beamline eq '23i') {$beamlineString .= 'IDD';}
  if ($beamline eq '23o') {$beamlineString .= 'IDB';}
  if ($beamline eq '23b') {$beamlineString .= 'BM';}

  if ($clearOnly == 0) {
     ### check for directories from the last 3 days
     $datadir = '';
     for ($i=0; $i<=3; $i++) {
        ### get dir name for $i days ago
        $testdir = &generateDataDir($i);
        ### if it exists, exit
        if (-d $testdir) {$datadir=$testdir; last;}
        if (-d "${testdir}_hdf") {$datadir=$testdir; last;}
        if (-d "${testdir}_cbf") {$datadir=$testdir; last;}
     }

  ### if no previous directories were found, set $datadir to today's date and create
     if ($datadir eq '') {
        if ($newUser eq $systemUser || $systemUser eq 'root') {
           ### set dir name for today's date
           $datadir = &generateDataDir(0);
           ### create the dir
           #$count = mkdir($datadir);
           #if ($count == 1) {
           #   if ($^O !~ /Win/i) {				 	#if UNIX
           #      $count = chmod(0755,$datadir);
           #	 if ($systemUser eq 'root') {$count = chown($uid,$gid,$datadir);}
           #   }
           #} else {
           #   print STDERR '!!! Failed to create data dir ['.$datadir.'].'."\n"
           #               .'!!! Setting JBluIce to ['.$homedir.'] instead.'."\n";
           #   $datadir = $homedir;
           #}
        } else {
           print STDERR '!!! Cannot create data dir ['.$datadir.']'."\n"
                       .'!!! because system user ['.$systemUser.'] cannot'."\n"
                       .'!!! write to new user ['.$newUser.'] directory.'."\n"
                       .'!!! Setting JBluIce to ['.$homedir.'] instead.'."\n";
           $datadir = $homedir;
        }
     }
     print STDOUT 'Using common dir: '.$datadir."\n";

     ### now write the final directory to the PV
     &pvwrite('bi:data_dir', $datadir,    'wfstring');
  }

  ### set defaults based on detector type
  if ($detector =~ /pilatus/i) {
      $defaultExposureTime = 0.2;
      $defaultCollectDelta = 0.2;
      $defaultRasterDelta = 0.0;
  } elsif ($detector =~ /eiger/i) {
      $defaultExposureTime = 0.2;
      $defaultCollectDelta = 0.2;
      $defaultRasterDelta = 0.0;
  } else {
      $defaultExposureTime = 1.0;
      $defaultCollectDelta = 1.0;
      $defaultRasterDelta = 1.0;
  }
  ### MH 10/5/2021: Apply raster delta limits
  $RASTER_DELTA_MIN = &pvget('bi:r_r15:delta_deg.DRVL');
  if ($RASTER_DELTA_MIN >= 0.001 && $defaultRasterDelta < $RASTER_DELTA_MIN) {
      $defaultRasterDelta = $RASTER_DELTA_MIN;
  }
  $RASTER_DELTA_MAX = &pvget('bi:r_r15:delta_deg.DRVH');
  if ($RASTER_DELTA_MAX >= 0.001 && $defaultRasterDelta > $RASTER_DELTA_MAX) {
      $defaultRasterDelta = $RASTER_DELTA_MAX;
  }

  if ($autodetectNruns) {
     ### Detect how many runs we have:
     for ($i=0; $i<$nRuns; $i++) {
        ($status,$val) = Pezca::GetDouble($beamline.':bi:r'.$i.':runLabel');
        if    ($status == 5 || $status == 6) {$nRuns=$i; last;}
        elsif ($status != 0) {print STDOUT "\n"; &report_only($script, $status,$beamline.':bi:r'.$i.':runLabel');}
     }
     ### Detect how many raster runs we have:
     for ($i=0; $i<$nRasterRuns; $i++) {
        ($status,$val) = Pezca::GetDouble($beamline.':bi:ar_r'.$i.':runLabel');
        if    ($status == 5 || $status == 6) {$nRasterRuns=$i; last;}
        elsif ($status != 0) {print STDOUT "\n"; &report_only($script, $status,$beamline.':bi:ar_r'.$i.':runLabel');}
     }
  }

### These values are needed by pvwrite below, but we cannot
### read them once the group is started:
  $CCD_D = &pvget('CCD:D:ActPos');
  $BS_D  = &pvget('BS:D:ActPos');
  $MO_E  = &pvget('MO:E:ActPos');

# - - - - - - - - - - - - - - - - - - - - -
  &pvwrite('bi:runs:currentUser', 		$newUser,  	'string');
  $bluiceUser = $newUser;

# - - - - - - - - - - - - - - - - - - - - -
  $nerr = 0;
  $msg = 'Clearing PVs failed:'."\n";
  $EZCA_GROUP = 0;
  if (! $slow) {Pezca::StartGroup(); $EZCA_GROUP=1;}

  print STDOUT 'Clearing individual PVs...'."\n";
  ### MH: added 2017/01/17
  &pvwrite('bi:runs:importDBUpdated',	        0			);
  &pvwrite('bi:runtimeConfig',		        0		        );
  &pvwrite('bi:imageFormat',			3			);
  &pvwrite('bi:hdfCleanupMode',			0			);

  &pvwrite('bi:collectSpotFinder',		0			);
  &pvwrite('bi:rasterSpotFinder',		0			);

  ### MH: added 2018/05/25
  &pvwrite('bi:sfCollectOptions',		"\0",	'wfstring',4000 );
  &pvwrite('bi:sfRasterOptions',		"\0",	'wfstring',4000 );
  &pvwrite('bi:imageServer',			1			);

  ### MH: added 2014/01/21
  if ($detector =~ /pilatus/i) {
    &pvwrite('bi:runs:enableShutterless',	1			);
    &pvwrite('bi:displayProgram',		3			);
  } elsif ($detector =~ /eiger/i) {
    &pvwrite('bi:runs:enableShutterless',	1			);
    &pvwrite('bi:displayProgram',		1			);
  } else {
    &pvwrite('bi:runs:enableShutterless',	0			);
    &pvwrite('bi:displayProgram',		0			);
  }

  ### MH: added 2013/10/22
  &pvwrite('FPE:filename',		        '',		'string');
  &pvwrite('FPE:lastfilename',		        '',		'string');
  &pvwrite('FPE:directory256',		        "\0",	  'wfstring',256);
  &pvwrite('FPE:lastfiledir256',		"\0",	  'wfstring',256);
  &pvwrite('FPE:viewDir256',			"\0",	  'wfstring',256);
  &pvwrite('FPE:viewFile256',		        "\0",	  'wfstring',256);

  ### MH: added 2011/05/24
  &pvwrite('bi:runs:currentLeftTab',	        0			);
  &pvwrite('bi:scr:ScreeningStatus',	        0			);
  &pvwrite('bi:scr:CurrentTask', 		'',		'string');

  &pvwrite('bi:enableVectorCollect',	        1			);
  &pvwrite('bi:runs:runCount',		        0			);
  &pvwrite('bi:runs:currentRun',	        0			);
  &pvwrite('bi:runs:collectState',	        0			);

  ### MH: added 2012/08/27
  &pvwrite('bi:runs:binMode',	                1			);

  &pvwrite('bi:runs:updated',		        0			);
  &pvwrite('bi:runs:updated',		        1			);
  &pvwrite('bi:runs:updated.PROC',	        1			);

  ### MH: added 2011/01/27
  &pvwrite('bi:r_runs:macroUpdated',	        0			);
  &pvwrite('bi:r_runs:autoMode',	        0			);
  &pvwrite('bi:r_runs:autoStep',	        0			);
  &pvwrite('bi:r_runs:autoStopAfterRun',	0			);
  &pvwrite('bi:r_runs:autoContourMethod',	0			);
  &pvwrite('bi:r_runs:autoSpotLimit',	        0			);
  &pvwrite('bi:r_runs:autoRunCount',	        0			);
  &pvwrite('bi:r_runs:autoCurrentRun',	        0			);
  &pvwrite('bi:r_runs:autoUpdated',	        0			);
  &pvwrite('bi:r_runs:autoUpdated',	        1			);

  ### MH: added 2011/05/24
  &pvwrite('bi:r_runs:siteListUpdated',		-1			);
  &pvwrite('bi:r_runs:siteListUpdated',	        -2			);
  &pvwrite('bi:r_runs:runCount',	        0			);
  &pvwrite('bi:r_runs:currentRun',	        0			);
  &pvwrite('bi:r_runs:rasterState',	        0			);

  ### MH: added 2014/12/18
  &pvwrite('bi:r_runs:viewMode',	        0			);
  &pvwrite('bi:r_runs:selectedCell',	        -1			);
  &pvwrite('bi:r_runs:fixResLimits',	        0			);

  ### MH: added 2012/08/27
  &pvwrite('bi:r_runs:binMode',	                2			);

  &pvwrite('bi:r_runs:updated',		        0			);
  &pvwrite('bi:r_runs:updated',		        1			);
  &pvwrite('bi:r_runs:updated.PROC',	        1			);

  ### MH: added 2011/09/28
  &pvwrite('bi:beamsize0',	                0			);
  &pvwrite('bi:beamsize1',	                0			);
  &pvwrite('bi:beamsize2',	                0			);
  &pvwrite('bi:beamsize3',	                0			);
  &pvwrite('bi:beamsize4',	                0			);
  &pvwrite('bi:beamsize5',	                0			);
  &pvwrite('bi:r_runs:currentMacro',            '',		'string');

  ### MH: added 2013/01/16
  &pvwrite('bi:request',                        0			);
  &pvwrite('bi:response',                       0			);

  if ($import_xyz) {
     &pvwrite('bi:runs:importDBUpdated',        0			);
     &pvwrite('bi:runtimeConfig',               0			);
  }

  ### added 2009/05/30:
  &pvwrite('bi:SCAN:currentTab',                0			);
  &pvwrite('bi:SCAN:autoStep',                  0			);
  &pvwrite('bi:SCAN:autoOptimized',             0			);
  &pvwrite('bi:SCAN:time',                      1			);
  &pvwrite('bi:SCAN:spectrumTime',              4			);
  ### added 2010/05/17 (0=fixed, 1=Adaptive, 2=User Input):
  &pvwrite('bi:SCAN:edgeMode',                  2			);
  ### added 2010/05/19:
  &pvwrite('bi:SCAN:edge',                      'Se-K',		'string');
  &pvwrite('bi:SCAN:edgeEnergy',                12658.0			);
  &pvwrite('bi:SCAN:scanCenter',                12658.0			);
  &pvwrite('bi:SCAN:roiLow',                    'NaN'			);
  &pvwrite('bi:SCAN:roiLow',                    'NaN'			);
  &pvwrite('bi:SCAN:roiHigh',                   'NaN'			);
  &pvwrite('bi:SCAN:choochInflection',		'NaN'			);
  ### Fp and Fpp PVs added 3/30/2015
  &pvwrite('bi:SCAN:choochInflFp',		'NaN'			);
  &pvwrite('bi:SCAN:choochInflFpp',     	'NaN'			);
  &pvwrite('bi:SCAN:choochPeak',		'NaN'			);
  &pvwrite('bi:SCAN:choochPeakFp',		'NaN'			);
  &pvwrite('bi:SCAN:choochPeakFpp',		'NaN'			);
  &pvwrite('bi:SCAN:choochRemote',		'NaN'			);
  &pvwrite('bi:SCAN:interStep',                 0			);
  &pvwrite('bi:SCAN:rightTab',                  0			);
  &pvwrite('bi:SCAN:prefix',                    'scan',		'string');
  &pvwrite('bi:SCAN:directory',                 "\0",	  'wfstring',256);
  &pvwrite('bi:SCAN:graphTitle',                0			);
  &pvwrite('bi:SCAN:graphMessage',              0			);
  &pvwrite('bi:SCAN:graphXMin',                 0			);
  &pvwrite('bi:SCAN:graphXMax',                 1			);
  &pvwrite('bi:SCAN:graphYMin',                 0			);
  &pvwrite('bi:SCAN:graphYMax',                 1			);
  &pvwrite('bi:SCAN:graphXLabel',               '',		'string');
  &pvwrite('bi:SCAN:graphYLabel',               '',		'string');
  &pvwrite('bi:SCAN:graphTraceX',               0			);
  &pvwrite('bi:SCAN:graphTraceY',               0			);
  &pvwrite('bi:SCAN:graphTraceCount',           0			);
  &pvwrite('bi:SCAN:choochPDF',                 0			);
  ### added 2012/02/07:
  &pvwrite('bi:SCAN:leftAxisTrace',             0			);
  &pvwrite('bi:SCAN:rightAxisTrace',            0			);
  &pvwrite('bi:SCAN:adaptiveScan',              0			);
  &pvwrite('bi:SCAN:plotMode',                  0			);

  &pvwrite('bi:SCAN:update',                    0			);
  &pvwrite('bi:SCAN:update',                    1			);

  for ($i=1; $i<=3; $i++) {
     ### added 2012/02/07:
     &pvwrite('bi:t'.$i.':graphTraceX',		0			);
     &pvwrite('bi:t'.$i.':graphTraceY',		0			);
     &pvwrite('bi:t'.$i.':graphYMin',		0			);
     &pvwrite('bi:t'.$i.':graphYMax',		0			);
     &pvwrite('bi:t'.$i.':graphYLabel',		'',			'string');
     &pvwrite('bi:t'.$i.':graphTraceCount',	0			);
     &pvwrite('bi:t'.$i.':isVisible',		0			);

     ### added 2012/06/02
     &pvwrite('bi:t'.$i.':defaultYMin',		0			);
     &pvwrite('bi:t'.$i.':defaultYMax',		0			);
  }

  ### Mark 2014/02/12:
  ### Sudhir 2014/10/05
  #&pvwrite('bi:s_r0:delta',			$defaultRasterDelta	);
  &pvwrite('bi:s_r0:delta',			$defaultCollectDelta	);
  &pvwrite('bi:s_r0:exposureTime',		$defaultExposureTime	);

  ### Sudhir
  &pvwrite('bi:s_r0:directory256',		"\0",	  'wfstring',256); # Sudhir Jun 4th 2013
  &pvwrite('bi:scr:spreadsheetFile', 		"\0",  	  'wfstring',128);
  &pvwrite('bi:scr:spreadsheetDir256',  	"\0",  	  'wfstring',256);
  &pvwrite('bi:scr:spreadsheetStatus',		0			);
  &pvwrite('bi:scr:ScreeningStatus',   		0			);
  &pvwrite('bi:scr:CurrentTask', 		'none',		'string');
  &pvwrite('bi:scr:directory256',		"\0",	  'wfstring',256);
  &pvwrite('bi:scr:sessionID',			'',		'string');
  &pvwrite('bi:scr:SILID',			0			);
# &pvwrite('bi:scr:spreadsheetScreened',	"\0",	  'wfstring,256');

  # MH: 2021/03/01
  &pvwrite('bi:s_r0:beamWidth_um',		100);
  &pvwrite('bi:s_r0:beamHeight_um',		100);
  &pvwrite('bi:s_r0:updated',		        0);
  &pvwrite('bi:s_r0:updated',		        1);

  ### MH: added 2011/05/24:
  &pvwrite('bi:scr:currentStrategyIndex',	0			);
  &pvwrite('bi:scr:dbUpdated',			-1			);
  &pvwrite('bi:scr:dbUpdated',			-2			);

  ### added 2012/05/29:
  ### Sudhir: added 2014/02/10
  if ($detector =~ /pilatus/i) {
        &pvwrite('bi:scr:strategyProgram',	1			);
        &pvwrite('bi:runs:bgProc',		2			);
   } elsif ($detector =~ /eiger/i) {
        &pvwrite('bi:scr:strategyProgram',	1			);
        &pvwrite('bi:runs:bgProc',		2			);
   } else {
        &pvwrite('bi:scr:strategyProgram',	0			);
        &pvwrite('bi:runs:bgProc',		2			);
  }
  &pvwrite('bi:analysis:currentDataProcIndex',	0			);
  &pvwrite('bi:analysis:dataprocUpdated',	0			);
  &pvwrite('bi:r_runs:scrClear',		1			);
  &pvwrite('bi:scr:MountNextXtal',		1			);
  &pvwrite('bi:scr:CenterXtal',			1			);
  &pvwrite('bi:scr:Pause4ctrng',		1			);
  &pvwrite('bi:scr:TakeImage0',			1			);
  &pvwrite('bi:scr:TakeImage45',		0			);
  &pvwrite('bi:scr:TakeImage90',		1			);
  &pvwrite('bi:scr:Pause4collect',		1			);

  ### added 2012/06/02:
  &pvwrite('bi:autoFind:coarseContour',		0			);
  &pvwrite('bi:autoFind:loopZ', 		0			);
  &pvwrite('bi:autoFind:faceOnOmega', 		0			);
  &pvwrite('bi:autoFind:complete', 		0			);
  &pvwrite('bi:alignDataH', 			0			);
  &pvwrite('bi:alignDataV', 			0			);
  &pvwrite('bi:alignState', 			0			);
  &pvwrite('bi:SCAN:displayedElements', 	0			);
  &pvwrite('bi:SCAN:otherElement', 		'',		'string');
  &pvwrite('bi:SCAN:otherElement2', 		'',		'string');
  &pvwrite('bi:SCAN:spectrumEnergy', 		12658.0			);
  &pvwrite('bi:SCAN:defaultInflection', 	0			);
  &pvwrite('bi:SCAN:defaultPeak', 		0			);
  &pvwrite('bi:SCAN:defaultRemote', 		0			);
  &pvwrite('bi:SCAN:defaultXMin', 		0			);
  &pvwrite('bi:SCAN:defaultXMax', 		0			);
  &pvwrite('bi:SCAN:showAllEdges', 		0			);

  ### added 2012/09/20:
  &pvwrite('bi:scr:centerType',	        	0			);

  ### MH: added 2014/12/18
  &pvwrite('bi:runs:hideCollectSF',        	0			);
  &pvwrite('bi:runs:hideSampleSF',	       	0			);

  ### Sudhir added 2013/01/28
  &pvwrite('bi:scr:diffimages0',		1			);
  &pvwrite('bi:scr:diffimages45',		1			);
  &pvwrite('bi:scr:diffimages90',		1			);
  &pvwrite('bi:scr:takeJpeg',			1			);
  &pvwrite('bi:scr:scalingSoftware',		0			);

  &pvwrite('bi:strategy:currentMultiStrategyIndex', 0			);
  &pvwrite('bi:strategy:multistrategyUpdated',	0			);
  &pvwrite('bi:strategy:referenceDataUpdated',	0			);
  &pvwrite('bi:strategy:mergeDatasetsUpdated',	0			);

  ### Sudhir added 2013/05/28
  &pvwrite('bi:scr:AngleImage0',		0.0			);
  &pvwrite('bi:scr:AngleImage45',		45.0			);
  &pvwrite('bi:scr:AngleImage90',		90.0			);

  ### Sudhir added 2014/10/02
  &pvwrite('bi:scr:prefix', 			'test',		'string');

  &commit_group(\$nerr,\$msg);

  print STDOUT 'Clearing collect PVs...'."\n".'   Tab(s):';
  for ($i=0; $i<$nRuns; $i++) {
     print STDOUT ' '.$i;
     &pvwrite('bi:r'.$i.':runStatus',		'inactive',		'string');
     &pvwrite('bi:r'.$i.':runLabel',		$i				);
     &pvwrite('bi:r'.$i.':fileroot',		'test',			'string');
     &pvwrite('bi:r'.$i.':directory256',        "\0",             'wfstring',256);
     &pvwrite('bi:r'.$i.':distance', 		$CCD_D				);
     &pvwrite('bi:r'.$i.':beamstop_mm', 	$BS_D				);
     &pvwrite('bi:r'.$i.':atten_factor',	1000.0				);
     &pvwrite('bi:r'.$i.':axisMotor',		'gonio_phi', 		'string');
     &pvwrite('bi:r'.$i.':delta',		$defaultCollectDelta		);
     &pvwrite('bi:r'.$i.':exposureTime', 	$defaultExposureTime		);
     &pvwrite('bi:r'.$i.':startAngle',    	0.0				);
     &pvwrite('bi:r'.$i.':endAngle',      	1.0				);
     &pvwrite('bi:r'.$i.':startFrameLabel',	1				);
     &pvwrite('bi:r'.$i.':start_x_mm',		0				);
     &pvwrite('bi:r'.$i.':start_y_mm',		0				);
     &pvwrite('bi:r'.$i.':start_z_mm',		0				);
     &pvwrite('bi:r'.$i.':start_set',		0				);
     &pvwrite('bi:r'.$i.':end_x_mm',		0				);
     &pvwrite('bi:r'.$i.':end_y_mm',		0				);
     &pvwrite('bi:r'.$i.':end_z_mm',		0				);
     &pvwrite('bi:r'.$i.':end_set',		0				);
     if ($import_xyz) {
        &pvwrite('bi:r'.$i.':pt1_x_mm',		0				);
        &pvwrite('bi:r'.$i.':pt1_y_mm',		0				);
        &pvwrite('bi:r'.$i.':pt1_z_mm',		0				);
        &pvwrite('bi:r'.$i.':pt1_set',		0				);
        &pvwrite('bi:r'.$i.':pt2_x_mm',		0				);
        &pvwrite('bi:r'.$i.':pt2_y_mm',		0				);
        &pvwrite('bi:r'.$i.':pt2_z_mm',		0				);
        &pvwrite('bi:r'.$i.':pt2_set',		0				);
        &pvwrite('bi:r'.$i.':pt3_x_mm',		0				);
        &pvwrite('bi:r'.$i.':pt3_y_mm',		0				);
        &pvwrite('bi:r'.$i.':pt3_z_mm',		0				);
        &pvwrite('bi:r'.$i.':pt3_set',		0				);
        &pvwrite('bi:r'.$i.':pt4_x_mm',		0				);
        &pvwrite('bi:r'.$i.':pt4_y_mm',		0				);
        &pvwrite('bi:r'.$i.':pt4_z_mm',		0				);
        &pvwrite('bi:r'.$i.':pt4_set',		0				);
     }
     &pvwrite('bi:r'.$i.':nextFrame',		0				);
     &pvwrite('bi:r'.$i.':reverseDir',		0				);
     &pvwrite('bi:r'.$i.':stepSize_um',		100.0				);
     &pvwrite('bi:r'.$i.':framesPerSite',	1				);
     &pvwrite('bi:r'.$i.':numberOfSites',       1				); # MH: added 2011/01/27
     &pvwrite('bi:r'.$i.':overlap',		0				);
     &pvwrite('bi:r'.$i.':inverseOn',       	0				);
     &pvwrite('bi:r'.$i.':wedgeSize',   	30.0				);
     &pvwrite('bi:r'.$i.':numEnergy',       	1				);
     &pvwrite('bi:r'.$i.':energy1',  		$MO_E				);
     &pvwrite('bi:r'.$i.':energy2',       	0.0				);
     &pvwrite('bi:r'.$i.':energy3',       	0.0				);
     &pvwrite('bi:r'.$i.':energy4',       	0.0				);
     &pvwrite('bi:r'.$i.':energy5',       	0.0				);

     ### MH: added 2011/05/24
     &pvwrite('bi:r'.$i.':strategyEnabled',    	0				);
     &pvwrite('bi:r'.$i.':startAngle2',    	90.0				);
     &pvwrite('bi:r'.$i.':scanMode',    	0				);
     &pvwrite('bi:r'.$i.':optionTab',    	0				);

     ### MH: added 2011/09/28
     &pvwrite('bi:r'.$i.':selectedSites',    	0				);
     &pvwrite('bi:r'.$i.':xdsProc',     	0				);
     &pvwrite('bi:r'.$i.':siteDeltaEnable',    	0				);
     &pvwrite('bi:r'.$i.':siteDelta_deg',    	0				);
     &pvwrite('bi:r'.$i.':savedFPS',    	0				);
     &pvwrite('bi:r'.$i.':savedOverlap',    	0				);
     &pvwrite('bi:r'.$i.':savedEndAngle',    	0				);

     ### MH: added 2012/02/07
     &pvwrite('bi:r'.$i.':vectorFillMode',      0				);
     &pvwrite('bi:r'.$i.':setupStep',    	0				);
     &pvwrite('bi:r'.$i.':potVecSiteCount',    	0				);

     ### MH: added 2012/06/02
     &pvwrite('bi:r'.$i.':beamWidth_um',    	20.0				);
     &pvwrite('bi:r'.$i.':beamHeight_um',    	20.0				);

     ### MH: added 2013/01/16
     &pvwrite('bi:r'.$i.':helicalMode',         0				);

     ### MH: added 2017/01/17
     &pvwrite('bi:r'.$i.':importSiteCount',     0				);
     &pvwrite('bi:r'.$i.':importPath256',       "\0",		  'wfstring',256);

     if ($import_xyz) {
        &pvwrite('bi:r'.$i.':importSiteCount',  0				);
        &pvwrite('bi:r'.$i.':importPath256',    "\0",             'wfstring',256);
     }
     
     # MH: added 2019/01/25
     &pvwrite('bi:r'.$i.':meshFilter',          1				);
     &pvwrite('bi:r'.$i.':meshCombine',         0				);
     &pvwrite('bi:r'.$i.':meshParam',		3				);
     &pvwrite('bi:r'.$i.':framesPerDir',        50000				);

     # MH: 2019/05/21
     &pvwrite('bi:r'.$i.':rasterRun',           0				);
     &pvwrite('bi:r'.$i.':rasterNumbering',     0				);
     &pvwrite('bi:r'.$i.':meshSpotFinder',      0				);

     &pvwrite('bi:r'.$i.':updated',    		0				);
     &pvwrite('bi:r'.$i.':updated',	    	1				);
     &pvwrite('bi:r'.$i.':updated.PROC',    	1				);

     &commit_group(\$nerr,\$msg);
  }
  print STDOUT "\n";

  print STDOUT 'Clearing raster PVs...'."\n".'   Tab(s):';
  for ($i=0; $i<$nRasterRuns; $i++) {
     print STDOUT ' '.$i;
     ### MH: added 2011/01/27
     @prefixList = ('ar_r','r_r');
     foreach (@prefixList) {
        &pvwrite('bi:'.$_.$i.':runStatus',		0			);
        &pvwrite('bi:'.$_.$i.':runType',		0			);
        &pvwrite('bi:'.$_.$i.':nextFrame',		0			);
        &pvwrite('bi:'.$_.$i.':runLabel',		$i			);
        &pvwrite('bi:'.$_.$i.':fileroot',		'raster',	'string');
        &pvwrite('bi:'.$_.$i.':directory256',	        "\0",	  'wfstring',256);
        &pvwrite('bi:'.$_.$i.':startFrameLabel',	1			);
        &pvwrite('bi:'.$_.$i.':gridWidth_um',	        350			);
        &pvwrite('bi:'.$_.$i.':gridHeight_um',		150			);
        &pvwrite('bi:'.$_.$i.':cellWidth_um',	        20			);
        &pvwrite('bi:'.$_.$i.':cellHeight_um',		20			);
        &pvwrite('bi:'.$_.$i.':enableProcessing',	1			);
        &pvwrite('bi:'.$_.$i.':roiMode',		1			); # MH: 2012/02/09
        &pvwrite('bi:'.$_.$i.':isCenterSet',	        0			);
        &pvwrite('bi:'.$_.$i.':center_omega_deg',	0			);
        &pvwrite('bi:'.$_.$i.':center_x_mm',	        0			);
        &pvwrite('bi:'.$_.$i.':center_y_mm',	        0			);
        &pvwrite('bi:'.$_.$i.':center_z_mm',	        0			);
        &pvwrite('bi:'.$_.$i.':delta_deg',		$defaultRasterDelta	);
        &pvwrite('bi:'.$_.$i.':exposure_sec',	        $defaultExposureTime	);
        &pvwrite('bi:'.$_.$i.':distance_mm',	        $CCD_D			);
        &pvwrite('bi:'.$_.$i.':energy_keV',		$MO_E			);
        &pvwrite('bi:'.$_.$i.':atten_factor',	        50.0			);
        ### MH: added 2011/01/27
        &pvwrite('bi:'.$_.$i.':contours',		0			);
        &pvwrite('bi:'.$_.$i.':selectedCells',		0			);
        &pvwrite('bi:'.$_.$i.':beamWidth_um',	        20			);
        &pvwrite('bi:'.$_.$i.':beamHeight_um',		20			);
        &pvwrite('bi:'.$_.$i.':gridTopOffset_um',	0			);
        &pvwrite('bi:'.$_.$i.':gridBottomOffset_um',	0			);
        &pvwrite('bi:'.$_.$i.':gridLeftOffset_um',	0			);
        &pvwrite('bi:'.$_.$i.':gridRightOffset_um',	0			);

        ### MH: added 2011/05/24
        &pvwrite('bi:'.$_.$i.':loopMeasurement_um',	200.0			);
        &pvwrite('bi:'.$_.$i.':userOffsetX',		0			);
        &pvwrite('bi:'.$_.$i.':userOffsetY',		0			);
        &pvwrite('bi:'.$_.$i.':userOffsetZ',		0			);
        &pvwrite('bi:'.$_.$i.':userOffsetOmega',	0			);
        &pvwrite('bi:'.$_.$i.':userOffsetSet',		0			);
        &pvwrite('bi:'.$_.$i.':referenceRun',		0			);
        &pvwrite('bi:'.$_.$i.':referenceHit',		0			);
        &pvwrite('bi:'.$_.$i.':refinementType',		0			);

        ### MH: added 2012/06/02
        &pvwrite('bi:'.$_.$i.':autoFind',	        1			);
        &pvwrite('bi:'.$_.$i.':lowResLimit_A',		50.0			);
        &pvwrite('bi:'.$_.$i.':highResLimit_A',		1.0			);
        &pvwrite('bi:'.$_.$i.':loopCentered',	        0			);

        # MH: added 2019/01/25
        &pvwrite('bi:'.$_.$i.':framesPerCell',	        1			);

        # MH: 2019/05/21
        &pvwrite('bi:'.$_.$i.':soniccBin',	        0			);
        &pvwrite('bi:'.$_.$i.':soniccRefPos',	        0			);

        &pvwrite('bi:'.$_.$i.':updated',		0			);
        &pvwrite('bi:'.$_.$i.':updated',		1			);
        &pvwrite('bi:'.$_.$i.':updated.PROC',	        1			);
     }
     &commit_group(\$nerr,\$msg);
  }
  &pvwrite('ioc3:saveTrigger.PROC', 1);					#trigger Keithley IOC autosave for saving new settings
  
  print STDOUT "\n";

# - - - - - - - - - - - - - - - - - - - - -
  if ($nerr > 0) {&zenity_error_exit($msg);}
  print STDOUT '... All done!'."\n";
  &clearAllCaMonitors();
  exit;

###########################################################################
sub pvwrite ($$;$$) {

  our $beamline;
  my ($pv, $val, $type, $nelm, $status);
  $pv   = shift(@_);
  $val  = shift(@_);
  $type = shift(@_);	if(!defined $type) {$type='double';}
  $nelm = shift(@_);
  if ($pv !~ /^${beamline}/) {$pv = $beamline.':'.$pv;}
# if    ($type =~ /^double$/i)   {$status = Pezca::PutDoubleOldCa($pv, $val);}
# elsif ($type =~ /^string$/i)   {$status = Pezca::PutStringOldCa($pv, $val);}
# elsif ($type =~ /^wfstring$/i) {$status = &PezcaPutStringWF($pv, $val, 'oldca', $nelm, 'group');}
  if    ($type =~ /^double$/i)   {$status = Pezca::PutDouble($pv, $val);}
  elsif ($type =~ /^string$/i)   {$status = Pezca::PutString($pv, $val);}
  elsif ($type =~ /^wfstring$/i) {$status = &PezcaPutStringWF($pv, $val, undef, $nelm, 'group');}
  else                           {&press_key_exit('*** '.$script.': incorrect PV type='.$type);}
  if (!defined $EZCA_GROUP) {$EZCA_GROUP=0;}
  if ($EZCA_GROUP == 0 && $status) {
     &report_only($script, $status, $pv);
  }
}

###########################################################################
sub pvget ($;$) {

  our $beamline;
  my ($pv, $val, $type, $status);
  $pv   = shift(@_);
  $type = shift(@_);	if(!defined $type) {$type='double';}
  if ($pv !~ /^${beamline}/) {$pv = $beamline.':'.$pv;}
  if    ($type =~ /^double$/i)   {($status,$val) = Pezca::GetDouble($pv);}
  elsif ($type =~ /^string$/i)   {($status,$val) = Pezca::GetString($pv);}
  elsif ($type =~ /^wfstring$/i) {($status,$val) = &PezcaGetStringWF($pv);}
  else                           {&press_key_exit('*** '.$script.': incorrect PV type='.$type);}
  if ($status) {&report_only($script, $status, $pv); $val=0;}
  return $val;
}

###########################################################################
sub generateDataDir {

  my ($daysAgo, $dateString, $result);

  $daysAgo = $_[0];

  # generate date string
  $dateString = &UnixDate($daysAgo.' days ago','%Y_%m_%d');

  # generate data dir
  $result = $homedir.'/'.$beamlineString.'_'.$dateString;
}

###########################################################################
sub restart_furka($) {

  my ($partial_dir, $prefix, $cmd, $msg, $question);
  $question = shift(@_);
  ### Run update_furka to change the furka uid/gid, create the link in /dataInt, and restart the furka service
  $cmd = 'perl -W '.${PEZCA}.'/furka_user.pl';
  print STDOUT $script.': querying Furka user with ['.$cmd.']...'."\n";
  $furkaUser = `$cmd`;
  if (${^CHILD_ERROR_NATIVE} || $?) {
     $msg = '*** '.$script.' command:'."\n".$cmd."\n".'failed with code='.${^CHILD_ERROR_NATIVE}.' ('.$?.').';
     &zenity_info($msg);
     $furkaUser = 'unknown';
  }
  elsif ($furkaUser !~ /^furka_user=/) {
     $msg = '!!! Failed to detect current FURKA user. Is FURKA running?';
#    print STDOUT $msg."\n";
     &zenity_info($msg);
     $furkaUser = 'unknown';
  } else {
     $furkaUser =~ s/^furka_user=//;
     $furkaUser =~ s/\n$//;  chomp $furkaUser;
  }
  if ($furkaUser ne $newUser && $furkaUser ne 'furka_is_disabled' && $furkaUser ne 'streamwriter') {
     if ($furkaUser eq 'unknown') {$msg='FURKA is either not running or running as unknown user'."\n";}
     else			  {$msg='FURKA is running as user "'.$furkaUser.'"'."\n";}
     if ($question) {
         $msg .= 'Do you want to restart FURKA as user "'.$newUser.'"?'."\n\n"
              .'ATTENTION: use with extreme care since all '.$detector."\n"
              .'images not yet copied to the shared storage will be lost!';
        $yesno = &zenity_question ($msg,'Restart FURKA','Skip this step');
     } else {
         $msg .= "\n".'ATTENTION: FURKA is running as user "'.$furkaUser.'" and'."\n"
              .'will be restarted as "'.$newUser.'". There is a slight chance'."\n"
              .'that very last '.$detector.' images collected last 10s are not'."\n"
              .'yet copied to the shared storage will be lost.'."\n\n"
              .'You can wait before pressing OK if you want to give FURKA some time.';
        &zenity_info($msg);
        $yesno = 1;
     }
     if ($yesno) {
        print STDOUT 'Reconfiguring FURKA...'."\n";
        if ($detector =~ /pilatus/i) {
           $furkacmd = 'sudo /usr/local/bin/update_furka '.$newUser;
        }
        elsif ($detector =~ /eiger/i) {
           $partial_dir = $homedir;
           $prefix = '/mnt/beegfs/';
           if ($partial_dir !~ /^${prefix}/) {&press_key_exit('*** '.$script.': home directory of '.$newUser.' is not on '.$prefix);}
           $partial_dir =~ s/^${prefix}//;
           $furkacmd = '/usr/local/bin/furka_start '.$uid.':'.$gid.':'.$partial_dir;
        }
        $host = $ENV{HOSTNAME};
        if (!defined $host) {$host=`hostname`; $host=~s/\n//g;}
        if (!defined $host) {print STDERR "!!! HOSTNAME is undefined!\n\n"; $host='undef';}
        $host = lc($host);
        $host =~ s/(-40g)*(\.gmca\.aps\.anl\.gov)*$//;
        if ($host !~ /^${bln}ppu$/) {
           $sshcmd   = 'ssh -aTxq -o BatchMode=yes -o StrictHostKeyChecking=no '.$server;
           $furkacmd = $sshcmd.' '.$furkacmd;
        }
#       print STDOUT $furkacmd."\n";
        $msg = `$furkacmd`;
        if (${^CHILD_ERROR_NATIVE} || $?) {print STDERR '*** '.$script.' command:'."\n".$furkacmd."\n".'failed with code='.${^CHILD_ERROR_NATIVE}.' ('.$?.').'."\n";}
        print STDOUT $msg."\n";
### Now check:
        $cmd = 'perl -W '.${PEZCA}.'/furka_user.pl';
        $furkaUser = `$cmd`;
        if (${^CHILD_ERROR_NATIVE} || $?) {
           $msg = '*** '.$script.' command:'."\n".$cmd."\n".'failed with code='.${^CHILD_ERROR_NATIVE}.' ('.$?.').'."\n"
                . '*** FURKA user could not be queried.';
           print STDOUT $msg."\n";
           $furkaUser = 'unknown';
        }
        elsif ($furkaUser !~ /^furka_user=/) {
           $msg = '!!! Failed to detect current FURKA user. Is FURKA running?';
           print STDOUT $msg."\n";
           $furkaUser = 'unknown';
        } else {
           $furkaUser =~ s/^furka_user=//;
           $furkaUser =~ s/\n$//;
        }
        if ($furkaUser ne $newUser) {&zenity_error_exit('FURKA update failed. It is running as '.$furkaUser);}
     }
  } 
  elsif ($furkaUser eq 'furka_is_disabled') {
     print STDOUT $script.': FURKA is disabled; images will not be transferred to shared storage.'."\n";
  }
  elsif ($furkaUser eq 'streamwriter') {
     print STDOUT $script.': FURKA is disabled, but it is OK because StreamWriter is running.'."\n";
  }
  else {
     print STDOUT $script.': Leaving FURKA alone as it is already runnung for user "'.$newUser.'"'."\n";
  }
}

###########################################################################
sub restart_marccd($) {

  my ($question, $marccdUser, $cmd, $msg, $killc1, $killc2, $killc3, $vnccmd);
  $question = shift(@_);
  ### Restart VNC on the MAR computer so that MARCCD will be running as current beamline user
  $cmd = 'perl -W '.${PEZCA}.'/marccd_user.pl';
  print STDOUT $script.': querying MARCCD user with ['.$cmd.']...'."\n";
  $marccdUser = `$cmd`;
  if (${^CHILD_ERROR_NATIVE} || $?) {
     $msg = '*** '.$script.' command:'."\n".$furkacmd."\n".'failed with code='.${^CHILD_ERROR_NATIVE}.' ('.$?.').';
#    print STDOUT $msg."\n";
     &zenity_info($msg);
     $marccdUser = 'unknown';
  }
  elsif ($marccdUser !~ /^marccd_user=/) {
     $msg = '!!! Failed to detect current MARCCD user. Is MARCCD running?';
#    print STDOUT $msg."\n";
     &zenity_info($msg);
     $marccdUser = 'unknown';
  } else {
     $marccdUser =~ s/^marccd_user=//;
     $marccdUser =~ s/\n$//;  chomp $marccdUser;
  }
  if ($marccdUser ne $newUser) {
     if ($marccdUser eq 'unknown') {$msg='MARCCD GUI is either not running or running without remote interface'."\n";}
     else                          {$msg='MARCCD GUI is running as user "'.$marccdUser.'"'."\n";}
     if ($question) {
         $msg .= 'Do you want to restart MARCCD as user "'.$newUser.'"?';
        $yesno = &zenity_question ($msg,'Restart MARCCD','Skip this step');
     } else {
         $msg .= "\n".'ATTENTION: MARCCD is running as user "'.$marccdUser.'" and'."\n"
              .'will be restarted as "'.$newUser.'"';
        &zenity_info($msg);
        $yesno = 1;
     }
     if ($yesno) {
        print STDOUT 'Restarting MARCCD...'."\n";
        $killc1 = 'sudo -n /bin/killall -9 -q marccd';
        $killc2 = 'sudo -n /bin/killall -9 -q marccd_server_socket';
        $killc3 = 'sudo -n /bin/killall -9 -q CCD_BluIce';
        $vnccmd = 'sudo -n /usr/local/bin/restart_fpe';
        $host = $ENV{HOSTNAME};
        if (!defined $host) {$host=`hostname`; $host=~s/\n//g;}
        if (!defined $host) {print STDERR "!!! HOSTNAME is undefined!\n\n"; $host='undef';}
        $host = lc($host);
        $host =~ s/(-40g)*(\.gmca\.aps\.anl\.gov)*$//;
        $host =~ s/(-10g)*(\.gmca\.aps\.anl\.gov)*$//;
        if ($host !~ /^${bln}mar$/) {
           $sshcmd   = 'ssh -aTxq -o BatchMode=yes -o StrictHostKeyChecking=no '.$server.' ';
           $killc1  = $sshcmd.$killc1;
           $killc2  = $sshcmd.$killc2;
           $killc3  = $sshcmd.$killc3;
           $vnccmd  = $sshcmd.$vnccmd;
        }
        print STDOUT $killc1."\n";
        $msg = `$killc1`;
        if (${^CHILD_ERROR_NATIVE} || $?) {print STDERR '*** '.$script.' command:'."\n".$killc1."\n".'failed with code='.${^CHILD_ERROR_NATIVE}.' ('.$?.').'."\n";}
        print STDOUT $killc2."\n";
        $msg = `$killc2`;
        if (${^CHILD_ERROR_NATIVE} || $?) {print STDERR '*** '.$script.' command:'."\n".$killc2."\n".'failed with code='.${^CHILD_ERROR_NATIVE}.' ('.$?.').'."\n";}
        print STDOUT $killc3."\n";
        $msg = `$killc3`;
        if (${^CHILD_ERROR_NATIVE} || $?) {print STDERR '*** '.$script.' command:'."\n".$killc3."\n".'failed with code='.${^CHILD_ERROR_NATIVE}.' ('.$?.').'."\n";}
        print STDOUT $vnccmd."\n";
        $msg = `$vnccmd`;
        if (${^CHILD_ERROR_NATIVE} || $?) {print STDERR '*** '.$script.' command:'."\n".$vnccmd."\n".'failed with code='.${^CHILD_ERROR_NATIVE}.' ('.$?.').'."\n";}
        print STDOUT $msg."\n";
        sleep(6);
### Now check:
        $cmd = 'perl -W '.${PEZCA}.'/marccd_user.pl';
        $marccdUser = `$cmd`;
        if (${^CHILD_ERROR_NATIVE} || $?) {
           $msg = '*** '.$script.' command:'."\n".$cmd."\n".'failed with code='.${^CHILD_ERROR_NATIVE}.' ('.$?.').'."\n"
                . '*** MARCCD user could not be queried.';
           print STDOUT $msg."\n";
           $marccdUser = 'unknown';
        }
        elsif ($marccdUser !~ /^maccd_user=/) {
           $msg = '!!! Failed to detect current MARCCD user. Is MARCCD running?';
           print STDOUT $msg."\n";
           $marccdUser = 'unknown';
        } else {
           $marccdUser =~ s/^marccd_user=//;
           $marccdUser =~ s/\n$//;
        }
        if ($marccdUser ne $newUser) {&zenity_error_exit('MARCCD user switching failed. It is running as '.$marccdUser);}
     }
  } 
  else {
     print STDOUT $script.': Leaving MARCCD alone as it is already runnung for user "'.$newUser.'"'."\n";
  }
}


###########################################################################
sub commit_group(\$\$;$) {							#\$nerr,\$msg,['last']
  our ($EZCA_GROUP);
  my ($nerr_ref, $msg_ref, $last, $buffer, @lines, $line);

  ($nerr_ref, $msg_ref, $last) = @_;
  if (!defined $last)                  {$last = 0;}
  if ($last ne 'last' && $last ne '1') {$last = 0;}

  if (!defined $EZCA_GROUP || $EZCA_GROUP == 0) {return;}

# print STDOUT 'Sending...'."\n";
  $status = Pezca::EndGroup();  $EZCA_GROUP=0;
# print STDOUT '... Status='.$status."\n";
  if ($status) {
#    &Pezca::Perror("\0");
     $buffer = Pezca::GetErrorString('');
     @lines = split(/[\r\n]/,$buffer);
     foreach $line (@lines) {
        if ($line !~ /OK/ && length($line) > 0) {$$nerr_ref++; $$msg_ref .= $$nerr_ref.'. '.$line."\n"; }
     }
  }
  if ($last ne 'last' && $last ne '1') {
     Pezca::StartGroup(); $EZCA_GROUP=1;
  } else {
     if ($$nerr_ref > 0) {&zenity_error_exit($$msg_ref);}
  }
  return;
}
