#!/usr/bin/perl -W
#--------------------------------------------------------------
# This script checks health of all beamline assemblies (amplifiers
# disabled, fatal following error, limit hit, inpos, open loop...)
#
#     		Author: Sergey Stepanov
#
# Version-1.0:  2020/06/25 using Pezca
#--------------------------------------------------------------
  use strict;
  use Pezca;

  our ($beamline, $NONINTERACTIVE, $USE_ZENITY_FOR_EXIT, $nRetry);
  my (@beamlinelist, $script, $syntax, $PEZCA, $bln_specified);
  my ($narg, $arg, $err, $status, $msg, $threshold);
  my (@assyList, $nAssy, $assy, @axes, @drives, @motors);
  my ($axis, $drive, $motor);
  my ($naxes, $rqsPos, $actPos, $diff, $inPos, @bits);

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;

  $nRetry = 5;
  do $PEZCA.'/Common/motor_subs.pl';
  do $PEZCA.'/Common/common_subs.pl';
  do $PEZCA.'/Common/beamline.pl';

  select STDERR; $|=1;					#set unbuffered output
  select STDOUT; $|=1;					#set unbuffered output

  $NONINTERACTIVE = 0;					#set as '1' to skip all dialogs
  $USE_ZENITY_FOR_EXIT = 1;
  @beamlinelist = ('23i:', '23o:', '23b:', '23d:');
  $script       = &makeScriptName();
  $syntax       = "\n\n".'Syntax: '.$script.' [beamline]'."\n\n";
  $threshold    = 0.2;

  foreach $arg (@ARGV) {
     $arg =~ s/^-{0,2}//;					#trim leading dashes
#    if (grep {$arg =~ /^$_:?$/} @beamlinelist) {		#use this wheh beamlinelist does not contain ':'
     if (grep(/^$arg:?$/, @beamlinelist)) {
	if (defined $bln_specified) {
           &press_key_exit('*** '.$script.': duplicate beamline specification in '.$arg.$syntax,1);
        } else {
	   $bln_specified = 1;
           $arg =~ s/:$//;					#remove ':' at the end
           if ($beamline ne $arg) {
              $beamline = $arg;
              print STDOUT "\n".'!!! '.$script.': Cross operation for '.$beamline."\n";
              do $PEZCA.'/Common/beamline_'.$beamline.'.pl';
           }
        }
     }
  }

  print STDOUT "\n".'This script checks health of all beamline '.$beamline.' motors'."\n";

### Generate the assemblies list from config<BL>.tcl:
  $nAssy = &generateAssyList($beamline,\@assyList,'assyOnly');
  if ($nAssy <= 0) {&press_key_exit('*** '.$script.': No assemblies found in config'.$beamline.'.tcl',1);}
# for ($i=0; $i<$nAssy; $i++) {print STDOUT "${assyList[$i]}\n";}

  foreach $assy (@assyList) {
     $err = 0;
     $assy   = &filterAssy($assy);
     printf STDOUT 'assy=%-20s',$assy;
     $status = &getAssyDataFromTsub($assy,\@axes,\@drives,\@motors);			#reads assembly tsub record
     if ($status) {&press_key_exit('*** '.$script.': error reading assembly data for assy='.$assy.' from Tsub.',1);}
     $inPos = &getInPos($assy);
     if ($inPos == 0) {
        if ($err == 0) {$msg.="\n".'Assembly='.$assy."\n";}
        $err++;
        $msg .= 'Assembly not InPos'."\n";
     }
     $naxes = @axes;
     if ($naxes > 0) {
        foreach $axis (@axes) {
           $actPos = &getAxisAct ($axis);						#fills status, if passed
           $rqsPos = &getAxisRqs ($axis);						#fills status, if passed
           $diff = $actPos-$rqsPos;
           if (abs($diff) > $threshold) {
              if ($err == 0) {$msg.="\n".'Assembly='.$assy."\n";}
              $err++;
              $msg .= sprintf 'Rqs/Act position mismatch for '.$axis.': Rqs=%.3f, Act=%.3f, Diff=%.3f'."\n",$rqsPos,$actPos,$diff;
           }
        }
     } else {
        foreach $drive (@drives) {
           $actPos = &getAxisAct ($drive);						#fills status, if passed
           $rqsPos = &getAxisRqs ($drive);						#fills status, if passed
           $diff = $actPos-$rqsPos;
           if (abs($diff) > $threshold) {
              if ($err == 0) {$msg.="\n".'Assembly='.$assy."\n";}
              $err++;
              $msg .= sprintf 'Rqs/Act position mismatch for '.$drive.': Rqs=%.3f, Act=%.3f, Diff=%.3f'."\n",$rqsPos,$actPos,$diff;
           }
        }
     }
     foreach $motor (@motors) {
# [b18] Open loop mode
# [b19] Amplifier enabled
# [b21] Positive end limit set (soft or hard)
# [b22] Negative end limit set (soft or hard)
        $status = &getMotorSrvStatus($motor,\@bits);
        if ($status) {&press_key_exit('*** '.$script.': error reading error reading status bits-1 for assy='.$assy.', motor='.$motor,1);}
        if ($bits[18] != 0) {
           if ($err == 0) {$msg.="\n".'Assembly='.$assy."\n";}
           $err++;
           $msg .= 'motor='.$motor.' is in open loop'."\n";
        }
        if ($bits[19] == 0) {
           if ($err == 0) {$msg.="\n".'Assembly='.$assy."\n";}
           $err++;
           $msg .= 'motor='.$motor.' amplifier is disabled'."\n";
        }
        if ($bits[21] != 0) {
           if ($err == 0) {$msg.="\n".'Assembly='.$assy."\n";}
           $err++;
           $msg .= 'motor='.$motor.' negative limit is hit'."\n";
        }
        if ($bits[22] != 0) {
           if ($err == 0) {$msg.="\n".'Assembly='.$assy."\n";}
           $err++;
           $msg .= 'motor='.$motor.' positive limit is hit'."\n";
        }
        $status = &getMotorSrvStat2($motor,\@bits);
        if ($status) {&press_key_exit('*** '.$script.': error reading error reading status bits-2 for assy='.$assy.', motor='.$motor,1);}
# [b0] In-position true
# [b1] Fatal following error exceeded
# [b3] Amplifier fault error
# [b5] I2T Amplifier fault error
# [b6] Integrated fatal following error
# [b8] Phasing searh error
# [b10] Home complete
        if ($bits[0] == 0) {
           if ($err == 0) {$msg.="\n".'Assembly='.$assy."\n";}
           $err++;
           $msg .= 'motor='.$motor.' is not positioned'."\n";
        }
        if ($bits[1] != 0) {
           if ($err == 0) {$msg.="\n".'Assembly='.$assy."\n";}
           $err++;
           $msg .= 'motor='.$motor.' has fatal following error'."\n";
        }
        if ($bits[3] != 0) {
           if ($err == 0) {$msg.="\n".'Assembly='.$assy."\n";}
           $err++;
           $msg .= 'motor='.$motor.' has amplifier fault error'."\n";
        }
        if ($bits[6] != 0) {
           if ($err == 0) {$msg.="\n".'Assembly='.$assy."\n";}
           $err++;
           $msg .= 'motor='.$motor.' has integrated fatal following error'."\n";
        }
        if ($bits[8] != 0) {
           if ($err == 0) {$msg.="\n".'Assembly='.$assy."\n";}
           $err++;
           $msg .= 'motor='.$motor.' has phasing search error'."\n";
        }
        if ($bits[10] == 0) {
           if ($err == 0) {$msg.="\n".'Assembly='.$assy."\n";}
           $err++;
           $msg .= 'motor='.$motor.' is not homed'."\n";
        }
     }
     if ($err > 0) {print STDOUT '  errors='.$err;}
     print STDOUT "\n";
  }
  &clearAllCaMonitors();
  if (length($msg) == 0) {$msg = "\n".'No errors found!'."\n";}
  exec('displayer.pl "'.$msg.'" "none" "Motors Health Report"');
# &press_key_exit($script.': All Done!',0);
