#!/usr/bin/perl -W
#--------------------------------------------------------------
# This script checks /var/log/messages on the Pilatus and Eiger 
# computer for failed today's interfaces
#
# ATTENTION:
#   this can only run via sudo -- add it to /etc/sudoers on Eiger & Pilatus DCU
#
#			Author: Sergey Stepanov
#
# Version-1.0: 2013/03/17 initial implementation for Pilatus
# Version-2.0: 2019/11/15 extended for Eiger and renamed from 
#                         checkpilatusconnections to checkdetectorconnections
# Version-2.1: 2021/01/20 corrections for new Eiger DCU OS (API 1.8.0)
#--------------------------------------------------------------
  use strict;

  my ($script, $user, $home, $host, $detector, $messages, @buf, $string);
  my (@months, @ct, $mday, $month, $year, $date);
  my (@nics, $nnic, @exist, @keys, $k, $found, $total);
  my ($ip, $grep, $awk, $i, $addr);

  select STDOUT; $|=1;					#set unbuffered output
  select STDERR; $|=1;					#set unbuffered output

  $script = $0;
  $script =~ s/\\/\//g;					#change \ to /
  $script =~ s/^.*\///g;				#remove everything before last /
  $script =~ s/\.pl$//i;

#---------------------------------------------------------------------
  if ($^O =~ /Win/i) {die $script.': cannot run under Windows.'."\n";}

### This script can run on either bl1pilatus or bl2dcu only:
  if    (defined $ENV{HOSTNAME}) {$host = $ENV{HOSTNAME};}
  elsif (defined $ENV{HOST})	 {$host = $ENV{HOST};}
  else                           {$host=`hostname`; $host=~s/\n//g;}
  if (!defined $host) {print STDOUT '!!! '.$script.': HOSTNAME is undefined!'."\n\n"; $host='undef';}
  $host = lc($host);
  $host =~ s/\..*$//;
  if ($host eq 'localhost') {$host = 'bl2dcu';}
  if ($host ne 'bl1pilatus' && $host ne 'bl1dcu' && $host ne 'bl2dcu') {
     print STDOUT 'host='.$host."\n";
     die '*** '.$script.': can run only on either bl1dcu (bl1pilatus) or bl2dcu'."\n";
  }
  if    ($host eq 'bl1pilatus') {$detector='Pilatus'; $host='bl1dcu';}
  elsif ($host eq 'bl1dcu')     {$detector='Pilatus';}
  elsif ($host eq 'bl2dcu')     {$detector='Eiger';}

### When initial user is not root, but he issued "su",
### the environment "USER" is preserved, but "HOME" is not.
  if    (defined $ENV{USER})     {$user = $ENV{USER};}
  elsif (defined $ENV{USERNAME}) {$user = $ENV{USERNAME};}
  else                           {$user = 'unknown';}
  if    (defined $ENV{HOME})     {$home = $ENV{HOME};}
  else                           {$home = 'unknown';}
  if ($user ne 'root' && $home !~ /^\/root/) {
     print STDOUT "\n".'User='.$user;
     print STDOUT "\n".'Home='.$home;
#    print STDOUT "\n".'This script should be executed as root or via sudo'."\n\n";
     print STDOUT "\n".'Re-launching as root via sudo...'."\n\n";
     exec('/usr/bin/sudo','-n',$0);
     exit 1;
  }

#---------------------------------------------------------------------

  @months = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
  @ct = localtime(time);
  ($mday,$year) = ($ct[3],$ct[5]+1900);
  $month = $months[$ct[4]];
  $date = sprintf $month.'%3d ', $mday;
# $date = 'Oct 31';							#for testing

  if ($detector eq 'Pilatus') {
     $messages = '/var/log/messages';
     if (! -e $messages) {die $script.': file "'.$messages.'" does not exist.'."\n";}
     if (! -r $messages) {die $script.': file "'.$messages.'" is not readable.'."\n";}
     if (&read_entire_file($messages,\@buf)) {die $script.': cannot open "'.$messages.'" for reading.'."\n";}
  } else {
     @buf = split(/\r?\n/,`journalctl -b`);
#    foreach $string (@buf) {print STDOUT $string,"\n";}; exit;
  }

### --------------
### Find active network interfaces:
  $ip   = &Which('ip');         if (!defined $ip)   {die 'Cannot find the "ip" program.'."\n";}
  $grep = &Which('grep');       if (!defined $grep) {die 'Cannot find the "grep" program.'."\n";}
  $awk  = &Which('awk');        if (!defined $awk)  {die 'Cannot find the "awk" program.'."\n";}
  @nics = split(/\r?\n/,`$ip link show | $grep "state UP" | $awk '{print \$2}'`);
  $nnic = @nics;
### Find only those interfaces that connect to detector (have IP 10.0.x.x):
  for ($i=0; $i<$nnic; $i++) {
     $nics[$i] =~ s/:$//;
     $addr = `$ip address show dev $nics[$i] | grep 'inet 10\.' | $awk '{print \$2}'`;
     chomp $addr;
#    print STDOUT $nics[$i].'   '.$addr,"\n";
     if ($addr !~ /^10\.[0-3]\./) {splice(@nics,$i,1); $i--; $nnic--;}
  }
# for ($i=0; $i<$nnic; $i++) {print STDOUT $nics[$i],"\n";}
  if (($detector eq 'Pilatus' && $nnic != 2) ||
      ($detector eq 'Eiger'   && $nnic != 4)) {
     die 'Incorrect number='.$nnic.' of detected '.$detector.' NICs: '."@nics\n";
  }
### --------------
###Example:
#Oct 31 14:42:27 bl1pilatus kernel: ixgbe 0000:04:00.0: p3p1: NIC Link is Down
#Oct 31 15:08:46 bl1pilatus kernel: ixgbe 0000:04:00.1: p3p2: NIC Link is Down
### --------------
  for ($i=0; $i<$nnic; $i++) {
     $exist[$i] = 0;
     $keys[$i]  = ' '.$nics[$i].': NIC Link is Down';
  }
 $total = 0;
 foreach $string (@buf) {
     if ($string !~ /^${date}/)					{next;}
     $found=0;
     foreach ($i=0; $i<$nnic; $i++) {
        $k = $keys[$i];
        if ($string =~ /$k$/) {
           $found++;
           $exist[$i]++;
        }
     }
     if ($found) {
        if ($total == 0) {                                              #first time, print header
           print STDOUT "\n".'## Today\'s failed 10Gbps connections between '
                        .$detector.' detector and its control computer '
                        .$host.' on interfaces '."[@nics]:\n\n";
        }
        print STDOUT $string,"\n";
        $total += $found;
     }
  }

  if ($total == 0) {
     print STDOUT "\n".'## No today\'s failed 10Gbps connections between '
                 .$detector.' detector and its control computer '
                 .$host.' on interfaces '."[@nics].\n\n";
  } else {
     print STDOUT "\n".'## SUMMARY:'."\n";
     foreach ($i=0; $i<$nnic; $i++) {
        print STDOUT '## Found '.$exist[$i].' failed connections on '.$nics[$i]."\n";
     }
  }
  exit 0;

#==========================================================================

# sub read_entire_file($file,\@lines) {
sub read_entire_file($\@) {
# use Fcntl;					#not used (defines O_RDONLY constant (O_RDONLY=0,O_WRONLY=1,O_RDWR=2))
  my ($file, $linesref, $buffer, $status);

  $file     = shift(@_);
  $linesref = shift(@_);

  @$linesref = ();
  $status = sysopen(FH, $file, 0);		#0 is used instead of O_RDONLY defined in "use Fcntl";
  if (!defined $status) {print STDOUT '*** '.$script.': Could not open '.$file."\n"; return 1;}
  sysread(FH, $buffer, -s FH);
  close FH;
  @$linesref = split(/\r?\n/,$buffer);
  return 0;
}


#=================================================================

#sub Which ($prg) {	#returns path to $prg or undef if not found.
sub Which ($) {

  my ($program, $path, @parr, $file, $filewin);
  $program = shift;				#get the passed value
  return if (not defined($program));		#return if nothing is provided
  if ($^O !~ /Win/i) {
     #use ENV;					#load the ENV
     $path = $ENV{PATH};			#load the path
     $path =~ s/\/:/:/g;			#replace all "/:" by ":"
     @parr = split(/:/,$path);			#now make an array (split by ":")
     foreach (@parr) {				#loop and find if the file is in one of the paths
        $file = $_ . '/' . $program;		#concatenate the file
        return $file if ((-e $file) && (-f $file));	#return the path if it was found
     }
  }
  return;
}
