#!/usr/bin/perl -W
#--------------------------------------------------------------
# By default, this script cleans up the JBluIce Clients database by erasing
# all orphan records referring to longer running instances of JBluIce.
# However, if called with the "forceKill" switch, it also attempts to kill
# all running versions of JBluIce and then erase their records from the
# Clients DB. The script relies on passwordless SSH between beamline machines
# and the inclusion into /etc/sudoers of killPID.pl, which it calls in order
# to kill JBluIce.
#
#     	            	Author: Sergey Stepanov
#
# Version-1.0: 2013/03/27 using Pezca
# Version-1.1: 2013/04/01 added stafflist and calling killPID.pl via sudo
# Version-1.2: 2022/09/23 added error handling for backticks operations
#--------------------------------------------------------------
  use strict;
  use Pezca;
# use Net::SSH2::Perl;


  our ($beamline, $IGNORE_ERRORS, $NONINTERACTIVE, $USE_ZENITY_FOR_EXIT);
  my ($PEZCA, $script, $narg, $user, $host, $force, $i, $PV, $status);
  my ($mode, $pid, $jhost, $juser, $state, $syntax, $cmd, $out, $pre);
  my (@beamlinelist, $arg, @staff, $canKill, $sshcmd, $perlcmd);

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;

  do ${PEZCA}.'/Common/common_subs.pl';
  do ${PEZCA}.'/Common/beamline.pl';

  select STDOUT; $|=1;					#set unbuffered output
  select STDERR; $|=1;					#set unbuffered output

  $IGNORE_ERRORS = 0;
  $NONINTERACTIVE = 1;
  $USE_ZENITY_FOR_EXIT = 1;
  @staff = qw(root dkissick makarov mbecker mhilgart ogata qxu scorcoran sergey vnagas gmcaadmin gmca user0 user1 user2);

  $script = &makeScriptName();
# $script =~ s/\.pl$//i;

  $narg = @ARGV;
  $syntax = "\n".'Syntax: '.$script.' [forceKill]'."\n";
  $force = 'checkOnly';
  @beamlinelist = ('23i:', '23o:', '23b:', '23d:');

  foreach $arg (@ARGV) {
     if (grep(/^$arg:?$/, @beamlinelist)) {
        $arg =~ s/:$//;						#trim ':'
        if ($arg ne $beamline) {&press_key_exit('*** '.$script.': cannot run cross beamlines!');}
     }
     elsif ($arg =~ /^-{0,2}forceKill$/i) {
        $force = 'forceKill';
     }
     else {print STDOUT $syntax; exit;}
  }

  if ($^O =~ /Win/i) {&press_key_exit('*** '.$script.': must run on Linux');}

### Determine who is running this:
  if    (defined $ENV{USER})     {$user = $ENV{USER};}
  elsif (defined $ENV{USERNAME}) {$user = $ENV{USERNAME};}
  else                           {&press_key_exit('*** '.$script.': cannot detect current user');}
  if (grep(/^${user}$/,@staff)) {
     print STDOUT 'This user='.$user.' is able to kill processes of other users'."\n";
     $canKill = 1;
  } else {
     print STDOUT 'This user='.$user.' can only view processes of other users'."\n";
     $canKill = 0;
     if ($force eq 'forceKill') {
        &press_key_exit('*** '.$script.': this user cannot kill JBluIce instances of other users.'."\n"
                              .'Please remove the "forceKill" flag or try again as a STAFF user');
        $force = 'checkOnly';
     }
  }

  $host = $ENV{HOSTNAME};
  if (!defined $host) {$host=`hostname`; $host=~s/\n//g;}
  if (!defined $host) {&press_key_exit('*** '.$script.': HOSTNAME is undefined!');}
  $host = lc($host);
  $host =~ s/\.gmca\.aps\.anl\.gov$//;

  if ($user eq 'root') {$pre = 'sudo -u gmca ';} else {$pre = '';}
  $sshcmd  = $pre.'/usr/bin/ssh -aTxq -o BatchMode=yes -o StrictHostKeyChecking=no ';
  $perlcmd = '/usr/bin/perl -W '.$PEZCA.'/killPID.pl --'.$force.' ';

  for ($i=0; $i<10; $i++) {
     print STDOUT "\n".'--------------------------------------------Processing Record='.$i."\n";
     $PV = $beamline.':bi:c'.$i.':';
     ($status,$mode)  = Pezca::GetDouble($PV.'mode');   &report_exit($script, $status, $PV.'mode');
     ($status,$pid)   = Pezca::GetDouble($PV.'pid');    &report_exit($script, $status, $PV.'pid');
     ($status,$jhost) = Pezca::GetString($PV.'host');   &report_exit($script, $status, $PV.'host');
     ($status,$juser) = Pezca::GetString($PV.'user');   &report_exit($script, $status, $PV.'user');
     ($status,$state) = Pezca::GetDouble($PV.'state');  &report_exit($script, $status, $PV.'state');

     if ($pid ne '0') {
        if ($jhost ne $host) {
           if ($force =~ /^forceKill$/i) {
              print STDOUT 'Record='.$i.'. Trying to kill PID='.$pid.' on host='.$jhost.' by user='.$juser."\n";
              $cmd = $sshcmd.$jhost.' /usr/bin/sudo '.$perlcmd.$pid
           } else {
              print STDOUT 'Record='.$i.'. Checking if active PID='.$pid.' on host='.$jhost.' by user='.$juser."\n";
              $cmd = $sshcmd.$jhost.' '.$perlcmd.$pid
           }
        } else {
           if ($force =~ /^forceKill$/i) {
              print STDOUT 'Record='.$i.'. Trying to kill PID='.$pid.' by user='.$juser."\n";
              $cmd = '/usr/bin/sudo '.$perlcmd.$pid
           } else {
              print STDOUT 'Record='.$i.'. Checking if active PID='.$pid.' by user='.$juser."\n";
              $cmd = $perlcmd.$pid
           }
        }
        print STDOUT 'cmd: '.$cmd."\n";
        $out = `$cmd`;
        if (${^CHILD_ERROR_NATIVE} || $?) {print STDOUT "\n".'*** '.$script.' command:'."\n".$cmd."\n".'for PID='.$pid.' failed with code='.${^CHILD_ERROR_NATIVE}.'.'."\n\n"; $status = 0;}
        elsif ($out =~ /PID=${pid} has been killed/)    {print STDOUT '  ...PID='.$pid.' has been killed'."\n"; $status = 1;}
        elsif ($out =~ /PID=${pid} does not exist/)     {print STDOUT '  ...PID='.$pid.' did not exist'."\n";   $status = 1;}
        elsif ($out =~ /PID=${pid} is running as user/) {print STDOUT '  ...PID='.$pid.' is running.'."\n";     $status = 0;}
        elsif ($out =~ /PID=${pid} is executed as user=${juser} while you are user/) {print STDOUT '  ...PID='.$pid.' could not be killed because run as different user'."\n"; $status = 0;}
        else	{print STDOUT '  ...PID='.$pid.' could not be killed because of ssh or another error -- see the dump:'."\n".'['.$out."\n".']'."\n"; $status = 0;}
     } else {
        print STDOUT 'Record='.$i.'. No active PID in record.'."\n";
        $status = 1;
     }
     if ($status) {
        print STDOUT 'Clearing record='.$i.': PID='.$pid.', host="'.$jhost.'", user="'.$juser.'".'."\n";
        $status = Pezca::PutDoubleOldCa($PV.'mode',0);   &report_exit($script, $status, $PV.'mode');
        $status = Pezca::PutDoubleOldCa($PV.'pid', 0);   &report_exit($script, $status, $PV.'pid');
        $status = Pezca::PutStringOldCa($PV.'host','');  &report_exit($script, $status, $PV.'host');
        $status = Pezca::PutStringOldCa($PV.'user','');  &report_exit($script, $status, $PV.'user');
        $status = Pezca::PutDoubleOldCa($PV.'state',0);  &report_exit($script, $status, $PV.'state');
        print STDOUT '  ...Record='.$i.' cleared.'."\n";
     } else {
        print STDOUT 'Skipping record='.$i.': PID='.$pid.', host='.$jhost.', user='.$juser."\n";
     }
  }
  print STDOUT "\n".'--------------------------------------------All Records processed.'."\n";
