#!/usr/bin/perl -W
#--------------------------------------------------------------
#
# This script interfaces EIGER detector via HTTP (via the detector JSON API)
# It can restart DAQ on the EIGER DCU and initialize the detector with
# watching the progress. If the above steps do not help, the script suggests
# to power-cycle EIGER and then repeat them.
#
#                       Author: Sergey Stepanov
#
# Version-1.0:  2019/01/28	initial implementation
# Version-1.1:  2019/01/31      added arm/disarm
# Version-1.2:  2020/10/24      added checking temperatures
#--------------------------------------------------------------
# Subs: &main
#       &refresh_gui();
#       &fpe_running();
#       &fixEiger();
#       &daqRestart([$nobuttons]);
#       &daqRestart_api([$nobuttons]);
#       &daqRestart_webex([$nobuttons]);
#       &checkTemperatures_webex([$nobuttons]);
#       &iniEiger([$nobuttons]);
#       &armDisarm([$nobuttons]);
#       &execEigerCmd($cmd,$target,$timeout,[$nobuttons]);
#       &wait_eiger_completion($timeout,[$target]);
#       &updateLogBox($msg,[$tag]);
#       &disable_buttons();
#       &enable_buttons();
#       &goodbye();
#--------------------------------------------------------------
# use if ! defined &LWP::UserAgent, 'LWP::UserAgent';
# use if ! defined &JSON,           'JSON';
# use if ! defined &Tk,             'Tk';
  use Pezca;						#(see below) 
  use Tk;
# use Tk::ROText;					#(see below) 
# use JSON;						#(see below) do not install json-xs as it is causing conflict
# use LWP::UserAgent;					#(see below) World-Wide Web library for Perl
  use strict;
  use warnings;
# use warnings::unused;

  our ($NONINTERACTIVE, $USE_ZENITY_FOR_EXIT, $IGNORE_ERRORS, $nRetry);
  our ($beamline, $mw, $repeater, $fBtns, $fLog, $logBox);
  our ($fixBtn, $daqBtn, $iniBtn, $armBtn, $tmprBtn);
  our ($fixLbl, $daqLbl, $iniLbl, $armLbl, $tmprLbl);
  my ($PEZCA, $script, $btnWidth, $lblWidth);
  my ($btnbkgcolor, $bpresscolor, $fborderwidth, $frelief);
  my ($greencolor, $goldcolor, $redcolor, $yellowcolor);
  my ($msg, $status, $site, $ua);
  my ($host, $line, $eigerPV, $daq_func);

  if (!defined $ENV{PEZCA}) {
     die 'Please define the PEZCA environment. It must point to the PERL scripts repository';
  }
  $PEZCA=$ENV{PEZCA};
  $PEZCA =~ s/\\/\//g;
  if ($PEZCA !~ /\/$/) {$PEZCA .='/';}

  do $PEZCA.'Common/common_subs.pl';
  if (defined $beamline) {do $PEZCA.'Common/beamline_'.$beamline.'.pl';}
  else                   {do $PEZCA.'Common/beamline.pl';}

  select(STDERR);  $|=1;							#Flush output after each print
  select(STDOUT);  $|=1;							#Flush output after each print

  $script = &makeScriptName();

  if (&try_load_module('LWP::UserAgent')) {&zenity_error_exit($script.': please install LWP::UserAgent Perl module');}
  if (&try_load_module('JSON'))           {&zenity_error_exit($script.': please install JSON Perl module');}
  if (&try_load_module('Tk::ROText'))     {&zenity_error_exit($script.': please install Tk::ROText Perl module');}

  $NONINTERACTIVE      = 1;
  $USE_ZENITY_FOR_EXIT = 1;
  $IGNORE_ERRORS       = 1;
  
  $site = 'http://bl2dcu.gmca.aps.anl.gov';
# $daq_func = 'api';				#this works, but not always (may need to reboot DCU)
  $daq_func = 'webex';				#this seems to work always?

  $btnbkgcolor = '#0097FF';
  $bpresscolor = '#00D8FF';
  $greencolor  = '#00BC00';
  $goldcolor   = '#D0D000';
  $redcolor    = '#FF0000';
  $yellowcolor = '#E0E000';

######################################################################################
  $host = &gethost('short');
  if ($host !~ /^bl2/) {
     &zenity_error_exit($script.': cannot run on host='.$host.'. Please use an IDB computer.');
  }
# $eigerPV = $beamline.':fpe:Acquire';
  $eigerPV = $beamline.':fpe:State_RBV';
  $nRetry  = 3;
  &setEpicsRetry();
  Pezca::AutoErrorMessageOff();
######################################################################################
  $fborderwidth = 0; $frelief = 'flat';						# this is normal setting
# $fborderwidth = 2; $frelief = 'groove';					# this is for debugging (comment out for normal operations)
######################################################################################
### Create a GUI:
  $mw         = MainWindow->new;

  $fBtns      = $mw->Frame(-borderwidth=>$fborderwidth,
                           -relief=>$frelief,                   )->grid(-row=>0, -column=>1, -sticky=>'n',);
  $btnWidth = 45;
  $lblWidth = 65;
  $line = 0;
  $fixBtn   = $fBtns->Button(-text=>'Fix EIGER',
                               -background=>$btnbkgcolor,
                               -activebackground=>$btnbkgcolor,
                               -width=>$btnWidth-1,
                               -command=>\&fixEiger,            )->grid(-row=>$line,   -column=>0, -pady=>2, -padx=>0, -sticky=>'ns');
  $fixLbl   = $fBtns->Label(-text=>'This restarts EIGER DAQ service, then initializes EIGER and sends.'."\n"
                                  .'arm/disarm. FPE must be off. If no success, suggests powercycling EIGER.',
                            -anchor=>'w',
                            -justify => 'left', 
                            -width=>$lblWidth-1,
                                                                )->grid(-row=>$line++, -column=>1, -pady=>2, -padx=>0, -sticky=>'ns');
###---------------------------
  $daqBtn   = $fBtns->Button(-text=>'Restart EIGER DAQ',
                               -background=>$btnbkgcolor,
                               -activebackground=>$btnbkgcolor,
                               -width=>$btnWidth-1,
                               -command=>\&daqRestart,          )->grid(-row=>$line,   -column=>0, -pady=>2, -padx=>0, -sticky=>'ns');
  $daqLbl   = $fBtns->Label(-text=>'This restarts EIGER DAQ service. You must initialize EIGER after that'."\n"
                                  .'and send arm/disarm. FPE must be off.',
                            -anchor=>'w',
                            -justify => 'left', 
                            -width=>$lblWidth-1,
                                                                )->grid(-row=>$line++, -column=>1, -pady=>2, -padx=>0, -sticky=>'ns');
###---------------------------
  $iniBtn   = $fBtns->Button(-text=>'Initialize EIGER',
                               -background=>$btnbkgcolor,
                               -activebackground=>$btnbkgcolor,
                               -width=>$btnWidth-1,
                               -command=>\&iniEiger,            )->grid(-row=>$line,   -column=>0, -pady=>2, -padx=>0, -sticky=>'ns');
  $iniLbl   = $fBtns->Label(-text=>'This initializes the EIGER detector. The process takes ~2 minutes.'."\n"
                                  .'You must send arm/disarm after that. FPE must be off.',
                            -anchor=>'w',
                            -justify => 'left', 
                            -width=>$lblWidth-1,
                                                                )->grid(-row=>$line++, -column=>1, -pady=>2, -padx=>0, -sticky=>'ns');
###---------------------------
  $armBtn   = $fBtns->Button(-text=>'Arm/disarm EIGER',
                               -background=>$btnbkgcolor,
                               -activebackground=>$btnbkgcolor,
                               -width=>$btnWidth-1,
                               -command=>\&armDisarm,           )->grid(-row=>$line,   -column=>0, -pady=>2, -padx=>0, -sticky=>'ns');
  $armLbl   = $fBtns->Label(-text=>'This sends the arm command followed by disarm to the EIGER detector.'."\n"
                                  .'It is needed because 1st arm after ini takes long time. FPE must be off.',
                            -anchor=>'w',
                            -justify => 'left', 
                            -width=>$lblWidth-1,
                                                                )->grid(-row=>$line++, -column=>1, -pady=>2, -padx=>0, -sticky=>'ns');
###---------------------------
  $tmprBtn  = $fBtns->Button(-text=>'Check EIGER temperatures',
                               -background=>$btnbkgcolor,
                               -activebackground=>$btnbkgcolor,
                               -width=>$btnWidth-1,
                               -command=>\&checkTemperatures_webex,)->grid(-row=>$line,   -column=>0, -pady=>2, -padx=>0, -sticky=>'ns');
  $tmprLbl  = $fBtns->Label(-text=>'This checks temperatures of all EIGER detector components.'."\n"
                                  .'Detector must be initialized.',
                            -anchor=>'w',
                            -justify => 'left', 
                            -width=>$lblWidth-1,
                                                                )->grid(-row=>$line++, -column=>1, -pady=>2, -padx=>0, -sticky=>'ns');
###===========================
  $fLog = $mw->Frame(-borderwidth=>$fborderwidth,
                     -relief=>$frelief,                                    )->grid(-row=>1, -column=>0, -columnspan=>2, -sticky=>'nwe');
# $logBox = $fLog->Scrolled('Listbox',
# $logBox = $fLog->Scrolled('Text',                                       			#see: http://docstore.mik.ua/orelly/perl3/tk/ch06_02.htm
  $logBox = $fLog->Scrolled('ROText',                                       			#see: http://docstore.mik.ua/orelly/perl3/tk/ch06_02.htm
                            -scrollbars=>'se',
                            -height=>'20',
			    -font=>[-weight=>'normal', -size=>12, -family=>'Courier'],		#see: http://docstore.mik.ua/orelly/perl3/tk/ch03_01.htm
 			    -foreground=>'#000077',
                            -background=>'#FFFFFF',                        )->pack(-side => 'top',-fill => 'x');
  $logBox->Subwidget("xscrollbar")->configure(-background=>$btnbkgcolor,);
  $logBox->Subwidget("yscrollbar")->configure(-background=>$btnbkgcolor,);
###===========================

### Set up binding:
  $mw->protocol('WM_DELETE_WINDOW', \&goodbye);

  $mw->resizable( 0, 0 );
  $mw->focus();

  $repeater = $mw->repeat(1000,\&refresh_gui);

  $ua = LWP::UserAgent->new(keep_alive=>1, timeout=>'2');
  $ua->cookie_jar( {} );					#create cookies DB
  $ua->agent('Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:64.0) Gecko/20100101 Firefox/64.0');
  $ua->default_header('REFERER' => $site, 'Accept-Language' => 'en');

  MainLoop;
  &goodbye;

#==========================================================================
sub refresh_gui() {
  if (Exists($mw)) {$mw->update;}
}

#==========================================================================
sub fpe_running() {
  my ($status, $val, $statusname);
  ($status, $val) = Pezca::GetString($eigerPV);
  if ($status == 0) {							#EZCA_OK
     &updateLogBox('FPE is running. Please stop it first.','error');
     return 1;
  }
  elsif ($status = 6) {							#EZCA_NOTIMELYRESPONSE
     &updateLogBox('OK, FPE is presumably stopped (PV='.$eigerPV.' is not responding)');
     return 0;
  }
  else {
     $statusname = &ezcaStatusCode($status);
     &updateLogBox('Unexpected response ['.$statusname.'] for FPE PV='.$eigerPV.'.'."\n"
                  .'Presuming that FPE is stopped.','error');
     return 0;
  }
}

#==========================================================================
sub fixEiger() {
  my ($status);
  if (&fpe_running()) {return;}
  &disable_buttons();
  $status = 0;
  if (&daqRestart('nobuttons')) {
     &updateLogBox('Restarting EIGER DAQ service failed. Fix did not succeed.','error');
     $status = 1;
  } else {
     if (&iniEiger('nobuttons')) {
        &updateLogBox('Initializing EIGER failed. Fix did not succeed.','error');
        $status = 1;
     } else {
        select(undef,undef,undef,2);                            #give Eiger a few extra secs to finish
        if (&armDisarm('nobuttons')) {
           &updateLogBox('Arming/disarming EIGER failed. Fix did not succeed.','error');
           $status = 1;
        }
     }
  }
  &enable_buttons();
  if ($status) {&zenity_error($script.': Fixing EIGER communication failed.'."\n\n"
                            .'Consider powercycling EIGER or rebootin DCU.',0,'nowait');}

  return $status;
}

#==========================================================================
#sub daqRestart([$nobuttons]) {
sub daqRestart(;$) {
   my ($status, $nobuttons);
   $nobuttons = shift(@_);
   if    ($daq_func eq 'api')   {$status = &daqRestart_api($nobuttons);}
   elsif ($daq_func eq 'webex') {$status = &daqRestart_webex($nobuttons);}
   else {
      if (!defined $daq_func) {$daq_func = 'undefined';}
      &zenity_error_exit($script.': Incorrect DAQ restart function: '.$daq_func);
      $status = 1;
   }
   return $status;
}

 #==========================================================================
#sub daqRestart_api([$nobuttons]) {
sub daqRestart_api(;$) {
 my ($url, $req, $rsp, $status, $timeout, $msg, $nobuttons);
  $nobuttons = shift(@_);
  $url = $site.'/system/api/1.6.0/command/restart';
  if (&fpe_running()) {return 1;}
  if (!defined $nobuttons) {&disable_buttons();}
  &updateLogBox('Sending EIGER API command: '.$url);
  $req = HTTP::Request->new(PUT => $url);
  $req->header('content-type' => 'application/json; charset=UTF-8');
  $req->header('Accept'       => 'application/json' );
  $rsp = $ua->request($req);
  if ($rsp->is_success|| $rsp->status_line =~ /timeout/) {
     $timeout = 60;
     $status = &wait_eiger_completion($timeout,'na');
     if ($status) {&updateLogBox('Timeout on restarting EIGER DAQ service','error');}
     else         {&updateLogBox('Restarting EIGER DAQ service completed OK');}
  } else {     
      $msg = sprintf 'Error reading response from '.$url.': %s', $rsp->status_line;
      &updateLogBox($msg,'error'); 
      $status = 1;
  }       
  if (!defined $nobuttons) {&enable_buttons();}
  return $status;
}

#==========================================================================
#sub daqRestart_webex([$nobuttons]) {
sub daqRestart_webex(;$) {
  my ($url, $ua10k, $site10k, $rsp, $status, $timeout, $msg, $nobuttons, %FORM);
  $nobuttons = shift(@_);
  if (&fpe_running()) {return 1;}
  $FORM{'user'} = 'eiger';
  $FORM{'pass'} = '#EIGER_Detector#';
  $FORM{'save'} = 1;
  $site10k = $site.':10000';
  $ua10k = LWP::UserAgent->new(keep_alive=>1, timeout=>'2');
  $ua10k->cookie_jar( {} );					#create cookies DB
  $ua10k->agent('Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:82.0) Gecko/20100101 Firefox/82.0');
  $ua10k->default_header('REFERER' => $site10k, 'Accept-Language' => 'en');
#
  $url = $site10k;
  &updateLogBox('Connecting to EIGER Webex service: '.$url);
  $rsp = $ua10k->get($url);
  if ($rsp->is_success || $rsp->status_line =~ /^302 Moved Temporarily/) {
     &updateLogBox('Connection successful.');
  } else {
     &updateLogBox('Connection failed: '.$rsp->status_line,'error');
     return 1;
  }
#
  $url = $site10k.'/session_login.cgi';
  &updateLogBox('Logging into EIGER Webex service: '.$url);
  $rsp = $ua10k->post($url,\%FORM);
  if ($rsp->is_success || $rsp->status_line =~ /^302 Moved Temporarily/) {
     &updateLogBox('Login successful.');
  } else {
     &updateLogBox('Login failed: '.$rsp->status_line,'error');
     return 1;
  }
#
  $url = $site10k.'/eiger-dcu-controls/run.cgi?id=1413362926';
  if (!defined $nobuttons) {&disable_buttons();}
  &updateLogBox('Sending EIGER Webex command: '.$url);
  $rsp = $ua10k->get($url);
  if ($rsp->is_success || $rsp->status_line =~ /^500/) {
     &updateLogBox('Command accepted with status='.$rsp->status_line);
     $timeout = 180;
     $status = &wait_eiger_completion($timeout,'na');
     if ($status) {&updateLogBox('Timeout on restarting EIGER DAQ service','error');}
     else         {&updateLogBox('Restarting EIGER DAQ service completed OK');}
  } else {     
      $msg = sprintf 'Error reading response from '.$url.': %s', $rsp->status_line;
      &updateLogBox($msg,'error'); 
      $status = 1;
  }       
  if (!defined $nobuttons) {&enable_buttons();}
  return $status;
}

#==========================================================================
#sub checkTemperatures_webex([$nobuttons]) {
sub checkTemperatures_webex(;$) {
  my ($url, $ua10k, $site10k, $rsp, $status, $timeout, $nobuttons, %FORM);
  my ($msg, @lines, $line, @words, $nwords, $i);
  my ($okfail, $name, $value, $state, $limits);

  $nobuttons = shift(@_);
# if (&fpe_running()) {return 1;}
  $FORM{'user'} = 'eiger';
  $FORM{'pass'} = '#EIGER_Detector#';
  $FORM{'save'} = 1;
  $site10k = $site.':10000';
  $ua10k = LWP::UserAgent->new(keep_alive=>1, timeout=>'300');
  $ua10k->cookie_jar( {} );					#create cookies DB
  $ua10k->agent('Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:82.0) Gecko/20100101 Firefox/82.0');
  $ua10k->default_header('Content-Type'=>'application/json', 'REFERER'=>$site10k, 'Accept-Language'=>'en');
#
  $url = $site10k;
  &updateLogBox('Connecting to EIGER Webex service: '.$url);
  $rsp = $ua10k->get($url);
  if ($rsp->is_success || $rsp->status_line =~ /^302 Moved Temporarily/) {
     &updateLogBox('Connection successful.');
  } else {
     &updateLogBox('Connection failed: '.$rsp->status_line,'error');
     return 1;
  }
#
  $url = $site10k.'/session_login.cgi';
  &updateLogBox('Logging into EIGER Webex service: '.$url);
  $rsp = $ua10k->post($url,\%FORM);
  if ($rsp->is_success || $rsp->status_line =~ /^302 Moved Temporarily/) {
     &updateLogBox('Login successful.');
  } else {
     &updateLogBox('Login failed: '.$rsp->status_line,'error');
     return 1;
  }
 
  $url = $site10k.'/eiger-tests/run.cgi?id=1415171921';
  if (!defined $nobuttons) {&disable_buttons();}
  &updateLogBox('Sending EIGER Webex command: '.$url);
  $rsp = $ua10k->get($url);
  if (!defined $nobuttons) {&enable_buttons();}
  if ($rsp->is_success || $rsp->status_line =~ /^500/) {
     &updateLogBox('Command accepted with status='.$rsp->status_line);
     $msg = $rsp->decoded_content;
     if    (! defined $msg)                 {&updateLogBox('Undefined response to EIGER status query.','error'); $msg='{}';}
     elsif (length($msg) == 0)              {&updateLogBox('Zero-length response to EIGER status query.','error'); $msg='{}';}
     elsif ($msg !~ /<html> <body><table>/) {&updateLogBox('No data (likely timeout) in the response to EIGER status query.','error'); $msg='{}';}
     else {
        &updateLogBox("\n".'Received reply:');
        @lines = split(/\r?\n/,$msg);
        $msg = '';
        foreach $line (@lines) {
           if (length($line) > 4000 && $line =~ /^<tr>/) {
              $msg = $line;
              last;
           }
        }
        if (length($msg) == 0) {&updateLogBox('Cannot find data in the response to EIGER status query','error'); return 1;}
        @lines = split('<tr>',$msg);
        #<td></td><td>Name</td><td>value</td><td>state</td><td>value_type</td><td>critical_limits</td><td>time</td><td>unit</td><td>critical_values</td></tr>
        #<td><font color="#00FF00">OK</font></td><td>link_0</td><td>up</td><td>normal</td><td>string</td><td>[None, None]</td><td>2020-09-29T12:00:13.541804</td><td></td><td>[u'down']</td></tr>
        shift @lines;							#remove first empty line because of leading <tr>
        shift @lines;							# remove header line
        # 0 = OK/fail
        # 1 = Name
        # 2 = value
        # 3 = state
        # 4 = value_type
        # 5 = critical_limits
        # 6 = time
        # 7 = unit
        # 8 = critical_values
        $msg = sprintf '%-4s %-24s %-14s %-8s %-12s','Sts','Name','Value','State','Crit.limits';
        &updateLogBox($msg);
        foreach $line (@lines) {
           @words = split('<td>',$line);
           shift @words;						#remove first empty word because of leading <td>
           $nwords = @words; 
           for ($i=$nwords; $i<8; $i++) {$words[$i] = '';}
           $okfail = $words[0];
           if ($okfail !~ /OK/) {
              $okfail =~ s/<\/font><\/td>$//;
              $okfail =~ s/^.*>//;
           } else {
              $okfail = 'OK';
           }
           $name   = $words[1]; $name   =~ s/<\/td>$//;
           $value  = $words[2]; $value  =~ s/<\/td>$//;
           $state  = $words[3]; $state  =~ s/<\/td>$//;
           $limits = $words[5]; $limits =~ s/<\/td>$//;
           if ($value !~ /^None$/i && ($okfail !~ /OK/ || $name =~ /_temp/)) {
              $msg = sprintf '%-4s %-24s %-14s %-8s %-12s',$okfail,$name,$value,$state,$limits;
              &updateLogBox($msg);
           }
        }
     }
  } else {     
      $msg = sprintf 'Error reading response from '.$url.': %s', $rsp->status_line;
      &updateLogBox($msg,'error'); 
      $status = 1;
  }       
  return $status;
}

#==========================================================================
#sub iniEiger([$nobuttons]) {
sub iniEiger(;$) {
  my ($cmd, $target, $timeout, $nobuttons, $status);
  $nobuttons = shift(@_); 
  $cmd     = 'initialize';
  $target  = 'idle';
  $timeout = 150;
  $status = &execEigerCmd($cmd,$target,$timeout,$nobuttons);
  return $status;
}

#==========================================================================
#sub armDisarm([$nobuttons]) {
sub armDisarm(;$) {
  my ($cmd, $target, $timeout, $nobuttons, $status);
  $nobuttons = shift(@_); 
  $cmd     = 'arm';
  $target  = 'ready';
  $timeout = 120;
  $status = &execEigerCmd($cmd,$target,$timeout,$nobuttons);
  if ($status) {return $status;}
  $cmd     = 'disarm';
  $target  = 'idle';
  $timeout = 30;
  $status = &execEigerCmd($cmd,$target,$timeout,$nobuttons);
  return $status;
}

#==========================================================================
#sub execEigerCmd($cmd,$target,$timeout,[$nobuttons]) {
sub execEigerCmd($$$;$) {
  my ($cmd, $target, $timeout, $nobuttons);
  my ($url, $req, $rsp, $status, $msg);
  $cmd       = shift(@_);
  $target    = shift(@_);
  $timeout   = shift(@_);
  $nobuttons = shift(@_); 
  $url = $site.'/detector/api/1.6.0/command/'.$cmd;
  if (&fpe_running()) {return 1;}
  if (!defined $nobuttons) {&disable_buttons();}
  &updateLogBox('Sending EIGER API command: '.$url);
  $req = HTTP::Request->new(PUT => $url);
  $req->header('content-type' => 'application/json; charset=UTF-8');
  $req->header('Accept'       => 'application/json' );
  $rsp = $ua->request($req);
  if ($rsp->is_success || $rsp->status_line =~ /^500/) {
     $status = &wait_eiger_completion($timeout,$target);
     if ($status) {&updateLogBox('Timeout on executing '.$cmd,'error');}
     else         {&updateLogBox('Command '.$cmd.' completed OK');}
  } else {     
      $msg = sprintf 'Error reading response from '.$url.': %s', $rsp->status_line;
      &updateLogBox($msg,'error'); 
      $status = 1;
  }       
  if (!defined $nobuttons) {&enable_buttons();}
  return $status;
}

#==========================================================================
#sub wait_eiger_completion($timeout,[$target]) {
sub wait_eiger_completion($;$) {
  my ($timeout, $target, $url, $req, $rsp);
  my ($wait, $step, $microstep, $nstep, $i, $msg, $fromjson);
  $timeout = shift(@_);
  $target  = shift(@_);  if (!defined $target) {$target = 'idle';}
  $url = $site.'/detector/api/1.6.0/status/state';
  $req = HTTP::Request->new(GET => $url);
  $req->header( 'content-type' => 'application/json; charset=UTF-8' );
  $req->header( 'Accept'       => 'application/json' );

  $wait = 0;
  $step = 5;
  $microstep = 0.1;
  $nstep = int($step/$microstep+0.5);
  do  {
     for ($i=0; $i<$nstep; $i++) {select(undef,undef,undef,$microstep);}
     $wait += $step;
     $rsp = $ua->request($req);
     if ($rsp->is_success) {
        $msg = $rsp->decoded_content;
        if    (! defined $msg)    {&updateLogBox('Undefined response to EIGER status query.','error'); $msg='{}';}
        elsif (length($msg) == 0) {&updateLogBox('Zero-length response to EIGER status query.','error'); $msg='{}';}
#       else                      {&updateLogBox('Received reply: '.$msg}
        $fromjson = decode_json($msg);				#this creates a hash of value (applies to utf8)
        if (!defined $fromjson->{'value'}) {
           &updateLogBox('Undefined EIGER state in response to status query.','error');
           $fromjson->{'value'} = 'unknown';
        }
     } 
#    elsif ($rsp->status_line =~ /500 read timeout/ || 
#           $rsp->status_line =~ /500 - Internal Server Error/) {
     elsif ($rsp->status_line =~ /^500/) {
        $fromjson->{'value'} = 'offline';
     }
     else {
        $msg = sprintf 'Error reading EIGER status query: %s', $rsp->status_line;
        &updateLogBox($msg,'error');
        return 1;
     }
     &updateLogBox('Waiting '.$wait.'s for EIGER command completion, current state='.$fromjson->{'value'}.'...');
  } while ($wait < $timeout && $fromjson->{'value'} ne $target);
  if ($fromjson->{'value'} ne $target) {
     &updateLogBox('Timeout waiting for the EIGER command completion.','error');
     return 1;
  } else {
     &updateLogBox('EIGER command completed OK');
     return 0;
  } 
}

#==========================================================================
#sub updateLogBox($msg,[$tag]) {
sub updateLogBox($;$) {
  my ($msg, $tag);
  $msg = shift(@_);
  $tag = shift(@_);
  if (!defined $tag) {				#auto-detect tag, if not specified
     if ($msg =~ /^\*\*\*/ ||
         $msg =~ / error/  ||
         $msg =~ /Error/) {$tag = 'error';}
     else                 {$tag = 'info';}
  }
  if (Exists($mw) && defined $logBox && Exists($logBox)) {
     $logBox->insert('end', $msg."\n", $tag);
     if    ($tag eq 'error') {$logBox->tagConfigure($tag, -foreground=>$redcolor);}
     elsif ($tag eq 'warn')  {$logBox->tagConfigure($tag, -foreground=>$goldcolor);}
     $logBox->see('end');
     if (Exists($mw)) {$mw->update;}
  } else {
     print STDOUT $msg."\n";
  }
}

#==========================================================================
sub disable_buttons() {
  $fixBtn->configure(-state=>'disabled');
  $daqBtn->configure(-state=>'disabled');
  $iniBtn->configure(-state=>'disabled');
  $armBtn->configure(-state=>'disabled');
  $tmprBtn->configure(-state=>'disabled');
}

#==========================================================================
sub enable_buttons() {
  my ($i);
  $fixBtn->configure(-state=>'normal',-relief=>'raised');
  $daqBtn->configure(-state=>'normal',-relief=>'raised');
  $iniBtn->configure(-state=>'normal',-relief=>'raised');
  $armBtn->configure(-state=>'normal',-relief=>'raised');
  $tmprBtn->configure(-state=>'normal',-relief=>'raised');
}

#==========================================================================
sub goodbye() {
  require Tk::ErrorDialog;
  &updateLogBox('Goodbye!'."\n");
  &clearAllCaMonitors();
# $mw->destroy;
# Tk::exit;
  exit 0;
}
