#!/usr/bin/perl -W
#---------------------------------------------------------------------------
#  This scipt prints CIDR for an IP. If works either on LST files or direct IPs
#              In the later case it is called from [myip.php]
#
#                         Author: Sergey Stepanov
# Version 1.0 2016/07/01  initial implementation
# Version 2.0 2016/07/02  added range2cidrlist
# Version 2.1 2018/03/13  corrected typo in regex on line 80
# Version 3.0 2018/03/13  added debugging, added protections against jwhois and too wide CIDR
#---------------------------------------------------------------------------
  use strict;
  use Fcntl;
  use Net::Netmask;  	#for @blocks = range2cidrlist($beginip, $endip);

  my ($script, $narg, $ip, $lstfile, $syntax, $cidr);
  my ($buffer, @lines, $line, @pieces, $debug, $whois);

  select STDERR; $|=1;                          #set unbuffered output
  select STDOUT; $|=1;                          #set unbuffered output

# $debug = 1;					#comment out

  $script = $0;
  $script =~ s/\\/\//g;				#change \ to /
  $script =~ s/^.*\///g;			#remove everything before last "/"

  $syntax = "\n".'Syntax: '.$script.' ip|lstfile'."\n\n";
  $narg = @ARGV;
  if ($narg != 1) {die $syntax;}

  if (-e '/usr/bin/jwhois') {
     if (-e '/usr/bin/whois.md') {$whois ='/usr/bin/whois.md';}
  } else {
     if (-e '/usr/bin/whois')    {$whois ='/usr/bin/whois';}
  }
  if ($debug) {
     if (defined $whois) {print STDERR '>> whois='.$whois."\n";}
     else                {print STDERR '>> whois is undefined, using CIDR ~ /16'."\n";}
  }

  $lstfile = $ARGV[0];

  if (!defined $debug) {$debug=0;}
  
  if (! -e $lstfile) {
     $ip = $lstfile;
     if ($debug) {print STDERR '>> argument is IP='.$ip."\n";}
     &process_ip($ip);
  } else {
     if ($debug) {print STDERR '>> argument is file='.$lstfile."\n";}
     sysopen(LST,$lstfile, O_RDONLY);
     sysread(LST, $buffer, -s LST);
     close(LST);
     @lines = split(/\n/,$buffer);
     foreach $line (@lines) {
        @pieces = split(/\s+/,$line);
	if (@pieces < 2) {next;}
        $ip = $pieces[1];
        if ($debug) {print STDERR '>> line=['.$line.'] IP='.$ip."\n";}
        &process_ip($ip);
     }
  }

#--------------------------------------------------------------------------------
sub process_ip($) {
  my ($ip, $buffer, @lines, $line, $fullline, $cidr);
  my (@blocks, $nblocks, @cidrlist, $toowide);
  $ip = shift(@_);
  if (! &is_ip4($ip)) {die '********* process_ip: Incorrect IP='.$ip.' *********'."\n";}

  if (defined $whois) {
     $buffer = `$whois $ip`;
     @lines = split(/\n/,$buffer);
     undef $cidr;
     foreach $line (@lines) {
        $fullline=$line; 
        if ($line =~ /^CIDR:/) {
           $line =~ s/^CIDR:\s+//;
           $cidr = $line;
           if ($debug) {print STDERR '>> ProcIP: line=['.$fullline.'] is ^CIDR, cidr='.$cidr."\n";}
           last;
        }
        elsif ($line =~ /^a. \[Network Number\]/) {
           $line =~ s/^a. \[Network Number\]\s+//;
           $cidr = $line;
           if ($debug) {print STDERR '>> ProcIP: line=['.$fullline.'] is [Network Number], cidr='.$cidr."\n";}
           last;
        }
        elsif ($line =~ /^inetnum:/) {
           $line =~ s/^inetnum:\s+//;
           @blocks = split(/\s+/,$line);
           $nblocks = @blocks;
           if ($nblocks >= 3 && &is_ip4($blocks[0]) && $blocks[1] eq '-' && &is_ip4($blocks[2])) {
              @cidrlist = &range2cidrlist($blocks[0],$blocks[2]);
              $cidr = join(', ',@cidrlist);
              if ($debug) {print STDERR '>> ProcIP: line=['.$fullline.'] is [inetnum] with words>=3, cidr='.$cidr."\n";}
              last;
	   }
           elsif ($nblocks == 1) {
              $line = $blocks[0];
              @blocks = split(/\//,$line);
              $nblocks = @blocks;
              if ($nblocks == 2 && $blocks[0] =~ /^\d{1,3}\.\d{1,3}(\.\d{1,3}){0,2}$/ && $blocks[1] =~ /^\d{1,2}$/) {
                 $cidr = $line;
                 if ($debug) {print STDERR '>> ProcIP: line=['.$fullline.'] is [inetnum] with words=1, cidr='.$cidr."\n";}
                 last;
              }
           }
           else {next;}
        }
        elsif ($line =~ /^IPv4 Address\s+:/) {
           $line =~ s/^IPv4 Address\s+:\s+//;
           @blocks = split(/\s+/,$line);
           $nblocks = @blocks;
           if ($nblocks >= 3 && &is_ip4($blocks[0]) && $blocks[1] eq '-' && &is_ip4($blocks[2])) {
              @cidrlist = &range2cidrlist($blocks[0],$blocks[2]);
              $cidr = join(', ',@cidrlist);
              if ($debug) {print STDERR '>> ProcIP: line=['.$fullline.'] is ^IPv4 Address, cidr='.$cidr."\n";}
              last;
           } else {next;}
        }
        elsif ($line =~ /\(NET-\d{1,3}-\d{1,3}-\d{1,3}-\d{1,3}-\d{1,3}\)/) {
           $line =~ s/^.*\(NET-\d{1,3}-\d{1,3}-\d{1,3}-\d{1,3}-\d{1,3}\)\s+//;
           @blocks = split(/\s+/,$line);
           $nblocks = @blocks;
           if ($nblocks >= 3 && &is_ip4($blocks[0]) && $blocks[1] eq '-' && &is_ip4($blocks[2])) {
              @cidrlist = &range2cidrlist($blocks[0],$blocks[2]);
              $cidr = join(', ',@cidrlist);
              if ($debug) {print STDERR '>> ProcIP: line=['.$fullline.'] is (NET-..., cidr='.$cidr."\n";}
 	      last;
           } else {next;}
        }
     }
     if (!defined $cidr) {
        print STDERR 'IP='.$ip."\t".'********* no CIDR *********'."\n";
        $cidr = ip2cidr16($ip);
     } else {
        @cidrlist = split(', ',$cidr);
        $toowide = 0;
        foreach $line (@cidrlist) {
           @blocks = split(/\//,$line);
           $nblocks = @blocks;
	   if ($nblocks == 2 && $blocks[1] =~ /^\d+$/ && $blocks[1] < 16) {
	      $toowide=1; 
              if ($debug) {print STDERR '>> ProcIP: cidr=['.$line.'] given by whois provides too wide range. Switching to /16'."\n";}
	      last;
	   }
        }
	if ($toowide) {$cidr = ip2cidr16($ip);}
     }
  } else {				########### no whois program ##################
     $cidr = ip2cidr16($ip);
  }
  print STDOUT 'IP='.$ip."\t".$cidr."\n";
}

#--------------------------------------------------------------------------------
sub is_ip4($) {
  my $ip = shift(@_);
  if ($ip =~ /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/) {return 1;}
  else                                               {return 0;}
}

#--------------------------------------------------------------------------------
sub ip2cidr16($) {
  my ($ip, $cidr, @parts, $nparts);
  $ip = shift(@_);
  @parts = split(/\./,$ip);
  $nparts = @parts;
  if ($nparts != 4) {die '********* ip2cidr16: Incorrect IP='.$ip.' *********'."\n";}
  $cidr = $parts[0].'.'.$parts[1].'.0.0/16';
  return $cidr;
}
