#!/usr/bin/perl -W
#--------------------------------------------------------------
#       This script checks if MARCCD is running and if yes,
#       returns to what user the MARCCD process belongs to
#
#     		Author: Sergey Stepanov
#
# Version-1.0:  2022/06/23 initial implementation for MarCCD in VNC window
# Version-1.1:  2022/09/23 added error catching and handling of the backticks
#--------------------------------------------------------------
  use strict;
  use warnings::unused;

  our ($beamline);
  my ($host, $pscmd1, $pscmd2, $pscmd3, $cmd, $msg, $str, @arr, $narr);
  my ($script, $maruser, $bl, $fhost, $status, $sshcmd, $debug, $stderr);
  my (@bmlist, $user, $mode);

  select STDOUT; $|=1;			# set unbuffered output
  select STDERR; $|=1;			# set unbuffered output
  $stderr = 1;				# output error to STDERR

# $debug = 1;
  if (!defined $debug) {$debug=0;}

  $script = $0;
  $script =~ s/\\/\//g;			#change \ to /
  $script =~ s/^.*\///g;		#remove everything before last /

  $sshcmd = '/usr/bin/ssh -aTxq -o BatchMode=yes -o StrictHostKeyChecking=no ';
  $pscmd1 = '/bin/ps -eo pid,user:19,comm | /bin/grep "marccd$"';
  $pscmd2 = '/bin/ps -eo pid,user:19,comm | /bin/grep "marccd_server_s$"';
  $pscmd3 = '/bin/ps -eo pid,user:19,comm | /bin/grep "CCD_BluIce$"';
  @bmlist = qw(23d 23i 23o 23b);

### Current OS user:
  $user = $ENV{USER};
  if (!defined $user) {$user = 'current';}

#---------
### Determine host and beamline
  $host = $ENV{HOSTNAME};
  if (!defined $host && $^O !~ /Win/i) {$host=`hostname`; $host=~s/\n//g;}
  if (!defined $host) {
     $str = '*** '.$script.': HOSTNAME is undefined!';
     &zenity_error_exit ($str,$stderr);
  }
  $host = lc($host);
  $host =~ s/(-40g)*(\.gmca\.aps\.anl\.gov)*$//;
  $host =~ s/(-10g)*(\.gmca\.aps\.anl\.gov)*$//;
  if ($debug) {print STDERR $script.': host='.$host."\n";}
  if    (defined $ENV{BEAMLINE}) {$bl = $ENV{BEAMLINE};}
  elsif ($host =~ /^bl1/) {$bl=1;}
  elsif ($host =~ /^bl2/) {$bl=2;}
  elsif ($host =~ /^bl3/) {$bl=3;}
  else  {
     $str = '*** '.$script.': Cannot determine beamline for host='.$host;
     &zenity_error_exit ($str,$stderr);
  }
  $beamline = $bmlist[$bl];
#---------

  if    ($bl == 3) {
     ($fhost,$status) = &getCCDfpehost();				# uses cainfo to get host (see below)
     if ($debug) {print STDERR $script.': FPEhost=['.$fhost.']'."\n";}
     if ($status || $fhost eq 'unknown') {$fhost = 'bl'.$bl.'mar';}	# MAR control computer
  }
  else  {
     $str = '*** '.$script.': unsupported beamline='.$bl.' (host='.$host.')';
     &zenity_error_exit ($str,$stderr);
  }
  $fhost =~ s/(-40g)*(\.gmca\.aps\.anl\.gov)*$//;
  $fhost =~ s/(-10g)*(\.gmca\.aps\.anl\.gov)*$//;
  if ($debug) {print STDERR $script.': FPEhost=['.$fhost.']'."\n";}

### If the script is running not on $fhost, but on a remote host, we need to ssh:
  if ($host !~ /^${fhost}/) {			# remote control of MAR computer
     $mode    = 'remote';
     $sshcmd .= $fhost.' ';
     $pscmd1  = $sshcmd.$pscmd1;
     $pscmd2  = $sshcmd.$pscmd2;
     $pscmd3  = $sshcmd.$pscmd3;
  } else {
     $mode = 'local';
  }

  $cmd = $pscmd1;
  if ($debug) {print STDERR $script.': Running cmd='.$cmd."\n";}
  $msg = `${cmd}`;
  if (${^CHILD_ERROR_NATIVE}) {
     $str = '*** '.$script.' command:'."\n".$cmd."\n".'failed with code='.${^CHILD_ERROR_NATIVE}.'.';
     if ($mode eq'remote') {$str .= "\n".'Please verify passwordless ssh connection to '.$fhost.' for the '.$user.' user.';}
     &zenity_error_exit ($str,$stderr);
  }
  @arr = split(/\n/,$msg);
  $narr = @arr;
  if    ($narr == 0) {
     print STDERR '!!! '.$script.': MARCCD is not running'."\n";
     print STDOUT 'marccd_user=none'."\n";
     exit;
  }
  elsif ($narr >  1) {
     print STDOUT 'marccd_user=multiple'."\n";
     $str = '!!! '.$script.': Multiple ('.$narr.') instances of MARCCD found'."\n";
     &zenity_error_exit ($str,$stderr);
  }
  else {
     $msg = $arr[0];
     $msg =~ s/^\s*//;
     @arr = split(/\s+/,$msg);
     $narr = @arr;
     if ($narr != 3) {
        print STDERR '!!! '.$script.': Cannot determine MARCCD user from ['.$msg.']'."\n";
        print STDOUT 'marccd_user=unknown'."\n";
        exit;
     } else {
        $maruser = $arr[1];
     }
  }

  $cmd = $pscmd2;
  if ($debug) {print STDERR $script.': Running cmd='.$cmd."\n";}
  $msg = `${cmd}`;
  if (${^CHILD_ERROR_NATIVE}) {
     $str = '*** '.$script.' command:'."\n".$cmd."\n".'failed with code='.${^CHILD_ERROR_NATIVE}.'.';
     if ($mode eq'remote') {$str .= "\n".'Please verify passwordless ssh connection to '.$fhost.' for the '.$user.' user.';}
     &zenity_error_exit ($str,$stderr);
  }
  @arr = split(/\n/,$msg);
  $narr = @arr;
  if    ($narr == 0) {
     print STDERR '!!! '.$script.': MARCCD-remote is not running'."\n";
     print STDOUT 'marccd_user=none'."\n";
     exit;
  }
  elsif ($narr >  1) {
     print STDOUT 'marccd_user=multiple'."\n";
     $str = '!!! '.$script.': Multiple ('.$narr.') instances of MARCCD-remote found'."\n";
     &zenity_error_exit ($str,$stderr);
  }

  $cmd = $pscmd3;
  if ($debug) {print STDERR $script.': Running cmd='.$cmd."\n";}
  $msg = `${cmd}`;
  if (${^CHILD_ERROR_NATIVE}) {
     $str = '*** '.$script.' command:'."\n".$cmd."\n".'failed with code='.${^CHILD_ERROR_NATIVE}.'.';
     if ($mode eq'remote') {$str .= "\n".'Please verify passwordless ssh connection to '.$fhost.' for the '.$user.' user.';}
     &zenity_error_exit ($str,$stderr);
  }
  @arr = split(/\n/,$msg);
  $narr = @arr;
  if    ($narr == 0) {
     print STDERR '!!! '.$script.': MARCCD-IOC is not running'."\n";
     print STDOUT 'marccd_user=none'."\n";
     exit;
  }
  elsif ($narr >  1) {
     print STDOUT 'marccd_user=multiple'."\n";
     $str = '!!! '.$script.': Multiple ('.$narr.') instances of MARCCD-IOC found'."\n";
     &zenity_error_exit ($str,$stderr);
  }
  print STDOUT 'marccd_user='.$maruser."\n";
  exit;

#=================================================================

#sub zenity_error_exit ($message,[$stderr]) {
sub zenity_error_exit ($;$) {

### Open graphical window to report

  my ($msg, $type, $stderr);
  my ($LUN, $status, $path, @cmd, $string);

  $msg    = shift(@_);                                          #pointer to message
  $stderr = shift(@_);						#0=stdout or 1=stderr
  if (!defined $stderr) {$stderr = 0;}

  $type   = 'error';
  $status = 0;
  $msg =~ s/\&/and/g;
  $msg =~ s/</\"/g;
  $msg =~ s/>/\"/g;
  @cmd = ("--title='".$script."'", '--'.$type, "--text='".$msg."'");
  push @cmd, '--no-wrap';						#this helps to reduce excessive white space below text
  push @cmd, '--ellipsize';						#this fixes the high window size with long texts
  push @cmd, ' >/dev/null', '2>&1';					#redirect both stdout and stderr (preferred)
  $path = '/usr/bin/zenity';
  if ($^O !~ /Win/i && -f $path) {					#Linux with zenity installed
### zenity --error --text="bla-bla-\rbla" --title="Beamstop-IN" --window-icon="/usr/share/pixmaps/applet-critical.png"
     unshift @cmd, $path;						#prepend with 'path'
     push @cmd, '&';
     $string = join(' ',@cmd);						#merge from array
     $status = system($string);						#only string form allows no-wait
     if ($status) {print STDOUT "\n".' !!! Error='.$status.' running '.$path."!\n\n";}
  }

  if ($stderr) {$LUN = \*STDERR;}
  else         {$LUN = \*STDOUT;}
  print $LUN "\n".'*** '.$script.':'."\n".$msg."\n\n";
  exit 1;
}

#=================================================================

# sub getCCDfpehost() {
sub getCCDfpehost() {			#returns ($fpehost,$status)

### This routine returns the name of beamline FPE host (e.g. bl1ppu, bl2epu, or bl3mar)

  our ($beamline);
  my ($script, $pv, $fpehost, $cmd, $str);

  $script   = 'getCCDfpehost';

  ### Here we need any PV hosted by FPE and common for both detectors:
# $pv = $beamline.':fpe:DetCmdStr';					#not present on Eiger2
# $pv = $beamline.':fpe:AcquireTime';					#not present on bl1ppu
  $pv = $beamline.':4:HOSTNAME';
  $cmd = 'cainfo -w 1 '.$pv.' | grep "^    Host:" 2>/dev/null';
  $fpehost = `$cmd`;
  if (${^CHILD_ERROR_NATIVE}) {
     $str = '*** '.$script.' command:'."\n".$cmd."\n".'failed with code='.${^CHILD_ERROR_NATIVE}.'.';
     &zenity_error_exit ($str,$stderr);
  }
  $fpehost =~ s/^\s+Host:\s+//;
  $fpehost =~ s/:\d+\n*$//;
  if ($fpehost !~ /\.gmca\.aps\.anl\.gov$/) {return ('unknown',1);}
  else                                      {return ($fpehost,0);}
}
