#!/usr/bin/perl -W
# _CALLED_FROM_BLUICE_
#--------------------------------------------------------------
# Calculates predicated mono-to-ID mismach based on the SQL 
# values and checks against actual. Returns "mismatch" if
# the difference exceeds tolerance. Also produses a visibility
# calc string for feedback_23x.adl.
#
#                Author: Sergey Stepanov
#
# Version-1.0:  2022/06/29 using Pezca
#--------------------------------------------------------------
# ATTENTION:
# To install MySQL support under Windows (ActivePerl)
# ppm>install DBI
# ppm>install DBD-mysql
#--------------------------------------------------------------
  use DBI;
  use Pezca;
  use warnings::unused;
  use strict;

  our ($IGNORE_ERRORS, $NONINTERACTIVE);
  our ($beamline, $undulator);
  my ($PEZCA, $script, $narg, $arg, $i, $syntax, @beamlinelist, $status);
  my ($bln_specified, $tol_specified, $clc_specified, $dbg_specified, $upd_specified);
  my ($dE_keV, $print_calc, $update_medm, $E_act, $ID_act, $ID_harm, $ID_rqs);
  my ($monoAxis, $nUH, $debug, $slope, $offset, $slope1, $slope3);
  my ($enmono13, @harmEn, @harmNum, @Elower, @Eupper, @Olower, @Oupper);
  my ($str1, $str2, $medm, @lines, $line, $pattern, $found);
# my ($monoAssy, $hkl);

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;

  $IGNORE_ERRORS  = 1;								#GLOBAL: do not break script if error occurred
  $NONINTERACTIVE = 1;								#GLOBAL: do not wait for any answers
 
  do $PEZCA.'/Common/motor_subs.pl';
  do $PEZCA.'/Common/mono_subs.pl';
  do $PEZCA.'/Common/common_subs.pl';
  do $PEZCA.'/Common/beamline.pl';

  select STDERR; $|=1;								#set unbuffered output
  select STDOUT; $|=1;								#set unbuffered output

  $script = &makeScriptName();
  $syntax = "\n\n".'Syntax: '.$script.' [beamline] [tolerance_keV] ["printcalc"] ["updatemedm"] ["debug"]'."\n";
  @beamlinelist = ('23i:', '23o:');
# ##############################
  $dE_keV      = 0.02;								#200eV
  $print_calc  = 0;								#do not print visibility calc for feedback_23x.adl
  $update_medm = 0;								#update feedback_23x.adl with new calc
  $debug       = 0;

  $narg = @ARGV;
### Scan for optional command line arguments:
  for ($i=0; $i<$narg; $i++) {
     $arg = $ARGV[$i];
     $arg =~ s/^-{0,2}//;							#trim leading dashes
     if (grep(/^$arg:?$/, @beamlinelist)) {
        if (!defined $bln_specified) {
           $bln_specified = 1;
           $arg =~ s/:$//;
           if ($arg ne $beamline) {
              print STDOUT "\n".'!!! '.$script.': Cross-beamline operation:'
                             .' controlling '.$arg.' from '.$beamline."\n\n";
              $beamline = $arg;
              do $PEZCA.'/Common/beamline_'.$beamline.'.pl';
           }
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     elsif (&is_decimal($arg)) {
        if (!defined $tol_specified) {
           $tol_specified = 1;
           $dE_keV = $arg;
           if ($dE_keV <= 0.001 || $dE_keV >= 0.1) {
              &press_key_exit('*** '.$script.': tolerance_keV ['.$ARGV[$i].'] is outside the range [0.001-0.1]!',1);
           }
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     elsif ($arg =~ /^printcalc$/i || $arg =~ /^print_calc$/i) {
        if (!defined $clc_specified) {
           $clc_specified = 1;
           $print_calc = 1;
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     elsif ($arg =~ /^updatemedm$/i || $arg =~ /^update_medm$/i) {
        if (!defined $upd_specified) {
           $upd_specified = 1;
           $update_medm = 1;
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     elsif ($arg =~ /^debug$/i) {
        if (!defined $dbg_specified) {
           $dbg_specified = 1;
           $debug = 1;
        } else {
           &press_key_exit('*** '.$script.': duplicate command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
        }
     }
     else {
        &press_key_exit('*** '.$script.': unrecognized command line argument-'
                                             .$i.' ['.$ARGV[$i].']!'.$syntax,1);
     }
  }
  if (! grep(/^$beamline:?$/, @beamlinelist)) {
     &press_key_exit('*** '.$script.': cannot run on beamline='.$beamline.'.',1);
  }

#---------------------------------------------------------------------------

### Configuration parameter -- change if beamline hardware changes:
  $monoAxis = $beamline.':MO:E';
# $monoAssy = $beamline.':MO:En';
### ------------------------------------------------------------------------
### Determine lane switch and harmonic switch parameters;
### read them from BluIce mySQL database:
###              ^^^^^^^^^^^^^^^^
### Read mySQL 'blc2004/Config' table:
  $status = &getIDparams(\$enmono13,\@harmEn,\@harmNum,
                         \@Elower,\@Olower,\@Eupper,\@Oupper);			# a more-friendly wrapper around &getVariableIDoffsets
  if ($status) {
     &press_key_exit('*** '.$script.': Cannot read Mono-ID Offsets data from mySQL',1);
  }

  $nUH = @harmNum;
  if ($nUH > 0 && $debug) {
     print STDOUT $script.': --- Current mySQL parameters ---'."\n";
     print STDOUT '  -- ID offsets:'."\n";
     for ($i=0; $i<$nUH; $i++) {
        printf STDOUT '      ID_harm=%d  at E=%5.2f keV  offset=%5.3f keV'."\n"
                      .'                 at E=%5.2f keV  offset=%5.3f keV'."\n",
                     $harmNum[$i], $Elower[$i], $Olower[$i], $Eupper[$i], $Oupper[$i];
     }
  }

  $E_act   = &getAxisAct ($monoAxis);
  $ID_act  = &getIDEnergy($undulator);
  $ID_harm = &getIDHarmonic($undulator);
# $hkl     = &getMonoHKL();

  $i = int(($ID_harm+1)/2 - 1);						#$i: 0=h1, 1=h3, 2=h5, 3=h7
  if ($i < @Eupper) {
     if ($Eupper[$i] != $Elower[$i]) {
           $slope = ($Oupper[$i] - $Olower[$i]) 
                  / ($Eupper[$i] - $Elower[$i]);
           $offset = $Olower[$i] + ($E_act - $Elower[$i]) * $slope;
        } else {
           $offset  = 0.5 * ($Olower[$i] + $Oupper[$i]);
        }
  } else {
     $offset  = 0;
  }
  $ID_rqs = $E_act + $offset;
  if (abs($ID_rqs-$ID_act) < $dE_keV) {printf STDOUT    'matched, diff=%.3fkeV'."\n", $ID_rqs-$ID_act;}
  else                                {printf STDOUT 'mismatched, diff=%.3fkeV'."\n", $ID_rqs-$ID_act;}

### VisibilityCalc: 
### (A=1)?(ABS(0.074+(B-6)*(0.125-0.074)/(13.299-6)-(C-B))>0.02):(ABS(0.193+(B-13.3)*(0.229-0.193)/(20-13.3)-(C-B))>0.02)
### A=$ID_harm
### B=$E_act
### C=$ID_act
  if ($print_calc || $update_medm) {
#    printf STDERR '(A=1)?(ABS(%.3f+(B-%.3f)*(%.3f-%.3f)/(%.3f-%.3f)-(C-B))>0.02):(ABS(%.3f+(B-%.3f)*(%.3f-%.3f)/(%.3f-%.3f)-(C-B))>0.02)'."\n",
#                      $Olower[0],$Elower[0],$Oupper[0],$Olower[0],$Eupper[0],$Elower[0],
#                      $Olower[1],$Elower[1],$Oupper[1],$Olower[1],$Eupper[1],$Elower[1];
     $slope1  = ($Oupper[0] - $Olower[0]) / ($Eupper[0] - $Elower[0]);
     $slope3  = ($Oupper[1] - $Olower[1]) / ($Eupper[1] - $Elower[1]);
     $str1 = sprintf '(A=1)?(ABS(B+%.3f+%.4f*(B-%.2f)-C)>0.02):(ABS(B+%.3f+%.4f*(B-%.2f)-C)>0.02)',
                                 $Olower[0],$slope1,$Elower[0],     $Olower[1],$slope3,$Elower[1];
     $str2 = $str1; 
     $str2 =~ s/\>/\</g;
     print STDERR $str1,"\n";
     print STDERR $str2,"\n";

     if ($update_medm) {
        $pattern = '(A=1)?(ABS(';
### Update MEDM file:
        $medm = '/home/gmca/epics/gmcaApp/adl/feedback_'.$beamline.'.adl';
        if (! -e $medm) {
           &press_key_exit('*** '.$script.': cannot find file: '.$medm,1);
        }
        $status = &read_entire_file($medm,\@lines);
        if ($status) {
           &press_key_exit('*** '.$script.': cannot read file: '.$medm,1);
        }
        $found = 0;
        foreach $line (@lines) {
           if ($line =~ /calc="\Q${pattern}\E/) {
              if ($line =~ /\>0.02/) {
                 if ($line !~ /\Q${str1}\E/) {
                    $line =~ s/\Q${pattern}\E.*$/$str1"/;
                    $found++;
                 } else {
                    print STDERR $script.': calc ['.$str1.'] in ['.$medm.'] is uptodate.'."\n";
                 } 
              }
              elsif ($line =~ /\<0.02/) {
                 if ($line !~ /\Q${str2}\E/) {
                    $line =~ s/\Q${pattern}\E.*$/$str2"/;
                    $found++;
                 } else {
                    print STDERR $script.': calc ['.$str2.'] in ['.$medm.'] is uptodate.'."\n";
                 }
              }
              else {&press_key_exit('*** '.$script.': unexpected line: ['.$line.'] in file: '.$medm,1);}
           }
        }
        if ($found) {
           if (-e $medm.'.BAK') {
              unlink($medm.'.BAK') || &press_key_exit('*** '.$script.': Cannot delete file ['.$medm.'.BAK].',1);
           }
           rename($medm, $medm.'.BAK') || &press_key_exit('*** '.$script.': Cannot renaem ['.$medm.'] into ['.$medm.'.BAK].',1);
           print STDERR $script.': updating calc record in file ['.$medm.']...'."\n";
           open (DAT,'> '.$medm) || &press_key_exit('*** '.$script.': Cannot open file ['.$medm.'] for wring.',1);
           select DAT; $|=1;						# set unbuffered output
           foreach $line (@lines) {print DAT $line."\n";}
           close(DAT);
        } else {
           print STDERR $script.': nothing to do in file ['.$medm.'].'."\n";
        }
### Update CaQtDM file:
        $medm =~ s/adl$/ui/;
        $str1 =~s/\>/\&gt\;/g;
        $str2 =~s/\</\&lt\;/g;
        if (! -e $medm) {
           &press_key_exit('*** '.$script.': cannot find file: '.$medm,1);
        }
        $status = &read_entire_file($medm,\@lines);
        if ($status) {
           &press_key_exit('*** '.$script.': cannot read file: '.$medm,1);
        }
        $found = 0;
        foreach $line (@lines) {
           if ($line =~ /\Q<string>${pattern}\E/) {
              if ($line =~ /\&gt\;0.02/) {
                 if ($line !~ /\Q${str1}\E/) {
                    $line =~ s/\Q${pattern}\E.*$/$str1\<\/string\>/;
                    $found++;
                 } else {
                    print STDERR $script.': calc ['.$str1.'] in ['.$medm.'] is uptodate.'."\n";
                 } 
              }
              elsif ($line =~ /\&lt\;0.02/) {
                 if ($line !~ /\Q${str2}\E/) {
                    $line =~ s/\Q${pattern}\E.*$/$str2\<\/string\>/;
                    $found++;
                 } else {
                    print STDERR $script.': calc ['.$str2.'] in ['.$medm.'] is uptodate.'."\n";
                 } 
              }
              else {&press_key_exit('*** '.$script.': unexpected line: ['.$line.'] in file: '.$medm,1);}
           }
        }
        if ($found) {
           if (-e $medm.'.BAK') {
              unlink($medm.'.BAK') || &press_key_exit('*** '.$script.': Cannot delete file ['.$medm.'.BAK].',1);
           }
           rename($medm, $medm.'.BAK') || &press_key_exit('*** '.$script.': Cannot renaem ['.$medm.'] into ['.$medm.'.BAK].',1);
           print STDERR $script.': updating calc record in file ['.$medm.']...'."\n";
           open (DAT,'> '.$medm) || &press_key_exit('*** '.$script.': Cannot open file ['.$medm.'] for wring.',1);
           select DAT; $|=1;						# set unbuffered output
           foreach $line (@lines) {print DAT $line."\n";}
           close(DAT);
        } else {
           print STDERR $script.': nothing to do in file ['.$medm.'].'."\n";
        }
     }
  }
