#!/usr/bin/perl -W
#-----------------------------------------------------------------------------------------------
# This is a monochromator feedback script.
# It reads a voltage signal from a specified ADC channel,
# and tries to find a maximum by driving a specified DAC channel.
# It has been updated to work with an MEDM screen.
#
#                Author:       dwy
# Version:
#	1.1	2005/08/05
#	1.2	2005/11/10	Added blinking active light
#	2.1	2006/05/18	Major revisions:	optimize P2
#							multiple search stages
#							peak locked PV
#							multiple beamline functionality
#							less messy code
#	2.15	2006/10/26	Minor revisions:	kill P2 while feedback is active
#							add more info to the log file
#	3.01	2006/11		Fix memory leak when ioc2 reboots
#	3.02	2006/11/17	Make changes so that intensity feedback can be easily
#				queried by JBluIce
#	3.03	2006/11/30	Perform intensity check before component check
#	4.00	2007/02/07	Release for interaction with JBluIce
#	4.00c	2007/06/27	Implements autoshift of P2 at DAC voltage limit
#				Implements averaging for intensity samples
#	4.00e	2007/08/02	Minor changes needed for features mentioned in 4.00c
#				Clean up code
#	4.00f	2007/08/06	Better response to CA error
#	4.00g	2008/06/16	Speed up feedback
#
#  *** This is the last experimental Derek's version dated 2012/06/02 ***
#
#	5.00	2013/01/15	Taken over by Sergey. General cleanup
#	5.01	2013/01/29	Added monitors on some critical PVs. May work faster.
#	5.02	2013/01/31	Changed $mode states from numbers to named constants
#	5.03	2013/05/22	Added a switch to work with CAEN AH401D quad amplifier instead of ADC
#				TODO: switching ranges (gains) in AH401D and checking overflow
#
#	6.00	2014/03/19	SZ: added DAC provided by lockin amplifier
#	6.01	2014/05/23	SS: some cleanup; began making lockin functions
#
#	7.00	2014/04/02	SZ: added hardware feedback: turning off the script if the feedback is on
#	7.01	2014/05/27	SS: HW feedback can be turned on and off while the script stays.
#	7.02	2014/06/09	SS: added posting to JBluIce; changed ADC/DAC mismatch from error to warning
#	7.05	2014/07/19	SS/SZ: merged scanrec-based fly search (SS) with smooth turning HW feedback on/off (SZ)
#	7.06	2014/10/10	SS/SZ: added Carp and added $subscript where missing
#	7.07	2015/02/05	SZ: changed hard feedback offsets on IDD
#	7.08	2015/03/17	SZ: added extra protection agains multiple instances and restricted to running on ws2 & ws3
#	7.08b	2015/06/17	SZ: minor equations tweaks
#	7.09	2016/01/25	SZ: Changes to incorporate hardware FPGA changes
#	7.10	2016/01/25	SS: added giving away Struck/Joerger scaler and pausing/restoring positional feedback
#	7.11	2017/03/01	SS: linked "DAC at limit" conditions to $lockin_use mode.
#	7.12	2018/03/19	SS: fixed 100% CPU usage when waiting for A-shutter, renamed log and improved logging
#	7.13	2018/03/29	SS: changed calculation of P2 scan range and step
#	7.14	2018/07/16	SS: minor cleanup (mostly commenting); no agorithm change
#	7.15	2019/10/18	DK: changed $tol_DAC_error=-1 (NO contol) for 23i because there is currently
#				no readback of that voltage; Change back when wiring is fixed.
#	7.16	2020/01/30	SS: changed turning off HardLock only when not in HW mode
#	7.17	2021/02/06	SS: added state for restoring P2.
#	7.18	2021/08/16	SS: added 333 watch
#	7.19	2021/08/26	SS: adapted to new &calcIDoffset which accounts for the 5th and 7th harmonic
#	7.20	2021/09/22	SS: added $trace
#	7.21	2021/09/22	SS: added beam component check and search after pausing;
#				merged (combined) &intensity_check() and &intensity_check_2();
#				modified &intensity_check() to check the signal of currently
#				selected BPM instead of always BPM1 before
#	7.22	2021/10/07	SS: changed $m1,$m2,$n1,$n2 to more readable BPM,
#				changed normalization for hybrid feedback step to conform max step of soft feedback,
#				removed beam component check and search after pausing (reverted back to 7.20),
#				implemented P2 search button,
#				added closing loop on mono piezo (strain gauge).
#	7.23	2021/10/26	SS: added search range PV and correction post PV
#	7.24	2021/11/02	SS: added $lockin_phase_hardware, $lockin_phase_hybrid, $lockin_recall;
#				added CAEN signal rescaling to what ADC reads and turned CAEN on for 23ID-D
#	7.25	2021/11/15	added option to use serialLIA:AUX instead of Systran DAC128, see: $use_LIA_DAC
#	7.26	2021/12/02	added reading LIA hybrid and hardware phases from PVs
#	7.27	2022/03/10	added monitoring and restarting CAEN
#	7.28	2022/03/27	added detecting when FPGA does not work ($FPGA_on) and fixed changing $CAEN_avertime
#                               which was causing restarting of CAEN acquiring.
#	7.29	2022/04/07	restricted hybrid factor
#	7.30	2022/07/19	reverted at 23i to using Systran DAC & disabled FPGA (see $use_LIA_DAC and $FPGA_on)
#	7.31	2022/09/22	added check for David's monostabilizer script
#	7.32	2022/10/06	removed FPGA
  my $version = '7.32';
#-----------------------------------------------------------------------------------------------
### List of Functions in this file:
#      01.   sub set_variables () {
#      02.   sub log_func_entry ($) {
#      03.   sub log_record ($) {
#      04.   sub open_logfile () {
#      05.   sub init_mstb () {
#      06.   sub set_monitors () {
#      07.   sub get_LIA_phases () {
#      08.   sub set_scanrec () {
#      09.   sub start_scaler () {
#      10.   sub stop_scaler () {
#      13.   sub medm_check ($) {
#      14.   sub feedback () {
#      15.   sub fast_feedback () {
#      16.   sub intensity_check ($) {
#      17.   sub component_check () {
#      18.   sub search () {
#      19.   sub pause_feedback () {
#      20.   sub no_beam_pause () {
#      21.   sub measure_intensity(;$) {
#      22.   sub a_auto_open () {
#      23.   sub post_message ($;$) {
#      24.   sub DAC_check ($) {
#      25.   sub status_write ($) {
#      26.   sub wait_for_dcm () {
#      27.   sub optimize_P2 () {
#      28.   sub recall_P2 () {
#      29.   sub bpm2 () {
#      30.   sub kill_p2 (;$) {
#      31.   sub closeloop_p2 () {
#      32.   sub calc_fwhm_x_E()
#      33.   sub putPV ($$;$$) {
#      34.   sub getPV ($;$) {
#      35.   sub func_error ($$$;$) {
#      36.   sub caget_error ($;$$) {
#      37.   sub caput_error ($$;$$) {
#      38.   sub timeout ($) {
#      39.   sub getPV_eh ($;$) {
#      40.   sub get_lockin_DAC_step ($) {
#      41.   sub get_lockin_DAC_abs ($) {
#      42.   sub get_lockin_ADC ($) {
#      43.   sub close_piezo_loop () {
#      44.   sub multiInstanceHostProtector () {
#      45.   sub signal_handler (;$) {

#-----------------------------------------------------------------------------------------------

  use strict;
  use Pezca;
  use DBI;				#for reading mono offsets from mySQL
  use Time::HiRes qw(gettimeofday tv_interval);
# use Carp qw(shortmess);
  use warnings::unused;

### Declare variables
  our ($beamline, $undulator, $NONINTERACTIVE, $IGNORE_ERRORS);
  our ($enmono13, @harmEn, @harmNum, @Elower, @Eupper, @Olower, @Oupper);	#make global to store SQL data between calls

### Mostab PVs
  my ($PEZCA, $script, $scr, $mostab_pv);
  my ($mostab_stop, $mostab_stop_pv, $mostab_pause, $mostab_pause_pv);
  my ($mostab_sttime, $mostab_sttime_pv, $mostab_stline, $mostab_stline_pv);
  my ($mostab_status, $mostab_status_pv, $mostab_lock, $mostab_lock_pv);
  my ($mostab_use2, $mostab_use2_pv, $last_use2);
  my ($mostab_ch2, $mostab_ch2_pv, $last_ch2);
  my ($mostab_search_mode, $mostab_search_mode_pv, $mostab_show_flow);
  my ($mostab_force_search, $mostab_force_search_pv);
  my ($mostab_search_range, $mostab_search_range_pv);
  my ($mostab_corr_pv);
  my ($mostab_inuse2, $mostab_inuse2_pv, $mostab_opt, $mostab_opt_pv);
  my ($mostab_recall_p2, $mostab_recall_p2_pv);
  my (@message, @message_pv, $nmessage);
  my ($have_CAEN, $CAEN_nch, $CAEN_range, $CAEN_range_pv);
  my ($CAEN_sum_pv, $CAEN_avertime_pv, $CAEN_acquire_pv);

### Other PVs
  my ($E, $E_pv, $id_pv, $E_mov, $E_mov_pv);
  my ($DAC_ch, $DAC_piezo, $DAC_piezo_last, $DAC_piezo_pv, $use_LIA_DAC);
  my ($DAC_piezo_DRVL, $DAC_piezo_DRVH, $DAC_piezo_CNTR);
  my ($DACmin, $DACmax, $DAC_offlimit);
  my ($DAC_rdbk, $DAC_rdbk_pv, $DAC_check_err);
  my ($DAC_piezo_norm1, $DAC_piezo_norm2, $DAC_piezo_offset, $DAC_error_scale);
  my ($ADC_bpm1_pv, $ADC_intrate1_pv, $ADC_ch1);
  my ($ADC_bpm2_pv, $ADC_intrate2_pv, @ADC_chs2);
  my ($ADC_rate_idx, %ADC_rate, $DAC_rdbkrate_pv, $ADC_DACcheck);
  my ($mbsh_pv, $current, $current_pv, $permit, $permit_pv);
  my ($fes_open_pv, $fes_close_pv, $ezcatimeout_default, @mname);
  my ($piezo_servoState_pv);

  my ($a_status, $a_status_pv, $c_status, $c_status_pv);
  my ($p2_assy_pv, $p2_axis_pv, $p2_mtr_pv, $p2_recall_pv);
  my ($FPE_start_pv, $FPE_exposure_pv, $BI_fdbk_pv);

### Misc vars
  my ($mSTARTING, $mACTFDBK1, $mINTCHK1, $mCOMPCHK, $mSEARCH, $mPAUSED);
  my ($mNOBEAM, $mACTFDBK2, $mAOPEN, $mHWMON, $mEXIT);
  my ($mOPTIMP2, $mINTCHK2, $mRECALLP2);
  my ($mWAITBEAM, $mWAITOPNA, $mWAITOPNC, $mWAITDCM, $mWAITEMIS);
  my ($int_low1, $int_ok1, $int_overflow1);
  my ($int_low2, $int_ok2, $int_overflow2);
  my ($int2_is_ok, $last_int2_is_ok);
  my ($multiplier, $multiplier_LIA, $scale0to10_LIA);
  my ($logdir, $logfile, $filehandle);
  my ($mode, @opt_steps, $pz_calib, $volt2urad, $urad2volt);
  my ($reversals, $v_div, $pitch_sig, $sigcount, $LOGLUN);
  my ($fdbk_rate, $message_time, $pause_time, $component_time, $current_time);
  my ($search_steps, $fwhm_x_E111, $fwhm_x_E333, $fwhm2sig, $t_max);
  my ($init_E, $fin_E, $tol_E, $tol_E_HW, $wait_dcm, $max_E_tol);
  my ($tol_DAC_error, $p2_autoshift);
# my ($p2_step);
  my ($mbsh, $last_mbsh, $ch_ok);
  my ($n_steps_DAC, $int_samples, $init_time, $maxStrLen);
  my ($hkl, $hkl_old, $trace, $trace_max);

### Lockin amplifier PVs
  my ($lockin_asyn_pv, $lockin_use_pv, $lockin_use, $lockin_prev);
  my ($DCM_VSUM, $DCM_VSUM_max, $status, $value);
  my ($asynTraceFile, $asynTraceDir);
  my ($lockin_ADC1_pv, $lockin_xOUT_pv);
  my ($lockin_ADC1, $lockin_recall); 
  my ($lockin_phase_hardware, $lockin_phase_hardware_pv);
  my ($lockin_phase_hybrid,   $lockin_phase_hybrid_pv);

### Joerger scaler & scan record PVs
  our ($joergerName, $scanrecName, $scanrecAbort, $scanrecPause);
  our ($struckName, $STRUCK_MODEL);
  my ($have_joerger, $joerger_time, $joerger_bpm1,  $joerger_bpm2);
  my ($scaler, $scalerNo, $p2_max_speed, $p2_speed, $halfrange);
  my ($scanrec, $scanrec_max_pts, $scan_in_progress);
  my ($struck, $scaler_mode);

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;

# do $PEZCA.'/Common/struck_subs.pl';          		# if we have a dedicated Struck
  do $PEZCA.'/Common/joerger_subs_v2.pl';		# if we have a dedicated Struck or Joerger
  do $PEZCA.'/Common/scanrec_subs.pl';         		# shall we do fly scans with scanrec?
  do $PEZCA.'/Common/struck_subs_v2.pl';       		# this is only for interacting with Struck as Joerger
# do $PEZCA.'/Common/shutter_subs.pl';
# do $PEZCA.'/Common/dac_subs.pl';
  do $PEZCA.'/Common/motor_subs.pl';
  do $PEZCA.'/Common/mono_subs.pl';
  do $PEZCA.'/Common/common_subs.pl';
  do $PEZCA.'/Common/beamline.pl';

  select STDOUT; $|=1;					# set unbuffered output
  select STDERR; $|=1;					# set unbuffered output
  $LOGLUN = \*STDERR;					# where to print log (either logfile or STDERR)

  $NONINTERACTIVE = 1;
  $IGNORE_ERRORS  = 1;

### Various debug flags:
  $trace        = 0;					# debugging: trace intensity variations (output to STDOUT)
  if (!defined $trace) {$trace=0;}
  $trace_max    = 5;					#max iterations to trace
  $mostab_show_flow = 0;				#1=log entry into each function

### Mode definitions:  (some only used within subroutines and not globally)
  $mSTARTING =   0; $mname[0]='STARTING';		# starting
  $mACTFDBK1 =   1; $mname[1]='ACTFDBK';		# active feedback
  $mINTCHK1  =   2; $mname[2]='INTCHK1';		# intensity check
  $mCOMPCHK  =   3; $mname[3]='COMPCHK'; 		# component status check
  $mSEARCH   =   4; $mname[4]='SEARCH'; 		# beam search
  $mPAUSED   =   5; $mname[5]='PAUSED'; 		# paused
  $mNOBEAM   =   7; $mname[7]='NOBEAM '; 		# no beam, paused
  $mACTFDBK2 =   8; $mname[8]='ACTFDBK2'; 		# active feedback; large step size
  $mOPTIMP2  =   9; $mname[9]='OPTIMP2'; 		# optimize P2
  $mINTCHK2  =  10; $mname[10]='INTCHK2'; 		# intensity check 2 (differs from the first intensity check in how it responds to a 'no-beam' condition)
  $mAOPEN    =  11; $mname[11]='AOPEN';			# open A-shutter after beam comes back
  $mHWMON    =  12; $mname[12]='HWMON';			# hardware feedback monitoring mode
  $mRECALLP2 =  13; $mname[13]='RECALLP2'; 		# recall P2
  $mEXIT     = 999; $mname[999]='EXIT'; 		# exiting;
  $mWAITBEAM =  31; $mname[31]='WAITBEAM';		# Waiting for APS to deliver beam
  $mWAITOPNA =  32; $mname[32]='WAITOPNA';		# Waiting for user to open A-shutter
  $mWAITOPNC =  33; $mname[33]='WAITOPNC';		# Waiting for user to open C-shutter
  $mWAITDCM  =  34; $mname[34]='WAITDCM';		# Waiting for DCM to finish moving
  $mWAITEMIS =  35; $mname[35]='WAITEMIS';		# Waiting for user to fix ID/DCM energy

  $SIG{HUP}  = 'signal_handler';	# (1) hangup (see signum.h)
  $SIG{INT}  = 'signal_handler';	# (2) interrupt (see signum.h)
  $SIG{QUIT} = 'signal_handler';	# (3) quit (see signum.h)
# $SIG{KILL} = 'signal_handler';	# (9) this is what is sent by kill -9 -- unblockable
  $SIG{TERM} = 'signal_handler';	# (15) terminate -- sent by kill without paramaters
# $SIG{STOP} = 'signal_handler';	# (19) stop -- unblockable

  $script = &makeScriptName();
  $script =~ s/\.pl$//i;
  $scr    = 'mostab';

  if ($^O =~ /Win/i) {die $script.': cannot run under Windows.'."\n";}

### Protect from multiple instances:
  &multiInstanceHostProtector();

### One-time subs:
  &set_variables();			# Set initial variable values (also sets $mode=$mSTARTING)
  &open_logfile();			# Open logfile
  &init_mstb();				# Initial status check, set PVs, etc.
  &set_monitors();			# Set monitors
  &get_LIA_phases();			# Read LIA phases from PVs
# &close_piezo_loop();			# Turn on close loop on mono piezo - SS added 2021/10
  if ($have_joerger) {&set_scanrec();}

  while (1) {

     if ($mode == $mINTCHK1  ||
         $mode == $mINTCHK2  ||
         $mode == $mCOMPCHK  ||
         $mode == $mSEARCH   ||
         $mode == $mPAUSED   ||
         $mode == $mNOBEAM   ||
         $mode == $mOPTIMP2  ||
         $mode == $mRECALLP2 ||
         $mode == $mAOPEN) {

        $lockin_use = &getPV($lockin_use_pv);
        if ($lockin_use > 1) {$lockin_use=1; &putPV($lockin_use_pv,$lockin_use,'double');}
     }
     $mode = &medm_check($mode);					# 1

     if    ($mode == $mACTFDBK1) {&feedback();}
     elsif ($mode == $mACTFDBK2) {&fast_feedback();}
     elsif ($mode == $mINTCHK1)  {&intensity_check($mode);}
     elsif ($mode == $mINTCHK2)  {&intensity_check($mode);}		# This added to aid in efficient flow of program
     elsif ($mode == $mCOMPCHK)  {&component_check();}
     elsif ($mode == $mSEARCH)   {&search();}
     elsif ($mode == $mPAUSED)   {&pause_feedback();}
     elsif ($mode == $mNOBEAM)   {&no_beam_pause();}
     elsif ($mode == $mOPTIMP2)  {&optimize_P2();}
     elsif ($mode == $mRECALLP2) {&recall_P2();}
     elsif ($mode == $mAOPEN)    {&a_auto_open();}			# wait for A-shutter to open

#    ### Sergey: this is temporary for testing the fly scan in the search function. Please comment out!
#    $NONINTERACTIVE = 0; &press_key_continue(); $NONINTERACTIVE = 1;
#    &post_message('------------ Going to search');
#    $mode = $mSEARCH;
#    &putPV($mostab_search_mode_pv,1,'double');			#set search mode to searching with P2
#    ### Sergey: end of temporary for testing the fly scan in the search function.

  }

  &signal_handler('Unexpected end of loop');

  exit;

#-----------------------------------------------------------------------------------------------

#==========================================================================
sub set_variables () {

  my ($subscript, $sbscrpt, $i);

  $sbscrpt = 'set_variables';
  $subscript = $script.'/'.$sbscrpt;
  &log_func_entry($subscript);

###--------------------------------------------------------------------------------
### User variables - These are the variables that staff might need to change
# $beamline     = '23i';				# Enter beamline here
  $have_CAEN    = 0;					# switch between CAEN quadAmp and ADC
  if ($beamline eq '23i') {$have_CAEN = 1;}		# SS: 2021.11.03
  if ($beamline eq '23o') {$have_CAEN = 1;}		# SS: 2021.11.09
  $have_joerger = 1;					# if BPM1 is connected to Joerger, then we can use P2 fly scans with scan record, otherwise we do step scans with ADC (no scan record)
  $logdir       = '/home/gmca/shared/logfiles/';	# Enter path for directory where logfile should be stored
  $logfile      = &makeScriptName('short');		# The root portion of the name of the logfile
  $logfile      = $beamline.'_'.$logfile;		# The root portion of the name of the logfile
  $ADC_rate_idx = 9;					# 6=1Hz; 7=2Hz; 8=5Hz; 9=10Hz
  if ($ADC_rate_idx<6 || $ADC_rate_idx>9) {die 'Incorrect ADC_rate';}
  $ADC_rate{6}  = 1.0;					# 6=1Hz
  $ADC_rate{7}  = 0.5;					# 7=2Hz
  $ADC_rate{8}  = 0.2;					# 8=5Hz
  $ADC_rate{9}  = 0.1;					# 9=10Hz
  $fdbk_rate    = 0.5;					# Pause in sec between active feedback steps; should be >= $ADC_rate
  if ($fdbk_rate < $ADC_rate{$ADC_rate_idx}) {die 'Incorrect fdbk_rate';}
  $search_steps = 2;					# Number of iterations before feedback is 'locked'; should be at least two; first, fast; last, slow
  if ($beamline ne '23i') {$use_LIA_DAC = 1;}		# use LIA DAC instead of Systran DAC
  $DAC_ch       = 1;					# Systran DAC channel to drive piezo (ignored if LIA AUX is used)
  $ADC_ch1      = 1;					# ADC channel for white hutch mono beam intensity (DCM_VSUM)
  @ADC_chs2     = (3,5,8);				# ADC channels which may be used for bpm_2; first channel is default (3=HDM_VSUM, 5=GS_UP, 8=IonChamber)
  $ADC_DACcheck = 13;					# Channel to monitor DAC voltage
  $DAC_offlimit = 0.05;					# in Volts: ($DACmin,$DACmax)=($DAC_piezo_DRVL+$DAC_offlimit,$DAC_piezo_DRVH-$DAC_offlimit)

  ### Joerger and scan record
  $scan_in_progress = 0;
  $joerger_time     = 0.1;				# counting time (autocount)
  $joerger_bpm1     = 9;				# input for BPM-1: DCM_VSUM
  $joerger_bpm2     = 0;				# input for BPM-2: none
  $scalerNo         = 2;				# Struck as Joerger in IOC-1. Should be in auto-count mode
  $STRUCK_MODEL = ('SIS3820','SIS3801')[$scalerNo-1];	#s1=SIS3820, s2=SIS3801

  ### DAC multipliers/dividers
  ### $multiplier_LIA - see &get_lockin_DAC_step()
  ### $DCM_VSUM_max   - see &feedback() and &fast_feedback() - Normalization terms to prevent oscillations near harmonic ID-B 1-3 switch.
  ###                   SS: why is &fast_feedback() normalization is 10x larger?
  ### $scale0to10_LIA - see &get_lockin_DAC_abs()
# if    ($beamline eq '23i') {$multiplier=1.00; $multiplier_LIA=15; $DCM_VSUM_max=2.8;  $scale0to10_LIA=1000;}	#SZ data used prior to 2021/10
# elsif ($beamline eq '23o') {$multiplier=1.00; $multiplier_LIA=50; $DCM_VSUM_max=0.8;  $scale0to10_LIA=1000;}	#SZ data used prior to 2021/10
  if    ($beamline eq '23i') {$multiplier=1.00; $multiplier_LIA=15; $DCM_VSUM_max=4.65; $scale0to10_LIA=1000;}	#SS entered 2021/10 per NV
  elsif ($beamline eq '23o') {$multiplier=1.00; $multiplier_LIA=50; $DCM_VSUM_max=1.10; $scale0to10_LIA=1000;}	#SS entered 2021/10 per NV
  else                       {$multiplier=1.00;                     $DCM_VSUM_max=9.9;  $scale0to10_LIA=1000;}

### If $lockin_recall=1 is defined, then LIA is instructed to recall the HW or Hybrid
### config sets on each mode switching
# $lockin_recall = 1;										#SS

### The lines below are no longer needed: since version 7.26 the phases are read from PVs
### If $lockin_recall is undefined or zero and $lockin_phase_hardware is defined, then LIA
### will set HW or Hybrid phases on each mode switching.
### Per Sioan/David the hybrid mode phase should be 45 degrees less than the HW mode phase:
  if (defined $lockin_phase_hardware) {$lockin_phase_hybrid = $lockin_phase_hardware - 45;}	#SS

  ### Voltage step calculation parameters:
  ### +--------------------------------------------------------------------+
  ### | $volt_step = ($fwhm_x_E/$E) * ($pitch_sig/$fwhm2sig) * $urad2volt; |
  ### +--------------------------------------------------------------------+
  ### Bragg curve FWHM for 111 and 333
  $fwhm_x_E111  = 400;					# 400/E is roughly the rocking curve FWHM in microrad at a given energy for Si-111
  $fwhm_x_E333  = 80;					# 80/E is roughly the rocking curve FWHM in microrad at a given energy for Si-333
  $pz_calib     = 197;					# Enter the calibration (urad/V) for the piezo without voltage dividers (Pitch motion of second xtal with 1 V change on Pz)
  $fwhm2sig     = 2.354;				# Conversion from FWHM to sigma
  
  ### Voltage dividers					# Enter ratio of voltage divider -> x:1
# if    ($beamline eq '23i') {$v_div = 10.0;}		# 23i -> use {(900 k / 100 k) = 10.0} with 12-bit DAC; set to 1.0 with 16-bit DAC
  if    ($beamline eq '23i') {$v_div = 1.0;}		# 2022_Jul_19 DK: changed to use with 20k Ohm Voltage divider
  elsif ($beamline eq '23o') {$v_div = 15.7;}		# 23o -> 1 M / 68.1 k
  else                       {$v_div =  1.0;}		# default -> No divider

  $volt2urad = $pz_calib / $v_div;
  $urad2volt = 1 / $volt2urad;

  ### Pitch step calculation: desired pitch change in terms of rocking curve sigma (0.14 sigma = 1% detune of Gaussian)
  if ($beamline eq '23i') {
#    $pitch_sig = 0.014;				# 23i -> 0.014 typically used; testing with smaller values
     $pitch_sig = 0.021;				# 23i -> 2011-10-17: Larger step actually reduces noise over long term
#    $pitch_sig = 0.028;				# 23i -> 2012-02-09: Even larger step?
#    $pitch_sig = 0.042;				# 23i -> 2021-10-05: Larger step actually reduces noise over long term
  }
  elsif ($beamline eq '23o') {
#    $pitch_sig = 0.012;				# 23o -> Used during 2006 run 1; motion not seen, even after lever arm
#    $pitch_sig = 0.048;				# 23o -> Test for high energy and secondary source
     $pitch_sig = 0.024;				# 23o -> Test for high energy and secondary source
  }
  else {
     $pitch_sig = 0.140;				# default -> Not optimized
  }

  ### See &DAC_check() - checking commanded DAC vs DAC_readback (optional)
  ### $DAC_error = $DAC_error_scale*($DAC_rdbk - ($DAC_piezo/$DAC_piezo_norm1 + $DAC_piezo_offset))/$DAC_piezo_norm2;
  if   ($beamline eq '23i') {$DAC_piezo_norm1=28.2; $DAC_piezo_norm2=3;  $DAC_piezo_offset=0.00517; $DAC_error_scale=1;}
  elsif($beamline eq '23o') {$DAC_piezo_norm1=75.4; $DAC_piezo_norm2=30; $DAC_piezo_offset=-0.0330; $DAC_error_scale=0.754;}
  else                      {$DAC_piezo_norm1=1;    $DAC_piezo_norm2=1;  $DAC_piezo_offset=0;       $DAC_error_scale=0;}

  ###-----------------------------------------------------------------------------
  ### Use this space to log comments regarding voltage dividers and pitch changes
  ###
  ### 23i ->	2006-1 used a divide-by-3 and a pitch change of 4.7% sigma.
  ###		2006-2 introduced divide-by-10; pitch change reduced to 1.4%
  ###
  ###-----------------------------------------------------------------------------

# @opt_steps       = (20,3,1);		# These are the step sizes during optimization in &fast_feedback()
  @opt_steps       = (20,10,5,2);	# These are the step sizes during optimization in &fast_feedback() -- SS changed 2021/10.
  $reversals       = 2;			# How many times to step over peak before 'lock'?
  $tol_E           = 0.1;		# How far can the mono move before a re-tune is needed?
# $tol_E_HW        = 0.4;		# same for HW mode. Should be larger than multistep (0.25 KeV)
  $tol_E_HW        = 50;		# same for HW mode. Should be larger than multistep (0.25 KeV). Large value disables pausing HW feedback at energy changes
  $int_ok1         = 0.1;		# How high does intensity need to be before confidently saying beam is present?
  $int_ok2         = $int_ok1;		# How high does intensity need to be before confidently saying beam is present?
  $int_overflow1   = 11.;		# when intensity overflows on bpm1
  $int_overflow2   = $int_overflow1;	# when intensity overflows on bpm2 - not used
# if ($have_CAEN) {
#    $int_overflow1 = 1048575;		# when intensity overflows on bpm1 (see page-9 of CAEN manual)
#    $int_ok1 = $int_overflow1/100;	# How high does intensity need to be before confidently saying beam is present (we take 1% of overflow)?
# }
  $int_low1        = $int_ok1;		# (May also be numeric)	# How low can intensity drop before beam is considered 'lost'?
  $int_low2        = $int_ok2;		# (May also be numeric)	# How low can intensity drop before beam is considered 'lost'?
  $message_time    = 0.0;		# How long between a series of messages?
  $pause_time      = 1;			# How long between PV checks when paused or waiting?
  $component_time  = 3;			# How long after beam loss until feedback checks status of components?
  $tol_DAC_error   = 0.02;		# How much error is allowed between DAC requested and actual? (set <=0 for no control)
  if ($beamline eq '23i') {$tol_DAC_error=-1;} #TEMP: disable control because no readback of that voltage at IDD
  $p2_autoshift    = 1;			# 1 = autoshift P2 at DAC limit; 0 = user intervention needed
  $n_steps_DAC     = 125;
  $int_samples     = 5;			# How many samples should be averaged for each intensity measurement?
# $int_samples     = 3;			# Change for high E
# $max_E_tol       = 0.02;		# Max error in E_ID compared to correct offset
  $max_E_tol       = 10.1;		# Increased 2009_01_28 because ID calibration has shifted
# $mostab_use2     = 1;			# Should a second bpm be used?
  $maxStrLen       = 39;		# maximum length of EPICS string PV (DO NOT CHANGE!)
  $t_max           = 300;		# max seconds to wait when PV connection is lost
  $DAC_piezo_last  = 5;			#this is just the middle, may be overwritten later
  $int2_is_ok      = 0;			# see &bpm()
  $last_int2_is_ok = 0;			# see &bpm()
  ###--------------------------------------------------------------------------------

  ### Insertion devices
  if (length($undulator) > 0) {$id_pv = $undulator.':Energy';}
  else                        {undef $id_pv;}

  ### Scaler and scanrec PVs:
  $scaler           = $joergerName.$scalerNo;
  if ($scaler !~ /^${beamline}/) {$scaler = $beamline.':'.$scaler;}
  $struck = $struckName;				#this is the same Struck scaler in the multiscaler mode
  $struck =~ s/[0-9]{1}$/$scalerNo/;
  if ($struck !~ /^${beamline}/) {$struck = $beamline.':'.$struck;}

  $scanrecName   = $beamline.':mostab:scan1';
  $scanrecAbort  = $beamline.':mostab:AbortScans';
  $scanrecPause  = $beamline.':mostab:scanPause';
  $scanrec       = $scanrecName;

  ### Mostab PVs
  $mostab_pv                = $beamline.':mostab:';
  $mostab_status_pv         = $mostab_pv.'status';		$mostab_status       = 2;	#0=exiting, 1=active, 2=doing.operation, 3=waiting, 4=lost
  $mostab_stline_pv         = $mostab_pv.'statusline';		$mostab_stline       = '';
  $mostab_sttime_pv         = $mostab_pv.'statustime';		$mostab_sttime       = '';
  $mostab_pause_pv          = $mostab_pv.'pause';		$mostab_pause        = 1;
  $mostab_stop_pv           = $mostab_pv.'stop';		$mostab_stop         = 1;
  $mostab_lock_pv           = $mostab_pv.'lock';		$mostab_lock         = 0;
  $mostab_opt_pv            = $mostab_pv.'opt';			$mostab_opt          = 0;
  $mostab_recall_p2_pv      = $mostab_pv.'recall_p2';		$mostab_recall_p2    = 0;
  $mostab_use2_pv           = $mostab_pv.'use_bpm2';		$mostab_use2         = 0;
  $mostab_ch2_pv            = $mostab_pv.'bpm2_ch';		$mostab_ch2          = $ADC_chs2[0];
  $mostab_inuse2_pv         = $mostab_pv.'bpm2_inuse';		$mostab_inuse2       = 0;
  $mostab_search_mode_pv    = $mostab_pv.'search_p2';		$mostab_search_mode  = 0;
  $mostab_force_search_pv   = $mostab_pv.'force_search';	$mostab_force_search = 0;
  $mostab_search_range_pv   = $mostab_pv.'search_range';	$mostab_search_range = 1;
  $mostab_corr_pv           = $mostab_pv.'dac_corr';
  $lockin_phase_hardware_pv = $mostab_pv.'phase_hw';
  $lockin_phase_hybrid_pv   = $mostab_pv.'phase_hybrid';

  if ($beamline eq '23i') {
     ### SS: 2021/10: why the piezo controller is not installed at IDB?
     $piezo_servoState_pv = $beamline.':MO:PIE709:PZ:servoState';		#PI E-709 piezo controller: for switching to close loop
  }

  $lockin_use_pv            = $beamline.':feedback:uselockin';			#0=software,1=hybrid,2=hardware
  $lockin_use               = 0;						#start with software
  $lockin_asyn_pv           = $beamline.':feedback:serialLIA';			#.TFIL,.TMOD,.AOUT,.IINP for controlling LIA
  $lockin_ADC1_pv           = $beamline.':feedback:LIA:ADC1';			#.TINP field; used for the voltage into the piezo (see $use_LIA_DAC)
  $lockin_xOUT_pv           = $beamline.':feedback:LIA:xout';			#.TINP field; DAC step in hybrid mode

  if    ($beamline eq '23i') {$asynTraceDir = 'bl1upper:/home/gmca/epics_synApps/synApps/dcm_d/iocBoot/iocdcm/';}
  elsif ($beamline eq '23o') {$asynTraceDir = 'bl2upper:/home/gmca/epics_synApps/synApps/dcm_b/iocBoot/iocdcm/';}
  elsif ($beamline eq '23b') {$asynTraceDir = 'bl3upper:/home/gmca/epics_synApps/synApps/dcm_bm/iocBoot/iocdcm/';}
  else                       {$asynTraceDir = '';}
  $asynTraceFile = 'atrace';

  $nmessage = 5;								#number of message PVs (message lines)
  for ($i=0; $i<$nmessage; $i++) {
     $message_pv[$i] = $mostab_pv.'message'.($i+1);
     $message[$i]    = '';
  }

  ### Beamline PVs
  $E_pv             = $beamline.':MO:E:ActPos';					# PV names associated with variables of interest
  $E_mov_pv         = $beamline.':MO:En:InPos';
  if (!defined $use_LIA_DAC || $use_LIA_DAC == 0) {
     $DAC_piezo_pv  = $beamline.':DAC1_'.$DAC_ch;				# 12 bit DAC
#    $DAC_piezo_pv  = $beamline.':DAC2_'.$DAC_ch;				# 16 bit DAC
     $use_LIA_DAC  = 0;
  } else {
     $DAC_piezo_pv  = $beamline.':serialLIA:AUX2';				# LIA analog output DAC
     $use_LIA_DAC  = 1;
  }
  &putPV($mostab_pv.'outopt',$use_LIA_DAC);

  $ADC_bpm1_pv      = $beamline.':ADC_'.$ADC_ch1;
  $ADC_intrate1_pv  = $ADC_bpm1_pv.'.SCAN';
  if ($have_CAEN) {
#    $CAEN_sum_pv      = $beamline.':AH501:SumAll:MeanValue_RBV';
     $CAEN_sum_pv      = $beamline.':AH501:SumX:MeanValue_RBV';
     $CAEN_avertime_pv = $beamline.':AH501:AveragingTime';
     $CAEN_range_pv    = $beamline.':AH501:Range';
     $CAEN_acquire_pv  = $beamline.':AH501:Acquire';
### SS: Bob & I empirically found factor of 2 for SumX; for 4 it should probably be 4:
     if    ($CAEN_sum_pv =~ /SumX/)   {$CAEN_nch = 2;}
     elsif ($CAEN_sum_pv =~ /SumY/)   {$CAEN_nch = 2;}
     elsif ($CAEN_sum_pv =~ /SumAll/) {$CAEN_nch = 4;}
     else                             {$CAEN_nch = 1;}
### SS: replace BPM reading via ADC with CAEN:
     $ADC_bpm1_pv = $CAEN_sum_pv;
  }
  $DAC_rdbk_pv      = $beamline.':ADC_'.$ADC_DACcheck;
  $DAC_rdbkrate_pv  = $DAC_rdbk_pv.'.SCAN';
  $current_pv       = 'S:SRcurrentAI.VAL';
  $permit_pv        = 'ACIS:ShutterPermit';
  $a_status_pv      = '23ID:FES:status';
  $c_status_pv      = '23i:IBSH:status';
  $fes_open_pv      = '23ID:V2000:0';
  $fes_close_pv     = '23ID:V2000:1';
  $last_mbsh        = 1;

  $p2_assy_pv       = $beamline.':MO:Tn2';
  $p2_axis_pv       = $beamline.':MO:P2';
  $p2_mtr_pv        = $beamline.':MO:mp2';
  $p2_recall_pv     = $beamline.':MO:P2:MemPosRcl.PROC';

  $FPE_start_pv     = $beamline.':FPE:start';
  $FPE_exposure_pv  = $beamline.':FPE:exposureTime';
  $BI_fdbk_pv       = $beamline.':bi:waitForFeedback';

  ### Mono beam shutter
  if    ($beamline eq '23i') {$mbsh_pv = '23i:MBSHI:status';}
  elsif ($beamline eq '23o') {$mbsh_pv = '23o:MBSHO:status';}
  else                       {undef $mbsh_pv;}

  ### Initial lockin mode:
  $lockin_prev     = -1;							#unknown

  $ezcatimeout_default = Pezca::GetTimeout();					#this is 50ms by default

  ### Get ID offsets and harmonic switching energies from the SQL/Config table
  $status = &getIDparams(\$enmono13,\@harmEn,\@harmNum,
                         \@Elower,\@Olower,\@Eupper,\@Oupper);

  ### Query mono Bragg reflection at start:
  $hkl     = &getMonoHKL();
  $hkl_old = $hkl;

  $mode    = $mSTARTING;
}

#==========================================================================
sub log_func_entry ($) {
    my ($subscript);
    $subscript = shift(@_);
    if (defined $mostab_show_flow && $mostab_show_flow) {&log_record('Entering '.$subscript);}
}

#==========================================================================
sub log_record ($) {
    my ($msg, @lines, $line);
    $msg = shift(@_);
    $current_time = scalar localtime;
    @lines = split(/\n/, $msg);
    foreach $line (@lines) {print $LOGLUN $current_time.' '.$line."\n";}
}
 
#==========================================================================
sub open_logfile () {
 ### Open file with timestamp included in filename

  my ($subscript, $sbscrpt, @ct, $sec, $min, $hour, $mday, $mon, $year);

  $sbscrpt = 'open_logfile';
  $subscript = $script.'/'.$sbscrpt;
  &log_func_entry($subscript);

  ### This is ($sec,$min,$hour,$mday,$mon,$year$,$wday,$yday,$isdst):
  @ct = localtime(time);
  ($sec,$min,$hour,$mday,$mon,$year) = ($ct[0],$ct[1],$ct[2],$ct[3],$ct[4]+1,$ct[5]+1900);
  $logfile = sprintf '%s%s_%04d.%02d.%02d_%02d.%02d.%02d.log',
                     $logdir,$logfile,$year,$mon,$mday,$hour,$min,$sec;
  $logfile =~ s/ /0/g;
  print STDOUT "\n".$script.': opening logfile: '.$logfile."\n\n";
  $filehandle = open (DAT,'> '.$logfile);				# open DAT file for writing
  if (!defined $filehandle) {
     &post_message('*** Could not open logfile: '.$logfile, 'error');
  } else {
     $LOGLUN=\*DAT;                                                	# redirect from STDERR to LOG
     select DAT; $|=1;							# set unbuffered output
  }
  $asynTraceFile = sprintf $asynTraceFile.'_%04d%02d%02d',$year,$mon,$mday;
  &post_message('Setting Asyn Record to use trace file: '.$asynTraceDir.$asynTraceFile);
  &putPV($lockin_asyn_pv.'.TFIL',$asynTraceFile,'string');
  &post_message('AsynDir='.$asynTraceDir);
  &post_message('AsynFile='.$asynTraceFile);
}

#==========================================================================
sub init_mstb () {

  my ($subscript, $sbscrpt, $wait_time, $msg);

  $sbscrpt = 'init_mstb';
  $subscript = $script.'/'.$sbscrpt;
  &log_func_entry($subscript);

  ### Write logfile header
  $current_time = scalar localtime;
  print $LOGLUN "\n".$current_time.'     '.$script.': intensity feedback started';
  if (defined $filehandle) {print $LOGLUN ', output logfile='.$logfile."\n\n";}
  else                     {print $LOGLUN ', output log to STDERR.'."\n\n";}

  ### Check for existing scripts: we set status=0, but the other instance,
  ### if present, should restore it to non-zero.
  &post_message('New script started.');
  &post_message('Checking for existing script.');
  &putPV($mostab_status_pv,0,'double');					#0=exiting, 1=active, 2=doing.operation, 3=waiting, 4=beam.lost
  $wait_time = $pause_time * 3;
  select(undef,undef,undef,$wait_time);
  $mostab_status = &getPV($mostab_status_pv,'double');
  if ($mostab_status != 0) {
     &post_message('Previous script detected.');
     &post_message('Halting new script.', 'warn');
     &signal_handler('DUPLICATE');
  }
  &post_message('No previous scripts detected.');
# &putPV($mostab_status_pv.'.DESC','soft/hybrid/HW-pause, '.$script,'string');
  &putPV($mostab_status_pv.'.DESC','version-'.$version,'string');
  ### Initialize bpm2_inuse pv
  $mostab_use2 = &getPV($mostab_use2_pv,'double');
  $last_use2 = $mostab_use2;
  if ($mostab_use2 == 1) {&putPV($mostab_inuse2_pv,1,'double');}
  else                   {&putPV($mostab_inuse2_pv,0,'double');}

  ### Which channel does the user wish to use for bpm2?
  ### If not on acceptable list, use default.
  $mostab_ch2  = &getPV($mostab_ch2_pv,'double');
  $ch_ok = 0;
  foreach (@ADC_chs2) {
     if ($_ == $mostab_ch2) {$ch_ok = 1; last;}
  }
  if (!$ch_ok) {
     $mostab_ch2 = $ADC_chs2[0];
     &putPV($mostab_ch2_pv,$mostab_ch2,'double');
     &post_message('Bad bpm2 channel. Using default.', 'warn');
  }
  $last_ch2 = $mostab_ch2;
  $ADC_bpm2_pv     = $beamline.':ADC_'.$mostab_ch2.'.VAL';
  $ADC_intrate2_pv = $beamline.':ADC_'.$mostab_ch2.'.SCAN';
  &putPV($ADC_intrate2_pv, $ADC_rate_idx,'double');

  $DAC_piezo_DRVL = &getPV($DAC_piezo_pv.'.DRVL','double');
  $DAC_piezo_DRVH = &getPV($DAC_piezo_pv.'.DRVH','double');
  if ($DAC_piezo_DRVL == 0 && $DAC_piezo_DRVH == 0) {
     $msg = 'No DRVL/DRVH set for DAC='.$DAC_piezo_pv;
     &post_message($msg, 'warn');
     &log_record('!!! '.$msg.', setting DAC limits [0-10]');
     $DAC_piezo_DRVL = 0;
     $DAC_piezo_DRVH = 10;
     &putPV($DAC_piezo_pv.'.DRVL',$DAC_piezo_DRVL,'double');
     &putPV($DAC_piezo_pv.'.DRVH',$DAC_piezo_DRVH,'double');
  }
  &putPV($DAC_piezo_pv.'.LOPR',$DAC_piezo_DRVL,'double');
  &putPV($DAC_piezo_pv.'.HOPR',$DAC_piezo_DRVH,'double');
  &putPV($DAC_piezo_pv.'.EGUL',$DAC_piezo_DRVL,'double');
  &putPV($DAC_piezo_pv.'.EGUF',$DAC_piezo_DRVH,'double');
#
  $DAC_piezo_CNTR = ($DAC_piezo_DRVL + $DAC_piezo_DRVH) / 2;
  $DACmin         = $DAC_piezo_DRVL + $DAC_offlimit;
  $DACmax         = $DAC_piezo_DRVH - $DAC_offlimit;
  $DAC_piezo_last = $DAC_piezo_CNTR;

  ### Enable write/read on LIA asyn:
  &putPV($lockin_asyn_pv.'.TMOD','Write/Read','string');

  ### Initialize feedback
  &post_message('Initializing feedback.');
  # $mode = undefined;
  &status_write($mACTFDBK1);
  $mostab_pause = 0;
  &putPV($mostab_pause_pv,$mostab_pause,'double');
  $mostab_stop = 1;
  &putPV($mostab_stop_pv,$mostab_stop,'double');
  $DAC_check_err = &DAC_check(0);
  if ($DAC_check_err) {$mode = $mPAUSED; return;}
  $mostab_lock = 0;
  &putPV($mostab_lock_pv,$mostab_lock,'double');
  select(undef,undef,undef,0.5);
  $mode = $mINTCHK2;
}

#==========================================================================
sub set_monitors () {

  my ($subscript, $sbscrpt, $status);

  $sbscrpt = 'set_monitors';
  $subscript = $script.'/'.$sbscrpt;
  &log_func_entry($subscript);

  $status = 0;

  $status += &addCaMonitor($id_pv,		     'double', $LOGLUN);
  $status += &addCaMonitor($E_pv,		     'double', $LOGLUN);
  $status += &addCaMonitor($E_mov_pv,		     'double', $LOGLUN);
  $status += &addCaMonitor($mostab_lock_pv,	     'double', $LOGLUN);
  $status += &addCaMonitor($mostab_stop_pv,	     'double', $LOGLUN);
  $status += &addCaMonitor($mostab_pause_pv,	     'double', $LOGLUN);
  $status += &addCaMonitor($mostab_status_pv,	     'double', $LOGLUN);
  $status += &addCaMonitor($mostab_use2_pv,	     'double', $LOGLUN);
  $status += &addCaMonitor($mostab_opt_pv,	     'double', $LOGLUN);
  $status += &addCaMonitor($mostab_recall_p2_pv,     'double', $LOGLUN);
  $status += &addCaMonitor($mostab_ch2_pv,	     'double', $LOGLUN);
  $status += &addCaMonitor($mostab_inuse2_pv,	     'double', $LOGLUN);
  $status += &addCaMonitor($mostab_search_mode_pv,   'double', $LOGLUN);
  $status += &addCaMonitor($mostab_force_search_pv,  'double', $LOGLUN);
  $status += &addCaMonitor($mostab_search_range_pv,  'double', $LOGLUN);
  $status += &addCaMonitor($lockin_phase_hardware_pv,'double', $LOGLUN);
  $status += &addCaMonitor($lockin_phase_hybrid_pv,  'double', $LOGLUN);

  $status += &addCaMonitor($lockin_use_pv,	     'double', $LOGLUN);
  $status += &addCaMonitor($current_pv,		     'double', $LOGLUN);
  $status += &addCaMonitor($permit_pv,		     'double', $LOGLUN);
  $status += &addCaMonitor($a_status_pv,	     'double', $LOGLUN);
  $status += &addCaMonitor($c_status_pv,	     'double', $LOGLUN);
  $status += &addCaMonitor($mbsh_pv,		     'double', $LOGLUN);
  $status += &addCaMonitor($p2_assy_pv.':InPos',     'double', $LOGLUN);
  $status += &addCaMonitor($p2_assy_pv.':RunPrg',    'double', $LOGLUN);
  $status += &addCaMonitor($p2_axis_pv.':AmpEna',    'double', $LOGLUN);
  $status += &addCaMonitor($FPE_start_pv,	     'double', $LOGLUN);
  $status += &addCaMonitor($FPE_exposure_pv,	     'double', $LOGLUN);
  $status += &addCaMonitor($BI_fdbk_pv,		     'double', $LOGLUN);
  ### Added by Sergey 2013/01/29 (should make Mostab faster)
  $status += &addCaMonitor($ADC_bpm1_pv,	     'double', $LOGLUN);
  $status += &addCaMonitor($ADC_bpm2_pv,	     'double', $LOGLUN);
  $status += &addCaMonitor($DAC_piezo_pv,	     'double', $LOGLUN);
  $status += &addCaMonitor($DAC_rdbk_pv,	     'double', $LOGLUN);
  if ($have_CAEN) {
     $status += &addCaMonitor($CAEN_range_pv,	     'string', $LOGLUN);	#amplification range string
     $status += &addCaMonitor($CAEN_avertime_pv,     'double', $LOGLUN);
  }
  if ($have_joerger) {
     $status += &addCaMonitor($scanrec.'.BUSY',	     'double', $LOGLUN);
     $status += &addCaMonitor($scanrec.'.CPT',	     'double', $LOGLUN);
  }
  if ($status != 0) {
     &log_record('CA error on establishing monitor(s) for '.$status.' PV(s).'."\n"
                .'Please investigate.'."\n"
                .'Feedback stopping.'."\n");
     &signal_handler($sbscrpt.' error');
  }
}

#==========================================================================
sub get_LIA_phases () {
  $lockin_phase_hardware = &getPV($lockin_phase_hardware_pv,'double');
  $lockin_phase_hybrid   = &getPV($lockin_phase_hybrid_pv,'double');
}

#==========================================================================
sub set_scanrec () {

  my ($subscript, $sbscrpt);

  $sbscrpt = 'set_scanrec';
  $subscript = $script.'/'.$sbscrpt;
  &log_func_entry($subscript);

  $scanrec_max_pts = &SRmaxPts($scanrec);
  &clearSR ($scanrec);
### $scanMode: 0=linear, 1=table, 2=fly
### $afterScan: 0=stay, 1=startpos, 2=priorpos 3=peakpos 4=valleypos 5=+edgepos 6=-edgepos
### $absRel: 0=abs, 1=rel
# &setupSRmotorsMore ($scanrec,     $numPts,  $scanMode, $settleTime, $afterScan, $numMotors,  \@actPosPV,              \@rqsPosPV,     \@begin,\@end, $absRel);
  &setupSRmotorsMore ($scanrec, $scanrec_max_pts, 2,          0.0,        0,           1, [$p2_axis_pv.':ActPos'], [$p2_axis_pv.':RqsPos'], [0],   [1],   0);
# &setupSRdetectors_v2 ($scanrec, $scaler, $ninputs, \@inputs,   $triggerPV, $triggerVAL, $settleTime);
  &setupSRdetectors_v2 ($scanrec, $scaler,    1,  [$joerger_bpm1], '',           1,     $joerger_time);

  $p2_max_speed = &getAxisHiSpeed ($p2_assy_pv, $p2_axis_pv);
}

#==========================================================================
sub start_scaler () {

  my ($subscript, $sbscrpt);

  $sbscrpt = 'start_scaler';
  $subscript = $script.'/'.$sbscrpt;
  &log_func_entry($subscript);
  $scaler_mode = &getStruckMode($struck,'string');
  if ($scaler_mode eq 'MCS') {&abortStruck ($struck);}

  &setJoerger ($scaler, $joerger_time);			#sets dwell for both auto & manual count
  &Joerger_autoStart ($scaler);				#turn on auto-count mode
}

#==========================================================================
sub stop_scaler () {

  my ($subscript, $sbscrpt);

  $sbscrpt = 'stop_scaler';
  $subscript = $script.'/'.$sbscrpt;
  &log_func_entry($subscript);

  &Joerger_autoStop ($scaler);
  if (defined $scaler_mode && $scaler_mode eq 'MCS') {&runStruck ($struck);}
}

#==========================================================================
sub medm_check ($) {                                                    #returns mode requested on MEDM

### $mode may be changed here. Although it is "returned", this is
### a global variable; so strictly speaking returning is not needed.

  my ($subscript, $sbscrpt, $present_mode, $lockin_phase);

  $sbscrpt = 'medm_check';
  $subscript = $script.'/'.$sbscrpt;
  &log_func_entry($subscript);

  $present_mode = shift(@_);

  $E_mov               = &getPV($E_mov_pv,'double');
  $mostab_status       = &getPV($mostab_status_pv,'double');		#0=exiting, 1=active, 2=doing.operation, 3=waiting, 4=lost
  $mostab_stop         = &getPV($mostab_stop_pv,'double');
  $mostab_pause	       = &getPV($mostab_pause_pv,'double');
  $mostab_force_search = &getPV($mostab_force_search_pv,'double');
  $mostab_opt          = &getPV($mostab_opt_pv,'double');		#this is shift P2 button (optimize DAC to center)
  $mostab_recall_p2    = &getPV($mostab_recall_p2_pv,'double');		#this is recalll P2 from memory button

  &putPV($lockin_asyn_pv.'.TMOD','Write','string');
  $lockin_phase = &getPV($lockin_phase_hybrid_pv,'double');
  if ($lockin_phase != $lockin_phase_hybrid) {
     $lockin_phase_hybrid = $lockin_phase;
     &post_message('Setting LIA Hybrid phase='.$lockin_phase_hybrid.' ('.$sbscrpt.')');
     &putPV($lockin_asyn_pv.'.AOUT','PHAS '.$lockin_phase_hybrid,'string','oldca');	#set hybrid-mode lockin phase
     select(undef,undef,undef,0.4);
  }

  if ($mostab_opt == 1) {						#Shift P2 to center DAC is requested
     $mostab_opt = 0;
     &putPV($mostab_opt_pv,$mostab_opt,'double');
#    &post_message('Pause. Then optimize.', 'warn');
     &post_message('Shifting P2 to center DAC @ 5V', 'warn');
     $mode = $mOPTIMP2;							# mode=9 <= optimize P2
  }
  elsif ($mostab_force_search == 1) {					#Searching with P2 or DAC is requested
     $mostab_force_search = 0;
     &putPV($mostab_force_search_pv,$mostab_force_search,'double');
     &post_message('Forcing beam search ('.$sbscrpt.')', 'warn');
     $mode = $mSEARCH;							# mode=4 <= beam search
  }
  elsif ($mostab_recall_p2 == 1) {
     $mostab_recall_p2 = 0;
     &putPV($mostab_recall_p2_pv,$mostab_recall_p2,'double');
     &post_message('Shifting P2 to MemPos', 'warn');
     $mode = $mRECALLP2;						# mode=9 <= optimize P2
  }
  if ($mostab_stop == 0)   {&signal_handler($sbscrpt.' mostab_stop=0');};
  if ($mostab_pause == 1)  {$mode = $mPAUSED;};
  if ($mostab_status == 0) {&status_write($present_mode);};		# restore status to confirm the script is alive (multiple instances check)
  if ($E_mov == 0)         {$wait_dcm = 1;}
  else                     {$wait_dcm = 0;}
  return $mode;
}


#==========================================================================
sub feedback () {

### This is a continuous feedback function where the script is supposed
### to sit once the beam is locked.

  my ($subscript, $sbscrpt, $DAC_piezo_new);
  my ($fwhm_x_E, $DAC_step, $DAC_step_max, $msg, $factor, $n);
  my ($ADC_bpm1_m1, $ADC_bpm1_m2);
  my ($ADC_bpm2_m1, $ADC_bpm2_m2);
  my ($volt_step, $bpm_choice);
  my ($hybrid_step_restriction_mode);						#0=off, 1=intensity/peak factor, 2=max=soft step, 3=intensity/small factor
  my ($soft_step_restriction_mode);						#0=off, 1=intensity/peak factor, 2=intensity/small factor
  my ($factor_min, $factor_max, $peak_steps);					#this is for $soft_step_restriction_mode and $hybrid_step_restriction_mode

  $sbscrpt = 'feedback';
  $subscript = $script.'/'.$sbscrpt;
  &log_func_entry($subscript);

### -------------------------------
  $hybrid_step_restriction_mode = 0;						#0=off, 1=old way (Sioan), 2=Sergey(2021/10), 3=Sergey(2021/10)
# $hybrid_step_restriction_mode = 0;						#0=off, 1=old way (Sioan), 2=Sergey(2021/10), 3=Sergey(2021/10)
#
# $soft_step_restriction_mode   = 3;						#0=off, 1=intensity/peak factor, 2=intensity/small factor, 3=account for ADC slop
  $soft_step_restriction_mode   = 0;						#0=off, 1=intensity/peak factor, 2=intensity/small factor, 3=account for ADC slop
### -------------------------------
  $factor_min = 0.25;								#this is for $soft_step_restriction_mode
  $factor_max = 3;								#this is for $hybrid_step_restriction_mode and $hybrid_step restriction itself
  $peak_steps = 1000;								#this is for $soft_step_restriction_mode

  &post_message('Active feedback.');
  &status_write($mode);
  $mode = &medm_check($mode);							# 2
  if ($mode == $mPAUSED) {return;}

# $ADC_bpm1_m1 = &getPV($ADC_bpm1_pv,'double');
# $ADC_bpm2_m1 = &getPV($ADC_bpm2_pv,'double');
  ($ADC_bpm1_m1,$ADC_bpm2_m1) = &measure_intensity($int_samples);
  if ($have_CAEN) {
     $CAEN_range = &getPV($CAEN_range_pv,'string');				#amplification range string
     if ($CAEN_range !~ /uA$/) {
        $msg = '!!! CAEN is in unexpected mode = '.$CAEN_range;
        &post_message($msg);
        &tellBluIce ($scr.': '.$msg, 'warn');
     }
  }
  ($ADC_bpm1_m2,$ADC_bpm2_m2) = ($ADC_bpm1_m1,$ADC_bpm2_m1);
  $factor = 1;

  $E = &getPV($E_pv,'double');

  $msg = sprintf 'E = %.2f',$E;
  &post_message($msg);
  $msg = sprintf 'Int1 = %.2f',$ADC_bpm1_m1;
  if ($have_CAEN) {$msg .= ' (CAEN quadEM)';}
  &post_message($msg);
  $msg = sprintf 'Int2 = %.2f',$ADC_bpm2_m1;
  &post_message($msg);
  $init_E = $E;

  $fwhm_x_E = &calc_fwhm_x_E();
  $volt_step = ($fwhm_x_E/$E) * ($pitch_sig/$fwhm2sig) * $urad2volt;
  $multiplier = abs($multiplier);
  $DAC_step_max = abs($factor_max * $volt_step * $multiplier);
  if ($trace) {
     $msg = sprintf $sbscrpt.': hkl=%s fwhm_x_E=%.1f volt_step=%.4f multiplier=%.1f volt_step_max=%.4f'."\n",
                                 $hkl,  $fwhm_x_E,     $volt_step,   $multiplier,   $DAC_step_max;
     &log_record($msg);
  }

  &status_write($mINTCHK2);

# $lockin_prev = -1;
  $n = 0;
  while ($mode == $mACTFDBK1) {	#=======================================================# mode=1 <= active feedback
     $n++;
     $mostab_lock = &getPV($mostab_lock_pv,'double');
     if ($mostab_lock == 0) {
        $mode = $mACTFDBK2;							# mode=8 <= active feedback; large step size
        &post_message('Please wait for beam lock - '.$sbscrpt);
        return;
     }
     &kill_p2();
     $DAC_piezo = &getPV($DAC_piezo_pv,'double');

     ### Measure intensity m1 --------------------------------------
     ($ADC_bpm1_m1,$ADC_bpm2_m1) = &measure_intensity($int_samples);
#    if ($trace && $n < $trace_max) {$msg = sprintf $sbscrpt.': ADC_bpm1=%.3f ADC_bpm2=%.3f'."\n",$ADC_bpm1_m1,$ADC_bpm2_m1; &log_record($msg);}

     $mode = &medm_check($mINTCHK2);
     if ($mode == $mPAUSED) {return;}

     ### Move to new DAC position ----------------------------------

     $lockin_use = &getPV($lockin_use_pv);
     if ($lockin_use > 1) {$lockin_use=1; &putPV($lockin_use_pv,$lockin_use,'double');}

     if ($lockin_use == 1 && $mostab_lock == 1) {	#-------------------------------# Hybrid feedback
        if ($lockin_use != $lockin_prev) {
           if ($lockin_prev != -1) {&post_message('Switching to hybrid feedback');}
           $lockin_prev = $lockin_use;
           $n = 0;
        }
        $value    = &get_lockin_DAC_step($lockin_xOUT_pv);
        $DCM_VSUM = &getPV($ADC_bpm1_pv,'double');
        if    ($DCM_VSUM < 0.)            {$DCM_VSUM = 0.;}
        elsif ($DCM_VSUM > $DCM_VSUM_max) {$DCM_VSUM_max = $DCM_VSUM;}
        ### Normalization terms to prevent oscillations near harmonic ID-B 1-3 switch.
        if ($hybrid_step_restriction_mode == 1) {
           ### At $DCM_VSUM=0 -> $factor=0.1*$DCM_VSUM_max/0.33 ~ 3/3.3 ~ 1
           ### At $DCM_VSUM=$DCM_VSUM_max -> $factor=0.1
           $factor = $int_ok1 * $DCM_VSUM_max / sqrt($DCM_VSUM**2+$int_ok1**2);		# $DCM_VSUM_max=~ 1 to 4
        }
        elsif ($hybrid_step_restriction_mode == 2) {
           ### SS: 2021/10: let us try is another approach: limit 
           ### the maximum hybrid step to the soft feedback step:
           if (abs($value) > $DAC_step_max) {$factor = $DAC_step_max/abs($value);}
           else                             {$factor = 1;}
        }
        elsif ($hybrid_step_restriction_mode == 3) {
           ### SS: 2021/10: At $DCM_VSUM=0 -> $factor=1
           ### At $DCM_VSUM=$DCM_VSUM_max  -> $factor=sqrt(1/((3/0.1)+1))=0.18	# $DCM_VSUM_max=~ 1 to 4
           $factor = sqrt($int_ok1/($DCM_VSUM+$int_ok1));			# $int_ok1 =~ 0.1
        }
        else {
           ### No normalization (assuming $hybrid_step_restriction_mode=0), basically same as in the HW mode
           $factor = 1;
        }
        if ($factor > $factor_max)  {
           $msg = sprintf $subscript.': Correction: DAC_step_calc(lockin)=%.4f, factor=%.3f mode=%s - restricting factor to %.3f',$value,$factor,$mname[$mode],$factor_max;
           &log_record($msg);
           $factor = $factor_max;
        }
        $DAC_step = $value *  $factor;
        if (abs($DAC_step) > $DAC_step_max) {
           $msg = sprintf $subscript.': Correction: DAC_step(lockin)=%.4f mode=%s - restricting step to %.4f',$DAC_step,$mname[$mode],$DAC_step_max*abs($DAC_step)/$DAC_step;
           &log_record($msg);
           $DAC_step = $DAC_step_max * abs($DAC_step)/$DAC_step;
        }
        $DAC_piezo_new = $DAC_piezo + $DAC_step;
        if ($trace && $n < $trace_max) {$msg = sprintf $sbscrpt.'(hybrid): DAC_step=%.4f factor=%.3f'."\n",$DAC_step,$factor; &log_record($msg);}

        if ($DAC_piezo_new > $DAC_piezo_DRVH) {$DAC_piezo_new = $DAC_piezo_DRVH;}
        if ($DAC_piezo_new < $DAC_piezo_DRVL) {$DAC_piezo_new = $DAC_piezo_DRVL;}
        &kill_p2();
        if ($trace) {
           $msg = sprintf $subscript.': Correction: DAC_step_calc(lockin)=%.4f factor=%.3f mode=%s',$value,$factor,$mname[$mode];
           &log_record($msg);
        }
        ###  writes the pv value
        &putPV($DAC_piezo_pv,$DAC_piezo_new,'double');
     } 
     else {	#-----------------------------------------------------------------------# software feedback
        $lockin_use = 0;
        if ($lockin_use != $lockin_prev){
           if ($lockin_prev != -1) {&post_message('Switching to software feedback');}
           $lockin_prev = $lockin_use;
           $n = 0;
        }
        $DCM_VSUM = &getPV($ADC_bpm1_pv,'double');
        if    ($DCM_VSUM < 0.)            {$DCM_VSUM = 0.;}
        elsif ($DCM_VSUM > $DCM_VSUM_max) {$DCM_VSUM_max = $DCM_VSUM;}
        $value = $volt_step;
        if ($soft_step_restriction_mode == 1) {
           ### At $DCM_VSUM=0 -> $factor=0.1*$DCM_VSUM_max/0.33 ~ 3/3.3 ~ 1
           ### At $DCM_VSUM=$DCM_VSUM_max -> $factor=0.1
           $factor *= $int_ok1 * $DCM_VSUM_max / sqrt($DCM_VSUM**2+$int_ok1**2);	# $DCM_VSUM_max=~ 1 to 4
#          printf STDOUT $sbscrpt.'(soft): factor=%.3f'."\n",$factor;
        }
        elsif ($soft_step_restriction_mode == 2) {
           ### SS: 2021/10: At $DCM_VSUM=0 -> $factor=1
           ### At $DCM_VSUM=$DCM_VSUM_max  -> $factor=sqrt(1/((3/0.1)+1))=0.18		# $DCM_VSUM_max=~ 1 to 4
           $factor = sqrt($int_ok1/($DCM_VSUM+$int_ok1));				# $int_ok1 =~ 0.1
#          printf STDOUT $sbscrpt.'(soft): factor=%.3f'."\n",$factor;
        }
        elsif ($soft_step_restriction_mode == 3) {
           ### SS: 2021/11: At $DCM_VSUM=0 -> $factor=1
           ### At $DCM_VSUM=$DCM_VSUM_max  -> $factor=sqrt(1/((3/0.1)+1))=0.18		# $DCM_VSUM_max=~ 1 to 4
           $factor *= abs($peak_steps*($ADC_bpm1_m1-$ADC_bpm1_m2)/($DCM_VSUM_max-$int_ok1));	# $int_ok1 =~ 0.1
#          printf STDOUT $sbscrpt.'(soft): factor_raw=%.3f  ',$factor;
           if    ($factor < $factor_min) {$factor = $factor_min;}
           elsif ($factor >   1)         {$factor = 1;}
#          printf STDOUT $sbscrpt.'(soft): factor_corr=%.3f'."\n",$factor;
        }
        else {
           ### No normalization (assuming $software_step_restriction_mode=0), basically same as in the HW mode
           $factor = 1;
        }
        $DAC_step = $value * $factor * $multiplier;
        $DAC_piezo_new = $DAC_piezo + $DAC_step;
        if ($trace && $n < $trace_max) {$msg = sprintf $sbscrpt.'(soft): DAC_step=%.4f multiplier=%.3f factor=%.4f'."\n",$DAC_step,$multiplier,$factor; &log_record($msg);}

        if ($DAC_piezo_new > $DAC_piezo_DRVH) {$DAC_piezo_new = $DAC_piezo_DRVH;}
        if ($DAC_piezo_new < $DAC_piezo_DRVL) {$DAC_piezo_new = $DAC_piezo_DRVL;}
        &kill_p2();
        if ($trace) {
           $msg = sprintf $subscript.': Correction: DAC_step_calc(soft)=%.4f factor=%.4f DAC_step=%.4f mode=%s',$volt_step,$factor,$DAC_step,$mname[$mode];
           &log_record($msg);
        }
        ###  writes the pv value
        &putPV($DAC_piezo_pv,$DAC_piezo_new,'double');
     } #--------------------------------------------------------------------------------# end of feedback choices
     &putPV($mostab_corr_pv,$DAC_step,'double');

     select(undef,undef,undef,0.1);
     &kill_p2();
     $mode = &medm_check($mINTCHK2);
     if ($mode == $mPAUSED) {return;}
     $DAC_check_err = &DAC_check(0);
     if ($DAC_check_err) {$mode = $mPAUSED; return;}

     ### Measure intensity m2 --------------------------------------
     ($ADC_bpm1_m2,$ADC_bpm2_m2) = &measure_intensity($int_samples);

     $mode = &medm_check($mINTCHK2);
     if ($mode == $mPAUSED) {return;}

     ### Determine which bpm to use and compare m1 and m2 ----------
     ###    0 - use bpm1, by choice
     ###    1 - use bpm1, because bpm2 isn't ready
     ###    2 - use bpm2, because all checks satisfied
     ###    3 - bpm2 recently available, optimize now on bpm2
     ###    4 - both signals bad; lost beam
     $bpm_choice = &bpm2();
     if ($bpm_choice == 0) {if ($ADC_bpm1_m2 < $ADC_bpm1_m1) {$multiplier *= -1;}}
     if ($bpm_choice == 1) {if ($ADC_bpm1_m2 < $ADC_bpm1_m1) {$multiplier *= -1;}}
     if ($bpm_choice == 2) {if ($ADC_bpm2_m2 < $ADC_bpm2_m1) {$multiplier *= -1;}}
     if ($bpm_choice == 3) {
        &post_message('BPM2 became available, optimise on it.');
        $mode = $mACTFDBK2; 						# mode=8 <= active feedback; large step size
        return;
     }
     if ($bpm_choice == 4) {
        &post_message('Both BPM signals are low, lost beam.');
        $mode = $mCOMPCHK; 
        return;								# mode=3 <= component status check
     }

     if ($trace && $n < $trace_max) {
#    if (1) {
        if ($bpm_choice < 2) {
           $msg = sprintf $sbscrpt.': BPM1_old=%.4f BPM1_new=%.4f BPM1_change=%.5f%s'
                                  .' DAC_step=%.4f multiplier=%.3f factor=%.4f'."\n",
                                  $ADC_bpm1_m1, $ADC_bpm1_m2,
                                  200*($ADC_bpm1_m2-$ADC_bpm1_m1)/($ADC_bpm1_m2+$ADC_bpm1_m1),
                                  '%', $DAC_step, $multiplier, $factor;
        } else {
           $msg = sprintf $sbscrpt.': BPM2_old=%.4f BPM2_new=%.4f BPM2_change=%.5f%s'
                                  .' DAC_step=%.4f multiplier=%.3f factor=%.4f'."\n",
                                  $ADC_bpm2_m1, $ADC_bpm2_m2,
                                  200*($ADC_bpm2_m2-$ADC_bpm2_m1)/($ADC_bpm2_m2+$ADC_bpm2_m1),
                                  '%', $DAC_step, $multiplier, $factor;
        }
        &log_record($msg);
     }   

     ### -----------------------------------------------------------
     ### Check for DAC at limit:
     if ($use_LIA_DAC) {$lockin_ADC1 = &get_lockin_ADC($lockin_ADC1_pv);}		#if using LIA DAC to drive piezo
     else              {$lockin_ADC1 = $DAC_piezo;}
     if    ($DAC_piezo < $DACmin || $lockin_ADC1 < $DACmin) {		#if non-HW mode & using LIA DAC to drive piezo
        &post_message('WARNING: Beam @ neg V limit.', 'warn');
        $msg = sprintf 'DAC_piezo=%.2f (low=%.2f)',$DAC_piezo,$DACmin;
        &post_message($msg);
        if ($use_LIA_DAC) {
           $msg = sprintf 'lockin_ADC1=%.2f (low=%.2f)',$lockin_ADC1,$DACmin;
           &post_message($msg);
        }
        if ($p2_autoshift == 1) {
           &post_message('Automatically shifting P2.', 'warn');
           $mode = $mOPTIMP2;						# mode=9 <= optimize P2
           return;
        } else {
           &post_message('Shift P2.', 'warn');
           $mode = $mPAUSED;						# mode=5 = paused
           return;
        }
     }
     elsif ($DAC_piezo > $DACmax || $lockin_ADC1 > $DACmax) {		#if non-HW mode & using LIA DAC to drive piezo
        &post_message('WARNING: Beam @ pos V limit.', 'warn');
        $msg = sprintf 'DAC_piezo=%.2f (high=%.2f)',$DAC_piezo,$DACmax;
        &post_message($msg);
        if ($use_LIA_DAC) {
           $msg = sprintf 'lockin_ADC1=%.2f (high=%.2f)',$lockin_ADC1,$DACmax;
           &post_message($msg);
        }
        if ($p2_autoshift == 1) {
           &post_message('Automatically shifting P2.', 'warn');
           $mode = $mOPTIMP2;						# mode=9 <= optimize P2
           return;
        } else {
           &post_message('Shift P2.', 'warn');
           $mode = $mPAUSED;
           return;
        }
     }
     $mode = &medm_check($mINTCHK2);					# 3
     if ($mode == $mPAUSED) {return;}
     &kill_p2();
     if ($wait_dcm) {&wait_for_dcm();}
     &kill_p2();
  }	#===============================================================================# end of active feedback loop
}

#==========================================================================
sub fast_feedback () {

### This is a feedback locking phase function. It is supposed to converge
### to the peak by successively decreasing the step and the proceed to to
### &feedback(). The number of iterations is:
###         $n_opt_steps * $reversals = 2 * $n_opt_steps.

  my ($subscript, $sbscrpt, $factor, $msg, $fwhm_x_E);
  my ($DAC_step, $DAC_piezo_new);
  my ($ADC_bpm1_m1, $ADC_bpm1_m2);
  my ($ADC_bpm2_m1, $ADC_bpm2_m2);
  my ($iter, $nrev, $opt_step, $n_opt_steps, $n);
  my ($volt_step, $bpm_choice, $factor_max, $DAC_step_max);

  $sbscrpt = 'fast_feedback';
  $subscript = $script.'/'.$sbscrpt;
  &log_func_entry($subscript);

  &post_message('Locking on peak - '.$sbscrpt);
  &status_write($mode);
  $mode = &medm_check($mode);						# 4
  if ($mode == $mPAUSED) {&post_message('Feedback paused.'); return;}

  $n_opt_steps  = @opt_steps;						#@opt_steps = (20,3,1);
  $multiplier   = abs($multiplier);
  $factor_max   = 3;							#this is for $hybrid_step restriction
  $mostab_lock  = 0;
  &putPV($mostab_lock_pv,$mostab_lock,'double');

  $E = &getPV($E_pv,'double');
  $fwhm_x_E = &calc_fwhm_x_E();
  $volt_step = ($fwhm_x_E/$E) * ($pitch_sig/$fwhm2sig) * $urad2volt;
  if ($trace) {$msg = sprintf $sbscrpt.': hkl=%s fwhm_x_E=%.1f volt_step=%.4f'."\n",$hkl,$fwhm_x_E,$volt_step; &log_record($msg);}

# $lockin_prev = -1;

  $n = 0;
  for ($iter=0; $iter < $n_opt_steps; $iter++) { #======================$n_opt_steps is number of optimization steps, with smaller & smaller sizes.
     $opt_step = $volt_step * $opt_steps[$iter];
     if ($trace && $n < $trace_max) {$msg = sprintf $sbscrpt.': opt_step=%d volt_step=%.4f'."\n",$opt_steps[$iter],$opt_step; &log_record($msg);}
     $mode = &medm_check($mode);
     if ($mode == $mPAUSED) {&post_message('Feedback paused.'); return;}

     $nrev = 0;
     while ($nrev < $reversals) { #= = = = = = = = = = = = = = = = = = =begin loop over $recersals=2 (left, right)
        $n++;
        &kill_p2();
        $mode = &medm_check($mode);
        if ($mode == $mPAUSED) {&post_message('Feedback paused.'); return;}

        ### Measurement 1
        $DAC_piezo = &getPV($DAC_piezo_pv,'double');
        $ADC_bpm1_m1 = &getPV($ADC_bpm1_pv,'double');
        $ADC_bpm2_m1 = &getPV($ADC_bpm2_pv,'double');
#       if ($trace && $n < $trace_max) {$msg = sprintf $sbscrpt.': ADC_bpm1=%.3f ADC_bpm2=%.3f'."\n",ADC_bpm1_1,ADC_bpm2_1; &log_record($msg);}

        ### New pitch position
        $lockin_use = &getPV($lockin_use_pv);
        if ($lockin_use > 1) {$lockin_use=1; &putPV($lockin_use_pv,$lockin_use,'double');}

        if ($lockin_use == 1) {		#------------------------------------if hybrid mode
           if ($lockin_use != $lockin_prev) {
              if ($lockin_prev != -1) {&post_message('Switching to hybrid feedback');}
              $lockin_prev = $lockin_use;
              $n = 0;
           }
           $value    = &get_lockin_DAC_step($lockin_xOUT_pv);
           $DCM_VSUM = &getPV($ADC_bpm1_pv,'double');
           if    ($DCM_VSUM < 0.)            {$DCM_VSUM = 0.;}
           elsif ($DCM_VSUM > $DCM_VSUM_max) {$DCM_VSUM_max = $DCM_VSUM;}
           ### Normalization to prevent oscillations near ID-B switch
           ### between ID harmonics 1 and  3.
           ### At $DCM_VSUM=0 -> $factor=0.33*$DCM_VSUM_max ~ 0.33*3 ~ 1
           ### At $DCM_VSUM=$DCM_VSUM_max -> $factor=0.1
#          $factor = 0.1*$DCM_VSUM_max/sqrt($DCM_VSUM**2+0.1);		# $DCM_VSUM_max=~ 1 to 4
           $factor = 1;							#SS: 2021/10/05, this routine does searching, no need to reduce oscillations at peak
           $DAC_step     = $value * $factor;
           $DAC_step_max = abs($factor_max * $opt_step * $multiplier);
           if (abs($DAC_step) > $DAC_step_max) {
              $msg = sprintf $subscript.': Correction: DAC_step(lockin)=%.4f mode=%s - restricting step to %.4f',$DAC_step,$mname[$mode],$DAC_step_max*abs($DAC_step)/$DAC_step;
              &log_record($msg);
              $DAC_step = $DAC_step_max * abs($DAC_step)/$DAC_step;
           }
           $DAC_piezo_new = $DAC_piezo + $DAC_step;			#SZ

           if ($DAC_piezo_new > $DAC_piezo_DRVH) {$DAC_piezo_new = $DAC_piezo_DRVH;}
           if ($DAC_piezo_new < $DAC_piezo_DRVL) {$DAC_piezo_new = $DAC_piezo_DRVL;}
           &kill_p2();
           if ($trace) {
              $msg = sprintf $subscript.': Correction: DAC_step_calc(lockin)=%.4f factor=%.4f DAC_step=%.4f mode=%s',
                                                                              $value,$factor,$DAC_step,$mname[$mode];
              &log_record($msg);
           }
           &putPV($DAC_piezo_pv,$DAC_piezo_new,'double');
        } 
        else {	#-------------------------------------------------------if software mode
           if ($lockin_use != $lockin_prev){
              if ($lockin_prev != -1) {&post_message('Switching to software feedback');}
              $lockin_prev = $lockin_use;
              $n = 0;
           }
           $factor = $multiplier;
           $DAC_step = $opt_step * $factor;
           $DAC_piezo_new = $DAC_piezo + $DAC_step;			#SZ
#          if ($trace && $n < $trace_max) {$msg = sprintf $sbscrpt.': DAC_step=%.4f'."\n",$DAC_step; &log_record($msg);}

           if ($DAC_piezo_new > $DAC_piezo_DRVH) {$DAC_piezo_new = $DAC_piezo_DRVH;}
           if ($DAC_piezo_new < $DAC_piezo_DRVL) {$DAC_piezo_new = $DAC_piezo_DRVH;}
           &kill_p2();
           if ($trace) {
              $msg = sprintf $subscript.': Correction: DAC_step_calc(soft)=%.4f factor=%.4f DAC_step=%.4f mode=%s',
                                                                       $opt_step,$factor,$DAC_step,$mname[$mode];
              &log_record($msg);
           }
           &putPV($DAC_piezo_pv,$DAC_piezo_new,'double');
        } #-------------------------------------------------------------end of fork over HW/Hybrid/Soft modes

        &putPV($mostab_corr_pv,$DAC_step,'double');

        $mode = &medm_check($mode);
        if ($mode == $mPAUSED) {&post_message('Feedback paused.'); return;}
        ### IMPORTANT NOTE: The rate may need to be longer because the piezo
        ### motion is not instantaneous!
        select(undef,undef,undef,$fdbk_rate);
        &kill_p2();
        $mode = &medm_check($mode);
        if ($mode == $mPAUSED) {&post_message('Feedback paused.'); return;}
        $DAC_check_err = &DAC_check(2*$opt_step);
        if ($DAC_check_err) {
           &post_message('Feedback paused: DAC error.'); 
           $mode = $mPAUSED; 
           return;
        }

        ### Measurement 2
        $ADC_bpm1_m2 = &getPV($ADC_bpm1_pv,'double');
        $ADC_bpm2_m2 = &getPV($ADC_bpm2_pv,'double');

        ### Determine which bpm to use and compare m1 and m2 ----------
        $bpm_choice = &bpm2();
        if ($trace && $n < $trace_max) {
           if ($bpm_choice < 2) {
              $msg = sprintf $sbscrpt.': BPM1_old=%.3f BPM1_new=%.3f'
                                     .' BPM1_change=%.5f%s DAC_step=%.4f'."\n",
                                     $ADC_bpm1_m1,$ADC_bpm1_m2,
                                     200*($ADC_bpm1_m2-$ADC_bpm1_m1)/($ADC_bpm1_m2+$ADC_bpm1_m1),
                                     '%',$DAC_step;
           } else {
              $msg = sprintf $sbscrpt.': BPM2_old=%.3f BPM2_new=%.3f'
                                     .' BPM2_change=%.5f%s DAC_step=%.4f'."\n",
                                     $ADC_bpm2_m1,$ADC_bpm2_m2,
                                     200*($ADC_bpm2_m2-$ADC_bpm2_m1)/($ADC_bpm2_m2+$ADC_bpm2_m1),
                                     '%',$DAC_step;
           }
           &log_record($msg);
        }   
        ### Compare signals depending on chosen bpm ----------
        ###    0 - use bpm1, by choice
        ###    1 - use bpm1, because bpm2 isn't ready
        ###    2 - use bpm2, because all checks satisfied
        ###    3 - bpm2 recently available, optimize now on bpm2
        ###    4 - both signals bad; lost beam
        if ($bpm_choice < 2 && $ADC_bpm1_m2 < $ADC_bpm1_m1) {
           $nrev += 1;
           $multiplier *= -1;
           $msg = sprintf 'Locking (iter=%d, step=%d, dir=%d), %s',$iter+1,$opt_steps[$iter],$multiplier,'bpm1';
           &post_message($msg);
        }
        elsif ($bpm_choice == 2 && $ADC_bpm2_m2 < $ADC_bpm2_m1) {
           $nrev += 1;
           $multiplier *= -1;
           $msg = sprintf 'Locking (iter=%d, step=%d, dir=%d), %s',$iter+1,$opt_steps[$iter],$multiplier,'bpm2';
           &post_message($msg);
        }
        elsif ($bpm_choice == 3) {
           &post_message('BPM2 became available, optimise on it.');
           $mode = $mACTFDBK2; 
           return;
        }
        elsif ($bpm_choice == 4) {
           &post_message('Both BPM signals are low, lost beam.');
           $mode = $mCOMPCHK;
           return;
        }

        ### ------------------------------------------------------

        $opt_step = $volt_step * $opt_steps[$iter];
        if ($use_LIA_DAC) {$lockin_ADC1 = &get_lockin_ADC($lockin_ADC1_pv);}		#if using LIA DAC to drive piezo
        else              {$lockin_ADC1 = $DAC_piezo;}
        if    ($DAC_piezo < $DACmin || $lockin_ADC1 < $DACmin) {	#if non-HW mode & using LIA DAC to drive piezo
           &post_message('WARNING: Beam @ neg V limit.', 'warn');
           $msg = sprintf 'DAC_piezo=%.2f (low=%.2f)',$DAC_piezo,$DACmin;
           &post_message($msg);
           if ($use_LIA_DAC) {
              $msg = sprintf 'lockin_ADC1=%.2f (low=%.2f)',$lockin_ADC1,$DACmin;
              &post_message($msg);
           }
           &post_message('Automatically shifting P2.', 'warn');
           $mode = $mOPTIMP2;
           return;
        }
        elsif ($DAC_piezo > $DACmax || $lockin_ADC1 > $DACmax) {
           &post_message('WARNING: Beam @ pos V limit.', 'warn');
           $msg = sprintf 'DAC_piezo=%.2f (high=%.2f)',$DAC_piezo,$DACmax;
           &post_message($msg);
           if ($use_LIA_DAC) {
              $msg = sprintf 'lockin_ADC1=%.2f (high=%.2f)',$lockin_ADC1,$DACmax;
              &post_message($msg);
           }
           &post_message('Automatically shifting P2.', 'warn');
           $mode = $mOPTIMP2;
           return;
        }
        $mode = &medm_check($mode);					# 5
        if ($mode == $mPAUSED) {&post_message('Feedback paused.'); return;}
        &kill_p2();
        if ($wait_dcm) {
           &post_message('DCM motion detected!', 'warn');
           $mode = $mCOMPCHK;
           return;
        }
        if ($mode == $mPAUSED) {&post_message('Feedback paused.'); return;}
     } #= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = end of loop over reversals
     &kill_p2();
     $mode = &medm_check($mode);
     if ($mode == $mPAUSED) {&post_message('Feedback paused.'); return;}
  } #===================================================================end of loop over iterations (step sizes)
  &post_message('Beam is locked - '.$sbscrpt);
  $mode = $mACTFDBK1;
  $mostab_lock = 1;
  &putPV($mostab_lock_pv,$mostab_lock,'double');
  &kill_p2();
  $mode = &medm_check($mode);
  if ($mode == $mPAUSED) {&post_message('Feedback paused.'); return;}
}

#==========================================================================
sub intensity_check ($) {
        
### Intensity check can be called in two modes: $mINTCHK1 and $mINTCHK2.
### If the beam intensity is above the low threshold, both go to &fast_feedback().
### The difference between them is when the beam is low:
###  - &intensity_check($mINTCHK1) goes to $mSEARCH  (&search();)
###  - &intensity_check($mINTCHK2) goes to $mCOMPCHK (&component_check();)

  my ($subscript, $sbscrpt, $message, $msg);
  my ($ADC_bpm, $ADC_bpm1, $ADC_bpm2);
  my ($int_overflow, $int_low, $pv, $bp);
  my ($checkMode, $bpmUse, $bpm_choice);

  $sbscrpt = 'intensity_check';
  $subscript = $script.'/'.$sbscrpt;
  &log_func_entry($subscript);

  $checkMode = shift(@_);

### In versions 7.20 and before intensity check was always checking signal on BPM1.
### Here we introduce a switch: checking signal on BPM1 or currently selected BPM
  $bpmUse = 'bpm1';
# $bpmUse = 'auto';

  &post_message('Checking for intensity.');
  &status_write($mode);
  $mode = &medm_check($mode);						# 6
  if ($mode == $mPAUSED) {&post_message('Feedback paused.'); return;}

  if ($have_CAEN) {
### SS: Should we change range in case of overflow with CAEN?
     $CAEN_range = &getPV($CAEN_range_pv,'string');			#amplification range string
     if ($CAEN_range !~ /uA$/) {
        $msg = '!!! CAEN is in unexpected mode = '.$CAEN_range;
        &post_message($msg);
        &tellBluIce ($scr.': '.$msg, 'warn');
     }
  }
  if ($bpmUse eq 'auto') {
     ### Determine which bpm to use and compare m1 and m2 ----------
     ###    0 - use bpm1, by choice
     ###    1 - use bpm1, because bpm2 isn't ready
     ###    2 - use bpm2, because all checks satisfied
     ###    3 - bpm2 recently available, optimize now on bpm2
     ###    4 - both signals bad; lost beam
     $bpm_choice = &bpm2();
     if ($bpm_choice == 4) {
        $mode = $mSEARCH; 						# mode=4 <= beam search
        $msg = 'Both BPM low.';
        &post_message($msg, 'none');
        $message = $msg;
        $msg = 'Performing search - '.$sbscrpt;
        &post_message($msg, 'none');
        $message .= ' '.$msg;
        &tellBluIce ($scr.': '.$message, 'warn');
        return;
     }
  } else {
     $bpm_choice = 0;							#use bpm1, by choice
  }

  ($ADC_bpm1,$ADC_bpm2) = &measure_intensity();
  if ($bpm_choice < 2)  {$bp=1; $ADC_bpm=$ADC_bpm1; $int_overflow=$int_overflow1; $int_low=$int_ok1; $pv=$ADC_bpm1_pv;}
  else                  {$bp=2; $ADC_bpm=$ADC_bpm2; $int_overflow=$int_overflow2; $int_low=$int_ok2; $pv=$ADC_bpm2_pv;}

  if ($ADC_bpm >= $int_overflow) {
     $mode = $mPAUSED;							# mode=5 <= feedback; paused
     $msg = 'Intensity overflow detected on '.$pv.'.';
     &post_message($msg, 'none');
     $message = $msg;
     $msg = sprintf 'Intensity=%.2f > max=%.2f', $ADC_bpm, $int_overflow;
     &post_message($msg, 'none');
     $message .= ' '.$msg;
     if ($bpm_choice < 2) {
        $msg = 'Please verify no YAG in - '.$sbscrpt;
        &post_message($msg, 'none');
        $message .= ' '.$msg;
     }
     $msg = 'Feedback is paused - '.$sbscrpt;
     &post_message($msg, 'none');
     $message .= ' '.$msg;
     &tellBluIce ($scr.': '.$message, 'error');
  }
  elsif ($ADC_bpm >= $int_low) {
     $mode = $mACTFDBK2;						# mode=8 <= active feedback (&fast_feedback); large step size
     &post_message('Beam detected bpm'.$bp.' - '.$sbscrpt);
     $msg = sprintf 'Intensity=%.2f > min=%.2f', $ADC_bpm, $int_low;
     &post_message($msg);
     &post_message('Please wait for beam lock - '.$sbscrpt);
  }
  elsif ($ADC_bpm < $int_low) {
     $msg = 'Little or no beam - '.$sbscrpt;
     &post_message($msg, 'none');
     $message = $msg;
     $msg = sprintf 'Intensity=%.2f < min=%.2f', $ADC_bpm, $int_low;
     &post_message($msg, 'none');
     $message .= ' '.$msg;
     if    ($checkMode == $mINTCHK1) {
        $mode = $mSEARCH;						# mode=4 <= beam search
        $msg = 'Performing search - '.$sbscrpt;
     }
     elsif ($checkMode == $mINTCHK2) {
        $mode = $mCOMPCHK;						# mode=3 = component status check;
        $msg = 'Checking beamline components - '.$sbscrpt;
     }
     &post_message($msg, 'none');
     $message .= ' '.$msg;
     &tellBluIce ($scr.': '.$message, 'warn')
  }
}

#==========================================================================
sub component_check () {

  my ($subscript, $sbscrpt, $ring_status, $enable_status, $mono_status);
  my ($status, $E_id, $E_diff, $E_off, $E_check, $id_status);
  my ($ADC_bpm1);

  $sbscrpt = 'component_check';
  $subscript = $script.'/'.$sbscrpt;
  &log_func_entry($subscript);

  &post_message('Checking: ring, shutters, DCM.');
  &status_write($mode);
  $mode = &medm_check($mode);						# 7
  if ($mode == $mPAUSED) {&post_message('Feedback paused.'); return;}
  select(undef,undef,undef,$component_time);

  $ring_status   = 1;
  $enable_status = 1;
  $a_status	 = 1;
  $c_status	 = 1;
  $mono_status   = 1;
  $id_status	 = 1;
  $status = $ring_status + $enable_status + $id_status + $a_status + $mono_status;
  if ($beamline eq '23i') {$status += $c_status;}

  while ($status > 0) {
     $mode = &medm_check($mode);					# 8
     if ($mode == $mPAUSED) {&post_message('Feedback paused.'); return;}

     ### Initial check of ring, shutter, and monochromator status
     $current = &getPV($current_pv,'double');

     if ($current < 10) {
        #&post_message('Ring dump.', 'warn');
        #&post_message('Please wait for A-shutter auto open.');
        &post_message('Low APS current, waiting for A-shutter'); 
        $mode = $mAOPEN;
        return;
     }

     if ($current > 50)  {$ring_status = 0;}
     else                {$ring_status = 1;}
     $permit = &getPV($permit_pv,'double');
     if ($permit == 1)   {$enable_status = 0;}
     else                {$enable_status = 1;}
     $a_status = &getPV($a_status_pv,'double');
     $c_status = &getPV($c_status_pv,'double');
     $E_mov = &getPV($E_mov_pv,'double');
     if ($E_mov == 1)    {$mono_status = 0;}
     else                {$mono_status = 1;}
     if (defined $id_pv) {
        $E_id = &getPV($id_pv,'double');
        $E = &getPV($E_pv,'double');
        $E_diff = $E_id - $E;
        $E_off = &calcIDoffset($E, 0,
                              \$enmono13, \@harmEn, \@harmNum, 
                              \@Elower, \@Eupper, \@Olower, \@Oupper);
        $E_check = abs($E_diff - $E_off);
        if ($E_check > $max_E_tol) {$id_status = 1;}
        else                       {$id_status = 0;}
     }
     else {$id_status = 0;};
#
     if ($ring_status) {
        &post_message('Waiting for ring current.');
        while ($ring_status == 1) {
           $current = &getPV($current_pv,'double');
           if ($current > 50) {$ring_status = 0;}
           else               {select(undef,undef,undef,$pause_time);}
           $mode = &medm_check($mWAITBEAM);				# 9
           if ($mode == $mPAUSED) {&post_message('Feedback paused.'); return;}
        }
        &post_message('Sufficient current present.');
     }
#
     if ($enable_status) {
        &post_message('Waiting for FES enable.');
        while ($enable_status == 1) {
           $permit = &getPV($permit_pv,'double');
           if ($permit == 1) {$enable_status = 0;}
           else              {select(undef,undef,undef,$pause_time);}
           $mode = &medm_check($mWAITBEAM);				# 10
           if ($mode == $mPAUSED) {&post_message('Feedback paused.'); return;}
        }
        &post_message('FES enabled.');
     }
#
     if ($id_status) {
        &post_message('ID/DCM mismatch!', 'warn');
        while ($id_status) {
           $ADC_bpm1 = &getPV($ADC_bpm1_pv,'double');
           if ($ADC_bpm1 > $int_ok1) {
              &post_message('Sufficient signal detected.');
              &post_message('Optimizing beam.');
              $mode = $mACTFDBK2;
              return;
           }
           $E_id = &getPV($id_pv,'double');
           $E = &getPV($E_pv,'double');
           $E_diff = $E_id - $E;
           $E_off = &calcIDoffset($E, 0,
                              \$enmono13, \@harmEn, \@harmNum, 
                              \@Elower, \@Eupper, \@Olower, \@Oupper);
           $E_check = abs($E_diff - $E_off);
           if ($E_check > $max_E_tol) {$id_status = 1; select(undef,undef,undef,$pause_time);}
           else                       {$id_status = 0;}
           $mode = &medm_check($mWAITEMIS);				# 10
           if ($mode == $mPAUSED) {&post_message('Feedback paused.'); return;}
        }
        &post_message('ID energy OK.');
     }
#
     if ($a_status == 1) {
        &post_message('Waiting for A-shutter.');
        while ($a_status == 1) {
           $a_status = &getPV($a_status_pv,'double');
           if ($a_status ==1) {select(undef,undef,undef,$pause_time);}
           $mode = &medm_check($mWAITOPNA);				# 11
           if ($mode == $mPAUSED) {&post_message('Feedback paused.'); return;}
        }
        &post_message('A-shutter open.');
     }
#
     if (($c_status == 1) && ($beamline eq '23i')) {
        &post_message('Waiting for C-shutter.');
        while ($c_status == 1) {
           $c_status = &getPV($c_status_pv,'double');
           if ($c_status == 1) {select(undef,undef,undef,$pause_time);}
           $mode = &medm_check($mWAITOPNC);				# 12
           if ($mode == $mPAUSED) {&post_message('Feedback paused.'); return;}
        }
        &post_message('C-shutter open.');
     }
#
     if ($mono_status == 1) {
        &post_message('Waiting for DCM.');
        while ($mono_status == 1) {
           $E_mov = &getPV($E_mov_pv,'double');
           if ($E_mov == 1) {$mono_status = 0;}
           else             {select(undef,undef,undef,$pause_time);}
           $mode = &medm_check($mWAITDCM);				# 13
           if ($mode == $mPAUSED) {&post_message('Feedback paused.'); return;}
        }
        &post_message('DCM positioned.');
     }

     ### Double check status
     select(undef,undef,undef,$component_time);
     $current = &getPV($current_pv,'double');
     if ($current > 50) {$ring_status = 0;}
     else               {$ring_status = 1;}
     $permit = &getPV($permit_pv,'double');
     if ($permit == 1) {$enable_status = 0;}
     else              {$enable_status = 1;}
     $a_status = &getPV($a_status_pv,'double');
     $c_status = &getPV($c_status_pv,'double');
     $E_mov = &getPV($E_mov_pv,'double');
     if ($E_mov eq 1) {$mono_status = 0;}
     else             {$mono_status = 1;}
     if (defined $id_pv) {
        $E_id = &getPV($id_pv,'double');
        $E = &getPV($E_pv,'double');
        $E_diff = $E_id - $E;
        $E_off = &calcIDoffset($E, 0,
                              \$enmono13, \@harmEn, \@harmNum, 
                              \@Elower, \@Eupper, \@Olower, \@Oupper);
        $E_check = abs($E_diff - $E_off);
        if ($E_check > $max_E_tol) {$id_status = 1;}			# 90 = current theoretical max + 10 - 7 Feb 2007
        else                       {$id_status = 0;}
     }
     else {$id_status = 0;};

     $status = $ring_status + $enable_status + $id_status + $a_status + $mono_status;
     if ($beamline eq '23i') {$status += $c_status;}

  }
  &post_message('Ring, shutters, DCM - OK.');
  $mode = $mINTCHK1;
}

#==========================================================================
sub search () {

  my ($subscript, $sbscrpt, $max_int, $DAC_piezo_max, $DAC_piezo_curr);
  my ($trip_level, $beam_int);
  my ($last_DAC_piezo, $last_E, $last_p2, $i, $fwhm_x_E, $fwhm);
  my ($scan_pts, $scan_step, $search_step, $search_start, $search_end);
  my ($status, $msg, $joerger_ts, $ADC_bpm1);
  my ($busy, $currPts, $totalPts, $runPrg, $inPos);
  my ($wait_, $wait_start, $wait_step, $wait_rep, $wait_rstep);
  my (@p2_positions, @p2_intensity, $p2_rqs, $p2_act, $i_max, $intensity_max);

  $sbscrpt = 'search';
  $subscript = $script.'/'.$sbscrpt;
  &log_func_entry($subscript);

  &post_message('Performing beam search.');

  ### Update status
  &status_write($mode);
  $mode = &medm_check($mode);						# 14
  if ($mode == $mPAUSED) {&post_message('Feedback paused.'); return;}	# if paused
  $mostab_lock = 0;
  &putPV($mostab_lock_pv,$mostab_lock,'double');
  &kill_p2();
  $mostab_search_mode = &getPV($mostab_search_mode_pv,'double');	#0=search with DAC, 1=search with P2

  if ($mostab_search_mode == 1) { #~~~~~~~~~Search-with-P2~~~~~~~~~~~~~~~~~~~~~~~~
     &post_message('Searching with P2.');
     ### Record last positions
     $last_DAC_piezo = &getPV($DAC_piezo_pv,'double');
     $last_E         = &getPV($E_pv,'double');
     $last_p2        = &getAxisAct($p2_axis_pv,\$status);
     &func_error($status,'getAxisAct',$p2_axis_pv,'pause');

     $mostab_search_range = &getPV($mostab_search_range_pv,'string');	#expect 1x, 2x, 4x
     $mostab_search_range =~ s/x//g;

     ### Leave DAC where it is

     ### Close loop and put motor in auto
     &post_message('Enabling/setting motor '.$p2_assy_pv);
     $status = &configAssy ($p2_assy_pv,'forceAuto');
     &func_error($status,'configAssy',$p2_assy_pv,'pause');

     ### Sync p2 actual and requested positions
     &post_message('Syncing motor P2 actual & requested pos.');
     $status = &syncAssy($p2_assy_pv);
     &func_error($status,'syncAssy',$p2_assy_pv,'pause');

     ### Calculate starting angle for search and step size
     $fwhm_x_E = &calc_fwhm_x_E();
     $hkl = &getMonoHKL();
     $fwhm = 0.001 * ($fwhm_x_E / $last_E);				# Rocking curve FWHM (urad -> mrad)
     $search_step = $fwhm/5.;						# Rocking curve step
     ### Search +/- 10 rocking curve widths (FWHM) around current position
#    $halfrange = 10.*$fwhm * $mostab_search_range;
     $halfrange = 20.*$fwhm * $mostab_search_range;			#Sergey: increased 2021.10.19

     if ($have_joerger) { #-------------Begin search with scanrec fly scan------#
        &start_scaler();                                                        #
        $scan_step  = $search_step;                                             #
        $joerger_ts = $joerger_time;                                            #
        $p2_speed = $scan_step/$joerger_ts;                                     #
#       if ($p2_speed > $p2_max_speed) {                                        #
#          $scan_step *= ($p2_max_speed/$p2_speed);                             # have to decrease the step
#          $p2_speed = $p2_max_speed;                                           #
#       }                                                                       #
        if ($p2_speed > $p2_max_speed) {                                        #
           $joerger_ts *= ($p2_speed/$p2_max_speed);                            # have to increase count time
           $p2_speed = $p2_max_speed;                                           #
           &changeSRdetectors_v2 ($scanrec, undef, undef, $joerger_ts);         #
           &setJoergerTime  ($scaler, $joerger_ts, 'autocount');                # change dwell for autocount
        }                                                                       #
        $scan_pts = int(2.*$halfrange/$scan_step + 1.5);                        #
        if ($scan_pts > $scanrec_max_pts) {                                     #
           $scan_pts = $scanrec_max_pts;                                        # have to restrict range
           $halfrange = 0.5*$scan_step*($scan_pts-1);                           #
        }                                                                       #
        $search_start = $last_p2 - $halfrange;                                  #
        $search_end   = $last_p2 + $halfrange;                                  #
                                                                                #
        $msg = sprintf 'P2 fly search, step=%.2g pts=%d range=+-%.2g '.         #
                       'start=%.2g end=%.2g '.                                  #
                       'speed=%.2g maxspeed=%.2g',                              #
                        $scan_step, $scan_pts,                                  #
                        $halfrange, $search_start, $search_end,                 #
                        $p2_speed, $p2_max_speed;                               #
        &log_record($msg);                                                      #
        &tellBluIce ($scr.': '.$msg);                                           #
                                                                                #
        #&changeSRmotors ($scanrec, $numPts, $numMotors, \@begin, \@end)        #
        &changeSRmotors ($scanrec,$scan_pts,1,[$search_start],[$search_end]);   #
        $status = &runMotor($p2_assy_pv,$p2_axis_pv,$search_start,'forceAuto'); #
        $status = &setAxisSpeed ($p2_assy_pv, $p2_axis_pv, $p2_speed);          #
                                                                                #
        &post_message('Starting motor P2 fly scan.');                           #
        &startSR ($scanrec);                                                    #
                                                                                #
        ### Wait to finish:                                                     #
        $wait_      = 0;                                                        #
        $wait_start = 4;                                                        #
        $wait_step  = 0.1;                                                      #
        $wait_rep   = 0;                                                        #
        $wait_rstep = 2;                                                        #
        $scan_in_progress = 1;                                                  #
        $p2_act = &getAxisAct ($p2_axis_pv);                                    #
        do {                                                                    #
           select(undef,undef,undef,$wait_step);                                # sleep($wait_step);
           $wait_    += $wait_step;                                             #
           $wait_rep += $wait_step;                                             #
           ($status,$busy)   = Pezca::GetDouble($scanrec.'.BUSY');              #
           ($status,$currPts)= Pezca::GetDouble($scanrec.'.CPT');               #
           ($status,$inPos)  = Pezca::GetDouble($p2_assy_pv.':InPos');          #
           ($status,$runPrg) = Pezca::GetDouble($p2_assy_pv.':RunPrg');         #
           if ($wait_rep >= $wait_rstep) {                                      #
              $wait_rep = 0;                                                    #
              ($status,$totalPts)= Pezca::GetDouble($scanrec.'.NPTS');          #
              &post_message($currPts.'/'.$totalPts.' scan steps done');         #
           }                                                                    #
        } while ((($busy == 1 || $currPts < 1) && ($runPrg || !$inPos)) ||      #
                                                   $wait_< $wait_start);        #
        $msg = sprintf 'P2 fly search finished busy=%d, currPts=%d,'            #
                      .' runPrg=%d, inpos=%d, wait=%.0f',                       #
                        $busy, $currPts, $runPrg, $inPos, $wait_;               #
        &log_record($msg);                                                      #
        &abortSR ($scanrecAbort, $scanrec);                                     #
        &abortAssy ($p2_assy_pv);                                               #
        &restoreAxis($p2_assy_pv, $p2_axis_pv);                                 #
        $scan_in_progress = 0;                                                  #
        $currPts = &waitSRdata ($scanrec);                                      #
        &post_message('Fly scan finished, pts='.$currPts);                      #
        @p2_positions = &readSRx ($scanrec, $currPts, 1);                       #
        @p2_positions = &calcSRxMiddle ($currPts, $p2_act, @p2_positions);      #
        @p2_intensity = &readSRy ($scanrec, $currPts, 1);                       #
        $i_max = &findIndexMaxRef(\@p2_intensity);                              #
        $intensity_max = (1E-5)*$p2_intensity[$i_max]/$joerger_ts;              #V2F converter gives (1E5)
        if ($intensity_max < $int_ok1) {                                        #
           ### Restore position                                                 #
           $status = &runMotor($p2_assy_pv,$p2_axis_pv,$last_p2,'forceAuto');   #
           &kill_p2(1);                                                         #
           $mode = $mNOBEAM;                                                    # no beam => pause feedback
           $msg = sprintf 'No beam within P2 scan range: max=%.2f, need=%.2f',  #
                                                    $intensity_max, $int_ok1,   #
           &post_message($msg, 'error');                                        #
           &post_message('Operator intervention needed.', 'error');             #
        } else {                                                                #
           ### Drive to max intensity                                           #
           $p2_rqs = $p2_positions[$i_max];                                     #
           $p2_rqs += +1.0*$scan_step;                                          # backlash correction
           $status = &runMotor($p2_assy_pv,$p2_axis_pv,$p2_rqs,'forceAuto');    #
           &post_message('Beam detected.');                                     #
           $beam_int = &getPV($scaler.'.S'.$joerger_bpm1,'double');             #
           $beam_int = (1E-5)*$beam_int/$joerger_ts,                            #
           $ADC_bpm1 = &getPV($ADC_bpm1_pv,'double');                           #
           $msg = sprintf 'Intensity: scan=%.2f scaler=%.2f ADC=%.2f',          #
                                        $intensity_max, $beam_int, $ADC_bpm1,   #
           &post_message($msg);                                                 #
           &post_message('Please wait for beam lock - '.$sbscrpt);              #
           &kill_p2(1);                                                         #
           $mode = $mACTFDBK2;                                                  # active feedback; large step size
        }                                                                       #
        &stop_scaler();                                                         #
        if ($joerger_ts != $joerger_time) {                                     # restore the defaults
           &changeSRdetectors_v2 ($scanrec, undef, undef, $joerger_time);       #
           &setJoergerTime  ($scaler, $joerger_time, 'autocount');              # change dwell for autocount
        }                                                                       #
        return;                                                                 #
     }  #----------------End search with scanrec fly scan-----------------------#
     else
     {  #----------------Begin search with step scan & ADC (no Joerger)---------#
                                                                                #
        $scan_pts = int(2.*$halfrange/$search_step + 1.5);                      #
        $search_start = $last_p2 - $halfrange;                                  #
        $search_end   = $last_p2 + $halfrange;                                  #
        $msg = sprintf 'P2 step search step=%.2g pts=%d range=+-%.2g '.         #
                       'start=%.2g end=%.2g',                                   #
                       $search_step, $scan_pts,                                 #
                       $halfrange, $search_start, $search_end;                  #
        &log_record($msg);                                                      #
        &tellBluIce ($scr.': '.$msg);                                           #
                                                                                #
        &post_message('Starting motor P2 step scan.');                          #
        for ($i=0; $i<$scan_pts; $i++) {  ### STEP-SCAN with P2 motor ~~~~~~~+  #
           $p2_rqs = $search_start + $search_step * $i;                     #|  #
           &post_message('Running motor '.$p2_assy_pv.' scan point '        #|  #
                         .($i+1).' of '.$scan_pts);                         #|  #
           $status = &runMotor ($p2_assy_pv, $p2_axis_pv,                   #|  #
                                              $p2_rqs, 'forceAuto');        #|  #
           if ($status) {                                                   #|  #
              &kill_p2(1);                                                  #|  #
              $mode = $mNOBEAM;                                             #|  # no beam => pause feedback
              &post_message('P2 failed to position.', 'error');             #|  #
              &post_message('Intervention needed to find beam.', 'error');  #|  #
              return;                                                       #|  #
           }                                                                #|  #
           $mode = &medm_check($mode);                                      #|  # if paused
           if ($mode == $mPAUSED) {                                         #|  #
              &kill_p2(1);                                                  #|  #
              &post_message('Feedback paused.');                            #|  #
              return;                                                       #|  #
           }                                                                #|  #
           ### Check for intensity                                          #|  #
           $ADC_bpm1 = &getPV($ADC_bpm1_pv,'double');                       #|  #
           if ($ADC_bpm1 >= $int_ok1) {                                     #|  #
              &kill_p2(1);                                                  #|  #
              $mode = $mACTFDBK2;                                           #|  # active feedback; large step size
              &post_message('Beam detected, aborting P2 scan');             #|  #
              $beam_int = sprintf '%.2f',$ADC_bpm1;                         #|  #
              &post_message('Intensity='.$beam_int);                        #|  #
              &post_message('Please wait for beam lock - '.$sbscrpt);       #|  #
              return;                                                       #|  #
           }                                                                #|  #
           $mode = &medm_check($mode);                                      #|  #
           if ($mode == $mPAUSED) {                                         #|  #
              &kill_p2(1);                                                  #|  #
              &post_message('Feedback paused.');                            #|  #
              return;                                                       #|  #
           }                                                                #|  # if paused
        }  ### End of STEP-SCAN with P2 motor  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~+  #
     }  #----------------End search with step scan & ADC (no Joerger)-----------#
     &post_message('P2 scan ended.');
     &kill_p2(1);
     $mode = $mNOBEAM;							# no beam; => pause feedback
     &post_message('No beam within P2 scan range.', 'error');
     &post_message('Intervention needed to find beam.', 'error');

  } else { #~~~~~~~~~~~~~~~~~~~~~~Search-with-DAC~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

     &post_message('Searching with DAC.');
     ### Record last DAC
     $DAC_piezo_last = &getPV($DAC_rdbk_pv,'double');

     $DAC_piezo = &getPV($DAC_piezo_pv,'double');
     while ($DAC_piezo > $DACmin) {					# SS: what is this? Smooth decrease of DAC?
        $DAC_piezo -= 0.2;
        if ($DAC_piezo < $DAC_piezo_DRVL) {$DAC_piezo = $DAC_piezo_DRVL;}
        &kill_p2();
        &putPV($DAC_piezo_pv,$DAC_piezo,'double');
        $mode = &medm_check($mode);					# 15
        if ($mode == $mPAUSED) {&post_message('Feedback paused.'); return;}	# if paused
        &kill_p2();
     }
     select(undef,undef,undef,1);
     &kill_p2();
     $DAC_check_err = &DAC_check(0);
     if ($DAC_check_err) {						# failure; pause feedback
        $mode = $mPAUSED; 
        &post_message('DAC error. Feedback paused.'); 
        return
     }
     $max_int = 0;
     $DAC_piezo_max = 0;
     $DAC_piezo_curr = 0;

     for ($i=0; $i<$n_steps_DAC; $i++) {  ### STEP-SCAN with DAC ~~~~~~~~~~~
        $DAC_piezo = (10.*$i)/$n_steps_DAC;
        &kill_p2();
        &putPV($DAC_piezo_pv,$DAC_piezo,'double');
        $DAC_check_err = &DAC_check(0);
        if ($DAC_check_err) {						# failure; pause feedback
           $mode = $mPAUSED; 
           &post_message('DAC error. Feedback paused.'); 
           return
        }
        select(undef,undef,undef,0.1);
        &kill_p2();
        $ADC_bpm1 = &getPV($ADC_bpm1_pv,'double');
        if ($ADC_bpm1 > $max_int) {
           $DAC_piezo_max = $DAC_piezo;
           $max_int = $ADC_bpm1;
        }
        $mode = &medm_check($mode);					# 16
        if ($mode == $mPAUSED) {&post_message('Feedback paused.'); return;}	# if paused
        if ($ADC_bpm1 > $int_ok1) {
           &post_message('Beam detected.');
           $mode = $mACTFDBK2;						# active feedback; large step size
           return;
        }
     }  ### End of STEP-SCAN with DAC ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     if ($max_int < $int_low1) {					# no beam; => pause feedback
        $mode = $mNOBEAM; 
        &post_message('No beam. Feedback paused.'); 
        return
     }		# no beam; => pause feedback
     if ($DAC_piezo_max < $DACmin) {
        &post_message('WARNING: Beam @ neg V limit.', 'warn');
        &post_message('Shift P2.', 'warn');
        &pause_feedback();
     }
     if ($DAC_piezo_max > $DACmax) {
        &post_message('WARNING: Beam @ pos V limit.', 'warn');
        &post_message('Shift P2.', 'warn');
        &pause_feedback();
     }
     $trip_level = 0.5 * $max_int;
     $beam_int = sprintf 'Max = %.2f @ DAC = %.2f',$max_int, $DAC_piezo_max;
     &post_message($beam_int);
     while ($ADC_bpm1 < $trip_level) {
        $DAC_piezo -= 0.05;
        &kill_p2();
        &putPV($DAC_piezo_pv,$DAC_piezo,'double');
        $DAC_check_err = &DAC_check(0);
        if ($DAC_check_err) {
           $mode = $mPAUSED; 
           &post_message('DAC error. Feedback paused.'); 
           return
        }
        $ADC_bpm1 = &getPV($ADC_bpm1_pv,'double');
        $mode = &medm_check($mode);					# 17
        if ($mode == $mPAUSED) {&post_message('Feedback paused.'); return;}
     }
     $mode = $mACTFDBK2;
  } #~~~~~~~~~~~~~~~~~~~~~~End of fork searching with DAC or P2~~~~~~~~~~~~~~~~~~~~~~~~~
  return;
}

#==========================================================================
sub pause_feedback () {

  my ($subscript, $sbscrpt);

  $sbscrpt = 'pause_feedback';
  $subscript = $script.'/'.$sbscrpt;
  &log_func_entry($subscript);

  &status_write($mode);
  &post_message('Script paused.', 'warn');
  $mostab_pause = 1;
  &putPV($mostab_pause_pv,$mostab_pause,'double');
  $mostab_lock = 0;
  &putPV($mostab_lock_pv,$mostab_lock,'double');

  while ($mode == $mPAUSED) {
     $mostab_stop         = &getPV($mostab_stop_pv,'double');
     $mostab_pause        = &getPV($mostab_pause_pv,'double');
     $mostab_status       = &getPV($mostab_status_pv,'double');
     $mostab_opt          = &getPV($mostab_opt_pv,'double');
     $mostab_recall_p2    = &getPV($mostab_recall_p2_pv,'double');
     $mostab_force_search = &getPV($mostab_force_search_pv,'double');
     if ($mostab_stop == 0) {&signal_handler($sbscrpt.' mostab_stop=0');}
     if ($mostab_status == 0) {&status_write($mode);}			# restore status to confirm the script is alive (multiple instances check)
     if ($mostab_force_search == 1) {					#Searching with P2 or DAC is requested
        $mostab_force_search = 0;
        &putPV($mostab_force_search_pv,$mostab_force_search,'double');
        &post_message('Forcing beam search ('.$sbscrpt.')', 'warn');
        $mode = $mSEARCH;						# mode=4 <= beam search
        $mostab_pause = 0;
        &putPV($mostab_pause_pv,$mostab_pause,'double');
        return;
     }
     elsif ($mostab_opt == 1) {						# shift P2 to optimize DAC at center requested
        $mostab_opt = 0;
        &putPV($mostab_opt_pv,$mostab_opt,'double');
        &post_message('Shifting P2 to cntr DAC ('.$sbscrpt.')', 'warn');
        $mode = $mOPTIMP2;
        $mostab_pause = 0;
        &putPV($mostab_pause_pv,$mostab_pause,'double');
        return;
     }
     elsif ($mostab_recall_p2 == 1) {					# recall P2 requested
        $mostab_recall_p2 = 0;
        &putPV($mostab_recall_p2_pv,$mostab_recall_p2,'double');
        &post_message('Recalling P2 ('.$sbscrpt.')', 'warn');
        $mode = $mRECALLP2;
        $mostab_pause = 0;
        &putPV($mostab_pause_pv,$mostab_pause,'double');
        return;
     }
     if ($mostab_pause == 0) {
### This will proceed to  &intensity_check($mINTCHK1) and then $mACTFDBK2 (&fast_feedback)
### if intensity above low threshold, otherwise to $mSEARCH (&search();) 
#       $mode = $mINTCHK1; 
### This will proceed to  &intensity_check($mINTCHK2) and then $mACTFDBK2 (&fast_feedback)
### if intensity above low threshold, otherwise to $mCOMPCHK (&component_check();)
#       $mode = $mINTCHK2; 	        				#2021/09/22 chose instead of $mACTFDBK1 used before, too slow for data collection pausing
### This will proceed to &feedback() directly, which is fastest in case 
### the feedback was paused for data collection. If the later fails due
### to low intensity, then it may go to &component_check() or &search(),
### but we need to start with the fastest option - the behaviour of version 
### 7.20 and prior.
        $mode = $mACTFDBK1;
### This will proceed to &fast_feedback directly, which is not right
#       $mode = $mACTFDBK2;

        ### Issue - what to do after coming out of a paused state, especially
        ### if beam was previously locked? Assume beam is still locked or
        ### re-verify lock? Modification of the following, or modification
        ### of intensity check would provide options.

        #$mostab_lock = &getPV($mostab_lock_pv,'double');
        #if ($mostab_lock) {$mode = $mCOMPCHK;}
        #else              {$mode = $mACTFDBK2;}
     }
     select(undef,undef,undef,$pause_time);
  }
  &post_message('Script resuming.');
}

#==========================================================================
sub no_beam_pause () {

  my ($subscript,$sbscrpt, $ADC_bpm1);

  $sbscrpt = 'no_beam_pause';
  $subscript = $script.'/'.$sbscrpt;
  &log_func_entry($subscript);

  &status_write($mode);
  $mode = &medm_check($mode);						# 18
  &post_message('No beam detected.', 'warn');
  &post_message('Moving to last known piezo DAC value');
  &kill_p2();
  &putPV($DAC_piezo_pv,$DAC_piezo_last,'double');
  $mostab_lock = 0;
  &putPV($mostab_lock_pv,$mostab_lock,'double');

  select(undef,undef,undef,$fdbk_rate);
  $DAC_check_err = &DAC_check(0);
  &post_message('Hibernating.', 'warn');
  &putPV($mostab_pause_pv,1,'double');

  while ($mode == $mNOBEAM) {
     $mostab_stop         = &getPV($mostab_stop_pv,'double');
     $mostab_pause        = &getPV($mostab_pause_pv,'double');
     $mostab_status       = &getPV($mostab_status_pv,'double');
     $mostab_opt          = &getPV($mostab_opt_pv,'double');
     $mostab_recall_p2    = &getPV($mostab_recall_p2_pv,'double');
     $mostab_force_search = &getPV($mostab_force_search_pv,'double');
     if ($mostab_stop == 0) {&signal_handler($sbscrpt.' mostab_stop');}
     if ($mostab_status == 0) {&status_write($mode);}			# restore status to confirm the script is alive (multiple instances check)
     if ($mostab_opt == 1) {
        $ADC_bpm1 = &getPV($ADC_bpm1_pv,'double');
        if ($ADC_bpm1 < $int_low1) {
           &post_message('Intensity too low to optimize P2.', 'error');
           $mostab_opt = 0;
           &putPV ($mostab_opt_pv,$mostab_opt,'double');
        } else {
           $mostab_opt = 0;
           &putPV($mostab_opt_pv,$mostab_opt,'double');
           &post_message('Shifting P2 to cntr DAC ('.$sbscrpt.')', 'warn');
           $mode = $mOPTIMP2;
           $mostab_pause = 0;
           &putPV($mostab_pause_pv,$mostab_pause,'double');
           return;
        }
     }
     elsif ($mostab_force_search == 1) {				#Searching with P2 or DAC is requested
        $mostab_force_search = 0;
        &putPV($mostab_force_search_pv,$mostab_force_search,'double');
        &post_message('Forcing beam search ('.$sbscrpt.')', 'warn');
        $mode = $mSEARCH;						# mode=4 <= beam search
        $mostab_pause = 0;
        &putPV($mostab_pause_pv,$mostab_pause,'double');
        return;
     }
     elsif ($mostab_recall_p2 == 1) {
        $mostab_recall_p2 = 0;
        &putPV($mostab_recall_p2_pv,$mostab_recall_p2,'double');
        &post_message('Recalling P2 ('.$sbscrpt.')', 'warn');
        $mode = $mRECALLP2;
        $mostab_pause = 0;
        &putPV($mostab_pause_pv,$mostab_pause,'double');
        return;
     }
     if ($mostab_pause == 0) {
        $ADC_bpm1 = &getPV($ADC_bpm1_pv,'double');
        if ($ADC_bpm1 < $int_low1) {
           &post_message('Intensity too low to start feedback.', 'error');
           $mostab_pause = 1;
           &putPV ($mostab_pause_pv,$mostab_pause,'double');
        } else {
### This will proceed to  &intensity_check($mINTCHK1) and then $mACTFDBK2 (&fast_feedback)
### if intensity above low threshold, otherwise to $mSEARCH (&search();) 
#          $mode = $mINTCHK1; 
### This will proceed to  &intensity_check($mINTCHK2) and then $mACTFDBK2 (&fast_feedback)
### if intensity above low threshold, otherwise to $mCOMPCHK (&component_check();)
#          $mode = $mINTCHK2;
### This will proceed to &feedback() directly(right since we have intensity?)
#          $mode = $mACTFDBK1; 
### This will proceed to &fast_feedback() directly (right since we have intensity?)
           $mode = $mACTFDBK2;
        }
     }
     select(undef,undef,undef,$pause_time);
  }
}

#==========================================================================
sub measure_intensity(;$) {

  my ($subscript, $sbscrpt, $int1, $int2, $i, $nsamples);
  my ($ADC_bpm1, $ADC_bpm2, $ADC_CAEN, $CAEN_alive, $msg);
  
  $sbscrpt = 'measure_intensity';
  $subscript = $script.'/'.$sbscrpt;
  &log_func_entry($subscript);

  $nsamples = shift (@_);
  if (!defined $nsamples) {$nsamples = 1;}
  if ($nsamples < 1)      {$nsamples = 1;}

  ### Measure intensity
  $int1 = 0;
  $int2 = 0;
  &kill_p2();
  if ($nsamples > 1) {$CAEN_alive = 0;}					#we will check below
  else               {$CAEN_alive = 1;}					#we cannot check
  for ($i=0; $i<$nsamples; $i++) {					#better use Struck instead of multiple ADC samples
     if ($i > 0) {select(undef,undef,undef,0.1);}
     $ADC_bpm1 = &getPV($ADC_bpm1_pv,'double');
     $ADC_bpm2 = &getPV($ADC_bpm2_pv,'double');
     $int1 += $ADC_bpm1;
     $int2 += $ADC_bpm2;
     if ($have_CAEN) {
        if ($i == 0) {$ADC_CAEN = $ADC_bpm1;}
        elsif ($ADC_bpm1 != $ADC_CAEN) {$CAEN_alive = 1;}		#CAEN is alive
     }
  }
  $int1 /= $nsamples;
  $int2 /= $nsamples;
  if ($have_CAEN) {
### Convert to uA (see the CAEN manual: AH501D_UsersManual_V1.3.pdf,
### page 11, 24-bit formula, or search there for "A/D converter").
### FSRx is Full Scale Range, from -2.5uA to +2.5uA, or 5uA.
#    if   ($int1 <= 8388607) {$int1 = -5 * $int1 / 16777215;}			# 16,777,215 = 2^24-1
#    else                    {$int1 =  5 * (16777216 - $int1) / 16777215;}	# 16,777,216 = 2^24
     $int1 = -5 * $int1 / 16777215;						# 16,777,215 = 2^24-1
     $int1 = abs($int1) * (4/$CAEN_nch); 
     if (! $CAEN_alive) {
        $msg = sprintf 'CAEN rd1=%.0f rd%d=%.0f PV=%s',$ADC_CAEN,$nsamples,$ADC_bpm1,$ADC_bpm1_pv;
        &post_message($msg, 'warn');
#       &post_message('CAEN may need restarting...', 'warn');
        &post_message('CAEN stopped. Restarting...', 'warn');
        &putPV($CAEN_acquire_pv,1,'double');
     }
  }  
# print STDOUT 'int(CAEN) = '.$int1."\n";
  &kill_p2();
  return ($int1,$int2);
}

#==========================================================================
sub a_auto_open () {

  my ($subscript, $sbscrpt, $status, $ring_status, $enable_status, $times_up);

  $sbscrpt = 'a_auto_open';
  $subscript = $script.'/'.$sbscrpt;
  &log_func_entry($subscript);

  &status_write($mode);
  &post_message('Ring dump.', 'warn');
  &post_message('Please wait for A-shutter auto open.');
  $mode = &medm_check($mode);
  if ($mode == $mPAUSED) {&post_message('Feedback paused.'); return;}

  $current = &getPV($current_pv,'double');
  if ($current > 50) {$ring_status = 0;}
  else               {$ring_status = 1;}
  $permit = &getPV($permit_pv,'double');
  if ($permit == 1)  {$enable_status = 0;}
  else               {$enable_status = 1;}

  $status = $ring_status + $enable_status;
  while ($status > 0) {
     select(undef,undef,undef,0.05);
     $mode = &medm_check($mode);
     if ($mode == $mPAUSED) {&post_message('Feedback paused.'); return;}
     $current = &getPV($current_pv,'double');
     if ($current > 50) {$ring_status = 0;}
     else               {$ring_status = 1;}
     $permit = &getPV($permit_pv,'double');
     if ($permit == 1)  {$enable_status = 0;}
     else               {$enable_status = 1;}
     $status = $ring_status + $enable_status;
  }

  &putPV($fes_open_pv,1,'double');

  &timeout(0);					#set ini time timeout

  while (1) {
     select(undef,undef,undef,0.05);
     $a_status = &getPV($a_status_pv,'double');
     if (!$a_status) {$mode = $mINTCHK2; return;}
     $times_up = &timeout(10);
     if ($times_up) {
        $mode = $mNOBEAM;
        &post_message('Timeout: A-shutter failed to open', 'error');
        return;
     }
     $mode = &medm_check($mode);
     if ($mode == $mPAUSED) {&post_message('Feedback paused.'); return;}
  }
}

#==========================================================================
# sub post_message ($message_new,['none'|$JbluIce_type) {
sub post_message ($;$) {

  my ($message_new, $jbluice_type, $status, $i);
# my ($subscript, $sbscrpt);

# $sbscrpt = 'post_message';
# $subscript = $script.'/'.$sbscrpt;
# &log_func_entry($subscript);

  $message_new  = shift(@_);   if (!defined $message_new)  {$message_new  = '';}
  $jbluice_type = shift(@_);   if (!defined $jbluice_type) {$jbluice_type = 'note';}

  if (length($message_new) > 0) {
     print STDOUT $message_new."\n";
     if ($jbluice_type !~ /^none$/i) {					#none, note,warn,error,popup_note,popup_error,direct
        &tellBluIce ($scr.': '.$message_new,$jbluice_type);
     }
  }
  for ($i=$nmessage-2; $i>=0; $i--) {
     if (!defined $message[$i]) {$message[$i] = '';}
     $message[$i+1] = $message[$i];
  }
  $message[0] = $message_new;
  for ($i=0; $i<$nmessage; $i++) {
     if (length($message[$i]) < 40) {$status = &putPV($message_pv[$i], $message[$i],'string');}
     else                           {$status = &putPV($message_pv[$i], substr($message[$i],0,39),'string');}
     if ($status) {last;}
  }
  &log_record($message_new);
  if ($message_time > 0) {select(undef,undef,undef,$message_time);}
}

#==========================================================================
sub DAC_check ($) {

  my ($subscript, $sbscrpt, $DAC_error, $CAEN_avertime);
  my ($tol_error, $msg, $message);

  $sbscrpt = 'DAC_check';
  $subscript = $script.'/'.$sbscrpt;
  &log_func_entry($subscript);

  $tol_error = shift (@_);

  if ($tol_DAC_error <= 0) {return 0;}					#no DAC control by reading back (for 23i, $tol_DAC_error=-1, no control)

### The error allowance cannot be smaller than the beamline constant,
### $tol_DAC_error. 
  if ($tol_error < $tol_DAC_error) {$tol_error = $tol_DAC_error;}

  if ($have_CAEN) {
     $CAEN_avertime = &getPV($CAEN_avertime_pv,'double');
     if (abs($CAEN_avertime-$ADC_rate{$ADC_rate_idx}) > 0.001) {
        $CAEN_avertime = $ADC_rate{$ADC_rate_idx};
        &post_message('Changing CAEN avertime='.$CAEN_avertime);
        &putPV($CAEN_acquire_pv,0,'double');				#pause for changing, just in case
        &putPV($CAEN_avertime_pv, $CAEN_avertime,'double');
        &putPV($CAEN_acquire_pv,1,'double');				#resume counting
     }
  } else {
     &putPV($ADC_intrate1_pv, $ADC_rate_idx,'double');
  }
  &putPV($DAC_rdbkrate_pv, $ADC_rate_idx,'double');
  $DAC_piezo = &getPV($DAC_piezo_pv,'double');				#$beamline.':DAC1_'.$DAC_ch      = 23x:DAC1_1 (12 bit DAC)
  $DAC_rdbk  = &getPV($DAC_rdbk_pv,'double');				#$beamline.':ADC_'.$ADC_DACcheck = 23x:ADC_13

  $DAC_error = $DAC_error_scale*($DAC_rdbk - ($DAC_piezo/$DAC_piezo_norm1 + $DAC_piezo_offset))/$DAC_piezo_norm2;

  if (abs($DAC_error) > $tol_error) {
     $msg = sprintf 'DAC/ADC mismatch=%.2fV. ', $DAC_error;  &post_message($msg);  $message  = $msg;
     $msg = 'IOC problem? Wiring? ';                         &post_message($msg);  $message .= $msg;
     $msg = 'Please investigate! ';                          &post_message($msg);  $message .= $msg;
     $message .= '(DAC='.$DAC_piezo_pv.' divided, ADC='.$DAC_rdbk_pv.')';
     &tellBluIce($scr.': '.$message, 'error');
#    return 1;								#SS: commented out 2014/06/09
  }
  return 0;
}

#==========================================================================
sub status_write ($) {

  my ($subscript, $sbscrpt, $write_option);
  my (@ct, $min, $hour, $mday, $mon, $stdate, $sttime, $status);

  $sbscrpt = 'status_write';
  $subscript = $script.'/'.$sbscrpt;
# &log_func_entry($subscript);

  $write_option = shift (@_);

  ### This is ($sec,$min,$hour,$mday,$mon,$year$,$wday,$yday,$isdst)
  @ct   = localtime(time);
  $min  = $ct[1];
  $hour = $ct[2];
  $mday = $ct[3];
  $mon  = ('Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec')[$ct[4]];
  $stdate = sprintf '%02d', $mday;
  $sttime = sprintf '%02d:%02d', $hour,$min;
  $stdate =~ s/ /0/g;
  $sttime =~ s/ /0/g;
  $mostab_sttime = $mon.' '.$stdate.' at '.$sttime;
### $mostab_status: 0=exiting, 1=active, 2=doing.operation, 3=waiting, 4=lost
  if    ($write_option == $mEXIT)     {$mostab_status = 0; $mostab_stline = 'Feedback: exiting';}
  elsif ($write_option == $mHWMON)    {$mostab_status = 1; $mostab_stline = 'Feedback: active, HW monitor';}
  elsif ($write_option == $mACTFDBK1) {$mostab_status = 2; $mostab_stline = 'Initializing active state';}
  elsif ($write_option == $mACTFDBK2) {$mostab_status = 2; $mostab_stline = 'Establishing beam lock';}
  elsif ($write_option == $mINTCHK1)  {$mostab_status = 2; $mostab_stline = 'Performing intensity check';}
  elsif ($write_option == $mINTCHK2)  {$mostab_status = 1; $mostab_stline = 'Feedback: active';}
  elsif ($write_option == $mCOMPCHK)  {$mostab_status = 2; $mostab_stline = 'Performing component check';}
  elsif ($write_option == $mSEARCH)   {$mostab_status = 2; $mostab_stline = 'Searching for beam';}
  elsif ($write_option == $mOPTIMP2)  {$mostab_status = 2; $mostab_stline = 'Optimizing DCM 2nd xtal pitch';}
  elsif ($write_option == $mRECALLP2) {$mostab_status = 2; $mostab_stline = 'Shifting DCM 2nd xtal pitch to MemPos';}
  elsif ($write_option == $mWAITDCM)  {$mostab_status = 2; $mostab_stline = 'Waiting for DCM to finish moving';}
  elsif ($write_option == $mAOPEN)    {$mostab_status = 3; $mostab_stline = 'Waiting to auto-open A-shutter';}
  elsif ($write_option == $mWAITBEAM) {$mostab_status = 3; $mostab_stline = 'Waiting for APS to deliver beam';}
  elsif ($write_option == $mWAITOPNA) {$mostab_status = 3; $mostab_stline = 'Waiting for user to open A-shutter';}
  elsif ($write_option == $mWAITOPNC) {$mostab_status = 3; $mostab_stline = 'Waiting for user to open C-shutter';}
  elsif ($write_option == $mWAITEMIS) {$mostab_status = 3; $mostab_stline = 'Waiting for user to fix ID/DCM energy';}
  elsif ($write_option == $mPAUSED)   {$mostab_status = 3; $mostab_stline = 'Paused; waiting for user start';}
  elsif ($write_option == $mNOBEAM)   {$mostab_status = 4; $mostab_stline = 'Paused; lost beam; intervention needed';}
  $status = 0;
  if ($status == 0) {$status = &putPV($mostab_stline_pv, $mostab_stline,'string');}
  if ($status == 0) {$status = &putPV($mostab_sttime_pv, $mostab_sttime,'string');}
  if ($status == 0) {$status = &putPV($mostab_status_pv, $mostab_status,'double');} 		# MH: moved status to last
  &log_record($mostab_stline);
  return $status;
}

#==========================================================================
sub wait_for_dcm () {

  my ($subscript, $sbscrpt, $del_E, $del_E_threshold, $msg);

  $sbscrpt = 'wait_for_dcm';
  $subscript = $script.'/'.$sbscrpt;
  &log_func_entry($subscript);

  $init_E = &getPV($E_pv,'double');					#added 2021/01
  &status_write($mWAITDCM);
  &post_message('Waiting for DCM');
  $E_mov = &getPV($E_mov_pv,'double');
  while ($E_mov == 0) {
     select(undef,undef,undef,0.05);
     $mostab_stop   = &getPV($mostab_stop_pv,'double');
     $mostab_pause  = &getPV($mostab_pause_pv,'double');
     $mostab_status = &getPV($mostab_status_pv,'double');
     $E_mov         = &getPV($E_mov_pv,'double');
     if ($mostab_stop   == 0)	{&signal_handler($sbscrpt.' mostab_stop');};
     if ($mostab_pause  == 1)	{&pause_feedback();};
     if ($mostab_status == 0)	{&status_write($mWAITDCM);};            # restore status to confirm the script is alive (multiple instances check)
     if ($E_mov == 1) {							# Double check, allow for DCM bounce
        select(undef,undef,undef,1);
        $E_mov = &getPV($E_mov_pv,'double');
     }
  }
  $E = &getPV($E_pv,'double');
  $fin_E = $E;
  $del_E = abs($fin_E - $init_E);
  &post_message('DCM finished.');
  $del_E_threshold = $tol_E;
  if ($del_E > $del_E_threshold) {
     $msg = sprintf $subscript.': E(ini)=%.2f. E(fin)=%.2f, dE=%.2f, dE_max=%.2f',
                                  $init_E,     $fin_E,      $del_E,  $del_E_threshold;
     &log_record($msg);
     $mode = $mCOMPCHK;
     &post_message('Energy change large.');
     &post_message('Checking components.');
  }
  else {
     &post_message('Continuing with feedback.');
     &status_write($mINTCHK2);
     $mode = $mACTFDBK1;
  }
}

#==========================================================================
sub optimize_P2 () {

  my ($subscript, $sbscrpt, $msg, $to);
  my ($FPE_start, $FPE_exposure, $BI_fdbk, $wait_, $wait_max);
  my ($ADC_bpm1);

  $sbscrpt = 'optimize_P2';
  $subscript = $script.'/'.$sbscrpt;
  &log_func_entry($subscript);

  $mostab_lock = 0;
  &putPV($mostab_lock_pv,$mostab_lock,'double');
  &status_write($mode);
  &post_message('Optimizing P2.');
  &post_message(' -- Unpausing itself.');
  $mostab_pause = 0;							#switch to "0=Run
  &putPV($mostab_pause_pv,$mostab_pause,'double');

  ### Wait for JBluIce ---------------------------------------
  $mode = &medm_check($mode);						# here mode=$mOPTIMP2
  select(undef,undef,undef,1);						# Wait - just in case too many things happen at once
  $mode = &medm_check($mode);
  &post_message(' -- Checking JBluIce status');
  $FPE_start = &getPV($FPE_start_pv,'double');
  $FPE_exposure = &getPV($FPE_exposure_pv,'double');
  $BI_fdbk = &getPV($BI_fdbk_pv,'double');
  if ($FPE_start) {
     &post_message('FPE active: taking frame', 'warn');
     if ($BI_fdbk == 0) {
        &post_message('JBluIce is not checking feedback!', 'warn');
        &post_message('Pausing', 'warn');
        $mode = $mPAUSED;
        return;
     }
     $wait_ = $FPE_exposure + 1;
     &timeout(0);
     &post_message('Waiting for frame to finish');
     while ($FPE_start) {
        $mode = &medm_check($mode);
        if ($mode == $mPAUSED) {return;}
        $to = &timeout($wait_);
        if ($to) {
           &post_message('Unexpected FPE delay', 'warn');
           &post_message('Pausing', 'warn');
           $mode = $mPAUSED;
           return;
        }
        $FPE_start = &getPV($FPE_start_pv,'double');
     }
  } else {
     &post_message('FPE inactive: no frame is being taken');
  }
  $mode = &medm_check($mode);
  select(undef,undef,undef,1);						# Wait to be sure it does not become active
  $mode = &medm_check($mode);
  $FPE_start = &getPV($FPE_start_pv,'double');
  if ($FPE_start) {
     &post_message('Warning: JBluIce is not responding to feedback!', 'warn');
     &post_message('Feedback is pausing!', 'warn');
     $mode = $mPAUSED;
     return;
  }
  ### End of waiting for JBluIce -----------------------------

  &post_message(' -- Reading intensity (BPM1)');
  $ADC_bpm1 = &getPV($ADC_bpm1_pv,'double');
  if ($ADC_bpm1 < $int_ok1) {
     &post_message('Insufficient intensity.', 'warn');
     &post_message('Optimization not possible.', 'warn');
     return;
  }
  $msg = sprintf ' -- Intensity=%.2f',$ADC_bpm1;
  &post_message($msg);

  &post_message(' -- Closing P2 loop');
  &closeloop_p2();							# Close loop at current p2 position
  &post_message(' -- Driving DAC to center='.int($DAC_piezo_CNTR).'V');
  &putPV($DAC_piezo_pv,$DAC_piezo_CNTR,'double');			# Move piezo to the middle => affects motor encoder

  ### Wait for inpos: this uses PMAC PID loop to
  ### compensate for the DAC change to 5V:
  &post_message('Please wait for P2 to position.');
  $wait_max = 30;
  $wait_ = &waitMotor ($p2_assy_pv,$wait_max);				# wait for motor to compensate position change
  if ($wait_ == 0) {&func_error(1,'waitMotor',$p2_assy_pv.','.$wait_max,'continue');}
  &kill_p2(1);								# Kill p2 motor

  $mode = &medm_check($mode);
  ### Check for beam
  &post_message(' -- Checking intensity (BPM1)');
  $ADC_bpm1 = &getPV($ADC_bpm1_pv,'double');
  $msg = sprintf ' -- Intensity=%.2f',$ADC_bpm1;
  &post_message($msg);
  if ($ADC_bpm1 < $int_ok1) {
     &post_message('Beam lost during optimization.', 'warn');
     &post_message('Human intervention needed.', 'warn');
     $mode = $mPAUSED;
     return;
  }
  ### Optimization complete
  $mode = $mINTCHK2;
  &post_message('P2 optimization complete.');
}

#==========================================================================
sub recall_P2 () {

  my ($subscript, $sbscrpt, $msg, $to);
  my ($FPE_start, $FPE_exposure, $BI_fdbk, $wait_, $wait_max);
  my ($ADC_bpm1);

  $sbscrpt = 'recall_P2';
  $subscript = $script.'/'.$sbscrpt;
  &log_func_entry($subscript);

  &status_write($mode);
  &post_message('MODE: Recalling P2 MemPos.');

  ### Wait for JBluIce ---------------------------------------
  $mode = &medm_check($mode);						# here mode=$mRECALLP2
  select(undef,undef,undef,1);						# Wait - just in case too many things happen at once
  $mode = &medm_check($mode);
  &post_message(' -- Checking JBluIce status');
  $FPE_start = &getPV($FPE_start_pv,'double');
  $FPE_exposure = &getPV($FPE_exposure_pv,'double');
  $BI_fdbk = &getPV($BI_fdbk_pv,'double');
  if ($FPE_start) {
     &post_message('FPE active: taking frame', 'warn');
     if ($BI_fdbk == 0) {
        &post_message('JBluIce is not checking feedback!', 'warn');
        &post_message('Pausing', 'warn');
        $mode = $mPAUSED;
        return;
     }
     $wait_ = $FPE_exposure + 1;
     &timeout(0);
     &post_message('Waiting for frame to finish');
     while ($FPE_start) {
        $mode = &medm_check($mode);
        if ($mode == $mPAUSED) {&post_message('Feedback paused.'); return;}
        $to = &timeout($wait_);
        if ($to) {
           &post_message('Unexpected FPE delay', 'warn');
           &post_message('Pausing', 'warn');
           $mode = $mPAUSED;
           return;
        }
        $FPE_start = &getPV($FPE_start_pv,'double');
     }
  } else {
     &post_message('FPE inactive: no frame is being taken');
  }
  $mode = &medm_check($mode);
  select(undef,undef,undef,1);						# Wait to be sure it does not become active
  $mode = &medm_check($mode);
  $FPE_start = &getPV($FPE_start_pv,'double');
  if ($FPE_start) {
     &post_message('Warning: JBluIce is not responding to feedback!', 'warn');
     &post_message('Feedback is pausing!', 'warn');
     $mode = $mPAUSED;
     return;
  }
  ### End of waiting for JBluIce -----------------------------

  &post_message(' -- Closing P2 loop');
  &closeloop_p2();							# Close loop at current p2 position
  &post_message(' -- Recalling P2 Mempos');
  &putPV($p2_recall_pv,1,'double','old');				# Move P2 to MemPos. Should HW feedback be paused?
  &post_message(' -- Waiting P2 to position');
  $wait_max = 5;
  $wait_ = &waitMotor ($p2_assy_pv,$wait_max);				# wait for motor to compensate position change
  if ($wait_ == 0) {&func_error(1,'waitMotor',$p2_assy_pv.','.$wait_max,'continue');}
  &kill_p2(1);								# Kill p2 motor

  $mode = &medm_check($mode);
  ### Check for beam
  &post_message(' -- Checking intensity (BPM1)');
  $ADC_bpm1 = &getPV($ADC_bpm1_pv,'double');
  $msg = sprintf ' -- Intensity=%.2f',$ADC_bpm1;
  &post_message($msg);
  if ($ADC_bpm1 < $int_ok1) {
     &post_message('Beam lost during optimization.', 'warn');
     &post_message('Human intervention needed.', 'warn');
     $mode = $mPAUSED;
     return;
  }
  ### Optimization complete
  $mode = $mINTCHK2;
  &post_message('Recallig P2 MemPos complete.');
}

#==========================================================================
sub bpm2 () {

  ### This sub decides which bpm signal to use for feedback
  ### Return options:
  ###    0 - use bpm1, by choice
  ###    1 - use bpm1, because bpm2 isn't ready (available)
  ###    2 - use bpm2, because all checks satisfied
  ###    3 - bpm2 recently available, optimize now on bpm2
  ###    4 - both signals bad; lost beam

  my ($subscript, $sbscrpt, $ch_ok, $msg);
  my ($ADC_bpm1, $ADC_bpm2);

  $sbscrpt = 'bpm2';
  $subscript = $script.'/'.$sbscrpt;
  &log_func_entry($subscript);

  ### Question 1 - Is the intensity on bpm1 sufficient?  If not, beam is lost.
  $ADC_bpm1 = &getPV($ADC_bpm1_pv,'double');
  if ($ADC_bpm1 < $int_ok1) {
     $msg = sprintf 'bpm1=%.2f < min=%.2f, low signal, PV=%s',$ADC_bpm1,$int_ok1,$ADC_bpm1_pv;
     &post_message($msg);
     return (4);
  }

  ### Question 2 - Does the user want a second bpm at all?  If no, there's
  ### no reason to continue.
  $mostab_use2 = &getPV($mostab_use2_pv,'double');
  if ($mostab_use2 == 0) {
     $mostab_inuse2 = 0;
     $last_use2 = $mostab_use2;
     &putPV($mostab_inuse2_pv,$mostab_inuse2,'double');
     return (0);
  }

  ### Question 3 - Should bpm2 really be used?
  #if (($beamline eq '23i') || ($beamline eq '23b')) {
  if ($beamline eq '23b') {
     $mostab_use2 = 0;
     &putPV($mostab_use2_pv,$mostab_use2,'double');
     &post_message('Beamline not configured for bpm2 use.', 'warn');
     $mostab_inuse2 = 0;
     &putPV($mostab_inuse2_pv,$mostab_inuse2,'double');
     return (0);
  }
  elsif ($lockin_use != 0) {
     $mostab_use2 = 0;
     &putPV($mostab_use2_pv,$mostab_use2,'double');
     &post_message('bpm2 works with software feedback only.', 'warn');
     $mostab_inuse2 = 0;
     &putPV($mostab_inuse2_pv,$mostab_inuse2,'double');
     return (0);
  }

  ### Question 4 - Is the mbsh open?  If not, use bpm1.
  $mbsh = &getPV($mbsh_pv,'double');		# Note mbsh open = 0
  if ($mbsh) {			#if MBSH closed
     $last_mbsh = $mbsh;
     $mostab_inuse2 = 0;
     &putPV($mostab_inuse2_pv,$mostab_inuse2,'double');
     &post_message('MBSH closed. Using bpm1.', 'warn');
     return (1);
  }

  ### Question 5 - Has the mbsh opened recently?  If so, optimize on bpm2.
  #if (!$mbsh && $last_mbsh) {
  #   $last_mbsh = $mbsh;
  #   $mostab_inuse2 = 2;
  #   &putPV($mostab_inuse2_pv,$mostab_inuse2,'double');
  #   return (3);
  #}

  ### Question 6 - Which channel does the user wish to use for bpm2?  If
  ### not on acceptable list, use default.  If new choice, re-optimize.
  $mostab_ch2  = &getPV($mostab_ch2_pv,'double');
  $ch_ok = 0;
  foreach (@ADC_chs2) {
     if ($_ == $mostab_ch2) {$ch_ok = 1;}
  }
  if (!$ch_ok) {
     $mostab_ch2 = $ADC_chs2[0];
     &putPV($mostab_ch2_pv,$mostab_ch2,'double');
     &post_message('Bad bpm2 channel. Using default.', 'warn');
  }
  &putPV($ADC_intrate2_pv, $ADC_rate_idx,'double');
  #if ($last_ch2 != $mostab_ch2) {
  #   $last_ch2 = $mostab_ch2;
  #   &post_message('New bpm2 choice.');
  #   return (3);
  #}

  ### Question 7 - Is the intensity on bpm2 sufficient?  If not, use bpm1.
  $ADC_bpm2 = &getPV($ADC_bpm2_pv,'double');
  if ($ADC_bpm2 < $int_ok2) {
     $mostab_inuse2 = 1;
     $last_int2_is_ok = 0;
     &putPV($mostab_inuse2_pv,$mostab_inuse2,'double');
     &post_message('Low bpm2 signal. Using bpm1.', 'warn');
     return (1);
  } else {$int2_is_ok = 1;}

  ### Question - Have any conditions changed that call for re-optimization?
  if ((!$mbsh && $last_mbsh)          ||
      ($last_use2    != $mostab_use2) ||
      ($last_ch2     != $mostab_ch2)  ||
      ($last_int2_is_ok != $int2_is_ok)) {
     $last_mbsh = $mbsh;
     $last_use2 = $mostab_use2;
     $last_ch2 = $mostab_ch2;
     $last_int2_is_ok = $int2_is_ok;
     $mostab_inuse2 = 2;
     &putPV($mostab_inuse2_pv,$mostab_inuse2,'double');
     &post_message('bpm changed, re-optimizing.');
     return (3);
  }

  ### Question 8 - Have you made it this far?  Optimize on bpm2
  $last_mbsh = $mbsh;
  $last_use2 = $mostab_use2;
  $last_ch2 = $mostab_ch2;
  $last_int2_is_ok = $int2_is_ok;
  $mostab_inuse2 = 2;
  &putPV($mostab_inuse2_pv,$mostab_inuse2,'double');
  return (2);

}

#==========================================================================
sub kill_p2 (;$) {

  my ($subscript, $sbscrpt, $p2_amp, $status, $kill_str);

  $sbscrpt = 'kill_p2';
  $subscript = $script.'/'.$sbscrpt;
  &log_func_entry($subscript);

  $p2_amp = shift(@_);

  if (!defined $p2_amp) {
     $status = 0;
     $p2_amp = &checkMotorAmp ($p2_axis_pv,0,\$status);
     &func_error($status,'checkMotorAmp',$p2_axis_pv,'continue');
  }
  if ($p2_amp) {
     $kill_str = 'K';
     $status = &sendMtrCmd($p2_mtr_pv,$kill_str);			# kill p2 motor
     &func_error($status,'sendMtrCmd',$p2_mtr_pv.','.$kill_str,'pause');
  }
}

#==========================================================================
sub closeloop_p2 () {

  my ($subscript, $sbscrpt, $stop_str, $status);

  $sbscrpt = 'closeloop_p2';
  $subscript = $script.'/'.$sbscrpt;
  &log_func_entry($subscript);

  $stop_str = 'J/';							# Turn ON amp & Close PID loop
  $status = &sendMtrCmd($p2_mtr_pv,$stop_str);
  &func_error($status,'sendMtrCmd',$p2_mtr_pv.','.$stop_str,'pause');
# &abortAssy($p2_assy_pv);						# Close loop at current p2 position
# &func_error($status,'abortAssy',$p2_assy_pv,'pause');
}

#==========================================================================
sub calc_fwhm_x_E () {
  $hkl = &getMonoHKL();
  if ($hkl ne $hkl_old) {
     &post_message('Switching '.$hkl_old.' -> '.$hkl); 
     $hkl_old = $hkl;
  }
  if ($hkl eq '333') {return $fwhm_x_E333;}
  else               {return $fwhm_x_E111;}
}

#==========================================================================
sub putPV ($$;$$) {

  my ($subscript, $sbscrpt, $pv2change, $new_pv_value);
  my ($ca_error, $pv_type, $put_type, $msg);

  $sbscrpt = 'putPV';
  $subscript = $script.'/'.$sbscrpt;
# &log_func_entry($subscript);

  # putPV($pv2change,$new_pv_value,$pv_type)
  #
  # $pv_type => string | double
  #
  # This sub not created for arrays.

  $pv2change    = shift (@_);
  $new_pv_value = shift (@_);
  $pv_type      = shift (@_);	if (!defined $pv_type)  {$pv_type  = 'double';}
  $put_type     = shift (@_);	if (!defined $put_type) {$put_type = 'normal';}

  if (!defined $pv2change) {
     &log_record($sbscrpt.': NO PV specified!');
     $msg = &call_trace();
     &log_record($sbscrpt.': '.$msg);
     return -1;
  }
  if    ($pv_type =~ /^double$/i) {					#number
     if ($put_type !~ /^old/i) {$ca_error = Pezca::PutDouble($pv2change,$new_pv_value);}
     else                      {$ca_error = Pezca::PutDoubleOldCa($pv2change,$new_pv_value);}
  }
  elsif ($pv_type =~ /^string$/i) {					#string
     if ($put_type !~ /^old/i) {$ca_error = Pezca::PutString     ($pv2change,substr($new_pv_value,0,$maxStrLen));}
     else                      {$ca_error = Pezca::PutStringOldCa($pv2change,substr($new_pv_value,0,$maxStrLen));}
  }
  elsif ($pv_type =~ /^wavestr/i) {					#wavestring
     if ($put_type !~ /^old/i) {$ca_error = &PezcaPutStringWF($pv2change,$new_pv_value);}
     else                      {$ca_error = &PezcaPutStringWF($pv2change,$new_pv_value,'oldca');}
  }
  else {								#unknown
     $ca_error = -1; &log_record($subscript.': Incorrect PVtype='.$pv_type);
  }
  if ($ca_error != 0) {
     &report_only($subscript, $ca_error, $pv2change);
     &caput_error($pv2change,$new_pv_value,$pv_type,$ca_error);
  }
  return $ca_error;
}

#==========================================================================
sub getPV ($;$) {

  my ($subscript, $sbscrpt, $pv2get, $pv_type, $ca_error, $pv_value, $msg);

  $sbscrpt = 'getPV';
  $subscript = $script.'/getPV';
# &log_func_entry($subscript);

  # ($pv_value) = getPV($pv2get,$pv_type)
  #
  # $pv_type => string | double
  #
  # This sub not created for arrays.

  $pv2get  = shift(@_);
  $pv_type = shift(@_);		if (!defined $pv_type) {$pv_type = 'double';}

  if (!defined $pv2get) {
     &log_record($sbscrpt.': NO PV specified!');
     $msg = &call_trace();
     &log_record($sbscrpt.': '.$msg);
     return 0;
  }
  if    ($pv_type =~ /^double$/i) {($ca_error,$pv_value) = Pezca::GetDouble($pv2get);}			#double
  elsif ($pv_type =~ /^string$/i) {($ca_error,$pv_value) = Pezca::GetString($pv2get);}			#wave
  elsif ($pv_type =~ /^wavestr/i) {($ca_error,$pv_value) = &PezcaGetStringWF($pv2get);}			#wavestring
  else                            {($ca_error,$pv_value) = (-1,0); &log_record($sbscrpt.': Incorrect PVtype='.$pv_type);}
  if ($ca_error != 0) {$pv_value = &caget_error($pv2get,$pv_type,$ca_error);}
  return $pv_value;
}

#==========================================================================
sub func_error ($$$;$) {

  my ($subscript, $sbscrpt, $error, $func, $error_string, $action);
  my (@msg, $status, $i);

  $sbscrpt = 'func_error';
  $subscript = $script.'/func_error';
# &log_func_entry($subscript);

  $error        = shift(@_);
  $func         = shift(@_);
  $error_string = shift(@_);
  $action       = shift(@_);		if (!defined $action) {$action='exit';}		#can be: exit, pause, continue

  if ($error == 0) {return;}

  $current_time = scalar localtime;

  $msg[0] = 'Error='.$error.' from function '.$func;
  $msg[1] = '';
  $msg[2] = 'Param/msg: '.$error_string;
  $msg[3] = '';
  if    ($action eq 'exit')  {$msg[4] = 'Exiting...';}
  elsif ($action eq 'pause') {$msg[4] = 'Pausing...';}
  else                       {$msg[4] = 'Continuing...';}

  $status = 0;
  for ($i=0; $i<5; $i++) {
     if ($status == 0) {
        $status = Pezca::PutString($message_pv[$i], substr($msg[$i],0,$maxStrLen));
        &report_only($subscript,$status,$message_pv[$i],$LOGLUN);
     }
     &log_record($msg[$i]);
  }
  if    ($action eq 'exit')  {&signal_handler($sbscrpt.' error from '.$func);}
  elsif ($action eq 'pause') {$mode = $mPAUSED; return;}
  else                       {return;}
}

#==========================================================================
# This is called when getPV receives EZCA error.
# Usage: $pv_value = &caget_error($pv2get,$pv_type,$ca_error)
sub caget_error ($;$$) {

  my ($subscript, $sbscrpt, $error_pv, $pv_type, $ca_error);
  my ($error, $pv_value, $times_up, @msg, $status, $i);

  $sbscrpt = 'caget_error';
  $subscript = $script.'/'.$sbscrpt;
# &log_func_entry($subscript);

  $error_pv = shift(@_);
  $pv_type  = shift(@_);	if (!defined $pv_type)  {$pv_type = 'double';}
  $ca_error = shift(@_);	if (!defined $ca_error) {$ca_error = 0;}

  if (defined $sigcount) {return 0;}

  $current_time = scalar localtime;
  $error = 1;
  $status = 0;
  for ($i=0; $i<$nmessage; $i++) {
     if ($error_pv eq $message_pv[$i]) {$status = 1; return '';}		# do not resport errors on messages
  }

  if ($ca_error >= 0) {
     $msg[0] = 'Error reading PV!';
     $msg[1] = $error_pv;
     $msg[2] = '';
     $msg[3] = 'If communication is not re-established,';
     $msg[4] = 'feedback will terminate in 5 minutes.';
  } else {
     $msg[0] = 'Incorrect PVtype='.$pv_type.'!';
     $msg[1] = $error_pv;
     $msg[2] = '';
     $msg[3] = '';
     $msg[4] = '';
  }

  for ($i=0; $i<5; $i++) {
     if ($status == 0) {
        $status = Pezca::PutString($message_pv[$i], substr($msg[$i],0,$maxStrLen));
        &report_only($subscript,$status,$message_pv[$i],$LOGLUN);
     }
     &log_record($msg[$i]);
  }

  &timeout(0);								# set ini time for timeout
  while ($error != 0) {
     for ($i=0; $i<100; $i++) {select(undef,undef,undef,0.1);}	# sleep 10s
     if ($pv_type == 0) {($error,$pv_value) = Pezca::GetString($error_pv);}
     else               {($error,$pv_value) = Pezca::GetDouble($error_pv);}
     &report_only($subscript,$error,$error_pv,$LOGLUN);
     $times_up = &timeout($t_max);
     if ($times_up) {
        $current_time = scalar localtime;
        $msg[0] = 'CA timeout!';
        if ($status == 0) {
           $status = Pezca::PutString($message_pv[0], substr($msg[0],0,$maxStrLen));
           &report_only($subscript,$status,$message_pv[0],$LOGLUN);
        }
        &log_record($msg[0]);
        &signal_handler($sbscrpt);
     }
  }
  &post_message('');
  &post_message('CA error: '.$error_pv, 'error');
  &post_message('Communication lost and re-established.', 'warn');
  return($pv_value);
}

#==========================================================================
sub caput_error ($$;$$) {

  my ($subscript, $sbscrpt, $error_pv, $pv_type, $ca_error);
  my ($error, $pv_value, $times_up, @msg, $status, $i);

  $sbscrpt = 'caput_error';
  $subscript = $script.'/'.$sbscrpt;
# &log_func_entry($subscript);

  $error_pv = shift(@_);
  $pv_value = shift(@_);
  $pv_type  = shift(@_);	if (!defined $pv_type)  {$pv_type = 'double';}
  $ca_error = shift(@_);	if (!defined $ca_error) {$ca_error = 0;}

  if (defined $sigcount) {return;}

  $current_time = scalar localtime;
  $error = 1;
  $status = 0;
  for ($i=0; $i<$nmessage; $i++) {
     if ($error_pv eq $message_pv[$i]) {$status = 1; return;}		# do not resport errors on messages
  }

  if ($ca_error >= 0) {
     $msg[0] = 'Error writing PV!';
     $msg[1] = $error_pv;
     $msg[2] = '';
     $msg[3] = 'If communication is not re-established,';
     $msg[4] = 'feedback will terminate in 5 minutes.';
  } else {
     $msg[0] = 'Incorrect PVtype='.$pv_type.'!';
     $msg[1] = $error_pv;
     $msg[2] = '';
     $msg[3] = '';
     $msg[4] = '';
  }

  for ($i=0; $i<5; $i++) {
     if ($status == 0) {
        $status = Pezca::PutString($message_pv[$i], substr($msg[$i],0,$maxStrLen));
        &report_only($subscript,$status,$message_pv[$i],$LOGLUN);
     }
     &log_record($msg[$i]);
  }

  &timeout(0);								# set ini time for timeout
  while ($error != 0) {
     for ($i=0; $i<100; $i++) {select(undef,undef,undef,0.1);}		#sleep 10s
     if    ($pv_type =~ /^string$/i) {$error = Pezca::PutString($error_pv,substr($pv_value,0,$maxStrLen));}
     elsif ($pv_type =~ /^double$/i) {$error = Pezca::PutDouble($error_pv,$pv_value);}
     else {
        &log_record('*** Incorrect PV type="'.$pv_type.'" for PV="'.$error_pv.'".');
        &signal_handler($sbscrpt);
     }
     &report_only($subscript,$error,$error_pv,$LOGLUN);
     $times_up = &timeout($t_max);
     if ($times_up) {
        $current_time = scalar localtime;
        $msg[0] = 'CA timeout!';
        if ($status == 0) {
           $status = Pezca::PutString($message_pv[0], substr($msg[0],0,$maxStrLen));
           &report_only($subscript,$status,$message_pv[0],$LOGLUN);
        }
        &log_record($msg[0]);
        &signal_handler($sbscrpt);
     }
  }
  &post_message('');
  &post_message('CA error: '.$error_pv, 'error');
  &post_message('Communication lost and re-established.', 'warn');
}

#==========================================================================
sub timeout ($) {

  my ($subscript, $sbscrpt, $max_time);

  $sbscrpt = 'timeout';
  $subscript = $script.'/'.$sbscrpt;
# &log_func_entry($subscript);

  $max_time = shift(@_);

  if ($max_time == 0) {							#set ini time for timeout
     $init_time = [gettimeofday];
     select(undef,undef,undef,1);
     return 0;
  } else {
     if (tv_interval($init_time) > $max_time) {return 1;}		# timeout yes
     else                                     {return 0;}		# timeout no
  }
}

#==========================================================================
sub getPV_eh ($;$) {

### In the hybrid mode this function returns a voltage step calculated by LIA
### which can be applied to the mirror piezo after scaling with $multiplier_LIA.
###
### In the hardware feedback mode this function returns absoluted DAC calculated
### by LIA, which can be scaled to the [0-10]V range after multiplying by
###  $scale0to10_LIA.

  my ($subscript, $sbscrpt, $lockin_asyn_pv, $corr_flag, $value, $msg);

  $sbscrpt = 'getPV_eh';
  $subscript = $script.'/'.$sbscrpt;
# &log_func_entry($subscript);

  $lockin_asyn_pv = shift(@_);
  $corr_flag      = shift(@_);
  if (!defined $corr_flag) {$corr_flag = 1;}

  ### request lock-in signal
  #Pezca::SetTimeout(0.2); &putPV($lockin_asyn_pv.'.AOUT','OUTP?1','string'); Pezca::SetTimeout($ezcatimeout_default);

  ### read lock-in signal and add to piezo control DAC
  $value = &getPV($lockin_asyn_pv.'.TINP','string');
  if (! &is_float($value)) {
     $msg = 'WARNING: incorrect lockin response=['.$value.']';
     &post_message($msg, 'warn');
     print STDOUT $msg.', PV='.$lockin_asyn_pv.'.TINP, expecting float'."\n";
     if (length($value) > 0 || $corr_flag) {$value = 0;}
     else                                  {$value = '';}
  }

  return $value;
}

#==========================================================================
sub get_lockin_DAC_step ($) {

  my ($subscript, $sbscrpt, $lockin_asyn_pv, $value);

  $sbscrpt = 'get_lockin_DAC_step';
  $subscript = $script.'/'.$sbscrpt;
# &log_func_entry($subscript);

  $lockin_asyn_pv = shift(@_);

  if (defined $multiplier_LIA) {
     $value = &getPV_eh ($lockin_asyn_pv);
     $value = $multiplier_LIA * $value;
  } else {
     &post_message('NO LIA, switching to SW feedback');
     $lockin_use = 0;
     &putPV($lockin_use_pv,$lockin_use,'double');
     $value = 0;
  }
  return $value;
}

#==========================================================================
sub get_lockin_DAC_abs ($) {			#THIS FUNCTION IS NOT USED

  my ($subscript, $sbscrpt, $lockin_asyn_pv, $value);

  $sbscrpt = 'get_lockin_DAC_abs';
  $subscript = $script.'/'.$sbscrpt;
# &log_func_entry($subscript);

  $lockin_asyn_pv = shift(@_);

  $value = &getPV_eh ($lockin_asyn_pv);
  $value = $scale0to10_LIA * $value;

  return $value;
}

#==========================================================================
sub get_lockin_ADC ($) {

  my ($subscript, $sbscrpt, $lockin_ADC_pv, $value);

  $sbscrpt = 'get_lockin_ADC';
  $subscript = $script.'/'.$sbscrpt;
# &log_func_entry($subscript);

  $lockin_ADC_pv = shift(@_);

  $value = &getPV_eh ($lockin_ADC_pv);
  if (length($value) == 0) {$value = 5;}				#some center voltage

  return $value;
}

#==========================================================================
sub close_piezo_loop () {
  my ($subscript, $sbscrpt, $closeloop);

  $sbscrpt   = 'close_piezo_loop';
  $subscript = $script.'/'.$sbscrpt;
  $closeloop = 1;				#0=Servo Open, 1=Servo Close
  if (defined $piezo_servoState_pv) {
     &post_message('Closing the piezo controller loop (turning ON strain gauge)');
     &putPV($piezo_servoState_pv,$closeloop,'double');
  }
}

#==========================================================================
sub multiInstanceHostProtector () {

### Multi-instance-host protection variables
my (@hosts_allowed, $host, $cmd, @processes, $nprocesses, $msg, $myname);
my ($sshcmd, @names, $nme, $hst, $filter, $mypid);

### Protection from running on multiple computers:
  $host = $ENV{HOSTNAME};
  if (!defined $host) {$host=`hostname`; $host=~s/\n//g;}
  if (!defined $host) {print STDOUT "!!! HOSTNAME is undefined!\n\n"; $host='undef';}
  $host = lc($host);
  $host =~ s/\.gmca\.aps\.anl\.gov$//;
  if    ($beamline eq '23i') {@hosts_allowed = qw(bl1ws2 bl1ws3);}
  elsif ($beamline eq '23o') {@hosts_allowed = qw(bl2ws2 bl2ws3);}
  elsif ($beamline eq '23b') {@hosts_allowed = qw(bl3ws2 bl3ws3);}
  elsif ($beamline eq '23d') {@hosts_allowed = qw(www px0 sergey x-server);}
  else                       {@hosts_allowed = 'undefined';}

  if (! grep(/^$host$/, @hosts_allowed)) {
     $msg = '*** '.$script.': cannot run on '.$host."\n"
          . 'Host(s) allowed for beamline '.$beamline.': '."@{hosts_allowed}";
     &zenity_error_exit ($msg);
  }

  $sshcmd = '/usr/bin/ssh -aTxq -o BatchMode=yes -o StrictHostKeyChecking=no ';
  $myname = &makeScriptName();
  @names  = ($myname, 'DCM_P2_LIA_feeback.*\.py');
  $filter = '/bin/ps -eo user:19,pid,cmd '
          .'| /bin/egrep -iv "grep|edit|geany|nano|emacs|term|onsole|bash|csh|zenity|pezca23id|defunct" '
          .'| /bin/egrep ';

#----------------------------------------------------------------
### Protection from running multiple instances:
  if ($^O !~ /Win/i) {
     foreach $hst (@hosts_allowed) {
        if ($hst ne $host) {
           ### Check connection with surely working command:
           $cmd = $sshcmd.$hst.' date';
           $msg = `$cmd`;
           if (${^CHILD_ERROR_NATIVE} != 0) {
              $msg = '*** '.$script.' command:'."\n".$cmd."\n".'failed with code='.${^CHILD_ERROR_NATIVE}.'.';
              if ($cmd =~ /ssh /) {$msg .= "\n".'Please verify passwordless ssh connection to '.$hst;}
              &zenity_error_exit ($msg);
           }
        }
        foreach $nme (@names) {
           $cmd = $filter.'"'.$nme.'"';
           if ($hst ne $host) {$cmd = $sshcmd.$hst.' '.$cmd;}
#          print STDOUT $script.': executing:'."\n".$cmd."\n";
           $msg = `$cmd`;
### grep returns "1" (translated to "256") when does not find anythiog; so it is not an error
           if (${^CHILD_ERROR_NATIVE} != 0 && ${^CHILD_ERROR_NATIVE} != 256) {
              $msg = '*** '.$script.' command:'."\n".$cmd."\n".'failed with code='.${^CHILD_ERROR_NATIVE}.'.';
              &zenity_error_exit ($msg);
           }
#          print STDOUT $script.': result:'."\n".$msg."\n";
           @processes = split(/\n/, $msg);
           if ($nme eq $myname) {
              $mypid = $$;
              @processes = grep(! / $mypid /,@processes);
           }
           $nprocesses = @processes;
#          print STDOUT $script.': nprocesses = '.$nprocesses."\n";
           if ($nprocesses == 1) {
              $msg =  '*** '.$script.': an instance of "'.$nme.'" is already running on host='.$hst.':'."\n\n".$msg;
              &zenity_error_exit ($msg);
           }
           if ($nprocesses > 1) {
              $msg =  '*** '.$script.': '.$nprocesses.' instances of "'.$nme.'" are already running on host='.$hst.':'."\n\n".$msg;
              &zenity_error_exit ($msg);
           }
        }
     }
  }
}

#==========================================================================
sub signal_handler (;$) {

  our ($ABORT_IN_PROCESS);
  my ($subscript, $sbscrpt, $signame, $status, $msg);

  $sbscrpt = 'signal_handler';
  $subscript = $script.'/'.$sbscrpt;
  &log_func_entry($subscript);
  $mostab_show_flow = 0;
  $current_time = scalar localtime;

  $signame = shift(@_);
  if (!defined $signame) {$signame = '';}
  if (length($signame) > 0) {
     $msg = "\n".$subscript.' "'.$signame.'" received.'
  } else {
     $msg = "\n".$subscript.' called.'
  }
  &log_record($msg);
  print STDOUT $msg."\n";

  if (defined $ABORT_IN_PROCESS) {return;}
  $ABORT_IN_PROCESS = 1;
  &abort_external_script();

  if (!defined $sigcount) {$sigcount=1;}
  else                    {$sigcount++;}
  if ($sigcount < 2) {
     if (!defined $lockin_use) {$lockin_use=0;}
     if ($lockin_use > 1) {$lockin_use=1; &putPV($lockin_use_pv,$lockin_use,'double');}
     if (defined $scaler) {&stop_scaler();}
     &clearAllCaMonitors($LOGLUN);
     if ($signame ne 'DUPLICATE' && defined $mostab_status_pv) {&putPV($mostab_status_pv.'.DESC','*none*','string');}
     $status = &status_write($mEXIT);
     if ($status == 0 && defined $mostab_lock_pv) {$mostab_lock=0; $status = &putPV($mostab_lock_pv,$mostab_lock,'double');}
     if ($status == 0) {&post_message('');  &post_message('Exiting.');}
  }
  &log_record("\n".$script.' stopped.');
  if (defined $filehandle) {close(DAT);}
  exit;
}

### ---------------------------------------------------------------
### To-do:
###
### Modify DAC check to quit when there is a DAC-ADC mismatch.
###
### Install Struck scaler in IOC-1 and use Struck auto-count instead of $int_samples times reading ADC
###
### Eliminate "smooth" DAC decreasing (i.e. $DAC_piezo -= 0.2) in search()
###
### Move P2 and DAC scans of search into scan record; then watch for progress?
###
### Analyze possible improvements to the variable-step scan in feedback()
###
### Any speed-ups in fast_feedback()?
