#!/usr/bin/perl -W
#--------------------------------------------------------------
#
#  This script implements remote start of MARCCD, Pilatus, or Eiger detector
#
#                       Author: Sergey Stepanov
#
# Version-1.0:  2014/01/26	initial implementation
# Version-1.1:  2014/02/05	added nomachine connection to mar computers
# Version-1.2:  2014/03/25	added launching camonitor for Pilatus
# Version-1.3:  2014/06/13	added second popup when user chooses to leave FURKA alone
# Version-1.4:  2014/10/28	added prepare_pilatus_fpe() function (switching between external_enable & internal_advance)
# Version-1.4a: 2015/01/14	switched to external_enable being default when cmdline is not specified
# Version-1.5:  2016/05/31	added Eiger for ID-B
# Version-1.6   2017/12/07	added added check if FURKA is anabled. FURKA may be disabled when filewriter
# 				is not in use and images are saved by StreamWriter. Then no need for FURKA.
# Version-1.7   2021/07/13	added &getCCDfpehost()
#
# This script calls external commands:	hostname			(for all)
#					changeBeamlineUser.pl		(for all; for mar it is called indirectly via marccdepics)
#					furka_user.pl			(for Pilatus and Eiger)
#					ppu: prepare_pilatus_fpe.pl	(for Pilatus: recompile for "external enable")
#					ppu: start_epics_pilatus_ioc	(for Pilatus)
#					ppu: update_furka		(for Pulatus)
#					epu: start_epics_eiger_ioc	(for Eiger)
#					epu: furka_start		(for Eiger)
#					mar: marccdepics		(for MarCCD)
#					mar: x11vnc_wrapper.pl		(for MarCCD)
#					mar: date			(for MarCCD)
#					mar: nxplayer			(for MarCCD)
#--------------------------------------------------------------
  use DBI;
  use Tk;
  use Tk::Balloon;
  use Tk::BrowseEntry;
  use Pezca;
  use strict;

  our ($NONINTERACTIVE, $USE_ZENITY_FOR_EXIT, $IGNORE_ERRORS, $PEZCA, $beamline);
  our($mw, $mheaderLbl, $mtextLbl, $mcancelBtn, $vncBtn, $nxBtn, $sshBtn, $mtooltip);
  our($ow, $oheaderLbl, $otextLbl, $ocancelBtn, $connectBtn, $otooltip);
  our($opassEntry, $opassBtn, $password, $needpsw, $passExists);
  our($oscaleLbl, $oscaleSpin, $ojpegLbl, $ojpegSpin, $scale, $jpeg);
  our($orestartLbl, $orestartBtn, $restart);
  my ($script, $user, $blnuser, $host, $server, $home, $bln, $b, $debug);
  my ($output, $status, $sshcmd, $cmd, $use_nx, $use_ssh, $detlabel);
  my ($fgcolor, $fgcolor2, $abgcolor, $afgcolor, $line, $ncol);
  my ($passworddir, $passwordfile, $logfile, $detector, $xterm, $gterm, $mterm);
  my (@clients, $port, $vncport, @beamlines, $tip, $master);
  my ($title, $hdr, $txt, $opt_cancel, $opt_vnc, $opt_nx, $opt_ssh);
  my ($nx_template, $nx_client, $ext_enable, $narg, $syntax, $sshkey);

  if (!defined $ENV{PEZCA}) {
     die 'Please define the PEZCA environment. It must point to the PERL scripts depository';
  }
  $PEZCA = $ENV{PEZCA};
  $PEZCA =~ s/\\/\//g;
  if ($PEZCA !~ /\/$/) {$PEZCA .='/';}

  do $PEZCA.'Common/ccd_subs.pl';
  do $PEZCA.'Common/common_subs.pl';

  select(STDERR);  $|=1;		# Flush output after each print
  select(STDOUT);  $|=1;		# Flush output after each print

  $script = &makeScriptName();

  $NONINTERACTIVE      = 1;
  $USE_ZENITY_FOR_EXIT = 1;
  $IGNORE_ERRORS       = 1;

  $debug    = 0;
  $port     = 5923;
  $vncport  = $port - 5900;
  $fgcolor  = '#000000';
  $fgcolor2 = '#BB0000';
  $abgcolor = '#DDDDDD';
  $afgcolor = '#770000';
  $use_nx   = 1;			# connect to MAR desktop via NX (default is VNC - Linux desktop sharing)
  $use_ssh  = 0;			# start MARCCD via SSH & X11    (default is VNC - Linux desktop sharing)
# $nx_client   = 'nxclient';
  $nx_client   = 'nxplayer';
  $nx_template = $PEZCA.'Data/'.$nx_client.'_mar_connection_template.nxs';
  $nx_client   = '/usr/NX/bin/'.$nx_client;

  $gterm = &Which('gnome-terminal');
  $mterm = &Which('mate-terminal');

  $host = &gethost('short');
# $host = 'bl1ws3';                                     #for testing
  if ($host !~ /^bl[123]/        &&
      $host !~ /^keithley[123]$/ &&
      $host !~ /^mar[123]$/      &&
      $host !~ /^px[013]$/) {
     $txt = 'Host "'.$host.'" is unsupported.';
     &zenity_error_exit($txt);
#    $mw = MainWindow->new(); $mw->withdraw();				#hide main window
#    $mw->messageBox(-title=>'Incorrect host',-message=>$txt,-wrap=>0,-type=>'ok',-icon=>'error');
#    exit 0;
  }
  if    ($host =~ /^keithley[123]$/) {$host =~ s/^keithley//; $host='bl'.$host.'keithley';}
  elsif ($host =~ /^mar[123]$/)      {$host =~ s/^mar//;      $host='bl'.$host.'mar';}

  if    ($host =~ /^bl[123]/)        {$b = substr($host,2,1);}
  else                               {$b = 0;}
  $beamline = qw(23d 23i 23o 23b)[$b];
  if (!defined $beamline) {die 'Unknown beamline for b='.$b;}
  $status = do $PEZCA.'Common/beamline_'.$beamline.'.pl';
  if    ($@)               {die 'Could not parse '.$PEZCA.'Common/beamline_'.$beamline.'.pl';}
  elsif (!defined $status) {die 'Could not do '.$PEZCA.'Common/beamline_'.$beamline.'.pl';}
  $bln = 'bl'.$b;
  if ($b != 0) {@clients = ($bln.'ws1', $bln.'ws2', $bln.'ws3', $bln.'ws6', $bln.'keithley', $bln.'mar');}
  else         {@clients = qw(px0 px1 px3);}

  if (! grep(/^$host$/,@clients)) {
     $txt = 'The detector program can be started from these computers only:'."\n"
          . "(@{clients})\n"
          . 'Host "'.$host.'" does not qualify.';
     &zenity_error_exit($txt);
#    $mw = MainWindow->new();  $mw->withdraw();				#hide main window
#    $mw->messageBox(-title=>'Incorrect host',-message=>$txt,-wrap=>0,-type=>'ok',-icon=>'error');
#    exit 0;
  }

  if ($host =~ /ws[23]$/ || $host eq 'www') {$master = 1;} else {$master = 0;}

  if (defined $ENV{HOME}) {$home = $ENV{HOME};}
  else                    {$home = '.'; print STDERR '!!! HOME is undefined!'."\n\n";}

  if    (defined $ENV{USER})     {$user=$ENV{USER};}
  elsif (defined $ENV{USERNAME}) {$user=$ENV{USERNAME};}
  else                           {$user='';}

  ($detector,$status) = &getCCDname();
  if ($status) {
     $txt = 'The getCCDname function failed to detect the detector (status='.$status.')';
     &zenity_error_exit($txt);
  }
  print STDOUT 'Beamline detector = '.$detector."\n";

  if ($detector =~ /^pilatus$/i) {
     $detlabel = 'Pilatus';
     ($server,$status) = &getCCDfpehost();
     if ($status || $server eq 'unknown') {$server = $bln.'ppu';}
     $server =~ s/(-40g)*(\.gmca\.aps\.anl\.gov)*$//;
     ### Parse command line (one can specify "external_enable" for Pilatus):
     $syntax = "\n".'Syntax: '.$script.' [EXTERNAL_ENABLE]'."\n";
     $narg = @ARGV;
     if ($narg > 0) {
        if ($ARGV[0] !~ /^external_enable$/i  &&
            $ARGV[0] !~ /^external_disable$/i &&
            $ARGV[0] !~ /^internal_advance$/i) {
           $txt = 'Incorrect command line: "'.$ARGV[0].'"'.$syntax;
           &zenity_error_exit($txt);
#          &press_key_exit($txt);
        }
        elsif ($ARGV[0] =~ /^external_enable$/i)  {$ext_enable=1;}
        elsif ($ARGV[0] =~ /^external_disable$/i) {$ext_enable=0;}
        elsif ($ARGV[0] =~ /^internal_advance$/i) {$ext_enable=0;}
     } else {
        $ext_enable = 0;				# The default when command line is NOT specified
     }
     &prepare_pilatus_fpe();
     &change_beamline_user_if_needed();
     &start_pilatus_or_eiger_fpe();
  }
  elsif ($detector =~ /^eiger$/i) {
     $detlabel = 'Eiger';
     ($server,$status) = &getCCDfpehost();
     if ($status || $server eq 'unknown') {$server = $bln.'epu';}
     $server =~ s/(-40g)*(\.gmca\.aps\.anl\.gov)*$//;
     $narg = @ARGV;
     if ($narg > 0) {
        &zenity_info('!!! '.$detlabel.' detector FPE does not support command line argiments including "external_enable"');
     }
     &change_beamline_user_if_needed();
     &start_pilatus_or_eiger_fpe();
  }
  elsif ($detector =~ /^marccd$/i) {
     $detlabel = 'MARCCD';
     $server = $bln.'mar';
     $restart      = 0;
     $password     = '';
     $passworddir  = $home.'/.vnc';
     $passwordfile = $passworddir.'/passwd_'.$server;
     $logfile      = $passworddir.'/x11vnclog_'.$server;
     if ($use_nx) {
        if (! -f $nx_client) {
           print STDOUT '!!! Cannot use NX connection method because NX client is not installed!'."\n";
           $use_nx = 0;
        }
        elsif (! -f $nx_template) {
           print STDOUT '!!! Cannot use NX connection method because NX template is not found!'."\n";
           $use_nx = 0;
        }
     }
     &draw_vncopt();
     $ow->withdraw();
     &draw_main_marccd();
     MainLoop;
  }

####################################END###############################################

#--------------------------------------------------------------------------
sub change_beamline_user_if_needed() {
  if ($master) {
     ($status,$blnuser) = Pezca::GetString($beamline.':bi:runs:currentUser');
     if ($status) {
        &report_only($script, $status, $beamline.':bi:runs:currentUser');
        $blnuser = 'unknown';
        $txt = $script.': failed to detect current JBluIce user!';
        &zenity_error($txt);
     }
     print STDOUT $script.': current SystemUser='.$user.'  JBluIceUser='.$blnuser."\n";
     if ($blnuser ne $user) {
        print STDOUT $script.': Changing JBluIce user to '.$user."\n";
        $title = 'ChangeBeamlineUser';
        if    (defined $gterm) {$xterm = $gterm.' --disable-factory --geometry 128x32+0+500 -t '.$title.' -e ';}
        elsif (defined $mterm) {$xterm = $mterm.' --disable-factory --geometry 128x32+0+500 -t '.$title.' -e ';}
        else                   {$xterm = 'xterm +cm -sb -rightbar -sl 5000 -geometry 128x32+0+500 -T '.$title.' -e ';}
        $cmd = $xterm.'"perl -W '.$PEZCA.'changeBeamlineUser.pl"';
        print STDOUT 'Executing: "'.$cmd.'"'."\n";
        $status = system ($cmd);
     } else {
        print STDOUT $script.': No JBluIce user change -- JBluIce runs will not be reset'."\n";
        &restart_furka();
     }
  } else {
     $txt = 'Host "'.$host.'" is not allowed to change beamline user,'."\n"
          . 'restart FURKA or start '.$detlabel.' FPE. Please use ws2 or ws3.';
     &zenity_error_exit($txt);
#    $mw = MainWindow->new();  $mw->withdraw();				#hide main window
#    $mw -> messageBox(-title=>'Incorrect host',-message=>$txt,-wrap=>0,-type=>'ok',-icon=>'error');
     &clearAllCaMonitors();
     exit;
  }
}

#--------------------------------------------------------------------------
sub prepare_pilatus_fpe() {
### Recompile the Pilatus FPE to turn EXTERNAL_ENABLE On or Off
  $sshkey = '/mnt/software/scripts/ppu_gmca_rsa';
  if ($master) {
     if (! -f $sshkey) {
        $txt = 'SSH key "'.$sshkey.'" is not found.';
        &zenity_error_exit($txt);
     }
     ### -X forces X11 forwarding
     ### -q quiet mode (IMPORTANT!: without it, the presence of banner forces ssh to return error ("1") in all cases)
     $sshcmd = 'ssh -Xq -o BatchMode=yes -o StrictHostKeyChecking=no -i '.$sshkey;
     $sshcmd .= ' gmca@'.$server.' perl -W '.$PEZCA.'prepare_pilatus_fpe.pl';
     if ($ext_enable) {$sshcmd .= ' external_enable';}
     else             {$sshcmd .= ' internal_advance';}
     print STDOUT 'Executing: "'.$sshcmd.'"'."\n";
     $status = system($sshcmd);
  } else {
     $txt = 'Host "'.$host.'" is not allowed to compile Pilatus FPE. Please use ws2 or ws3.';
     &zenity_error_exit($txt);
#    $mw = MainWindow->new();  $mw->withdraw();				#hide main window
#    $mw -> messageBox(-title=>'Incorrect host',-message=>$txt,-wrap=>0,-type=>'ok',-icon=>'error');
  }
}

#--------------------------------------------------------------------------
sub start_pilatus_or_eiger_fpe() {
  if ($master) {
     $title = uc($detector).'_FPE';
#    $xterm = 'gnome-terminal --disable-factory     --geometry 128x40+200+300          -t '.$title.' -e ';
#    $xterm = 'mate-terminal  --disable-factory     --geometry 128x40+200+300          -t '.$title.' -e ';
#    $xterm = 'konsole --noxft --nomenubar --noframe -geometry 128x40+200+300          -T '.$title.' -e ';
#    $xterm = 'xterm -hold         +cm -sb -rightbar -geometry 128x40+200+300 -sl 5000 -T '.$title.' -e ';
     $xterm = 'xterm -hold -iconic +cm -sb -rightbar -geometry 128x40+200+300 -sl 5000 -T '.$title.' -e ';
     ### -X forces X11 forwarding
     ### -q quiet mode (IMPORTANT!: without it, the presence of banner forces ssh to return error ("1") in all cases)
     $sshcmd = 'ssh -Xq -o BatchMode=yes -o StrictHostKeyChecking=no';
     $sshcmd .= ' '.$server.' start_epics_'.$detector.'_ioc';
     print STDOUT 'Executing: '.$xterm.'"'.$sshcmd.'" &'."\n";
     $status = system($xterm.'"'.$sshcmd.'" &');
  } else {
     $txt = 'Host "'.$host.'" is not allowed to start '.$detlabel.' FPE. Please use ws2 or ws3.';
     &zenity_error_exit($txt);
#    $mw = MainWindow->new();  $mw->withdraw();				#hide main window
#    $mw -> messageBox(-title=>'Incorrect host',-message=>$txt,-wrap=>0,-type=>'ok',-icon=>'error');
  }
  &clearAllCaMonitors();
  exit;
}

#--------------------------------------------------------------------------
sub restart_furka() {
### run update_furka to change the furka uid/gid, create the link in /dataInt, and restart the furka service
  my ($furkaUser, $msg, $yesno, $furkacmd, $cmd, $prefix, $partial_dir);
  my ($user_pw, $passwd, $uid, $gid, $quota, $comment, $gcos, $homedir, $shell);
  $cmd = 'perl -W '.$PEZCA.'furka_user.pl';
  $furkaUser = `$cmd`;
  if ($furkaUser !~ /^furka_user=/) {
     $msg = '!!! Failed to detect current FURKA user. Is FURKA running?';
     &zenity_info($msg);
     $furkaUser = 'unknown';
   } else {
     $furkaUser =~ s/^furka_user=//;
     $furkaUser =~ s/\n$//;  chomp $furkaUser;
  }
  if ($furkaUser ne $user && $furkaUser ne 'furka_is_disabled' && $furkaUser ne 'streamwriter') {
     if    ($furkaUser eq 'none')    {$msg='FURKA is not running'."\n";}
     elsif ($furkaUser eq 'unknown') {$msg='FURKA is running as unknown user'."\n";}
     else			     {$msg='FURKA is running as user "'.$furkaUser.'"'."\n";}
     if ($furkaUser eq 'none') {
        $msg .= 'Do you want to start FURKA as user "'.$user.'"?'."\n";
     } else {
        $msg .= 'Do you want to restart FURKA as user "'.$user.'"?'."\n";
        if ($detector =~ /^pilatus$/) {
           $msg .= "\n"
                .'ATTENTION: exercise extreme care when restarting FURKA since all'."\n"
                .$detlabel.' images not yet copied to the shared storage will be lost!'."\n"
                .'Make sure at least 5 minutes have passed after the last image was collected.';
        }
     }
     $yesno = &zenity_question ($msg);
     if ($yesno) {
        print STDOUT 'Reconfiguring FURKA...'."\n";
        if    ($detector =~ /^pilatus$/) {
           $furkacmd = 'sudo /usr/local/bin/update_furka '.$user;
        }
        elsif ($detector =~ /^eiger$/) {
           ($user_pw, $passwd, $uid, $gid, $quota, $comment, $gcos, $homedir, $shell) = getpwnam ($user);
           $partial_dir = $homedir;
           $prefix = '/mnt/beegfs/';
           if ($partial_dir !~ /^${prefix}/) {&zenity_error_exit('*** '.$script.': home directory of '.$user.' is not on '.$prefix);}
           $partial_dir =~ s/^${prefix}//;
           $furkacmd = '/usr/local/bin/furka_start '.$uid.':'.$gid.':'.$partial_dir;
        }
        $host =~ s/(-40g)*(\.gmca\.aps\.anl\.gov)*$//;
        if ($host !~ /^${server}/i) {
           ### -a disables forwarding of the authentication agent connection (do we need it?)
           ### -T disables pseudo-tty allocation (do we need it?)
           ### -x disables X11 forwarding
           ### -q quiet mode (IMPORTANT!: without it, the presence of banner forces ssh to return error ("1") in all cases)
           $sshcmd   = 'ssh -aTxq -o BatchMode=yes -o StrictHostKeyChecking=no '.$server.' ';
           $furkacmd = $sshcmd.$furkacmd;
        }
        $msg = `${furkacmd}`;
        print STDOUT $msg."\n";
     } else {
        if ($furkaUser eq 'none') {
           $msg = 'You chose to skip starting FURKA. Images will still be collected,'."\n"
                 .'but not copied to your home directory and only remain in the '.$server."\n"
                 .'RAM until either FURKA is started or the '.$server.' computer is restarted.'."\n";
        } else {
           $msg = 'You chose to keep FURKA running as another user. When FURKA runs'."\n"
                 .'as another user, images will still be collected and copied to your'."\n"
                 .'home directory. They will also be readable by you, but owned by the'."\n"
                 .'FURKA user.'."\n";
        }
        $msg .= "\n"
               .'Do you want to continue starting the '.$detlabel.' software?';
        $yesno = &zenity_question ($msg);
        if ($yesno) {print STDOUT 'Leaving FURKA alone and continuing with the '.$detlabel.' startup'."\n";}
        else        {print STDOUT 'Exiting...'."\n"; &clearAllCaMonitors(); exit;}
     }
  } 
  elsif ($furkaUser eq 'furka_is_disabled') {
     print STDOUT 'FURKA is disabled; images will not be transferred to shared storage.'."\n";
  }
  elsif ($furkaUser eq 'streamwriter') {
     print STDOUT $script.': FURKA is disabled, but it is OK because StreamWriter is running.'."\n";
  }
  else {
     print STDOUT 'Leaving FURKA alone as it is already started for user "'.$user.'"'."\n";
  }
}

#--------------------------------------------------------------------------
sub draw_main_marccd () {
  my ($col);
  if (! defined $mw) {
     ### Create a GUI:
     if ($master) {$title = 'Remote start of '.$detlabel.' on'.$server;}
     else         {$title = 'Remote connection to '.$detlabel.' on'.$server;}

     $txt = "\n"
          . 'This tool is designed for remote users.'
          . ' Local users are strongly advised to login on '.$server
          . ' desktop and launch/display the '.$detlabel.' program there. ';
     if ($master) {
        $txt .= 'Continue reading if you are remote and still need to see '.$detlabel
              . ' program graphical interface. There might be two scenarios:'."\n\n"
              . '1. Your host or collaborator is already logged on the physical'
              . ' desktop of '.$server.' computer and possibly started MACCCD program'
              . ' there. Then you want '.$server.' desktop to be shared with you.'."\n\n"
              . '2. You are operating beamline in remote-only mode and no one has'
              .'  logged on the physical desktop of '.$server.' for you. Then, you want'
              . ' to start '.$detlabel.' program via NX or ssh and display it on your screen.'
              . ' If needed, your host or colaborators may be able to see '.$detlabel
              . ' inerface along with your desktop via NX shadow connection or'
              .' TeamViewer'."\n\n"
              . 'What do you want to do?'."\n";
     } else {
        $txt .= "\n\n".'Remote users can start '.$detlabel.' from '.$bln.'ws2 or '.$bln.'ws3'
              . ' running only. From this computer ('.$host.') you can only connect to'
              . ' already session after it has been opened from one of the above two'
              . ' workstations.'."\n\n";
     }

     $hdr  = "\n".$detlabel.' Remote ';
     $opt_cancel = 'Cancel (local user)';
     if ($master) {
        $hdr .= 'Start';
        $ncol = 1;
        $opt_vnc = $ncol.'. Connect to '.$server.' desktop'; $ncol++;
        if ($use_nx)  {$opt_nx  = $ncol.'. Connect to '.$server.' via NX'; $ncol++;}
        if ($use_ssh) {$opt_ssh = $ncol.'. Start '.$detlabel.' via SSH'; $ncol++;}
     } else {
        $hdr .= $detlabel.' View';
        $opt_vnc = 'Connect to '.$server.' desktop';
        $ncol = 2;
     }

     $mw = MainWindow->new(-width=>800,-title=>$title);
     $mw->resizable( 0, 0 );
     $mtooltip = $mw->Balloon(-background=>'yellow');
     $line = 0;
     $mheaderLbl   = $mw->Label(-text =>$hdr, -foreground => 'blue', -font=>[-size=>18, -weight=>'bold']
                               )->grid(-row=>$line++,   -column=>0, -columnspan=>$ncol, -padx=>10, -pady=>10);
     $mtextLbl     = $mw->Label(-text =>$txt, -foreground=>$fgcolor2, -font=>[-size=>9, -weight=>'bold'],
                                -wraplength=>750, -justify=>'left'
                               )->grid(-row=>$line++, -column=>0, -columnspan=>$ncol, -padx=>10, -pady=>10);

     $col = 0;
     $mcancelBtn   = $mw->Button(-text=>$opt_cancel, -font=>[-size=>10,-weight=>'bold'], -command=>\&goodbye,
                                 -activebackground=>$abgcolor, -foreground=>$fgcolor, -activeforeground=>$afgcolor
                               )->grid(-row=>$line, -column=>$col++, -pady=>5, -padx=>5);
     $tip = 'Press this to close this screen (you can also hit the ESC key)';
     $mtooltip->attach($mcancelBtn, -initwait=>0,-balloonmsg=>$tip);

     $vncBtn       = $mw->Button(-text=>$opt_vnc, -font=>[-size=>10,-weight=>'bold'], -command=>\&show_vncopt,
                                 -activebackground=>$abgcolor, -foreground=>$fgcolor, -activeforeground=>$afgcolor
                               )->grid(-row=>$line, -column=>$col++, -pady=>5, -padx=>5);
     $tip = 'Press this to connect to '.$server.' desktop via VNC and access '.$detlabel.' program as a part of the desktop.'."\n\n"
          . '**NOTE-1: At first start you may be asked to generate a password protecting remote access to '.$server.'.'."\n"
          . '**NOTE-2: For this option to work, user '.$user.' must be already logged on '.$server.' desktop.';
     $mtooltip->attach($vncBtn,   -initwait=>0,-balloonmsg=>$tip);

     if ($master) {
        if ($use_nx) {
           $nxBtn = $mw->Button(-text=>$opt_nx,  -font=>[-size=>10,-weight=>'bold'], -command=>\&nx2mar,
                                 -activebackground=>$abgcolor, -foreground=>$fgcolor, -activeforeground=>$afgcolor
                               )->grid(-row=>$line, -column=>$col++, -pady=>5, -padx=>5);
           $tip = 'Press this to login on '.$server.' via NX and start or display '.$detlabel.' program.'."\n\n"
                . '**NOTE: if '.$detlabel.' program is already running on '.$server.', then you'
                . ' must connect to existing destop where '.$detlabel.' was started.'."\n";
           $mtooltip->attach($nxBtn,-initwait=>0,-balloonmsg=>$tip);
        }
        if ($use_ssh) {
           $sshBtn = $mw->Button(-text=>$opt_ssh,  -font=>[-size=>10,-weight=>'bold'], -command=>\&ssh2mar,
                                 -activebackground=>$abgcolor, -foreground=>$fgcolor, -activeforeground=>$afgcolor
                               )->grid(-row=>$line, -column=>$col++, -pady=>5, -padx=>5);
           $tip = 'Press this to start '.$detlabel.' program on '.$server.' via ssh and display it on this screen.'."\n\n"
                . '**NOTE-1: For this option to work, '.$detlabel.' program must not be running on '.$server.' yet.'."\n"
                . '**NOTE-2: No password is required.';
           $mtooltip->attach($sshBtn,-initwait=>0,-balloonmsg=>$tip);
        }
     }
     $mw->bind('<Destroy>', [\&goodbye]);
     $mw->bind('<KeyPress-Escape>', [\&goodbye]);
#    $mw->bind('<KeyPress-Return>', [\&goodbye]);
     $mw->protocol('WM_DELETE_WINDOW' => \&goodbye);

  } else {
     $mw->deiconify();
     $mw->raise();
  }
  $mw->focus();
}

#--------------------------------------------------------------------------
sub  draw_vncopt () {
  my (@sclList, @jpgList, $pswprompt);
  if (! defined $ow) {
     $needpsw = 0;
     if (-f $passwordfile) {$passExists = 1;}
     else                  {$passExists = 0;}

     $hdr = "\n".'Remote Connection Options'."\n";
     if ($passExists) {
        $title     = 'Set options';
        $pswprompt = 'Reset remote connection password'
     } else {
        $title     = 'Set password and options';
        $pswprompt = 'Set remote connection password';
     }

     if ($master) {
        $txt = 'INFO: ';
        if ($passExists) {
           $txt .= 'Password file already exists. It will not be overwritten '
                 . 'unless you choose to enter a new password in the field below.';
        } else {
           $txt .= 'Password is not set. Automatic password will be generated '
                 . 'unless you choose to enter your own password in the field below.';
        }
        $txt .= "\n\n"
              . 'You will not be prompted for password when connecting to '.$server.' with '
              . 'this tool. Therefore you do not need to know the password unless you '
              . 'wish to share '.$server.' desktop viewing on VNC port '.$vncport.' with '
              .' someone else.'."\n";
     }


     @sclList = ('1/4', '1/3', '1/2', '2/3', '3/4', '1');
     @jpgList = ('1=poor', 2, 3, 4, 5, 6, 7, 8, '9=best');
     $scale = '3/4';
     $jpeg  = '6';

     ### Create options window:
     $ow = MainWindow->new(-width=>10,-height=>4,-title=>$title);
     $ow->resizable( 0, 0 );
     $otooltip = $ow->Balloon(-background=>'yellow');

     $ncol = 2;
     $line = 0;
     $oheaderLbl   = $ow->Label(-text =>$hdr, -foreground => 'blue', -font=>[-size=>18, -weight=>'bold']
                               )->grid(-row=>$line++, -column=>0, -columnspan=>$ncol, -padx=>10, -pady=>10);
     $otextLbl     = $ow->Label(-text =>$txt, -foreground=>$fgcolor2, -font=>[-size=>9,-weight=>'bold'],
                                -wraplength=>750, -justify=>'left'
                               )->grid(-row=>$line++, -column=>0, -columnspan=>$ncol, -padx=>10, -pady=>10);
     if ($master) {
        $opassBtn   = $ow->Checkbutton(-text=>$pswprompt, -variable=>\$needpsw, -command=>\&reconfPassEntry,
                                -font=>[-size=>9,-weight=>'bold'], -justify=>'left'
                               )->grid(-row=>$line, -column=>0, -padx=>10,-pady=>10, -sticky=>'e');
        ### When the password is empty, it is auto-generated with random function:
        $opassEntry = $ow->Entry(-width=>21, -exportselection=>1,
                                 -font=>[-size=>9,-weight=>'bold'], -justify=>'left',
                                 -textvariable=>\$password, -show=>'*',
                                 -disabledbackground=>'#999999',-disabledforeground=>'#999999',
                                 -background=>'white', -highlightbackground=>'white'
                               )->grid(-row=>$line++, -column=>1, -padx=>10,-pady=>10, -sticky=>'w');
        &reconfPassEntry();
        $txt = 'Set desktop scaling';
        $oscaleLbl   = $ow->Label(-text =>$txt,
                                  -font=>[-size=>9,-weight=>'bold'], -justify=>'left'
                               )->grid(-row=>$line, -column=>0, -padx=>10, -pady=>10, -sticky=>'e');
        $oscaleSpin  = $ow->BrowseEntry(-variable =>\$scale,-choices=>\@sclList,-state=>'readonly',
                                 -font=>[-size=>9,-weight=>'bold'], -justify=>'left',
                                 -disabledbackground=>'white', -disabledforeground=>'black',
                                 -background=>'white', -foreground=>'black', -highlightbackground=>'white'
                               )->grid(-row=>$line++, -column=>1, -padx=>10, -pady=>10, -sticky=>'w');
        $tip = 'This is applied at the VNC server start. If after connecting to '.$server.' you find that the'."\n"
             . 'scaling is different from your request, re-launch this GUI and check "Restart VNC Server"';
        $otooltip->attach($oscaleSpin, -initwait=>0,-balloonmsg=>$tip);
     }
     $txt = 'Set JPEG quality level';
     $ojpegLbl       = $ow->Label(-text =>$txt,
                                 -font=>[-size=>9,-weight=>'bold'], -justify=>'left'
                               )->grid(-row=>$line, -column=>0, -padx=>10, -pady=>10, -sticky=>'e');
     $ojpegSpin      = $ow->BrowseEntry(-variable =>\$jpeg,-choices=>\@jpgList,-state=>'readonly',
                                 -font=>[-size=>9,-weight=>'bold'], -justify=>'left',
                                 -disabledbackground=>'white', -disabledforeground=>'black',
                                 -background=>'white', -foreground=>'black', -highlightbackground=>'white'
                               )->grid(-row=>$line++, -column=>1, -padx=>10, -pady=>10, -sticky=>'w');

     if ($master) {
        $txt = 'Restart VNC server';
        $orestartLbl = $ow->Label(-text =>$txt,
                                  -font=>[-size=>9,-weight=>'bold'], -justify=>'left'
                               )->grid(-row=>$line, -column=>0, -padx=>10, -pady=>10, -sticky=>'e');
        $orestartBtn = $ow->Checkbutton(-variable=>\$restart,
                                -font=>[-size=>9,-weight=>'bold'], -justify=>'left'
                               )->grid(-row=>$line++, -column=>1, -padx=>10,-pady=>10, -sticky=>'w');
     }


     $opt_cancel = 'Cancel';
     $opt_vnc    = 'Connect';
     $ocancelBtn   = $ow->Button(-text=>$opt_cancel, -font=>[-size=>10,-weight=>'bold'], -command=>\&back2main,
                                 -activebackground=>$abgcolor, -foreground=>$fgcolor, -activeforeground=>$afgcolor
                                )->grid(-row=>$line, -column=>0, -pady=>5, -padx=>5, -sticky=>'e',);
     $tip = 'Press this to return to previous menu';
     $otooltip->attach($ocancelBtn, -initwait=>0,-balloonmsg=>$tip);

     $connectBtn   = $ow->Button(-text=>$opt_vnc, -font=>[-size=>10,-weight=>'bold'], -command=>\&vnc_connect,
                                 -activebackground=>$abgcolor, -foreground=>$fgcolor, -activeforeground=>$afgcolor
                                )->grid(-row=>$line, -column=>1, -pady=>5, -padx=>5, -sticky=>'w',);
     $tip = 'Press this to connect to '.$server.' desktop via VNC and access '.$detlabel.' program as a part of the desktop.'."\n\n"
          . '**NOTE: For this option to work, user "'.$user.'" must be already logged on '.$server.' desktop.';
     $otooltip->attach($connectBtn,   -initwait=>0,-balloonmsg=>$tip);
     $ow->bind('<Destroy>', [\&goodbye]);
     $ow->bind('<KeyPress-Escape>', [\&back2main]);
     $ow->bind('<KeyPress-Return>', [\&vnc_connect]);
  } else {
     $ow->deiconify();
     $ow->raise();
  }
  $ow->focus();
}

#--------------------------------------------------------------------------
sub  show_vncopt () {
  if (defined $mw) {$mw->withdraw();}
  &draw_vncopt();
}

#--------------------------------------------------------------------------
sub  back2main () {
  if (defined $mw) {&draw_main_marccd(); $ow->withdraw();}
  else             {&goodbye();}
}

#--------------------------------------------------------------------------
sub  reconfPassEntry () {
  if ($needpsw) {$opassEntry->configure(-state=>'normal');}
  else          {$opassEntry->configure(-state=>'disabled'); $password='';}
}

#--------------------------------------------------------------------------
sub goodbye () {
  my ($db, $answer);
### Confirmation dialog:
# $db = $mw->Dialog(-title=>'Exit?', -buttons=>['OK', 'Cancel'], -bitmap=>'question');
# $answer = $db->Show(-width=>20);
  $answer = 'OK';
  if ($answer eq 'OK') {
     &clearAllCaMonitors();
     if (defined $ow) {
#       $ow->bindings(undef);
#       $ow->destroy();
#       undef $ow;
     }
     if (defined $mw) {
#       $mw->bindings(undef);
#       $mw->destroy();
#       undef $mw;
     }
#    Tk::exit;
     exit;
  }
}

#--------------------------------------------------------------------------
#sub random ($range) {
sub random ($) {
   my $range = shift(@_);
   return int(rand()*$range+0.5);
}

#--------------------------------------------------------------------------
sub generate_pass () {
  my ($plen, $upper, $loop, $pass, $letter);
  $plen  = 8;
  $upper = 1;
  $loop  = 1;
  $pass  = '';
  for ($loop=0; $loop<$plen; $loop++) {
     if (&random(4) == 1) {
        $pass .= sprintf '%c',  (&random(9)+48);		#ASCII 48-57 are numbers 0-9
     } else {
        $letter = sprintf '%c', (&random(25)+97);		#ASCII 97-122 are letter a-z
        if ($upper && &random(1)) {$letter = uc($letter);}
        $pass .= $letter;
     }
  }
  return $pass;
}

#--------------------------------------------------------------------------
sub  nx2mar () {
  my ($nx_file, $string, $fhi, $fho);
  if ($server eq 'mar0') {return;}
  $nx_file = $home.'/nxclient.session.'.$server.'.'.$user.'.nxs';
  if (-f $nx_file) {unlink $nx_file;}
  $fhi = open (IN, ' < ', $nx_template);
  if (!defined $fhi) {
     $txt = 'Could not open template file="'.$nx_template.'"';
     &zenity_error($txt);
#    $ow->messageBox(-title=>'Password file error',-message=>$txt,-wrap=>0,-type=>'ok',-icon=>'error');
     return;
  }
  $fho = open (OU, ' > ', $nx_file);
  if (!defined $fho) {
     $txt = 'Could not create NX session file="'.$nx_file.'"';
     &zenity_error($txt);
#    $ow->messageBox(-title=>'Password file error',-message=>$txt,-wrap=>0,-type=>'ok',-icon=>'error');
     return;
  }
  while (<IN>) {
     $string = $_;
     if    ($string =~ /mar\.gmca\.aps\.anl\.gov/) {$string =~ s/mar/${server}/;}
     elsif ($string =~ /Enter your username/) {$string =~ s/Enter your username/${user}/;}
     print OU $string;
  }
  close(IN);
  close(OU);
  $cmd = $nx_client.' --session '.$nx_file.' &';
  print STDOUT 'Executing: "'.$cmd.'"'."\n";
  $status = system ($cmd);
  &clearAllCaMonitors();
  exit;
}

#--------------------------------------------------------------------------
sub  ssh2mar () {
  if ($server eq 'mar0') {return;}
  $title  = 'marccdepics';
# -a Disables forwarding of the authentication agent connection
# -C Requests compression of all data
# -f Requests ssh to go to background just before command execution.
# -n Redirects stdin from /dev/null (actually, prevents reading from stdin). This must be used when ssh is run in the background
# -q Quiet mode. Causes most warning and diagnostic messages to be suppressed
# -T Disable pseudo-tty allocation
# -X Enables X11 forwarding. This can also be specified on a per-host basis in a configuration file.
# -c arcfour is the fastest crypto algorithm, about 7-8 faster than the default one -- see
#    http://xmodulo.com/2013/07/how-to-speed-up-x11-forwarding-in-ssh.html
#    However, Argonne disabled it due to security flaw
### ATTENTION: option "-q" is important since without it, the presence of the
### ANL banner forces ssh to return error ("1") in all cases.
# $sshcmd = 'ssh -aqTX -o BatchMode=yes -o StrictHostKeyChecking=no';
# $sshcmd = 'ssh -qX   -o BatchMode=yes -o StrictHostKeyChecking=no';
# $sshcmd = 'ssh -CqX  -o BatchMode=yes -o StrictHostKeyChecking=no';
  $sshcmd = 'ssh -qX   -o BatchMode=yes -o StrictHostKeyChecking=no';
  $sshcmd .= ' '.$server.' marccdepics';
  if    (defined $gterm) {$xterm = $gterm.' --disable-factory --geometry 128x40+200+300       -t '.$title.' -e ';}
  elsif (defined $mterm) {$xterm = $mterm.' --disable-factory --geometry 128x40+200+300       -t '.$title.' -e ';}
  else                   {$xterm = 'xterm +cm -sb -rightbar -sl 5000 -geometry 128x40+200+300 -T '.$title.' -e ';}
#                         $xterm = 'xterm +cm -sb -rightbar -sl 5000 -geometry 128x40+200+300 -T '.$title.' -hold -e ';
#                         $xterm = 'xterm +cm -sb -rightbar -sl 5000 -geometry 128x40+200+300 -T '.$title.' -iconic -e ';
  $cmd = $xterm.'"'.$sshcmd.'" &';
  print STDOUT 'Executing: '.$cmd."\n";
  $status = system ($cmd);
  &clearAllCaMonitors();
  exit;
}

#--------------------------------------------------------------------------
sub  vnc_connect () {
  my ($mode, @vncpars, $clientlist, $amp, $jpg);
  if    ($needpsw      && length($password) == 0)    {$password = &generate_pass();}
  elsif (! $passExists && length($password) == 0)    {$password = &generate_pass();}
  if (length($password) != 0)			     {$needpsw  = 1;}

  $mode = 'startasneeded';
  if ($master) {
     if ($restart || $needpsw) {$mode = 'restart';}
  }
  @vncpars = qw(-no6 -quiet -shared);
  push @vncpars, '-wait 100';				#event polling in ms (default=20)
  push @vncpars, '-rfbport '.$port;
  push @vncpars, '-rfbauth '.$passwordfile;
  push @vncpars, '-logappend '.$logfile;
  if ($master) {push @vncpars, '-scale '.$scale;}
  $clientlist = join(',',@clients);
  push @vncpars, '-allow '.$clientlist;
### Using either of these two options blocks return from x11vnc_wrapper.pl
## and then we need to start the wrapper with ampersand (see below):
# push @vncpars, '-forever';
# push @vncpars, '-gui tray';
  push @vncpars, '-display :0 -bg';
# push @vncpars, '&';
  if ($debug) {print STDOUT 'VNCPARS=('."@{vncpars})\n";}
  if (grep(/^\-gui$/,@vncpars) || grep(/^\-forever$/,@vncpars)) {$amp = 1;}
  else                                                          {$amp = 0;}

  if (length($password) != 0) {
     if ($debug) {print STDOUT 'Storing password='.$password.' in file='.$passwordfile."\n";}
     if (-e $passwordfile) {unlink($passwordfile) or warn 'Could not delete '.$passwordfile."$!\n";}
     elsif (! -d $passworddir) {mkdir $passworddir;}
     $status = system('x11vnc -storepasswd '.$password.' '.$passwordfile);
     sleep 1;

  }
  if (! -e $passwordfile) {
     $txt = 'Could not find or generate shared password file="'.$passwordfile.'"';
     &zenity_error($txt);
#    $ow->messageBox(-title=>'Password file error',-message=>$txt,-wrap=>0,-type=>'ok',-icon=>'error');
     return;
  }
### Test ssh connection.
### ATTENTION: option "-q" is important since without it, the presence of
### banner forces ssh to return error ("1") in all cases.
  $sshcmd = 'ssh -q -o BatchMode=yes -o StrictHostKeyChecking=no '.$server;
  $cmd = $sshcmd.' date 2>&1';
# $cmd = $sshcmd.' date';
  print STDOUT 'Executing: '.$cmd."\n";
  $output = `$cmd`;
  $status = ${^CHILD_ERROR_NATIVE} >> 8;
  if ($status) {
     $txt = 'Could not establish shared-key ssh connection to '.$server."\n".'Output:'."\n".$output;
     &zenity_error($txt);
#    $ow->messageBox(-title=>'SSH error',-message=>$txt,-wrap=>0,-type=>'ok', -icon=>'error');
     return
  }
  $cmd = join (' ',@vncpars);
  if ($amp == 1) {$cmd .= ' &';}
  $cmd = $sshcmd.' x11vnc_wrapper.pl -mode='.$mode.' '.$cmd.' 2>&1';
  print STDOUT 'Executing '.$cmd."\n";
  $output = `$cmd`;
  $status = ${^CHILD_ERROR_NATIVE} >> 8;
  if ($status){
     $txt = 'Failed to start VNC server on '.$server."\n".'Output:'."\n".$output."\n\n".'Check "'.$logfile.'" for additional details.';
     &zenity_error($txt);
#    $ow->messageBox(-title=>'SSH error',-message=>$txt,-wrap=>0,-type=>'ok',-icon=>'error');
     return
  }
  if ($amp) {sleep 4;}			#wait for 4s (if we started with ampersand, we need to give wrapper time to finish)

### Start vncviewer:
  $jpg = substr($jpeg,0,1);
  $cmd = 'vncviewer -PasswordFile '.$passwordfile.' -Shared -QualityLevel '.$jpg.' '.$server.':'.$vncport.' &';
  print STDOUT 'Executing '.$cmd."\n";
  $status = system($cmd);
  &goodbye();
}
