#!/usr/bin/perl -W

# -----------------------------------------------------------------
# This script is triggered by the "UserScriptAfterLogin" parameter in
# /usr/NX/etc/server.cfg. It receives two command  line arguments:
# $1 is username and $2 is remote IP (works correctly even in case of
# webplayer) and checks if login of given user from given IP is allowed.
# If the script exits with non-zero status, the session is terminated.
# Unfortunately, there is no way to display dialogs like those with zenity.
# The script runs as user/group "nx". 
#
# More details are here: https://www.nomachine.com/DT11R00174#11.5
# Some examples are here: https://www.nomachine.com/AR02L00787
#
# ATTENTION: this can only run via sudo -- add it to /etc/sudoers
#
#         	Author: Sergey Stepanov
#
# Version-1.0: 2021/02/17 initial implementation
# -----------------------------------------------------------------
  use strict;
  use POSIX;
# use Net::CIDR::Lite;					#this is loaded below via &try_load_module()					
  use warnings::unused;

  my ($script, $user, $nxip, $host, $loginAllowed);

  select STDOUT; $|=1;					#set unbuffered output
  select STDERR; $|=1;					#set unbuffered output

  $script = $0 =~ s/^.*[\/\\]//gr;
  $script =~ s/\.pl$//i;

  if ($^O =~ /Win/i) {die $script.': cannot run under Windows.'."\n";}

  if    (defined $ENV{HOSTNAME})  {$host = $ENV{HOSTNAME};}
  elsif (defined $ENV{HOST})	  {$host = $ENV{HOST};}
  else                            {$host=`hostname`; $host=~s/\n//g;}
  if (!defined $host) {die $script.': undefined HOST'."\n";}

  $host = lc($host);
  $host =~ s/[.-]\Qgmca.aps.anl.gov\E$//;
  if ($host !~ /bl[123]ws[256]$/ && $host ne 'px0') {die $script.': unlisted host='.$host."\n";}

  if (geteuid() ne '0') {
### ----------------------------------------------------------
### /etc/sudoers should have
# ALL ALL = (root) NOPASSWD:  /home/gmca/epics/pezca/nxallowedcheck.pl
### ----------------------------------------------------------
     if (exec('/usr/bin/sudo','-n',$0,@ARGV)) {die $script.': failed restarting with sudo.'."\n";}
  }
  $user = $ARGV[0];	if (!defined $user) {$user = 'unknown';}
  $nxip = $ARGV[1];	if (!defined $nxip) {$nxip = '0.0.0.0';}

  &try_load_module('Net::CIDR::Lite');					#this will exit if no module found  (see below)

  $loginAllowed = &checkLoginAllowed($user,$nxip);

  if (! $loginAllowed) {die $script.': login of user='.$user.' from ip='.$nxip.' is not allowed.'."\n";}
  exit 0;

#==========================================================================

sub checkLoginAllowed($$) {						#$user,$ip
  my ($user, $ip, $sshcfg, @buf, $key);
  my ($str, @words, $nwords, $word, $usr, $adr);
  my ($cidr, $nranges, @parts, $nparts);
  my ($ipmin, $ipmax, $found, $i, $restrict);

  ($user,$ip) = @_;

  if (!defined $ip || length($ip)==0 || $ip eq '-') {return 0;}
  
  $sshcfg = '/etc/ssh/sshd_config';
  if (! -e $sshcfg || ! -r $sshcfg) {die $script.': cannot access '.$sshcfg."\n";}

  &read_entire_file($sshcfg,\@buf);

  $key = 'AllowUsers';
  $cidr = Net::CIDR::Lite->new;
  $restrict = 0;
  $nranges = 0;
  foreach $str (@buf) {
     if ($str !~ /$key/) {next;}
     $str =~ s/DO_NOT_MODIFY_AFTER_HERE//g;	# remove divider
     $str =~ s/\t/ /g;				# replace tabs by spaces
     $str =~ s/\s+/ /g;				# compress spaces ("+" means match 1 or more times)
     $str =~ s/^ *//g;				# remove beginning-of-the-line spaces
     $str =~ s/ *$//g;				# remove end-of-the-line spaces
     $str =~ s/^#.*$//g;			# remove comment lines
     if ($str =~ /^${key}/) {
        $restrict = 1;				#there is at least one AllowUsers statement
        $str =~ s/^${key} //;
        @words = split(/\s+/,$str);
        $nwords = @words;
        foreach $word (@words) {
           if ($word =~ /\@/) {($usr,$adr) = split(/\@/,$word);}
           else               {$usr = '*'; $adr = $word;}
           if ($usr eq $user || $usr eq '*') {
              if    ($adr =~ /^([0-9]{1,3}\.){1,3}[0-9]{1,3}\/{1}[0-9]{1,2}$/) {			#record is CIDR: 1.2.3/24
#                if ($debug > 1) {print LOG 'Found CIDR=',$adr,"\n";}
                 $cidr->add($adr);
                 $nranges++;
              }
              elsif ($adr =~ /^([0-9]{1,3}\.){1,3}[0-9]{1,3}$/) {			#record is IP or range: 1.2.3 (which means 1.2.3.1-1.2.3.255)
                 @parts = split(/\./,$adr);
                 $nparts = @parts;
                 if    ($nparts < 4 && $nparts > 1) {
                    $ipmin = $ipmax = $adr;
                    for ($i=0; $i<(4-$nparts); $i++) {$ipmin .= '.1'; $ipmax .= '.255';}
#                   if ($debug > 1) {print LOG 'Found RANGE=',$ipmin.'-'.$ipmax,"\n";}
                    $cidr->add_range($ipmin.'-'.$ipmax);
                    $nranges++;
                 } 
                 elsif ($nparts == 4) {
#                   if ($debug > 1) {print LOG 'Found IP=',$adr,"\n";}
                    $cidr->add_ip($adr);
                    $nranges++;
                 }
                 else {
#                   if ($debug > 1) {print LOG $script.': Unexpected '.$key.' address='.$adr."\n";}
                 }
              }
              elsif ($adr =~ /^([0-9]{1,3}\.){3}\*$/) {					#record is IP range: 1.2.3.*
                 $adr =~ s/\*$//;
                 $ipmin = $adr.'1';
                 $ipmax = $adr.'255';
#                if ($debug > 1) {print LOG 'Found RANGE=',$ipmin.'-'.$ipmax,"\n";}
                 $cidr->add_range($ipmin.'-'.$ipmax);
                 $nranges++;
              }
              elsif ($adr =~ /^([0-9]{1,3}\.){2}\*\.\*$/) {				#record is IP range: 1.2.*.*
                 $adr =~ s/\*\.\*$//;
                 $ipmin = $adr.'1.1';
                 $ipmax = $adr.'255.255';
#                if ($debug > 1) {print LOG 'Found RANGE=',$ipmin.'-'.$ipmax,"\n";}
                 $cidr->add_range($ipmin.'-'.$ipmax);
                 $nranges++;
              }
           }
        }   
     }
  }
  if ($nranges > 0) {
#    if ($debug > 1) {
#       my @cidr_list = $cidr->list();
#       print LOG 'CIDR list:'."\n";
#       foreach $i (@cidr_list) {print LOG '   '.$i."\n";}
#    }
     $found = $cidr->find($ip);						#returns 1 or empty string
     if ($found ne '1') {$found = 0;}
  } else {
     $found = 1 - $restrict;
  }
# if ($debug > 1) {print LOG 'nranges='.$nranges.'  restrict='.$restrict.' found='.$found."\n";}
  return $found;
}

#=================================================================

sub try_load_module($) {
  my $module = shift(@_);
  eval('use '.$module);
  if ($@) {die '*** '.$script.': module "'.$module.'" failed to load.';}
}

#==========================================================================

# sub read_entire_file($file,\@lines) {
sub read_entire_file($\@) {
# use Fcntl;					#not used (defines O_RDONLY constant (O_RDONLY=0,O_WRONLY=1,O_RDWR=2))
  my ($file, $linesref, $buffer, $status);

  $file     = shift(@_);
  $linesref = shift(@_);

  @$linesref = ();
  $status = sysopen(FH, $file, 0);		#0 is used instead of O_RDONLY defined in "use Fcntl";
  if (!defined $status) {die $script.': Could not open '.$file."\n";}
  sysread(FH, $buffer, -s FH);
  close FH;
  @$linesref = split(/\r?\n/,$buffer);
}

