#!/usr/bin/perl -W
#--------------------------------------------------------------
# This script checks closed NX session history and displays 
# versions of NX player used on current computer. 
#
# ATTENTION: this can only run via sudo -- add it to /etc/sudoers
#
#                  Author: Sergey Stepanov
#
# Version-1.0: 2018/07/13 initial implementation
# Version-1.1: 2018/08/16 added print
# Version-1.2: 2018/11/12 added -d[ebug] & -a[ll] flags and improved nxwebclient detection,
# Version-1.3  2019/04/22 added -s[um] flag to sum sessions with the same IP,user,date,client,version
# Version-1.4  2021/01/22 added support for NX-7 (Redis)
#--------------------------------------------------------------
# use Socket;                                   #inet_ntoa (unpacks binary IP into d.d.d.d, legacy)
  use strict;
# use warnings::unused;

  my ($script, $cmd, $out, $host, $user, $home);
  my ($nxclosed, @closedsessions, $sesfile, $sesname, @buf, @msg, $i, $j);
  my ($nxuser, $nxip, $nxversion, $nxOS, $nxOSw, $nxclienttype, $nxbrowser);
  my ($nxtype, $nsessions, @ct, $nxstamp, $nxstart, $nxend, $nxdur);
  my ($debug, $allip, $sum, $narg, $arg, @w0, @w1, $t0, $t1);
  my ($nxserver, $serverVersion, $serverMajor);

  select STDOUT; $|=1;                                  #set unbuffered output
  select STDERR; $|=1;                                  #set unbuffered output

  $script = $0;
  $script =~ s/\\/\//g;                                 #change \ to /
  $script =~ s/^.*\///g;                                #remove everything before last /
  $script =~ s/\.pl$//i;
  $debug = 0;

#---------------------------------------------------------------------
  if ($^O =~ /Win/i) {die $script.': cannot run under Windows.'."\n";}
  if    (defined $ENV{USER})     {$user = $ENV{USER};}
  elsif (defined $ENV{USERNAME}) {$user = $ENV{USERNAME};}
  else                           {$user = 'unknown';}
  if    (defined $ENV{HOME})     {$home = $ENV{HOME};}
  else                           {$home = 'unknown';}

### When initial user is not root, but he issued "su",
### the environment "USER" is preserved, but "HOME" is not.
  if ($user ne 'root' && $home !~ /^\/root/) {
     if ($debug) {
        print STDOUT "\n".'User='.$user;
        print STDOUT "\n".'Home='.$home;
#       print STDOUT "\n".'This script should be executed as root or via sudo'."\n\n";
        print STDOUT "\n".'Re-launching as root via sudo...'."\n\n";
     }
     exec('/usr/bin/sudo','-n',$0,@ARGV);
     exit;
  }
  if ($user eq 'root' && defined $ENV{SUDO_USER}) {$user = $ENV{SUDO_USER};}
  if    (defined $ENV{HOSTNAME}) {$host = $ENV{HOSTNAME};}
  elsif (defined $ENV{HOST})	 {$host = $ENV{HOST};}
  else                           {$host=`hostname`; $host=~s/\n//g;}
  if (!defined $host) {print STDOUT '!!! '.$script.': HOSTNAME is undefined!'."\n\n"; $host='undef';}
  $host = lc($host);
  $host =~ s/\Q.gmca.aps.anl.gov\E$//;
#---------------------------------------------------------------------

  $nxserver  = '/usr/NX/bin/nxserver';
  $nxclosed  = '/usr/NX/var/db/closed';
  if (! -e $nxserver) {
     print STDOUT $script.': cannot find NX Server at '.$nxserver."\n";
     exit;
  }
  $serverVersion = `$nxserver --version 2>/dev/null | grep ' - Version ' | grep -v ' WARNING: '`; 
  $serverVersion =~ s/^.* - Version //;
  chomp $serverVersion;
  $serverMajor = $serverVersion;
  $serverMajor =~ s/\..*//g;
  if ($serverMajor !~ /^\d*$/) {
     print STDOUT '!!! '.$script.': Error determining NX server Generation='.$serverMajor.'. Assuming 6.'."\n";
     $serverMajor = 6;
  }

  $allip = 0;                                                                   #do not include ANL IPs (normally GMCA staff shadowing)
  $sum   = 0;                                                                   #sum sessions with the same IP,user,date,client,version
  $narg = @ARGV;
  foreach $arg (@ARGV) {
     if    ($arg =~ /^-\?$/  || $arg =~ /^-?help$/i)  {die 'Syntax: '.$script.' [-d[ebug]] [a[llip] [-s[sum]]'."\n";}
     elsif ($arg =~ /^-?d$/i || $arg =~ /^-?debug$/i) {$debug = 1;}
     elsif ($arg =~ /^-?a$/i || $arg =~ /^-?allip$/i) {$allip = 1;}             #include GMCA shadowing
     elsif ($arg =~ /^-?s$/i || $arg =~ /^?sum$/i)    {$sum   = 1;}             #sum sessions with the same IP,user,date,client,version
     else {print STDOUT '!!! '.$script.': unrecognized command line parameter "'.$arg.'"'."\n";}
  }

### ========================================================================
### Check all closed NX logins:
### ========================================================================
  if ($serverMajor > 6) {
     $cmd = $nxserver.' --redis cmd smembers closed';
     $out = `$cmd 2>/dev/null`;
     @closedsessions = split(/[\s\n]+/,$out);					#these are session IDs
  } else {
     $cmd = '/usr/bin/find '.$nxclosed.' -maxdepth 1 -type f 2>/dev/null';
     $out = `$cmd`;
     @closedsessions = split(/\n/,$out);					#these are session files
  }
  $nsessions = 0;
  @msg = ();
  foreach $sesfile (@closedsessions) {
     if ($serverMajor > 6) {
        $sesname = $sesfile;
        if (&read_nx_redis($nxserver,'closed',$sesname,\@buf) != 0) {next;}
     } else {
        if ($sesfile !~ /^$nxclosed\/sessionId/) {next;}
        if (&read_entire_file($sesfile,\@buf) != 0) {next;}			#read session file data into @buf
        $sesname = $sesfile;
        $sesname =~ s/^.*\{//;
        $sesname =~ s/\}.*$//;
     }
     $nxclienttype = &findkey2('clientType',\@buf);				#nxclient|nxwebclient
     if ($nxclienttype eq 'nxwebclient') {
        $nxOS      = 'web';
        $nxOSw     = &findkey2('clientPlatform',\@buf);
        $nxbrowser = &findkey2('clientBrowser',\@buf);
        if ($nxOSw     eq '' || $nxOSw     eq 'notset') {$nxOSw     = '-';}
        if ($nxbrowser eq '' || $nxbrowser eq 'notset') {$nxbrowser = '-';}
        $nxOS.='('.$nxbrowser.':'.$nxOSw.')';
     } else {
        $nxOS      = &findkey2('clientPlatform',\@buf);
     }
     $nxip         = &findkey2('foreignAddres',\@buf);
     $nxuser       = &findkey2('userName',\@buf);
#    $nxid         = &findkey2('sessionId',\@buf);
     $nxversion    = &findkey2('clientVersion',\@buf);  $nxversion=~s/notset/-/g;
     $nxtype       = &findkey2('type',\@buf);					#virtualDefault|virtualAttach|nodeConnectionMonitor
#    if ($serverMajor <= 6) {
#       @ct        = localtime((stat($sesfile))[9]);				#find file date (time when the session ended)
#       $nxstamp   = sprintf '%04d/%02d/%02d %02d:%02d', $ct[5]+1900,$ct[4]+1,$ct[3],$ct[2],$ct[1];
#    }
     $nxstart      = &findkey2('startTime',\@buf);				#startTime is epoch
     if (length($nxstart) == 0) {
        $nxstart   = &findkey2('creationTime',\@buf);
     }
     $nxend        = &findkey2('endTime',\@buf);				#endTime is epoch
     if (length($nxend) == 0) {
        $nxend     = &findkey2('suspendTime',\@buf);
     }
     if (length($nxstart) > 0 && length($nxend) > 0) {$nxdur = sprintf '%.1fm',($nxend-$nxstart)/60.;}
     else                                            {$nxdur = '-';}
     if (length($nxstart) > 0) {
        @ct        = localtime($nxstart);
        if (! $sum) {
           $nxstart = sprintf '%04d/%02d/%02d %02d:%02d', $ct[5]+1900,$ct[4]+1,$ct[3],$ct[2],$ct[1];
        } else {
           $nxstart = sprintf '%04d/%02d/%02d',           $ct[5]+1900,$ct[4]+1,$ct[3];
        }
     } else {
        $nxstart   = '-';
     }
     if (length($nxend) > 0) {
        @ct        = localtime($nxend);
        if (! $sum) {
           $nxend  = sprintf '%04d/%02d/%02d %02d:%02d', $ct[5]+1900,$ct[4]+1,$ct[3],$ct[2],$ct[1];
        } else {
           $nxend  = sprintf '%04d/%02d/%02d',           $ct[5]+1900,$ct[4]+1,$ct[3];
        }
     } else {
        $nxend     = '-';
     }
### APS Citrix Servers IPs:
### 164.54.22.111 carme.aps.anl.gov
### 164.54.22.112 elara.aps.anl.gov
### 164.54.22.113  leda.aps.anl.gov
     if ($nxtype =~ /nodeConnectionMonitor/)     {next;}				#not really a session
     if ($nxip =~ /^\Q164.54.103\E/ && ! $allip) {next;}				#exclude GMCA local shadowing
     if ($nxip =~ /^\Q164.54.22\E/  && ! $allip) {next;}				#exclude GMCA shadowing via APS Citrix servers
     if (! $sum) {$msg[$nsessions].= sprintf 'start=[%-16s]', $nxstart;}
     else        {$msg[$nsessions] = sprintf 'start=%-10s',   $nxstart;}
     $msg[$nsessions] .= sprintf ' time=%-8s user=%-16s ip=%-15s ver=%-7s OS=%-30s', 
                                   $nxdur,   $nxuser,   $nxip,   $nxversion, $nxOS;
     if (! $sum) {$msg[$nsessions] .= sprintf ' ses=%s',$sesname;}
     if ($debug) {$msg[$nsessions] .= ' session='.$sesfile;}
     $nsessions++;
  }
  print STDOUT "\n".'### '.$script.': Formerly executed and now closed NX sessions on host='.$host;
  if ($serverMajor > 6) {print STDOUT ' (see "'.$nxserver.' --redis cmd smembers closed"):'."\n";}
  else                  {print STDOUT ' (see "'.$nxclosed.'/sessionId{..}"):'."\n";}
  if ($nsessions > 0) {
     @msg = sort (@msg);
     if ($sum) {
        for ($i=0; $i<$nsessions-1; $i++) {
           @w0 = split(/\s+/,$msg[$i]);
           for ($j=$i+1; $j<$nsessions; $j++) {
              @w1 = split(/\s+/,$msg[$j]);
              if ($w0[0] eq $w1[0] &&							#date
                  $w0[2] eq $w1[2] &&							#user
                  $w0[3] eq $w1[3] &&							#ip
                  $w0[4] eq $w1[4] &&							#version
                  $w0[5] eq $w1[5]) {							#OS
                 $t0 = $w0[1]; $t0 =~ s/^time=//; $t0 =~ s/m$//;
                 $t1 = $w1[1]; $t1 =~ s/^time=//; $t1 =~ s/m$//;
                 if ($t0 ne '-' && $t1 ne '-') {$w0[1] = sprintf 'time=%.1fm',($t0+$t1);}
                 else                          {$w0[1] = '-';}
                 $msg[$i] = sprintf '%-16s %-13s %-21s %-18s %-11s %-19s',$w0[0],$w0[1],$w0[2],$w0[3],$w0[4],$w0[5];
                 if ($debug) {$msg[$i] .= ' '.$w0[6];}
                 splice(@msg,$j,1);
                 $nsessions--;
                 $j--;
              }
           }
        }
     }
     for ($i=0; $i<$nsessions; $i++) {print STDOUT $msg[$i],"\n";}
  } else {
     print STDOUT '  None found in '.$nxclosed."\n";
  }
  exit;

#==========================================================================

sub findkey2($\@) {					#$key,\@buf
  my ($key, $bufref, $line);
  ($key,$bufref) = @_;
  foreach $line (@$bufref) {
     if ($line =~ /^$key=/) {
        $line =~ s/^$key=//;
        return $line;
     }
  }
  return '';
}

#==========================================================================

# sub read_entire_file($file,\@lines) {
sub read_entire_file($\@) {
# use Fcntl;						#not used (defines O_RDONLY constant (O_RDONLY=0,O_WRONLY=1,O_RDWR=2))
  my ($file, $linesref, $buffer, $status);

  $file     = shift(@_);
  $linesref = shift(@_);

  @$linesref = ();
  $status = sysopen(FH, $file, 0);			#0 is used instead of O_RDONLY defined in "use Fcntl";
  if (!defined $status) {print STDOUT '*** '.$script.': Could not open '.$file."\n"; return 1;}
  sysread(FH, $buffer, -s FH);
  close FH;
  @$linesref = split(/\r?\n/,$buffer);
  return 0;
}

#==========================================================================

# sub read_nx_redis($nxserver,$sesfolder,$session,\@lines) {
sub read_nx_redis($$$\@) {
  my ($nxserver, $sesfolder, $session, $linesref);
  my ($cmd, $buffer, @buf, $nbuf, $i);

  $nxserver  = shift(@_);
  $sesfolder = shift(@_);
  $session   = shift(@_);
  $linesref  = shift(@_);

  @$linesref = ();
  if ($sesfolder ne 'running' && 
      $sesfolder ne 'closed'  && 
      $sesfolder ne 'failed') {
     print STDOUT '*** '.$script.': incorrect NX Redis folder ['.$sesfolder.']'."\n";
     return 1;
  }
  $cmd = $nxserver.' --redis cmd hgetall '.$sesfolder.'.'.$session;
  $buffer = `$cmd 2>/dev/null`;
  if (${^CHILD_ERROR_NATIVE}) {
     print STDOUT '*** '.$script.': Error executing '.$cmd."\n";
     return 1;
  }
  @buf = split(/\s+/,$buffer);
  $nbuf = @buf;
  if ($nbuf == 0 || $nbuf != 2*int($nbuf/2)) {
     print STDOUT '*** '.$script.': incorrect words number='.$nbuf.' in the output of '.$cmd."\n";
     return 1;
  }
  for ($i=0; $i<$nbuf-1; $i+=2) {
     push @$linesref,$buf[$i].'='.$buf[$i+1];
  }
# foreach $buffer (@$linesref) {print STDOUT $buffer,"\n";}; exit;
  return 0;
}
