#!/usr/bin/perl -W
#--------------------------------------------------------------
# This script displays a plot of data in a given file along with
# respective Gaussian and Loretzian fits and prints respective fitting
# parameters. It deploys the 'peakFit' function provided by common_PDL.pl.
# The later makes use of the Perl PDL lmfit function
# (Marquardt-Levenberg non-linear least squares fitting)
#
# See: http://search.cpan.org/~chm/PDL-2.4.6/Lib/Fit/LM.pm
#
#     		     Author: Sergey Stepanov
#
# Version-1.0:  2017/02/17 initial implementation
# Version-1.1:  2018/11/16 changed internal calls from &findNoisyPeakRef to &findPeakRef
#--------------------------------------------------------------
  use PDL;
  use PDL::Fit::LM;
  use strict;

  my ($PEZCA, $script, $file, $xCol, $yCol, @xdata, @ydata);
  my ($syntax, $xFound, $yFound, $dFound, $i);
  my ($statusG, $y0G, $AG, $xcG, $wG, $FWHMG);
  my ($statusL, $y0L, $AL, $xcL, $wL, $FWHML);
  my ($npts, $gnuplot, $scan_start, $scan_end);
  my ($dest, $title, $narg, $arg, $msg, $status);

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;
  do $PEZCA.'/Common/gnuplot_subs.pl';
  do $PEZCA.'/Common/common_PDL.pl';
  do $PEZCA.'/Common/common_subs.pl';

  select STDOUT; $|=1;			# set unbuffered output

  $script = &makeScriptName();
  $narg = @ARGV;
  $syntax = "\n".'Syntax: '.$script.' <file> [X] [Y1] [IMG|PNG|PS|PDF]'."\n";
  if ($narg>0) {$syntax=$syntax."\n".'Received: '.$script." @{ARGV}\n\n";}

  if ($narg < 1 || $narg > 4) {
    print STDOUT "\n".'*** '.$script.': Incorrect number of command-line arguments: '.$narg."\n".$syntax;
    exit;
  }

  $file = $ARGV[0];
  if (!(-f $file)) {die '*** '.$script.': File='.$file.' does not exist!'."\n";}

  $xCol  = 1;				# default X-column number
  $yCol  = 2;				# default Y-column number
  $dest  = 'SCREEN';			# default destination
  $title = '';				# default plot title

  for ($i=1; $i<$narg; $i++) {
    $arg = $ARGV[$i];
    if ($arg =~ /^[0-9]*$/) {
      if    (!defined $xFound) {$xCol=$arg;  $xFound=1;}
      elsif (!defined $yFound) {$yCol=$arg;  $yFound=1;}
      else {die "\n".'*** '.$script.': incorrect command line argument-'.$i.' ['.$arg.']!'."\n".$syntax;}
    }
    elsif ($arg =~ /^IMG|PNG|PS|PDF$/i) {
      if    (!defined $dFound) {$dest=uc($arg);  $dFound=1;}
      else {die "\n".'*** '.$script.': incorrect command line argument-'.$i.' ['.$arg.']!'."\n".$syntax;}
    }
    else {
      print STDOUT "\n";
      if  (!defined $xFound)  {$xFound=0;}
      if  (!defined $yFound)  {$yFound=0;}
      if  (!defined $dFound)  {$dFound=0;}
      print STDOUT '    xFound='.$xFound.'  xCol='.$xCol."\n";
      print STDOUT '    yFound='.$yFound.'  yCol='.$yCol."\n";
      print STDOUT '    dFound='.$dFound.'  dest='.$dest."\n";
      die "\n".'*** '.$script.': incorrect command line argument-'.$i.' ['.$arg.']!'."\n".$syntax;
    }
  }

  ($status,$msg) = &read_data_file ($file,\@xdata,\@ydata,$xCol,$yCol);
  if ($status) {&press_key_exit($msg);}

  $npts = @xdata;
  $scan_start = $xdata[0];
  $scan_end = $xdata[$npts-1];
  $gnuplot = &open_gnuplot($scan_start,$scan_end); 		#uses LUN=GNUPLOT
  if ($gnuplot) {
     if ($dest eq 'IMG' || $dest eq 'PNG') {
        print GNUPLOT 'set terminal png small size 640,480'."\n";
        print GNUPLOT "set output '".$file.".png'\n";
     }
     elsif ($dest eq 'PS' || $dest eq 'PDF') {
        print GNUPLOT 'set terminal postscript portrait enhanced color blacktext'."\n";
        print GNUPLOT 'set output "'.$file.'.ps"'."\n";
     }
     print  GNUPLOT 'set xlabel "Xcol data"'."\n";
     print  GNUPLOT 'set ylabel "Ycol data"'."\n";
     print  GNUPLOT 'set samples 500'."\n";
  }

### Find FWHM:
  $msg = '';
  ($statusG,$y0G,$AG,$xcG,$wG) = &peakFit(\@xdata, \@ydata, 'Gaussian');
  if ($statusG) {
     print STDOUT '*** '.$script.': Gaussian fitting failed with status='.$statusG."\n";
     ($y0G,$AG,$xcG,$wG) = (0,0,0,0);
  } else {
     $FWHMG = $wG*2.35482;					#$FWHM=$w*(2*sqrt(2*ln(2)))
     $msg .= sprintf 'Fit=Gaussian  xc=%.3f  w=%.3f  FWHM=%.3f'."\n",$xcG,$wG,$FWHMG
  }

  ($statusL,$y0L,$AL,$xcL,$wL) = &peakFit(\@xdata, \@ydata, 'Lorentzian');
  if ($statusL) {
     print STDOUT '*** '.$script.': Lorentzian fitting failed with status='.$statusL."\n";
     ($y0L,$AL,$xcL,$wL) = (0,0,0,0);
  }  else {
     $FWHML = $wL*2.;
     $msg .= sprintf 'Fit=Lorentzian  xc=%.3f  w=%.3f  FWHM=%.3f'."\n",$xcL,$wL,$FWHML
  }
  if ($gnuplot) {
     if ($wG) {printf  GNUPLOT 'set label 1 "FWHM(G)=%.3f" at graph 0.01,0.95'."\n",$FWHMG;}
     if ($wL) {printf  GNUPLOT 'set label 2 "FWHM(L)=%.3f" at graph 0.01,0.90'."\n",$FWHML;}
     printf GNUPLOT 'plot "%s" using %d:%d with lines',$file,$xCol,$yCol;
     if ($wG) {printf GNUPLOT ', %.3f+%.3f*exp(-0.5*((x-%.3f)/%.3f)**2)',$y0G,$AG,$xcG,$wG;}
     if ($wL) {printf GNUPLOT ', %.3f+%.3f/(1+((x-%.3f)/%.3f)**2)',$y0L,$AL,$xcL,$wL;}
     print GNUPLOT "\n";
  }
  if (length($msg)>0) {
     &zenity_info($msg);
#    if ($gnuplot) {&print_gnuplot();}
     if ($gnuplot) {
        if    ($dest eq 'SCREEN') {&print_gnuplot();}
        else                      {&close_gnuplot();}
     }

     if ($dest eq 'PDF') {
        $status = system('ps2pdf', $file.'.ps', $file.'.pdf');
        if ($status) {die '*** '.$script.': failed to convert PS to PDF'."\n";}
        else         {unlink $file.'.ps';}		#delete postscript file
     }
  }
