#!/usr/bin/perl -W
#--------------------------------------------------------------
# This script is a frontend to GNUplot. Depending on command line
# switches it either plots specified file on the screen, or creates
# the plot as PNG or postscript file. The X-Y columns in the file
# are selectable via command line arguments.
#
#     		     Author: Sergey Stepanov
# v.1.0 2006/10/31
# v.1.1 2006/12/07 added header
# v.2.0 2006/12/12 added plotting multiple Y's
#                  added plotting multiple Y2's (plots linked to the right Y-axis)
# v.2.1 2011/02/21 switched to full use of gnuplot library, simplified viewing on
#                  screen (no more "print?" question) and added viewing of PNG/PS/PDF
#                  with respective software.
# v.2.2 2011/10/03 added batch mode (no display)
#--------------------------------------------------------------
  use strict;
  use File::Spec;
# use Cwd 'abs_path';

  my ($PEZCA, $script, $narg, $arg, $syntax, $file, $fullname);
  my ($xCol, @yCol, @y2Col, $dest, $title, $batch, $viewerPDF);
  my ($xFound, $yFound, $y2Found, $dFound, $tFound, $i);
  my ($nyCol, $ny2Col, $status, $eog, $ps2pdf);

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;

  do $PEZCA.'/Common/gnuplot_subs.pl';
  do $PEZCA.'/Common/common_subs.pl';

  select STDERR; $|=1;			# set unbuffered output
  select STDOUT; $|=1;			# set unbuffered output

  $script = &makeScriptName();
  $batch  = 0;
  $narg   = @ARGV;
  $syntax = "\n".'Syntax: '.$script.' <file> [X] [Y1,Y1,Y1...] [Y2,Y2,Y2...] [[IMG|PNG|PS|PDF][batch]] ["title"]'."\n";
  if ($narg>0) {$syntax .= '  Received: '.$script." @{ARGV}\n\n";}

  if ($narg < 1 || $narg > 6) {
    print STDOUT "\n".'*** '.$script.': Incorrect number of command-line arguments: '.$narg."\n".$syntax;
    exit;
  }

  $file = $ARGV[0];
  if (!(-f $file)) {die '*** '.$script.': File='.$file.' does not exist!'."\n";}

  if ($^O !~ /Win/i) {
    $viewerPDF = &Which('okular');				# this is preferred (part of KDE)
  if (!defined $viewerPDF) {$viewerPDF = &Which('atril');}	# this is a MATE fork of evince. To be checked if it works with DELL-5100cn
    if (!defined $viewerPDF) {$viewerPDF = &Which('evince');}	# has problem with printing to DELL-5100cn (see: https://www.centos.org/forums/viewtopic.php?f=13&t=49706 and  https://bugs.launchpad.net/ubuntu/+source/cups/+bug/419143)
    if (!defined $viewerPDF) {$viewerPDF = &Which('acroread');}	# rarely installed
    if (!defined $viewerPDF) {$viewerPDF = &Which('gv');}	# does not allow choosing printer (just default)
    $ps2pdf = &Which('$ps2pdf');
    $eog    = &Which('eog');
    if (!defined $eog) {$eog = &Which('eom');}			# eye of MATE
  } else {
    $viewerPDF = &Which('gsview32');
    if (!defined $viewerPDF) {$viewerPDF = &Which('gsview64');}
    if (!defined $viewerPDF) {$viewerPDF = &Which('acroread');}
    if (!defined $viewerPDF) {$viewerPDF = &Which('acrobat');}
    if    (-f 'C:/Progra~1/IrfanView/i_view32.exe') {$eog = 'C:/Progra~1/IrfanView/i_view32.exe';}
    elsif (-f 'C:/Progra~2/IrfanView/i_view32.exe') {$eog = 'C:/Progra~2/IrfanView/i_view32.exe';}
    elsif (-f 'C:/Progra~1/IrfanView/i_view64.exe') {$eog = 'C:/Progra~1/IrfanView/i_view64.exe';}
  }
  $xCol  = 1;				# default X-column number
  @yCol  = (2);				# default Y-column number
  @y2Col = ();				# default Y-column number (none)
  $dest  = 'SCREEN';			# default destination
  $title = '';				# default plot title

  for ($i=1; $i<$narg; $i++) {
    $arg = $ARGV[$i];
    if ($arg =~ /^[0-9]*$/) {
      if    (!defined $xFound) {$xCol=$arg;  $xFound=1;}
      elsif (!defined $yFound) {@yCol=$arg;  $yFound=1;}
      elsif (!defined $y2Found) {@y2Col=$arg; $y2Found=1;}
      else {die "\n".'*** '.$script.': incorrect command line argument-'.$i.' ['.$arg.']!'."\n".$syntax;}
    }
    elsif ($arg =~ /^([0-9]*\,)*[0-9]*$/) {
      if    (!defined $yFound) {@yCol=split /\,/, $arg;    $yFound=1;}
      elsif (!defined $y2Found) {@y2Col=split /\,/, $arg;  $y2Found=1;}
      else {die "\n".'*** '.$script.': incorrect command line argument-'.$i.' ['.$arg.']!'."\n".$syntax;}
    }
    elsif ($arg =~ /^IMG|PNG|PS|PDF(batch)*$/i) {
      if    (!defined $dFound) {
         $dest=uc($arg);  $dFound=1;
	 if ($dest =~ /batch/i) {$batch=1; $dest =~ s/batch//i;}
	 else                   {$batch=0;}
      }
      else {die "\n".'*** '.$script.': incorrect command line argument-'.$i.' ['.$arg.']!'."\n".$syntax;}
    }
    else {
      if (!defined $tFound) {$title=$arg;  $tFound=1;}
      else {
         print STDOUT "\n";
         if  (!defined $xFound)  {$xFound=0;}
         if  (!defined $yFound)  {$yFound=0;}
         if  (!defined $y2Found) {$y2Found=0;}
         if  (!defined $dFound)  {$dFound=0;}
         if  (!defined $tFound)  {$tFound=0;}
	 print STDOUT '    xFound='.$xFound .'   xCol='.$xCol."\n";
         print STDOUT '    yFound='.$yFound .'   yCol='."(@{yCol})\n";
         print STDOUT '   y2Found='.$y2Found.'  y2Col='."(@{y2Col})\n";
         print STDOUT '    dFound='.$dFound .'   dest='.$dest.'  (batch='.$batch.')'."\n";
         print STDOUT '    tFound='.$tFound .'  title='.$title."\n";
         die "\n".'*** '.$script.': incorrect command line argument-'.$i.' ['.$arg.']!'."\n".$syntax;
      }
    }
  }

  $nyCol  = @yCol;
  $ny2Col = @y2Col;

  if ($dest eq 'SCREEN') {$status = &open_gnuplot(undef,undef,undef,['-persist']);}
  else                   {$status = &open_gnuplot();}
  if (!$status) {die '*** '.$script.': Cannot open pipe to gnuplot'."\n";}

  if ($dest eq 'IMG' || $dest eq 'PNG') {
#    print GNUPLOT 'set terminal png small picsize 640 480'."\n";	# old GnuPlot versions
     print GNUPLOT 'set terminal png small    size 640,480'."\n";
     print GNUPLOT 'set output "'.$file.'.png"'."\n";
  }
  elsif ($dest eq 'PS' || $dest eq 'PDF') {
     print GNUPLOT 'set terminal postscript portrait enhanced color blacktext'."\n";
     print GNUPLOT 'set output "'.$file.'.ps"'."\n";
  }
  if (length($title) > 0) {print GNUPLOT 'set title "'.$title.'"'."\n";}
  if ($ny2Col > 0)        {print GNUPLOT 'set y2tics'."\n";}

  print GNUPLOT 'plot "'.$file.'" using '.$xCol.':'.$yCol[0].' axes x1y1';
  for ($i=1; $i<$nyCol; $i++) {
     print GNUPLOT ', "'.$file.'" using '.$xCol.':'.$yCol[$i].' axes x1y1';
  }
  for ($i=0; $i<$ny2Col; $i++) {
     print GNUPLOT ', "'.$file.'" using '.$xCol.':'.$y2Col[$i].' axes x1y2';
  }
  print GNUPLOT "\n";

### This is replacing "-persist" that is not supported on Win32
# if ($dest eq 'SCREEN' && $^O =~ /Win/i) {print GNUPLOT 'pause mouse "Hit any key to close"'."\n";}
  if ($dest eq 'SCREEN' && $^O =~ /Win/i) {print GNUPLOT 'pause -1 "Hit any key to close"'."\n";}
  &close_gnuplot();
  if ($dest eq 'PDF') {
     if (defined $ps2pdf) {$status = system($ps2pdf, $file.'.ps', $file.'.pdf');}
     else                 {$status = -1;}
     if ($status) {die '*** '.$script.': failed to convert PS to PDF'."\n";}
     else         {unlink $file.'.ps';}		#delete postscript file
  }

### Display the plot (if not in a batch mode)
  if (!$batch) {
     if ($dest eq 'PDF') {
	if (defined $viewerPDF) {
           if ($^O !~ /Win/i) {
              system($viewerPDF.' '.$file.'.pdf &');
           } else {
              $fullname = File::Spec->rel2abs($file.'.pdf');
#             $fullname = abs_path($file.'.pdf');
              system(1, $viewerPDF, $fullname);
           }
	} else {
           print STDOUT '!!! '.$script.': no PDF viewer found in the system path!'."\n";
        }
     }
     elsif ($dest eq 'IMG' || $dest eq 'PNG') {
        if ($^O !~ /Win/i) {
	   if (defined $eog) {system($eog.' '.$file.'.png &');}
	   else  {print STDOUT '!!! '.$script.': no PNG viewer found in the system path!'."\n";}
        } else {
           $fullname = File::Spec->rel2abs($file.'.png');
#          $fullname = abs_path($file.'.png');
	   if    (defined $eog) {system(1, $eog, $fullname);}
	   elsif (-f 'C:/WINDOWS/system32/shimgvw.dll') {
              system(1, 'rundll32.exe','C:/WINDOWS/system32/shimgvw.dll,ImageView_Fullscreen', $fullname);
           }
        }
     }
     elsif ($dest eq 'PS') {
        if ($^O !~ /Win/i) {
	   if (defined $viewerPDF) {system($viewerPDF.' '.$file.'.ps &');}
	   else {print STDOUT '!!! '.$script.': no PS viwewer found in the system path!'."\n";}
        } else {
           $fullname = File::Spec->rel2abs($file.'.ps');
#          $fullname = abs_path($file.'.ps');
#          system(1, 'gsview32.exe', $fullname);
	   if    (defined $viewerPDF) {system(1, $viewerPDF, $fullname);}
	   elsif (defined $eog)       {system(1, $eog,       $fullname);}	#Irfan can also view PS files
	   else {print STDOUT '!!! '.$script.': no PS viewer found in the system path!'."\n";}
        }
     }
  }

  exit;							### End of program
