#!/usr/bin/perl -W
#--------------------------------------------------------------
#
# This script provides a switch between "external enable" synchronization
# and start-start synchronization of gonio and detector in the Pilatus FPE.
# The script:
# 1. Reads command line to find if EXT_ENABLE should be enabled (default is NOT)
# 2. Checks for "#define EXT_ENABLE" in fpe.st is commented out or not.
# 3. If the state of EXT_ENABLE corresponds to what was requested in the command
#    line and the executable in bin is newer than the source, the script exits.
# 4. Othewice, the script changes the state of "#define EXT_ENABLE" in fpe.st
#    to what is requested in the command line and recompiles the FPE
# 5. In any case the script also uploads the status and the date of FPE into a PV
#
# The script is supposed to be started (via ssh on ppu? or via sudo?) as user=gmca by
# nomarccdepics.pl. For that, it should be added to sudoers on respective
# computer (bl1ppu? blws2? bl1ws3?).
# ATTENTION: adding to /etc/sudoers is tricky. This does not work because for
# security reasons sudo wipes LD_LIBRARY_PATH and then libezca.so cannot be found:
##ALL ALL=(gmca) NOPASSWD: /usr/bin/perl -W /home/gmca/epics/pezca/prepare_pilatus_fpe.pl [a-z]*
# Contrary to the above, this works except for X11 (no Zenity popups):
##ALL ALL=(gmca) NOPASSWD: /bin/bash -lc /home/gmca/epics/pezca/prepare_pilatus_fpe.pl [A-z]*
# in the above line "-l" is important as it forces bash to include the environment.
# The sudo call will look like this (pay attention to quites around script+argument):
##sudo -u gmca /bin/bash -lc '/home/gmca/epics/pezca/prepare_pilatus_fpe.pl fddfdfdf'
#
# If running via ssh to bl1ppu:
# 1) Generate a pair in ~gmca/.ssh : ssh-keygen -t rsa -b 4096 -f ~gmca/.ssh/ppu_gmca_rsa
# 2) Add public key to authorized_keys: cat ~gmca/.ssh/ppu_gmca_rsa.pub >> ~gmca/.ssh/authorized_keys
# 3) Copy the private key to shared location: cp -p ~gmca/.ssh/ppu_gmca_rsa /mnt/software/scripts
# 4) chmod a+r-w /mnt/software/scripts/ppu_gmca_rsa
# 5) chown root:root /mnt/software/scripts/ppu_gmca_rsa
# 6) When doing ssh, point to the private key: ssh -i /mnt/software/scripts/ppu_gmca_rsa gmca@bl1ppu
# 7) Or better:
# ssh -i /mnt/software/scripts/ppu_gmca_rsa -X -o BatchMode=yes -o StrictHostKeyChecking=no -q gmca@bl1ppu
#
#                       Author: Sergey Stepanov
#
# Version-1.0:  2014/10/24	initial implementation
#--------------------------------------------------------------
  use Pezca;
  use strict;

  our ($NONINTERACTIVE, $USE_ZENITY_FOR_EXIT, $IGNORE_ERRORS, $beamline);
  my ($PEZCA, $script, $host, $bln, $b, $user, $status, $txt, @clients);
  my ($fpedir, $fpesrc, $fpebin, $datesrc, $datebin, $arch, $timestamp);
  my ($narg, $syntax, $ext_enable, $need_rebuild, $fpeinfo_PV, @ct);
  my (@lines, $nlines, $i, $found, $commented, $modified, $rebuild);

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;

  do $PEZCA.'/Common/common_subs.pl';

  select(STDERR);  $|=1;		# Flush output after each print
  select(STDOUT);  $|=1;		# Flush output after each print

  $script = &makeScriptName();

  if ($^O =~ /Win/i) {
     $txt = $script.': cannot run on Windows.';
     &press_key_exit($txt);
  }

  $NONINTERACTIVE      = 1;
  $USE_ZENITY_FOR_EXIT = 1;
  $IGNORE_ERRORS       = 1;

  if (defined $ENV{EPICS_HOST_ARCH}) {$arch=$ENV{EPICS_HOST_ARCH};}
  else                               {$arch='linux-x86_64';}
  $fpedir = '/mnt/software/epics_synApps_ppu/synApps/pilatus-1-0';
# $fpedir = '/mnt/software/epics_synApps_ppu/synApps/pilatus-1-0-test';
  $fpesrc = $fpedir.'/FPEApp/src/fpe.st';
  $fpebin = $fpedir.'/bin/'.$arch.'/FPE';
  $fpeinfo_PV = 'FPE:version';

  if    (defined $ENV{HOSTNAME}) {$host = $ENV{HOSTNAME};}
  elsif (defined $ENV{HOST})     {$host = $ENV{HOST};}
  if (!defined $host) {$host=`hostname`; $host=~s/\n//g;}
  if (!defined $host) {print STDERR '!!! HOSTNAME is undefined!'."\n\n"; $host='undef';}
  $host = lc($host);
  $host =~ s/\.gmca\.aps\.anl\.gov$//;
  if ($host !~ /^bl[123]/ && $host !~ /^keithley[123]$/) {
     $txt = 'Host "'.$host.'" is unsupported.';
     &press_key_exit($txt);
  }
  if ($host =~ /^keithley[123]$/) {$host =~ s/^keithley//; $host='bl'.$host.'keithley';}
  $b = substr($host,2,1);
  $beamline = qw(23i 23o 23b)[$b-1];
  do $PEZCA.'Common/beamline_'.$beamline.'.pl';
  $bln = 'bl'.$b;
  if ($b != 0) {@clients = ($bln.'ws1', $bln.'ws2', $bln.'ws3', $bln.'ws6', $bln.'keithley', $bln.'ppu', $bln.'upper');}
  else         {@clients = qw(px0 px1 px2 px3 px4 www);}

  if (! grep(/^$host$/,@clients)) {
     $txt = 'Pilatus FPE can be compiled from these computers only:'."\n"
	  ."(@{clients})\n"
          .'Host "'.$host.'" does not qualify.';
     &press_key_exit($txt);
  }

  if    (defined $ENV{USER})     {$user=$ENV{USER};}
  elsif (defined $ENV{USERNAME}) {$user=$ENV{USERNAME};}
  else                           {$user='';}
  if ($user ne 'gmca') {
     $txt = 'User "'.$user.'" does not qualify.';
     &press_key_exit($txt);
  }

### Get the command line:
  $syntax = "\n".'Syntax: '.$script.' [EXTERNAL_ENABLE]'."\n";
  $narg = @ARGV;
  if ($narg > 1) {
     $txt = 'Incorrect command line.'.$syntax;
     &press_key_exit($txt);
  }
  if ($narg > 0) {
     if ($ARGV[0] !~ /^external_enable$/i  &&
         $ARGV[0] !~ /^external_disable$/i &&
         $ARGV[0] !~ /^internal_advance$/i) {
       $txt = 'Incorrect command line: "'.$ARGV[0].'"'.$syntax;
       &press_key_exit($txt);
     }
     elsif ($ARGV[0] =~ /^external_enable$/i)  {$ext_enable=1;}
     elsif ($ARGV[0] =~ /^external_disable$/i) {$ext_enable=0;}
     elsif ($ARGV[0] =~ /^internal_advance$/i) {$ext_enable=0;}
  } else {
     $ext_enable=0;
  }
  if ($ext_enable) {print STDOUT $script.': requested FPE option is EXTERNAL_ENABLE'."\n";}
  else             {print STDOUT $script.': requested FPE option is INTERNAL_ADVANCE'."\n";}

  if (! -f $fpesrc) {
     $txt = 'FPE source "'.$fpesrc.'" not found.';
     &press_key_exit($txt);
  }

### Get file statistics:
# ($deviceNum,$inodeNum,$fileMode,$nHardLinks,		#0-3
#  $uid,$gid,$rdev,$size,$atime,$mtime,			#4-9
#  $ctime,$blksize,$blocks) = stat($fpesrc);		#10-12
  $datesrc = (stat($fpesrc))[9];			# this is $mtime
  $timestamp = localtime($datesrc);
  print STDOUT 'FPE source file='.$fpesrc."\n"
              .' -- Modification date: '.$timestamp."\n"
              .' -- Epoch='.$datesrc."\n";

### Read the FPE source file and find the flag:
  $status = &read_entire_file($fpesrc,\@lines);
  if ($status) {
     $txt = 'Failed to read FPE source "'.$fpesrc.'".';
     &press_key_exit($txt);
  }
  $nlines = @lines;
  $found = 0;
  $commented = 0;
  $modified = 0;
  for ($i=0; $i<$nlines; $i++) {
     if ($lines[$i] =~ /\#define\s+EXT_ENABLE/) {
	$found = 1;
	$lines[$i] =~ s/^\s+//;
	$lines[$i] =~ s/\s+$//;
	$lines[$i] =~ s/\s+/ /g;
	if ($lines[$i] =~ /^\/\*/ && $lines[$i] =~ /\*\/$/) {
           $commented=1;
        }
	elsif ($lines[$i] =~ /^\/\*/ || $lines[$i] =~ /\*\/$/) {
	   $txt = 'FPE source line: "'.$lines[$i].'"'."\n"
                 .'cannot be handled. Please place /* .. */ on one line';
           &press_key_exit($txt);
	}
	if ($ext_enable && $commented) {
	   ### Uncomment
	   $lines[$i] =~ s/\/\*//g;			#remove /*
	   $lines[$i] =~ s/\*\///g;			#remove */
	   $lines[$i] =~ s/^\s+//;			#remove leadong spaces
	   $modified = 1;
	}
	elsif (!$ext_enable && !$commented) {
	   $lines[$i] = '/* '.$lines[$i].' */';
	   $modified = 1;
	}
	last;
     }
  }
  if (! $found) {
     $txt = 'Failed to find "define EXT_ENABLE" in the FPE source "'.$fpesrc.'".';
     &press_key_exit($txt);
  }

  $rebuild = 0;
  if ($modified) {
     print STDOUT "\n".$script.': FPE will be rebult because source does not'."\n"
     .' correspond to the requested option EXT_ENABLE='.$ext_enable."\n\n";
     $rebuild = 1;
     ### Rewrite source:
     print STDOUT 'Rewriting '.$fpesrc.' ...'."\n";
     $status = &write_entire_file($fpesrc,\@lines);
     if ($status) {
        $txt = 'Failed to modify the FPE source "'.$fpesrc.'".';
        &press_key_exit($txt);
     }
     utime($datesrc,$datesrc,$fpesrc);			#update the file time to what it wads before
  } else {
     if (! -f $fpebin) {
        print STDOUT "\n".$script.': FPE rebuild requied because'."\n"
        .'FPE binary does not exist!.'."\n\n";
        $rebuild = 1;
     } else {
        $datebin = (stat($fpebin))[9];			# this is $mtime
        if ($datebin <= $datesrc) {
           print STDOUT "\n".$script.': FPE rebuild requied because'."\n"
           .'FPE binary is older than the source!.'."\n\n";
           $rebuild = 1;
        }
     }
  }
  if ($rebuild) {
     chdir($fpedir);
     print STDOUT "\n".$script.': Rebuilding FPE...'."\n\n";
     $status = system('/usr/bin/make', 'clean');
     $status = system('/usr/bin/make');
     print STDOUT "\n".$script.': ...Done rebuilding FPE'."\n\n";
     if ($status) {
        $txt = '*** '.$script.': error during rebuilding the FPE source!';
        &press_key_exit($txt);
     }
  } else {
     print STDOUT "\n".$script.': All looks good -- no FPE rebuilding required'."\n\n";
  }

### Uploading version to PV:
  $fpeinfo_PV = $beamline.':'.$fpeinfo_PV;
  @ct = localtime($datesrc);
# ($sec,$minute,$hour,$mday,$month,$year) = ($ct[0],$ct[1],$ct[2],$ct[3],$ct[4]+1,$ct[5]+1900);
  $timestamp = sprintf '%04d/%02d/%02d %02d:%02d',$ct[5]+1900,$ct[4]+1,$ct[3],$ct[2],$ct[1];
  if ($ext_enable) {$timestamp .= ' Mode=EXTERNAL_ENABLE';}
  else             {$timestamp .= ' Mode=INTERNAL_ADVANCE';}
  if (length($timestamp) > 39) {$timestamp = substr($timestamp,0,39);}
  $status = Pezca::PutStringOldCa($fpeinfo_PV, $timestamp);
  &report_only($script, $status, $fpeinfo_PV);

#==========================================================================

# sub write_entire_file($file,\@lines) {
sub write_entire_file($\@) {
  my ($script, $file, $linesref, $status, $nlines, $i);

  $script   = 'write_entire_file';
  $file     = shift(@_);
  $linesref = shift(@_);

  $status = open (FH,'> '.$file);
  if (!defined $status) {
     print STDERR '*** '.$script.': Could not open '.$file.' for writing'."\n";
     return 1;
  }
  $nlines = @$linesref;
  for ($i=0; $i<$nlines; $i++) {
     $status = print FH @$linesref[$i]."\n";
     if (! $status) {
        print STDERR '*** '.$script.': Could not write to '.$file."\n";
        return 2;
     }
  }
  close (FH);
  return 0;
}
