#!/usr/bin/perl -W
#----------------------------------------------------
# This script tunes networks parameters for 10Gbs and
# faster networks interfaces, The tuning parameters are
# per Dave Leibfritz. See also /etc/sysctl.conf
#
# Author: Sergey Stepanov
# Version 1.0 2015/09/29 initial implementation
# Version 1.1 2018/01/22 added watchdog control to reduce reboot times
# Version 1.2 2020/10/09 added host and checking for root
#----------------------------------------------------
  use strict;
  my (@inets, $inet, $str, @hispeednets, $user, $home, $net, $host);
  my ($LS, $IFCONFIG, $ETHTOOL, $MODPROBE, $SYSCTL, $WDCTL);

  select STDERR; $|=1;                          #set unbuffered output
  select STDOUT; $|=1;                          #set unbuffered output

  if ($^O =~ /Win/i) {
     print STDOUT "\n".'This script cannot run on Windows'."\n\n";
     exit;
  }

  if    (defined $ENV{USER})     {$user = $ENV{USER};}
  elsif (defined $ENV{USERNAME}) {$user = $ENV{USERNAME};}
  else                           {$user = 'unknown';}
  if    (defined $ENV{HOME})     {$home = $ENV{HOME};}
  else                           {$home = 'unknown';}

### When initial user is not root, but he issued "su",
### the environment "USER" is preserved, but "HOME" is not.
  if ($user ne 'root' && $home !~ /^\/root/) {
     print STDOUT "\n".'User='.$user;
     print STDOUT "\n".'Home='.$home;
     print STDOUT "\n".'This script should be executed as root or via sudo'."\n\n";
     exit;
  }

  if    (defined $ENV{HOSTNAME}) {$host = $ENV{HOSTNAME};}
  elsif (defined $ENV{HOST})	 {$host = $ENV{HOST};}
  else                           {$host=`hostname`; $host=~s/\n//g;}
  if (!defined $host) {print STDOUT '!!! HOSTNAME is undefined!'."\n\n"; $host='undef';}
  $host = lc($host);
  $host =~ s/\Q.gmca.aps.anl.gov\E$//;

  if    (-e '/bin/ls')            {$LS = '/bin/ls';}
  elsif (-e '/usr/bin/ls')        {$LS = '/usr/bin/ls';}
  elsif (-e `which ls`)           {$LS = `which ls`;}
  else                            {die 'ls command not found';}

  if    (-e '/sbin/ifconfig')     {$IFCONFIG = '/sbin/ifconfig';}
  elsif (-e '/usr/sbin/ifconfig') {$IFCONFIG = '/usr/sbin/ifconfig';}
  elsif (-e `which ifconfig`)     {$IFCONFIG = `which ifconfig`;}
  else                            {die 'ifconfig command not found';}

  if    (-e '/usr/sbin/ethtool')  {$ETHTOOL = '/usr/sbin/ethtool';}
  elsif (-e '/sbin/ethtool')      {$ETHTOOL = '/sbin/ethtool';}
  elsif (-e `which ethtool`)      {$ETHTOOL = `which ethtool`;}
  else                            {print STDERR 'ethtool command not found'."\n";}

  if    (-e '/sbin/modprobe')     {$MODPROBE = '/sbin/modprobe';}
  elsif (-e '/usr/sbin/modprobe') {$MODPROBE = '/usr/sbin/modprobe';}
  elsif (-e `which modprobe`)     {$MODPROBE = `which modprobe`;}
  else                            {print STDERR 'modprobe command not found'."\n";}

  if    (-e '/sbin/sysctl')       {$SYSCTL = '/sbin/sysctl';}
  elsif (-e '/usr/sbin/sysctl')   {$SYSCTL = '/usr/sbin/sysctl';}
  elsif (-e `which sysctl`)       {$SYSCTL = `which sysctl`;}
  else                            {print STDERR 'sysctl command not found'."\n";}

  if    (-e '/usr/bin/wdctl')     {$WDCTL = '/usr/bin/wdctl';}
  elsif (-e `which wdctl`)        {$WDCTL = `which wdctl`;}
  else                            {print STDERR 'wdctl command not found'."\n";}

### 164.54.103. is 10Gbp GMCA-APS (Chelsio)
### 10.10.10.   is 40Gbs link between PPU and Pilatus computer (Melanox)
### 10.20.103.  is 40Gbs GMCA internal storage access network (Melanox)
### 10.30.103   is 10Gbp BB-beegfs (Chelsio)
### 192.5.180.  is 10Gbs ScienceDMZ (Chelsio, ws8 only)
  @hispeednets = qw(164.54.103. 10.10.10. 10.20.103 10.30.103 192.5.180.);
  @inets = split(/\n/, `$LS -1 /sys/class/net`);

  foreach $inet (@inets) {
     $str = `$IFCONFIG $inet`;
     foreach $net (@hispeednets) {
        if (index($str,$net) != -1) {
           print STDOUT "\n\n".'Interface '.$inet.' connects to '.$net.' subnet'."\n";
           &runcmd($ETHTOOL, '-K '.$inet.' tso on');
	   &runcmd($ETHTOOL, '-C '.$inet.' rx-usecs 75');
	}
     }
  }

### See: https://fasterdata.es.net/host-tuning/linux/
  print STDOUT "\n\n";
  &runcmd($MODPROBE, 'tcp_htcp');
  &runcmd($SYSCTL, '-w net.ipv4.tcp_congestion_control=htcp');
# &runcmd($SYSCTL, 'net.ipv4.tcp_available_congestion_control');
  ### Increase TCP buffers up to 64MB (for 10Gbps)
  &runcmd($SYSCTL, '-w net.core.rmem_max=134217728');
  &runcmd($SYSCTL, '-w net.core.wmem_max=134217728');
  ### Increase TCP buffers up to 2GB (for 40Gbps and 100Gbps)
# &runcmd($SYSCTL, '-w net.core.rmem_max=2147482548');
# &runcmd($SYSCTL, '-w net.core.wmem_max=2147482548');
  ### Increase Linux autotuning TCP buffer limit to 32MB (for 10Gbps)
  &runcmd($SYSCTL, '-w net.ipv4.tcp_rmem="4096 87380 67108864"');
  &runcmd($SYSCTL, '-w net.ipv4.tcp_wmem="4096 65536 67108864"');
  ### Increase Linux autotuning TCP buffer limit to 1GB (for 40Gbps and 100Gbps -untested)
# &runcmd($SYSCTL, '-w net.ipv4.tcp_rmem="4096 87380 1073741274"');
# &runcmd($SYSCTL, '-w net.ipv4.tcp_wmem="4096 65536 1073741274"');
  &runcmd($SYSCTL, '-w net.core.netdev_max_backlog=250000');
  &runcmd($SYSCTL, '-w net.ipv4.tcp_mtu_probing=1');
  ### SS: added 2021.09.07
  &runcmd($SYSCTL, '-w net.core.default_qdisc=fq');

### On upper rsh server cannot be restarted with ipv6 disabled.
### We can handle it in restart_rsh.sh
# if ($host !~ /bl?upper/) {
     &runcmd($SYSCTL, '-w net.ipv6.conf.all.disable_ipv6=1');
#    &runcmd($SYSCTL, '-w net.ipv6.conf.default.disable_ipv6=1');
# }

# &runcmd($WDCTL, '-s 5');				#change watchdog timeout from 30s to 5s
#=======================================================
sub runcmd($$) {
  my ($cmd, $arg, $str);

  $cmd = shift(@_);
  $arg = shift(@_);
  if (! defined $cmd) {return;}
  if (defined $arg) {$cmd .= ' '.$arg;}

  print STDOUT 'Running: '.$cmd."\n";
  $str = `$cmd 2>&1`;
  if (length($str) > 0) {print STDOUT 'Response: '.$str."\n";}
}
