#!/usr/bin/perl -W
#--------------------------------------------------------------
# This script reads the JBluIce 'PmacMotorAssembly' table in mySQL
# and fills it with InPos band data for most of motors except for
# complex ones (detector and mono energy). For the list of exceptions
# (skipped axes) see @axesExceptions below.
# The script can set InPos data for JBluIce a bit higher than in EPICS
# -- see $factor below.
#
#       		Author: Sergey Stepanov
#
# Version-1.0:  2011/08/04 using Pezca and DBI
#--------------------------------------------------------------
# There are four simple steps to follow when executing an SQLquery with the Perl DBI.
#
#1. Begin by initializing a database handle with a call to the connect()
#   method. This method accepts connection parameters as a string containing
#   the database type ("mysql"), the host name ("localhost") and the database
#   name ("db2"). The database user name ("joe") and password ("guessme") is
#   also provided to the connect() method as the second and third argument
#
#2. Create the SQL query string, and execute it with either the do() method
#   or the prepare() and execute() methods. Use the do() method for one-off
#   INSERT, UPDATE or DELETE queries, but use the prepare() and execute()
#   methods for SELECT queries. The result object of these methods will be
#   different depending on the query type and whether or not it was successful.
#   Successful SELECT queries return a result object; successful
#   INSERT/UPDATE/DELETE queries with do() return the number of rows affected;
#   and unsuccessful queries return an error.
#
#3. For SELECT queries, the result object may be processed further to
#   extract data from it. Used in a loop, the fetchrow_hashref() method
#   retrieves each record as a Perl hash. You access individual fields of the
#   record by calling the appropriate key of the hash.
#
#4. End the session with a call to the disconnect() method.
#--------------------------------------------------------------
  use strict;
  use Pezca;
  use DBI;

  our ($beamline);
  our ($mysqlHost, $mysqlUser, $mysqlPass, $mysqlDb);
  our ($NONINTERACTIVE, $EXIT_WITH_SIGNAL_HANDLER);
  my ($PEZCA, $script, $syntax, $arg, @beamlinelist);
  my ($mysqlTable, $dbh, $sth, $hnd, $nrecords, $nrows, $i, $j);
  my (@name, @bl, @grp, @assy, @axis, @jassy, @inpos);
  my ($status, $pmac, $aod, $aodIndex, @axes, @drives, @motors, @mtNumbers);
  my ($as, $ax, $dr, $mt, $nAxes, $nDrives, $nMotors);
  my (@inposmtr, $scale, $inposband, @axesExceptions, $factor);

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}

  do $PEZCA.'/Common/motor_subs.pl';
  do $PEZCA.'/Common/common_subs.pl';
  do $PEZCA.'/Common/beamline.pl';

  select STDOUT; $|=1;					#set unbuffered output

################Change this if needed###################
### Defaults:
  @beamlinelist   = ('23i:', '23o:', '23b:', '23d:');
# @axesExceptions = ('MO:E', 'CCD:2Q', 'CCD:D');
  @axesExceptions = ('MO:E', 'CCD:2Q');
  $factor = 1.5;					#InPosBand(mySQL)/InPosBand(PMAC), must be >=1
################End of editable part####################

  $script = &makeScriptName();
  $syntax = "\n\n".'Syntax: '.$script.' [beamline]'."\n";

#----------------------------------------------------------------
  foreach $arg (@ARGV) {
     $arg =~ s/^-{0,2}//;					#trim leading dashes
     if (grep(/^$arg:?$/, @beamlinelist)) {
        if ($arg ne $beamline) {
           $beamline = $arg;
           $beamline =~ s/:$//g;
           print STDOUT "\n".'!!! '.$script.': cross-beamline operation for '.$beamline.'!'."\n\n";
           do $PEZCA.'/Common/beamline_'.$beamline.'.pl';
	}
     }
     else {
        &press_key_exit('*** '.$script.': incorrect cmdline.'.$syntax);
     }
  }
  &ask_to_continue('I M P O R T A N T !!!'."\n".
'This script will update Inpos bands for all motors in JBluIce mySQL config.'."\n".
'Any previously entered values will be automatically overwritten.'."\n".
'Please proceed only if you fully understand what you are doing.');
#----------------------------------------------------------------
  $SIG{HUP}  = 'signal_handler';				#(1) hangup (see signum.h)
  $SIG{INT}  = 'signal_handler';				#(2) interrupt (see signum.h)
  $SIG{QUIT} = 'signal_handler';				#(3) quit (see signum.h)
# $SIG{KILL} = 'signal_handler';				#(9) this is what is sent by kill -9 -- unblockable
  $SIG{TERM} = 'signal_handler';				#(15) terminate -- sent by kill without paramaters
# $SIG{STOP} = 'signal_handler';				#(19) stop -- unblockable
  $EXIT_WITH_SIGNAL_HANDLER = 1;
#----------------------------------------------------------------
  if    (!defined $factor) {$factor=1;}
  elsif ($factor < 1)      {$factor=1;}

### ------------------------------------------------------------------------
### Read list of motors from mySQL
### ------------------------------------------------------------------------
  $mysqlTable = 'PmacMotorAssembly';
  print STDOUT $script.': connecting to mySQL database='
                      .$mysqlDb.' on host='.$mysqlHost."\n";
  $dbh = DBI->connect('DBI:mysql:host='.$mysqlHost.';database='.$mysqlDb,
                      $mysqlUser,
                      $mysqlPass,
                      {AutoCommit=>1,PrintError=>0,});
  if (!defined $dbh) {
     print STDOUT '*** '.$script
                        .': cannot connect to mySQL DB='
                        .$mysqlDb.' on '.$mysqlHost."\n";
     &signal_handler('quit');
  }

  print STDOUT $script.': opening mySQL table='.$mysqlTable."\n";
  $sth = $dbh->prepare('SELECT * FROM '.$mysqlTable);
  if (!defined $sth) {
     print STDOUT '*** '.$script
                        .': error on mySQL table='.$mysqlTable
                        .' -- '.$dbh->errstr."\n";
     &signal_handler('quit');
  }

  print STDOUT $script.': requesting all records'."\n";
  $nrecords = $sth->execute();				#nrecords works for MySQL only. For SQLite result is def/undef only.
  if (!defined $nrecords) {
     print STDOUT '*** '.$script
                        .': could not execute SELECT on Table="'
                        .$mysqlTable.'" -- '.$sth->errstr."\n";
     &signal_handler('quit');
  }
  print STDOUT $script.': found '.$nrecords.' record'."\n";
  for ($i=0; $i<$nrecords; $i++) {
     $hnd = $sth->fetchrow_hashref();
     if (!defined $hnd) {
        print STDOUT '*** '.$script.': Error reading line='.$i
                           .' in table '.$mysqlTable
                           .' on mySQL server '.$mysqlHost."\n";
        &signal_handler('quit');
     }
     $name[$i]  = $hnd->{'name'};
     $bl[$i]    = $hnd->{'beamline'};
     $grp[$i]   = $hnd->{'device'};
     $assy[$i]  = $hnd->{'control'};
     $axis[$i]  = $hnd->{'motor'};
     $jassy[$i] = $hnd->{'assembly'};	if (!defined $jassy[$i])  {$jassy[$i]='[NONE]';}
					if (length($jassy[$i])==0){$jassy[$i]='[NONE]';}
     $inpos[$i] = $hnd->{'inPosBand'};	if (!defined $inpos[$i])  {$inpos[$i]='[NULL]';}
  }
  $sth->finish();

### Now we have the full list of records in the mySQL 'PmacMotorAssembly' table.
### Go over each record and calculate InPos band as max over all drives when
### the record corresponds to an axis or for the particular drive when the
### record corresponds to a drive.
  $sth = $dbh->prepare('UPDATE '.$mysqlTable.' SET inPosBand=? WHERE name=?');

  for ($i=0; $i<$nrecords; $i++) {	#===================================
     print STDOUT "\n".'Processing Name='.$name[$i]
                               .'  assy='.$bl[$i].$grp[$i].$assy[$i]
                               .'  axis='.$bl[$i].$grp[$i].$axis[$i]
                               .'  JBluIce_assembly='.$jassy[$i]
                               .'  InposBand='.$inpos[$i]."\n";
     if ($bl[$i] !~ /^$beamline/) {
        print STDOUT '*** '.$script.': incorrect beamline='.$bl[$i]
                           .' for name='.$name[$i]
                           .' in table '.$mysqlTable
                           .' on mySQL server '.$mysqlHost."\n";
	next;
#       &signal_handler('quit');
     }
     $as = $bl[$i].$grp[$i].$assy[$i]; $as =~ s/^${beamline}://; $as =~ s/:$//;
     $ax = $bl[$i].$grp[$i].$axis[$i]; $ax =~ s/^${beamline}://; $ax =~ s/:$//;
     if (grep(/^$ax:?$/, @axesExceptions)) {
        print STDOUT '...skipping exception: '.$ax."\n";
        next;
     }
     $status = &searchConfigTcl ($bl[$i],$as,$ax,
                                \$pmac,\$aod,\$aodIndex,
                                \@axes,\@drives,\@motors,\@mtNumbers);
     if ($status) {
        print STDOUT '*** '.$script.': error from searchConfigTcl'
                           .' for name='.$name[$i]
                           .', assembly='.$as
                           .', axis='.$ax."\n";
        next;
#       &signal_handler('quit');				#skip this record
     }
     $nAxes   = @axes;
     $nDrives = @drives;
     $nMotors = @motors;
     if ($nDrives != $nMotors) {
        print STDOUT '*** '.$script.': motors='.$nMotors.' and drives='.$nDrives
                           .' mismatch in data returned by searchConfigTcl'
                           .' for name='.$name[$i]
                           .', assembly='.$as
                           .', axis='.$ax."\n";
        next;
#       &signal_handler('quit');				#skip this record
     }
     @inposmtr = ();
     if ($aod) {   #--------------------------------------------#if axis
	$inposband = 0.;
        for ($j=0; $j<$nDrives; $j++) {
	   $dr = $drives[$j];
	   $mt = $motors[$j];
	   if ($dr !~ /^${beamline}/) {$dr = $beamline.':'.$dr;}  if ($dr !~ /:$/) {$dr .= ':';}
	   if ($mt !~ /^${beamline}/) {$mt = $beamline.':'.$mt;}  if ($mt !~ /:$/) {$mt .= ':';}
### Request InPosBand as Ixx28 (to be divided by 16):
           $inposmtr[$j] = &readMtrRsp($mt, 'i'.$mtNumbers[$j].'28');
### Request scale factor 'motors to drives':
           ($status,$scale) = Pezca::GetDouble($dr.'Scale');
	   if ($status) {
	      &report_only($script,$status,$dr.'Scale');
	      next;                   				#skip this record
#             &signal_handler('quit');
	   }
	   $inposmtr[$j] = $scale * $inposmtr[$j] / 16.;
	   $inposband += $inposmtr[$j]**2;
        }
	$inposband = $factor * sqrt($inposband);                #take as a sum of squares
#	$inposband = $factor * &findMaxRef(\@inposmtr);		#take as the maximum band
     }
     else {	   #--------------------------------------------#if drive
	$dr = $drives[$aodIndex];
	$mt = $motors[$aodIndex];
	   if ($dr !~ /^${beamline}/) {$dr = $beamline.':'.$dr;}  if ($dr !~ /:$/) {$dr .= ':';}
	   if ($mt !~ /^${beamline}/) {$mt = $beamline.':'.$mt;}  if ($mt !~ /:$/) {$mt .= ':';}
### Request InPosBand as Ixx28 (to be divided by 16):
        $inposmtr[0] = &readMtrRsp($mt, 'i'.$mtNumbers[$aodIndex].'28');
### Request scale factor 'motors to drives':
        ($status,$scale) = Pezca::GetDouble($dr.'Scale');
	if ($status) {
	   &report_only($script,$status,$dr.'Scale');
	   next;                   				#skip this record
#          &signal_handler('quit');
	}
	$inposband = $factor * $scale * $inposmtr[0] / 16.;
     }             #--------------------------------------------#end inposband calc.
### Round (?):
     $inposband = sprintf '%.2g', $inposband;
     print STDOUT 'Updating mySQL '.$name[$i].' ('.$ax.') --> InposBand='
                                   .$inposband.', was '.$inpos[$i]."\n";
### The do-version of update is simpler (does not require prepare-execute-finish):
#    $nrows = $dbh->do('UPDATE '.$mysqlTable.' SET inPosBand='.$inposband.' WHERE name="'.$name[$i].'"');
     $nrows = $sth->execute($inposband, $name[$i]);				#nrecords works for MySQL only. For SQLite result is def/undef only.
     if (!defined $nrows || $nrows != 1) {
        print STDOUT '*** '.$script
                           .': could not execute UPDATE on Table="'
                           .$mysqlTable.'" -- '.$sth->errstr."\n";
 	next;
#       &signal_handler('quit');
     }
  }                                     #===================================
  $sth->finish();
  $dbh->disconnect();

### ------------------------------------------------------------------------
### Cleanup
### ------------------------------------------------------------------------
  &signal_handler('quit');

#=================================================================

sub signal_handler {
# local($sig) = @_;
  my $sig = shift;
  if ($sig ne 'quit') {print STDOUT "\n".'*** '.$script.': caught a SIG'.$sig."\n\n";}

  $NONINTERACTIVE = 1;
  $EXIT_WITH_SIGNAL_HANDLER = 2;

  if (defined $dbh) {$dbh->disconnect();}

  print STDOUT $script.': clearing all CA monitors'."\n";
  &clearAllCaMonitors();
  print STDOUT $script.': done with exiting!'."\n";
  exit;
}
