#!/usr/bin/perl -W
#--------------------------------------------------------------
# This script verifies PMAC configuration lists in configNNN.tcl
# against EPICS data (tsub, pcs, mtrno, mtr/drv/axs names & count).
#
#                       Author: Sergey Stepanov
#
# Version-1.0:  2019/02/25	initial implementation
# Version-1.1:  2020/02/10      change that order of axes should not matter
#--------------------------------------------------------------

  use Pezca;
  use strict;
  use warnings;

  our ($USE_ZENITY_FOR_EXIT, $IGNORE_ERRORS, $beamline);
  my ($PEZCA, $script, $arg, @hash, $label, $assy, $axs, $drv); 
  my ($mtr, $mno, $pmac, $pcs, $nassy, $naxs, $ndrv, $nmtr);
  my (@beamlinelist, $syntax, $i, $j, $m);
  my (@axsEPICS, @drvEPICS, @mtrEPICS, $status, $pcsEPICS, $pmacEPICS, $mnoEPICS);
  my (@axsCONFG, @drvCONFG, @mtrCONFG, @mtrEPHSH, $str, $d);

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;

  do $PEZCA.'/Common/motor_subs.pl';
  do $PEZCA.'/Common/common_subs.pl';
  do $PEZCA.'/Common/beamline.pl';

  select STDERR; $|=1;					#set unbuffered output
  select STDOUT; $|=1;					#set unbuffered output

  $USE_ZENITY_FOR_EXIT = 1;
  $IGNORE_ERRORS       = 1;

  $script = &makeScriptName();
  @beamlinelist  = qw(23i 23o 23b 23d);
  $syntax = "\n".'Syntax: '.$script.' [beamline]'."\n\n";

# @ARGV = qw(23b);                                      #for testing
  foreach $arg (@ARGV) {
     if (grep(/^${arg}/, @beamlinelist)) {
        $arg =~ s/:$//;						#trim ':'
        if ($arg ne $beamline) {
           print STDOUT "\n".'!!! '.$script.': Cross-beamline operation: controlling '.$arg.' from '.$beamline."\n\n";
           $beamline = $arg;
           do $PEZCA.'/Common/beamline_'.$beamline.'.pl';
	}
     } else {
        &press_key_exit('*** '.$script.': Unrecognized cmdline parameter "'.$arg.'"'.$syntax,1);
     }
  }

  $nassy = &generateComponentList($beamline,\@hash);
  if ($nassy < 0) {
     &press_key_exit('*** '.$script.': error='.$nassy.' from generateComponentList',1);
  }

  for ($i=0; $i<$nassy; $i++) {
     $label = $hash[$i]->{'label'};
     $assy  = $hash[$i]->{'assy'};
     $pmac  = $hash[$i]->{'pmac'};
     $pcs   = $hash[$i]->{'pcs'};
     $naxs  = $hash[$i]->{'naxes'};  
     $ndrv  = $hash[$i]->{'ndrives'};  
     $nmtr  = $hash[$i]->{'nmotors'};
     printf STDOUT 'Verifying assembly=%s, (%s)'."\n", $assy, $label;
#    printf STDOUT "\n".'--------------'."\n".'label=%s'."\n",$label;
#    printf STDOUT 'assy=%s, pmac=%s, pcs=%s, naxes=%s, ndrives=%s, nmotors=%s'."\n",$assy,$pmac,$pcs,$nxs,$ndrv,$nmtr;
     if ($assy !~ /^${beamline}.*:$/)           {&press_key_exit('*** '.$script.': incorrect assy='.$assy,1);}
     if ($pmac !~ /^pmac[0-9]{2}:$/)            {&press_key_exit('*** '.$script.': incorrect pmac='.$pmac.' for assy='.$assy,1);}
     if (!&is_int($pcs)  || $pcs<1  || $pcs>16) {&press_key_exit('*** '.$script.': incorrect pcs='.$pcs.' for assy='.$assy,1);}
     if (!&is_int($naxs) || $naxs<0 || $naxs>3) {&press_key_exit('*** '.$script.': incorrect naxes='.$naxs.' for assy='.$assy,1);}
     if (!&is_int($ndrv) || $ndrv<1 || $ndrv>3) {&press_key_exit('*** '.$script.': incorrect ndrives='.$ndrv.' for assy='.$assy,1);}
     if (!&is_int($nmtr) || $nmtr != $ndrv)     {&press_key_exit('*** '.$script.': incorrect nmotors='.$nmtr.' for assy='.$assy,1);}

     $pmacEPICS = &getPmac($assy,\$status);
     if ($status) {&press_key_exit('*** '.$script.': getPmac failed for assy='.$assy,1);}
     if ($pmacEPICS !~ /:$/) {$pmacEPICS .= ':';}
     $pmacEPICS =~ s/^${beamline}://;
     if ($pmacEPICS ne $pmac) {&press_key_exit('*** '.$script.': TCL-Tsub pmac mismatch: '.$pmac.' != '.$pmacEPICS. ' for assy='.$assy,1);}

     $pcsEPICS = &getPcs($assy,\$status);
     if ($status) {&press_key_exit('*** '.$script.': getPCS failed for assy='.$assy,1);}
     if ($pcsEPICS != $pcs) {&press_key_exit('*** '.$script.': TCL-Tsub PCS mismatch: '.$pcs.' != '.$pcsEPICS. ' for assy='.$assy,1);}

     $status = &getAssyDataFromTsub($assy,\@axsEPICS,\@drvEPICS,\@mtrEPICS); 
     if ($status) {&press_key_exit('*** '.$script.': getAssyDataFromTsub failed for assy='.$assy,1);}
     if (scalar(@axsEPICS) != $naxs) {&press_key_exit('*** '.$script.': TCL-Tsub naxes mismatch: '  .$naxs.' != '.scalar(@axsEPICS). ' for assy='.$assy,1);}
     if (scalar(@drvEPICS) != $ndrv) {&press_key_exit('*** '.$script.': TCL-Tsub ndrives mismatch: '.$ndrv.' != '.scalar(@drvEPICS). ' for assy='.$assy,1);}
     if (scalar(@mtrEPICS) != $nmtr) {&press_key_exit('*** '.$script.': TCL-Tsub nmotors mismatch: '.$nmtr.' != '.scalar(@mtrEPICS). ' for assy='.$assy,1);}
     
     @axsCONFG = ();
     for ($j=0; $j<$naxs; $j++) {
        $m = $j+1;
        $axs = $hash[$i]->{'axis'.$m};
        if ($axs !~ /^${beamline}.*:$/) {&press_key_exit('*** '.$script.': incorrect axis'.$m.'='.$axs.' for assy='.$assy,1);}
        push(@axsCONFG,$axs);
        if ($axsEPICS[$j] !~ /:$/) {$axsEPICS[$j] .= ':';}
     }
     @axsCONFG = sort(@axsCONFG);
     @axsEPICS = sort(@axsEPICS);
     for ($j=0; $j<$naxs; $j++) {
        if ($axsCONFG[$j] ne $axsEPICS[$j]) {
           $str = 'AXES: Config=('; foreach $d (@axsCONFG) {$str .= $d.' ';} $str =~ s/ $/\)/;
           $str .=       ' Tsub=('; foreach $d (@axsEPICS) {$str .= $d.' ';} $str =~ s/ $/\)./;
           &press_key_exit('*** '.$script.': TCL-Tsub axis mismatch: '.$axsCONFG[$j].' != '.$axsEPICS[$j]
                                 . ' for assy='.$assy."\n".$str,1);
        }
#       printf STDOUT '    axis%d=%s'."\n", $m, $axs;
     }

     @drvCONFG = ();
     for ($j=0; $j<$ndrv; $j++) {
        $m = $j+1;  
        $drv = $hash[$i]->{'drive'.$m};
        if ($drv !~ /^${beamline}.*:$/) {&press_key_exit('*** '.$script.': incorrect drive'.$m.'='.$drv.' for assy='.$assy,1);}
        push(@drvCONFG,$drv);
        if ($drvEPICS[$j] !~ /:$/) {$drvEPICS[$j] .= ':';}
     }        
     @drvCONFG = sort(@drvCONFG);
     @drvEPICS = sort(@drvEPICS);
     for ($j=0; $j<$ndrv; $j++) {
        if ($drvCONFG[$j] ne $drvEPICS[$j]) {
           $str = 'DRIVES: Config=('; foreach $d (@drvCONFG) {$str .= $d.' ';} $str =~ s/ $/\)/;
           $str .=         ' Tsub=('; foreach $d (@drvEPICS) {$str .= $d.' ';} $str =~ s/ $/\)./;
           &press_key_exit('*** '.$script.': TCL-Tsub drive mismatch: '.$drvCONFG[$j].' != '.$drvEPICS[$j]
                                 . ' for assy='.$assy."\n".$str,1);
        }
#       printf STDOUT '   drive%d=%s'."\n", $m, $drv;
     }

     @mtrCONFG = ();
     @mtrEPHSH = ();
     for ($j=0; $j<$nmtr; $j++) {
        $m = $j+1; 
        $mtr = $hash[$i]->{'motor'.$m};
        $mno = $hash[$i]->{'mtr'.$m.'_no'};
        if ($mtr !~ /^${beamline}.*:$/) {&press_key_exit('*** '.$script.': incorrect motor'.$m.'='.$mtr.' for assy='.$assy,1);}
        if (!&is_int($mno) || $mno<1 || $mno>32) {&press_key_exit('*** '.$script.': incorrect motor'.$m.'_number='.$mno.' for assy='.$assy,1);}
        push @mtrCONFG, {name=>$mtr, number=>$mno};
        if ($mtrEPICS[$j] !~ /:$/) {$mtrEPICS[$j] .= ':';}
        $mnoEPICS = &getMtrNo($mtr,\$status);
        if ($status) {&press_key_exit('*** '.$script.': getMtrNo failed for assy='.$assy.', motor='.$mtr,1);}
        push @mtrEPHSH, {name=>$mtrEPICS[$j], number=>$mnoEPICS};
     }        
     @mtrCONFG = sort{$a->{name} cmp $b->{name}} @mtrCONFG;
     @mtrEPHSH = sort{$a->{name} cmp $b->{name}} @mtrEPHSH;
     for ($j=0; $j<$nmtr; $j++) {
        if ($mtrCONFG[$j]->{name} ne $mtrEPHSH[$j]->{name}) {
           $str = 'MOTORS: Config=('; foreach $d (@mtrCONFG) {$str .= $d->{name}.' ';} $str =~ s/ $/\)/;
           $str .=         ' Tsub=('; foreach $d (@mtrEPHSH) {$str .= $d->{name}.' ';} $str =~ s/ $/\)./;
           &press_key_exit('*** '.$script.': TCL-Tsub motor mismatch: '.$mtrCONFG[$j]->{name}.' != '.$mtrEPHSH[$j]->{name}
                                 . ' for assy='.$assy."\n".$str,1);
        }
        if ($mno ne $mnoEPICS) {
           &press_key_exit('*** '.$script.': TCL-Tsub mtrNo mismatch: '.$mtrCONFG[$j]->{number}.' != '.$mtrEPHSH[$j]->{number}
                                 . ' for assy='.$assy.', motor='.$mtrCONFG[$j]->{name},1);
        }
#       printf STDOUT '   motor%d=%s, no=%d'."\n", $m, $mtr, $mno;
     }
  }
  print STDOUT "\n".'All OK'."\n\n";
  &clearAllCaMonitors();
  exit 0;
