package gov.anl.gmca.analysis;

import gov.anl.gmca.analysis.DB2Html.RESULT_PAGE;
import gov.anl.gmca.analysis.helpers.ExportStrategy;
import gov.anl.gmca.analysis.helpers.StrategyWithSpaceGroup;
import gov.anl.gmca.io.Database;
import gov.anl.gmca.io.Database.DBName;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.ExecTask;
import gov.anl.gmca.io.PVLookup;
import gov.anl.gmca.syslog.Log;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.math.NumberUtils;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.browser.Browser;
import org.eclipse.swt.browser.LocationEvent;
import org.eclipse.swt.browser.LocationListener;
import org.eclipse.swt.custom.CTabFolder;
import org.eclipse.swt.widgets.Item;

import java.io.File;

public class UrlTaskHandler implements LocationListener {
	private CTabFolder folder;
	private Item item;
	private Browser browser;
	private RESULT_PAGE title;

	public UrlTaskHandler(CTabBrowser cb) {
		this.folder = cb.getFolder();
		this.item = cb.getItem();
		this.browser = cb.getBrowser();
		this.title = cb.getPageType();
		//Log.error("UrlTaskHandler invoked with title = " + title);
	}

	private Double getValue(String strValue, Double defaultValue) {
		Double value = defaultValue;
		if (NumberUtils.isCreatable(strValue)) {
			value = Double.valueOf(strValue);
		} else {
			Log.error(String.format("strategy: %s is not a number, use default value %.2f.", strValue, defaultValue));
		}
		return value;
	}

	@Override
	public void changing(LocationEvent event) {
		// NB a special url process:id to handle process form here
		//Log.error("location listener : " + title + " " + event);
		if (event.location.startsWith("process:")) {
			event.doit = false;
			item.dispose(); // TODO item dispose causing the focus of the dialog to pin to the wrong page
			folder.setSelection(1);
			int id = Integer.valueOf(event.location.replace("process:", ""));
			Log.note("open process page for " + id + " from " + this.title + " page " + event.location);
			new WdgDataProcDialog(browser.getShell(), id, RESULT_PAGE.PROCESSING_RESULTS).open();
			folder.setSelection(0);
		} else if (event.location.startsWith("strategy:")) {
			event.doit = false;
			item.dispose();
			folder.setSelection(1);
			int id = Integer.valueOf(event.location.replace("strategy:", ""));
			Log.note("calculate strategy for " + id + " from " + this.title + " page " + event.location);
			new WdgDataProcDialog(browser.getShell(), id, RESULT_PAGE.SCREENING_RESULTS).open();
		} else if (event.location.startsWith("strategy2:")) {
			event.doit = false;
			// item.dispose();
			folder.setSelection(1);

			String[] params = event.location.replace("strategy2:", "").split(":");
			new StrategyWithSpaceGroup(params, browser.getShell());
		} else if (event.location.startsWith("refresh:")) {
			event.doit = false;
			String html = new DB2Html(this.title).html();
			browser.setText(html);
			item.setText(this.title.getPageTitle());
		} else if (event.location.startsWith("export:")) {
			event.doit = false;
			folder.setSelection(1);
			String[] values = event.location.replace("export:", "").split(":");
			
			try {
				int id = Integer.valueOf(values[0]);
				Double startosc = getValue(values[1], 0.0);
				Double endosc = getValue(values[2], 180.0);
				Double delta = getValue(values[3], 0.2);
				Double distance = getValue(values[4], 300.0);
				Log.note(String.format("DEBUG export requested with values = %s, start=%s, end=%s, delta=%s, distance=%s",
						event.location, startosc, endosc, delta, distance));
				new ExportStrategy(startosc, endosc, delta, distance, true);
				item.dispose();
			} catch (Exception e) {
				Log.error("export strategy failed for request " + event.location);
				e.printStackTrace();				
			}

		} else if (event.location.startsWith("delete:")) {
			//item.dispose();
			event.doit = false;

			// for some reason title is incorrect here, why????
			//Log.error("before dialog " + folder.getSelectionIndex());

			int id = Integer.valueOf(event.location.replace("delete:", ""));
			String pvMessage;
			if (folder.getSelectionIndex() == 0) {
				pvMessage = id + " PROCESS DELETE";
			} else {
				pvMessage = id + " STRATEGY DELETE";
			}
			boolean confirmed = MessageDialog.openConfirm(browser.getShell(), "Confirm",
					"Please confirm that you want to delete this row permanently.");

			if (confirmed) {
				Database.exec(DBName.USERDATA, String.format("delete from pipelinestatus where id=%d;", id));
				Epics.put(PVLookup.PV_ANALYSIS.pipelineStatus, pvMessage);
				Log.note("row id=" + id + " deleted from " + this.title + " page " + event.location + " pv="+pvMessage);
			}
			//Log.error("1 after dialog " + folder.getSelectionIndex());
		} else if (event.location.startsWith("file://") && (event.location.endsWith(".cbf")
				|| event.location.endsWith(".h5") || event.location.endsWith(".hdf5"))) {
			event.doit = false;
			startProgram("adxv -small_spots " + event.location.replace("file://", ""), null);
		} else if (event.location.startsWith("file://") && (event.location.endsWith("coot.sh"))) {
			// launch coot.sh
			event.doit = false;
			String cootsh = event.location.replace("file://", "");
			new File(cootsh).setExecutable(true);
			String cootdir = cootsh.substring(0, cootsh.lastIndexOf("/"));
			startProgram(cootsh, cootdir);	
			
			if (item.getText().length() <= 1) { // disable creating new tabs if no title is set
				item.dispose();
				folder.setSelection(0);
			}
		
		} else if (event.location.startsWith("file://") && (event.location.endsWith("XDS.INP"))) {
			// using xdsgui to explore xds files
			String xdsdir = event.location.replace("file://", "");
			xdsdir = xdsdir.substring(0, xdsdir.lastIndexOf("/"));
			startProgram("xdsgui", xdsdir);
		} else if (event.location.startsWith("file://") && (event.location.contains("aimless")
				|| event.location.contains("truncate") || event.location.contains("pointless")
				|| event.location.contains("XDSSTAT.LP") || event.location.contains("XDSCC12.LP"))
				&& (event.location.endsWith(".log"))) {
			// open ccp4 log
			event.doit = false;
			startProgram("/mnt/software/px/CCP4/ccp4/bin/logview " + event.location.replace("file://", ""), null);

			if (item.getText().length() <= 1) { // disable creating new tabs if no title is set
				item.dispose();
				folder.setSelection(0);
			}
		} else if (event.location.startsWith("file://") && (event.location.endsWith(".mtz"))) {
			// deal with mtz download or open mtz view hklview
			event.doit = true;
			boolean viewhkl = false;
			if (viewhkl) {
				startProgram("/mnt/software/px/CCP4/ccp4/bin/viewhkl " + event.location.replace("file://", ""), null);
			}
			if (item.getText().length() <= 1) { // disable creating new tabs if no title is set
				item.dispose();
				folder.setSelection(0);
			}
		} else if (event.location.startsWith("file://") && (event.location.endsWith(".sh")
				|| event.location.contains(".sh.") || (event.location.endsWith(".xml")))) {
			event.doit = false;
			String fn = event.location.replace("file://", "");
			startProgram("nedit "+fn, null);
		} else if (event.location.startsWith("file://")
				&& (event.location.endsWith(".pdf") || event.location.endsWith(".PDF"))) {
			event.doit = false;
			String fn = event.location.replace("file://", "");
			startProgram("atril "+fn, null);
		} else if (event.location.startsWith("file://") && (event.location.endsWith(".pdb"))) {
			String fn = event.location.replace("file://", "");
			event.doit = false;
			startProgram("pymol "+fn, null);
		} else if (event.location.startsWith("file://")) {
			// if it is a directory, sorted the list of content (webkit default order appears
			// random)
			String location = event.location.replace("file://", "/").replaceAll("//", "/");
			if (new File(location).isDirectory()) {
				new SortDirectoryContentInBrowser(browser);
			}

			String locationDisplay = location;
			int level5 = StringUtils.ordinalIndexOf(location, "/", 5);
			if (level5 > 0) {
				locationDisplay = "..." + location.substring(level5);
			}

			item.setText(locationDisplay);

		} else {
			item.setText(event.location);
		}
	}

	private void startProgram(String program, String wdir) {
		ExecTask task = ExecTask.new_ExecTask(program, program, false);
		if (wdir !=null) {
			task.setWorkingDir(wdir);
		}
		task.start();
		Log.note(program + " started in "+ wdir);
	}
	
	@Override
	public void changed(LocationEvent arg0) {
	}
}
