package gov.anl.gmca.analysis.helpers;

import gov.anl.gmca.io.Database;
import gov.anl.gmca.io.Database.DBName;
import gov.anl.gmca.io.Database.Result;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.PV_SCR_RUN;
import gov.anl.gmca.syslog.Log;

import java.util.Map;

// this class find out the strategy results for the latest crystal mounted, or from a pipeline id

public class StrategyResults {
    Double oscStart = null;
    Double oscEnd = null;
    Double oscDelta = null;
    Double detectorDistance = null;
    String spaceGroup = null;
    String cell = null;
    String software = null;
    String fileRoot = null;
    String sampleName = null;

    public StrategyResults(Integer id) {
        String query = "select osc_start, osc_end, osc_delta, detectordistance "
                + "from screenstrategyresults where pipelinestatus_id="
                + id + ";";
        Result r = Database.exec(DBName.USERDATA, query);

        if (!r.isOK() || r.numRows() != 1) {
            Log.error("failed to find screening results for pipelinestatus_id=" + id);
            return;
        }

        getStrategy(r.getRow(0));
    }

    public StrategyResults(String fileRoot, String spacegroup, Integer maxEntries) {
        this.fileRoot = fileRoot;
        if (fileRoot == null || fileRoot.equals("")) {
            Log.warn("no sample specified for StrategyResults.");
            return;
        }

        String symmQuery = "";
        if (spacegroup != null) {
            symmQuery = String.format("AND LOWER(t.spacegroup) LIKE LOWER('%s') ", spacegroup);
        }

        String query = String.format(
                "SELECT sampleName, software, unitcell, osc_start, "
                + "osc_end, osc_delta, detectordistance, spacegroup FROM "
                + "(SELECT * FROM `screenstrategyresults` ORDER BY pipelinestatus_id DESC LIMIT %d) AS t "
                + "WHERE t.`state`='STRATEGY' %s "
                + "AND t.sampleName LIKE '%s%%' "
                + "ORDER BY CAST(t.score as DECIMAL(10,6) ) DESC LIMIT 1;", maxEntries, symmQuery, fileRoot);

        Result r = Database.exec(DBName.USERDATA, query);

        if (!r.isOK() || r.numRows() != 1) {
            Log.warn("Failed to retrieve strategy results for requested " + fileRoot);
            return;
        }

        getStrategy(r.getRow(0));
    }

    public StrategyResults() {
        this(Epics.getString(PV_SCR_RUN.fileroot.run(0)));
    }

    public StrategyResults(String fileRoot) {
        this(fileRoot, null, 4);
    }

    private void getStrategy(Map<String, String> row) {
        try {
            oscStart = Double.parseDouble(row.get("osc_start"));
            oscEnd = Double.parseDouble(row.get("osc_end"));
        } catch (NumberFormatException nfe) {
            Log.error("fail to extract start end end angle of strategy");
            return;
        }

        try {
            oscDelta = Double.parseDouble(row.get("osc_delta"));
            detectorDistance = Double.parseDouble(row.get("detectordistance"));
        } catch (NumberFormatException nfe) {
            Log.warn("no valid oscillation angle or detector distance found, set default as 0.2 deg and 350mm respectively.");
            oscDelta = 0.2;
            detectorDistance = 350.0;
        }

        spaceGroup = row.get("spacegroup");
        cell = MiscUtils.trimCell(row.get("unitcell"));
        software = row.get("software");
        sampleName = row.get("sampleName");
        if (software != null) {
            software = software.replace("_strategy", "");
        }
    }

    public Double getOscStart() {
        return oscStart;
    }

    public Double getOscEnd() {
        return oscEnd;
    }

    public Double getOscDelta() {
        return oscDelta;
    }

    public Double getDetectorDistance() {
        return detectorDistance;
    }

    public String getSpaceGroup() { return spaceGroup; }

    public String getSoftware() {
        return software;
    }

    public String getCell() {
        return cell;
    }

	@Override
	public String toString() {
		return String.format("%s: %s\nsymm: %s\ncell: %s,\ns=%.0f, e=%.0f, osc=%.1f, d=%.0f",
				sampleName,
				software,
				spaceGroup,
				cell,
				oscStart,
				oscEnd,
				oscDelta,
				detectorDistance
		);
	}

    public static void main(String... args) {
        StrategyResults sr = new StrategyResults();
        System.out.println("start=" + sr.getOscStart() + " " + sr.getDetectorDistance() + " " + sr.getOscEnd() + " " + sr.getOscDelta());
    }
}
