package gov.anl.gmca.collect;

import gov.anl.gmca.collect.ModelCollectRunList.CollectState;
import gov.anl.gmca.common.Command;
import gov.anl.gmca.common.DBConfig;
import gov.anl.gmca.common.EpicsDevice;
import gov.anl.gmca.hutch.DevHutchShutter;
import gov.anl.gmca.hutch.DevKeithleyAmp;
import gov.anl.gmca.hutch.DevKeithleyAmp.KeithleyState;
import gov.anl.gmca.hutch.DevPmacMotor;
import gov.anl.gmca.hutch.DevUndulator;
import gov.anl.gmca.hutch.DevUndulator.HarmonicAndOffset;
import gov.anl.gmca.raster.ModelRasterState;
import gov.anl.gmca.raster.ModelRasterState.RasterState;
import gov.anl.gmca.server.messaging.Popup;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.users.ModelUserTable.Client;
import gov.anl.gmca.util.Util;
import gov.aps.jca.CAException;

import java.util.List;

/*
 * Checks that are used before important actions like collection or scans.
 * 
 * This command can be used either on its own thread or just by calling run().
 */
public class CmdCheck extends Command
{
	public final static int FLAG_BEAMSTOP					= 0x0001;
	public final static int FLAG_ENERGYGAP					= 0x0002;
	public final static int FLAG_ENDSTATIONSHUTTER			= 0x0004;
	public final static int FLAG_WAITFORFEEDBACKWITHPOLL	= 0x0008;
	public final static int FLAG_WAITFORFEEDBACKNOPOLL		= 0x0010;
	public final static int FLAG_WAITFORKEITHLEYAMP			= 0x0020;
	
	public final static int FLAGS_PRECOLLECT 		= FLAG_BEAMSTOP|FLAG_ENERGYGAP|FLAG_ENDSTATIONSHUTTER|FLAG_WAITFORFEEDBACKWITHPOLL;
	public final static int FLAGS_COLLECTFRAME 		= FLAG_ENDSTATIONSHUTTER|FLAG_WAITFORFEEDBACKWITHPOLL|FLAG_WAITFORKEITHLEYAMP;
	public final static int FLAGS_ENERGYCHANGE		= FLAG_WAITFORFEEDBACKWITHPOLL;
	public final static int FLAGS_PRESCAN 			= FLAG_BEAMSTOP|FLAG_ENDSTATIONSHUTTER|FLAG_WAITFORFEEDBACKWITHPOLL;
	public final static int FLAGS_PRERASTER 		= FLAG_BEAMSTOP|FLAG_ENERGYGAP|FLAG_ENDSTATIONSHUTTER|FLAG_WAITFORFEEDBACKWITHPOLL;
	
	protected int checkFlags = 0;
	
	protected boolean _checksPassed = false;
	public boolean checksPassed() { return _checksPassed; }
	
	public CmdCheck(int _checkFlags)
	{
		checkFlags = _checkFlags;
	}

	protected boolean endstationShutterCheck()
	{
		while (true)
		{
			if (!DevCollectSettings.getEndstationCheckEnabled())
			{
				Log.note("Endstation shutter check disabled, returning proceed");
				return true;
			}
			
			DevHutchShutter esShutter = DevHutchShutter.get("endStation");
			if (esShutter.isOpen()) return true;
			
			// Try to open it
			try
			{
				Log.note("Opening endstation shutter");
				esShutter.open();
			}
			catch (CAException cae) { cae.printStackTrace(); }
			
			boolean shutterOpen=false;
			for (int i=0; i<20; i++)
			{
				if (esShutter.isOpen())
				{
					shutterOpen=true;
					break;
				}
				Util.sleep(250);
			}
			
			if (!shutterOpen)
			{
				String msg = "The endstation shutter failed to open. "+
						"Click \"retry\" to try again, otherwise cancel.";
				if (DBConfig.serverMode) {
					int result = Popup.syncPopup("question", msg, new String[]{"OK","Cancel"}, 0);
					if (0 != result) return false;
				} else {
					boolean ok = Util.popupOkCancel("Shutter",msg);
					if (!ok) return false;
				}
			}
			else
			{
		        Log.note("Endstation shutter successfully opened.");
		        return true;
			}
		}
    }
	
	public static boolean energyGapCheck(boolean beSilent)
	{
		///////////////////////////////////////////////////////////////////////////////////////////////////
		///////////////////////////////////// verify the check is enabled /////////////////////////////////
		///////////////////////////////////////////////////////////////////////////////////////////////////
		
		if (!DevCollectSettings.getEnergyGapCheckEnabled())
		{
			if (!beSilent) System.out.println("Energy gap check disabled, returning proceed");
			return true;
		}
		
		///////////////////////////////////////////////////////////////////////////////////////////////////
		///////////////////////////////////// verify this is an ID beamline ///////////////////////////////
		///////////////////////////////////////////////////////////////////////////////////////////////////
		
		if (!Util.getBID().equals("23i") && !Util.getBID().equals("23o"))
		{
			if (!beSilent) System.out.println("Energy gap check not applicable at beamline "+Util.getBID());
			return true;
		}

		///////////////////////////////////////////////////////////////////////////////////////////////////
		//////////////////////// verify the mono and undulator are connected and not moving ///////////////
		///////////////////////////////////////////////////////////////////////////////////////////////////
		
		DevPmacMotor monoDev = DevPmacMotor.get("mo_En_E");
		DevUndulator undDev = DevUndulator.getDevice();

		if (null == monoDev || null == undDev)
		{
			if (!beSilent) System.out.println("Mono ("+monoDev+") or undulator ("+undDev+") are null, returning true for energy gap check.");
		    return true;
		}
		
		if (monoDev.getState() != EpicsDevice.State.READY ||
			undDev.getState() != EpicsDevice.State.READY)
		{
			if (!beSilent) System.out.println("Mono state="+monoDev.getState()+", undulator state="+undDev.getState()+", aborting energy gap check.");
			return true;
		}
		
		///////////////////////////////////////////////////////////////////////////////////////////////////
		//////////////////////////// verify we are not collecting or rastering ////////////////////////////
		///////////////////////////////////////////////////////////////////////////////////////////////////
		
		CollectState collectState = null;
		RasterState rasterState = null;
		if (null != ModelCollectRunList.get())
		{
			collectState = ModelCollectRunList.get().getCollectState();
		}
		if (null != ModelRasterState.get())
		{
			rasterState = ModelRasterState.get().getRasterState();
		}
		
		if (   ( null != collectState && collectState != CollectState.INACTIVE )
			|| ( null != rasterState  && rasterState != RasterState.INACTIVE) )
		{
			if (!beSilent) System.out.println("Collect state = "+collectState+", raster state = "+rasterState+", aborting energy gap check.");
			return true;
		}

		///////////////////////////////////////////////////////////////////////////////////////////////////
		////////////////////////////////// verify this client is master ///////////////////////////////////
		///////////////////////////////////////////////////////////////////////////////////////////////////
		
		if (!Client.isMaster())
		{
			if (!beSilent) System.out.println("Skipping energy gap check since this client is not master.");
			return true;
		}

		///////////////////////////////////////////////////////////////////////////////////////////////////
		/////////////////////////////////// everything is OK, do the check ////////////////////////////////
		///////////////////////////////////////////////////////////////////////////////////////////////////
		
		double offsetTolerance = 0.010;

		double differenceFromExpected = Double.MAX_VALUE;
		List<HarmonicAndOffset> calculatedOffset = DevUndulator.calcHarmonicAndOffset(monoDev.getPosition());
		for (HarmonicAndOffset harmonicInfo : calculatedOffset)
		{
			double thisDifference = Math.abs(undDev.getPosition() - harmonicInfo.offset - monoDev.getPosition());
			differenceFromExpected = Math.min(thisDifference, differenceFromExpected);
		}
		
		if (!beSilent) System.out.println("checkEnergyGap: monoVal="+monoDev.getPosition()+" undVal="+undDev.getPosition()+" calculatedOffset="+calculatedOffset+" differenceFromExpected="+differenceFromExpected);
		
    	if ( differenceFromExpected > offsetTolerance )
    	{
    		if (!beSilent)
    		{
    			String offsetString = Util.sprintf("%.03f", calculatedOffset.get(0).offset);
    			if (calculatedOffset.size()>1)
    				offsetString = Util.sprintf("%s or %.03f", offsetString, calculatedOffset.get(1).offset);
    			
    			Util.asyncPopupOk
    			(
					"Undulator/mono sync error",
					"At the current energy, the undulator energy should be "+
					offsetString+
					"keV higher than the monochromator energy.  Choosing a new energy "+
					"in the hutch tab will re-sync both devices."
				);
    		}
			return false;
    	}
    	else
    	{
    		if (!beSilent) System.out.println(	"checkEnergyGap: difference "+Util.sprintf("%.03f",differenceFromExpected)+
    							" less than tolerance "+Util.sprintf("%.03f",offsetTolerance)+
    							", returning proceed.");
    		return true;
    	}
	}
	
	protected boolean beamstopCheck()
	{
		Log.note("Beamstop check not implemented, returning proceed.");
		return true;
	}
	
	protected boolean waitForKeithleyAmpCheck()
	{
		KeithleyState state = DevKeithleyAmp.get().getKeithleyState();
		if (KeithleyState.IDLE == state) return true;

		int secondsLeftToWait = 10;
		do
		{
			Log.note("Waiting for Keithley Amplifier, state = "+state);
			
			for (int i=0;i<10;i++)
			{
				Util.sleep(100);
				if (KeithleyState.IDLE == state) return true;
			}
			
			secondsLeftToWait--;
			if (secondsLeftToWait==0)
			{
				Log.warn("Keithley amplifier reported busy for 5 seconds, continuing anyway...");
				return true;
			}
		}
		while (true);
	}
	
	@Override
	public void run()
	{
		if ( 0 != (checkFlags & FLAG_ENERGYGAP) )
		{
			if (!energyGapCheck(false)) { _checksPassed = false; return; }
		}
		if ( 0 != (checkFlags & FLAG_ENDSTATIONSHUTTER) )
		{
			if (!endstationShutterCheck()) { _checksPassed = false; return; }
		}
		if ( 0 != (checkFlags & FLAG_BEAMSTOP) )
		{
			if (!beamstopCheck()) { _checksPassed = false; return; }
		}
		if ( 0 != (checkFlags & FLAG_WAITFORFEEDBACKWITHPOLL) )
		{
			if (!CmdFeedbackCheck.runCheck(true,true)) { _checksPassed = false; return; }
		}
		if ( 0 != (checkFlags & FLAG_WAITFORFEEDBACKNOPOLL) )
		{
			if (!CmdFeedbackCheck.runCheck(false,false)) { _checksPassed = false; return; }
		}
		if ( 0 != (checkFlags & FLAG_WAITFORKEITHLEYAMP) )
		{
			if (!waitForKeithleyAmpCheck()) { _checksPassed = false; return; }
		}
		
		_checksPassed = true;
	}
}
