package gov.anl.gmca.collect;

import gov.anl.gmca.collect.DevFPE.FrameListResult;
import gov.anl.gmca.collect.ModelCollectRun.Axis;
import gov.anl.gmca.collect.ModelCollectRunList.CollectState;
import gov.anl.gmca.common.DevDataDir;
import gov.anl.gmca.common.Redis;
import gov.anl.gmca.common.client_server.RPC;
import gov.anl.gmca.hutch.DetectorConfig;
import gov.anl.gmca.hutch.DevPmacMotor;
import gov.anl.gmca.hutch.DevShutter;
import gov.anl.gmca.io.Disk;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.ExecTask;
import gov.anl.gmca.io.PVLookup;
import gov.anl.gmca.io.PVLookup.*;
import gov.anl.gmca.sample.DevCollimatorCtrl;
import gov.anl.gmca.sample.DevInOut;
import gov.anl.gmca.sample.DevSlider;
import gov.anl.gmca.sample.ModelBeamsize;
import gov.anl.gmca.spotfinder.DBImageStats;
import gov.anl.gmca.spotfinder.DevSFType;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.ActionResult;
import gov.anl.gmca.util.ListenerList;
import gov.anl.gmca.util.Util;
import redis.clients.jedis.Jedis;

import java.nio.file.FileSystems;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.attribute.GroupPrincipal;
import java.nio.file.attribute.UserPrincipal;
import java.nio.file.attribute.UserPrincipalLookupService;
import java.util.ArrayList;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class CmdCollectJet implements Runnable
{
    public static Thread thread = new Thread();
    public boolean pauseFlag = false;
    public ListenerList.Listener<DevFPE> frameListener;
    public static Pattern cbfPattern = Pattern.compile("(.*)_(\\d+)\\.cbf");

    public void start()
    {
        if (null!=thread && thread.isAlive()) {
            System.out.println("ERROR: CmdCollectJet already running.");
            return;
        }
        thread = new Thread(this,"CmdCollectJet");
        thread.start();
    }

    public static String dir, currentDir, prefix;
    public double expTime_sec;
    public int framesPerDir, startFrame, startSubDir, subDirLimit;
    public String error = null;

    public CmdCollectJet(String dir, String subDir, String prefix, double expTime_sec, int framesPerDir, int startFrame, int subDirLimit)
    {
        this.dir=dir;
        this.startSubDir=Integer.parseInt(Util.stripLeadingZeroes(subDir));
        this.subDirLimit=subDirLimit;
        this.prefix=prefix;
        this.expTime_sec=expTime_sec;
        this.framesPerDir=framesPerDir;
        this.startFrame=startFrame;
    }

    public void pause()
    {
        pauseFlag = true;
        Epics.putInteger(PV_RUNS.collectState, CollectState.PAUSING.ordinal());
    }

    public static String getFilename(int frameIndex)
    {
        return getFilename(prefix, frameIndex);
    }

    public static String getFullDir()
    {
        return currentDir;
    }

    public static String getFilename(String prefix, int frameIndex)
    {
        return Util.sprintf("%s_%06d.cbf", prefix, frameIndex);
    }

    @Override
    public void run()
    {
        try (Jedis jedis = Redis.get()) {
            jedis.hset(Redis.collect_state, "info", "Running checks...");
        }

        // check FPE connection
        if (!DevFPE.get().checkFPEState(null, false)) {
            error = "FPE is not connected";
            return;
        }
        if (!DevFPE.get().isIdle()) {
            error = "FPE is not idle";
            return;
        }

        // check filename lengths
        if (DetectorConfig.isEiger())
        {
            // A1_000001.cbf
            int filenameLength =
                    prefix.length()+
                            1+ // underscore
                            8+ // image index
                            4; // .cbf

            if (filenameLength>CmdCollectSequence.MAX_FILENAME_LENGTH) {
                int exceedsBy = filenameLength - CmdCollectSequence.MAX_FILENAME_LENGTH;
                error = "Prefix too long by "+exceedsBy+" characters";
                return;
            }
        }

        // set states for collecting
        Epics.putInteger(PV_RUNS.collectState, CollectState.COLLECTING.ordinal());
        try (Jedis jedis = Redis.get()) {
            jedis.hset(Redis.collect_run(0),"state","COLLECTING");
        }

        Double ringlightInitialSetting = null;
        try {
            ArrayList<Path> filesToDelete = new ArrayList<>();
            ArrayList<Path> dirsToDelete = new ArrayList<>();

            // check for files in the current subdir
            String _relDir = Util.sprintf("%s/%03d", dir, startSubDir);
            String firstFullDir = DevDataDir.addToFPEDir(_relDir);
            Path firstFullDirPath = FileSystems.getDefault().getPath(firstFullDir);
            if (Files.exists(firstFullDirPath) && Files.isDirectory(firstFullDirPath)) {
                Files.list(firstFullDirPath).forEach(path->{
                    Matcher m = cbfPattern.matcher(path.getFileName().toString());
                    if (m.matches()) {
                        try {
                            String thisPrefix = m.group(1);
                            int imageIndex = Integer.parseInt(m.group(2));
                            if (prefix.equals(thisPrefix) && imageIndex > startFrame) {
                                filesToDelete.add(path);
                            }
                        } catch (NumberFormatException e) {}
                    }
                });
            }

            // check for future subdirs
            Path parentDirPath = FileSystems.getDefault().getPath(DevDataDir.addToFPEDir(dir));
            if (Files.exists(parentDirPath) && Files.isDirectory(parentDirPath)) {
                Files.list(parentDirPath).forEach(path->{
                    boolean isDir = Files.isDirectory(path);
                    String dirName = path.getFileName().toString();
                    if (isDir && dirName.length()==3) {
                        try {
                            int subdirNumber = Integer.parseInt(dirName);
                            if (subdirNumber > startSubDir) {
                                dirsToDelete.add(path);
                            }
                        } catch (NumberFormatException e) {}
                    }
                });
            }

            // for now, exit if any existing files/subdirs found
            if (filesToDelete.size()>0 || dirsToDelete.size()>0) {
                String errorMsg = "";
                if (filesToDelete.size()>0) {
                    errorMsg += filesToDelete.size()+" existing files found.\n";
                }
                if (dirsToDelete.size()>0) {
                    errorMsg += dirsToDelete.size()+" existing subdirectories found.\n";
                }
                errorMsg += "\nDelete isn't currently supported for jet collect.\n";
                errorMsg += "Please change the output settings to avoid existing files.";
                Util.asyncPopupError("Existing files", errorMsg);
                error = "Existing files found";
                return;
            }

            // close shutter
            if (DevShutter.get("shutter").getIsOpen()) {
                Log.note("Closing shutter");
                DevShutter.get("shutter").close();
                DevShutter.get("shutter").waitFor();
            }

            // pre-collect checks
            Log.note("Pre-collect checks");

            CmdCheck preCollectChecks = new CmdCheck(CmdCheck.FLAGS_PRECOLLECT);
            preCollectChecks.run();
            if (!preCollectChecks.checksPassed()) {
                error = "Pre-collect checks failed";
                return;
            }

            // ringlight to max
            DevSlider ringLight = DevSlider.get("ring");
            ringlightInitialSetting = ringLight.getIntensity();
            ringLight.setIntensity(ringLight.getHigherLimit());

            // sample env
            DevCollimatorCtrl.CCMove ccMove = null;
            if (DevCollectSettings.getEnableSampleEnvMove())
            {
                ccMove = DevCollimatorCtrl.get().newMove();
                ccMove.setDest("collimator",	DevInOut.InOut.IN);
                ccMove.setDest("backlight",		DevInOut.InOut.OUT);
                ccMove.setDest("beamstop",		DevInOut.InOut.IN);
                ccMove.setDest("pindiode",		DevInOut.InOut.OUT);
                ccMove.start();
                ccMove.waitFor();
                if (!ccMove.getResult().isOK()) {
                    error = "Sample env. setup failed";
                    return;
                }
            }

            for (int subDir = startSubDir; subDir <= subDirLimit; subDir++) {
                if (pauseFlag) break;
                int thisDirStartFrame = subDir == startSubDir ? startFrame : 0;
                if (thisDirStartFrame == framesPerDir) continue;

                try (Jedis jedis = Redis.get()) {
                    jedis.hset(Redis.collect_run(0), "subDir", "" + subDir);
                    jedis.hset(Redis.collect_state, "info", "Collecting subdir "+subDir);
                }

                // create the directory
                String relDir = Util.sprintf("%s/%03d", dir, subDir);
                currentDir = DevDataDir.addToFPEDir(relDir);
                Path currentPath = Paths.get(currentDir);

                try {
                    for (int i = 0; i < currentPath.getNameCount(); i++) {
                        String subPath = "/"+currentPath.subpath(0, 1 + i).toString();
                        if (!Disk.isDirectory(subPath)) {
                            if (!Disk.mkdir(subPath)) {
                                error = "Couldn't create directory: " + currentDir;
                                return;
                            }
                            if (Util.getUserLogin().equals("root")) {
                            }
                        }
                    }
                } catch (Exception e) { e.printStackTrace(); }

                // clear existing SF stats
                for (DevSFType.SPOTFINDER_OUTPUT output : DevSFType.SPOTFINDER_OUTPUT.values()) {
                    DBImageStats.delete(relDir, prefix, 16, output.toString());
                }

                // start spot finder
                RPC.RPCCommand ih = RPC.sendCommand("cmd:image_handler","dir:"+currentDir,"prefix:"+prefix);

                ////////////////////////////////////////////////////////////////////////////////
                ////////////////////////////////// start detector //////////////////////////////
                ////////////////////////////////////////////////////////////////////////////////

                try {
                    frameListener = null;

                    try (Jedis jedis = Redis.get()) {

                        // set counter for this collect to zero
                        jedis.hset(Redis.image_info(relDir,prefix), "complete_counter", "0");

                        // set expected image count for this collect
                        jedis.hset(Redis.image_info(relDir,prefix), "total_images", ""+framesPerDir);

                        // set FPE to copy current frame to this run
                        jedis.hset(Redis.server_fpe, "copy_frame_to_key", Redis.image_info(relDir,prefix)+"$complete_counter");
                    }

                    // send FPE start command
                    DevFPE.FrameList fpeFrame = DevFPE.get().newFrameList();
                    fpeFrame.delta = 0.0;
                    fpeFrame.disableEigerPerImageTrigger = true;
                    fpeFrame.prefix = prefix + "_";
                    fpeFrame.energy = DevPmacMotor.get("mo_En_E").getPosition();
                    fpeFrame.exposureTime = expTime_sec;
                    fpeFrame.startAngle = DevPmacMotor.get("gonio_omega").getPosition();
                    fpeFrame.nextAngle = fpeFrame.startAngle;
                    fpeFrame.endAngle = fpeFrame.startAngle;
                    fpeFrame.firstFrameIndex = thisDirStartFrame;
                    fpeFrame.runNumber = 1;
                    fpeFrame.filename = getFilename(prefix, thisDirStartFrame+1);
                    fpeFrame.firstFrameLabel = thisDirStartFrame + 1;
                    fpeFrame.frameCount = framesPerDir - thisDirStartFrame;
                    fpeFrame.directory = currentDir;
                    fpeFrame.axis = Axis.GONIO_OMEGA;
                    fpeFrame.fpeMode = DevFPE.FPEMode.SHUTTERLESS_ARM_DISARM;
                    fpeFrame.shutteredMotion = DevFPE.XYZMotion.current();
                    fpeFrame.shutterlessMotion = DevFPE.XYZMotion.current();
                    fpeFrame.start();

                    do {
                        // check for pause
                        if (pauseFlag) { DevFPE.get().pause(true); }

                        try (Jedis jedis = Redis.get()) {
                            int counter = Epics.getInteger(PV_FPE.frameNumber);
                            jedis.hset(Redis.collect_state, "info", "Collecting subdir "+subDir+" frame "+counter);
                        }

                        // sleep
                        Util.sleep(100);
                    }
                    while (fpeFrame.isRunning());

                    FrameListResult result = fpeFrame.getResult();
                    System.out.println("Collect result = " + result);

                    // check for frame error
                    if (!result.isOK()) {
                        error = "FPE reported error";
                        return;
                    }

                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
            if (null != ringlightInitialSetting) {
                ringLight.setIntensity(ringlightInitialSetting);
            }
        } catch(Exception e){
            e.printStackTrace();
        } finally{
            // back light in
            DevCollimatorCtrl.CCMove ccMove = DevCollimatorCtrl.get().newMove();
            ccMove.setArguments("OUT backlightIN bluice");
            ccMove.start();

            Epics.putInteger(PV_RUNS.collectState, CollectState.INACTIVE.ordinal());
            try (Jedis jedis = Redis.get()) {
                jedis.hset(Redis.collect_run(0), "state", "INACTIVE");
            }
        }
    }
}
