package gov.anl.gmca.collect;

import gov.anl.gmca.common.DBConfig;
import gov.anl.gmca.common.client_server.CounterState;
import gov.anl.gmca.gui.state.SWServer;
import gov.anl.gmca.hutch.DetectorConfig;
import gov.anl.gmca.hutch.DevPmacMotor;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.PV_FPE;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.users.ModelUserTable.Client;
import gov.anl.gmca.util.Util;
import gov.aps.jca.dbr.DBR;

import java.util.LinkedList;
import java.util.concurrent.atomic.AtomicInteger;

public class DevFPE2 extends DevFPE
{
	@Override public boolean isIdle()
	{
		int opStat = getOperationStatus();
		return getFPEState()==FPEState.idle && (opStat<=0 || opStat==2);
	}
	@Override public boolean isCacheIdle()
	{
		int opStat = getCacheOpStat();
		return getCacheFPEState()==FPEState.idle && (opStat<=0 || opStat==2);
	}
	
	////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////// normal operation //////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////
	
	private LinkedList<FPEState> stateCache = new LinkedList<FPEState>();
	
	private boolean stateMatches(FPEState state, FPEState[] statesToCompare)
	{
		for (FPEState s : statesToCompare)
		{
			if (s == state) return true;
		}
		return false;
	}
	
	// checks both the list of states that have passed as well as the current state
	// 
	// this will be fairly reliable but may still miss states:
	// the state cache is updated from the EPICS thread, so it may be many seconds slow
	// if an EPICS callback holds up the thread.  getFPEState() is updated instantly as
	// results come in.  so if a state passes too quickly AND the epics thread is hung up,
	// then this method incorrectly returns false.
	//
	// to be fully reliable, we need to add a method to EPICS such as
	// Epics.getHistory() which returns a list of the last 100 values
	// and then we can call Epics.getHistory().contains(stateToCheck)
	//
	private boolean stateReached(FPEState[] statesToCheck)
	{
		synchronized(stateCache)
		{
			while (!stateCache.isEmpty())
			{
				FPEState firstState = stateCache.removeFirst();
				//Log.warn("Checking state "+firstState+" vs. "+stateToCheck+" (cache size="+stateCache.size()+")");
				if (stateMatches(firstState, statesToCheck))
				{
					//stateCache.clear();
					return true;
				}
			}
			return stateMatches(getFPEState(),statesToCheck);
		}
	}

	private long fpeStartMs = 0;
	
	@Override public void collectPartition(FrameList frameList)
	{
		abortSent = false;
		pauseSent = false;
		retryNumber = 0;
		synchronized(stateCache) { stateCache.clear(); }
		
		try
		{
			// check that FPE is running and MAR connection is OK
			if (!checkFPEState(frameList, true)) return;

			// check that the stream writer is running if applicable
			if (DBConfig.getBoolean("streamwriter_check_status") && !DevEiger.checkStreamWriter()) return;
			
			// FPE must be idle before starting
			long idleWaitStartMs = System.currentTimeMillis();
			//Log.warn("Checking FPE state before starting: opStat="+getOperationStatus()+" state="+getFPEState()+" opStatCache="+getCacheOpStat());
			while (!isIdle())
			{
				long nowMs = System.currentTimeMillis();
				if (nowMs-idleWaitStartMs>5000)
				{
					Log.error("Could not start partition, FPE not idle after 5 seconds");
					frameList.result=FrameListResult.FAILED;
					frameList.disconnectListeners();
					return;
				}
				Util.sleep(10);
				//Log.warn("Checking FPE state before starting: opStat="+getOperationStatus()+" state="+getFPEState()+" opStatCache="+getCacheOpStat());
			}
			
			// check if mode256 is clear before sending it - this is to determine if keithley was restarted
			int firstTimeoutMs = 20000;//isFirstFrameAfterKeithleyRestart() ? 15000 : 5000;
			
			// send opStat=2 and wait for it to be read back
			// (this prevents reading the previous value of opStat in the initial wait and
			//  immediately exiting the collection of this partition)
			Epics.putInteger(PV_FPE.operationStatus, 2);
			long startMs = System.currentTimeMillis();
			fpeStartMs = startMs;
			int twoCount=0;
			while(true)
			{
				int opStat = Epics.getInteger(PV_FPE.operationStatus);
				if (opStat==2) twoCount++;
				if (twoCount==3) break;
				long nowMs = System.currentTimeMillis();
				long waitMs = nowMs-startMs;
				if (waitMs > 5000)
				{
					Log.error("Sent opStat=2 to FPE but didn't read back the value after 5 seconds");
					frameList.result=FrameListResult.FAILED;
					frameList.disconnectListeners();
					return;
				}
				Util.sleep(10);
				System.out.println("Waiting for opStat==2 (opStat="+opStat+")");
			}

			// send command to FPE
			boolean commandSent = frameList.sendToFPE();
			if (!commandSent)
			{
				Log.error("Aborting collect due to FPE command send failure");
				frameList.result=FrameListResult.FAILED;
				frameList.disconnectListeners();
				return;
			}
			Log.note("FPE: command sent");
			
			long waitStartMs=0, waitMs=0;

			// How many times does FPE go through the st_pos, arm, updates and stop_motion states?
			// TODO: check if this affects Pilatus, and probably add shuttered vector mode
			int nLoops = 1;
			if (DetectorConfig.isEiger() && frameList.fpeMode==FPEMode.SHUTTERED_RASTER)
			{
				nLoops = frameList.frameCount;
			}
			
			// start the loop
			for (int i=0;i<nLoops;i++)
			{
				// FPE receives command
				waitStartMs = System.currentTimeMillis();
				if (!waitForState(frameList, new StateCheck(){public void check(CheckResult result, long stateElapsedMs)
				{
					result.abortReason = "Waiting for FPE to acknowledge command";
					result.okToWait = false;
					result.complete = stateReached(new FPEState[]{FPEState.st_pos});
					result.timeoutMs = firstTimeoutMs;
				}})) return;
		
				waitMs = System.currentTimeMillis()-waitStartMs;
				Log.note("FPE: command received in "+waitMs+"ms");
	
				if (DetectorConfig.isEiger())
				{
				    // detector arm
				    waitStartMs = System.currentTimeMillis();
				    if (!waitForState(frameList, new StateCheck(){public void check(CheckResult result, long stateElapsedMs)
				    {
					result.abortReason = "Waiting for detector arm";
					result.okToWait = false;
					result.complete = stateReached(new FPEState[]{FPEState.arm_eiger, FPEState.gonio_still});
					result.timeoutMs = 35000*(retryNumber+1)+5000;
				    }})) return;
				    
					waitMs = System.currentTimeMillis()-waitStartMs;
					Log.note("FPE: Detector armed in "+waitMs+" ms");
				}
				
				// FPE moves gonio to start position
				waitStartMs = System.currentTimeMillis();
				if (!waitForState(frameList, new StateCheck(){public void check(CheckResult result, long stateElapsedMs)
				{
					result.abortReason = "Waiting for gonio to move to start";
					result.okToWait = DevPmacMotor.get("gonio_omega").getMoving();
					result.complete = stateReached(new FPEState[]{FPEState.blu_ice_updates});
					result.timeoutMs = 60000; // increased from 5 seconds because FPE sometimes takes longer - is 10 seconds enough?
				}})) return;
		
				waitMs = System.currentTimeMillis()-waitStartMs;
				Log.note("FPE: gonio at start position in "+waitMs+"ms");
				
				// FPE takes requested images
				final AtomicInteger lastFrame = new AtomicInteger(-1);
				final int timeoutMs = (int)(frameList.exposureTime*1000.0)+5000;
				System.out.println("For taking images, timeout = "+timeoutMs+"ms");
				waitStartMs = System.currentTimeMillis();
				if (!waitForState(frameList, new StateCheck(){public void check(CheckResult result, long stateElapsedMs)
				{
					boolean gonioMoving = DevPmacMotor.get("gonio_omega").getMoving() || DevPmacMotor.get("sample_z").getMoving();
					
					int currentFrame = DevFPE.get().getCurrentFrameIndex();
					boolean frameChanged = (lastFrame.get()!=currentFrame);
					lastFrame.set(currentFrame);
					
					result.abortReason = "Waiting for images to be taken";
					result.okToWait = gonioMoving || frameChanged;
					result.complete = stateReached(new FPEState[]{FPEState.stop_motion});
					result.timeoutMs = timeoutMs;
				}})) return;
		
				waitMs = System.currentTimeMillis()-waitStartMs;
				Log.note("FPE: image taking complete in "+waitMs+"ms");
			}
			
			// FPE moves gonio to next start position and on MAR, checks for previous frame on disk
			final AtomicInteger opStatResult = new AtomicInteger();
			waitStartMs = System.currentTimeMillis();
			if (!waitForState(frameList, new StateCheck(){public void check(CheckResult result, long stateElapsedMs)
			{
				boolean fileCheckUpdated = getFileCheckValue() != lastFileCheckValue;
				lastFileCheckValue = getFileCheckValue();
				boolean gonioMoving = DevPmacMotor.get("gonio_omega").getMoving();
				opStatResult.set(getOperationStatus());
				
				result.abortReason = "Waiting for stop motion step";
				result.okToWait = true;
				result.complete = opStatResult.get()<=0;
				result.timeoutMs = 60000;
			}})) return;

			waitMs = System.currentTimeMillis()-waitStartMs;
			Log.note("FPE: ready, stop motion took "+waitMs+"ms");
 			
			// Wait for all CBFs to be written
			if (usingServerCounter())
			{
				waitStartMs = System.currentTimeMillis();
				final AtomicInteger counterValue = new AtomicInteger();
				final int expectedFinalCounterValue = currentLastFrameIndex+1;
				if (!waitForState(frameList, new StateCheck(){public void check(CheckResult result, long stateElapsedMs)
				{
					counterValue.set(CounterState.current.counter.completeCounter);

					result.abortReason = "Waiting for counter to reach final frame";
					result.complete = counterValue.get()>=expectedFinalCounterValue;
					result.timeoutMs = 5000;
					result.okToWait = SWServer.elapsedSinceCounterChangeMs() < result.timeoutMs;
				}})) return;
				
				waitMs = System.currentTimeMillis()-waitStartMs;
				if (counterValue.get() == expectedFinalCounterValue) {
					Log.note("Counter value "+counterValue.get()+" equals expected final value "+expectedFinalCounterValue);
				} else {
					Log.warn("Counter value "+counterValue.get()+" does not equal expected final value "+expectedFinalCounterValue);
				}
			}
			
			// return the result
			switch (opStatResult.get())
			{
			case 0: frameList.result = FrameListResult.SUCCEEDED; break;
			case -1: frameList.result = FrameListResult.FAILED; break;
			case -2: frameList.result = FrameListResult.SUCCEEDED_WITH_WARNING; break;
			}
		}
		catch (Exception e) { e.printStackTrace(); frameList.result = FrameListResult.FAILED; }
		
		frameList.disconnectListeners();
	}
	
	private int lastFileCheckValue = -1;
	private int getFileCheckValue() { return 0; } // TODO: when implemented: return Epics.getInteger(PV_FPE.fileCheck); }
	
	private int retryNumber=0;
	
	protected abstract static class StateCheck
	{
		public abstract void check(CheckResult result, long stateElapsedMs);
	}
	protected class CheckResult
	{
		public String abortReason;
		public boolean okToWait, complete;
		public int timeoutMs;
	}
	
	private boolean waitForState(FrameList frameList, StateCheck sc)
	{
		CheckResult result = new CheckResult();
		long waitStartMs = System.currentTimeMillis();
		while (true)
		{
			long nowMs = System.currentTimeMillis();
			sc.check(result, nowMs-waitStartMs);

			boolean timedOut = (nowMs-waitStartMs>result.timeoutMs);
			if (result.okToWait) { timedOut=false; waitStartMs=nowMs; }
			
			if (result.complete) return true;
			
			// abort sent, exit loop
			if (abortSent)
			{
				System.out.println("Abort flag set, returning false from waitForState");
				frameList.result = FrameListResult.FAILED;
				frameList.disconnectListeners();
				return false;
			}
			
			// protocol timeout, send abort and exit loop
			if (timedOut)
			{
				Log.note("FPE timeout, sending abort.  timeout="+result.timeoutMs+" okToWait="+result.okToWait+" complete="+result.complete+" wait="+(nowMs-waitStartMs)+" reason="+result.abortReason);
				Util.asyncPopupError("FPE timeout", "FPE timeout, sending abort.  Reason="+result.abortReason+".  Please check the FPE window for further details.");
				abort();
				frameList.result = FrameListResult.FAILED;
				frameList.disconnectListeners();
				return false;
			}
			
			// handle FPE end of operation values
			int opStat = Epics.getInteger(PV_FPE.operationStatus);
			int opStatCache = Epics.getCacheInteger(PV_FPE.operationStatus);

			int completeFlag = 0;
			if (usingServerCounter()) {
			    int completeCounter = CounterState.current.counter.completeCounter;
			    boolean streamInProgress = CounterState.current.counter.streamInProgress;
			    int expectedFinalCounterValue = currentLastFrameIndex+1;
			    boolean serverCounterDone = completeCounter==expectedFinalCounterValue || (!streamInProgress && pauseSent);
			    completeFlag = serverCounterDone?0:1;
			}

			// temporary fix: EPICS is saving opStat=0 in the callback, which can happen ~.5 seconds after opStat is
			// set to 2.  This needs to be ignored until that bug can be fixed.
			long msSinceFPEStart = System.currentTimeMillis() - fpeStartMs;
			boolean ignorePilatusZero = DetectorConfig.isPilatus() && msSinceFPEStart < 2000;
			if (opStat<=0 && completeFlag<=0 && !ignorePilatusZero)
			{
			        //System.out.println("opStat="+opStat+" completeCounter="+completeCounter+" expectedFinalCounterValue="+expectedFinalCounterValue+", returning false from waitForState");
				switch (opStat)
				{
				case 0: frameList.result = FrameListResult.SUCCEEDED; break;
				case -1: frameList.result = FrameListResult.FAILED; break;
				case -2: frameList.result = FrameListResult.SUCCEEDED_WITH_WARNING; break;
				}
				
				if (opStat==-1)
				{
					Util.asyncPopupError("FPE reported error", "FPE returned an error code (operationStatus=-1).  Please check the FPE window for details.");
				}
				
				frameList.disconnectListeners();
				Log.warn("Unexpected end of collection, returning result: "+frameList.result.toString());
				return false; // even in "success", returning false means the thread should exit immediately
			}
			
			Util.sleep(10);
		}
	}
	
	private void handleLogMessages(String message)
	{
		if (message.contains("log_error TIMEOUT: detector arm") && Client.isMaster())
		{
			int retryIndex = message.indexOf("retry");
			int retryNumber = Integer.parseInt(message.substring(retryIndex+"retry".length()+1));
			this.retryNumber=retryNumber;
			
			if (retryNumber==1)
			{
				Util.asyncPopupError("Eiger detector error", "The Eiger detector failed to arm, FPE is retrying.  Details are in the FPE log window.");
			}
		}
		else if (message.contains("log_error initialize the EIGER") && Client.isMaster())
		{
			Util.asyncPopupError("Eiger detector error", "Out of retries for arming the Eiger detector.  Please try again or contact your host.");
		}
	}
	
	////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////// interruptions ///////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////
	
	private volatile boolean abortSent = false, pauseSent = false;
	@Override public void abort()
	{
		abortSent=true;
		Epics.putInteger(PV_FPE.abort, 1);
	}

	@Override public boolean pause(boolean enforcePauseSpacing)
	{
		Log.note("Pause requested, enforceSpacing="+enforcePauseSpacing);
		
		if (getFPEState()==FPEState.idle)
		{
			System.out.println("Pause ignored, FPE is idle"); return false;
		}
		
		long sinceLastPause = Util.msSince(lastModeCmdSent);
		if (sinceLastPause<PAUSE_SPACING_MS)
		{
			System.out.println("Pause not sent, last command was sent "+sinceLastPause+" ms ago");
			return false;
		}
		
		pauseSent=true;
		lastModeCmdSent = System.currentTimeMillis();
		Epics.put(PV_FPE.mode256, "0 0");
		return true;
	}

	////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////// interface implementation ////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////////////
	
	public DevFPE2()
	{
		super("FPE");
		Epics.addMonitor(fpeStatePV, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				//System.out.println("camonitor: "+pvName+" "+Epics.dbrToString(value));
				synchronized(stateCache)
				{
					FPEState state = FPEState.lookup(Epics.dbrToString(value));
					//Log.warn("Adding state "+state);
					stateCache.addLast(state);
				}
			}
		});
		Epics.addMonitor(PV_FPE.cmd_output256, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				handleLogMessages(Epics.dbrToString(value));
			}
		});
	}
}
