package gov.anl.gmca.collect;

import gov.anl.gmca.collect.ModelCollectRun.ScanMode;
import gov.anl.gmca.collect.ModelCollectRun.SetupStep;
import gov.anl.gmca.hutch.DevPmacMotor;

// Methods here share a lot of code and in general all affect each other.
// Because changes here normally require reviewing the entire class, it is
// kept as small and simple as possible.
//
public class FieldInteractionDependent
{
	public void changeStartFrame(int newStartFrame)
	{
		int newEndFrame = run.getEndFrameLabel();
		
		// correct end frame if < start frame or for run 0
		if (0 == run.dev().ri) newEndFrame = newStartFrame;
		if (newEndFrame < newStartFrame) newEndFrame = newStartFrame;
		
		// fixes
		fixEndAngle(newStartFrame, newEndFrame, run.startAngle);
		fixFPSNOSWedge(newStartFrame, newEndFrame, run.getDelta(), false);
		
		// set PVs
		run.dev().setStartFrameLabel(newStartFrame, false);
		done(true);
	}
	
	public void changeEndFrame(int newEndFrame)
	{
		// correct end frame to be >= start frame
		if (newEndFrame < run.startFrameLabel) newEndFrame = run.startFrameLabel;
		
		// fixes
		fixEndAngle(run.startFrameLabel, newEndFrame, run.startAngle);
		fixFPSNOSWedge(run.getStartFrameLabel(), newEndFrame, run.getDelta(), false);

		done(true);
	}
	
	public void changeStartAngle(double newStartAngle)
	{
		int newEndFrame = run.getEndFrameLabel();
		double newStartAngle2 = Double.NaN;
		
		// correct end frame for run 0
		if (0 == run.dev().ri) newEndFrame = run.getStartFrameLabel();
		
		// calculate strategy frame angle
		if (0 == run.dev().ri && run.getStrategyEnabled())
		{
			double upperGonioLimit = DevPmacMotor.get("gonio_omega").getLimits().getUpper();
			boolean roomForSecondFrame = newStartAngle + 90.0 <= upperGonioLimit;
			if (roomForSecondFrame)	newStartAngle2 = newStartAngle + 90.0;
			else					newStartAngle2 = newStartAngle - 90.0;
		}

		// fixes
		fixEndAngle(run.getStartFrameLabel(), newEndFrame, newStartAngle);
		fixFPSNOSWedge(run.getStartFrameLabel(), newEndFrame, run.getDelta(), false);
		
		// set PVs
		run.dev().setStartAngle(newStartAngle, false);
		if (!Double.isNaN(newStartAngle2)) run.dev().setStartAngle2(newStartAngle2, false);
		done(true);
	}
	
	public void changeEndAngle(double newEndAngle)
	{
		// get acceptable end angle
		if (newEndAngle < run.getStartAngle())
		{
			newEndAngle = run.getStartAngle();
		}

		// calculate new end frame
		int newEndFrame = (int)((newEndAngle - run.getStartAngle()) / run.getDelta() - 0.01 + run.getStartFrameLabel());

		// fixes
		fixEndAngle(run.getStartFrameLabel(), newEndFrame, run.getStartAngle());
		fixFPSNOSWedge(run.getStartFrameLabel(), newEndFrame, run.getDelta(), false);
		
		// set PVs
		run.dev().setEndAngle(newEndAngle, false);
		done(true);
	}
	
	public void changeDelta(double newDelta)
	{
		// correct end frame and angle
		double newEndAngle = run.getEndAngle();
		int newEndFrame = (int)((newEndAngle - run.getStartAngle()) / newDelta - 0.01 + run.getStartFrameLabel());
		if (0 == run.dev().ri) newEndFrame = run.getStartFrameLabel();
		
		// fixes
		fixEndAngle(run.getStartFrameLabel(), newEndFrame, run.getStartAngle());
		fixFPSNOSWedge(run.getStartFrameLabel(), newEndFrame, newDelta, false);

		// write PVs
		run.dev().setEndAngle(newEndAngle, false);
		run.dev().setDelta(newDelta, false);
		done(true);
	}
	
	// called from vector mode only, helical FPS always = 1
	public void changeFramesPerSite(int newFPS, boolean changeWedge, boolean update)
	{
		double wedge_deg = run.getWedgeSize_deg();

		// overlap must be <= FPS
		int newOverlap=run.getOverlap();
		if (newOverlap>newFPS)
		{
			newOverlap = newFPS;
			run.dev().setOverlap(newOverlap, false);
		}
		
		// don't use "fix" methods here because changing FPS is different from other changes
		double newEndAngle = run.endAngle;
		if (run.setupStep == SetupStep.COLLECT_PARAM)
		{
			// totalFrames = NOS * FPS
			int newFrameTotal = run.getSelectedSiteCount() * (newFPS-newOverlap);
			int newEndFrame = run.getStartFrameLabel()+newFrameTotal-1;
			
			// calculate new end angle from end frame
			int startFrame = run.getStartFrameLabel();
			newEndAngle = run.getAllowedEndAngle(startFrame,newEndFrame,run.getDelta(),run.getStartAngle());
			run.dev().setEndAngle(newEndAngle, false);
		}

		// set PVs
		int newWedgeSites = (int)Math.round(wedge_deg / (newFPS*run.getDelta())); // maintain wedge degrees
		if (newWedgeSites>run.getSelectedSiteCount()) newWedgeSites=run.getSelectedSiteCount();
		if (newWedgeSites<1) newWedgeSites=1;
		if (changeWedge) run.dev().setWedgeSize(newWedgeSites*newFPS*run.getDelta(), false);
		run.dev().setFramesPerSite(newFPS,false);
		done(update);
	}

	private void done(boolean update)
	{
		run.dev().setNextFrame(0);
		if (update) run.dev().sendUpdate();
	}
	
	private double fixEndAngle(int startFrame, int endFrame, double startAngle)
	{
		// calculate new end angle from end frame and correct if necessary
		double newEndAngle = run.getAllowedEndAngle(startFrame,endFrame,run.delta,startAngle);
		run.dev().setEndAngle(newEndAngle, false);
		return newEndAngle;
	}
	
	private void fixFPSNOSWedge(int newStartFrame, int newEndFrame, double newDelta, boolean update)
	{
		// update either fps (discrete mode) or step size (helical mode)
		int newFPS = run.framesPerSite;
		double newStepSize = run.stepSize_mm;
		if (run.getHelicalMode())
		{
			// change site spacing to set site count to new simple frame count
			int frameCount = (newEndFrame - newStartFrame + 1);
			Double vectorLength = run.getVectorLength_mm();
			if (null != vectorLength)
			{
				newStepSize = vectorLength / (frameCount<=1?1:frameCount-1);
				run.dev().setPotentialVectorSites(frameCount, false);
				run.selectAllVectorSites(false);
			}
			newFPS = 1;
		}
		else
		{
			// correct FPS to spread new frame count over number of sites
			if (run.setupStep == SetupStep.COLLECT_PARAM)
			{
				int frameCount = (newEndFrame - run.getStartFrameLabel() + 1);
				newFPS = (int)Math.ceil((double)frameCount / (double)run.getSelectedSiteCount());
			}
		}
		
		// try to keep the same wedge size in degrees but make sure it is an integer number of sites
		double newWedgeSize = run.wedgeSize_deg;
		if (ScanMode.SINGLE != run.getScanMode() &&
			ScanMode.JET != run.getScanMode() &&
			ScanMode.MESH != run.getScanMode())
		{
			int newWedgeSites = (int)Math.round(run.wedgeSize_deg / (newFPS*run.getDelta()));
			if (newWedgeSites>run.getSelectedSiteCount()) newWedgeSites=run.getSelectedSiteCount();
			if (newWedgeSites<1) newWedgeSites=1;
			newWedgeSize = newWedgeSites*newFPS*run.getDelta();
		}
		
		run.dev().setStepSize_mm(newStepSize, false);
		run.dev().setWedgeSize(newWedgeSize, false);
		run.dev().setFramesPerSite(newFPS, false);
	}

	// constructor and run reference
	private ModelCollectRun run; public FieldInteractionDependent(ModelCollectRun run) { this.run=run; }
}
