package gov.anl.gmca.collect;

import gov.anl.gmca.collect.ModelCollectRun.*;
import gov.anl.gmca.collect.ui.WdgCommitButton.ButtonAction;
import gov.anl.gmca.common.DBConfig;
import gov.anl.gmca.hutch.DevPmacMotor;
import gov.anl.gmca.raster.ModelRasterGrid;
import gov.anl.gmca.sample.ModelBeamsize;
import gov.anl.gmca.spotfinder.SFData.SFType;
import gov.anl.gmca.spotfinder.SpotFinderQueue;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.Point2D;
import gov.anl.gmca.util.Util;

import java.util.ArrayList;

// These change methods are fairly straightforward and share little code.
// They can be studied independently.
//
// (ctrl+shift+numpad / to collapse all methods)
//
public class FieldInteractionIndependent
{
	public void changeBeamWidth_um(double beamWidth, boolean update)
	{
		Point2D beamSize = ModelRasterGrid.calcBeamSize(beamWidth, run.getBeamHeight_um(), ModelBeamsize.getActualBeamsize_um().y);
		run.dev().setBeamWidth_um(beamSize.x, false);
		run.dev().setBeamHeight_um(beamSize.y, false);
		if (update) run.dev().sendUpdate();
	}
	public void changeBeamHeight_um(double beamHeight, boolean update)
	{
		Point2D beamSize = ModelRasterGrid.calcBeamSize(beamHeight, run.getBeamWidth_um(), ModelBeamsize.getActualBeamsize_um().x);
		run.dev().setBeamHeight_um(beamSize.x, false);
		run.dev().setBeamWidth_um(beamSize.y, false);
		if (update) run.dev().sendUpdate();
	}
	public void changeWedge(double newWedgeSize)
	{
		newWedgeSize = (int)(newWedgeSize/run.getDelta()) * run.getDelta();
		
		run.dev().setWedgeSize(newWedgeSize, false);
		run.dev().setNextFrame(0);
		run.dev().sendUpdate();
	}
	public void changeWedgeSites(int newWedgeInSites)
	{
		double wedge_deg = newWedgeInSites * (run.framesPerSite * run.delta);
		
		run.dev().setWedgeSize(wedge_deg, false);
		run.dev().setNextFrame(0);
		run.dev().sendUpdate();
	}
	public void changeEnergies(ArrayList<Double> newEnergies, boolean update)
	{
		int numEnergies = newEnergies.size();
		if (numEnergies > 1 && XDSProcMode.NATIVE == run.getXDSProcMode())
		{
			Log.note("Changing XDS processing mode to anomalous due to multiple energy setting");
			run.dev().setXDSProcMode(XDSProcMode.ANOMALOUS, false);
		}
		
		run.dev().setEnergies(newEnergies, update);
	}
	public void changeInverseBeam(Boolean inverseOn, boolean update)
	{
		if (inverseOn && XDSProcMode.NATIVE == run.getXDSProcMode())
		{
			Log.note("Changing XDS processing mode to anomalous due to inverse beam setting");
			run.dev().setXDSProcMode(XDSProcMode.ANOMALOUS, false);
		}
		
		run.dev().setInverseOn(inverseOn, update);
	}
	public void changeStrategyEnabled(boolean enable)
	{
		double upperGonioLimit = DevPmacMotor.get("gonio_omega").getLimits().getUpper();
		double _startAngle = run.getStartAngle();
		if (_startAngle + 90.0 <= upperGonioLimit)
		{
			run.dev().setStartAngle2(_startAngle + 90.0, false);
		}
		else
		{
			run.dev().setStartAngle2(_startAngle - 90.0, false);
		}

		run.dev().setNextFrame(0);
		run.dev().setStrategyEnabled(enable, true);
	}
	
	public void changeStartPoint(Point3D newStartPoint, boolean update)
	{
		boolean vectorCompleted = (null != run.pt2 && null == run.pt1 && null != newStartPoint);
		run.dev().setPt1(newStartPoint, false);
		if (vectorCompleted)
		{
			run.fitSitesToVector(false, false, false);
			run.selectAllVectorSites(false);
		}
		else if (null == newStartPoint)
		{
			run.dev().clearSelectedSites(false);
			run.dev().setPotentialVectorSites(0, false);
		}
		if (update) run.dev().sendUpdate();
	}
	public void changeEndPoint(Point3D newEndPoint, boolean update)
	{
		boolean vectorCompleted = (null != run.pt1 && null == run.pt2 && null != newEndPoint);
		run.dev().setPt2(newEndPoint, false);
		if (vectorCompleted)
		{
			run.fitSitesToVector(false, false, false);
			run.selectAllVectorSites(false);
		}
		else if (null == newEndPoint)
		{
			run.dev().clearSelectedSites(false);
			run.dev().setPotentialVectorSites(0, false);
		}
		if (update) run.dev().sendUpdate();
	}
	public void changeFiducial1(Point3D newPt1, boolean update)
	{
		run.dev().setPt1(newPt1, update);
	}
	public void changeFiducial2(Point3D newPt2, boolean update)
	{
		run.dev().setPt2(newPt2, update);
	}
	public void changeFiducial3(Point3D newPt3, boolean update)
	{
		run.dev().setPt3(newPt3, update);
	}
	public void changeFiducial4(Point3D newPt4, boolean update)
	{
		run.dev().setPt4(newPt4, update);
	}
	public void changeStepSize_mm(double newStepSize_mm, boolean update)
	{
		run.dev().setStepSize_mm(newStepSize_mm, false);
		
		run.fitSitesToVector(true, false, false);
		run.selectAllVectorSites(false);
	
		if (update) run.dev().sendUpdate();
	}
	public void changeSiteCheck(boolean enable, boolean update)
	{
		run.dev().setSiteCheckEnable(enable, false);
		
		if (enable)
		{
			// save standard settings
			run.dev().setSavedFPS(run.getFramesPerSite(), false);
			run.dev().setSavedOverlap(run.getOverlap(), false);
			run.dev().setSavedEndAngle(run.getEndAngle(), false);

			// set up for 1 frame per site
			if (run.getHelicalMode())
			{
				int newSiteSkip = (int)Math.round((double)run.getPotentialVectorSites() / 10.0);
				if (newSiteSkip<1) newSiteSkip=1;
				run.dev().setSiteSkip(newSiteSkip, false);
				run.dev().setSiteDelta_deg(run.getDelta(), false);
			}
			else
			{
				double newSiteDelta = run.getFramesPerSite() * run.getDelta();
				run.dev().setSiteSkip(0, false);
				run.dev().setSiteDelta_deg(newSiteDelta, false);
			}
			run.dev().setOverlap(0, false);
			run.fid.changeFramesPerSite(1, false, false);
		}
		else
		{
			// restore the saved standard settings
			run.dev().setSiteDelta_deg(0.0, false);
			run.dev().setSiteSkip(0, false);
			run.dev().setFramesPerSite(run.getSavedFPS(), false);
			run.dev().setOverlap(run.getSavedOverlap(), false);
			run.dev().setEndAngle(run.getSavedEndAngle(), false);
		}
		
		if (update) run.dev().sendUpdate();
	}
	public void changeSiteCount(Integer newCount, boolean update)
	{
		int currentCount = run.getPotentialVectorSites();
		
		ArrayList<Boolean> newSelectedSites = new ArrayList<Boolean>();
		if (newCount >= currentCount)
		{
			for (int i=0; i<currentCount; i++) 			newSelectedSites.add(run.isSiteSelected(i+1));
			for (int i=currentCount; i<newCount; i++) 	newSelectedSites.add(true);
		}
		else
		{
			for (int i=0; i<newCount; i++)				newSelectedSites.add(run.isSiteSelected(i+1));
		}
		
		run.dev().setSelectedSites(newSelectedSites, false);
		run.dev().setPotentialVectorSites(newCount, false);
		
		if (update) run.dev().sendUpdate();
	}
	public void changeSiteDelta(double siteDelta)
	{
		run.dev().setSiteDelta_deg(siteDelta, true);
	}

	////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////// button handlers ///////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////////
	
	public void handleCalculateButton(ButtonAction buttonAction)
	{
		switch (buttonAction)
		{
		case FINISH:
			run.dev().setSetupStep(SetupStep.COLLECT_PARAM, false);

			run.calculateCrystalPositions();
			run.dev().sendUpdate();
			break;
		case RETURN:
			DBImportSites.setSitesForRun(run.runIndex, null);
			run.dev().clearSelectedSites(false);
			run.dev().setSetupStep(SetupStep.SITE_SETUP, true);
			break;
		}
	}
	
	public void handleFinishButton(ButtonAction buttonAction) 
	{
		switch (buttonAction)
		{
		case FINISH:
			run.dev().setSetupStep(SetupStep.COLLECT_PARAM, false);
			run.dev().setOptionTab(OptionTab.COLLECT_PARAM, false);
			
			if (ScanMode.VECTOR == run.scanMode) run.setVectorCollectParamDefaults(false);
			if (ScanMode.RASTER == run.scanMode) run.setRasterDefaults(false);
			
			run.dev().sendUpdate();
			break;
		case RETURN:
			boolean okToGoBack = Util.popupOkCancel("Are you sure?", "Collect parameters will be lost if you return to site setup.  Proceed?");
			if (!okToGoBack) return;
			
			if (run.siteCheck) changeSiteCheck(false, false);
			run.dev().setSetupStep(SetupStep.SITE_SETUP, true);
			break;
		}
	}
	public void handleScanModeChange(ScanMode newMode)
	{
		if (run.getScanMode() == newMode) return;

		// clean up for current mode
		if (run.getScanMode()==ScanMode.VECTOR)
		{
			boolean okToDeleteVector = true;
			boolean vectorSet = (run.getPt1()!=null || run.getPt2()!=null);
			
			if (vectorSet)
			{
				okToDeleteVector = Util.popupYesNo
				(
					"Delete vector?",
					"Changing from vector mode will DELETE your current vector.  OK to delete?"
				);
			}
			
			if (okToDeleteVector)
			{
				run.dev().setPt1(null, false);
				run.dev().setPt2(null, false);
				run.dev().setScanMode(newMode, false);
				run.dev().setHelicalMode(false, false);
				run.dev().setSiteCheckEnable(false, false);
				run.dev().clearSelectedSites(false);
			}
			else
			{
				run.dev().sendUpdate(); // sends an update with the non-vector mode it was before
				return; // don't proceed with new mode setup
			}
		}

		// set up for new mode
		if (newMode == ScanMode.MESH)
		{
			run.updateMeshParameters(-1, false);
		}

		Object[][] defaultModes = new Object[][]{
				{ScanMode.SINGLE,	SetupStep.COLLECT_PARAM, 	OptionTab.COLLECT_PARAM	},
				{ScanMode.VECTOR,	SetupStep.SITE_SETUP, 		OptionTab.SITE_SETUP	},
				{ScanMode.MESH, 	SetupStep.COLLECT_PARAM, 	OptionTab.SITE_SETUP	},
				{ScanMode.JET, 		SetupStep.COLLECT_PARAM, 	OptionTab.COLLECT_PARAM	}
		};

		for (Object[] modeInfo : defaultModes) {
			ScanMode mode = (ScanMode)modeInfo[0];
			if (mode==newMode) {
				run.dev().setSetupStep((SetupStep)modeInfo[1], false);
				run.dev().setOptionTab((OptionTab)modeInfo[2], false);
			}
		}
		run.dev().setScanMode(newMode, true);
		run.dev().sendUpdate();
	}

	public void reset(boolean forgetSites, boolean update)
	{
		// first stop any running threads
		if (DBConfig.getBoolean("spotfinder_enable_collect"))
		{
			// TODO: put this in a separate thread?
			SpotFinderQueue.stopActiveThreads(SFType.COLLECT, run.getRunIndex());
		}
		
		// clear connected raster sites
		if (forgetSites)
		{
			run.dev().clearSelectedSites(false);
			run.dev().setPt1(null, false);
			run.dev().setPt2(null, false);
			run.dev().setScanMode(ScanMode.SINGLE, false);
			run.dev().setSetupStep(SetupStep.SITE_SETUP, false);
			run.dev().setOptionTab(OptionTab.COLLECT_PARAM, false);
		}

		// reset run state
		run.dev().setRunStatus(RunState.INACTIVE, false);
		run.dev().setNextFrame(0);
		
		run.dev().sendUpdate();
	}
	public void changeHelicalMode(boolean newModeIsHelical, boolean update)
	{
		if (newModeIsHelical)
		{
			run.dev().setOverlap(0, false);
			run.fid.changeFramesPerSite(1, true, false);
			run.fitSitesToVector(false, false, false);
			run.selectAllVectorSites(false);
		}
		run.dev().setHelicalMode(newModeIsHelical, update);
	}

	////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////// constructor /////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////////
	
	private ModelCollectRun run;
	public FieldInteractionIndependent(ModelCollectRun run)
	{
		this.run=run;
	}
}
