package gov.anl.gmca.collect;

import gov.anl.gmca.collect.ModelCollectRun.ScanMode;
import gov.anl.gmca.collect.RunCalculator.Frame;
import gov.anl.gmca.diffimage.HDF5MasterFile;
import gov.anl.gmca.hutch.DetectorConfig;
import gov.anl.gmca.util.Util;

import java.util.regex.Pattern;

public class Filenames
{
	public static Pattern HDF5MasterFilePattern = Pattern.compile
	(
	//   prefix  run  start_frm --------------options--------------
		"(\\S+)_(\\d+)_(\\d+)_(INV_|)(scr_|)(chk_|)(E\\d_|)(S\\d+_|)master.h5"
	);
	public static Pattern HDF5DataFilePattern = Pattern.compile("(\\S+)_data_(\\d+).h5");
	
	//////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////
	
	public static class FilenameInfo
	{
		public FilenameInfo(Frame frame)
		{
			isInverse = frame.isInverse;
			isSiteCheck = frame.siteCheckEnable;
			if (frame.runEnergyCount>1) energyNumber = frame.energyNumber;

	    	boolean vectorWithOverlap = (ScanMode.VECTOR == frame.scanMode && frame.overlap>0);
	    	boolean rasterMode = ScanMode.RASTER == frame.scanMode;
	    	boolean meshMode = ScanMode.MESH == frame.scanMode;
	    	boolean addSiteSuffix = vectorWithOverlap || rasterMode || meshMode;
	    	if (addSiteSuffix) siteNumber = frame.siteNumber+1;
	    	
	    	fileroot = frame.fileroot;
	    	runLabel = frame.runLabel;
	    	firstFrameLabelInPartition = frame.firstFrameLabelInPartition;
	    	indexInPartition = frame.indexInPartition;
		}

		public FilenameInfo(HDF5MasterFile masterFile, int imageNumber)
		{
			firstFrameLabelInPartition = masterFile.firstFrameNumber;
			indexInPartition = imageNumber-1;
			fileroot = masterFile.prefix;
			runLabel = masterFile.runLabel;
			isInverse = masterFile.isInverse;
			isScreening = masterFile.isScreening;
			isSiteCheck = masterFile.isSiteCheck;
			energyNumber = masterFile.energyNumber;
			siteNumber = masterFile.siteNumber;
		}

		public FilenameInfo(){}
		
		public boolean isInverse, isSiteCheck, isScreening;
		public Integer siteNumber=null, energyNumber=null;
		public String fileroot;
		public int runLabel, firstFrameLabelInPartition, indexInPartition;
	}

	// returns options with trailing underscore
	private static String generateOptionsString(FilenameInfo info)
	{
    	String optionsString = "";
    	if (info.isInverse) optionsString += "INV_";
    	if (info.isScreening) optionsString += "scr_";
    	if (info.isSiteCheck) optionsString += "chk_";
    	if (null!=info.energyNumber) optionsString += "E"+info.energyNumber+"_";
    	if (null!=info.siteNumber) optionsString += "S"+info.siteNumber+"_";
    	return optionsString;
	}

	public static String generateFilename(FilenameInfo info)
	{
        if (DetectorConfig.isEiger())
        {
        	String filename = Util.sprintf
        	(
        		"%s_%d_%d_%smaster.h5",
        		info.fileroot,
        		info.runLabel,
        		info.firstFrameLabelInPartition,
        		generateOptionsString(info)
        	);
        	return filename;
        }
        else if (DetectorConfig.isPilatus())
        {
        	String filename = Util.sprintf
        	(
        		"%s_%d_%s%05d.cbf",
        		info.fileroot,
        		info.runLabel,
        		generateOptionsString(info),
        		info.firstFrameLabelInPartition+info.indexInPartition
        	);
        	return filename;
        }
        else if (DetectorConfig.isMAR())
        {
        	String filename = Util.sprintf
        	(
        		"%s_%d%s.%05d",
        		info.fileroot,
        		info.runLabel,
        		generateOptionsString(info),
        		info.firstFrameLabelInPartition
        	);
        	return filename;
        }
        else throw new IllegalStateException("Unsupported detector type: "+DetectorConfig.getType());
	}
	
	public static String generateCBFFilename(FilenameInfo info)
	{
		switch (DetectorConfig.getType())
		{
		case EIGER:
			return String.format
			(
				"%s_%d_%s%06d.cbf",
				info.fileroot,
				info.runLabel,
				generateOptionsString(info),
				info.firstFrameLabelInPartition+info.indexInPartition
			);
		case PILATUS:
			return generateFilename(info);
		case MAR:
			return null;
		default:
			throw new IllegalStateException();
		}
	}

	public static String generateCBFPrefix(FilenameInfo info)
	{
		return String.format
		(
			"%s_%d_%s",
			info.fileroot,
			info.runLabel,
			generateOptionsString(info)
		);
	}
}
