package gov.anl.gmca.collect;

import gov.anl.gmca.collect.DevFPE.XYZMotion;
import gov.anl.gmca.collect.Filenames.FilenameInfo;
import gov.anl.gmca.collect.ModelCollectRun.Point3D;
import gov.anl.gmca.collect.ModelCollectRun.ScanMode;
import gov.anl.gmca.collect.RunCalculator.Frame;
import gov.anl.gmca.gui.Main;
import gov.anl.gmca.gui.options.DevUserConfig;
import gov.anl.gmca.hutch.DetectorConfig;
import gov.anl.gmca.hutch.DevEnergy;
import gov.anl.gmca.util.Util;

import java.util.ArrayList;

// Takes a frame list from the run calculator and creates partitions or sub-groups of frames
// that FPE can collect on.
//
// Sample x/y/z motion is not yet supported - a new partition is created for each new position.
//
// Why create a new array of partition objects instead of adding a partition index to the
// existing frame list? (Add answer here, I forget it at the moment but there is a reason..)
//
public class FramePartitioner
{
	public static class FramePartition
	{
		// start a partition based on a given first frame
		// (motion is set for stationary partitions but not used)
		public FramePartition(Frame frame, int partitionIndex, int originalPartitionIndex)
		{
			this.firstFilename		= frame.hdfFilename;
			this.startAngle			= frame.angle;
			this.endAngle			= frame.angle+frame.delta;
			this.nextAngle			= this.endAngle;
			this.energy				= frame.energy;
			this.energyNumber		= frame.energyNumber;
			this.position 			= frame.position;
			this.firstFrameIndex	= frame.frameIndex;
			this.firstSiteIndex		= frame.siteNumber;
			this.siteIncrement		= 0;
			this.firstFrameLabel	= frame.frameLabel;
			this.frameCount			= 1;
			this.delta				= frame.delta;
			this.partitionIndex		= partitionIndex;
			this.originalPartitionIndex = originalPartitionIndex;
			
			if (null!=frame.position)
			{
				this.shutteredMotion = new DevFPE.XYZMotion(frame.position.x, frame.position.x, frame.position.y, frame.position.y, frame.position.z, frame.position.z);
				this.shutterlessMotion = new DevFPE.XYZMotion(frame.position.x, frame.position.x, frame.position.y, frame.position.y, frame.position.z, frame.position.z);
			}
		}
		
		// should only be called during frame calculations!
		void addFrameToPartition(Frame frame)
		{
			frameCount++;
			endAngle = startAngle + frameCount * delta;
			
			// calculate siteIncrement if this is the second frame
			if (frameCount==2)
			{
				siteIncrement = frame.siteNumber - firstSiteIndex;
			}
			// otherwise verify that the increment is correct
			else
			{
				int expectedSiteIndex = firstSiteIndex + (frameCount-1)*siteIncrement;
				if (expectedSiteIndex != frame.siteNumber)
				{
					throw new IllegalArgumentException("First site index = "+firstSiteIndex+", increment = "+siteIncrement+", existing count = "+frameCount+", expected site index "+expectedSiteIndex+", got "+frame.siteNumber);
				}
			}
			
			// update shutteredMotion variable
			if (null != shutteredMotion && null != frame.position)
			{
				shutteredMotion.endX = frame.position.x;
				shutteredMotion.endY = frame.position.y;
				shutteredMotion.endZ = frame.position.z;
			}
			
			// update shutterlessMotion variable
			if (null != frame.position)
			{
				if (frameCount==2)
				{
					vectorBetweenSites = frame.position.minus(position);
					shutterlessMotion = new XYZMotion
					(
						position.minus(vectorBetweenSites.times(0.5)),
						frame.position.plus(vectorBetweenSites.times(0.5))
					);
				}
				else
				{
					shutterlessMotion.setEndpoint(frame.position.plus(vectorBetweenSites.times(0.5)));
				}
			}
		}
		
		public String 	firstFilename, prefix;
		public double 	startAngle;
		public double	endAngle;
		public double	nextAngle;
		public double 	energy;
		public int		energyNumber;
		public Point3D 	position, vectorBetweenSites;
		public int 		firstFrameIndex;	// 0-based
		public int 		firstSiteIndex;		// 0-based
		public int		siteIncrement;
		public int		firstFrameLabel;
		public int		frameCount;
		public double	delta;
		public int		partitionIndex;		// -1 means it was already collected
		public int		originalPartitionIndex;
		public DevFPE.XYZMotion shutteredMotion, shutterlessMotion;
		public DevFPE.FPEMode fpeMode;
		
		public String getFilename(int frameSubIndex)
		{
			int frameLabel = firstFrameLabel + frameSubIndex;
			int suffixLength = DetectorConfig.isDectris() ? 8 : 4;
			String fileExtension = DetectorConfig.isDectris() ? ".cbf" : "";
			
			String filenamePrefix = firstFilename.substring(0, firstFilename.length()-suffixLength);
			String fullFilename = filenamePrefix+Util.sprintf("%05d"+fileExtension,frameLabel);
			return fullFilename;
		}

		public int getSiteIndexForFrameIndex(int frameIndex)
		{
			return firstSiteIndex + frameIndex * siteIncrement;
		}

		public String getFilenameToSendToFPE()
		{
			if (DetectorConfig.isEiger())
			{
				int prefixEndIndex = firstFilename.indexOf("_master.h5");
				return firstFilename.substring(0, prefixEndIndex);
			}
			else
			{
				return firstFilename;
			}
		}
	}
	
	// This is the key to this class - it looks at a frame and sees if it lines up
	// with the last frame in the partition.
	//
	private static boolean frameBelongs(Frame frame, FramePartition partition, ScanMode scanMode, boolean siteCheckMode)
	{
		if (null==frame || null==partition) return false;
		
		if (DetectorConfig.isEiger() && !DevUserConfig.get().getShutterlessEnable())
		{
			return false;
		}

		// filenames should match except for the frame number
		if (!DetectorConfig.isEiger() && frame.hdfFilename.length()>=4)
		{
			int suffixLength=0;
			if (DetectorConfig.isEiger()) suffixLength = 9; // 6 digit image number + .cbf
			if (DetectorConfig.isPilatus()) suffixLength = 8; // 5 digit image number + .cbf
			if (DetectorConfig.isMAR()) suffixLength = 4; // 4 digit image number
			if (0==suffixLength) throw new IllegalStateException("Unknown detector type");
			
			String framePrefix = frame.hdfFilename.substring(0, frame.hdfFilename.length()-suffixLength);
			String partitionPrefix = partition.firstFilename.substring(0, partition.firstFilename.length()-suffixLength);
			if (!framePrefix.equals(partitionPrefix)) return false;
		}
		
		// frame numbers should increment one per frame
		int frameNumber = frame.frameLabel;
		int expectedFrameNumber = partition.firstFrameLabel + partition.frameCount;
		if (frameNumber != expectedFrameNumber) return false;
		
		// energy shouldn't change
		if (frame.energyNumber != partition.energyNumber) return false;
		if (Math.abs(frame.energy-partition.energy)>0.0001) return false;

		// don't group frames in site check mode
		if (siteCheckMode) return false;
		
		// in vector mode, if fps>1 then site change means new partition
		// we detect this by looking back two frames - if the last two were the
		// same site and this one is different, then start a new partition
		if
		(
			scanMode == ScanMode.VECTOR &&
			partition.frameCount>=2 &&
			partition.siteIncrement == 0 &&
			frame.siteNumber != partition.firstSiteIndex
		)
		return false;
		
		if (Math.abs(partition.endAngle-frame.angle)>0.01) return false;
		
		return true;
	}

	public static FramePartition findPartitionIndex(ArrayList<FramePartition> list, int index) {
		for (FramePartition part : list) {
			if (part.partitionIndex==index) return part;
		}
		return null;
	}

	// the only public method in this class
	//
	public static synchronized ArrayList<FramePartition> partition(ArrayList<Frame> frameList, RunCalculator runCalc, int nextFrameIndex, ScanMode scanMode, boolean siteCheckMode)
	{
		// check for 0 length frame list
		if (0 == frameList.size()) return new ArrayList<FramePartition>();
		
		// create a blank partition list
		ArrayList<FramePartition> partitionList = new ArrayList<FramePartition>();
		
		/////////////////////////////////// partition the frames /////////////////////////////////

		FramePartition lastPartition = null;
		int notTakenPartitionIndex=0, takenPartitionIndex=-1, originalIndexCounter=0;
		for (Frame frame : frameList)
		{
			boolean frameTaken 	= (frame.frameIndex <  nextFrameIndex);
			boolean isNextFrame = (frame.frameIndex == nextFrameIndex);

			// for the first frame, create a partition for it
			if (null == lastPartition)
			{
				int partitionIndex = frameTaken ? takenPartitionIndex-- : notTakenPartitionIndex++;
				int originalPartitionIndex = originalIndexCounter++;
				lastPartition = new FramePartition(frame, partitionIndex, originalPartitionIndex);
			}
			else if (!isNextFrame && frameBelongs(frame, lastPartition, scanMode, siteCheckMode)) // add to an existing part. if possible
			{
				lastPartition.addFrameToPartition(frame);
			}
			else // otherwise create a new partition and add the existing one to the list
			{
				lastPartition.nextAngle = frame.angle; // set the next angle of the last partition to the start angle of this frame
				partitionList.add(lastPartition);
				
				int partitionIndex = frameTaken ? takenPartitionIndex-- : notTakenPartitionIndex++;
				int originalPartitionIndex = originalIndexCounter++;
				lastPartition = new FramePartition(frame, partitionIndex, originalPartitionIndex);
			}
		}
		
		// add the last partition to the list
		if (null != lastPartition)
		{
			lastPartition.nextAngle = lastPartition.endAngle; // have the last partition just stop at the last end angle
			partitionList.add(lastPartition);
		}

		// reverse the negative partition numbers
		for (FramePartition partition : partitionList)
		{
			if (partition.partitionIndex < 0)
			{
				partition.partitionIndex = ++takenPartitionIndex;
			}
		}

		// add FPE modes
		for (FramePartition part : partitionList)
		{
			if (part.partitionIndex<0) continue;

			FramePartition lastPart = part.partitionIndex==0?null:findPartitionIndex(partitionList, part.partitionIndex-1);
			FramePartition nextPart = findPartitionIndex(partitionList, part.partitionIndex+1);

			boolean isHelical = scanMode==ScanMode.VECTOR && part.siteIncrement!=0;
			if
			(
				DetectorConfig.isDectris() &&
				DevUserConfig.get().getShutterlessEnable() &&
				part.frameCount>1
			)
			{
				boolean isFirstPart = null==lastPart;
				boolean detectorWasDisarmed = null != lastPart && lastPart.fpeMode != DevFPE.FPEMode.SHUTTERLESS_ARM_NODISARM && lastPart.fpeMode != DevFPE.FPEMode.SHUTTERLESS_NOARM_NODISARM;
				boolean arm = isFirstPart || detectorWasDisarmed;

				boolean isLastPart = null==nextPart;
				boolean nextPartHasEnergyChange = null != nextPart && Math.abs(part.energy - nextPart.energy) >= DevEnergy.getDevice().getInPosBand();
				boolean disarm = isLastPart || nextPartHasEnergyChange;

				if (Main.STILL_SINGLE_ARM_NOT_IMPLEMENTED) { arm=true; disarm=true; }

				if (arm  &&  disarm) part.fpeMode = DevFPE.FPEMode.SHUTTERLESS_ARM_DISARM;
				if (arm  && !disarm) part.fpeMode = DevFPE.FPEMode.SHUTTERLESS_ARM_NODISARM;
				if (!arm &&  disarm) part.fpeMode = DevFPE.FPEMode.SHUTTERLESS_NOARM_DISARM;
				if (!arm && !disarm) part.fpeMode = DevFPE.FPEMode.SHUTTERLESS_NOARM_NODISARM;
			}
			else
			{
				part.fpeMode = isHelical ? DevFPE.FPEMode.SHUTTERED_VECTOR : DevFPE.FPEMode.STANDARD;
			}
		}

		// copy partition indices to individual frames
		int frameIndex=0;
		for (FramePartition partition : partitionList)
		{
			String firstFilename = null, prefix = null;
			int firstFrameLabel = -1;
			for (int i=0; i<partition.frameCount; i++)
			{
				Frame frame = frameList.get(frameIndex);
				if (-1==firstFrameLabel) firstFrameLabel = frame.frameLabel;
				
				frame.partitionIndex = partition.partitionIndex;
				frame.originalPartitionIndex = partition.originalPartitionIndex;
				frame.indexInPartition = i;
				frame.firstFrameLabelInPartition = firstFrameLabel;
				frame.partitionSize = partition.frameCount;
				frame.hdfFilename = runCalc.generateFilename(frame);
				FilenameInfo fileInfo = new FilenameInfo(frame);
				frame.cbfFilename = Filenames.generateCBFFilename(fileInfo);
				frame.cbfPrefix = Filenames.generateCBFPrefix(fileInfo);
				if (null == firstFilename) firstFilename = frame.hdfFilename;
				if (null == prefix) prefix = frame.cbfPrefix;
				frameIndex++;
			}
			partition.firstFilename = firstFilename;
			partition.prefix = prefix;
		}
		
		// done!
		return partitionList;
	}
}
