package gov.anl.gmca.collect;

import gov.anl.gmca.collect.DBImportSites.ImportSite;
import gov.anl.gmca.collect.Filenames.FilenameInfo;
import gov.anl.gmca.collect.FramePartitioner.FramePartition;
import gov.anl.gmca.collect.ModelCollectRun.Point3D;
import gov.anl.gmca.collect.ModelCollectRun.ScanMode;
import gov.anl.gmca.collect.ModelCollectRun.SetupStep;
import gov.anl.gmca.common.Limits;
import gov.anl.gmca.hutch.DetectorConfig;
import gov.anl.gmca.hutch.DevPmacMotor;
import gov.anl.gmca.raster.ModelRasterHits.Hit;
import gov.anl.gmca.raster.ModelRasterSiteList.Site;
import gov.anl.gmca.syslog.Log;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.Semaphore;

/* 
 * Calculates a list of frames given a run.
 */
public class RunCalculator
{
	public static class Frame
	{
		public Frame(String hdfFilename, double angle, double energy, int energyNumber,
					 Point3D position, int frameNumber, int siteIndex, int frameLabel,
					 double delta, boolean isInverse,
					 boolean siteCheckEnable, int runEnergyCount, ScanMode scanMode,
					 int overlap, String fileroot, int runLabel)
					 
		{
			this.hdfFilename	= hdfFilename;
			this.angle			= angle;
			this.energy			= energy;
			this.energyNumber	= energyNumber;
			this.position 		= position;
			this.frameIndex		= frameNumber;
			this.siteNumber		= siteNumber;
			this.frameLabel		= frameLabel;
			this.delta			= delta;
			this.isInverse		= isInverse;
			
			this.siteCheckEnable = siteCheckEnable;
			this.runEnergyCount	 = runEnergyCount;
			this.scanMode		 = scanMode;
			this.overlap		 = overlap;
			this.fileroot		 = fileroot;
			this.runLabel		 = runLabel;
		}

		public Frame(Frame frame)
		{
			this.hdfFilename 	= frame.hdfFilename;
			this.cbfFilename	= frame.cbfFilename;
			this.cbfPrefix		= frame.cbfPrefix;
			this.angle 			= frame.angle;
			this.energy 		= frame.energy;
			this.energyNumber	= frame.energyNumber;
			this.position 		= (null==frame.position)?null:new Point3D(frame.position.x, frame.position.y, frame.position.z);
			this.frameIndex 	= frame.frameIndex;
			this.siteNumber		= frame.siteNumber;
			this.frameLabel		= frame.frameLabel;
			this.delta			= frame.delta;
			this.isInverse		= frame.isInverse;
			
			this.siteCheckEnable = frame.siteCheckEnable;
			this.runEnergyCount	 = frame.runEnergyCount;
			this.scanMode		 = frame.scanMode;
			this.overlap		 = frame.overlap;
			this.fileroot		 = frame.fileroot;
			this.runLabel		 = frame.runLabel;
		}

		public int		indexInPartition;
		public int		partitionSize;
		public String 	hdfFilename; // null if not written
		public String	cbfFilename; // null if not written
		public String	cbfPrefix;
		public double 	angle;
		public double 	energy;
		public int		energyNumber;
		public Point3D 	position;
		public int		partitionIndex;
		public int		originalPartitionIndex;
		public int 		frameIndex;		// 0-based
		public int 		firstFrameLabelInPartition;
		public int 		siteNumber;		// 0-based
		protected int	frameLabel;	
		public double	delta;			// doesn't change between frames but is needed for frame partitioning
		public boolean 	isInverse;
		public Frame	inverseFrame;	// original and inverse frames point to each other

		public boolean 	siteCheckEnable;
		public int 		runEnergyCount;
		public ScanMode scanMode;
		public int 		overlap;
		public String 	fileroot;
		public int 		runLabel;
		
		public String getDisplayFilename()
		{
			if (!DetectorConfig.isEiger()) return hdfFilename;
			if (-1==frameIndex) return hdfFilename; // -1 for frameIndex indicates the end marker in the run sequence
			
			//return filename+":"+(indexInPartition+1); // image numbering should start with 1
			return cbfFilename;
		}
	}
	
	private ArrayList<Frame> frameList = new ArrayList<Frame>();
	private ArrayList<FramePartition> partitionList = new ArrayList<FramePartition>();
	private Semaphore frameListLock = new Semaphore(1);
	
	public ArrayList<FramePartition> getFramePartitions()
	{
		ArrayList<FramePartition> ret = null;

		frameListLock.acquireUninterruptibly();
		try
		{
			ret = new ArrayList<FramePartition>(partitionList);
		}
		catch (Throwable t)
		{
			t.printStackTrace();
		}
		finally
		{
			frameListLock.release();
		}
		
		return ret;
	}
	
	public ArrayList<Frame> getFrameListCopy() 
	{
		ArrayList<Frame> ret = null;

		frameListLock.acquireUninterruptibly();
		try
		{
			ret = new ArrayList<Frame>(frameList);
		}
		catch (Throwable t)
		{
			t.printStackTrace();
		}
		finally
		{
			frameListLock.release();
		}
		
		return ret;
	}
	public List<Frame> getRemainingFrameList()
	{
		return getFrameListCopy().subList(run.getNextFrame(), frameList.size());
	}
	public int getRemainingFrameCount()
	{
		return frameList.size()-run.getNextFrame();
	}
	public int getTotalFrameCount()
	{
		return frameList.size();
	}
	
	protected 	ModelCollectRun 	run;
	protected 	long 				versionNumber 			= 0;
	private		Limits				omegaLimits				= null;
	private		boolean				isNonHelicalVector		= false;
	public		boolean				getIsNonHelicalVector() { return isNonHelicalVector; }
	public 		long 				getVersion() 			{ return versionNumber; }
	
	public RunCalculator(ModelCollectRun _run)
	{
		this.run=_run;
		omegaLimits = DevPmacMotor.get("gonio_omega").getLimits();
	}
	
	public void recalc()
	{
		frameListLock.acquireUninterruptibly();
		
		try
		{
			frameList.clear();
			
			// don't calculate in site setup mode
			if (ScanMode.SINGLE != run.getScanMode() && ScanMode.JET != run.getScanMode() && SetupStep.SITE_SETUP == run.getSetupStep())
			{
				System.out.println("scanMode="+run.getScanMode()+" and setupStep="+run.getSetupStep()+" so recalc is skipped.");
				versionNumber++; return;
			}
			// run sequence is empty in jet mode
			if (run.getScanMode()==ScanMode.JET) {
				versionNumber++; return;
			}
			
			generateInitialFrameList();
			addStrategyFrame();
			addSites();
			applySiteSkip();
			addOverlap();
			addInverseBeam();
			addImportSites();
			addMeshSites();
			addEnergies();
			generateFrameNumbers();
			partitionList = FramePartitioner.partition(frameList, this, run.getNextFrame(), run.getScanMode(), run.getSiteCheckEnable());
			determineHelical();
			
			versionNumber++;
		}
		catch (Throwable t)
		{
        	frameList.clear();
        	partitionList.clear();
        	versionNumber++;
        	
			Log.error("Error calculating frame sequence: "+t.getMessage());
			t.printStackTrace();
		}
		finally
		{
			frameListLock.release();
		}
	}

	public static int calcFrameCount(double delta, double startAngle, double endAngle)
	{
		return (int)Math.floor( (endAngle - startAngle + 0.0000001) / delta );
	}
	
	private void generateInitialFrameList()
	{
		int frameCount = calcFrameCount(run.getDelta(), run.getStartAngle(), run.getEndAngle());

		for (int frameIndex = 0; frameIndex < frameCount; frameIndex++)
		{
			// determine filename, energy, angle
	        int frameLabel = run.getStartFrameLabel() + frameIndex;
	        double energy = run.getEnergies().get(0);
	    	double delta = run.getSiteCheckEnable() ? run.getSiteDelta_deg() : run.getDelta();
	        double angle = run.getStartAngle() + frameIndex * delta;
	    	
	    	// create the frame
			Frame frame = new Frame("", angle,energy,0,null,frameIndex,0,frameLabel,delta,false,
									run.getSiteCheckEnable(), run.getEnergies().size(), run.getScanMode(),
									run.getOverlap(),run.getFileRoot(),run.getRunLabel());
	    	frameList.add(frame);
		}
	}
	
	private void applySiteSkip()
	{
		if (run.getSiteSkip()==0) return;
		
		ArrayList<Frame> newFrameList = new ArrayList<Frame>();
		for (int i=0; i<frameList.size(); i+=run.getSiteSkip())
		{
			newFrameList.add(frameList.get(i));
		}
		frameList = newFrameList;
	}
	
	private void addStrategyFrame()
	{
		if (run.getStrategyEnabled() && frameList.size() > 0)
		{
			Frame frame0 = frameList.get(0);
			Frame frame = new Frame("", run.getStartAngle2(),frame0.energy,0,frame0.position,1,0,
									run.getStartFrameLabel()+2,run.getDelta(),false,
									run.getSiteCheckEnable(), run.getEnergies().size(), run.getScanMode(),
									run.getOverlap(),run.getFileRoot(),run.getRunLabel());
			frameList.add(frame);
		}
	}
	
	private void addSites()
	{
		if (ScanMode.VECTOR == run.getScanMode())
		{
			Point3D u = run.getUnitVector();
			Point3D startPoint = run.getReverseDir() ? run.getPt2() : run.getPt1();
			
	        // loop preparation
			int siteIndex = 0, framesLeftThisSite = run.getFramesPerSite();
			boolean framesFit = true;
			
			// for each frame
			for (int frameIndex = 0; frameIndex < frameList.size(); frameIndex++)
			{
				// set the site number
				while (!run.isSiteSelected(siteIndex+1)) { siteIndex++; if (siteIndex > run.getPotentialVectorSites()) break; }
				frameList.get(frameIndex).siteNumber = siteIndex;
				
				if (siteIndex > run.getPotentialVectorSites()) framesFit = false;
				
				// set the frame position
	            double x = startPoint.x + u.x * (run.getStepSize_mm() * siteIndex);
	            double y = startPoint.y + u.y * (run.getStepSize_mm() * siteIndex);
	            double z = startPoint.z + u.z * (run.getStepSize_mm() * siteIndex);
				frameList.get(frameIndex).position = new Point3D(x,y,z);
				
				// if this site is done, increment the site index
				framesLeftThisSite--;
				if (framesLeftThisSite==0)
				{
					framesLeftThisSite = run.getFramesPerSite();
					siteIndex++;
				}
			}
			
			if (!framesFit) Log.error("Frames did not fit on given vector sites");
		}
		else if (ScanMode.RASTER == run.getScanMode())
		{
	        // loop preparation
			int siteIndex = 0, framesLeftThisSite = run.getFramesPerSite();
			ArrayList<Site> selectedSites = run.getSelectedRasterSiteObjects();
			
			// for each frame
			for (int frameIndex = 0; frameIndex < frameList.size(); frameIndex++)
			{
				if (siteIndex >= selectedSites.size()) { Log.error("Frames did not fit in selected sites ("+selectedSites.size()+")"); break; }
				
				Frame frame = frameList.get(frameIndex);
				
				// set the site number
				frame.siteNumber = siteIndex;
				
				Site site = selectedSites.get(siteIndex);
				
				// set the frame position
				frame.position = new Point3D(site.motorPos.x, site.motorPos.y, site.motorPos.z);
				
				// if this site is done, increment the site index
				framesLeftThisSite--;
				if (framesLeftThisSite==0)
				{
					framesLeftThisSite = run.getFramesPerSite();
					siteIndex++;
				}
			}
		}
	}
	
	private void addOverlap()
	{
		if (ScanMode.SINGLE == run.getScanMode() || ScanMode.JET == run.getScanMode() || run.getOverlap() <= 0) return;
		
		// loop prep
		int indexWithoutOverlap = 0;
		
		// for each frame
		for (int frameIndex = 0; frameIndex < frameList.size(); frameIndex++)
		{
			// if this is the last frame for this site
			if (frameIndex == frameList.size()-1 ||
				frameList.get(frameIndex).siteNumber != frameList.get(frameIndex+1).siteNumber)
			{
				Frame lastFrame = frameList.get(frameIndex);
				
				// add on the extra overlap frames
				for (int overlapIter=0; overlapIter<run.getOverlap(); overlapIter++)
				{
					Frame newFrame = new Frame(lastFrame);
					newFrame.frameLabel = lastFrame.frameLabel + 1 + overlapIter;
					newFrame.angle = run.getStartAngle() + (indexWithoutOverlap+1+overlapIter) * run.getDelta();
					frameList.add(frameIndex+1+overlapIter, newFrame);
				}
				
				frameIndex += run.getOverlap();
			}
			
			indexWithoutOverlap++;
		}		
	}
	
	private void addInverseBeam()
	{
		if (!run.getInverseOn()) return;
		
		// make a separate array to create a new frame list in
		ArrayList<Frame> newFrameList = new ArrayList<Frame>();

		// this is a temporary list which collects a wedge of frames and then
		// is emptied at the end of each wedge
		ArrayList<Frame> wedgeOfInverseFrames = new ArrayList<Frame>();
		int framesPerWedge = (int)Math.round(run.getWedgeSize_deg() / run.getDelta());
		if (run.getSiteCheckEnable()) framesPerWedge = run.getWedgeSize_sites();
		
		// calculate n, the number of extra 360's to add to ensure no frame label collisions
		/*int lastNonInverseLabel = frameList.get(frameList.size()-1).frameLabel;
		int n=0;
		while (lastNonInverseLabel >= 180+n*360 && n<50) n++;
		if (n==50) { Log.error("Inverse beam frames would not fit under 9999, not adding."); return; }*/

		// for each frame in the original list...
		for (Frame frame : frameList)
		{
			// add each frame as it's iterated over
			newFrameList.add(frame);
			
			// gather a wedge worth of frames and add 180
			Frame f = new Frame(frame);
			f.isInverse = true;
			f.angle += 180.0; if (f.angle > omegaLimits.getUpper()) f.angle -= 360.0;
			//f.frameLabel = f.frameLabel+180+360*n;
			wedgeOfInverseFrames.add(f);
			
			// point the inverse frames to each other
			frame.inverseFrame = f;
			f.inverseFrame = frame;

			// when the wedge is complete, add the +180 frames and start a new wedge
			if (wedgeOfInverseFrames.size() == framesPerWedge)
			{
				newFrameList.addAll(wedgeOfInverseFrames);
				wedgeOfInverseFrames.clear();
			}
		}

		// add the remainder of this wedge
		newFrameList.addAll(wedgeOfInverseFrames);

		// replace the frame list with the new one
		frameList = newFrameList;
	}

	private void addMeshSites()
	{
		if (run.getScanMode()!=ScanMode.MESH) return;

		// convert hit locations to a list of motor positions
		ArrayList<Hit> hits = run.getHits();
		ArrayList<Point3D> positions = new ArrayList<>();
		for (Hit hit : hits) {
			positions.add(hit.getMotorXYZO().to3D());
		}

		// repeat the frame list for each motor position
		addRepeatedSites(positions);
	}

	private void addImportSites()
	{
		if (run.getScanMode()!=ScanMode.IMPORT) return;
		
		// get site count and positions
		int numImportSites = run.getImportSiteCount();
		ArrayList<ImportSite> siteInfo = DBImportSites.getAllImportSites();
		if (siteInfo.size()!=numImportSites)
		{
			String err = "Import site count PV value ("+numImportSites+") doesn't equal "+
						 " DB count ("+siteInfo.size()+")";
			throw new IllegalStateException(err);
		}

		// take the positions out of the site list
		ArrayList<Point3D> positions = new ArrayList<>();
		for (int importSiteIndex=0; importSiteIndex<numImportSites; importSiteIndex++) {
			if (!run.isSiteSelected(importSiteIndex + 1)) continue;
			positions.add(siteInfo.get(importSiteIndex).xyz);
		}

		// repeat the frame list for each site
		addRepeatedSites(positions);
	}

	private void addRepeatedSites(ArrayList<Point3D> positions)
	{
		// make a separate array to create a new frame list in
		ArrayList<Frame> newFrameList = new ArrayList<Frame>();

		int siteNumber=0;
		for (Point3D position : positions)
		{
			// for each frame in the original list...
			for (Frame frame : frameList)
			{
				// create a new frame with the current site number
				Frame newFrame = new Frame(frame);
				newFrame.siteNumber = siteNumber;

				// store the site position
				newFrame.position = position;

				// add the frame to the list
				newFrameList.add(newFrame);
			}
			siteNumber++;
		}

		// replace the frame list with the new one
		frameList = newFrameList;
	}
	
	// TODO: take into account wedge size
	private void addEnergies()
	{
		if (run.getEnergies().size()==1) return;
		
		ArrayList<Frame> newFrameList = new ArrayList<Frame>();
		
		ArrayList<Frame> wedgeOfFrames = new ArrayList<Frame>();
		int framesPerWedge = (int)Math.round(run.getWedgeSize_deg() / run.getDelta());
		if (run.getInverseOn()) framesPerWedge *= 2; // inv. beam frames are duplicated too
		
		for (Frame frame : frameList)
		{
			// add each frame as it's iterated over, adding the energy prefix to filenames if needed
			frame.energyNumber = 1;
			newFrameList.add(frame);
			
			// gather a wedge worth of frames
			wedgeOfFrames.add(new Frame(frame));

			// when the wedge is complete, duplicate them for each energy
			if (wedgeOfFrames.size() == framesPerWedge)
			{
				addWedgeCopiesForEachEnergy(wedgeOfFrames, newFrameList);
				wedgeOfFrames.clear();
			}
		}

		// copy energies for the last partial wedge
		addWedgeCopiesForEachEnergy(wedgeOfFrames, newFrameList);
		
		// replace the frame list with the new one
		frameList = newFrameList;
	}

	// helper method for addEnergies()
	private void addWedgeCopiesForEachEnergy(ArrayList<Frame> wedgeOfFrames, ArrayList<Frame> newFrameList)
	{
		// for each energy
		for (int energyIndex=1; energyIndex<run.getEnergies().size(); energyIndex++)
		{
			double energy = run.getEnergies().get(energyIndex);
			
			// create a copy wedge with a different energy
			ArrayList<Frame> wedgeCopy = new ArrayList<Frame>();
			for (Frame f2 : wedgeOfFrames)
			{
				Frame f2Copy = new Frame(f2);
				f2Copy.energy = energy;
				f2Copy.energyNumber = energyIndex+1;
				wedgeCopy.add(f2Copy);
			}
			
			// and add it to the main list
			newFrameList.addAll(wedgeCopy);
		}
	}
	
	private void generateFrameNumbers()
	{
		for (int i=0; i<frameList.size(); i++)
		{
			Frame frame = frameList.get(i);
			frame.frameIndex=i;
		}
	}
	
	public String generateFilename(Frame frame)
	{
		return Filenames.generateFilename(new FilenameInfo(frame));
	}
	
	private void determineHelical()
	{
		isNonHelicalVector=false;
		for (FramePartition partition : partitionList)
		{
			if (partition.siteIncrement==0 && partition.frameCount>1) isNonHelicalVector=true;
		}
	}
}
