package gov.anl.gmca.collect.ui.cam;

import gov.anl.gmca.collect.ModelCollectRun.Point3D;
import gov.anl.gmca.collect.ModelCollectRun.ScanMode;
import gov.anl.gmca.collect.ModelCollectRunList;
import gov.anl.gmca.common.DevMove;
import gov.anl.gmca.common.ModelRunList.CurrentRunListener;
import gov.anl.gmca.common.ModelRunList.ModelRun;
import gov.anl.gmca.gui.GUI;
import gov.anl.gmca.hutch.DevPmacMotor;
import gov.anl.gmca.util.ListenerList;
import gov.anl.gmca.util.Point4D;
import org.eclipse.swt.events.ControlEvent;
import org.eclipse.swt.events.ControlListener;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Composite;

import java.awt.*;
import java.util.ArrayList;
import java.util.concurrent.Semaphore;
import java.util.concurrent.atomic.AtomicBoolean;

// Since this class is only used from the GUI thread we don't have to worry about
// multiple overlapping calls.  Therefore we store temporary info in member variables
// knowing they won't change within a call.
//
public class WdgVectorCamera extends WdgSiteCamera
{
	private static final String[] motorsToWatch = new String[]{"camera_zoom_2","sample_x","sample_y","sample_z","gonio_omega"};
	private ListenerList.Listener<DevMove> positionListener;
	private CurrentRunListener currentRunListener;
	
	private ArrayList<AtomicBoolean> numberedSites = null, drawnSites = null;
	private int pointSize, lineSize, borderWidth;
	private Color vectorBorder, vectorBG;
	private Point3D start, end;
	private Point startPtScreen, endPtScreen;
	private double omega;
	private Semaphore siteNumberingLock = new Semaphore(1);
	
	/////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////// top-level draw methods ////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	
	@Override protected void drawAll()
	{
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		vectorInit();
		drawVector();
		drawVectorSites();
	}

	private void vectorInit()
	{
		// set sizes and colors
		drawNumbers	 		= ModelCollectRunList.get().getShowVectorNumbers();
		pointSize 			= drawNumbers?20:10;
		lineSize 			= drawNumbers?20:5;
		borderWidth 		= 2;
		vectorBorder 		= drawNumbers?java.awt.Color.white		:java.awt.Color.black;
		vectorBG			= drawNumbers?GUI.COLOR_LIGHT_BLUE5_AWT	:GUI.COLOR_BRIGHT_GREEN_AWT;
		
		// save omega
		omega = DevPmacMotor.get("gonio_omega").getPosition();
		
		// determine start/end points
		start = run.getReverseDir() ? run.getPt2() : run.getPt1();
		end = run.getReverseDir() ? run.getPt1() : run.getPt2();
		if (null!=start) startPtScreen = motorMMtoScreenIntPx(new Point4D(start.x,start.y,start.z,omega));
		if (null!=end) endPtScreen = motorMMtoScreenIntPx(new Point4D(end.x,end.y,end.z,omega));
	}

	private void drawVector()
	{
		// in no-number mode, the line is drawn behind the endpoints
		if (!drawNumbers) drawConnectingLine();
		
		// start point
		if (null != start)
		{
			g2.setColor(vectorBorder);
			g2.fillOval((int)(startPtScreen.x-(pointSize/2))-1, (int)(startPtScreen.y-(pointSize/2))-1, pointSize+borderWidth, pointSize+borderWidth);
			g2.setColor(vectorBG);
			g2.fillOval((int)(startPtScreen.x-(pointSize/2)), (int)(startPtScreen.y-(pointSize/2)), pointSize, pointSize);
		}
		
		// end point
		if (null != end)
		{
			g2.setColor(vectorBorder);
			g2.fillOval((int)(endPtScreen.x-(pointSize/2))-1, (int)(endPtScreen.y-(pointSize/2))-1, pointSize+borderWidth, pointSize+borderWidth);
			g2.setColor(vectorBG);
			g2.fillOval((int)(endPtScreen.x-(pointSize/2)), (int)(endPtScreen.y-(pointSize/2)), pointSize, pointSize);
		}
		
		// in number mode, the line is drawn in front
		if (drawNumbers) drawConnectingLine();
	}

	protected void drawConnectingLine()
	{
		if (null != start && null != end)
		{
			g2.setStroke(new BasicStroke(lineSize+borderWidth));
			g2.setColor(vectorBorder);
			g2.drawLine(startPtScreen.x,startPtScreen.y,endPtScreen.x,endPtScreen.y);
			
			g2.setStroke(new BasicStroke(lineSize));
			g2.setColor(vectorBG);
			g2.drawLine(startPtScreen.x,startPtScreen.y,endPtScreen.x,endPtScreen.y);

			g2.setStroke(new BasicStroke(1));
		}
	}
	
	protected void drawVectorSites()
	{
		if (null==start || null==end) return;
		
		if (null == numberedSites) recalcSiteVisibility();
		
		try
		{
			siteNumberingLock.acquire();
			
			// draw sites marked as visible
			for (int i=0; i<numberedSites.size(); i++)
			{
				if (drawnSites.get(i).get() || !drawNumbers) outlineSite(getSiteXY(i));
			}
			for (int i=0; i<numberedSites.size(); i++)
			{
				if (drawnSites.get(i).get() || !drawNumbers) fillSite(i,getSiteXY(i));
			}
			for (int i=0; i<numberedSites.size(); i++)
			{
				if (numberedSites.get(i).get()) drawSiteNumber(i,getSiteXY(i));
			}
		}
		catch (Throwable t) { t.printStackTrace(); }
		finally
		{
			siteNumberingLock.release();
		}
	}

	// called for:
	// 1) end of move for motors: high-res zoom, sample x, sample y, sample z, gonio omega
	// 2) camera resize
	//
	private void recalcSiteVisibility()
	{
		if (null == run || null == start || null == end) return;
		if (run.getScanMode()!=ScanMode.VECTOR) return;
		
		try
		{
			siteNumberingLock.acquire();
			
			vectorInit(); // update positions
			
			// site count/size
			int 	siteCount 		= run.getPotentialVectorSites();
			double 	siteSizePx 		= siteCount > 20 ? pointSize/0.8 : pointSize/1.5;
	
			// vector length
			Point a = getSiteXY(0);
			Point b = getSiteXY(siteCount-1);
			double dx = a.x-b.x;
			double dy = a.y-b.y;
			
			// (add one extra site size to account for the extra room we get by having half site ends
			// past each end of the vector)
			double 	vectorLengthPx 	= Math.sqrt(dx*dx+dy*dy)+siteSizePx;
			
			// calc numbered site step
			double sitesThatFit = vectorLengthPx/siteSizePx;
			int numberedSiteStep = (int)Math.ceil(siteCount/sitesThatFit); // rounding down would cause overlap
			if (numberedSiteStep>siteCount-1) numberedSiteStep=siteCount-1;
			if (numberedSiteStep<1) numberedSiteStep=1;
			
			// calc drawn site step
			int drawnSiteStep = numberedSiteStep/2;
			if (drawnSiteStep<1) drawnSiteStep=1;
	
			// create the arrays
			numberedSites = new ArrayList<AtomicBoolean>();
			drawnSites = new ArrayList<AtomicBoolean>();
			for (int i=0; i<siteCount; i++)
			{
				numberedSites.add(new AtomicBoolean(false));
				drawnSites.add(new AtomicBoolean(false));
			}
			
			// fill the arrays
			for (int i=0; i<siteCount; i += numberedSiteStep)
			{
				numberedSites.get(i).set(true);
			}
			numberedSites.get(numberedSites.size()-1).set(true); // always number the last site
			if (drawnSiteStep >= 0)
			{
				for (int i=0; i<siteCount; i+=drawnSiteStep)
				{
					drawnSites.get(i).set(true);
				}
			}
			drawnSites.get(drawnSites.size()-1).set(true); // always draw the last site
		}
		catch (Throwable t) { t.printStackTrace(); }
		finally
		{
			siteNumberingLock.release();
		}
	}
	
	//////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////// site helper methods /////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////
	
	private Point getSiteXY(int siteIndex)
	{
		// calculate site motor positions
		Point4D siteMotorPos = new Point4D(start.x,start.y,start.z,omega);
		Point3D unitVector = run.getUnitVector(start,end);
		siteMotorPos.x += siteIndex * run.getStepSize_mm() * unitVector.x;
		siteMotorPos.y += siteIndex * run.getStepSize_mm() * unitVector.y;
		siteMotorPos.z += siteIndex * run.getStepSize_mm() * unitVector.z;

		return motorMMtoScreenIntPx(siteMotorPos);
	}
	

	/////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////// constructor and drawOverlays override ///////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	
	public WdgVectorCamera(Composite parent, int style, String urlString, CameraType camType)
	{
		super(parent, style, urlString, camType, CameraSize.LARGE);
		
		connectMonitors();
	}
	
	private void connectMonitors()
	{
		///////////////////////////////////////////////////////////////////////////////
		/////////////////////////////// recalc on resize //////////////////////////////
		///////////////////////////////////////////////////////////////////////////////
		
		ControlListener controlListener = new ControlListener()
		{
			@Override public void controlResized(ControlEvent arg0)
			{
				if (null == run || run.getScanMode()!=ScanMode.VECTOR) return;
				update(); // temporary fix because the camera ratio is updated only in draw()
				recalcSiteVisibility();
			}
			@Override public void controlMoved(ControlEvent arg0) {}
		};
		addControlListener(controlListener);

		///////////////////////////////////////////////////////////////////////////////
		/////////////////////////////// recalc on run change //////////////////////////
		///////////////////////////////////////////////////////////////////////////////
		
		currentRunListener = new CurrentRunListener()
		{
			@Override public void currentRunChanged(ModelRun run)
			{
				recalcSiteVisibility();
			}
		};
		ModelCollectRunList.get().addCurrentRunListener(currentRunListener);
		
		///////////////////////////////////////////////////////////////////////////////
		/////////////////////////// recalc on move start or stop //////////////////////
		///////////////////////////////////////////////////////////////////////////////
		
		positionListener = new ListenerList.Listener<DevMove>()
		{
			@Override public void handleEvent(DevMove object, Object data)
			{
				recalcSiteVisibility();
			}
		};
		for (String motorName : motorsToWatch)
		{
			DevPmacMotor.get(motorName).llMoveStartStop.addListener(positionListener);
		}

		///////////////////////////////////////////////////////////////////////////////
		////////////////////////////////// dispose listener ///////////////////////////
		///////////////////////////////////////////////////////////////////////////////
		
		addDisposeListener(new DisposeListener()
		{
			@Override public void widgetDisposed(DisposeEvent arg0)
			{
				removeMonitors();
			}
		});
	}
	
	private void removeMonitors()
	{
		for (String motorName : motorsToWatch)
		{
			DevPmacMotor.get(motorName).llMoveStartStop.removeListener(positionListener);
		}
		ModelCollectRunList.get().removeCurrentRunListener(currentRunListener);
	}
}
