package gov.anl.gmca.common;

import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.Epics.EpicsMonitor;
import gov.anl.gmca.io.Epics.EpicsValueMonitor;
import gov.anl.gmca.io.PVLookup.PVName;
import gov.aps.jca.dbr.DBR;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.concurrent.Semaphore;

/**
 * This device ensures a notification is sent when all PVs are connected,
 * and once they are connected then whenever the update PV changes.
 * 
 * For now, setUpdatePV must be called, and it must be called after all registerPV calls.
 */
public abstract class DevEpicsUpdate
{
	protected Map<PVName,Boolean> pvState = new HashMap<PVName,Boolean>();
	protected boolean allConnected = false;
	protected boolean firstUpdateSent = false;
	protected Semaphore firstUpdateSentLock = new Semaphore(1);
	protected PVName updatePV = null;
	
	public boolean connected() { return firstUpdateSent; }
	
	protected Epics.EpicsMonitor updateMonitor;
	protected HashMap<PVName, EpicsMonitor> valueMonitorMap = new HashMap<PVName, EpicsMonitor>();
	
	// collect values from PVs and update the model
	protected abstract void sendNotification();

	protected synchronized void registerPV(PVName pvName)
	{
		synchronized(pvState) { pvState.put(pvName,false); }
	}
	
	protected void setUpdatePV(PVName pvName)
	{
		updatePV = pvName;
		synchronized(pvState) { if (pvState.containsKey(pvName)) pvState.remove(pvName); }
	}
	
	protected void reportConnected(PVName pvName)
	{
		synchronized(pvState)
		{
			pvState.put(pvName, true);
			boolean _allConnected = true;
			for (PVName key : pvState.keySet())
			{
				if (!pvState.get(key)) _allConnected = false;
			}
			allConnected=_allConnected;
		}
	}

	// connect all the PVs and set their monitors to call sendNotification() from whoever connects last
	// additionally, once that first notification is sent, the update PV should send notifications whenever it changes
	protected void registrationComplete()
	{
		if (null == updatePV) throw new IllegalArgumentException("No update PV specified");
		
		updateMonitor = new EpicsValueMonitor() { @Override public void valueChanged(String pvName, DBR value)
		{
			reportConnected(updatePV);

			if (allConnected)
			{
    			firstUpdateSentLock.acquireUninterruptibly();
    			try
    			{
    				if (!firstUpdateSent || Epics.dbrToInt(value)==1)
    				{
    					sendNotification();
            			firstUpdateSent=true;
    				}
    			}
    			catch (Exception e)
    			{
    				e.printStackTrace();
    			}
    			finally
    			{
    				firstUpdateSentLock.release();
    			}
			}
		}};
		Epics.addMonitor(updatePV, updateMonitor);
		if (Epics.isConnected(updatePV)) reportConnected(updatePV);

		synchronized(pvState)
		{
			for (final PVName pvName : new HashSet<PVName>(pvState.keySet()))
			{
				EpicsValueMonitor valueMonitor = new EpicsValueMonitor() { @Override public void valueChanged(String _pvName, DBR value)
				{
					reportConnected(pvName);
					
					firstUpdateSentLock.acquireUninterruptibly();
					try
					{
		    			if (allConnected && !firstUpdateSent)
		    			{
		    				sendNotification();
		    				firstUpdateSent=true;
		    			}
					}
					catch (Exception e) { e.printStackTrace(); }
					finally
					{
						firstUpdateSentLock.release();
					}
				}};
	
				Epics.addMonitor(pvName,valueMonitor);
				valueMonitorMap.put(pvName, valueMonitor);
				if (Epics.isConnected(pvName)) reportConnected(pvName);
			}
		}
	}
	
	public void stopEpicsMonitors()
	{
		synchronized(pvState)
		{
			for (PVName pvName : pvState.keySet())
			{
				Epics.removeMonitor(pvName, valueMonitorMap.get(pvName));
			}
			Epics.removeMonitor(updatePV, updateMonitor);
		}
	}
	
	public void sendUpdate()
	{
		Epics.putInteger(updatePV, 0);
		Epics.putInteger(updatePV, 1);
	}
}
