package gov.anl.gmca.common;

import gov.anl.gmca.hutch.DevShutter;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.PVName;
import gov.anl.gmca.syslog.Log;
import gov.aps.jca.dbr.DBR;

import java.util.HashMap;
import java.util.Map.Entry;

/**
 * This device handles registering which PVs and other devices this one depends on.
 * Then, when you call registrationComplete, it connects the PVs and creates the dependent devices,
 * and sets the state to READY when it's done.
 * 
 */
public abstract class EpicsDevice
{
	///////////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////// interface //////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////////////
	
	abstract protected void onAllPVsConnected();
	abstract protected void valueChangedAction(String pvName, DBR value);
	abstract public void abort();
	
	///////////////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////// state ////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////////////
	
	public enum State
	{
		OFFLINE, READY, MOVING, ERROR;
		public static State fromString(String i) {
			for (State s : State.values()) {
				if (s.toString().toLowerCase().equals(i.toLowerCase())) return s;
			}
			return null;
		}
	};
	
	protected State state = State.OFFLINE;
	public State	getState()							{ return state; }
	public boolean	isOnline()							{ return (state == State.READY || state == State.MOVING); }
	public void 	setState(State state) 				{ this.state = state; }
	
	///////////////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////// abort ////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////////////
	
	public static void abortAll()
	{
		DevShutter shutter = DevShutter.get("shutter");
		if (null != shutter)
		{
			Log.note("Closing shutter");
			shutter.close();
		}
		else
		{
			Log.warn("Shutter device not found");
		}
		
		Log.note("Aborting all devices.");
		for (EpicsDevice dev : devMap.values())
		{
			dev.abort();
		}
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////// ctor ////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////////////
	
	protected boolean deviceAlreadyExists = false;
	
	protected String motorID;
	public String getDeviceName() { return motorID; }
	
	public EpicsDevice(String motorID)
	{
		if (null != devMap.get(motorID))
		{
			Log.error("Device "+motorID+" already exists");
			deviceAlreadyExists=true;
			return;
		}
		
		System.out.println("Creating device: "+motorID);
		this.motorID = motorID;
		devMap.put(motorID, this);
	}
	
	public static EpicsDevice getDevice(String devName)
	{
		return devMap.get(devName);
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////// PV reg ///////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////////////

	protected static HashMap<String,EpicsDevice> devMap = new HashMap<String,EpicsDevice>(); 

	// List of PVs this device needs in order to consider itself "online"
	protected HashMap<PVName,Boolean> pvList = new HashMap<PVName,Boolean>(); 
	
	protected void registerDevicePV(PVName pvName)
	{
		pvList.put(pvName,false);
	}

	protected void registrationComplete()
	{
		// connect PVs and watch for when they all connect
		// also, forward valueChanged events
		for (final Entry<PVName, Boolean> e : pvList.entrySet())
		{
			Epics.addMonitor(e.getKey(), new Epics.EpicsValueMonitor()
			{
				@Override
				public void valueChanged(String _pvName, DBR value)
				{
					boolean isAlreadyConnected = pvList.get(e.getKey());
					
					if (!isAlreadyConnected)
					{
    					pvList.put(e.getKey(), true);
    					
    					if (State.READY != state)
    					{
    						// did all PVs get connected?
    						boolean allConnected = true;
    						int numConnected=0;
    						
    						for (Entry<PVName, Boolean> e : pvList.entrySet())
    						{
    							if (!e.getValue())
    							{
    								//System.out.println(e.getKey()+" not connected");
    								allConnected = false;
    							}
    							else
    							{
    								numConnected++;
    							}
    						}
    						
    						System.out.println("dev "+getDeviceName()+" connected "+numConnected+" of "+pvList.entrySet().size()+": "+_pvName);
    						
    						if (allConnected && State.READY != state)
    						{
    							System.out.println("Device connected: "+getDeviceName());
    							
    							setState(State.READY);
    							onAllPVsConnected();
    						}
    					}
					}
					
					valueChangedAction(_pvName, value);
				}
			});
		}
	}
}
