package gov.anl.gmca.common;

import gov.anl.gmca.gui.Main;
import gov.anl.gmca.io.Disk;
import gov.anl.gmca.io.ExecTask;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.users.ModelUserTable.Client;
import gov.anl.gmca.users.ModelUserTable.ClientListener;
import gov.anl.gmca.util.PerfTimer;
import gov.anl.gmca.util.Util;

import java.util.List;
import java.util.concurrent.atomic.AtomicBoolean;

public class SafeImageDelete
{
	private static final String scriptPath = Main.getAppDir()+"/tools/scripts/safe_image_delete.py";
	public static final String sshOptions = "-o StrictHostKeyChecking=no -o PubkeyAuthentication=no -o NumberOfPasswordPrompts=0";
	public static String ppuHost;
	private static boolean enablePPUDelete = false;
	
	public static class FileToDelete implements Comparable<FileToDelete>
	{
		public String dir, filename;
		public long modTime;
		public FileToDelete(String dir, String filename, long modTime) { this.dir=dir; this.filename=filename; this.modTime=modTime; }
		public String getFullPath() { return dir+"/"+filename; }
		
		@Override public int compareTo(FileToDelete ftd)
		{
			return getFullPath().compareTo(ftd.getFullPath());
		}
	}

	public static void init()
	{
		ClientListener startupListener = new ClientListener()
		{
			@Override public void masterChanged(boolean isMaster) 
			{
				if (isMaster) { startupChecks(); Client.removeListener(this); }
			}
		};
		Client.addListener(startupListener);
	}
	
	private static void startupChecks()
	{
		new Thread(new Runnable(){public void run()
		{
	        if (!DBConfig.getBoolean("ppu_clear_ramdisk")) return;
			ppuHost = DBConfig.get("host_ppu");
	        if (ppuHost.trim().length()==0) return;
	        
	        ExecTask task = ExecTask.new_ExecTask("ssh "+sshOptions+" "+ppuHost+" "+scriptPath+" check","safe_image_delete", false, false, true);
	        task.start();
	        task.waitFor();
	        for (String line : task.stdout)
	        {
	        	if (line.contains("check OK"))
	        	{
	        		System.out.println("safe_image_delete script check OK");
	        		enablePPUDelete=true;
	        		return;
	        	}
	        }
	        Util.popupError("Script error", "Error calling safe_image_delete script, images\nwill not be deleted from the PPU RAM disk");
		}},"safe_image_delete startup check").start();
	}
	
	public static void delete(List<FileToDelete> files, boolean deleteOnGFS, boolean deleteOnPPU)
	{
		AtomicBoolean gfsDone = new AtomicBoolean(false);
		AtomicBoolean ppuDone = new AtomicBoolean(false);
		
		if (deleteOnGFS)
		{
			new Thread(new Runnable(){public void run()
			{
				Log.note("Deleting files on GFS...");
				PerfTimer pt = new PerfTimer();
				String[] pathsToDelete = new String[files.size()];
				for (int i=0; i<files.size(); i++)
				{
					pathsToDelete[i]=files.get(i).getFullPath();
				}
				Disk.delete(pathsToDelete);
				Log.note("GFS file deletion complete in "+Util.printDouble(pt.getElapsedMs()/1000.0,2)+" seconds");
				gfsDone.set(true);
			}},"GFS delete").start();
		}
		else
		{
			gfsDone.set(true);
		}
		
		if (deleteOnPPU && enablePPUDelete && files.size()>0)
		{
			new Thread(new Runnable(){public void run()
			{
				Log.note("Deleting files on PPU RAM disk...");
				PerfTimer pt = new PerfTimer();
				
		        String dir = files.get(0).dir;
				StringBuilder fileList = new StringBuilder();
		        int fileCount=0;
				
				for (SafeImageDelete.FileToDelete file : files)
				{
					if (!file.dir.equals(dir) || fileCount>2000)
					{
						SafeImageDelete.runCmd(ppuHost, fileCount, dir, fileList.toString());
						fileCount=0;
						dir=file.dir;
						fileList=new StringBuilder();
					}
					
					if (ValidateDirPrefix.isValidFullFilename(file.filename))
					{
						fileList.append(file.filename);
						fileList.append(" ");
						fileList.append(file.modTime);
						fileList.append(" ");
						fileCount++;
					}
					else
					{
						System.out.println("ERROR: Skipping deletion of image "+file.filename+" because it failed validation");
					}
				}
				SafeImageDelete.runCmd(ppuHost, fileCount, dir, fileList.toString());
				Log.note("PPU file deletion complete in "+Util.printDouble(pt.getElapsedMs()/1000.0,2)+" seconds");
				ppuDone.set(true);
			}},"PPU delete").start();
		}
		else
		{
			ppuDone.set(true);
		}
		
		while (!gfsDone.get() || !ppuDone.get()) Util.sleep(50);
	}
	
	private static boolean runCmd(String ppuHost, int fileCount, String dir, String fileList)
	{
		String cmd = "ssh "+sshOptions+" "+ppuHost+" "+scriptPath+" "+fileCount+" "+dir+" "+fileList;
        ExecTask task = ExecTask.new_ExecTask(cmd, "safe_image_delete", false, false, true);
        task.start();
        task.waitFor();
        
        for (String line : task.stdout)
        {
        	if (line.indexOf("Done.")>=0) return true;
        }
        
        Log.warn("Deletion of files on the PPU failed");
        return false;
	}
}
