/**
 * @author spothineni
 * This class prepares run_XDS.csh script and submits to Grid Engine through DRMAA.
 * It also write .xdsparams file with parameters from Settings tab into
 * data processing directory
 * 
 */


package gov.anl.gmca.dataproc;

import gov.anl.gmca.collect.ModelCollectRunList;
import gov.anl.gmca.common.Command;
import gov.anl.gmca.common.DevDataDir;
import gov.anl.gmca.diffimage.MarImageHeader;
import gov.anl.gmca.gridengine.CheckSubmitHost;
import gov.anl.gmca.gridengine.DRMAA;
import gov.anl.gmca.io.Disk;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.ExecTask;
import gov.anl.gmca.io.PVLookup.PV_SCR;
import gov.anl.gmca.multistrategy.DevMultiStrategy;
import gov.anl.gmca.syslog.Log;
import org.ggf.drmaa.DrmaaException;

import java.io.*;
import java.util.ArrayList;

public class CmdDataProc extends Command {

	public String workDir = "";
	public String sampleName = "";
	public String images = "";
	public String firstImage = "";
	public String imageSet = "1-10";
	public String firstImgNo = "1";
	public String lastImgNo = "2";
	public String imageDir = "";
	public String collectType = "Standard";
	public String overlap = "0";

	public String log = "";
	public static String tempLogDir = ".JBluIce";
	
	public boolean software = true;
	public boolean inverseOn = false;
	public boolean dontdisplay = false;
	public boolean second_pipeline_update = false;
	public boolean interface_update = false;
	protected int sampNumber;
	
	public static boolean THREAD_RUNNING = false;
	public static String previousImageSet = "";
	
	public ExecTask task;
	public DRMAA drmaa;
	protected static MarImageHeader currentHeader = null;
	private ModelDataProcParams dataprocResult;
	private DevDataProc devDataProc = new DevDataProc();

	public void setSampleName(String _sampleName) {
		sampleName = _sampleName;
	}

	@Override
	public void run() {

		THREAD_RUNNING = true;
		
		/* prepare a Dataprocparams which contains all the parameters from input to output
		 */
		dataprocResult = new ModelDataProcParams();
		ModelAimlessSummary modelAimlessSummary = new ModelAimlessSummary();

		dataprocResult.workDir = this.workDir;
		dataprocResult.imageDir = this.imageDir;
		dataprocResult.collectType = this.collectType;
		dataprocResult.sampleName = this.sampleName;
		dataprocResult.images = this.images;
		dataprocResult.state = "PROCESSING";
		dataprocResult.imageSet = this.imageSet;
		dataprocResult.inverseOn = this.inverseOn;
		dataprocResult.overlap = this.overlap;
		dataprocResult.firstImage = this.firstImage;
		dataprocResult.log = this.log;
		dataprocResult.software = this.software;
		
		int overlap_proc = ModelCollectRunList.get().getCurrentRunModel()
				.getOverlap();
		
		/* this scalingSoftware parameter can be removed as processing will use only AIMLESS */
		int scalingSoftware = Epics.getInteger(PV_SCR.scalingSoftware);
		
		dataprocResult.scalingsoftware = "AIMLESS";
		if (this.imageSet.contains("Final_")) {
			devDataProc.setResult(sampNumber, dataprocResult, true);
		} else {
			devDataProc.setResult(sampNumber, dataprocResult, this.interface_update && this.second_pipeline_update);
		}
		/* TODO display the warning about current processing image set*/
		
		if(!this.inverseOn) {
			dataprocResult.warning = "\nProcessing image set: "+ this.imageSet;
			devDataProc.updateWarning(sampNumber, dataprocResult, previousImageSet, true && this.second_pipeline_update);
		}
		/* Check the working directory exists or not 
		 */
		File dirFile = new File(this.workDir);
		if (!dirFile.exists()) {

			if (!dirFile.mkdirs()) {
				THREAD_RUNNING = false;
				Log.error("Couldn't create directory " + workDir);
				return;
			}
		}
		Log.note("XDS data reduction in progress, check the dir. " + workDir);

		/* write user changed XDS parameters (from Settings tab) into a separate file *
		 * .xdsparams
		 * run_XDS.csh script will read these parameters and writes inot XDS.INP
		 */
		if(ModelXDSSettings.takeXDSSettings && ModelXDSSettings.changedValues.size() > 0) {
			try {
				BufferedWriter runxdsOut = new BufferedWriter(new FileWriter(
					this.workDir + "/" + ".xdsparams", false));
				
				runxdsOut.write("\n!XDS parameters from Settings tab of JBluIce\n");
				for(ModelXDSSettings.DefaultParam param : ModelXDSSettings.changedValues.keySet()) {
					String val = ModelXDSSettings.changedValues.get(param);
					if(val == null || val.equals(param.getValue()))
						continue;
					if(param.getXDSParam().equals("FRIEDELS_LAW")) {
						runxdsOut.write("FRIEDEL'S_LAW"+"="+ModelXDSSettings.changedValues.get(param)+"\n");
						continue;
					}
					if(param.getXDSParam().equals("SPACE_GROUP_NUMBER")) {
						BufferedWriter runxds = new BufferedWriter(new FileWriter(
								workDir + "/" + "run_XDS.sh", true));
						
						String sp = ModelXDSSettings.changedValues.get(param);
						
						if(sp.equals(" ") || sp.equals("")) {
							continue;
						}
						try {
							sp = DevMultiStrategy.get().getSpacegroupFromNumber(Integer.parseInt(sp));
							String uc = ModelXDSSettings.changedValues.get(ModelXDSSettings.DefaultParam.UNIT_CELL_CONSTANTS);
						
							if(uc!=null && !uc.equals("")){
								runxds.append(' ');
								runxds.append(uc.replace(" ", ","));
								runxds.append(' ');
								runxds.append(sp);
							}
							runxds.close();
						} catch (NumberFormatException e) {
							e.printStackTrace();
						}
					}
					if(!val.equals("")) {
						runxdsOut.write(param.getXDSParam()+"="+ModelXDSSettings.changedValues.get(param)+"\n");
					}
				}
				runxdsOut.close();
			} catch (IOException e1) {
				e1.printStackTrace();
			}
		}
		
		/*check for gridengine environment and quit if not a submit host */
		boolean isSubmitHost = CheckSubmitHost.get().isSubmitHost();
		
		System.out.println("Gridengine env detected: "+isSubmitHost);
		if(!isSubmitHost) {
			quitWithException(new Exception());
			return;
		}
		
		/* submit the processing job to Grid Engine using DRMAA 
		 * waits 900 seconds before it deletes the job
		 * 
		 */
		
		if (!isSubmitHost) {
		//if (true) { // force local since sge does not run	
			task = ExecTask.new_ExecTask(workDir + "/" + "run_XDS.sh", "dataProcessing", false, false);
			task.start();
			task.waitFor();
		} else {
			try {
				/*
				 * wait 900 sec before quitting the job
				 * 
				 */
				drmaa = new DRMAA(workDir + "/" + "run_XDS.sh", workDir, WdgDataProc.TIMEWAIT_FOR_PROCESSINGJOB); 
				drmaa.start();
				int jobStatus = drmaa.waitFor();
	
			} catch (NoClassDefFoundError  e) {
				quitWithException(new Exception());
				return;
			} catch (UnsatisfiedLinkError e) {
				System.out.println(e.getMessage());
				quitWithException(new Exception());
				return;
			} catch (DrmaaException e) {
				quitWithException(e);
				return;
			} catch (Exception e) {
				quitWithException(e);
				return;
			} 
		}
		/* check if the status still says processing */
		if (dataprocResult.warning.contains("Processing") || dataprocResult.warning.contains("processing"))
			dataprocResult.warning = "DONE";
		if (dataprocResult.state.equals("PROCESSING"))
			dataprocResult.state = "DONE";

		/*dataprocResult.scalaLog = workDir+"/"+tempLogDir + "/"
				+ dataprocResult.sampleName + "_aimless.log";
			
		dataprocResult.subdir = workDir + "/" + dataprocResult.sampleName + "_"
				+ (dataprocResult.software?"fast_dp":"GMCAproc") + "_" + dataprocResult.imageSet + "/";*/
		
		/* read the subdirectory where the aimless log files is stored */
		BufferedReader aimlessbr;
		String aimlesslogFile = workDir  +"/"+tempLogDir + "/subdirfile";
		try {
			aimlessbr = new BufferedReader(new FileReader(aimlesslogFile));
			dataprocResult.subdir = aimlessbr.readLine();
			
		} catch (FileNotFoundException e) {
				System.out
				.println("imagesets file not found " + e.getMessage());
				dataprocResult.imageSet = this.imageSet;
		} catch (IOException e) {
				System.out
				.println("imagesets file not found " + e.getMessage());
				dataprocResult.imageSet = this.imageSet;
		}
		
		/* save the log files into parameters which are writeen to MySQL */
		dataprocResult.scalaLog = dataprocResult.subdir+ "/"+dataprocResult.sampleName + "_aimless.log";
		
		if(workDir.contains("fast_dp")) {
			dataprocResult.log = dataprocResult.subdir+"/fast_dp.log";
		} else {
			dataprocResult.log = dataprocResult.subdir+"/GMCAproc.LOG";
		}
		
		/* read the <prefix>_aimless log for summary of data quality params */
		try {
			modelAimlessSummary.processResult(workDir, dataprocResult);
		} catch (Exception e) {
				e.printStackTrace();
		}
		
		/* truncate log */
		dataprocResult.truncateLog = dataprocResult.subdir+"/"
				+ dataprocResult.sampleName + "_ctruncate.log";
		
		/* submitted image set may be different from the image set which
		 * run_XDS.csh processed, there may be new data available
		 * so re-read the imageset processed.
		 */
		dataprocResult.imageSet = "";
		String imageSetsFile = workDir  +"/"+tempLogDir + "/imagesets";
		BufferedReader br;
		try {
				br = new BufferedReader(new FileReader(imageSetsFile));

				ArrayList<String> lines = Disk.readText(imageSetsFile);
				
				if (this.imageSet.contains("Final_"))
					dataprocResult.imageSet += "Final_";
				
				dataprocResult.images = dataprocResult.images.substring(0,
						dataprocResult.images.indexOf(" "));
				
				if (lines.size() > 2) {
					for (String line :lines) {
						dataprocResult.imageSet += line + "_";
						dataprocResult.images += line + " ";
					}	
				} else if (lines.size() == 2){
					dataprocResult.imageSet += lines.get(0)+"_"+lines.get(1);
					dataprocResult.images += lines.get(0)+" "+lines.get(1);
				} else {
					dataprocResult.imageSet += lines.get(0);
					dataprocResult.images += " "+lines.get(0);
				}
				br.close();
		} catch (FileNotFoundException e) {
				System.out
						.println("imagesets file not found " + e.getMessage());
				dataprocResult.imageSet = this.imageSet;
		} catch (IOException e) {
				System.out
						.println("imagesets file not found " + e.getMessage());
				dataprocResult.imageSet = this.imageSet;
		}
		
		/* always fast_dp is default unless GMCAproc selected by user */
		if(ModelDataProcList.currentSoftwareSelection.equals("GMCAproc")) {
			this.second_pipeline_update = true;
		}
		devDataProc.setDataProcResult(sampNumber, dataprocResult,
				this.imageSet, this.second_pipeline_update);
		
		/* save the new result to Excel spreadsheet */
		String savePath = DevDataDir.getPrimary();
		devDataProc.saveAllResults(savePath + "/dataProcessingResults.xls");
		
		previousImageSet = dataprocResult.imageSet;
		THREAD_RUNNING = false;
		Log.note("Data processing with "+(dataprocResult.software?"fast_dp":"GMCAproc")+" finished");
	}
	
	private void quitWithException(Exception e) {
		THREAD_RUNNING = false;
		dataprocResult.warning = "GridEngine/DRMAA exception for submitting processing job\n report this to your host";
		dataprocResult.state = "ERROR";
		e.printStackTrace();
		devDataProc.setDataProcResult(sampNumber, dataprocResult,
				this.imageSet, true && this.second_pipeline_update);
		
		return;
	}
}
