/**
 * @author spothineni, mhilgart
 * 
 * 
 * 'dataproc' package takes care of automated data processing in parallel to data collection.
 * It handles two pipelines fast_dp developed at Diamond Light Source, GMCAproc is inhouse
 * development. Both pipelines runs XDS,POINTLESS,AIMLESS,CTRUNCATE in that order.
 * 
 * data collection jobs are submitted to grid engine which in turn run csh scripts from
 *`/mnt/software/bluice/tools/scripts/dataproc` folder, main script is run_XDS.csh
 *
 * Results are displayed on Analysis tab, multiple results can be browsered using <<, <, >, >>
 * buttons. All the results are exported to spreadsheet using Export button.
 * 
 * Data processing threads are initiated from the gov.anl.gmca.log.TabLog. When ever collect run
 * status (ex: 23i:bi:r1:runStatus) changed to 'collecting' a thread is created in startProcessingThread
 * function of TabLog.
 * 
 * Data processing job threads are created in InitDataProcessing one by one by looking at the call backs
 * from PV 23i:FPE:lastfilename, monitor for this pv is in TabLog.
 * 
 * CmdDataProc will prepare a script run_XDS.csh in a sub directory of the data and submits 
 * to Grid Engine using DRMAA. After job completion ModelAimlessSummary parses aimless log and updates
 * MySQL table for persistence of the data. It also created plots using gnuplot and saves as png
 * files which are displayed on Analysis tab.
 * 
 * Multiple sub-tabs are available, which are initiated from TabAnalysis.java, actual sub-tabs 
 * are initiated from WdgDataProc
 * 
 * XDSSettings sub-tab user can select already known unitcell,
 * spacegroup or if any parameters different from defaults can be set. 
 * 
 * 			GUI: WdgXDSSettins
 * 			Model: ModelXDSSettings
 * 			MySQL interface: DevXDSSettings
 * 
 *
 * Data quality sub-tab (WdgDataProc) displayed the summary from Analysis tab.
 * 
 * 			GUI: WdgDataProc
 * 			Model: ModelDataproc.java, ModelDataProcList
 * 			MySQL interface: DevDataProc
 * 
 * Reprocessing can be done on existing results on the Reprocessing tab. User can import
 * unit cell, spacegroup from Reprocessing tab. XDSGUI also can open by clicking on corresponding 
 * button. When clicking on the reprocess button CmdReporcessing prepare a script and submit the job
 * to Grid Engine using DRMAA.
 * 
 * 			GUI: WdgReporcessing, WdgUnitCell
 * 			Model: ModelReprocessing
 * 			MySQL interface: DevReprocessing
 * 
 * When clicked on AIMLESS log, CTRUNCATE log buttons it will run qtrview (ccp4 program) using CmdLogView
 * 
 * 
 */

package gov.anl.gmca.dataproc;


/*
 Automated Data processing at GMCA@APS, Sector 23 beamlines

Data are processed as soon as collected using two software pipelines fast_dp and GMCAproc, which are wrappers around XDS and generate intensities/structure factor amplitudes in mtz format.

Processing results are in the sub directory of data, <prefix>_fast_dp, <prefix>_GMCAproc and for manual processing <prefix>_XDS contains input file of XDS: XDS.INP, by changing current directory to <prefix>_XDS and using xds_par command one can process data manually.

In <prefix>_fast_dp, <prefix>_GMCAproc there are subdirectories for different sets of images, sub directories named with frmae numbers at the end, in case of 1800 frames collected final results are in <prefix>_GMCAproc_Final_1-1800, there are other subsets give more information while colleting data.

For example first frame is F1_1_00001.cbf, 1800 frames collected, in this case F1_1 is the prefix, directory structure is like following,

F1_1_fast_dp

    F1_1_fast_dp_Final_1-1800
    |
    ---> all the XDS files are here

F1_1_GMCAproc

    F1_1_GMCAproc_Final_1-1800
         XDS
	 |
	 ---> all the XDS files are here for pass JOB=ALL
	 XDS_REINTEGRATE
	 |
	 ---> all the XDS reintegration files are here for pass JOB= DEFPIX INTEGRATE CORRECT
	 POINTLESS_AIMLESS
	 |
	 ---> POINTLESS, AIMLESS, CTRUNCATE logs are here, check the summary information of data quality at the end of <prefix>_aimless.log

Below is breif description of algorithm of software pipelines

fast_dp

1. Index the data from three wedges of ~ 5 degrees around the start, start + 45, start + 90 or as close as can be reached, take the P1 solution.
2. Integrate the data in P1 using the cluster.
3. Postrefine / scale in P1 using CORRECT - record results of post refinement.
4. take output of this into pointless, compare with post refinement results, select lattice, point group.
5. Estimate resolution limit where I/sigma ~ 1 for unmerged measurements.
6. Repeat post refinement / scaling with these settings.
7. Merge data with AIMLESS. (scaling is off with keyword "scales constant", since scaling already done in CORRECT step)

GMCAproc

1. Run XDS with JOB=ALL or JOB= XYCORR INIT COLSPOT IDXREF DEFPIX INTEGRATE CORRECT(Results are in XDS direcotry).
2. Obtain resolution cutoff from CORRECT.LP for I/Sigma ~ 1, lets say it as $resol
3. Copy GXPARM.XDS into XDS_REINTEGRATE direcotry. Run XDS with JOB= XYCORR INIT DEFPIX INTEGRATE CORRECT & INCLUDE_RESOLUTION_RANGE=30.0 $resol (Results are in XDS_REINTEGRATE directory).
4. Take XDS_ASCII.HKL from step 3 and run POINTLESS which gives <prefix>_xds.mtz, see pointless.sh (Results are in POINTLESS_AIMLESS directory)
5. Run AIMLESS on mtz file from step 4 will gives <prefix>_aimless.mtz, scaling is off with keyword "scales constant", since scaling already done in CORRECT step, ANOMOLUS ON, see aimless.sh (Results are in POINTLESS_AIMLESS directory)
6. Run ctruncate using ctruncate.sh, mtz from step 5 will give <prefix>_ctruncate.mtz

Either <prefix>_aimless.mtz or <prefix>_ctruncate.mtz can be used for structure solution with PHENIX or CCP4

In-case of inverse beam, vector data collection with overlaps step 1 to 6 performed in a separate sub-directories, observe directory names with frame numbers. Orientation matrix is preserved between the datasets by using first dataset as REFERENCE_DATA_SET. Multiple datasets will be merged in <prefix>_COMBINED directory.

1. POINTLESS_AIMLESS directory: multiple datasets combined with POINTLESS, see poinetless.sh and merged and scaled together with AIMLESS. either <prefix>_aimless.mtz or <prefix>_ctruncate.mtz can be used for structure solution with PHENIX or CCP4
2. XSCALE directory: This step is separate from step 1 gives different perspective of data merged and scaled together using XSCALE.
Reference: "Tightly integrated single- and multi-crystal data collection strategy calculation and parallelized data processing in JBluIce beamline control system." Pothineni S. B., Venugopalan N., Ogata C. M., Hilgart M. C., Stepanov S., Sanishvili R., Becker M., Winter G., Sauter N. K., Smith J. L. and Fischetti R. F., J. Appl. Cryst. 47 (6), 1992-1999 (2014). 
*/