package gov.anl.gmca.diffimage;

import gov.anl.gmca.util.Util;
import hdf.hdf5lib.H5;
import hdf.hdf5lib.HDF5Constants;

public class HDF5Image extends DiffractionImage32Bit
{
    public HDF5Image(String masterImagePath, int imagesPerFile, int imageIndex) throws Exception
    {
    	// get the file prefix
    	int masterIndex = masterImagePath.indexOf("master.h5");
    	if (masterIndex <= 0 || masterIndex != masterImagePath.length()-"master.h5".length())
    		throw new IllegalArgumentException("Invalid master HDF5 image name: "+masterImagePath);
    	String filePrefix = masterImagePath.substring(0, masterIndex);
    	
    	// open the data file
    	int datafileNumber = imageIndex/imagesPerFile + 1;
    	String dataFile = Util.sprintf("%sdata_%06d.h5", filePrefix, datafileNumber);
    	System.out.println("Loading image "+imageIndex+" from file "+dataFile);
        long file_id = H5.H5Fopen(masterImagePath, HDF5Constants.H5F_ACC_RDONLY, HDF5Constants.H5P_DEFAULT);
        if (file_id == 0) throw new Exception("error opening image "+masterImagePath);

        // open the dataset
        String dataset_name = Util.sprintf("/entry/data/data_%06d", datafileNumber);
        long dataset_id = H5.H5Dopen(file_id, dataset_name, HDF5Constants.H5P_DEFAULT);
        if (dataset_id == 0) throw new Exception("error opening dataset");
        
        // load the dataspace
        long dataspace_id = H5.H5Dget_space(dataset_id);
        if (dataspace_id == 0) throw new Exception ("error getting dataspace");

        // get dataspace dimensions
        long[] dims = {0,0,0};
        H5.H5Sget_simple_extent_dims(dataspace_id, dims, null);
        originalImgWidth=(int)dims[1];
        originalImgHeight=(int)dims[2];

        // select first image
        int file_status = H5.H5Sselect_hyperslab
        (
        	dataspace_id,
        	HDF5Constants.H5S_SELECT_SET,	// define this slab with one call, no combinations
        	new long[]{0,0,0},				// offsets - first number is the image number
        	null,							// stride
        	new long[]{1,dims[1],dims[2]},	// element count for each dimension
        	null							// block
        );
        
        // create memory space
        long memory_space_id = H5.H5Screate_simple(2, new long[]{originalImgWidth,originalImgHeight}, null);
        
        // select memory hyperslab
        int memory_status = H5.H5Sselect_hyperslab
        (
        	memory_space_id,
        	HDF5Constants.H5S_SELECT_SET,
        	new long[]{0,0},
        	null,
        	new long[]{dims[1],dims[2]},
        	null
        );
        
        // read image data
        originalImg = new int[originalImgWidth*originalImgHeight];
        H5.H5Dread
        (
                dataset_id,
                HDF5Constants.H5T_STD_U32LE,	// memory type
                memory_space_id,				// memory space
                dataspace_id,					// file space
                HDF5Constants.H5P_DEFAULT,		// transfer properties
                originalImg
        );
        
        detector 						="EIGER";//getStringDataset(file_id, "/entry/instrument/detector", "description");

        // TODO: verify units
        //beam_center_mm_from_left	=getFloatDataset(file_id, "/entry/instrument/detector", "beam_center_x");
        //beam_center_mm_from_top	=getFloatDataset(file_id, "/entry/instrument/detector", "beam_center_y");
        distance_mm						=getFloatDataset(file_id, "/entry/instrument/detector", "detector_distance")*1000.0;
        exposureTime_sec				=getFloatDataset(file_id, "/entry/instrument/detector", "count_time");
        mmPerPixel						=getFloatDataset(file_id, "/entry/instrument/detector", "x_pixel_size")*1000.0;
        wavelength_A					=getFloatDataset(file_id, "/entry/instrument/beam",     "incident_wavelength");
        
        Util.printf("Detector description: %s", detector);
        Util.printf("displayOriginX: %.2f px displayOriginY: %.2f px", beam_center_mm_from_left, beam_center_mm_from_top);
        Util.printf("distance: %.2f mm", distance_mm);
        Util.printf("exposureTime: %.2f sec", exposureTime_sec);
        Util.printf("mmPerPixel: %.2f mm", mmPerPixel);
        Util.printf("wavelength: %.2f A", wavelength_A);
        
        //Log.note("Loaded image "+imageIndex);
        
        // close
        H5.H5Sclose(dataspace_id);
        H5.H5Dclose(dataset_id);
        H5.H5Fclose(file_id);
    }
    
    public static String getStringDataset(long file_id, String dataset_name, String attributeName) throws Exception
    {
        long dataset_id = H5.H5Dopen(file_id, dataset_name+"/"+attributeName, HDF5Constants.H5P_DEFAULT);
        if (dataset_id <= 0) throw new Exception("error opening dataset: "+attributeName);

        long type_id = H5.H5Dget_type(dataset_id);
        long space_id = H5.H5Dget_space(dataset_id);
        long size = H5.H5Tget_size(space_id);
        
        byte[] buf = new byte[(int)size];
        H5.H5Dread
        (
                dataset_id,
                type_id,						// memory type
                space_id,						// memory space
                space_id,						// file space
                HDF5Constants.H5P_DEFAULT,		// transfer properties
                buf
        );
        
        H5.H5Sclose(space_id);
        H5.H5Dclose(dataset_id);
        return new String(buf);
    }
    
    public static float getFloatDataset(long file_id, String dataset_name, String attributeName) throws Exception
    {
        long dataset_id = H5.H5Dopen(file_id, dataset_name+"/"+attributeName, HDF5Constants.H5P_DEFAULT);
        long dataspace_id = H5.H5Screate_simple(1, new long[]{1l}, null);
        
        float[] buf = new float[1];
        H5.H5Dread
        (
                dataset_id,
                HDF5Constants.H5T_IEEE_F32LE,	// memory type
                dataspace_id,					// memory space
                dataspace_id,					// file space
                HDF5Constants.H5P_DEFAULT,		// transfer properties
                buf
        );
        
        H5.H5Dclose(dataset_id);
        return buf[0];
    }
    
    public static long[] getDatasetDims(long file_id, String dataset_name, String attributeName) throws Exception
    {
    	System.out.println("Reading "+dataset_name+"/"+attributeName);
		long dataset_id = H5.H5Dopen(file_id, dataset_name+"/"+attributeName, HDF5Constants.H5P_DEFAULT);
		long dataspace_id = H5.H5Dget_space(dataset_id);

		// get dataspace dimensions
		long[] dims = { 0, 0, 0 };
		int dimensionCount = H5.H5Sget_simple_extent_dims(dataspace_id, dims, null);
		
		// close
		H5.H5Sclose(dataspace_id);
        H5.H5Dclose(dataset_id);
		
        // single element datasets return zero for the number of dimensions
        if (dimensionCount==0) return new long[]{1};

        return dims;
    }
}
