package gov.anl.gmca.diffimage;

import gov.anl.gmca.collect.Filenames;
import gov.anl.gmca.common.Units;
import gov.anl.gmca.io.sim.Disk;
import gov.anl.gmca.util.Util;
import hdf.hdf5lib.H5;
import hdf.hdf5lib.HDF5Constants;

import java.nio.file.FileSystems;
import java.nio.file.Path;
import java.nio.file.attribute.GroupPrincipal;
import java.nio.file.attribute.PosixFileAttributes;
import java.nio.file.attribute.UserPrincipal;
import java.util.regex.Matcher;

public class HDF5MasterFile
{
	private static final int MAX_RETRIES=3, RETRY_WAIT_MS=500;

	public static class HDF5Header
	{
		public String detector;
		public double beam_center_pixels_from_left_x;
		public double beam_center_pixels_from_top_y;
		public double distance_mm;
		public double exposureTime_sec;
		public double mmPerPixel;
		public double wavelength_A;
		public double delta_degr;
		public int imagesPerDataFile;
		public int totalImageCount;
		
		public double getWavelength_keV() { return Units.angstrom_to_keV(wavelength_A); }
	}
	private HDF5Header header = null;
	public synchronized HDF5Header getHeader() { if (null==header) readHeader(); return header; }
	
	public Path path;
	
	public String prefix = null;
	public Integer runLabel = null, firstFrameNumber = null;
	public Boolean isInverse = null, isSiteCheck = null;
	public Boolean isScreening = null;
	public Integer energyNumber = null, siteNumber = null;
	
	public HDF5MasterFile(String path)
	{
		this.path = FileSystems.getDefault().getPath(path);
		init();
	}
	
	public HDF5MasterFile(Path path)
	{
		this.path=path;
		init();
	}
	
	private void init()
	{
		String filename = path.getFileName().toString();
		Matcher matcher = Filenames.HDF5MasterFilePattern.matcher(filename);
		if (!matcher.matches() || matcher.groupCount()<5)
			throw new IllegalStateException("Invalid master file name: " + path.getFileName());

		prefix = matcher.group(1);
		runLabel = Integer.parseInt(matcher.group(2));
		firstFrameNumber = Integer.parseInt(matcher.group(3));
		isInverse = matcher.group(4).equals("INV_");
		isScreening = matcher.group(5).equals("scr_");
		isSiteCheck = matcher.group(6).equals("chk_");
		
		String energyOption = matcher.group(7);
		if (energyOption.length()>0)
		{
			energyNumber = Integer.parseInt(energyOption.substring(1,2));
		}
		
		String siteOption = matcher.group(8);
		if (siteOption.length()>0)
		{
			siteNumber = Integer.parseInt(siteOption.substring(1,siteOption.length()-1));
		}
	}
	
	public void readHeader()
	{
		for (int i=0;i<MAX_RETRIES;i++)
		{
			try
			{
				if (!Disk.exists(path.toString()))
				{
					System.out.println(path.toString()+" doesn't exist, waiting...");
					continue;
				}
				
				// open the data file
				long file_id = H5.H5Fopen(path.toString(), HDF5Constants.H5F_ACC_RDONLY, HDF5Constants.H5P_DEFAULT);
				
				// save dimensions
				header = new HDF5Header();
			    header.detector 						="eiger-16M";//HDF5Image.getStringDataset(file_id, "/entry/instrument/detector", "description");
				header.totalImageCount 					=(int)(HDF5Image.getDatasetDims (file_id, "/entry/sample/goniometer", "omega")[0]);
			    header.beam_center_pixels_from_left_x	=HDF5Image.getFloatDataset(file_id, "/entry/instrument/detector", "beam_center_x");
			    header.beam_center_pixels_from_top_y	=HDF5Image.getFloatDataset(file_id, "/entry/instrument/detector", "beam_center_y");
			    header.distance_mm						=HDF5Image.getFloatDataset(file_id, "/entry/instrument/detector", "detector_distance")*1000.0;
			    header.exposureTime_sec					=HDF5Image.getFloatDataset(file_id, "/entry/instrument/detector", "count_time");
			    header.mmPerPixel						=HDF5Image.getFloatDataset(file_id, "/entry/instrument/detector", "x_pixel_size")*1000.0;
			    header.wavelength_A						=HDF5Image.getFloatDataset(file_id, "/entry/instrument/beam",     "incident_wavelength");
			    header.delta_degr						=HDF5Image.getFloatDataset(file_id, "/entry/sample/goniometer",   "omega_increment");

			    System.out.println(path.getFileName().toString()+" has "+header.totalImageCount+" images");
			    
				// close
				H5.H5Fclose(file_id);
				
				return;
			}
			catch (Exception e)
			{
				System.out.println("Exception "+e.getMessage()+" for file "+path);
				e.printStackTrace();
				Util.sleep(RETRY_WAIT_MS);
			}
		}
	}

	private Integer imagesPerDataFile=null;
	public int getImagesPerDataFile()
	{
		if (null!=imagesPerDataFile) return imagesPerDataFile;
		
		for (int i=0;i<MAX_RETRIES;i++)
		{
			try
			{
				long file_id = H5.H5Fopen(path.toString(), HDF5Constants.H5F_ACC_RDONLY, HDF5Constants.H5P_DEFAULT);
				imagesPerDataFile = (int)(HDF5Image.getDatasetDims (file_id, "/entry/data", "data_000001")[0]);
				H5.H5Fclose(file_id);
			}
			catch (Exception e) { e.printStackTrace(); }
			Util.sleep(RETRY_WAIT_MS);
			System.out.println("Retrying to read first data file for "+path.toString());
		}
		
		return imagesPerDataFile;
	}

	
	///////////////////////////////// read file owner /////////////////////////////////////////
	
	public static class OwnerInfo
	{
		public UserPrincipal user;
		public GroupPrincipal group;
	}
	private OwnerInfo ownerInfo = null;
	private static Object ownerInfoLock = new Object();
	public OwnerInfo getOwnerInfo()
	{
		if (null!=ownerInfo) return ownerInfo;
		synchronized(ownerInfoLock)
		{
			PosixFileAttributes attributes = Disk.getPosixAttributes(path);
			ownerInfo = new OwnerInfo();
			ownerInfo.user = attributes.owner();
			ownerInfo.group = attributes.group();
			return ownerInfo;
		}
	}
	
	/////////////////////////////////////////////////////////////////////////////////////////////
	
	public Path get_BeeGFS_Path()
	{
		String ret = "/mnt/beegfs"+path.toString().substring("/ramdisk".length());
		return FileSystems.getDefault().getPath(ret);
	}
}
