package gov.anl.gmca.gui;

import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.Util;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.concurrent.ConcurrentHashMap;

public class GUIThreadMonitor implements Runnable
{
	protected final static int MESSAGE_TIMEOUT_SEC = 7;
	
	private static GUIThreadMonitor inst;
	protected volatile long currentGUITaskStart = 0;
	protected volatile boolean hangDetected = false;
	private Thread thread;
	private GUI gui;
	private volatile int messagesProcessedThisSec = 0;
	private volatile static int messagesQueuedThisSec = 0;
	private volatile long longestProcTimeThisSec = 0;
	public volatile static String currentTrace = null;
	private static ConcurrentHashMap<String,Integer> callCounts = new ConcurrentHashMap<String,Integer>();
	
	public static GUIThreadMonitor start(GUI gui)
	{
		if (null!=inst) { Log.error("GUI hang check thread already started"); return null; }
		inst = new GUIThreadMonitor(gui);
		
		inst.thread = new Thread(inst, "GUI Thread Monitor");
		inst.thread.start();
		
		return inst;
	}
	
	public GUIThreadMonitor(GUI gui)
	{
		this.gui=gui;
	}
	
	public static GUIThreadMonitor get() { return inst; }

	public static void handleRunnableStart()
	{
		if (null==inst) return;
		inst.currentGUITaskStart = System.currentTimeMillis();
	}
	
	public static void handleRunnableEnd()
	{
		if (null==inst) return;
		
		long thisWait = System.currentTimeMillis() - inst.currentGUITaskStart;
		if (thisWait>200)
		{
			System.out.println("procTime="+Util.printDouble(thisWait/1000.0,3)+" sec "+currentTrace);
		}
		
		inst.longestProcTimeThisSec = Math.max(inst.longestProcTimeThisSec,thisWait);
		inst.currentGUITaskStart = 0;
		inst.messagesProcessedThisSec++;
	}
	
	public static String logCall()
	{
		GUIThreadMonitor.guiThreadQueueSize++;
		GUIThreadMonitor.messagesQueuedThisSec++;
		
		String bluiceCall = Thread.currentThread().getStackTrace()[3].toString();
		
		Integer existingCount = callCounts.get(bluiceCall);
		if (null==existingCount) existingCount=0;
		callCounts.put(bluiceCall, existingCount+1);
		
		return bluiceCall;
	}

	// only called at the end of runnables queued by this application
	public static void handleAppRunnableEnd()
	{
		GUIThreadMonitor.guiThreadQueueSize--;
	}
	
	protected static volatile int guiThreadQueueSize = 0;
	
	@Override public void run()
	{
		// while the pump is working
		while (!gui.messagePumpExited)
		{
			// if there is a message being processed
			if (0 != currentGUITaskStart)
			{
				// how long has this message been waiting?
    			long now = System.currentTimeMillis();
    			int secSinceMsgStart = (int)((now-currentGUITaskStart)/1000.0);
    			
    			// too long?
    			if (secSinceMsgStart >= MESSAGE_TIMEOUT_SEC && !hangDetected)
    			{
    				hangDetected = true;
    				
    				System.out.println("GUI HANG DETECTED");
    				
    				Thread[] tArray = new Thread[100];
    				Thread.enumerate(tArray);
    				for (Thread thread : tArray)
    				{
    					if (null == thread) break;
    					System.out.println("["+thread.getName()+"]");
    					for (StackTraceElement ste : thread.getStackTrace())
    					{
    						System.out.println("\t"+ste.toString());
    					}
    					System.out.println();
    				}
    			}
			}

			// if there is GUI lag (>75% CPU usage by the GUI thread)
			if (longestProcTimeThisSec > 750 || messagesQueuedThisSec > 2000)
			{
				// print message count and queue size
				System.out.println("Processed "+messagesProcessedThisSec+" messages, queued "+messagesQueuedThisSec+" messages, max wait time "+Util.printDouble(longestProcTimeThisSec/1000.0, 2)+" sec, queue size = "+guiThreadQueueSize+", waiting on thread: "+GUIUtil.getTopBluIceMethodFromStack(GUI.getDisplay().getThread()));
				 
				// print sorted call counts
	            System.out.println();
	            ConcurrentHashMap<String,Integer> mapCopy = new ConcurrentHashMap<String,Integer>(callCounts);
	            for (String key : sortByValue(mapCopy))
				 {
		            System.out.println("["+callCounts.get(key)+"] Call: "+key);
		        }
			}
			
			messagesQueuedThisSec = 0;
			messagesProcessedThisSec = 0;
			longestProcTimeThisSec = 0;
			callCounts = new ConcurrentHashMap<String,Integer>();
			Util.sleep(1000);
		}
	}
	
	public static ArrayList<String> sortByValue(final ConcurrentHashMap<String,Integer> callCounts)
    {
        ArrayList<String> keys = new ArrayList<String>();
        keys.addAll(callCounts.keySet());
        Collections.sort(keys, new Comparator<Object>()
        {
            @SuppressWarnings("unchecked")
			public int compare(Object o1, Object o2)
            {
                Object v1 = callCounts.get(o1);
                Object v2 = callCounts.get(o2);
                return ((Comparable<Object>)v1).compareTo(v2);
            }
        });
        return keys;
    }
}
