package gov.anl.gmca.hutch;

import gov.anl.gmca.common.DevMoveFloat;
import gov.anl.gmca.common.Limits;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.PV_BLUICE;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.Util;
import gov.aps.jca.dbr.DBR;

import java.util.*;

/**
 * Ported from GM/CA's engine/g2eAttenuation.tcl with Satish's comments
 * 
 * Important: This device must be created -after- the filter devices
 */
public class DevAttenuation extends DevMoveFloat implements DevShutter.Listener
{
	protected static final int WAIT_TIMEOUT = 3000;

	// At 1000x, attenuation fluctuates about 0.1 eV
	// At 10,000x, attenuation fluctuates about 1 eV
	// So we set the inPosBand to 0.1% of the attenuation factor
	private static final double ATTEN_INPOS_BAND_PERCENT = 0.1;

	///////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////// state //////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////////

	protected boolean inLocalMove=false, inRemoteMove=false;
	
	protected void updateState()
	{
		State newState = null;
		
		if (!isOnline())					newState = State.OFFLINE;
		else if (inLocalMove||inRemoteMove)	newState = State.MOVING;
		else								newState = State.READY;
		
		setState(newState);

    	for (Listener listener : listenerSet) listener.handleStateChange(getState());
	}
	
	public void notifyActPosListeners()
	{
		for (Listener l : listenerSet)
		{
			l.handleActPosChange(this, attenuation_factor);
		}
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////// DevMove methods /////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////////

	public enum Mode { FACTORS, PERCENTAGE };
	
	protected Mode mode = Mode.FACTORS;

	double attenuation_factor = 1.0;
	public double getFactors() { attenuation_update(); return attenuation_factor; }

	HashMap<DevFilter, Boolean> filterConnectState = new HashMap<DevFilter, Boolean>();
	
	public DevAttenuation()
	{
		super("attenuation");
		if (deviceAlreadyExists) return;
		
		// assume this device requires all filters to be online
		for (DevFilter filter : DevFilter.getEnabledFilters())
		{
			filterConnectState.put(filter, false);
			filter.addListener(this);
			if (filter.getState() == State.READY) handleStateChange(filter, State.READY);
		}
	}
	
	protected Timer moveTimer = null;
	
	public class MoveTimer extends TimerTask
	{
		@Override
		public void run()
		{
			Epics.putDouble(PV_BLUICE.attenuation, attenuation_factor);
			inRemoteMove = false;
			updateState();
		}
	}
	
	protected HashSet<DevShutter> shutterFirstCB = new HashSet<DevShutter>();
	
	@Override
	public void handleOpenChange(DevShutter shutter, boolean open)
	{
		lastMoveTime = System.currentTimeMillis();

		if (!shutterFirstCB.contains(shutter))
		{
			shutterFirstCB.add(shutter);
			handleActPosChange(true);
		}
		else
		{
			handleActPosChange(false);
		}
	}
	
	public synchronized void handleActPosChange(boolean onStartup)
	{
		if (!onStartup)
		{
    		if (null != moveTimer) moveTimer.cancel();
    		moveTimer = new Timer("Attenuation move monitor");
    		moveTimer.schedule(new MoveTimer(), 500);
    		
    		inRemoteMove = true;
    		
    		updateState();
		}
		
		attenuation_update();
		
		// The attenuation device doesn't always know where it will end up
		// when the current "move" ends, so the mechanism where the motor widget
		// sets the rqspos at the beginning of the move will not always work.
		// Therefore, update the rqspos on every filter change.  This is OK for
		// the attenuation device because oscillation (or any unrequested motion)
		// is not a problem.
		for (Listener l : listenerSet)
		{
			l.handleActPosChange(this, attenuation_factor);
			l.handleRqsPosChange(attenuation_factor);
		}
	}

	@Override
	public void handleStateChange(DevShutter shutter, State state)
	{
		System.out.println(shutter.getDeviceName()+" state = "+state);
		
		if (state == State.READY)
		{
			filterConnectState.put((DevFilter)shutter, true);
		}
		
		int connected=0;
		
		boolean allConnected = true;
		for (DevFilter filter : filterConnectState.keySet())
		{
			if (!filterConnectState.get(filter))
			{
				allConnected=false;
			}
			else
			{
				connected++;
			}
		}
		
		//System.out.println("DevAttenuation: "+connected+" of "+filterConnectState.keySet().size()+" connected");
		
		if (allConnected)
		{
			attenuation_update();
			System.out.println("Initial attenuation = "+attenuation_factor);
			
			Epics.putDouble(PV_BLUICE.attenuation, attenuation_factor);
			
			setState(State.READY);
			
			final double rqspos = attenuation_factor;
			
			for (Listener l : listenerSet)
			{
				l.handleStateChange(State.READY);
				l.handleActPosChange(this, attenuation_factor);
				l.handleRqsPosChange(rqspos);
			}
		}
	}
	
	public static DevAttenuation get()
	{
		if (null == getDevice("attenuation"))
		{
			new DevAttenuation();
		}
		return ((DevAttenuation)getDevice("attenuation"));
	}

	public Limits getLimits()
	{
		return new Limits(1,10000000,"attenuation");
	}
	
	@Override protected void onAllPVsConnected() {}
	@Override protected void valueChangedAction(String pvName, DBR value) {}

	@Override
	public double getPosition()
	{
		attenuation_update();
		return attenuation_factor;
	}

	// attenuation changes are not abortable
	@Override public void abort() {}

	// TODO: fix these names!
	// attenuation changes are nearly immediate
	@Override public boolean getMoving() 	{ return inRemoteMove; }
	@Override public boolean isMoving() 	{ return inLocalMove; }

	// there is no rqspos PV, so just report the computed position
	@Override public double getRqsPos()		{ return getPosition(); }

	///////////////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////// public methods //////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////////

	public double getInPosBand()
	{
		double attenuation_magnitude = Math.pow(10,Math.round(Math.log10(attenuation_factor)));
		return ATTEN_INPOS_BAND_PERCENT/100.0 * attenuation_magnitude;
	}

	// dest<0, dest becomes 0, dest>100, dest becomes 100, so no requests fail
	@Override
	public boolean moveChecks(double dest) { return true; }

	// attenuation_move in Tcl
	@Override
	public void moveCmd(double newAttenuation, MoveResult result)
	{
		setState(State.MOVING);
		
		inLocalMove = true;
		
		System.out.println("Changing attenuation to "+newAttenuation);
		
		// convert to percentage if needed
		if (mode == Mode.FACTORS)
		{
			newAttenuation = (1.0-(1.0/newAttenuation))*100.0;
		}

		double energy = DevEnergy.getDevice().getPosition();
		
		// handle out of bounds requests
		if (newAttenuation >= 100.0)
		{
			for (DevFilter filter : DevFilter.getEnabledFilters())
			{
				filter.close();
			}
			for (DevFilter filter : DevFilter.getEnabledFilters())
			{
				filter.waitFor();
			}
			return;
		}

		if (newAttenuation < 0.0)
		{
			newAttenuation = 0.0;
		}

		// figure out how much attenuation each filter provides at the current energy
		int filterIndex = 0;
		ArrayList<Double> mutList = new ArrayList<Double>();
		for (DevFilter filter : DevFilter.getEnabledFilters())
		{
			double thickness = filter.getThickness();
			Material material = filter.getMaterial();
			double filterMuT = calculateFilterMuT(material, thickness, energy);
			
			mutList.add(filterMuT);
			filterIndex++;
		}
		
		double requiredMuT = calculateRequiredMuT(newAttenuation, energy);

		SortAndSearchAttenuationResult sasaResult = sortAndSearchAttenuation(requiredMuT, mutList);
		
		// move filters in that are going in
		for (int i=0; i < sasaResult.originalPattern.size(); i++)
		{
			if (sasaResult.originalPattern.get(i) != 0)
			{
				DevFilter.getEnabledFilters().get(i).close();
			}
		}
		
		// move filters out that are going out
		for (int i=0; i < sasaResult.originalPattern.size(); i++)
		{
			if (sasaResult.originalPattern.get(i) == 0)
			{
				DevFilter.getEnabledFilters().get(i).open();
			}
		}
		
		// move out disabled filters
		for (DevFilter filter : DevFilter.getDisabledFilters())
		{
			filter.open();
		}

		long startTime = System.currentTimeMillis();
		while (true)
		{
			long now = System.currentTimeMillis();
			if (now-startTime>WAIT_TIMEOUT)
			{
				String msg = "Attenuation move timed out";
				
				Log.error(msg);
				setStateAndResult(result, State.READY, ResultType.ERROR);
				result.msg = msg;
				
				return;
			}
			
			boolean allMatch = true;
			for (int i=0; i < sasaResult.originalPattern.size(); i++)
			{
				boolean requestedOpen = (0==sasaResult.originalPattern.get(i));
				DevFilter filter = DevFilter.getEnabledFilters().get(i);

				// are the filters that are going out actually out?
				if (requestedOpen  && !filter.getIsOpen()) allMatch=false;
				
				// are the filters that are going in actually in?
				if (!requestedOpen &&  filter.getIsOpen()) allMatch=false;
			}
			for (DevFilter filter : DevFilter.getDisabledFilters())
			{
				// are disabled filters all out?
				if (!filter.getIsOpen()) allMatch=false;
			}
			if (allMatch) break;
			Util.sleep(10);
		}
		
		inLocalMove = false;
		
		setStateAndResult(result, State.READY, ResultType.SUCCESS);
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////// calculation methods ///////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////////

	protected void attenuation_update()
	{
		// calculate attenuation from filter states
		double newattenuation_value = attenuation_calculate();

		if ( newattenuation_value != 100.0 )
		{
			attenuation_factor = 100.0 / (100.0 - newattenuation_value);
		}
		else
		{
			attenuation_factor = 100.0 / (100.0 - 99.99999999999999);
		}
	}	
	
	protected double attenuation_calculate()
	{
		// sd: device::energy
		double energy = DevPmacMotor.get("mo_En_E").getPosition();

		// mh: if energy is zero it will cause a divide by zero in proc absorption
	    if (0.0 == energy) return 1.0;

		double totalMuT = 0.0;
		double attenuationResult = 0.0;

		// Loop over all of the filters and sum up the total thickness.
		for (DevFilter filter : DevFilter.getEnabledFilters())
		{
			// sd: device::$shutter
			if (!filter.getIsOpen())
			{
				double filterThickness = filter.getThickness();
				Material filterMaterial = filter.getMaterial();
				double filterMuT = calculateFilterMuT(filterMaterial, filterThickness, energy);
				totalMuT += filterMuT;
			}
		}

		//Calculate the attenuation based on the total thickness.
		double totalTransmission = Math.exp(-1 * totalMuT);
		
		attenuationResult = (1.0 - totalTransmission) * 100.0;
		return attenuationResult;
	}
	
	protected  double calculateTransmission (Material material, double energy, double length)
	{
		double mu = material.absorption(energy);
		return Math.exp(-mu*length);
	}
	
	//sd:
		// ported from bluice04/dcss/scripts/devices
		// NB: some bugs were fixed as commented below

		// This function searches for a combination of filters that is closest to, but
		// not exceeding the desired thickness.
		// Inputs:
		//  targetThickness -> This it the
		//  thicknessList   -> This is the list of thicknesses for available filters.
		//                     The thickness list must be ordered by decreasing thickness
		//                     for this function to work.
		// Outputs:
		//           delta  -> The difference between the target thickness and the best
		//                     achievable thickness from the component filters.
		//           pattern-> This will be a list of binary numbers indicating whether
		//                     the component filter should be used or not.
		//
		// The function will recursively call itself while searching for the answer.
		//
	protected class SearchAttenuationResult
	{
		public double bestDelta;
		public ArrayList<Integer> bestPattern;
	}
	
	public SearchAttenuationResult searchAttenuation (double targetThickness, List<Double> thicknessList)
	{
		SearchAttenuationResult result = new SearchAttenuationResult();
		
		int totalShutters = thicknessList.size();
		result.bestDelta = targetThickness;
		result.bestPattern = new ArrayList<Integer>();
		ArrayList<Integer> thisPattern = new ArrayList<Integer>();
		
		//sd:set best pattern to all 0's
		for (int i=0; i<thicknessList.size(); i++)
		{
			result.bestPattern.add(0);
		}
		
		//sd: 
			//Shortcircuit the recursive calling if the target thickness
			//is less than or equals to zero
		if (targetThickness <= 0)
		{
			return result;
		}
		
		//Shortcircuit the recursive calling if the sum of the thicknesses
		// is less than the target thickness.
		double thicknessSum=0.0;
		ArrayList<Integer> possiblePattern = new ArrayList<Integer>();
		
		for (double shutter : thicknessList)
		{
			thicknessSum += shutter;
			possiblePattern.add(1);
		}

		if (thicknessSum <= targetThickness)
		{
			result.bestDelta = targetThickness - thicknessSum;
			result.bestPattern = possiblePattern;
			return result;
		}
		
		//Loop over all the filters, except the last filter in the list.

		int count=0;
		while ( (count+1) < totalShutters )
		{
			double thisThickness = thicknessList.get(count);
			
			//search for best resulting pattern for sub string
			SearchAttenuationResult saResult = searchAttenuation(targetThickness - thisThickness, thicknessList.subList(count+1, thicknessList.size()));
			double delta = Math.abs(saResult.bestDelta);

			if (delta >= result.bestDelta)
			{
				//that one didn't work
				thisPattern.add(0);
			}
			else
			{
				result.bestPattern = new ArrayList<Integer>();
				
				result.bestPattern.addAll(thisPattern);
				result.bestPattern.add(1);
				result.bestPattern.addAll(saResult.bestPattern);
				
				result.bestDelta = delta;
				thisPattern.add(0);
			}
			
			count++;
		}
		
		//Look at the last filter in the list.
		double delta = Math.abs(targetThickness - thicknessList.get(thicknessList.size()-1));
		
		if (delta < result.bestDelta)
		{
			result.bestPattern = new ArrayList<Integer>();
			result.bestPattern.addAll(thisPattern);
			result.bestPattern.add(1);

			result.bestDelta = delta;
		}

		return result;
	}
	
	// returns the sorted list, including duplicates, in sortedThicknessList
	// returns the corresponding original positions in sortedPosition
	protected class SortArrayResult
	{
		public ArrayList<Double> sortedThicknessList;
		public ArrayList<Integer> sortedPosition;
	}
	
	public SortArrayResult sortArray(ArrayList<Double> elements)
	{
		SortArrayResult result = new SortArrayResult();
		
		// copy elements
		result.sortedThicknessList = new ArrayList<Double>();
		for (Double d : elements) result.sortedThicknessList.add(d);
		
		int[] position = new int[result.sortedThicknessList.size()];
		
     	// position- position of element in unsorted array
		for (int i=0; i<result.sortedThicknessList.size(); i++)
		{
			position[i]=i;
		}

		for (int i=0; i<result.sortedThicknessList.size()-1; i++)
		{
			for (int j=0; j<result.sortedThicknessList.size()-1-i; j++)
			{
				double a = result.sortedThicknessList.get(j);
				double b = result.sortedThicknessList.get(j+1);
				
				if (a<b)
				{
					// swap result.sortedThicknessList[j] and result.sortedThicknessList[j+1]
					double dTemp = result.sortedThicknessList.get(j);
					result.sortedThicknessList.set(j, result.sortedThicknessList.get(j+1));
					result.sortedThicknessList.set(j+1, dTemp);
					
					// swap position[j] and position[j+1]
					int iTemp = position[j];
					position[j] = position[j+1];
					position[j+1] = iTemp;
				}
			}
		}
		
     	//sd:
     	//Example: {20 2 4 1 25}
     	//sorted array:     {25 20 4 2 1}
     	//unsortedPosition: {4 0 2 1 3 }
     	//                   ^ 25 is at position 4 in unsorted array
     	//position of element in the unsorted/input array
		ArrayList<Integer> unsortedPosition = new ArrayList<Integer>();
		for (int i=0; i<result.sortedThicknessList.size(); i++)
		{
			unsortedPosition.add(position[i]);
		}
		
     	//result.sortedPosition: position of element in sorted array
     	//Example:          {20 2 4 1 25}
     	//sorted array:     {25 20 4 2 1}
     	//result.sortedPosition:     1 3  2 4 0
     	//                    ^ 20 is at index 1 in sorted array
		result.sortedPosition = new ArrayList<Integer>();
		for (int i=0; i<result.sortedThicknessList.size(); i++)
		{
			int index = unsortedPosition.indexOf(i);
			result.sortedPosition.add(index);
		}
		
		return result;
	}
	
	public double calculateFilterAttenuation(Material filterMaterial, double filterThickness, double energy)
	{
		return (1.0 - calculateTransmission(filterMaterial, energy, filterThickness)) * 100;
	}

	protected class SortAndSearchAttenuationResult
	{
		public ArrayList<Integer> originalPattern;
		public double delta;
	}
	
	public SortAndSearchAttenuationResult sortAndSearchAttenuation (double targetThickness, ArrayList<Double> thicknessList)
	{
		SortAndSearchAttenuationResult result = new SortAndSearchAttenuationResult();
		
		SortArrayResult sortedArray = sortArray(thicknessList);
		
		SearchAttenuationResult saResult = searchAttenuation(targetThickness, sortedArray.sortedThicknessList);
		result.delta = saResult.bestDelta;
		
		result.originalPattern = new ArrayList<Integer>();
		
		for (Integer position : sortedArray.sortedPosition)
		{
			result.originalPattern.add(saResult.bestPattern.get(position));
		}
		
		return result;
	}
	
	public double calculateFilterMuT (Material filterMaterial, double filterThickness, double energy)
	{
		double mu = filterMaterial.absorption(energy);
		double muT = mu * filterThickness;
		
		return muT;
	}
	
	public double calculateRequiredMuT(double requestedAttenuation, double energy)
	{
		double transmission = (100.0 - requestedAttenuation) / 100.0;
		double requiredMuT = -1.0 * Math.log(transmission);
		
		return requiredMuT;
	}
}
