package gov.anl.gmca.hutch;

import gov.anl.gmca.common.DevMove.MoveResult;
import gov.anl.gmca.common.DevMove.ResultType;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.PVName;
import gov.anl.gmca.io.PVLookup.PV_PMAC;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.Util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.concurrent.atomic.AtomicBoolean;

// Used to move multiple motors in an assembly at once.
//
// Example usage:
// AssemblyMove move = DevPmacAssembly.get("sampleXY").newMove();
// move.setDest("sample_x", run.grid().getCenter().x+targetXYZ[0]);
// move.setDest("sample_y", run.grid().getCenter().y+targetXYZ[1]);
// MoveResult xyResult = move.start().waitFor();
//
// Note: It is not required to set a destination for each motor in an assembly.  If a destination
//       is not set, that motor's rqspos will be set to its actpos.
//
public class DevPmacAssembly
{
	public class AssemblyMove implements Runnable
	{
		protected HashMap<DevPmacMotor, Double> destMap = new HashMap<DevPmacMotor, Double>();
		protected Thread moveThread = null;
		protected MoveResult result = new MoveResult(ResultType.NOT_STARTED);
		
		public void setDest(String motorName, Double dest)
		{
			DevPmacMotor motor = DevPmacMotor.get(motorName);

			if (null == motor)
			{
				throw new IllegalArgumentException("Motor "+motorName+" does not exist.");
			}
			
			if (!motorList.contains(motor))
			{
				throw new IllegalArgumentException("Assembly "+assemblyName+" does not contain motor "+motor.getDeviceName());
			}
			
			if (destMap.containsKey(motor))
			{
				System.out.println("WARNING: Motor "+motor.getDeviceName()+"'s destination is being set more than once");
			}
			
			destMap.put(motor, dest);
		}
		
		public AssemblyMove start()
		{
			moveThread = new Thread(this, "Assembly "+assemblyName+" move");
			moveThread.start();
			return this;
		}
		
		@Override
		public void run()
		{
			if (!moveChecks(this))
			{
				result = new MoveResult(ResultType.INVALID);
				moveThread = null;
				return;
			}
			
			moveCmd(this, result);
			moveThread = null;
		}
		
		public DevPmacMotor.MoveResult waitFor()
		{
			while (null != moveThread)
			{
				Util.sleep(10);
			}
			return result;
		}
	}
	
	protected String assemblyName;
	protected ArrayList<DevPmacMotor> motorList;
	protected static HashMap<String, DevPmacAssembly> assemblyMap = new HashMap<String, DevPmacAssembly>();
	protected boolean useRunPrg;
	
	public DevPmacAssembly(String assemblyName, ArrayList<DevPmacMotor> motorList)
	{
		// check arguments
		if (null == motorList || motorList.isEmpty() || null == assemblyName)
		{
			throw new IllegalArgumentException("DevPmacAssembly("+assemblyName+","+motorList+") called with invalid arguments");
		}

		// make sure all motors really are in the same assembly
		// TODO: make this work again? add dev/ctl queries to PV_PMAC?
		/*String devPVSegment = motorList.get(0).getDevPVSegment();
		String ctlPVSegment = motorList.get(0).getCtlPVSegment();
		for (DevPmacMotor motor : motorList)
		{
			if (!motor.getDevPVSegment().equals(devPVSegment) ||
				!motor.getCtlPVSegment().equals(ctlPVSegment))
			{
				throw new IllegalArgumentException("Motor prefixes "+devPVSegment+":"+ctlPVSegment+" and "+motor.getDevPVSegment()+":"+motor.getCtlPVSegment()+" don't match for assembly "+assemblyName);
			}
		}*/
		
		this.motorList = new ArrayList<DevPmacMotor>(motorList);
		this.assemblyName = assemblyName;
		
		// set useRunPrg and make sure all motors match
		this.useRunPrg = motorList.get(0).useRunPrg;
		for (DevPmacMotor motor : motorList)
		{
			if (motor.useRunPrg != this.useRunPrg) throw new IllegalArgumentException("Motors in assembly "+assemblyName+" have mismatched runPrg setting in SQL");
		}
		
		assemblyMap.put(assemblyName,this);
	}
	
	public static DevPmacAssembly get(String assemblyName)
	{
		if (assemblyMap.containsKey(assemblyName))
		{
			return assemblyMap.get(assemblyName);
		}
		return null;
	}
	
	public AssemblyMove newMove()
	{
		return new AssemblyMove();
	}
	
	protected double estimateMoveTimeSec(AssemblyMove moveDest)
	{
		double maxEstimate = 0.0;
		
		for (DevPmacMotor motor : motorList)
		{
			double moveEstimate = motor.estimateMoveTimeSec(moveDest.destMap.get(motor));
			if (moveEstimate > maxEstimate) maxEstimate = moveEstimate;
		}
		
		return maxEstimate;
	}
	
	protected boolean moveChecks(AssemblyMove moveDest)
	{
		if (moveDest.destMap.isEmpty())
		{
			System.out.println("ERROR: Assembly "+assemblyName+" has no destinations set.");
			return false;
		}

		// NaN checks
		for (DevPmacMotor motor : moveDest.destMap.keySet())
		{
    		if (Double.isNaN(moveDest.destMap.get(motor)))
    		{
    			Log.error("Motor "+motor.getDeviceName()+" requested to move to NaN, aborted.");
    			return false;
    		}
		}
		
		// set destinations to actpos for any motors not set by the user
		for (DevPmacMotor motor : motorList)
		{
			if (!moveDest.destMap.containsKey(motor))
			{
				moveDest.destMap.put(motor, motor.getPosition());
			}
		}

		// run each motor's checks individually
		for (DevPmacMotor motor : moveDest.destMap.keySet())
		{
			// some are repeated since they apply to the assembly, but that's ok
			if (!motor.moveChecks(moveDest.destMap.get(motor)))
			{
				return false;
			}
		}
		
		return true;
	}
	
	protected AtomicBoolean moveAborted = new AtomicBoolean(false);
	
	protected void moveCmd(AssemblyMove moveDest, MoveResult result)
	{
		// check if all motors are already at their destinations
		if (!DevPmacMotor.MOVE_ON_POSITION_MATCH)
		{
			boolean allMatch = true;
			
			for (DevPmacMotor motor : motorList)
			{
            	double startingActPos = motor.getPosition();
            	double deltaMotion = moveDest.destMap.get(motor) - startingActPos;
            	if (Math.abs(deltaMotion) > motor.inPosBand)
            	{
            		allMatch = false;
            	}
			}
			
			if (allMatch)
			{
    			Log.note(Util.sprintf("Assembly %s move skipped because moves are less than tolerances",assemblyName));
    	        result.type = ResultType.SUCCESS;
    			return;
			}
		}
		
		for (DevPmacMotor motor : motorList)
		{
			motor.inLocalMove = true;
			motor.updateState();
		}
		
		// needed in post-processing
		boolean stillAtStart=false, allAmpsEnabled=true, progressing=false;
		double estimatedMoveTimeSec=0, maxTime=0, elapsedSec=0;
		boolean hasStarted		= false;
		boolean	hitAnyLimit		= false;
		boolean autoModeWasSet	= false;
		PVName autoPVName		= PV_PMAC.AutoStart.motor(motorList.get(0).getDeviceName());
		
		try
		{
            long	moveStartTimeMs 	= System.currentTimeMillis();
    				elapsedSec 			= 0.0;
    		int 	positionedCount		= 0;
    		double	motor_startup_time	= 1.0;
    
    		// add a map of status objects to track the state of each motor
    		class MotorStatus
    		{
    			double oldPos, prvPos, actPos, rqsPos, rqsDistance;
    			int limSign;
    			boolean limStatus;
    			boolean ampEna;
    			DevPmacMotor motor;
    		}
    		HashMap<DevPmacMotor,MotorStatus> motorStatusMap = new HashMap<DevPmacMotor,MotorStatus>();
    		for (DevPmacMotor motor : motorList)
    		{
    			// where is the motor and where is it going?
    			MotorStatus ms = new MotorStatus();
    			ms.oldPos = motor.getPosition(); 					// start position
    			ms.prvPos = ms.oldPos;           					// updated at the end of each iteration
    			ms.actPos = ms.oldPos;								// updated at the start of each iteration
    			ms.rqsPos = moveDest.destMap.get(motor);			// requested position
    			ms.rqsDistance = Math.abs(ms.rqsPos - ms.oldPos);	// requested distance
    			
    			// which direction is the requested motion in?
    			double signedRqsDistance = ms.rqsPos - ms.oldPos;
    			if (Math.abs(signedRqsDistance) < motor.inPosBand) ms.limSign=0;
    			else if (signedRqsDistance > 0) ms.limSign=1;
    			else ms.limSign=-1;
    
    			// limit and amp status
    			ms.limStatus = false;
    			ms.ampEna = true;
    			
    			// store the motor for easy iteration
    			ms.motor = motor;
    			
    			motorStatusMap.put(motor, ms);
    		}
    		
    				estimatedMoveTimeSec	= estimateMoveTimeSec(moveDest);
    		double	warnTime				= estimatedMoveTimeSec + 3.0;
        	double	overheadTime			= 5.0;
    				maxTime					= 2.0 * estimatedMoveTimeSec + overheadTime;
    		
    		// save auto mode
    		autoModeWasSet = (0==Epics.getInteger(autoPVName));
    		
    		// set auto mode to manual if necessary
    		if (!autoModeWasSet)
    		{
    			Epics.putInteger(autoPVName, 1);
    		}
    		
        	// if amp is disabled or runprg is running, send abort to stop motion
            boolean ampEnabled = 		(1==Epics.getInteger(PV_PMAC.AmpEna.motor(motorList.get(0).getDeviceName())));
            boolean runPrgRunning =  	(1==Epics.getInteger(PV_PMAC.RunPrg.motor(motorList.get(0).getDeviceName())));
     		if (!ampEnabled || runPrgRunning)
    		{
    			System.out.println("ampEna="+ampEnabled+" runPrg="+runPrgRunning+", sending abort=1 for assembly "+assemblyName);
    			Epics.putInteger(PV_PMAC.Abort.motor(motorList.get(0).getDeviceName()), 1);
    			
    			// wait to ensure abort doesn't interfere with start
    			Util.sleep(250);
    		}
        	
    		// set rqspos
    		for (DevPmacMotor motor : motorList)
    		{
    			if (moveDest.destMap.containsKey(motor))
    			{
    				double dest = moveDest.destMap.get(motor);
    				System.out.println("Setting rqspos="+dest+" for motor "+motor.getDeviceName()+" in assembly "+assemblyName);
    				Epics.putDouble(PV_PMAC.RqsPos.motor(motor.getDeviceName()), dest);
    			}
    		}
    		
    		// reset abort flags
			moveAborted.set(false);
			for (DevPmacMotor motor : motorList)
			{
				motor.moveAborted.set(false);
			}
    		
    		// send start
			System.out.println("Setting start=1 for assembly "+assemblyName);
    		Epics.putInteger(PV_PMAC.Start.motor(motorList.get(0).getDeviceName()), 1);
    
    		// track
    		while (true)
    		{
        		long currentTimeMs = System.currentTimeMillis();
        		
                // calculate elapsed and remaining time
                long    elapsedMs               = currentTimeMs - moveStartTimeMs;
                		elapsedSec              = ((double)elapsedMs)/1000.0;
                double  estimatedRemainingSec   = estimatedMoveTimeSec - elapsedSec;
                
    			// check assembly status
    			int inPos 	= Epics.getInteger(PV_PMAC.InPos.motor(motorList.get(0).getDeviceName()));
    			int runPrg 	= Epics.getInteger(PV_PMAC.RunPrg.motor(motorList.get(0).getDeviceName()));
    			if (1 == runPrg)	hasStarted = true;
    			if (1 == inPos)		positionedCount++;
    			else				positionedCount = 0;
    			
    			// check status of individual motors
    			stillAtStart	= false;
    			allAmpsEnabled	= true;
    			progressing		= false;
    			for (MotorStatus status : motorStatusMap.values())
    			{
    				// Check if at least one motor has amplifier failure
    				status.ampEna = (1==Epics.getInteger(PV_PMAC.AmpEna.motor(status.motor.getDeviceName())));
    				if (!status.ampEna)
    				{
    					Log.error("Amplifier fault for motor "+status.motor.getDeviceName());
    					allAmpsEnabled = false;
    				}
    				
    				// Check if at least one motor is still at start
    				status.actPos = Epics.getDouble(PV_PMAC.ActPos.motor(status.motor.getDeviceName()));
    				if (status.rqsDistance > status.motor.inPosBand)
    				{
    					double d1 = Math.abs(status.actPos - status.oldPos);
    					double d2 = Math.abs(status.actPos - status.rqsPos);
    					if (d1 < d2) stillAtStart = true;
    				}
    				
    				// Check if at least one motor is progressing
    				if (status.prvPos < status.actPos && status.actPos < status.rqsPos)
					{
						progressing = true;
					}
    				else if (status.prvPos > status.actPos && status.actPos > status.rqsPos)
    				{
    					progressing = true;
    				}
    				status.prvPos = status.actPos;
    				
    				// Check if at least one motor hit a limit
    				if (-1 == status.limSign)
    				{
    					boolean negLimitHit = (1==Epics.getInteger(PV_PMAC.NgLimSet.motor(status.motor.getDeviceName())));
    					if (negLimitHit)
    					{
    						System.out.println("Motor "+status.motor.getDeviceName()+" hit negative limit.");
    						hitAnyLimit = true;
    					}
    				}
    				else if (1 == status.limSign)
    				{
    					boolean posLimitHit = (1==Epics.getInteger(PV_PMAC.PsLimSet.motor(status.motor.getDeviceName())));
    					if (posLimitHit)
    					{
    						System.out.println("Motor "+status.motor.getDeviceName()+" hit positive limit.");
    						hitAnyLimit = true;
    					}
    				}
    			}
    			
    			// set abort flag if any motors in this assembly are aborted
    			for (DevPmacMotor motor : motorList)
    			{
    				if (motor.moveAborted.get())
    				{
    					moveAborted.set(true);
    				}
    			}
    			
    			// print move progress
    			String line = Util.sprintf
    			(
            		"track_assembly_motion %12s: " +
            		"elap.=%5.2fs  " +
            		"est.rem.=%5.1fs  " +
            		"allow.rem.=%5.1fs " +
            		"inPos=%d " +
            		"runPrg=%d ",
            		assemblyName,
                    elapsedSec,
                    estimatedRemainingSec,
                    maxTime-elapsedSec,
                    inPos,
                    runPrg
                );

    			for (MotorStatus status : motorStatusMap.values())
    			{
    				line += Util.sprintf
    				(
                		"********** "+
                		status.motor.getDeviceName()+" "+
                		"actPos=%5.4f " +
                		"oldPos=%5.4f " +
                		"rqsPos=%5.4f " +
                		"limitHit=%d " +
                		"ampEna=%d ",
                		status.actPos,
                		status.oldPos, 
                		status.rqsPos,
                		status.limStatus?1:0,
                		status.ampEna?1:0
                     );
                 }
    			
    			System.out.println(line);
    			
    			// determine if the loop can be broken
    			if
    			(
    				0 == runPrg &&
    				!stillAtStart &&
    				(useRunPrg || positionedCount >= 3) &&
    				(hasStarted || elapsedSec > motor_startup_time)
    			)
    			{
    				System.out.println("Quit on motion success.");
    				break;
    			}
    			if (elapsedSec > maxTime && !progressing)
    			{
    				System.out.println("Quit tracking on motion timeout.");
    				break;
    			}
    			if (!allAmpsEnabled)
    			{
    				System.out.println("Quit tracking on amp. failure.");
    				break;
    			}
    			if (hitAnyLimit)
    			{
    				System.out.println("Quit tracking on mtr. hit limit.");
    				break;
    			}
    			if (moveAborted.get())
    			{
    				System.out.println("Quit tracking on abort.");
    				break;
    			}
    			
    			int sleepTimeMs = useRunPrg ? 10 : 100;
    			Util.sleep(sleepTimeMs);
    		}
		}
		catch (Exception e) { e.printStackTrace(); }
		finally
		{
			for (DevPmacMotor motor : motorList)
			{
    			motor.inLocalMove = false;
    			motor.updateState();
			}
		}
		
		////////////////////////// post processing //////////////////////////
		
        //////////////////////////////////////////////////////////////////////////////////////////////
        // Check for exit conditions
        //////////////////////////////////////////////////////////////////////////////////////////////
		if (stillAtStart && !hasStarted)
		{
			Log.warn(assemblyName+" motion not detected. Possible late start? ...");
		}
	    if (moveAborted.get())
		{
			Log.warn("track_motion: Move aborted for assembly "+assemblyName+", tracking stopped.");
		}
		if (!allAmpsEnabled)
		{
			Log.warn("track_motion: Amplifier disabled in assembly "+assemblyName+", tracking stopped.");
		}
		if (hitAnyLimit)
		{
            Log.warn("track_motion: Hard limit hit in assembly "+assemblyName+", tracking stopped.");
		}

    	//////////////////////////////////////////////////////////////////////////////////////////////
    	// CHECK: Did the move take too long?
    	//////////////////////////////////////////////////////////////////////////////////////////////
		if (elapsedSec > maxTime)
		{
    		//Util.printf("track_motion: Motor %s motionTime=%3.1fs exceeded allowed time=%3.1fs",
             //       deviceName, elapsedSec, maxTimeSec);
		}

    	//////////////////////////////////////////////////////////////////////////////////////////////
    	// Check final position
    	//////////////////////////////////////////////////////////////////////////////////////////////
		for (DevPmacMotor motor : motorList)
		{
			double actpos = motor.getPosition();
			double diff = Math.abs(actpos - moveDest.destMap.get(motor));
			if (diff > motor.inPosBand)
			{
				// wait up to 250ms more for position to match, printing inpos and mismatch every 10ms
				for (int i=0; i<25; i++)
				{
					Util.sleep(10);
					actpos = motor.getPosition();
					diff = Math.abs(actpos - moveDest.destMap.get(motor));
					
					if (diff <= motor.inPosBand) break;
				}
				
				// check again
				if (diff > motor.inPosBand)
				{
				    System.out.println("Extra waiting for motor "+motor.getDeviceName()+" didn't help - position mismatch.");
				    if (DevPmacMotor.ERROR_ON_POSITION_MISMATCH)
				    {
					result.type = ResultType.ERROR;
					result.msg = "Position mismatch";
					return;
				    }
				}
				else
				{
					System.out.println("Motor "+motor+" finally arrived.");
				}
			}
		}
		
		// restore auto mode if necessary
		if (autoModeWasSet)
		{
			Epics.putInteger(autoPVName, 0);
		}
		
        // return OK
        Util.printf("track_motion: Assembly %s, no errors detected.", assemblyName);
		result.type = ResultType.SUCCESS;
	}
	
	public void abort()
	{
		moveAborted.set(true);
		PVName abortPV = PV_PMAC.Abort.motor(motorList.get(0).getDeviceName());
		System.out.println("Sending 1 to abort PV: "+abortPV);
		Epics.put(abortPV, "1");
	}
}
