package gov.anl.gmca.hutch;

import gov.anl.gmca.collect.DevCollectSettings;
import gov.anl.gmca.common.DevMoveFloat;
import gov.anl.gmca.common.Limits;
import gov.anl.gmca.gui.GUI;
import gov.anl.gmca.gui.GUIStatus;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup;
import gov.anl.gmca.io.PVLookup.PV_PMAC;
import gov.anl.gmca.scan.ModelScan;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.Util;
import gov.aps.jca.dbr.DBR;

import java.util.StringTokenizer;
import java.util.concurrent.atomic.AtomicBoolean;

public class DevPmacMotor extends DevMoveFloat
{
	protected static final boolean	ERROR_ON_POSITION_MISMATCH	= true;
	protected static final boolean	MOVE_ON_POSITION_MATCH		= false;
	protected static final double	BACKLASH_COEFFICIENT		= 0.20;
	protected static final double	DEFAULT_POSITION_ALLOWANCE	= 0.010;
	
	protected double inPosBand;
	protected boolean useRunPrg;
	@Override public double getInPosBand() { return inPosBand; }
	
	/////////////////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////// state ///////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////////////

	protected boolean 	inLocalMove 	= false;
	protected boolean	inRemoteMove 	= false;
	protected String 	error 			= null;
	
	protected void updateState()
	{
		State newState = null;
		
		if (!isOnline())						newState = State.OFFLINE;
		else if (null != error)					newState = State.ERROR;
		else if (inLocalMove || inRemoteMove)	newState = State.MOVING;
		else									newState = State.READY;

		//System.out.println("Dev "+motorID+" oldState="+getState()+" newState="+newState);

		setState(newState);

    	for (Listener listener : listenerSet) listener.handleStateChange(getState());
	}
	
	/////////////////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////// PV handlers /////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	@Override
	protected void onAllPVsConnected()
	{
		updateState();
		
		System.out.println("All PVs connected for device "+motorID+". State="+getState()+" inPosBand="+Util.sprintf("%.5f",inPosBand));
	}
	
	protected void determineInPosBand(String inPosBandString)
	{
		if (null != inPosBandString)
		{
			try
			{
				inPosBand = Double.parseDouble(inPosBandString);
				return;
			}
			catch (NumberFormatException nfe)
			{
				System.out.println("ERROR: Couldn't format inPosBand string \""+inPosBandString+"\"");
			}
		}

		// see if a backlash is set
		double backlash	= Math.abs(Epics.getDouble(PV_PMAC.RqsBkl.motor(motorID)));
    	if (0.0 != backlash)
    	{
    		inPosBand = BACKLASH_COEFFICIENT * backlash;
    	}
    	else
    	{
    		// fall back to the default
    		inPosBand = DEFAULT_POSITION_ALLOWANCE;
    	}
	}
	
	public boolean isMoveNeededTo(double destination)
	{
		double currentPos = getPosition();
		boolean needed = !(Math.abs(destination-currentPos) <= inPosBand);
		
		Util.printf("Move from current position %.6f to requested destination %.6f IS "+(needed?"":"NOT ")+" needed.", currentPos, destination);
		
		return needed;
	}
	
	protected double stableActPos = Double.MAX_VALUE;
	
	// TODO: remove this once GUI no longer uses the device directly
	@Override
	public void valueChangedAction(String pvName, final DBR value)
	{
		if (pvName.equals(PV_PMAC.ActPos.motor(motorID).toString()))
		{
        	final double actpos = Epics.dbrToDouble(value);
			
        	//System.out.println(Util.sprintf("Motor %12s act=%12.6f last=%12.6f inPos=%12.6f",getDeviceName(),actpos,stableActPos,inPosBand));
        	if (Math.abs(actpos- stableActPos)<inPosBand && !inRemoteMove && !inLocalMove)
        	{
        		//System.out.println("Skipping "+deviceName+" update");
        		return;
        	}
        	
    		if (!DevCollectSettings.getEnableEnergyMove() && getDeviceName().equals("mo_En_E"))
    		{
    			return;
    		}
        	
        	stableActPos = actpos;
        	
			final DevMoveFloat dev = this;
			if ( GUIStatus.inGUIMode) GUI.asyncExec(new Runnable()
			{
	            public void run()
	            {
	            	for (Listener listener : listenerSet) listener.handleActPosChange(dev, actpos);
	            }
			});

			// not exact, but any actpos updates will trigger updating lastMoveTime during a move
			if (Epics.getEnum(PV_PMAC.InPos.motor(motorID))==0)
			{
				lastMoveTime = System.currentTimeMillis();
			}
		}
		else if (pvName.equals(PV_PMAC.RqsPos.motor(motorID).toString()))
		{
			if (GUIStatus.inGUIMode) GUI.asyncExec(new Runnable()
			{
	            public void run()
	            {
	            	double rqspos = Epics.dbrToDouble(value);
	            	for (Listener listener : listenerSet) listener.handleRqsPosChange(rqspos);
	            }
			});
		}
		else if (pvName.equals(PV_PMAC.InPos.motor(motorID).toString()))
		{
			boolean energyMoveDuringScan = motorID.equals("mo_En_E") &&
			                               (Epics.getInteger(PVLookup.PV_SCAN.autoStep) == ModelScan.AutoStep.EDGE_SCAN.ordinal() ||
										   Epics.getInteger(PVLookup.PV_SCAN.interStep) == ModelScan.InterStep.EDGE.ordinal());
			if (energyMoveDuringScan) {
				System.out.print("mono_updates: Skipping in-pos update due to edge scan");
			} else {
				inRemoteMove = (0 == Epics.dbrToInt(value));
				llMoveStartStop.notifyListeners();
				updateState();
			}
		}
		else if (pvName.equals(PVLookup.PV_SCAN.autoStep.toString()) || pvName.equals(PVLookup.PV_SCAN.interStep.toString()))
		{
			if (motorID.equals("mo_En_E") && 0 == Epics.dbrToInt(value)) {
				System.out.print("mono_updates: Sending in-pos update on edge scan complete.");
				inRemoteMove = (0 == Epics.getInteger(PV_PMAC.InPos.motor(motorID)));
				llMoveStartStop.notifyListeners();
				updateState();
			}
		}
		else if (pvName.equals(PV_PMAC.ActPos_EGU.motor(motorID).toString()))
		{
			String unitsWithParens = Epics.dbrToString(value);
			StringTokenizer st = new StringTokenizer(unitsWithParens,"()");
			final String units = st.hasMoreTokens() ? st.nextToken() : "";
			if (GUIStatus.inGUIMode) GUI.asyncExec(new Runnable()
			{
	            public void run()
	            {
	            	for (Listener listener : listenerSet) listener.handleUnitsChange(units);
	            }
			});
		}
		
		// the following turns off ActPos monitors after their first callback
		// TODO: test this
		/*if (pvName.indexOf("ActPos") != -1)
		{
			Epics.stopMonitor(pvName);
		}*/
	}
	
	////////////////////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////// move commands //////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////////////////////////

	public boolean syncAssembly()
	{
		System.out.println("syncAssembly: Starting");
		Epics.putInteger(PV_PMAC.Sync_PROC.motor(motorID), 0);
		Util.sleep(300); // wait for any moves triggered by sync to start run program
		long start = System.currentTimeMillis();
		while (true) {
			long elapsedMs = System.currentTimeMillis() - start;
			if (Epics.getInteger(PV_PMAC.RunPrg.motor(motorID)) == 0) {
				System.out.println("syncAssembly: Success after "+elapsedMs+" ms");
				return true;
			}
			if (elapsedMs > 10000) {
				System.out.println("syncAssembly: Failed after "+elapsedMs+" ms");
				return false;
			}
			Util.sleep(20);
		}
	}

	@Override
	public boolean moveChecks(double dest)
	{
		// skip checks for disabled motors
		if (!DevCollectSettings.getEnableEnergyMove() && getDeviceName().equals("mo_En_E"))
		{
			System.out.println("Mono moves are disabled, returning OK for checks.");
			return true;
		}
		if (!DevCollectSettings.getEnableDetectorMove() && getDeviceName().equals("detector_z"))
		{
			System.out.println("CCD/PAD detector moves are disabled, returning OK for checks.");
			return true;
		}
		
		// NaN check
		if (Double.isNaN(dest))
		{
			Log.error("Motor "+motorID+" requested to move to NaN, aborted.");
			return false;
		}
		
		// check if request is within limits
		double lowerLimit = Epics.getDouble(PV_PMAC.RqsPos_DRVL.motor(motorID));
		if (0.0 != lowerLimit && dest < lowerLimit)
		{
			Log.error("Motor "+motorID+" move to "+dest+" aborted because lower limit is "+lowerLimit);
			return false;
		}
		double upperLimit = Epics.getDouble(PV_PMAC.RqsPos_DRVH.motor(motorID));
		if (0.0 != upperLimit && dest > upperLimit)
		{
			Log.error("Motor "+motorID+" move to "+dest+" aborted because upper limit is "+upperLimit);
			return false;
		}
    	
    	// check if homed
        int mHomed = Epics.getInteger(PV_PMAC.Homed.motor(motorID));

        if (0 == mHomed)
        {
			Log.error("Motor "+motorID+" move to "+dest+" aborted because homed=0");
    		return false;
        }
        
        // check that motion program is loaded
        int mMotion = Epics.getInteger(PV_PMAC.Motion.motor(motorID));
        if (mMotion==0)
        {
			Log.error("Motor "+motorID+" move to "+dest+" aborted because motion=0");
    		return false;
        }
        
		// check if amp is enabled
		if (0 == Epics.getInteger(PV_PMAC.AmpEna.motor(motorID)))
		{
			Log.error("Motor "+motorID+" move to "+dest+" aborted because AmpEna=0");
    		return false;
		}
		
		System.out.println("Device "+motorID+" move checks pass.");
		return true;
	}
	
	protected double estimateMoveTimeSec(double dest)
	{
    	double motionTime=0.0;
    	double accelTime = Epics.getDouble(PV_PMAC.AccelTime.motor(motorID));
    	double accelCurve = Epics.getDouble(PV_PMAC.AccelCurve.motor(motorID));
    	
    	if (accelTime < 2.0*accelCurve)
    	{
    		accelTime = 2.0*accelCurve;
    	}

        double actPos, rqsPos, speed;

        actPos = getPosition();
        rqsPos = dest;
        speed = Epics.getDouble(PV_PMAC.speed.motor(motorID));
        
        if (speed == 0.0)
        {
        	System.out.println("Speed is zero, requesting actual speed...");
        	
        	// enable speedSDIS
        	Epics.putInteger(PV_PMAC.speedSDIS.motor(motorID), 0);
        	Util.sleep(20);
        	
        	// re-enter zero for speed
        	Epics.putDouble(PV_PMAC.speed.motor(motorID), 0.0);
            Util.sleep(20);
            
            // re-read speed
            speed = Epics.getDouble(PV_PMAC.speed.motor(motorID));
            
            System.out.println("Speed re-read as "+speed);
        }
        
        // if speed is still zero, warn
        if (speed == 0.0)
        {
        	Log.warn("Speed for motor "+motorID+" is zero!");
        }
        else
        {
        	motionTime = Math.abs((rqsPos - actPos) / speed);
        }
    	
    	if (motionTime > 30.0*60.0)
    	{
    		motionTime = 30.0*60.0;
    	}
    	
    	// acceleration + deceleration time
    	motionTime += 0.002 * accelTime;
    	
    	// backlash acceleration
    	motionTime += 0.002 * accelTime;
            
        // time to get the three in-position readings in a row
        motionTime += 0.3;

    	System.out.println(Util.sprintf("track_motion: estimateMoveTime(): " +
    			"motionTime=%3.1fs " +
    			"accelTime=%4.2fs " +
    			"accelCurve=%4.2fs " +
    			"speed=%6.4f " +
    			"rqs=%4.4f " +
    			"act=%4.4f " +
    			"motor=%s " +
    			"totalEstimate=%3.1fs",
    			motionTime,
    			accelTime,
    			accelCurve,
    			speed,
    			rqsPos,
    			actPos,
    			motorID,
    			motionTime));
    	
    	return motionTime;
	}
	
	protected AtomicBoolean moveAborted = new AtomicBoolean(false);
	
	@Override
	public void moveCmd(double dest, MoveResult result)
	{
		Log.note("moveMotor("+motorID+", "+dest+")");
		
		double maxAllowedFinalDiff = inPosBand;
    	
		// skip moves for disabled motors
		if (!DevCollectSettings.getEnableEnergyMove() && getDeviceName().equals("mo_En_E"))
		{
			System.out.println("Mono moves are disabled, returning success for move.");
	        result.type = ResultType.SUCCESS;
			return;
		}
		if 
		(
			!DevCollectSettings.getEnableDetectorMove() &&
			(
				getDeviceName().equals("detector_z") ||
				getDeviceName().equals("detector_2t")
			)
		)
		{
			System.out.println("CCD/PAD detector moves are disabled, returning success for move.");
	        result.type = ResultType.SUCCESS;
			return;
		}
		
		// check if motor is already at destination
		if (!MOVE_ON_POSITION_MATCH || useRunPrg)
		{
        	double startingActPos = getPosition();
        	double deltaMotion = dest - startingActPos;
        	if (Math.abs(deltaMotion) <= maxAllowedFinalDiff)
        	{
    			Log.note(Util.sprintf("Motor %s move from %.4f to %.4f skipped because move is less than tolerance=%.4f",motorID, startingActPos, dest, maxAllowedFinalDiff));
    	        result.type = ResultType.SUCCESS;
    			return;
        	}
		}
		
        inLocalMove = true; updateState();
        
    	// if amp is disabled or runprg is running, send abort to stop motion
        boolean ampEnabled = 		(1==Epics.getInteger(PV_PMAC.AmpEna.motor(motorID)));
        boolean runPrgRunning =  	(1==Epics.getInteger(PV_PMAC.RunPrg.motor(motorID)));
 		if (!ampEnabled || runPrgRunning)
		{
			System.out.println("ampEna="+ampEnabled+" runPrg="+runPrgRunning+", sending abort=1 for motor "+motorID);
			Epics.putInteger(PV_PMAC.Abort.motor(motorID), 1);
			
			// wait to ensure abort doesn't interfere with start
			Util.sleep(250);
		}
    	
		// set the new rqspos
		System.out.println("Setting rqspos="+dest+" for motor "+motorID);
		Epics.put(PV_PMAC.RqsPos.motor(motorID), Double.toString(dest));

		// reset abort flag
		moveAborted.set(false);
		
		// set start=1 if necessary (autostart=1 means manual!)
		if (1 == Epics.getInteger(PV_PMAC.AutoStart.motor(motorID)))
		{
			System.out.println("Setting start=1 for motor "+motorID);
			Epics.put(PV_PMAC.Start.motor(motorID), "1");
		}
		else
		{
			System.out.println("Motor "+motorID+" has autostart set, not sending start=1");
		}
		
        long moveStartTimeMs = System.currentTimeMillis();
		double actPosAtMoveStart = Epics.getDouble(PV_PMAC.ActPos.motor(motorID));
		boolean runPrgShowedStart = false;
		int positionedCount = 0;
		double elapsedSec = 0.0;
		boolean upperLimitHit=false, lowerLimitHit=false;
		boolean isAtGoal=false;
		double rqsPos=Double.NaN, actPos=Double.NaN;
        boolean ampEna = (1==Epics.getInteger(PV_PMAC.AmpEna.motor(motorID)));

		double estimatedMoveTimeSec = estimateMoveTimeSec(dest);
    	double motorStartupTimeSec	= 1.0;
    	double safeOverheadTimeSec	= 5.0;
    	double maxTimeSec			= 2.0 * estimatedMoveTimeSec + safeOverheadTimeSec;
		
        // use try/catch to make sure inLocalMove gets reset to false
        try
        {
    		while (true)
    		{
        		long currentTimeMs = System.currentTimeMillis();
        		
                // calculate elapsed and remaining time
                long    elapsedMs               = currentTimeMs - moveStartTimeMs;
                elapsedSec              		= ((double)elapsedMs)/1000.0;
                double  estimatedRemainingSec   = estimatedMoveTimeSec - elapsedSec;
                
            	// PV status
            	int	inPos		    = Epics.getEnum(PV_PMAC.InPos.motor(motorID));
            	int	runPrg		    = Epics.getEnum(PV_PMAC.RunPrg.motor(motorID));
                actPos      		= getPosition();
                double  oldPos      = actPosAtMoveStart;
                rqsPos      		= Epics.getDouble(PV_PMAC.RqsPos.motor(motorID));
                ampEna				= (1==Epics.getInteger(PV_PMAC.AmpEna.motor(motorID)));
                
                int moveDirection = 0;
                if (rqsPos > oldPos) moveDirection=1;
                if (rqsPos < oldPos) moveDirection=-1;
                
                // limit status: allow limits to be hit if the move is in the opposite direction
                // note: although the values are always 1 or 0, the type is double
                lowerLimitHit = (Epics.getEnum(PV_PMAC.NgLimSet.motor(motorID))==1 && moveDirection == -1);
                upperLimitHit = (Epics.getEnum(PV_PMAC.PsLimSet.motor(motorID))==1 && moveDirection == +1);
        
            	// runPrgShowedStart
            	if (runPrg == 1/*moving*/) runPrgShowedStart = true;
        
            	// positionedCount
            	if (inPos == 1)		positionedCount++;
            	else				positionedCount = 0;
            	
            	// isAtGoal
            	double distanceThisMove = Math.abs(actPos - oldPos);
            	double distanceFromGoal = Math.abs(actPos - rqsPos);
            	isAtGoal =
        		(
        			(
        				distanceThisMove >= distanceFromGoal || // is it half way?
        				Math.abs(rqsPos-oldPos) <= 0.1			// is the move very small (zero?)
        			)
        			&&
        			(
        				distanceFromGoal <= maxAllowedFinalDiff // is it within goal tolerance?
        			)
        		);
            	
                // print move progress
                Util.printf(
                		"track_motion %12s: " +
                		"elap.=%5.1fs  " +
                		"est.rem.=%5.1fs  " +
                		"allow.rem.=%5.1fs " +
                		"inPos=%d " +
                		"runPrg=%d " +
                		"actPos=%5.4f " +
                		"oldPos=%5.4f " +
                		"rqsPos=%5.4f " +
                		"ngLim=%d " +
                		"psLim=%d " +
                		"runPrgStrt=%d " +
                		"posCt=%3d " +
                		"distSoFar=%5.4f " +
                		"fromGoal=%5.4f " +
                		"isAtGoal=%d",
                		motorID,
                        elapsedSec,
                        estimatedRemainingSec,
                        maxTimeSec-elapsedSec,
                        inPos,
                        runPrg,
                        actPos, 
                        oldPos, 
                        rqsPos, 
                        lowerLimitHit?1:0,
                        upperLimitHit?1:0,
                        runPrgShowedStart?1:0,
                        positionedCount, 
                        distanceThisMove, 
                        distanceFromGoal, 
                        isAtGoal?1:0
                );
          
            	//////////////////////////////////////////////////////////////////////////////////////////////
            	// Set up the next callback if the move hasn't finished but there's still time left
            	//////////////////////////////////////////////////////////////////////////////////////////////
            	if
                (
                	!(
                        runPrg == 0 /* stopped */ &&
                        (useRunPrg || positionedCount >= 3) &&
                        isAtGoal &&
                        (
                    		runPrgShowedStart ||
                    		elapsedSec > motorStartupTimeSec
                        )
                	)
                    &&
                    !(
                    	elapsedSec > 1.0 && elapsedSec > maxTimeSec && 0 == runPrg
                    )
                    &&
                    !(
                        !ampEna || upperLimitHit || lowerLimitHit || moveAborted.get()
                    )
                )
            	{
                    int sleepTimeMs = useRunPrg ? 10 : 100;
            		Util.sleep(sleepTimeMs);
                    continue;
            	}
            	else
            	{
            		break;
            	}
    		}
        }
        catch (Exception e) { e.printStackTrace(); }
        finally
        {
        	inLocalMove = false; updateState();
        }
        
    	//////////////////////////////////////////////////////////////////////////////////////////////
    	// Show final move time.
    	//////////////////////////////////////////////////////////////////////////////////////////////
    	Util.printf("track_motion: Tracking ended. Real in-motion time for motor %s was %3.1fs (%4.1f%% of est.)",
                              motorID,
                              elapsedSec,
                              elapsedSec/estimatedMoveTimeSec*100.0);
    
        //////////////////////////////////////////////////////////////////////////////////////////////
        // Check for exit conditions
        //////////////////////////////////////////////////////////////////////////////////////////////
	    if (moveAborted.get())
		{
			Log.warn("track_motion: Move aborted for motor "+motorID+", tracking stopped.");
		}
        if (upperLimitHit)
        {
            Log.warn("track_motion: Upper hard limit hit for motor "+motorID+", tracking stopped.");
        }
        if (lowerLimitHit)
        {
            Log.warn("track_motion: Lower hard limit hit for motor "+motorID+", tracking stopped.");
        }
        if (!ampEna)
        {
            Log.warn("track_motion: Amplifier disabled for motor "+motorID+", tracking stopped.");
        }
    
    	//////////////////////////////////////////////////////////////////////////////////////////////
    	// CHECK: Did the move take too long?
    	//////////////////////////////////////////////////////////////////////////////////////////////
    	if (elapsedSec > maxTimeSec)
    	{
    		Util.printf("track_motion: Motor %s motionTime=%3.1fs exceeded allowed time=%3.1fs",
                        motorID, elapsedSec, maxTimeSec);
		}
		
    	//////////////////////////////////////////////////////////////////////////////////////////////
    	// Check final position
    	//////////////////////////////////////////////////////////////////////////////////////////////
    	double diff = Math.abs(rqsPos-actPos);
    	if (diff > maxAllowedFinalDiff)
    	{
	    if (useRunPrg)
	    {
		// wait up to 250ms more for position to match, printing inpos and mismatch every 10ms
		for (int i=0; i<25; i++)
		{
		    Util.sleep(10);
		    actPos = getPosition();
		    diff = Math.abs(rqsPos-actPos);
		    
		    if (diff <= maxAllowedFinalDiff) break;
		}
	    }

	    // check again
	    if (diff > maxAllowedFinalDiff)
	    {
		Log.error(Util.sprintf("Move failed for "+motorID+", "+
				       "mismatch=%5.4f, " +
				       "req=%5.4f, " +
				       "act=%5.4f, " +
				       "allowance=%5.4f, " +
				       "moveDetected="+runPrgShowedStart,
				       diff,
				       rqsPos,
				       actPos,
				       maxAllowedFinalDiff));
		
		if (ERROR_ON_POSITION_MISMATCH)
		{
		    result.type = ResultType.ERROR;
		    result.msg = "Position mismatch";
		    return;
		}
		else
		{
		    Util.printf("Ignoring position mismatch for "+motorID+", "+
				"mismatch=%5.4f, " +
				"req=%5.4f, " +
				"act=%5.4f, " +
				"allowance=%5.4f, " +
				"moveDetected="+runPrgShowedStart,
				diff,
				rqsPos,
				actPos,
				maxAllowedFinalDiff);
		    
		    result.type = ResultType.SUCCESS;
		    return;
		}
	    }
	}
    	
        // no errors!
        Util.printf("track_motion: Motor %s, motion completed successfully.", motorID);
		result.type = ResultType.SUCCESS;
	}
	
	//@Override
	public void abort()
	{
		moveAborted.set(true);
		System.out.println("Sending 1 to abort PV for device "+motorID);
		Epics.put(PV_PMAC.Abort.motor(motorID), "1");
	}
	
	/////////////////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////// ctor ////////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	public DevPmacMotor(String motorID, boolean useRunPrg, String inPosBandString)
	{
		super(motorID);
		if (deviceAlreadyExists)
		{
			throw new IllegalStateException("Device "+motorID+" already exists");
		}
		
		this.useRunPrg = useRunPrg;
		determineInPosBand(inPosBandString);
		
		for (PV_PMAC pv : PV_PMAC.values())
		{
			if (motorID.contains("crl") && pv == PV_PMAC.RqsBkl) continue; // this PV appears to not be available
			registerDevicePV(pv.motor(motorID));
		}
		if (motorID.equals("mo_En_E")) {
			registerDevicePV(PVLookup.PV_SCAN.interStep);
			registerDevicePV(PVLookup.PV_SCAN.autoStep);
		}
		registrationComplete();
	}
	
	/////////////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////// register PVs /////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	@Override public double getPosition()
	{
		return Epics.getDouble(PV_PMAC.ActPos.motor(motorID));
	}
	public double getStablePosition()
	{
		return stableActPos;
	}

	public boolean getHomed() { return 1==Epics.getInteger(PV_PMAC.Homed.motor(motorID)); }
	public boolean getAmpEnabled() { return 1==Epics.getInteger(PV_PMAC.AmpEna.motor(motorID)); }
	public boolean getHasMotionPrg() { return 0!=Epics.getInteger(PV_PMAC.Motion.motor(motorID)); }

	@Override public double getRqsPos()
	{
		return Epics.getDouble(PV_PMAC.RqsPos.motor(motorID));
	}

	public String getUnits()
	{
		return Epics.getString(PV_PMAC.ActPos_EGU.motor(motorID));
	}

	public Boolean getInPos()
	{
		return Epics.getBoolean(PV_PMAC.InPos.motor(motorID));
	}
	
	public boolean getRunPrg()
	{
		return (0==Epics.getInteger(PV_PMAC.RunPrg.motor(motorID)));
	}
	
	@Override public boolean getMoving()
	{
		return (0 == Epics.getInteger(PV_PMAC.InPos.motor(motorID)));
	}
	
	@Override public Limits getLimits()
	{
		return new Limits
		(
			Epics.getDouble(PV_PMAC.RqsPos_DRVL.motor(motorID)),
			Epics.getDouble(PV_PMAC.RqsPos_DRVH.motor(motorID)),
			"soft limit for device "+motorID
		);
	}
	
	/////////////////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////// etc /////////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	public static DevPmacMotor get(String deviceName)
	{
		return (DevPmacMotor)getDevice(deviceName);
	}
	
	public void connectionChanged(String pvName, boolean isConnected) {}
}
	
