package gov.anl.gmca.hutch;

// This device handles the 23x:bi:s[0-7] PVs which store which scripts are running

import gov.anl.gmca.common.EpicsDevice;
import gov.anl.gmca.hutch.ModelScriptTable.Script;
import gov.anl.gmca.hutch.ModelScriptTable.ScriptState;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.PV_SCRIPT;
import gov.anl.gmca.syslog.Log;
import gov.aps.jca.dbr.DBR;

import java.util.concurrent.Semaphore;

public class DevScriptTable extends EpicsDevice
{
	protected void readScriptTable()
	{
		tableLock.acquireUninterruptibly();

		model.resetTable();
		
		String usersString = "Read scripts table from EPICS: ";
		
		for (int j=0; j < ModelScriptTable.MAX_SCRIPTS; j++)
		{
			// read the row from EPICS
			String		name	= Epics.getString(PV_SCRIPT.name.row(j));
			int			pid		= Epics.getInteger(PV_SCRIPT.pid.row(j));
			String		host	= Epics.getString(PV_SCRIPT.host.row(j));
			ModelScriptTable.ScriptState	state	= ModelScriptTable.ScriptState.values()[Epics.getInteger(PV_SCRIPT.state.row(j))];

			// set this row in the user table
			model.scriptTable[j].name 	= name;
			model.scriptTable[j].pid 	= pid;
			model.scriptTable[j].host	= host;
			model.scriptTable[j].state 	= state;
			
			if (null != name && name.length()>0 || null != host && host.length()>0)
			{
				usersString += "["+name+"@"+host+"] ";
			}
		}
		tableLock.release();
		
		System.out.println(usersString);
		
		model.ll.notifyListeners();
	}
	
	public int registerScript(Script script)
	{
		readScriptTable();
		
		int firstEmptyRow = -1;
		for(int j = 0; j < ModelScriptTable.MAX_SCRIPTS; j++)
		{
	 		if (0 == model.scriptTable[j].pid)
	 		{
	 			firstEmptyRow = j;
	 			break;
	 		}
		}

		if (-1 == firstEmptyRow)
		{
			Log.error("Scripts table is full, can't register client.");
			return -1;
		}

		Log.note("Registering script at row "+firstEmptyRow);

		Epics.put		(PV_SCRIPT.name.row	(firstEmptyRow), script.name);
		Epics.putInteger(PV_SCRIPT.pid.row	(firstEmptyRow), script.pid);
		Epics.put		(PV_SCRIPT.host.row	(firstEmptyRow), script.host);
		Epics.putInteger(PV_SCRIPT.state.row(firstEmptyRow), script.state.ordinal());

		return firstEmptyRow;
	}
	
	public boolean unregisterScript(Script script)
	{
		readScriptTable();

		int row = findScript(script);
		
		if (row < 0) return false;

		Log.note("Releasing script slot "+row);
		
		Epics.put		(PV_SCRIPT.name.row	(row), "");
		Epics.putInteger(PV_SCRIPT.pid.row	(row), 0);
		Epics.put		(PV_SCRIPT.host.row	(row), "");
		Epics.putInteger(PV_SCRIPT.state.row(row), 0);

		return true;
	}
	
	public int findScript(Script script)
	{
		int scriptRow = -1;
		for (int j = 0; j < ModelScriptTable.MAX_SCRIPTS; j++)
		{
	 		if
	 		(
	 			script.pid == model.scriptTable[j].pid &&
	 			script.host.equals(model.scriptTable[j].host)
	 		)
	 		{
	 			scriptRow = j;
	 			break;
	 		}
		}
		return scriptRow;
	}
	
	/////////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////// end public methods /////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////
	
	protected 			ModelScriptTable 	model;
	final 				Semaphore 			tableLock 	= new Semaphore(1);
	
	protected static	DevScriptTable	inst		= null;
	public static DevScriptTable get() { return inst; }
	
	public DevScriptTable(ModelScriptTable scriptTable)
	{
		super("scriptTable");
		inst = this;
		model = scriptTable;
		
		for (int row=0; row < ModelScriptTable.MAX_SCRIPTS; row++)
		{
			for (PV_SCRIPT pv : PV_SCRIPT.values())
			{
				registerDevicePV(pv.row(row));
			}
		}
		registrationComplete();
	}
	
	@Override protected void onAllPVsConnected()
	{
		//System.out.println("User table: Cleaning up clients...");
		//cleanupClients();
		
		//System.out.println("User table: Registering client...");
		//registerClient();
		
		readScriptTable();
	}
	
	@Override protected void valueChangedAction(String pvName, DBR value)
	{
		int updateRow=-1;
		PV_SCRIPT updatePV = null;
		
		for (int row=0; row < ModelScriptTable.MAX_SCRIPTS; row++)
		{
			for (PV_SCRIPT pv : PV_SCRIPT.values())
			{
				if (pv.row(row).toString().equals(pvName))
				{
					updateRow = row;
					updatePV = pv;
				}
			}
		}
		
		if (null == updatePV || updateRow < 0)
		{
			throw new IllegalStateException("Unknown PV: "+pvName);
		}
		
		switch (updatePV)
		{
		case name:	model.scriptTable[updateRow].name	= Epics.dbrToString(value); break;
		case pid:	model.scriptTable[updateRow].pid 	= Epics.dbrToInt(value); break;
		case host:	model.scriptTable[updateRow].host 	= Epics.dbrToString(value); break;
		case state:	model.scriptTable[updateRow].state 	= ScriptState.values()[Epics.dbrToInt(value)]; break;
		default:	break;
		}
		
		ModelScriptTable.get().ll.notifyListeners();
	}
	
	@Override public void abort() {}
}
