package gov.anl.gmca.hutch;

import gov.anl.gmca.common.EpicsDevice;
import gov.anl.gmca.io.Database;
import gov.anl.gmca.io.Database.DBName;
import gov.anl.gmca.io.Database.Result;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.PVName;
import gov.anl.gmca.io.PVLookup.PV_MISC;
import gov.anl.gmca.io.PVLookup.PV_SHUTTER;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.Util;
import gov.aps.jca.dbr.DBR;

import java.util.HashMap;
import java.util.HashSet;

public class DevShutter extends EpicsDevice
{
	protected static final int timeoutMs = 1000;

	public static void init()
	{
		Result pmaResult = Database.exec(DBName.CONFIG, "select * from EpicsShutter");
		for (int i=0;i<pmaResult.numRows();i++)
		{
			HashMap<String,String> row = pmaResult.getRow(i);

			String devName = row.get("name");

			// attributes for attenuators only
			String materialString = row.get("material");
			String thicknessString = row.get("thickness");

			// used to construct the two PV names
			String control = 	row.get("recordID");
			String readback =	row.get("readback");
			String channel = 	row.get("channel");

			// some filters are turned off if they're not installed
			boolean enable = row.get("enable").equals("1");

			if (null != materialString && materialString.length() > 0 && null != thicknessString && thicknessString.length() > 0)
			{
				Material material = Material.lookup(materialString);
				double thickness = Double.parseDouble(thicknessString);
				new DevFilter(devName, material, thickness, enable);
			}
			else
			{
				if (enable) new DevShutter(devName);
			}
		}
	}

	////////////////////////////////////////////////////// Listener ////////////////////////////////////////////
	
	public interface Listener
	{
		public abstract void handleOpenChange(DevShutter shutter, boolean open);
		public abstract void handleStateChange(DevShutter shutter, State state);
	}
	
	public abstract static class OpenListener implements Listener
	{
		@Override public void handleStateChange(DevShutter shutter, State state) {}
	}
	
	protected HashSet<Listener> listenerSet = new HashSet<Listener>();
	public void addListener(Listener listener)
	{
		System.out.println("Adding listener for device "+motorID);
		listenerSet.add(listener);
	}
	public void removeListener(Listener listener)
	{
		listenerSet.remove(listener);
	}

	////////////////////////////////////////////////////// Interface ////////////////////////////////////////////

	public DevShutter(String name)
	{
		super(name);
		
		registerDevicePV(PV_SHUTTER.control.shutter(name));
		registerDevicePV(PV_SHUTTER.readback.shutter(name));
		registrationComplete();
	}
	
	public static DevShutter get(String name)
	{
		return ((DevShutter)getDevice(name));
	}

	public boolean getIsOpen()
	{
		boolean isOpen = (1 == Epics.getInteger(PV_SHUTTER.readback.shutter(getDeviceName())));
		return isOpen;
	}
	
	public void open()
	{
		Epics.putInteger(PV_SHUTTER.control.shutter(getDeviceName()), 1);
	}
	
	public void close()
	{
		int fsRunPrg=-1;
		
		// this hard-codes the fast shutter name but the better solution would be to
		// create a whole new device class just for the fast shutter, since it's the only
		// shutter also looks like a PMAC motor?
		if (getDeviceName().equals("shutter") &&
			(fsRunPrg=Epics.getInteger(PV_MISC.fastShutterRunPrg))!=0)
		{
			final int f_fsRunPrg=fsRunPrg;
			new Thread(new Runnable(){public void run()
			{
				Log.note("Fast shutter runprg="+f_fsRunPrg+", sending abort before sending close");
				Epics.putInteger(PV_MISC.fastShutterAbort, 1);
				Util.sleep(100);
				Epics.putInteger(PV_SHUTTER.control.shutter(getDeviceName()), 0);
			}}).start();
		}
		else
		{
			Epics.putInteger(PV_SHUTTER.control.shutter(getDeviceName()), 0);
		}
	}
	
	@Override
	protected void onAllPVsConnected()
	{
		state = State.READY;
		for (Listener l : listenerSet)
		{
			l.handleStateChange(this, state);
		}
	}

	@Override
	protected void valueChangedAction(String pvName, DBR value)
	{
		if (pvName.equals(PV_SHUTTER.readback.shutter(getDeviceName()).toString()))
		{
			boolean isOpen = (1 == Epics.dbrToInt(value));
			for (Listener l : listenerSet)
			{
				l.handleOpenChange(this, isOpen);
			}
		}
	}

	public void waitFor()
	{
		long waitStartMs = System.currentTimeMillis();
		
		PVName controlPV = PV_SHUTTER.control.shutter(getDeviceName());
		PVName readbackPV = PV_SHUTTER.readback.shutter(getDeviceName());
		
		while (true)
		{
			boolean moveComplete = true;
			if (Epics.getInteger(controlPV) != Epics.getInteger(readbackPV)) moveComplete = false;
			if (getDeviceName().equals("shutter") && Epics.getInteger(PV_MISC.fastShutterRunPrg)!=0) moveComplete = false;
			if (moveComplete) break;

			Util.sleep(10);
			if (System.currentTimeMillis() - waitStartMs > timeoutMs)
			{
				Log.error("Shutter "+motorID+" timed out");
				return;
			}
		}
		Util.sleep(20); // wait for any other EPICS servers to receive the "closed" signal
	}

	@Override public void abort() {}
}
