package gov.anl.gmca.io;

import gov.anl.gmca.common.*;
import gov.anl.gmca.io.Database.DBName;
import gov.anl.gmca.io.Database.Result;
import gov.anl.gmca.raster.ModelRasterState.RasterListType;
import gov.anl.gmca.scan.ModelScan.TRACE_ID;

import java.time.chrono.*;
import java.util.*;

// All PV names should be determined in this file
//
// If they are generated elsewhere, they are easy to track down (see below)
//
public class PVLookup
{
	// Although PVName is a trivial interface, by only accepting PVName objects in the Epics class
	// we can keep track of any PV name creation and ensure it is only done in this file
	//
	// This interface also allows suffixes to be generated on the fly, but we are keeping
	// all suffixes as enums for now so they are easy to find.
	public static interface PVName
	{
		// each PV enum can decide how to construct the full PV name
		public String toString();
	}

	// covers simple prefixes that don't change based on e.g. which motor is being referenced
	private static boolean PREFIX_init = false;
	public static boolean simMode = false;
	public static HashMap<String,String> PREFIX_map = new HashMap<String,String>();
	public static PVPrefix[] simPrefixes = new PVPrefix[]{PVPrefix.SONICC,PVPrefix.FPE,PVPrefix.MAR};

	public static enum PVPrefix
	{
		BEAMLINE,		BLUICE,			COLLIMATOR,	COUNTER,		FPE,			MCA,			MOSTAB,		ROBOT,		RUN, 		SCRIPT,
		RUNS,			R_RUN,			R_RUNS,		SCR_RUN,		SCREENING,		SCAN,		USER, 		UNDULATOR,
		ANNEAL_FLOW,	ANNEAL_SHIFT, 	PMAC21,		FRAMEAUDIT,		KEITHLEY,		TRACE,		AUTOFIND,	ANALYSIS,
		MAR, 			STRATEGY,		FEEDBACK,	EIGER,			SW,				SONICC;
		
		public String getPrefix()
		{
			String prefix = PREFIX_map.get(toString().toLowerCase());
			
			// add the beamline prefix to all other prefixes
			if (this != BEAMLINE && this != UNDULATOR)
			{
				boolean useSimPrefix = DBConfig.sim5 && Arrays.asList(simPrefixes).contains(this);
				String beamlinePrefix = useSimPrefix ? "23sim:" : PREFIX_map.get("beamline");
				prefix = beamlinePrefix + prefix;
			}
			
			return prefix;
			
		}
		public String add(String suffix) { return getPrefix()+suffix; }
		
		PVPrefix()
		{
			if (!PREFIX_init)
			{
				PREFIX_init = true;
				
    			Result result = Database.exec(DBName.CONFIG, "select * from PVPrefix");
    			for (HashMap<String,String> row : result.getRows())
    			{
    				PREFIX_map.put(row.get("Label"), row.get("Prefix"));
    			}
			}
		}
	}

	public static enum PV_AUTOFIND implements PVName
	{
		coarseContour, loopZ, faceOnOmega, complete;
		
		@Override public String toString() { return PVPrefix.AUTOFIND.add(super.toString()); }
	}
	
	public static enum PV_BLUICE implements PVName
	{
		enableEndShutterCheck,		enableBeamstopCheck,
		enableEnergyGapCheck,		pointclick,					px2mm_h_lo$VAL,
		px2mm_v_lo$VAL,				px2mm_h_hi$VAL,				px2mm_v_hi$VAL,
		waitForFeedback, 			attenuation,				beamsize_h_act1$VAL,
		beamsize_v_act1$VAL,		beamsizeOption, 			currentCollimatorSet,
		beamsize_h, 				beamsize_v, 				autoCenterRun,
		beamsize, 					nopin,						nopin$B,
		alignDataV,					alignDataH,					alignState,
		alignLastRunTime, 			alignLastEnergy, 			alignLastMirrorV,
		alignLastMirrorH,			request,					response,
		robotWarmupPending,			data_dir,					sampleEnv,
		reloadSQL, 					displayProgram,				beam_xy,
		run0LastFrame0,				run0LastFrame1,				runtimeConfig,
		hdfCleanupMode, 			collectSpotFinder,			rasterSpotFinder,
		imageFormat, 				sfCollectOptions,			sfRasterOptions,
		imageServer,				enableEnergyMove,			enableDetectorMove,
		enableSampleEnvMove,		enableCryojetMonitor,		AlignBeamHistory;
		
		@Override public String toString()
		{
			String str = super.toString().replace('$', '.');
			String[] simPVs = new String[]{
					"pointclick",
					"px2mm_h_lo.VAL","px2mm_h_hi.VAL",
					"px2mm_v_lo.VAL","px2mm_v_hi.VAL",
					"data_dir"
			};
			boolean isSimPV = Arrays.asList(simPVs).contains(str);
			if (DBConfig.sim5 && isSimPV) return "23sim:bi:"+str;
			return PVPrefix.BLUICE.add(str);
		}
		public PVName getBeamsizePV (int size)
		{
			final String str = super.toString()+size;
			return new PVName()
			{
				@Override public String 	toString()
				{
					return PVPrefix.BLUICE.add(str);
				}
			};
		}
	}
	
	public static enum PV_COLLIMATOR implements PVName
	{
		ActivePreset, 	RunPrg;
		
		@Override public String toString() { return PVPrefix.COLLIMATOR.add(super.toString()); }
	}
	
	public static enum PV_FEEDBACK implements PVName
	{
		serialLIA$TINP, serialLIA$AOUT, uselockin;
		
		@Override public String toString() { return PVPrefix.FEEDBACK.add(super.toString().replace("$",".")); }
	}
	
	public static enum PV_FRAMEAUDIT implements PVName
	{
		sts, 	msg;
		@Override public String toString() { return PVPrefix.FRAMEAUDIT.add(super.toString()); }
	}
	
	public static enum PV_FPE implements PVName
	{
		abort,			axisMotor,		cmd_output256,		delta,		
		directory256,	energy,			etc,				exposureTime,
		filename, 		frameNumber,	lastfilename,		lastfiledir256,
		lastfileupdate,	modeIndex,		operationStatus,	runNumber, 	
		startAngle, 	start, 			logMessage256, 		nextAngle,
		mode256, 		viewFile256,	viewDir256,			viewUpdate,
		fileCheck;
		
		@Override public String toString() { return PVPrefix.FPE.add(super.toString().replace("$",".")); }
	}

	private static boolean HS_init = false;
	private static HashMap<String,String>
		HS_openMap  = new HashMap<String,String>(),
		HS_closeMap = new HashMap<String,String>(),
		HS_statMap  = new HashMap<String,String>();
	
	public static enum PV_HUTCH_SHUTTER
	{
		stat, open, close;
		
		PV_HUTCH_SHUTTER()
		{
			if (!HS_init)
			{
				HS_init=true;
				
				Result result = Database.exec(DBName.CONFIG, "select * from HutchShutter");
				for (HashMap<String,String> row : result.getRows())
				{
					String shutter = row.get("Shutter");
					HS_openMap.put(		shutter, row.get("openPV"));
					HS_closeMap.put(	shutter, row.get("closePV"));
					HS_statMap.put(		shutter, row.get("statPV"));
				}
			}
		}
		
		public PVName shutter(final String shutterID)
		{
			final PV_HUTCH_SHUTTER pv = this;
			return new PVName()
			{
				@Override public String toString()
				{
					switch (pv)
					{
					case stat:	return HS_statMap.get(shutterID);
					case open:	return HS_openMap.get(shutterID);
					case close:	return HS_closeMap.get(shutterID);
					default:	throw new IllegalStateException();
					}
				}
			};
		}
	}

	// Usage:
	// PV_KEITHLEY.Set
	// PV_KEITHLEY.ActGain.index(3)
	//
	public static enum PV_KEITHLEY implements PVName
	{
		ActGain, Set;
		
		@Override public String toString()
		{
		    if (super.toString().equals("Set"))
		    {
		    	return PVPrefix.KEITHLEY.getPrefix().replace("#","") + "Set";
		    }
		    else
		    {
		    	return super.toString();
		    }
		}
		
		public PVName index(final int keithleyIndex)
		{
			final String enumString = toString();
			return new PVName()
			{
				@Override public String toString()
				{
					return PVPrefix.KEITHLEY.getPrefix().replace("#", ""+keithleyIndex) + enumString;
				}
			};
		}
	}
	
	public static enum PV_MAR implements PVName
	{
		DetCmdStr, DetInStr, DetOutStr, AsynIO$CNCT, FPE_state;
		
		@Override public String toString()
		{
			String str = super.toString().replace('$', '.');
			return PVPrefix.MAR.add(str);
		}
	}
	
	public static enum CompressionAlgo { LZ4, BSLZ4 };
	public static enum StreamHeaderDetail { NONE, BASIC, ALL };
	
	public static enum PV_EIGER implements PVName
	{
		DetCmdStr, DetInStr, DetOutStr, AsynIO$CNCT, FPE_state,
		ArrayCounter_RBV, ArrayCounter, Acquire_RBV, ImageId, PhotonEnergy, PhotonEnergy_RBV,
		StatusMessage_RBV, CompressionAlgo, CompressionAlgo_RBV, StreamHeaderDetail,
		StreamHeaderDetail_RBV, TriggerMode;
		
		@Override public String toString()
		{
			String str = super.toString().replace('$', '.');
			return PVPrefix.EIGER.add(str);
		}
	}
	
	public static enum PV_MCA implements PVName
	{
		EraseStart,	Erase,		Start,		Stop,
		GAIN,		OFFSET,		_ACQG,		_CALO,
		_CALS,		_ELTM,		_ERTM,		_NMAX,
		_NUSE,		_PLTM,		_PRTM,		_PCTL,
		_PCTH,		_R0LO,		_R0HI,		_R0IP,
		_READ,		_VAL;
		
		@Override public String toString()
		{
			boolean isFalcon = PVPrefix.MCA.getPrefix().toLowerCase().contains("falcon");
			String str = super.toString();

			if (isFalcon) {
				// map PV names to falcon
				if (str.equals("EraseStart")) str=".ERST";
				else if (str.equals("Erase")) str=".ERAS";
				else if (str.equals("Start")) str = ".STRT";
				else if (str.equals("Stop")) str=".STOP";
			}
			str = str.replace('_','.');
			return PVPrefix.MCA.add(str);
		}
	}

	protected static boolean MISC_init = false;
	protected static HashMap<String,String> MISC_map = new HashMap<String,String>();
	
	public static enum PV_MISC implements PVName
	{
		apsCurrent, 		apsShutterStat,			localIZero,			verticalPiezoTweak,
		stationSearched, 	dcm, 					Iabs, 				verticalPiezo, 	
		verticalPiezoLowerLimit, verticalPiezoHigherLimit,	horizontalPiezo, 	horizontalPiezoTweak, 	
		horizontalPiezoLowerLimit, horizontalPiezoHigherLimit, 	backlightInt, 		backlightIntTweak,
		backlightLowerLimit, backlightHigherLimit,  lowlightInt, 		lowlightIntTweak, 		
		lowlightLowerLimit, lowlightHigherLimit, 	highlightInt, 		highlightIntTweak, 
		highlightLowerLimit, highlightHigherLimit, 	ringlightInt, 		ringlightIntTweak,		
		ringlightLowerLimit, ringlightHigherLimit,	pindiode_actpos, 	pindiode_mem1, 			pindiode_mem2, 		pindiode_run,			
		collimator_mem1, 	collimator_mem2, 		collimator_mem3, 	collimator_mem4,		
		collimator_mem5, 	p2Pos,					ctrlVoltage, 		bpm1, 					p2rqsPos,
		fastShutterRunPrg,	fastShutterAbort,		mostab_readback,		cryojetTempReadback, cryojetFlowInner,
		falcon_roi_copy;
		
		PV_MISC()
		{
			if (!MISC_init)
			{
				MISC_init = true;
				
				Result result = Database.exec(DBName.CONFIG, "select * from MiscPV");
				for (HashMap<String,String> row : result.getRows())
				{
					String id = row.get("ID");
					MISC_map.put(id, row.get("FullPVName"));
				}
			}
		}
		
		@Override public String toString()
		{
			return MISC_map.get(super.toString());
		}
	}
	
	public static enum PV_MOSTAB implements PVName
	{
		pause,		status,		statusline;
		
		@Override public String toString() { return PVPrefix.MOSTAB.add(super.toString()); }
	}
	
	// dev and ctl prefixes for each motor
	private static boolean PMAC_init = false;
	private static HashMap<String,String> PMAC_dev_map = new HashMap<String,String>();
	private static HashMap<String,String> PMAC_ctl_map = new HashMap<String,String>();
	
	public static enum PV_PMAC
	{
		// Motor PVs
		ActPos(0),		RqsPos(0),			AmpEna(0),			PsLimSet(0),		NgLimSet(0),
		ActPos_EGU(0),	RqsPos_DRVL(0),		RqsPos_DRVH(0),		RqsBkl(0),			speed(0),
		
		// Control PVs
		Abort(1),		InPos(1),			Start(1),			Motion(1),			Homed(1),
		AutoStart(1),	AccelTime(1),		AccelCurve(1),		RunPrg(1),			speedSDIS(1),
		Sync_PROC(1);
		
		protected boolean isControlPV;
		
		PV_PMAC(int pvType)
		{
			isControlPV = (1==pvType);
			if (!PMAC_init)
			{
				PMAC_init = true;

				Result result = Database.exec(DBName.CONFIG, "select * from PmacMotorAssembly");
				for (HashMap<String,String> row : result.getRows())
				{
					String motorName = row.get("name");

					String beamlinePrefix = DBConfig.sim5 ? "23sim" : row.get("beamline");
					PMAC_dev_map.put(motorName,beamlinePrefix+row.get("device")+row.get("motor"));
					PMAC_ctl_map.put(motorName,beamlinePrefix+row.get("device")+row.get("control"));
				}
			}
		}
		
		public PVName motor(final String motorID)
		{
			final PV_PMAC pv = this;
			return new PVName()
			{
				@Override public String 	toString()
				{
					if (isControlPV)	return PMAC_ctl_map.get(motorID)+pv.toString();
					else				return PMAC_dev_map.get(motorID)+pv.toString();
				}
			};
		}
		
		@Override public String toString() { return super.toString().replace('_', '.'); }
	}
	
	public static enum PV_PMAC21 implements PVName
	{
		// beamstop backlight
		bo14_VAL, 	bi21_VAL, 	bo13_VAL, 	bi20_VAL;
		
		@Override public String toString() { return PVPrefix.PMAC21.add(super.toString().replace('_', '.')); }
		
	}
	
	public static enum PV_ROBOT implements PVName
	{
		sta,	sta$1,	sta$2, 	mounted,	puckType, 	
		mnt$0,	pause,	cmd, w_tim;
		
		//@Override public String toString() { return PVPrefix.ROBOT.add(super.toString()); }
		@Override public String toString() { return PVPrefix.ROBOT.add(super.toString().replace('$', '-')); }
	}
	public static enum PV_ANALYSIS implements PVName
	{
		currentDataProcIndex, 	dataprocUpdated, 	updateXDSparams,
		// the following is used by new analysis
		pipelineStatus;
		
		@Override public String toString() { return PVPrefix.ANALYSIS.add(super.toString()); }
	}
	public static enum PV_STRATEGY implements PVName
	{
		// new strategy pvs by QX 10/17/2019
		strategyLastXtal, strategyJobStatus,
		// old strategy pvs
		currentMultiStrategyIndex, 	multistrategyUpdated, isReference, referenceDataUpdated,
		referencedata, unitCell, spaceGroup, strategyImages_VAL, lowRes, highRes, mergeDatasetsUpdated, showMultiStrategy;
		
		@Override public String toString() { return PVPrefix.STRATEGY.add(super.toString().replace('_', '.')); }
	}
	public static enum PV_RUN_LIMITS implements PVName
	{
		wedgeSize_DRVL,		wedgeSize_DRVH,		exposureTime_DRVL,
		exposureTime_DRVH,	delta_DRVL,			delta_DRVH;
		
		// collect limits are stored in run 0
		public PVName collectLimit()
		{
			final PV_RUN_LIMITS pv = this;
			return new PVName(){@Override public String toString()
			{
				String prefix = PVPrefix.RUN.getPrefix().replace('#', '0');
				String suffix = pv.toString().replace('_', '.');
				return prefix + suffix;
			}};
		}
		// raster limits are stored in run 1
		public PVName rasterLimit()
		{
			final PV_RUN_LIMITS pv = this;
			return new PVName(){@Override public String toString()
			{
				String prefix = PVPrefix.R_RUN.getPrefix().replace("#", "15").replace("$","");
				String suffix = pv.toString().replace('_', '.').replace("delta","delta_deg").replace("exposureTime", "exposure_sec");
				return prefix + suffix;
			}};
		}
	}
	
	public static enum PV_RUNS implements PVName
	{
		runCount,		currentRun,		collectState,
		currentLeftTab,	updated, 		currentUser,
		binMode,		bgProc, 		enableShutterless,
		hideCollectSF,	hideSampleSF,	traceVisCollect,
		traceVisSample, importDBUpdated;
		
		@Override public String toString() { return PVPrefix.RUNS.add(super.toString()); }
	}
	
	public static enum PV_RUN
	{
        runStatus,		runLabel,		fileroot,		directory256,
        beamstop_mm,	atten_factor,	axisMotor,    	delta,	
        exposureTime,	startAngle,		endAngle,     	startFrameLabel, 
        pt1_x_mm,   	pt1_y_mm,		pt1_z_mm,   	pt1_set,	
        pt2_x_mm,     	pt2_y_mm,		pt2_z_mm,     	pt2_set,	
        pt3_x_mm,     	pt3_y_mm,		pt3_z_mm,     	pt3_set,	
        pt4_x_mm,     	pt4_y_mm,		pt4_z_mm,     	pt4_set,
        nextFrame,    	reverseDir,		stepSize_um,  	framesPerSite,	
        overlap,   		inverseOn,		wedgeSize,    	numEnergy,
        energy1,      	energy2,		energy3,      	energy4,
        energy5,       	distance,		savedEndAngle,	strategyEnabled,
        startAngle2,	scanMode,		optionTab,		selectedSites,
        updated,        xdsProc, 		siteDelta_deg,	siteDeltaEnable,
        savedFPS,		savedOverlap,	setupStep,		vectorFillMode,
        potVecSiteCount,savedInverseBeam, beamHeight_um,beamWidth_um,
        helicalMode,	siteSkip, 		dbUpdated,		importSiteCount,
        importPath256,	meshFilter,		meshCombine,	framesPerDir,
		rasterRun,		meshParam,		rasterNumbering,meshSpotFinder;
        
        public static PV_RUN energy(int energy)
        {
        	return valueOf("energy"+energy);
        }
        
		public PVName run(final int runIndex)
		{
			final String enumString = toString();
			return new PVName()
			{
				@Override public String toString()
				{
					return PVPrefix.RUN.getPrefix().replace("#", ""+runIndex) + enumString;
				}
			};
		}
	}
		
	public static enum PV_R_RUNS implements PVName
	{
		runCount,		currentRun,			updated,		autoRunCount,
		autoCurrentRun,	autoUpdated,		selectedCells, 	rasterState,
		autoMode, 		siteListUpdated, 	macroUpdated, 	currentMacro,
		scrClear,		binMode,			fixResLimits, 	hitDBUpdated,
		viewMode,		selectedCell;
		
		@Override public String toString() { return PVPrefix.R_RUNS.add(super.toString()); }
	}
	
	public static enum PV_R_RUN
	{
	    runStatus,				runType,			nextFrame,			runLabel,
	    fileroot,				directory256,   	roiMode,			savedZoom,
	    delta_deg,				exposure_sec,		distance_mm,		energy_keV,
	    atten_factor, 			gridLeftOffset_um,	gridRightOffset_um,	cellWidth_um,
	    cellHeight_um,			isCenterSet,		center_omega_deg,	center_x_mm,
	    center_y_mm,  			center_z_mm,      	startFrameLabel,	contours,
	    selectedCells,			beamWidth_um,		beamHeight_um,		gridTopOffset_um,
	    gridBottomOffset_um,	referenceRun,		referenceHit,		refinementType,
	    userOffsetX,			userOffsetY,		userOffsetSet,		loopMeasurement_um,
	    hideRunParameters,		updated,			userOffsetZ,		userOffsetOmega,
	    autoFind,				dbUpdated, 			lowResLimit_A,		highResLimit_A,
	    loopCentered,			soniccBin,			soniccRefPos;
		
		public PVName run(final RasterListType rType, final int runIndex)
		{
			final String enumString = toString();
			return new PVName()
			{
				@Override public String toString()
				{
					boolean isAuto = (rType == RasterListType.AUTO);
					
					String ret = PVPrefix.R_RUN.getPrefix().replace("#", ""+runIndex);
					ret = ret.replace("$", isAuto?"a":"");
					ret += enumString;
					
					return ret;
				}
			};
		}
	}
	public static Object[][] pvMap = { 
		{ "runStatus", PV_SCR_RUN.runStatus, PV_RUN.runStatus },
		{ "fileroot", PV_SCR_RUN.fileroot, PV_RUN.fileroot },
		{ "directory256", PV_SCR_RUN.directory256, PV_RUN.directory256 },
		{ "beamstop_mm", PV_SCR_RUN.beamstop_mm, PV_RUN.beamstop_mm },
		{ "atten_factor", PV_SCR_RUN.atten_factor, PV_RUN.atten_factor },
		{ "axisMotor", PV_SCR_RUN.axisMotor, PV_RUN.axisMotor },
		{ "delta", PV_SCR_RUN.delta, PV_RUN.delta },
		{ "exposureTime", PV_SCR_RUN.exposureTime, PV_RUN.exposureTime },
		{ "startAngle", PV_SCR_RUN.startAngle, PV_RUN.startAngle },
		{ "startFrameLabel", PV_SCR_RUN.startFrameLabel, PV_RUN.startFrameLabel },
		{ "start_x_mm", PV_SCR_RUN.start_x_mm, PV_RUN.pt1_x_mm },
		{ "start_y_mm", PV_SCR_RUN.start_y_mm, PV_RUN.pt1_y_mm },
		{ "start_z_mm", PV_SCR_RUN.start_z_mm, PV_RUN.pt1_z_mm },
		{ "start_set", PV_SCR_RUN.start_set, PV_RUN.pt1_set },
		{ "end_x_mm", PV_SCR_RUN.end_x_mm, PV_RUN.pt2_x_mm },
		{ "end_y_mm", PV_SCR_RUN.end_y_mm, PV_RUN.pt2_y_mm },
		{ "end_z_mm", PV_SCR_RUN.end_z_mm, PV_RUN.pt2_z_mm },
		{ "end_set", PV_SCR_RUN.end_set, PV_RUN.pt2_set },
		{ "nextFrame", PV_SCR_RUN.nextFrame, PV_RUN.nextFrame },
		{ "reverseDir", PV_SCR_RUN.reverseDir, PV_RUN.reverseDir },
		{ "stepSize_um", PV_SCR_RUN.stepSize_um, PV_RUN.stepSize_um },
		{ "framesPerSite", PV_SCR_RUN.framesPerSite, PV_RUN.framesPerSite },
		{ "overlap", PV_SCR_RUN.overlap, PV_RUN.overlap },
		{ "inverseOn", PV_SCR_RUN.inverseOn, PV_RUN.inverseOn },
		{ "wedgeSize", PV_SCR_RUN.wedgeSize, PV_RUN.wedgeSize },
		{ "numEnergy", PV_SCR_RUN.numEnergy, PV_RUN.numEnergy },
		{ "energy1", PV_SCR_RUN.energy1, PV_RUN.energy1 },
		{ "energy2", PV_SCR_RUN.energy2, PV_RUN.energy2 },
		{ "energy3", PV_SCR_RUN.energy3, PV_RUN.energy3 },
		{ "energy4", PV_SCR_RUN.energy4, PV_RUN.energy4 },
		{ "energy5", PV_SCR_RUN.energy5, PV_RUN.energy5 },
		{ "distance", PV_SCR_RUN.distance, PV_RUN.distance },
		{ "strategyEnabled", PV_SCR_RUN.strategyEnabled, PV_RUN.strategyEnabled },
		{ "startAngle2", PV_SCR_RUN.startAngle2, PV_RUN.startAngle2 },
		{ "scanMode", PV_SCR_RUN.scanMode, PV_RUN.scanMode },
		{ "optionTab", PV_SCR_RUN.optionTab, PV_RUN.optionTab },
		{ "updated", PV_SCR_RUN.updated, PV_RUN.updated },
		{ "beamWidth_um", PV_SCR_RUN.beamWidth_um, PV_RUN.beamWidth_um },
		{ "beamHeight_um", PV_SCR_RUN.beamHeight_um, PV_RUN.beamHeight_um }
	};
	public static enum PV_SCR_RUN
	{
        runStatus,		runLabel,		fileroot,		directory256,
        beamstop_mm,	atten_factor,	axisMotor,    	delta,	
        exposureTime,	startAngle,		endAngle,     	startFrameLabel, 
        start_x_mm,   	start_y_mm,		start_z_mm,   	start_set,	
        end_x_mm,     	end_y_mm,		end_z_mm,     	end_set,	
        nextFrame,    	reverseDir,		stepSize_um,  	framesPerSite,	
        overlap,   		inverseOn,		wedgeSize,    	numEnergy,
        energy1,      	energy2,		energy3,      	energy4,
        energy5,       	distance,		numberOfSites,	strategyEnabled,
        startAngle2,	scanMode,		optionTab,		updated,
        beamHeight_um,	beamWidth_um;
        
        public static PV_SCR_RUN energy(int energy)
        {
        	return valueOf("energy"+energy);
        }
        public static PVName getPV(String scr, String pv) {
        	PVName tempPV = null;
        	for(Object[] row : pvMap) {
        		if(pv.equals(row[0])) {
        			if(scr.equals("_scr"))
        				tempPV = (PVName)row[1];
        			else
        				tempPV = (PVName)row[2];
        		} 
        	}
        	return tempPV;
        }
		public PVName run(final int runIndex)
		{
			final String enumString = toString();
			return new PVName()
			{
				@Override public String toString()
				{
					return PVPrefix.SCR_RUN.getPrefix().replace("#", ""+runIndex) + enumString;
				}
			};
		}
	}
	
	public static enum PV_SCR implements PVName
	{
		spreadsheetFile,	spreadsheetStatus,
		spreadsheetDir256, 	prefix, 					dbUpdated,
		MountNextXtal, 		CenterXtal, 				Pause4ctrng, 				
		TakeImage90, 		Pause4collect, 				RunWebIce, 
		AngleImage0,		AngleImage45, 				AngleImage90, 
		TakeImage0, 		TakeImage45,		 		ScreeningStatus, 	
		CurrentTask, 		spreadsheetSelected_VAL,	spreadsheetScreened_VAL,
		directory256,		SILID,						spreadsheetStatus_PROC,
		sessionID,			currentStrategyIndex, 		silUpdated, 
		currentDataProcIndex, dataprocUpdated,			strategyProgram,
		centerType,			diffimages0,				diffimages45,
        diffimages90,		takeJpeg, 					scalingSoftware,
        sampleNumber, 		enablePuckOwnership,		pucksUpdated;
		
		@Override public String toString() { return PVPrefix.SCREENING.add(super.toString().replace('_', '.')); }
		//@Override public String toString() { return super.toString().replace('_', '.'); }
	}

	public enum PV_SONICC implements PVName
	{
		omegaOffset,			// degrees of omega offset of the image (0=on axis)
		ycenter_px, xcenter_px, // high-res camera center in x/y pixels on the sonicc image
		ypixel2um, xpixel2um,	// um/px of the image
		axes_rotation,			// rotate the image by this many degrees, then use xcenter/ycenter
		focusOffset,			// um to move sample along the beam to put its center in the sonicc FOV
		serverCMD;				// send commands to this waveform PV, listen for responses

		@Override public String toString() {
			String pvSuffix = super.toString();
			if (!pvSuffix.equals("serverCMD")) pvSuffix += ":SHG";
			return PVPrefix.SONICC.add(pvSuffix);
		}
	}

	public static enum PV_A_SHIFT implements PVName
	{
		Anneal, Cancel, A_Time,	inboard;
		@Override public String toString() { return PVPrefix.ANNEAL_SHIFT.add(super.toString()); }
	}
	
	public static enum PV_A_FLOW implements PVName
	{
		Anneal, Cancel, A_Time;
		@Override public String toString() { return PVPrefix.ANNEAL_FLOW.add(super.toString()); }
	}
	
	public static enum PV_SCAN implements PVName
	{
		interStep, 		edge, 				edgeEnergy, 		time, 				spectrumTime,
		roiLow, 		roiHigh,			choochInflection,	choochPeak,			choochRemote,
		currentTab,		autoStep,			rightTab,			graphTitle,			graphMessage,
		graphXMax,		graphXLabel,		choochPDF,			autoOptimized,		scanCenter,
		prefix,			directory,			graphXMin,			fastscanData,		plotMode,
		leftAxisTrace,	rightAxisTrace,		adaptiveScan,		update,				displayedElements,
		otherElement,	spectrumEnergy,		defaultInflection,	defaultPeak,		defaultRemote,
		defaultXMin,	defaultXMax,        otherElement2, 		showAllEdges,		choochInflFp,
		choochInflFpp,	choochPeakFp,		choochPeakFpp;
		
		@Override public String toString() { return PVPrefix.SCAN.add(super.toString()); }
	}
	
	public static enum PV_SCAN_TRACE
	{
		graphTraceX, 	graphTraceY, 	graphYMin, 	graphYMax, 	graphYLabel, 	graphTraceCount,
		isVisible,		defaultYMin,	defaultYMax;
		
		public PVName id(final TRACE_ID id)
		{
			final String enumString = toString();
			return new PVName()
			{
				protected TRACE_ID _id = id;
				@Override public String toString()
				{
					int traceNum = _id.ordinal()+1;
					return PVPrefix.TRACE.getPrefix().replace("#", ""+traceNum) + enumString;
				}
			};
		}
	}
	
	public static enum PV_SCRIPT
	{
		name,		pid,		host,
		state,		heartbeat;
		
		public PVName row(final int row)
		{
			final String enumString = toString();
			return new PVName()
			{
				protected int _row = row;
				@Override public String toString()
				{
					String prefix = PVPrefix.SCRIPT.getPrefix().replace("#", Integer.toString(_row));
					return prefix + enumString;
				}
			};
		}
	}

	// control and readback PVs for each shutter
	private static boolean SHUTTER_init = false;
	private static HashMap<String,String> SHUTTER_ctl_map = new HashMap<String,String>();
	private static HashMap<String,String> SHUTTER_rbk_map = new HashMap<String,String>();
	
	public static enum PV_SHUTTER
	{
		control, readback;
		
		PV_SHUTTER()
		{
			if (!SHUTTER_init)
			{
				SHUTTER_init = true;
				
				Result result = Database.exec(DBName.CONFIG, "select * from EpicsShutter");
				for (int i=0;i<result.numRows();i++)
				{
					HashMap<String,String> row = result.getRow(i);
					
					String devName = row.get("name");
					
					// used to construct the two PV names
					String control = 	row.get("recordID");
					String readback =	row.get("readback");
					String channel = 	row.get("channel");
					
					if (control.charAt(0)==':') 	control = control.substring(1);
					if (readback.charAt(0)==':') 	readback = readback.substring(1);
					
					// some filters are turned off if they're not installed
					boolean enable = row.get("enable").equals("1");
					
					//if (enable)
					{
						SHUTTER_ctl_map.put(devName, PVPrefix.BEAMLINE.add(control+channel));
						SHUTTER_rbk_map.put(devName, PVPrefix.BEAMLINE.add(readback+channel));
					}
				}
			}
		}
		
		public PVName shutter(final String shutterID)
		{
			final PV_SHUTTER pv = this;
			return new PVName()
			{
				@Override public String 	toString()
				{
					switch (pv)
					{
					case control:	return SHUTTER_ctl_map.get(shutterID);
					case readback:	return SHUTTER_rbk_map.get(shutterID);
					default:		throw new IllegalArgumentException("Invalid PV: "+pv.toString());
					}
				}
			};
		}
	}

	
	public static enum PV_SW implements PVName
	{
		status,		filename;
		
		@Override public String toString() { return PVPrefix.SW.add(super.toString()); }
	}
	
	public static enum PV_UNDULATOR implements PVName
	{
		Busy_VAL,		Energy_VAL,		EnergySet_VAL,
		HarmonicValue,	Start_VAL,		Stop_VAL;
		
		@Override public String toString()
		{
			return PVPrefix.UNDULATOR.add(super.toString()).replace('_', '.');
		}
	}
	
	public static enum PV_USER
	{
		mode,		pid,		host,
		user,		state;
		
		public PVName row(final int row)
		{
			final String enumString = toString();
			return new PVName()
			{
				protected int _row = row;
				@Override public String toString()
				{
					String prefix = PVPrefix.USER.getPrefix().replace("#", Integer.toString(_row));
					return prefix + enumString;
				}
			};
		}
	}
	
	//////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////// logged PV list ////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////

	// these PV's values are printed to the file log whenever they change
	public static final PVName[] loggedPVs = new PVName[]
	{
		PV_FPE.operationStatus, PV_FPE.start, PV_FPE.cmd_output256,
		PV_FPE.logMessage256, PV_MAR.FPE_state, PV_FPE.abort,
		PV_FPE.fileCheck, PV_PMAC.InPos.motor("gonio_omega"),
		PV_EIGER.ArrayCounter_RBV, PV_EIGER.ImageId,
		PV_EIGER.Acquire_RBV,
		PV_PMAC.RunPrg.motor("gonio_omega"),
		PV_R_RUN.contours.run(RasterListType.INTER, 0),
		PV_R_RUN.contours.run(RasterListType.INTER, 5),
		PV_SCAN.displayedElements
	};
	
	public static HashSet<String> waveformPVs = new HashSet<>();
	
	public static void initWaveformPVs()
	{
		waveformPVs.add(PV_SCAN.displayedElements.toString());
		//waveformPVs.add(PV_BLUICE.sampleEnv.toString());
		
		for (int i=0;i<16;i++)
		{
			waveformPVs.add(PV_R_RUN.contours.run(RasterListType.AUTO, i).toString());
			waveformPVs.add(PV_R_RUN.contours.run(RasterListType.INTER, i).toString());
		}
	};
	
	public static boolean isWaveform(String pvName)
	{
		return waveformPVs.contains(pvName);
	}
}
