package gov.anl.gmca.io.hw;

import gov.anl.gmca.common.*;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.Util;

import java.sql.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;

public class Database extends gov.anl.gmca.io.Database
{
	/////////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////// settings ////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////
	
	protected static final int RETRY_MAX = 5;

	// TODO: move this to a config file?
	protected static String[] sqlHostMap =
	{
		"bl1upper",
		"bl2upper",
		"bl3upper"
	};
	
	public static String getSqlHost()
	{
		return sqlHostMap[Util.getBeamlineNumberFromHostname()];
	}
	
	protected static final String sqlUsername 	= "dhs";
	protected static final String sqlPassword 	= "";
	
	/////////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////// query string builders ///////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////
	
	public static String q_create(String tableName, ArrayList<DBField> fields)
	{
		ArrayList<DBField> primaryKeys = new ArrayList<DBField>(), foreignKeys = new ArrayList<DBField>();
		
		String stmt = "CREATE TABLE `"+tableName+"` (";
		
		for (DBField field : fields)
		{
			stmt += "`"+field.fieldName+"` ";
			if 		(field.dataType==SQLDataType.BOOL)			stmt += "TINYINT"; 					// 8-bit
			else if (field.dataType==SQLDataType.INT)			stmt += "BIGINT"; 					// 32-bit
			else if (field.dataType==SQLDataType.AUTOINC)		stmt += "BIGINT AUTO_INCREMENT";	// set to 0 when sending
			else if (field.dataType==SQLDataType.FLOAT)			stmt += "FLOAT"; 					// single precision
			else if (field.dataType==SQLDataType.STRING)		stmt += "VARCHAR(255)";				// 255 char max
			else if (field.dataType==SQLDataType.VARCHAR)		stmt += "VARCHAR(10000)";			// higher max
			else if (field.dataType==SQLDataType.TIMESTAMP)		stmt += "DATETIME";					// e.g. "2013-04-22 12:01:00"
			else if (field.dataType.toString().startsWith("ENUM_"))
			{
				stmt += "ENUM(";
				for (String enumString : field.dataType.values)
				{
					stmt += "'"+enumString+"',";
				}
				stmt = stmt.substring(0,stmt.length()-1); // remove last comma
				stmt += ")";
			}
			else throw new IllegalArgumentException("Unsupported field type "+field);

			if (!field.allowNull) stmt += " NOT NULL";
			
			stmt += ",";
			
			// track which keys are primary
			if (field.primaryKey) primaryKeys.add(field);
			
			// track foreign keys
			if (null != field.foreignKey) foreignKeys.add(field);
		}
		stmt=stmt.substring(0,stmt.length()-1); // remove last comma
		
		// add primary key clause if specified
		if (primaryKeys.size()>0)
		{
			stmt += ", PRIMARY KEY (";
			for (DBField pkField : primaryKeys)
			{
				stmt += "`"+pkField.fieldName+"`,";
			}
			stmt=stmt.substring(0,stmt.length()-1); // remove last comma
			stmt += ")";
		}
		
		// add foreign key clauses if specified
		for (DBField fkField : foreignKeys)
		{
			stmt += ", FOREIGN KEY ("+fkField.fieldName+") ";
			stmt += "REFERENCES "+fkField.foreignKey.tableName+" ("+fkField.foreignKey.fieldName+")";
		}
		
		stmt += ");";
		
		return stmt;
	}
	
	public static String q_delete(String table, WhereList whereList)
	{
		String stmt = "DELETE FROM "+table;
		stmt += constructWhereClause(whereList);
		stmt += ";";
		return stmt;
	}
	
	public static String q_select(String table, ArrayList<String> requestedColumns, WhereList whereList, String orderBy)
	{
		String stmt = "SELECT ";
		if (requestedColumns==null) stmt += "* ";
		else
		{
			for (String col : requestedColumns)
			{
				stmt += col+",";
			}
			stmt=stmt.substring(0,stmt.length()-1)+" ";
		}
		stmt += "FROM "+table;
		stmt += constructWhereClause(whereList);
		
		if (null != orderBy)
		{
			stmt += " ORDER BY "+orderBy;
		}
		
		stmt += ";";
		return stmt;
	}
	
	public static String q_insert(String table, ArrayList<HashMap<String,String>> rows)
	{
		// single row non-standard syntax
		// (id1=val1, id2=val2...) instead of (id1, id2...) values (val1, val2...)
		/*String stmt = "";
		for (HashMap<String,String> row : rows)
		{
			stmt += "INSERT INTO "+table+" VALUES (";
			for (String key : row.keySet())
			{
				stmt += key+"='"+row.get(key)+"',";
			}
			stmt = stmt.substring(0, stmt.length()-1); // remove last comma
			stmt += ");";
		}*/
		
		// multi-row standard syntax

		// first choose an order for the columns, since the hashmaps don't have one
		ArrayList<String> columnOrder = new ArrayList<String>();
		for (String colName : rows.get(0).keySet())
		{
			columnOrder.add(colName);
		}
		
		String stmt = "INSERT INTO "+table+" (";
		
		for (String colName : columnOrder)
		{
			stmt += colName+",";
		}
		stmt = stmt.substring(0, stmt.length()-1); // remove last comma
		stmt += ") VALUES ";
		
		//int rowIndex=0;
		for (HashMap<String,String> row : rows)
		{
			//if (rowIndex++%10==0) System.out.println("On row "+(rowIndex+1));
			stmt += "(";
			for (String column : columnOrder) // use the chosen column order
			{
				String value = row.get(column);
				if (null!=value) value=value.replace('\'', '"');
				stmt += "'"+value+"',";
			}
			stmt = stmt.substring(0, stmt.length()-1); // remove last comma
			stmt += "),";
		}
		stmt = stmt.substring(0, stmt.length()-1); // remove last comma
		stmt += ";";
		
		return stmt;
	}
	
	public static String q_insert(String table, DBRows rows)
	{
		StringBuilder stmt = new StringBuilder(rows.values.length*50);
		stmt.append("INSERT INTO "+table+" (");
		
		for (String colName : rows.columns)
		{
			stmt.append(colName+",");
		}
		stmt.setLength(stmt.length()-1); // remove last comma
		stmt.append(") VALUES ");
		
		int rowIndex=0;
		for (String[] rowValues : rows.values)
		{
			if (++rowIndex%100==0) System.out.println("Adding row "+rowIndex);
			stmt.append("(");
			for (String value : rowValues) // use the chosen column order
			{
				if (null!=value) value=value.replace('\'', '"');
				stmt.append("'"+value+"',");
			}
			stmt.setLength(stmt.length()-1);
			stmt.append("),");
		}
		stmt.setLength(stmt.length()-1); // remove last comma
		stmt.append(";");
		
		return stmt.toString();
	}
	
	public static String q_update(String tableName, HashMap<String,String> setMap, WhereList whereList)
	{
		// update
		String query = "UPDATE "+tableName+" SET ";
		
		// set
		for (String key : setMap.keySet())
		{
			query += key+"='"+setMap.get(key)+"',";
		}
		if (query.endsWith(",")) query=query.substring(0, query.length()-1);
		
		// where
		query += constructWhereClause(whereList);
		
		query += ";";
		
		return query;
	}
	
	public static String constructWhereClause(HashMap<String,String> whereMap)
	{
		WhereList whereList = new WhereList();
		for (String key : whereMap.keySet())
		{
			whereList.add(key, WhereOp.EQUALS, whereMap.get(key));
		}
		return constructWhereClause(whereList);
	}
	
	private static String constructWhereClause(WhereList whereList)
	{
		if (null == whereList) return "";
		
		String ret = "";
		
		if (whereList.whereList.size()!=0) ret += " WHERE ";
		
		boolean first=true;
		for (Where where : whereList.whereList)
		{
			if (!first) ret += " AND "; first=false;
			ret += where.fieldName+where.whereOp.getSymbol()+"'"+where.value+"'";
		}
		
		return ret;
	}
	
	/////////////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////// interface implementation //////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////
	
	@Override protected boolean i_create(DBName dbName, String tableName, ArrayList<DBField> fields)
	{
		String stmt = q_create(tableName, fields);
		((gov.anl.gmca.io.hw.Database)inst).execQuery(dbName, stmt);
		return true;
	}
	
	@Override protected boolean i_delete(DBName dbName, String table, WhereList whereList)
	{
		String stmt = q_delete(table, whereList);
		((gov.anl.gmca.io.hw.Database)inst).execQuery(dbName, stmt);
		return true;
	}
	
	@Override protected Result i_select(DBName dbName, String tableName, ArrayList<String> requestedColumns, WhereList whereList, String orderBy)
	{
		String stmt = q_select(tableName, requestedColumns, whereList, orderBy);
		return ((gov.anl.gmca.io.hw.Database)inst).execQuery(dbName, stmt);
	}
	
	@Override protected Result i_insert(DBName dbName, String table, ArrayList<HashMap<String,String>> rows)
	{
		String stmt = q_insert(table, rows);
		return ((gov.anl.gmca.io.hw.Database)inst).execQuery(dbName, stmt);
	}
	
	@Override protected Result i_insert(DBName dbName, String table, DBRows rows)
	{
		String stmt = q_insert(table, rows);
		return ((gov.anl.gmca.io.hw.Database)inst).execQuery(dbName, stmt);
	}
	
	@Override protected boolean i_update(DBName dbName, String tableName, HashMap<String,String> setMap, WhereList whereList)
	{
		String query = q_update(tableName, setMap, whereList);
		((gov.anl.gmca.io.hw.Database)inst).execQuery(dbName, query);
		return true;
	}
	
	@Override protected Result i_exec(DBName dbName, String query)
	{
		return ((gov.anl.gmca.io.hw.Database)inst).execQuery(dbName, query);
	}
	
	@Override protected void i_init()
	{
		for (DBName name : DBName.values())
		{
			((gov.anl.gmca.io.hw.Database)inst).verifyConnected(name);
		}
	}
	
	/////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////// helper methods ///////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////
	
	// db variables
    protected HashMap<DBName, Connection> connectionMap = new HashMap<DBName, Connection>();
	public static String sqlLitePath = System.getProperty("application.home")+"/data/blc2004.sqlite";

	protected void verifyConnected(DBName dbName)
	{
		try
		{
			if (connectionMap.containsKey(dbName) && !connectionMap.get(dbName).isClosed())
			{
				return;
			}
		}
		catch(SQLException se) { se.printStackTrace(); }
		
	    try
	    {
	    	String sqlHost = getSqlHost();
	    	Log.note("Connecting to SQL host "+sqlHost+" database "+dbName.getSQLName());

	    	Connection conn;
	    	if (DBConfig.sim5 && dbName==DBName.CONFIG) {
	    		conn = DriverManager.getConnection("jdbc:sqlite:"+sqlLitePath);
			} else {
				Class.forName("com.mysql.jdbc.Driver").getConstructor().newInstance();
				conn = DriverManager.getConnection("jdbc:mysql://" + sqlHost + "/" + dbName.getSQLName(), sqlUsername, sqlPassword);
			}

			if(conn.isClosed())
			{
				throw new Exception("Error opening database connection.");
			}
			else
			{
				connectionMap.put(dbName, conn);
			}
	    }
		catch (Exception e)
		{
			e.printStackTrace();
		}
	}

	protected synchronized Result execQuery(DBName dbName, String query)
	{
		String logMsg = "MySQL: "+query;
		if (logMsg.length()>200) logMsg=logMsg.substring(0, 200);
		// QX allow disabling printing of frequent pooling queries by adding a comment /*NOPRINT*/ in query string
		if (!logMsg.contains("NOPRINT")) { 
			System.out.println(logMsg);
		}
		
		if (query == null) {
			Log.error("empty query");
			return new Result(null, null);
		}
		
		verifyConnected(dbName);

		int retriesLeft = RETRY_MAX;
		while (true)
		{
			Statement st=null;
			ResultSet rs=null;
    		try
    		{
    			st = connectionMap.get(dbName).createStatement();
    			if (DBConfig.sim5) { // return_generated_keys not supported in sqlite (what is it used for?)
					st.execute(query);
				} else {
					st.execute(query, Statement.RETURN_GENERATED_KEYS);
				}

				Integer generatedKey = null;
    			if (!DBConfig.sim5) {
					// first get any generated keys
					rs = st.getGeneratedKeys();
					if (null != rs && rs.next()) {
						generatedKey = rs.getInt(1);
					}
				}

    			// now process normal results
				rs = st.getResultSet();
				if (null != rs)
				{
		    		// get column names
		    		ResultSetMetaData rsmd = rs.getMetaData();
		    		int columnCount = rsmd.getColumnCount();
		    		String[] colNames = new String[columnCount];
		    		for (int i=0;i<columnCount;i++)
		    		{
		    			// getColumnName uses 1-based indices!
		    			colNames[i] = rsmd.getColumnName(i+1);
		    		}
	
		    		// process rows
		    		ArrayList<HashMap<String,String>> ret = new ArrayList<HashMap<String,String>>(); 
					while (rs.next())
			    	{
			    		HashMap<String,String> hm = new LinkedHashMap<String,String>();
			    		for (int i=0;i<colNames.length;i++)
			    		{
			    			// ResultSet.getString also uses 1-based indices
			    			hm.put(colNames[i], rs.getString(i+1));
			    		}
			    		ret.add(hm);
			    	}
					
					closeQuery(st,rs);
					return new Result(ret, generatedKey);
				}
				else
				{
					closeQuery(st, rs);
					return new Result(null, generatedKey);
				}
    		}
    		catch (SQLException se)
    		{
    			closeQuery(st,rs);
    			
    			// Example error message: "Table 'gmca_data.RasterEvents' doesn't exist"
    			// The exception type here is something like MySQLSyntaxErrorException, so
    			// the actual message has to be parsed.
    			String errMsg = se.getMessage();
    			if (errMsg.indexOf("Table")==0 && errMsg.endsWith("doesn't exist"))
    			{
    				return new Result(RESULT_ERROR.TABLE_DOESNT_EXIST);
    			}
    			
    			retriesLeft--;
    			if (retriesLeft<0)
    			{
    				Log.error("Out of SQL retries, returning error.");
    				return new Result(RESULT_ERROR.SYNTAX);
    			}
    			
    			Util.sleep(50);
    			
    			Log.note("Caught SQL exception, retrying. (query="+query+" msg="+se.getMessage()+")");
    			Connection conn = connectionMap.get(dbName);

    			// close the connection
    			try
    			{
    				if (null != conn) conn.close();
    			}
    			catch (SQLException se2)
    			{
    				se.printStackTrace();
    			}
    			
    			// remove it to ensure it's reconnected
    			connectionMap.remove(dbName);
    			
    			// reconnect
    			verifyConnected(dbName);
    		}
		}
	}

    protected void closeQuery(Statement st, ResultSet rs)
    {
    	if (st != null)
    	{
    		try { st.close(); st = null; } catch (SQLException sqle) { sqle.printStackTrace(); }
    	}
    	if (rs != null)
    	{
    		try { rs.close(); rs = null; } catch (SQLException sqle) { sqle.printStackTrace(); }
    	}
    }
}
