package gov.anl.gmca.io.hw;

import gov.anl.gmca.gui.Main;
import gov.anl.gmca.io.PVLookup.PVName;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.syslog.TextLog;
import gov.anl.gmca.util.ProcessHandling;
import gov.anl.gmca.util.ProcessHandling.Signal;
import gov.anl.gmca.util.Util;

import java.io.*;
import java.util.ArrayList;

public class ExecTask extends gov.anl.gmca.io.ExecTask
{
	protected static final int SIGINT_TIMEOUT_MS = 10000;
	protected static final String execScript = Main.getAppDir()+"/tools/scripts/exec.sh";
	
	public ExecTask (final String cmd, final String cmdFilename, final boolean logOutput)
	{
		init(cmd, cmdFilename, null, logOutput, true, true);
	}
	
	public ExecTask (final String cmd, final String cmdFilename, final boolean logOutput, final boolean logError, final boolean logToFile)
	{
		init(cmd, cmdFilename, null, logOutput, logError, logToFile);
	}
	
	public ExecTask (final String cmd, final String cmdFilename, final PVName heartbeatPV, final boolean logOutput, final boolean logError, final boolean logToFile)
	{
		init(cmd, cmdFilename, heartbeatPV, logOutput, logError, logToFile);
	}
	
	protected void init(final String cmd, final String cmdFilename, final PVName heartbeatPV, final boolean logOutput, final boolean logError, final boolean logToFile)
	{
        if (TextLog.isLogConnected()) System.out.println("Exec: "+cmd);
		
		this.cmdFilename = cmdFilename;
		
		thread = new Thread(new Runnable(){public void run()
		{
			BufferedReader bufferedInput = null;
			BufferedReader bufferedError = null;

			InputStream inputstream = null;
			InputStream errorstream = null;
			
			try
			{
				ProcessBuilder pb = new ProcessBuilder(execScript);
				if (null != workingDir) pb.directory(new File(workingDir));

				System.out.println("LD_LIBRARY_PATH is "+pb.environment().get("LD_LIBRARY_PATH"));
				pb.environment().put("BluIcePID", ""+Util.getPid());
				pb.environment().put("ExecCommand", cmd);
				pb.environment().put("LD_PRELOAD", "");

				pb.environment().put
				(
					"HeartbeatPV",
					(null==heartbeatPV) ? "none" : heartbeatPV.toString()
				);

				if (null != inputFilename) pb.redirectInput(new File(inputFilename));

				proc = pb.start();

				if (null != inputText)
				{
					BufferedWriter bw = new BufferedWriter(new OutputStreamWriter(proc.getOutputStream()));
					bw.write(inputText);
					bw.flush();
					bw.close();
				}

				inputstream = proc.getInputStream();
				InputStreamReader inputstreamreader = new InputStreamReader(inputstream);
				bufferedInput = new BufferedReader(inputstreamreader);
				
				errorstream = proc.getErrorStream();
				InputStreamReader errorstreamreader = new InputStreamReader(errorstream);
				bufferedError = new BufferedReader(errorstreamreader);
				
				// get output
				String line;
				while ((line = bufferedInput.readLine()) != null)
				{
					String logLine = "["+cmdFilename+"] "+line;
					if (logOutput)		Log.note(logLine);
					else if (logToFile)	System.out.println(logLine);
					stdout.add(line);
				}
				while ((line = bufferedError.readLine()) != null) {
					String logLine = "["+cmdFilename+"] "+line;
					if (logError)		Log.error(logLine);
					else if (logToFile)	System.out.println("stderr: "+logLine);
					stderr.add(line);
				}
				
				// wait for proc to finish
				exitValue = proc.waitFor();
				if (exitValue != 0)
				{
					System.out.println("ExecTask: "+cmd+" returned " +proc.exitValue());
				}
			}
			catch (IOException ioe) { ioe.printStackTrace(); }
			catch (InterruptedException ie) { ie.printStackTrace(); }
			finally
			{
				try
				{
	    			if (null != bufferedInput) bufferedInput.close();
	    			if (null != bufferedError) bufferedError.close();
	    			//if (null != inputstream) inputstream.close();
	    			//if (null != errorstream) errorstream.close();
				}
				catch (IOException ioe) { ioe.printStackTrace(); }
			}
		}}, "ExecTask "+cmdFilename);
	}

	@Override public void syncKill()
	{
   		ArrayList<Integer> pidList = ProcessHandling.getChildPids(cmdFilename, true);
   		if (0==pidList.size())
   		{
   			Log.error("Couldn't find pid for "+cmdFilename+", can't stop the process.");
   			return;
   		}
   		else
   		{
			ProcessHandling.kill(pidList.get(0), Signal.KILL);
   		}		
	}
	
	@Override public void syncStop()
	{
		if (!isRunning()) return;
		
    	// send control+C
		ArrayList<Integer> pidList = ProcessHandling.getChildPids(cmdFilename, true);
   		if (0 == pidList.size())
   		{
   			Log.error("Couldn't find pid for "+cmdFilename+", can't stop the process.");
   			return;
   		}
   		else
   		{
   			for (Integer pid : pidList)
   			{
   				System.out.println("Sending SIGINT to "+pid);
   				ProcessHandling.kill(pid, Signal.TERM);
   			}
		
			// wait to see if it terminates
			Log.note("Waiting for script to finish...");
			boolean procDead;
			int tries = SIGINT_TIMEOUT_MS/100;
			do
			{
				procDead = true;
				try
				{
					Util.sleep(100);
					proc.exitValue();
				}
				catch (IllegalThreadStateException itse)
				{
					procDead = false;
				}
			}
			while (!procDead && tries-- > 0);
			
			// if not, kill it
			if (!procDead)
			{
				Log.warn(cmdFilename+" did not terminate.  Process will be killed.");
				
	   			for (Integer pid : pidList)
	   			{
	   				ProcessHandling.kill(pid, Signal.KILL);
	   			}

				Util.sleep(100);
			}
   		}
	
    	try
    	{
    		int ev = proc.exitValue();
    		Log.note("Script ended, exit value = "+ev);
    		return;
    	}
    	catch (IllegalThreadStateException itse)
    	{
    		Log.error("Couldn't kill script!");
    	}
	}
}
