package gov.anl.gmca.io.sim.dev;

import gov.anl.gmca.collect.DevBinMode;
import gov.anl.gmca.collect.DevBinMode.BinMode;
import gov.anl.gmca.common.DevMove.MoveCmd;
import gov.anl.gmca.hutch.DevPmacMotor;
import gov.anl.gmca.hutch.DevShutter;
import gov.anl.gmca.io.PVLookup.PV_EIGER;
import gov.anl.gmca.io.PVLookup.PV_FPE;
import gov.anl.gmca.io.PVLookup.PV_MAR;
import gov.anl.gmca.io.PVLookup.PV_PMAC;
import gov.anl.gmca.io.sim.Epics;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.Point2D;
import gov.anl.gmca.util.Util;
import gov.aps.jca.dbr.DBR;

import java.util.StringTokenizer;

public class SimFPE
{
	private Thread thread;
	private BinMode binMode = BinMode.UM_72;
	
	public SimFPE()
	{
		Epics.putInteger(PV_MAR.AsynIO$CNCT,1);
		Epics.putInteger(PV_EIGER.AsynIO$CNCT, 1);
		Epics.put(PV_EIGER.FPE_state, "idle");
		
		Epics.addMonitor(PV_MAR.DetCmdStr, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, final DBR value)
			{
				String cmd = Epics.dbrToString(value);
				if (cmd.length()==0) return;
				
				// just signals we got the message
				Epics.put(PV_MAR.DetCmdStr, "");
				
				if (cmd.equals("get_bin"))
				{
					Epics.put(PV_MAR.DetInStr, DevBinMode.modeToStr(binMode));
				}
				else if (cmd.equals("set_bin"))
				{
					String modeStr=cmd.substring(cmd.indexOf(',')+1);
					binMode=DevBinMode.strToMode(modeStr);
				}
						
			}
		});
		Epics.addMonitor(PV_FPE.start, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, final DBR value)
			{
				if (Epics.dbrToInt(value) != 1) return;
				startFrameList();
			}
		});
		Epics.addMonitor(PV_FPE.mode256, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, final DBR value)
			{
				Epics.putInteger(PV_FPE.start, 1);
			}
		});
	}
	
	// This method doesn't need to be sync as long as it's only called from the
	// single-threaded EPICS queue.  Otherwise it should be set synchronized to
	// prevent multiple threads from being started.
	protected void startFrameList()
	{
		if (null != thread && thread.isAlive()) return;
		
		thread = new Thread(new Runnable(){public void run()
		{
			// parse mode string
			String modeString = Epics.getString(PV_FPE.mode256);
			if (modeString.equals("0 0")) return; // pause
			StringTokenizer st = new StringTokenizer(modeString);
			if (st.countTokens()==4)
			{
				runMARCollection(st);
			}
			else if (st.countTokens()==11)
			{
				runEigerCollection(st);
			}
			else
			{
				Log.error("Unsupported mode string format: "+modeString);
				Epics.putInteger(PV_FPE.start, 0);
				return;
			}
		}}, "SimFPE");
		thread.start();
	}
	
	private void runEigerCollection(StringTokenizer st)
	{
		int mode = Integer.parseInt(st.nextToken());
		int frameCount = Integer.parseInt(st.nextToken());
		Point2D omegaMove = new Point2D(Double.parseDouble(st.nextToken()),Double.parseDouble(st.nextToken()));
		Point2D sampleXMove = new Point2D(Double.parseDouble(st.nextToken()),Double.parseDouble(st.nextToken()));
		Point2D sampleYMove = new Point2D(Double.parseDouble(st.nextToken()),Double.parseDouble(st.nextToken()));
		Point2D sampleZMove = new Point2D(Double.parseDouble(st.nextToken()),Double.parseDouble(st.nextToken()));
		int framesPerFile = Integer.parseInt(st.nextToken());

		Util.sleep(500);

		// gonio arrives at start position
		Epics.putInteger(PV_FPE.abort, 0);
		Epics.putInteger(PV_FPE.start, 0);
		Epics.putInteger(PV_FPE.operationStatus, 1);
		//Epics.putInteger(PV_COUNTER.cbfCompleteFlag, 1);
		Epics.put(PV_EIGER.FPE_state, "st_pos");
		Util.sleep(1000);

		// collection starts
		Epics.put(PV_EIGER.FPE_state, "bluice_updates");
		Util.sleep(3000);
		
		// collection ending..
		Epics.put(PV_EIGER.FPE_state, "stop_motion");
		Util.sleep(500);
		
		// collection ends
		Epics.putInteger(PV_FPE.operationStatus, 0);
		Epics.put(PV_EIGER.FPE_state, "idle");
		Util.sleep(500);

		// all files arrived on disk
		//Epics.putInteger(PV_COUNTER.cbfCompleteFlag, 0);
	}
	
	private void runMARCollection(StringTokenizer st)
	{
		int mode = Integer.parseInt(st.nextToken()); // only 0 is supported for now
		int frameCount = Integer.parseInt(st.nextToken());
		double gonioStart = Double.parseDouble(st.nextToken());
		double gonioEnd = Double.parseDouble(st.nextToken());
		double delta = (gonioEnd-gonioStart)/frameCount;
		
		Epics.putInteger(PV_FPE.operationStatus, 1);
		Epics.putInteger(PV_FPE.abort, 0);
		Epics.putInteger(PV_FPE.start, 0);

		DevPmacMotor devOmega = DevPmacMotor.get("gonio_omega");

		// get parameters
		int startFrameNumber = Epics.getInteger(PV_FPE.frameNumber);
		double expTime = Epics.getInteger(PV_FPE.exposureTime);
		double frameSpeed = delta/expTime;

		for (int frameIndex=0; frameIndex<frameCount; frameIndex++)
		{
			double frameStartAngle = gonioStart + (delta*frameIndex);
			double frameEndAngle = frameStartAngle + delta;
			
			// rewind gonio
			devOmega.startMove(frameStartAngle-0.2).waitFor();
			
			// set speed for frame taking
			Epics.putDouble(PV_PMAC.speed.motor("gonio_omega"), frameSpeed);
			
			// start gonio
			MoveCmd gonioMove = devOmega.startMove(frameEndAngle+0.2);
			
			boolean shutterOpened=false, shutterClosed=false;
			while (true)
			{
				// open shutter when gonio hits start angle
				if (!shutterOpened && devOmega.getPosition() >= gonioStart)
				{
					DevShutter.get("shutter").open();
					shutterOpened = true;
				}
				
				// close shutter when gonio hits end angle
				if (!shutterClosed && devOmega.getPosition() >= gonioEnd)
				{
					DevShutter.get("shutter").close();
					shutterClosed = true;
				}
				
				// wait for gonio stop
				if (gonioMove.isComplete())
				{
					if (!shutterOpened) { Log.warn("Shutter did not open for FPE frame"); }
					if (!shutterClosed) { Log.warn("Shutter did not close for FPE frame"); }
					
					// break the loop on gonio stop
					break;
				}
				
				Util.sleep(10);
			}
			
			// reset to normal speed
			Epics.putDouble(PV_PMAC.speed.motor("gonio_omega"), 90.0);
			
			// update frameNumber
			Epics.putInteger(PV_FPE.frameNumber, startFrameNumber+frameIndex+1);

			final String filename = Epics.getString(PV_FPE.filename);
			final String dir = Epics.getString(PV_FPE.directory256);
			new Thread(new Runnable(){public void run()
			{
				// 3 seconds later, update lastfilename and lastdirectory
				Util.sleep(3000);
				Epics.put(PV_FPE.lastfilename, filename);
				Epics.put(PV_FPE.lastfiledir256, dir);
				Epics.putInteger(PV_FPE.lastfileupdate, 1);
			}}).start();
			
			// check for pause
			String modeString = Epics.getString(PV_FPE.mode256);
			if (modeString.equals("0 0")) break;
		}
		
		Epics.putInteger(PV_FPE.operationStatus, 0);
	}
}
