package gov.anl.gmca.log.db;

import gov.anl.gmca.common.Redis;
import gov.anl.gmca.common.SQLDataType;
import gov.anl.gmca.common.Units;
import gov.anl.gmca.hutch.DevAPSStatus;
import gov.anl.gmca.hutch.DevEnergy;
import gov.anl.gmca.hutch.DevPmacMotor;
import gov.anl.gmca.io.Database.DBField;
import gov.anl.gmca.io.Database.ForeignKey;
import gov.anl.gmca.log.db.DevEvents.EventParam;
import gov.anl.gmca.log.db.DevFrameauditMsg.FrameauditParam;
import gov.anl.gmca.raster.ModelRasterRun;
import gov.anl.gmca.raster.ModelRasterRun.RunType;
import gov.anl.gmca.raster.ModelRasterState;
import gov.anl.gmca.spotfinder.DBImageStats;
import gov.anl.gmca.spotfinder.DevSFType;
import gov.anl.gmca.users.ModelUserTable.Client;
import gov.anl.gmca.util.Util;
import org.apache.commons.lang3.ArrayUtils;

import java.nio.FloatBuffer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;

public class RasterEvent
{
	public static final String GMCA_RASTER_DATA_TABLE = "RasterEvents";
	
	public static enum RasterParam implements EventParam
	{
							// db string		// units			// datatype						// allowNull	// primary key		// foreign key
		SAMPLE_NUMBER		("sampleNumber",	Units.NONE,			SQLDataType.INT,				false,			false,				null),
		RASTER_NUMBER		("rasterNumber",	Units.NONE,			SQLDataType.AUTOINC,			false,			true,				null),
		USERNAME			("userName",		Units.NONE,			SQLDataType.STRING,				false,			false,				null),
		RASTER_TYPE			("rasterType",		Units.NONE,			SQLDataType.ENUM_RASTER_TYPE,	false,			false,				null),
		SEARCH_TYPE			("searchType",		Units.NONE,			SQLDataType.ENUM_SEARCH_TYPE,	false,			false,				null),
		REFERENCE_RASTER	("referenceRaster",	Units.NONE,			SQLDataType.INT,				false,			false,				null),
		PREFIX				("prefix",			Units.NONE,			SQLDataType.STRING,				false,			false,				null),
		WORKDIR				("workDir",			Units.NONE,			SQLDataType.STRING,				false,			false,				null),
		NOOFCELLS			("noOfCells",		Units.NONE,			SQLDataType.INT,				false, 			false,				null),
		CELLSIZE_UM			("cellSize_um",		Units.UM,			SQLDataType.STRING,				false, 			false,				null),
		COLLIMATOR			("collimator",		Units.NONE,			SQLDataType.ENUM_COLLIMATOR,	false, 			false,				null),
		BEAMWIDTH_UM		("beamWidth_um",	Units.UM,			SQLDataType.FLOAT,				false, 			false,				null),
		BEAMHEIGHT_UM		("beamHeight_um",	Units.UM,			SQLDataType.FLOAT,				false, 			false,				null),
		ATTENUATION_FACTOR	("attenFactor",		Units.NONE,			SQLDataType.INT,				false, 			false,				null),
		CRYSTAL_XYZ_MM		("crystalXYZ_mm",	Units.MM,			SQLDataType.STRING,				false, 			false,				null),
		DET_DIST_MM			("detDist_mm",		Units.MM,			SQLDataType.FLOAT,				false, 			false,				null),
		DET_2THETA_DEG		("det2theta_deg",	Units.DEG,			SQLDataType.FLOAT,				false, 			false,				null),
		OMEGA_START_DEG		("omegStart_deg",	Units.DEG,			SQLDataType.FLOAT,				false, 			false,				null),
		DELTA_DEG			("delta_deg",		Units.DEG,			SQLDataType.FLOAT,				false, 			false,				null),
		EXPOSURE_SEC		("exposure_sec",	Units.SEC,			SQLDataType.FLOAT,				false, 			false,				null),
		SIZE_XYZ_UM			("sizeXYZ_um",		Units.UM,			SQLDataType.STRING,				false, 			false,				null),
		RES_LIMITS_A		("resLimits_A",		Units.A,			SQLDataType.STRING,				false, 			false,				null),
		PIXELSIZE_UM		("pixelSize_um",	Units.UM,			SQLDataType.FLOAT,				false, 			false,				null),
		ROI_TYPE			("roiType",			Units.NONE,			SQLDataType.ENUM_ROI_TYPE,		false, 			false,				null),
		ROI_KEV				("roi_keV",			Units.KEV,			SQLDataType.STRING,				false, 			false,				null),
		AUTOFIND_ENABLED	("autoFind",		Units.NONE,			SQLDataType.BOOL,				false, 			false,				null),
		LOOP_CENTERED		("loopCentered",	Units.NONE,			SQLDataType.BOOL,				false, 			false,				null),
		AUTO_HITCOUNT		("autoHitCount",	Units.NONE,			SQLDataType.INT,				false, 			false,				null),
		MAX_SPOTS			("maxSpots",		Units.NONE,			SQLDataType.INT,				false, 			false,				null),
		TIMESTAMP			("timestamp",		Units.TIMESTAMP,	SQLDataType.TIMESTAMP,			false, 			false,				null),
		IZERO				("iZero",			Units.V,			SQLDataType.FLOAT,				false,			false,				null),
		ACTIVE_BS			("active_bs",		Units.V,			SQLDataType.FLOAT,				false,			false,				null),
		APS_CURRENT_MA		("apsCurrent_mA",	Units.MAMP,			SQLDataType.FLOAT,				false,			false,				null),
		DCM_VSUM			("dcmVsum",			Units.V,			SQLDataType.FLOAT,				false,			false,				null),
		ENERGY1_KEV			("energy1_keV",		Units.KEV,			SQLDataType.FLOAT,				false,			false,				null),
		FINAL_ENERGY_KEV	("finalEnergy_keV",	Units.KEV,			SQLDataType.FLOAT,				false,			false,				null);
		
		public String dbString;
		public Units units;
		public SQLDataType dataType;
		public boolean allowNull;
		public boolean isPrimaryKey;
		public ForeignKey foreignKey;
		
		RasterParam(String dbString, Units units, SQLDataType dataType, boolean allowNull, boolean isPrimaryKey, ForeignKey foreignKey)
		{
			this.dbString=dbString;
			this.units=units;
			this.dataType=dataType;
			this.allowNull=allowNull;
			this.isPrimaryKey=isPrimaryKey;
			this.foreignKey=foreignKey;
		}
		
		public DBField getDBField() { return new DBField(dbString, dataType, allowNull, isPrimaryKey, foreignKey); }
	};
	
	public static void writeEvent(ModelRasterRun run, double startOmega)
	{
		ArrayList<HashMap<String,String>> rows = new ArrayList<HashMap<String,String>>(); 
		HashMap<String,String> row = new HashMap<String,String>();
		for (RasterParam param : RasterParam.values())
		{
			switch (param)
			{
			case ATTENUATION_FACTOR:	DevEvents.addDouble	(row, param, run.getAttenuation());			break;
			/**/case AUTO_HITCOUNT:			DevEvents.addInt	(row, param, 0);							break;
			case AUTOFIND_ENABLED:		DevEvents.addBool	(row, param, run.getAutoFind());			break;
			case BEAMWIDTH_UM:			DevEvents.addString	(row, param, Integer.toString((int)run.getBeamWidth_um()));		break;
			case BEAMHEIGHT_UM:			DevEvents.addString	(row, param, Integer.toString((int)run.getBeamHeight_um()));	break;
			case CELLSIZE_UM:			DevEvents.addString	(row, param, Integer.toString((int)run.grid().getCellWidth_um())+"x"+Integer.toString((int)run.grid().getCellWidth_um()));	break;
			case COLLIMATOR:			DevEvents.addString	(row, param, run.getBeamsizeSetup().collimatorSize.toString()); 														break;
			/**/case CRYSTAL_XYZ_MM:		DevEvents.addString	(row, param, "0x0x0");						break;
			case DELTA_DEG:				DevEvents.addDouble	(row, param, run.getDelta());				break;
			case DET_DIST_MM:			DevEvents.addDouble	(row, param, run.getDetectorDist());		break;
			case EXPOSURE_SEC:			DevEvents.addDouble	(row, param, run.getExposureTime());		break;
			case LOOP_CENTERED:			DevEvents.addBool	(row, param, run.getLoopCentered());		break;
			case MAX_SPOTS:				DevEvents.addInt	(row, param, getMaxSpots()); break;
			case NOOFCELLS:				DevEvents.addInt	(row, param, run.grid().getSelectedCellCount()); break;
			case OMEGA_START_DEG:		DevEvents.addDouble	(row, param, startOmega);							break;
			case PIXELSIZE_UM:			DevEvents.addInt	(row, param, ModelRasterState.get().getBinMode().getIntSize()); break;
			case PREFIX:				DevEvents.addString	(row, param, run.getFileRoot());			break;
			case RASTER_NUMBER:			DevEvents.addInt	(row, param, 0);							break;
			case RASTER_TYPE:			DevEvents.addString	(row, param, run.getRunType().toString()); 	break;
			/**/case REFERENCE_RASTER:		DevEvents.addInt	(row, param, 0);							break;
			case RES_LIMITS_A:			DevEvents.addString	(row, param, Util.printDouble(run.getLowResLimit_A(),2)+"x"+Util.printDouble(run.getHighResLimit_A(),2)); break;
			/**/case ROI_KEV:				DevEvents.addString	(row, param, "0,0");					break;
			case ROI_TYPE:				DevEvents.addString	(row, param, run.getROIMode().toString());  break;
			case SAMPLE_NUMBER:			DevEvents.addInt	(row, param, DevEvents.getSampleNumber());	break;
			case SEARCH_TYPE:			DevEvents.addString	(row, param, run.getRefinementType().toString()); break;
			/**/case SIZE_XYZ_UM:			DevEvents.addString	(row, param, "0x0x0");						break;
			case TIMESTAMP:				DevEvents.addString	(row, param, SQLDataType.getCurrentTimeString()); break;
			case WORKDIR:				DevEvents.addString	(row, param, run.getFullDirectory());		break;
			case IZERO:					String s_i0 = DevFrameauditMsg.fieldMap.get(FrameauditParam.I0_SUM); DevEvents.addDouble(row, param, Double.parseDouble(s_i0==null?"0":s_i0)); break;
			case ACTIVE_BS:				String s_bs = DevFrameauditMsg.fieldMap.get(FrameauditParam.BS_SUM); DevEvents.addDouble(row, param, Double.parseDouble(s_bs==null?"0":s_bs)); break;
			case APS_CURRENT_MA:		DevEvents.addDouble	(row, param, DevAPSStatus.getCurrent());		break;
			/**/case DCM_VSUM:				DevEvents.addDouble	(row, param, 0);		break;
			case ENERGY1_KEV:			DevEvents.addDouble	(row, param, run.getRunType()==RunType.DIFFRACTION?DevEnergy.getDevice().getPosition():run.getEnergy()); break;
			case FINAL_ENERGY_KEV:		DevEvents.addDouble	(row, param, DevEnergy.getDevice().getPosition()); break;
			case DET_2THETA_DEG:		DevEvents.addDouble	(row, param, DevPmacMotor.get("detector_2t").getPosition()); break;
			case USERNAME:				DevEvents.addString	(row, param, Client.getUser());		break;
			default:					throw new IllegalArgumentException("Unknown raster event field: "+param);
			}
		}
		rows.add(row);

		DevEvents.addEvent(GMCA_RASTER_DATA_TABLE, RasterParam.values(), rows);
	}
	
	private static int getMaxSpots()
	{
		try
		{
			ModelRasterRun rrun = ModelRasterState.get().getCurrentRunModel();
			DevSFType.SPOTFINDER_OUTPUT output = ModelRasterState.get().getDiffDisplayParameter();
			FloatBuffer data = DBImageStats.load(Redis.image_stats(rrun.getDirectorySuffix(), rrun.getFileRoot(), rrun.getRunLabel(), output.toString()));
			float max = DBImageStats.max(data,0);
			return Math.round(max);
		}
		catch (Exception e) { e.printStackTrace(); return 0; }
	}
}
