package gov.anl.gmca.log.db;

import gov.anl.gmca.common.SQLDataType;
import gov.anl.gmca.common.Units;
import gov.anl.gmca.hutch.DevAPSStatus;
import gov.anl.gmca.hutch.DevAttenuation;
import gov.anl.gmca.hutch.DevEnergy;
import gov.anl.gmca.hutch.DevPmacMotor;
import gov.anl.gmca.io.Database.DBField;
import gov.anl.gmca.io.Database.ForeignKey;
import gov.anl.gmca.log.db.DevEvents.EventParam;
import gov.anl.gmca.log.db.DevFrameauditMsg.FrameauditParam;
import gov.anl.gmca.sample.DevCollimatorCtrl;
import gov.anl.gmca.scan.ModelScan;
import gov.anl.gmca.users.ModelUserTable.Client;

import java.util.ArrayList;
import java.util.HashMap;

public class ScanEvent
{
	public static final String GMCA_SCAN_DATA_TABLE = "ScanEvents";
	
	public static enum ScanParam implements EventParam
	{
							// db string		// units			// datatype						// allowNull	// primary key		// foreign key
		SAMPLE_NUMBER		("sampleNumber",	Units.NONE,			SQLDataType.INT,				false,			false,				null),
		SCAN_NUMBER			("scanNumber",		Units.NONE,			SQLDataType.AUTOINC,			false,			true,				null),
		USERNAME			("userName",		Units.NONE,			SQLDataType.STRING,				false,			false,				null),
		AUTOSCAN_STEP		("autoScanStep",	Units.NONE,			SQLDataType.ENUM_AUTOSCAN_STEP,	false,			false,				null),
		INTERSCAN_STEP		("interScanStep",	Units.NONE,			SQLDataType.ENUM_INTERSCAN_STEP,false,			false,				null),
		SPECTRUM_EXP_SEC	("spectrumExpSec",	Units.SEC,			SQLDataType.FLOAT,				false, 			false,				null),
		EDGE_EXP_SEC		("edgeExpSec",		Units.SEC,			SQLDataType.FLOAT,				false, 			false,				null),
		SCAN_CENTER_KEV		("scanCenter_keV",	Units.KEV,			SQLDataType.FLOAT,				false, 			false,				null),
		PREFIX				("prefix",			Units.NONE,			SQLDataType.STRING,				false,			false,				null),
		WORKDIR				("workDir",			Units.NONE,			SQLDataType.STRING,				false,			false,				null),
		PCT_COMPLETE		("pctComplete",		Units.NONE,			SQLDataType.INT,				false,			false,				null),
		ELEMENT_EDGE		("elementEdge",		Units.NONE,			SQLDataType.STRING,				false, 			false,				null),
		ROI_LOW_KEV			("roiLow_keV",		Units.KEV,			SQLDataType.FLOAT,				false, 			false,				null),
		ROI_HIGH_KEV		("roiHigh_keV",		Units.KEV,			SQLDataType.FLOAT,				false, 			false,				null),
		PEAK_KEV			("peak_keV",		Units.KEV,			SQLDataType.FLOAT,				false, 			false,				null),
		PEAK_FP				("peak_fP",			Units.NONE,			SQLDataType.FLOAT,				false, 			false,				null),
		PEAK_FPP			("peak_fPP",		Units.NONE,			SQLDataType.FLOAT,				false, 			false,				null),
		INFL_KEV			("infl_keV",		Units.KEV,			SQLDataType.FLOAT,				false, 			false,				null),
		INFL_FP				("infl_fP",			Units.NONE,			SQLDataType.FLOAT,				false, 			false,				null),
		INFL_FPP			("infl_fPP",		Units.NONE,			SQLDataType.FLOAT,				false, 			false,				null),
		SPECTRUM_KEV		("spectrum_keV",	Units.KEV,			SQLDataType.FLOAT,				false, 			false,				null),
		COLLIMATOR			("collimator",		Units.NONE,			SQLDataType.ENUM_COLLIMATOR,	false, 			false,				null),
		BEAMWIDTH_UM		("beamWidth_um",	Units.UM,			SQLDataType.FLOAT,				false, 			false,				null),
		BEAMHEIGHT_UM		("beamHeight_um",	Units.UM,			SQLDataType.FLOAT,				false, 			false,				null),
		APS_CURRENT_MA		("apsCurrent_mA",	Units.MAMP,			SQLDataType.FLOAT,				false,			false,				null),
		IZERO				("iZero",			Units.V,			SQLDataType.FLOAT,				false,			false,				null),
		DCM_VSUM			("dcmVsum",			Units.V,			SQLDataType.FLOAT,				false,			false,				null),
		FINAL_ENERGY_KEV	("finalEnergy_keV",	Units.KEV,			SQLDataType.FLOAT,				false,			false,				null),
		OMEGA_START_DEG		("omegStart_deg",	Units.DEG,			SQLDataType.FLOAT,				false, 			false,				null),
		TIMESTAMP			("timestamp",		Units.TIMESTAMP,	SQLDataType.TIMESTAMP,			false, 			false,				null),
		ATTENUATION_FACTOR	("attenFactor",		Units.NONE,			SQLDataType.INT,				false, 			false,				null);
		
		public String dbString;
		public Units units;
		public SQLDataType dataType;
		public boolean allowNull;
		public boolean isPrimaryKey;
		public ForeignKey foreignKey;
		
		ScanParam(String dbString, Units units, SQLDataType dataType, boolean allowNull, boolean isPrimaryKey, ForeignKey foreignKey)
		{
			this.dbString=dbString;
			this.units=units;
			this.dataType=dataType;
			this.allowNull=allowNull;
			this.isPrimaryKey=isPrimaryKey;
			this.foreignKey=foreignKey;
		}
		
		public DBField getDBField() { return new DBField(dbString, dataType, allowNull, isPrimaryKey, foreignKey); }
	};
	
	public static void writeEvent(double dcmVsum, int percentComplete)
	{
		ArrayList<HashMap<String,String>> rows = new ArrayList<HashMap<String,String>>(); 
		HashMap<String,String> row = new HashMap<String,String>();
		for (ScanParam param : ScanParam.values())
		{
			switch (param)
			{
			case ATTENUATION_FACTOR:	DevEvents.addDouble	(row, param, DevAttenuation.get().getFactors());	break;
			case BEAMWIDTH_UM:			DevEvents.addString	(row, param, Integer.toString((int)DevPmacMotor.get("beam_size_x").getPosition()));	break;
			case BEAMHEIGHT_UM:			DevEvents.addString	(row, param, Integer.toString((int)DevPmacMotor.get("beam_size_y").getPosition()));	break;
			case COLLIMATOR:			DevEvents.addString	(row, param, DevCollimatorCtrl.getCollimatorSize().toString()); 														break;
			case OMEGA_START_DEG:		DevEvents.addDouble	(row, param, DevPmacMotor.get("gonio_omega").getPosition()); break;
			case PREFIX:				DevEvents.addString	(row, param, ModelScan.get().getPrefix());	break;
			case SAMPLE_NUMBER:			DevEvents.addInt	(row, param, DevEvents.getSampleNumber());	break;
			case TIMESTAMP:				DevEvents.addString	(row, param, SQLDataType.getCurrentTimeString()); break;
			case WORKDIR:				DevEvents.addString	(row, param, ModelScan.get().getFullDirectory()); break;
			case IZERO:					String s_i0 = DevFrameauditMsg.fieldMap.get(FrameauditParam.I0_SUM); DevEvents.addDouble(row, param, Double.parseDouble(s_i0==null?"0":s_i0)); break;
			case APS_CURRENT_MA:		DevEvents.addDouble	(row, param, DevAPSStatus.getCurrent());		break;
			case DCM_VSUM:				DevEvents.addDouble	(row, param, dcmVsum); break;
			case FINAL_ENERGY_KEV:		DevEvents.addDouble	(row, param, DevEnergy.getDevice().getPosition()); break;
			case AUTOSCAN_STEP:			DevEvents.addString	(row, param, ModelScan.get().getAutoStep().toString());		break;
			case EDGE_EXP_SEC:			DevEvents.addDouble	(row, param, ModelScan.get().getEdgeTime());		break;
			case ELEMENT_EDGE:			DevEvents.addString	(row, param, ModelScan.get().getElementEdge().toString());		break;
			case INFL_FP:				DevEvents.addDouble	(row, param, ModelScan.get().getInfl_fP()); break;
			case INFL_FPP:				DevEvents.addDouble	(row, param, ModelScan.get().getInfl_fPP()); break;
			case INFL_KEV:				DevEvents.addDouble	(row, param, ModelScan.get().getInflection()); break;
			case INTERSCAN_STEP:		DevEvents.addString	(row, param, ModelScan.get().getInterStep().toString()); break;
			case PCT_COMPLETE:			DevEvents.addInt	(row, param, percentComplete);	break;
			case PEAK_FP:				DevEvents.addDouble	(row, param, ModelScan.get().getPeak_fP()); break;
			case PEAK_FPP:				DevEvents.addDouble	(row, param, ModelScan.get().getPeak_fPP()); break;
			case PEAK_KEV:				DevEvents.addDouble	(row, param, ModelScan.get().getPeak()); break;
			case ROI_HIGH_KEV:			DevEvents.addDouble	(row, param, ModelScan.get().getROIHigh()); break;
			case ROI_LOW_KEV:			DevEvents.addDouble	(row, param, ModelScan.get().getROILow()); break;
			case SCAN_CENTER_KEV:		DevEvents.addDouble	(row, param, ModelScan.get().getScanCenter()); break;
			case SCAN_NUMBER:			DevEvents.addInt	(row, param, 0); break;
			case SPECTRUM_EXP_SEC:		DevEvents.addDouble	(row, param, ModelScan.get().getSpectrumTime()); break;
			case SPECTRUM_KEV:			DevEvents.addDouble	(row, param, ModelScan.get().getSpectrumEnergy()); break;
			case USERNAME:				DevEvents.addString	(row, param, Client.getUser());		break;
			default:					throw new IllegalArgumentException("Unknown scan event field: "+param);
			}
		}
		rows.add(row);

		DevEvents.addEvent(GMCA_SCAN_DATA_TABLE, ScanParam.values(), rows);
	}
}
