/**
 * Model for XPLAN stratgey tab
 * 
 */
package gov.anl.gmca.multistrategy;

import gov.anl.gmca.collect.DevCollectRun;
import gov.anl.gmca.collect.DevCollectRunList;
import gov.anl.gmca.collect.ModelCollectRun;
import gov.anl.gmca.collect.ModelCollectRun.XDSProcMode;
import gov.anl.gmca.collect.ModelCollectRunList;
import gov.anl.gmca.common.DevRunList.RUN_LABEL_SETTING;
import gov.anl.gmca.strategy.ModelStrategyResult;
import gov.anl.gmca.strategy.ModelStrategyResult.StrategyField;
import gov.anl.gmca.strategy.ModelStrategyResultList;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.Util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.StringTokenizer;

public class ModelMultiStrategy implements Comparable<ModelMultiStrategy> {
	public static enum MultiStrategyField {
		               // db string // user string // display lines
		SAMPLE_NUMBER ("sampleNumber", "Sample #", 3), 
		STATE        ("state", "State", 2), 
		WORKDIR      ("workdir", "Dir", 2), 
		WARNING      ("warning", "warning", 3), 
		REFERENCEDATA("referencedata", "Ref. Data", 2), 
		MINRANGE     ("min_range","Min. Range", 2), 
		IMAGES       ("images", "Images", 2), 
		UNIT_CELL    ("unitcell", "Unit Cell", 2), 
		SPACE_GROUP  ("spacegroup","Space Group", 2), 
		LOWRESOLUTION("lowresolution", "Low Res.", 1), 
		HIGHRESOLUTION("highresolution", "High Res.", 1), 
		DISPLAYTEXT   ("displaytext","Strategy", 2),
		STRATEGY	  ("strategy", "Strategy", 2),
		XDSLOG       ("xdslog", "XDS Log", 1), 
		XPLANLOG     ("xplanlog", "XPLAN Log", 1);

		protected String dbString, userString;
		protected int displayLines;
		boolean exportable;

		public String getDBString() {
			return dbString;
		}

		public String getUserString() {
			return userString;
		}

		public int getDisplayLines() {
			return displayLines;
		}

		MultiStrategyField(String dbString, String userString, int displayLines) {
			this.dbString = dbString;
			this.userString = userString;
			this.displayLines = displayLines;
		}
	}

	public static enum MultiStrategyState {
		INACTIVE, ERROR, PROCESSING, DONE;

		public static MultiStrategyState stringToState(String stateString) {
			if (null == stateString)
				return null;

			for (MultiStrategyState state : MultiStrategyState.values()) {
				if (stateString.toLowerCase().equals(
						state.toString().toLowerCase()))
					return state;
			}

			return null;
		}
	}

	public MultiStrategyState getState() {
		return MultiStrategyState
				.stringToState(getField(MultiStrategyField.STATE));
	}

	public static ArrayList<String> strategyFileList = new ArrayList<String>();
	protected HashMap<MultiStrategyField, String> fieldMap = new HashMap<MultiStrategyField, String>();

	public void setField(MultiStrategyField field, String value) {
		fieldMap.put(field, value);
	}

	public String getField(MultiStrategyField field) {
		return fieldMap.get(field);
	}

	public HashMap<MultiStrategyField, String> getAllFields() {
		return fieldMap;
	}

	protected Integer resultID = null;

	public Integer getID() {
		return resultID;
	}

	public void setID(int id) {
		resultID = id;
	}

	public ModelMultiStrategy(HashMap<MultiStrategyField, String> fieldMap) {
		this.fieldMap = new HashMap<MultiStrategyField, String>(fieldMap);
		String sampleNumber = fieldMap.get(MultiStrategyField.SAMPLE_NUMBER);
		if (null != sampleNumber) {
			try {
				resultID = Integer.parseInt(sampleNumber);
			} catch (NumberFormatException nfe) {
			}
		}
	}
	
	public void exportToRun(String strategy, boolean isSingle, boolean createNewRun) {
		double oscStart, oscEnd, oscDelta = 0.5, detDist = 300, wedge;
		ModelStrategyResult singleStrategy = null;
		
		if(strategy== null || strategy.equals("") || strategy.equals(" ")) {
			
			Log.warn("Multi crystal startegy not found");
			return;
		}
		
		try {
			StringTokenizer str = new StringTokenizer(strategy);
			String angleSet = str.nextToken().trim().replaceAll("\\s+","");
		
			String[] angles = angleSet.split("\\.\\.\\.");
			oscStart = Double.parseDouble(angles[0].trim());
			oscEnd = Double.parseDouble(angles[1].trim());
			Log.error(oscStart+ " "+oscEnd);
			
			if(isSingle) {
				singleStrategy = ModelStrategyResultList.get().getCurrentResult();
				
				if(singleStrategy != null) {
					oscDelta = Double
							.parseDouble(singleStrategy.getField(StrategyField.OSC_DELTA));
					detDist = Double.parseDouble(singleStrategy.getField(StrategyField.DETECTOR_DIST));
					
					Log.note("Osc. Dleta and Detector disatnce exported from single crystal strategy");
					
				} else {
					
					Log.warn("nothing exported from single crystal strategy");
				}
			} 
		} catch (NumberFormatException nfe) {
			Util.asyncPopupOk(
					"Error",
					"Error reading strategy field, export aborted. "
							+ nfe.getMessage());
			return;
		}

		DevCollectRunList listDev = DevCollectRunList.get();
		ModelCollectRunList list = ModelCollectRunList.get();

		int targetIndex = createNewRun ? list.getRunCount() + 1 : list
				.getCurrentRunIndex();

		Boolean inverseOn = null;
		Integer xdsProcMode = null;
		Double wedgeSize = null;

		if (createNewRun) {
			// copy last run's contents
			DevCollectRunList.getIndex(list.getRunCount()).copyRunContents(
					targetIndex, RUN_LABEL_SETTING.INCREMENT_SRC, true, false);

			// set run device directly
			DevCollectRun.setStartFrameLabel(targetIndex, 1);
			DevCollectRun.setStartAngle(targetIndex, oscStart);
			DevCollectRun.setEndAngle(targetIndex, oscEnd);
			if(isSingle && singleStrategy!=null) {
				DevCollectRun.setDelta(targetIndex, oscDelta);
				DevCollectRun.setDistance(targetIndex, detDist);
			}
			if (null != inverseOn)
				DevCollectRun.setInverseOn(targetIndex, inverseOn);
			if (null != xdsProcMode)
				DevCollectRun.setXDSProcMode(targetIndex, xdsProcMode);
			if (null != wedgeSize)
				DevCollectRun.setWedgeSize(targetIndex, wedgeSize);

			// increase run count to load the new run
			listDev.setRunCount(targetIndex, false);
			listDev.setCurrentRun(targetIndex, true);
		} else {
			ModelCollectRun run = (ModelCollectRun) ModelCollectRunList.get()
					.getRun(targetIndex);
			run.dev().setStartFrameLabel(1, false);
			run.dev().setStartAngle(oscStart, false);
			run.dev().setDelta(oscDelta, false);
			run.dev().setDistance(detDist, false);
			if (null != inverseOn)
				run.dev().setInverseOn(inverseOn, false);
			if (null != xdsProcMode)
				run.dev().setXDSProcMode(XDSProcMode.values()[xdsProcMode],
						false);
			if (null != wedgeSize)
				run.dev().setWedgeSize(wedgeSize, false);

			run.fid.changeEndAngle(oscEnd); // sends update
		}

		Log.note("Multi Crystal Strategy exported to run "
				+ DevCollectRun.getRunLabel(targetIndex));
	}

	@Override
	public int compareTo(ModelMultiStrategy r2) {
		return getID() - r2.getID();
	}
}
