/**
 * GUI for Multi Crystal Strategy tab
 * all the GUI of sub-tabs are initiated from this class
 */

package gov.anl.gmca.multistrategy;

import gov.anl.gmca.collect.ModelCollectRun;
import gov.anl.gmca.collect.ModelCollectRunList;
import gov.anl.gmca.common.DevDataDir;
import gov.anl.gmca.common.ModelRunList;
import gov.anl.gmca.common.ModelRunList.ModelRun;
import gov.anl.gmca.common.ui.WdgControl;
import gov.anl.gmca.gui.GUI;
import gov.anl.gmca.gui.Main;
import gov.anl.gmca.io.ExecTask;
import gov.anl.gmca.multistrategy.ModelMultiStrategy.MultiStrategyField;
import gov.anl.gmca.multistrategy.ModelMultiStrategy.MultiStrategyState;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.users.ModelUserTable.Client;
import gov.anl.gmca.users.ModelUserTable.ClientListener;
import gov.anl.gmca.util.ListenerList;
import gov.anl.gmca.util.Util;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.*;

import java.io.File;
import java.util.HashMap;

public class WdgMultiStrategy extends Composite implements
		ListenerList.Listener<ModelMultiStrategyList>, 
		ModelCollectRunList.Listener, ModelCollectRunList.CurrentRunListener, ClientListener {
	protected static final ModelMultiStrategy.MultiStrategyField[] displayFields = {
			ModelMultiStrategy.MultiStrategyField.WORKDIR,
			ModelMultiStrategy.MultiStrategyField.REFERENCEDATA,
			ModelMultiStrategy.MultiStrategyField.IMAGES,
			ModelMultiStrategy.MultiStrategyField.WARNING,
			ModelMultiStrategy.MultiStrategyField.DISPLAYTEXT,
			ModelMultiStrategy.MultiStrategyField.STRATEGY,
	};

	protected HashMap<MultiStrategyField, Text> fieldTexts = new HashMap<MultiStrategyField, Text>();
	protected Composite referenceComp, genComp, labelComp, expandComp, logComp, controlComp;
	protected Label positionLabel, stateLabel;
	protected Button exportStrategy, xplanLogButton, xdsLogButton, deleteButton, mcsHelpButton, mcsVideoButton;
	protected Button startButton, prevButton, nextButton, lastButton, singleStrategyButton;
	protected Button[] strategyButtons;

	// sudhir
	protected TabFolder ctFolder;
	protected TabItem referencedataTab, allTab;
	protected WdgMultiStrategyTable allTable = null;
	protected WdgReferenceData stratReference = null;

	protected static final boolean UPDATE_CURRENT_INDEX_ON_RESULT_CREATION = true;
	protected static final boolean SHOW_MultiStrategy_TAB_ON_RESULT_CREATION = true;
	protected static final int NUMBER_OF_ANGLES = 6;

	public WdgMultiStrategy(Composite parent) {
		
		super(parent, SWT.NONE);
		setBackground(GUI.BGCOLOR);
		setLayout(new GridLayout(1, true));

		expandComp = new Composite(this, SWT.NONE);
		expandComp.setBackground(GUI.BGCOLOR);
		GridLayout gl1 = new GridLayout(4, false);
		expandComp.setLayout(gl1);
		GridData gd = new GridData(SWT.FILL, SWT.FILL, true, true);
		expandComp.setLayoutData(gd);
		
		Composite helpGroup = new Composite(expandComp, SWT.NONE);
		helpGroup.setBackground(GUI.BGCOLOR);
		
		GridLayout actionGrid = new GridLayout(4, false);
		helpGroup.setLayout(actionGrid);

		gd = new GridData(SWT.FILL, SWT.DEFAULT, true, false);
		gd.horizontalSpan = 4;

		helpGroup.setLayoutData(gd);
		helpGroup.setFont(GUI.FONT_COLLECT_GROUP);

		mcsHelpButton = new Button(helpGroup, SWT.PUSH);
		mcsHelpButton.setText("Help");
		mcsHelpButton.setBackground(GUI.COLOR_BLUE_BUTTON);
		gd = new GridData(SWT.CENTER, SWT.DEFAULT, true, false);
		gd.horizontalSpan = 2;
		gd.minimumWidth = 120;
		mcsHelpButton.setLayoutData(gd);
		
		mcsHelpButton.addSelectionListener(new SelectionListener() {

			@Override
			public void widgetDefaultSelected(SelectionEvent arg0) {}

			@Override
			public void widgetSelected(SelectionEvent arg0) {
				
				String filename = Main.getAppDir()+"/java/docs/mcs_instructions.pdf";
				ExecTask.new_ExecTask("firefox " + filename + "", "Firefox",
						false, false).start();
			}	
		});
		
		mcsVideoButton = new Button(helpGroup, SWT.PUSH);
		mcsVideoButton.setText("Video Tutorial");
		mcsVideoButton.setBackground(GUI.COLOR_BLUE_BUTTON);
		gd = new GridData(SWT.CENTER, SWT.DEFAULT, true, false);
		gd.horizontalSpan = 2;
		mcsVideoButton.setLayoutData(gd);
		
		mcsVideoButton.addSelectionListener(new SelectionListener() {

			@Override
			public void widgetDefaultSelected(SelectionEvent arg0) {}

			@Override
			public void widgetSelected(SelectionEvent arg0) {
				
				String filename = "https://www.gmca.aps.anl.gov/video/html5-guides/collect-mcs-vjs.html";
				ExecTask.new_ExecTask("firefox " + filename + "", "firefox",
						false, false).start();
			}
			
		});
		
		createControls();

		createMultiStrategyDisplay();
		createExportSolutionButtons();
	}

	public void connectListeners() {
		ModelMultiStrategyList.get().ll.addListener(this);
		DevMultiStrategy.get().verifyConnected();
		Client.addListener(this);
		ModelCollectRunList.get().addListener(this);
		ModelCollectRunList.get().addCurrentRunListener(this); // needed to
																// track run
																// label
		updateLabels();

		// allTable.updateTable();
		ctFolder.setSelection(0);
	}

	protected void createExportSolutionButtons() {
		
		logComp = new Composite(expandComp, SWT.NONE);
		logComp.setBackground(GUI.BGCOLOR);
		GridLayout gl = new GridLayout(4, false);
		gl.marginWidth = 0;
		gl.marginHeight = 0;
		logComp.setLayout(gl);
		GridData gd = new GridData(SWT.FILL, SWT.FILL, true, false);
		gd.horizontalSpan = 4;
		logComp.setLayoutData(gd);
		
		exportStrategy = new Button(logComp, SWT.PUSH);
		exportStrategy.setText("Export");
		exportStrategy.setBackground(GUI.COLOR_BLUE_BUTTON);
		gd = new GridData(SWT.CENTER, SWT.FILL, true, true);
		gd.widthHint = 150;
		exportStrategy.setLayoutData(gd);
		exportStrategy.addSelectionListener(new SelectionListener() {
			@Override
			public void widgetDefaultSelected(SelectionEvent se) {
			}

			@Override
			public void widgetSelected(SelectionEvent se) {
				ModelCollectRunList runList = ModelCollectRunList.get();

				int runCount = runList.getRunCount();
				if (runCount == 15) {
					Util.popupOk("Tab limit",
							"You have reached the limit of 16 run tabs.");
					return;
				}
				String strategy = "";
				for(int j = 0; j < NUMBER_OF_ANGLES; j++) {
					if(strategyButtons[j].getSelection()) {
						strategy = strategyButtons[j].getText();
					}	
				}
				
				ModelMultiStrategyList.get().getCurrentResult()
						.exportToRun(strategy, singleStrategyButton.getSelection(), true);
			}
		});
		
		xdsLogButton = new Button(logComp, SWT.PUSH);
		xdsLogButton.setText("XDS Log");
		xdsLogButton.setBackground(GUI.COLOR_BLUE_BUTTON);
		gd = new GridData(SWT.CENTER, SWT.DEFAULT, true, false);
		gd.widthHint = 150;
		xdsLogButton.setLayoutData(gd);
		xdsLogButton.addSelectionListener(new SelectionListener() {
			@Override
			public void widgetDefaultSelected(SelectionEvent se) {
			}

			@Override
			public void widgetSelected(SelectionEvent se) {
				String logfile = "XDS.LOG";
				String filename = ModelMultiStrategyList.get()
						.getCurrentResult()
						.getField(MultiStrategyField.WORKDIR)
						+ "/" + logfile;

				// TODO: handle spaces
				File xdsFile = new File(filename);
				if (xdsFile.exists())
					ExecTask.new_ExecTask("nedit " + filename + "", "nedit",
							false, false).start();
				else
					Log.error(filename + " doesn't exists");
			}
		});

		xplanLogButton = new Button(logComp, SWT.PUSH);
		xplanLogButton.setText("XPLAN Log");
		xplanLogButton.setBackground(GUI.COLOR_BLUE_BUTTON);
		gd = new GridData(SWT.CENTER, SWT.DEFAULT, true, false);
		gd.widthHint = 150;
		xplanLogButton.setLayoutData(gd);
		xplanLogButton.addSelectionListener(new SelectionListener() {
			@Override
			public void widgetDefaultSelected(SelectionEvent se) {
			}

			@Override
			public void widgetSelected(SelectionEvent se) {
				String logfile = "XPLAN.LP";
				String filename = ModelMultiStrategyList.get()
						.getCurrentResult()
						.getField(MultiStrategyField.WORKDIR)
						+ "/" + logfile;

				// TODO: handle spaces
				File xdsFile = new File(filename);
				if (xdsFile.exists())
					ExecTask.new_ExecTask("nedit " + filename + "", "nedit",
							false, false).start();
				else
					Log.error(filename + " doesn't exists");
			}
		});
	}

	protected void createMultiStrategyDisplay() {

		GridData gd = new GridData(SWT.FILL, SWT.FILL, true, true, 4, 1);
		ctFolder = new TabFolder(expandComp, SWT.NONE);
		ctFolder.setFont(GUI.FONT_SAMPLE_TABS);
		ctFolder.setLayoutData(gd);

		ctFolder.addSelectionListener(new SelectionListener() {
			@Override
			public void widgetDefaultSelected(SelectionEvent se) {
			}

			@Override
			public void widgetSelected(SelectionEvent se) {
				GUI.asyncExec(new Runnable() {
					public void run() {
							setButtonsVisible(ctFolder.getSelectionIndex() != 1);
					}
				});
			}
		});
		
		referenceComp = new Composite(ctFolder, SWT.NONE);
		referenceComp.setBackground(GUI.BGCOLOR);
		referenceComp.setLayoutData(new GridData(SWT.DEFAULT, SWT.FILL, false, true));
		GridLayout glstr = new GridLayout(1, false);
		referenceComp.setLayout(glstr);

		new GridData(SWT.DEFAULT, SWT.FILL, false, true);
		stratReference = new WdgReferenceData(referenceComp, SWT.NONE);
		stratReference.setLayoutData(gd);

		referencedataTab = new TabItem(ctFolder, SWT.NONE);
		referencedataTab.setText("Reference Data");
		referencedataTab.setControl(referenceComp);

		labelComp = new Composite(ctFolder, SWT.BORDER);
		labelComp.setBackground(GUI.COLOR_WHITE);
		labelComp.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false,
				4, 1));
		GridLayout gl = new GridLayout(4, false);
		gl.verticalSpacing = 5;
		gl.horizontalSpacing = 15;
		labelComp.setLayout(gl);

		stateLabel = new Label(labelComp, SWT.BORDER | SWT.CENTER);
		gd = new GridData(SWT.FILL, SWT.DEFAULT, true, false, 4, 1);
		stateLabel.setLayoutData(gd);
		stateLabel.setText("... Loading ...");

		Font font = GUI.FONT_10;
		strategyButtons = new Button[NUMBER_OF_ANGLES];
		for (ModelMultiStrategy.MultiStrategyField field : displayFields) {
			
			Text valueText = (field.getDisplayLines() > 2) ? new Text(
					labelComp, SWT.WRAP | SWT.V_SCROLL | SWT.MULTI | SWT.BORDER)
					: new Text(labelComp, SWT.WRAP | SWT.MULTI);

			valueText.setFont(font);
			gd = new GridData(SWT.FILL, SWT.CENTER, true, false);
			if (field.getDisplayLines() > 1)
				gd.horizontalSpan = 3;
			if (field.getDisplayLines() > 2) {
				GC gc = new GC(labelComp);
				gc.setFont(font);
				int textHeight = gc.textExtent("1234567890").y;
				gd.heightHint = textHeight * field.getDisplayLines();
			}
			valueText.setEditable(false);
			valueText.setLayoutData(gd);

			if (null != field)
				fieldTexts.put(field, valueText);
			if (field == ModelMultiStrategy.MultiStrategyField.WARNING)
				valueText.setForeground(GUI.COLOR_RED);
			
			if (field == ModelMultiStrategy.MultiStrategyField.STRATEGY) {
			
				for(int j = 0; j < NUMBER_OF_ANGLES; j++) {
					gd = new GridData(SWT.FILL, SWT.CENTER, true, false);
					gd.horizontalSpan = 4;
					strategyButtons[j] = new Button(labelComp, SWT.RADIO);
					strategyButtons[j].setLayoutData(gd);
				}
			}
			
			valueText.setBackground(GUI.COLOR_WHITE);
		}
		
		gd = new GridData(SWT.FILL, SWT.DEFAULT, true, false);
		gd.horizontalSpan = 3;

		singleStrategyButton = new Button(expandComp, SWT.CHECK | SWT.BORDER);
		singleStrategyButton.setText("Consider single crystal strategy");
		singleStrategyButton.setFont(GUI.FONT_COLLECT_GROUP);
		singleStrategyButton.setLayoutData(gd);
		singleStrategyButton.setBackground(GUI.BGCOLOR);
		singleStrategyButton.setToolTipText("Check this button if you want to export single crystal startegy.\nIt will export detector distance and oscillation angle shown in Strategy tab");
		singleStrategyButton.setSelection(true);
		
		TabItem strategyTab = new TabItem(ctFolder, SWT.NONE);
		strategyTab.setText("XPLAN Strategy");
		strategyTab.setControl(labelComp);

		Composite allComp = new Composite(ctFolder, SWT.NONE);
		allComp.setBackground(GUI.COLOR_WHITE);
		allComp.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
		GridLayout glall = new GridLayout(1, false);
		allComp.setLayout(glall);
		/*
		 * new GridData(SWT.FILL, SWT.FILL, true, true); allTable = new
		 * WdgMultiStrategyTable(allComp, SWT.NONE); allTable.setLayoutData(gd);
		 * 
		 * allTab = new CTabItem(ctFolder, SWT.NONE);
		 * allTab.setText("All Results"); allTab.setControl(allComp);
		 */

		/*Composite mergeComp = new Composite(ctFolder, SWT.NONE);
		mergeComp.setBackground(GUI.BGCOLOR);
		mergeComp.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
		GridLayout glmerge = new GridLayout(1, false);
		mergeComp.setLayout(glmerge);

		new GridData(SWT.FILL, SWT.FILL, true, true);
		WdgMergeDatasets mergeD = new WdgMergeDatasets(mergeComp, SWT.NONE);
		mergeD.setLayoutData(gd);

		TabItem mergeTab = new TabItem(ctFolder, SWT.NONE);
		mergeTab.setText("Combine Datasets");
		mergeTab.setControl(mergeComp);*/

		ctFolder.setSelection(0);
	}

	protected void createControls() {
		
		controlComp = new Composite(expandComp, SWT.NONE);
		controlComp.setBackground(GUI.BGCOLOR);
		GridData gd = new GridData(SWT.FILL, SWT.DEFAULT, true, false, 2, 1);
		controlComp.setLayoutData(gd);
		GridLayout gl = new GridLayout(6, false);
		gl.horizontalSpacing = 0;
		controlComp.setLayout(gl);

		int clearButtonWidth = 80;
		int moveButtonWidth = 70;
		Font buttonFont = GUI.FONT_12;

		startButton = new Button(controlComp, SWT.PUSH);
		startButton.setText("|<<");
		startButton.setFont(buttonFont);
		startButton.setBackground(GUI.COLOR_BLUE_BUTTON);
		gd = new GridData();
		gd.widthHint = moveButtonWidth;
		startButton.setLayoutData(gd);
		startButton.addSelectionListener(new SelectionListener() {
			@Override
			public void widgetDefaultSelected(SelectionEvent se) {
			}

			@Override
			public void widgetSelected(SelectionEvent se) {
				ModelMultiStrategy currentResult = ModelMultiStrategyList.get()
						.getResultIndex(0);
				if (currentResult != null) {

					DevMultiStrategy.setCurrentMultiStrategyIndex(0);
				}
			}
		});

		prevButton = new Button(controlComp, SWT.PUSH);
		prevButton.setText("<");
		prevButton.setFont(buttonFont);
		prevButton.setBackground(GUI.COLOR_BLUE_BUTTON);
		gd = new GridData();
		gd.widthHint = moveButtonWidth;
		prevButton.setLayoutData(gd);
		prevButton.addSelectionListener(new SelectionListener() {
			@Override
			public void widgetDefaultSelected(SelectionEvent se) {
			}

			@Override
			public void widgetSelected(SelectionEvent se) {

				int prev = ModelMultiStrategyList.get()
						.getCurrentMultiStrategyIndex() - 1;
				ModelMultiStrategy currentResult = ModelMultiStrategyList.get()
						.getResultIndex(prev);
				if (currentResult != null) {
					DevMultiStrategy.setCurrentMultiStrategyIndex(prev);
				}
			}
		});

		nextButton = new Button(controlComp, SWT.PUSH);
		nextButton.setText(">");
		nextButton.setFont(buttonFont);
		nextButton.setBackground(GUI.COLOR_BLUE_BUTTON);
		gd = new GridData();
		gd.widthHint = moveButtonWidth;
		nextButton.setLayoutData(gd);
		nextButton.addSelectionListener(new SelectionListener() {
			@Override
			public void widgetDefaultSelected(SelectionEvent se) {
			}

			@Override
			public void widgetSelected(SelectionEvent se) {

				int next = ModelMultiStrategyList.get()
						.getCurrentMultiStrategyIndex() + 1;
				ModelMultiStrategy currentResult = ModelMultiStrategyList.get()
						.getResultIndex(next);
				if (currentResult != null) {

					DevMultiStrategy.setCurrentMultiStrategyIndex(next);
				}
			}
		});

		lastButton = new Button(controlComp, SWT.PUSH);
		lastButton.setText(">>|");
		lastButton.setFont(buttonFont);
		lastButton.setBackground(GUI.COLOR_BLUE_BUTTON);
		gd = new GridData();
		gd.widthHint = moveButtonWidth;
		lastButton.setLayoutData(gd);
		lastButton.addSelectionListener(new SelectionListener() {
			@Override
			public void widgetDefaultSelected(SelectionEvent se) {
			}

			@Override
			public void widgetSelected(SelectionEvent se) {

				int last = ModelMultiStrategyList.get().getMultiStrategyList()
						.size() - 1;
				ModelMultiStrategy currentResult = ModelMultiStrategyList.get()
						.getResultIndex(last);
				if (currentResult != null) {

					DevMultiStrategy.setCurrentMultiStrategyIndex(last);
				}
			}
		});

		positionLabel = new Label(controlComp, SWT.CENTER);
		positionLabel.setText("");
		positionLabel.setBackground(GUI.BGCOLOR);
		gd = new GridData(GridData.FILL_HORIZONTAL);
		gd.minimumWidth = 100;
		positionLabel.setLayoutData(gd);

		deleteButton = new Button(controlComp, SWT.PUSH);
		deleteButton.setText("Export All");
		deleteButton.setToolTipText("save all strategy results to excel");
		deleteButton.setBackground(GUI.COLOR_BLUE_BUTTON);
		gd = new GridData();
		gd.widthHint = clearButtonWidth;
		deleteButton.setLayoutData(gd);
		deleteButton.addSelectionListener(new SelectionListener() {
			@Override
			public void widgetDefaultSelected(SelectionEvent se) {
			}

			@Override
			public void widgetSelected(SelectionEvent se) {
				exportToSpreadsheet();
			}
		});
	}
	protected void setButtonsVisible(boolean visible) {
		WdgControl.setControlVisible(logComp, !visible);
		WdgControl.setControlVisible(controlComp, !visible);
		expandComp.layout();
	}
	protected void setStateLabel(MultiStrategyState ss) {
		if (null == ss) {
			stateLabel.setBackground(GUI.COLOR_LIGHT_GRAY);
			stateLabel.setForeground(GUI.COLOR_BLACK);
			stateLabel.setText("---");
			return;
		}

		switch (ss) {
		case INACTIVE:
			stateLabel.setBackground(GUI.COLOR_LIGHT_GRAY);
			stateLabel.setForeground(GUI.COLOR_BLACK);
			stateLabel.setText("Inactive");
			break;

		case ERROR:
			stateLabel.setBackground(GUI.COLOR_RED_BUTTON);
			stateLabel.setText("Error");
			break;

		case PROCESSING:
			stateLabel.setBackground(GUI.COLOR_DARK_CYAN);
			stateLabel.setForeground(GUI.COLOR_WHITE);
			stateLabel.setText("Processing");
			break;

		case DONE:
			stateLabel.setBackground(GUI.COLOR_DARK_GREEN);
			stateLabel.setForeground(GUI.COLOR_WHITE);
			stateLabel.setText("Done");

			break;
		}
	}

	@Override
	public void handleEvent(ModelMultiStrategyList resultList, Object data) {
		updateLabels();
	}

	protected void updateLabels() {
		GUI.asyncExec(new Runnable() {
			public void run() {
				ModelMultiStrategyList resultList = ModelMultiStrategyList
						.get();

				// position in list
				int listSize = resultList.getMultiStrategyList().size();

				if (listSize >= 1) {
					positionLabel.setText((resultList
							.getCurrentMultiStrategyIndex() + 1)
							+ " of "
							+ listSize);
				} else {
					positionLabel.setText("---");
				}

				// fields
				ModelMultiStrategy currentResult = resultList
						.getCurrentResult();
				if (null != currentResult) {
					setStateLabel(currentResult.getState());

					for (MultiStrategyField field : currentResult
							.getAllFields().keySet()) {
						
						if (field == ModelMultiStrategy.MultiStrategyField.STRATEGY) {
							
							String strategy = currentResult.getField(field).trim();
							String lines[] = strategy.replaceAll("\\r", "").split("\\n");
							int i=0;
							
							if(lines.length == 0) {
								for(int j = 0; j < NUMBER_OF_ANGLES; j++) {
									if(strategyButtons[j]!=null)
										strategyButtons[j].setText("");
								}
							}
							
							for(String line:lines) {

								if(strategyButtons[i]!=null)
									strategyButtons[i].setText(line);
								i++;
							}
							
						} else {
							
							Text text = fieldTexts.get(field);

							if (null != text) {
								String value = currentResult.getField(field);

								if (null == value) {
									text.setText("---");
								} else {
									if (!value.equals(""))
										text.setText(value.trim());
									else
										text.setText("---");
								}
							}
						}
					}

				} else {
					setStateLabel(null);
					xplanLogButton.setEnabled(false);
					xdsLogButton.setEnabled(false);

					for (Text text : fieldTexts.values()) {
						if (null != text)
							text.setText("---");
					}
				}

				updateEnabled();
				ctFolder.setSelection(1);
				setButtonsVisible(false);
				
				labelComp.layout(true);
			}
			
		});
	}

	protected void updateEnabled() {
		
		ModelMultiStrategy currentResult = ModelMultiStrategyList.get()
				.getCurrentResult();
		final MultiStrategyState state = (null == currentResult) ? null
				: currentResult.getState();

		GUI.asyncExec(new Runnable() {
			public void run() {
				boolean enableButtons = (null != state && state == MultiStrategyState.DONE);

				xplanLogButton.setEnabled(enableButtons);

				boolean zeroResults = ModelMultiStrategyList.get()
						.getMultiStrategyList().size() == 0;
				boolean atStart = ModelMultiStrategyList.get()
						.getCurrentMultiStrategyIndex() <= 0;
				boolean atEnd = ModelMultiStrategyList.get()
						.getCurrentMultiStrategyIndex() >= ModelMultiStrategyList
						.get().getMultiStrategyList().size() - 1;

				xdsLogButton.setEnabled(Client.isMaster() && !zeroResults);
				xplanLogButton.setEnabled(Client.isMaster() && !zeroResults);
				exportStrategy.setEnabled(Client.isMaster() && !zeroResults);

				deleteButton.setEnabled(Client.isMaster() && !zeroResults);
				startButton.setEnabled(Client.isMaster() && !atStart
						&& !zeroResults);
				prevButton.setEnabled(Client.isMaster() && !atStart
						&& !zeroResults);
				nextButton.setEnabled(Client.isMaster() && !atEnd
						&& !zeroResults);
				lastButton.setEnabled(Client.isMaster() && !atEnd
						&& !zeroResults);
			}
		});
	}

	protected void exportToSpreadsheet() {
		
		FileDialog dialog = new FileDialog(GUI.getShell(), SWT.SAVE);
		dialog.setFilterNames(new String[] { "Excel Files", "All Files (*.*)" });
		dialog.setFilterExtensions(new String[] { "*.xls" });

		String savePath = "";
		savePath = DevDataDir.getPrimary();

		dialog.setFilterPath(savePath);
		dialog.setFileName("MultiStrategyessingResults.xls");

		ModelMultiStrategyList.get().save(dialog.open());
	}

	@Override
	public void masterChanged(boolean isMaster) {
		
		updateEnabled();
	}
	
	@Override
	public void currentRunChanged(ModelRun run) {
		updateExportLabel();
	}

	protected void updateExportLabel() {
		final ModelCollectRunList runList = ModelCollectRunList.get();

		int runCount = runList.getRunCount();
		if (runCount >= 0) {
			ModelCollectRun lastRun = (ModelCollectRun) runList
					.getRun(runCount);
			if (null != lastRun) {
				final int nextRunLabel = lastRun.getRunLabel() + 1;
				
				GUI.asyncExec(new Runnable() {
					public void run() {
						exportStrategy.setText("Create run "
								+ nextRunLabel);

					}
				});
			}
		}
	}

	@Override
	public void runListChanged(ModelRunList runList) {
		updateExportLabel();
		
	}
}
