package gov.anl.gmca.raster;

import gov.anl.gmca.collect.CmdCheck;
import gov.anl.gmca.collect.DevBinMode;
import gov.anl.gmca.collect.DevBinMode.BinMode;
import gov.anl.gmca.collect.DevBinMode.ModeChange;
import gov.anl.gmca.common.Command;
import gov.anl.gmca.common.DevMove.MoveCmd;
import gov.anl.gmca.common.DevMove.MoveResult;
import gov.anl.gmca.common.SafeImageDelete;
import gov.anl.gmca.common.ui.NBDlgDelete;
import gov.anl.gmca.hutch.DetectorConfig;
import gov.anl.gmca.hutch.DevAttenuation;
import gov.anl.gmca.hutch.DevPmacMotor;
import gov.anl.gmca.io.Disk;
import gov.anl.gmca.raster.ModelRasterRun.RunState;
import gov.anl.gmca.raster.ModelRasterRun.RunType;
import gov.anl.gmca.raster.ModelRasterState.RasterListType;
import gov.anl.gmca.raster.ModelRasterState.RasterState;
import gov.anl.gmca.sample.DevCollimatorCtrl;
import gov.anl.gmca.sample.DevCollimatorCtrl.CCMove;
import gov.anl.gmca.sample.DevCollimatorCtrl.CollimatorSize;
import gov.anl.gmca.sample.DevSlider;
import gov.anl.gmca.sample.ModelBeamsize;
import gov.anl.gmca.spotfinder.DBImageStats;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.ActionResult;
import gov.anl.gmca.util.ListenerList;
import gov.anl.gmca.util.Point2D;
import gov.anl.gmca.util.Util;
import org.eclipse.swt.SWT;

import java.io.File;
import java.io.FilenameFilter;
import java.util.ArrayList;
import java.util.Collections;

public abstract class CmdRaster extends Command
{
	public final static int STOP = 1;
	protected boolean DISABLE_PROBLEM_MOTIONS = false;
	
	@Override public abstract void run();
	
	protected ModelRasterRun run = null;
	protected ListenerList.Listener<ModelRasterRun> runListener = null;
	protected ListenerList.Listener<ModelRasterState> stateListener = null;
	
	protected Point2D origGSSize;
	protected CollimatorSize origCollimator;
	
	protected CmdRaster(ModelRasterRun run)
	{
		this.run=run;
		DISABLE_PROBLEM_MOTIONS = Util.getBID().equals("23b");
		
		// check that the list and run states haven't been reset
		runListener = new ListenerList.Listener<ModelRasterRun>()
		{
			@Override public void handleEvent(ModelRasterRun run, Object data)
			{
				if (run.getRunState() != RunState.RASTERING)
				{
					run.dev().setRunStatus(RunState.RASTERING, true);
				}
			}
		};
		stateListener = new ListenerList.Listener<ModelRasterState>()
		{
			@Override public void handleEvent(ModelRasterState object, Object data)
			{
				if (ModelRasterState.get().getRasterState() != RasterState.RASTERING &&
					ModelRasterState.get().getRasterState() != RasterState.STOPPING)
				{
					DevRasterState.get().setRasterState(RasterState.RASTERING);
				}
			}
		};
	}
	
	protected void startStateCheck()
	{
		run.llRunChanged.addListener(runListener);
		ModelRasterState.get().llCurrentRun.addListener(stateListener);
	}
	
	protected void stopStateCheck()
	{
		if (null != runListener) { run.llRunChanged.removeListener(runListener); runListener = null; }
		if (null != stateListener) { ModelRasterState.get().llCurrentRun.removeListener(stateListener); stateListener = null; }
	}
	
	protected ActionResult preRasterChecks()
	{
		CmdCheck cc = new CmdCheck(CmdCheck.FLAGS_PRERASTER);
		cc.run();
		
		if (!cc.checksPassed()) return new ActionResult("Checks failed, aborting rastering.");
		return ActionResult.SUCCESS;
	}
	
	protected ActionResult checkDirectory(final ModelRasterRun run)
	{
		if (DetectorConfig.isEiger() && run.getNextFrame()!=0) {
			return new ActionResult("Resuming raster isn't currently supported with the Eiger");
		}

		// check filename length
		if (run.getFileRoot().length() + 7 > 38) // 7 = length of "_0.0001" filename suffix
		{
			return new ActionResult("Filename length may not exceed 38 characters");
		}
		
		// check that a file doesn't exist with the same name as the collect dir
		File rasterDir = new File(run.getFullDirectory());
		if (rasterDir.isFile())
		{
			boolean okToDelete = Util.popupOkCancel("File deletion required","A file, "+run.getFullDirectory()+", exists with the same name as the output directory.  Delete this file?");
			
			if (!okToDelete) 			return new ActionResult("Deletion declined, rastering aborted");
			if (!rasterDir.delete())	return new ActionResult("Error creating the directory "+run.getFullDirectory()+", rastering aborted.");
		}
		
		// create the run dir if necessary
		if (!rasterDir.exists())
		{
			if (!rasterDir.mkdirs()) return new ActionResult("Error creating the directory "+run.getFullDirectory()+", rastering aborted.");
		}
		
		// make sure it's writable
		if (!rasterDir.canWrite()) return new ActionResult("The directory "+run.getFullDirectory()+" is not writable, rastering aborted.");
		
		if (run.getRunType() == RunType.DIFFRACTION)
		{
    		// remove a temporary file leftover from DISTL
    		File pickleFile = new File(run.getFullDirectory()+"/DISTL_pickle");
    		if (pickleFile.exists())
    		{
        		boolean pickleDeleted = pickleFile.delete();
        		if (!pickleDeleted) Log.warn("Could not delete DISTL_pickle from the selected raster directory");
    		}
    		
			// make a list of files to check
			ArrayList<SafeImageDelete.FileToDelete> filesToCheck = new ArrayList<SafeImageDelete.FileToDelete>();

			if (run.getRunList().getListType() == RasterListType.INTER)
			{
				for (ModelRasterCell cell : run.grid().getGridCalc().getCellListCopy())
				{
					if (cell.getIndex() < run.getNextFrame()) continue;

					long modTime=(new File(run.getFullDirectory()+"/"+cell.getImageName())).lastModified() / 1000;
					String filename = DetectorConfig.isEiger() ? cell.getCBFImageName() :  cell.getImageName();
					SafeImageDelete.FileToDelete fileToDelete = new SafeImageDelete.FileToDelete(run.getFullDirectory(), filename, modTime);
					filesToCheck.add(fileToDelete);
				}
			}
			else
			{
				File runDir = new File(run.getFullDirectory());
				File[] fileList = runDir.listFiles(new FilenameFilter()
				{
					@Override public boolean accept(File dir, String name)
					{
						try
						{
							// prefix
							String prefix = run.getFileRoot();
							if (name.indexOf(prefix) != 0) return false;
							name = name.substring(prefix.length());

							// _
							if (name.length()==0 || name.charAt(0) != '_') return false;
							name = name.substring(1);

							// "A" + run number + '.'
							if (name.length() == 0 || name.indexOf('A') != 0 || name.indexOf('.') < 0) return false;
							String token = name.substring(0, name.indexOf('.'));
							try
							{
								Integer.parseInt(token);
							}
							catch (NumberFormatException nfe) { return false; }
							name = name.substring(name.indexOf('.')+1);

							// frame number
							if (name.length() == 0) return false;
							try
							{
								Integer.parseInt(name);
							}
							catch (NumberFormatException nfe) { return false; }

							return true;
						}
						catch (Exception e) { e.printStackTrace(); }
						return true;
					}
				});

				for (File file : fileList)
				{
					long modTime=(new File(run.getFullDirectory()+"/"+file.getName())).lastModified() / 1000;
					SafeImageDelete.FileToDelete fileToDelete = new SafeImageDelete.FileToDelete(run.getFullDirectory(), file.getName(), modTime);
					filesToCheck.add(fileToDelete);
				}
				Collections.sort(filesToCheck);
			}

			// check if they exist
			ArrayList<SafeImageDelete.FileToDelete> existingFiles = new ArrayList<SafeImageDelete.FileToDelete>();
			for (SafeImageDelete.FileToDelete fileToCheck : filesToCheck)
			{
				if (Disk.exists(fileToCheck.getFullPath()))
				{
					//System.out.println("CmdRaster.run(): "+fileToCheck+" exists");
					existingFiles.add(fileToCheck);
				}
			}

			// prompt user if they exist, abort if user declines
			if (0 != existingFiles.size())
			{
				Log.note("Asking to delete existing files");

				NBDlgDelete dlgDelete = new NBDlgDelete(existingFiles);
				int okToDelete = dlgDelete.open();

				Log.note("User chose "+(okToDelete==SWT.OK?"":"not ")+" to delete files");

				if (SWT.OK == okToDelete)
				{
					SafeImageDelete.delete(existingFiles, true, DetectorConfig.isPilatus());
					DBImageStats.delete(run.directorySuffix, run.fileRoot, run.runLabel, null);
				}
				else
				{
					return new ActionResult("Deletion declined, aborting rastering.");
				}
			}
		}

		return ActionResult.SUCCESS;
	}
	
	protected ActionResult preRasterMoves()
	{
		ModeChange binModeChange = null;
		if (run.getRunType()==RunType.DIFFRACTION)
		{
			// set the detector binning mode
			BinMode binMode = ModelRasterState.get().getBinMode();
			binModeChange = DevBinMode.get().startChange(binMode);
		}
		
	    // move the detector if needed
		MoveCmd moveDetector = null;
		if (!DISABLE_PROBLEM_MOTIONS && DevPmacMotor.get("detector_z").isMoveNeededTo(run.getDetectorDist()))
		{
			moveDetector = DevPmacMotor.get("detector_z").startMove(run.getDetectorDist());
		}

		// set up guardslits and collimator (wait is included)
		Object[] setupRetVal = ModelBeamsize.setUpBeamsize(run.getBeamWidth_um(), run.getBeamHeight_um(), DISABLE_PROBLEM_MOTIONS);
									origCollimator 	= (CollimatorSize)	setupRetVal[0];
		if (null!=setupRetVal[1]) 	origGSSize 		= (Point2D)			setupRetVal[1];
		ActionResult 				setupResult 	= (ActionResult)	setupRetVal[2];
		if (!setupResult.succeeded()) return setupResult;
		
		// wait for detector move
	    if (null != moveDetector) moveDetector.waitFor();
	    if (null != moveDetector && !moveDetector.getResult().isOK()) return new ActionResult("Pre-raster detector move failed, aborting raster.");

	    // change attenuation (synchronous)
	    // attenuation move needs to be after energy move
		MoveCmd moveAtten 			= DevAttenuation.get().startMove(run.getAttenuation());
		MoveResult attenResult 		= moveAtten.waitFor();
		if (!attenResult.isOK())	return new ActionResult("Pre-raster attenuation move failed, aborting raster.");
		
		// wait for the binning mode change
		if (null != binModeChange)
		{
			ActionResult binModeResult;
			try
			{
				binModeResult = binModeChange.waitFor();
			}
			catch (Exception e)
			{
				e.printStackTrace();
				binModeResult = ActionResult.FAILURE;
			}
			if (!binModeResult.succeeded())	return new ActionResult
			(
				"Binning mode change failed, aborting raster.\n\n"+DevBinMode.RESTART_MESSAGE
			);
		}
		
		if (run.getRunList().getListType()!=RasterListType.AUTO)
		{
			// turn the ringlight on
			DevSlider ringLight = DevSlider.get("ring");
			ringLight.setIntensity(ringLight.getHigherLimit());
		}
		
		return ActionResult.SUCCESS;
	}
	
	protected CCMove ccMove;
	protected MoveCmd moveGSX = null, moveGSY = null;
	
	protected void postRasterMoves()
	{
		if (run.getRunList().getListType()!=RasterListType.AUTO)
		{
			// ringlight off
			DevSlider ringLight = DevSlider.get("ring");
			ringLight.setIntensity(ringLight.getLowerLimit());
		}
	}
	
	protected void startRestoreBeamsize()
	{
		if (null != origGSSize)
		{
			moveGSX = DevPmacMotor.get("beam_size_x").startMove(origGSSize.x);
			moveGSY = DevPmacMotor.get("beam_size_y").startMove(origGSSize.y);
		}
		if (null != origCollimator)
		{
			DevCollimatorCtrl.setCollimatorSize(origCollimator);
			
			// backlight in
		    ccMove = DevCollimatorCtrl.get().newMove();
		    ccMove.setArguments("OUT backlightIN bluice");
		    /*ccMove.setDest("collimator",	InOut.OUT);
		    ccMove.setDest("backlight",		InOut.IN);
		    ccMove.setDest("beamstop", 		InOut.OUT);*/
		    ccMove.start();
		}
	}
	
	protected void waitForRestoreBeamsize()
	{
		if (null != moveGSX) moveGSX.waitFor();
		if (null != moveGSY) moveGSX.waitFor();
		if (null != ccMove) ccMove.waitFor();
		
		moveGSX = null;
		moveGSY = null;
		ccMove = null;
	}
}
