package gov.anl.gmca.raster;

import gov.anl.gmca.collect.DevFPE;
import gov.anl.gmca.collect.DevFPE.FPEMode;
import gov.anl.gmca.collect.ETA;
import gov.anl.gmca.collect.ModelCollectRun.Axis;
import gov.anl.gmca.common.DBConfig;
import gov.anl.gmca.common.DevDataDir;
import gov.anl.gmca.common.DevMove.MoveCmd;
import gov.anl.gmca.common.Redis;
import gov.anl.gmca.common.ui.NBMessageBox;
import gov.anl.gmca.gui.*;
import gov.anl.gmca.gui.state.SWServer;
import gov.anl.gmca.hutch.DetectorConfig;
import gov.anl.gmca.hutch.DevPmacMotor;
import gov.anl.gmca.hutch.DevXYZ;
import gov.anl.gmca.log.db.RasterEvent;
import gov.anl.gmca.raster.ModelRasterRun.RunState;
import gov.anl.gmca.raster.ModelRasterState.RasterListType;
import gov.anl.gmca.raster.ModelRasterState.RasterState;
import gov.anl.gmca.raster.calc.FramePartitioner.FramePartition;
import gov.anl.gmca.raster.ui.TabRaster;
import gov.anl.gmca.screening.DevRobotWarmup;
import gov.anl.gmca.spotfinder.*;
import gov.anl.gmca.spotfinder.SFData.SFType;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.users.ModelUserTable.Client;
import gov.anl.gmca.util.ActionResult;
import gov.anl.gmca.util.ListenerList;
import gov.anl.gmca.util.Util;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Shell;

import java.io.File;
import java.util.ArrayList;

// The 3D cell positions are calculated here
public class CmdRasterDiff extends CmdRaster
{
	protected volatile boolean	finalMovesComplete	= false;
	protected boolean 			requestStop 		= false;
	protected boolean			isProcessingDone	= false;
	protected boolean 			isInteractiveRun;
	protected ActionResult		rasterResult		= null;
	protected DevFPE.FrameList 	currentFPEBatch		= null;
	protected FramePartition	currentPartition	= null;
	protected double			startOmega			= 0;
	protected int lastProcessedFrameIndex = 0;
	protected ArrayList<FramePartition> currentPartitionList = null;

	protected ArrayList<ModelRasterCell> cellListCopy;
	protected ListenerList.Listener<DevFPE> frameListener;
	
	///////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////// end variables //////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////
	
	public CmdRasterDiff(ModelRasterRun _run)
	{
		super(_run);

		// get a copy of the cell list so it doesn't change during rastering
		cellListCopy = run.grid().getGridCalc().getCellListCopy();
		
		RasterListType mode=_run.getRunList().getListType();
		isInteractiveRun = (mode == RasterListType.INTER);
		
		frameListener = new ListenerList.Listener<DevFPE>()
		{
			@Override public void handleEvent(DevFPE dev, Object frameNumber)
			{
				handleCellFrameAdvance();
			}
		};
		DevFPE.get().frameListener.addListener(frameListener);
	}
	
	// used for stand-alone (interactive) raster
	@Override public void run()
	{
		// check FPE connection
		if (!DevFPE.get().checkFPEState(null, false))
		{
			endRaster(false, false);
			return;
		}
		
		// check total helical raster angle
		if (run.isShutterless()) {
			double largestTotalRotation = 0.0;
			for (FramePartition part : run.grid().getFramePartitions())
			{
				double totalRotation = part.frameCount * run.getDelta();
				if (totalRotation>largestTotalRotation) largestTotalRotation = totalRotation;
			}
			if (largestTotalRotation>90.0) {
				boolean okToProceed = Util.popupOkCancel
				(
					"Large total rotation",
					"This raster run has a large total rotation within a row of "+
					Util.printDouble(largestTotalRotation, 1)+
					" degrees.  Proceed?"
				);
				if (!okToProceed) {
					Log.note("Raster cancelled by user.");
					endRaster(false,false);
					return;
				}
			}
		}
		
		ActionResult result = runAll();
		if (!result.succeeded())
		{
			errorPopupAfterExit("Raster error", result.getError());
			endRaster(false, false);
		}
	}
	
	// called directly from auto-raster
	public ActionResult runAll()
	{
		// set states
		setInitialStates();
		startStateCheck();

		// check directory
		if (isInteractiveRun)
		{
    		ActionResult dirResult 			= checkDirectory(run);
    		if (!dirResult.succeeded()) 	return dirResult;
		}

		if (!Main.SIM_FPE) {
			// pre-raster checks
			ActionResult checksResult = preRasterChecks();
			if (!checksResult.succeeded()) {
				if (!isInteractiveRun) endRaster(false, false);
				return checksResult;
			}

			// pre-raster moves
			ActionResult moveResult = preRasterMoves();
			if (!moveResult.succeeded()) {
				if (!isInteractiveRun) endRaster(false, false);
				return moveResult;
			}
		}

		// raster
		return raster();
	}

	// the main raster loop
	protected ActionResult raster()
	{
		// stop previous threads when raster starts
		SFType sfType = run.getRunList().getListType()==RasterListType.AUTO?SFType.AUTO_RASTER:SFType.INTER_RASTER;
		SpotFinderQueue.stopActiveThreads(sfType, run.getRunIndex());
		Util.sleep(200);
		
		// set up processing monitor and cell list
		rasterStartup();
		
		// send robot warm-up if needed
		int etaSec = ETA.calcActiveETA();
		Log.note("ETA for robot warmup check: "+etaSec);
		DevRobotWarmup.get().sendWarmupIfNeeded(etaSec);

		// raster!
		currentPartitionList = run.grid().getFramePartitions();
		for (FramePartition partition : currentPartitionList)
		{
			if (partition.partitionIndex < 0) continue; // frames already taken are in partition -1
			currentPartition = partition;

			if (!activeCheck()) break;
			checkForPause();
			if (checkForBreak()) break;

			startPartition();
			waitForCurrentPartition();
		}

		// cleanup
		endRaster(!checkForBreak(), false);
		DevFPE.get().setETA(0);
		
		// see if the loop was broken with a result
		if (checkForBreak()) return rasterResult;

		return ActionResult.SUCCESS;
	}
	
	protected boolean checkForBreak() { return (null != rasterResult); }

	// returns false if raster ended
	protected void checkForPause()
	{
		if (requestStop) { endRaster(false, false); rasterResult = ActionResult.SUCCESS; return; }

		if (RasterState.STOPPING == ModelRasterState.get().getRasterState())
		{
			pauseRaster("Pause requested, ending raster.", false);
			run.dev().sendUpdate();
			rasterResult = ActionResult.SUCCESS;
			return;
		}
	}
	
	private boolean activeCheck()
	{
		if (!Client.isMaster())
		{
			Log.note("Client became slave during collection, exiting raster thread.");
			return false;
		}
		return true;
	}
	
	/////////////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////// helper methods ///////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////////////

	protected void setInitialStates()
	{
		// initialize image stats
		totalCellCount = cellListCopy.size();
		DBImageStats.initialize(Redis.image_stats(run.getDirectorySuffix(), run.getFileRoot(), run.getRunLabel(), ""), totalCellCount);

		run.dev().setRunStatus(RunState.RASTERING, true);
		if (isInteractiveRun) DevRasterState.get().setRasterState(RasterState.RASTERING);
	}
	
	protected void rasterStartup()
	{
		startOmega = DevPmacMotor.get("gonio_omega").getPosition();
		
		// start the monitor to watch for processing to end
		processedCellCount = 0;
		lastProcessedFrameIndex = -1;
		startProcessingMonitor();

		// clear the user position
		if (isInteractiveRun) run.dev().setUserPosition(null, false);
	}

	private void handleCellFrameAdvance()
	{
		try
		{
			if (null == cellListCopy) return;

			// keep sending pause each frame to make sure FPE does in fact pause
			/*if (RasterState.STOPPING == ModelRasterState.get().getRasterState())
			{
				DevFPE.get().sendPause();
			}*/
			
			int currentFrameIndex = DevFPE.get().getCurrentFrameIndex();
			Log.note("Current FPE frame is "+currentFrameIndex);

			// wait until a batch is ready
			int writtenFrameIndex = currentFrameIndex-1;
			int numFramesAvailable = writtenFrameIndex - lastProcessedFrameIndex;
			//double batchLengthSec = 0.2;
			int batchSize = 1;//(int)Math.round(batchLengthSec / run.getExposureTime());
			//if (batchSize<3) batchSize=3;
			//if (batchSize>10) batchSize=10;
			if (numFramesAvailable < batchSize && currentFrameIndex != run.grid().getSelectedCellCount()) return;

			// send batches if available
			int startIndex = lastProcessedFrameIndex +1;
			for (int frameIndex = startIndex; frameIndex<=writtenFrameIndex; frameIndex += batchSize)
			{
				ArrayList<String> filenames = new ArrayList<>();
				for (int i=0;i<batchSize&&frameIndex+i<run.grid().getSelectedCellCount();i++)
				{
					ModelRasterCell c = cellListCopy.get(frameIndex+i);
					filenames.add(c.getCBFImageName());
					lastProcessedFrameIndex = frameIndex+i;
				}

				for (DevSFType.SFConfig sfConfig : DevSFType.configList)
				{
					CmdSpotFinder cmdProcess = new CmdSpotFinderRemote
					(
						SFType.INTER_RASTER,
						DevSFType.SFRunType.RASTER,
						run.getRunIndex(),
						cellListCopy.get(frameIndex).index,
						run.getFullDirectory(),
						filenames,
						run.getFileRoot(),
						run.getRunLabel(),
						run.getLowResLimit_A(),
						run.getHighResLimit_A(),
						-1,
						sfConfig.id,
						null,
						CmdSpotFinder.TIMEOUT_DEFAULT_MS
					);
					SpotFinderQueue.submitTask(cmdProcess);
				}
				processedCellCount += filenames.size();
			}
		}
		catch (Exception e) { e.printStackTrace(); }
	}
	
	protected void startPartition()
	{
		if (requestStop) { endRaster(false, false); rasterResult = ActionResult.SUCCESS; return; }

		// create dir if needed
		File dir = new File(run.getFullDirectory());
		if (!dir.exists()) dir.mkdir();

		// determine FPE mode number
		FPEMode fpeMode = null;
		int frameCount = currentPartition.frameCount;
		if (run.isShutterless())
		{
			// one row raster
			if (currentPartitionList.size()==1 || !DetectorConfig.isEiger() || !DBConfig.getBoolean("raster_single_arm") || run.getDelta() > 0.0 )
			{
				fpeMode = FPEMode.SHUTTERLESS_ARM_DISARM;
			}
			// first row of a multi-row raster
			else if (currentPartition.partitionIndex==0)
			{
				fpeMode = FPEMode.SHUTTERLESS_ARM_NODISARM;
				//frameCount = cellListCopy.size();
			}
			// last row of a multi-row raster
			else if (currentPartition.partitionIndex==currentPartitionList.size()-1)
			{
				fpeMode = FPEMode.SHUTTERLESS_NOARM_DISARM;
				//frameCount = cellListCopy.size();
			}
			// middle row of a multi-row raster (not the first or last row)
			else
			{
				fpeMode = FPEMode.SHUTTERLESS_NOARM_NODISARM;
				//frameCount = cellListCopy.size();
			}
		}
		else
		{
			fpeMode = FPEMode.SHUTTERED_RASTER;
		}
		
		// calculate helical raster parameters if set
		double delta = run.getDelta();
		double startAngle = run.grid().getCenter().omega;
		double endAngle = startAngle;
		if (DBConfig.getBoolean("raster_helical")) {
			delta *= currentPartition.frameCount;
			startAngle -= delta/2.0;
			endAngle += delta/2.0;
		}
		
		// send partition to FPE
		currentFPEBatch 				= DevFPE.get().newFrameList();
		currentFPEBatch.partition		= currentPartition;
		currentFPEBatch.firstFrameLabel	= currentPartition.firstFrameIndex+1;
		currentFPEBatch.firstFrameIndex	= currentPartition.firstFrameIndex;
		currentFPEBatch.frameCount		= frameCount;
		currentFPEBatch.delta 			= run.getDelta();
		currentFPEBatch.energy 			= DevPmacMotor.get("mo_En_E").getPosition();
		currentFPEBatch.exposureTime 	= run.getExposureTime();
		currentFPEBatch.startAngle 		= startAngle;
		currentFPEBatch.nextAngle		= endAngle;
		currentFPEBatch.endAngle		= endAngle;
		currentFPEBatch.runNumber 		= run.getRunLabel();
		currentFPEBatch.filename 		= currentPartition.getFilenameToSendToFPE();
		currentFPEBatch.prefix			= currentPartition.prefix;
		currentFPEBatch.directory 		= DevDataDir.addToFPEDir(run.getDirectorySuffix());
		currentFPEBatch.axis 			= Axis.GONIO_PHI;
		currentFPEBatch.fpeMode			= fpeMode;
		currentFPEBatch.shutteredMotion	= currentPartition.shutteredMotion;
		currentFPEBatch.shutterlessMotion = currentPartition.shutterlessMotion;
		currentFPEBatch.start();
	}
	
	protected void waitForCurrentPartition()
	{
		if (null == currentFPEBatch) return;
		
        do
        {
            if (requestStop) currentFPEBatch.requestStop();
            Util.sleep(10);
        }
        while (currentFPEBatch.isRunning());
        
        if (!currentFPEBatch.getResult().isOK())
        {
			endRaster(true, false);
			rasterResult = new ActionResult("FPE reported a frame error.  Check the FPE window for details.");
        }
	}

	protected void errorPopupAfterExit(String title, String msg)
	{
		final String f_title = title;
		final String f_msg = msg;
		
		new Thread(new Runnable(){public void run()
		{
			//while (run.runList.isRasterThreadAlive()) Util.sleep(50);
			
			Shell shell = (null != GUI.getShell())?(GUI.getShell()):(new Shell(GUI.getDisplay()));
			NBMessageBox mb2 = new NBMessageBox(shell, SWT.APPLICATION_MODAL | SWT.ICON_INFORMATION);
			mb2.setText(f_title);
			mb2.setMessage(f_msg);
			mb2.open();
			Log.note(f_msg);
		}}, "CmdRasterDiff error popup").start();
	}

	protected void pauseRaster(String message, boolean moveToStart)
	{
		Log.note(message);
		endRaster(moveToStart, true);
	}
	
	protected void endRaster(boolean moveToStart, boolean paused)
	{
		stopStateCheck();
		
		// wait for all updates to process for this raster
		while (!DevFPE.get().isCacheIdle()) Util.sleep(10);

		// process remaining frames
		/*int remainingFrameCount = totalCellCount-lastProcessedFrameIndex;
		if (!paused && remainingFrameCount <=5)
		{
			for (int i=lastProcessedFrameIndex; i<totalCellCount; i++)
			{
				System.out.println("endRaster sending remaining cell to processing: "+i);
				ModelRasterCell cellToProcess = cellListCopy.get(i);
				startProcessing(cellToProcess);
			}
			lastProcessedFrameIndex=totalCellCount;
		}*/
		
		// remove listeners
		Log.note("Removing frame listener for run "+run.getRunIndex());
		if (DetectorConfig.isEiger()) SWServer.get().update();
		DevFPE.get().frameListener.removeListener(frameListener);

		// don't look for any more cell data
		if (paused) totalCellCount = processedCellCount;
		
		if (isInteractiveRun) DevRasterState.get().setRasterState(RasterState.STOPPING);

		if (!Main.SIM_FPE) {
			postRasterMoves();
			startRestoreBeamsize();
		}

		if (moveToStart)
		{
			// wait for FPE, since it could be moving omega and we need to return omega to its original position
			if (!DevFPE.get().isIdle())
			{
				Log.note("Waiting for FPE to finish...");

				int maxWait = 100;
				while (!DevFPE.get().isIdle())
				{
					Util.sleep(100); maxWait--; 
					if (0==maxWait)
					{
						Log.warn("Returning to center but FPE appears to still be running.");
						break;
					}
				}
			}
			
    		Log.note("Returning to center...");
    		
			// start omega move
			MoveCmd omegaMove = DevPmacMotor.get("gonio_omega").startMove(run.grid().getCenter().omega);
			
    		// do XYZ move
    		DevXYZ.syncMove(run.grid().getCenter().to3D(), true);

			// wait for gonio and beamsize moves
			omegaMove.waitFor();
			waitForRestoreBeamsize();
			
			finalMovesComplete = true;
    		Log.note("Done.");
		}
		else
		{
			waitForRestoreBeamsize();
		}
		
		// update the run state based on how many frames were completed
		final int cellCount = cellListCopy.size();
		RunState newState = RunState.PAUSED;
		if (run.getNextFrame() == 0) 			newState = RunState.INACTIVE;
		if (run.getNextFrame() == cellCount)	newState = RunState.COMPLETE;
		run.dev().setRunStatus(newState, true);
		
		// update the run list state
		if (isInteractiveRun) DevRasterState.get().setRasterState(RasterState.INACTIVE);
	}

	public void requestStop()
	{
		requestStop = true;
	}
	
	/////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////// processing monitor ///////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////
	
	protected ArrayList<CmdSpotFinder> processCmdList = new ArrayList<CmdSpotFinder>();
	protected int totalCellCount = 0, processedCellCount = 0;
	
	// TODO: add timeout so if no cells are processed within say 30 seconds, just write the log
	
	protected void startProcessingMonitor()
	{
		new Thread(new Runnable(){public void run()
		{
			while (true)
			{
				if (requestStop) return;
				
				// if all the processing commands started.. (don't check arraylist until it's full)
				if (processedCellCount >= totalCellCount && finalMovesComplete)
				{
					// break if all processing commands are done
					boolean allDone = true;
					for (CmdSpotFinder cmd : processCmdList)
					{
						//if (cmd.isAlive()) allDone = false;
					}
					if (allDone) break;
				}
				Util.sleep(100);
			}
			
			isProcessingDone = true;
			
			handleRasterEnd();
		}	
		}, "CmdRasterDiff processing monitor").start();
	}
	
	public void waitForProcessing()
	{
		while (!isProcessingDone)
		{
			System.out.println("Waiting for run "+run.getRunIndex()+" processing: "+processedCellCount+" of "+totalCellCount+" isProcDone="+isProcessingDone+" finalMovesComplete="+finalMovesComplete);
			Util.sleep(1000);
		}
	}
	
	private void handleRasterEnd()
	{
		// write raster log
		SFLog.writeLogFile(run.getRunList().getListType()==RasterListType.INTER?SFType.INTER_RASTER:SFType.AUTO_RASTER, run.getRunIndex(), null);
		
		// write to SQL log
		RasterEvent.writeEvent(run, startOmega);
		
		// save screenshot to a default (non-existant) filename
		String ssFilename = run.generateRasterScreenshotFilename();
		TabRaster.get().cam.saveScreenshot(ssFilename, false);
	}
}
