package gov.anl.gmca.raster;

import gov.anl.gmca.io.Database;
import gov.anl.gmca.io.Database.DBName;
import gov.anl.gmca.io.Database.Result;
import gov.anl.gmca.io.Database.WhereList;
import gov.anl.gmca.io.Database.WhereOp;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.PV_R_RUNS;
import gov.anl.gmca.raster.ModelRasterHits.Hit;
import gov.anl.gmca.raster.ModelRasterSiteList.Site;
import gov.anl.gmca.raster.ModelRasterState.RasterListType;
import gov.anl.gmca.util.Util;
import gov.aps.jca.dbr.DBR;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.StringTokenizer;

public class DBRasterHits extends DevDB
{
	protected static enum COLS
	{
		HIT_INDEX,
		INTER_RUN,	AUTO_RUN,
		CELL_INDEX_LIST;
		
		public String dbString() { return toString().toLowerCase(); }
	}
	
	public static synchronized boolean verifyTableExists()
	{
		if (!doesTableExist())
		{
			String createCmd =
				"CREATE TABLE `"+tableName+"` ("+
				"`hit_index` int(11) NOT NULL,"+
				"`inter_run` int(11) NOT NULL,"+
				"`auto_run` int(11) NOT NULL,"+
				"`cell_index_list` varchar(16000) NOT NULL,";
			createCmd += "PRIMARY KEY (`hit_index`,`inter_run`,`auto_run`))";
			Database.exec(DBName.USERDATA, createCmd);
		}
		return doesTableExist();
	}
	
	private int getLastHitIndex(int interRunIndex, int autoRunIndex)
	{
		Result result = Database.exec
		(
			DBName.USERDATA,
			"SELECT hit_index from "+tableName+
			" where inter_run='"+interRunIndex+"' and auto_run='"+autoRunIndex+"'"
		);
		
		int lastNumber=0;
		for (HashMap<String,String> row : result.getRows())
		{
			int number = 0;
			try
			{
				number = Integer.parseInt(row.get("hit_index"));
			}
			catch (NumberFormatException nfe) {}
			if (number > lastNumber) lastNumber = number;
		}
		
		return lastNumber;
	}
	
	// this updates both remotely and locally so the hit is available immediately to query
	public void addHit(int hitIndex, int interRunIndex, int autoRunIndex, Integer[] cellIndexList, boolean sendUpdate)
	{
		// add locally - it will be overwritten once the SQL update occurs
		ArrayList<Integer> cellList = new ArrayList<Integer>();
		for (Integer index : cellIndexList) cellList.add(index);
		Hit hit = new Hit(hitIndex, interRunIndex, autoRunIndex, cellList, ModelRasterState.get().getDiffDisplayParameter());
		ModelRasterHits.get().addHit(hit, false);
		
		verifyTableExists();
		
		// set last hit number if requested
		if (hitIndex < 1)
		{
			hitIndex = getLastHitIndex(interRunIndex,autoRunIndex)+1;
		}
		else // delete existing hit
		{
			WhereList wl = new WhereList();
			wl.add("hit_index", WhereOp.EQUALS, ""+hitIndex);
			wl.add("inter_run", WhereOp.EQUALS, ""+interRunIndex);
			wl.add("auto_run", WhereOp.EQUALS, ""+autoRunIndex);
			Database.delete(DBName.USERDATA, tableName, wl);
		}
		
		// construct hit array
		StringBuilder sb = new StringBuilder();
		for (int index : cellIndexList)
		{
			sb.append(index);
			sb.append(',');
		}
		sb.deleteCharAt(sb.length()-1);
		
		HashMap<String,String> row = new HashMap<String,String>();
		row.put("hit_index", 		""+hitIndex);
		row.put("inter_run",		""+interRunIndex);
		row.put("auto_run",			""+autoRunIndex);
		row.put("cell_index_list",	""+sb.toString());
		Database.insert(DBName.USERDATA, tableName, row);
		
		if (sendUpdate) sendUpdate();
	}
	
	public void deleteHit(int hitIndex, int interRunIndex, int autoRunIndex, boolean update)
	{
		// delete requested hit number
		WhereList whereList = new WhereList();
		whereList.add("hit_index", WhereOp.EQUALS, ""+hitIndex);
		whereList.add("inter_run", WhereOp.EQUALS, ""+interRunIndex);
		whereList.add("auto_run", WhereOp.EQUALS, ""+autoRunIndex);
		Database.delete(DBName.USERDATA, tableName, whereList);
		
		// update subsequent hit numbers
		for (int indexToUpdate=hitIndex+1; indexToUpdate<=getLastHitIndex(interRunIndex,autoRunIndex); indexToUpdate++)
		{
			WhereList whereList2 = new WhereList();
			whereList2.add("hit_index", WhereOp.EQUALS, ""+indexToUpdate);
			HashMap<String,String> setMap = new HashMap<String,String>();
			setMap.put("hit_index",""+(indexToUpdate-1));
			Database.update(DBName.USERDATA, tableName, setMap, whereList2);
			
			// update sites pointing to the hit whose number changed
			for (Site site : ModelRasterSiteList.get().getSiteListCopy())
			{
				if (site.hitNumber==indexToUpdate)
				{
					DBRasterSiteList.get().updateHitNumber(site, indexToUpdate-1);
				}
			}
		}
		
		if (update) { sendUpdate(); DBRasterSiteList.get().sendUpdate(); }
	}
	
	public void clearHits(boolean sendUpdate)
	{
		verifyTableExists();
		
		Database.delete(DBName.USERDATA, tableName, null);
		
		if (sendUpdate) sendUpdate();
	}
	
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	protected boolean connected = false;
	protected Epics.EpicsValueMonitor dbUpdatedMonitor = null;
	protected static DBRasterHits inst = null;
	
	public static DBRasterHits get() { if (null==inst) inst=new DBRasterHits(); return inst; }
	
	public DBRasterHits()
	{
		inst = this;
		tableName = "RasterHits";
		
		Epics.addMonitor(PV_R_RUNS.hitDBUpdated, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				if (-2 == Epics.dbrToInt(value))
				{
					reload();
				}
			}
		});
	}
	
	public void verifyConnected()
	{
		if (connected) return;
		
		reload();
		
		dbUpdatedMonitor = new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				int intVal = Epics.dbrToInt(value);
				if (-1 != intVal) handlePVUpdate(intVal);
			}
		};
		Epics.addMonitor(PV_R_RUNS.hitDBUpdated, dbUpdatedMonitor);
		
		connected = true;
	}
	
	protected ModelRasterHits.Hit sqlRowToHit(HashMap<String,String> row)
	{
		try
		{
			// convert string list "1,2,3" to list of integers
			ArrayList<Integer> cellIndexList = new ArrayList<>();
			StringTokenizer st = new StringTokenizer(row.get(COLS.CELL_INDEX_LIST.dbString()),",");
			while (st.hasMoreTokens())
			{
				int cellIndex = Integer.parseInt(st.nextToken());
				cellIndexList.add(cellIndex);
			}

			// create the hit
			ModelRasterHits.Hit hit = new ModelRasterHits.Hit
			(
				Integer.parseInt(row.get(COLS.HIT_INDEX.dbString())),
				Integer.parseInt(row.get(COLS.INTER_RUN.dbString())),
				Integer.parseInt(row.get(COLS.AUTO_RUN.dbString())),
				cellIndexList,
				ModelRasterState.get().getDiffDisplayParameter()
			);
			
			return hit;
		}
		catch (NumberFormatException nfe)
		{
			nfe.printStackTrace();
			return null;
		}
	}
	
	public int getHitCount()
	{
		Database.Result result = Database.exec
		(
			DBName.USERDATA,
			"select * from "+tableName
		);
		return result.numRows();
	}
	
	public void reload()
	{
		verifyTableExists();
		
		synchronized(ModelRasterHits.get())
		{
			final Database.Result result = Database.exec
			(
				DBName.USERDATA,
				"select * from "+tableName+" order by hit_index"
			);
			
			ModelRasterHits.get().clearRows(false);
			for (HashMap<String,String> row : result.getRows())
			{
				Hit hit = sqlRowToHit(row);
				if (null!=hit) ModelRasterHits.get().addHit(hit,false);
			}
		}
		ModelRasterHits.get().ll.notifyListeners();
	}
	
	protected void handlePVUpdate(int pvValue)
	{
		verifyTableExists();
		if (-2 == pvValue) reload();
	}
	public void sendUpdate()
	{
		Epics.putInteger(PV_R_RUNS.hitDBUpdated, -1);
		Epics.putInteger(PV_R_RUNS.hitDBUpdated, -2);
	}

	public static class RunIndices
	{
		public int interRunIndex, autoRunIndex;
		public RunIndices(ModelRasterRun run)
		{
			interRunIndex = run.getRunList().getListType()==RasterListType.INTER?run.getRunIndex():-1;
			autoRunIndex = run.getRunList().getListType()==RasterListType.AUTO?run.getRunIndex():-1;
		}
	}
	
	public void handleDeletedRun(ModelRasterRun run)
	{
		RunIndices ri = new RunIndices(run);
		
		// first delete hits for the run being deleted
		deleteHitsForRun(run, false);

		// then update hits for subsequent runs
		for (int runIndex=run.getRunIndex()+1; runIndex<=run.getRunList().getRunCount(); runIndex++)
		{
			HashMap<String,String> setMap = new HashMap<String,String>();
			WhereList whereList = new WhereList();
			if (ri.interRunIndex>=0)
			{
				setMap.put("inter_run",""+(runIndex-1));
				whereList.add("inter_run",WhereOp.EQUALS,""+runIndex);
			}
			else
			{
				setMap.put("auto_run",""+(runIndex-1));
				whereList.add("auto_run",WhereOp.EQUALS,""+runIndex);
			}
			Database.update(DBName.USERDATA, tableName, setMap, whereList);
		}
		
		// TODO: remove timer!
		new Thread(new Runnable(){public void run()
		{
			Util.sleep(500);
			DBRasterHits.get().sendUpdate();
		}}).start();
	}

	public void deleteHitsForRun(ModelRasterRun run, boolean update)
	{
		RunIndices ri = new RunIndices(run);
		WhereList whereList1 = new WhereList();
		whereList1.add("inter_run",WhereOp.EQUALS,""+ri.interRunIndex);
		whereList1.add("auto_run",WhereOp.EQUALS,""+ri.autoRunIndex);
		Database.delete(DBName.USERDATA, tableName, whereList1);
		
		if (update) sendUpdate();
	}

	public void deleteHitsForSite(Site site, boolean update)
	{
		if (site.hitNumber>=0)
		{
			deleteHit(site.hitNumber, site.interRunIndex, -1, update);
		}
	}

	public void deleteAutoHits(boolean update)
	{
		WhereList whereList1 = new WhereList();
		whereList1.add("auto_run",WhereOp.GREATER_THAN,"0");
		Database.delete(DBName.USERDATA, tableName, whereList1);
		
		if (update) sendUpdate();
	}
}
