package gov.anl.gmca.raster;

import gov.anl.gmca.collect.DevAdxvFile;
import gov.anl.gmca.collect.DevFPE;
import gov.anl.gmca.collect.ModelCollectRun.Point3D;
import gov.anl.gmca.common.AppState;
import gov.anl.gmca.common.AppState.Operation;
import gov.anl.gmca.hutch.DetectorConfig;
import gov.anl.gmca.hutch.DevPmacMotor;
import gov.anl.gmca.hutch.DevXYZ;
import gov.anl.gmca.raster.ModelRasterSiteList.Site;
import gov.anl.gmca.raster.ModelRasterState.RasterListType;
import gov.anl.gmca.raster.calc.FramePartitioner.FramePartition;
import gov.anl.gmca.spotfinder.CmdSpotFinder;
import gov.anl.gmca.spotfinder.CmdSpotFinderRemote;
import gov.anl.gmca.spotfinder.DevSFType;
import gov.anl.gmca.spotfinder.DevSFType.SFConfig;
import gov.anl.gmca.spotfinder.DevSFType.SFRunType;
import gov.anl.gmca.spotfinder.SFData.SFType;
import gov.anl.gmca.spotfinder.SpotFinderQueue;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.users.ModelUserTable.Client;
import gov.anl.gmca.util.Point4D;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

public class ModelRasterCell
{
	// the grid this cell is part of
	ModelRasterGrid grid = null;
	public ModelRasterGrid grid() { return grid; }
	
	Point4D gridCenter;
	
	protected int version = -1;
	public void setVersion(int version) { this.version = version; grid.llCellChanged.notifyListeners(this); }
	public int getVersion() { return version; }
	
	// coordinates
	public int index=-1;
	protected int row=-1;
	protected int col=-1;
	protected double[] xyz;

	public int		getIndex()					{ return index;							} // 0-based
	public int		getRow()					{ return row;   						} // 0-based
	public int		getCol()					{ return col; 							} // 0-based
	public double[] getRelativeXYZ()			{ return xyz;							}
	public double   getOmega()      			{ return grid.getCenter().omega;		}

	//////////////////////////////////////////// methods ////////////////////////////////////////////
	
	public ModelRasterCell(ModelRasterCell obj, boolean copyData)
	{
		if (null != obj)
		{
			this.grid		=obj.grid;
			this.index		=obj.index;
			this.row		=obj.row;
			this.col		=obj.col;
			this.gridCenter =obj.gridCenter;
			this.xyz		=new double[]{obj.xyz[0],obj.xyz[1],obj.xyz[2]};
		}
	}
	
	public ModelRasterCell(ModelRasterGrid _grid, int _number, int _row, int _col, double[] _xyz, Point4D _gridCenter)
	{
		grid		= _grid;
		index		= _number;
		row			= _row;
		col			= _col;
		xyz			= _xyz;
		gridCenter  = _gridCenter;
	}

	public static String generateImagePrefix(boolean isAutoRun, String fileRoot, int runLabel) {
		String autoPrefix = isAutoRun?"A":"";
		String prefix = fileRoot+"_"+autoPrefix+runLabel;
		return prefix;
	}

	public static String generateImageFilename(boolean isAutoRun, int cellIndex, String fileRoot, int runLabel, int rowNumber, int columnIndex)
	{
		String prefix = generateImagePrefix(isAutoRun, fileRoot, runLabel);
		if (DetectorConfig.isEiger())
		{
			// Eiger example for the cell in the third row, fifth column, run 0
			// raster_0_3_master.h5:4
			// 
			// TODO: currently this only works for the first cell in a row
			//
			return String.format("%s_%d_master.h5:1",prefix,cellIndex+1);
		}
		else if (DetectorConfig.isPilatus())
		{
			// Pilatus example: raster_0_00001.cbf
			return String.format("%s_%05d.cbf", prefix, cellIndex+1);
		}
		else
		{
			// MAR example: raster_0.00001
			return String.format("%s.%05d", prefix, cellIndex+1);
		}
	}
	
	public String getCBFImageName()
	{
		boolean isAutoRun = grid().run().runList.getListType()==RasterListType.AUTO;
		String autoPrefix = isAutoRun ? "A" : "";
		String fileRoot = grid.run().getFileRoot();
		int runLabel = grid.run().getRunLabel();
		int cellNumber = index+1;
		
		int digits = DetectorConfig.isPilatus() ? 5 : 6;
		String format = "%s_%s%d_%0"+digits+"d.cbf";
		return String.format(format, fileRoot, autoPrefix, runLabel, cellNumber);
	}

	public String getImagePrefix()
	{
		boolean isAutoRun = grid().run().runList.getListType()==RasterListType.AUTO;
		return generateImagePrefix(isAutoRun, grid.run().getFileRoot(), grid.run().getRunLabel());
	}

	public String getImageName()
	{
		boolean isAutoRun = grid().run().runList.getListType()==RasterListType.AUTO;
		String fileRoot = grid.run().getFileRoot();
		int runLabel = grid.run().getRunLabel();
		int rowNumber = getRow()+1;
		int columnIndex = grid.cellCountX-1-getCol();
		
		return generateImageFilename(isAutoRun, index, fileRoot, runLabel, rowNumber, columnIndex);
	}

	public double[] getMotorXYZ()
	{
		double[] motorXYZ = new double[3];
		Point4D gridCenter = grid().getCenter();
		if (null==gridCenter) return null;
		motorXYZ[0] = gridCenter.x + xyz[0];
		motorXYZ[1] = gridCenter.y + xyz[1];
		motorXYZ[2] = gridCenter.z + xyz[2];
		return motorXYZ;
	}
	
	public void center()
	{
		if (!AppState.isOperationAllowed(Operation.MOVE_SAMPLE))
		{
			Log.note("Sample motion not allowed at this time");
			return;
		}
		
		DevPmacMotor.get("gonio_omega").startMove(getOmega());
		double[] motorXYZ = getMotorXYZ();
		if (null==motorXYZ)
		{
			Log.warn("Center skipped, no grid center.");
			return;
		}
		DevXYZ.syncMove(new Point3D(motorXYZ), false);
	}
	
	public void addSiteFromCell()
	{
		// get motor position
		double[] motorXYZ = getMotorXYZ();
		if (null==motorXYZ)
		{
			Log.warn("Add site skipped, no grid center");
			return;
		}
		Point4D motorPos = new Point4D(motorXYZ[0], motorXYZ[1], motorXYZ[2], getOmega());
		
		// gather size info
		double w = grid().getCellWidth_um();
		double h = grid().getCellHeight_um();
		double l = Math.min(w,h); // TODO: correct?
		Point3D size = new Point3D(w,h,l);
		
		// first create a hit
		int hitIndex=ModelRasterHits.get().getHighestHitIndex(grid().run())+1;
		ModelRasterRun run = grid().run();
		boolean isInterRun = run.getRunList().getListType()==RasterListType.INTER;
		int interRunIndex = isInterRun ? run.getRunIndex() : -1;
		int autoRunIndex = isInterRun ? -1 : run.getRunIndex();
		DBRasterHits.get().addHit(hitIndex, interRunIndex, autoRunIndex, new Integer[]{getIndex()}, true);
		
		// now create the site containing the hit
		Site site = new Site
		(
			0,									// row, 0=auto increment
			grid().run().getRunIndex(),			// inter run
			index,								// cell number
			-1, -1,								// reference fine run, ortho run - for auto raster
			hitIndex,
			motorPos,
			size
		);
		DBRasterSiteList.get().addSite(site, true);
	}

	public void loadImage()
	{
		if (!Client.isMaster()) { Log.note("Skipping raster image load command since client is not master"); return; }
		
		ModelRasterRun currentRun = ModelRasterState.get().getCurrentRunModel();
		String imageName = getImageName();

		if (DetectorConfig.isEiger())
		{
			String cbfDir = currentRun.getFullDirectory();
			String cbfName = getCBFImageName();
			DevAdxvFile.update(cbfDir+"/"+cbfName);
			DevFPE.get().setImageViewerFilePath(cbfDir,cbfName);
		}
		else
		{
			DevFPE.get().setImageViewerFilePath(currentRun.getFullDirectory(),imageName);
		}
	}

	public int getPartitionIndex()
	{
		for (FramePartition partition : grid().getGridCalc().getPartitionList())
		{
			int firstIndex = partition.firstFrameIndex;
			int lastIndex = firstIndex + partition.frameCount - 1;
			if (index>=firstIndex && index<=lastIndex) return partition.partitionIndex;
		}
		throw new IllegalStateException("Cell "+index+" not found in partition list");
	}
}
