package gov.anl.gmca.raster;

import gov.anl.gmca.common.Redis;
import gov.anl.gmca.raster.ModelRasterState.RasterListType;
import gov.anl.gmca.raster.calc.GridCalculator;
import gov.anl.gmca.sample.DBCamera;
import gov.anl.gmca.sample.WdgCamera;
import gov.anl.gmca.sample.WdgCamera.CameraSize;
import gov.anl.gmca.sample.WdgCamera.CameraType;
import gov.anl.gmca.spotfinder.DBImageStats;
import gov.anl.gmca.spotfinder.DevSFType;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.ListenerList;
import gov.anl.gmca.util.Point2D;
import gov.anl.gmca.util.Point4D;
import org.eclipse.swt.graphics.Point;

import java.nio.FloatBuffer;
import java.util.ArrayList;
import java.util.HashMap;

public class ModelRasterHits
{
	public static class Hit implements Comparable<Hit>
	{
		private int interRunIndex=-1,autoRunIndex=-1;
		public ArrayList<Integer> cellList = new ArrayList<Integer>();

		public int hitNumber;
		public int getNumber() { return hitNumber; }
		
		public Point getScreenXYFromGridCenter(Point4D gridCenter)
		{
			// get difference between current position and hit position
			double[] xyzo = new double[]
			{
				motorXYZO.x - gridCenter.x,
				motorXYZO.y - gridCenter.y,
				motorXYZO.z - gridCenter.z,
				motorXYZO.omega
	        };
	
			// convert to screen coordinates
			double[] screenXY = WdgCamera.motorMMtoAbsScreen(CameraType.HIGHRES, true, xyzo, 1.0, null);
			
			// move to center of screen
			Point cameraDim = DBCamera.getCameraDimensions(CameraType.HIGHRES, CameraSize.LARGE);
			screenXY[0] += cameraDim.x/2;
			screenXY[1] += cameraDim.y/2;
	
			return new Point((int)Math.round(screenXY[0]),(int)Math.round(screenXY[1]));
		}
		
		private Point4D motorXYZO;
		public Point4D getMotorXYZO() { return motorXYZO; }

		public int spotMax=0;
		private DevSFType.SPOTFINDER_OUTPUT output;
		private int cellCountX,cellCountY,selectedCellCount;
		public int getCellCountX() { return cellCountX; }
		public int getCellCountY() { return cellCountY; }
		public int getSelectedCellCount() { return selectedCellCount; }
		
		public Hit(int hitNumber, int interRunIndex, int autoRunIndex, ArrayList<Integer> cellIndexList, DevSFType.SPOTFINDER_OUTPUT output)
		{
			this.interRunIndex=interRunIndex;
			this.autoRunIndex=autoRunIndex;
			this.output=output;
			GridCalculator calc = getRun().grid().getGridCalc();
			ArrayList<ModelRasterCell> cellList = new ArrayList<>();
			for (Integer cellIndex : cellIndexList)
			{
				cellList.add(calc.getCellCopy(cellIndex));
			}
			init(hitNumber, interRunIndex, autoRunIndex, cellList);
		}
		
		private ModelRasterRun getRun()
		{
			ModelRasterRun run = null;
			if (interRunIndex>=0) run = (ModelRasterRun)ModelRasterRunList.get(RasterListType.INTER).getRun(interRunIndex);
			if (autoRunIndex>=0) run = (ModelRasterRun)ModelRasterRunList.get(RasterListType.AUTO).getRun(autoRunIndex);
			return run;
		}
		
		private void init(int hitNumber, int interRunIndex, int autoRunIndex, ArrayList<ModelRasterCell> cellList)
		{
			this.hitNumber=hitNumber;

			ArrayList<Integer> cellIndices = new ArrayList<Integer>();
			for (ModelRasterCell cell : cellList) cellIndices.add(cell.getIndex());
			this.cellList=cellIndices;
			
			int minCol = cellList.get(0).getCol(), maxCol = minCol;
			int minRow = cellList.get(0).getRow(), maxRow = minRow;
			
			double sumMotorX=0,sumMotorY=0,sumMotorZ=0;
			int totalSpotTotal=0;
			ModelRasterRun rrun = cellList.get(0).grid().run();

			// read spot finder output for the selected output type
			DevSFType.SPOTFINDER_OUTPUT output = ModelRasterState.get().getDiffDisplayParameter();
			FloatBuffer data;
			if (output == null) {
				data = null;
			} else {
				String outputStr = rrun.getRunType() == ModelRasterRun.RunType.DIFFRACTION ? output.toString() : "sonicc";
				data = DBImageStats.load(Redis.image_stats(rrun.getDirectorySuffix(), rrun.getFileRoot(), rrun.getRunLabel(), outputStr));
			}

			for (ModelRasterCell cell : cellList)
			{
				// look up spot finder data for this cell
				int _spotTotal = 1;
				if (null != data) {
					_spotTotal = (int)data.get(cell.index);
					if (spotMax<_spotTotal) spotMax=_spotTotal;
					if (_spotTotal<1) _spotTotal=1;
				}
				totalSpotTotal += _spotTotal;

				double[] motorXYZ = cell.getMotorXYZ();
				if (null==motorXYZ)
				{
					Log.warn("No grid center");
					break;
				}
				sumMotorX += motorXYZ[0] * _spotTotal;
				sumMotorY += motorXYZ[1] * _spotTotal;
				sumMotorZ += motorXYZ[2] * _spotTotal;
				
				if (cell.getRow()<minRow) minRow=cell.getRow();
				if (cell.getRow()>maxRow) maxRow=cell.getRow();
				if (cell.getCol()<minCol) minCol=cell.getCol();
				if (cell.getCol()>maxCol) maxCol=cell.getCol();
			}
			
			motorXYZO = new Point4D
			(
				sumMotorX/totalSpotTotal,
				sumMotorY/totalSpotTotal,
				sumMotorZ/totalSpotTotal,
				cellList.get(0).getOmega()
			);
			
			cellCountX = maxCol - minCol + 1;
			cellCountY = maxRow - minRow + 1;
			selectedCellCount = cellList.size();
		}

		@Override public int compareTo(Hit h)
		{
			return Integer.compare(spotMax, h.spotMax);
		}
	}
	
	public ListenerList<ModelRasterHits> ll;
	public static ModelRasterHits get() { if (null==inst) inst=new ModelRasterHits(); return inst; }
	private static ModelRasterHits inst = null;
	
	public ModelRasterHits()
	{
		inst = this;
		ll = new ListenerList<ModelRasterHits>(this);
	}
	
	private ArrayList<Hit> hits = new ArrayList<Hit>();
	public ArrayList<Hit> getCopyOfHits()
	{
		synchronized(hits) { return new ArrayList<Hit>(hits); }
	}
	// only return hits for the given run
	public ArrayList<Hit> getCopyOfHits(ModelRasterRun run)
	{
		synchronized(hits)
		{
			ArrayList<Hit> ret = new ArrayList<Hit>();
			for (Hit hit : hits)
			{
				if (run.getRunList().getListType()==RasterListType.INTER &&	run.getRunIndex()==hit.interRunIndex)
					ret.add(hit);
				if (run.getRunList().getListType()==RasterListType.AUTO && run.getRunIndex()==hit.autoRunIndex)
					ret.add(hit);
			}
			return ret;
		}
	}
	public void clearRows(boolean b)
	{
		hits = new ArrayList<Hit>();
		hitMap.clear();
	}
	public void addHit(Hit hit, boolean notifyListeners)
	{
		hits.add(hit);
		for (int cellIndex : hit.cellList)
		{
			HitSource src = new HitSource(hit.interRunIndex, hit.autoRunIndex, cellIndex);
			hitMap.put(src,hit.getNumber());
		}
		if (notifyListeners) ll.notifyListeners();
	}
	
	public static class HitSource
	{
		int interRunIndex=-1, autoRunIndex=-1, cellIndex=-1;
		public HitSource(int interRunIndex, int autoRunIndex, int cellIndex)
		{
			this.interRunIndex=interRunIndex; this.autoRunIndex=autoRunIndex; this.cellIndex=cellIndex;
		}
		public HitSource(ModelRasterCell cell)
		{
			setRunIndices(cell.grid().run());
			this.cellIndex = cell.getIndex();
		}
		public HitSource(ModelRasterRun run, int cellIndex)
		{
			setRunIndices(run);
			this.cellIndex=cellIndex;
		}
		private void setRunIndices(ModelRasterRun run)
		{
			if (run.getRunList().getListType()==RasterListType.INTER)
			{
				this.interRunIndex=run.getRunIndex();
			}
			else
			{
				this.autoRunIndex=run.getRunIndex();
			}
		}
		@Override public boolean equals(Object _rhs)
		{
			if (!(_rhs instanceof HitSource)) return false;
			
			HitSource rhs = (HitSource)_rhs;
			if (rhs.interRunIndex!=this.interRunIndex) return false;
			if (rhs.autoRunIndex!=this.autoRunIndex) return false;
			return rhs.cellIndex==this.cellIndex;
		}
		@Override public int hashCode()
		{
			return (interRunIndex*100000)+(autoRunIndex*1000)+cellIndex;
		}
	}
	
	private HashMap<HitSource,Integer> hitMap = new HashMap<HitSource,Integer>();
	public int getHitNumber(HitSource src)
	{
		synchronized(hits)
		{
			if (hitMap.containsKey(src))
			{
				return hitMap.get(src);
			}
			return -1;
		}
	}
	
	public int getHighestHitIndex(ModelRasterRun run)
	{
		synchronized(hits)
		{
			int ret=0;
			for (Hit hit : hits)
			{
				if (hit.interRunIndex==run.getInterRunIndex() && hit.autoRunIndex==run.getAutoRunIndex())
				{
					ret = Math.max(ret, hit.hitNumber);
				}
			}
			return ret;
		}
	}
}
