package gov.anl.gmca.raster.cam;

import gov.anl.gmca.collect.ModelCollectRun.Point3D;
import gov.anl.gmca.gui.GUI;
import gov.anl.gmca.raster.*;
import gov.anl.gmca.raster.ModelRasterHits.Hit;
import gov.anl.gmca.raster.ModelRasterHits.HitSource;
import gov.anl.gmca.raster.ModelRasterSiteList.Site;
import gov.anl.gmca.raster.ModelRasterState.RasterListType;
import gov.anl.gmca.raster.cam.ModelRasterVideo.VideoState;
import gov.anl.gmca.spotfinder.DevSFType;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.users.ModelUserTable.Client;
import gov.anl.gmca.users.ModelUserTable.ClientListener;
import gov.anl.gmca.util.ListenerList;
import org.eclipse.swt.graphics.Rectangle;

import java.util.ArrayList;
import java.util.concurrent.atomic.AtomicBoolean;

// VCCreateSite interacts with WdgCreateSite and ModelRasterVideo to manage the editing
// of interactive raster sites.
//
// 1) Selection mode is entered based on ModelRasterVideo sending a mode change event.
// 2) User interacts with either video or WdgCreateSite which make calls to this class
// 3) Selection mode is exited by clicking cancel or done on WdgCreateSite which
//    calls a cancel or done method in this class
// 4) The cancel or done method here changes the ModelRasterVideo mode which also calls
//    a callback here to do cleanup
//
public class VCCreateSite
{
	// are we creating a site at the moment?
	private boolean inMode = false;
	
	protected ModelRasterVideo model;
	protected ModelRasterRun run;
	DrawCreateSite draw;
	
	ArrayList<AtomicBoolean> hit;

	protected int cellCount;
	protected int currentHitIndex;
	protected int getCurrentHit() { return currentHitIndex; }
	protected int getHitCount()
	{
		if (null==run||null==run.grid()) return 0;
		ArrayList<ModelRasterHits.Hit> hits=ModelRasterHits.get().getCopyOfHits(run);
		int committedHitCount = null==hits?0:hits.size();
		return Math.max(committedHitCount, currentHitIndex);
	}

	//////////////////////////////////////////////////////////////////////////////////////
	////////////////////////// entering/exiting selection mode ///////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////

	// When entering selection mode:
	// 1) prompt to reset runs if hits are in use
	// 2) load first existing hit or set up an empty hit
	//
	public void startSelection()
	{
		// prompt to delete existing sites
		/*boolean deletionOK = ClearSampleState.deleteInterSites(ModelRasterState.get().getCurrentRunModel(), "", true, true);
		if (!deletionOK)
		{
			cancelSelection();
			return;
		}*/

		// initialize variables
		run = ModelRasterState.get().getCurrentRunModel();
		cellCount = run.grid().getSelectedCellCount();
		resetHitArray();

		int hitCount = getHitCount();
		
		// if there are existing hits, load the first one and start in edit mode
		if (hitCount>0)
		{
			Log.note("Resuming hit editing starting with #1");
			
			// set current hit and load it
			currentHitIndex = 1;
			int cellsLoaded = loadHit(0);
			
			if (0 == cellsLoaded) hitCount = 0;
		}
		
		// if there are no existing hits (or the first hit is empty), start in 'click create' mode
		if (0 == hitCount)
		{
			Log.note("Starting new hit editor - click create to begin");
			
			// set current hit
			currentHitIndex = 0;
		}
		
		model.llStateChange.notifyListeners();
	}
	
	// (re-)create array to hold the current hit selection
	private void resetHitArray()
	{
		hit = new ArrayList<AtomicBoolean>();
		for (int j=0;j<cellCount;j++)
		{
			hit.add(new AtomicBoolean(false));
		}
	}

	// On mode cancellation, simply exit and forget the changes
	//
	public void cancelSelection()
	{
		hit = null;
		model.changeToDefaultState();
	}

	// On mode completion, send the selected hits to SQL
	//
	public void handleDone()
	{
		sendHitToDevice();
		DBRasterHits.get().sendUpdate();
		hit = null;
		model.changeToDefaultState();
	}
	
	/////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////// helpers ///////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////

	// returns # of cells loaded for the hit
	private int loadHit(int hitIndexToLoad)
	{
		int sc=0;
		
		// for each cell
		for (int cellIndex=0;cellIndex<cellCount;cellIndex++)
		{
			// see which hit this cell is part of
			int hitIndex = ModelRasterHits.get().getHitNumber(new HitSource(run, cellIndex))-1;
			
			// set the boolean accordingly
			boolean selected = (hitIndex==hitIndexToLoad);
			hit.get(cellIndex).set(selected);
			if (selected) sc++;
		}
		
		System.out.println("Loaded "+sc+" cells for hit index "+hitIndexToLoad);
		return sc;
	}
	
	private boolean sendHitToDevice()
	{
		if (currentHitIndex==0) return false;
		
		// debugging info for exception seen in a log file
		if (null==run) 					{ System.out.println("run is null!"); 			}
		if (null==run.dev()) 			{ System.out.println("run.dev() is null!"); 	System.out.println("index="+run.getRunIndex()); }

		// create the cell list
		ArrayList<Integer> cellList = new ArrayList<Integer>();
		for (int cellIndex=0;cellIndex<hit.size();cellIndex++)
		{
			boolean cellSelected = hit.get(cellIndex).get();
			
			// if the cell is part of the new hit
			if (cellSelected)
			{
				cellList.add(cellIndex);
			}
		}

		if (cellList.size()>0)
		{
			// send hit to device
			boolean isInterRun = run.getRunList().getListType()==RasterListType.INTER;
			int interRunIndex = isInterRun ? run.getRunIndex() : -1;
			int autoRunIndex = isInterRun ? -1 : run.getRunIndex();
			DBRasterHits.get().addHit(currentHitIndex, interRunIndex, autoRunIndex, cellList.toArray(new Integer[0]), true);
			
			// if the site doesn't already exist...
			Site existingSite = ModelRasterSiteList.get().getSiteForHit(currentHitIndex, interRunIndex, autoRunIndex);
			if (null==existingSite)
			{
				// create the site
				DevSFType.SPOTFINDER_OUTPUT output = ModelRasterState.get().getDiffDisplayParameter();
				Hit hitObj = new Hit(currentHitIndex, interRunIndex, autoRunIndex, cellList, output);
				double sizeX = hitObj.getCellCountX()*run.grid().getCellWidth_um();
				double sizeY = hitObj.getCellCountY()*run.grid().getCellHeight_um();
				double sizeZ = Math.min(sizeX, sizeY);
				
				Log.note("Adding site from hit "+currentHitIndex+" with "+hitObj.getSelectedCellCount()+" cells");
				Site site = new Site
				(
					0,								// row, 0=auto increment
					run.getRunIndex(),
					-1,								// cell number
					-1, -1, currentHitIndex,		// reference fine run, ortho run, hit # - for auto raster
					hitObj.getMotorXYZO(),
					new Point3D(sizeX,sizeY,sizeZ)
				);
				
				DBRasterSiteList.get().addSite(site, true);
			}
			
			return true;
		}
		else { return false; }
	}
	
	// given a mouse position, find the cell under it and update its hit number
	//
	private void setCellSelection(int mouseX, int mouseY, boolean select)
	{
		if (0 == currentHitIndex) return; // don't allow selection before clicking create
		
		try
		{
			ModelRasterRun currentRun = ModelRasterState.get().getCurrentRunModel();
			if (null == currentRun) return;
			ModelRasterGrid currentGrid = currentRun.grid();
			
		    for (ModelRasterCell cell : currentGrid.getGridCalc().getCellListCopy())
		    {
		    	Rectangle cellBounds = model.getCellScreenBounds(cell);
		    	if (null==cellBounds) continue;
		    	if (cellBounds.contains(mouseX,mouseY) && hit.get(cell.index).get() != select)
		    	{
		    		hit.get(cell.index).set(select);
		    		System.out.println((select?"Selected":"Deselected")+" cell "+cell.index);
		    		model.cam.redraw();
		    		break;
		    	}
		    }
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}
	}

	/////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////// handle mouse events ///////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////

	public void onMove(int x, int y, int buttonDown)
	{
		if (buttonDown==0) return;
		setCellSelection(x,y,buttonDown==1);
	}
	
	public void onClick(int x, int y, int buttonDown)
	{
		setCellSelection(x,y,buttonDown==1);
	}
	
	/////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////// handle model events ///////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////

	private void handleModelChange(ModelRasterVideo mrv)
	{
		boolean enteringSiteMode = mrv.getState()==VideoState.CREATE_SITE;
		if (enteringSiteMode && !inMode)
		{
			Log.note("Entering site creation mode");
			//draw.saveFreezeFrame();
			startSelection();
		}
		if (!enteringSiteMode && inMode)
		{
			Log.note("Leaving site creation mode");
			//draw.disposeFreezeFrame();
			// cancel/done method is called by the button handler
		}
		inMode = enteringSiteMode;
	}
	
	private void handleMasterSlave(boolean isMaster)
	{
		if (!isMaster && inMode)
		{
			Log.note("Client became slave, exiting raster site selection.");
			handleDone();
		}
	}
	
	public int getSelectedCellCount()
	{
		int ret = 0;
		for (int i=0;i<hit.size();i++)
		{
			if (hit.get(i).get()) ret++;
		}
		return ret;
	}
	
	public void handleCreateButton()
	{
		if (0 == currentHitIndex)
		{
			// user clicked create for the first hit, no real hits are created yet
			currentHitIndex = 1;
			model.cam.redraw();
			model.llStateChange.notifyListeners();
			return;
		}
		
		if (0 == getSelectedCellCount())
		{
			Log.warn("No cells are selected, can't create a new hit");
			return;
		}
		
		// send hit to SQL
		sendHitToDevice();

		// start new hit
		currentHitIndex = getHitCount()+1;
		
		// clear current hit
		for (int j=0;j<hit.size();j++)
		{
			hit.get(j).set(false);
		}

		model.cam.redraw();
		model.llStateChange.notifyListeners();
	}

	public void deleteHit()
	{
		// decrement current hit if it was the last one
		int newHitNumber;
		if (getHitCount() == 1)
		{
			newHitNumber = -1; // signals the only hit is being deleted
		}
		else if (currentHitIndex==getHitCount())
		{
			newHitNumber = currentHitIndex-1; // last hit is being deleted, subtract 1
		}
		else
		{
			newHitNumber = currentHitIndex;
		}
		
		// delete the hit
		int interRunIndex = run.getRunList().getListType()==RasterListType.INTER?run.getRunIndex():-1;
		int autoRunIndex = run.getRunList().getListType()==RasterListType.AUTO?run.getRunIndex():-1;
		/*DBRasterHits.get().deleteHit(currentHit, interRunIndex, autoRunIndex, true);
		DBRasterHits.get().reload();*/
		
		// delete the corresponding site, which will delete the hit
		DBRasterSiteList.get().deleteSiteForHit(currentHitIndex, interRunIndex, autoRunIndex);
		
		// if there are no more hits, exit site creation mode
		if (newHitNumber == -1)
		{
			hit = null;
			model.changeToDefaultState();
			return;
		}
		
		// update the hit number
		currentHitIndex = newHitNumber;
		
		// update the highlighted cells to the new current hit
		loadHit(currentHitIndex-1);
		
		// redraw new hits
		model.cam.redraw();
		model.llStateChange.notifyListeners();
	}
	
	public void goToNextHit()
	{
		if (currentHitIndex<getHitCount())
		{
			sendHitToDevice();

			currentHitIndex++;
			loadHit(currentHitIndex-1);
			model.cam.redraw();
			model.llStateChange.notifyListeners();
		}
	}
	
	public void goToPrevHit()
	{
		if (currentHitIndex>0)
		{
			sendHitToDevice();

			currentHitIndex--;
			loadHit(currentHitIndex-1);
			model.cam.redraw();
			model.llStateChange.notifyListeners();
		}
	}

	/////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////// constructor ///////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////
	
	public VCCreateSite(ModelRasterVideo model)
	{
		this.model=model;
		draw = new DrawCreateSite(model);
		
		model.llStateChange.addListener(new ListenerList.Listener<ModelRasterVideo>()
		{
			@Override public void handleEvent(final ModelRasterVideo vc, Object data)
			{
				GUI.asyncExec(new Runnable(){public void run()
				{
					handleModelChange(vc);
				}});
			}
		});
		
		Client.addListener(new ClientListener()
		{
			@Override public void masterChanged(boolean isMaster)
			{
				handleMasterSlave(isMaster);
			}
		});
	}
}
