package gov.anl.gmca.raster.cam;

import gov.anl.gmca.common.ui.NBFileDialog;
import gov.anl.gmca.gui.GUI;
import gov.anl.gmca.io.Disk;
import gov.anl.gmca.raster.ModelRasterRun;
import gov.anl.gmca.raster.ModelRasterState;
import gov.anl.gmca.raster.cam.ModelRasterVideo.VideoState;
import gov.anl.gmca.sample.ImageViewer;
import gov.anl.gmca.sample.WdgCamera;
import gov.anl.gmca.spotfinder.DevSFType;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.ListenerList;
import gov.anl.gmca.util.Util;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.*;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.widgets.Composite;

import java.awt.*;

// 1) WdgRasterCamera is created during startup, possibly more than one
// 2) This widget object creates a ModelRasterVideo
// 3) ModelRasterVideo creates VC objects to handle controls for different modes
// 4) VC objects create Draw objects to handle drawing for each mode
//
// Note that all objects are linked to a single camera/model pair and are unique for each pair.

public class WdgRasterCamera extends WdgCamera implements MouseMoveListener, KeyListener, MouseWheelListener, ControlListener
{
	protected 	boolean 				clickOverride				= false;
	protected 	int 					mouseX, mouseY;
	public 		ModelRasterVideo 		model						= null;
	public 		RasterMenu 				menu						= null;
	protected 	int 					lastMouseX, lastMouseY;
	
	/////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////// overrides ///////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	
	@Override public void draw(GC gc, int targetWidth, int targetHeight)
	{
		ModelRasterRun currentRun = ModelRasterState.get().getCurrentRunModel();
		if (currentRun.getRunType()== ModelRasterRun.RunType.DIFFRACTION) {
			DevSFType.SPOTFINDER_OUTPUT output = ModelRasterState.get().getDiffDisplayParameter();
			model.displayParam = null==output?"NONE":output.toString();
			model.displayNumbering = ModelRasterState.get().getNumberingParameter();
		} else {
			model.displayParam = ModelRasterState.get().getSONICCDisplay().toString().toLowerCase();
			model.displayNumbering = ModelRasterState.NumberingParameter.NONE;
		}

		ModelRasterState state = ModelRasterState.get();
		if (null==state) return;
		ModelRasterRun run = state.getCurrentRunModel();
		if (null==run) return;
		if (run.getRunType()== ModelRasterRun.RunType.SONICC && !run.getSONICCBin()) {
			SONICCView.draw(gc, this, targetWidth, targetHeight);
			return;
		}

		switch (model.getState())
		{
		case POINT_AND_CLICK:	super.draw(gc, targetWidth, targetHeight);							break;
		case CREATE_CONTOUR:	super.draw(gc, targetWidth, targetHeight);							break;
		case CREATE_SITE:		super.draw(gc, targetWidth, targetHeight);							break;
		case VIEW_IMAGES:		super.draw(gc, targetWidth, targetHeight);
		}
	}

	public void drawSuper(GC gc, int w, int h) { super.draw(gc, w, h); }
	
	@Override public void drawOverlays(Graphics2D g2)
	{
       	model.drawStandard.drawOverlays(g2);
	}

	public void drawSimpleOverlays(Graphics2D g2) { super.drawOverlays(g2); }
	
	@Override public void mouseDown(MouseEvent me)
	{
		if (me.button==1)
		{
			System.out.println("Left click");
			if (!model.handleLeftClick(this, me)) super.mouseDown(me);
		}
		else if (me.button==3) // SWT calls right button "3"
		{
			System.out.println("Right click");
			if (!model.handleRightClick(this, me)) super.mouseDown(me);
		}
	}

	@Override public void mouseScrolled(MouseEvent me)
	{
		model.handleScroll(me.count>0 ? SWT.UP : SWT.DOWN);
	}
	
	@Override public void mouseMove(MouseEvent me)
	{
		model.handleMouseMove(me);
	}

	@Override public void keyPressed(KeyEvent ke)
	{
		model.handleKeyPressed(this, ke);
	}
	@Override public void keyReleased(KeyEvent ke) {}
	
	////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////// Constructor and Misc. functions //////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////
	
	public WdgRasterCamera(Composite parent, int style, String urlString, CameraType camera, VideoState startupState)
	{
		super(parent,style,urlString,camera, CameraSize.LARGE);
		
		model = new ModelRasterVideo(this, startupState);
		menu = new RasterMenu(model);
		menu.setEnabled(true);
		addMouseMoveListener(this);
		addKeyListener(this);
		addMouseWheelListener(this);
		addControlListener(this);
		ModelRasterState.get().llCurrentRun.addListener(new ListenerList.Listener<ModelRasterState>()
		{
			@Override
			public void handleEvent(ModelRasterState object, Object data)
			{
				model.setDisplayRunIndex(ModelRasterState.get().getCurrentRunModel().getRunIndex());
			}
		});
	}
	
	// used for taking a standard freeze frame when not in standard mode
	protected void drawStandard(GC gc, int targetWidth, int targetHeight)
	{
		super.draw(gc, targetWidth, targetHeight);
	}
	
	public void saveScreenshot(String fullImagePath, final boolean viewImage)
	{
		if (null == fullImagePath)
		{
			String imgPath = ModelRasterState.get().getCurrentRunModel().getFullDirectory();
			
			NBFileDialog fileDialog = new NBFileDialog(GUI.getShell(), SWT.SAVE);
			fileDialog.setFilterPath(imgPath);
			fullImagePath = fileDialog.openStr();
			
			if (null == fullImagePath)
			{
				fullImagePath = ModelRasterState.get().getCurrentRunModel().generateRasterScreenshotFilename();
			}
			int suffix=1;
			while (Disk.exists(fullImagePath))
			{
				if (suffix==1000)
				{
					Log.error("Couldn't find a unique filename to save the raster screenshot (tried up to "+fullImagePath+"), aborting.");
					return;
				}
				fullImagePath = imgPath+Util.sprintf("test_%05d.png",suffix++);
			}
			if (fullImagePath.indexOf('.')<0) fullImagePath += ".png";
		}
		
		final String f_filename = fullImagePath;
		GUI.asyncExec(new Runnable(){public void run()
		{
			saveScreenshot(f_filename);
			if (viewImage) ImageViewer.show(f_filename);
		}});
	}

	@Override public void controlMoved(ControlEvent e) {}
	@Override public void controlResized(ControlEvent e)
	{
	}
}
