package gov.anl.gmca.raster.sonicc;

import com.mashape.unirest.http.HttpResponse;
import com.mashape.unirest.http.Unirest;
import gov.anl.gmca.gui.Main;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.PV_SONICC;
import gov.anl.gmca.raster.ModelRasterRun;
import gov.anl.gmca.raster.calc.RasterGeometry;
import gov.anl.gmca.raster.cam.RasterMap;
import gov.anl.gmca.raster.cam.WdgRasterCamera;
import gov.anl.gmca.raster.ui.WdgSONICCTab;
import gov.anl.gmca.sample.DBCamera;
import gov.anl.gmca.sample.DevCameraSettings;
import gov.anl.gmca.sample.WdgCamera;
import gov.anl.gmca.util.Point2D;
import gov.anl.gmca.util.Point4D;
import gov.anl.gmca.util.Util;

import javax.imageio.ImageIO;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;

public class SONICCImage
{
    private static final boolean DEBUG = false;

    public String soniccModeString;
    public BufferedImage originalImg, temp1, temp2, transformedImg;
    public Point hrTopLeft = new Point();
    public Point2D cropped_size_um;

    public SONICCImage(String soniccModeString)
    {
        this.soniccModeString=soniccModeString;
    }

    public void loadFromHTTP()
    {
        try {
            String hr_url = "http://192.168.1.50:4000/Channel"+soniccModeString+"Last.jpg";
            HttpResponse resp = Unirest.get(hr_url).asBinary();
            originalImg = ImageIO.read(resp.getRawBody());
        } catch (Exception e) {
            e.printStackTrace();
            originalImg = null;
        }
    }

    public void loadFromDisk()
    {
        String imagePath = System.getProperty("application.home")+"/data/Channel"+soniccModeString+"Last.jpg";
        if (DEBUG) System.out.println("Loading "+imagePath);
        File f = new File(imagePath);
        try {
            originalImg = ImageIO.read(f);
        } catch (IOException ioe) {
            ioe.printStackTrace();
            originalImg = null;
        }
    }

    private static double sin(double a) { return Math.sin(a); }
    private static double cos(double a) { return Math.cos(a); }

    //
    // convenience function which uses current EPICS values
    //
    public void transform(Point4D sonicc_ref, boolean cropToRectangle, boolean cameraTransform)
    {
        // get SONICC configuration
        double xpixel2um = Main.SONICC_TEST_MODE ? 2.73 : Epics.getDouble(PV_SONICC.xpixel2um);
        double ypixel2um = Main.SONICC_TEST_MODE ? 2.5 : Epics.getDouble(PV_SONICC.ypixel2um);
        double xcenter_px = Main.SONICC_TEST_MODE ? 277 : Epics.getInteger(PV_SONICC.xcenter_px);
        double ycenter_px = Main.SONICC_TEST_MODE ? 299 : Epics.getInteger(PV_SONICC.ycenter_px);

        // calc on-screen offset due to current motor positions
        if (null==sonicc_ref) sonicc_ref = RasterGeometry.getCurrentMotorPositions();
        Point4D curPos = RasterGeometry.getCurrentMotorPositions();
        Point4D motor_delta = new Point4D(curPos.x-sonicc_ref.x,curPos.y-sonicc_ref.y,curPos.z-sonicc_ref.z,sonicc_ref.omega);
        double[] offsetPx = WdgRasterCamera.motorMMtoAbsScreen(WdgCamera.CameraType.HIGHRES, true, motor_delta.toArray(), 1.0, null);
        Point motorOffset = new Point((int)offsetPx[0],(int)offsetPx[1]);

        // prepare settings for passing to transform()
        double angle_r = DevSONICC.getAngle_rad();
        Point2D hr_px2mm = new Point2D(DevCameraSettings.getPx2mm_h_hi()*1000.0, DevCameraSettings.getPx2mm_v_hi()*1000.0);
        Point2D sonicc_px2um = new Point2D(xpixel2um, ypixel2um);
        Point2D hr_center_offset = new Point2D(xcenter_px, ycenter_px);
        org.eclipse.swt.graphics.Point _hr_size = DBCamera.getCameraDimensions(WdgCamera.CameraType.HIGHRES, WdgCamera.CameraSize.LARGE);
        Point hr_size = new Point(_hr_size.x,_hr_size.y);

        transform(angle_r, motorOffset, hr_size, hr_px2mm, sonicc_px2um, hr_center_offset, cropToRectangle, cameraTransform);
    }

    //
    // Definitions
    //
    // Images:
    //   originalImg is the original image from the web server
    //   targetImg is the transformed image, either cropped or expanded to fit the rotated image
    //   hrImg is the high-res camera image
    //
    // Input variables:
    //
    //   angle_r: degrees clockwise to rotate originalImg to align axes with hrImg
    //   hr_center_offset: after angle_r is applied, hr_center_offset on originalImg
    //   sonicc_umPerPx: pixel size on originalImg
    //   hr_size: size of hrImg
    //   hr_umPerPx: pixel size on hrImg
    //
    // Output variables:
    //
    //   hrTopLeft: where to draw targetImg on hrImg
    //
    public boolean transform(double angle_r, Point motorOffset, Point hr_size, Point2D hr_umPerPx, Point2D sonicc_umPerPx, Point2D hr_center_offset, boolean cropToRectangle, boolean cameraTransform)
    {
        if (null==originalImg) return false;

        if (DEBUG) Util.printf("Inputs: angle_deg=%.2f hr_umPerPx=%s sonicc_umPerPx=%s hr_center_offset=%s cropToRectangle=%d cameraTransform=%d",
                    Math.toDegrees(angle_r), hr_umPerPx.toString(2), sonicc_umPerPx.toString(2), hr_center_offset.toString(2), cropToRectangle?1:0, cameraTransform?1:0);

        int origW = originalImg.getWidth(), origH = originalImg.getHeight();
        int tgtW, tgtH, tgtXOffset, tgtYOffset, scaledTgtH, scaledTgtW;
        double scaleX=1.0, scaleY=1.0;

        if (WdgSONICCTab.showCenter) {
            Graphics2D _g2 = originalImg.createGraphics();
            _g2.setColor(Color.RED);
            _g2.fillRect((int)hr_center_offset.x-10,(int)hr_center_offset.y-10,20,20);
            _g2.dispose();
        }

        if (cropToRectangle) {
            //
            // With cropToRectangle the target image is smaller than the original image.  The entire
            // target image is filled with data, there are no transparent/unfilled sections.  This
            // means some data is lost.  For example, with a 3 degree rotation, a 512x512 image
            // is clipped to 486x486.  xOffset/yOffset are needed here because the center is shifted
            // toward the origin due to the smaller image size.
            //
            tgtW=(int)Math.round((origW*origH*cos(angle_r))/(origH+origW*sin(angle_r)*cos(angle_r)));
            tgtH=tgtW*origH/origW;
            tgtXOffset=(int)Math.round((tgtW-origW)/2.0);
            tgtYOffset=(int)Math.round((tgtH-origH)/2.0);
            cropped_size_um = new Point2D(tgtW*sonicc_umPerPx.x, tgtH*sonicc_umPerPx.y);
        } else {
            //
            // Without cropToRectangle, the target image is large enough to contain the entire
            // rotated original image.  Its center is moved by a positive xOffset/yOffset to
            // accomodate the new larger size.
            //
            tgtW = (int)Math.round(origH*sin(angle_r) + origW*cos(angle_r));
            tgtH = (int)Math.round(origW*sin(angle_r) + origH*cos(angle_r));
            tgtXOffset = (int)Math.round((tgtW-origW)/2.0);
            tgtYOffset = (int)Math.round((tgtH-origH)/2.0);
        }
        if (DEBUG) Util.printf("cropToRectangle=%d tgtW=%d tgtH=%d tgtXOffset=%d tgtYOffset=%d", cropToRectangle?1:0, tgtW, tgtH, tgtXOffset, tgtYOffset);

        // If cameraTransform is true, the image will be zoomed to match the high-res camera scale
        // and a returned double[] will contain the x/y coordinates to draw the target image
        // in position on the high-res camera image.
        //
        if (cameraTransform) {
            //
            // input: hr_center_offset is the center pixel of the high-res camera on the ROTATED sonicc image
            //
            // To convert to target sonicc image offset on the HR camera:
            // (1) Find the vector SCV from the sonicc image center to hr_center_offset in sonicc image coordinates
            // (2) SCV2 = Rotate SCV by -1 * angle_r to get back to HR camera orientation
            // (3) Find scaleX/Y to scale sonicc pixels to hr pixels
            // (4) Apply scaleX/Y to SCV2 and tgtH/W
            // (5) SCV3 = Add tgt[X|Y]Offset to finish moving the center to target sonicc image coordinates
            // (6) Let soniccSize be the size of the original sonicc image.  Subtract soniccSize/2 to get the
            //     top left of where to draw the image.

            // Step one
            Point2D origSize = new Point2D(originalImg.getWidth(), originalImg.getHeight());
            Point2D origCenter = origSize.div(2.0);
            Point2D SCV = hr_center_offset.minus(origCenter);
            if (DEBUG) Util.printf("(T1) origSize=%s origCenter=%s, SCV=%s", origSize.toString(0), origCenter.toString(0), SCV.toString(0));

            // Step two
            Point2D SCV2 = SCV.rotate(angle_r);
            if (DEBUG) Util.printf("(T2) Rotated SCV1 = SCV2 = "+SCV2.toString(0));

            // Step three
            //
            // what to multiply sonicc image size by to match scale to HR image
            // e.g.: 3um sonicc pixel, 2um HR pixel, need to multiply sonicc image size by 1.33 to match scales
            // Therefore scale = sonicc pixel size / HR pixel size
            scaleX = sonicc_umPerPx.x / hr_umPerPx.x;
            scaleY = sonicc_umPerPx.y / hr_umPerPx.y;
            if (DEBUG) Util.printf("(T3) Multiplier for sonicc image size to match HR image = scale = %.2f,%.2f", scaleX, scaleY);

            // Step four: Apply scale to target size, SCV
            scaledTgtW = (int)(tgtW*scaleX);
            scaledTgtH = (int)(tgtH*scaleY);
            Point2D SCV3 = new Point2D(SCV2.x*scaleX, SCV2.y*scaleY);
            if (DEBUG) Util.printf("(T4) Target sonicc image size (scaled and cropped/enlarged): scaledTgtW=%d scaledTgtH=%d", scaledTgtW, scaledTgtH);
            if (DEBUG) Util.printf("(T4) Scaled SCV2 = SCV3 = %s", SCV3.toString(0));

            // Step five: find target x/y coordinates on HR camera
            Point2D hr_center = new Point2D(hr_size.x/2.0, hr_size.y/2.0);
            hrTopLeft.x = (int)Math.round(hr_center.x - SCV3.x - scaledTgtW/2);
            hrTopLeft.y = (int)Math.round(hr_center.y - SCV3.y - scaledTgtH/2);
            if (DEBUG) Util.printf("(T5) hrTopLeft.x = %d - %d - %d = %d", (int)hr_center.x, (int)SCV3.x, (int)scaledTgtW/2, hrTopLeft.x);
            if (DEBUG) Util.printf("(T5) hrTopLeft.y = %d - %d - %d = %d", (int)hr_center.y, (int)SCV3.y, (int)scaledTgtH/2, hrTopLeft.y);

            // Step six: Add motor offset
            hrTopLeft.x -= motorOffset.x;
            hrTopLeft.y -= motorOffset.y;
            if (DEBUG) Util.printf("(T6) hrTopLeft.x - %d = %d", motorOffset.x, hrTopLeft.x);
            if (DEBUG) Util.printf("(T6) hrTopLeft.y - %d = %d", motorOffset.y, hrTopLeft.y);

        } else {
            scaledTgtW=tgtW;
            scaledTgtH=tgtH;
        }

        if (DEBUG) Util.printf("tgtW=%d tgtH=%d tgtXOffset=%d tgtYOffset=%d", tgtW, tgtH, tgtXOffset, tgtYOffset);

        temp1 = new BufferedImage(tgtW, tgtH, BufferedImage.TYPE_BYTE_GRAY);
        Graphics2D g2 = temp1.createGraphics();
        g2.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BICUBIC);
        g2.rotate(angle_r, origW/2, origH/2);
        g2.drawImage(originalImg, null, tgtXOffset, tgtYOffset);
        g2.dispose();

        if (DEBUG) Util.printf("scaledTgtW=%d scaledTgtH=%d scaleX=%.2f scaleY=%.2f", scaledTgtW, scaledTgtH, scaleX, scaleY);
        temp2 = new BufferedImage(scaledTgtW, scaledTgtH, BufferedImage.TYPE_BYTE_GRAY);
        g2 = temp2.createGraphics();
        g2.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BICUBIC);
        g2.scale(scaleX, scaleY);
        g2.drawImage(temp1, null, 0, 0);
        g2.dispose();

        transformedImg = temp2;

        return true;
    }

    public static int composite(int soniccVal, int hrVal, int maxSoniccVal)
    {
        int dt = WdgSONICCTab.displayIndex;
        if (dt==0) return hrVal;

        int hrR = (hrVal&0x00FF0000) >> 16;
        int hrG = (hrVal&0x0000FF00) >> 8;
        int hrB = (hrVal&0x000000FF) >> 0;

        int destR=0,destG=0,destB=0;

        if (dt==1) { // highlight 1 (alpha blend)
            double _hrR = hrR; _hrR = _hrR * 55.0 / 200.0;
            double _hrG = hrG; _hrG = _hrG * 55.0 / 200.0;
            double _hrB = hrB; _hrB = _hrB * 55.0 / 200.0;
            double _soniccVal = soniccVal; _soniccVal = _soniccVal * 200.0 / ((double)maxSoniccVal);

            destR = (int)(((double)hrR)*55.0/200.0);
            destG = (int)_hrG + (int)_soniccVal;
            destB = (int)(((double)hrB)*55.0/200.0);
        }
        else if (dt==2) { // highlight 2 (add only)
            double _soniccVal = soniccVal; _soniccVal = _soniccVal * 200.0 / ((double)maxSoniccVal);
            double _hrG = hrG + (int)_soniccVal;
            if (_hrG>255) _hrG=255;

            destR = hrR;
            destG = (int)_hrG;
            destB = hrB;
        }
        else if (dt>=3) { // heatmaps

            double alpha = dt==3 ? 200.0/255.0 : 1.0;

            // scale soniccVal to 0-200
            double _soniccVal = soniccVal;
            _soniccVal = _soniccVal * (255.0*alpha)/((double)maxSoniccVal);

            // scale hrRGB to 0-55
            double _hrR = hrR;
            _hrR = _hrR*(1.0-alpha);
            double _hrG = hrG;
            _hrG = _hrG*(1.0-alpha);
            double _hrB = hrB;
            _hrB = _hrB*(1.0-alpha);

            destR = ((int)_hrR)+(int)RasterMap.colormap[0][(int)_soniccVal];
            destG = ((int)_hrG)+(int)RasterMap.colormap[1][(int)_soniccVal];
            destB = ((int)_hrB)+(int)RasterMap.colormap[2][(int)_soniccVal];
        }

        int ret = 255<<24 | destR<<16 | destG<<8 | destB;
        return ret;
    }

    public static BufferedImage applyToHRImg(BufferedImage transformedImg, Point hrTopLeft)
    {
        try {
            boolean loadFromDisk = false; // Main.FAKE_DATA
            BufferedImage hrImg = loadFromDisk ? loadHRFromDisk() : loadHRFromHTTP();
            BufferedImage destImg = new BufferedImage(hrImg.getWidth(), hrImg.getHeight(), BufferedImage.TYPE_INT_ARGB);
            Graphics2D g2 = destImg.createGraphics();
            g2.drawImage(hrImg, null, 0, 0);
            g2.dispose();
            if (null!=transformedImg) {
                if (DEBUG) Util.printf("Drawing transformedImg at %d,%d", hrTopLeft.x, hrTopLeft.y);
                int maxSoniccVal = 0;
                for (int x = 0; x < transformedImg.getWidth(); x++) {
                    for (int y = 0; y < transformedImg.getHeight(); y++) {
                        int soniccVal = 0xFF & transformedImg.getRGB(x, y);
                        if (soniccVal > maxSoniccVal) maxSoniccVal = soniccVal;
                    }
                }
                for (int x = 0; x < transformedImg.getWidth(); x++) {
                    for (int y = 0; y < transformedImg.getHeight(); y++) {
                        int destX = x + hrTopLeft.x, destY = y + hrTopLeft.y;
                        if (destX < 0 || destX >= hrImg.getWidth() || destY < 0 || destY >= hrImg.getHeight()) continue;

                        int soniccVal = 0xFF & transformedImg.getRGB(x, y);
                        int hrVal = destImg.getRGB(destX, destY);
                        int destVal = composite(soniccVal, hrVal, maxSoniccVal);

                        destImg.setRGB(destX, destY, destVal);
                    }
                }
                if (DEBUG) System.out.println("max sonicc val " + maxSoniccVal);
            }
            return destImg;
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }
    }

    public static BufferedImage loadHRFromDisk()
    {
        String imagePath = System.getProperty("application.home")+"/data/highres_cam.jpg";
        if (DEBUG) System.out.println("Loading "+imagePath);
        File f = new File(imagePath);
        try {
            BufferedImage img = ImageIO.read(f);
            return img;
        } catch (IOException ioe) {
            ioe.printStackTrace();
            return null;
        }
    }

    public static BufferedImage loadHRFromHTTP()
    {
        try {
            String hr_url = "http://192.168.1.22/now.jpg?ds=4";
            HttpResponse resp = Unirest.get(hr_url).asBinary();
            BufferedImage hrImg = ImageIO.read(resp.getRawBody());
            return hrImg;
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }
    }}
