package gov.anl.gmca.raster.util;

import gov.anl.gmca.raster.DevRasterRun;
import gov.anl.gmca.raster.ModelRasterRun;
import gov.anl.gmca.raster.ModelRasterRun.ROIMode;
import gov.anl.gmca.raster.ModelRasterRun.RunType;
import gov.anl.gmca.raster.ModelRasterRunList;
import gov.anl.gmca.raster.ModelRasterState.RasterListType;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.ActionResult;
import gov.anl.gmca.util.ListenerList;
import gov.anl.gmca.util.Util;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;

public class ModelMacroList
{
	public static class Macro implements Comparable<Macro>
	{
		public String ID, prefix, dir;
		public int rasterMode, ROI;
		public double beamWidth, beamHeight, cellWidth, cellHeight, exposureTime, delta, ccdDistance, attenuation, energy;
		
		public ROIMode getROI() 					{ return ROIMode.values()[ROI]; 		}
		public RunType getRasterType()				{ return RunType.values()[rasterMode]; 	}
		
		public void setROI(ROIMode r)				{ ROI = r.ordinal(); 					}
		public void setRasterType(RunType rt)		{ rasterMode = rt.ordinal();			}
		
		@Override public int compareTo(Macro o)
		{
			return ID.compareTo(o.ID);
		}
	}
	
	public		ListenerList<ModelMacroList> ll;
	
				HashMap<String,Macro>	macroMap								= new HashMap<String,Macro>();
	public		ArrayList<Macro>		getAllMacros()							{ return new ArrayList<Macro>(macroMap.values()); }
	
	protected	String					currentMacro							= "";
	public 		String 					getCurrentMacro()						{ return currentMacro;	}
	public		void					setCurrentMacro(String s)				{ currentMacro = s;		}
	
	public static ModelMacroList inst;
	public static ModelMacroList get() { return inst; }
	
	public ModelMacroList()
	{
		inst = this;
		ll = new ListenerList<ModelMacroList>(this);
	}
	
	public ActionResult saveMacro(String macroName)
	{
		// check for empty string
		if (null == macroName || macroName.length() == 0)
		{
			return new ActionResult("Empty string entered for macro name.");
		}
		
		// check that the macro name is valid
		boolean nameIsValid = true;
		for (char c : macroName.toCharArray())
		{
			boolean charOK = false;
			if (c >= 'a' && c <= 'z') charOK = true;
			if (c >= 'A' && c <= 'Z') charOK = true;
			if (c >= '0' && c <= '9') charOK = true;
			if (c == '-' || c == '_') charOK = true;
			
			if (!charOK) nameIsValid = false;
		}
		if (!nameIsValid) return new ActionResult("Macro name '"+macroName+"' must contain only alpha-numeric characters, dashes or underscores.");
		
		// check that the macro name doesn't already exist
		for (Macro m : macroMap.values())
		{
			if (m.ID.equals(macroName))
			{
				return new ActionResult("A macro with the name '"+macroName+"' already exists."); 
			}
		}
		
		// create the macro
		ModelRasterRun run 			= ModelRasterRunList.get(RasterListType.INTER).getCurrentRunModel();
		Macro macro 				= new Macro();
		macro.ID					= macroName;
		macro.prefix 				= run.getFileRoot();
		macro.dir 					= run.getDirectorySuffix();
		macro.setRasterType			(run.getRunType());
		macro.setROI				(run.getROIMode());
		macro.cellWidth 			= run.grid().getCellWidth_um();
		macro.cellHeight 			= run.grid().getCellHeight_um();
		macro.beamWidth 			= run.getBeamWidth_um();
		macro.beamHeight 			= run.getBeamHeight_um();
		macro.exposureTime 			= run.getExposureTime();
		macro.delta 				= run.getDelta();
		macro.ccdDistance 			= run.getDetectorDist();
		macro.attenuation 			= run.getAttenuation();
		macro.energy 				= run.getEnergy();
		
		// send it to the device
		DevMacroList.get().addMacro(macro);
		
		Log.note("Macro '"+macroName+"' added.");
		return ActionResult.SUCCESS;
	}
	
	public void loadMacro(String macroName)
	{
		if (!macroMap.containsKey(macroName))
		{
			Log.warn("Macro "+macroName+" not found");
			return;
		}
		
		Macro macro = macroMap.get(macroName);
		DevRasterRun dev = ModelRasterRunList.get(RasterListType.INTER).getCurrentRunModel().dev();
		
		//dev.setFileroot			(macro.prefix,					false);
		//dev.setDirectory			(macro.dir,						false);
		dev.setRunType				(macro.getRasterType(),			false);
		dev.setROIMode				(macro.getROI(),				false);
		dev.setCellWidth_um			(macro.cellWidth,				false);
		dev.setCellHeight_um		(macro.cellHeight,				false);
		dev.setBeamWidth_um			(macro.beamWidth,				false);
		dev.setBeamHeight_um		(macro.beamHeight,				false);
		dev.setExposureTime			(macro.exposureTime,			false);
		dev.setDelta				(macro.delta,					false);
		dev.setDistance_mm			(macro.ccdDistance,				false);
		dev.setAttenuation_factors	(macro.attenuation,				false);
		dev.setEnergy_keV			(macro.energy,					false);
		
		dev.getRun().grid().getGridCalc().fitGridToContour(false, true); // sends update
		
		Log.note("Macro '"+macroName+"' is loaded.");
	}
	
	public void deleteMacro(final String macroName)
	{
		new Thread(new Runnable(){public void run()
		{
			boolean okToDelete = Util.popupOkCancel("Delete macro", "Do you want to delete macro '"+macroName+"'?");
			
			if (okToDelete)
			{
				DevMacroList.get().removeMacro(macroName);
				Log.note("Macro '"+macroName+"' deleted.");
			}
			else
			{
				Log.note("Macro deletion declined.");
			}
		}
		}, "Delete macro").start();
	}
	
	public String getAdjacentMacroID(String macroID)
	{
		ArrayList<Macro> macros = new ArrayList<Macro>(macroMap.values());
		Collections.sort(macros);
		int givenMacroIndex = -1;
		for (int i=0; i<macros.size(); i++)
		{
			if (macros.get(i).ID.equals(macroID)) { givenMacroIndex = i; break; }
		}
		if (givenMacroIndex >= 1) return macros.get(givenMacroIndex-1).ID;
		if (givenMacroIndex == 0 && macros.size()>1) return macros.get(1).ID;
		return "";
	}
}
