package gov.anl.gmca.sample;

import gov.anl.gmca.gui.GUI;
import gov.anl.gmca.util.Util;

import java.util.ArrayList;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

// This is a single thread that runs all the time.  It contains an executor queue
// to load multiple images at once.  It also sends redraw calls to the GUI thread.
//
public class CameraThread implements Runnable
{
	private static final 	int 				HTTP_THREAD_MAX = 2;
	private static final	int					SLEEP_MS		= 20;
	protected 				ThreadPoolExecutor 	executor;
	private static 			Thread 				thread 			= null;
	
	public synchronized static void startIfNeeded()
	{
		if (null == thread || !thread.isAlive())
		{
			thread = new Thread(new CameraThread(), "Camera");
			thread.start();
		}
	}
	
	public CameraThread()
	{
		ThreadFactory factory =  new ThreadFactory()
		{
			@Override public Thread newThread(Runnable r)
			{
				return new Thread(r, "Camera http load");
			}
		};
		executor = new ThreadPoolExecutor(HTTP_THREAD_MAX,HTTP_THREAD_MAX,0,TimeUnit.MILLISECONDS,new LinkedBlockingQueue<Runnable>(),factory);
	}

	private boolean anyCameraEnabled = false;
	private long lastLogMsg = 0L;
	private long drawCount=0, loadCount=0, totalLoadMs=0, totalDrawMs=0, maxLoadMs=0, maxDrawMs=0;

	@Override public void run()
	{
		while (!Util.atExit())
		{
			anyCameraEnabled = false;
			for (final WdgCamera camera : new ArrayList<WdgCamera>(WdgCamera.cameraList))
			{
				if (camera.isDisposed()) continue;

				try
				{
					long now = System.currentTimeMillis();
					
					// load images
					if (camera.bLoadImages.get())
					{
						anyCameraEnabled = true;
						if (now >= camera.nextImageLoadTime)
						{
							camera.currentImageLoadStartTime = now;
							camera.nextImageLoadTime = Long.MAX_VALUE;

							int active = executor.getActiveCount();
							if (active==HTTP_THREAD_MAX) {
								System.out.println("CameraStat: WARNING: submitting to full load queue");
							}
							executor.submit(new Runnable(){public void run()
							{
								try
								{
									long start = System.currentTimeMillis();
									camera.getImage();
									long elapsed = System.currentTimeMillis() - start;
									totalLoadMs += elapsed;
									maxLoadMs = Math.max(maxLoadMs, elapsed);
									loadCount++;
								}
								catch (Throwable t) { t.printStackTrace(); }
							}});
						}
					}
					
					// redraw calls
					// NOTE: The isWindows() calls are a temporary workaround until we figure out
					//       why redraw calls aren't being sent in Windows
					if (Util.isWindows() || (camera.bLoadImages.get() && camera.bNewImage.get()))
					{
						if (now >= camera.nextImageDrawTime)
						{
							camera.currentImageDrawStartTime = now;
							
						    GUI.asyncExec(new Runnable(){public void run()
						    {
								long start = System.currentTimeMillis();

						    	camera.redraw();
						    	camera.update();
								camera.bNewImage.set(false);

								long elapsed = System.currentTimeMillis() - start;
								totalDrawMs += elapsed;
								maxDrawMs = Math.max(maxDrawMs, elapsed);
								drawCount++;
							}});
							
							camera.nextImageDrawTime = Util.isWindows() ? now+100 : Long.MAX_VALUE;
						}
					}
				}
				catch (Exception e) { e.printStackTrace(); }
			}

			long now = System.currentTimeMillis();
			if (now-lastLogMsg > 5000) {
				lastLogMsg = System.currentTimeMillis();

				if (anyCameraEnabled) {
					float avgLoad = (float) totalLoadMs / (float) loadCount;
					float avgDraw = (float) totalDrawMs / (float) drawCount;
					String msg = String.format("CameraStat: load_count=%d draw_count=%d avg_load_ms=%.0f avg_draw_ms=%.0f max_load_ms=%d max_draw_ms=%d",
							loadCount, drawCount, avgLoad, avgDraw, maxLoadMs, maxDrawMs);
					//System.out.println(msg);
				}

				loadCount=0;drawCount=0;maxLoadMs=0;maxDrawMs=0;totalLoadMs=0;totalDrawMs=0;
			}

			Util.sleep(SLEEP_MS);
		}
	}
}
