package gov.anl.gmca.sample;

import gov.anl.gmca.collect.DevCollectSettings;
import gov.anl.gmca.common.DevMove.MoveResult;
import gov.anl.gmca.common.DevMove.ResultType;
import gov.anl.gmca.hutch.DevPmacMotor;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.PV_BLUICE;
import gov.anl.gmca.io.PVLookup.PV_COLLIMATOR;
import gov.anl.gmca.sample.DevInOut.InOut;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.ListenerList;
import gov.anl.gmca.util.Util;
import gov.aps.jca.dbr.DBR;

import java.util.HashMap;


/*
 * Moves multiple devices at once that are controlled by the collimator script:
 * - collimator in/out
 * - pin diode  in/out
 * - backlight  in/out
 * - beamstop   in/out
 * 
 * Example usage (this example is the equivalent of no arguments to the script):
 * 
 *	CCMove move = DevCollimatorCtrl.get().newMove();
 *	move.setDest("collimator",		InOut.IN);
 *	move.setDest("backlight",		InOut.OUT);
 *	move.setDest("beamstop",		InOut.IN);
 *	MoveResult ccResult = move.start().waitFor();
 */
public class DevCollimatorCtrl
{
	//////////////////////////////////  variables //////////////////////////////////////
	
	protected final static	int								MSG_FREQ_MS				= 2000;
	protected final static 	String 							collimatorCtrlScript 	= 	"collimatorCtrl.pl";
	protected final static 	String[] 						stringDevices 			= { "ringlight_value", "backlight_value" };
    protected final static	String[]						idleCommandStrings		= { "", " ", "finished", "failed" };
	public static enum 		CollimatorSize
	{
		NONE(-1),
		S100(100),
		S20(20),
		S10(10),
		S5(5);
		public int size_um;
		CollimatorSize(int _size_um) { size_um=_size_um; }
	};
	protected static 		CCMove							ccMove 					= null;
	protected static 		DevCollimatorCtrl 				inst 					= null;
	public					ListenerList<DevCollimatorCtrl> ll;
	
	private static enum COLLIMATOR_RESULT 
	{
		SUCCESS, FAILURE;
	
		public static COLLIMATOR_RESULT fromCommandString(String cmdStr)
		{
			if (cmdStr.equals("finished")) return SUCCESS;
			if (cmdStr.equals("failed")) return FAILURE;
			return null;
		}
	}
	
	///////////////////////////////////  methods ///////////////////////////////////////
	
	public static 	DevCollimatorCtrl 			get() 						{ return inst; 								}
	protected 		boolean 					moveChecks(CCMove moveDest) { return true; 								}
	public 			CCMove 						newMove() 					{ return new CCMove(); 						}
	public 			void 						abort() 					{ abortRequested = true; }
	public			String						getCommandString()			{ return Epics.getString(PV_BLUICE.sampleEnv); }
	public 			boolean 					inLocalMove()				{ return (null != ccMove && null != ccMove.moveThread && ccMove.moveThread.isAlive()); }
	public 			boolean 					isRunning()					{ if (inLocalMove()) return true; return inRemoteMove(); }
	private 		volatile COLLIMATOR_RESULT	serverResult				= null;
	private			volatile boolean			abortRequested				= false;
	
	public DevCollimatorCtrl()
	{
		inst = this;
		ll = new ListenerList<DevCollimatorCtrl>(this);
		Epics.addMonitor(PV_BLUICE.sampleEnv, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				COLLIMATOR_RESULT tempResult = COLLIMATOR_RESULT.fromCommandString(Epics.dbrToString(value));

				// only update the result variable if it matches a known string
				// indicating the action has completed
				if (null != tempResult)
				{
					serverResult = tempResult;
				}
			}
		});
	}
	
	public boolean inRemoteMove()
	{
		String cmdString = getCommandString();
		for (String idleString : idleCommandStrings)
		{
			if (cmdString.equals(idleString)) return false;
		}
		return true;
	}
	
	public static void setCollimatorSize(CollimatorSize reqSize)
	{
		// The collimatorCtrl script moves to the set position whenever it's moved to "in"
		// If the collimator is already in, setting a new size and moving it to "in" will move
		// it to the new size
		Epics.putInteger(PV_COLLIMATOR.ActivePreset, reqSize.ordinal());
	}
	public static CollimatorSize getCollimatorSize()
	{
		int index = Epics.getInteger(PV_COLLIMATOR.ActivePreset);
		//if (DBConfig.get("app_collimatorSet").equals("TRIPLE") && index>=2) index++;
		return CollimatorSize.values()[index];
	}

	////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////// Move class /////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////
	
	public class CCMove implements Runnable
	{
		protected HashMap<String, DevInOut.InOut> inoutMap = new HashMap<String, DevInOut.InOut>();
		protected HashMap<String, String> stringMap = new HashMap<String, String>();
		protected Thread moveThread = null;
		protected String args = null;
		protected MoveResult result = new MoveResult(ResultType.NOT_STARTED);
		
		public MoveResult getResult() { return result; }
		
		public void setArguments(String args)
		{
			this.args = args;
		}
		
		public void setDest(String devName, String dest)
		{
			if (null == devName || 0 == devName.length())
			{
				throw new IllegalArgumentException("\""+devName+"\" is not a valid device name");
			}
			
			boolean devOK = false;
			for (String stringDev : stringDevices)
			{
				if (stringDev.equals(devName)) devOK = true;
			}
			if (!devOK)
			{
				throw new IllegalArgumentException(devName+" is not a valid string device name");
			}
			
			stringMap.put(devName, dest);
		}
		
		public void setDest(String devName, DevInOut.InOut dest)
		{
			if (inoutMap.containsKey(devName))
			{
				System.out.println("WARNING: Motor "+devName+"'s destination is being set more than once");
			}
			
			boolean collimatorEnable = DevCollectSettings.getEnableSampleEnvMove();
			if (!collimatorEnable && devName.equals("collimator"))
			{
				Log.warn("Collimator is disabled, requested destination of "+dest+" is ignored.");
				return;
			}
			
			inoutMap.put(devName, dest);
		}
		
		public CCMove start()
		{
			if (null != ccMove && null != ccMove.moveThread && ccMove.moveThread.isAlive())
			{
				Log.error("CollimatorCtrl is already running, call aborted.");
				result = new MoveResult(ResultType.ERROR, "CollimatorCtrl already running");
				return this;
			}
			
			ccMove = this;
			
			moveThread = new Thread(this, "collimator move");
			moveThread.start();
			ll.notifyListeners();
			
			// TODO: temporary
			new Thread(new Runnable(){public void run()
			{
				while (isRunning()) Util.sleep(10);
				ll.notifyListeners();
			}}, "collimator move monitor").start();
			
			return this;
		}
		
		@Override
		public void run()
		{
			if (!DevCollectSettings.getEnableSampleEnvMove())
			{
				Log.note("Sample environment moves disabled, skipping move.");
				result = new MoveResult(ResultType.SUCCESS);
				moveThread = null;
				return;
			}
			
			if (!moveChecks(this))
			{
				result = new MoveResult(ResultType.INVALID);
				moveThread = null;
				return;
			}
			
			moveCmd(this, result);
			moveThread = null;
		}
		
		public DevPmacMotor.MoveResult waitFor()
		{
		    try {
			while (null != moveThread && moveThread.isAlive())
			{
				Util.sleep(10);
			}
		    } catch (NullPointerException npe) {}

			return result;
		}
	}
	
	/////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////// Actual move method //////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////
	
	public void moveCmd(CCMove dest, MoveResult moveResult)
	{
		String deviceCmd = "";
		abortRequested = false;

		HashMap<String, DevInOut.InOut> map = dest.inoutMap;

		for (String devName : map.keySet())
		{
			System.out.println("moveCmd: "+devName+"="+map.get(devName));
		}
		
		InOut collimatorDir = null;
		if (map.containsKey("collimator")) collimatorDir = map.get("collimator");

		//////////////////////////// start argument construction //////////////////////////////

		if (null != dest.args)
		{
			deviceCmd += " "+dest.args;
		}
		else
		{
			// collimator argument
			if (null == collimatorDir)						deviceCmd += " IN noCollimator";
			else if (InOut.IN == collimatorDir)				deviceCmd += " IN";
			else											deviceCmd += " OUT";
			
			// backlight argument
			if (!map.containsKey("backlight"))			deviceCmd += " noBacklight";
			else if (map.get("backlight")==InOut.OUT)	deviceCmd += " backlightOUT";
			else if (null == collimatorDir || InOut.IN == collimatorDir)
			{
				moveResult.type = ResultType.ERROR;
	            moveResult.msg = "Collimator in/none and backlight in can't be requested at the same time.";
	            return;
			}
			else {} // add nothing - backlight in is default for collimator out

			// beamstop argument
			if (!map.containsKey("beamstop"))			deviceCmd += " noBeamstop";
			else if (map.get("beamstop")==InOut.OUT)	deviceCmd += " beamstopOUT";
			else if (InOut.OUT == collimatorDir)
			{
				moveResult.type = ResultType.ERROR;
	            moveResult.msg = "Collimator out and beamstop in can't be requested at the same time.";
	            return;
			}
			else {} // add nothing - beamstop in is default for collimator in
			
			// pin diode argument
			if (!map.containsKey("pindiode"))			deviceCmd += "";
			else if (map.get("pindiode")==InOut.OUT)	deviceCmd += " pindiodeOUT";
			else										deviceCmd += " pindiodeIN";
			
			// string devices
			for (String stringDev : dest.stringMap.keySet())
			{
			    // TODO: temporary fix for arguments not having the _value suffix
			    deviceCmd += " "+stringDev.substring(0,stringDev.length()-6)+"="+dest.stringMap.get(stringDev);
			}
			deviceCmd += " bluice";
		}
		
		////////////////////////////// end argument construction //////////////////////////////
		
		// wait for device to be ready
		long lastMsg = System.currentTimeMillis()-MSG_FREQ_MS;
		while (inRemoteMove())
		{
			if (abortRequested)
			{
				Log.error("Sample environment move aborted, returning failure.");
				moveResult.type = ResultType.ABORTED;
				moveResult.msg = "Abort requested";
				return;
			}
			
			long now = System.currentTimeMillis();
			long sinceLastMsg = now-lastMsg;
			if (sinceLastMsg>MSG_FREQ_MS)
			{
				Log.note("Waiting for sample environment remove move ("+getCommandString()+")");
				lastMsg = now;
			}
			Util.sleep(100);
		}
		
		// clear the result
		serverResult = null;
		
		// write command (note someone could have written a command before this command gets
		// accepted - in this case, our command is ignored!)
		Epics.put(PV_BLUICE.sampleEnv, deviceCmd);
		
		// wait for completion
		lastMsg = System.currentTimeMillis()-MSG_FREQ_MS;
		while (null == serverResult)
		{
			if (abortRequested)
			{
				Log.error("Sample environment move aborted, returning failure.");
				moveResult.type = ResultType.ABORTED;
				moveResult.msg = "Abort requested";
				return;
			}
			
			long now = System.currentTimeMillis();
			long sinceLastMsg = now-lastMsg;
			if (sinceLastMsg>MSG_FREQ_MS)
			{
				Log.note("Waiting for sample environment move result ("+getCommandString()+")");
				lastMsg = now;
			}
			Util.sleep(100);
		}
		
		System.out.println("Returning sample environment move result "+serverResult);

		moveResult.type =  (serverResult == COLLIMATOR_RESULT.SUCCESS) ? ResultType.SUCCESS : ResultType.ERROR;
	}
}
