package gov.anl.gmca.sample;

import gov.anl.gmca.collect.DevCollectSettings;
import gov.anl.gmca.common.DBConfig;
import gov.anl.gmca.common.DevMove.MoveCmd;
import gov.anl.gmca.common.DevMoveFloat;
import gov.anl.gmca.common.DevMoveFloat.Listener;
import gov.anl.gmca.common.EpicsDevice.State;
import gov.anl.gmca.hutch.DevPmacMotor;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.PV_BLUICE;
import gov.anl.gmca.sample.DevCollimatorCtrl.CCMove;
import gov.anl.gmca.sample.DevCollimatorCtrl.CollimatorSize;
import gov.anl.gmca.sample.DevInOut.InOut;
import gov.anl.gmca.util.ActionResult;
import gov.anl.gmca.util.ListenerList;
import gov.anl.gmca.util.Point2D;
import gov.anl.gmca.util.Util;

// Replaced by DBBeamSize for fixed choices that vary with energy (ID-D currently)
//
public class ModelBeamsize
{
	public static class BeamsizeSetup
	{
		public BeamsizeSetup(CollimatorSize cs, double gsWidth, double gsHeight)
		{
			this.collimatorSize=cs;
			this.guardslitWidth=gsWidth;
			this.guardslitHeight=gsHeight;
		}
		
		public CollimatorSize collimatorSize;
		public double guardslitWidth, guardslitHeight;
	}

	public static double DEFAULT_BEAM_WIDTH_UM = 150.0;
	public static double DEFAULT_BEAM_HEIGHT_UM = 50.0;

	public static ModelBeamsize get() { return inst; }
	private static ModelBeamsize inst;
	
	public ListenerList<ModelBeamsize> ll;
	
	public ModelBeamsize()
	{
		inst = this;
		ll = new ListenerList<ModelBeamsize>(this);
		
		Listener gsxListener = new Listener()
		{
			@Override public void handleStateChange(State state) {}
			@Override public void handleUnitsChange(String units) {}
			@Override public void handleRqsPosChange(double newPos) {}
			@Override public void handleActPosChange(DevMoveFloat dev, double newPos)
			{
				ll.notifyListeners(getCurrentBeamsize());
			}
		};
		Listener gsyListener = new Listener()
		{
			@Override public void handleStateChange(State state) {}
			@Override public void handleUnitsChange(String units) {}
			@Override public void handleRqsPosChange(double newPos) {}
			@Override public void handleActPosChange(DevMoveFloat dev, double newPos)
			{
				ll.notifyListeners(getCurrentBeamsize());
			}
		};
		ListenerList.Listener<DevCollimatorCtrl> colliListener = new ListenerList.Listener<DevCollimatorCtrl>()
		{
			@Override public void handleEvent(DevCollimatorCtrl object, Object data)
			{
				ll.notifyListeners(getCurrentBeamsize());
			}
		};
		
		DevPmacMotor.get("beam_size_x").addListener(gsxListener);
		DevPmacMotor.get("beam_size_y").addListener(gsyListener);
		DevCollimatorCtrl.get().ll.addListener(colliListener);
	}
	
	public static Point2D getCurrentBeamsize()
	{
		switch (DevCollimatorCtrl.getCollimatorSize())
		{
		case S20:	return new Point2D(20,20);
		case S10:	return new Point2D(10,10);
		case S5:	return new Point2D(5,5);
		default:
			double beamX = DevPmacMotor.get("beam_size_x").getPosition()*1000.0;
			double beamY = DevPmacMotor.get("beam_size_y").getPosition()*1000.0;
			if (beamX > getActualBeamsize_um().x) beamX = getActualBeamsize_um().x;
			if (beamY > getActualBeamsize_um().y) beamY = getActualBeamsize_um().y;
			
			return new Point2D(beamX,beamY);
		}
	}
	
	public static ModelBeamsize.BeamsizeSetup getBeamsizeSetup(double beamWidth_um, double beamHeight_um)
	{
		if (beamWidth_um <= 20)
		{
			if      (beamWidth_um > 10)	return new ModelBeamsize.BeamsizeSetup(DevCollimatorCtrl.CollimatorSize.S20, 0.200, 0.200);
			else if (beamWidth_um > 5)	return new ModelBeamsize.BeamsizeSetup(DevCollimatorCtrl.CollimatorSize.S10, 0.200, 0.200);
			else			   			return new ModelBeamsize.BeamsizeSetup(DevCollimatorCtrl.CollimatorSize.S5,  0.200, 0.200);
		}
		else
		{
			double gs_x_mm, gs_y_mm;
			gs_x_mm = beamWidth_um / 1000.0;
			gs_y_mm = beamHeight_um / 1000.0;
			return new ModelBeamsize.BeamsizeSetup(DevCollimatorCtrl.CollimatorSize.S100, gs_x_mm, gs_y_mm);
		}
	}

	// synchronous move
	// for fixed beam sizes, method parameters are collimator size not actual beam size
	//
	// returns:
	// ret[0] = origCollimator (CollimatorSize)
	// ret[1] = origGS (Point2D)
	// ret[2] = ActionResult
	//
	public static Object[] setUpBeamsize(double beamWidth_um, double beamHeight_um, boolean disableProblemMotions)
	{
		DevCollimatorCtrl.CollimatorSize origCollimator = DevCollimatorCtrl.getCollimatorSize();
		Point2D origGSSize = null;
		ActionResult result;
	
		// set collimator size
		ModelBeamsize.BeamsizeSetup setup = getBeamsizeSetup(beamWidth_um, beamHeight_um);
		DevCollimatorCtrl.setCollimatorSize(setup.collimatorSize);
		
		// move guardslits if needed
		DevPmacMotor devGSX = DevPmacMotor.get("beam_size_x");
		DevPmacMotor devGSY = DevPmacMotor.get("beam_size_y");
		boolean doGSMove =
		(
			DevCollimatorCtrl.CollimatorSize.S100 == setup.collimatorSize ||
			DevCollimatorCtrl.CollimatorSize.NONE == setup.collimatorSize
		);
		// don't move GS for ID-D scatterguard if larger than minimum
		if (Util.beamSizeCalculated()) {
			if (devGSX.getPosition() > 0.2 - devGSX.getInPosBand()) doGSMove=false;
			if (devGSY.getPosition() > 0.2 - devGSY.getInPosBand()) doGSMove=false;
		}
	
		MoveCmd moveGSX=null, moveGSY=null;
		if (doGSMove)
		{
			origGSSize = new Point2D(devGSX.getPosition(), devGSY.getPosition());

			if (Util.beamSizeCalculated()) {
				moveGSX = devGSX.startMove(0.200);
				moveGSY = devGSY.startMove(0.200);
			} else {
				moveGSX = devGSX.startMove(setup.guardslitWidth);
				moveGSY = devGSY.startMove(setup.guardslitHeight);
			}
		}
		
		// beam size setup
		CCMove ccMove = null;
		if (DevCollectSettings.getEnableSampleEnvMove())
		{
		    ccMove = DevCollimatorCtrl.get().newMove();
		    ccMove.setDest("collimator",	InOut.IN);
		    ccMove.setDest("backlight",		InOut.OUT);
		    ccMove.setDest("beamstop",		InOut.IN);
		    ccMove.setDest("pindiode",		InOut.OUT);
			ccMove.start();
		}

	    // wait for all moves
	    if (doGSMove) moveGSX.waitFor();
	    if (doGSMove) moveGSY.waitFor();
	    if (!disableProblemMotions && null != ccMove) ccMove.waitFor();
		
	    if (null != ccMove && !ccMove.getResult().isOK())			result = new ActionResult("Collimator move failed.");
	    else if (doGSMove && !moveGSX.getResult().isOK())			result = new ActionResult("Horizontal guardslit move failed.");
	    else if (doGSMove && !moveGSY.getResult().isOK())			result = new ActionResult("Vertical guardslit move failed.");
	    else														result = ActionResult.SUCCESS;
		
		return new Object[]{origCollimator, origGSSize, result};
	}

	public static Point2D getActualBeamsize_um()
	{
		double x_1_sigma = Epics.getDouble(PV_BLUICE.beamsize_h_act1$VAL) / 2.35482;
		double y_1_sigma = Epics.getDouble(PV_BLUICE.beamsize_v_act1$VAL) / 2.35482;
		
		if (x_1_sigma <= 0.0) x_1_sigma = DEFAULT_BEAM_WIDTH_UM;
		if (y_1_sigma <= 0.0) y_1_sigma = DEFAULT_BEAM_HEIGHT_UM;
		
		Point2D actualSize = new Point2D
		(
			Math.round(x_1_sigma * Double.parseDouble(DBConfig.get("beam_size_factor"))),
			Math.round(y_1_sigma * Double.parseDouble(DBConfig.get("beam_size_factor")))
		);
		
		return actualSize;
	}
}
