package gov.anl.gmca.scan;

import gov.anl.gmca.common.Command;
import gov.anl.gmca.io.ExecTask;
import gov.anl.gmca.scan.ModelScan.TRACE_ID;
import gov.anl.gmca.scan.ui.TabScan;
import gov.anl.gmca.scan.ui.WdgGraph;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.Point2D;
import gov.anl.gmca.util.Util;

import java.io.*;
import java.util.ArrayList;
import java.util.StringTokenizer;

// BluIce writes a .dat file containing the edge scan and sends it to chooch
// 
// Chooch outputs:
// 1) an .efs file which is a numerical version of the f' and f'' plots
// 2) a .ps visual version of #1 
// 
// BluIce displays the f' and f'' plots from the .efs file
// BluIce converts the .ps file to a .pdf
// BluIce saves a .png of the edge scan
// BluIce saves a .bip or numerical version of the edge scan (and eventually the f' and f'' plots)

public class CmdChooch extends Command
{
	public static final String choochDir = "/mnt/software/px/bin/";
	public static final String choochFile = "chooch";
	protected static final String gsDir = "/usr/bin/";
	protected static final String gsFile = "gs";
	
	protected String 	filePrefix;
	protected File 		choochInput, choochOutputPS, choochOutputPDF, choochOutputEFS;
	protected ExecTask 	choochTask;
	
	protected ModelScan.ElementEdge ee;
	ArrayList<Point2D> trace;
	
	// element-edge is passed to chooch
	// ordinates are used only for the chooch comment
	public CmdChooch(ModelScan.ElementEdge ee, ArrayList<Point2D> trace)
	{
		this.ee = ee;
		this.trace = trace;
	}
	
	@Override public void run()
	{
		if (0 == trace.size()) { Log.error("Chooch aborted, graph has no points."); return; }

		writeScanForChooch();
		runChooch();
		processChoochOutput();
		//convertPStoPDF();
		sendChoochPlotsToDisplay();
		saveScreenshot();
		saveBIPFile();
	}

	protected String writeScanForChooch()
	{
		// write scan to temporary file for chooch
		filePrefix = CmdScan.createDirAndGeneratePrefixWithDir(ee, true);
		choochInput = new File(filePrefix+".dat");

		choochInput.delete();
		
		double startEnergy		= trace.get(0).x;
		double endEnergy		= trace.get(trace.size()-1).x;
		double centerEnergy		= ee.getEnergy();
		
		String choochComment = Util.sprintf("%s Start=%.1f End=%.1f Center=%.1f",ee,startEnergy,endEnergy,centerEnergy);
	   	saveTraceForChooch(choochInput, choochComment);
	   	
	   	return filePrefix;
	}
	
	// helper to writeScanForChooch
	protected void saveTraceForChooch(File file, String comment)
	{
		System.out.println("Saving trace to "+file.getAbsolutePath()+" with comment "+comment);
		
		PrintWriter pw;
		try
		{
			pw = new PrintWriter(new FileOutputStream(file));
		}
		catch (FileNotFoundException fnfe)
		{
			fnfe.printStackTrace();
			return;
		}

		//First line must be a comment
		pw.println(comment);

		//Second line must be the number of data points
		pw.println(trace.size());

		// write out the vector data itself
		for (Point2D point : trace)
		{
			pw.println(point.x+" "+point.y);
		}
		pw.close();
	}
	
	protected void runChooch()
	{
		choochOutputPS = new File(filePrefix+"_chooch.ps");
		choochOutputPS.delete();
		
		choochOutputEFS = new File(filePrefix+".efs");
		choochOutputEFS.delete();
		
		String taskString = Util.sprintf("%s%s -e %s -a %s -o %s %s", choochDir, choochFile, ee.element, ee.edge, filePrefix+".efs", choochInput.getAbsolutePath());
		
		choochTask = ExecTask.new_ExecTask(taskString, choochFile, true, false);
		choochTask.start();
		choochTask.waitFor();
	}
	
	protected void processChoochOutput()
	{
		// extract MAD energies from Chooch output
		for (String line : choochTask.stdout)
		{
			StringTokenizer st = new StringTokenizer(line," |\t");
			if (!st.hasMoreTokens()) continue;
			String nextToken = st.nextToken();
			
			if (nextToken.equals("peak"))
			{
				if (st.hasMoreTokens())
				{
    				double peakEnergy 	= Double.parseDouble(st.nextToken());
    				double peak_fPP		= Double.parseDouble(st.nextToken());
    				double peak_fP		= Double.parseDouble(st.nextToken());
    				
    				//sd: This version of chooch program does not provide "Remote" energy
    				//    Ward and Nukri suggested to set 
    				//    Remote energy to an approximate value of peak + 400
    				double remoteEnergy = peakEnergy + 400.0;
    				
    				DevScan.get().setPeak(peakEnergy, false);
    				DevScan.get().setDefaultPeak(peakEnergy, false);
    				DevScan.get().setPeak_fP(peak_fP, false);
    				DevScan.get().setPeak_fPP(peak_fPP, false);
    				DevScan.get().setRemote(remoteEnergy, false);
    				DevScan.get().setDefaultRemote(remoteEnergy, false);
				}
			}
			else if (nextToken.equals("infl"))
			{
				if (st.hasMoreTokens())
				{
					double inflEnergy 	= Double.parseDouble(st.nextToken());
					double infl_fPP		= Double.parseDouble(st.nextToken());
					double infl_fP		= Double.parseDouble(st.nextToken());
					
    				DevScan.get().setInflection(inflEnergy, false);
    				DevScan.get().setDefaultInflection(inflEnergy, false);
    				DevScan.get().setInfl_fP(infl_fP, false);
    				DevScan.get().setInfl_fPP(infl_fPP, false);
				}
			}
		}
		
		DevScan.get().sendUpdate();
	}
	
	/*protected void convertPStoPDF()
	{
		// show PDF
		if (!choochOutputPS.exists())
		{
		        Log.error("Chooch output doesn't exist, skipping PDF display.  Expected path: "+choochOutputPS.getAbsolutePath());
			return;
		}

		choochOutputPDF = new File(filePrefix+"_chooch.pdf");
		String pdfCmd = 
			Main.getAppDir()+"/tools/scripts/ps2pdf_chooch.sh "+
			choochOutputPDF.getAbsolutePath()+" "+
			choochOutputPS.getAbsolutePath();
		ExecTask.syncExec(pdfCmd, gsFile, true);
		
		DevScan.get().setChoochPDFPath(choochOutputPDF.getAbsolutePath(), true);
		Util.sleep(50); // separate from potential updates in the near future
	}*/
	
	protected void sendChoochPlotsToDisplay()
	{
		if (!choochOutputEFS.exists())
		{
			Log.error("Chooch EFS output doesn't exist, skipping plot display.  Expected path: "+choochOutputEFS.getAbsolutePath());
			return;
		}
		
		BufferedReader br;
		ArrayList<Point2D> fP = new ArrayList<Point2D>(), fPP = new ArrayList<Point2D>();
		double fP_lowVal=Double.POSITIVE_INFINITY, fP_highVal=Double.NEGATIVE_INFINITY, fPP_lowVal=Double.POSITIVE_INFINITY, fPP_highVal=Double.NEGATIVE_INFINITY;
		try
		{
			br = new BufferedReader(new FileReader(choochOutputEFS));
			String line;
			while ( (line=br.readLine()) != null)
			{
				StringTokenizer st = new StringTokenizer(line);
				if (st.hasMoreTokens())
				{
					double energy  = Double.parseDouble(st.nextToken());
					double fPP_val  = Double.parseDouble(st.nextToken());
					double fP_val = Double.parseDouble(st.nextToken());
					
					fP.add(new Point2D(energy, fP_val));
					fPP.add(new Point2D(energy, fPP_val));

					if (fP_val  < fP_lowVal)   fP_lowVal   = fP_val;
					if (fP_val  > fP_highVal)  fP_highVal  = fP_val;
					if (fPP_val < fPP_lowVal)  fPP_lowVal  = fPP_val;
					if (fPP_val > fPP_highVal) fPP_highVal = fPP_val;
				}
			}
		}
		catch (FileNotFoundException e) { e.printStackTrace(); }
		catch (IOException e)			{ e.printStackTrace(); }
		catch (NumberFormatException e) { e.printStackTrace(); }
		
		if (fP.size()>0 && fPP.size()>0)
		{
			DevScan.get().setGraphYMin(TRACE_ID.fPrime, fP_lowVal, false);
			DevScan.get().setGraphYMax(TRACE_ID.fPrime, fP_highVal, false);
			DevScan.get().setDefaultYMin(TRACE_ID.fPrime, fP_lowVal, false);
			DevScan.get().setDefaultYMax(TRACE_ID.fPrime, fP_highVal, false);
			
			DevScan.get().setGraphYMin(TRACE_ID.fDoublePrime, fPP_lowVal, false);
			DevScan.get().setGraphYMax(TRACE_ID.fDoublePrime, fPP_highVal, false);
			DevScan.get().setDefaultYMin(TRACE_ID.fDoublePrime, fPP_lowVal, false);
			DevScan.get().setDefaultYMax(TRACE_ID.fDoublePrime, fPP_highVal, false);

			DevScan.get().setLeftAxisTrace(TRACE_ID.fPrime, false);
			DevScan.get().setRightAxisTrace(TRACE_ID.fDoublePrime, false);
			
			DevScan.get().setGraphYLabel(TRACE_ID.fPrime, "f Prime", false);
			DevScan.get().setGraphYLabel(TRACE_ID.fDoublePrime, "f Double Prime", false);
			
			DevScan.get().setGraphTrace(TRACE_ID.fPrime, fP, false);
			DevScan.get().setGraphTrace(TRACE_ID.fDoublePrime, fPP, false);
			
			DevScan.get().setTraceVisible(TRACE_ID.edge, 			false, false);
			DevScan.get().setTraceVisible(TRACE_ID.fPrime, 			true, false);
			DevScan.get().setTraceVisible(TRACE_ID.fDoublePrime, 	true, false);
			
			DevScan.get().sendUpdate();
		}
	}
	
	protected void saveScreenshot()
	{
		final WdgGraph graph = TabScan.get().graphWidget;
		graph.saveScreenshot(filePrefix+".png");
	}
	
	protected void saveBIPFile()
	{
		FileBIP.saveFile(filePrefix+".bip", TabScan.get().graphModel, ModelScan.get());
	}
}
