package gov.anl.gmca.scan;

import gov.anl.gmca.common.EpicsDevice;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.*;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.ActionResult;
import gov.anl.gmca.util.Util;
import gov.aps.jca.dbr.DBR;

// handles 23x:aim_adc1* PVs
//
// Request and read methods are blocking
//
// Portions ported from user_scan.tcl

public class DevMCA extends EpicsDevice
{
	public int channelMax;
	public static Boolean isFalcon = null;

	private double lastDcmVsum = 0;
	private void updateDcmVsum()
	{
		double currentSum = Epics.getDouble(PV_MISC.dcm);
		if (currentSum>lastDcmVsum) lastDcmVsum = currentSum;
	}
	public double getLastSpectrumDcmVsum() { return lastDcmVsum; }
	
	///////////////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////// public methods ////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////////////
	
	public synchronized ActionResult requestSpectrum(int startChannel, int endChannel, double scanTimeSec)
	{
		lastDcmVsum = 0;
		int channelMax = Epics.getInteger(PV_MCA._NMAX);
		
		if (startChannel < 0 || startChannel > channelMax)
		{
			String errString = "startChannel "+startChannel+" not between 0 and "+channelMax;
			Log.error(errString);
			return new ActionResult(errString);
		}
		if (endChannel < 1 || endChannel < startChannel || endChannel > channelMax)
		{
			String errString = "endChannel "+endChannel+" not between 0 and "+channelMax+" or less than startChannel "+startChannel;
			Log.error(errString);
			return new ActionResult(errString);
		}
		
		Epics.putInteger(	PV_MCA._PCTL,			startChannel);
		Epics.putInteger(	PV_MCA._PCTH, 			endChannel);
		Epics.putInteger(	PV_MCA._NUSE,			endChannel-startChannel+1); // channels used
		Epics.putDouble(	PV_MCA._PRTM,			scanTimeSec); 				// preset realtime
		Epics.putDouble(	PV_MCA._PLTM,			scanTimeSec); 				// preset livetime
		Epics.putInteger(	PV_MCA._R0IP,			0);							// disable ROI
		
		resetStateVariables();
		
		Epics.putInteger(PV_MCA.EraseStart, 1);
		
		// wait for acq->1 then acqg->0
		long startTime = System.currentTimeMillis();
		double timeoutSec = scanTimeSec+5;
		while (!acqg0)
		{
			long now = System.currentTimeMillis();
			if (now-startTime > timeoutSec*1000.0)
			{
				String errString = "Timeout after "+timeoutSec+" seconds waiting for ACQG->0";
				Log.error(errString);
				return new ActionResult(errString);
			}
			
			updateDcmVsum();
			Epics.poll();
			Util.sleep(5);
		}
		
		// send 1->read
		Epics.putInteger(PV_MCA._READ, 1);
		
		// wait for read->1 then read->0
		startTime = System.currentTimeMillis();
		timeoutSec = 5;
		while (!read0)
		{
			long now = System.currentTimeMillis();
			if (now-startTime > timeoutSec*1000.0)
			{
				String errString = "Timeout after "+timeoutSec+" seconds waiting for READ->0";
				Log.error(errString);
				return new ActionResult(errString);
			}
			
			updateDcmVsum();
			Epics.poll();
			Util.sleep(5);
		}
		
		return ActionResult.SUCCESS;
	}
	
	public int[] readSpectrum()
	{
		return Epics.getIntegerArray(PV_MCA._VAL);
	}
	
	public static class SumNet { public int sum, net; }
	
	public void requestSumNet()
	{
	}
	
	public SumNet readSumNet()
	{
		return null;
	}
	
	protected static class OptimizeResult
	{
		public int gain, offset, N1, N2;
		public OptimizeResult(int gain, int offset, int N1, int N2) { this.gain=gain; this.offset=offset; this.N1=N1; this.N2=N2; }
	}

	// "zoom" the MCA to the specified ROI
	public void optimize(double roiLow, double roiHigh)
	{
		OptimizeResult result = optimizeMCA_ADC(roiLow, roiHigh);
		            
      	setROILow(result.N1);
      	setROIHigh(result.N2);
	}
	
	public void setROILow(int lowChannel)			{ Epics.putInteger(PV_MCA._R0LO, lowChannel); }
	public void setROIHigh(int highChannel)			{
		Epics.putInteger(PV_MCA._R0HI, highChannel);
		Epics.putInteger(PV_MISC.falcon_roi_copy, 1);
		Util.sleep(300); // ROI copy needs about 100ms
	}
	
	protected static class Calibration
	{
		public double gainSlope, intercept, offsetSlope;
	}
	
	//sd: routine to convert channel to energy
	//sd: Derek's New formula- 02/02/05
	//
	//	Energy = 2048 * slope * { Channel + 128 * [int(offset/128)] }
	//		------------------------------------------------------ - offsetEnergy
	//					Gain
	public double channelToEnergy(int channel, int gain, int offset)
	{
		if (isFalcon) {
			double energy = fldSlope * channel + fldOffset;
			return energy;
		} else {
			Calibration cb = MCAcalibDB(gain);
			double energy = cb.gainSlope * channel + (int) (offset / 128) * cb.offsetSlope + cb.intercept;
			return energy;
		}
	}
	
	public int energyToChannel(double energy, int gain, double offset)
	{
		if (isFalcon) {
			int channel = (int)( (energy - fldOffset) / fldSlope + 0.5);
			return channel;
		} else {
			Calibration calib = MCAcalibDB(gain);
			int channel =
					(int)
							(
									(
											energy
													- (
													(offset / 128)
															* calib.offsetSlope
											)
													- calib.intercept
									)
											/ calib.gainSlope
							);

			if (channel < 0 || channel >= channelMax) {
				Log.error("EnergyToChannel: energy=" + energy + " gain=" + gain + " offset=" + offset + " channel=" + channel);
				return 0;
			}
			return channel;
		}
	}
	
	public int getGain()
	{
		return enumToGain(Epics.getInteger(PV_MCA.GAIN));
	}
	
	public int getOffset()
	{
		return Epics.getInteger(PV_MCA.OFFSET);
	}
	
	public void setGain(int gain)
	{
		if (isGainOK(gain))
		{
			Epics.putInteger(PV_MCA.GAIN, gainToEnum(gain));
		}
		else
		{
			throw new IllegalArgumentException("Invalid gain: "+gain);
		}
	}
	
	public void setOffset(int offset)
	{
		if (isOffsetOK(offset))
		{
			Epics.putInteger(PV_MCA.OFFSET, offset);
		}
		else
		{
			throw new IllegalArgumentException("Invalid offset: "+offset);
		}
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////// private methods ////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////////////
	
	protected double fldSlope = Double.NaN, fldOffset = Double.NaN;

	protected Calibration MCAcalibDB(int gain)
	{
		//sd: validate input gain
		if (!isGainOK(gain))
		{
			return null;
		}
		
		Calibration ret = new Calibration();
		
		ret.gainSlope 	= (2048.0 * fldSlope) / gain;
		ret.intercept 	= -1.0 * fldOffset;
		ret.offsetSlope = (2048.0 * fldSlope * 128.0) / gain;
		
		return ret;
	}
	
	public DevMCA()
	{
		super("MCA");

		isFalcon = PVPrefix.MCA.getPrefix().toLowerCase().contains("falcon");

		channelMax = Epics.getInteger(PV_MCA._NMAX);
		fldSlope 	= Epics.getDouble(PV_MCA._CALS);
		fldOffset 	= Epics.getDouble(PV_MCA._CALO);
		
		for (PVName pv : PV_MCA.values())
		{
			boolean reg = true;
			if (isFalcon && pv.toString().contains("GAIN")) reg=false;
			if (isFalcon && pv.toString().contains("OFFSET")) reg=false;
			if (reg) registerDevicePV(pv);
		}
		registrationComplete();
	}
	
	public static DevMCA get() { return (DevMCA)getDevice("MCA"); }
	
	// These four booleans represent the four states, in order, that are needed before the spectrum
	// can be read from aim_adc1.VAL.
	// They are set to true, only in the correct order, when each state is received.
	protected volatile boolean acqg1=false,acqg0=false,read1=false,read0=false;
	protected void resetStateVariables() { acqg1=false; acqg0=false; read1=false; read0=false; }

	@Override
	public void valueChangedAction(String pvName, DBR pvValue)
	{
		if (pvName.equals(PV_MCA._ACQG.toString()))
		{
			int value = Epics.dbrToInt(pvValue);
			if (1==value)          {System.out.println("acqg1=true"); acqg1=true;}
			if (0==value && acqg1) {System.out.println("acqg0=true"); acqg0=true;}
		}
		else if (pvName.equals(PV_MCA._READ.toString()))
		{
			int value = Epics.dbrToInt(pvValue);
			if (acqg0 && 1==value) {System.out.println("read1=true"); read1=true;}
			if (0==value && read1) {System.out.println("read0=true"); read0=true;}
		}
	}

	@Override
	public void abort()
	{
		Epics.putInteger(PV_MCA.Stop,1);
	}
	
	protected boolean isGainOK(int gain)
	{
		for (int i=256; i<=16384; i*=2)
		{
			if (i == gain) return true;
		}
		return false;
	}
	
	int[][] gainLookup = new int[][]
	{
    	{ 0, 256},
    	{ 1, 512},
    	{ 2, 1024},
    	{ 3, 2048},
    	{ 4, 4096},
    	{ 5, 8192},
    	{ 6, 16384}
	};
	
	protected int gainToEnum(int gain)
	{
		for (int[] gainEntry : gainLookup)
		{
			if (gainEntry[1]==gain) return gainEntry[0];
		}
		Log.error("Invalid gain "+gain);
		return -1;
	}
	
	protected int enumToGain(int enumGain)
	{
		for (int[] gainEntry : gainLookup)
		{
			if (gainEntry[0]==enumGain) return gainEntry[1];
		}
		Log.error("Invalid enum gain "+enumGain);
		return -1;
	}
	
	protected boolean isOffsetOK(int offset)
	{
		return (offset >= 0 && offset <= 16128);
	}
	
	protected OptimizeResult optimizeMCA_ADC(double E1, double E2)
	{
		double E_center = (E1+E2)/2.0;
		double dE = Math.abs(E2-E1);
		
		// Find the maximum gain at which the MCA scale is wider that $dE
		int gain = 16384;
		
		while (scaleMCA(gain) <= dE && gain > 1024)
		{
			gain /= 2;
		}
		
		Calibration calib = MCAcalibDB(gain);
		
		// E_scale=gainSlope
		double E_scale = calib.gainSlope;
		
		// Find $offset at which $E_center appears at the center ($channel=1023)
		// of the MCA (follows from: $E=$E_scale*$channel+$E_offset;)
		int midChannel = (int)((channelMax-1.0)/2.0);
		
		double E_offset = E_center - (E_scale * midChannel);
		
		// This follows from: $E_offset=$Intercept+int($offset/128)*$offsetSlope;
		int offset = 128 * (int)((E_offset - calib.intercept) / calib.offsetSlope);
		
		if (offset < 0)
		{
			Log.note("optimizeMCA_ADC: offset="+offset+", resetting to 0");
			offset = 0;
		}
		
		if (offset > 16128) // 16128 = 256 * 63
		{
			Log.note("optimizeMCA_ADC: offset="+offset+", resetting to 16128");
			offset = 16128;
		}
		
		int N1 = energyToChannel(E1, gain, offset);
		int N2 = energyToChannel(E2, gain, offset);
		
		if (N1 < 0 || N1 > channelMax-1)
		{
			Log.error("optimizeMCA_ADC: Calculated N1="+N1+" not in the range 0 to "+(channelMax-1));
			return null;
		}
		
		if (N2 < 0 || N2 > channelMax-1)
		{
			Log.error("optimizeMCA_ADC: Calculated N2="+N2+" not in the range 0 to "+(channelMax-1));
			return null;
		}
		
		Log.note("optimizeMCA_ADC: optimal gain="+gain+" offset="+offset+" N1="+N1+" N2="+N2);
		
		return new OptimizeResult(gain, offset, N1, N2);
	}
	
	//-------------------------------------------------------------------
	//sd:
	// This function returns the MCA energy scale ($dE) at given $gain.
	// -------------------------------------------------------------------
	protected double scaleMCA(int gain)
	{
		// sd: scaleMCA is calculating energy of the maximum channel. So the conversion is
		// channel to energy; conv=1
		Calibration calib = MCAcalibDB(gain);
		
		double mcaScale = (channelMax-1) * calib.gainSlope;
		
		return mcaScale;
	}
	
	@Override protected void onAllPVsConnected() {}
}
