package gov.anl.gmca.scan.ui;

import gov.anl.gmca.gui.GUI;
import gov.anl.gmca.io.ExecTask;
import gov.anl.gmca.scan.ModelScan;
import gov.anl.gmca.scan.ModelScan.ScanPlotMode;
import gov.anl.gmca.util.Util;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ControlEvent;
import org.eclipse.swt.events.ControlListener;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.graphics.*;
import org.eclipse.swt.widgets.Canvas;
import org.eclipse.swt.widgets.Composite;

import java.util.ArrayList;
import java.util.concurrent.atomic.AtomicBoolean;

public class WdgGraph extends Canvas implements PaintListener
{
	private static final Font EXTRA_INFO_FONT = GUI.FONT_12_FIXED;
	private static final int EXTRA_INFO_LEFT_MARGIN = 10;
	private static final int EXTRA_INFO_LINE_COUNT = 5;
	private int extraInfoTextHeight;
	
	protected ModelGraph graph;
	protected Image buffer;
	
	protected static WdgGraph inst;
	public static WdgGraph get() { return inst; }
	
	protected LayerLabels layerLabels;
	protected LayerAxis layerAxisX, layerAxisYLeft, layerAxisYRight;
	protected LayerPlot layerPlot;
	protected LayerOverlays layerOverlays;
	
	WdgGraph(Composite parent, int style, ModelGraph graphModel)
	{
		super(parent, SWT.DOUBLE_BUFFERED);
		inst = this;
		
		this.graph = graphModel;
		graph.setWidget(this);
		
		setBackground(GUI.COLOR_LIGHT_BLUE2);
		
		layerLabels = new LayerLabels(graph, getBackground());
		layerAxisX = new LayerAxis(graph, SWT.HORIZONTAL, SWT.BOTTOM, getBackground());
		layerAxisYLeft = new LayerAxis(graph, SWT.VERTICAL, SWT.LEFT, getBackground());
		layerAxisYRight = new LayerAxis(graph, SWT.VERTICAL, SWT.RIGHT, getBackground());
		layerPlot = new LayerPlot(graph);
		layerOverlays = new LayerOverlays(graph);
		
		ListenerEdgeDisplay.get().addListeners(this);
		ListenerBarDrag.get().addListeners(this);
		ListenerZoom.get().addListeners(this);
		//ListenerScroll.get().addListeners(this);
		ContextMenu.get().addMenu(this);
		
		addPaintListener(this);
		
		addControlListener(new ControlListener()
		{
			@Override public void controlResized(ControlEvent arg0)
			{
				graph.redraw(true,true,true,true,true,false);
			}
			@Override public void controlMoved(ControlEvent arg0) {}
		});
	}
	
	protected Point lastSize = null;
	
	///////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////// composite layers into a single image ///////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////////

	public void composite()
	{
		//System.out.println("Compositing scan display");
		
		GC imgGC = (buffer==null)?null:new GC(buffer);
		
		// if window resized
		if (null == lastSize || lastSize.x != getBounds().width || lastSize.y != getBounds().height)
		{
			// determine new size
			Point size = new Point(getBounds().width, getBounds().height);
			if (size.x<=0 || size.y<=0) return;
			lastSize = size;
			
			// recreate buffer
			if (null!=buffer && !buffer.isDisposed()) buffer.dispose();
			buffer = new Image(GUI.getDisplay(), size.x, size.y);
			
			if (null!=imgGC && !imgGC.isDisposed()) imgGC.dispose();
			imgGC = new GC(buffer);
			
			// recalculate layout
			graph.recalc(imgGC, size);
		}
		
		// draw layers
		imgGC.setAntialias(SWT.ON);
		
		layerLabels.draw(imgGC);
		layerAxisX.draw(imgGC);
		layerAxisYLeft.draw(imgGC);
		layerAxisYRight.draw(imgGC);
		layerPlot.draw(imgGC);
		
		imgGC.dispose();
		
		redraw();
	}
	
	///////////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////// main paint method ////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////////
	
	private boolean needsRedraw = false;
	
	@Override
	public void paintControl(PaintEvent pe)
	{
		Rectangle pb = ModelGraph.get().graphBounds;
		if (pe.x==pb.x && pe.y==pb.y && pe.width==pb.width && pe.height==pb.height)
		{
			GC imgGC = (buffer==null)?null:new GC(buffer);
			imgGC.setClipping(pb);
			layerPlot.drawDirect(imgGC);
			layerOverlays.drawDirect(imgGC);
			try
			{
				pe.gc.drawImage(buffer, pe.x, pe.y, pe.width, pe.height, pe.x, pe.y, pe.width, pe.height);
			}
			catch (IllegalArgumentException e) {}
			imgGC.dispose();
		}
		else
		{
			if (getBounds().width<=0 || getBounds().height<=0) return;
			
			if (null==buffer||buffer.isDisposed()) composite();
			
			if (null!=buffer&&!buffer.isDisposed())
			{
				GC gc = pe.gc;
				try
				{
					gc.drawImage(buffer, pe.x, pe.y, pe.width, pe.height, pe.x, pe.y, pe.width, pe.height);
				}
				catch (IllegalArgumentException e) {} // TODO: what is causing this?
				layerOverlays.draw(gc);
			}
		}
	}
	
	protected void paintToImage(Image image, boolean addExtraInfo)
	{
		GC gc = new GC(image);
		boolean isEdgeScan = ModelScan.get().getScanPlotMode()==ScanPlotMode.EDGE;
		
		gc.setFont(EXTRA_INFO_FONT);
		extraInfoTextHeight = isEdgeScan ? gc.textExtent("Wy").y : 0;
		
		int textBoxY = image.getBounds().height - extraInfoTextHeight*EXTRA_INFO_LINE_COUNT - 2;
		Point size = new Point(image.getBounds().width, textBoxY);
		lastSize = size;
		graph.recalc(gc, size);
		
		gc.setAntialias(SWT.ON);
		
		layerLabels.drawDirect(gc);
		layerAxisX.drawDirect(gc);
		layerAxisYLeft.drawDirect(gc);
		layerAxisYRight.drawDirect(gc);
		layerPlot.drawDirect(gc);
		layerOverlays.drawDirect(gc);
		
		if (isEdgeScan) drawExtraInfo(gc, image.getBounds(), textBoxY);
		
		gc.setAntialias(SWT.OFF);
		gc.dispose();
		
		// TODO: temporary fix - real fix is to have a second ModelGraph for off-screen drawing
		graph.redraw(false, false, false, false, true, false);
	}
	
	// on return from this method, the file must be written
	public void saveScreenshot(final String screenshotPath)
	{
		final WdgGraph inst = this;
		final AtomicBoolean fileWritten = new AtomicBoolean(false);
		
		GUI.asyncExec(new Runnable(){public void run()
		{
			int x = inst.getSize().x, y = inst.getSize().y;
			if (x<781) x=781;
			if (y<605) y=605; // +20 of minimum for extra info
			
			Image screenshot = new Image(GUI.getDisplay(), x, y);
			try
			{
				paintToImage(screenshot, true);
	    		
	    		ImageLoader loader = new ImageLoader();
	    		loader.data = new ImageData[]{screenshot.getImageData()};
	    		loader.save(screenshotPath, SWT.IMAGE_PNG);
			}
			catch (Exception e) { e.printStackTrace(); }
			finally
			{
				screenshot.dispose();
			}
    		
    		fileWritten.set(true);
		}});

		while (!fileWritten.get()) Util.sleep(50);
	}
	
	public void printScreenshot()
	{
		new Thread(new Runnable(){public void run()
		{
    		String filepath = System.getProperty("user.home")+"/bluice_temp/temp_screenshot";
    		saveScreenshot(filepath+".png");
    		ExecTask.syncExec("/usr/bin/convert -border 50x50 -bordercolor white -page 792x612 "+filepath+".png "+filepath+".pdf", "convert", true);
    		//ExecTask.asyncExec("firefox "+filepath+".pdf", "firefox", true);
    		ExecTask.syncExec("/usr/bin/lpr "+filepath+".pdf", "lpr", true);
		}}, "Print screenshot").start();
	}
	
	private void drawExtraInfo(GC gc, Rectangle bounds, int textBoxY)
	{
		// setup
		gc.setClipping(bounds);
		gc.setFont(EXTRA_INFO_FONT);
		gc.setForeground(GUI.COLOR_BLACK);
		gc.setBackground(GUI.COLOR_WHITE);
		gc.drawRectangle(0, bounds.height-extraInfoTextHeight*EXTRA_INFO_LINE_COUNT-2, bounds.width-1, extraInfoTextHeight*EXTRA_INFO_LINE_COUNT+1);

		ArrayList<String> lines = new ArrayList<String>();
		lines.add(Util.sprintf("-------------------------------------"));
		lines.add(Util.sprintf("|      |  energy  |    f'' |   f'   |"));
		lines.add(Util.sprintf("| peak | %7.2f |   %4.2f | %6.2f |", ModelScan.get().getPeak(), ModelScan.get().getPeak_fPP(), ModelScan.get().getPeak_fP()));
		lines.add(Util.sprintf("| infl | %7.2f |   %4.2f | %6.2f |", ModelScan.get().getInflection(), ModelScan.get().getInfl_fPP(), ModelScan.get().getInfl_fP()));
		lines.add(Util.sprintf("-------------------------------------"));
		
		for (int lineIndex=0; lineIndex < lines.size(); lineIndex++)
		{
			int y = textBoxY + extraInfoTextHeight*lineIndex;
			
			gc.drawString(lines.get(lineIndex), EXTRA_INFO_LEFT_MARGIN, y, true);
		}
	}
}
