/**
 * @author spothineni
 * 
 * All the screening Tasks handled in this file
 * 
 * Mount and unmount samples using automounter
 * Autocentering
 * Wait for center
 * collect images with user specified orientation
 * 
 * Checks directory
 * Performs pre-collect moves (detector, collimator, ring light, beam size) on its own (could be merged with collect)
 * Calls DevFPE synchronously for each frame (could be made shutterless by changing the loop)
 * 
 * First half of file is for collection (and image taking)
 * 
 * task flag is the current task
 * repeat flag overrides this and changes the current task
 * 
 * CrystalList is a list of strings (E1,E2) in the order the user selected them.
 * 
 * screening thread has a while loop over the crystalList
 * (singleSampleRepeat and moun.equals("none") to be documented later)
 * - cur is set to the first selected sample row in the table
 * - screenCrystal handles mount and other operations
 * - handles return flags from screenCrystal
 * 
 * finishScreening sets final states
 * 
 * on pause, screening thread keeps running
 * 
 * for normal operations screening thread runs for a long time, unless stopThread is called e.g. if robot goes to pause
 * 
 * in screenCrystal:
 *   case statement for each of the 8 operations (mount, center, pause, etc.)
 *   auto-center is synchronous (since screening is a thread)
 *   pause moves screening thread to the while loop
 *   snapshots moves backlight in if checked and takes images for the checked orientations
 */

package gov.anl.gmca.screening;

import gov.anl.gmca.collect.DevBinMode;
import gov.anl.gmca.collect.DevBinMode.BinMode;
import gov.anl.gmca.collect.DevBinMode.ModeChange;
import gov.anl.gmca.collect.DevFPE;
import gov.anl.gmca.collect.DevFPE.FrameListResult;
import gov.anl.gmca.collect.DevFPE.XYZMotion;
import gov.anl.gmca.collect.Filenames;
import gov.anl.gmca.collect.ModelCollectRun.Axis;
import gov.anl.gmca.collect.ModelCollectRun.Point3D;
import gov.anl.gmca.collect.ModelCollectRunList;
import gov.anl.gmca.common.DevDataDir;
import gov.anl.gmca.common.DevMove.MoveCmd;
import gov.anl.gmca.common.DevMove.MoveResult;
import gov.anl.gmca.common.ui.NBMessageBox;
import gov.anl.gmca.gui.GUI;
import gov.anl.gmca.hutch.*;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.ExecTask;
import gov.anl.gmca.io.PVLookup.*;
import gov.anl.gmca.log.db.ScreeningEvents;
import gov.anl.gmca.multistrategy.DevReferenceData;
import gov.anl.gmca.multistrategy.ModelReferenceData;
import gov.anl.gmca.raster.CmdRasterAuto;
import gov.anl.gmca.raster.ModelRasterSiteList.Site;
import gov.anl.gmca.raster.ModelRasterState.RasterListType;
import gov.anl.gmca.raster.calc.RasterGeometry;
import gov.anl.gmca.sample.DevCollimatorCtrl;
import gov.anl.gmca.sample.DevCollimatorCtrl.CCMove;
import gov.anl.gmca.sample.DevCollimatorCtrl.CollimatorSize;
import gov.anl.gmca.sample.DevInOut.InOut;
import gov.anl.gmca.sample.DevSlider;
import gov.anl.gmca.sample.ModelBeamsize;
import gov.anl.gmca.screening.ui.WdgScreeningTasks;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.users.ModelUserTable.Client;
import gov.anl.gmca.util.ActionResult;
import gov.anl.gmca.util.Point2D;
import gov.anl.gmca.util.Util;
import org.eclipse.swt.SWT;

import java.io.File;
import java.util.ArrayList;
import java.util.StringTokenizer;
import java.util.regex.Matcher;

public class CmdScreening implements Runnable {

	public static int group = 1;
	public static String snapImage1 = "", snapImage2 = "", snapImage3 = "";
	public static String image1 = "", image2 = "", image3 = "";
	public static int manualRepeatFlag = -1;
	public static int noofImagesCollected = 0;
	public static ExecTask runningTask = null;
	public static String imageDirectory;
	public static String screenDirectory;
	public static String primaryDirectory;

	public static boolean stopThread = true;

	public volatile boolean _requestStop = false;
	protected CollimatorSize origCollimator;
	protected Point2D origGSSize = null;

	boolean mount, autoCenter, pause1, takejpeg, collect1, collect2, collect3, pause2;

	int loopCenter = 0, fpeError = 0, repeatFlag = 0;
	int fileExists = 0;

	ModelCrystalList crystalList = new ModelCrystalList();
	ModelScreening scr;

	public CmdScreening() {
	}

	public void requestStop() {
		_requestStop = true;
	}

	public CmdScreening(ModelCrystalList crystalList) {

		this.crystalList = crystalList;
	}

	public void updateActions() {

		mount = Epics.getBoolean(PV_SCR.MountNextXtal);
		autoCenter = Epics.getBoolean(PV_SCR.CenterXtal);
		pause1 = Epics.getBoolean(PV_SCR.Pause4ctrng);
		takejpeg = Epics.getBoolean(PV_SCR.takeJpeg);
		collect1 = Epics.getBoolean(PV_SCR.TakeImage0);
		collect2 = Epics.getBoolean(PV_SCR.TakeImage45);
		collect3 = Epics.getBoolean(PV_SCR.TakeImage90);
		pause2 = Epics.getBoolean(PV_SCR.Pause4collect);
	}

	public static int labelToindex(String label) {
		String pinNumber = label.substring(1, label.length());
		int puckNumber = 1;
		int pinIndex = 1;
		for (int i = 0; i < ModelScreening.NoofPucks; i++) {
			if (ModelScreening.puckLabels[i] == label.charAt(0))
				puckNumber = i;
		}
		pinIndex = puckNumber * ModelScreening.getPinsPerPuck()
				+ Integer.valueOf(pinNumber);
		return (pinIndex - 1);
	}

	public static String getMounted(int index) {

		int pinsPerPuck = ModelScreening.getPinsPerPuck();
		int puckNumber = index / pinsPerPuck;
		int pinNumber = index % pinsPerPuck;

		String puckLetter = Character
				.toString(ModelScreening.puckLabels[puckNumber]);

		// String params = puckLetter + " " + Integer.toString(pinNumber+1);
		String params = puckLetter + Integer.toString(pinNumber + 1);
		return (params);
	}

	public String getDir() {
		String scrDir = Epics.getString(PV_SCR_RUN.directory256.run(0));
		// Log.error(scrDir);
		if (scrDir == null || scrDir.equals("null"))
			scrDir = "";
		return DevDataDir.getPrimary() + "/" + scrDir;
	}

	public String getFPEDir() {
		String scrDir = Epics.getString(PV_SCR_RUN.directory256.run(0));
		// Log.error(scrDir);
		if (scrDir == null || scrDir.equals("null"))
			scrDir = "";
		String FPEDir = DevDataDir.getFPEDir() + "/" + scrDir;
		FPEDir = FPEDir.replace("//","/");

		// create it if it does not exist
		File dirFile = new File(FPEDir);
		if (!dirFile.exists()) {
			System.out.println("CollectImage mkdir " + FPEDir);
			if (!dirFile.mkdirs()) {
				Log.error("Couldn't create directory " + FPEDir);
				return getDir();
			}
		}		
		return FPEDir;
	}

	public double getDelta() {
		double delta = Epics.getDouble(PV_SCR_RUN.delta.run(0));
		return (delta);
	}

	public double getExpTime() {
		double delta = Epics.getDouble(PV_SCR_RUN.exposureTime.run(0));
		return (delta);
	}

	public double getDistance() {
		double distance = Epics.getDouble(PV_SCR_RUN.distance.run(0));
		return (distance);
	}

	public double getAttenuation_factors() {
		double distance = Epics.getDouble(PV_SCR_RUN.atten_factor.run(0));
		return (distance);
	}
	
	public String getFilename(int frameNumber) {
		return getFilenameReal(frameNumber, "");
	}

	public String getFilenameReal(int frameNumber, String newFilePrefix) {
		String filePrefix;
		if (newFilePrefix.isEmpty()) {
			filePrefix = Epics.getString(PV_SCR_RUN.fileroot.run(0));
		} else {
			filePrefix = newFilePrefix;		
		}

		if (frameNumber < 0)
			frameNumber = -frameNumber;
		/*
		String fileFrame = Integer.toString(frameNumber);
		while (fileFrame.length() < 5)
			fileFrame = "0" + fileFrame;
		*/
		
		String fileFrame=String.format("%05d", frameNumber);
		if (DetectorConfig.isEiger()) {
			fileFrame=String.format("%06d", frameNumber);
		}
		
		// Generate filename
		// MAR: A1_scr.00001
		// Pilatus: A1_scr_00001.cbf
		// Eiger: A1_0_000001_scr_master.h5

		String filename;
		int runNumber = 0; // screening is always run 0
		
		if (DetectorConfig.isMAR())
			filename = filePrefix + "_scr." + fileFrame;
		else if (DetectorConfig.isPilatus())
			filename = filePrefix + "_scr_" + fileFrame+".cbf";
		else if (DetectorConfig.isEiger())
			// eiger hdf5 mode
			//filename = filePrefix + "_" + runNumber + "_" + frameNumber + "_scr_master.h5";
			
			// eiger cbf mode
			filename = filePrefix + "_scr_" + fileFrame+".cbf";
		else
			throw new IllegalStateException("Unknown detector type");
		
		return (filename);
	}
	
	public static int getFrameNumber(int angleIndex, double angle, int img) {
		int frameNumber1;
		if (angle >= 0)
			frameNumber1 = (int) getAngle(angleIndex) + img + 1;
		else {
			if (angle < -360)
				frameNumber1 = (int) getAngle(angleIndex) + img + 1 + 720;
			else if (angle > -360 && angle < 0)
				frameNumber1 = (int) getAngle(angleIndex) + img + 1 + 360;
			else
				frameNumber1 = (int) getAngle(angleIndex) - img - 1;
		}
		return frameNumber1;
	}

	public static double getAngle(int angleIndex) {
		double angle = 0.0;
		if (angleIndex == 0)
			angle = Double.parseDouble(Epics.getString(PV_SCR.AngleImage0));
		else if (angleIndex == 1)
			angle = Double.parseDouble(Epics.getString(PV_SCR.AngleImage45));
		if (angleIndex == 2)
			angle = Double.parseDouble(Epics.getString(PV_SCR.AngleImage90));

		return angle;
	}

	public static int getTotalImagesScreen() {
		int images = 0;
		if (Epics.getBoolean(PV_SCR.TakeImage0))
			images = Epics.getInteger(PV_SCR.diffimages0);
		if (Epics.getBoolean(PV_SCR.TakeImage45))
			images += Epics.getInteger(PV_SCR.diffimages45);
		if (Epics.getBoolean(PV_SCR.TakeImage90))
			images += Epics.getInteger(PV_SCR.diffimages90);

		return images;

	}

	public int getNoofImages(int angleIndex) {

		int images = 1;
		if (angleIndex == 0)
			images = Epics.getInteger(PV_SCR.diffimages0);
		else if (angleIndex == 1)
			images = Epics.getInteger(PV_SCR.diffimages45);
		else if (angleIndex == 2)
			images = Epics.getInteger(PV_SCR.diffimages90);
		else {
			images = 0;
		}

		return images;
	}

	public void takeJpegSnapshot(double angle, String sample, String collectDir) {

		int frameNumber = (int) angle + 1;
		int angleInt = frameNumber;
		Epics.put(PV_SCR.CurrentTask, sample + " JPEGsnapshot" + angleInt);
		String filename_0 = getFilename(frameNumber);

		MoveCmd gonioMove = DevPmacMotor.get("gonio_omega").startMove(angle);
		gonioMove.waitFor();

		Util.takeSnapshot(collectDir + "/" + filename_0);
	}

	public int CollectImage(int angleIndex, String sample) {

		String filename;
		if (!Client.isMaster()) {
			stopThread = true;
			return 0;
		}

		if (noofImagesCollected == 1) {
			Epics.put(PV_SCR.CurrentTask, "RenameGroup");
			Util.sleep(100);
		}

		String collectDir = getDir();
		/* Take home directory if getDir() returns null or empty */
		if (collectDir == null || collectDir.equals("")) {
			collectDir = DevDataDir.getPrimary();
		}
		collectDir = collectDir.replace("//", "/");
		/* create directory if does's exist */
		File dirFile = new File(collectDir);
		if (!dirFile.exists()) {

			System.out.println("CollectImage mkdir "+collectDir);
			if (!dirFile.mkdirs()) {
				Log.error("Couldn't create directory " + collectDir);
			}
		}
		imageDirectory = collectDir;
		double delta = getDelta();
		double exposureTime = getExpTime();

		/* set screening status */
		Epics.putInteger(PV_SCR.ScreeningStatus, 1);

		/* Move the detector */
		double distance = getDistance();
		MoveCmd detectorMove = null;
		if (Math.abs(DevPmacMotor.get("detector_z").getPosition()-distance) >= 0.001)
		{
			Log.note("Moving detector to "+Util.sprintf("%.3f",distance)+" mm");
			detectorMove = DevPmacMotor.get("detector_z").startMove(distance);
		}

		/*MoveCmd detectorMove = DevPmacMotor.get("detector_z").startMove(
				distance);*/

		/* change the ring light intensity */
		double ringInt = DevSlider.get("ring").getHigherLimit();
		DevSlider.get("ring").setIntensity(ringInt);


		double angle = getAngle(angleIndex);
		System.out.println("angle for collect : " + angle);
		int frameNumber = (int) angle + 1;
		int angleInt = (int) angle + 1;

		Epics.put(PV_SCR.CurrentTask, sample + " Collect" + angleInt);

		/* collimator in */
		CCMove ccMove = DevCollimatorCtrl.get().newMove();
		ccMove.setDest("collimator", InOut.IN);
		ccMove.setDest("backlight", InOut.OUT);
		ccMove.setDest("beamstop", InOut.IN);
		ccMove.setDest("pindiode", InOut.OUT);
		MoveResult ccResult = ccMove.start().waitFor();

		if (!ccResult.isOK()) {
			Log.error("Screening : Collimator script failed , collection aborted");
			stopThread = true;
			return 0;
		}

		/* move collimator, backlight, beamstop */
		if (!Util.getBID().equals("23b")) {
			/* set up guardslits and collimator */
			Object[] setupRetVal = ModelBeamsize.setUpBeamsize(
					Epics.getDouble(PV_SCR_RUN.beamWidth_um.run(0)),
					Epics.getDouble(PV_SCR_RUN.beamHeight_um.run(0)), false);
			origCollimator = (CollimatorSize) setupRetVal[0];
			if (null != setupRetVal[1])
				origGSSize = (Point2D) setupRetVal[1];
			ActionResult setupResult = (ActionResult) setupRetVal[2];

			if (!setupResult.succeeded()) {
				Log.error("Beam size setup error Beam size setup failed, collection aborted.");
				return 0;
			}
		}

		/* wait for detector here to make detector move/ taking jpeg in parallel */
		if(detectorMove != null) {
			MoveResult detectorResult = detectorMove.waitFor();
			if (!detectorResult.isOK())
			{
				Log.error("Detector move to "+Util.sprintf("%.3f",distance)+" failed, collection aborted.");
				stopThread = true;
				return 0;
			}
		}

		if (stopThread || ModelScreening.pauseCollect) {
			ModelScreening.pauseCollect = false;
			Log.note("paused");
			return 0;
		}
		// if(stopThread || ModelScreening.pauseCollect) {
		// waitForResume("pause");
		// Epics.putInteger(PV_SCR.ScreeningStatus, 2);
		// }

		// set the detector binning mode
		final BinMode reqBinMode = ModelCollectRunList.get().getBinMode();
		ModeChange binModeChange = DevBinMode.get().startChange(reqBinMode);
		// wait for the binning mode change
		ActionResult binModeResult;
		try {
			binModeResult = binModeChange.waitFor();
		} catch (Exception e) {
			e.printStackTrace();
			binModeResult = ActionResult.FAILURE;
		}
		if (!binModeResult.succeeded()) {
			Log.error("Changing the CCD binning mode to " + reqBinMode
					+ " failed, collection aborted.");
			final boolean[] done = new boolean[1];
			GUI.asyncExec(new Runnable() {
				public void run() {

					NBMessageBox mb = new NBMessageBox(GUI.getShell(), SWT.OK
							| SWT.ICON_ERROR);
					mb.setText("CCD Bin Mode Error");
					mb.setMessage("Changing the CCD binning mode to "
							+ reqBinMode
							+ " failed, collection aborted.\n\n"
							+ "Please close all windows on the MAR computer and then click the MAR icon to\n"
							+ "restart FPE and the MAR GUI.  If the problem persists, restart JBluIce as well.");
					int result = mb.open();
					done[0] = true;

				}
			});
			while (!done[0]) {
				try {
					Thread.sleep(200);
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
			stopThread = true;
			return 0;
		}

		// set attenuation
		Log.note("Setting attenuation to " + getAttenuation_factors());
		MoveCmd attenMove = DevAttenuation.get().startMove(
				getAttenuation_factors());
		MoveResult attenResult = attenMove.waitFor();
		if (!attenResult.isOK()) {
			Log.error("Attenuation move to "
					+ Util.printDouble(getAttenuation_factors(), 4)
					+ " (factors) failed, aborting screening");
			stopThread = true;
			return 0;
		}
		
		Log.note("Noofimages : " + getNoofImages(angleIndex));
		int isMultiStrategy = Epics.getInteger(PV_STRATEGY.showMultiStrategy);
		/* clear multi crystal strategy images field */
		if (noofImagesCollected == 1 && isMultiStrategy == 0) {
			ModelReferenceData.IMAGES = "";
			DevReferenceData.get().setResult(true);
		}

		boolean singleArmMode = true;
		for (int img = 0; img < getNoofImages(angleIndex); img++) {

			double angle1 = getAngle(angleIndex) + img * delta;

			int frameNumber1 = getFrameNumber(angleIndex, angle1, img);
			Log.note("Angle for collect: " + angle1 + ", Framenumber=" + frameNumber1);

			//int angleInt1 = (int) angle1 + 1;

			/* set filename */
			filename = getFilename(frameNumber1);

			// update frame names to multi strategy tab
			if (isMultiStrategy == 0) {
				ModelReferenceData.IMAGES += collectDir + "/" + filename + " ";
				DevReferenceData.get().setResult(true);
			}

			/* Move gonio to the angle selected */
			// Epics.put(PV_SCR.CurrentTask,sample+ " Rotate"+angleInt1);
			// Log.note("FileName : " +filename);
			Epics.put(PV_SCR.CurrentTask, sample + " Collect" + frameNumber1);
			fileExists = 0;

			// alternate fix with version number at the 2nd position, 
			// it will be an issue 
			// if the user were to mount the same crystal several times, 
			// but choose different sets of images to collect.
			//
			// NB: Ideally it is best to check whether the crystal was mounted/screened
			// previously before the mount operation, and update prefix accordingly,
			// otherwise, it will be too late to properly fix it here
			File collectFile = new File(getFPEDir() + "/" + filename);
			Integer iTry = 0;
			while (collectFile.exists() && iTry < 10 ) {
				String prefix = Epics.getString(PV_SCR_RUN.fileroot.run(0));
				String newPrefix;
				Integer pos = prefix.lastIndexOf("_");

				if (pos == -1) {
					newPrefix = prefix + "_1";
				} else {
					Integer group;
					try {
                                                //Log.note("DEBUG: substring = " + prefix.substring(pos));
						Integer oldGroup = Integer.valueOf(prefix.substring(pos+1));
						group = oldGroup + 1;
					} catch (NumberFormatException e) {
						group = 100;
					}
					newPrefix = prefix.substring(0, pos) + "_" + group;
				}
				
				filename = getFilenameReal(frameNumber1, newPrefix);
				Log.warn("image file already exists, new file will be named as " + filename);
				
				// update PV so that next image will get an updated file root
				Epics.put(PV_RUN.fileroot.run(0), newPrefix);
				Epics.put(PV_SCR_RUN.fileroot.run(0), newPrefix);
				Epics.put(PV_SCR.prefix, newPrefix);
				WdgScreeningTasks.updateRuns();

				// update not working
				//Epics.putInteger(PV_SCR_RUN.updated.run(0), 0);
				//Epics.putInteger(PV_SCR_RUN.updated.run(0), 1);
				// synchronization check, if put wasn't successful, issue a warning
				try {
					Thread.sleep(100);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
				if (!newPrefix.equals(Epics.getString(PV_SCR_RUN.fileroot.run(0)))) {
					Log.warn("new fileroot PV sent but not set yet: "+newPrefix);
				} else {
					Log.note("new fileroot PV was set as: "+newPrefix);
				}
				// check again
				collectFile = new File(getFPEDir() + "/" + filename);
				iTry++;
				
			}
			
			if (angleIndex == 0 && img == 0) {
				image1 = filename;
				snapImage1 = filename + "_cam1.jpg";
			}
			if (angleIndex == 1 && img == 0) {
				image2 = filename;
				snapImage2 = filename + "_cam1.jpg";
			}
			if (angleIndex == 2 && img == 0) {
				image3 = filename;
				snapImage3 = filename + "_cam1.jpg";
			}

			/* This should before collecting image */
			Epics.putDouble(PV_SCR_RUN.startAngle.run(0), angle1);
			Epics.put(PV_SCR_RUN.runStatus.run(0), "Collecting");

			// save motion for standard mode (this may cause motion since rqspos may not equal actpos)
			Point3D currentPos = RasterGeometry.getCurrentMotorPositions().to3D();
			XYZMotion standardMotion = new XYZMotion(currentPos,currentPos);

			/* Collect image */
			String fpeFilename = filename;
			if (filename.endsWith("_master.h5")) {
				int masterIndex = fpeFilename.lastIndexOf("_master.h5");
				if (masterIndex>=0) fpeFilename = fpeFilename.substring(0, masterIndex);
			}

			// TODO: File naming in screening needs review, this block should be temporary
            String prefix = fpeFilename;
			try {
				int suffixesToRemove = 0;
				if (fpeFilename.endsWith("_master.h5")) {
					suffixesToRemove = 2;
				} else if (fpeFilename.endsWith(".cbf")) {
					suffixesToRemove = 1;
				}
				for (int i = 0; i < suffixesToRemove; i++) {
					int lastUnderscoreIndex = prefix.lastIndexOf('_');
					if (lastUnderscoreIndex >= 0)
						prefix = prefix.substring(0, lastUnderscoreIndex);
				}
			} catch (Exception e) {
				e.printStackTrace();
			}
			if (fpeFilename.endsWith("_master.h5")) {
				prefix += "_scr";
			}

			DevFPE.FrameList fpeFrame = DevFPE.get().newFrameList();
			fpeFrame.delta = delta;
			fpeFrame.energy = DevPmacMotor.get("mo_En_E").getPosition();
			fpeFrame.exposureTime = exposureTime;
			fpeFrame.prefix = prefix+"_";
			fpeFrame.firstFrameIndex = frameNumber;
			fpeFrame.firstFrameLabel = frameNumber1;
			fpeFrame.runNumber = 0;
			fpeFrame.filename = fpeFilename;
			fpeFrame.directory = getFPEDir();
			fpeFrame.axis = Axis.GONIO_PHI;
			fpeFrame.shutteredMotion = standardMotion;
			fpeFrame.shutterlessMotion = standardMotion;
			
			if (singleArmMode) { // single arm mode to collect all images
				fpeFrame.fpeMode = DevFPE.FPEMode.SHUTTERLESS_ARM_DISARM;
				fpeFrame.startAngle = getAngle(angleIndex);
				fpeFrame.nextAngle = getAngle(angleIndex) + delta;
				fpeFrame.endAngle = getAngle(angleIndex) + getNoofImages(angleIndex) * delta;
				fpeFrame.frameCount = getNoofImages(angleIndex);
				// fpeFrame.shutterlessMotion = 
			} else { // collect one image at a time
				fpeFrame.startAngle = angle1;
				fpeFrame.nextAngle = angle1 + delta;
				fpeFrame.endAngle = angle1 + delta;
			}
			fpeFrame.start().waitFor();

			// update image viewer
			if (DetectorConfig.isEiger() && fpeFilename.endsWith("_master.h5")) {
				Matcher matcher = Filenames.HDF5MasterFilePattern.matcher(fpeFilename+"_master.h5");
				if (matcher.matches())
				{
					String prefix2 = matcher.group(1);
					int runLabel = Integer.parseInt(matcher.group(2));
					int frameLabel = Integer.parseInt(matcher.group(3));
					String cbfFilename = String.format
							(
									"%s_%d_scr_%06d.cbf",
									prefix2,
									runLabel,
									frameLabel
									);

					DevFPE.get().setImageViewerFilePath(collectDir, cbfFilename);
				}
			} else {
				DevFPE.get().setImageViewerFilePath(collectDir, fpeFilename);
			}

			if (fpeFrame.getResult() == FrameListResult.FAILED) {
				stopThread = true;
				ModelScreening.pauseCollect = false;
				return 1;
			}
			if (stopThread || ModelScreening.pauseCollect) {
				ModelScreening.pauseCollect = false;
				// Epics.put(PV_SCR.CurrentTask,"Recollect "+6);
				break;
			}
			
			if (singleArmMode) {
				break;
			}
		}
		//ringInt = DevSlider.get("ring").getLowerLimit();
		//DevSlider.get("ring").setIntensity(ringInt);
		Epics.put(PV_SCR_RUN.runStatus.run(0), "Complete");
		return 0;
	}

	public int getRepeatFlag(int repeatFlag) {
		if (manualRepeatFlag == -1)
			return repeatFlag;
		else
			return manualRepeatFlag;
	}

	/**
	 * Get the starting position according to the current task
	 * @return
	 */
	public int getStartingFlag() {

		int angle1 = Epics.getInteger(PV_SCR.AngleImage0);
		int angle2 = Epics.getInteger(PV_SCR.AngleImage45);
		int angle3 = Epics.getInteger(PV_SCR.AngleImage90);
		int task = 0;

		String currentStatus1 = Epics.getString(PV_SCR.CurrentTask);
		if (currentStatus1.equals(""))
			return 0;
		int space = currentStatus1.indexOf(' ');

		String currentStatus = currentStatus1.substring(space + 1);
		if (currentStatus.equals("MountNextXtal")) {

			task = 0;
		} else if (currentStatus.equals("CenterXtal")) {

			task = 1;
		} else if (currentStatus.equals("Paused4Center")) {
			Log.note("Paused4resume");
			task = 3;
		} else if (currentStatus.contains("JPEGsnapshot")) {

			task = 3;
		}else if (currentStatus.equals("Collect" + angle1)
				|| currentStatus.equals("Rotate" + angle1)) {

			task = 4;
		} else if (currentStatus.equals("Collect2" + angle2)
				|| currentStatus.equals("Rotate" + angle2)) {

			task = 5;
		} else if (currentStatus.equals("Collect3" + angle3)
				|| currentStatus.equals("Rotate" + angle3)) {

			task = 6;
		} else if (currentStatus.equals("Paused4Inspection")) {

			task = 0;
		} else if (currentStatus1.contains("Recollect")) {
			StringTokenizer str = new StringTokenizer(currentStatus1);
			str.nextToken();
			task = Integer.parseInt(str.nextToken());
		} else {

			// Log.note("returned4resume");
			String Pin = Epics.getString(PV_ROBOT.mounted);

			if (Pin.equals("none"))
				return 0;
			else
				return 2;
		}
		if (manualRepeatFlag == -1)
			return task;
		else
			return manualRepeatFlag;
	}

	@Override
	public void run() {
		group = 1;
		int returnFlag = 0;
		int index;

		int manualCrystalIndex = 0;
		_requestStop = false;
		stopThread = false;
		boolean singleSampleRepeat = false;

		System.out.println("No. of crystals " + crystalList.getNoofCrystals());
		if (!Client.isMaster()) {
			stopThread = true;
			return;
		}

		Log.note("ScreeningStarted");
		repeatFlag = getStartingFlag();
		String moun = Epics.getString(PV_ROBOT.mounted);

		while (crystalList.getNoofCrystals() > 0 || singleSampleRepeat == true || !moun.equals("none")) {

			if (stopThread)
				return;

			image1 = "";
			image2 = "";
			image3 = "";

			int cur = -1;
			/* the sample will be removed from List as soon as it mounted */
			if (repeatFlag != 0) {
				moun = Epics.getString(PV_ROBOT.mounted);
				if (moun.equals("none")) {
					repeatFlag = 0;
					cur = Integer.parseInt(crystalList.getElement(0));
				} else {
					cur = CmdScreening.labelToindex(moun);
				}
			} else {
				group = 1;

				if (crystalList.getElement(0) != null) {
					cur = Integer.parseInt(crystalList.getElement(0));
					Log.note("Repeat" + cur);
				}
			}

			/* for manual selected sample 
			 * 
			 * Sudhir: This block of code is confusing for myself 
			 * what is manual selected sample?
			 * make a comment inside if statement and test if it ever come into this block
			 * check with Naga, if he has any explanation
			 * 
			 * */
			if (ModelScreening.crystalStartIndex > 0) {
				manualCrystalIndex = ModelScreening.crystalStartIndex;
				index = crystalList.getIndex(Integer.toString(manualCrystalIndex));

				if (index < crystalList.getNoofCrystals() && index != -1) {

					if (manualCrystalIndex != -1 && manualCrystalIndex != 0) {

						if (repeatFlag == 0) {
							group = 1;
							cur = manualCrystalIndex;

							System.out.println("manualCrystalIndex " + manualCrystalIndex + " " + index + 1);
							try {
								if (index + 1 < crystalList.getNoofCrystals())
									ModelScreening.crystalStartIndex = Integer.parseInt(crystalList.getElement(index + 1));
							} catch (Exception e) {

							}
						}
					}
				} else
					ModelScreening.crystalStartIndex = 0;
			}

			if (cur > -1)
				returnFlag = screenCrystal(cur, repeatFlag);
			else {
				finishScreening();
				return;
			}
			if (returnFlag == 7) {
				returnFlag = screenCrystal(cur, 7);
			}
			if (returnFlag == 1) {
				_requestStop = true;
				return;
			} else if (returnFlag == 2) {
				_requestStop = true;
				return;
			} else if (returnFlag == 3) {
				_requestStop = true;
				return;
			} else if (returnFlag == 4) {
				stopThread = true;
				if (Client.isMaster())
					Epics.putInteger(PV_SCR.ScreeningStatus, 0);
				_requestStop = true;
				return;
			}

			if (manualRepeatFlag == -1)
				repeatFlag = 0;
			else
				repeatFlag = manualRepeatFlag;
			moun = Epics.getString(PV_ROBOT.mounted);

			if (!moun.equals("none")) {
				if (manualRepeatFlag == -1)
					singleSampleRepeat = false;
				else
					singleSampleRepeat = true;
			}
		}

		finishScreening();
		_requestStop = true;
		Log.note("ScreeningFinished");

		return;
	}

	public void finishScreening() {
		if (!Client.isMaster()) {
			stopThread = true;
			return;
		}

		Epics.put(PV_SCR.CurrentTask, "Completed");
		String mounted = Epics.getString(PV_ROBOT.mounted);

		if (!mounted.equals("none")) {
			Epics.put(PV_SCR.CurrentTask, "UnloadLastSample");
			CmdRobot cmdRobot = new CmdRobot("unload");
			if (!cmdRobot.unload()) {
				PVName mountedPV = PV_ROBOT.mounted;
				String mount = Epics.getString(mountedPV);
				Epics.put(PV_SCR.CurrentTask, mount + " UnloadFailed");
			} else {
				Epics.put(PV_SCR.CurrentTask, "Unloaded");
			}
		}

		if (Client.isMaster())
			Epics.putInteger(PV_SCR.ScreeningStatus, 0);

	}

	public int screenCrystal(int sample, int repeatFlag) {

		String pin = ModelScreening.indexToLabel(sample);
		String mounted = "";
		noofImagesCollected = 0;
		updateActions();
		if (!Client.isMaster()) {
			stopThread = true;
			return 4;
		}

		switch (repeatFlag) {
		case 0:
			if (mount) {
				/*
				 * load returns 0, when unload current mounted sample failed
				 * returns 1, when load failed 
				 * returns 2, success 
				 * returns 3, when robot not Ready 
				 * returns 4, already mounted
				 */
				manualRepeatFlag = -1;

				CmdRobot cmdRobot = new CmdRobot();

				int success = cmdRobot.load(sample,
						ModelScreening.getRobotServerFormat(sample));

				if (success == 0) {
					PVName mountedPV = PV_ROBOT.mounted;
					String mount = Epics.getString(mountedPV);

					if (mount.equals("none"))
						return (1);
					int inde = labelToindex(mount);
					mount = ModelScreening.indexToLabel(inde);
					Epics.put(PV_SCR.CurrentTask, mount + " UnloadFailed");
					Epics.putInteger(PV_SCR.ScreeningStatus, 0);
					Log.note("unload failed");

					/* write GM/CA database */
					ModelScreening.robotSuccess = false;
					ScreeningEvents.writeEvent(scr);
					return (1);
				}
				if (success == 1) {
					Epics.put(PV_SCR.CurrentTask, pin + " SampleLoadFailed");
					Util.sleep(10);
					Epics.putInteger(PV_SCR.ScreeningStatus, 0);
					Util.sleep(10);
					Log.note("load failed");

					/* write GM/CA database */
					ModelScreening.robotSuccess = false;
					ScreeningEvents.writeEvent(scr);

					return (2);
				}
				if (success == 2) {

					ModelReferenceData.IMAGES = "";
					DevReferenceData.get().setResult(true);
				}
				if (success == 3) {

					Epics.put(PV_SCR.CurrentTask, "Robot not ready");

					Epics.putInteger(PV_SCR.ScreeningStatus, 0);
					return (3);
				}
				if (success == 4) {
					mounted = Epics.getString(PV_ROBOT.mounted);
					Log.note("already mounted");
					Epics.put(PV_ROBOT.mounted, "none");
					Epics.put(PV_ROBOT.mounted, mounted);
				}
			}
			repeatFlag++;
			updateActions();
			repeatFlag = getRepeatFlag(repeatFlag);

			/* write GM/CA database : robot success */
			ModelScreening.robotError = "";
			ScreeningEvents.writeEvent(scr);

			if (stopThread)
				return 4;
			if (repeatFlag <= manualRepeatFlag)
				break;
		case 1:
			if (autoCenter) {
				manualRepeatFlag = -1;
				Epics.put(PV_SCR.CurrentTask,
						ModelScreening.indexToLabel(sample) + " CenterXtal");
				Log.note(pin + " AutoCenter selected");
				loopCenter = Epics.getInteger(PV_SCR.centerType);
				if (loopCenter == 0)
					DevAutocenter.get().syncExecLoop();
				else if (loopCenter == 1)
					DevAutocenter.get().syncExecCrystal();
				else if (loopCenter == 2) {
					Epics.putInteger(
							PV_R_RUN.autoFind.run(RasterListType.AUTO, 0), 1);
					ArrayList<Site> allSites = CmdRasterAuto.runUnattended();

					/*
					 * center the first site TODO: should automatically index
					 * all sites
					 */
					if (allSites.size() > 0 && allSites != null) {
						Site e = allSites.get(0);
						DevXYZ.syncMove(e.motorPos.to3D(), false);
					}
				} else
					Log.error("wrong option for autocentering in screening tab");
			} else {
				manualRepeatFlag = -1;
			}

			repeatFlag++;
			updateActions();
			repeatFlag = getRepeatFlag(repeatFlag);

			if (stopThread)
				return 4;
			if (repeatFlag <= manualRepeatFlag)
				break;
		case 2:
			if (pause1) {
				manualRepeatFlag = -1;
				Epics.putInteger(PV_SCR.ScreeningStatus, 2);
				Epics.put(PV_SCR.CurrentTask,
						ModelScreening.indexToLabel(sample) + " Paused4Center");
				waitForResume("pause1");
			}

			repeatFlag++;
			updateActions();
			repeatFlag = getRepeatFlag(repeatFlag);

			if (stopThread)
				return 4;
			if (repeatFlag <= manualRepeatFlag)
				break;
		case 3:

			mounted = Epics.getString(PV_ROBOT.mounted);
			String collectDir = getDir();

			if (!mounted.equals("none")) {

				/* Taking 3 snapshots of the crystal */
				int isJpeg = Epics.getInteger(PV_SCR.takeJpeg);
				if (isJpeg == 1)
					WdgScreeningTasks.takeJpeg = true;
				else
					WdgScreeningTasks.takeJpeg = false;

				while (DevCollimatorCtrl.get().isRunning()) {
					Util.sleep(2000);
					Log.warn("Collimator script still running");
				}

				if (WdgScreeningTasks.takeJpeg) {

					/* collimator move */

					Log.note("DebugPindiode Movement CI_NOBEAMSTOP falgs CI_REVERSE_NOBEAMSTOP");
					CCMove ccMove = DevCollimatorCtrl.get().newMove();
					ccMove.setArguments("OUT backlightIN bluice");
					MoveResult ccResult = ccMove.start().waitFor();

					if (!ccResult.isOK()) {
						Log.error("Screening : Collimator script failed , collection aborted");
						stopThread = true;
						return 4;
					}

					/* create directory if does's exist */
					File dirFile = new File(collectDir);
					if (!dirFile.exists()) {

						System.out.println("screenCrystal mkdir "+collectDir);
						if (!dirFile.mkdirs()) {
							Log.error("Couldn't create directory " + collectDir);
						}
					}
					/* if diffraction images selected */
					if(!collect1 && !collect2 && !collect3) {
						/* take image in the current angle */
						double current_angle = Epics.getDouble(PV_PMAC.ActPos.motor("gonio_omega"));
						takeJpegSnapshot(current_angle, pin, collectDir);

						/* take image in the current angle */
						takeJpegSnapshot(current_angle+90.0, pin, collectDir);
					}
					if (collect1) {
						double angle_0 = getAngle(0);
						takeJpegSnapshot(angle_0, pin, collectDir);
					}
					if (collect2) {

						double angle_1 = getAngle(1);
						takeJpegSnapshot(angle_1, pin, collectDir);
					}
					if (collect3) {

						double angle_2 = getAngle(2);
						takeJpegSnapshot(angle_2, pin, collectDir);
					}
				}
			} else {
				manualRepeatFlag = -1;
				Log.note("No sample mounted. skipping collection");
				return 6;
			}

			repeatFlag++;
			updateActions();
			repeatFlag = getRepeatFlag(repeatFlag);

			if (stopThread)
				return 4;
		case 4:
			mounted = Epics.getString(PV_ROBOT.mounted);
			if (!mounted.equals("none")) {
				if (collect1) {
					noofImagesCollected++;
					manualRepeatFlag = -1;
					Log.note(pin + " collect1 selected");
					fpeError = CollectImage(0,
							ModelScreening.indexToLabel(sample));
				}
			} else {
				manualRepeatFlag = -1;
				Log.note("No sample mounted. skipping collection");
				return 6;
			}

			repeatFlag++;
			updateActions();
			repeatFlag = getRepeatFlag(repeatFlag);

			if (ModelScreening.pauseCollect) {
				ModelScreening.pauseCollect = false;
				Epics.put(PV_SCR.CurrentTask, "Recollect " + 3);
				Epics.putInteger(PV_SCR.ScreeningStatus, 2);
				waitForResume("pause");
				return 0;
			}

			if (stopThread)
				return 4;

		case 5:
			mounted = Epics.getString(PV_ROBOT.mounted);
			if (!mounted.equals("none")) {
				if (collect2) {
					noofImagesCollected++;
					manualRepeatFlag = -1;
					Log.note(pin + " collect2 selected");
					fpeError = CollectImage(1,
							ModelScreening.indexToLabel(sample));

				}
			} else {
				manualRepeatFlag = -1;
				Log.note("No sample mounted. skipping collection");
				return 6;
			}

			repeatFlag++;
			updateActions();
			repeatFlag = getRepeatFlag(repeatFlag);

			if (ModelScreening.pauseCollect) {
				ModelScreening.pauseCollect = false;
				Epics.put(PV_SCR.CurrentTask, "Recollect " + 4);
				Epics.putInteger(PV_SCR.ScreeningStatus, 2);
				waitForResume("pause");
				return 0;
			}

			if (stopThread)
				return 4;
		case 6:
			mounted = Epics.getString(PV_ROBOT.mounted);
			if (!mounted.equals("none")) {
				if (collect3) {
					noofImagesCollected++;
					manualRepeatFlag = -1;
					Log.note(pin + " collect3 selected");
					fpeError = CollectImage(2,
							ModelScreening.indexToLabel(sample));

				}
			} else {
				manualRepeatFlag = -1;
				Log.note("No sample mounted. skipping collection");
				return 6;
			}

			repeatFlag++;
			updateActions();
			repeatFlag = getRepeatFlag(repeatFlag);

			if (ModelScreening.pauseCollect) {
				ModelScreening.pauseCollect = false;
				Epics.put(PV_SCR.CurrentTask, "Recollect " + 5);
				Epics.putInteger(PV_SCR.ScreeningStatus, 2);
				waitForResume("pause");
				return 0;
			}

			if (stopThread)
				return 4;

		case 7:
			if (pause2) {
				manualRepeatFlag = -1;

				/* collimator move */
				Log.note("DebugPindiode Movement CI_NOBEAMSTOP falgs CI_REVERSE_NOBEAMSTOP");

				CCMove ccMove = DevCollimatorCtrl.get().newMove();
				ccMove.setArguments("OUT backlightIN bluice");
				ccMove.start().waitFor();

				Epics.putInteger(PV_SCR.ScreeningStatus, 2);
				Epics.put(PV_SCR.CurrentTask,
						ModelScreening.indexToLabel(sample)
						+ " Paused4Inspection");
				Log.note(pin + " paused for inspection");
				waitForResume("pause2");
			}
			repeatFlag++;
			updateActions();
			repeatFlag = getRepeatFlag(repeatFlag);

			ModelScreening.pauseCollect = false;

			if (stopThread)
				return 4;
			if (repeatFlag <= manualRepeatFlag)
				break;
		}
		if (Client.isMaster())
			Epics.put(PV_SCR.CurrentTask, ModelScreening.indexToLabel(sample)
					+ " Paused4Inspection");
		Util.sleep(100);
		return (5);
	}

	public void waitForResume(String pause) {

		Util.sleep(50);
		int screeningStatus = Epics.getInteger(PV_SCR.ScreeningStatus);

		while (screeningStatus != 1) {
			if (stopThread)
				return;
			screeningStatus = Epics.getInteger(PV_SCR.ScreeningStatus);
			Util.sleep(50);
		}
	}
}
