package gov.anl.gmca.server;

import gov.anl.gmca.collect.CmdCollectJet;
import gov.anl.gmca.collect.DevFPE;
import gov.anl.gmca.collect.ModelCollectRunList;
import gov.anl.gmca.common.*;
import gov.anl.gmca.common.client_server.DeviceMap;
import gov.anl.gmca.common.client_server.StateMap;
import gov.anl.gmca.common.client_server.StateMap.MotorCommands;
import gov.anl.gmca.common.client_server.StateMap.StateMotor;
import gov.anl.gmca.common.client_server.StateMap.WorkerState;
import gov.anl.gmca.hutch.*;
import gov.anl.gmca.hutch.DevAlignMinibeam.AlignData;
import gov.anl.gmca.hutch.DevAttenuation.*;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.PVLookup.*;
import gov.anl.gmca.server.messaging.MessageQueue;
import redis.clients.jedis.Jedis;

import java.util.*;

public class StateMapUpdate
{
	private static Timer redisTimer;

	public static void init()
	{
		redisTimer = new Timer("Redis update task");
		redisTimer.schedule(new TimerTask(){public void run(){updateRedis();}}, 1000);
	}

	// Update Redis based on server state and EPICS values
	public static void updateRedis()
	{
		try (Jedis jedis = Redis.get()) {

			// HR scale
			jedis.hset(Redis.config, "mm_per_px_hr_h", String.format("%.8f", Epics.getDouble(PV_BLUICE.px2mm_h_hi$VAL)));
			jedis.hset(Redis.config, "mm_per_px_hr_v", String.format("%.8f", Epics.getDouble(PV_BLUICE.px2mm_v_hi$VAL)));

			// collect
			String collectState = ModelCollectRunList.CollectState.values()[Epics.getInteger(PV_RUNS.collectState)].toString();
			jedis.hset(Redis.collect_state, "data_dir", DevDataDir.getCBF());
			jedis.hset(Redis.collect_state, "state", collectState);

			// FPE
			DevFPE.FPEState fs = DevFPE.get().getFPEState();
			String stateString;
			if (null==fs) stateString="OFF";
			else if (fs==DevFPE.FPEState.idle) stateString="IDLE";
			else stateString = "RUNNING";
			jedis.hset(Redis.server_fpe, "state", stateString);

			// PMAC motors
			for (String motorId : DeviceMap.pmacMotorMap.keySet()) {
				DevPmacMotor dev = DeviceMap.pmacMotorMap.get(motorId);
				Limits limits = dev.getLimits();

				String key = Redis.hutch_motor(motorId);
				jedis.hset(key, "state", dev.getState().toString());
				jedis.hset(key, "actPos", String.format("%.4f", dev.getStablePosition()));
				jedis.hset(key, "rqsPos", String.format("%.4f", dev.getRqsPos()));
				jedis.hset(key, "inPos", dev.getInPos() ? "1" : "0");
				jedis.hset(key, "status", "");
				jedis.hset(key, "lowerLimit", limits == null ? "" : String.format("%.4f", limits.getLower()));
				jedis.hset(key, "upperLimit", limits == null ? "" : String.format("%.4f", limits.getUpper()));
				jedis.hset(key, "inPosBand", String.format("%.4f", dev.getInPosBand()));
				jedis.hset(key, "units", dev.getUnits());
				jedis.hset(key, "components", "");
			}

			// energy
			DevEnergy devEnergy = DevEnergy.getDevice();
			String key = Redis.hutch_motor("energy");
			Limits limits = devEnergy.getLimits();
			jedis.hset(key, "state", 		devEnergy.getState().toString());
			jedis.hset(key, "actPos", 		String.format("%.4f",devEnergy.getPosition()));
			jedis.hset(key, "rqsPos", 		String.format("%.4f",devEnergy.getRqsPos()));
			jedis.hset(key, "inPos", 		devEnergy.getMoving()?"1":"0");
			jedis.hset(key, "status", 		"");
			jedis.hset(key, "lowerLimit",	limits == null ? "" : String.format("%.4f",limits.getLower()));
			jedis.hset(key, "upperLimit", 	limits == null ? "" : String.format("%.4f",limits.getUpper()));
			jedis.hset(key, "inPosBand", 	String.format("%.4f",devEnergy.getInPosBand()));
			jedis.hset(key, "units", 		"keV");
			jedis.hset(key, "components", 	"");

			// attenuation
			DevAttenuation devAtten = DevAttenuation.get();
			key = Redis.hutch_motor("attenuation");
			limits = devAtten.getLimits();
			jedis.hset(key, "state", 		devAtten.getState().toString());
			jedis.hset(key, "actPos", 		String.format("%.4f",devAtten.getPosition()));
			jedis.hset(key, "rqsPos", 		String.format("%.4f",devAtten.getRqsPos()));
			jedis.hset(key, "inPos", 		devAtten.getMoving()?"1":"0");
			jedis.hset(key, "status", 		"");
			jedis.hset(key, "lowerLimit",	limits == null ? "" : String.format("%.4f",limits.getLower()));
			jedis.hset(key, "upperLimit", 	limits == null ? "" : String.format("%.4f",limits.getUpper()));
			jedis.hset(key, "inPosBand", 	String.format("%.4f",devAtten.getInPosBand()));
			jedis.hset(key, "units", 		"factors");
			jedis.hset(key, "components", 	"");

			// align minibeam
			DevAlignMinibeam devOpt = DevAlignMinibeam.get();
			jedis.hset(Redis.auto_beamAlign, "busy", devOpt.isRunning()?"1":"0");

			// app state (busy map)
			AppState.sendStateToRedis();

		} catch (Exception e) {
			e.printStackTrace();
		} finally {
			redisTimer.schedule(new TimerTask() {
				public void run() {
					updateRedis();
				}
			}, 100);
		}
	}

	public static StateMap createNewState(String clientID)
	{
		StateMap newState = new StateMap();

		try (Jedis jedis = Redis.get())
		{
			///////////////////////////////// workers ////////////////////////////////////

			WorkerState alignMiniBeam = new WorkerState();
			alignMiniBeam.name = "alignMiniBeam";
			alignMiniBeam.state = DevAlignMinibeam.get().isRunning() ? "RUNNING" : "IDLE";
			newState.workerStates.add(alignMiniBeam);

			////////////////////////////// image viewer /////////////////////////////////

			newState.imageViewer.fullDir = DevFPE.get().getImageViewerDir();
			newState.imageViewer.sampleDir = DevFPE.getPathAfterCommonDir(DevFPE.get().getImageViewerDir());
			newState.imageViewer.filename = DevFPE.get().getImageViewerFile().filename;
			newState.imageViewer.contrast = 10;
			newState.imageViewer.zoom = 1.0;
			newState.imageViewer.serverIndex = 1;
			newState.imageViewer.viewerIndex = 1;

			///////////////////////////////// collect ///////////////////////////////////

			Map<String, String> fpeState = jedis.hgetAll(Redis.server_fpe);
			DevFPE.FPEState fs = DevFPE.get().getFPEState();
			newState.fpe.state = null == fs ? "disconnected" : fs.toString();
			int nextFrame = 0;
			try { String strNextFrame = fpeState.get("nextFrame"); if (null!=strNextFrame && strNextFrame.length()>0) nextFrame = Integer.parseInt(strNextFrame); }
			catch (Exception e) { e.printStackTrace(); }
			newState.fpe.nextFrame = nextFrame;

			newState.collect.collectState = ModelCollectRunList.CollectState.values()[Epics.getInteger(PV_RUNS.collectState)].toString();
			newState.collect.dataDir = DevDataDir.getCBF();

			Map<String, String> jetConfig = jedis.hgetAll(Redis.collect_run(0));
			newState.jetCollect.state = CmdCollectJet.thread.isAlive() ? "collecting" : "inactive";
			newState.jetCollect.dir = jetConfig.get("rel_dir");
			newState.jetCollect.prefix = jetConfig.get("prefix");
			try {
				newState.jetCollect.expTime_sec = Double.parseDouble(jetConfig.get("expTime_sec"));
				newState.jetCollect.framesPerDir = Integer.parseInt(jetConfig.get("framesPerDir"));
				newState.jetCollect.smxProc = Integer.parseInt(jetConfig.get("smxProc"));
				newState.jetCollect.subDir = Integer.parseInt(jetConfig.get("subDir"));
			} catch (Exception e) {
				newState.jetCollect.expTime_sec = 0.0;
				newState.jetCollect.framesPerDir = 0;
				newState.jetCollect.smxProc = 0;
				newState.jetCollect.subDir = 0;
			}

			////////////////////////////////// motors ///////////////////////////////////

			for (String motorId : DeviceMap.pmacMotorMap.keySet()) {
				DevPmacMotor dev = DeviceMap.pmacMotorMap.get(motorId);
				StateMotor motorState = new StateMotor();

				// set large inPosBand for large (not operational) energies
				double inPosBand = dev.getInPosBand();
				if (dev.getDeviceName().equals("energy") && dev.getStablePosition()>50) {
					inPosBand = 10;
				}

				Limits limits = dev.getLimits();
				motorState.motorID = dev.getDeviceName();
				motorState.state = dev.getState();
				motorState.actPos = dev.getStablePosition();
				motorState.rqsPos = dev.getRqsPos();
				motorState.inPos = dev.getInPos();
				motorState.lowerLimit = limits == null ? null : limits.getLower();
				motorState.upperLimit = limits == null ? null : limits.getUpper();
				motorState.inPosBand = inPosBand;
				motorState.units = dev.getUnits();
				motorState.error = "";
				if (!dev.getHasMotionPrg()) motorState.error += " NO_MOTION_PRG";
				if (!dev.getHomed()) motorState.error += " NOT_HOMED";
				if (!dev.getAmpEnabled()) motorState.error += " AMP_DISABLED";
				motorState.error = motorState.error.trim();
				newState.motorList.put(motorId, motorState);
			}
			for (MotorCommands cmd : StateMap.MotorCommands.values()) {
				newState.motorCommands.add(cmd.name());
			}

			DevEnergy devEnergy = DevEnergy.getDevice();
			DevPmacMotor devMono = DeviceMap.pmacMotorMap.get("mo_En_E");
			StateMotor energyState = new StateMotor();
			energyState.motorID = "energy";
			energyState.state = devEnergy.getState();
			energyState.actPos = devEnergy.getPosition(); // TODO: add getStablePosition()?
			energyState.rqsPos = devEnergy.getRqsPos();
			energyState.inPos = devEnergy.getMoving();
			energyState.lowerLimit = devEnergy.getLimits().getLower();
			energyState.upperLimit = devEnergy.getLimits().getUpper();
			energyState.inPosBand = devEnergy.getInPosBand();
			energyState.units = "keV";
			energyState.error = ""; // TODO: add errors for undulator not connected or not writeable
			if (!devMono.getHasMotionPrg()) energyState.error += " NO_MOTION_PRG";
			if (!devMono.getHomed()) energyState.error += " NOT_HOMED";
			if (!devMono.getAmpEnabled()) energyState.error += " AMP_DISABLED";
			energyState.error = energyState.error.trim();
			newState.motorList.put(energyState.motorID, energyState);

			DevAttenuation devAtten = DevAttenuation.get();
			StateMotor attenState = new StateMotor();
			attenState.motorID = "attenuation";
			attenState.state = devAtten.getState();
			attenState.actPos = devAtten.getPosition();
			attenState.rqsPos = devAtten.getRqsPos();
			attenState.inPos = devAtten.getMoving();
			attenState.lowerLimit = devAtten.getLimits().getLower();
			attenState.upperLimit = devAtten.getLimits().getUpper();
			attenState.inPosBand = devAtten.getInPosBand();
			attenState.units = "factors";
			attenState.error = "";
			newState.motorList.put(attenState.motorID, attenState);

			///////////////////////////////// Optimize Beam //////////////////////////////////

			DevAlignMinibeam devOpt = DevAlignMinibeam.get();
			AlignData hData = devOpt.getDataH(), vData = devOpt.getDataV();
			Date lastRunTime = devOpt.getLastRunTime();
			long lastRunTimeInt = 0;
			if (null != lastRunTime) lastRunTimeInt = lastRunTime.getTime() / 1000L;

			newState.optimizeBeam.cmdRunning = devOpt.isRunning();
			newState.optimizeBeam.lastRunTime_unix = lastRunTimeInt;
			newState.optimizeBeam.lastEnergy_keV = devOpt.getLastEnergy();
			newState.kbm.xPos_V = Epics.getDouble(PV_MISC.horizontalPiezo);
			newState.kbm.yPos_V = Epics.getDouble(PV_MISC.verticalPiezo);
			newState.optimizeBeam.vGraph.pointCount = vData.pointCount;
			newState.optimizeBeam.vGraph.xPoints = vData.xPoints;
			newState.optimizeBeam.vGraph.yPoints = vData.yPoints;
			newState.optimizeBeam.vGraph.scanFlag = vData.scanFlag;
			newState.optimizeBeam.vGraph.vHigh = vData.vHigh;
			newState.optimizeBeam.vGraph.vLow = vData.vLow;
			newState.optimizeBeam.hGraph.pointCount = hData.pointCount;
			newState.optimizeBeam.hGraph.scanFlag = hData.scanFlag;
			newState.optimizeBeam.hGraph.vHigh = hData.vHigh;
			newState.optimizeBeam.hGraph.vLow = hData.vLow;
			newState.optimizeBeam.hGraph.xPoints = hData.xPoints;
			newState.optimizeBeam.hGraph.yPoints = hData.yPoints;

			//////////////////////////////////// Workers ////////////////////////////////////////

			newState.autoCenter.cmdRunning = DevAutocenter.get().isRunning();
			newState.autoCenter.mode = DevAutocenter.mode.toString();

			newState.enterHutch.cmdRunning = CmdEnterHutch.isRunning();

			/////////////////////////////////// client table ////////////////////////////////////

			newState.clientList = ClientTable.getClientList();

			newState.messages = MessageQueue.get(clientID);
		}
		catch (Exception e) { e.printStackTrace(); }

		return newState;
	}
	
	public static void update(String clientID)
	{
		StateMap.current = createNewState(clientID);
	}
}
