package gov.anl.gmca.server.messaging;

import gov.anl.gmca.server.ClientTable;
import gov.anl.gmca.server.HTTP;
import gov.anl.gmca.util.Util;
import spark.Request;
import spark.Response;

import java.util.ArrayList;
import java.util.StringTokenizer;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;

public class Popup
{
    // HTTP protocol:
    //
    // URL: POST to /popup
    // type=[question|note|progress]
    // text="Popup text\nwith multiple lines"
    //
    // For question:
    // options parameter: "comma,separated,options", e.g. "cancel,OK"
    // blocks for user response, returns -1 for error, otherwise option index, e.g. 1 for OK in above example
    //
    // For note:
    // No other parameters, returns immediately with 0 for delivered or -1 for error
    //
    // For progress:
    // wait_sec parameter: Number of seconds to show the progress bar
    // blocks for the wait time, returns 1 for delivered and user waited, 0 for user cancelled early, -1 for error
    //
    // curl example:
    //
    // curl -X POST -d "text=OK to move sample?" -d "options=OK,Cancel" \
    //      http://10.20.103.207:8001/popup
    //
    public static class PopupMessage
    {
        String popupID;
        String type, text;
        String[] options = {};
        Integer wait_sec = 0;
        Integer response = null;
    }

    // Keeps track of current popups
    public static ConcurrentHashMap<String,PopupMessage> popupMap = new ConcurrentHashMap<>();

    public static int syncPopup(PopupMessage message)
    {
        // look for master client
        String masterID = ClientTable.getMasterID();
        if (null==masterID) {
            System.out.println("ERROR: No master client found");
            return -1;
        }

        boolean typeOK = false;
        for (String t : new String[]{"question","note","progress"}){
            if (t.equals(message.type)) typeOK=true;
        }
        if (!typeOK) {
            System.out.println("ERROR: Invalid popup type: "+message.type);
            return -1;
        }

        // add question to master client queue
        popupMap.put(message.popupID, message);
        int timeoutMs = 10000;
        boolean messageAck = MessageQueue.syncSend(masterID, HTTP.gson.toJson(message), timeoutMs);
        if (!messageAck) return -1;

        // wait for response
        long waitStartMs = System.currentTimeMillis(), popupTimeoutMs = 3600*1000;
        while (true)
        {
            // check for response, return if found
            synchronized(popupMap) {
                Integer resp = popupMap.get(message.popupID).response;
                if (resp != null) {
                    popupMap.remove(message.popupID);
                    return resp;
                }
            }

            // if timeout reached, return error
            long now = System.currentTimeMillis();
            if (waitStartMs+popupTimeoutMs < now) {
                return -1;
            }

            Util.sleep(100);
        }
    }

    public static int syncPopup(String type, String text, String[] options, int wait_sec)
    {
        PopupMessage message = new PopupMessage();
        message.popupID = UUID.randomUUID().toString();
        message.type = type;
        message.text = text;
        message.options = options;
        message.wait_sec = wait_sec;
        return syncPopup(message);
    }

    public static byte[] handleRequest(Request req, Response res)
    {
        // parse query
        PopupMessage message = new PopupMessage();
        message.popupID = UUID.randomUUID().toString();
        message.type = req.queryParams("type");
        message.text = req.queryParams("text");
        if (req.queryParams("options")!=null) {
            ArrayList<String> options = new ArrayList<>();
            StringTokenizer st = new StringTokenizer(req.queryParams("options"),",");
            while (st.hasMoreTokens()) options.add(st.nextToken());
            message.options = options.toArray(new String[]{});
        }
        if (req.queryParams("wait_sec")!=null) {
            message.wait_sec = (int)Math.round(Double.parseDouble(req.queryParams("wait_sec")));
        }
        return Integer.toString(syncPopup(message)).getBytes();
    }

    // store the response from the client
    //
    public static byte[] handleResponse(Request req, Response resp)
    {
        String popupID = req.queryParams("popup_id");
        String popup_resp = req.queryParams("response");
        popupMap.get(popupID).response = Integer.parseInt(popup_resp);
        return "".getBytes();
    }
}
