package gov.anl.gmca.server.services;

import gov.anl.gmca.hutch.DetectorConfig;
import gov.anl.gmca.util.Point2D;
import spark.Request;
import spark.Response;

import javax.imageio.ImageIO;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.io.ByteArrayOutputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Formatter;
import java.util.Locale;

// This is a partial port of Resolution.tcl from SSRL BluIce 4.2
//
// Only rectangular detectors are supported
// Direct beam spot, detector partitions and beamstop are not drawn

public class ResPredictor
{
    protected double pixelsPerMM = Double.NaN;
    protected int canvasOriginX=0, canvasOriginY=0;
    protected int detectorWidthInPixels=1, detectorHeightInPixels=1;
    protected int detectorHalfWidthInPixels=1, detectorHalfHeightInPixels=1;
    protected double detectorXInMM=0.0, detectorYInMM=0.0; // default is centered
    protected int detectorXInPixels, detectorYInPixels;
    protected double detectorOffsetInMM;
    protected int detectorOffsetInPixels;
    protected double detectorOffsetAngle, beamOffsetAngle;
    protected final static int minRingGapInPixels = 1;
    protected final static int minRingSizeInPixels = 1;
    protected final static int maxRingSizeInPixels = 2000;

    protected int twoThetaOffsetInPixels = 0;

    protected double detectorWidthInMM, detectorHeightInMM;

    protected double detectorZ_mm = Double.NaN;
    protected double beamEnergy_eV = Double.NaN;
    protected double detector2t_deg = Double.NaN;

    protected int border, width, height;
    private static Font resTextFont = new Font("Helvetica", Font.BOLD, 15);
    protected Graphics2D g2;
    private Color bgColor, circleColor;

    private final static Color LIGHT_BLUE = new Color(173, 216, 230);
    private final static Color LIGHT_GRAY = new Color(235, 235, 235);

    public synchronized static byte[] handleRequest(Request req, Response res)
    {
        // parse arguments
        int imgWidthPx=Integer.parseInt(req.queryParams("width"));
        int imgHeightPx=Integer.parseInt(req.queryParams("height"));
        int borderPx = Integer.parseInt(req.queryParams("border"));
        double detector_z_mm = Double.parseDouble(req.queryParams("detector_z_mm"));
        double detector_2t_deg = Double.parseDouble(req.queryParams("detector_2t_deg"));
        double energy_eV = Double.parseDouble(req.queryParams("energy_eV"));

        // create temporary image
        BufferedImage img = new BufferedImage(imgWidthPx, imgHeightPx, BufferedImage.TYPE_INT_RGB);
        Graphics2D g2 = img.createGraphics();

        // draw
        ResPredictor rp = new ResPredictor(g2, imgWidthPx, imgHeightPx, borderPx, detector_z_mm, detector_2t_deg, energy_eV);
        rp.draw();

        // save to PNG and return
        try {
            ByteArrayOutputStream os = new ByteArrayOutputStream();
            ImageIO.write(img, "png", os);
            os.flush();
            byte[] ret = os.toByteArray();
            os.close();
            return ret;
        } catch (Exception e) {
            e.printStackTrace();
            return new byte[]{};
        }
    }

    public ResPredictor(Graphics2D g2, int imageWidthPx, int imageHeightPx, int borderPx, double det_z_mm, double det_2t_deg, double energy_eV)
    {
        this.g2 = g2;
        this.detectorZ_mm = det_z_mm;
        this.detector2t_deg = det_2t_deg;
        this.beamEnergy_eV = energy_eV;
        this.border = borderPx;
        this.width = imageWidthPx;
        this.height = imageHeightPx;
        circleColor = Color.RED;
        bgColor = LIGHT_BLUE;
    }

    public synchronized void draw()
    {
        g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        g2.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);

        Point2D detectorSize = DetectorConfig.getDimensionsMM();
        detectorWidthInMM = detectorSize.x;
        detectorHeightInMM = detectorSize.y;

        g2.setColor(bgColor);
        g2.fillRect(0,0,width,height);

        recalc2ThetaOffset();
        calcPixelsPerMM();
        calculateDetectorParameters();

        drawDetector();
        drawCrosshair();
        drawResolutionRings();
    }

    protected void recalc2ThetaOffset()
    {
        double detector2t_rad = detector2t_deg * Math.PI / 180.0;
        double twoThetaOffsetInMM = detectorZ_mm * Math.tan(detector2t_rad);
        detectorYInMM = twoThetaOffsetInMM;

        twoThetaOffsetInPixels = (int)Math.round(twoThetaOffsetInMM * pixelsPerMM);
    }

    protected void calcPixelsPerMM()
    {
        int conservativeSizeX = width - border;
        int conservativeSizeY = height - border;

        if (conservativeSizeX <= 0) return;
        if (conservativeSizeY <= 0) return;

        if (width > height)
        {
            pixelsPerMM = conservativeSizeY / detectorWidthInMM;
        }
        else
        {
            pixelsPerMM = conservativeSizeX / detectorHeightInMM;
        }

        canvasOriginX = width/2;
        canvasOriginY = height/2;
    }

    protected void calculateDetectorParameters()
    {
        ////////// DCS::DetectorShapeOverlay::calculateDetectorParameters ////////////

        // convert detector coordinates from mm to pixels
        detectorXInPixels = (int)(detectorXInMM * pixelsPerMM);
        detectorYInPixels = (int)(detectorYInMM * pixelsPerMM);

        // calculate detector offset distance in mm and pixels
        detectorOffsetInMM = Math.sqrt(detectorXInMM*detectorXInMM) + (detectorYInMM*detectorYInMM);
        detectorOffsetInPixels = (int)(detectorOffsetInMM * pixelsPerMM);

        // calculate angle of detector offset from beam and vice versa
        if ( detectorOffsetInMM > 1 )
        {
            detectorOffsetAngle = Math.atan2(detectorYInMM, detectorXInMM);
            beamOffsetAngle = Math.atan2(detectorYInMM, detectorXInMM) + Math.acos(-1);
        }
        else
        {
            detectorOffsetAngle = 0.0;
            beamOffsetAngle = 0.0;
        }

        ///////// DCS::DetectorFace::calculateDetectorParameters ///////////

        detectorWidthInPixels = (int)(pixelsPerMM * detectorWidthInMM);
        detectorHeightInPixels = (int)(pixelsPerMM * detectorHeightInMM);
        detectorHalfWidthInPixels = detectorWidthInPixels / 2;
        detectorHalfHeightInPixels = detectorHeightInPixels / 2;
    }

    // draw the detector face
    protected void drawDetector()
    {
        Rectangle face = new Rectangle(
                canvasOriginX - detectorWidthInPixels/2,
                canvasOriginY - detectorHeightInPixels/2,
                detectorWidthInPixels,
                detectorHeightInPixels);

        g2.setColor(Color.WHITE);
        g2.fillRect(face.x, face.y, face.width, face.height);

        g2.setColor(Color.BLACK);
        g2.drawRect(face.x, face.y, face.width, face.height);
    }

    protected void drawCrosshair()
    {
        g2.setColor(Color.BLACK);

        int crossHairY = canvasOriginY+twoThetaOffsetInPixels;

        // horizontal line
        if (crossHairY < canvasOriginY+detectorHeightInPixels/2)
        {
            g2.drawLine
            (
                canvasOriginX-detectorWidthInPixels/2,
                crossHairY,
                canvasOriginX+detectorWidthInPixels/2,
                crossHairY
            );
        }

        // vertical line
        g2.drawLine
        (
            canvasOriginX,
            canvasOriginY-detectorHeightInPixels/2,
            canvasOriginX,
            canvasOriginY+detectorHeightInPixels/2
        );
    }

    protected static class Edge implements Comparable<Edge>
    {
        enum Position { LEFT, TOP, RIGHT, BOTTOM };

        double radius;
        Edge.Position position;

        Edge(double _radius, Edge.Position _position) { radius = _radius; position = _position; }

        @Override
        public int compareTo(Edge rhs)
        {
            return Double.valueOf(radius).compareTo(rhs.radius);
        }
    }

    static class Ring
    {
        public double radius;
        int x,y;
        Ring(double _r, int _x, int _y) { radius=_r; x=_x; y=_y; }
        Ring() {}
    }

    protected Ring getCornerRing()
    {
        Ring ring = new Ring();

        // handle top left corner
        if ( detectorXInMM < 0 && detectorYInMM < 0 )
        {
            ring.x = canvasOriginX - detectorWidthInPixels/2;
            ring.y = canvasOriginY - detectorHeightInPixels/2;
        }

        // handle top right corner
        if ( detectorXInMM > 0 && detectorYInMM < 0 )
        {
            ring.x = canvasOriginX + detectorWidthInPixels/2;
            ring.y = canvasOriginY - detectorHeightInPixels/2;
        }

        // handle bottom left corner
        if ( detectorXInMM < 0 && detectorYInMM > 0 )
        {
            ring.x = canvasOriginX - detectorWidthInPixels/2;
            ring.y = canvasOriginY + detectorHeightInPixels/2;
        }

        // handle bottom right corner
        if ( detectorXInMM > 0 && detectorYInMM > 0 )
        {
            ring.x = canvasOriginX + detectorWidthInPixels/2;
            ring.y = canvasOriginY + detectorHeightInPixels/2;
        }

        double deltaX = ring.x - detectorXInPixels - canvasOriginX;
        double deltaY = ring.y - detectorYInPixels - canvasOriginY;
        ring.radius = (int)(Math.sqrt( (deltaX * deltaX) + (deltaY * deltaY) ) / pixelsPerMM);

        return ring;
    }

    protected ArrayList<Ring> getResolutionRings()
    {
        // calculate the four distances from the beam center to the detector edges
        ArrayList<Edge> edgeList = new ArrayList<Edge>();
        edgeList.add(new Edge(Math.abs(detectorXInMM + detectorWidthInMM / 2.0),  Edge.Position.LEFT));
        edgeList.add(new Edge(Math.abs(detectorYInMM + detectorHeightInMM / 2.0), Edge.Position.TOP));
        edgeList.add(new Edge(Math.abs(detectorXInMM - detectorWidthInMM / 2.0),  Edge.Position.RIGHT));
        edgeList.add(new Edge(Math.abs(detectorYInMM - detectorHeightInMM / 2.0), Edge.Position.BOTTOM));

        // sort in ascending order
        Collections.sort(edgeList);

        // initialize a pruned list of radii
        ArrayList<Ring> textVisibleList = new ArrayList<Ring>();

        // eliminate rings not tangent to actual edges of detector
        for (Edge edge : edgeList)
        {
            double radius = edge.radius;
            Edge.Position position = edge.position;

            RVResult ringVisibleResult = getResolutionTextCoordinates(position,radius);

            if (ringVisibleResult.ringVisible)
            {
                Ring ring = new Ring();
                ring.radius = radius;
                ring.x = ringVisibleResult.x;
                ring.y = ringVisibleResult.y;
                textVisibleList.add(ring);
            }
        }

        // calculate corner ring if no tangent rings
        if (0 == textVisibleList.size())
        {
            textVisibleList.add(getCornerRing());
        }

        double lastRadius = 0.0;

        // initialize a pruned list of radii
        ArrayList<Ring> ringVisibleList = new ArrayList<Ring>();

        // eliminate rings too close to the edge or to each other
        for (Ring ring : textVisibleList)
        {
            if (Math.abs(ring.radius - lastRadius) > minRingGapInPixels / pixelsPerMM)
            {
                ringVisibleList.add(ring);
                lastRadius = ring.radius;
            }
        }

        return ringVisibleList;
    }

    static class RVResult
    {
        public boolean ringVisible;
        public int x,y;
    }

    protected RVResult getResolutionTextCoordinates(Edge.Position position, double ringRadiusInMM)
    {
        RVResult result = new RVResult();

        switch (position)
        {
            case LEFT:
                result.x = canvasOriginX - detectorWidthInPixels / 2;
                result.y = canvasOriginY + detectorYInPixels;
                if (Math.abs(canvasOriginY - result.y) < detectorHeightInPixels / 2 )
                {
                    result.ringVisible = true;
                }
                break;
            case RIGHT:
                result.x = canvasOriginX + detectorWidthInPixels / 2;
                result.y = canvasOriginY + detectorYInPixels;
                if (Math.abs(canvasOriginY - result.y) < detectorHeightInPixels / 2)
                {
                    result.ringVisible = true;
                }
                break;
            case TOP:
                result.x = canvasOriginX + detectorXInPixels;
                result.y = canvasOriginY - detectorHeightInPixels / 2;
                if (Math.abs(canvasOriginX - result.x) < detectorWidthInPixels / 2)
                {
                    result.ringVisible = true;
                }
                break;
            case BOTTOM:
                result.x = canvasOriginX + detectorXInPixels;
                result.y = canvasOriginY + detectorHeightInPixels / 2;
                if ( Math.abs(canvasOriginX - result.x) < detectorWidthInPixels / 2 )
                {
                    result.ringVisible  = true;
                }
                break;
        }

        return result;
    }

    protected void drawResolutionRing(double ringRadiusInMM, boolean fill)
    {
        // get ring radius in units of pixels
        int ringRadiusInPixels = (int)(ringRadiusInMM * pixelsPerMM);

        // draw the ring if scaled radius is reasonable
        if ( ringRadiusInPixels > minRingSizeInPixels && ringRadiusInPixels < maxRingSizeInPixels )
        {
            if (fill)
            {
                g2.setColor(Color.DARK_GRAY);
                g2.setBackground(Color.RED);
                g2.fillOval
                (
                    canvasOriginX - ringRadiusInPixels + detectorXInPixels,
                    canvasOriginY - ringRadiusInPixels + detectorYInPixels,
                    ringRadiusInPixels*2,
                    ringRadiusInPixels*2
                );
            }
            else
            {
                g2.setColor(circleColor);
                g2.drawOval
                (
                    canvasOriginX - ringRadiusInPixels + detectorXInPixels,
                    canvasOriginY - ringRadiusInPixels + detectorYInPixels,
                    ringRadiusInPixels*2,
                    ringRadiusInPixels*2
                );
            }
        }
    }

    public static double calcLimit(double detectorZ_mm, double detector2t_deg, double beamEnergy_eV, double ringRadius_mm)
    {
        // added for GM/CA
        double detector2TInRad = 0.0174532925199 * detector2t_deg;

        // calculate the resolution of the ring from its radius
        double twoTheta = Math.atan(ringRadius_mm / detectorZ_mm);
        double wavelength = 12398.0 / beamEnergy_eV;
        double dSpacing = wavelength / (2.0*Math.sin(twoTheta/2.0 + detector2TInRad/2.0));

        return dSpacing;
    }

    protected void drawResolutionText(int textX, int textY, double ringRadiusInMM)
    {
        //System.out.println("DRT: z="+detectorZ_mm+" 2t="+detector2t_deg+" energy="+beamEnergy_eV);

        if (!Double.isNaN(detectorZ_mm) 	&&
                !Double.isNaN(beamEnergy_eV) 	&&
                !Double.isNaN(detector2t_deg))
        {
            double dSpacing = calcLimit(detectorZ_mm, detector2t_deg, beamEnergy_eV, ringRadiusInMM);

            // draw the resolution text on the canvas
            StringBuilder sb = new StringBuilder();
            Formatter formatter = new Formatter(sb, Locale.US);
            String label = formatter.format("%.2f", dSpacing).toString();

            g2.setFont(resTextFont);

            int stringWidth = g2.getFontMetrics().stringWidth(label);
            int stringHeight = g2.getFontMetrics().getHeight();
            int maxDescent = g2.getFontMetrics().getMaxDescent();
            Point center = new Point(textX - stringWidth/2, textY + stringHeight/2 - maxDescent);

            // draw a pseudo-border
            g2.setColor(LIGHT_GRAY);
            for (int i=-1; i<=1; i+=2)
                for (int j=-1; j<=1; j+=2)
                {
                    g2.drawString(label, center.x + i, center.y + j);
                }

            // draw the text
            g2.setColor(Color.BLACK);
            g2.drawString(label, center.x, center.y);
        }
    }

    protected void drawOppositeCornerArc(double largestRingRadius)
    {
        int x=0,y=0;

        // handle top left corner
        if ( detectorXInMM <= 0 && detectorYInMM <= 0 )
        {
            x = canvasOriginX + detectorWidthInPixels/2;
            y = canvasOriginY + detectorHeightInPixels/2;
        }

        // handle top right corner
        if ( detectorXInMM >= 0 && detectorYInMM <= 0 )
        {
            x = canvasOriginX - detectorWidthInPixels/2;
            y = canvasOriginY + detectorHeightInPixels/2;
        }

        // handle bottom left corner
        if ( detectorXInMM <= 0 && detectorYInMM >= 0 )
        {
            x = canvasOriginX + detectorWidthInPixels/2;
            y = canvasOriginY - detectorHeightInPixels/2;
        }

        // handle bottom right corner
        if ( detectorXInMM >= 0 && detectorYInMM >= 0 )
        {
            x = canvasOriginX - detectorWidthInPixels/2;
            y = canvasOriginY - detectorHeightInPixels/2;
        }

        int deltaX = x - detectorXInPixels - canvasOriginX;
        int deltaY = y - detectorYInPixels - canvasOriginY;
        int radius = (int)(Math.sqrt( (deltaX * deltaX) + (deltaY * deltaY) ) / pixelsPerMM);

        if ( radius - largestRingRadius > minRingGapInPixels )
        {
            double angle = Math.atan2(-deltaY, deltaX);
            drawResolutionArc(radius, angle);
            drawResolutionText(x, y, radius);
        }
    }

    protected void drawResolutionArc(double ringRadiusInMM, double angleAtMidpoint)
    {
        // get ring radius in units of pixels
        int ringRadiusInPixels = (int)(ringRadiusInMM * pixelsPerMM);

        // calculate angular start of arc in degrees
        int startingAngle = (int)(angleAtMidpoint * 180.0 / Math.acos(-1) - 20);

        // draw the arc if scaled radius is reasonable
        if ( ringRadiusInPixels > minRingSizeInPixels && ringRadiusInPixels < maxRingSizeInPixels )
        {
            g2.setColor(circleColor);
            g2.drawArc
            (
                canvasOriginX - ringRadiusInPixels + detectorXInPixels,
                canvasOriginY - ringRadiusInPixels + detectorYInPixels,
                ringRadiusInPixels*2,
                ringRadiusInPixels*2,
                startingAngle,
                40
            );
        }
    }

    protected void drawResolutionRings()
    {
        ArrayList<Ring> ringList = getResolutionRings();
        double radius = 0;

        for (Ring ring : ringList)
        {
            drawResolutionRing(ring.radius, false);
            drawResolutionText(ring.x, ring.y, ring.radius);
        }

        drawOppositeCornerArc(radius);
    }
}
