package gov.anl.gmca.spotfinder;

import gov.anl.gmca.io.Database;
import gov.anl.gmca.io.Database.DBName;
import gov.anl.gmca.io.Database.Result;
import gov.anl.gmca.io.hw.Epics;
import gov.anl.gmca.raster.ModelRasterState;
import gov.anl.gmca.util.ListenerList;
import gov.aps.jca.dbr.DBR;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.StringTokenizer;

public class DevSFType
{
	//////////////////////////////////////////////////////////////////////////////
	/////////////////// which SF is selected for a given run type ////////////////
	//////////////////////////////////////////////////////////////////////////////

	// notifies for SF selection change
	// there are two of these, one for collect and one for raster
	public ListenerList<DevSFType> llChange;

	public static int getConfigIndex(SFRunType sfRunType)
	{
		return Epics.getInteger(DevSFOptions.getSFPV(sfRunType));
	}
	public static void setConfigIndex(SFRunType sfRunType, int i)
	{
		Epics.putInteger(DevSFOptions.getSFPV(sfRunType), i);
	}

	//////////////////////////////////////////////////////////////////////////////
	/////////////////////// manage collect and raster instances //////////////////
	//////////////////////////////////////////////////////////////////////////////

	public static enum SFRunType { COLLECT, RASTER };
	public SFRunType sfRunType;

	public static DevSFType get(SFRunType type) { return type==SFRunType.COLLECT?collectInst:rasterInst; }
	private static DevSFType collectInst, rasterInst;

	public static List<SFConfig> configList = new ArrayList<SFConfig>();

	public static SFConfig getConfig(SPOTFINDER_TYPE sfType) {
		for (SFConfig config : configList) {
			if (config.id==sfType) return config;
		}
		return null;
	}

	public static int getActiveIndex(SPOTFINDER_TYPE type) {
		for (int i=0;i<configList.size();i++) {
			if (configList.get(i).id==type) return i;
		}
		return -1;
	}
	public static SPOTFINDER_TYPE[] getActiveTypes() {
		SPOTFINDER_TYPE[] ret = new SPOTFINDER_TYPE[configList.size()];
		for (int i=0;i<configList.size();i++) {
			ret[i]=configList.get(i).id;
		}
		return ret;
	}

	public DevSFType(SFRunType type)
	{
		// save the reference
		if (type==SFRunType.COLLECT) collectInst=this; else rasterInst=this;
		this.sfRunType=type;

		// connect the change listener
		llChange = new ListenerList<>(this);
		Epics.addMonitor(DevSFOptions.getSFPV(type), new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				int selectedIndex = Epics.dbrToInt(value);
				llChange.notifyListeners(selectedIndex);

				// check that the selected SF output matches the selected SF
				if (sfRunType==SFRunType.RASTER) {
					SPOTFINDER_TYPE currentSF = DevSFType.get(sfRunType).configList.get(selectedIndex).id;
					SPOTFINDER_OUTPUT output = ModelRasterState.get().getDiffDisplayParameter();
					if (null!=ModelRasterState.get() && null!=output && output.sf_type != currentSF) {
						ModelRasterState.get().setDiffDisplayParameter(currentSF.defaultOutput);
					}
				}
			}
		});

		// load configuration from SQL
		if (configList.size()==0)
		{
			Result pmaResult = Database.exec(DBName.CONFIG, "select * from SpotFinder order by id");
			for (int i = 0; i < pmaResult.numRows(); i++) {
				HashMap<String, String> row = pmaResult.getRow(i);

				SFConfig config = new SFConfig();
				config.index = Integer.parseInt(row.get("index"));
				config.id = SPOTFINDER_TYPE.fromString(row.get("id"));
				config.name = row.get("name");
				config.process_max = Integer.parseInt(row.get("process_max"));
				config.url = row.get("url");

				String hosts_string = row.get("hosts");
				StringTokenizer st = new StringTokenizer(hosts_string, " ");
				while (st.hasMoreTokens()) {
					config.hosts.add(st.nextToken());
				}
				configList.add(config);
			}
		}
	}
	//////////////////////////////////////////////////////////////////////////////
	///////////////// configuration of each SF type and its outputs //////////////
	//////////////////////////////////////////////////////////////////////////////

	public enum SPOTFINDER_OUTPUT
	{
		//						Full name				Short name		Short SF+name		Decimals	Reverse display
		DIALS_SPOTS_NO_ICE		("Spots no ice",		"SpotsNoIce",	"dials:spots_no_ice",	0, 		false),
		DIALS_SPOT_TOTAL		("Spot total",		 	"Spots", 		"dials:spots",		0, 			false),
		DIALS_TOTAL_INT			("Total intensity",		"TotalInt",		"dials:total_int",	0, 			true),
		DOZOR_MAIN_SCORE		("Main score", 			"Score", 		"dozor:score",		2, 			false),
		DOZOR_RES				("Resolution", 			"Res", 			"dozor:res",		2, 			true);
		
		public SPOTFINDER_TYPE sf_type;
		public String name, shortName, shortSF_Name;
		public int decimals;
		public boolean reverseDisplay;
		SPOTFINDER_OUTPUT(String name, String shortName, String shortSF_Name, int decimals, boolean reverseDisplay)
		{
			this.name=name;
			this.shortName=shortName;
			this.shortSF_Name=shortSF_Name;
			this.decimals=decimals;
			this.reverseDisplay=reverseDisplay;
		}
		public static SPOTFINDER_OUTPUT fromShortSF_Name(String shortSF_Name) {
			for (SPOTFINDER_OUTPUT output : values()) {
				if (output.shortSF_Name.equals(shortSF_Name)) return output;
			}
			return null;
		}
		public String toString() { return shortSF_Name; }
	}

	public static enum SPOTFINDER_TYPE
	{
		DIALS_SPOTFINDER(SPOTFINDER_OUTPUT.DIALS_SPOT_TOTAL, SPOTFINDER_OUTPUT.DIALS_TOTAL_INT),
		DOZOR(SPOTFINDER_OUTPUT.DOZOR_MAIN_SCORE, SPOTFINDER_OUTPUT.DOZOR_RES);
		
		public List<SPOTFINDER_OUTPUT> getOutputs()
		{
			ArrayList<SPOTFINDER_OUTPUT> ret = new ArrayList<>();
			for (SPOTFINDER_OUTPUT output : SPOTFINDER_OUTPUT.values()) {
				if (output.sf_type==this) ret.add(output);
			}
			return ret;
		}
		public String toRedisString()
		{
			if (this==DIALS_SPOTFINDER) return "dials";
			if (this==DOZOR) return "dozor";
			throw new IllegalStateException("Unknown spotfinder type: "+this.toString());
		}
		public int getOutputIndex(SPOTFINDER_OUTPUT output) {
			List<SPOTFINDER_OUTPUT> outputs = getOutputs();
			for (int i=0;i<outputs.size();i++) {
				if (outputs.get(i)==output) return i;
			}
			return -1;
		}
		public static SPOTFINDER_TYPE fromString(String s)
		{
			for (SPOTFINDER_TYPE t : SPOTFINDER_TYPE.values())
			{
				if (t.toString().toLowerCase().equals(s)) return t;
			}
			return null;
		}
		public SPOTFINDER_OUTPUT defaultOutput, secondDefaultOutput;
		SPOTFINDER_TYPE(SPOTFINDER_OUTPUT defaultOutput, SPOTFINDER_OUTPUT secondDefault) { this.defaultOutput=defaultOutput; this.secondDefaultOutput=secondDefault; }
		
		static
		{
			SPOTFINDER_OUTPUT.DIALS_SPOTS_NO_ICE.sf_type 			= DIALS_SPOTFINDER;
			SPOTFINDER_OUTPUT.DIALS_SPOT_TOTAL.sf_type 			= DIALS_SPOTFINDER;
			SPOTFINDER_OUTPUT.DIALS_TOTAL_INT.sf_type 			= DIALS_SPOTFINDER;
			SPOTFINDER_OUTPUT.DOZOR_MAIN_SCORE.sf_type 			= DOZOR;
			SPOTFINDER_OUTPUT.DOZOR_RES.sf_type 				= DOZOR;
		}
	}
	
	public static class SFConfig
	{
		public int index;
		public SPOTFINDER_TYPE id;
		public String name;
		public int process_max;
		public List<String> hosts = new ArrayList<>();
		public String url;
	}

	public static SFConfig getConfig(SFRunType type) { return get(type)._getConfig(); }

	public SFConfig _getConfig()
	{
		int currentConfig = Epics.getInteger(DevSFOptions.getSFPV(sfRunType));
		return configList.get(currentConfig);
	}
	
	public static String[] getNames(SFRunType type)
	{
		ArrayList<String> ret = new ArrayList<>();
		for (SFConfig config : get(type).configList)
		{
			ret.add(config.name);
		}
		return ret.toArray(new String[]{});
	}
	
}
