package gov.anl.gmca.spotfinder;

import gov.anl.gmca.collect.ModelCollectRun;
import gov.anl.gmca.collect.ModelCollectRunList;
import gov.anl.gmca.collect.RunCalculator.Frame;
import gov.anl.gmca.io.Disk;
import gov.anl.gmca.raster.ModelRasterRun;
import gov.anl.gmca.raster.ModelRasterState;
import gov.anl.gmca.spotfinder.DevSFType.SPOTFINDER_OUTPUT;
import gov.anl.gmca.spotfinder.SFData.SFImageData;
import gov.anl.gmca.spotfinder.SFData.SFType;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.SpreadsheetViewer;
import gov.anl.gmca.util.Util;
import org.apache.poi.hssf.usermodel.HSSFRow;
import org.apache.poi.hssf.usermodel.HSSFSheet;
import org.apache.poi.hssf.usermodel.HSSFWorkbook;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;

public class SFLog
{
	public static enum SFLogColumn
	{
		FRAME_NUMBER, STATE, ENERGY, ANGLE, SITE;
		public String getTitle()
		{
			String ret = super.toString().replace('_', ' ').replace('$', '.');
			return Util.capitalizeFirstLetters(ret);
		}
	};
	public static final SFLogColumn[] RASTER_COLUMNS = new SFLogColumn[]
	{
		SFLogColumn.FRAME_NUMBER,
		SFLogColumn.STATE, 
	};
	public static final SFLogColumn[] COLLECT_COLUMNS = new SFLogColumn[]
	{
		SFLogColumn.FRAME_NUMBER,
		SFLogColumn.STATE,
		SFLogColumn.ANGLE,
		SFLogColumn.SITE,
		SFLogColumn.ENERGY
	};
	public static SFLogColumn[] getColumns(SFType sfType)
	{
		return sfType==SFType.COLLECT?COLLECT_COLUMNS:RASTER_COLUMNS;
	}
	
	public static void openLog(SFType sfType, int runIndex)
	{
		String logPath = generateSFLogName(sfType, runIndex);
		writeLogFile(sfType, runIndex, logPath);
		SpreadsheetViewer.openSpreadsheet(logPath);
	}
	
	public static void writeLogFile(SFType sfType, int runIndex, String logPath)
	{
		try
		{
			if (null == logPath) logPath = generateSFLogName(sfType, runIndex);
			
			// create the workbook
			HSSFWorkbook wb = new HSSFWorkbook();

			// write raster results
			HSSFSheet resultsSheet = wb.createSheet("Results");
			resultsSheet.setDisplayRowColHeadings(false);
			writeSFResults(sfType, runIndex, resultsSheet);
			
			// write to disk
			File of = new File(logPath);
			of.createNewFile();
			FileOutputStream outStream = new FileOutputStream(of);
			wb.write(outStream);
			outStream.close();
		}
		catch (FileNotFoundException e) { e.printStackTrace(); }
		catch (IOException e)			{ e.printStackTrace(); }
	}
	
	// helper
	private static void writeSFResults(SFType sfType, int runIndex, HSSFSheet resultsSheet)
	{
		/*DBSpotFinder db = DBSpotFinder.getDB(sfType, runIndex);
		ArrayList<SFImageData> data = db.getAllPointsCopy();
		SFLogColumn[] columns = getColumns(sfType);
		
		// add the header row
		int rowCounter=0, colCounter=0;
		HSSFRow headerRow = resultsSheet.createRow(rowCounter++);
		for (SFLogColumn col : columns)
		{
			headerRow.createCell(colCounter++).setCellValue(col.getTitle());
		}
		for (SPOTFINDER_OUTPUT output : SPOTFINDER_OUTPUT.values())
		{
			headerRow.createCell(colCounter++).setCellValue(output.name);
		}

		ArrayList<Frame> frameList = null;
		if (sfType==SFType.COLLECT)
		{
			try
			{
				frameList = ((ModelCollectRun)ModelCollectRunList.get().getRun(runIndex)).seq().getFrameListCopy();
			}
			catch (Exception e) { e.printStackTrace(); }
		}
		
		// add the data
		for (SFImageData dataRow : data)
		{
			if (dataRow.state==null) continue;
			
			Frame frame = null;
			try
			{
				if (sfType==SFType.COLLECT) frame = frameList.get(dataRow.frameIndex);
			}
			catch (Exception e) { e.printStackTrace(); }
			
			colCounter=0;
			HSSFRow row = resultsSheet.createRow(rowCounter++);

			for (SFLogColumn col : columns)
			{
				String cellValue = "";
				switch (col)
				{
				case FRAME_NUMBER: 		cellValue = ""+(dataRow.frameIndex+1); break;
				case STATE: 			cellValue = (dataRow.state==null)?"":dataRow.state.toString(); break;
				case ANGLE:				cellValue = Util.printDouble(frame.angle, 2); break;
				case ENERGY:			cellValue = Util.printDouble(frame.energy, 4); break;
				case SITE:				cellValue = ""+frame.siteNumber; break;
				}
				
				row.createCell(colCounter++).setCellValue(cellValue);
			}
			for (SPOTFINDER_OUTPUT output : SPOTFINDER_OUTPUT.values())
			{
				Double value = dataRow.values.get(output);
				if (null!=value) row.createCell(colCounter++).setCellValue(value);
			}
		}*/
	}
	
	private static String generateSFLogName(SFType sfType, int runIndex)
	{
		String fileroot, fullDirectory;
		int runLabel;
		
		// get prefix, run label and directory based on run type
		switch (sfType)
		{
		case COLLECT:
			ModelCollectRun collectRun = ModelCollectRunList.get().getCurrentRunModel();
			fileroot = collectRun.getFileRoot();
			runLabel = collectRun.getRunLabel();
			fullDirectory = collectRun.getFullDirectory();
			break;
		case INTER_RASTER:
		case AUTO_RASTER:
			ModelRasterRun rasterRun = ModelRasterState.get().getCurrentRunModel();
			fileroot = rasterRun.getFileRoot();
			runLabel = rasterRun.getRunLabel();
			fullDirectory = rasterRun.getFullDirectory();
			break;
		default:
			throw new IllegalStateException("Invalid SFType: "+sfType);
		}
		
		if (!Disk.exists(fullDirectory)) Disk.mkdirs(fullDirectory);
		
		String logPrefix = fullDirectory+"/"+fileroot+"_"+runLabel+"_SpotFinder";
		String logSuffix=".xls";
		
		// add a suffix if necessary so files aren't overwritten
		int logSuffixNumber=1;
		while (new File(logPrefix+logSuffix).exists())
		{
			if (logSuffixNumber>1000)
			{
				Log.error("Error writing log file "+logPrefix+logSuffix);
				return null;
			}
			logSuffix = "_"+logSuffixNumber+".xls";
			logSuffixNumber++;
		}
		System.out.println("Returning filename "+logPrefix+logSuffix);
		return logPrefix+logSuffix;
	}
}
