/**
 * Model for StrategyResults MySQL table
 */

package gov.anl.gmca.strategy;

import gov.anl.gmca.collect.DevCollectRun;
import gov.anl.gmca.collect.DevCollectRunList;
import gov.anl.gmca.collect.ModelCollectRun;
import gov.anl.gmca.collect.ModelCollectRun.XDSProcMode;
import gov.anl.gmca.collect.ModelCollectRunList;
import gov.anl.gmca.common.DevRunList.RUN_LABEL_SETTING;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.util.Util;

import java.util.ArrayList;
import java.util.HashMap;

public class ModelStrategyResult implements Comparable<ModelStrategyResult> {
	public ArrayList<String> spaceGroups = new ArrayList();

	public static enum StrategyField {
		// db string // user string // display lines //exportable
		SAMPLE_NUMBER("sampleNumber", "Sample #", 1, false), 
		SAMPLE_NAME  ("sampleName", "Sample Name", 1, false), 
		SCREENING_SAMPLE("screeningSample", "Screening Sample", 1, false), 
		SOLUTION("solutionNumber", "Solution Number", 1, false), 
		GOOD("good","Labelit", 1, false), STATE("state", "State", 1, false), 
		WORKDIR("workdir", "Dir", 2, false), 
		SOFTWARE("software", "Software",1, false), 
		WARNING("warning", "warning", 3, false), 
		IMAGES("images", "Images", 1, false), SCORE("score", "Score", 1, false), 
		SPACE_GROUP("spaceGroup", "Space group", 1, false), 
		SPACE_GROUP_SH("spaceGroup_short", "Space group", 1, false),
		OSC_START("oscstart", "Osc. start", 0, true), 
		ANOM_OSC_START("anom_oscstart", "Anom Osc. start", -1, true), 
		INV_OSC_START("oscstart", "Inv Osc.start", -1, true), 
		MIN_END("oscend","Osc. End", 0, false), 
		MIN_ANOM_END("anom_oscend","Anom Osc. End", -1, false), 
		MIN_INV_END("oscend","Min inv end", -1, false), 
		OSC_END("oscend", "Sug. Osc. end", 0,true), 
		ANOM_OSC_END("continuous", "Anom Osc. end", -1, true), 
		INV_OSC_END("oscend", "Inv Osc. end", -1, true), 
		COMPLETE("complete","Completeness", 0, false), 
		ANOM_COMPLETE("anom_complete","Anom Completeness", -1, false), 
		INV_COMPLETE("complete","Inv Completeness", -1, false), 
		OSC_DELTA("oscdelta","Osc. delta", 1, true), 
		DETECTOR_DIST("detectordistance","Det. dist.(pred)", 1, true), 
		DETECTOR_WARN("detectorwarning","Det. warn.", 2, false), 
		RESOLUTION("resolution", "Res. (pred)", 1, false), 
		MOSAICITY("mosaicity", "Mosaicity",1, false), 
		UNIT_CELL("unitcell", "Unit cell", 1, false), 
		RMSD("rmsd", "RMSD", 1, false), 
		BEAM_X("beamx", "Beam X", 1, false), 
		BEAM_Y("beamy", "Beam Y", 1, false);

		protected String dbString, userString;
		protected int displayLines;
		boolean exportable;

		public String getDBString() {
			return dbString;
		}

		public String getUserString() {
			return userString;
		}

		public int getDisplayLines() {
			return displayLines;
		}

		public boolean getExportable() {
			return exportable;
		}

		StrategyField(String dbString, String userString, int displayLines,
				boolean exportable) {
			this.dbString = dbString;
			this.userString = userString;
			this.displayLines = displayLines;
			this.exportable = exportable;
		}
	}

	public static enum StrategyState {
		INACTIVE, ERROR, PROCESSING, DONE;

		public static StrategyState stringToState(String stateString) {
			if (null == stateString)
				return null;

			for (StrategyState state : StrategyState.values()) {
				if (stateString.toLowerCase().equals(
						state.toString().toLowerCase()))
					return state;
			}

			return null;
		}
	}

	public StrategyState getState() {
		return StrategyState.stringToState(getField(StrategyField.STATE));
	}

	protected HashMap<StrategyField, String> fieldMap = new HashMap<StrategyField, String>();

	public void setField(StrategyField field, String value) {
		fieldMap.put(field, value);
	}

	public String getField(StrategyField field) {
		return fieldMap.get(field);
	}

	public HashMap<StrategyField, String> getAllFields() {
		return fieldMap;
	}

	protected Integer resultID = null;

	public Integer getID() {
		return resultID;
	}

	public void setID(int id) {
		resultID = id;
	}

	public ModelStrategyResult(HashMap<StrategyField, String> fieldMap) {
		this.fieldMap = new HashMap<StrategyField, String>(fieldMap);
		String sampleNumber = fieldMap.get(StrategyField.SAMPLE_NUMBER);
		if (null != sampleNumber) {
			try {
				resultID = Integer.parseInt(sampleNumber);
			} catch (NumberFormatException nfe) {
			}
		}
	}

	public void exportToRun(int isAnom, boolean createNewRun) {
		double oscStart, oscEnd, oscDelta, detDist, wedge;
		try {
			if (isAnom == 1) {
				oscStart = Double.parseDouble(fieldMap
						.get(StrategyField.ANOM_OSC_START));
				oscEnd = Double.parseDouble(fieldMap
						.get(StrategyField.ANOM_OSC_END));
				// wedge =
				// Double.parseDouble(fieldMap.get(StrategyField.WEDGE));
			} else if (isAnom == 2) {
				oscStart = Double.parseDouble(fieldMap
						.get(StrategyField.OSC_START));
				oscEnd = Double
						.parseDouble(fieldMap.get(StrategyField.OSC_END));
			} else {
				oscStart = Double.parseDouble(fieldMap
						.get(StrategyField.OSC_START));
				oscEnd = Double
						.parseDouble(fieldMap.get(StrategyField.OSC_END));
			}
			oscDelta = Double
					.parseDouble(fieldMap.get(StrategyField.OSC_DELTA));
			detDist = Double.parseDouble(fieldMap
					.get(StrategyField.DETECTOR_DIST));
		} catch (NumberFormatException nfe) {
			Util.asyncPopupOk(
					"Error",
					"Error reading strategy field, export aborted. "
							+ nfe.getMessage());
			return;
		}

		DevCollectRunList listDev = DevCollectRunList.get();
		ModelCollectRunList list = ModelCollectRunList.get();

		int targetIndex = createNewRun ? list.getRunCount() + 1 : list
				.getCurrentRunIndex();

		Boolean inverseOn = null;
		Integer xdsProcMode = null;
		Double wedgeSize = null;
		switch (isAnom) {
		case 2:
			inverseOn = true;
			xdsProcMode = 2;
			wedgeSize = 30.0; // calculate later/experimentally determine
			break;
		case 1:
			xdsProcMode = 2;
			break;
		default:
			inverseOn = false;
			break;
		}

		// get fileRoot and directory suffix from run 0
		String fileRoot0 = "changeMe";
		String directorySuffix0 = "changeMe";
		boolean run0Ok = false;
		try {
			ModelCollectRun run0 = (ModelCollectRun) ModelCollectRunList.get().getRun(0);
			fileRoot0 = run0.getFileRoot();
			directorySuffix0 = run0.getDirectorySuffix();
			run0Ok = true;
		} catch (Exception e) {
			Log.error("failed to extract info from run0 during strategy preparation.");
			e.printStackTrace();
		}
		
		if (createNewRun) {
			// copy last run's contents
			DevCollectRunList.getIndex(list.getRunCount()).copyRunContents(
					targetIndex, RUN_LABEL_SETTING.INCREMENT_SRC, true, false);

			// set run device directly
			DevCollectRun.setStartFrameLabel(targetIndex, 1);
			DevCollectRun.setStartAngle(targetIndex, oscStart);
			DevCollectRun.setEndAngle(targetIndex, oscEnd);
			DevCollectRun.setDelta(targetIndex, oscDelta);
			DevCollectRun.setDistance(targetIndex, detDist);
			if (null != inverseOn)
				DevCollectRun.setInverseOn(targetIndex, inverseOn);
			if (null != xdsProcMode)
				DevCollectRun.setXDSProcMode(targetIndex, xdsProcMode);
			if (null != wedgeSize)
				DevCollectRun.setWedgeSize(targetIndex, wedgeSize);

			if (run0Ok) {
				DevCollectRun.setFileroot(targetIndex, fileRoot0);
				DevCollectRun.setDirectorySuffix(targetIndex, directorySuffix0);
			}
			
			// increase run count to load the new run
			listDev.setRunCount(targetIndex, false);
			listDev.setCurrentRun(targetIndex, true);
		} else {
			ModelCollectRun run = (ModelCollectRun) ModelCollectRunList.get()
					.getRun(targetIndex);
			if (run0Ok) {
				run.setFileRoot(fileRoot0);
				run.setDirectorySuffix(directorySuffix0);
			}

			run.dev().setStartFrameLabel(1, false);
			run.dev().setStartAngle(oscStart, false);
			run.dev().setDelta(oscDelta, false);
			run.dev().setDistance(detDist, false);
			if (null != inverseOn)
				run.dev().setInverseOn(inverseOn, false);
			if (null != xdsProcMode)
				run.dev().setXDSProcMode(XDSProcMode.values()[xdsProcMode],
						false);
			if (null != wedgeSize)
				run.dev().setWedgeSize(wedgeSize, false);

			run.fid.changeEndAngle(oscEnd); // sends update
		}

		Log.note("Strategy exported to run "
				+ DevCollectRun.getRunLabel(targetIndex));
	}

	@Override
	public int compareTo(ModelStrategyResult r2) {
		return getID() - r2.getID();
	}
}
