package gov.anl.gmca.users;

// This device handles the 23x:bi:c[0-9] PVs which store which clients are running and which is master
//
// Partially ported from GMCA's clientList.tcl

import gov.anl.gmca.collect.DevFPE;
import gov.anl.gmca.collect.ModelCollectRunList;
import gov.anl.gmca.collect.ModelCollectRunList.CollectState;
import gov.anl.gmca.common.*;
import gov.anl.gmca.gui.Main;
import gov.anl.gmca.io.Epics;
import gov.anl.gmca.io.ExecTask;
import gov.anl.gmca.io.PVLookup.PV_RUNS;
import gov.anl.gmca.io.PVLookup.PV_USER;
import gov.anl.gmca.syslog.Log;
import gov.anl.gmca.users.ModelUserTable.Client;
import gov.anl.gmca.users.ModelUserTable.User;
import gov.anl.gmca.util.ListenerList;
import gov.anl.gmca.util.ListenerList.Listener;
import gov.anl.gmca.util.Util;
import gov.aps.jca.dbr.DBR;
import redis.clients.jedis.*;

import java.util.ArrayList;
import java.util.StringTokenizer;
import java.util.concurrent.Semaphore;

public class DevUserTable extends EpicsDevice
{
	protected void readClientTable()
	{
		tableLock.acquireUninterruptibly();

		try
		{
			Client.slot 		= -1;
			model.masterSlot 	= -1;
			model.masterExists 	= false;
			
			model.resetTable();
			
			String usersString = "Read users table from EPICS: ";
			
			for (int j=0; j < ModelUserTable.MAX_CLIENTS; j++)
			{
				// read the row from EPICS
				boolean isMaster 	= (Epics.getInteger(PV_USER.mode.row(j)) == 1);
				int 	pid 		= Epics.getInteger(PV_USER.pid.row(j));
				String 	host 		= Epics.getString(PV_USER.host.row(j));
				String 	user 		= Epics.getString(PV_USER.user.row(j));
	
				// set this row in the user table
				model.userTable[j].isMaster = isMaster;
				model.userTable[j].pid 		= pid;
				model.userTable[j].host 	= host;
				model.userTable[j].user 	= user;
				
				if (null != user && user.length()>0 || null != host && host.length()>0)
				{
					usersString += "["+user+"@"+host+"] ";
				}
				
				// does this row match our client's info?
				if (pid != 0 &&
					Client.pid == pid &&
	    			(
						Client.getHostname().equals(host) ||
						Client.getShortHostname().equals(host)
	   				))
				{
					Client.slot = j;
				}
	
				// if this row is master, save its info
				if (isMaster)
				{
					model.masterSlot = j;
					model.masterExists = true;
				}
			}
			System.out.println(usersString);
		}
		catch (Exception e) { e.printStackTrace(); }
		finally
		{
			tableLock.release();
		}
		
		ModelUserTable.get().tableListenerList.notifyListeners();
	}
	
	public boolean becomeMaster()
	{
		readClientTable();

		if (Client.slot != model.masterSlot)
		{
			if (DevPyBluIceUsers.activeSlot() >= 0) {
				Log.warn("A PyBluIce client is active.");
				return false;
			}

			if (model.masterExists)
			{
				Log.warn("Another client is already master.");
				return false;
			}
			else
			{
				Epics.putInteger(PV_USER.mode.row(Client.slot), 1);
				Epics.putInteger(PV_USER.state.row(Client.slot),1);
				try (Jedis jedis = Redis.get()) {
					jedis.hset(Redis.clients(Client.getID()), "isMaster", "1");
				}
				Client.isMaster = true;
				ModelUserTable.notifyClientListeners();
				return true;
			}
		}
		return true; // we're already master
	}
	
	public void becomeMasterByForce()
	{
		if (ModelCollectRunList.get().getCollectState()!=CollectState.INACTIVE)
		{
			boolean okToProceed = Util.popupOkCancel("Collection will pause", "Data collection is in progress.  If this client becomes active, collection will be paused.  Proceed?");
			if (!okToProceed) return;
		}
		
		readClientTable();
		int pybluiceActiveSlot = DevPyBluIceUsers.activeSlot();
		
		if (Client.slot != model.masterSlot)
		{
			if (!model.masterExists && pybluiceActiveSlot < 0)
			{
				becomeMaster();
			}
			else
			{
    			new Thread(new Runnable() { public void run()
    			{
					if (pybluiceActiveSlot >= 0) {
						DevPyBluIceUsers.takeOver();
					}

					for (int i=0; i<10; i++)
    				{
    					DevFPE.get().setCmdOutput(" ");
    					DevFPE.get().setCmdOutput("g2eBecomeSlave");
    					
    					Util.sleep(1000);
    					
    					readClientTable();
    					
    					if (!model.masterExists) break;
    					
    					Log.note("Waiting for current master to become slave "+(9-i)+" seconds left");
    				}
    				
    				if (model.masterExists)
    				{
    					// master must be frozen or crashed, clear its status line
    					Epics.putInteger(PV_USER.mode.row(model.masterSlot),0);
    					Epics.putInteger(PV_USER.pid.row(model.masterSlot),0);
    					Epics.put(PV_USER.host.row(model.masterSlot),"");
    					Epics.put(PV_USER.user.row(model.masterSlot),"");
    					Epics.putInteger(PV_USER.state.row(model.masterSlot),0);
    				}
    				
    				becomeMaster();
    			}
    			}, "becomeMasterByForce").start();
			}
		}
	}
	
	public void becomeSlave(boolean popupOnInvalidState)
	{
		if (!Client.isMaster()) return;
		
		if (popupOnInvalidState)
		{
			if (ModelCollectRunList.get().isCollectThreadAlive())
			{
				Util.popupError("Action not possible", "This client can't become passive during collection.  Please pause collection first.");
				return;
			}
		}
		
		readClientTable();
		
		if (model.masterSlot != Client.slot && Client.slot != -1)
		{
			Log.warn("This client is already passive");
			return;
		}
		
		if (-1 == Client.slot)
		{
			// This client's row in the client table was erased!
			//
			// This can happen if it was removed by force for being
			// unresponsive during a byForce command on another client
			//
			// Solution: Add it back in slave mode, then call
			//           stog_become_slave.
			//

			// wait half a second, in case this client just became
			// responsive again and is processing a g2eBecomeSlave msg
			new Thread(new Runnable() { public void run()
			{
				Util.sleep(500);
				
				int emptyRow = -1;
				for (int i=0; i<ModelUserTable.MAX_CLIENTS; i++)
				{
					if (Epics.getInteger(PV_USER.pid.row(i)) == 0)
					{
						emptyRow = i;
						break;
					}
				}
				
				if (-1 == emptyRow)
				{
					Client.isMaster = false;
					ModelUserTable.notifyClientListeners();
					Util.popupError("Client table error", "The clients table is full, this instance of BluIce is no longer listed and can not become master.  Please restart.");
					return;
				}
				
				Log.note("Client was erased, adding back to "+emptyRow);
				addClientToRow(emptyRow);
				
				Client.isMaster = false;
				ModelUserTable.notifyClientListeners();
			}
			}, "becomeSlave").start();
		}
		else
		{
			Epics.putInteger(PV_USER.mode.row(Client.slot), 0);
			Epics.putInteger(PV_USER.state.row(Client.slot),0);
			try (Jedis jedis = Redis.get()) {
				jedis.hset(Redis.clients(Client.getID()), "isMaster", "0");
			}
			Client.isMaster = false;
			ModelUserTable.notifyClientListeners();
		}
	}
	
	public void cleanupClients()
	{
		readClientTable();
		
		ArrayList<String> psOutput = ExecTask.syncExec("/bin/ps -ef", null, false).stdout;
		ArrayList<Integer> javaPids = new ArrayList<Integer>();
		
		for (String line : psOutput)
		{
			if (line.contains("jbluice") || line.contains("gov.anl.gmca.gui.Main"))
			{
				StringTokenizer st = new StringTokenizer(line);
				if (st.countTokens() < 2) continue;
				st.nextToken();
				int javaPid = Integer.parseInt(st.nextToken());
				
				System.out.println("Found running Java instance: "+javaPid);
				javaPids.add(javaPid);
			}
		}
		
		System.out.println("Starting client check, this client's pid="+Client.pid+" host="+Client.hostname+" short="+Client.getShortHostname());
		for (int i=0; i < ModelUserTable.MAX_CLIENTS; i++)
		{
			User iUser = model.userTable[i];
			
			System.out.println("Checking client pid="+iUser.pid+" host="+iUser.host);
			
			if (!javaPids.contains(iUser.pid) &&
				iUser.pid != Client.pid &&
				(
					iUser.host.equals(Client.hostname) ||
					iUser.host.equals(Client.getShortHostname())
				))
			{
				Log.note("Clearing dead client at row "+i);
				
				Epics.putInteger(PV_USER.mode.row(i), 	0);
				Epics.putInteger(PV_USER.pid.row(i),	0);
				Epics.put		(PV_USER.host.row(i), 	"");
				Epics.put		(PV_USER.user.row(i), 	"");
				Epics.putInteger(PV_USER.state.row(i),	0);
				Epics.putInteger(PV_USER.state.row(i),	1);
			}
		}
	}
	
	public void registerClient()
	{
		readClientTable();

		int firstEmptyRow = -1;
		for(int j = 0; j < ModelUserTable.MAX_CLIENTS; j++)
		{
	 		if (0 == model.userTable[j].pid)
	 		{
	 			firstEmptyRow = j;
	 			break;
	 		}
		}

		if (-1 == firstEmptyRow)
		{
			Log.error("Clients table is full, can't register client.");
			return;
		}

		Log.note("Registering client at row "+firstEmptyRow);

		Epics.putInteger(PV_USER.mode.row(firstEmptyRow), 		0);
		Epics.putInteger(PV_USER.pid.row(firstEmptyRow), 		Client.pid);
		Epics.put		(PV_USER.host.row(firstEmptyRow), 		Client.getShortHostname());
		Epics.put		(PV_USER.user.row(firstEmptyRow), 		Client.user);
		Epics.putInteger(PV_USER.state.row(firstEmptyRow), 		0);
		Epics.putInteger(PV_USER.state.row(firstEmptyRow),		1);

		try (Jedis jedis = Redis.get()) {
			String key = Redis.clients(Client.getID());
			jedis.hset(key, "isMaster", 	"0");
			jedis.hset(key, "currentTab", 	"hutch");
		}

		Client.slot = firstEmptyRow;
	}
	
	public void unregisterClient()
	{
		readClientTable();

		if (Client.slot < 0 || Client.slot >= ModelUserTable.MAX_CLIENTS)
		{
			Log.note("Client slot = "+Client.slot+", returning.");
			return;
		}

		Log.note("Releasing client slot "+Client.slot);
		
		Epics.putInteger(PV_USER.mode.row(Client.slot), 		0);
		Epics.putInteger(PV_USER.pid.row(Client.slot), 			0);
		Epics.put		(PV_USER.host.row(Client.slot), 		"");
		Epics.put		(PV_USER.user.row(Client.slot), 		"");
		Epics.putInteger(PV_USER.state.row(Client.slot), 		0);
		Epics.putInteger(PV_USER.state.row(Client.slot),		1);

		try (Jedis jedis = Redis.get()) {
			jedis.del(Redis.clients(Client.getID()));
		}

		Client.slot = -1;
	}
	
	/////////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////// end public methods /////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////
	
	protected 			ModelUserTable 				model;
	final 				Semaphore 					tableLock 		= new Semaphore(1);
	protected static	DevUserTable				inst			= null;
	public				ListenerList<DevUserTable>	llCurrentUser	= null;
	
	public DevUserTable()
	{
		super("userTable");
		inst = this;
		model = ModelUserTable.get();
		llCurrentUser = new ListenerList<DevUserTable>(this);

		Epics.addMonitor(PV_RUNS.currentUser, new Epics.EpicsValueMonitor()
		{
			@Override public void valueChanged(String pvName, DBR value)
			{
				llCurrentUser.notifyListeners(Epics.dbrToString(value));
			}
		});
		
		for (int row=0; row < ModelUserTable.MAX_CLIENTS; row++)
		{
			for (PV_USER pvName : PV_USER.values())
			{
				registerDevicePV(pvName.row(row));
			}
		}
		registrationComplete();
	}
	
	public String getCurrentUser()
	{
		return Epics.getString(PV_RUNS.currentUser);
	}
	
	public void connectListeners()
	{
		DevFPE.get().commandListener.addListener(new Listener<DevFPE>()
		{
			@Override public void handleEvent(DevFPE object, Object data)
			{
				String cmd = (String)data;
				if (cmd.contains("g2eBecomeSlave"))
				{
					becomeSlave(false);
				}
			}
		});
	}
	
	public static DevUserTable get() { return inst; }
	
	protected void addClientToRow(int row)
	{
		Epics.putInteger(	PV_USER.mode.row(row),	0);
		Epics.putInteger(	PV_USER.pid.row(row), 	Client.pid);
		Epics.put(			PV_USER.host.row(row),	Client.getShortHostname());
		Epics.put(			PV_USER.user.row(row), 	Client.user);
		Epics.putInteger(	PV_USER.state.row(row),	1);
	}

	@Override
	protected void onAllPVsConnected()
	{
		System.out.println("User table: Cleaning up clients...");
		cleanupClients();
		
		System.out.println("User table: Registering client...");
		registerClient();
		
		System.out.println("User table: Rereading client table.");
		readClientTable();
	}
	
	@Override protected void valueChangedAction(String pvName, DBR value)
	{
		int row = pvName.charAt(8)-'0';
		String col = pvName.substring(10);
		
		if (col.equals("mode"))		model.userTable[row].isMaster 	= (1==Epics.dbrToInt(value));
		if (col.equals("pid"))		model.userTable[row].pid 		= Epics.dbrToInt(value);
		if (col.equals("host"))		model.userTable[row].host 		= Epics.dbrToString(value);
		if (col.equals("user"))		model.userTable[row].user 		= Epics.dbrToString(value);
		
		if (col.equals("state") && Epics.dbrToInt(value)==10)
		{
			System.out.println("EXITING: Got value "+Epics.dbrToInt(value)+" on PV "+pvName);
			Epics.putInteger(PV_USER.state.row(Client.slot), 11);
			Main.shutdown();
		}
		
		ModelUserTable.get().tableListenerList.notifyListeners();
	}
	
	@Override public void abort() {}
}
