package gov.anl.gmca.util;

import gov.anl.gmca.common.DBConfig;
import gov.anl.gmca.common.ui.NBMessageBox;
import gov.anl.gmca.gui.GUI;
import gov.anl.gmca.gui.GUIStatus;
import gov.anl.gmca.gui.Main;
import gov.anl.gmca.gui.WdgEmergencyStop;
import gov.anl.gmca.hutch.DetectorConfig;
import gov.anl.gmca.io.Disk;
import gov.anl.gmca.io.ExecTask;
import gov.anl.gmca.server.messaging.Popup;
import gov.anl.gmca.users.ModelUserTable.Client;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.*;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableItem;

import java.io.*;
import java.lang.management.ManagementFactory;
import java.lang.management.RuntimeMXBean;
import java.net.InetAddress;
import java.net.NetworkInterface;
import java.net.URL;
import java.net.URLConnection;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class Util
{
	private static final String marFPESrcPath = 	"/mnt/software/epics_synApps_mar/synApps/ccd-2-0/ccdApp/marCCDSrc/marCCD_BluIce.st";
	private static final String pilatusFPESrcPath = "/mnt/software/epics_synApps_ppu/synApps/pilatus-1-0/FPEApp/src/fpe.st";
	private static final String eigerFPESrcPath = 	"/mnt/software/epics_synApps_eiger/synApps/eiger-1-0/FPEApp/src/fpe.st";
	
	public static final double epsilon = 0.001;
	
	public static boolean sigLessThan(double a, double b)
	{
		return (a + epsilon < b);
	}
	
	public static boolean sigGreaterThan(double a, double b)
	{
		return (a - epsilon > b);
	}

	public static void printAllStackTraces()
	{
		Map<Thread, StackTraceElement[]> traces = Thread.getAllStackTraces();
		for (Iterator<Thread> it = traces.keySet().iterator(); it.hasNext(); )
		{
			Thread t = it.next();
			StackTraceElement[] stack = traces.get(t);
			
			System.out.println("Thread "+t.getName());
			for (int i=0;i<stack.length;i++)
			{
				System.out.println("\t"+stack[i]);
			}
		}
	}
	
	public static void printStackTrace()
	{
		printStackTrace(Thread.currentThread(), "");
	}
	
	public static void printStackTrace(String prefix)
	{
		printStackTrace(Thread.currentThread(), prefix);
	}
	
	public static String getStackTrace(String prefix)
	{
		return getStackTrace(Thread.currentThread(), prefix);
	}
	
	public static void printStackTrace(Thread thread, String prefix)
	{
		System.out.println(getStackTrace(thread, prefix));
	}
	
	public static String getStackTrace(Thread thread, String prefix)
	{
		String ret = new String();
		for (StackTraceElement element : thread.getStackTrace())
		{
			ret += prefix+": "+element.toString()+"\n";
		}
		return ret;
	}
	
	public static int getWorkstationNumberFromHostname()
	{
		String hostname = Client.getShortHostname();
		
		if (hostname.length()>=9 && hostname.indexOf("keithley")>=0)
		{
			return 11;
		}
		if (hostname.length()>=4 && hostname.substring(0, 3).equals("mar"))
		{
			return 12;
		}
		if (hostname.length()>=3 && hostname.substring(0, 2).equals("bl"))
		{
			return Integer.parseInt(hostname.substring(5,6));
		}
		//System.out.println("ERROR: Couldn't determine workstation number from host "+hostname+", exiting.");
		//System.exit(0);
		return 3;
	}
	
	public static int getBeamlineNumberFromHostname()
	{
		String hostname = Client.getShortHostname();
		
		try
		{
			if (DBConfig.sim5)
			{
				return 1;
			}
			if (hostname.length()>=9 && hostname.substring(0, 8).equals("keithley"))
			{
				return Integer.parseInt(hostname.substring(8,9))-1;
			}
			if (hostname.length()>=9 && hostname.substring(3, 11).equals("keithley"))
			{
				return Integer.parseInt(hostname.substring(2,3))-1;
			}
			if (hostname.length()>=4 && hostname.substring(0, 3).equals("mar"))
			{
				return Integer.parseInt(hostname.substring(3,4))-1;
			}
			if (hostname.length()>=3 && hostname.substring(0, 2).equals("bl"))
			{
				return Integer.parseInt(hostname.substring(2,3))-1;
			}
		}
		catch (NumberFormatException nfe) {}
		//System.out.println("ERROR: Couldn't determine beamline from host "+hostname+", exiting.");
		//System.exit(0);
		return 2;
	}
	
	public static String getBID()
	{
		return DBConfig.get("app_beamlineID");
	}
	
	public static boolean nearlyEquals(double d1, double d2, int decimalPrecision)
	{
		if (Double.isNaN(d1) || Double.isNaN(d2)) return false;
		return round(d1,decimalPrecision) == round(d2,decimalPrecision);
	}
	
	protected static String userLogin = null;
	public static String getUserLogin()
	{
		if (null != userLogin) return userLogin;
		
		userLogin = ExecTask.syncExec("whoami","",false).stdoutString();
		userLogin = userLogin.trim();
		return userLogin;
	}
	
	public static boolean isUserNonProprietary() {
		String stdout = ExecTask.syncExec("/usr/bin/groups " + getUserLogin(),"",false).stdoutString().toLowerCase();
		return stdout.contains("nonproprietary");
	}
	
	public static boolean isUserNonProprietary(String username) {
		String stdout = ExecTask.syncExec("/usr/bin/groups " + username,"",true).stdoutString().toLowerCase();
		return stdout.contains("nonproprietary");
	}
	
	public static void sleep(int ms)
	{
		try { Thread.sleep(ms); } catch (InterruptedException ie) { ie.printStackTrace(); }
	}
	
	public static String sprintf(String format, Object... args)
	{
		StringBuilder sb = new StringBuilder();
		Formatter formatter = new Formatter(sb, Locale.US);
		return formatter.format(format, args).toString();
	}
	
	public static void printf(String format, Object... args)
	{
		System.out.println(sprintf(format,args));
	}
	
	public static String printDouble(Double d, int digits)
	{
		if (d==null) return "null";
		if (digits <= 0)
		{
			return Integer.toString((int)round(d,digits));
		}
		else
		{
			return sprintf("%."+digits+"f",d);
		}
	}
	
	public static boolean waitForFileToExist(String filename)
	{
		File f = new File(filename);
		int maxwait = 100; // 10 seconds
		while (!f.exists())
		{
		    sleep(100);
		    if (maxwait-- == 0) return false;
		}
		return true;
	}
	
	public static double round(double d, int digits)
	{
		return Math.round(d*Math.pow(10,digits))/Math.pow(10,digits);
	}
	
	public static boolean joinThread(Thread thread, double timeoutSeconds)
	{
		int triesLeft = (int)(timeoutSeconds/0.100);
		do
		{
			try
			{
				thread.join(100);
			}
			catch (InterruptedException ie) { ie.printStackTrace(); }
		}
		while (thread.isAlive() && (triesLeft-- > 0));
		
		return thread.isAlive();
	}
	
	public static void takeSnapshot (String filename) {
		
		WdgEmergencyStop.registerProcess("axis-snapshot_v4.pl");
		
		//blt::bgexec snapshotRun /home/gmca/epics/pezca/axis-snapshot_v2.pl $snapfile $env(DCS_BEAMLINE_ID) 1 &
		String snapshotCmd = Util.getPerlPrefix()+"axis-snapshot_v4.pl";
		snapshotCmd = snapshotCmd.concat(" " +filename+".jpg -noninteractive");
		snapshotCmd = snapshotCmd.concat(" "+ DBConfig.get("app_beamlineID"));
		snapshotCmd = snapshotCmd.concat(" HighRes");
		
		ExecTask.syncExec(snapshotCmd,"axis-snapshot_v2.pl", true);
	}
	
	public static void take2Snapshot (String filename) {
		if (filename != null) {
			ExecTask runningTask = null;
			String snapshotCmd = Util.getPerlPrefix()+"axis-snapshot_v4.pl";
			snapshotCmd = snapshotCmd.concat(" " + filename+".jpg");
			snapshotCmd = snapshotCmd.concat(" " + "-display -noninteractive");
			snapshotCmd = snapshotCmd.concat(" HighRes");
			runningTask = ExecTask.new_ExecTask(snapshotCmd,"axis-snapshot_v4.pl", false);
			runningTask.start();
		}
	}

	protected static boolean _atExit = false;
	public static void atExit(boolean val) { _atExit = val; }
	public static boolean atExit()
	{
		if (!GUIStatus.inGUIMode || null == GUI.getShell())
		{
			return _atExit;
		}
		else
		{
			return (_atExit || GUI.getShell().isDisposed());
		}
	}
	
	// This function is from http://golesny.de/wiki/code:javahowtogetpid
	public static Integer getPid()
	{
        RuntimeMXBean rtb = ManagementFactory.getRuntimeMXBean();
        String processName = rtb.getName();	   	

        Integer result = null;
        		
        /* tested on: */
        /* - windows xp sp 2, java 1.5.0_13 */
        /* - mac os x 10.4.10, java 1.5.0 */
        /* - debian linux, java 1.5.0_13 */
        /* all return pid@host, e.g 2204@antonius */
        	
        Pattern pattern = Pattern.compile("^([0-9]+)@.+$", Pattern.CASE_INSENSITIVE);
        Matcher matcher = pattern.matcher(processName);
        if (matcher.matches())
        {
        	result = Integer.parseInt(matcher.group(1));
        }
    	return result;
	}
	
	public static boolean popupOkCancel(final String title, final String msg)
	{
		final AtomicInteger result = new AtomicInteger();
		
		System.out.println("popupOkCancel: title="+title+" msg="+msg);
		
		NBMessageBox mb = new NBMessageBox(GUI.getShell(), SWT.APPLICATION_MODAL | SWT.OK | SWT.CANCEL);
        mb.setText(title);
        mb.setMessage(msg);
        result.set(mb.open());
		
		boolean bResult = (result.get() == SWT.OK);
		
		System.out.println("popupOkCancel: result="+bResult);
		
        return bResult;
	}

	public static void asyncPopupError(final String title, final String msg)
	{
		if (DBConfig.serverMode) {
			new Thread(()->Popup.syncPopup("note", msg, null, 0),"Server popup error").start();
		} else {
			new Thread(()->popupError(title,msg),"Local popup error").start();
		}
	}
	
	public static void popupError(final String title, final String msg)
	{
		final AtomicInteger result = new AtomicInteger();
		
		System.out.println("popupError: title="+title+" msg="+msg);
		
		NBMessageBox mb = new NBMessageBox(GUI.getShell(), SWT.APPLICATION_MODAL | SWT.OK | SWT.ERROR);
        mb.setText(title);
        mb.setMessage(msg);
        result.set(mb.open());
	}
	
	public static boolean popupYesNo(final String title, final String msg)
	{
		final AtomicInteger result = new AtomicInteger();
		
		System.out.println("popupYesNo: title="+title+" msg="+msg);
		
		NBMessageBox mb = new NBMessageBox(GUI.getShell(), SWT.APPLICATION_MODAL | SWT.YES | SWT.NO);
        mb.setText(title);
        mb.setMessage(msg);
        result.set(mb.open());
		
		boolean bResult = (result.get() == SWT.YES);
		
		System.out.println("popupYesNo: result="+bResult);
		
        return bResult;
	}
	
	public static void asyncPopupOk(final String title, final String msg)
	{
		new Thread(new Runnable(){public void run()
		{
			popupOk(title,msg);
		}
		}, "Util.asyncPopupOk").start();
	}
	
	public static void popupOk(final String title, final String msg)
	{
		System.out.println("popupOk: title="+title+" msg="+msg);
		
        NBMessageBox mb = new NBMessageBox(GUI.getShell(), SWT.APPLICATION_MODAL | SWT.OK);
        mb.setText(title);
        mb.setMessage(msg);
        mb.open();
	}
	
	public static ArrayList<String> getURLText(String queryURL, String logPrefix)
	{
		ArrayList<String> ret = new ArrayList<String>();
		
		try
		{
    		URL url = new URL(queryURL);
    		URLConnection connection = url.openConnection();
    		connection.setConnectTimeout(1000);
    		InputStream is = connection.getInputStream();
    		BufferedReader br = new BufferedReader(new InputStreamReader(is));
    		
    		String line;
    		while ( (line=br.readLine()) != null )
    		{
    			ret.add(line);
    			if (null != logPrefix) System.out.println(logPrefix+line);
    		}
    		is.close();
		}
		catch (java.net.ConnectException ce)
		{
			System.out.println("ERROR: HTTP request failed for "+queryURL);
			ce.printStackTrace();
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}
		
		return ret;
	}

	public static void allowOnlyAlpha (Event e) {
		String string = e.text;
		char [] chars = new char [string.length ()];
		string.getChars (0, chars.length, chars, 0);
		for (int i=0; i<chars.length; i++) {
			if(chars[i] == '.')
				return;
			if(chars[i] == '-')
				return;
			if (!('0' <= chars [i] && chars [i] <= '9')) {
				e.doit = false;
				return;
			}
		}
	}
	
	public static double[] toPrimitiveArray(List<Double> a)
	{
		double[] ret = new double[a.size()];
		for (int i=0; i<a.size(); i++)
		{
			ret[i]=a.get(i);
		}
		return ret;
	}
	
	public static boolean doStringsMatch(String a, String b)
	{
		if (null == a && null == b) return true;
		if (null == a || null == b) return false;
		return a.equals(b);
	}

	public static enum OS_VER { RH5, RH6, UNKNOWN };
	public static OS_VER getOSVer()
	{
		File osFile = new File("/etc/redhat-release");
		if (osFile.exists())
		{
			BufferedReader br;
			try
			{
				br = new BufferedReader(new FileReader(osFile));
				String line = br.readLine();
				
				if (line.contains("CentOS Linux release 6.")) return OS_VER.RH6;
				if (line.contains("Scientific Linux release 6.")) return OS_VER.RH6;
				if (line.contains("Fedora release 1")) return OS_VER.RH6;
				if (line.contains("CentOS release 5.")) return OS_VER.RH5;
				if (line.contains("Fedora release")) return OS_VER.RH5;
			}
			catch (FileNotFoundException e) { e.printStackTrace(); return OS_VER.UNKNOWN; }
			catch (IOException e) { e.printStackTrace(); return OS_VER.UNKNOWN; }
		}
		return OS_VER.UNKNOWN;
	}

	
	// git method
	public static String getVersionString()
	{
		String gitPath = "/usr/bin/git";
		String gitDir = System.getProperty("application.home")+"/.git";
		
		/////////////////////////////////// get branch name ///////////////////////////////////////
		
		ExecTask task = ExecTask.syncExec("git --git-dir="+gitDir+" rev-parse --abbrev-ref HEAD", "git rev-parse", false, false);
		String branchName = task.stdoutString().trim();
		
		////////////////////////// get revision count for this branch /////////////////////////////
		
		ExecTask task2 = ExecTask.syncExec(gitPath+" --git-dir="+gitDir+" rev-list HEAD", "git rev-list", false, false);
		String revListString = task2.stdoutString();
		
		int lastNewlineIndex=-1, lineCount=1;
		while (true)
		{
			int nextNewlineIndex = revListString.indexOf('\n', lastNewlineIndex+1);
			if (nextNewlineIndex>=0)
			{
				lineCount++;
				lastNewlineIndex=nextNewlineIndex;
			}
			else break;
		}
		
		///////////////////////////////// get FPE last build date /////////////////////////////

		String fpeSrcPath;
		switch (DetectorConfig.getType())
		{
		case MAR: 		fpeSrcPath = marFPESrcPath; break;
		case PILATUS: 	fpeSrcPath = pilatusFPESrcPath; break;
		case EIGER:		fpeSrcPath = eigerFPESrcPath; break;
		default:		fpeSrcPath = null;
		}
		
		String fpeDate = "???";
		if (null != fpeSrcPath)
		{
			Calendar lastModified = Disk.getLastModified(fpeSrcPath);
			if (null!=lastModified)
			{
				SimpleDateFormat dateFormat = new SimpleDateFormat("MMM-dd");
				fpeDate = dateFormat.format(lastModified.getTime());
			}
		}
		
		return "Version "+branchName+" Build "+lineCount+" FPE "+fpeDate;
	}
	
	public static String getSVNVersionString()
	{
		String toolsDir = Main.getAppDir()+"/tools/";
		 
		String biDir = null;
		try
		{
		    biDir = (new File(Main.getAppDir())).getCanonicalPath();
		}
		catch (IOException ioe)
		{
		    System.out.println("ERROR: Could not determine canonical path for "+Main.getAppDir());
		    ioe.printStackTrace();
		    return "";
		}

		String svnPath = "";
		switch (getOSVer())
		{
		case RH5: 	svnPath = toolsDir+"svn-rh5"; break;
		case RH6: 	svnPath = toolsDir+"svn-rh6"; break;
		default: 	svnPath = "/usr/bin/svn";
		}
		ExecTask task = ExecTask.syncExec(svnPath+" info "+biDir, null, false);

		String versionString = "Version ";
		StringTokenizer st = new StringTokenizer(task.stdoutString());
		
		while (st.hasMoreTokens())
		{
			String token = st.nextToken();
			
			if (token.equals("URL:"))
			{
				// get the branch name
				String fullURL = st.nextToken();
				int lastSlashIndex = fullURL.lastIndexOf('/');
				String branchName = fullURL.substring(lastSlashIndex+1);

				// rename trunk to beta
				if (branchName.equals("trunk")) branchName = "Beta";

				// rename v20##.# to 20##.#
				if (branchName.substring(0, 3).equals("v20"))
				{
					branchName = branchName.substring(1);
				}
			    versionString += branchName;
			}

			if (token.equals("Revision:"))
			{
			    String revisionNumber = st.nextToken();
				versionString += " Build "+revisionNumber;
			}
		}
		
		return versionString;
	}
	
	public static String intToWord(int size)
	{
		String[] words = {"Zero","One","Two","Three","Four","Five","Six","Seven","Eight","Nine","Ten"};
		if (size>=0 && size<words.length) return words[size];
		return Integer.toString(size);
	}

	public static void saveImage(Image img, String pngPath)
	{
        ImageLoader loader = new ImageLoader();
        loader.data = new ImageData[]{img.getImageData()};
        loader.save(pngPath, SWT.IMAGE_PNG);
	}

	public static String getPerlPrefix()
	{
		return "/usr/bin/perl -W "+getPerlDir();
	}

	public static String getPerlDir()
	{
		String pezcaDir = System.getenv("PEZCA");
		if (null == pezcaDir) pezcaDir = "/home/gmca/epics/pezca";
		return pezcaDir+"/";
	}

	public static String capitalizeFirstLetters(String s)
	{
		String ret = new String();
		
		for (int i=0;i<s.length();i++)
		{
			if (0==i || s.charAt(i-1)==' ') ret+=Character.toUpperCase(s.charAt(i));
			else ret += Character.toLowerCase(s.charAt(i));
		}
		
		return ret;
	}
	
	public static void safeSyncExec(final Runnable runnable)
	{
		Thread guiThread = GUI.getDisplay().getThread();
		Thread currentThread = Thread.currentThread();
		if (guiThread == currentThread)
		{
			runnable.run();
		}
		else
		{
			final AtomicBoolean runnableComplete = new AtomicBoolean(false);
			
			Runnable runnableWithBoolean = new Runnable(){public void run()
			{
				runnable.run();
				runnableComplete.set(true);
			}};
			
			
    		GUI.asyncExec(runnableWithBoolean);
    		while (!runnableComplete.get()) Util.sleep(10);
		}
	}
	
	// there are many more valid filename characters but this method is conservative
	public static boolean isValidFilenameChar(char c)
	{
		if (c >= '0' && c <= '9') 				return true;
		if (c == '_' || c == '-' || c == '.')	return true;
		if (c >= 'a' && c <= 'z')				return true;
		if (c >= 'A' && c <= 'Z')				return true;
		
		return false;
	}
	
	public static boolean isValidDirChar(char c)
	{
		if (isValidFilenameChar(c)) return true;
		if (c == '/') return true;
		return false;
	}
	
	public static String arrayToEnglishList(ArrayList<Integer> list)
	{
		String msg = "";
		
		if (list.size()==1) msg += list.get(0);
		else
		{
    		for (int i=0; i<list.size()-1; i++)
    		{
    			msg += list.get(i);
    			int lastIndex = list.size()-1;
    			int lastIndexInLoop = lastIndex-1;
    			msg += (i<lastIndexInLoop)?", ":" ";
    		}
    		msg += "and "+list.get(list.size()-1);
		}
		
		return msg;
	}
	
	public static Point getClickedTableRowAndColumn(Table table, int x, int y)
	{
		Rectangle clientArea = table.getClientArea ();
		Point pt = new Point(x,y);
		int index = table.getTopIndex();
		while (index < table.getItemCount ())
		{
			boolean visible = false;
			TableItem item = table.getItem (index);
			
			if (table.getColumnCount()==0)
			{
				Rectangle rect = item.getBounds(0);
				if (rect.contains(pt))
				{
					return new Point(index,0);
				}
				if (!visible && rect.intersects (clientArea))
				{
					visible = true;
				}
			}
			else for (int i=0; i < table.getColumnCount(); i++)
			{
				Rectangle rect = item.getBounds (i);
				if (rect.contains (pt))
				{
					return new Point(index,i);
				}
				if (!visible && rect.intersects (clientArea))
				{
					visible = true;
				}
			}
			if (!visible) return null;
			index++;
		}
		return null;
	}
	
	private static Boolean isWindows = null;
	public static boolean isWindows()
	{
		if (null==isWindows)
		{
			String os = System.getProperty("os.name").toLowerCase();
			isWindows = (os.indexOf("win") >= 0) && (os.indexOf("linux") < 0);
			System.out.println("isWin="+isWindows+" OS="+os);
		}
		return isWindows;
	}
	
	public static boolean checkPassword(String user, String password)
	{
		// run checkpw
		String cmd = "/usr/local/bin/checkpw_stdout "+user+" "+password;
		ExecTask task = ExecTask.syncExec(cmd, "checkpw", false);
		
		// parse result
		if (null==task.stdout || task.stdout.size()==0)
		{
			System.out.println("checkpw for user "+user+" returned no output");
			return false;
		}
		
		StringTokenizer st = new StringTokenizer(task.stdout.get(0));
		if (st.hasMoreTokens())
		{
			int result = Integer.parseInt(st.nextToken());
			System.out.println("checkpw for user "+user+" returned "+result);
			return result==0;
		}
		return false;
	}

	// groups strings inside double quotes - doesn't yet handle escaped quotes
	public static ArrayList<String> parseStringWithQuotes(String s)
	{
		ArrayList<String> ret = new ArrayList<String>();
		
		Matcher m = Pattern.compile("\"([^\"]*)\"|(\\S+)").matcher(s);
		while (m.find())
		{
			if (m.group(1) != null)
			{
				String f = m.group(1);
				if (f.trim().length()==0) continue;
				ret.add(f.trim());
			}
			else
			{
				StringTokenizer st = new StringTokenizer(m.group(2));
				while (st.hasMoreTokens())
				{
					ret.add(st.nextToken());
				}
			}
		}
		return ret;
	}

	public static String getPathOnly(String pathWithFilename)
	{
		File f = new File(pathWithFilename);
		if (f.isDirectory()) return f.toString();
		try
		{
			return f.getParentFile().getCanonicalPath().toString();
		}
		catch (IOException ioe)
		{
			ioe.printStackTrace();
			return null;
		}
	}

	public static String getFilenameOnly(String pathWithFilename)
	{
		if (Disk.isDirectory(pathWithFilename)) return null;
		
		java.io.File f = new java.io.File(pathWithFilename);
		return f.getName();
	}

	public static long msSince(long startMs)
	{
		long nowMs = System.currentTimeMillis();
		return nowMs-startMs;
	}

	public static String readableSeconds(int totalSec)
	{
		// Examples of output:
		// 14hr, 1.6hr, 37min, 4.2min, 50sec, 9sec
		if (totalSec>=36000)	return (totalSec/3600)+" hr";
		if (totalSec>=3600) 	return Util.sprintf("%.1f hr",totalSec/3600.0);
		if (totalSec>=600)		return (totalSec/60)+" min";
		if (totalSec>=60) 		return Util.sprintf("%.1f min",totalSec/60.0);
		return totalSec+" sec";
	}
	
	// ex. name=run, strings=3,5,8
	// output = "runs 3, 5, and 8"
	public static String commaSeparatedList(String name, ArrayList<? extends Object> objects, boolean addPeriod)
	{
		StringBuilder sb = new StringBuilder();
		int count = objects.size();
		sb.append(name);
		if (count>1) sb.append("s");
		sb.append(" ");
		for (int i=0;i<count;i++)
		{
			sb.append(objects.get(i).toString());
			if (count-i>=3) sb.append(", ");
			else if (count-i==2) sb.append(", and ");
			else if (addPeriod) sb.append(".");
		}
		return sb.toString();
	}

	public static String getServerIP()
	{
		String override = System.getProperty("server_ip");
		if (null!=override) return override;
		if (null!=System.getProperty("server_local_ip")) {
			return getLocalPrivateAddress();
		}
		return DBConfig.get("bluice_server_ip");
	}

	public static String getLocalPrivateAddress()
	{
		try {
			for (NetworkInterface i : Collections.list(NetworkInterface.getNetworkInterfaces())) {
				for (InetAddress a : Collections.list(i.getInetAddresses())) {
					String a_str = a.toString().substring(1);
					if (a_str.startsWith("10.20.103")) {
						return a_str;
					}
				}
			}
		}
		catch (Exception e) {
			e.printStackTrace();
			return null;
		}
		return null;
	}

	public static String getServerURL()
	{
		return String.format("http://%s:%s", getServerIP(), DBConfig.get("bluice_server_port"));
	}

	public static boolean beamSizeCalculated()
	{
		return DBConfig.get("beam_size_calculated").equals("1");
	}

	public static String stripLeadingZeroes(String s)
	{
		while (s.length()>=2 && s.charAt(0)=='0')
			s=s.substring(1);
		return s;
	}

	public static void signal(long pid, String signal)
	{
		try {
			Runtime.getRuntime().exec("kill -"+signal+" " + pid);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public static String getSrcPath(String suffix)
	{
		return System.getProperty("application.home")+"/java/src/gov/anl/gmca/"+suffix;
	}
}
