#!/opt/anaconda/bin/python

# This script is self-contained and can be run directly, no -m or cwd setting needed.
# It is run by the /usr/local/bin/bluice launcher or anywhere else it's needed.
#
# Return values:
#   0: OK - no mismatch, possibly as result of clear
#   1: ERROR - mismatch, don't allow proceeding
#   2: WARN - mismatch but user is allowed to proceed and was warned

import socket
import getpass
import grp
import sys
import os
import epics
import PyQt5.QtWidgets as qw
import threading
import subprocess

# Configuration
script_name = 'changeBeamlineUser.pl'
always_ask = len(sys.argv)>1 and sys.argv[1]=='always_ask'
local_user = getpass.getuser()
is_staff = 'staffGroup' in [g.gr_name for g in grp.getgrall() if local_user in g.gr_mem]
app = qw.QApplication(sys.argv)

# 1) compare 23x:fpe:ccduser to $USER
beamline_index = int(socket.gethostname()[2:3])-1
beamline_prefix = ['23i','23o','23b'][beamline_index]
pv_name = '{}:bi:runs:currentUser'.format(beamline_prefix)
ccduser = epics.caget(pv_name)
if ccduser is None:
    qw.QMessageBox.information(None, 'PV not found', 'Could not find EPICS PV '+pv_name+', can\'t start JBluIce')
    sys.exit(1)
if local_user == ccduser and not always_ask:
    print('Local user {} equals ccd user {}, exiting.'.format(local_user, ccduser))
    sys.exit(0)

# 2) handle mismatch based on user

# 2a) just exit for gmca, clear is not useful
if local_user == 'gmca':
    print('User is gmca, skipping user change.')
    sys.exit(2)

# 2b) ask for all other users
msg = "Beamline user: {}\nYour login: {}\n\nDo you wish to RESET the beamline setup and become the active user?"
msg = msg.format(ccduser, local_user)
ok = qw.QMessageBox.question(None, 'User mismatch', msg) == qw.QMessageBox.Yes
if not ok:
    if is_staff:
        qw.QMessageBox.information(None, 'Reset declined', 'Reset declined.  Continuing as staff, major operations may not work.')
        sys.exit(2)
    else:
        qw.QMessageBox.information(None, 'Reset declined', 'Reset declined, JBluIce will not start.')
        sys.exit(1)

# 3) run script
# 3a) dialog during script run
mb = qw.QMessageBox()
mb.setWindowTitle('Change user')
mb.setText('Resetting, please wait...')
mb.setStandardButtons(qw.QMessageBox.Cancel)
mb.buttonClicked.connect(lambda:process.terminate())
mb.show()

# 3b) run the script then close the dialog
def run_script():
    global mb, process
    cmd_list = [script_name, '--force'] if always_ask else [script_name]
    process = subprocess.Popen(cmd_list, stdout=subprocess.PIPE, stderr=subprocess.PIPE, preexec_fn=os.setsid)
    process.wait()
    print(process.communicate()[0].decode('utf-8'))
    mb.close()
    print('exit code is '+str(process.returncode))

script_thread = threading.Thread(target=run_script)
script_thread.start()

# 3c) start app loop
app.exec()

# exit with process's return code
sys.exit(process.returncode)
