// logDirect.cc
//
// Usage: bluice | logDirect
//
// logDirect enables standard log rotation by handling SIGHUP and
// reopening the log after it's been rotated out of the way.  It
// also adds a timestamp to each line.
//
// Logs are stored in /var/tmp/bluice-[username]

#include <cstdio>
#include <csignal>
#include <cstring>
#include <iostream>
#include <fstream>
#include <sstream>
#include <pwd.h>
#include <cstdarg>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
using namespace std;

const int MAX_CHAR_ARRAY_SIZE=4096*16;
string logPrefix = "/var/tmp/";
ofstream log;
string username;
string logDir;
int maxFileSize = 1 * 1024 * 1024 * 1024; // 1GB

void error(const char *fmt, ...)
{
  char	buf[MAX_CHAR_ARRAY_SIZE];
  va_list 	ap;
  
  va_start(ap, fmt);
  vsprintf(buf, fmt, ap);
  strcat(buf, "\n");
  fflush(stderr);
  fputs(buf, stderr);
  fflush(NULL);
  va_end(ap);
}

string generateLogname()
{
  // get timestamp
  const time_t now = time(0);
  char timestampBuf[20];
  strftime(timestampBuf,20,"%Y-%m-%d-%H-%M-%S",localtime(&now));

  pid_t pid = getpid();
  
  // construct filename
  ostringstream os;
  os << logDir << username << "-" << timestampBuf << "-" << pid << ".log";
  return os.str();
}

void logOpen()
{
  struct stat st;
  int statRet = stat(logDir.c_str(),&st);
  int acc = access(logDir.c_str(),W_OK|X_OK);

  if (statRet==-1 && errno == ENOENT)
  {
    if (mkdir(logDir.c_str(), 00770) != 0)
    {
      error("Error creating directory %s by user %s", logDir.c_str(), username.c_str());
      return;
    }
  }
  else if (!S_ISDIR(st.st_mode) || (S_ISDIR(st.st_mode) && acc != 0))
  {
    error("Directory %s is either not a directory or can't be accessed by user %s", logDir.c_str(), username.c_str());
    return;
  }
  else if (statRet == -1)
  {
    error("Directory %s caused an unknown error when being checked by stat() by user %s", logDir.c_str(), username.c_str());
    return;
  }

  string logName = generateLogname();
  log.open(logName.c_str(), ofstream::out | ofstream::app);
  if (!log)
  {
    error("Log file %s could not be opened by user %s", logName.c_str(), username.c_str());
    return;
  }
}

// close/reopen the log file to allow rotation of logs
void onSigHup(int sigNum)
{
  log.close();
  logOpen();
}

int main()
{
  umask(0);

  // get username
  username = getpwuid(getuid())->pw_name;

  // set logDir
  logDir = logPrefix + "bluice-" + username + "/";

  // register signal handler for log close/reopen
  signal(SIGHUP, onSigHup);

  // open the log
  logOpen();

  // loop to handle input
  char timestampBuf[16];
  char msBuf[4];
  timeval tv;
  time_t tt;
  long ms;
  //tm *ptm;
  string s;
  int fileSize=0;
  while (getline(cin, s))
  {
    // get current time
    gettimeofday (&tv, NULL);
    tt = tv.tv_sec;
    ms = tv.tv_usec / 1000;

    // format date/time and ms
    tm *ptm = localtime(&tt);
    strftime (timestampBuf, 16, "%b-%d %H:%M:%S", ptm);
    sprintf(msBuf,"%03ld",ms);

    // combine and print
    log << "[" << timestampBuf << "." << msBuf << "] " << s << endl;

    // add to file size and rotate log if it's too big
    fileSize += 1 + 16 + 1 + 3 + 2 + s.length() + 1;
    if (fileSize > maxFileSize)
    {
      log << "WARNING: File size limit reached, rotating log." << endl;
      
      log.close();
      logOpen();

	  fileSize=0;
    }
  }

  // close the log
  log.close();

  return 0;
}
