#!/opt/anaconda/bin/python

# Put this in front of any command with CBF files individually listed (no wildcards).
# It will wait until all files are complete to the last byte, then run the command.
#
# CBF filenames are read from the end of the command, and parsing ends when any
# argument doesn't end in ".cbf".
#
# Wildcards are not allowed because files may not be there yet to find.
#
# With minimal imports and only checking the last few bytes of each file,
# this script runs in 70ms for one image, 270ms for 10 images at BM if all
# images are already complete.
#
# Ex.: ./cbf_footer_check.py dials.find_spots A1_000001.cbf A1_000002.cbf
#
# -MH

import os
import sys
import subprocess
import time

footer_ref = b'--CIF-BINARY-FORMAT-SECTION----\r\n;\r\n\r\n'
timeout_sec = 20  # approximate
if 'TIMEOUT_MS' in os.environ:
    timeout_sec = int(os.environ['TIMEOUT_MS'])/1000
    print('Timeout set to {:2f} sec'.format(timeout_sec))

cbf_list = []
for i in range(len(sys.argv)-1,0,-1):
    if sys.argv[i].endswith('.cbf'):
        cbf_list.append(sys.argv[i])
    else:
        break

print('Read CBF list: '+(' '.join(sorted(cbf_list))))

# for each image
retry = 0
while True:
    all_match = True
    for img in sorted(cbf_list):
        try:
            with open(img, 'rb') as file:
                file.seek(-1*len(footer_ref), os.SEEK_END)
                footer = file.read()
                match = footer_ref==footer
                if not match:
                    print('Waiting for partial file: {}'.format(img))
                    all_match = False
                    break
        except:
            print('Waiting for non-existant file: {}'.format(img))
            all_match = False
            break
    if all_match:
        break
    else:
        retry += 1
        if retry==timeout_sec:
            print('Out of retries, exiting with error.')
            sys.exit(-1)
        else:
            time.sleep(1)

sys.exit(subprocess.call(sys.argv[1:]))
