#!/usr/bin/env python

#  get resolution cutoff from CORRECT.LP using CC(1/2) as a criterior
#
#  @ qxu Sept-13-2016
#  @ modified to use 2 pass correct, or aimless to get a better resolution estimation

import os

def run_aimless_get_resolution(correct_lp_file="CORRECT.LP", target_CC=30.0):
    # this version of resolution cutoff use aimless results
    if os.path.isfile("XDS_ASCII.HKL"):
        os.system("pointless -copy xdsin XDS_ASCII.HKL hklout XDS_ASCII.mtz > pointless4res.log")
        other_outputs = "ROGUES {0} ROGUEPLOT {0} SCALES {0} NORMPLOT {0} ANOMPLOT {0} CORRELPLOT {0}".format("/dev/null")
        os.system("echo ONLYMERGE | aimless hklin XDS_ASCII.mtz hklout junk_scaled.mtz {} > aimless4res.log".format(other_outputs))
        os.remove("junk_scaled.mtz")
        reslimit = []
        with open("aimless4res.log", 'r') as log:
            for line in log:
                if " == maximum resolution" in line:
                    return line.split("=")[1].replace("A", "").replace(" ", "")
                elif "limit = " in line:
                    res = line.split("=")[1].replace("A", "").replace(" ", "")
                    reslimit.append(res)
        return min( [ float(r) for r in reslimit] )

    # aimless or pointless failed, run the failsafe method
    return one_pass_correct_get_resolution(correct_lp_file, target_CC)


def getBlock(inputStr, firstStr, lastStr, order='o'):
    # extract string between two strings: firstStr and lastStr
    #   if order is given as 'r', the last occurrence will be extracted
    #   tailing and beginning new lines will be stripped
    # return: a text block or an empty str
    try:
        if order == 'r':
            start = inputStr.rindex(firstStr) + len(firstStr)
            end = inputStr.rindex(lastStr, start)
        else:
            start = inputStr.index(firstStr) + len(firstStr)
            end = inputStr.index(lastStr, start)
        return inputStr[start:end].rstrip().strip()
    except ValueError:
        return None

def parse_correct_table(correct_lp_file="CORRECT.LP", target_CC=30.0):
    '''Read the CORRECT.LP file and get an estimate of the resolution limit.
    This should then be recycled to a rerun of CORRECT, from which the
    reflections will be merged to get the statistics.'''

    correct_lp = open(correct_lp_file, 'r').read()
    
    str1 = "RESOLUTION     NUMBER OF REFLECTIONS    COMPLETENESS R-FACTOR  R-FACTOR COMPARED I/SIGMA   R-meas  CC(1/2)  Anomal  SigAno   Nano"
    str2 = "NUMBER OF REFLECTIONS IN SELECTED SUBSET OF IMAGES"
    table = getBlock(correct_lp, str1, str2, order="r")
    lines = table.split("\n") [2:-1]

    # !! fail to deal with cases that CC fall below threshold in the middle!!
    res_cc = []
    for i, line in enumerate(lines):
        els = line.split()
        resol = float(els[0])
        cc = float(els[-4].replace("*", ""))
        res_cc.append([resol, cc])
        if cc < target_CC:
            break

    return res_cc

def one_pass_correct_get_resolution(correct_lp_file="CORRECT.LP", target_CC=30.0):
    '''Read the CORRECT.LP file and get an estimate of the resolution limit.
    This should then be recycled to a rerun of CORRECT, from which the
    reflections will be merged to get the statistics.'''
    res_cc = parse_correct_table(correct_lp_file, target_CC)

    # if all shell cc > target_CC, return max resol
    if res_cc[-1][1] >= target_CC:
       return res_cc[-1][0]
    elif len(res_cc) >= 2:
       # interpolate to cc around 30%
       x1, y1 = res_cc[-2]
       x2, y2 = res_cc[-1]
       slope = (y2-y1) / (x2-x1)
       intercept = y2 - slope * x2 
       # print y1, y2, x1, x2, slope, intercept
       # note assuming y (CC) = 0 for x (res) = 0 is violated
       return round((target_CC - intercept)/slope, 2)
    else:
       print "Something is wrong here..."
       return None

def two_pass_correct_get_resolution(correct_lp_file="CORRECT.LP", target_CC=30.0):
    res_cc = parse_correct_table(correct_lp_file, target_CC)
    # if all greater than target CC return the last resolution
    if abs(res_cc[-1][1] - target_CC) < 5.0:
       return res_cc[-1][0]

    # otherwise take the last entry and rerun CORRECT
    maxres = res_cc[-1][0]

    lines = open("XDS.INP", "r").readlines()
    for i, line in enumerate(lines):
        if line.startswith("JOB="):
            lines[i] = "JOB= CORRECT\n"
        if line.startswith("INCLUDE_RESOLUTION_RANGE="):
            lines[i] = "INCLUDE_RESOLUTION_RANGE= 30.00 {}\n".format(maxres) 

    os.rename("XDS.INP", "XDS.INP.bk")
    os.rename("CORRECT.LP", "CORRECT.LP.bk")
    with open("XDS.INP", "w") as xdsinp:
        xdsinp.write("".join(lines))

    # rerun xds_par and parse CORRECT.LP
    os.system("xds_par > /dev/null")
    return one_pass_correct_get_resolution(correct_lp_file, target_CC)

if __name__ == "__main__":
    #print(one_pass_correct_get_resolution())
    print(run_aimless_get_resolution())
    #print(two_pass_correct_get_resolution())
