import string
import pandas as pd


class ScreenTemplate:
    def __init__(self, filename='ALS_template.xls', number_of_positions=6, crystals_per_puck=16):
        self.crystals_per_puck = crystals_per_puck
        self.number_of_positions = number_of_positions
        self.header = ['Port', 'CrystalID', 'Protein', 'Comment',
                       'Directory', 'FreezingCondition', 'CrystalCondition',
                       'Metal', 'Priority', 'Person']
        self.nrows = crystals_per_puck * number_of_positions
        self.ncols = len(self.header)
        self.filename = filename
        self.data = self.generate_rows()
        self.worksheet = None

    def generate_rows(self):
        data = []
        positions_labels = string.ascii_uppercase[:self.number_of_positions]
        for i in range(self.number_of_positions):
            pl = positions_labels[i]
            for j in range(self.crystals_per_puck):
                crystal_number = str(j + 1)
                label = pl + crystal_number
                row = [label, label, "", "", label, "", "", "", "", ""]
                data.append(row)
        return data

    def create(self):
        self.generate_rows()
        df = pd.DataFrame(data=self.data, columns=self.header)
        writer = pd.ExcelWriter(self.filename, engine='xlwt')
        df.to_excel(writer, sheet_name='Sheet1', index=False)
        writer.save()


class ScreenTemplate2:
    def __init__(self, filename='ALS_template.xls', number_of_positions=6, crystals_per_puck=16):
        self.crystals_per_puck = crystals_per_puck
        self.number_of_positions = number_of_positions
        self.header = ['Port', 'CrystalID', 'Protein', 'Comment', 'Directory', 
                       'FreezingCondition', 'CrystalCondition', 'Metal', 
                       'Spacegroup', 'ModelPath', 'SequencePath', 'Priority', 'Person']
        self.nrows = crystals_per_puck * number_of_positions
        self.ncols = len(self.header)
        self.filename = filename
        self.error_type = 'warning'
        self.data = self.generate_rows()

    def index2port(self, index):
        # index 0-287
        positions_labels = string.ascii_uppercase[:self.number_of_positions]
        ipuck, posi = divmod(index, self.crystals_per_puck)
        return positions_labels[ipuck], posi+1

    def generate_rows(self):
        data = []
        positions_labels = string.ascii_uppercase[:self.number_of_positions]
        for i in range(self.number_of_positions):
            pl = positions_labels[i]
            for j in range(self.crystals_per_puck):
                crystal_number = str(j + 1)
                label = pl + crystal_number
                row = [label, label, "", "", label, "", "", "", "", "", "", "", ""]
                data.append(row)
        return data

    def create(self):
        self.generate_rows()
        df = pd.DataFrame(data=self.data, columns=self.header)
        #writer = pd.ExcelWriter(self.filename, engine='xlwt')
        writer = pd.ExcelWriter(self.filename, engine='xlsxwriter')


        df.to_excel(writer, sheet_name='Sheet1', index=False)
        self.worksheet = writer.sheets['Sheet1']

        self.add_notation()

        self.worksheet.set_column("A:A", 4)
        self.worksheet.set_column("B:B", 19)
        self.worksheet.set_column("C:C", 20)
        self.worksheet.set_column("D:D", 20)
        self.worksheet.set_column("E:E", 40)
        self.worksheet.set_column("F:F", 40)
        self.worksheet.set_column("G:G", 20)
        self.worksheet.set_column("H:H", 10)
        self.worksheet.set_column("I:I", 10)
        self.worksheet.set_column("J:J", 30)
        self.worksheet.set_column("K:K", 30)
        self.worksheet.set_column("L:L", 6)
        self.worksheet.set_column("M:M", 15)
        self.worksheet.set_default_row(hide_unused_rows=True)

        self.attach_column_name_validator()
        self.attach_Port_validator()
        self.attach_CrystalID_validator()
        self.attach_Directory_validator()

        writer.save()

    def add_notation(self):
        text = ("The spreadsheet must contains 289 rows including the header row. \n\n"
                "The Port column should not be modified. \n\n"
                "The CrystalID column length must be less than 20 chars, \n"
                "only alphanumberic, _ (underscore), and - is allowed. \n\n"
                "The Directory column define subdirectory for the crystal, \n"
                "only alphanumberic, _ (underscore), /, and - is allowed. \n\n"
                "If invalid values are inputted, a warning will be shown. \n\n"
                "To validate the spreadsheet in Excel, "
                "Go to Data/Data Validation/Circle Invalid Data; "
                "in case of # of rows is incorrect, the header row will be circled.\n\n"
                "Please save the spreadsheet in excel 1997-2003 format (.xls, not xslx) "
                "to ensure compatibility."
                )
        options = {
            'fill': {'color': 'yellow'},
            'width': 400,
            'height': 300,
        }
        self.worksheet.insert_textbox(4, 5, text, options)        

    def attach_column_name_validator(self):
        # A1 = Port
        # B1 = CrystalID
        # E1 = Directory
        self.worksheet.data_validation('A1', {'validate':'custom', 
                                        'value': '=AND(COUNTA(A:A)=289, EXACT(A1, "Port"))',
                                        'ignore_blank': False,
                                        'error_type':self.error_type,
                                        'input_message': "must contain 289 rows including header",
                                        'error_message': "must contain 289 rows including header", 
                                        })

        self.worksheet.data_validation('B1', {'validate':'custom', 
                                        'value': '=AND(COUNTA(A:A)=289, EXACT(B1, "CrystalID"))',
                                        'ignore_blank': False,
                                        'error_type':self.error_type,
                                        'input_message': "must contain 289 rows including header",
                                        'error_message': "must contain 289 rows including header", 
                                        })

        self.worksheet.data_validation('E1', {'validate':'custom', 
                                        'value': '=AND(COUNTA(A:A)=289, EXACT(E1, "Directory"))',
                                        'ignore_blank': False,
                                        'error_type':self.error_type,
                                        'input_message': "must contain 289 rows including header",
                                        'error_message': "must contain 289 rows including header", 
                                        })

    def attach_Port_validator(self):
        # column A2 to A289
        #condition_template = """=AND(COUNTIF({cells}, {cell})<2, ROWS({cells})=288, COUNTBLANK({cells})=0, EXACT({cell}, "{expected_value}") )"""
        condition_template = """=AND(COUNTIF({cells}, {cell})<2, EXACT({cell}, "{expected_value}") )"""
        cells = "$A$2:$A$289"

        for row in range(2,290):
            label = "{}{}".format(*self.index2port(row-2))
            cell = "$A${}".format(row)
            value = condition_template.format(cell=cell, cells=cells, expected_value=label)
            self.worksheet.data_validation(cell, {'validate': 'custom', 
                                                    'value': value,
                                                    'input_title': 'Port',
                                                    'ignore_blank': False,
                                                    'error_type':self.error_type,
                                                    'input_message': "robot port, no need to change",
                                                    'error_message': "robot port, no need to change"
                                                    })


    def attach_CrystalID_validator(self):
        # column B2 to B289
        condition_template = """=AND(len({cell})>1, len({cell})<20, COUNTIF({cells}, {cell})<2, ISNUMBER(SUMPRODUCT(FIND(MID({cell},ROW(INDIRECT("1:"&LEN({cell}))),1),"0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_-"))) )"""
        cells = "B2:B289"
        for row in range(2,290):
            cell = "B{}".format(row)
            value = condition_template.format(cell=cell, cells=cells)
            self.worksheet.data_validation(cell, {'validate': 'custom', 
                                                    'value': value,
                                                    'input_title': 'crystalID',
                                                    'ignore_blank': False,
                                                    'error_type':self.error_type,
                                                    'input_message': "1<len<20, used as prefix, unique, allowed chars: alphanumeric and _-",
                                                    'error_message': "1<len<20, unique, allowed chars: alphanumeric and _-"
                                                    })
        

    def attach_Directory_validator(self):
        # column E2 to E289
        condition_template = """=AND(len({cell})>1, len({cell})<60, COUNTIF({cells}, {cell})<2, ISNUMBER(SUMPRODUCT(FIND(MID({cell},ROW(INDIRECT("1:"&LEN({cell}))),1),"0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_-/"))) )"""
        cells = "E2:E289"
        for row in range(2,290):
            cell = "E{}".format(row)
            value = condition_template.format(cell=cell, cells=cells)
            self.worksheet.data_validation(cell, {'validate': 'custom', 
                                                    'value': value,
                                                    'input_title': 'relative path',
                                                    'ignore_blank': False,
                                                    'error_type':self.error_type,
                                                    'input_message': "subdirectory path, 1<len<60, unique, allowed chars: alphanumeric and _-/",
                                                    'error_message': "subdirectory path, 1<len<60, unique, allowed chars: alphanumeric and _-/"
                                                    })



if __name__ == '__main__':
    #ScreenTemplate(filename='ALS_template_18pucks.xls', number_of_positions=18,crystals_per_puck=16).create()
    #ScreenTemplate(filename='ALS_template_6pucks.xls', number_of_positions=6,crystals_per_puck=16).create()
    ScreenTemplate2(filename='ALS_template_18pucks.xlsx', number_of_positions=18,crystals_per_puck=16).create()
    print("NB: Open the result xlxs file using libreOffice and save it into xls format (excel prior 2003)")
    #ScreenTemplate2(filename='ALS_template_6pucks.xls', number_of_positions=6,crystals_per_puck=16).create()

    #ScreenTemplate(filename='Rigaku_template_18pucks.xls', number_of_positions=18,crystals_per_puck=12).create()
    #ScreenTemplate(filename='Rigaku_template_6pucks.xls', number_of_positions=6,crystals_per_puck=12).create()
