#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 25 15:02:10 2016

@author: qxu

NB only tested for gmca idd beamlines, here the beamx and beamy 
are used directly from header, no transformation is necessary
"""

from __future__ import absolute_import

import logging
import os
import re
from subprocess import Popen, PIPE
import math

import h5py

from ..config import config
from ..common.utils import getNumbers

logger = logging.getLogger(__name__)


class Header:
    def __init__(self, image_file_name):
        self.header = {}
        if image_file_name.endswith('.cbf'):
            self.header = extract_minicbf_header(image_file_name)

        if image_file_name.endswith('.h5'):
            self.header = extract_hdf5_header(image_file_name)

        if len(self.header) < 10:
            self.header = extract_header_failback(image_file_name)

        if len(self.header) < 10:
            logger.error("unable to understand the header for {}"
                         .format(image_file_name))
        self.get_edge_corner_res()

    def get_item_by_key(self, key):
        if key in self.header:
            return self.header[key]
        return None

    def get_keys(self):
        return self.header.keys()

    def get_header_as_dict(self):
        return self.header

    def get_edge_corner_res(self):
        """ get resolution at the edge of the detector"""
        nx, ny = self.header['size']
        qx, qy = self.header['pixel_size']
        orgx, orgy = self.header['beam_in_pixel']
        distance = self.header['distance']
        wavelength = self.header['wavelength']
        two_theta = self.header['two_theta']

        edge_res, corner_res = get_max_resolution_on_detector((nx, ny),
                                                              (orgx, orgy),
                                                              qx,
                                                              wavelength,
                                                              distance,
                                                              detector_tilt=two_theta)
        self.header['edge_res'] = edge_res
        self.header['corner_res'] = corner_res
        return edge_res, corner_res


def extract_header_failback(image_file_name):
    def parse_header(input_str):
        lines = input_str.split(";\n")
        adict = {}
        for line in lines:
            elements = line.split("=")
            if len(elements) == 2:
                key, value = elements
                adict[key] = value.lstrip(" ")
        return adict

    def dump_header(filename):
        """dump header of an image using an external program, defined in the config file"""
        if os.path.exists(config.IMG_HEADER_DUMPER):
            process = Popen([config.IMG_HEADER_DUMPER, filename], stdout=PIPE)
            (output, err) = process.communicate()
            exit_code = process.wait()
            if exit_code == 0:
                return parse_header(output)
            else:
                logger.warning("failed to extract header information")
        return None

    h = dump_header(image_file_name)
    header_dict = dict()
    if h:
        header_dict['size'] = int(h['SIZE2']), int(h['SIZE1'])
        header_dict['saturation'] = float(h['OVERLOAD_CUTOFF'])
        header_dict['distance'] = float(h['DISTANCE'])
        header_dict['detector_class'] = h['DETECTOR']
        header_dict['serial_number'] = h['DETECTOR_SN']
        header_dict['wavelength'] = float(h['WAVELENGTH'])
        header_dict['beam_in_mm'] = float(h['BEAM_CENTER_X']), float(h['BEAM_CENTER_Y'])
        header_dict['pixel_size'] = float(h['PIXEL_SIZE']), float(h['PIXEL_SIZE'])
        header_dict['phi_start'] = float(h['OSC_START'])
        header_dict['phi_width'] = float(h['OSC_RANGE'])
        header_dict['exposure_time'] = float(h['EXPOSURE TIME'])

    if header_dict:
        header_dict['beam_in_pixel'] = (header_dict['beam_in_mm'][0] / header_dict['pixel_size'][0],
                                        header_dict['beam_in_mm'][1] / header_dict['pixel_size'][1])
        header_dict['phi_end'] = header_dict['phi_start'] + header_dict['phi_width']
        header_dict['goniometer_is_vertical'] = False
        header_dict['nimgs'] = 1
        header_dict['oscillation'] = header_dict['phi_start'], header_dict['phi_width']
        header_dict['oscillation_axis'] = 'OMEGA'
    return header_dict


def extract_minicbf_header(image_file_name):
    with open(image_file_name, "rb") as cbf:
        bdata = cbf.read(2048)

    i1 = bdata.find(b"###CBF")
    i2 = bdata.find(b"X-Binary-Size-Padding") + 27
    lines = [line.decode('utf-8') for line in bdata[i1:i2].splitlines() if len(line) > 2]

    # artificial bug for debug for testing parallelize hanging problem
    # lines = [line for line in bdata[i1:i2].splitlines() if len(line) > 2]

    header_dict = dict()

    for l in lines:
        if l.startswith("# Detector:"):
            dclass, sn = l.split(':')[-1].split(',')[0:2]
            header_dict['detector_class'] = dclass.lstrip(" ")
            header_dict['detector'] = 'dectris'
            header_dict['serial_number'] = sn.lstrip(" ")

        if l.startswith("# Beam_xy"):
            header_dict['beam_in_pixel'] = tuple(map(float, getNumbers(l)))

        if l.startswith("# Detector_2theta"):
            header_dict['two_theta'] = float(getNumbers(l)[1])
        else:
            header_dict['two_theta'] = 0.0

        if l.startswith("# Detector_distance"):
            header_dict['distance'] = 1000 * float(getNumbers(l)[0])

        if l.startswith("# Exposure_time"):
            header_dict['exposure_time'] = float(getNumbers(l)[0])

        if l.startswith("# Wavelength"):
            header_dict['wavelength'] = float(getNumbers(l)[0])

        if l.startswith("# CdTe sensor, thickness"):
            header_dict['sensor'] = 1000 * float(getNumbers(l)[0])

        if l.startswith("# Silicon sensor, thickness"):
            header_dict['sensor'] = 1000 * float(getNumbers(l)[0])

        if l.startswith("# Pixel_size"):
            header_dict['pixel_size'] = tuple(map(lambda x: round(1000 * float(x), 4), getNumbers(l)))

        if l.startswith("# Count_cutoff"):
            header_dict['saturation'] = float(getNumbers(l)[0])

        if l.startswith("# Start_angle"):
            header_dict['phi_start'] = float(getNumbers(l)[0])

        if l.startswith("# Angle_increment"):
            header_dict['phi_width'] = float(getNumbers(l)[0])
            if header_dict['phi_width'] <= config.MIN_OSC_VALUE:
                header_dict['phi_width'] = config.MIN_OSC_VALUE

        if re.match(r'# (\d{4}-\d{2}-\d{2})T(\d{2}:\d{2}:\d{2}.\d+)', l):
            header_dict['date'] = l

        if l.startswith("X-Binary-Size-Fastest-Dimension:"):
            dim_fast = int(getNumbers(l)[0])

        if l.startswith("X-Binary-Size-Second-Dimension:"):
            dim_slow = int(getNumbers(l)[0])

    if 'date' not in header_dict:
        header_dict['date'] = ''

    if header_dict:
        header_dict['phi_end'] = header_dict['phi_start'] + header_dict['phi_width']
        header_dict['goniometer_is_vertical'] = False
        header_dict['nimgs'] = 1
        header_dict['oscillation'] = header_dict['phi_start'], header_dict['phi_width']
        header_dict['size'] = (dim_fast, dim_slow)
        header_dict['epoch'] = header_dict['date']
        header_dict['oscillation_axis'] = 'OMEGA'
        header_dict['beam_in_mm'] = (header_dict['beam_in_pixel'][0] * header_dict['pixel_size'][0],
                                     header_dict['beam_in_pixel'][1] * header_dict['pixel_size'][1])

    return header_dict


def extract_hdf5_header(master_file_name):
    logger.info("extract header information from master file: {}.".format(master_file_name))
    sensor_thickness = "/entry/instrument/detector/sensor_thickness"
    nimages = "/entry/instrument/detector/detectorSpecific/nimages"
    description = "/entry/instrument/detector/description"
    detector_number = "/entry/instrument/detector/detector_number"
    x_pixel_size = "/entry/instrument/detector/x_pixel_size"
    y_pixel_size = "/entry/instrument/detector/y_pixel_size"
    data_collection_date = "/entry/instrument/detector/detectorSpecific/data_collection_date"
    count_time = "/entry/instrument/detector/count_time"
    countrate_correction_count_cutoff = "/entry/instrument/detector/detectorSpecific/countrate_correction_count_cutoff"
    incident_wavelength = "/entry/instrument/beam/incident_wavelength"
    detector_distance = "/entry/instrument/detector/detector_distance"
    beam_center_x = "/entry/instrument/detector/beam_center_x"
    beam_center_y = "/entry/instrument/detector/beam_center_y"
    omega_range_average = "/entry/sample/goniometer/omega_range_average"
    two_theta = "/entry/instrument/detector/goniometer/two_theta_range_average"
    omega_start = "/entry/sample/goniometer/omega_start"
    omega = "/entry/sample/goniometer/omega"
    nx = "/entry/instrument/detector/detectorSpecific/x_pixels_in_detector"
    ny = "/entry/instrument/detector/detectorSpecific/y_pixels_in_detector"
    nbits = "/entry/instrument/detector/bit_depth_image"

    master_file = h5py.File(master_file_name)

    header_dict = dict()
    header_dict['two_theta'] = master_file[two_theta].value
    detector = master_file[description].value
    header_dict['size'] = master_file[nx].value, master_file[ny].value
    header_dict['serial_number'] = master_file[detector_number].value
    if b'EIGER 16M' in detector.upper():
        header_dict['detector_class'] = 'eiger 16M'
        header_dict['detector'] = 'dectris'
        # header['size'] = (4150, 4371)
        # header['serial_number'] = master_file[detector_number].value
    elif b'EIGER 9M' in detector.upper():
        header_dict['detector_class'] = 'eiger 9M'
        header_dict['detector'] = 'dectris'
        # header['size'] = (4150, 4371)
        # header['serial_number'] = master_file[detector_number].value
    else:
        logger.error("unknown detector, not implemented.")
        raise Exception("unrecognized detector")

    header_dict['distance'] = 1000 * master_file[detector_distance].value
    header_dict['exposure_time'] = master_file[count_time].value
    header_dict['oscillation_axis'] = get_oscillation_axis(master_file)
    header_dict['wavelength'] = master_file[incident_wavelength].value
    header_dict['pixel_size'] = (1000 * master_file[x_pixel_size].value,
                                 1000 * master_file[y_pixel_size].value)
    header_dict['sensor'] = 1000 * master_file[sensor_thickness].value
    try:
        header_dict['saturation'] = master_file[countrate_correction_count_cutoff].value
    except KeyError:
        logger.warning("saturation cutoff not set countrate_correction_count_cutoff.")
        header_dict['saturation'] = 2 ** 20

    header_dict['beam_in_mm'] = (master_file[beam_center_x].value * header_dict['pixel_size'][0],
                                 master_file[beam_center_y].value * header_dict['pixel_size'][1])
    header_dict['beam_in_pixel'] = (master_file[beam_center_x].value,
                                    master_file[beam_center_y].value)
    header_dict['date'] = master_file[data_collection_date].value
    header_dict['epoch'] = header_dict['date']
    header_dict['goniometer_is_vertical'] = False
    header_dict['nimgs'] = master_file[nimages].value
    header_dict['nbits'] = master_file[nbits].value

    if header_dict['oscillation_axis'] == 'OMEGA':
        try:
            header_dict['phi_start'] = master_file[omega_start].value
        except:
            logger.warning('omega may not be set, using default')
            header_dict['phi_start'] = master_file[omega].value[0]

        header_dict['phi_width'] = master_file[omega_range_average].value
        if header_dict['phi_width'] <= config.MIN_OSC_VALUE:
            header_dict['phi_width'] = config.MIN_OSC_VALUE

        header_dict['phi_end'] = header_dict['phi_start'] + header_dict['phi_width']
        header_dict['oscillation'] = header_dict['phi_start'], header_dict['phi_width']
    else:
        logger.error("the sample rotation axis is not omega, not implemented yet")

    header_dict['data_files'] = sorted(master_file['/entry/data'].keys())
    # header_dict['data'] = master_file['/entry/data'].values()

    return header_dict


def get_oscillation_axis(master_file_name):
    if master_file_name['/entry/sample/goniometer/omega_range_total'].value > 0:
        return 'OMEGA'
    if master_file_name['/entry/sample/goniometer/phi_range_total'].value > 0:
        return 'PHI'
    if master_file_name['/entry/sample/goniometer/kappa_range_total'].value > 0:
        return 'KAPPA'
    if master_file_name['/entry/sample/goniometer/chi_range_total'].value > 0:
        return 'CHI'
    return 'OMEGA'


def get_point_resolution_on_detector(point_xy, beam_xy, pixel_size, wavelength, distance, detector_tilt):
    """ calculate resolution on a point (in pixels) on the detector, given 
    pixel_size (mm), beam_xy in pixels, wavelength (A), distance (mm) 
    tan (2*theta) = radius / distance, d = wavelengh / 2*sin(theta),
    !!!for detector_tilt, we assume a particular configure that is adopted by GMCA!!!
    """
    if abs(distance) < 0.0001:
        return None
    tilt = detector_tilt * math.pi / 180.0
    beamy = beam_xy[1] + distance * math.atan(tilt) / pixel_size

    dist = math.sqrt((point_xy[0] - beam_xy[0]) ** 2 +
                     (point_xy[1] - beamy) ** 2)

    radius = dist * pixel_size
    theta = math.atan(radius / distance)
    return wavelength / (2.0 * math.sin(0.5 * (theta + tilt)))


def get_max_resolution_on_detector(size_xy, beam_xy, pixel_size, wavelength, distance, detector_tilt=0,
                                   edge_only=False):
    """ return maximum edge and corner resolution for a detector given diffraction parameter """
    if distance < 0.001 or wavelength < 0.001:
        print(f"Warning: distance {distance} wavelength {wavelength}")
        return 0.001, 0.001

    nx, ny = size_xy
    edges = [(nx // 2, 0), (0, ny // 2), (nx // 2, ny), (nx, ny // 2)]
    max_edge_reso = min(
        get_point_resolution_on_detector(
            p, beam_xy,
            pixel_size,
            wavelength,
            distance,
            detector_tilt) for p in edges)

    if not edge_only:
        corners = [(0, 0), (nx, 0), (0, ny), (nx, ny)]
        max_corner_reso = min(
            get_point_resolution_on_detector(
                p, beam_xy,
                pixel_size,
                wavelength,
                distance,
                detector_tilt) for p in corners)
    else:
        max_corner_reso = -1
    return max_edge_reso, max_corner_reso


if __name__ == "__main__":
    #    h1 = header('idd-test.cbf')
    #    import pprint; pprint.pprint(h1.get_header_as_dict())
    #
    #    #h2 = header('lyso_7_1_master.h5')
    #    #pprint.pprint(h2.get_header_as_dict())
    #
    #    h3 = header('a_cbf.img')
    #    pprint.pprint(h3.get_header_as_dict())
    #
    #    # the same image as a_cbf.img
    #    h4 = header('AEF_Se_MalA_1_2_1_INV_000206.cbf')
    #    pprint.pprint(h4.get_header_as_dict())
    #
    #
    #    h5 = header('/home/qxu/test-data/16eiger-2/2015_11_10/insu6_1_master.h5')
    #    pprint.pprint(h5.get_header_as_dict())
    #
    #    h2 = header('lyso_7_1_master.h5')
    #    pprint.pprint(h2.get_header_as_dict())
    #
    #    print extract_hdf5_header('/home/qxu/test-data/B3-spg-check/B3_1_0_91_scr_master.h5')
    #    h3 = header('/home/qxu/test-data/B3-spg-check/B3_1_0_91_scr_master.h5')
    #    print h3.get_header_as_dict()['two_theta']
    #
    #
    # h4 = Header('/home/qxu/data-for-tests/jcsg-p3-pilatus-6m/SP13538A_4qrl/196603_1_E1_00001.cbf')
    # print(h4.get_header_as_dict())
    h = Header('/mnt/beegfs/qxu/eiger4mCDTE_000001.cbf')
    print(h.get_header_as_dict())
