from __future__ import absolute_import
from collections import namedtuple
import logging

from . import smartie
from .plots import make_plot_vs_resolution
from .fileopener import FileOpener
from ..common.utils import getNumbers

logger = logging.getLogger(__name__)


class AimlessParser(FileOpener):
    def __init__(self, wdir=".", filename='aimless.log'):
        super(AimlessParser, self).__init__(wdir, filename)
        self.smartie = smartie.parselog(self.logfile)

    def get_resolution_estimation(self, criterion="cchalf~30%"):
        signature = "from half-dataset correlation CC(1/2) >  0.30: limit ="
        if criterion == "cchalf~30%":
            signature = "from half-dataset correlation CC(1/2) >  0.30: limit ="
        elif criterion == "<i/sigmai>~1.5":
            signature = "from Mn(I/sd) >  1.50:                         limit ="
        elif criterion == "<i/sigmai>~2.0":
            signature = "from Mn(I/sd) >  2.00:                         limit ="

        lines = self.parser.get_lines(startIndex=-500)
        reslimit = []
        for line in lines:
            if signature in line:
                res = line.split("=")[1]
                res = res[:res.find("A")].replace(" ", "")
                if float(res) > 0.2:
                    reslimit.append(res)
                else:
                    logger.warning("maximum resolution gt than {}, ignored".format(res))

        return min([float(r) for r in reslimit]) if reslimit else 0.0

    def summary(self):
        # return smartie.summarise(smartie.parselog(self.logfile))
        return smartie.parselog(self.logfile).summary(0).retrieve()

    def get_table1(self, title=True, output='text'):
        """ extract the final summary table (aka table 1) """
        summary = self.summary().split('\n')

        start = 3
        dataset_keys = list()
        for i, line in enumerate(summary):
            if "Project:" in line:
                dataset_keys.append(line.split(':')[-1].replace(" ", ""))
            if "Overall   Inner   Outer" in line or "Overall  InnerShell  OuterShell" in line:
                start = i
                break

        if not title:
            start += 1

        table1 = summary[start:start + 22]
        ndatasets = len(dataset_keys)
        if ndatasets > 1:
            logger.warning("found more than one datasets in the file!")

        if output == 'text':
            return "\n".join(table1)
        elif output == 'list':
            return table1
        elif output == 'namedtuple':
            table1row = namedtuple('TABLE1', "paraName, Overall,  InnerShell,  OuterShell")
            ndtable1 = []
            for line in table1:
                values = line.rsplit()
                if len(values) >= 4:
                    parameter_name, cols = '_'.join(values[:-3]), values[-3:]
                    row = table1row(parameter_name, *cols)
                    ndtable1.append(row)
            return ndtable1
        else:
            logger.warning("unknown output format defined in aimless_parser.get_table1")
            return None

    def get_cell_space_group(self):
        """ extract cell and spg from the end of the summary table"""
        summary = self.summary().split('\n')
        result = {}
        for line in summary[-10:]:
            if 'Average unit cell:' in line:
                result['unitcell'] = line.split(':')[-1]
            elif 'Space group:' in line:
                result['spacegroup'] = line.split(':')[-1].replace(" ", "")
        return result

    def plot(self, prefix=None):
        prefix = '' if not prefix else prefix

        logfile = smartie.parselog(self.logfile)
        table = logfile.tables(select_title='Analysis against resolution')[0]
        one_d2 = table.col("1/d^2")
        Rmerge = table.col("Rmrg")
        Rmeas = table.col("Rmeas")
        avgIsigI = table.col("Mn(I/sd)")
        make_plot_vs_resolution(self.wdir,
                                one_d2,
                                Rmerge,
                                ylabel="Rmerge",
                                outfig=prefix + "rmergevsresolution.png",
                                markers=[0.5])
        make_plot_vs_resolution(self.wdir,
                                one_d2,
                                avgIsigI,
                                ylabel=r"mean I/$\sigma$I",
                                outfig=prefix + "isigmavsresolution.png",
                                markers=[2.0])
        make_plot_vs_resolution(self.wdir,
                                one_d2,
                                Rmeas,
                                ylabel="Rmeas",
                                outfig=prefix + "rmeasvsresolution.png")

        table = logfile.tables(select_title='Completeness & multiplicity v. resolution')[0]
        one_d2 = table.col("1/d^2")
        completeness = table.col("%poss")
        multiplicity = table.col("Mlplct")
        make_plot_vs_resolution(self.wdir,
                                one_d2,
                                completeness,
                                ylabel="Completeness (%)",
                                outfig=prefix + "completevsresolution.png",
                                markers=[50.0])
        make_plot_vs_resolution(self.wdir,
                                one_d2,
                                multiplicity,
                                ylabel="Multiplicity",
                                outfig=prefix + "multivsresolution.png",
                                markers=[2.0])

        table = logfile.tables(select_title='Correlations CC(1/2) within dataset')[0]
        one_d2 = table.col("1/d^2")
        cchalf = table.col("CC1/2")
        make_plot_vs_resolution(self.wdir,
                                one_d2,
                                cchalf,
                                ylabel="CC1/2",
                                outfig=prefix + "cchalfvsresolution.png",
                                markers=[.30])

    def summarize(self):
        table1 = self.get_table1(title=False, output='namedtuple')

        # gather information for database, take the row assuming I+ = I-
        db_dict = {'table1': self.get_table1(title=True)}
        db_dict.update(self.get_cell_space_group())

        for e in table1:
            if "Low" in e.paraName and "resolution" in e.paraName:
                db_dict['lowresolution'] = e.Overall
                db_dict['lowresolution_inner'] = e.InnerShell
                db_dict['lowresolution_outer'] = e.OuterShell
                continue

            if "High" in e.paraName and "resolution" in e.paraName:
                db_dict['highresolution'] = e.Overall
                db_dict['highresolution_inner'] = e.InnerShell
                db_dict['highresolution_outer'] = e.OuterShell
                continue

            if "Rmerge" in e.paraName and "all" in e.paraName:
                db_dict['rmerge'] = e.Overall
                db_dict['rmerge_inner'] = e.InnerShell
                db_dict['rmerge_outer'] = e.OuterShell
                continue

            if "Rmerge" in e.paraName and "within" in e.paraName:
                db_dict['anom_rmerge'] = e.Overall
                db_dict['anom_rmerge_inner'] = e.InnerShell
                db_dict['anom_rmerge_outer'] = e.OuterShell
                continue

            if "Rmeas" in e.paraName and "all" in e.paraName:
                db_dict['rmeas'] = e.Overall
                db_dict['rmeas_inner'] = e.InnerShell
                db_dict['rmeas_outer'] = e.OuterShell
                continue

            if "Rpim" in e.paraName and "all" in e.paraName:
                db_dict['rpim'] = e.Overall
                db_dict['rpim_inner'] = e.InnerShell
                db_dict['rpim_outer'] = e.OuterShell
                continue

            if "Mean((I)/sd(I))" in e.paraName:
                db_dict['isigmai'] = e.Overall
                db_dict['isigmai_inner'] = e.InnerShell
                db_dict['isigmai_outer'] = e.OuterShell
                continue

            if e.paraName.startswith("Completeness"):
                db_dict['completeness'] = e.Overall
                db_dict['completeness_inner'] = e.InnerShell
                db_dict['completeness_outer'] = e.OuterShell
                continue

            if "Multiplicity" in e.paraName:
                db_dict['multiplicity'] = e.Overall
                db_dict['multiplicity_inner'] = e.InnerShell
                db_dict['multiplicity_outer'] = e.OuterShell
                continue

            if "Mn(I)" in e.paraName and "CC(1/2)" in e.paraName:
                db_dict['cchalf'] = e.Overall
                db_dict['cchalf_inner'] = e.InnerShell
                db_dict['cchalf_outer'] = e.OuterShell
                continue

            if "Anomalous" in e.paraName and "completeness" in e.paraName:
                db_dict['anom_completeness'] = e.Overall
                db_dict['anom_completeness_inner'] = e.InnerShell
                db_dict['anom_completeness_outer'] = e.OuterShell
                continue

            if "Anomalous" in e.paraName and "multiplicity" in e.paraName:
                db_dict['anom_multiplicity'] = e.Overall
                db_dict['anom_multiplicity_inner'] = e.InnerShell
                db_dict['anom_multiplicity_outer'] = e.OuterShell
                continue

            if "DelAnom" in e.paraName and "correlation" in e.paraName:
                db_dict['anom_cchalf'] = e.Overall
                db_dict['anom_cchalf_inner'] = e.InnerShell
                db_dict['anom_cchalf_outer'] = e.OuterShell
                continue

            if "Total" in e.paraName and "observations" in e.paraName:
                db_dict['Nobs'] = e.Overall
                db_dict['Nobs_inner'] = e.InnerShell
                db_dict['Nobs_outer'] = e.OuterShell
                continue

            if "Total" in e.paraName and "unique" in e.paraName:
                db_dict['Nuniq'] = e.Overall
                db_dict['Nuniq_inner'] = e.InnerShell
                db_dict['Nuniq_outer'] = e.OuterShell
                continue

        db_dict["data_quality"] = AimlessParser.evaluate_data_quality(db_dict)
        db_dict['anomalous_signal'] = self.evaluate_anomalous_signal()

        # report to log, needs to rewritten
        result = {'resolution_range': (table1[0].Overall, table1[1].Overall),
                  'Rmerge': table1[3].Overall,
                  'Rmeas': table1[5].Overall,
                  'Rpim': table1[7].Overall,
                  'Nobs': table1[9].Overall,
                  'Nuniq': table1[10].Overall,
                  'I/sigmaI': table1[11].Overall,
                  'CChalf': table1[12].Overall,
                  'Completeness': table1[13].Overall,
                  'Multiplicity': table1[14].Overall,
                  'h_resolution_range': (table1[0].OuterShell, table1[1].OuterShell),
                  'h_Rmerge': table1[3].OuterShell,
                  'h_Rmeas': table1[5].OuterShell,
                  'h_Rpim': table1[7].OuterShell,
                  'h_Nobs': table1[9].OuterShell,
                  'h_Nuniq': table1[10].OuterShell,
                  'h_I/sigmaI': table1[11].OuterShell,
                  'h_CChalf': table1[12].OuterShell,
                  'h_Completeness': table1[13].OuterShell,
                  'h_Multiplicity': table1[14].OuterShell}

        result.update(self.get_cell_space_group())

        # print summary string
        text = (
            "\n**Final results (assuming that I+=I-):\n"
            "  Data within {resolution_range} A scaled in space group spg={spg} \n"
            "  Cell={cell}\n"
            "  Rmerge={Rmerge} ({h_Rmerge}), Rmeas={Rmeas} ({h_Rmeas}), Rpim={Rpim} ({h_Rpim}) \n"
            "  Nobs={Nobs} ({h_Nobs}), Nuniq={Nuniq} ({h_Nuniq}), \n"
            "  Completeness={Completeness}% ({h_Completeness}%), <I/sigmaI>={isigi} ({h_isigi}), \n"
            "  <Multiplicity>={multiplicity} ({h_multiplicity}), <CC1/2>={CChalf} ({h_CChalf}), \n"
            "  values in () represent high resolution bin defined as {h_resolution_range} A. \n"
        ).format(
            resolution_range='-'.join(result['resolution_range']),
            spg=result['spacegroup'],
            cell=result['unitcell'],
            Rmerge=result['Rmerge'],
            h_Rmerge=result['h_Rmerge'],
            Rmeas=result['Rmeas'],
            h_Rmeas=result['h_Rmeas'],
            Rpim=result['Rpim'],
            h_Rpim=result['h_Rpim'],
            Nobs=result['Nobs'],
            h_Nobs=result['h_Nobs'],
            Nuniq=result['Nuniq'],
            h_Nuniq=result['h_Nuniq'],
            Completeness=result['Completeness'],
            h_Completeness=result['h_Completeness'],
            isigi=result['I/sigmaI'],
            h_isigi=result['h_I/sigmaI'],
            multiplicity=result['Multiplicity'],
            h_multiplicity=result['h_Multiplicity'],
            CChalf=result['CChalf'],
            h_CChalf=result['h_CChalf'],
            h_resolution_range='-'.join(result['h_resolution_range']),
        )

        logger.info(text)

        return db_dict

    def evaluate_anomalous_signal(self):
        """ aimless/scala using the following methods to determine whether there is significant anomalous signal
              1) Anomplot slope > anomslopethreshold (1.3)
              2) CCanom > anomCCthreshold (0.3) in more than anomNbinthreshold bins (2), or overall,
                   or interdataset CCanom
              3) RCRanom > anomRCRthreshold (1.3) in more than anomNbinthreshold bins (2), or overall
        return 0: weak anomalous signal
        return num: strong anomalous signal, suggested resolution cutoff for anomalous signal
        return 10: strong anomalous signal but no cutoff found, shouldn't happen!
        """
        str1 = "strong anomalous signal found"
        str1a = "Anomalous flag switched OFF in input but there appears to be a significant anomalous signal"
        str2 = "Estimate of the resolution limit for a significant anomalous signal "

        anomalous = self.parser.get_lines_contain(str1)
        anomalous2 = self.parser.get_lines_contain(str1a)
        if anomalous2:
            logger.warning("There is anomalous signal, but the data is be scaled as native, this is probably not optimal.")

        if anomalous + anomalous2:
            lines = self.parser.get_lines_contain(str2)
            if lines:
                numbers = getNumbers(lines[0])
                logger.info("anomalous signal resolution cutoff: {}A".format(numbers[0]))
                return numbers[0] if numbers else 10
            return 10
        return 0

    @staticmethod
    def evaluate_data_quality(data_dict):
        return True
        resol = data_dict.get('highresolution')
        completeness = data_dict.get('completeness')
        rmerge = data_dict.get('rmerge')

        try:
            if float(rmerge) < 0.3 and float(resol) < 3.5 and float(completeness) > 90:
                return True
        except Exception:
            return False

        return False

if __name__ == '__main__':
    '''
    # ap = AimlessParser(wdir='/home/qxu/work/quickProcess/quickProcess/tests', filename='aimless.3wav.log')
    # print(ap.summary())
    # print(ap.summarize())
    #print(ap.get_table1())
    ap = AimlessParser(wdir='/mnt/beegfs/qxu/bluice/tools/scripts/quickProcess/quickProcess/tests/',
                        filename='A2_2_w1.0332-aimless-TRIMDATA.log')
    #print(ap.get_table1())
    #print(ap.summarize())
    
    ap = AimlessParser(wdir="/mnt/beegfs/qxu/bluice/tools/scripts/quickProcess/quickProcess/tests/",
                       filename="178832_1_w0.9794-aimless-TRIMDATA.log")
    '''

    ap = AimlessParser(wdir="/mnt/beegfs/qxu/23BM_2017_08_08/SP13538A_4qrl/process1/repo/xia2_dials_aimless/LogFiles",
                       filename="xia2_dials_aimless_p_196603__aimless.log")

    from pprint import pprint
    pprint(ap.summarize())
