from __future__ import absolute_import
import logging
import os

from ..common import utils
from .fileopener import FileOpener
from .table import Table
from .plots import make_plot_vs_resolution

logger = logging.getLogger(__name__)


class CorrectParser(FileOpener):
    def __init__(self, wdir=".", filename='CORRECT.LP'):
        super(CorrectParser, self).__init__(wdir, filename)

        if self.OK and self.get_ISa() and self.get_ISa() > 1:
            self.OK = True
        else:
            logger.warning("ISa value is low for {}/{}".format(self.wdir, filename))
            self.OK = False
        self.summary = None

    def get_point_group_cell_orientation(self):
        result = {}
        lines = self.parser.get_lines_after("SELECTED SPACE GROUP AND UNIT CELL FOR THIS DATA SET",
                                            nLines=6)
        spg = list(map(int, utils.getNumbers(lines[1])))[0]
        cell = list(map(float, utils.getNumbers(lines[2])))
        unit_cell_a_axis = list(map(float, utils.getNumbers(lines[3])))
        unit_cell_b_axis = list(map(float, utils.getNumbers(lines[4])))
        unit_cell_c_axis = list(map(float, utils.getNumbers(lines[5])))
        result['cell'] = cell
        result['point_group_number'] = spg
        result['unit_cell_axes'] = (unit_cell_a_axis,
                                 unit_cell_b_axis,
                                 unit_cell_c_axis)
        result['xdsinp_update_string'] = ";".join(lines[1:])
        return result

    def get_refined_point_group_cell_orientation(self):
        result = {}
        lines = self.parser.get_lines_around("REFINEMENT OF DIFFRACTION PARAMETERS USING ALL IMAGES",
                                             offset=4,
                                             nLines=19)
        result['lines'] = lines
        result['point_group_number'] = int(lines[4].split()[-1])
        result['cell'] = list(map(float, lines[5].split()[-6:]))
        unit_cell_a_axis = list(map(float, lines[8].split()[-3:]))
        unit_cell_b_axis = list(map(float, lines[9].split()[-3:]))
        unit_cell_c_axis = list(map(float, lines[10].split()[-3:]))
        result['unit_cell_axes'] = (unit_cell_a_axis, unit_cell_b_axis, unit_cell_c_axis)
        result['xdsinp_update_string'] = ('SPACE_GROUP_NUMBER={spg};UNIT_CELL_CONSTANTS={cell}'
                                        ';UNIT_CELL_A-AXIS={ax};UNIT_CELL_B-AXIS={bx};UNIT_CELL_C-AXIS={cx}').format(
                                        spg=result['point_group_number'],
                                        cell=' '.join(map(str, result['cell'])),
                                        ax=' '.join(map(str, unit_cell_a_axis)),
                                        bx=' '.join(map(str, unit_cell_b_axis)),
                                        cx=' '.join(map(str, unit_cell_c_axis)),)
        return result

    def get_REMOVE_HKL(self, filename='REMOVE.HKL', mode='a', max_res=None):
        """ create REMOVE.HKL in the wdir, return 0 if no aliens or number of aliens written"""
        aliens = self.parser.get_lines_contain('alien')

        rejected_hkl = []
        for i, a in enumerate(aliens):
            items = a.split()
            z = float(items[4])
            res = float(items[3])
            if z >= 20 or (max_res and res < max_res):
                rejected_hkl.append(a)

        if rejected_hkl:
            remove_hkl = os.path.join(self.wdir, filename)
            with open(remove_hkl, mode) as fh:
                fh.write(''.join(rejected_hkl))
            return len(rejected_hkl)
        else:
            return 0

    def get_progress_tables(self):
        titles = self.parser.get_lines_contain('SUMMARY OF DATA SET STATISTICS FOR IMAGE   DATA_RANGE=')
        titles.append(self.parser.get_lines_contain('STATISTICS OF SAVED DATA SET'))
        textblocks = self.parser.get_blocks_between("RESOLUTION     NUMBER OF REFLECTIONS    COMPLETENESS",
                                                    "total",
                                                    iMatch=-1)
        tables = []
        for i, b in enumerate(textblocks):
            tbl = Table()
            tbl.set_title(''.join(titles[i]).strip('\n'))
            tbl.set_table_by_lines(b, header=True)
            tables.append(tbl)
        return tables

    def get_final_table(self):
        return self.get_progress_tables()[-1]

    def get_correlation_to_reference_hkl(self):
        anchor1 = 'CORRELATIONS BETWEEN INPUT DATA SETS AFTER CORRECTIONS'
        anchor2 = 'Factor applied to intensities'
        lines = self.parser.get_blocks_between(anchor1, anchor2, skip1=5, skip2=3)
        if lines and lines[0]:
            line = lines[0][0]
        else:
            return None

        res = {}
        fields = line.split()
        res['correlation'] = float(fields[3])
        res['common_reflections'] = int(fields[2])
        return res

    def get_ISa(self):
        try:
            ISa = self.parser.get_lines_after('a        b          ISa', nLines=1)[0].split()[-1]
            return float(ISa)
        except TypeError:
            logger.error("failed to extract ISa in {}/{}".format(self.wdir, self.filename))
            return None

    def get_resolution_estimation(self, target_name='CC(1/2)', target_value=30, interpolate=True):
        # alternative: target_name="I/SIGMA", target_value=0.1
        """will try to get high resol estimation by interpolation, overestimate a bit otherwise"""
        last_table = self.get_final_table()
        # get the resolution and cc column and skip the last row (total)
        res = last_table.get_table_col_by_name('RESOL')[:-1]
        target_col = last_table.get_table_col_by_name(target_name)[:-1]
        res = list(map(float, res))[::-1]
        target_col = list(map(float, [el.replace("*", "") for el in target_col]))[::-1]

        if all([e < target_value for e in target_col]):
            logger.warning("the data processing is likely failed since CC value in all resolution shells are low")
            return res[-1]

        compare_to_target = [e >= target_value for e in target_col]
        if all(compare_to_target):
            logger.warning("the data resolution might be improved (e.g. the detector is moved closer to crystal?)")
            return res[0]

        # look for index with False, False, True or False, True, True
        for i in range(len(compare_to_target) - 2):
            if not compare_to_target[i] and compare_to_target[i+1] and compare_to_target[i+2]:
                position = i
                if interpolate:
                    x1, y1 = res[position], target_col[position]
                    x2, y2 = res[position + 1], target_col[position + 1]
                    slope = (y2 - y1) / (x2 - x1)
                    intercept = y2 - slope * x2
                    return round((target_value - intercept) / slope, 2)
                else:
                    return res[position]

        # otherwise just return the first one with desired cc value
        for i, v in enumerate(compare_to_target):
            if v:
                return res[i]

    def get_summary(self):
        result = {}
        ISa = self.get_ISa()
        estimated_resolution = self.get_resolution_estimation()
        final_table = self.get_final_table()
        res = final_table.get_table_col_by_name('RESOLUTION')[:-1]  # skip last element since it correspnds to overall
        one_d2 = [round(1.0 / float(r) ** 2, 2) for r in res]

        result['ISa'] = ISa
        result['estimated_resolution'] = estimated_resolution
        result['final_table'] = final_table

        overall = [float(utils.getNumbers(v)[0]) for v in final_table.get_table_row_by_index(-1)[1:]]
        result['multiplicity'] = round(overall[0] / overall[1], 2)
        result['Nobs'] = int(overall[0])
        result['Nuniq'] = int(overall[1])
        result['Completeness'] = overall[3]
        result['Rmerge'] = overall[4]
        result['I/SigmaI'] = overall[7]
        result['maxRes'] = res[-1]
        result['one_d2'] = one_d2

        spg_cell = self.get_point_group_cell_orientation()
        result['spg'] = spg_cell['point_group_number']
        result['cell'] = ' '.join(str(v) for v in spg_cell['cell'])
        self.summary = result
        return self.summary

    def summarize(self, plot=False):
        prefix = ''
        if not self.summary:
            self.get_summary()
        final_table = self.summary['final_table']
        ISa = self.summary['ISa']
        one_d2 = self.summary['one_d2']

        if plot:
            Rmerge = final_table.get_table_col_by_name('R-FACTOR')[:-1]
            Rmerge = [float(utils.getNumbers(v)[0]) for v in Rmerge]
            Rmerge = [v if 0. < v < 300. else 300. for v in Rmerge]

            make_plot_vs_resolution(self.wdir,
                                    one_d2,
                                    Rmerge,
                                    ylabel="Rmerge (%)",
                                    outfig=prefix + "Rmerge_vs_resolution.png",
                                    markers=[50])

            completeness = [float(utils.getNumbers(v)[0])
                            for v in final_table.get_table_col_by_name('COMPLETENESS')[:-1]]
            make_plot_vs_resolution(self.wdir,
                                    one_d2,
                                    completeness,
                                    ylabel="Completeness (%)",
                                    outfig=prefix + "completeness_vs_resolution.png",
                                    markers=[50.0])

            avgIsigI = [float(utils.getNumbers(v)[0])
                        for v in final_table.get_table_col_by_name('I/SIGMA')[:-1]]
            make_plot_vs_resolution(self.wdir,
                                    one_d2,
                                    avgIsigI,
                                    ylabel=r"mean I/$\sigma$I",
                                    outfig=prefix + "isigma_vs_resolution.png",
                                    markers=[2.0])

            cchalf = [float(utils.getNumbers(v)[0])
                      for v in final_table.get_table_col_by_name('CC(1/2)')[:-1]]
            make_plot_vs_resolution(self.wdir,
                                    one_d2, cchalf,
                                    ylabel="CC1/2 (%)",
                                    outfig=prefix + "ccHalf_vs_resolution.png",
                                    markers=[.30])

            anomal_corr = [float(utils.getNumbers(v)[0])
                           for v in final_table.get_table_col_by_name('Anomal')[:-1]]
            make_plot_vs_resolution(self.wdir,
                                    one_d2,
                                    anomal_corr,
                                    ylabel="Anomal Correlation (%)",
                                    outfig=prefix + "anomalCorr_vs_resolution.png",
                                    markers=[.30])

        if ISa < 10:
            logger.warning(
                "ISa value is low, for a good crystal, this may indicate "
                "strong systematic errors (e.g. BL, wrong parameters etc).")

        return CorrectParser.print_result_overview(self.summary)

    @staticmethod
    def print_result_overview(result):
        text = (
            "  Images integrated to {maxRes} A in space group spg={spg} \n"
            "  Cell={cell}\n"
            "  with ISa={ISa}, \n"
            "  Rmerge={Rmerge}%, Nobs={Nobs}, Nuniq={Nuniq}, \n"
            "  Completeness={Completeness}%, <I/sigmaI>={isigi}, \n"
            "  <Multiplicity>={multiplicity} \n"
            "  Resolution is estimated to be {estimated_resolution} A \n"
        ).format(
            maxRes=result['maxRes'],
            spg=result['spg'],
            cell=result['cell'],
            ISa=result['ISa'],
            Rmerge=result['Rmerge'],
            Nobs=result['Nobs'],
            Nuniq=result['Nuniq'],
            Completeness=result['Completeness'],
            isigi=result['I/SigmaI'],
            multiplicity=result['multiplicity'],
            estimated_resolution=result['estimated_resolution'],
        )
        return text


if __name__ == '__main__':
    p = CorrectParser(wdir='/mnt/share/qxu/bluice/tools/scripts/quickProcess/./quickProcess/misc/mpi-xds/test/ref-results/')
    result = p.get_resolution_estimation()
    print(result)

    result = p.get_resolution_estimation(target_name="I/SIGMA", target_value=0.5)
    print("i/sigmai=0.1 cutoff = {}".format(result))

    p= CorrectParser(wdir='/mnt/share/qxu/23BM_2017_08_08/rlefkowitz/process/gmcaproc/C8_4_00001_00050/XDS-p1/')
    print(p.get_correlation_to_reference_hkl())