from __future__ import absolute_import

import os
import logging
import json
from operator import itemgetter

from ..common.spacegroup import Spacegroup
from .fileopener import FileOpener
from ..common.utils import getNumbers

logger = logging.getLogger(__name__)


def loadjson(jsonfile):
    if os.path.isfile(jsonfile):
        return json.load(open(jsonfile, 'r'))
    else:
        logger.error('failed to open json file {}'.format(jsonfile))
        return None


class DialsParser(FileOpener):
    def __init__(self, wdir=".", filename='dials.sh.out'):
        super(DialsParser, self).__init__(wdir, filename)

        self.datablock = loadjson(os.path.join(self.wdir, 'imported.expt'))
        self.spot_stat = loadjson(os.path.join(self.wdir, 'spot_stat.json'))
        self.experiments = loadjson(os.path.join(self.wdir, 'indexed.expt'))
        self.bravais_summary = loadjson(os.path.join(self.wdir, 'bravais_summary.json'))
        self.refined_experiments = None

    def set_refined_experiments(self):
        #  this is generated in dials2.sh
        self.refined_experiments = loadjson(os.path.join(self.wdir, 'refined.expt'))

    def get_datablock(self):
        return self.datablock

    def get_experiments(self):
        return self.experiments

    def get_bravais_summary(self, solNum=None):
        if solNum:
            return self.bravais_summary[solNum]

        return self.bravais_summary

    def get_spot_stat(self):
        lines = self.parser.get_lines_after('Overall statistics', nLines=7)
        reso = n_spots = ""
        for line in lines:
            if "#spots " in line:
                n_spots = getNumbers(line)[-1]

            if 'd_min     ' in line:
                reso = getNumbers(line)[-1]

        return {'n_spots': n_spots, 'resolution_from_spots':reso}

    def get_index_table(self):
        lines = self.parser.get_blocks_between("Chiral space groups corresponding to each Bravais lattice:",
                                               "* = recommended solution")[0]
        return ''.join(lines)

    def get_possible_space_groups(self, sgtype='spacegroup'):
        if not self.bravais_summary:
            return None

        possible_space_groups = []
        for sol in self.bravais_summary:
            if self.bravais_summary[sol]['recommended']:
                lattice = self.bravais_summary[sol]['bravais']
                rmsd = "{}".format(round(self.bravais_summary[sol]['rmsd'], 3))
                cell = ','.join([str(round(n, 1)) for n in self.bravais_summary[sol]['unit_cell']])

                if sgtype.startswith('space'):
                    pgs = Spacegroup.get_space_groups_from_lattice(lattice, sgtype=sgtype, output_mode='symbol')
                else:
                    pgs = Spacegroup.get_space_groups_from_lattice(lattice, sgtype='pointgroup', output_mode='symbol')

                for spg in pgs:
                    # the first value is lattice order used for sorting purpose
                    tpl = [Spacegroup.LATTICE[lattice][0], ' '.join([spg, 'lattice:', lattice, 'rmsd:', rmsd, 'cell:', cell])]
                    possible_space_groups.append(tpl)

        solutions = sorted(possible_space_groups, key=itemgetter(0), reverse=True)
        return ';'.join(x[1] for x in solutions)

    @property
    def choose_best(self):
        # find the highest lattice among recommended solution
        if not self.bravais_summary:
            return None

        d = {}
        for sol in self.bravais_summary:
            if self.bravais_summary[sol]['recommended']:
                lattice = self.bravais_summary[sol]['bravais']
                d[sol] = Spacegroup.LATTICE[lattice][0]

        solNum = max(d, key=d.get)  # return key with max value
        return solNum
        # return self.get_bravais_summary(solNum)

    def choose_symm(self, space_group):
        if not self.bravais_summary:
            return None

        # choose a solution based on the given symmetry, multiple solution may arise
        lattice = Spacegroup.get_lattice_from_spacegroup(space_group)

        # deal with C2/I2 case
        if lattice in ['mC', 'mI']:
            lattice = ['mC', 'mI']

        results = list()
        for sol in self.bravais_summary:
            if self.bravais_summary[sol]['recommended']:
                if self.bravais_summary[sol]['bravais'] in lattice:
                    results.append(sol)

        if len(results) > 1:
            logger.warning('more than one results consistent with spg, last one will be used')
        elif not results:
            logger.error('no results consistent with spg')
            return None

        return results[-1]


if __name__ == '__main__':
    p = DialsParser('/tmp/qxu/j6/D1_2_0_scr_00061_00151/dials_strategy/')
    from pprint import pprint
    #pprint(p.get_bravais_summary(solNum='9'))
    print(p.choose_best)
    print(p.choose_symm('i422'))