from __future__ import absolute_import

import logging
import os

from ..common import utils
from .fileparser import FileParser

logger = logging.getLogger(__name__)


class FileOpener(object):
    def __init__(self, wdir=".", filename=None):
        self.wdir = os.path.abspath(wdir)
        filename = os.path.join(self.wdir, filename)
        self.filename = filename
        self.logfile = filename
        if os.path.isfile(filename):
            self.parser = FileParser(filename)
        else:
            logger.error("file {} does not exist.".format(filename))
            raise Exception("file {} does not exist.".format(filename))

        self.error = None
        if filename.endswith('.LP'):
            self.error = self.__xds_error()
        elif filename.endswith('.log'):
            self.error = self.__ccp4_error()

        self.OK = True if not self.error else False

    def check_error(self):
        return self.error

    def __xds_error(self):
        """check for errors near the end of file."""
        colspot_error = '!!! ERROR !!! INSUFFICIENT NUMBER OF SPOTS.'
        unresolved_error = '!!! ERROR !!! ASSERT VIOLATION IN "RESTHKL"'
        idxref_error1 = '!!! ERROR !!! SOLUTION IS INACCURATE'
        idxref_error2 = '!!! ERROR !!! INSUFFICIENT PERCENTAGE (< 50%) OF INDEXED REFLECTIONS'
        idxref_error3 = '!!! ERROR !!! INSUFFICIENT NUMBER OF ACCEPTED SPOTS.'
        idxref_error4 = '!!! ERROR IN REFINE !!! RETURN CODE IS IER=           0'

        idxref_warning = '!!! WARNING !!! REFINEMENT DID NOT CONVERGE'
        integ_error1 = '!!! ERROR !!! AUTOMATIC DETERMINATION OF SPOT SIZE PARAMETERS HAS FAILED.'
        integ_error2 = '!!! ERROR !!! CANNOT ALLOCATE MEMORY'

        correct_error1 = '!!! ERROR !!! CANNOT INTERPRETE LATTICE BY THE GIVEN UNIT CELL'

        endblock = self.parser.get_lines(startIndex=-100)
        for line in endblock:
            if unresolved_error in line:
                logger.error("unhandled XDS error, contact qxu@anl.gov,\nmessage by XDS: {}"
                             .format(unresolved_error))
                raise SystemExit(1)

            if idxref_error1 in line or idxref_error2 in line:
                error = idxref_error1 if idxref_error1 in line else idxref_error2
                message = ('AUTOMATIC DATA PROCESSING STOPPED. AS THE CRITERIA FOR A GOOD \n'
                           'SOLUTION ARE RATHER STRICT, YOU MAY CHOOSE TO CONTINUE DATA \n'
                           'PROCESSING AFTER CHANGING THE "JOB="-CARD IN "XDS.INP" TO \n'
                           '"JOB= DEFPIX INTEGRATE CORRECT". \n'
                           'IF THE BEST SOLUTION IS REALLY NONSENSE YOU SHOULD FIRST HAVE \n'
                           'A LOOK AT THE ASCII-FILE "SPOT.XDS". THIS FILE CONTAINS THE \n'
                           'INITIAL SPOT LIST SORTED IN DECREASING SPOT INTENSITY. SPOTS \n'
                           'NEAR THE END OF THE FILE MAY BE ARTEFACTS AND SHOULD BE ERASED. \n'
                           'ALTERNATIVELY YOU MAY TRY DIFFERENT VALUES FOR "INDEX_ORIGIN" \n'
                           'AS SUGGESTED IN THE ABOVE LISTING. \n'
                           'IF THE CRYSTAL HAS SLIPPED AT THE BEGINNING OF DATA COLLECTION \n'
                           'YOU MAY CHOOSE TO SKIP SOME OF THE FIRST FRAMES BY CHANGING \n'
                           'THE "DATA_RANGE=" IN FILE "XDS.INP" AND START ALL OVER AGAIN.\n')
                logger.warning("XDS indexing stopped with error: {}\n{}.\n{}\n".format(self.filename, error, message))
                return error, message, "JOB= DEFPIX INTEGRATE CORRECT"

            if idxref_error3 in line:
                error = idxref_error3
                message = 'POSSIBLE CAUSE: NOT ENOUGH SPOTS FOUND, WEAK DIFFRACTION OR BLANK IMAGE.'
                logger.error("XDS indexing stopped with error: {}\n{}.\n{}\n".format(self.filename, error, message))
                return error, message, "ABORT"

            if idxref_error4 in line:
                error = idxref_error4
                message = 'POSSIBLE CAUSE: NOT ENOUGH SPOTS FOUND, WEAK DIFFRACTION OR BLANK IMAGE.'
                logger.error("XDS indexing stopped with error: {}\n{}.\n{}\n".format(self.filename, error, message))
                return error, message, "ABORT"

            if idxref_warning in line:
                message = 'TURN OFF REFINEMENT OF POSITION AND/OR AXIS.'
                logger.error("XDS indexing stopped with error: {}\n{}. {}\n".format(self.filename, idxref_warning, message))
                return "REFINE(IDXREF)=CELL BEAM ORIENTATION AXIS"

            if integ_error1 in line:
                message = ('YOU MAY RESTART THIS STEP AFTER SPECIFYING VALUES IN XDS.INP FOR: \n'
                           'REFLECTING_RANGE=    REFLECTING_RANGE_E.S.D.= \n'
                           'BEAM_DIVERGENCE=     BEAM_DIVERGENCE_E.S.D.= . \n'
                           'YOU SHOULD ALSO REDUCE THE UPPER LIMIT OF THE DATA_RANGE, \n'
                           'TO STOP BEFORE THE PROBLEM FRAMES.\n')
                logger.error("Warning: XDS integration stopped with error: {}\n{}.\n{}\n".format(self.filename, integ_error1, message))
                good_frame = utils.getNumbers(self.parser.get_lines_contain('PROCESSING OF IMAGES')[-1])[0]
                return integ_error1, message, "DATA_RANGE=1 {good_frame};JOB=INTEGRATE CORRECT".format(
                    good_frame=good_frame)

            if integ_error2 in line:
                message = 'TOO FEW REFLECTIONS TO BUILD PROFILE.'
                logger.error("XDS integration stopped with error: {}\n{}.\n{}\n".format(self.filename, integ_error2, message))
                good_frame = utils.getNumbers(self.parser.get_lines_contain('PROCESSING OF IMAGES')[-1])[0]
                return integ_error2, message, "DELHI=20"

            if correct_error1 in line:
                message = "UNABLE TO RUN CORRECT in user-defined space group and cell."
                logger.error("CORRECT failed with error: {} in {}.".format(correct_error1, self.wdir))
                return correct_error1, message, ""

            if '!!! ERROR !!!' in line:
                error = line
                logger.error("unhandled XDS error in file {}, \nmessage by XDS: {}\n"
                             .format(self.filename, line))
                return error, "", ""

        return None

    def __ccp4_error(self):
        endblock = self.parser.get_lines(startIndex=-30)
        for i,line in enumerate(endblock[::-1]):
            if "Normal termination" in line:
                return None
            if "FATAL ERROR" in line or "**** ERROR ****" in line:
                message = "".join(endblock)
                logger.error("program crashed with error: {}\n{}".format(self.filename, message))
                return "CCP4_ERROR", message, ""


if __name__ == '__main__':
    pass
