from __future__ import absolute_import

import logging
import os
from collections import namedtuple

from .fileopener import FileOpener
from .labelitparser import screening_score
from .table import Table
from ..common.spacegroup import Spacegroup
from ..common.header import get_point_resolution_on_detector
from ..common.utils import getNumbers

logger = logging.getLogger(__name__)


class IdxrefParser(FileOpener):
    def __init__(self, wdir=".", filename='IDXREF.LP'):
        super(IdxrefParser, self).__init__(wdir, filename)
        self.resolution_from_spots = self.estimate_resolution_from_spot_xds()
        self.n_spots = None
        self.solution = None
        self.p1_solution = None
        self.index_table = None

        self.index_solutions = None
        self.possible_point_groups = None
        self.possible_space_groups = None

    def get_lattice_candidates_as_table(self, n_extend=0, reduce=True):
        """
        # the potential solution are marked by " *" at the beginning of the lines
        m1 = 'DETERMINATION OF LATTICE CHARACTER AND BRAVAIS LATTICE'
        m2 = 'For protein crystals the possible space'
        block_all = self.parser.get_blocks_between(m1, m2, skip1=16, skip2=1)[0]
        block = [line[2:] for line in block_all if line.startswith(" *")]
        """
        
        # in some case, it is necessary to look for information in solution
        # no marked by XDS (i.e. solution not considered as good candidates by xds)
        m1 = 'DETERMINATION OF LATTICE CHARACTER AND BRAVAIS LATTICE'
        m2 = 'For protein crystals the possible space'
        blocks = self.parser.get_blocks_between(m1, m2, skip1=16, skip2=1)
        if blocks:
            block_all = blocks[0]
        else:
            logger.error("failed to found indexing table section, index likely failed. workdir={}".format(self.wdir))
            return None

        ix_recommended_lines = [ix for ix, line in enumerate(block_all) if line.startswith(" *")]

        nlines = len(block_all)
        block = [block_all[ix][2:] for ix in ix_recommended_lines]

        if n_extend > 0 and ix_recommended_lines:
            ix_first = ix_recommended_lines[0]
            ix_last = ix_recommended_lines[-1]
            count = 0
            for ix in range(ix_last+1, nlines):
                if len(block_all[ix]) <= 10 or count >= n_extend:
                    break
                line = block_all[ix]
                # print(line.split())
                penality = float(line.split()[2])
                if penality < 250:
                    block.append(line[2:])
                    count += 1

        if not block:
            logger.error("failed to extract indexing results @ {}.".format(self.wdir))
            self.index_table = None
            return None

        logger.info("indexing results:")

        tbl = Table()
        tbl.set_title("IDXREF indexing results")
        tbl.set_table_by_lines(block, header=False, hasRowNames=False)

        tbl.set_headers(['LATTICE',
                         'BRAVIS-LATTICE',
                         'FIT',
                         'a', 'b', 'c',
                         'alpha', 'beta', 'gamma'])

        if reduce:
            # if the same Bravais lattice and similar cells, only keep the one with smallest penalty
            data = tbl.data
            lattice_tbl = Spacegroup.LATTICE
            if len(data) >= 1:
                # first sort by lattice order (decrease), then by penalty if same lattice
                idx_tbl = sorted(data,
                                 key=lambda row: (-lattice_tbl[row[1]][0], float(row[2])))

                new_data = [idx_tbl[0]]
                for ix in range(1, len(idx_tbl)):
                    if all(not (idx_tbl[ix][1] == e[1] and Spacegroup.similar_cell(idx_tbl[ix][3:], e[3:])) for e in
                           new_data):
                        new_data.append(idx_tbl[ix])

                # sort in the original order
                new_data = sorted(new_data,
                                  key=lambda row: (lattice_tbl[row[1]][0], float(row[2])))
                tbl.set_data(new_data)
                self.index_solutions = new_data[::-1]

        tbl.print()
        self.index_table = tbl
        return tbl

    '''
    # extract all point groups from index solutions into the following format:
    # point_group lattice penalty cell;...
    def get_possible_point_groups(self):
        if not self.index_solutions:
            self.get_lattice_candidates_as_table(reduce=True)

        possible_point_groups = []
        for line in self.index_solutions:
            lattice = line[1]
            pgs = Spacegroup.get_point_groups_from_lattice(lattice, output_mode='symbol')
            for pg in pgs:
                possible_point_groups.append(' '.join([pg, lattice, line[2], ','.join(line[-6:])]))
        return ';'.join(possible_point_groups)
    '''

    # extract all possible space or point groups from index solutions into the following format:
    # point_group lattice penalty cell;...
    def get_possible_space_groups(self, sgtype='spacegroup'):
        if not self.index_solutions:
            self.get_lattice_candidates_as_table(reduce=True)
            if not self.index_solutions:
                return None

        possible_space_groups = []
        for line in self.index_solutions:
            lattice = line[1]
            if sgtype.startswith('space'):
                pgs = Spacegroup.get_space_groups_from_lattice(lattice, sgtype=sgtype, output_mode='symbol')
            else:
                pgs = Spacegroup.get_space_groups_from_lattice(lattice, sgtype='pointgroup', output_mode='symbol')

            for spg in pgs:
                possible_space_groups.append(' '.join([spg, 'lattice:', lattice, 'penalty:', line[2], 'cell:', ','.join(line[-6:])]))
        return ';'.join(possible_space_groups)

    def get_p1_unit_cell(self, output='string'):
        if not self.index_table:
            self.get_lattice_candidates_as_table()
        cell = self.index_table.get_table_row_by_index(0)[-6:]
        assert len(cell) == 6
        cell = tuple(map(float, cell))
        logger.info("triclinic cell: {}".format(cell))
        return cell

    def get_unit_cell_from_spg(self, user_spg, n_extra_solutions=0):
        """ choose unit cell and spg number, based on user defined spg"""
        if n_extra_solutions > 0:  # consider extra solutions not marked by *
            self.get_lattice_candidates_as_table(n_extend=n_extra_solutions)
        else:
            if not self.index_table:
                self.get_lattice_candidates_as_table()

        # print("+++++ {} {}".format(len(self.index_table.data), self.index_table.data))

        lattice = Spacegroup.get_lattice_from_spacegroup(user_spg)
        res = [row for row in self.index_table.data if row[1] == lattice]
        # if there are more than one solution, return the solution with lower penalty
        if len(res) >= 1:
            solution = sorted(res, key=lambda x: float(x[2]))[0]
            cell = Spacegroup.cell_enforcer(user_spg, map(float, solution[-6:]))
            logger.info("choose cell {} based on given space group {}".format(cell, user_spg))
            return cell, Spacegroup.get_spg_number_from_symbol(user_spg)
        else:
            logger.error("no solution found for the given spg {}".format(user_spg))
            return None, Spacegroup.get_spg_number_from_symbol(user_spg)

    def get_unit_cell_and_spg(self):
        """ select a candidate unit cell and space group number for use in integration."""
        if not self.index_table:
            self.get_lattice_candidates_as_table()

        idx_tbl = self.index_table.data
        lattice_tbl = Spacegroup.LATTICE
        # print lattice_tbl, idx_tbl
        # for row in idx_tbl:
        #    print row[1], (lattice_tbl[row[1]][0], -float(row[2]))
        if len(idx_tbl) >= 1:
            # first sort by lattice order (increasing), then -penalty if same lattice
            idx_tbl = sorted(idx_tbl,
                             key=lambda row: (lattice_tbl[row[1]][0], -float(row[2])))
            solution = idx_tbl[-1]
            selected_lattice = solution[1]
            spg_num = min(lattice_tbl[selected_lattice][2])
            cell = Spacegroup.cell_enforcer(spg_num, map(float, solution[-6:]))
            logger.info("selected space group {} with cell {}".format(spg_num, cell))
            return cell, spg_num
        else:
            logger.error("no index solution found: {}".format(idx_tbl))
            return None

    @staticmethod
    def validate_cell_spg(unit_cell, space_group):
        if isinstance(unit_cell, str):
            unit_cell = unit_cell.split()

        spg_num = Spacegroup.get_spg_number_from_symbol(space_group)
        cell = map(float, unit_cell)
        cell = Spacegroup.cell_enforcer(space_group, cell)
        return cell, spg_num

    def n_spots(self):
        """AUTOINDEXING IS BASED ON    259 OUT OF     259 SPOTS"""
        lines = self.parser.get_lines_contain("AUTOINDEXING IS BASED ON")
        if lines:
            return getNumbers(lines[0])[-1]
        else:
            logger.error("# of spots info not found in {}".format(self.logfile))

    def get_index_solution_spg(self,
                               a1='REFINED SOLUTION BASED ON GIVEN CELL CONSTANTS AND SPACE GROUP',
                               a2='DIFFRACTION PARAMETERS USED AT START OF INTEGRATION'):
        """get results from user specified space group"""
        # TODO need to deal with cases of failures
        rms = unit_cell = mosaicity = index_table = spg_symbol = n_spots = n_indexed = lattice = ""
        blocks = self.parser.get_blocks_between(a1, a2)
        if blocks:
            for line in blocks[0]:
                if 'OUT OF' in line:
                    n_indexed, n_spots = getNumbers(line)

                if 'UNIT CELL PARAMETERS' in line:
                    unit_cell = ' '.join(str(round(float(n), 1)) for n in getNumbers(line))

                if 'CRYSTAL MOSAICITY (DEGREES)' in line:
                    mosaicity = getNumbers(line)[0]
                    # correct it to match MOSFLM:
                    # https://strucbio.biologie.uni-konstanz.de/xdswiki/index.php/FAQ#You_said_that_the_XDS_deals_with_high_mosaicity._How_high_mosaicity_is_still_manageable.3F
                    if mosaicity:
                        mosaicity = str(round(float(mosaicity) * 2.355, 2))

                if 'STANDARD DEVIATION OF SPOT    POSITION (PIXELS)' in line:
                    rms = getNumbers(line)[0]
                    # convert it into mm to match mosflm convention
                    pixel_size_in_mm = getNumbers(self.parser.get_lines_contain('QX=')[0])[-2]
                    rms = str(round(float(rms) * float(pixel_size_in_mm), 3))

                if 'SPACE GROUP NUMBER' in line:
                    spg_num = getNumbers(line)[0]
                    spg_symbol = Spacegroup.get_symbol_from_spg_number(int(spg_num))
                    lattice = Spacegroup.get_lattice_from_spacegroup(spg_symbol)

        a1 = 'LATTICE-  BRAVAIS-   QUALITY  UNIT CELL CONSTANTS (ANGSTROEM & DEGREES)'
        a2 = 'For protein crystals the'
        blocks = self.parser.get_blocks_between(a1, a2, skip1=0, skip2=1)
        if blocks:
            index_table = ''.join(blocks[0])

        solution = namedtuple("XDSSolution", "n_spots, n_indexed, sdxy, Spacegroup, Refined_cell, Bravais_lattice, Mosaicity, index_table, Resolution_from_spots, Score")
        score = screening_score(self.resolution_from_spots, rms, mosaicity)
        return solution(n_spots=n_spots,
                        n_indexed=n_indexed,
                        sdxy=rms,
                        Spacegroup=spg_symbol,
                        Refined_cell=unit_cell,
                        Mosaicity=mosaicity,
                        Resolution_from_spots=self.resolution_from_spots,
                        Score=score,
                        Bravais_lattice=lattice,
                        index_table=index_table)

    def get_index_solutions(self):
        solution = self.get_index_solution_spg()
        p1_solution = self.get_index_solution_spg(a1="INDEXING OF OBSERVED SPOTS IN SPACE GROUP #   1",
                                                  a2="DETERMINATION OF LATTICE CHARACTER AND BRAVAIS LATTICE")
        self.solution = solution
        self.p1_solution = p1_solution
        return solution, p1_solution

    def estimate_resolution_from_spot_xds(self):
        """ read xy coordinates and calculate maximum resolution for SPOT.XDS, assuming 2theta=0 """
        # TODO need to grab these values from the detector object and implement calculation in numpy to make it faster
        two_theta = 0
        pixel_size_in_mm = float(getNumbers(self.parser.get_lines_contain('QX=')[0])[-2])
        beam_xy = list(map(float, getNumbers(self.parser.get_lines_contain('ORGX=')[0])))
        dist = float(getNumbers(self.parser.get_lines_contain('DETECTOR_DISTANCE=')[0])[0])
        wave = float(getNumbers(self.parser.get_lines_contain('X-RAY_WAVELENGTH=')[0])[0])
        spot_xds = os.path.join(self.wdir, 'SPOT.XDS')
        if os.path.isfile(spot_xds):
            spots = [list(map(float, line.split()[0:2])) for line in open(spot_xds, 'r')]
            self.n_spots = len(spots)
            res = sorted(
                get_point_resolution_on_detector(spot, beam_xy, pixel_size_in_mm, wave, dist, two_theta) for spot in
                spots)
            if res:
                return str(round(sum(res[0:5]) / len(res[0:5]), 2))
        return None

    def get_nspots(self):
        if not self.n_spots:
            self.estimate_resolution_from_spot_xds()
        return self.n_spots

    def get_resolution(self):
        if not self.resolution_from_spots:
            return self.estimate_resolution_from_spot_xds()
        return self.resolution_from_spots


if __name__ == '__main__':
    i = IdxrefParser(wdir='/mnt/share/qxu/bluice/tools/scripts/quickProcess/quickProcess/tests/', filename='IDXREF1.LP')
    print(i.get_index_solutions())

    #print(i.get_lattice_candidates_as_table(n_extend=2))
    print(i.get_unit_cell_from_spg('p1'))
    print(i.get_unit_cell_from_spg('p212121', n_extra_solutions=2))