from __future__ import absolute_import

import logging
import os
import xml.etree.ElementTree

from .fileopener import FileOpener
from ..common.spacegroup import Spacegroup

logger = logging.getLogger(__name__)


class PointlessParser(FileOpener):
    def __init__(self, wdir=".", filename='pointless.log'):
        super(PointlessParser, self).__init__(wdir, filename)

    def get_laue_group_from_xml(self, xmlfile='pointless-spg.xml'):
        xmlfile = os.path.join(self.wdir, xmlfile)
        root = xml.etree.ElementTree.parse(xmlfile).getroot()
        solutions = {}

        try:
            e = root.find('BestSolution')
            solutions['type'] = e.attrib['Type']
            pgn = e.find('GroupName').text.replace(" ", "").lower()
        except (TypeError, AttributeError):
            logger.error("pointless failed: {}.".format(self.wdir))
            return None

        if pgn == 'i121' or pgn == 'c121':
            pgn = 'c2'
        elif pgn == 'p1211':
            pgn = 'p21'
        elif pgn == 'p121':
            pgn = 'p2'
        solutions['point_group_name'] = pgn

        # solutions['point_group_number'] = int(e.find('SGnumber').text)
        solutions['TotalProb'] = float(e.find('TotalProb').text)
        solutions['Confidence'] = float(e.find('Confidence').text)
        solutions['reindex_ops'] = e.find('ReindexOperator').text

        point_group_number = Spacegroup.get_laue_number_from_symbol(solutions['point_group_name'])
        solutions['point_group_number'] = point_group_number

        # cell
        e = root.find('BestCell/cell')
        a = e.find('a').text
        b = e.find('b').text
        c = e.find('c').text
        al = e.find('alpha').text
        be = e.find('beta').text
        ga = e.find('gamma').text
        cell = map(float, [a, b, c, al, be, ga])
        cell = Spacegroup.cell_enforcer(solutions['point_group_name'], cell)
        solutions['cell'] = cell
        return solutions

    def get_best_reindex_from_xml(self, xmlfile='pointless-ref.xml', tag='BestReindex'):
        xmlfile = os.path.join(self.wdir, xmlfile)
        root = xml.etree.ElementTree.parse(xmlfile).getroot()
        solutions = {}

        e = root.find(tag)
        if e is not None:
            # note we added 0 to make 3x4 matrix for XDS
            opm = ' 0 '.join(e.find('ReindexMatrix').text.split('\n'))
            solutions['reindex_matrix'] = ' '.join(opm.split())
            solutions['reindex_operator'] = e.find('ReindexOperator').text
            solutions['CC'] = float(e.find('CC').text)
            solutions['likelihood'] = float(e.find('Likelihood').text)
            solutions['confidence'] = float(e.find('Confidence').text)

        return solutions

    def get_all_reindex_from_xml(self, xmlfile='pointless-ref.xml', tag='IndexScores'):
        xmlfile = os.path.join(self.wdir, xmlfile)
        root = xml.etree.ElementTree.parse(xmlfile).getroot()

        e = root.find(tag)
        res = []
        for c in e:
            solutions = {}
            if c.tag == 'Index':
                opm = ' 0 '.join(c.find('ReindexMatrix').text.split('\n'))
                solutions['reindex_matrix'] = ' '.join(opm.split())
                solutions['reindex_operator'] = c.find('ReindexOperator').text
                solutions['CC'] = float(c.find('CC').text)
                solutions['likelihood'] = float(c.find('Likelihood').text)
                res.append(solutions)
        return res


if __name__ == '__main__':
    fo = FileOpener(filename='file_opener.py')
