from __future__ import absolute_import

import logging

from .fileopener import FileOpener
from .table import Table
from ..common.utils import getNumbers

logger = logging.getLogger(__name__)


class XplanParser(FileOpener):
    def __init__(self, wdir=".", filename='XPLAN.LP'):
        super(XplanParser, self).__init__(wdir, filename)
        self.rot_angle = None
        self.rot_start = None
        self.completeness = None

    def nref_existing(self):
        p = self.parser
        nref_existing = getNumbers(p.get_lines_contain('NUMBER OF UNIQUE REFLECTIONS IN REFERENCE_DATA_SET')[0])[0]
        return nref_existing

    def anomalous(self):
        p = self.parser

        anomalous = False
        if p.get_lines_contain("FRIEDEL'S_LAW=FALSE"):
            anomalous = True
        return anomalous

    def get_resolution_npossible(self):
        p = self.parser
        res, nref_poss = getNumbers(
            p.get_lines_after(
                'RESOLUTION       NUMBER OF UNIQUE POSSIBLE REFLECTIONS',
                nLines=1)[0])
        return res, nref_poss

    def plans(self, DESIRED_COMPLETENESS=95):
        p = self.parser

        # extract the plan table
        plans = p.get_blocks_between("starting at     total rotation     completeness        multiplicity of",
                                     "************************************************************************",
                                     iMatch=1, skip1=3, skip2=2)[0]

        tbl = Table()
        tbl.set_headers(['start_angle',
                         'total_osc',
                         'completeness_newonly',
                         'multiplicity'])
        tbl.set_table_by_lines(plans, header=False)

        new_data = tbl.data

        # get old and new values corresponding to the plan table above
        ilines = p.get_line_nums_contain('TOTAL ROTATION RANGE =')
        ilines.append(-1)

        curr_best = 0
        curr_cmpl = 0
        multi_crystal_mode = False
        for i, v in enumerate(ilines[:-1]):
            rotation = getNumbers(p.get_line(v))
            block = p.get_lines(v, ilines[i + 1])

            # corresponding items in the xplan table
            plans_rot = tbl.data[i][1]
            plans_start = tbl.data[i][0]

            for j, line in enumerate(block):
                if 'COMPLETENESS OF DATA COLLECTED IN THE OSCILLATION RANGE' in line:
                    if abs(float(plans_start) - float(getNumbers(line)[0])) < 0.1:
                        blk = block[j + 2:j + 6]
                        completeness_existing_percent = getNumbers(blk[0])[0]
                        completeness_fromnew_percent = getNumbers(blk[2])[0]
                        completeness_combined_percent = getNumbers(blk[2])[0]
                        common_reflection_percent = getNumbers(blk[-1])[0]

                        if float(completeness_combined_percent) >= DESIRED_COMPLETENESS and not self.rot_start:
                            self.rot_start = plans_start
                            self.rot_angle = plans_rot
                            self.completeness = completeness_combined_percent

                        if curr_cmpl < float(completeness_combined_percent):
                            curr_best = i
                            curr_cmpl = float(completeness_combined_percent)

                        if float(completeness_existing_percent) > 1:
                            multi_crystal_mode = True

                        new_data[i].extend([
                            completeness_existing_percent,
                            completeness_fromnew_percent,
                            common_reflection_percent,
                            completeness_combined_percent]
                        )

        # if no entries great than desired completeness, take the 2nd row
        if not self.rot_start:
            if multi_crystal_mode:
                curr_best = 1

            logger.warning(
                "no entries great than desired completeness {}, choose {}th row.".format(DESIRED_COMPLETENESS,
                                                                                         curr_best + 1))

            self.rot_angle = tbl.data[curr_best][1]
            self.rot_start = tbl.data[curr_best][0]
            self.completeness = tbl.data[curr_best][-1]

            '''
            # if no entries great than desired completeness, take the last row
            self.rot_angle = tbl.data[-1][1]
            self.rot_start = tbl.data[-1][0]
            self.completeness = tbl.data[-1][-1]
            '''

        newtbl = Table()
        newtbl.set_headers(['start_angle',
                            'total_osc',
                            'completeness_newonly',
                            'multiplicity',
                            'completeness_existing',
                            'completeness_from_new',
                            'common_oldandnew',
                            'complete_combined'
                            ])
        newtbl.set_data(new_data)
        newtbl.set_title("XDS strategy results")
        newtbl.print()
        return newtbl.format(tablefmt='simple', numalign='right', stralign='right')

    def get_osc_start(self):
        return self.rot_start

    def get_osc_range(self):
        return self.rot_angle

    def get_osc_end(self):
        return str(round(float(self.rot_start) + float(self.rot_angle), 2))

    def get_osc_delta(self):
        return 0.2

    def get_completeness(self):
        return self.completeness


if __name__ == '__main__':
    # p = XplanParser('/home/qxu/work/quickProcess', filename='XPLAN.LP.wREF')
    # p.plans()
    p = XplanParser('/mnt/beegfs/qxu/', filename='XPLAN.LP')
    print(p.plans())
    print(p.get_osc_start())
