#!/usr/bin/env python

"""
For basic applications, MPI is as easy to use as any other 
message-passing system. The sample code below contains the complete 
communications skeleton for a dynamically load balanced master/slave 
application. Following the code is a description of the few functions 
necessary to write typical parallel applications.

important parameters
--------------------

status = MPI.Status()               # where all info is stored

# Receive results from a slave
result = comm.recv(                 # message buffer
         source=MPI.ANY_SOURCE,     # receive from any sender (-1)
         tag=MPI.ANY_TAG,           # any type of message (-1)
         status=status)             # info about the received msg (class)

# Send the slave a new work unit
comm.send(work,                     # message buffer
         dest=status.Get_source(),  # to whom we just received from
         tag=WORKTAG)               # user chosen message tag

# Original author: Fernandoo Paolo <fpaolo@ucsd.edu> Jan 15, 2013
# Adapted for GMCA usage by Qingping Xu <qxu@anl.gov> Oct 12, 2016
"""

import os
import sys
import numpy as np
from mpi4py import MPI
from Queue import Queue

WORKTAG = 1
DIETAG = 0

class Work(object):
    def __init__(self, prog, params): 
        q = Queue()
        for param in params:
            job = "echo "+param + "|" + prog 
            q.put(job)
            #q.put(' '.join([prog, param]))
        self.work = q

    def get_next(self):
        if self.work.empty():
            return None
        return self.work.get()
    
    def qsize(self):
        return self.work.qsize()


def do_work(work):
    print "run job {}".format(work)
    if '.py' in work:
        os.system('python ' + work)
    else:
        print "run {}".format(work)
        os.system(work)  # for './'
    print "finish job {}".format(work)
    return

def process_result(result):
    pass


def get_params():
    # generate work queue
    # print "len=", len(sys.argv), sys.argv
    if len(sys.argv) == 3:
        program = 'mcolspot_par'
        ntask = int(sys.argv[1])
        maxcpu = int(sys.argv[2])
        params = map(str, range(1, ntask+1))
    elif len(sys.argv) == 6:
        program = 'mintegrate_par'
        fframe = int(sys.argv[1])
        ni = int(sys.argv[2])
        ntask = int(sys.argv[3])
        niba0 = int(sys.argv[4])
        maxcpu = int(sys.argv[5])
        r1 = integ_range(fframe, ni, ntask, niba0, maxcpu)
        r2 = integ_range2(fframe, ni, ntask, niba0, maxcpu)
        assert(r1 == r2)
        params = integ_range(fframe, ni, ntask, niba0, maxcpu)
    else:
        print "Error: unexpected number of argument passed into the program: {}".format(sys.argv)
        exit(1)

    return program, params

def integ_range(fframe, ni, ntask, niba0, maxcpu):
    # fframe id number of the first image
    # ni     #number of images in the data set
    # niba0  #minimum number of images in a batch
    # maxcpu #maximum number of processors used by each job
    print "integ_range parameters: {} {} {} {} {}".format(fframe, ni, ntask, niba0, maxcpu)
    # translated from xds forkintegrate script
    minitask = ni // ntask  # minimum number of images in a job
    mtask = ni % ntask      # number of jobs with minitask+1 images
    litask, itask = 0, 1 
    #nba = 0

    params = []
    while itask <= ntask:
        if itask > mtask:
            nitask = minitask
        else:
            nitask = minitask + 1
        
        fitask = litask + 1
        litask = litask + nitask

        if nitask < niba0:
            n = nitask
        else:
            n = niba0

        if n < 1:
            n = 1
        #n = max(1, min(nitask, niba0))
        
        nbatask = nitask // n
        #nba = nba + nbatask
        image1 = fframe + fitask - 1

        params.append("{image1} {nitask} {itask} {nbatask}".format(image1=image1, nitask=nitask, itask=itask, nbatask=nbatask))

        itask += 1

    return params

def integ_range2(fframe, ni, ntask, niba0, maxcpu):
    # version 2, equivalent to abve 
    minitask = ni // ntask 
    mtask = ni % ntask
    itasks = [ i + 1 for i in range(ntask)]
    nitasks = [ minitask if i > mtask else minitask + 1 for i in itasks ]
    nbatasks = [ nitask // n for n in [ max(1, min(nitask, niba0)) for nitask in nitasks ] ]
    image1 = [ fframe + sum(nitasks[:i]) for i in range(ntask) ]
    return [ "{} {} {} {}".format(a, b, c, d) for a, b, c, d in zip(image1, nitasks, itasks, nbatasks) ]

def master(comm):
    num_procs = comm.Get_size()
    status = MPI.Status()

    program, params = get_params()
    print program, params
    wq = Work(program, params)
    print wq.qsize()

    # Seed the slaves, send one unit of work to each slave (rank)
    for rank in xrange(1, num_procs):
        work = wq.get_next()
        comm.send(work, dest=rank, tag=WORKTAG)

    # Loop over getting new work requests until there is no more work to be done
    while True:
        work = wq.get_next()
        if not work: break

        # Receive results from a slave
        print "receive result {} from a slave".format(work)
        result = comm.recv(source=MPI.ANY_SOURCE, tag=MPI.ANY_TAG, status=status)
        print "receive result {} from a slave".format(work)
        #process_result(result)

        # Send the slave a new work unit
        print "sent {} to slave".format(work)
        comm.send(work, dest=status.Get_source(), tag=WORKTAG)

    # No more work to be done, receive all outstanding results from slaves
    for rank in xrange(1, num_procs): 
        result = comm.recv(source=MPI.ANY_SOURCE, tag=MPI.ANY_TAG, status=status)
        #process_result(result)

    # Tell all the slaves to exit by sending an empty message with DIETAG
    for rank in xrange(1, num_procs):
        comm.send(0, dest=rank, tag=DIETAG)


def slave(comm):
    my_rank = comm.Get_rank()
    status = MPI.Status()
    #print "I am slave {} {}".format(my_rank, status)

    while True:
        # Receive a message from the master
        work = comm.recv(source=0, tag=MPI.ANY_TAG, status=status)
        print "slave {} received {}".format(my_rank, work)

        # Check the tag of the received message
        if status.Get_tag() == DIETAG: break 

        # Do the work
        result = do_work(work)

        # Send the result back
        comm.send(result, dest=0, tag=0)


def main():
    comm = MPI.COMM_WORLD
    my_rank = comm.Get_rank()
    my_name = MPI.Get_processor_name()
    #comm.Barrier()
    #start = MPI.Wtime()

    if my_rank == 0:
        master(comm)
    else:
        slave(comm)

    #comm.Barrier()
    #end = MPI.Wtime()
    #print 'time:', end - start

if __name__ == '__main__':
    main()
