#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Thu Nov  3 16:34:51 2016

@author: qxu
"""
from __future__ import absolute_import

from .pipelines import *
from .get_args import ARGS
from ..log.logginghandler import FileDBHandler
from ..common.pv import update_pipelineStatus_PV

logger = logging.getLogger(__name__)


def setup_logging():
    # set up logging configuration, in this case, modified python's
    # file logger and add a db component to it, the default stream handler
    # is removed
    FORMAT = "%(asctime)s %(levelname)s %(name)s: %(message)s"
    DATE_FORMAT = "%Y-%m-%d %H:%M:%S"
    logging.basicConfig(level=logging.INFO, format=FORMAT)

    # adding a custom file/db handler
    rootlog = logging.getLogger()
    mh = FileDBHandler()

    FORMAT = "%(levelname)s %(name)s: %(message)s"
    format = logging.Formatter(FORMAT, datefmt=DATE_FORMAT)
    mh.setFormatter(format)
    rootlog.addHandler(mh)

    if ARGS.disable_console_logger:
        rootlog.removeHandler(rootlog.handlers[0])

    # attach pipeline status id to args
    if ARGS.log2db:
        ARGS.pipelinestatus_id = mh.get_pipelinestatus_id()
        ARGS.pipelinestatus_table_reference = mh.get_pipelinestatus()

    ARGS.lockfile = mh.getLockFilename()

    job_type = "PROCESS"
    if "strategy" in ARGS.pipeline[0]:
        job_type = "STRATEGY"

    update_pipelineStatus_PV(ARGS.pipelinestatus_id, job_type, "STARTED")

    return


def pipeline_runner():
    try:
        if any('kamo_' in x for x in ARGS.pipeline):
            # skip parsing images in case Kamo since it takes a long time
            logger.info("skipped directory parsing for pipeline {}".format(ARGS.pipeline))
            sweeps = None
            imgset = "kamoSet"
            if os.path.exists(os.path.join(ARGS.data_dir, "datasets.txt")):
                imgset = "datasets.txt"
        else:
            sweeps = get_file_sweeps()
            if sweeps:
                imgset = ','.join(s.get_imageSet() for s in sweeps)
            else:
                imgset = 'NoData'

        # update imageSet of pipeline status table,
        # TODO how to better handle this, since the follow will not work for hdf5
        if ARGS.log2db:
            ARGS.imageSet = imgset
            ARGS.pipelinestatus_table_reference.imageSet = imgset
            ARGS.pipelinestatus_table_reference.save()
    except Exception:
        logger.error("error occured during parsing image directories.", exc_info=True)
        update_job_status("DONE")
        return

    # or sweeps = get_images_sweeps(args)
    if sweeps:
        if len(sweeps) < 1 or ARGS.showsweeps:
            update_job_status("DONE")
            logger.info("nothing to process, job completed.")
            return
    else:
        if any('kamo_' in x for x in ARGS.pipeline):
            pass
        else:
            logger.info("nothing to process, job completed.")
            update_job_status("DONE")
            return

    # if generate XDS.INP is requested
    if ARGS.xdsinp:
        for sweep in sweeps:
            create_xdsinp(sweep)
        logger.info("created XDS.INP files, job completed.")
        update_job_status("DONE")
        return

    if not os.path.isdir(ARGS.proc_dir):
        os.makedirs(ARGS.proc_dir)

    try:
        update_job_status("RUN")
        jobs = []
        if "xia2_dials" in ARGS.pipeline:
            logger.info("running xia2_dials....")
            process = Process(target=xia2, args=(sweeps, "xia2_dials"))
            process.start()
            jobs.append(process)

        if "xia2_xds" in ARGS.pipeline:
            logger.info("running xia2_xds....")
            process = Process(target=xia2, args=(sweeps, "xia2_xds"))
            process.start()
            jobs.append(process)

        if "xia2_dials_aimless" in ARGS.pipeline:
            logger.info("running xia2_dials_aimless....")
            process = Process(target=xia2, args=(sweeps, "xia2_dials_aimless"))
            process.start()
            jobs.append(process)

        if "kamo_dials" in ARGS.pipeline:
            logger.info("running kamo_dials....")
            process = Process(target=kamo, args=(sweeps, "kamo_dials"))
            process.start()
            jobs.append(process)

        if "kamo_xds" in ARGS.pipeline:
            logger.info("running kamo_xds....")
            process = Process(target=kamo, args=(sweeps, "kamo_xds"))
            process.start()
            jobs.append(process)

        if "autoproc" in ARGS.pipeline:
            process = Process(target=autoproc, args=(sweeps,))
            process.start()
            jobs.append(process)

        if "gmcaproc" in ARGS.pipeline:
            if ARGS.space_group or ARGS.xds_refhkl:
                process = Process(target=gmca_process, args=(sweeps,))
            else:
                process = Process(target=quick_process, args=(sweeps,))

            process.start()
            jobs.append(process)

        if 'fastdp' in ARGS.pipeline:
            process = Process(target=fastdp, args=(sweeps,))
            process.start()
            jobs.append(process)

        for j in jobs:
            j.join()

        update_job_status("DONE")

        return
    except SystemExit as e:
        logger.info("program exited.")
    except Exception:
        logger.error("program failed with error.", exc_info=True)

    if ARGS.log2db:
        update_job_status("FAIL")

    return


def main():
    setup_logging()

    pipeline_runner()

    if os.path.exists(ARGS.lockfile):
        logger.info("lock file {} deleted.".format(ARGS.lockfile))
        os.remove(ARGS.lockfile)

    return


if __name__ == '__main__':
    main()
