#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Fri Mar  3 17:33:56 2017

@author: qxu
"""

import argparse
import datetime
import logging
import os
import re
import sys

from ..common.utils import isFloat
from ..file_parser.mtzfile import Mtzfile
from ..file_parser.xdsasciiparser import XdsAsciiParser
from ..file_parser.xls_reader import xlsReader
from ..tasks.sca2xds import SCA2XDS

logger = logging.getLogger(__name__)


def get_args():
    """ command line options """

    parser = argparse.ArgumentParser(prog='GMCAproc')

    g1 = parser.add_argument_group("traditional data processing")

    g1.add_argument('--data_dir', '-i',
                    required=False,
                    help='directory for image files')
    g1.add_argument("--filelist",
                    default=None,
                    help="a list of collected files, one file per line, in absolute path")
    g1.add_argument("--proc_dir", "-d", "-o",
                    default=os.getcwd(),
                    help="root processing directory")
    g1.add_argument("--recursive",
                    action="store_true",
                    required=False,
                    help="parse directory recursively to find all images")
    g1.add_argument("--small_wedges",
                    action="store_true",
                    required=False,
                    help="small wedges mode for KAMO, data consisting of many small wedges")
    g1.add_argument("--exclude_dirs", "-exdir",
                    type=str,
                    default=None,
                    help='listing excluded directories in the recursive mode, e.g. "process;raster"')
    g1.add_argument("--include_dirs", "-indir",
                    type=str,
                    default=None,
                    help='listing included directories in the recursive mode, e.g. "collect"')
    g1.add_argument("--showsweeps", "--collectinfo",
                    action="store_true",
                    required=False,
                    help="parse directory and show sweeps to be processed")
    g1.add_argument("--xdsinp",
                    action="store_true",
                    required=False,
                    help="create XDS.INP files, but don't run it")
    g1.add_argument("--pipeline", "-p",
                    required=False,
                    choices=["gmcaproc",
                             "autoproc",
                             "xia2_xds",
                             "xia2_dials",
                             "xia2_dials_aimless",
                             "kamo_xds",
                             "kamo_dials",
                             "fastdp",
                             "mosflm_strategy",
                             "labelit_strategy",
                             "dials_strategy",
                             "xds_strategy",
                             ],
                    action='append',
                    help="select processing pipeline")

    g1.add_argument("--beamx",
                    required=False,
                    type=str,
                    default=None,
                    help="define beamx in pixel, overwrite header value")
    g1.add_argument("--beamy",
                    required=False,
                    type=str,
                    default=None,
                    help="define beamy in pixel, overwrite header value")

    g1.add_argument("--start", "-s",
                    type=int,
                    default=None,
                    help="start frame number")
    g1.add_argument("--end", "-e",
                    type=int,
                    help="last frame number")
    g1.add_argument("--percent",
                    type=int,
                    help="percent of frame from start to process")
    g1.add_argument("--highres", "-hr",
                    type=str,
                    default=None,
                    help='high resolution cutoff')
    g1.add_argument('--space_group', '-symm',
                    type=str,
                    default=None,
                    help='space group')
    g1.add_argument("--unit_cell", "-cell",
                    type=str,
                    default=None,
                    help="unit cell parameter")
    g1.add_argument("--prefix", "--startwith",
                    help="image filename prefix; use semicolon to separate multiple prefixes, e.g. A1;B1_1")
    g1.add_argument('--native', '-noanomalous',
                    action="store_true",
                    help="turn off anomalous (native data)")
    g1.add_argument("--gmca_optimize", "-optimize",
                    action="store_true",
                    help="turn on optimization step for quickProcess")
    g1.add_argument("--min_sweep_size", "-min",
                    type=int,
                    default=5,
                    help="ignore images if minimum # of frames in a sweep is less than defined size")
    g1.add_argument("--max_sweep_size", "-max",
                    type=int,
                    default=99999,
                    help="max # of frames in a sweep")
    g1.add_argument("--timeout",
                    type=int,
                    default=600,
                    help="wait until timeout is reached for filelist based data processing, in sec")
    g1.add_argument("--runner", "-r",
                    default="local",
                    choices=["local", "sge", "slurm"],
                    help="use local cpus, sge, or slurm")
    g1.add_argument("--username",
                    default=os.environ.get("USER"),
                    help="username job belongs to, used by web")
    g1.add_argument("--nproc", "-np",
                    type=int,
                    default=4,
                    help="number of processors per machine")
    g1.add_argument("--njobs", "-nj",
                    type=int,
                    default=1,
                    help="number of machines to use")
    g1.add_argument("--xds_refhkl",
                    type=str,
                    default=None,
                    help="reference data used by xds XPLAN or CORRECT")
    g1.add_argument("--xdscmds",
                    type=str,
                    default="",
                    help="specific xds commands separated by semicolon")
    g1.add_argument("--auto_exclude_frames",
                    action="store_true",
                    required=False,
                    help="automatic exclude bad frames")
    g1.add_argument("--sampleNumber",
                    type=int,
                    default=0,
                    help="jbluice specific, sample number")
    g1.add_argument("--sampleName",
                    type=str,
                    default="",
                    help="jbluice specific, sample name")
    g1.add_argument("--xlsFile",
                    type=str,
                    default=None,
                    help="jbluice specific, spreadsheet for screening")
    g1.add_argument("--mountedCrystal",
                    type=str,
                    default=None,
                    help="jbluice specific, mounted crystal")
    g1.add_argument("--model", "--pdb_file",
                    type=str,
                    default=None,
                    help="model for dimple")
    g1.add_argument("--sequence", "--seq_file",
                    type=str,
                    default=None,
                    help="sequence file in fasta format")
    g1.add_argument("--hatom",
                    type=str,
                    default=None,
                    help="heavy atom type")
    g1.add_argument("--imageSet",
                    type=str,
                    default=None,
                    help="jbluice specific, sets of images processed")

    g1.add_argument('--image', dest='images',
                    action='append',
                    type=str,
                    required=False,
                    help="adding individual images one by one")

    g1.add_argument("--beamline",
                    type=str,
                    choices=["idb",
                             "23o",
                             "idd",
                             "23i",
                             "23b",
                             "23bm",
                             "config"
                             ],
                    default="config",
                    help="jbluice specific, beamline submitted the job, decides which db to write to")
    g1.add_argument("--logfile",
                    type=str,
                    default="",
                    help="name of the logfile")
    g1.add_argument("--redis_key",
                    type=str,
                    default="",
                    help="redis_key used to communicate results with pybluice")
    g1.add_argument("--log2db",
                    action="store_true",
                    required=False,
                    help="jbluice specific, logging program state/results to a database")
    g1.add_argument("--htmlout",
                    action="store_true",
                    required=False,
                    help="save html elements into the db, instead of plain text")
    g1.add_argument("--disable_console_logger",
                    action="store_true",
                    required=False,
                    help="disable stream the log to stdout")
    g1.add_argument("--debug",
                    action="store_true",
                    required=False,
                    help="turn on debug mode with more verbose output")

    args = parser.parse_args()

    if not args.data_dir and not args.filelist and not args.hitlist_dir:
        print("at least one of --data_dir, --filelist or --hitlist_dir is required.")
        parser.print_help()
        sys.exit(0)

    if args.unit_cell:
        cell = re.split(r"[,;\s]+", args.unit_cell.strip(" "))
        if len(cell) == 6:
            args.unit_cell = ",".join(cell)
        else:
            logger.error("cell dimensions is not defined correctly, ignored.")
            args.unit_cell = None

    if args.xlsFile and args.mountedCrystal and os.path.exists(args.xlsFile):
        xls_handler(args)

    if args.auto_exclude_frames:
        if not args.cmds:
            args.xdscmds = "!SPOT_RANGE=DATA_RANGE"
        elif "!SPOT_RANGE=DATA_RANGE" not in args.xdscmds:
            args.xdscmds += ";!SPOT_RANGE=DATA_RANGE"

    if args.xds_refhkl and os.path.exists(args.xds_refhkl):
        try:
            referencedata_handler(args)
        except Exception:
            raise RuntimeError("unable to handle reference data: {}".format(args.xds_refhkl))

    if not args.pipeline:
        args.pipeline = ["gmcaproc"]
        print("no pipeline selected, defaulted to GMCAproc")

    # formula based resol cutoff mode only supported in gmcaproc mode only
    if len(args.pipeline) == 1 and args.highres:
        if 'gmcaproc' in args.pipeline:
            if args.highres not in ("cchalf~30%", "<i/sigmai>~1.5", "<i/sigmai>~2.0") and not isFloat(args.highres):
                print("highres was not defined correctly, resorted to default of using cc~30%")
                args.highres = "cchalf~30%"
        else:
            if not isFloat(args.highres):
                args.highres = None

    args.hostname = os.getenv('HOSTNAME')
    args.pid = os.getpid()
    args.processing_id = os.getpid()
    args.command = " ".join(sys.argv)
    args.starttime = datetime.datetime.now()

    if not args.imageSet:
        imageSet = ""
        if args.start:
            imageSet = "s:{}".format(args.start)

        if args.end:
            imageSet += "e:{}".format(args.end)

        if args.percent:
            imageSet += "p:{}%".format(args.percent)

        if args.filelist:
            path, file = os.path.split(args.filelist)
            imageSet += "f:{}".format(file.replace('filelist', 'chkpt'))

        if not imageSet:
            imageSet = "All"
        args.imageSet = imageSet

    if not args.logfile:
        args.logfile = "GMCAproc_{}_{}_{}.log".format('-'.join(args.pipeline), args.hostname, args.pid)
    print("PID = {} on host {}".format(args.pid, args.hostname))

    args.DATA = dict()
    return args


def namespace2dict(ns, values_as_string=True):
    ns_copy = argparse.Namespace(**vars(ns))  # make a copy, otherwise, original copy got changed as well
    if isinstance(ns_copy, argparse.Namespace):
        runtime_info = vars(ns_copy)
        if values_as_string:
            for k, v in runtime_info.items():
                runtime_info[k] = str(v)
    return runtime_info


def xls_handler(args):
    xls = xlsReader(args.xlsFile)
    port = args.mountedCrystal
    hatom = xls.get_element(port, 'Metal')
    model = xls.get_element(port, 'ModelPath')
    sequence = xls.get_element(port, 'Sequences')
    space_group = xls.get_element(port, 'Spacegroup')
    reference_hkl = xls.get_element(port, "ReferenceDataSet")
    protein = xls.get_element(port, "Protein")
    xtal_id = xls.get_element(port, 'CrystalID')
    sample_id = "{}{}{}".format(protein, xtal_id, port)

    if hatom:
        args.hatom = hatom
        print("user provided heavy atom in spreadsheet: {}.".format(args.hatom))

    if space_group:
        args.space_group = space_group
        print("user provided space group in spreadsheet: {}.".format(args.space_group))

    if model:
        model_file = looking_for_file(model, ext='.pdb')
        if model_file:
            args.model = model_file
            print("user provided model file in spreadsheet: {}.".format(args.model))

    if sequence:
        seq_file = looking_for_file(sequence, ext='')
        if seq_file:
            print("user provided sequence file in spreadsheet: {}.".format(args.seq_file))
            args.sequence = seq_file

    if reference_hkl and os.path.exists(reference_hkl):
        args.reference_hkl = reference_hkl
        print("user provided a reference reflection file in spreadsheet: {}.".format(args.reference_hkl))

    if sample_id:
        args.sample_id = sample_id

    if hatom:
        args.anomalous = True
    else:
        args.anomalous = False


def looking_for_file(filename, ext='.pdb'):
    """ look for user supplied files, if abs path is given, use it if exists,
    otherwise,  $HOME, $HOME/Desktop, $HOME/Downloads"""

    filename = filename.strip()
    if "$HOME" in filename:
        filename = filename.replace("$HOME", os.environ['HOME'])

    if os.path.isabs(filename):
        if os.path.exists(filename):
            return filename
    elif ext in filename.lower():
        default_directories = [os.path.join(os.environ['HOME'], d) for d in ['Downloads', 'Desktop', '']]
        for d in default_directories:
            fn = os.path.join(d, filename)
            if os.path.exists(fn):
                return fn

    return None


def referencedata_handler(args):
    if args.xds_refhkl.endswith('.sca') or args.xds_refhkl.endswith('.SCA'):
        print("converting scalepack format to xds format")
        args.xds_refhkl = SCA2XDS(sca_file=args.xds_refhkl).convert()
    elif args.xds_refhkl.endswith('.mtz'):
        print("converting mtz format to xds format")
        mtzfile = Mtzfile(args.xds_refhkl)
        args.xds_refhkl = mtzfile.convert_to_XDS()
        args.space_group = mtzfile.get_spacegroup()[0]
        args.unit_cell = mtzfile.get_cell()
    else:
        hkldir, hklname = os.path.split(args.xds_refhkl)
        spg, cell = XdsAsciiParser(hkldir, filename=hklname).get_spg_cell(spg_type='symbol')
        args.space_group = spg
        args.unit_cell = " ".join(cell)
        print("user supplied reference hkl with spg={}, cell={}".format(spg, cell))


# run function and make it available when user import this module
ARGS = get_args()
