#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Thu Mar  2 09:03:32 2017

@author: qxu
"""
from __future__ import absolute_import

import logging
import os
from functools import partial
from multiprocessing import Process
import datetime

from .get_args import ARGS
from ..xds.scaler import Scaler
from ..common.generatesweeps import GenerateSweeps
from ..common.imagesweep import ImageSweep
from ..common.utils import timeit, parallelize, isFloat
from ..tasks.autoproc import AutoPROC
from ..tasks.kamo import Kamo
from ..tasks.fastdp import FastDP
from ..tasks.xia2 import Xia2
from ..xds.xds import XDS
from ..tasks.solver import Solver
from ..common.pv import update_pipelineStatus_PV

logger = logging.getLogger(__name__)


def update_job_status(status="RUN", jobname=None):
    if ARGS.log2db:
        ARGS.pipelinestatus_table_reference.state = status
        ARGS.pipelinestatus_table_reference.elapsedtime = (datetime.datetime.now() - ARGS.starttime).total_seconds()

        if jobname:
            ARGS.pipelinestatus_table_reference.sampleName = jobname

        ARGS.pipelinestatus_table_reference.save()

        update_pipelineStatus_PV(ARGS.pipelinestatus_id, "PROCESS", status)

@timeit
# parse directory recursively to find all sweeps, if requested
def get_file_sweeps():
    xs = list()

    dirs_include = []
    dirs_exclude = []
    if ARGS.include_dirs:
        dirs_include = ARGS.include_dirs.replace(" ", "").split(";")

    if ARGS.exclude_dirs:
        dirs_exclude = ARGS.exclude_dirs.replace(" ", "").split(";")

    for directory, subdirList, fileList in os.walk(ARGS.data_dir, topdown=True):
        if os.path.isfile(os.path.join(ARGS.data_dir, "datasets.txt")) and any('kamo_' in x for x in ARGS.pipeline):
            logger.info("use datasets definition datasets.txt @ {} for KAMO path".format(ARGS.data_dir))
            return None

        # skipping directories
        if dirs_exclude and any(xd in directory for xd in dirs_exclude):
            continue

        # if user specified certain directories
        if dirs_include and not any(xd in directory for xd in dirs_include):
            continue

        imgs = GenerateSweeps(directory,
                              filelist=ARGS.filelist,
                              images=ARGS.images,
                              split_gap=True,
                              start=ARGS.start,
                              end=ARGS.end,
                              percent=ARGS.percent,
                              max_images_per_sweep=ARGS.max_sweep_size,
                              prefix=ARGS.prefix)

        xs.extend(imgs.get_file_sweeps())
        if not ARGS.recursive:
            break

    xs = [sweep for sweep in xs if sweep.nfiles >= ARGS.min_sweep_size]
    for s in xs:
        s.print_summary()

    logger.info("Total # of sweeps found: {}".format(len(xs)))
    logger.info("-" * 80)
    return xs


def sweep_convert_file2image(fs):
    return ImageSweep(fs, start_frame=ARGS.start, end_frame=ARGS.end)


@timeit
def autoproc(sweeps):
    if not isinstance(sweeps[0], ImageSweep):
        sweeps = [sweep_convert_file2image(fs) for fs in sweeps]

    sweep_def = '  \\\n'.join([s.get_autoPROC_ids() for s in sweeps])

    difficult = False
    if ARGS.recursive:
        difficult = True
    elif ARGS.highres:
        if isFloat(ARGS.highres) and float(ARGS.highres) > 3.0:
            difficult = True

    subdir = "autoPROC"
    if "/autoproc" in ARGS.proc_dir or "/autoPROC" in ARGS.proc_dir:
        subdir = ""  # do not create a subdirectory if such directory is already created

    # switch to display native log
    if ARGS.log2db:
        ARGS.pipelinestatus_table_reference.logfile = "{}/{}/summary.html".format(ARGS.proc_dir, subdir)
        ARGS.pipelinestatus_table_reference.save()

    p = AutoPROC(sweep_def,
                 wdir=ARGS.proc_dir,
                 subdir=subdir,
                 space_group=ARGS.space_group,
                 unit_cell=ARGS.unit_cell,
                 highres=ARGS.highres,
                 native=ARGS.native,
                 log2db=ARGS.log2db,
                 image_sweeps=sweeps,
                 difficulty=difficult,
                 runner=ARGS.runner,
                 nproc=ARGS.nproc,
                 njobs=ARGS.njobs)
    retcode = p.run_process()
    logger.info("{} job completed".format("autoPROC @ Globalphasing"))
    return retcode


@timeit
def xia2(sweeps, pipeline='xia2_dials'):
    if not isinstance(sweeps[0], ImageSweep):
        sweeps = [sweep_convert_file2image(fs) for fs in sweeps]

    sweep_def = '  \\\n'.join([s.get_xia2_sweep_defintion() for s in sweeps])

    subdir = pipeline
    if pipeline in ARGS.proc_dir:
        subdir = ""  # do not create a subdirectory if such directory is already present upstream

    # switch to display native log
    if ARGS.log2db:
        ARGS.pipelinestatus_table_reference.logfile = "{}/{}/xia2.txt".format(ARGS.proc_dir, subdir)
        ARGS.pipelinestatus_table_reference.save()

    try:
        p = Xia2(sweep_def,
                 wdir=ARGS.proc_dir,
                 subdir=subdir,
                 space_group=ARGS.space_group,
                 unit_cell=ARGS.unit_cell,
                 highres=ARGS.highres,
                 pipeline=pipeline,
                 native=ARGS.native,
                 log2db=ARGS.log2db,
                 image_sweeps=sweeps,
                 runner=ARGS.runner,
                 nproc=ARGS.nproc,
                 njobs=ARGS.njobs)

        retcode = p.run_xia2()
        logger.info("{} job completed".format(pipeline))
        return retcode
    except Exception as e:
        logger.error("xia2 failed", exc_info=True)


@timeit
def kamo(sweeps, pipeline='kamo_dials'):
    """
    # use filesweep directly since this step take too long
    if not isinstance(sweeps[0], ImageSweep):
        logger.info("converting to image sweep....")
        sweeps = [sweep_convert_file2image(fs, args) for fs in sweeps]
        logger.info("convert to image sweep done.")
    """

    subdir = pipeline
    if pipeline in ARGS.proc_dir:
        subdir = ""  # do not create a subdirectory if such directory is already present upstream

    if ARGS.log2db:
        ARGS.pipelinestatus_table_reference.logfile = "{}/{}/report.html".format(ARGS.proc_dir, subdir)
        ARGS.pipelinestatus_table_reference.save()

    '''
    # file sweep has no such property
    oscs = [s.get_total_osc() for s in sweeps]
    n_small_wedges = sum(1 if osc < 25 else 0 for osc in oscs)
    small_wedge = True if float(n_small_wedges)/len(oscs) >= 0.6 else False
    '''
    try:
        p = Kamo(wdir=ARGS.proc_dir,
                 data_dir=ARGS.data_dir,
                 subdir=subdir,
                 space_group=ARGS.space_group,
                 unit_cell=ARGS.unit_cell,
                 highres=ARGS.highres,
                 pipeline=pipeline,
                 native=ARGS.native,
                 small_wedges=ARGS.small_wedges,
                 log2db=ARGS.log2db,
                 image_sweeps=sweeps,
                 runner=ARGS.runner,
                 nproc=ARGS.nproc,
                 njobs=ARGS.njobs)

        retcode = p.run_kamo()
        logger.info("{} job completed".format(pipeline))
        return retcode
    except Exception as e:
        logger.error("kamo failed", exc_info=True)


@timeit
def fastdp(sweeps):
    def process_sweep(s):
        first_image = s.get_first_file(abs_path=True)
        subdir = "{}-fastdp".format(s.get_sweep_id())

        # switch to display native log
        if ARGS.log2db:
            ARGS.pipelinestatus_table_reference.logfile = "{}/{}/fast_dp.log".format(ARGS.proc_dir, subdir)
            ARGS.pipelinestatus_table_reference.save()

        p = FastDP(first_image,
                   start=ARGS.start,
                   end=ARGS.end,
                   wdir=ARGS.proc_dir,
                   subdir=subdir,
                   space_group=ARGS.space_group,
                   unit_cell=ARGS.unit_cell,
                   highres=ARGS.highres,
                   nproc=ARGS.nproc,
                   njobs=ARGS.njobs,
                   image_sweeps=s,
                   log2db=ARGS.log2db)
        return p.run_process()

    if not isinstance(sweeps[0], ImageSweep):
        sweeps = [sweep_convert_file2image(fs) for fs in sweeps]

    if len(sweeps) > 1:
        logger.warning("fastdp will process each sweep independently, "
                       "for data with multiple sweeps, you will need to "
                       "check indexing consistency between sweeps.")

    try:
        parallelize(process_sweep, sweeps)
    except Exception as e:
        logger.error("fastdp pipeline failed to complete", exc_info=True)
        return 1
    logger.info("{} job completed.".format("fastdp @ Diamond"))

    # solve structure
    Solver().solve()

    return 0

@timeit
def GMCA_proc(sweep, subdir='XDS', xds_ref=None):
    if not isinstance(sweep, ImageSweep):
        sweep = sweep_convert_file2image(sweep)

    xds = XDS(wdir=ARGS.proc_dir,
              runner=ARGS.runner,
              user_space_group=ARGS.space_group,
              user_unit_cell=ARGS.unit_cell,
              user_xdscmds=ARGS.xdscmds,
              user_refhkl=ARGS.xds_refhkl,
              highres=ARGS.highres,
              native=ARGS.native,
              nproc=ARGS.nproc,
              njobs=ARGS.njobs
              ).set_xdsinp_from_image_sweep(sweep)
    xds.process(ref_XDS=xds_ref, subdir=subdir)
    xds.refine(max_cycles=1)
    return xds


def create_xdsinp(sweep):
    if not isinstance(sweep, ImageSweep):
        sweep = sweep_convert_file2image(sweep)

    xds = XDS(wdir=ARGS.proc_dir,
              nproc=ARGS.nproc,
              njobs=ARGS.njobs,
              native=ARGS.native,
              user_xdscmds=ARGS.xdscmds,
              user_space_group=ARGS.space_group,
              user_unit_cell=ARGS.unit_cell,
              ).set_xdsinp_from_image_sweep(sweep, save=True)
    return xds


@timeit
def scale_xds(xds_objs, subdir, reindex_check):
    scaler_obj = Scaler(xds_objs,
                        wdir=ARGS.proc_dir,
                        subdir=subdir,
                        nproc=ARGS.nproc,
                        log2db=ARGS.log2db,
                        reindex_check=reindex_check,
                        anomalous=not ARGS.native,
                        use_parallelize=True)

    try:
        scaler_obj.merge_processed_data()
    except Exception as e:
        logger.error("scale_xds failed", exc_info=True)
        raise e


@timeit
def quick_process(sweeps):
    # run all sweeps in p1 in parallel
    # run the following two steps in parallel
    #   1) check reindex and scaling them together
    #   2) select one of them as a reference and run xds optimization (integrate and correct) 
    #      and scale them together
    @timeit
    def GMCA_proc_p1(sweep):
        if not isinstance(sweep, ImageSweep):
            sweep = sweep_convert_file2image(sweep)

        try:
            xds = XDS(wdir=ARGS.proc_dir,
                      runner=ARGS.runner,
                      nproc=ARGS.nproc,
                      highres=ARGS.highres,
                      native=ARGS.native,
                      user_xdscmds=ARGS.xdscmds,
                      njobs=ARGS.njobs).set_xdsinp_from_image_sweep(sweep)

            xds.process_in_p1()
            return xds
        except Exception as e:
            logger.error("gmcaproc p1 failed", exc_info=True)
            raise

    subdir = "gmcaproc"
    if "/gmcaproc" in ARGS.proc_dir:
        subdir = ""

    ARGS.proc_dir = os.path.join(ARGS.proc_dir, subdir)

    if not os.path.isdir(ARGS.proc_dir):
        os.makedirs(ARGS.proc_dir)

    logger.info("processing all sweeps of data in parallel in p1 ...")

    xds1 = parallelize(GMCA_proc_p1,
                       sweeps,
                       nproc=8)

    logger.info("scale data sweeps processed in p1 async ...")
    s1 = Process(target=scale_xds, args=(xds1, "scale-1", True))
    s1.start()
    async_flag = False
    if not async_flag:
        s1.join()

    if ARGS.gmca_optimize:
        update_job_status("REFINE")

        logger.info("optimize all data sweeps in parallel using parameters determined above ...")
        # use the sweep with maximum i/sigmai as reference
        xds_ref = sorted(xds1, key=lambda xds: xds.correct_summary['I/SigmaI'])[-1]
        logger.info("using sweep with maximum i/sigmaI as reference in {}".format(xds_ref.get_process_dir()))

        func = partial(GMCA_proc, xds_ref=xds_ref)
        xds2 = parallelize(func, sweeps, nproc=8)

        logger.info("scale data sweeps in pass 2  ...")
        s2 = Process(target=scale_xds, args=(xds2, "scale-2", False))
        s2.start()
        s2.join()

    if async_flag:
        s1.join()  # turn on if run scaling async

    # solve structure
    Solver().solve()

    logger.info("{} job completed.".format("GMCAproc quickProcess"))

    return 0


@timeit
def gmca_process(sweeps):
    # handle reprocess mode of GMCA pipeline
    # create a subdirectory for pipelinename if necessary
    subdir = "gmcaproc"
    if "/gmcaproc" in ARGS.proc_dir:
        subdir = ""

    ARGS.proc_dir = os.path.join(ARGS.proc_dir, subdir)
    if not os.path.isdir(ARGS.proc_dir):
        os.makedirs(ARGS.proc_dir)

    try:
        logger.info("processing all sweeps of data in parallel in requested space group {} ..."
                    .format(ARGS.space_group))
        func = partial(GMCA_proc, subdir='XDS-{}'.format(ARGS.space_group))
        xds1 = parallelize(func, sweeps, nproc=ARGS.nproc)

        logger.info("scale data sweeps processed together ...")

        s1 = Process(target=scale_xds, args=(xds1, "scale-1", True))
        s1.start()
        s1.join()

        logger.info("{} job completed.".format("GMCAproc Process"))
        return 0
    except Exception as e:
        raise e
