from __future__ import absolute_import

import os, logging
from .geometry import Geometry
from .indexamajig import Indexamajig
from ..common.utils import parallelize
from .smx_get_args import SMX_ARGS
from .cell import Cell
from ..process.GMCAproc import setup_logging

logger = logging.getLogger(__name__)


class Sweep:
    def __init__(self, id, run_name, image_list, directory):
        self.run_name = run_name
        self.id = id
        self.image_list = image_list
        self.directory = directory
        self.geometry = None
        self.sweeps = []

    def set_geometry(self, geometry_file):
        self.geometry = geometry_file

    def write_filelist(self, dest='filelist'):
        out = [os.path.join(self.directory, f) for f in self.image_list]
        with open(dest, "w") as fh:
            fh.write("\n".join(out))

        return dest

    def __repr__(self):
        return "sweep_id={} directory={} run_name={} image_list={}".format(self.id,
                                                                           self.directory,
                                                                           self.run_name,
                                                                           self.image_list)


class Scanner:
    def __init__(self,
                 directory,
                 proc_dir=".",
                 cal_geometry_per_run=False,
                 chunk_size=1000,
                 nproc=4):
        self.directory = os.path.abspath(directory)
        self.max_size = chunk_size
        self.runs = None
        self.proc_dir = os.path.abspath(proc_dir)
        self.cal_geometry_per_run = cal_geometry_per_run
        self.nproc = nproc
        self.stream_dir = None
        self.sweeps = []
        self.cell = Cell().get_option_p(dest=os.path.join(self.proc_dir, "calib", "pdb", "crystfel.cell"))

    def get_runs(self):

        runs = sorted([d for d in os.listdir(self.directory)
                       if d.startswith("run")
                       and os.path.isdir(os.path.join(self.directory, d))])

        if SMX_ARGS.manage:
            allowed_run = ["run{:04d}".format(int(d)) for d in SMX_ARGS.manage]
            return [r for r in runs if r in allowed_run]

        return runs

    def get_images(self, d):
        if not os.path.isabs(d):
            d = os.path.join(self.directory, d)
        return sorted([f for f in os.listdir(d) if f.endswith(".cbf")])

    def split_filelist(self, flist):
        l = len(flist)
        n = l // self.max_size
        if l % self.max_size != 0:
            n += 1

        return [flist[i * self.max_size:(i + 1) * self.max_size] for i in range(n)]

    def scan(self):
        self.runs = self.get_runs()
        for run_dir in self.runs:
            flist = self.get_images(run_dir)

            if not flist:
                continue

            if not self.cal_geometry_per_run:
                geom_prefix = "current_geometry"
            else:
                geom_prefix = run_dir

            f0 = os.path.join(self.directory, run_dir, flist[0])
            geom = "{}/calib/geometry/{}.geom".format(os.path.abspath(self.proc_dir), geom_prefix)
            if not os.path.exists(geom):
                Scanner.cal_geometry(f0, geom)

            for i, image_list in enumerate(self.split_filelist(flist)):
                sweep_id = "{}-{}".format(run_dir, i + 1)
                new_sweep = Sweep(sweep_id, run_dir, image_list, os.path.join(self.directory, run_dir))
                new_sweep.set_geometry(geom)
                self.sweeps.append(new_sweep)

        return self.sweeps

    def index_all(self):
        '''run indexamajig for each sweep in a separate directory'''
        self.scan()
        logger.info("nruns={} nsweeps = {}".format(len(self.runs), len(self.sweeps)))
        if not self.sweeps:
            logger.info("nothing to process at ".format(self.directory))
            return

        self.stream_dir = os.path.join(self.proc_dir, "indexing", "streams")
        if not os.path.exists(self.stream_dir):
            os.makedirs(self.stream_dir)

        parallelize(self.index, self.sweeps, nproc=self.nproc)

    def index(self, sweep):
        outdir = os.path.abspath(os.path.join(self.proc_dir, "indexing", sweep.id))
        if not os.path.exists(outdir):
            os.makedirs(outdir)
        filelist = sweep.write_filelist(os.path.join(outdir, "filelist"))
        index = Indexamajig(filelist,
                            wdir=outdir,
                            pdb_file=self.cell,
                            geometry_file=sweep.geometry,
                            internal_peakfinding_parameters={"algorithm": SMX_ARGS.peak_algorithm,
                                                             "min_snr": SMX_ARGS.min_snr,
                                                             "min_pix_count": SMX_ARGS.min_pix,
                                                             "max_pix_count": SMX_ARGS.max_pix,
                                                             "threshold": SMX_ARGS.threshold,
                                                             },
                            indexing_parameters={"indexing_methods": SMX_ARGS.indexing_methods},
                            integration_paramters={'int_radius': SMX_ARGS.int_radius},
                            output_stream="{}/{}.stream".format(self.stream_dir, sweep.id),
                            )
        index.run()

    @staticmethod
    def cal_geometry(cbf_file, out_geom_file):
        base, ext = os.path.split(out_geom_file)
        if not os.path.exists(base):
            os.makedirs(base)
        Geometry(cbf_file, geom_file=out_geom_file).generate(disable_data_label=True)


if __name__ == '__main__':
    setup_logging()
    s = Scanner(SMX_ARGS.data_dir)
    s.index_all()