from __future__ import absolute_import

import logging, re, os

from ..common.spacegroup import Spacegroup
from .smx_get_args import SMX_ARGS

logger = logging.getLogger(__name__)

cell_template1 = \
    """CrystFEL unit cell file version 1.0

lattice_type = {}
centering = {}
{}
"""

cell_template2 = \
    """CrystFEL unit cell file version 1.0

lattice_type = {}
centering = {}
{}
a = {} A
b = {} A
c = {} A
al = {} deg
be = {} deg
ga = {} deg
"""

pdb_template = \
    """REMARK
REMARK 1 -  6        Record name    "CRYST1"
REMARK 7 - 15        Real(9.3)      a (Angstroms)
REMARK 16 - 24       Real(9.3)      b (Angstroms)     
REMARK 25 - 33       Real(9.3)      c (Angstroms)     
REMARK 34 - 40       Real(7.2)      alpha (degrees)   
REMARK 41 - 47       Real(7.2)      beta (degrees)    
REMARK 48 - 54       Real(7.2)      gamma (degrees)   
REMARK 56 - 66       LString        Space group       
REMARK 67 - 70       Integer        Z value           
REMARK CRYST1   45.230   50.140   50.720  90.00  90.00  90.00 P 21 21 21    4
CRYST1{:9.3f}{:9.3f}{:9.3f}{:7.2f}{:7.2f}{:7.2f} {:<10}    8
"""


class Cell:
    def __init__(self):
        self.lattice_type = SMX_ARGS.lattice_type
        self.centering = SMX_ARGS.centering
        self.pdb_file = SMX_ARGS.model
        self.cell_file = SMX_ARGS.cell_file
        self.unit_cell = None
        self.space_group = None

        if SMX_ARGS.unit_cell:
            self.unit_cell = [float(e) for e in re.split(r"[,;\s]+", SMX_ARGS.unit_cell)]
            if len(self.unit_cell) != 6:
                logger.error("unable to extract unit cell")
                self.unit_cell = None

        if SMX_ARGS.space_group:
            lattice = Spacegroup.get_lattice_from_spacegroup(SMX_ARGS.space_group)
            if lattice:
                self.centering = lattice[-1]

                first_letter = lattice[0]
                if first_letter == 'm':
                    self.lattice_type = 'monoclinic'
                elif first_letter == 'a':
                    self.lattice_type = 'triclinic'
                elif first_letter == 't':
                    self.lattice_type = 'tetragonal'
                elif first_letter == 'o':
                    self.lattice_type = 'orthorhombic'
                elif first_letter == 'c':
                    self.lattice_type = 'cubic'
                elif first_letter == 'h':
                    self.lattice_type = 'hexagonal'

                # special case
                if lattice == 'hR':
                    self.centering = 'P'
                    self.lattice_type = 'rhombohedral'

        if SMX_ARGS.space_group and SMX_ARGS.space_group in Spacegroup.HM:
            self.space_group = Spacegroup.HM.get(SMX_ARGS.space_group)

    def get_cryst1(self, dest=None):
        if self.unit_cell and self.space_group:
            cryst1 = pdb_template.format(*[*self.unit_cell, self.space_group])

            if dest:
                base, file = os.path.split(dest)
                if not os.path.exists(base):
                    os.makedirs(base)

                with open(dest, "w") as fh:
                    fh.write(cryst1)
                logger.info("CRYST1 was saved as {}".format(dest))
                return dest

            return cryst1
        else:
            logger.info("cannot write CRYST1 record, check unit_cell or space_group definition.")

    def get_cell_file(self, dest=None):
        out = None

        if self.lattice_type:
            if self.lattice_type == 'monoclinic':
                unique_axis = 'unique_axis = b'
            else:
                unique_axis = 'unique_axis = c'

            if self.lattice_type and self.centering:
                out = cell_template1.format(self.lattice_type, self.centering, unique_axis)

            if self.centering and self.unit_cell:
                out = cell_template2.format(self.lattice_type, self.centering, unique_axis, *self.unit_cell)

            if dest and out:
                base, file = os.path.split(dest)
                if not os.path.exists(base):
                    os.makedirs(base)

                with open(dest, "w") as fh:
                    fh.write(out)
                logger.info("crystfel cell file was saved as {}".format(dest))
                return dest

        return out

    def get_option_p(self, fmt='cell', dest=None):

        if self.cell_file and os.path.exists(self.cell_file):
            return self.cell_file

        if self.pdb_file and os.path.exists(self.pdb_file):
            return self.pdb_file

        if fmt == 'pdb' and self.unit_cell and self.space_group:
            return self.get_cryst1(dest=dest)

        return self.get_cell_file(dest=dest)


if __name__ == '__main__':
    # print(Cell().get_cryst1())
    # print(Cell().get_cell_file())
    print(Cell().get_option_p(dest="x.cell"))
