import glob
import json
import logging
import os
import time
from collections import defaultdict

import pyqtgraph as pg
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import QThread, pyqtSignal
from sortedcollections import SortedDict

from .adxv import AdxvSocket, AdxvServer
from .command import Command
from .crystfel_symm import CRYST_SYMMETRY
from .smx_get_args import SMX_ARGS
from ..tasks.smx_merger import SmxMerger

logger = logging.getLogger(__name__)

'''
self.data memory usage:

80kB per image, 800MB 10000 images, 80GB per million images (assuming all diffract well)

'''


class FileChooser(QtGui.QWidget):

    def __init__(self, pattern="CrystFEL cell files (*.cell);;All files (*)"):
        super().__init__()
        self.title = 'PyQt5 file dialogs - pythonspot.com'
        self.pattern = pattern
        self.left = 10
        self.top = 10
        self.width = 640
        self.height = 480
        self.initUI()

    def initUI(self):
        self.setWindowTitle(self.title)
        self.setGeometry(self.left, self.top, self.width, self.height)

        # self.openFileNameDialog()
        # self.openFileNamesDialog()
        # self.saveFileDialog()

        # self.show()

    def openFileNameDialog(self):
        options = QtWidgets.QFileDialog.Options()
        options |= QtWidgets.QFileDialog.DontUseNativeDialog
        fileName, _ = QtWidgets.QFileDialog.getOpenFileName(self, "QFileDialog.getOpenFileName()", SMX_ARGS.proc_dir,
                                                            self.pattern,
                                                            options=options)
        return fileName

    def openFileNamesDialog(self):
        options = QtWidgets.QFileDialog.Options()
        options |= QtWidgets.QFileDialog.DontUseNativeDialog
        files, _ = QtWidgets.QFileDialog.getOpenFileNames(self, "QFileDialog.getOpenFileNames()", "",
                                                          "All Files (*);;Python Files (*.py)", options=options)
        if files:
            print(files)
        return files

    def saveFileDialog(self):
        options = QtWidgets.QFileDialog.Options()
        options |= QtWidgets.QFileDialog.DontUseNativeDialog
        fileName, _ = QtWidgets.QFileDialog.getSaveFileName(self, "QFileDialog.getSaveFileName()", "",
                                                            "All Files (*);;Text Files (*.txt)", options=options)
        if fileName:
            print(fileName)
        return fileName


class Merger(QThread):
    signal = pyqtSignal('PyQt_PyObject')

    def __init__(self):
        QThread.__init__(self)
        self.input_stream = None
        self.real_symmetry = None
        self.cell_file = None
        self.apparent_symmetry = None
        self.refine_model = None
        self.ambiguity_operator = None
        self.highres = None
        self.wdir = None
        self.subdir = "merge"

    def run(self):
        merger = SmxMerger(self.input_stream,
                           self.real_symmetry,
                           self.cell_file,
                           apparent_symmetry=self.apparent_symmetry,
                           ambiguity_operator=self.ambiguity_operator,
                           refine_model=self.refine_model,
                           wdir=self.wdir,
                           subdir=self.subdir,
                           highres=self.highres,
                           )
        merger.run()
        self.signal.emit(self.wdir)


class Plotter(QtGui.QWidget):
    # interval is sent in milliseconds
    def __init__(self, queue, stream_file, index_queue=None, size=(1200, 800), interval=200):
        super(Plotter, self).__init__()
        # Data stuff
        self.size = size
        self._interval = interval
        self.queue = queue
        self.data = defaultdict(SortedDict)
        self.x = [0]
        self.y = [0]
        self.n_data = 0
        self.index_queue = index_queue

        self.minimum_spots = 10
        self.stream_file = stream_file
        self.force_update = False
        self.last_image_index = None
        self.last_prefix = None
        self.all_prefices = list()
        self.adxv_server = None
        self.adxv_client = None
        self.timer = None
        self.merge_subdir = "merge"

        self.red = QtGui.QColor(255, 0, 0)
        self.green = QtGui.QColor(0, 255, 0)

        self.key = 'num_peaks'
        self.filter = 'Show images w >=10 peaks'
        self.hitlist_version = 0


        self.merge_thread = Merger()
        self.merge_thread.signal.connect(self.on_merge_finished)

        self.init_gui()
        self.init_adxv()
        self.start_update_timer()

    def init_gui(self):
        self.pushButtonColor = "QPushButton:checked { background-color: rgb(180, 215, 245); }"

        self.vbox = QtGui.QVBoxLayout()
        self.setLayout(self.vbox)

        # PyQtGraph stuff
        self.pltw = pg.PlotWidget()
        self.pltw.resize(*self.size)
        self.pltw.showGrid(x=True, y=True)
        self.pltw.setLabel('left', 'no. spots', '')
        self.pltw.enableAutoScale()
        self.pltw.setLabel('bottom', 'frame no', '')

        self.info_label = QtGui.QTextEdit("Info")
        self.info_label.setReadOnly(True)
        self.info_label.setFrameShape(QtGui.QFrame.Panel)
        self.info_label.setFrameShadow(QtGui.QFrame.Plain)
        self.info_label.setLineWidth(1)
        self.info_label.setStyleSheet(
            """QTextEdit {background-color: #333;
                               color: #00FF00;
                               text-decoration: underline;
                               font-family: Courier;}""")

        self.info_label.setSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Expanding)

        hbox_info = QtGui.QHBoxLayout()
        label_info = QtWidgets.QLabel("Info  ")
        hbox_info.addWidget(label_info)
        hbox_info.addWidget(self.info_label)

        self.status_label = QtWidgets.QLabel("Status")
        self.status_label.setFrameShape(QtGui.QFrame.Panel)
        self.status_label.setFrameShadow(QtGui.QFrame.Plain)
        self.status_label.setLineWidth(1)
        self.status_label.setStyleSheet('color: red')
        self.status_label.setSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Expanding)

        hbox_status = QtGui.QHBoxLayout()
        label_status = QtWidgets.QLabel("Status")
        hbox_status.addWidget(label_status)
        hbox_status.addWidget(self.status_label)

        self.vbox.addWidget(self.pltw)
        self.vbox.addLayout(hbox_info)
        self.vbox.addLayout(hbox_status)

        self.curve = self.pltw.plot(self.x, self.y, pen=(200, 200, 200), symbolBrush=(255, 0, 0), symbolPen='w')

        # make the points clickable
        self.pltw.scene().sigMouseClicked.connect(self.show_image_with_adxv)

        self.tabs = QtWidgets.QTabWidget()
        self.tab1 = QtWidgets.QWidget()
        self.tab2 = QtWidgets.QWidget()
        self.tab3 = QtWidgets.QWidget()

        self.tabs.addTab(self.tab1, "Graph Control")
        self.tabs.addTab(self.tab2, "Adxv Control")
        self.tabs.addTab(self.tab3, "Merging")

        self.plot_control_ui()  # tab1
        self.adxv_ui()  # tab2
        self.merge_ui()  # tab3

        self.vbox.addWidget(self.tabs)

        self.show()

    def plot_control_ui(self):

        self.tab1.layout = QtGui.QGridLayout()

        l_yvalue = QtWidgets.QLabel("Choose y-axis:")
        self.y_value = QtGui.QComboBox()
        self.y_value.addItem('num_peaks')
        self.y_value.addItem('peak_resolution')
        self.y_value.addItem('num_reflections')

        l_select_run = QtWidgets.QLabel("Select prefix")
        self.selectRun = QtGui.QComboBox()

        self.filterImages = QtGui.QComboBox()
        self.filterImages.addItem('Show all images')
        self.filterImages.addItem('Show images w >={} peaks'.format(self.minimum_spots))
        self.filterImages.addItem('Show indexed images')
        self.filterImages.setCurrentIndex(1)

        self.pauseUpdateButton = QtGui.QPushButton("Pause update")
        self.pauseUpdateButton.setCheckable(True)
        self.pauseUpdateButton.setStyleSheet(self.pushButtonColor)

        self.sviewButton = QtGui.QPushButton("Sview")
        self.cellExplorerButton = QtGui.QPushButton("cell_explorer")
        self.exportButton = QtGui.QPushButton("Export hitlist")

        self.tab1.layout.addWidget(l_yvalue, 0, 0, alignment=QtCore.Qt.AlignRight)
        self.tab1.layout.addWidget(self.y_value, 0, 1)

        self.tab1.layout.addWidget(l_select_run, 1, 0, alignment=QtCore.Qt.AlignRight)
        self.tab1.layout.addWidget(self.selectRun, 1, 1)
        self.tab1.layout.addWidget(self.filterImages, 2, 1)

        self.tab1.layout.addWidget(self.pauseUpdateButton, 3, 0)
        self.tab1.layout.addWidget(self.sviewButton, 3, 1)
        self.tab1.layout.addWidget(self.cellExplorerButton, 4, 0)
        self.tab1.layout.addWidget(self.exportButton, 4, 1)

        self.tab1.setLayout(self.tab1.layout)

        # actions
        self.y_value.currentIndexChanged.connect(self.on_yaxis_selection)
        self.selectRun.currentIndexChanged.connect(self.on_run_change)
        self.filterImages.currentIndexChanged.connect(self.on_image_filters)
        # self.terminalButton.clicked.connect(self.on_terminal_clicked)
        self.pauseUpdateButton.clicked.connect(self.on_pause)
        self.sviewButton.clicked.connect(self.on_sview_clicked)
        self.cellExplorerButton.clicked.connect(self.on_cellexplorer_clicked)
        self.exportButton.clicked.connect(self.on_export_hitlist)

    def adxv_ui(self):
        self.adxvButton = QtGui.QPushButton("Adxv")

        self.adxvCtrlButton = QtGui.QPushButton("Adxv Control Panel")
        self.adxvCtrlButton.setCheckable(True)
        self.adxvCtrlButton.setStyleSheet(self.pushButtonColor)

        self.showResolutionRings = QtGui.QPushButton("Resolution Rings")
        self.showResolutionRings.setCheckable(True)
        self.showResolutionRings.setStyleSheet(self.pushButtonColor)

        self.showPeaks = QtGui.QPushButton("Show peaks")
        self.showPeaks.setCheckable(True)
        self.showPeaks.setStyleSheet(self.pushButtonColor)

        self.showPredictions = QtGui.QPushButton("Show predictions")
        self.showPredictions.setCheckable(True)
        self.showPredictions.setStyleSheet(self.pushButtonColor)

        l_spot_radius = QtWidgets.QLabel("Peak radius")
        self.spotRadius = QtWidgets.QLineEdit()
        self.spotRadius.setText("10")
        self.spotRadius.setFixedWidth(40)

        l_reflection_radius = QtWidgets.QLabel("Reflection radius")
        self.reflectionRadius = QtWidgets.QLineEdit()
        self.reflectionRadius.setText("4")
        self.reflectionRadius.setFixedWidth(40)

        # adxv controls
        self.tab2.layout = QtGui.QGridLayout()
        self.tab2.layout.addWidget(self.adxvButton, 0, 0)
        self.tab2.layout.addWidget(self.adxvCtrlButton, 0, 1)
        self.tab2.layout.addWidget(self.showResolutionRings, 1, 0)
        self.tab2.layout.addWidget(self.showPeaks, 2, 0)
        self.tab2.layout.addWidget(l_spot_radius, 2, 1, alignment=QtCore.Qt.AlignRight)
        self.tab2.layout.addWidget(self.spotRadius, 2, 2, 1, 1)

        self.tab2.layout.addWidget(self.showPredictions, 3, 0)
        self.tab2.layout.addWidget(l_reflection_radius, 3, 1, alignment=QtCore.Qt.AlignRight)
        self.tab2.layout.addWidget(self.reflectionRadius, 3, 2, 1, 1)

        self.tab2.setLayout(self.tab2.layout)

        # actions
        self.adxvButton.clicked.connect(self.on_adxv_clicked)
        self.adxvCtrlButton.clicked.connect(self.on_adxv_ctrl_clicked)
        self.showResolutionRings.toggled.connect(self.on_resolution_toggle)
        self.showPeaks.toggled.connect(self.redraw_spots)
        self.showPredictions.toggled.connect(self.redraw_spots)
        self.spotRadius.textChanged.connect(self.redraw_spots)
        self.reflectionRadius.textChanged.connect(self.redraw_spots)

    def merge_ui(self):
        # merging button & create cell file
        self.tab3.layout = QtWidgets.QGridLayout()

        self.merge_btn = QtGui.QPushButton("Partialator")
        self.tab3.layout.addWidget(self.merge_btn, 0, 0, 5, 1)

        self.tab3.layout.addWidget(QtWidgets.QLabel(" in sub-dir: "), 0, 1)

        self.merge_subdir = QtWidgets.QLineEdit()
        self.merge_subdir.setText("merge")
        self.merge_subdir.setFixedWidth(50)
        self.tab3.layout.addWidget(self.merge_subdir, 0, 2)

        self.tab3.layout.addWidget(QtWidgets.QLabel(" with "), 1, 1)

        self.merged_stream = QtWidgets.QLineEdit()
        self.merged_stream.setText(self.stream_file)
        self.merged_stream.setFixedWidth(200)

        self.choose_merge_file_btn = QtGui.QPushButton("stream file: ")

        self.tab3.layout.addWidget(self.choose_merge_file_btn, 1, 2)
        self.tab3.layout.addWidget(self.merged_stream, 1, 3)

        # merging parameters
        self.l_rsymm = QtWidgets.QLabel("Real symm")
        self.rsymm = QtGui.QComboBox()
        self.rsymm.addItems(CRYST_SYMMETRY)
        self.tab3.layout.addWidget(self.l_rsymm, 2, 1)
        self.tab3.layout.addWidget(self.rsymm, 2, 2)

        self.l_asymm = QtWidgets.QLabel("Apparent symm")
        self.l_asymm = QtWidgets.QComboBox()
        self.l_asymm.addItem("Apparent symmetry")
        self.l_asymm.addItem("Ambiguity operator")

        self.asymm = QtWidgets.QLineEdit()
        self.asymm.setFixedWidth(100)
        self.tab3.layout.addWidget(self.l_asymm, 2, 3)
        self.tab3.layout.addWidget(self.asymm, 2, 4)

        self.highres = QtWidgets.QLineEdit()
        self.highres.setText("")
        self.highres.setFixedWidth(40)
        self.tab3.layout.addWidget(QtWidgets.QLabel("h.res"), 3, 1)
        self.tab3.layout.addWidget(self.highres, 3, 2)

        self.l_refine = QtWidgets.QLabel("Partiality model")
        self.refine = QtGui.QComboBox()
        self.refine.addItem("--iterations=1 --model=unity")
        self.refine.addItem("--iterations=1 --model=xsphere")
        self.refine.setEditable(True)
        self.tab3.layout.addWidget(self.l_refine, 3, 3)
        self.tab3.layout.addWidget(self.refine, 3, 4)

        self.cell_file_btn = QtGui.QPushButton("Cell file")
        self.cell_file_fld = QtWidgets.QLineEdit()
        self.cell_file_fld.setFixedWidth(200)
        cell_files = glob.glob("{}/*.cell".format(SMX_ARGS.proc_dir))
        if cell_files:
            self.cell_file_fld.setText(cell_files[0])

        self.tab3.layout.addWidget(self.cell_file_btn, 4, 1)
        self.tab3.layout.addWidget(self.cell_file_fld, 4, 2)

        self.tab3.setLayout(self.tab3.layout)

        self.cell_file_btn.clicked.connect(self.on_select_cell_file)
        self.merge_btn.clicked.connect(self.on_merge)
        self.choose_merge_file_btn.clicked.connect(self.on_select_merged_stream)

    def init_adxv(self):
        self.adxv_server = AdxvServer()
        self.adxv_client = AdxvSocket()
        self.adxv_client.close("Load")
        self.adxv_client.close('Control')
        self.adxv_client.add_type(0, 'red', 10)
        self.adxv_client.add_type(1, 'green', 4)
        self.adxv_client.rings('off')
        self.adxv_client.remove_spots()

    def start_update_timer(self):
        # QTimer
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.updateplot)
        self.timer.start(self._interval)

    def set_stream_file(self, stream_file):
        self.stream_file = stream_file

    def on_pause(self):
        if not self.pauseUpdateButton.isChecked():
            self.updateplot()

    def on_merge(self):
        self.merge_thread.input_stream = self.merged_stream.text()
        self.merge_thread.real_symmetry = self.rsymm.currentText()
        self.merge_thread.cell_file = self.cell_file_fld.text()

        if "operator" in self.l_asymm.currentText().lower():
            self.merge.ambiguity_operator = self.asymm.text()
            self.merge_thread.apparent_symmetry = None
        else:
            self.merge_thread.apparent_symmetry = self.asymm.text()
            self.merge_thread.ambiguity_operator = None

        self.merge_thread.refine_model = self.refine.currentText()
        self.merge_thread.highres = self.highres.text()
        self.merge_thread.wdir = SMX_ARGS.proc_dir
        self.merge_thread.subdir = self.merge_subdir.text()
        self.merge_btn.setEnabled(False)
        self.merge_subdir.setEnabled(False)
        self.merge_thread.start()
        self.info_label.append("merge process started")

    def on_merge_finished(self):

        merge_out = os.path.join(SMX_ARGS.proc_dir, self.merge_subdir.text(), 'smx_merge.sh.out')
        if os.path.exists(merge_out):
            self.info_label.setText("merge process finished")
            with open(merge_out, 'r') as fh:
                merge_log = fh.read()
                self.info_label.append(merge_log)
        else:
            self.info_label.append("merge failed.")

        self.merge_subdir.setEnabled(True)
        self.merge_btn.setEnabled(True)

    def on_select_cell_file(self):
        cell_file = FileChooser().openFileNameDialog()
        self.cell_file_fld.setText(cell_file)

    def on_select_merged_stream(self):
        stream_file = FileChooser(pattern="CrystFEL stream files (*.streams);;All files (*)").openFileNameDialog()
        self.merged_stream.setText(stream_file)

    def on_run_change(self):
        self.force_update = True
        self.last_prefix = self.selectRun.currentText()
        self.updateplot()

        self.info_label.setText(
            ("Overall: # images to be indexed {}, # images processed: {}.\n"
             "Run {}: processed: {}, hit: {}%, indexed: {}%.").format(
                self.index_queue.qsize(),
                sum(len(v) for k, v in self.data.items()),
                self.last_prefix,
                len(self.data[self.last_prefix]),
                self.cal_hit_rate(self.last_prefix),
                self.cal_index_rate(self.last_prefix)
            ))

        self.force_update = False

    def on_image_filters(self):
        self.force_update = True
        self.filter = self.filterImages.currentText()
        self.updateplot()
        self.force_update = False

    def on_yaxis_selection(self):
        self.status_label.setText("changing y-axis")
        self.key = self.y_value.currentText()
        if self.key == 'num_peaks':
            self.pltw.setLabel('left', 'no. spots', '')
            # self.pltw.setYRange(0, 300)
        elif self.key == 'peak_resolution':
            self.pltw.setLabel('left', 'max peak resolution', '')
            # self.pltw.setYRange(0, 20)
        elif self.key == 'num_reflections':
            self.pltw.setLabel('left', 'no. predicted reflections', '')
            # self.pltw.setYRange(0, 3000)

        # self.pltw.setAutoScale()

        self.force_update = True
        self.updateplot()
        self.force_update = False

    def on_cheetah_clicked(self):
        self.status_label.setText("start cheetah-gui")
        Command('{}'.format('cheetah-gui-gmca')).run()

    def on_terminal_clicked(self):
        self.status_label.setText("start a terminal with enviroments set serial crystallography programs")
        Command('{}'.format('xterm')).run()

    def on_resolution_toggle(self):
        self.status_label.setText("resolution ring {}".format(self.showResolutionRings.isChecked()))
        if self.showResolutionRings.isChecked():
            self.adxv_client.rings('on')
        else:
            self.adxv_client.rings('off')

    def on_cellexplorer_clicked(self):
        self.status_label.setText("start cell_explorer using {}".format(self.stream_file))
        Command('{} {}'.format('cell_explorer', self.stream_file)).run()

    def on_export_hitlist(self):
        hitlist = list()
        for _, v1 in self.data.items():
            for _, v2 in v1.items():
                if v2.get('num_peaks', 0) > self.minimum_spots:
                    hitlist.append(v2.get('filename'))

        outfile = os.path.join(SMX_ARGS.proc_dir, "hitlist-v{}.txt".format(self.hitlist_version))
        with open(outfile, 'w') as fh:
            fh.write("\n".join(hitlist))
            self.hitlist_version += 1
        self.status_label.setText("hitlist (image with >={}) saved to {}, total {} images.".format(self.minimum_spots,
                                                                                                   outfile,
                                                                                                   len(hitlist)))

    def on_sview_clicked(self):
        self.status_label.setText("start sview to check for running jobs.")
        Command('{}'.format('sview')).run()

    def on_adxv_clicked(self):
        if self.adxv_server.is_running():
            self.status_label.setText('adxv server is already running, request ignored.')
        else:
            self.adxv_server = AdxvServer()
            del self.adxv_client
            time.sleep(6)
            self.adxv_client = AdxvSocket()
            self.adxv_client.close('Control')

    def on_adxv_ctrl_clicked(self):
        if not self.adxv_server.is_running():
            return

        if self.adxvCtrlButton.isChecked():
            self.adxv_client.raise_window('Control')
            self.status_label.setText('open adxv control menu.')
        else:
            self.adxv_client.close('Control')
            self.status_label.setText('close adxv control menu.')

    def show_image_with_adxv(self, event):
        index = self.get_index(event)
        my_data = self.data[self.last_prefix]

        if index in my_data:
            image_filename = my_data[index]['filename']
            if os.path.exists(image_filename):
                self.adxv_client.load_image(image_filename)
                self.info_label.append(self.update_info_label(index))
                self.status_label.setText("load {} into adxv".format(image_filename))
            else:
                self.status_label.setText("file not present: {}".format(image_filename))

            self.last_image_index = index
            self.redraw_spots()

    def update_info_label(self, index):
        my_data = self.data[self.last_prefix]

        filename = my_data[index]['filename']
        npeaks = my_data[index]['num_peaks']
        nrefls = my_data[index].get('num_reflections', 0)
        lattice = my_data[index].get('lattice_type', '')
        centering = my_data[index].get('centering', '')
        cell = my_data[index].get('cell', [])
        indexed_by = my_data[index].get('indexed_by', '')
        return "image={} n_peaks={} n_reflections={}\nlattice={} centering={}\ncell={} indexed_by={}".format(filename,
                                                                                                             npeaks,
                                                                                                             nrefls,
                                                                                                             lattice,
                                                                                                             centering,
                                                                                                             " ".join(
                                                                                                                 cell),
                                                                                                             indexed_by)

    def redraw_spots(self):
        my_data = self.data[self.last_prefix]

        if not self.last_image_index or self.last_image_index not in my_data:
            return

        index = self.last_image_index

        spots = list()
        if self.showPeaks.isChecked() and 'list_of_peaks' in my_data[index]:
            if my_data[index]['list_of_peaks']:
                spots += my_data[index]['list_of_peaks']

        # print("show peaks, refls=", self.showPeaks.isChecked(), self.showPredictions.isChecked())
        if self.showPredictions.isChecked() and 'list_of_reflections' in my_data[index]:
            spots += my_data[index]['list_of_reflections']

        self.adxv_client.add_type(0, 'red', self.spotRadius.text())
        self.adxv_client.add_type(1, 'green', self.reflectionRadius.text())
        if spots:
            self.adxv_client.add_spots(spots)
        else:
            self.adxv_client.remove_spots()

    def get_index(self, event):
        if event.button() == 1:
            items = self.pltw.scene().items(event.scenePos())
            for item in items:
                if isinstance(item, pg.ViewBox):
                    pos_x = item.mapSceneToView(event.scenePos()).x()
                    index = int(round(pos_x))
                    return index

    def get_data(self):
        result = None
        while not self.queue.empty():
            result = self.queue.get()
            prefix = result['run'] + "/" + result['prefix']
            img_num = result['img_num']
            self.data[prefix].update({img_num: result})

            if prefix not in self.all_prefices:
                self.all_prefices.append(prefix)
                self.selectRun.addItem(prefix)

            self.last_prefix = prefix
            self.selectRun.setCurrentIndex(self.all_prefices.index(prefix))

        if not result:
            return False

        self.info_label.setText(
            "Run {}: processed: {}, hit: {}%, indexed: {}%.".format(
                prefix,
                len(self.data[prefix]),
                self.cal_hit_rate(prefix),
                self.cal_index_rate(prefix)
            ))

        if not self.pauseUpdateButton.isChecked():
            if (('peaks' in self.filter and img_num >= self.minimum_spots) or
                    ('indexed' in self.filter and result.get('indexed_by', 'none') != 'none') or
                    'all' in self.filter):
                self.adxv_client.load_image(result['filename'])
                self.status_label.setText("load {} in adxv".format(result['filename']))
                self.last_image_index = img_num

                self.redraw_spots()

        # if self.n_data != len(self.data[prefix]):
        #     self.n_data = len(self.data[prefix])
        #     return True
        return True

    def cal_index_rate(self, prefix):
        if not self.data.get(prefix, None):
            return 0

        count = len(self.data[prefix])
        indexed = sum(1 for p in self.data[prefix].values() if p.get('indexed_by', 'none') != 'none')
        return round(float(indexed) * 100 / count, 2)

    def cal_hit_rate(self, prefix):
        if not self.data.get(prefix, None):
            return 0

        count = len(self.data[prefix])
        hits = sum(1 for p in self.data[prefix].values() if p['num_peaks'] >= self.minimum_spots)
        return round(float(hits) * 100 / count, 2)

    def updateplot(self):
        if self.pauseUpdateButton.isChecked():
            return

        if self.get_data() or self.force_update:

            my_data = self.data[self.last_prefix]
            sym_brush = []
            if 'all' in self.filter:
                self.x[:] = my_data.keys()
                self.y[:] = [v.get(self.key, 0) for v in my_data.values()]
                sym_brush[:] = [self.green if v.get('indexed_by', 'none') != 'none' else self.red for v in
                                my_data.values()]
            elif 'peaks' in self.filter:
                data = [(v['img_num'], v.get(self.key, 0)) for v in my_data.values() if
                        v.get('num_peaks', 0) >= self.minimum_spots]
                if not data:
                    return
                self.x[:], self.y[:] = zip(*data)
                sym_brush[:] = [self.green if my_data[x].get('indexed_by', 'none') != 'none' else self.red for x in
                                self.x]
            elif 'indexed' in self.filter:
                data = [(v['img_num'], v.get(self.key, 0), self.green) for v in my_data.values() if
                        v.get('indexed_by', 'none') != 'none']
                if not data:
                    return
                self.x[:], self.y[:], sym_brush[:] = zip(*data)

            if self.x and self.y:
                self.curve.setData(self.x, self.y, symbol='o', symbolBrush=sym_brush)

                QtWidgets.QApplication.processEvents()

    def closeEvent(self, event):
        self.adxv_client.exit()
        self.save_json()

    def save_json(self):
        print("saving streams.json to {}".format(SMX_ARGS.proc_dir))
        json_file = os.path.join(SMX_ARGS.proc_dir, "streams.json")
        with open(json_file, "w") as fh:
            json.dump(self.data, fh, indent=4)
