import os
import time
import shutil
from multiprocessing import Pool

from .cell import Cell
from .indexamajig import Indexamajig
from .geometry import Geometry
from ..file_parser.streamparser import StreamParser
from .smx_get_args import SMX_ARGS


def indexing(filelist, outq, mergeq, job_id, cell, geometry, saved_stream_dir, clean_up=True):
    tempdir = os.path.join(SMX_ARGS.proc_dir, "s{:07d}".format(job_id))

    if os.path.exists(tempdir):
        os.rmdir(tempdir)

    os.makedirs(tempdir)

    stream_fname = "s{:07d}.stream".format(job_id)
    out_stream = os.path.join(tempdir, stream_fname)

    filelist_name = os.path.join(tempdir, "filelist")
    with open(filelist_name, "w") as fh:
        fh.write("\n".join(filelist))

    index_it = Indexamajig(filelist_name,
                           wdir=tempdir,
                           pdb_file=cell,
                           geometry_file=geometry,
                           internal_peakfinding_parameters={"algorithm": SMX_ARGS.peak_algorithm,
                                                            "min_snr": SMX_ARGS.min_snr,
                                                            "min_pix_count": SMX_ARGS.min_pix,
                                                            "max_pix_count": SMX_ARGS.max_pix,
                                                            "threshold": SMX_ARGS.threshold,
                                                            },
                           indexing_parameters={"indexing_methods": SMX_ARGS.indexing_methods},
                           integration_paramters={'int_radius': SMX_ARGS.int_radius},
                           high_res=SMX_ARGS.highres,
                           temp_dir='/dev/shm',
                           output_stream=out_stream,
                           )
    index_it.run()

    # parse result & put it into queue
    sp = StreamParser(wdir=tempdir, filename=stream_fname)
    all_results = sp.get_all_results()
    save_stream = False
    for r in all_results:
        outq.put(r)
        if r['num_peaks'] >= 5:
            save_stream = True

    if save_stream:
        # print("move from {} to {}".format(out_stream, saved_stream_dir))
        shutil.move(out_stream, saved_stream_dir)
        # put it into merger for merging

        indexed_chunks = sp.get_indexed_chunks()
        if indexed_chunks:
            mergeq.put(indexed_chunks)

    if clean_up:
        shutil.rmtree(tempdir)


class ProcessManager:
    def __init__(self, task_queue, out_queue, merge_queue, chunk_size=2, geometry_file=None, pool_size=40):
        self.chunk_size = chunk_size
        self.task_queue = task_queue
        self.out_queue = out_queue
        self.merge_queue = merge_queue
        self.cell = Cell().get_option_p(dest=os.path.join(SMX_ARGS.proc_dir, "crystfel.cell"))
        self.geometry = geometry_file
        self.saved_stream_dir = os.path.join(SMX_ARGS.proc_dir, "streams")
        if not os.path.exists(self.saved_stream_dir):
            os.makedirs(self.saved_stream_dir)
        self.pool = Pool(processes=pool_size)

    def manage(self):
        job_id = 0
        try:
            filelist = list()
            waited_for = 0
            while True:
                if len(filelist) == self.chunk_size or (waited_for >= 30 and filelist):
                    job_id += 1

                    # generate geometry for the first image unless predefined
                    if not self.geometry and job_id == 1:
                        self.geometry = os.path.join(SMX_ARGS.proc_dir, "current.geom")
                        g = Geometry(filelist[0], geom_file=self.geometry)
                        g.generate(disable_data_label=True)

                        # send the text to the stream merger as well
                        self.merge_queue.put(g.get_geometry_block())

                    serial = False
                    if serial:
                        indexing(filelist,
                                 self.out_queue,
                                 self.merge_queue,
                                 job_id,
                                 self.cell,
                                 self.geometry,
                                 self.saved_stream_dir,
                                 not SMX_ARGS.debug)
                    else:
                        self.pool.apply_async(indexing, args=(filelist,
                                                              self.out_queue,
                                                              self.merge_queue,
                                                              job_id,
                                                              self.cell,
                                                              self.geometry,
                                                              self.saved_stream_dir,
                                                              not SMX_ARGS.debug
                                                              ))

                    filelist = list()
                    waited_for = 0

                if not self.task_queue.empty():
                    filelist.append(self.task_queue.get())
                    waited_for = 0
                else:
                    time.sleep(5)
                    waited_for += 5

        except:
            raise Exception("ProcessManager exception")

    def stop(self):
        if self.pool:
            self.pool.terminate()
