#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import os
import re
import shutil
import subprocess
import tempfile

import StringIO
from mod_python import apache, util


SPACE_EQUIVALENT_CHARACTERS = '#:=,()'

NON_OPTIONAL_KEYWORDS = {
    # key: (pattern, [value_indeces], type)
    'Detector_identifier': ('Detector ', [slice(1, None)], str),
    'Pixel_size': ('Pixel_size', [1, 4], float),
    'Silicon': ('Silicon', [3], float),
    'Exposure_time': ('Exposure_time', [1], float),
    'Exposure_period': ('Exposure_period', [1], float),
    'Tau': ('Tau', [1], float),
    'Count_cutoff': ('Count_cutoff', [1], int),
    'Threshold_setting': ('Threshold_setting', [1], float),
    'Gain_setting': ('Gain_setting', [1, 2], str),
    'N_excluded_pixels': ('N_excluded_pixels', [1], int),
    'Excluded_pixels': ('Excluded_pixels', [1], str),
    'Flat_field': ('Flat_field', [1], str),
    'Trim_file': ('Trim_file', [1], str),
    'Image_path': ('Image_path', [1], str),
}

OPTIONAL_KEYWORDS = {
    'Wavelength': ('Wavelength', [1], float),
    'Energy_range': ('Energy_range', [1, 2], float),
    'Detector_distance': ('Detector_distance', [1], float),
    'Detector_Voffset': ('Detector_Voffset', [1], float),
    'Beam_xy': ('Beam_xy', [1, 2], float),
    'Beam_x': ('Beam_xy', [1], float),
    'Beam_y': ('Beam_xy', [2], float),
    'Flux': ('Flux', [1], str),
    'Filter_transmission': ('Filter_transmission', [1], float),
    'Start_angle': ('Start_angle', [1], float),
    'Angle_increment': ('Angle_increment', [1], float),
    'Detector_2theta': ('Detector_2theta', [1], float),
    'Polarization': ('Polarization', [1], float),
    'Alpha': ('Alpha', [1], float),
    'Kappa': ('Kappa', [1], float),
    'Phi': ('Phi', [1], float),
    'Phi_increment': ('Phi_increment', [1], float),
    'Chi': ('Chi', [1], float),
    'Chi_increment': ('Chi_increment', [1], float),
    'Oscillation_axis': ('Oscillation_axis', [slice(1, None)], str),
    'N_oscillations': ('N_oscillations', [1], int),
    'Start_position': ('Start_position', [1], float),
    'Position_increment': ('Position_increment', [1], float),
    'Shutter_time': ('Shutter_time', [1], float),
}

ALL_KEYWORDS = {}
ALL_KEYWORDS.update(NON_OPTIONAL_KEYWORDS)
ALL_KEYWORDS.update(OPTIONAL_KEYWORDS)


class PilatusHeader(object):
    """
    Class for parsing contents of a Pilatus cbf header from a minimal
    or full cbf file.

    Parsing the Pilatus cbf header populates the header_dict dictionary,
    using Pilatus cbf header keywords as keys.
    """

    def __init__(self, filepath, non_binary_length=4096):
        self.filepath = filepath
        self.non_binary_length = non_binary_length
        self.header_dict = {}
        self.header_lines = []

        self.read_header_lines()
        self.parse_header()

    def _rawheader(self):
        return open(self.filepath, 'rb').read(self.non_binary_length)

    def has_pilatus_cbf_convention(self):
        # Check for the _array_data.header_convention data item
        pilatus_header_pattern = re.compile(
            r'''_array_data.header_convention +["']?(SLS|PILATUS)'''
            r'''_\d+(\.?\d*)*["']?''')
        return bool(pilatus_header_pattern.search(self._rawheader()))

    def read_header_lines(self):
        """
        Populate the self.header_lines list.
        """
        contents_pattern = re.compile(
            r'''_array_data.header_contents\s+;.*?;''',
            re.DOTALL)
        contents_match = contents_pattern.search(self._rawheader())
        assert contents_match is not None
        self.header_lines = contents_match.group().splitlines()

    def _spaced_header_lines(self):
        """
        Return header_lines with all space equivalent charecters converted
        to space.
        """
        spaced_header_lines = []
        for line in self.header_lines:
            for space_equivalent in SPACE_EQUIVALENT_CHARACTERS:
                line = line.replace(space_equivalent, ' ')
            spaced_header_lines.append(line)
        return spaced_header_lines

    def parse_header(self):
        """
        Populate self.header_dict with contents of Pilatus cbf header
        """
        assert self.has_pilatus_cbf_convention()
        if len(self.header_lines) == 0:
            self.read_header_lines()
        # parse the header lines
        for key, (pattern, valueindices, datatype) in ALL_KEYWORDS.items():
            for line in self._spaced_header_lines():
                if pattern in line:
                    values = []
                    for idx in valueindices:
                        try:
                            # handle multiple or single values
                            if isinstance(idx, slice):
                                values += line.split()[idx]
                            else:
                                values.append(line.split()[idx])
                        except IndexError:
                            print('No value at index %d on header line: \n'
                                  '%s' % (idx, line))
                    value = self._datatype_handling(values, key, datatype)
                    if value is not None:
                        self.header_dict[key] = value

    def _datatype_handling(self, values, key, datatype):
        # handle rare cases of value "not set"
        if datatype is float and values[0] == 'not':
            # NON_OPTIONAL_KEYWORDS should always have value, at least NaN
            if key in NON_OPTIONAL_KEYWORDS:
                return float('NaN')
            else:
                return None
        # do the conversion for standard cases
        if len(values) == 1:
            values = datatype(values[0])
        else:
            if datatype is str:
                values = ' '.join(values)
            else:
                values = tuple([datatype(v) for v in values])
        return values

    def get_beam_xy_mm(self, factor=1000):
        return tuple([n * size * factor for n, size in zip(
            self.header_dict['Beam_xy'], self.header_dict['Pixel_size'])])

    def get_date_time(self):
        """
        Return date and time of image acquistion.

        Works for format of current camserver versions
        2011-06-04T04:57:02.976
        or format of old camserver versions
        2011/Sep/12 09:21:27.252
        """
        date_time_pattern = re.compile(
            r'(\d{4}-\d{2}-\d{2}T|\d{4}/\D+/\d{2} )\d{2}:\d{2}:\d{2}.\d+')
        return date_time_pattern.search(self._rawheader()).group()

    def get_time(self):
        time_pattern = re.compile(r'\d{2}:\d{2}:\d{2}.\d+')
        return time_pattern.search(self.get_date_time()).group()

    date_time = property(get_date_time)
    time = property(get_time)


def get_beamstop_shadow(xbeam, ybeam, radius=100):
    # radius of beam stop in pixels
    xbeam = float(xbeam)
    ybeam = float(ybeam)
    return tuple([0, int(ybeam - radius), int(xbeam + radius), int(ybeam + radius)])


pilatus_template = \
"""
photon_energy = {energy} ; in eV
adu_per_photon = 1.0
clen = {distance}  ; in m
coffset = 0.
res = 5814.0  ; in pixels

;data = /entry_1/data_1/data ; /entry_1/data_1/data
;dim0 = %
;dim1 = ss
;dim2 = fs

0/min_fs = 0
0/max_fs = 2462
0/min_ss = 0
0/max_ss = 2526
0/corner_x = -{xbeam} 
0/corner_y = -{ybeam}
0/fs = x
0/ss = y


bad_beamstop/min_fs = {xmin}
bad_beamstop/min_ss = {ymin}
bad_beamstop/max_fs = {xmax}
bad_beamstop/max_ss = {ymax}

bad_v1/min_fs = 487
bad_v1/min_ss = 0
bad_v1/max_fs = 493
bad_v1/max_ss = 2526
bad_v2/min_fs = 981
bad_v2/min_ss = 0
bad_v2/max_fs = 987
bad_v2/max_ss = 2526
bad_v3/min_fs = 1475
bad_v3/min_ss = 0
bad_v3/max_fs = 1481
bad_v3/max_ss = 2526
bad_v4/min_fs = 1969
bad_v4/min_ss = 0
bad_v4/max_fs = 1975
bad_v4/max_ss = 2526

bad_h1/min_fs = 0
bad_h1/min_ss = 195
bad_h1/max_fs = 2462
bad_h1/max_ss = 211
bad_h2/min_fs = 0
bad_h2/min_ss = 407
bad_h2/max_fs = 2462
bad_h2/max_ss = 423
bad_h3/min_fs = 0
bad_h3/min_ss = 619
bad_h3/max_fs = 2462
bad_h3/max_ss = 635
bad_h4/min_fs = 0
bad_h4/min_ss = 831
bad_h4/max_fs = 2462
bad_h4/max_ss = 847
bad_h5/min_fs = 0
bad_h5/min_ss = 1043
bad_h5/max_fs = 2462
bad_h5/max_ss = 1059
bad_h6/min_fs = 0
bad_h6/min_ss = 1255
bad_h6/max_fs = 2462
bad_h6/max_ss = 1271
bad_h7/min_fs = 0
bad_h7/min_ss = 1467
bad_h7/max_fs = 2462
bad_h7/max_ss = 1483
bad_h8/min_fs = 0
bad_h8/min_ss = 1679
bad_h8/max_fs = 2462
bad_h8/max_ss = 1695
bad_h9/min_fs = 0
bad_h9/min_ss = 1891
bad_h9/max_fs = 2462
bad_h9/max_ss = 1907
bad_ha/min_fs = 0
bad_ha/min_ss = 2103
bad_ha/max_fs = 2462
bad_ha/max_ss = 2119
bad_hb/min_fs = 0
bad_hb/min_ss = 2315
bad_hb/max_fs = 2462
bad_hb/max_ss = 2331
"""

eiger_template = \
"""
photon_energy = {energy} ; in eV
adu_per_photon = 1.0
clen = {distance}  ; in m
coffset = 0.
res = 13333.3  ; in pixels

;data = /entry_1/data_1/data ; /entry_1/data_1/data
;dim0 = %
;dim1 = ss
;dim2 = fs

0/min_fs = 0
0/max_fs = 4149
0/min_ss = 0
0/max_ss = 4370
0/corner_x = -{xbeam}
0/corner_y = -{ybeam}
0/fs = x
0/ss = y


bad_beamstop/min_fs = {xmin}
bad_beamstop/min_ss = {ymin}
bad_beamstop/max_fs = {xmax}
bad_beamstop/max_ss = {ymax}


bad_v1/min_fs = 1030
bad_v1/min_ss = 0
bad_v1/max_fs = 1039
bad_v1/max_ss = 4370
bad_v2/min_fs = 2070
bad_v2/min_ss = 0
bad_v2/max_fs = 2079
bad_v2/max_ss = 4370
bad_v3/min_fs = 3110
bad_v3/min_ss = 0
bad_v3/max_fs = 3119
bad_v3/max_ss = 4370

bad_h1/min_fs = 0
bad_h1/min_ss = 514
bad_h1/max_fs = 4149
bad_h1/max_ss = 550
bad_h2/min_fs = 0
bad_h2/min_ss = 1065
bad_h2/max_fs = 4149
bad_h2/max_ss = 1101
bad_h3/min_fs = 0
bad_h3/min_ss = 1616
bad_h3/max_fs = 4149
bad_h3/max_ss = 1652
bad_h4/min_fs = 0
bad_h4/min_ss = 2167
bad_h4/max_fs = 4149
bad_h4/max_ss = 2203
bad_h5/min_fs = 0
bad_h5/min_ss = 2718
bad_h5/max_fs = 4149
bad_h5/max_ss = 2754
bad_h6/min_fs = 0
bad_h6/min_ss = 3269
bad_h6/max_fs = 4149
bad_h6/max_ss = 3305
bad_h7/min_fs = 0
bad_h7/min_ss = 3820
bad_h7/max_fs = 4149
bad_h7/max_ss = 3856
"""


def get_geometry(wdir, detector, energy, distance, xbeam, ybeam):
    template = pilatus_template if 'pilatus' in detector.lower() else eiger_template
    xmin, ymin, xmax, ymax = get_beamstop_shadow(xbeam, ybeam, radius=80)
    geometry = template.format(energy=energy,
                               distance=distance,
                               xbeam=xbeam,
                               ybeam=ybeam,
                               xmin=xmin,
                               ymin=ymin,
                               xmax=xmax,
                               ymax=ymax
                               )

    geometry_file = os.path.join(wdir, "geometry.geom")
    with open(geometry_file, 'w') as fh:
        fh.write(geometry)

    return geometry_file


def get_filelist(wdir, filelist):
    filelist_fn = os.path.join(wdir, "filelist")
    with open(filelist_fn, 'w') as fh:
        fh.write(filelist)

    return filelist_fn


def run_indexamajig(args):
    ''' run indexamajig directly '''

    filename = args['file_name']
    if not os.path.isfile(filename):
        print("Error: file {} does not exist.".format(filename))
    filename = os.path.abspath(filename)

    h_dict = PilatusHeader(filename, 4096).header_dict
    distance = h_dict['Detector_distance']
    energy = round(12398.42 / h_dict['Wavelength'], 2)
    xbeam = h_dict['Beam_x']
    ybeam = h_dict['Beam_y']
    detector = h_dict['Detector_identifier']

    wdir = tempfile.mkdtemp(prefix='indexamajig_tmp')

    geometry_file = get_geometry(wdir, detector, energy, distance, xbeam, ybeam)
    filelist = get_filelist(wdir, filename)

    try:
        ostream = '{}/out.stream'.format(wdir)
        cmd = "source /mnt/software/px/serial-x/serialx-setup.sh; indexamajig --indexing=mosflm -g {} -i {} -o {}".format(
            geometry_file, filelist, ostream)
        out = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
        return json.dumps(parse_indexamajig(ostream), indent=4)
    except subprocess.CalledProcessError as e:
        raise Exception('ERROR: {}'.format(e.output))
    finally:
        shutil.rmtree(wdir)


def round1(num):
    return str(round(float(num), 1))


def round_nm2ang(index, num):
    if index < 3:
        return str(round(float(num) * 10, 1))
    else:
        return str(round(float(num), 1))


# reflection is shown as type 0 spots
def get_adxv_spots(peaks, maxs=1000):
    if len(peaks) <= 2:
        return []

    xy_type = [peak.split()[:2] for peak in peaks[2:]][:maxs]
    xy_type = [list(map(round1, p)) + ['0'] for p in xy_type]
    return [" ".join(xyt) for xyt in xy_type]


# reflection is shown as type 1 spots
def get_adxv_predictions(reflections, maxp=2000):
    if len(reflections) <= 2:
        return []

    xy_type = [reflection.split()[-3:-1] for reflection in reflections[2:]]
    if len(xy_type) > maxp:
        mid = len(xy_type) // 2
        shift = maxp // 2
        xy_type = xy_type[mid - shift:mid + shift]

    xy_type = [list(map(round1, p)) + ['1'] for p in xy_type]
    return [" ".join(xyt) for xyt in xy_type]


def parse_indexamajig(outfile):
    # only deals with one image

    with open(outfile, 'r') as fh:
        lines = fh.readlines()

    results = {}
    peak_end = None
    peak_start = None

    refl_start = None
    refl_end = None

    chunk_start = None
    chunk_end = None
    for i, line in enumerate(lines):
        line = line.strip("\n")

        if '---- Begin chunk -----' in line:
            chunk_start = i

        if '----- End chunk -----' in line:
            chunk_end = i

        if "Image filename: " in line:
            full_filename = line.split()[-1]
            short_filename = full_filename[full_filename.rindex('/') + 1:]
            results['filename'] = full_filename
            results['run'] = full_filename.split('/')[-2]
            results['img_num'] = int(line[:-4].split('_')[-1])
            results['prefix'] = short_filename[:short_filename.rindex('_')]

        if "hit = " in line:
            results['hit'] = int(line.split()[-1])

        if "num_peaks = " in line:
            results['num_peaks'] = int(line.split()[-1])

        if "peak_resolution = " in line:
            peak_resol = line.split()[-2]
            results['peak_resolution'] = 0.0 if peak_resol == 'inf' else float(peak_resol)

        if "Peaks from peak search" in line:
            peak_start = i

        if "End of peak list" in line:
            peak_end = i

        # extra if indexed
        if 'Cell parameters' in line:
            cell = line.replace("nm,", "").split()[2:-1]
            results['cell'] = [round_nm2ang(i, e) for i, e in enumerate(cell)]  # convert to angstrom, instead of nm

        if 'lattice_type' in line:
            results['lattice_type'] = line.split()[-1]

        if 'centering' in line:
            results['centering'] = line.split()[-1]

        if 'unique_axis =' in line:
            results['unique_axis'] = line.split()[-1]

        if 'predict_refine/det_shift x' in line:
            flds = line.split()
            results['delta_x'] = flds[-5]
            results['delta_y'] = flds[-2]

        if 'predict_refine/final_residual' in line:
            results['fit_residual'] = line.split()[-1]

        if 'num_reflections' in line:
            results['num_reflections'] = int(line.split()[-1])

        if 'indexed_by' in line:
            results['indexed_by'] = line.split()[-1]

        if 'Reflections measured after indexing' in line:
            refl_start = i

        if 'End of reflections' in line:
            refl_end = i

    if peak_end and peak_start:
        results['list_of_peaks'] = get_adxv_spots(lines[peak_start:peak_end])

    if refl_start and refl_end:
        results['list_of_reflections'] = get_adxv_predictions(lines[refl_start:refl_end])

    if chunk_start and chunk_end:
        results['chunk'] = lines[chunk_start:chunk_end+1]

    return results


def handler(req):
    d = util.FieldStorage(req)
    args = {'file_name': d.get('file_name', ''), 'binning': d.get('binning', '1'), 'nimages': d.get('nimages', '1'),
            "spots": d.get('spots', '')}
    logfile = StringIO.StringIO()
    logfile.write(run_indexamajig(args))
    log = logfile.getvalue()
    req.set_content_length(len(log))
    req.write(log)
    return apache.OK


def test(args):
    return run_indexamajig(args)


if __name__ == '__main__':
    args = {'file_name': '/mnt/beegfs/qxu/test/C1_8_000270.cbf', 'spots': 1}
    print(test(args))
