from __future__ import absolute_import

import logging, os, sys
from multiprocessing import Process
import datetime

from ..common.generatesweeps import GenerateSweeps
from ..common.utils import timeit
from ..xds.xds import XDS
from ..process.GMCAproc import setup_logging
from ..tasks.mosflm import Mosflm
from ..tasks.dialsindex import DialsIndex
from ..tasks.labelit import Labelit
from ..process.get_args import ARGS
from ..common.pv import update_pipelineStatus_PV

logger = logging.getLogger(__name__)


def update_job_status(status="RUN", jobname=None):
    if ARGS.log2db:
        ARGS.pipelinestatus_table_reference.state = status
        ARGS.pipelinestatus_table_reference.elapsedtime = (datetime.datetime.now() - ARGS.starttime).total_seconds()

        if jobname:
            ARGS.pipelinestatus_table_reference.sampleName = jobname
        ARGS.pipelinestatus_table_reference.save()

        update_pipelineStatus_PV(ARGS.pipelinestatus_id, "STRATEGY", status)

@timeit
def get_sweeps():
    imgs = GenerateSweeps(ARGS.data_dir,
                          filelist=ARGS.filelist,
                          images=ARGS.images,
                          skip_sweeps_with_nimages_ge=10,
                          split_gap=False,
                          prefix=ARGS.prefix)
    xs = imgs.get_image_sweeps(sort_by_mtime=True)

    if len(xs) < 1:
        logger.error("no images or too many (>10) found for indexing/strategy calculation.")
        update_job_status("DONE")
        raise SystemExit
    xs[-1].print_summary()
    return xs


@timeit
def xds_strategy(sweep):
    xds = XDS(wdir=ARGS.proc_dir,
              runner=ARGS.runner,
              user_space_group=ARGS.space_group,
              user_unit_cell=ARGS.unit_cell,
              user_xdscmds=ARGS.xdscmds,
              user_refhkl=ARGS.xds_refhkl,
              highres=ARGS.highres,
              native=ARGS.native,
              nproc=ARGS.nproc,
              njobs=ARGS.njobs,
              sampleName=ARGS.sampleName
              ).set_xdsinp_from_image_sweep(sweep)
    xds.strategy(subdir="xds_strategy")
    logger.info("xds strategy job completed.")


@timeit
def mosflm_strategy(sweep):
    mosflm = Mosflm(sweep,
                    log2db=ARGS.log2db,
                    runner=ARGS.runner,
                    highres=ARGS.highres,
                    space_group=ARGS.space_group,
                    native=ARGS.native,
                    wdir=ARGS.proc_dir)
    mosflm.standard_strategy()
    logger.info("mosflm strategy job completed.")


@timeit
def labelit_strategy(sweep):
    labelit = Labelit(sweep,
                      log2db=ARGS.log2db,
                      runner=ARGS.runner,
                      space_group=ARGS.space_group,
                      native=ARGS.native,
                      wdir=ARGS.proc_dir)
    strategy = labelit.standard_strategy()
    logger.info("labelit strategy job completed.")


@timeit
def dials_strategy(sweep):
    dials = DialsIndex(sweep,
                       log2db=ARGS.log2db,
                       runner=ARGS.runner,
                       space_group=ARGS.space_group,
                       native=ARGS.native,
                       wdir=ARGS.proc_dir)
    strategy = dials.standard_strategy()
    logger.info("dial strategy job completed.")


def main():
    try:
        if not os.path.isdir(ARGS.data_dir):
            logger.error("ERROR: directory {} does not exist.".format(ARGS.data_dir))
            sys.exit(1)

        setup_logging()
        sweeps = get_sweeps()

        # update imageSet of pipeline status table,
        # TODO how to better handle this, since the follow will not work for hdf5
        if ARGS.log2db:
            imgset = sweeps[-1].get_imageSet(mode="screen")   # ','.join(s.get_imageSet(mode="screen") for s in sweeps)
            ARGS.imageSet = imgset

            ARGS.pipelinestatus_table_reference.imageSet = imgset
            ARGS.pipelinestatus_table_reference.state = "RUN"
            ARGS.pipelinestatus_table_reference.save()

        sweep = sweeps[-1]

        if len(sweeps) < 1:
            logger.info("nothing to process, job completed.")
            return

        if not os.path.isdir(ARGS.proc_dir):
            os.makedirs(ARGS.proc_dir)

        jobs = []

        update_job_status("RUN")

        if "mosflm_strategy" in ARGS.pipeline:
            logger.info("running mosflm strategy calculation ....")
            process = Process(target=mosflm_strategy, args=(sweep, ))
            process.start()
            jobs.append(process)

        if "dials_strategy" in ARGS.pipeline:
            logger.info("running dials/mosflm strategy calculation ....")
            process = Process(target=dials_strategy, args=(sweep, ))
            process.start()
            jobs.append(process)

        if "labelit_strategy" in ARGS.pipeline:
            logger.info("running labelit/mosflm strategy calculation ....")
            process = Process(target=labelit_strategy, args=(sweep, ))
            process.start()
            jobs.append(process)

        if "xds_strategy" in ARGS.pipeline:
            logger.info("running xds strategy calculation ....")
            sweep = get_sweeps()[-1]
            process = Process(target=xds_strategy, args=(sweep, ))
            process.start()
            jobs.append(process)

        for j in jobs:
            j.join()

        update_job_status("DONE")

        if os.path.exists(ARGS.lockfile):
            logger.info("lock file {} deleted.".format(ARGS.lockfile))
            os.remove(ARGS.lockfile)

    except SystemExit as e:
        logging.info("program exited.")
    except Exception:
        logging.error("program failed with error.", exc_info=True)
        update_job_status("FAIL")

    return


if __name__ == '__main__':
    main()
