from __future__ import absolute_import
import os
import logging

from .scriptjob import ScriptJob
from ..process.get_args import ARGS, namespace2dict

logger = logging.getLogger(__name__)


class Blend(ScriptJob):
    def __init__(self, xds_objs,
                 wdir='.',
                 subdir='.',
                 script_name='blend.sh',
                 highres=None,
                 native=False,
                 space_group=None,
                 unit_cell=None,
                 njobs=1,
                 script_text=None,
                 nproc=2,
                 runner='slurm',
                 pe_name='smp',
                 image_sweeps=None,
                 log2db=False,
                 ):
        super(Blend, self).__init__(wdir,
                                    script_name,
                                    script_text,
                                    nproc,
                                    runner,
                                    pe_name)
        self.xds_objs = xds_objs
        self.sort_xds_objs()
        self.highres = highres
        self.native = native
        self.space_group = space_group
        self.unit_cell = unit_cell
        self.njobs = njobs
        self.runner = runner
        self.image_sweeps = image_sweeps

        self.subdir = subdir

        self.wdir = os.path.join(os.path.abspath(self.wdir), self.subdir)
        if not os.path.isdir(self.wdir):
            os.makedirs(self.wdir)

        self.log2db = log2db
        self.data_dict = namespace2dict(ARGS, values_as_string=True)

    def get_most_common_spg(self):
        pass

    def filter_datasets(self):
        pass

    # sort xds objects by some useful criterion
    def sort_xds_objs(self):
        self.xds_objs = sorted(self.xds_objs, key=lambda xds: -float(xds.correct_summary.get('ISa', '0')))

    def get_datafile_list(self):

        datafilename = os.path.join(self.wdir, "blend.dat")
        with open(datafilename, "w") as f:
            for xds in self.xds_objs:
                isa = xds.correct_summary.get('ISa', '0')

                pdir = xds.get_process_dir()
                integ = os.path.join(pdir, "INTEGRATE.HKL")
                if os.path.exists(integ) and float(isa) > 5.0:
                    f.write(integ)
                    f.write("\n")
                else:
                    logger.warning("sweep {} ignore due to integ failure or low ISa value: {}".format(xds.get_imagesweep().get_prefix(), isa))

        return datafilename

    def cmd_for_blend(self, mode="analysis", arguments=""):
        # set up blend env
        cmd = list([". /mnt/software/px/CCP4/ccp4/bin/ccp4.setup-sh\n"])

        # add cmd
        mode = mode.lower()
        if mode.startswith("analysis"):
            flag = "-a"
        elif mode.startWith("synthesis"):
            flag = "-s"
        elif mode.startswith("combination"):
            flag = "-c"
        elif mode.startswith("gui"):
            flag = "-g"

        if mode.startswith("analysis"):
            arguments = self.get_datafile_list()

        cmd.append('blend {} {} <<eof\n'.format(flag, arguments))

        # boilerplate other blend keywords may be added
        if self.native:
            cmd.append("ANOMALOUS OFF\n")
        else:
            cmd.append("ANOMALOUS ON\n")

        cmd.append("eof\n")

        self.script_text = ''.join(cmd)
        return self

    def run_blend(self):
        if len(set(xo.wavelength for xo in self.xds_objs)) > 1:
            logger.info("blend does not support multiple wavelength dataset yet.")
            return

        self.cmd_for_blend(mode="analysis").run_sync()
