import json
import os
import shutil
import subprocess
import tempfile

from ..smx.cbf_header import PilatusHeader


class Dozor:
    def __init__(self,
                 filelist,
                 header=None,
                 wdir=None,
                 clean_up=False
                 ):

        self.wdir = wdir
        if not wdir or not os.path.exists(self.wdir):
            self.wdir = tempfile.mkdtemp(prefix='dozor_tmp')

        self.filelist = filelist
        self.first_image = filelist[0]
        if not os.path.isfile(self.first_image):
            print("Error: file {} does not exist.".format(self.first_image))

        if not header:
            self.hdict = PilatusHeader(self.first_image, 4096).header_dict
        else:
            self.hdict = header

        self.clean_up = clean_up

        self.prepare_images()

        self.all_results = self.run_dozor()

    def prepare_images(self):
        # NB create links since the file #'s are not continuous
        # the following only works for abc_0xxxxx.cbf type of filenames
        fn_template = "{}/image_{:06d}.cbf"
        for i, f in enumerate(self.filelist):
            os.symlink(f, fn_template.format(self.wdir, i + 1))

        self.hdict['Nimages'] = len(self.filelist)
        self.hdict['N_first'] = 1
        self.hdict['Template'] = "{}/image_??????.cbf".format(self.wdir)

    def get_dozor_in(self):
        DOZOR = ("job single\n"
                 "detector {detector}\n"
                 "pixel_max {cutoff}\n"
                 "fraction_polarization 0.99\n"
                 "detector_distance {distance}\n"
                 "X-ray_wavelength {wavelength}\n"
                 "orgx {orgx}\n"
                 "orgy {orgy}\n"
                 "spot_size 3\n"
                 "exposure {exposure}\n"
                 "oscillation_range {osc}\n"
                 "starting_angle {start_angle}\n"
                 "name_template_image {template}\n"
                 "first_image_number  {n_first}\n"
                 "number_images {nimages}\n"
                 "pixel_min -1\n"
                 "ix_min 1\n"
                 "ix_max {ix_max}\n"
                 "iy_min {iy_min}\n"
                 "iy_max {iy_max}\n")

        if 'PILATUS' in self.hdict['Detector_identifier'].upper() and \
                '6M' in self.hdict['Detector_identifier'].upper():
            detector = 'pilatus6m'
            beamstopsizepixel = 50
        elif 'EIGER 16M' in self.hdict['Detector_identifier'].upper():
            detector = 'eiger16m'
            beamstopsizepixel = 100
        else:
            print("Error: detector type {} not implemented.".format(self.hdict['Detector_identifier']))

        orgx, orgy = self.hdict['Beam_xy']
        ix_max = int(orgx + beamstopsizepixel)
        iy_min = int(orgy - beamstopsizepixel)
        iy_max = int(orgy + beamstopsizepixel)
        start_angle = self.hdict['Start_angle']
        osc = self.hdict['Angle_increment']
        wavelength = self.hdict['Wavelength']
        nimages = self.hdict['Nimages']
        cutoff = self.hdict['Count_cutoff']
        exposure = self.hdict['Exposure_time']
        distance = self.hdict['Detector_distance'] * 1000
        template = self.hdict['Template']
        n_first = self.hdict['N_first']
        return DOZOR.format(detector=detector, cutoff=cutoff,
                            distance=distance, wavelength=wavelength,
                            orgx=orgx, orgy=orgy, exposure=exposure,
                            osc=osc, start_angle=start_angle,
                            template=template, n_first=n_first,
                            nimages=nimages, ix_max=ix_max, iy_min=iy_min,
                            iy_max=iy_max)

    def run_dozor(self):
        """ run dozor directly """
        if "besthome" not in os.environ:
            os.environ["besthome"] = "/mnt/software/px/best"

        dozor_input = self.get_dozor_in()
        d_out = {}
        try:
            dozor_in = "{}/dozor.in".format(self.wdir)
            with open(dozor_in, "w") as fh:
                fh.write(dozor_input)

            cmd = ['/mnt/software/px/DOZOR/dozor', '-p', dozor_in]

            dozor_out = subprocess.check_output(cmd, stderr=subprocess.STDOUT, shell=False, cwd=self.wdir)
            d_out = Dozor.parse_dozor(dozor_out.decode('utf-8', errors='ignore'))

            if not isinstance(d_out, list) or not d_out:
                return d_out

            all_results = list()
            for i, d in enumerate(d_out):
                # only deal with 1st image for now, need to modify to deal with nimages > 1

                d['filename'] = self.filelist[i]
                d['run'], rel_filename = d['filename'].split('/')[-2:]
                d['prefix'] = rel_filename[:rel_filename.rindex('_')]

                spotfile = "{}/{:05d}.spot".format(self.wdir, int(d['img_num']))
                spots = Dozor.parse_spots(spotfile)
                d['list_of_peaks'] = spots

                d['img_num'] = int(rel_filename[:-4].split('_')[-1])  # update img_num to real one

                all_results.append(d)
        except subprocess.CalledProcessError as e:
            raise Exception('ERROR: {}'.format(e.output))
        finally:
            if self.clean_up:
                shutil.rmtree(self.wdir)

        #from pprint import pprint
        #pprint(all_results)

        return all_results

    def get_hits(self):
        return [r['filename'] for r in self.all_results if int(r.get('num_peaks', 0)) >= 10]

    def get_all_results(self):
        return self.all_results

    @staticmethod
    def parse_spots(spotfile):
        if not os.path.exists(spotfile):
            return

        with open(spotfile, 'r') as f:
            lines = f.readlines()

        if len(lines) <= 3:
            return

        out = []
        for l in lines[3:]:
            x = l.split()
            if len(x) == 5:
                # o = "{} {} {}".format(x[1], x[2], round(float(x[3]) / float(x[4]), 2))
                o = "{} {} {}".format(x[1], x[2], '0')  # adxv spot format
                out.append(o)

        return out

    @staticmethod
    def parse_dozor(outlog):
        lines = outlog.split("\n")[6:-2]
        out = []
        if len(lines) >= 1:
            start = False
            for line in lines:
                if line.startswith('-'):
                    start = not start
                    continue

                if start:
                    results = list()
                    groups = line.split("|")
                    results.append(groups[0])

                    nfields_per_group = [1, 3, 5, 3]
                    for i in range(1, 4):
                        if "no results" in groups[i]:
                            results.extend(['0'] * nfields_per_group[i])
                        else:
                            results.extend(groups[i].split())

                    keys = [
                        'img_num',
                        'num_peaks',
                        'Average Intensity',
                        'peak_resolution',
                        'Scale',
                        'B factor',
                        'Resolution Wilson',
                        'Correlation',
                        'R-factor',
                        'Main Score',
                        'Spot Score',
                        'Resolution Visible'
                    ]
                    if len(results) == 12:
                        results = [float(r) for r in results]
                        results[0] = int(results[0])
                        results[1] = int(results[1])
                        results[3] = int(results[3])
                        out.append(dict(zip(keys, results)))
        return out if out else outlog


if __name__ == '__main__':
    def get_filelist(directory, nimages=2):
        files = os.listdir(directory)
        return [os.path.join(directory, f) for f in files if f.endswith('.cbf')][:nimages]

    d = '/mnt/beegfs/qxu/23BM_2017_08_08/4z2x-p21212'
    d = '/mnt/beegfs/qxu/23BM_2017_08_08/naga-screening-23IDB_2017_04_06/F16/cbf'
    filelist = get_filelist(d, 10)
    dozor = Dozor(filelist)
    from pprint import pprint
    pprint(dozor.get_all_results())
    pprint(dozor.get_hits())
