from __future__ import absolute_import
import os
import logging
import re

from .scriptjob import ScriptJob
from ..config.config import PROGRAM_FAST_DP
from ..file_parser.aimlessparser import AimlessParser
from ..process.get_args import ARGS, namespace2dict
from ..tasks.solver import Solver

logger = logging.getLogger(__name__)


class FastDP(ScriptJob):
    def __init__(self,
                 first_image,
                 start=None,
                 end=None,
                 atom_type=None,
                 wdir='.',
                 data_dir=None,
                 subdir='fastdp',
                 script_name='fastdp.sh',
                 scaler=None,
                 lowres=None,
                 highres=None,
                 native=False,
                 space_group=None,
                 unit_cell=None,
                 njobs=1,
                 script_text=None,
                 nproc=1,
                 runner='slurm',
                 pe_name='smp',
                 log2db=False,
                 image_sweeps=None
                 ):

        super(FastDP, self).__init__(wdir,
                                     script_name,
                                     script_text,
                                     nproc,
                                     runner,
                                     pe_name)
        self.log2db = log2db
        self.first_image = first_image
        self.start = start
        self.end = end
        self.atom_type = atom_type
        self.data_dir = data_dir
        self.lowres = lowres
        self.highres = highres
        self.native = native
        self.space_group = space_group.replace(" ", "").upper() if space_group else None
        self.unit_cell = re.sub("\s+", ",", unit_cell).replace(",,", ",") if unit_cell else None  # separated by comma
        self.njobs = njobs
        self.scaler = scaler
        self.runner = runner
        self.wdir = os.path.abspath(wdir)
        self.subdir = subdir
        self.wdir = os.path.join(os.path.abspath(self.wdir), self.subdir)
        self.image_sweeps = image_sweeps
        if not os.path.isdir(self.wdir):
            os.makedirs(self.wdir)

        self.data_dict = namespace2dict(ARGS, values_as_string=True)
        self.summary_html = None
        self.truncate_mtz = "{}/{}_fast_dp_ctruncate.mtz".format(self.wdir, ARGS.sampleName)
        self.truncate_log = "{}/{}_fast_dp_ctruncate.log".format(self.wdir, ARGS.sampleName)
        self.dimple_dir = self.wdir  # os.path.join(self.wdir, "dimple")

    def cmd_for_process(self):
        cmd = ['\n{} '.format(PROGRAM_FAST_DP)]

        if self.first_image and os.path.isfile(self.first_image):
            cmd.append('{}'.format(self.first_image))
        else:
            logger.error("fastdp need first image name to run, check {} exists."
                         .format(self.first_image))
            raise Exception

        if self.start:
            cmd.append('-1 {}'.format(self.start))

        if self.end:
            cmd.append('-N {}'.format(self.end))

        if self.highres:
            cmd.append('-r {}'.format(self.highres))

        if self.lowres:
            cmd.append('-R {}'.format(self.lowres))

        if self.atom_type:
            cmd.append('-a {}'.format(self.atom_type))

        if self.space_group:
            cmd.append('-s "{}"'.format(self.space_group))

        if self.njobs > 1:
            cmd.append('-j {}'.format(self.njobs))

        if self.nproc > 1:
            cmd.append('-k {}'.format(self.nproc))

        # scala is not an option in github version
        if '_qxu' in os.path.realpath(PROGRAM_FAST_DP):
            cmd.append('-S {}'.format('aimless'))

        if self.unit_cell:
            cmd.append('-c "{}"'.format(self.unit_cell))

        self.script_text = ' '.join(cmd)
        return self

    def report(self):
        aimless_log = os.path.join(self.wdir, "aimless.log")
        if os.path.isfile(aimless_log):
            # gather aimless statistics
            ap = AimlessParser(wdir=self.wdir, filename='aimless.log')
            self.data_dict.update(ap.summarize())

            self.data_dict['imagedir'], self.data_dict['firstFrame'] = os.path.split(self.first_image)
            self.data_dict['pipeline'] = 'fastdp'
            self.data_dict['state'] = 'DONE'
            self.data_dict['software'] = 'fastdp'
            self.data_dict['scalingsoftware'] = 'aimless'
            self.data_dict['workdir'] = self.wdir
            self.data_dict['imageSet'] = self.data_dict.get('imageSet')
            self.data_dict['scale_dir'] = self.wdir
            self.data_dict['truncate_mtz'] = self.truncate_mtz
            self.data_dict['truncate_log'] = self.truncate_log
            self.data_dict['scale_mtz'] = "{}/{}".format(self.wdir, 'fast_dp.mtz')
            self.data_dict['scale_log'] = '{}/{}'.format(self.wdir, 'aimless.log')
            self.data_dict['logfile'] = "{}/fast_dp.log".format(self.wdir)
            self.data_dict['imageSet'] = ARGS.imageSet

            # if no sampleName was set, use prefix
            sample_name = self.data_dict.get('sampleName')
            if sample_name == 'sampleName' or not sample_name:
                self.data_dict['sampleName'] = self.data_dict['prefix']

            if self.summary_html:
                self.data_dict['report_url'] = '{}'.format(self.summary_html)

            if self.log2db:
                from ..log.database import save_to_db

                dp = save_to_db(self.data_dict)
                self.data_dict['dataprocessresults_id'] = dp.id

            # make some plots, note that matplotlib is not threadsafe, so don't use it in a thread
            # ap.plot()
        else:
            logger.info("{} does not exist, fastdp may have failed.")

    def run_process(self):
        self.cmd_for_process().run_sync()

        from ..tasks.xdsstat import XDSstat

        tag_name = "{}/{}/{}".format(ARGS.sampleName, "fastdp", self.image_sweeps.get_prefix())
        stat = XDSstat(wdir=self.wdir, tag_name=tag_name)
        stat.run()
        self.summary_html = stat.get_summary_html()

        self.report()

        from ..tasks.truncate import Truncate
        Truncate(wdir=self.wdir,
                 hklin='fast_dp.mtz',
                 hklout=self.truncate_mtz,
                 logfile=self.truncate_log).run_ctruncate(async_flag=True)

        Solver(data_info=self.data_dict).solve()
