from __future__ import absolute_import

import glob
import logging
import os
import threading
import time
from datetime import datetime

from watchdog.events import PatternMatchingEventHandler, DirCreatedEvent
from watchdog.observers import Observer

from .scriptjob import ScriptJob
from ..common.utils import get_directories_contains_file, get_html_table
from ..config.config import KAMO_SETUP, PX
from ..file_parser import fileparser
from ..process.get_args import ARGS, namespace2dict

logger = logging.getLogger(__name__)

MERGE_SCRIPT_TEMPLATE = \
"""
#!/bin/sh
# settings

export PX={PX}

{kamo_setup}

dmin={dmin}  # resolution
anomalous={anomalous} # true or false
lstin={merge_list}  # list of XDS_ASCII.HKL files, e.g. formerge_reindexed.lst
min_cmpl={min_cmpl}
min_redun={min_redun}
workdir={workdir}
use_ramdisk=true # set false if there is few memory or few space in /tmp

# _______/setting


kamo.multi_merge \
        workdir=${{workdir}} \\
        lstin=${{lstin}} d_min=${{dmin}} anomalous=${{anomalous}} \\
        space_group=None reference.data=None \\
        program=xscale xscale.reference=bmin xscale.degrees_per_batch=None \\
        reject_method=framecc+lpstats rejection.lpstats.stats=em.b+bfactor \\
        clustering=blend blend.min_cmpl=${{min_cmpl}} blend.min_redun=${{min_redun}} blend.max_LCV=None blend.max_aLCV=None \\
        max_clusters=None xscale.use_tmpdir_if_available=${{use_ramdisk}} \\
        nproc={nproc} \\
#       batch.engine=sge batch.par_run=merging batch.nproc_each=8 nproc=8 batch.sge_pe_name=smp
"""


def update_db_status(status):
    if ARGS.log2db:
        ARGS.pipelinestatus_table_reference.state = status
        ARGS.pipelinestatus_table_reference.save()

        from ..common.pv import  update_pipelineStatus_PV
        update_pipelineStatus_PV(ARGS.pipelinestatus_table_reference.id, "PROCESS", status)


class MergeObserver(Observer):
    def dispatch_events(self, event_queue, timeout):
        event, watch = event_queue.get(block=True, timeout=timeout)
        '''
        if not ("merge_" in event.src_path and type(event) in [DirCreatedEvent]):
            event_queue.task_done()
            return
        '''

        if type(event) in [DirCreatedEvent]:
            sleepedfor = 0
            time.sleep(2)

            while True:
                f1 = os.path.join(event.src_path, "formerge.lst")
                f2 = os.path.join(event.src_path, "merge_blend.sh")
                f3 = os.path.join(event.src_path, "input_files")
                if any([os.path.isdir(f3), os.path.exists(f1), os.path.exists(f2)]):
                    break

                if sleepedfor >= 5:
                    logger.info("needed merge files not present, skipped @ {}.".format(event.src_path))
                    event_queue.task_done()
                    return
                time.sleep(1)
                sleepedfor += 1
        else:
            event_queue.task_done()
            return

        try:
            for handler in list(self._handlers.get(watch, [])):
                if handler in self._handlers.get(watch, []):
                    handler.dispatch(event)
            # self._dispatch_event(event, watch)
        except KeyError:
            pass
        event_queue.task_done()


def do_merging(merge_dir, kamo_obj):
    update_db_status("MERGE")
    logger.info("merging job started: {}".format(merge_dir))
    sleepedfor = 0
    time.sleep(1)
    while True:
        f1 = os.path.join(merge_dir, "formerge.lst")
        f2 = os.path.join(merge_dir, "merge_blend.sh")
        if os.path.exists(f1) and os.path.exists(f2):
            break
        if sleepedfor >= 300:
            logger.info("needed files not present, so abort the merging process @ {}.".format(merge_dir))
            return
        time.sleep(1)
        sleepedfor += 1

    kamo_obj.reindex(merge_dir)
    kamo_obj.merge(merge_dir)
    logger.info("merging {} finished, saving results.".format(merge_dir))
    kamo_obj.save_results(merge_dir)
    logger.info("saving {} to db finished.".format(merge_dir))
    update_db_status("MERGED")


class MergeHandler(PatternMatchingEventHandler):
    def __init__(self, kamo):
        PatternMatchingEventHandler.__init__(self, patterns=["*"],
                                             ignore_directories=False,
                                             case_sensitive=True)
        self.kamo = kamo

    def on_any_event(self, event):
        merge_dir = event.src_path
        logger.info("new merging directory created @ {} ...".format(merge_dir))

        background = True
        if not background:
            logger.info("started merging job in foreground @ {} ...".format(merge_dir))
            do_merging(merge_dir, self.kamo)
        else:
            logger.info("started merging job in background @ {} ...".format(merge_dir))
            t = threading.Thread(target=do_merging, args=(merge_dir, self.kamo,))
            t.start()


class Kamo(ScriptJob):
    def __init__(self,
                 wdir='.',
                 data_dir=None,
                 subdir='kamo_dials',
                 script_name='kamo_dials.sh',
                 highres=None,
                 native=True,
                 atom_type=None,
                 space_group=None,
                 unit_cell=None,
                 pipeline='kamo_dials',
                 script_text=None,
                 small_wedges=True,
                 runner='slurm',
                 pe_name='smp',
                 njobs=1,
                 nproc=2,
                 reference_file=None,
                 image_sweeps=None,
                 log2db=False,
                 anomalous=False,
                 exclude_dir=['process','raster', 'screen','scan'],
                 ):
        super(Kamo, self).__init__(wdir,
                                   script_name,
                                   script_text,
                                   nproc,
                                   runner,
                                   pe_name)
        self.highres = highres
        self.native = native
        self.atom_type = atom_type
        self.space_group = space_group
        self.unit_cell = unit_cell
        self.njobs = njobs
        self.runner = runner
        self.pipeline = pipeline
        self.image_sweeps = image_sweeps
        self.small_wedges = small_wedges
        self.subdir = subdir
        self.script_name = "{}.sh".format(self.pipeline)

        self.process_root_dir = self.wdir
        self.wdir = os.path.join(os.path.abspath(self.wdir), self.subdir)
        if not os.path.isdir(self.wdir):
            os.makedirs(self.wdir)

        self.data_dir = data_dir
        self.exclude_dir = exclude_dir
        self.log2db = log2db
        self.data_dict = namespace2dict(ARGS, values_as_string=True)
        self.reference_file = reference_file

        self.merge_subdir = "blend_framecc_b+B"
        self.anomalous = anomalous

        # use watchdog module to monitor merging requests
        self.m_observer = MergeObserver()
        self.m_observer.schedule(MergeHandler(self),
                                 path=self.wdir,
                                 recursive=False)
        self.m_observer.start()
        self.stop_observer = False

        self.ndatasets = 2
        datasets = os.path.join(self.data_dir, "datasets.txt")
        if os.path.exists(datasets):
            self.datasets = datasets
        else:
            self.datasets = None

        self.kamo_report_html = os.path.join(self.wdir, "report.html")

    def estimate_best_resolution(self, default=3.0, pad=0.1):
        if os.path.exists(self.kamo_report_html):
            with open(self.kamo_report_html) as fh:
                html = fh.read()
                col = get_html_table(html, column_name='Resn (Å)')
                if col:
                    res = [float(x) for x in col if x.replace(".", "", 1).isdigit()]
                    if res:
                        return round(min(res) - pad, 1)

        return default

    def estimate_mean_completeness(self, default=3):
        # mean completeness per dataset
        if os.path.exists(self.kamo_report_html):
            with open(self.kamo_report_html) as fh:
                html = fh.read()
                col = get_html_table(html, column_name='Cmpl (%)')
                if col:
                    cmpls = [float(x) for x in col if x.replace(".", "", 1).isdigit()]
                    if cmpls:
                        return round(sum(cmpls)/len(cmpls), 2)

        return default

    def cmd_for_kamo(self):
        # set up dial env
        cmd = list(["unset HDF5_PLUGIN_PATH\n", "export PX={PX}\n".format(PX=PX), "{}\n".format(KAMO_SETUP)])

        cmd.append("cd {}\n".format(self.data_dir))  # kamo must be start from top level data_dir
        # add cmd
        cmd.append('kamo bl=other logwatch_interval=10 use_tmpdir_if_available=true')

        if not self.small_wedges:
            logger.info("small_wedges mode off")
            cmd.append("small_wedges=false")
        else:
            logger.info("small_wedges mode on")
            cmd.append("small_wedges=true")

        if self.datasets:
            cmd.append("dataset_paths_txt={}".format(self.datasets))

        if self.data_dir:
            cmd.append("topdir={}".format(self.data_dir))

        cmd.append("workdir={}".format(self.wdir))

        if self.pipeline == 'kamo_dials':
            cmd.append('engine=dials')
        else:
            cmd.append('engine=xds')

        if self.anomalous:
            cmd.append("anomalous=true")
        else:
            cmd.append("anomalous=false")

        if self.space_group and self.unit_cell:
            cmd.append("known.space_group={}".format(self.space_group))
            cmd.append('known.unit_cell="{}"'.format(self.unit_cell))

        if self.runner:
            cmd.append("batch.engine={} batch.nproc_each={}".format("slurm", self.nproc))

        if self.pe_name and 'sge' in self.runner:
            cmd.append('batch.sge_pe_name={}'.format(self.pe_name))

        # skipping subdirectories, this requires a modification of KAMO itself
        if self.exclude_dir:
            for d in self.exclude_dir:
                cmd.append("exclude_dir={}".format(d))

        self.script_text = ' '.join(cmd)
        return self

    def get_merge_dir(self):
        merge_dirs = get_directories_contains_file("merge_.*.sh", root_directory=self.wdir, ftype='filename')
        if merge_dirs:
            merge_dir = merge_dirs[-1]
        else:
            merge_dir = None

        return merge_dir

    def reindex(self, merge_dir):
        logger.info("starting reindex in {}".format(merge_dir))

        if not merge_dir:
            logger.warning("no merging directory was found.")
            return

        # resolve index ambiguity
        reindex_cmds = list()
        reindex_cmds.append(KAMO_SETUP)
        if self.highres:
            dmin = "dmin={}".format(self.highres)
        else:
            dmin = ""

        if self.reference_file:
            reference = "method=reference reference_file={}".format(self.reference_file)
        else:
            reference = ""

        reindex_cmds.append(
            "kamo.resolve_indexing_ambiguity formerge.lst skip_bad_files=true {} {}".format(dmin, reference))

        ScriptJob(wdir=merge_dir,
                  script_text="\n".join(reindex_cmds),
                  script_name="reindex.sh",
                  runner='local').run_sync()

    def merge(self, merge_dir):
        logger.info("starting merging in {}".format(merge_dir))

        if not merge_dir:
            logger.info("merge failed, no merge directory")
            return

        if os.path.exists(os.path.join(merge_dir, "formerge_reindexed.lst")):
            merge_list = "formerge_reindexed.lst"
        elif os.path.exists(os.path.join(merge_dir, "formerge.lst")):
            merge_list = "formerge.lst"
        else:
            logger.warning("no file list present for merging @ {}".format(merge_dir))
            return

        with open(os.path.join(merge_dir, "formerge.lst")) as ml:
            self.ndatasets = len(ml.readlines())
            logger.info("number of datasets used: {}".format(self.ndatasets))

        if self.highres:
            dmin = self.highres
            logger.info("user  has defined resolution cutoff: {}".format(dmin))
        else:
            dmin = self.estimate_best_resolution()
            logger.info("estimated resolution from report.html: {}".format(dmin))

        self.merge_subdir = "blend_{dmin}a_framecc_b+B".format(dmin=dmin)

        try:
            mean_cmpl = self.estimate_mean_completeness()
            min_cmpl = mean_cmpl * self.ndatasets / 2.0
        except:
            if self.small_wedges:
                min_cmpl = 5 * self.ndatasets * 0.8
            else:
                min_cmpl = 10 * self.ndatasets * 0.8

        min_cmpl = min(round(min_cmpl, 0), 90)
        min_redun = 2 if min_cmpl > 60 else 1

        if self.anomalous:
            anomalous_v = "true"
        else:
            anomalous_v = "false"

        mdir = os.path.join(merge_dir, self.merge_subdir)
        if os.path.exists(mdir):
            time_stamp = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
            new_dir = "{}.backup-{}".format(mdir, time_stamp)
            os.rename(mdir, new_dir)
            logger.info("directory {} already exists, renamed to {}".format(mdir, new_dir))

        merge_script = MERGE_SCRIPT_TEMPLATE.format(kamo_setup=KAMO_SETUP,
                                                    PX=PX,
                                                    dmin=dmin,
                                                    merge_list=merge_list,
                                                    anomalous=anomalous_v,
                                                    min_cmpl=min_cmpl,
                                                    min_redun=min_redun,
                                                    nproc=16,
                                                    workdir=self.merge_subdir)

        ScriptJob(wdir=merge_dir,
                  script_name="merge_{}a.sh".format(dmin),
                  script_text=merge_script,
                  nproc=16,
                  pe_name='smp',
                  runner='slurm').run_sync()

    def save_results(self, merge_dir):
        mdir = os.path.join(merge_dir, self.merge_subdir)
        cluster_dirs = glob.glob("{}/{}".format(mdir, "cluster*/"))
        if not cluster_dirs:
            logger.error("cluster scaling failed. ")

        table1 = ""
        scalelog = os.path.join(mdir, "cluster_summary.dat")
        if os.path.exists(scalelog):
            with open(scalelog) as fh:
                table1 = fh.read()

        if not table1 or len(table1.splitlines()) <= 3:
            logger.info("merge failed, clustering stage skipped, very few datasets or low completeness.")
            merge_log = os.path.join(mdir, "multi_merge.log")
            scalelog = merge_log
            if os.path.exists(merge_log):
                fp = fileparser.FileParser(filename=merge_log)
                table1 = "".join(fp.get_blocks_between(text1='Space group for merging:', text2='Normal exit')[0])

        logger.info("DEBUG: table1 {} = ".format(table1))
        if self.log2db:
            from ..log.database import save_to_db

            self.data_dict['imagedir'], self.data_dict['firstFrame'] = self.data_dir, ""
            self.data_dict['pipeline'] = self.pipeline
            self.data_dict['state'] = 'DONE'
            self.data_dict['software'] = 'kamo'
            self.data_dict['scalingsoftware'] = 'XSCALE'
            self.data_dict['workdir'] = mdir
            self.data_dict['run_stats'] = "{}/report.html".format(self.wdir)

            self.data_dict['truncate_mtz'] = 'NA'
            self.data_dict['truncate_log'] = 'NA'

            self.data_dict['scale_log'] = "{}".format(scalelog)
            self.data_dict['table1'] = table1
            self.data_dict['report_url'] = "{}/report.html".format(mdir)
            self.data_dict['logfile'] = ''
            self.data_dict['imageSet'] = ARGS.imageSet

            # if no sampleName was set, use prefix
            sample_name = ARGS.sampleName
            if sample_name == 'sampleName' or not sample_name:
                self.data_dict['sampleName'] = self.data_dict['prefix']

            save_to_db(self.data_dict)

    def run_kamo(self):
        self.cmd_for_kamo()
        if not os.path.exists(os.path.join(self.wdir, self.script_name)):
            use_existing_script = False
            # self.cmd_for_kamo()
        else:
            logger.info("use existing script to start kamo.")
            use_existing_script = True

        self.runner = 'local'  # run kamo locally since it needs X-windows
        self.run_sync(timeout=10800, write_script_to_file=not use_existing_script)

        # check if all tasks are unfinished, if there are none, stop; otherwise wait
        while True:

            if not self.m_observer.event_queue.empty():
                logger.debug("=>=>=>=> empty={} unfinished_tasks={}".format(
                    self.m_observer.event_queue.empty(),
                    self.m_observer.event_queue.unfinished_tasks,
                ))

                logger.info("waiting m_observer to finish")
                time.sleep(10)
            else:
                self.m_observer.stop()
                break

        self.m_observer.join()
        logger.info("kamo job finished.")
