from __future__ import absolute_import
import logging

from .scriptjob import ScriptJob
from ..file_parser.matthewscoefparser import MatthewsCoefParser

logger = logging.getLogger(__name__)


class MatthewCoef(ScriptJob):
    def __init__(self,
                 spacegroup,
                 unitcell,
                 wdir='.',
                 script_name='matthews_coef.sh',
                 script_text=None,
                 nproc=1,
                 molsize=None,
                 nmol=None,
                 molw=None,
                 moltype=None,
                 highres=None,
                 logfile='matthews_coef.log',
                 runner='local',
                 pe_name='smp'):

        super(MatthewCoef, self).__init__(wdir,
                                          script_name,
                                          script_text,
                                          nproc,
                                          runner,
                                          pe_name)

        if not molsize:
            molsize = 50

        self.spacegroup = spacegroup
        self.unitcell = unitcell
        self.molsize = molsize
        self.molw = molw
        self.moltype = moltype
        self.nmol = nmol
        self.highres = highres
        self.table = None
        self.estimated_asu = None
        self.logfile = logfile

    def cmd_run_matthewscoef(self):
        cmd = 'matthews_coef <<eof >{}\n'.format(self.logfile)
        cmd += 'symmetry {}\n'.format(self.spacegroup)
        cmd += 'cell {}\n'.format(self.unitcell)

        if self.nmol:
            cmd += 'nmol {}\n'.format(self.nmol)
        else:
            cmd += 'auto\n'

        if self.molw:
            cmd += 'molw {}\n'.format(self.molw)
        elif self.molsize:
            cmd += 'nres {}\n'.format(self.molsize)

        # default to protein, D-DNA, C-Protein/DNA complex
        if self.moltype in ['D', 'C']:
            cmd += 'mode {}\n'.format(self.moltype)

        if self.highres:
            cmd += 'reso {}\n'.format(self.highres)

        cmd += "eof\n"

        self.script_text = cmd
        return self

    def run_matthewscoef(self):
        return self.cmd_run_matthewscoef().run_sync(timeout=1800)

    def run(self):
        self.run_matthewscoef()
        mp = MatthewsCoefParser(self.wdir, filename='matthews_coef.log')
        self.table = mp.get_table()
        self.estimated_asu = mp.get_estimated_asu_content()

    def get_table(self):
        if not self.table:
            self.run()
        return self.table

    def get_asu_content(self):
        if not self.table:
            self.run()

        return self.estimated_asu


if __name__ == '__main__':
    mc = MatthewCoef('p321','65.57 65.57 65.63 90.00 90.00 120.00')
    print(mc.get_table())
    print(mc.get_asu_content())