from __future__ import absolute_import
import os
import logging

from .scriptjob import ScriptJob
from ..config.config import PHENIX_SETUP
from ..process.get_args import ARGS

logger = logging.getLogger(__name__)


template = """
hklin = {hklin}
labin = {labin} 
resolution_cutoff = 2.5
no_tncs = False
crystal_symmetry {{
  unit_cell = None
  space_group = {space_group}
}}
mode = *quick full
symmetry_exploration = {symmetry_exploration}
output {{
  root = {sample_name}
  max_solutions_to_write = 1
  gui_base_dir = {workdir}
  save_aniso_data = False
  job_title = None
}}
composition {{
  count = None
  component {{
    sequence = {seq_file}
    mtype = *protein dna rna
    stoichiometry = 1
    ensemble {{
      coordinates {{
        pdb = None
        identity = None
        error_translation = *default floored_chothia_and_lesk rmsd
      }}
      alignment = None
    }}
    model_collection {{
      trim = False
      coordinates {{
        pdb = None
        identity = None
        error_translation = *default floored_chothia_and_lesk rmsd
      }}
    }}
    template {{
      pdb = None
      alignment = None
    }}
    template_list = None
    homology {{
      file_name = None
      max_hits = 3
    }}
    search {{
      services = local *ncbi
      max_hits = 3
    }}
  }}
}}
assembly {{
  use_assembled = False
  local_search {{
    sweep = 5
    extent = 5
  }}
  component {{
    file_name = None
    operation {{
      euler = 0 0 0
      displacement = 0 0 0
    }}
  }}
}}

queue {{
  technology = lsf threading *sge pbs
  cpus = {nproc}
  submission_command = None
  qslot_cpus = {nmachines}
}}
packing_pool = 20
rotation_peaks_cutoff = 0.75
post_refinement_cutoff = 0.75
final_selection_cutoff = 0.75
b_factor_refinement = True
significant_peak_threshold = 7.0
sculptor_protocols = 11 10 13 *12 1 3 2 5 4 7 6 9 8 all minimal
template_equivalence = True
assembly_acceptance_policy = observed *always never
assembly_ensemble_creation_policy = *observed always never
pdb_mirror = *rcsb pdbe pdbj
exclude_pdb_ids = None
simple_run {{
  enable = False
  component_sequence = None
  template_model = None
  alignment = None
  blast_services = local ncbi
  max_blast_hits = 3
}}
"""


class MRage(ScriptJob):
    def __init__(self,
                 wdir='.',
                 script_name='mrage.sh',
                 script_text=None,
                 sample_name='sampleName',
                 space_group=None,
                 nproc=8,
                 nmachines=4,
                 runner='local',
                 pe_name='smp',
                 hklin='aimless.mtz',
                 labin=None,
                 seq_file=None,
                 logfile='mrage.log'):

        super(MRage, self).__init__(wdir,
                                    script_name,
                                    script_text,
                                    nproc,
                                    runner,
                                    pe_name)
        self.hklin = hklin
        self.labin = labin
        self.wdir = wdir
        self.nproc = nproc
        self.nmachines = nmachines
        self.sample_name = sample_name
        self.space_group = space_group
        self.logfile = logfile
        self.seq_file = seq_file

        if not os.path.exists(self.wdir):
            os.makedirs(self.wdir)

    def cmd_run_mrage(self):
        # only deal with enantiomorph, not pointgroup since it would too expensive computationally
        symmetry_exploration = "pointgroup enantiomorph *dataset" if ARGS.space_group else "pointgroup *enantiomorph dataset"
        eff = template.format(
            workdir=self.wdir,
            hklin=self.hklin,
            seq_file=self.seq_file,
            sample_name=self.sample_name,
            labin=self.labin if self.labin else "I(+),SIGI(+),I(-),SIGI(-),merged",
            nproc=self.nproc,
            nmachines=self.nmachines,
            space_group=self.space_group if self.space_group else "None",
            symmetry_exploration=symmetry_exploration,
        )
        eff_file = os.path.join(self.wdir, "mrage.eff")
        with open(eff_file, "w") as fh:
            fh.write(eff)

        cmd = "{}\n".format(PHENIX_SETUP)
        if os.path.isfile(self.hklin):
            cmd += "phenix.mrage mrage.eff\n"
            self.script_text = cmd
            return self
        else:
            logger.error("{} does not exist".format(self.hklin))

    def run(self, async_flag=False):
        logger.info("run phenix.mrage pipeline for structure determination.")
        if async_flag:
            self.cmd_run_mrage().run_async()
        else:
            return self.cmd_run_mrage().run_sync(timeout=7200)
