from __future__ import absolute_import

import logging
import os
import datetime

from .scriptjob import ScriptJob
from ..smx.smx_get_args import SMX_ARGS, namespace2dict

logger = logging.getLogger(__name__)


class SmxMerger(ScriptJob):
    def __init__(self,
                 input_stream,
                 real_symmetry,
                 cell_file,
                 apparent_symmetry=None,
                 refine_model="--model=unity --iterations=1",
                 wdir='.',
                 subdir='merge',
                 script_name='smx_merge.sh',
                 lowres=None,
                 highres=None,
                 native=False,
                 ambiguity_operator=None,
                 njobs=1,
                 script_text=None,
                 nproc=48,
                 runner='slurm',
                 pe_name='smp',
                 log2db=False,
                 ):

        super(SmxMerger, self).__init__(wdir,
                                        script_name,
                                        script_text,
                                        nproc,
                                        runner,
                                        pe_name)
        self.log2db = log2db
        self.input_stream = input_stream
        self.cell_file = cell_file
        self.real_symmetry = real_symmetry
        self.refine_model = refine_model
        self.lowres = lowres
        self.highres = highres
        self.native = native

        prefix = "{}_{}".format(self.real_symmetry, self.highres if self.highres else "norescut")
        self.prefix = prefix.replace("/", "_")

        self.output_hkl = "{}_output.hkl".format(self.prefix)
        self.halfset_hkl1 = "{}1".format(self.output_hkl)
        self.halfset_hkl2 = "{}2".format(self.output_hkl)

        self.apparent_symmetry = apparent_symmetry
        self.ambiguity_operator = ambiguity_operator
        self.ambiguitor_output_stream = self.input_stream
        self.njobs = njobs
        self.runner = runner
        self.wdir = os.path.abspath(wdir)
        self.subdir = subdir
        self.wdir = os.path.join(os.path.abspath(self.wdir), self.subdir)

        if os.path.isdir(self.wdir):
            timestamp = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
            os.rename(self.wdir, "{}_{}".format(self.wdir, timestamp))

        os.makedirs(self.wdir)

        self.data_dict = namespace2dict(SMX_ARGS, values_as_string=True)

    def cmd_for_merge(self):
        cmd = ['{}'.format('source /mnt/software/px/serial-x/serialx-setup.sh'),
               self.cmd_for_partialator(),
               self.cmd_for_check_hkl(),
               self.cmd_for_compare_hkl(fom='rsplit'),
               self.cmd_for_compare_hkl(fom='ccstar')]

        self.script_text = '\n'.join(cmd)
        return self

    def cmd_for_partialator(self):
        cmd = ['partialator -i {} -o {} -y {}  -j {}'.format(self.ambiguitor_output_stream,
                                                             self.output_hkl,
                                                             self.real_symmetry,
                                                             self.nproc)]

        if self.real_symmetry:
            cmd.append("-y {}".format(self.real_symmetry))

        if self.apparent_symmetry:
            cmd.append("-w {}".format(self.apparent_symmetry))

        if self.ambiguity_operator:
            cmd.append("--operator={}".format(self.ambiguity_operator))

        cmd.append("{}".format(self.refine_model))

        return " ".join(cmd)

    def cmd_for_stats_only(self):
        cmd = ['{}'.format('source /mnt/software/px/serial-x/serialx-setup.sh'),
               self.cmd_for_check_hkl(),
               self.cmd_for_compare_hkl(fom='rsplit'),
               self.cmd_for_compare_hkl(fom='ccstar')]

        self.script_text = '\n'.join(cmd)

    def cmd_for_check_hkl(self):
        highres = ""

        if self.highres:
            highres = "--highres={}".format(self.highres)

        check_hkl = "check_hkl {} -y {} {} -p {}".format(self.output_hkl,
                                                         self.real_symmetry,
                                                         highres,
                                                         self.cell_file
                                                         )
        cmd = ['printf "\n\n"',
               'echo === output for check_hkl ===',
               'echo cmd= {}'.format(check_hkl),
               check_hkl,
               "cat shells.dat",
               "mv shells.dat {}_{}.dat".format(self.prefix, "check_hkl_shells")]

        return "\n".join(cmd)

    def cmd_for_compare_hkl(self, fom='rsplit'):
        highres = ""
        if self.highres:
            highres = "--highres={}".format(self.highres)

        compare_hkl = "compare_hkl {} {} -y {} --fom={} {} -p {}".format(self.halfset_hkl1,
                                                                         self.halfset_hkl2,
                                                                         self.real_symmetry,
                                                                         fom,
                                                                         highres,
                                                                         self.cell_file
                                                                         )
        cmd = ['printf "\n\n"',
               'echo === calculate statististics for {} ==='.format(fom),
               'echo cmd= {}'.format(compare_hkl),
               compare_hkl,
               "cat shells.dat",
               "mv shells.dat {}_{}_shells.dat".format(self.prefix, fom)]

        return "\n".join(cmd)

    def run(self):
        if not self.input_stream or not os.path.exists(self.input_stream):
            print("input stream was not set.")
            return

        if not self.real_symmetry:
            print("symmetry was not set.")
            return

        if not self.cell_file or not os.path.exists(self.cell_file):
            print("cell_file was not set.")
            return

        self.cmd_for_merge().run_sync()
