from __future__ import absolute_import
import os
import logging
import glob

from .scriptjob import ScriptJob
from ..file_parser.aimlessparser import AimlessParser
from ..file_parser.xia2parser import Xia2Parser
from ..process.get_args import ARGS, namespace2dict
from ..config.config import DIALS_SETUP

logger = logging.getLogger(__name__)


class Xia2(ScriptJob):
    def __init__(self,
                 sweep_def,
                 wdir='.',
                 data_dir=None,
                 subdir='xia2_dials',
                 script_name='xia2_dials.sh',
                 highres=None,
                 native=False,
                 atom_type=None,
                 space_group=None,
                 unit_cell=None,
                 njobs=1,
                 fastmode=False,
                 pipeline='xia2_dials',
                 script_text=None,
                 nproc=2,
                 runner='slurm',
                 pe_name='smp',
                 image_sweeps=None,
                 log2db=False):
        super(Xia2, self).__init__(wdir,
                                   script_name,
                                   script_text,
                                   nproc,
                                   runner,
                                   pe_name)
        self.sweep_def = sweep_def
        self.highres = highres
        self.native = native
        self.atom_type = atom_type
        self.space_group = space_group
        self.unit_cell = unit_cell
        self.njobs = njobs
        self.runner = runner
        self.pipeline = pipeline
        self.image_sweeps = image_sweeps
        self.fast = fastmode

        self.subdir = subdir
        self.script_name = "{}.sh".format(self.pipeline)

        self.wdir = os.path.join(os.path.abspath(self.wdir), self.subdir)
        if not os.path.isdir(self.wdir):
            os.makedirs(self.wdir)

        self.data_dir = data_dir
        self.log2db = log2db
        self.data_dict = namespace2dict(ARGS, values_as_string=True)

        self.project = self.pipeline
        self.crystal = self.get_crystal_name()

    def cmd_for_xia2(self):
        # set up dial env
        cmd = list(["unset HDF5_PLUGIN_PATH\n",
                    ". /mnt/software/px/CCP4/ccp4/bin/ccp4.setup-sh\n",
                    "#source /mnt/software/px/dials/dials_env.sh\n",
                    "# "+DIALS_SETUP+"\n"])
        # add cmd
        cmd.append('xia2 failover=True read_all_image_headers=False trust_beam_centre=True')

        cmd.append('{}'.format(self.sweep_def))
        # if self.runner == 'sge':
        #    cmd.append('multiprocessing.type=qsub multiprocessing.qsub_command="qsub -q all.q"')
        #    if self.njobs > 1:
        #        cmd.append('multiprocessing.njobs={}'.format(self.njobs))

        if self.nproc > 1:
            cmd.append('multiprocessing.nproc={}'.format(self.nproc))
            cmd.append('multiprocessing.mode=parallel')
            nsweeps = self.sweep_def.count("image=")
            cmd.append('multiprocessing.njob={}'.format(min(self.njobs, nsweeps)))

            if self.runner == 'sge' and nsweeps > 1 and self.njobs > 1:
                cmd.append('multiprocessing.type=qsub')
                cmd.append("multiprocessing.qsub_command='qsub'")
            '''
            # xia2 does not support slurm
            elif self.runner == 'slurm' and nsweeps > 1 and self.njobs > 1:
                cmd.append('multiprocessing.type=sbatch')
                cmd.append("multiprocessing.qsub_command='sbatch'")
            '''

        if self.space_group:
            cmd.append("xia2.settings.space_group='{}'".format(self.space_group))
            if self.unit_cell:
                cmd.append('xia2.settings.unit_cell="{}"'.format(self.unit_cell))

        if self.highres:
            #  cmd.append('resolution={}'.format(self.highres))
            #  cmd.append('d_min={}'.format(self.highres))
            cmd.append('xia2.settings.resolution.d_min={}'.format(self.highres))

        if self.atom_type:
            cmd.append('atom={}'.format(self.atom_type))

        if not self.native:
            cmd.append('xia2.settings.input.anomalous=True')

        # use aimless for scaling
        if self.pipeline == 'xia2_dials_aimless':
            cmd.append('pipeline=dials-aimless')
        elif self.pipeline == "xia2_dials":
            cmd.append('pipeline=dials')
        elif self.pipeline == 'xia2_xds':
            cmd.append('pipeline=3d')
        elif self.pipeline == 'xia2_xds-all-spots':
            cmd.append('pipeline=3dii')

        if self.fast:
            cmd.append('dials.fast_mode=True')  # according to Graeme, don't use this

        cmd.append('project={}'.format(self.pipeline))
        cmd.append('crystal={}'.format(self.crystal))

        self.script_text = ' '.join(cmd)
        return self

    def report(self):
        xia2_results = Xia2Parser(wdir=self.wdir, filename="xia2.txt").summarize()
        # parse xia2.txt for summary
        if xia2_results:
            self.data_dict.update(xia2_results)

            self.data_dict['imagedir'] = ARGS.data_dir
            self.data_dict['firstFrame'] = ':'.join(s.get_first_file() for s in self.image_sweeps)
            self.data_dict['pipeline'] = self.pipeline
            self.data_dict['state'] = 'DONE'
            self.data_dict['software'] = self.pipeline
            self.data_dict['logfile'] = "{}/xia2.txt".format(self.wdir)
            self.data_dict['imageSet'] = ARGS.imageSet
            self.data_dict['workdir'] = self.wdir
            
            if os.path.exists("{}/xia2.html".format(self.wdir)):
                self.data_dict['report_url'] = "{}/xia2.html".format(self.wdir)
            else:
                self.data_dict['report_url'] = "{}/xia2-summary.dat".format(self.wdir)

            # self.data_dict['imageSet'] = self.runtime_info.get('imageSet')
            self.data_dict['scalingsoftware'] = 'dials'

            scale_dir = os.path.join(self.wdir, self.crystal, "scale")
            self.data_dict['scale_log'] = scale_dir
            self.data_dict['truncate_log'] = scale_dir
            self.data_dict['truncate_mtz'] = '{}/{}'.format(self.wdir, '{}/scale/{}_{}_free.mtz'.format(self.crystal, self.project, self.crystal))

            # if no sampleName was set, use prefix
            sample_name = ARGS.sampleName
            if sample_name == 'sampleName' or not sample_name:
                self.data_dict['sampleName'] = self.data_dict['prefix']

            # if aimless log exist use it.
            aimless_log = os.path.join(self.wdir, "LogFiles/{}_{}_aimless.log".format(self.project, self.crystal))
            if os.path.isfile(aimless_log):
                ap = AimlessParser(wdir=self.wdir, filename=aimless_log)
                self.data_dict.update(ap.summarize())

                self.data_dict['scalingsoftware'] = 'aimless'
                self.data_dict['scale_log'] = '{}/{}'.format(self.wdir, 'LogFiles/{}_{}_aimless.log'.format(self.project, self.crystal))

                truncate_logs = glob.glob(self.wdir + "/LogFiles/*truncate.log")
                if len(truncate_logs) == 1:
                    truncate_log = truncate_logs[0]
                else:
                    truncate_log = self.wdir + "/LogFiles/"

                self.data_dict['truncate_log'] = truncate_log  # '{}/{}'.format(self.wdir, 'LogFiles/AUTOMATIC_DEFAULT_NATIVE_truncate.log')
                self.data_dict['truncate_mtz'] = '{}/{}'.format(self.wdir, 'DataFiles/{}_{}_free.mtz'.format(self.project, self.crystal))

            if self.log2db:
                from ..log.database import save_to_db
                save_to_db(self.data_dict)

        else:
            logger.error('no results found in xia2.txt, {} may have failed.'.format(self.pipeline))

    def run_xia2(self):
        self.cmd_for_xia2().run_sync(timeout=10800)
        self.report()

    def get_crystal_name(self):
        crystal_name = ARGS.sampleName
        if self.image_sweeps:
            prefices = [s.get_prefix().rstrip("_") for s in self.image_sweeps]
            crystal_name = os.path.commonprefix(prefices)

            if len(crystal_name) <= 1:
                crystal_name = "_".join(prefices)[:30]

        if crystal_name[0].isdigit():
            crystal_name = "p_{}".format(crystal_name)
            logger.info("xia2 does not allow crystal name {}, prepend with p_".format(crystal_name))
    
        #crystal name must consist only of alphanumeric characters and underscores. The first character must be a non-digit character.
        crystal_name = crystal_name.replace("-", "_").replace(".","_")

        return crystal_name
