#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Thu Dec  1 08:38:51 2016

@author: qxu
"""
from __future__ import absolute_import

import logging
import os
import shutil
from collections import Counter

from .xds2mtz import XDSASCII2MTZ
from .xdsutils import XDSutils
from .xscale_inp import XSCALE_INP
from ..common.utils import parallelize
from ..file_parser.pointlessparser import PointlessParser
from ..process.get_args import ARGS, namespace2dict
from ..tasks.pointless import Pointless

logger = logging.getLogger(__name__)


class Scaler(object):
    def __init__(self,
                 xds_objs,
                 wdir='.',
                 subdir='scale',
                 nproc=4,
                 runner='local',
                 reindex_check=True,
                 use_parallelize=True,
                 log2db=False,
                 anomalous=False):
        self.wdir = os.path.join(os.path.abspath(wdir), subdir)
        if not os.path.isdir(self.wdir):
            os.makedirs(self.wdir, exist_ok=True)
        self.xds_objs = xds_objs
        self.scale_out_hkls = []
        self.nproc = nproc
        self.runner = runner
        self.reindex_check = reindex_check
        self.use_parallelize = use_parallelize
        self.log2db = log2db
        self.anomalous = anomalous
        self.metadata = {}
        self.n_wavelength = None

        self.runtime_info = namespace2dict(ARGS, values_as_string=True)
        self.data_dict = self.runtime_info.copy()
        self.OK = True
        self.highres_cutoff = None

        post_xds = XDSutils(self.xds_objs, wdir=self.wdir, nproc=self.nproc, use_parallelize=self.use_parallelize)
        self.ref_process, self.grouped_xds_objs = post_xds.get_xscale_input_objs()
        self.summary_html = post_xds.sweep_summary_html

        if self.summary_html and os.path.exists(self.summary_html):
            self.data_dict['report_url'] = "file://{}".format(self.summary_html)

        self.xds_objs_filtered = list()

        self.most_common_space_group = None
        self.multiple_crystal_mode = None
        if self.ref_process.get_possible_space_groups():
            self.data_dict['spacegroup_choices'] = self.ref_process.get_possible_space_groups()

    def get_multiple_crystal_mode(self):
        if (ARGS.recursive or len(self.xds_objs) > 3) and self.get_nwavelength() == 1:
            self.multiple_crystal_mode = True
        else:
            self.multiple_crystal_mode = False

        return self.multiple_crystal_mode

    def get_nwavelength(self):
        return len(self.grouped_xds_objs)

    def run_xscale_v3(self, ncycles=2):
        icycle = 0
        xds_objs = self.grouped_xds_objs
        xscale_inp = XSCALE_INP(wdir=self.wdir,
                                runner=self.runner,
                                anomalous=self.anomalous,
                                rejection_method=1,
                                nproc=2 * self.nproc).create_from_xds_objects(xds_objs)

        angle_cutoff = 20
        while icycle <= ncycles:
            logger.info("scaling cycle: {}".format(icycle))
            xscale_inp.run_xscale()
            xscale_inp.remove_outliers(icycle,
                                       max_resolution=self.ref_process.correct_summary['estimated_resolution'],
                                       use_xdscc12=True)

            icycle += 1
            angle_cutoff *= 0.9
            if xscale_inp.n_outliers == 0:
                break

        return xscale_inp.get_outfiles()

    def xds_ascii_2mtz(self, xds_ascii_hkl, wdir=None):
        # copy  meta data of xds obj members
        if not wdir:
            wdir = self.wdir

        data_dict = self.data_dict.copy()
        for wave in self.grouped_xds_objs:
            if str(wave) in xds_ascii_hkl:
                xds_metadata = self.get_metadata_as_dict()[wave]
                data_dict.update(xds_metadata)

        user_spacegroup = self.ref_process.pointless_spg_name if self.ref_process else None

        if self.most_common_space_group:
            user_spacegroup = self.most_common_space_group

        converter = XDSASCII2MTZ(xds_ascii_hkl,
                                 wdir=wdir,
                                 data_dict=data_dict,
                                 user_spacegroup=user_spacegroup,
                                 user_highres=self.runtime_info.get("highres"),
                                 sample_name=self.runtime_info.get("sampleName"),
                                 anomalous=self.anomalous
                                 )
        converter.convert()

    def get_metadata_as_dict(self):
        for key, xdsobjs in self.grouped_xds_objs.items():
            dicts = []
            res = {}
            for xds in xdsobjs:
                dicts.append(xds.get_metadata_as_dict())

            keys = dicts[0].keys()
            for k in keys:
                dlist = [str(d[k]) for d in dicts]
                if 'workdir' in k:
                    res[k] = os.path.commonpath(dlist)
                elif 'wavelength' in k:
                    res[k] = ':'.join(set(dlist))
                else:
                    res[k] = ':'.join(dlist)

            res['scaledir'] = self.wdir
            self.metadata[key] = res

        return self.metadata

    def merge_processed_data(self):
        if len(self.xds_objs) > 1:
            logger.info("multiple sweep scaling:")

            self.scale_out_hkls = self.run_xscale_v3()

            # rerun pointless to check space group assignments
            # this is useful in case of multiple sweeps merged together

            try:
                self.most_common_space_group = ChooseSpaceGroup(self.scale_out_hkls,
                                                                wdir=self.wdir).get_most_common_spacegroup()
                logger.info("most common space group: " + self.most_common_space_group)
            except Exception:
                logger.info("failed to get most common space group")

            if self.scale_out_hkls:
                if self.use_parallelize:
                    parallelize(self.xds_ascii_2mtz,
                                self.scale_out_hkls,
                                nproc=self.nproc)
                else:
                    for hkl in self.scale_out_hkls:
                        self.xds_ascii_2mtz(hkl)
        # only a sweep, bypass scaling to save time
        # copy XDS_ASCII.HKL into the scale dir and rename it to be consistent with
        # multiple sweep and convert it into mtz and make graphs
        elif len(self.xds_objs) == 1:
            logger.info("resolution cutoff and convert to mtz for single sweep:")
            (outfile, xds), = self.grouped_xds_objs.items()
            xds_dir = xds[0].get_process_dir()
            file_xds_ascii_hkl = os.path.join(xds_dir, 'XDS_ASCII.HKL')
            target_file_name = outfile
            shutil.copy(file_xds_ascii_hkl, os.path.join(self.wdir, target_file_name))

            self.xds_ascii_2mtz(target_file_name)
        else:
            logger.error("nothing to scale")

    def get_highres_cutoff(self):
        return self.highres_cutoff


def pointless2spg(xds_hklin, wdir=None):
    path, filename = os.path.split(xds_hklin)
    if not path:
        wdir = wdir
    else:
        wdir = path

    Pointless(wdir=wdir,
              script_name=filename + '-pointless-spg.sh'
              ).check_space_group(xds_hklin,
                                  logfile=filename + 'pointless-spg.log',
                                  outxml=filename + '-pointless-spg.xml')
    spg_info = PointlessParser(wdir=wdir, filename=filename + 'pointless-spg.log'
                               ).get_laue_group_from_xml(xmlfile=filename + '-pointless-spg.xml')
    return spg_info


class ChooseSpaceGroup:
    # rerun pointless to check the space group
    def __init__(self, xds_hklins, wdir=None):
        self.wdir = wdir
        self.xds_hklins = xds_hklins
        self.spg_infos = {}

    def get_all_pointless(self):
        for xdsin in self.xds_hklins:
            self.spg_infos[xdsin] = pointless2spg(xdsin, wdir=self.wdir)
        return self.spg_infos

    def get_most_common_spacegroup(self):
        if not self.spg_infos:
            self.get_all_pointless()

        print(self.spg_infos)
        return Counter(e['point_group_name'] for e in self.spg_infos.values()).most_common()[0][0]
