#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Thu Dec  1 08:36:49 2016

@author: qxu
"""
from __future__ import absolute_import
import logging
import os
import shutil

from ..common.spacegroup import Spacegroup
from ..common import utils
from ..common.imagesweep import ImageSweep
from ..file_parser.idxrefparser import IdxrefParser
from ..file_parser.integrationparser import IntegrationParser
from ..file_parser.pointlessparser import PointlessParser
from ..file_parser.correctparser import CorrectParser
from ..file_parser.xdsasciiparser import XdsAsciiParser
from ..file_parser.xparm import Xparm
from ..file_parser.xplanpaser import XplanParser
from ..file_parser.spotxds import SpotXDS

from ..tasks.scriptjob import ScriptJob
from ..tasks.pointless import Pointless
from ..tasks.xdsstat import XDSstat
from ..xds.xds_inp import XDS_INP
from ..xds.xds_info import XDS_INFO
from ..process.get_args import ARGS
from ..common.pv import update_strategy_ready_PV

from ..tasks.matthews_coef import MatthewCoef

logger = logging.getLogger(__name__)


class XDS(object):
    """ the runner class that rules over other classes"""

    def __init__(self,
                 wdir='.',
                 runner='local',
                 nproc=2,
                 njobs=1,
                 user_xdscmds=None,
                 user_space_group=None,
                 user_unit_cell=None,
                 user_refhkl=None,
                 highres=None,
                 native=False,
                 sampleName="sampleName"):
        self.image_sweep = None
        self.sweep_subdir = '.'
        self.pdir = os.path.abspath(wdir)
        self.wdir = self.pdir
        self.nproc = nproc
        self.njobs = njobs
        self.sampleName = sampleName

        if user_space_group:
            self.user_space_group = Spacegroup.get_spg_number_from_symbol(user_space_group)
        else:
            self.user_space_group = user_space_group

        if user_xdscmds:
            self.user_xdscmds = user_xdscmds
        else:
            self.user_xdscmds = ""

        self.user_refhkl = user_refhkl
        self.user_unit_cell = user_unit_cell
        self.native = native
        self.highres = highres

        self.pointless_spg = None
        self.pointless_spg_name = None
        self.pointless_cell = None
        self.correct_refined_results_dic = None
        self.wavelength = None

        self.process_dir = None
        if not os.path.isdir(self.wdir):
            os.makedirs(self.wdir)

        self.xdsinp = None
        self.runner = runner
        self.metadata = {}
        self.subdir = None
        self.correct_summary = None
        self.possible_space_groups = None  # hold point or space group all possible solutions
        self.idx_parser = None

        self.OK = True
        self.dbrow = None
        self.summary_html = None
        self.pointless_solution = None

    def get_summary_html(self):
        return self.summary_html

    def get_pointless_solution(self):
        return self.pointless_solution

    @property
    def status(self):
        return self.OK

    def get_imagesweep(self):
        return self.image_sweep

    def get_possible_space_groups(self):
        return self.possible_space_groups

    def get_metadata_as_dict(self):
        self.metadata = self.image_sweep.get_metadata_as_dict()
        self.metadata['pipeline'] = 'gmcaproc'
        self.metadata['xdsdir'] = self.get_process_dir()
        self.metadata['correctdir'] = self.get_process_dir()
        self.metadata['workdir'] = self.pdir
        self.metadata['subdir'] = self.subdir
        return self.metadata

    def get_process_dir(self):
        return self.process_dir

    def get_resolution_estimate(self, target_name="CC(1/2)", target_value=30):
        if os.path.exists(os.path.join(self.wdir, "CORRECT.LP")):
            cparser = CorrectParser(wdir=self.wdir)
            return cparser.get_resolution_estimation(target_value=target_value, target_name=target_name)
        else:
            return None

    def get_pointless_spg_cell(self):
        return self.pointless_spg, self.pointless_cell

    def set_xdsinp_from_image_sweep(self, a_image_sweep, save=False):
        if not isinstance(a_image_sweep, ImageSweep):
            self.image_sweep = ImageSweep(a_image_sweep)
        else:
            self.image_sweep = a_image_sweep

        self.sweep_subdir = self.image_sweep.get_id()
        self.wdir = os.path.join(self.pdir, self.sweep_subdir)
        self.wavelength = self.image_sweep.get_wavelength()
        self.xdsinp = XDS_INP(wdir=self.wdir,
                              njobs=self.njobs,
                              nproc=self.nproc,
                              native=self.native,
                              spot_range_mode='firsthalf',
                              xdscmds=self.user_xdscmds,
                              user_refhkl=self.user_refhkl,
                              user_space_group=self.user_space_group,
                              user_unit_cell=self.user_unit_cell,
                              highres=self.highres).create_from_sweep(self.image_sweep)
        if not os.path.isdir(self.wdir):
            os.makedirs(self.wdir)

        if save:
            self.xdsinp.save(dest_dir=self.wdir)

            # save db entry
            if ARGS.log2db:
                from ..log.database import DataProcessResults
                result = dict(pipelinestatus_id=ARGS.pipelinestatus_id, workdir=self.wdir)
                DataProcessResults.create(**result)

        return self

    def set_xdsinp_from_previous_run(self, src_dir, filename='XDS.INP'):
        self.xdsinp = (XDS_INP(wdir=self.wdir)
                       .load_from_file(from_dir=src_dir,
                                       filename=filename))
        if self.user_xdscmds:
            self.xdsinp.update(self.user_xdscmds)
        self.wdir = self.pdir
        if not os.path.isdir(self.wdir):
            os.makedirs(self.wdir)
        return self

    def set_xdsinp(self, xdsinp_obj):
        self.xdsinp = xdsinp_obj
        return self

    def run_xds_task(self, task='setup'):
        xdsjob = 'JOB='
        if task == 'setup':
            xdsjob += 'XYCORR INIT'
        elif task == 'spots':
            xdsjob += 'COLSPOT'
        elif task == 'index':
            xdsjob += 'IDXREF'
        elif task == 'spots-index':
            xdsjob += 'COLSPOT IDXREF'
        elif task == 'setup-2-index':
            xdsjob += 'XYCORR INIT COLSPOT IDXREF'
        elif task == 'xplan':
            xdsjob += 'DEFPIX XPLAN'
        elif task == 'index-xplan':
            xdsjob += 'IDXREF DEFPIX XPLAN'
        elif task == 'setup-2-spot':
            xdsjob += 'XYCORR INIT COLSPOT'
        elif task == 'defpix':
            xdsjob += 'DEFPIX'
        elif task == 'integrate':
            xdsjob += 'INTEGRATE'
        elif task == 'correct':
            xdsjob += 'CORRECT'
        elif task == 'defpix-integrate-correct':
            xdsjob += 'DEFPIX INTEGRATE CORRECT'
        elif task == 'integrate-correct':
            xdsjob += 'INTEGRATE CORRECT'
        elif task == 'index-integrate-correct':
            xdsjob += 'IDXREF DEFPIX INTEGRATE CORRECT'
        else:
            xdsjob += 'ALL'

        jobscript_name = '{}.sh'.format(task)
        self.xdsinp.update(xdsjob)
        self.xdsinp.save(dest_dir=self.wdir)

        jobstatus = ScriptJob(self.wdir, jobscript_name,
                              script_text='xds_par\n',
                              runner=self.runner,
                              nproc=self.nproc).run_sync(timeout=3600)
        return jobstatus

    def setup(self):
        return self.run_xds_task(task='setup')

    def index(self):
        return self.run_xds_task(task='index')

    def integrate(self):
        return self.run_xds_task(task='integrate')

    def correct(self):
        return self.run_xds_task(task='correct')

    def indexing_trials(self):
        """try different indexing combination, first using different spots, then different parameter"""
        indexed = False
        for spot_ranges in self.xdsinp.ranges.generate_xds_spot_ranges():
            last_spot_frame = spot_ranges[-1][-1]
            spot_ranges = ";".join(f"SPOT_RANGE= {r[0]} {r[1]}" for r in spot_ranges)
            self.xdsinp.update(spot_ranges)
            utils.wait_until_files_exist(self.image_sweep.get_image(last_spot_frame), timeout=ARGS.timeout)
            logger.info(f"run COLSPOT using {spot_ranges}")
            self.run_xds_task(task='spots')
            for parameter in ["INDEX_ERROR= 0.05",
                              "INDEX_ERROR= 0.10",
                              "INDEX_ERROR= 0.20",
                              "INDEX_ERROR= 0.05;MAXIMUM_ERROR_OF_SPOT_POSITION= 6.0;MAXIMUM_ERROR_OF_SPINDLE_POSITION= 4.0",
                              "INDEX_ERROR= 0.10;SEPMIN= 4.00;CLUSTER_RADIUS= 2",
                              ]:
                logger.info(f"run index using {parameter}")
                self.xdsinp.update(parameter)
                self.run_xds_task(task='index')

                # get triclinic cell in p1, run integrate and correct
                self.idx_parser = IdxrefParser(self.wdir)

                self.possible_space_groups = self.idx_parser.get_possible_space_groups(sgtype='spacegroup')
                if self.possible_space_groups:
                    indexed = True
                    p1cell = " ".join(map(str, self.idx_parser.get_p1_unit_cell()))
                    p1_cell = "SPACE_GROUP_NUMBER=1;UNIT_CELL_CONSTANTS={}".format(p1cell)
                    self.xdsinp.update(p1_cell)
                    return indexed

        if not indexed:
            self.OK = False
            logger.error("abandon processing sweep: {} since indexing failed.".format(self.image_sweep.get_prefix()))

        return indexed

    def integrate_trials(self):
        integrated = False
        last_frame = self.xdsinp.ranges.get_data_range()[-1]

        rescue_integ_fixes = ["DELPHI=20",
                              "DELPHI=40",
                              "DELPHI=80",
                              "DELPHI=180",
                              "REFINE(INTEGRATE)= BEAM POSITION ORIENTATION ! CELL"]

        utils.wait_until_files_exist(self.image_sweep.get_image(last_frame), timeout=ARGS.timeout)

        new_last_frame = False
        exclude_data_ranges_prev = None
        while True:
            self.run_xds_task(task='integrate')

            integrate_parser = IntegrationParser(self.wdir)
            if not os.path.exists(os.path.join(self.wdir, "INTEGRATE.HKL")):
                if not new_last_frame:  # only run once
                    integ_params_fix = integrate_parser.get_redo_suggestions()
                    rescue_integ_fixes.insert(0, integ_params_fix)
                    new_last_frame = True

                if rescue_integ_fixes:
                    new_params = rescue_integ_fixes.pop(0)
                    logger.info(f"rerun integrate with {new_params}")
                    self.xdsinp.update(new_params)
                else:
                    break
            else:  # integration didn't fail, check bad frames
                exclude_ranges = integrate_parser.get_exclude_range(start_threshold=2,
                                                                    max_threshold=3,
                                                                    min_reject=0,
                                                                    min_gap=5)
                if exclude_ranges and exclude_ranges != exclude_data_ranges_prev:
                    exclude_data_ranges = ";".join(
                        f"EXCLUDE_DATA_RANGE= {start} {end}" for start, end in exclude_ranges)
                    logger.info(f"rerun integrate with {exclude_data_ranges}")
                    exclude_data_ranges_prev = exclude_ranges
                    self.xdsinp.update(exclude_data_ranges)
                    # self.xdsinp.update(integrate_parser.get_refined_params_for_beam_reflections())
                else:
                    return True

        return integrated

    def process_in_p1(self, subdir='XDS-p1'):
        self.wdir = os.path.join(self.pdir, self.sweep_subdir, subdir)
        self.process_dir = self.wdir
        self.subdir = subdir

        if not self.xdsinp:
            self.set_xdsinp_from_image_sweep(self.image_sweep)

        utils.wait_until_files_exist(self.image_sweep.get_image(self.xdsinp.ranges.get_bg_range()[-1]), timeout=300)
        # run initialization
        self.run_xds_task(task='setup')

        if not self.indexing_trials():
            return

        self.run_xds_task(task='defpix')

        if not self.integrate_trials():
            return

        self.run_xds_task(task='correct')

        utils.rename_files_with_label(self.wdir,
                                      ['XDS_ASCII.HKL', 'CORRECT.LP'],
                                      label='-p1')

        # use pointless to determine the spg/cell
        Pointless(wdir=self.wdir,
                  script_name='pointless-spg.sh'
                  ).check_space_group('XDS_ASCII-p1.HKL')
        spg_res = PointlessParser(wdir=self.wdir,
                                  filename='pointless-spg.log'
                                  ).get_laue_group_from_xml(xmlfile='pointless-spg.xml')

        if not spg_res:
            logger.error("pointless @ {} failed".format(self.wdir))
            return

        # run correct in pointless determined cell/spg
        '''
        # NB not working for C2 since pointless gives a different cell with permutation which xds does not like
        spg_num = spg_res['point_group_number']
        cell = ' '.join(map(str, spg_res['cell']))
        px_cell = "SPACE_GROUP_NUMBER={};UNIT_CELL_CONSTANTS={}".format(spg_num, cell)
        '''
        # just use pointless spg but use xds cell
        cell, spg_num = self.idx_parser.get_unit_cell_from_spg(spg_res['point_group_name'], n_extra_solutions=5)
        if not cell:
            logger.error("failed to get cell for spg {} in {}/IDXREF.LP, aborted.".format(spg_res['point_group_name'],
                                                                                          self.wdir))
            return

        cell = ' '.join(map(str, cell))
        px_cell = "SPACE_GROUP_NUMBER={};UNIT_CELL_CONSTANTS={}".format(spg_num, cell)

        self.pointless_solution = spg_res
        self.pointless_cell = cell
        self.pointless_spg = spg_num
        self.pointless_spg_name = spg_res['point_group_name']
        logger.info("pointless suggested space group as {}, unit_cell = {}.".format(spg_num, cell))
        self.xdsinp.update(px_cell)
        self.run_xds_task(task='correct')

        if not os.path.exists(os.path.join(self.wdir, "CORRECT.LP")):
            logger.error("failed to integrate: {}".format(self.wdir))
            return

        cparser = CorrectParser(wdir=self.wdir)
        self.OK = cparser.OK

        self.correct_refined_results_dic = cparser.get_refined_point_group_cell_orientation()
        self.correct_summary = cparser.get_summary()

        logger.info("Processing result for {}:\n{}".format(self.wdir, cparser.summarize()))

        # run xdsstat and xdscc12
        stat = XDSstat(wdir=self.wdir, lowres=50.0, highres=cparser.get_resolution_estimation())
        stat.run()
        self.summary_html = stat.get_summary_html()

        return self

    def process(self, subdir='XDS', ref_XDS=None, single_crystal=True):
        self.wdir = os.path.join(self.pdir, self.sweep_subdir, subdir)
        self.process_dir = self.wdir
        self.subdir = subdir

        if not os.path.isdir(self.wdir):
            os.makedirs(self.wdir)

        if not self.xdsinp:
            self.set_xdsinp_from_image_sweep(self.image_sweep)

        # optimization, or use the same  orientation to run integrate/correct
        if ref_XDS and single_crystal:
            src_dir = ref_XDS.get_process_dir()
            ref_hkl = 'REFERENCE_DATA_SET={}'.format(os.path.join(src_dir, 'XDS_ASCII.HKL'))

            # TODO copy from parent instead of ref???
            parent_dir = os.path.join(self.wdir, "..", "XDS-p1")
            if os.path.exists(parent_dir) and os.path.exists(os.path.join(parent_dir, 'XDS_ASCII.HKL')):
                src_dir = parent_dir

            file_gxparm = os.path.join(src_dir, 'GXPARM.XDS')
            if os.path.isfile(file_gxparm):
                if self.image_sweep:
                    new_start_angle = self.image_sweep.get_start_angle()
                    new_distance = self.image_sweep.get_distance()
                    new_wavelength = self.image_sweep.get_wavelength()
                    new_osc = self.image_sweep.get_osc()
                    Xparm().rewrite_gxparm_xds_with_new_params(new_start_angle,
                                                               new_distance,
                                                               new_wavelength,
                                                               new_osc,
                                                               src_dir=src_dir,
                                                               dest_dir=self.wdir)
                    self.xdsinp.update(ref_hkl)

                if ref_XDS.correct_refined_results_dic:
                    self.xdsinp.update(ref_XDS.correct_refined_results_dic['xdsinp_update_string'])
                else:
                    logger.error('unhandled situation in the xds class.')
            else:
                logger.error("cannot find file {}.".format(file_gxparm))

            for f in XDS_INFO['files_needed_by_INTEGRATE']:
                if f not in ['XPARM.XDS', 'XDS.INP']:
                    shutil.copy(os.path.join(src_dir, f), self.wdir)

            self.xdsinp.update("NUMBER_OF_PROFILE_GRID_POINTS_ALONG_ALPHA/BETA=13")
            self.xdsinp.update("NUMBER_OF_PROFILE_GRID_POINTS_ALONG_GAMMA=13")
            parser = IntegrationParser(wdir=src_dir,
                                       filename='INTEGRATE.LP')
            newparams = parser.get_refined_params_for_beam_reflections()
            self.xdsinp.update(newparams)
            self.run_xds_task('integrate-correct')
        # if there is a reference xds processing, but sweeps not having the same crystal orientation
        elif ref_XDS and not single_crystal:
            self.run_xds_task(task='setup-2-spot')
            ref_dir = ref_XDS.get_process_dir()
            refhkl = os.path.join(ref_dir, 'XDS_ASCII.HKL')
            if os.path.isfile(refhkl):
                self.xdsinp.update('REFERENCE_DATA_SET={}'.format(refhkl))

            self.run_xds_task(task='index')
            self.run_xds_task(task='defpix-integrate-correct')
        # a regular run of xds with or without user given spg, cell with reference data
        # NB this option is obsolete since the spg and unit_cell are already parsed in get_args
        elif self.user_refhkl and os.path.isfile(self.user_refhkl):
            self.run_xds_task(task='setup-2-spot')

            hkldir, hklname = os.path.split(self.user_refhkl)
            spg_num, cell = XdsAsciiParser(hkldir, filename=hklname).get_spg_cell(spg_type='number')
            if spg_num and cell:
                spg_cell = "SPACE_GROUP_NUMBER={};UNIT_CELL_CONSTANTS={}".format(spg_num, ' '.join(cell))
                self.xdsinp.update(spg_cell)
                self.xdsinp.update('REFERENCE_DATA_SET={}'.format(self.user_refhkl))
                self.run_xds_task(task='index')
                self.run_xds_task(task='defpix-integrate-correct')
            else:
                logger.error("reference data {} is not in XDS_ASCII format.".format(self.user_refhkl))
                return
        else:
            cell = spg_num = None

            self.run_xds_task(task='setup-2-spot')
            self.run_xds_task(task='index')
            idx_parser = IdxrefParser(self.wdir)
            self.possible_space_groups = idx_parser.get_possible_space_groups(sgtype='spacegroup')

            if not self.possible_space_groups:
                logger.error("failed to extract space groups from {}/IDXREF.LP".format(self.wdir))
                return

            if not self.user_space_group and not self.user_unit_cell:
                cell, spg_num = idx_parser.get_unit_cell_and_spg()
            elif self.user_space_group and not self.user_unit_cell:
                cell, spg_num = idx_parser.get_unit_cell_from_spg(self.user_space_group, n_extra_solutions=5)
            elif self.user_space_group and self.user_unit_cell:
                cell, spg_num = IdxrefParser.validate_cell_spg(self.user_unit_cell, self.user_space_group)
            else:
                logger.error("if cell is given, you must provide the space group as well.")

            if not cell:
                logger.error(
                    "failed to extract space group {} from {}/IDXREF.LP".format(self.user_space_group, self.wdir))
                return

            cell = " ".join(map(str, cell))
            logger.info('processing data in space group {} using cell {}.'.format(spg_num, cell))

            spg_cell = "SPACE_GROUP_NUMBER={};UNIT_CELL_CONSTANTS={}".format(spg_num, cell)
            if self.user_refhkl and os.path.exists(self.user_refhkl):
                self.xdsinp.update('REFERENCE_DATA_SET={}'.format(self.user_refhkl))

            self.xdsinp.update(spg_cell)
            self.run_xds_task(task='index')
            self.run_xds_task(task='defpix-integrate-correct')

        cparser = CorrectParser(wdir=self.wdir)
        self.correct_refined_results_dic = cparser.get_refined_point_group_cell_orientation()
        self.correct_summary = cparser.get_summary()
        logger.info("Processing result for {}:\n{}".format(self.wdir, cparser.summarize()))

        # XDSstat(wdir=self.wdir, lowres=50.0, highres=cparser.get_resolution_estimation()).run()

        return self

    def update_correct_summary(self):
        cparser = CorrectParser(wdir=self.wdir)

        try:
            self.correct_summary = cparser.get_summary()
        except Exception:
            logger.warning("Failed to update correct_summary in {}".format(self.wdir))
            return

    def refine(self, max_cycles=3):
        """ carry out refine protocol of xds by iterating integration-correct
        1) copy GXPARM.XDS into GXPARM
        2) update the XDS.INP with beam convergence and reflections
        3) rerun integrate-correct in the same directory
        """
        ncycles = 0
        while ncycles < max_cycles:
            file_gxparm = os.path.join(self.wdir, 'GXPARM.XDS')
            if not os.path.exists(file_gxparm):
                logger.error("{} does not exist, cannot proceed with refinement".format(file_gxparm))
                return

            shutil.copy(file_gxparm, os.path.join(self.wdir, 'XPARMS.XDS'))

            cparser = CorrectParser(wdir=self.wdir)
            self.correct_refined_results_dic = cparser.get_refined_point_group_cell_orientation()
            highres_cutoff = cparser.get_resolution_estimation(target_name="I/SIGMA", target_value=0.25)
            if ncycles > 0:
                logger.info("Processing result for cycle {} at {}:\n{}"
                            .format(ncycles, self.wdir, cparser.summarize()))

            self.xdsinp.update(self.correct_refined_results_dic['xdsinp_update_string'])
            self.xdsinp.update("NUMBER_OF_PROFILE_GRID_POINTS_ALONG_ALPHA/BETA=15")
            self.xdsinp.update("NUMBER_OF_PROFILE_GRID_POINTS_ALONG_GAMMA=15")
            # update resolution cutoff as well in refinement
            logger.info("high resolution cutoff was set as {}".format(highres_cutoff))
            self.xdsinp.update("INCLUDE_RESOLUTION_RANGE=50.0 {}".format(highres_cutoff))

            if float(highres_cutoff) >= 3.0:
                logger.info("for resolution better than 2.5, turn on cell refinement.")
                self.xdsinp.update("REFINE(INTEGRATE)= ORIENTATION CELL")

            parser = IntegrationParser(wdir=self.wdir, filename='INTEGRATE.LP')
            newparams = parser.get_refined_params_for_beam_reflections()
            self.xdsinp.update(newparams)
            self.run_xds_task('integrate-correct')
            ncycles += 1

        stat = XDSstat(wdir=self.wdir)
        stat.run()
        self.summary_html = stat.get_summary_html()

    def test_space_group_xds(self):
        self.xdsinp.update('SPACE_GROUP_NUMBER=0')
        self.run_xds_task(task='correct')
        return CorrectParser(wdir=self.wdir,
                             filename='CORRECT.LP'
                             ).get_refined_point_group_cell_orientation()

    # strategy calculation using xds
    def strategy(self, subdir="strategy", log2db=True):
        self.wdir = os.path.join(self.pdir, self.sweep_subdir, subdir)
        self.process_dir = self.wdir
        self.subdir = subdir

        if not os.path.isdir(self.wdir):
            os.makedirs(self.wdir)

        if not self.xdsinp:
            self.set_xdsinp_from_image_sweep(self.image_sweep)

        if os.path.exists(os.path.join(self.wdir, "SPOT.XDS")):  # save time if previous step was already run
            self.run_xds_task(task="index")
        else:
            self.run_xds_task(task='setup-2-index')

        # pick a solution if user has not provide it
        idx_parser = IdxrefParser(self.wdir)

        if log2db:
            from ..log.database import ScreenStrategyResults
            result = dict(pipelinestatus_id=ARGS.pipelinestatus_id,
                          workdir=self.wdir,
                          directory=self.image_sweep.get_absdir(),
                          sampleName=self.sampleName,
                          software="xds_strategy",
                          resolution_from_spots=idx_parser.get_resolution(),
                          n_spots=idx_parser.get_nspots(),
                          state='SPOT',
                          images=';'.join(self.image_sweep.get_all_files(abs_path=False)))
            self.dbrow = ScreenStrategyResults.create(**result)

        if idx_parser.check_error() and 'ABORT' in idx_parser.check_error()[-1]:
            logger.error('XDS indexing failed.')
            return

        if self.user_refhkl:
            hkldir, hklname = os.path.split(self.user_refhkl)
            self.user_space_group, self.user_unit_cell = XdsAsciiParser(hkldir,
                                                                        filename=hklname
                                                                        ).get_spg_cell(spg_type='symbol')

        spg_num, cell = None, None
        if not self.user_space_group and not self.user_unit_cell:
            cell, spg_num = idx_parser.get_unit_cell_and_spg()
        elif self.user_space_group and not self.user_unit_cell:
            cell, spg_num = idx_parser.get_unit_cell_from_spg(self.user_space_group, n_extra_solutions=5)
        elif self.user_space_group and self.user_unit_cell:
            cell, spg_num = IdxrefParser.validate_cell_spg(self.user_unit_cell, self.user_space_group)
        else:
            logger.error("if cell is given, you must provide the space group as well.")

        self.possible_space_groups = idx_parser.get_possible_space_groups(sgtype='pointgroup')

        if cell and spg_num:
            cell = " ".join(map(str, cell))
            spg_cell = "SPACE_GROUP_NUMBER={};UNIT_CELL_CONSTANTS={}".format(spg_num, cell)
            self.xdsinp.update(spg_cell)
            edge_res = self.highres
            if not edge_res:
                edge_res = self.image_sweep.get_edge_res()
            self.xdsinp.update("INCLUDE_RESOLUTION_RANGE=50.0 {}".format(edge_res))
            self.run_xds_task(task='index')  # rerun index in the desired spg/cell

            idx_parser = IdxrefParser(self.wdir)
            if idx_parser.check_error() and 'ABORT' in idx_parser.check_error()[-1]:
                logger.error('XDS indexing using {}/{}failed.'.format(spg_num, cell))
                return

            solution, p1_solution = idx_parser.get_index_solutions()

            self.run_xds_task(task='xplan')
            xplan = XplanParser(self.wdir, filename='XPLAN.LP')

            # calculate matthews coef
            mc = MatthewCoef(spg_num, cell, wdir=self.wdir)
            solvent_content = mc.get_table()
            asu_content = mc.get_asu_content()

            try:
                max_res = float(solution.Resolution_from_spots)
                distance = self.image_sweep.get_detector_distance_given_max_res(max_res, default_distance=350,
                                                                                tolerance=0.6)
            except:
                logger.warning("detector distance estimation failed.")
                distance = 350

            # save index and xplan results into db
            if log2db and self.dbrow:
                from ..log.database import ScreenStrategyResults
                self.dbrow.resolution_from_spots = solution.Resolution_from_spots
                self.dbrow.rmsd = solution.sdxy
                self.dbrow.n_spots = solution.n_spots
                self.dbrow.spacegroup = solution.Spacegroup
                self.dbrow.unitcell = solution.Refined_cell
                self.dbrow.score = solution.Score
                self.dbrow.state = 'INDEX'
                self.dbrow.bravais_lattice = solution.Bravais_lattice
                self.dbrow.index_table = solution.index_table
                self.dbrow.mosaicity = solution.Mosaicity
                self.dbrow.workdir = self.wdir
                self.dbrow.solvent_content = solvent_content
                self.dbrow.estimated_asu_content_aa = asu_content
                if self.possible_space_groups:
                    self.dbrow.pointgroup_choices = self.possible_space_groups

                if self.user_refhkl:
                    self.dbrow.referencedata = self.user_refhkl
                # update strategy field
                self.dbrow.sampleName = self.image_sweep.get_imageSet(mode="prefix")

                self.dbrow.xplanlog = xplan.plans(DESIRED_COMPLETENESS=95)
                self.dbrow.osc_start = xplan.get_osc_start()
                self.dbrow.osc_end = xplan.get_osc_end()
                self.dbrow.osc_delta = "0.2"
                self.dbrow.detectordistance = distance
                self.dbrow.state = 'STRATEGY'
                if xplan.anomalous():
                    self.dbrow.anomalous = 1
                    self.dbrow.completeness_anomalous = xplan.get_completeness()
                else:
                    self.dbrow.completeness_native = xplan.get_completeness()
                    self.dbrow.anomalous = 0

                self.dbrow.save()
                update_strategy_ready_PV()

                logger.info("strategy calculation by XDS: DONE.")

        else:
            logger.error("illegal cell {} or space group {}.".format(cell, spg_num))
