from __future__ import absolute_import
import logging

logger = logging.getLogger(__name__)


class XDS_ranges:
    def __init__(self, start, end, osc, delta=5.0, mode='firsthalf'):
        self.start = start
        self.end = end
        self.osc = osc
        self.delta = delta

        if osc < 0.001:
            logger.error("osc is close to zero.")

        self.totalosc = (start - end + 1) * osc

        ni = int(delta / osc)
        last = min(start + ni - 1, start + 19)  # 5 is usually enough, 20 ample
        if last > end:
            last = end

        self.backgroundRange = (start, start+5)
        self.spotRanges = self.set_spot_ranges(mode=mode)
        self.dataRange = (start, end)

    def get_bg_range(self):
        return self.backgroundRange

    def get_data_range(self):
        return self.dataRange

    def set_data_range(self, start, end):
        self.dataRange = (start, end)

    def set_bg_range(self, start, end):
        self.backgroundRange = (start, end)

    def get_spot_ranges(self):
        return self.spotRanges

    def set_spot_ranges(self, mode, userSpotRanges=None, n=3, m=5):
        mode = mode.lower()
        if mode == 'standard':
            spotRanges = XDS_ranges.calSpotRange(self.start,
                                                 self.end,
                                                 self.osc,
                                                 spotrange_wedge=3.0,
                                                 wedge=15.0)
        elif mode == 'wedge45':
            spotRanges = XDS_ranges.calSpotRange(self.start,
                                                 self.end,
                                                 self.osc,
                                                 spotrange_wedge=3.0,
                                                 wedge=45.0)
        elif mode == 'wedge90':
            spotRanges = XDS_ranges.calSpotRange(self.start,
                                                 self.end,
                                                 self.osc,
                                                 spotrange_wedge=3.0,
                                                 wedge=90.0)
        elif mode == 'firsthalf':
            spotRanges = [(self.start, (self.start + self.end) // 2)]
        elif mode == 'full':
            spotRanges = [(self.start, self.end)]
        elif mode == 'sparse':
            spotRanges = self.get_n_of_m_images(n, m)
        elif mode == 'user' and userSpotRanges:
            spotRanges = userSpotRanges
        else:
            spotRanges = [(self.start, self.end)]
        return spotRanges

    def alternative_xds_spot_ranges(self):
        """return alternative spot ranges (can be used in case of indexing failure) """
        alternative_spot_ranges = []
        for mode in ['sparse', 'standard', 'wedge45', 'wedge90', 'full']:
            ranges = self.set_spot_ranges(mode=mode)
            alternative_spot_ranges.append(";".join(f"SPOT_RANGE= {r[0]} {r[1]}" for r in ranges))

        return alternative_spot_ranges

    def generate_xds_spot_ranges(self):
        """return alternative spot ranges (can be used in case of indexing failure) """
        all_ranges = []
        for mode in ['firsthalf', 'full', 'sparse', 'standard', 'wedge45', 'wedge90']:
            r = self.set_spot_ranges(mode=mode)
            if r not in all_ranges:
                all_ranges.append(r)
        return all_ranges

    @staticmethod
    def calSpotRange(start, end, osc, spotrange_wedge=3.0, wedge=90.0):
        # spot ranges samples a portion of data (default 3 degrees of data)
        # every certain degree of data
        totalimg = end - start + 1
        totalosc = totalimg * osc

        if totalosc < 5.0:
            return [(start, end)]

        binsize = spotrange_wedge
        # fine slice data, reduce bin size
        # binsize = spotrange_wedge if osc > 0.4 else osc * 5
        # if totalosc > 90:
        #    wedge = totalosc / 10.0
        # else:
        #    wedge = totalosc / 6.0

        if wedge < 2 * binsize:
            wedge = 2 * binsize

        nImgPerBin = int(binsize / osc)
        nWedge = int(totalosc / wedge) + 1
        nImgPerWedge = int(wedge / osc)

        count = 0
        res = []
        for i in range(nWedge):
            x1 = start + i * nImgPerWedge
            x2 = x1 + nImgPerBin
            if x1 >= end:
                x1 = end - nImgPerBin
            if x2 >= end:
                x2 = end
            res.append((x1, x2))
            count += 1
            if count >= 10:
                break
        return res

    # n segments of m images each, even distributed, e.g. in fast_dp n=3, m=5
    def get_n_of_m_images(self, n, m):
        return XDS_ranges.calSparseSpotRange(self.start, self.end, n, m)

    # n segments of m images each, even distributed, e.g. in fast_dp n=3, m=5
    @staticmethod
    def calSparseSpotRange(s, e, n, m):
        nlen = e - s + 1
        if n == 1:
            return [((nlen - m) // 2, (nlen + m) // 2)]

        gap = (nlen - m * n) // (n - 1) + m
        return [(s + i * gap, s + i * gap + m - 1) for i in range(n)]

    def __str__(self):
        return ("background range: " + str(self.backgroundRange)
                + " data range: " + str(self.dataRange)
                + " spot ranges: " + str(self.spotRanges))
