from __future__ import absolute_import
import logging
import os
from collections import defaultdict, Counter
from functools import partial
from operator import itemgetter

from ..common.utils import parallelize, rel_path, isFloat
from ..file_parser.correctparser import CorrectParser
from ..xds.xds import XDS
from ..xds.xds_inp import XDS_INP
from ..file_parser.table import Table
from ..common.spacegroup import Spacegroup

logger = logging.getLogger(__name__)


class XDSutils(object):
    def __init__(self, xds_objs, wdir=".",
                 nproc=1,
                 minimum_ISa=0.5,
                 use_parallelize=True,):
        self.xds_objs = xds_objs
        self.wdir = wdir
        self.n_wavelength = 1
        self.use_parallelize = use_parallelize
        self.nproc = nproc
        self.ref_process = None
        self.grouped_xds_objs = None
        self.xds_objs_filtered = defaultdict(list)
        self.sweep_summary_html = None
        self.minimum_ISa = minimum_ISa

        self.process()

    def get_ref_process(self):
        if not self.ref_process:
            self.process()
        return self.ref_process

    def get_xscale_input_objs(self):
        # return reference dataset and a dictionary with outfile as key, input files as arrays
        return self.ref_process, self.xds_objs_filtered

    def process(self):
        # step 1, choose a reference and reindex if necessary
        if len(self.xds_objs) > 1:
            self.ref_process = self.get_reference_xds()
            self.rerun_correct_with_reference_all()
        else:
            self.ref_process = self.xds_objs[0]  # single sweep case

        # generate a grouped xds (outfile is key, each with a list of xds objects
        self.write_xds_summary()

        # step 2, group xds objects by wavelength
        self.grouped_xds_objs = self.xds_groupby_wavelength()
        self.filter_grouped_XDSs()

    def filter_grouped_XDSs(self):
        all_spgs = set([x.correct_summary['spg'] for x in self.xds_objs if x.correct_summary])
        if not all_spgs:
            logger.info("no valid sweeps found")
            raise SystemExit(1)

        ref_spg = self.ref_process.correct_summary['spg']
        ref_cell = self.ref_process.correct_summary['cell']

        if len(all_spgs) != 1:
            logger.critical("!!!WARNING!!! space group ({}) does not match between sweeps!!!".format(all_spgs))
            logger.info(
                "will try to proceed with spg {}, based on i/sigma".format(self.ref_process.correct_summary['spg']))

        for k, v in self.grouped_xds_objs.items():
            outfile = "{}w{}.HKL".format(XDSutils.combine_prefices(v), str(k))

            xds_objs = list()
            for xds in v:
                proc_dir = xds.get_process_dir()
                scale_hkl = os.path.join(proc_dir, 'XDS_ASCII.HKL')

                if os.path.isfile(scale_hkl):
                    spg_curr = xds.correct_summary.get('spg', 'NA')
                    cell_curr = xds.correct_summary.get('cell', 'NA')
                    ISa_curr = xds.correct_summary.get('ISa', 'NA')
                    reso_curr = xds.correct_summary.get('estimated_resolution', 'NA')

                    # discard because space group mismatch, but will allow if laue group is the same
                    if not Spacegroup.is_pointgroup_equal(spg_curr, ref_spg):
                        logger.warning("{} ignored in xscale due to space group mismatch.".format(proc_dir, spg_curr))
                        continue

                    if Spacegroup.is_cell_equal(ref_cell, cell_curr) is False:
                        logger.warning("{} ignored in xscale due to cell mismatch.".format(proc_dir, cell_curr))
                        continue

                    if reso_curr >= 8.0:
                        logger.warning("{} ignored in xscale due to low resolution {}A.".format(proc_dir, reso_curr))
                        continue

                    if isFloat(ISa_curr) and float(ISa_curr) < self.minimum_ISa:
                        logger.warning("{} ignored in xscale due to low ISa {}.".format(proc_dir, ISa_curr))
                        continue

                    xds_objs.append(xds)
                else:
                    logger.error('file {} does not exist.'.format(scale_hkl))

            if len(xds_objs) > 0:  # only add lines of there are valid input lines
                self.xds_objs_filtered[outfile] = xds_objs

        if not self.xds_objs_filtered:
            raise SystemExit("Nothing to scaling. Aborted.")

    def xds_groupby_wavelength(self, rounded_by=4):
        """ group sweeps by wavelength, rounded to 4 digits """
        # wave4 = [ round(s.get_wavelength(), rounded_by) for s in image_sweeps ]
        res = defaultdict(list)
        for xds in self.xds_objs:
            if xds.wavelength:
                key = round(float(xds.wavelength), rounded_by)
                res[key].append(xds)
            else:
                logger.error("in scaler, no wavelength was set for {}"
                             .format(xds.get_process_dir()))
        self.n_wavelength = len(res)
        return res

    def get_reference_xds(self, mode="completeness"):
        if not self.ref_process:
            if not self.xds_objs:
                self.ref_process = None
                return

            all_spgs = [x.correct_summary['spg'] for x in self.xds_objs if x.correct_summary]
            if not all_spgs:
                logger.info("no successfully processed sweeps, aborting.")
                raise SystemExit(1)
                return

            spg_counter = Counter(all_spgs)
            # assume all spg is a number, first sort by frequency, then by number (higher symmetry)
            # most_frequent_spg = Counter(all_spgs).most_common(1)[0][0]
            most_frequent_spg = [v for v in sorted(spg_counter.items(), key=itemgetter(1, 0), reverse=True)][0][0]
            logger.info('most frequent space group observed: {}'.format(most_frequent_spg))

            # all processing with the same space group
            xdsobjs_filtered_by_spg = [xds for xds in self.xds_objs
                                       if xds.correct_summary and xds.correct_summary['spg'] == most_frequent_spg
                                       and xds.correct_summary['ISa'] <= 48.0
                                       and xds.correct_summary['estimated_resolution'] <= 5.0
                                       ]

            if not xdsobjs_filtered_by_spg:
                self.ref_process = self.xds_objs[0]
            else:
                if mode == "isigma":
                    self.ref_process = \
                        sorted(xdsobjs_filtered_by_spg, key=lambda xds: -xds.correct_summary['I/SigmaI'])[0]
                    logger.info("using sweep with top 3 ISa with max i/sigmai as reference in {}.".format(
                        self.ref_process.get_process_dir()))
                else:
                    # data with best ISa values
                    xdsobjs_filtered_by_ISa = sorted(xdsobjs_filtered_by_spg,
                                                     key=lambda xds: -xds.correct_summary['ISa'])[:3]

                    if mode == "completeness":
                        xdsobjs_filtered_by_completeness = sorted(xdsobjs_filtered_by_ISa,
                                                                  key=lambda xds: -xds.correct_summary['Completeness'])
                        self.ref_process = xdsobjs_filtered_by_completeness[0]
                        logger.info("using sweep with top 3 ISa with max completeness as reference in {}.".format(
                            self.ref_process.get_process_dir()))
                    else:
                        self.ref_process = xdsobjs_filtered_by_ISa[0]
                        logger.info("using sweep with maximum ISa as reference in {}.".format(
                            self.ref_process.get_process_dir()))

        return self.ref_process

    def rerun_correct_with_reference_all(self):
        # rerun correct to add comparison to the reference
        # NB XDS attempts resolve indexing ambiguity between reference data and target data
        # so a specific reindexing is not necessary
        # ref_proc = self.xds_objs[0]
        ref_proc = self.ref_process

        if self.use_parallelize:
            func = partial(XDSutils.rerun_correct_with_reference,
                           ref_proc=ref_proc)
            parallelize(func, self.xds_objs, nproc=self.nproc)
        else:
            [XDSutils.rerun_correct_with_reference(p, ref_proc) for p in self.xds_objs]

        return [x.update_correct_summary() for x in self.xds_objs if os.path.exists(os.path.join(x.get_process_dir(), "CORRECT.LP"))]

    # probably should move to other places
    def write_xds_summary(self):
        tbl = Table()
        tbl.set_headers(["#", "Sweep", "Images", "XDSsummary", "Rmerge", "Resolution", "Completeness",
                         "ISa", "I/sigmaI", "SpaceGroup",
                         "UnitCell", "XDS.INP", "CORRECT.LP", "XDSSTAT", "XDSCC12"])

        for ix, xds in enumerate(self.xds_objs, 1):
            sweep = xds.get_imagesweep().get_prefix()
            nimages = xds.get_imagesweep().nimages()

            xdsinp = os.path.join(xds.get_process_dir(), "XDS.INP")
            correctlog = os.path.join(xds.get_process_dir(), "CORRECT.LP")
            xdsstat = os.path.join(xds.get_process_dir(), "XDSSTAT.LP.log")
            xdscc12 = os.path.join(xds.get_process_dir(), "XDSCC12.LP")

            summary = os.path.join(xds.get_process_dir(), "XDSsummary.html")
            if os.path.exists(summary):
                summary = XDSutils.linkify(summary)
            else:
                summary = 'NA'

            spg, cell, ISa, Rmerge, resol, completeness, isigma = ["NA"]*7

            if xds.correct_summary:
                spg = xds.correct_summary.get('spg', 'NA')
                cell = xds.correct_summary.get('cell', 'NA')
                ISa = xds.correct_summary.get('ISa', 'NA')
                Rmerge = xds.correct_summary.get('Rmerge', 'NA')
                resol = xds.correct_summary.get('estimated_resolution', 'NA')
                completeness = xds.correct_summary.get('Completeness', 'NA')
                isigma = xds.correct_summary.get('I/SigmaI', 'NA')

            tbl.add_data_by_a_list([ix] +
                                   [str(e).strip() for e in
                                    [sweep, nimages, summary, Rmerge, resol, completeness, ISa, isigma, spg, cell]] +
                                   [XDSutils.linkify(path) if os.path.exists(path) else "NA" for path in
                                    [xdsinp, correctlog, xdsstat, xdscc12]]
                                   )

        datafilestat = os.path.join(self.wdir, "SWEEP_PROCESSING_SUMMARY.html")
        with open(datafilestat, "w") as f:
            f.write(tbl.htmlify())

        if len(self.xds_objs) == 1:
            self.sweep_summary_html = self.xds_objs[0].get_summary_html()
        else:
            self.sweep_summary_html = datafilestat

    # turn a directory into a link
    @staticmethod
    def linkify(path):
        return "<a href='file://{}'>Open</a>".format(path)

    @staticmethod
    def combine_prefices(xds_objs, max_prefix_length=39):
        prefices = [xds.get_imagesweep().get_prefix() for xds in xds_objs]

        combined_prefix = os.path.commonprefix(prefices)

        if not combined_prefix:
            if len(prefices) <= 2:
                combined_prefix = ''.join(xds.get_imagesweep().get_prefix() for xds in xds_objs)
            else:
                combined_prefix = ''.join(xds.get_imagesweep().get_prefix() for xds in xds_objs[:2])
                combined_prefix += "..."

            if len(combined_prefix) > max_prefix_length:  # OUTPUT_FILE in XDS filename is limited to 50 chars
                logger.warning("combined prefices {} is more than {} chars ...".format(combined_prefix, max_prefix_length))
                combined_prefix = combined_prefix[:max_prefix_length]
        return combined_prefix

    @staticmethod
    def rerun_correct_with_reference(proc, ref_proc):
        ref_dir = ref_proc.get_process_dir()
        reference_hkl = os.path.join(ref_dir, 'XDS_ASCII.HKL')

        wdir = proc.get_process_dir()
        xdsinp = XDS_INP(wdir=wdir
                         ).load_from_file(from_dir=wdir,
                                          filename='XDS.INP')
        xdsinp.update('REFERENCE_DATA_SET={}'.format(reference_hkl))

        # XDS update resolution cutoff so to remove outshell junk
        highres = proc.get_resolution_estimate(target_name="I/SIGMA", target_value=0.25)
        if highres:
            xdsinp.update('INCLUDE_RESOLUTION_RANGE={} {}'.format(50.0, highres))
        else:
            logger.warning("rerun CORRECT with res cutoff, unable to extract highres estimate @ {}".format(wdir))
            xdsinp.update('INCLUDE_RESOLUTION_RANGE={} {}'.format(50.0, 3.5))

        # if the space group for this sweep is not the same as ref, use reference cell and spg
        ref_spg = ref_proc.correct_summary['spg']
        ref_cell = ref_proc.correct_summary['cell']

        if not proc.correct_summary:
            return None

        spg = proc.correct_summary['spg']
        if ref_spg != spg:
            logger.warning(
                "{}: space group chosen by pointless is different from spg of reference, "
                "use reference crystal parameters to rerun index to correct.".format(wdir))
            xdsinp.update('SPACE_GROUP_NUMBER={}'.format(ref_spg))
            xdsinp.update('UNIT_CELL_CONSTANTS={}'.format(ref_cell))

            highres_ref = ref_proc.correct_summary['estimated_resolution']
            xdsinp.update('INCLUDE_RESOLUTION_RANGE={} {}'.format(50.0, highres_ref))

            xds = XDS(wdir=wdir).set_xdsinp(xdsinp)
            xds.index()
            xds.integrate()
            xds.correct()
        else:
            xds = XDS(wdir=wdir).set_xdsinp(xdsinp)
            xds.correct()

        try:
            parser = CorrectParser(wdir=wdir)

            if parser.error:
                logger.error("CORRECT with reference @ {} failed.".format(wdir))
                return proc

            corr = parser.get_correlation_to_reference_hkl()
            if corr:
                nref = corr['common_reflections']
                cc = corr['correlation']
                logger.info("{} correlation to reference, CC= {} for {} reflections"
                            .format(wdir, cc, nref))

                if cc < .90:
                    logger.warning("correlation to reference is rather low, "
                                   "something may be wrong (e.g. spg assignment, radiation decay etc).")
        except Exception:
            logger.error("failed to obtain correlation between ref: {}.".format(wdir))

        return proc