from __future__ import absolute_import

import datetime
import logging
import os
from collections import OrderedDict, Counter
from shutil import copy2


from ..common.spacegroup import Spacegroup
from ..common.utils import rel_path
from ..file_parser.correctparser import CorrectParser
from ..file_parser.xdscc12parser import XDSCC12Parser
from ..file_parser.xscale_1_inp import XSCALE1INP
from ..tasks.scriptjob import ScriptJob

logger = logging.getLogger(__name__)


class INPUT_FILE(object):
    """ one XSCALE INPUT FILE line"""

    def __init__(self, xds_ascii_file, include_resolution_range=None, crystal_name=None, score=None):
        self.file = xds_ascii_file
        self.filepath = None
        self.filename = None
        if os.path.exists(self.file) and os.path.isabs(self.file):
            self.filepath, self.filename = os.path.split(xds_ascii_file)
        else:
            logger.warning(
                "failed to initialize INPUT_FILE, {} does not exists or not an absolute path".format(self.file))

        self.include_resolution_range = include_resolution_range
        self.crystal_name = crystal_name
        self.score = score

    def set_include_resolution_range(self, resolution_range):
        self.include_resolution_range = resolution_range

    def set_crystal_name(self, crystal_name):
        self.crystal_name = crystal_name

    def set_score(self, score):
        self.score = score

    def get_relative_path(self, wdir):
        # return file in relative path of wdir
        rel_file = rel_path(wdir, self.file)
        if len(rel_file) > 120:
            logger.warning("filename too long {}".format(self.file))
        return rel_file

    def guess_resolution_range(self, correct_dir, correct_lp="CORRECT.LP", target_name='CC(1/2)', target_value=15):
        cparser = CorrectParser(wdir=correct_dir, filename=correct_lp)
        highres = cparser.get_resolution_estimation(target_name=target_name, target_value=target_value)
        if highres:
            self.include_resolution_range = "50 {}".format(highres)
        else:
            self.include_resolution_range = "50 3.5"

    def __str__(self):
        return "file: {}, resolution_range: {}, score: {}".format(self.file, self.include_resolution_range, self.score)

    def __repr__(self):
        return self.__str__()


class XSCALE_INP(object):
    def __init__(self,
                 filename="XSCALE.INP",
                 anomalous=False,
                 merge=False,
                 wdir=".",
                 spacegroup=None,
                 unitcell=None,
                 nproc=16,
                 runner="slurm",
                 rejection_method=1,
                 minimum_isigma=None,
                 reference_hkl=None
                 ):
        self.wdir = wdir
        self.filename = filename
        self.spacegroup = spacegroup
        self.unitcell = unitcell
        self.nproc = nproc
        self.minimum_isigma = minimum_isigma
        self.anomalous = anomalous
        self.merge = merge
        self.reference_hkl = reference_hkl
        self.runner = runner
        self.rejection_method = rejection_method

        self.data = OrderedDict()
        self.xscale_inp = None
        self.n_outliers = 0

    def create_from_template(self, xscale_inp_file):
        # parse an XSCALE.INP into the XSCALE_INP object
        if not os.path.exists(xscale_inp_file):
            logger.error("{} does not exist.".format(xscale_inp_file))
            return
        else:
            print(xscale_inp_file)

        with open(xscale_inp_file, 'r') as fh:
            lines = fh.readlines()

        lines = [line.lstrip() for line in lines]
        # lines = [line for line in lines if not line.startswith("!") and len(line) > 10]

        for line in lines:
            if line.startswith("SPACE_GROUP_NUMBER="):
                self.spacegroup = line.split("=")[1].replace(" ", "")

            # assume at least one more character at the end of input, including \n
            if line.startswith("UNIT_CELL_CONSTANTS="):
                self.unitcell = line[line.find("=") + 1: line.find("!")].lstrip()

            if line.startswith("REFERENCE_DATA_SET="):
                self.spacegroup = line.split("=")[1].replace(" ", "")

            if not line.startswith("!") and "FRIEDEL'S_LAW=" in line:
                if line.split("FRIEDEL'S_LAW=")[-1].split()[0].upper() == "TRUE":
                    self.anomalous = False
                else:
                    self.anomalous = True

            if not line.startswith("!") and "MERGE=" in line:
                if line.split("MERGE=")[-1].split()[0].upper() == "TRUE":
                    self.merge = True
                else:
                    self.merge = False

            if line.startswith("OUTPUT_FILE="):
                outfile = line.split("=")[1].replace(" ", "")
                self.data.setdefault(outfile, [])

            if line.startswith("INPUT_FILE="):
                infile = line.split("=")[1].replace(" ", "")
                self.data[outfile].append(INPUT_FILE(infile))

            if "length=CC*,angle,cluster=" in line:
                angle = line.split("length=CC*,angle,cluster=")[-1].lstrip().split()[1]
                self.data[outfile][-1].set_score(angle)

            if line.startswith("INCLUDE_RESOLUTION_RANGE="):
                included_reso = line[line.find("=") + 1: line.find("!")].lstrip()
                self.data[outfile][-1].set_include_resolution_range(included_reso)

        return self

    def create_from_xds_objects(self, grouped_xds_objs):
        # create XSCALE.INP from XDS objects, assuming the objects are grouped already
        spgs = list()
        cells = list()
        for k, v in grouped_xds_objs.items():
            if k not in self.data:
                self.data.setdefault(k, [])

            for xds in v:
                proc_dir = xds.get_process_dir()
                ifile = INPUT_FILE(os.path.join(proc_dir, 'XDS_ASCII.HKL'))
                high_res = xds.correct_summary.get('estimated_resolution', None)
                spgs.append(xds.correct_summary.get('spg', None))
                cells.append(xds.correct_summary.get('cell', None))
                if high_res:
                    ifile.set_include_resolution_range("50 {}".format(high_res))
                self.data[k].append(ifile)

        self.spacegroup = Counter(spgs).most_common(1)[0][0]
        self.unitcell = ' '.join(str(round(x, 2)) for x in Spacegroup.cal_average_cell(cells))

        return self

    def add_output_file(self, outfile):
        if outfile not in self.data:
            self.data.setdefault(outfile, [])

    def add_inputfile(self, outfile, inputfile):
        if outfile not in self.data:
            self.add_output_file(outfile)

        self.data[outfile].append(inputfile)

    # remove an inputfile from output file group, i is the ith file to delete (starting from 1)
    def delete_input_file_by_number(self, outfile, ith_file):
        logger.info("{} was removed from XSCALE.".format(self.data[outfile][ith_file - 1]))
        new_inputfiles = [x for i, x in enumerate(self.data[outfile], 1) if i != ith_file]
        self.data[outfile] = new_inputfiles

    # remove several inputfiles from an output file group, i is the ith file to delete (starting from 1)
    def delete_input_file_by_numbers(self, outfile, numbers):
        logger.info(
            "{} was removed from XSCALE.".format([x.file for i, x in enumerate(self.data[outfile], 1) if i in numbers]))
        new_inputfiles = [x for i, x in enumerate(self.data[outfile], 1) if i not in numbers]
        self.data[outfile] = new_inputfiles

    def set_spacegroup(self, spacegroup):
        self.spacegroup = spacegroup

    def set_unitcell(self, unitcell):
        self.unitcell = unitcell

    def create_xscale_inp(self):
        xscale_inp = list()

        xscale_inp.append("MAXIMUM_NUMBER_OF_PROCESSORS= {}".format(self.nproc))
        xscale_inp.append("SAVE_CORRECTION_IMAGES=FALSE")
        xscale_inp.append("WFAC1=1.20")
        xscale_inp.append("PRINT_CORRELATIONS=FALSE")

        if self.minimum_isigma:
            xscale_inp.append("MINIMUM_I/SIGMA= {}".format(self.minimum_isigma))

        if self.spacegroup:
            xscale_inp.append("SPACE_GROUP_NUMBER= {}".format(self.spacegroup))

        if self.unitcell:
            xscale_inp.append("UNIT_CELL_CONSTANTS= {}".format(self.unitcell))

        if self.reference_hkl:
            xscale_inp.append("REFERENCE_DATA_SET= {}".format(self.reference_hkl))

        for outfile, inputfiles in self.data.items():
            if not inputfiles:
                continue

            xscale_inp.append("OUTPUT_FILE={}".format(outfile))
            if self.merge:
                xscale_inp.append("MERGE=TRUE")
            else:
                xscale_inp.append("MERGE=FALSE")

            if self.anomalous:
                xscale_inp.append("FRIEDEL'S_LAW=FALSE")
            else:
                xscale_inp.append("FRIEDEL'S_LAW=TRUE")

            for inputfile in inputfiles:
                xscale_inp.append("INPUT_FILE={}".format(inputfile.get_relative_path(self.wdir)))
                if inputfile.include_resolution_range:
                    xscale_inp.append("INCLUDE_RESOLUTION_RANGE={}".format(inputfile.include_resolution_range))

        self.xscale_inp = xscale_inp
        self.save(saveBackup=False)

    def save(self, sep="\n", saveBackup=False, dest_dir=None, dest_fname="XSCALE.INP"):
        if dest_dir and dest_fname:
            dest_dir = os.path.abspath(dest_dir)
            if not os.path.isdir(dest_dir):
                os.makedirs(dest_dir)

            fname = os.path.join(dest_dir, dest_fname)
        else:
            fname = os.path.join(self.wdir, self.filename)

        if os.path.exists(fname) and saveBackup:
            # check versions and make backups
            timeStamp = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
            newname = fname + "." + timeStamp
            os.rename(fname, newname)

        # print(sep.join(self.xscale_inp))
        with open(fname, "w") as fh:
            fh.write(sep.join(self.xscale_inp))

    def run_xscale(self):
        self.create_xscale_inp()
        xscale = ScriptJob(wdir=self.wdir,
                           script_name='xscale.sh',
                           script_text='xscale_par',
                           runner=self.runner)
        xscale.run_sync()

    def get_outfiles(self):
        return self.data.keys()

    def remove_outliers(self, icycle, max_resolution=None, use_xdscc12=False):
        if not max_resolution:
            max_resol = ""
        else:
            max_resol = "-dmax {}".format(max_resolution)

        self.n_outliers = 0
        for outfile, infiles in self.data.items():
            if use_xdscc12:
                # run xdscc12 to find outliers, only single wavelength case
                if os.path.exists(os.path.join(self.wdir, outfile)):
                    logger.info("running xdscc12 to identify outliers")
                    # backup xscalecc outfile
                    if os.path.exists(os.path.join(self.wdir, "xdscc12.sh.out")):
                        # t_stamp = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
                        t_stamp = "cycle{}".format(icycle)
                        copy2(os.path.join(self.wdir, "xdscc12.sh.out"), os.path.join(self.wdir, "xdscc12.sh.out_{}".format(t_stamp)))

                    ScriptJob(self.wdir,
                              script_name="xdscc12.sh",
                              script_text="xdscc12 -w {} {}".format(outfile, max_resol),
                              runner="local"
                              ).run_sync()

                # parse xdscc12.sh.out to identify outlier
                outliers = XDSCC12Parser(wdir=self.wdir,
                                         filename="xdscc12.sh.out"
                                         ).get_outliers_using_a(method=self.rejection_method)
            else:
                # use xscale_isocluster method
                if os.path.exists(os.path.join(self.wdir, outfile)):
                    logger.info("running xscale_isocluster to identify outliers")
                    # backup xscalecc outfile
                    if os.path.exists(os.path.join(self.wdir, "XSCALE.1.INP")):
                        # t_stamp = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
                        t_stamp = "cycle{}".format(icycle)
                        copy2(os.path.join(self.wdir, "XSCALE.1.INP"), os.path.join(self.wdir, "XSCALE.1.INP_{}".format(t_stamp)))
                        copy2(os.path.join(self.wdir, "XSCALE.INP"), os.path.join(self.wdir, "XSCALE.INP_{}".format(t_stamp)))
                        copy2(os.path.join(self.wdir, "XSCALE.LP"), os.path.join(self.wdir, "XSCALE.LP_{}".format(t_stamp)))
                        copy2(os.path.join(self.wdir, outfile), os.path.join(self.wdir, "{}_{}".format(outfile, t_stamp)))

                    ScriptJob(self.wdir,
                              script_name="xscale_isocluster.sh",
                              script_text="xscale_isocluster -w {} {}".format(outfile, max_resol),
                              runner="local"
                              ).run_sync()

                # parse XSCALE.INP.1 to identify outliers
                if os.path.exists(os.path.join(self.wdir, "XSCALE.1.INP")):
                    outliers = XSCALE1INP(wdir=self.wdir,
                                          filename="XSCALE.1.INP"
                                          ).get_outliers_using_angle(method='elbow', threshold=1.5)
                else:
                    logger.info("xscale_isocluster failed to generate XSCALE.1.INP")
                    outliers = []

            logger.info("outliers {} identified for {}".format(outliers, outfile))
            if not outliers:
                continue
            else:
                self.delete_input_file_by_numbers(outfile, outliers)

                # self.data[outfile] = [f for i,f in enumerate(self.data[outfile]) if i+1 not in outliers]
                # remove outfile group if nothing left
                if not self.data[outfile]:
                    del self.data[outfile]

            self.n_outliers += len(outliers)


if '__main__' == __name__:
    print("===")
    input1 = INPUT_FILE("/tmp/a.1")
    input2 = INPUT_FILE("/tmp/a.2")
    input3 = INPUT_FILE("/tmp/a.3")

    x = XSCALE_INP(wdir="/tmp")
    # x.set_spacegroup(5)
    # x.set_unitcell("20 20 20 90 90 90")
    # x.add_inputfile("a", input1)
    # x.add_inputfile("a", input2)
    # x.add_inputfile("a", input3)
    # print(x.data)
    # x.create_xscale_inp()

    x.create_from_template("/mnt/share/qxu/bluice/tools/scripts/quickProcess/quickProcess/tests/XSCALE.INP")
    x.create_xscale_inp()
    print(x.data)
