#!/usr/bin/perl -W
# _CALLED_FROM_BLUICE_
#--------------------------------------------------------------
## TEST:
# alignMiniBeam.pl autoPinholeOrGuardSlits bluice nognuplot nomedm warnSample fastScan centerV=Imax centerH=Imax useCRL
#--------------------------------------------------------------
# This is a minibeam alignment script aimed at threading the beam
# through the collimator by step-scanning mirror piezos (DACs) both
# horizontally and vertically.
#
# The algorithm (according to Bob Fischetti):
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#  0. Close guard slits or change collimator from 300um to 10um
#  1. Move collimator to "IN" position, raise diode support to
#     "IN" and lower beamstop to "OUT" (external script)
#  2. Pause intensity and positional feedback
#  3. Open shutters
#  4. Scan VFM Pz (DAC) and then move to maximum intensity (+-2V)
#  5. Scan HFM Pz (DAC) and then move to maximum intensity (+-2V)
#  6. Store new set points in the positional feedback
#  7. Start intensity and positional feedback
#  8. Close fast shutter
#  9. Raise beamstop to "IN" and lower diode support to "OUT" (external script)
# 10. Restore guard slits or change collimator from 300um to 10um
# 11. Done!
#
#			Author: Sergey Stepanov
#
# Version-1.00:  2007/06/08 using Pezca
# Version-2.00:  2007/07/05 added option to close guard slits before scanning
#                           (purpose: using the scipt with non-minibeam)
#                           added command line parameters to define H&V step/range
#                           added restoring the DACs and slits to the original
#                           state when script is interrupted.
# Version-3.00:  2007/11/15 moved Mostab/Postab related functions into mostab_subs.pl
# Version-3.01:  2007/11/21 added auto-selection of Pinhole vs Guard Slits depending
#                           on current beamline configuration.
# Version-3.02:  2007/11/27 added communication with BluIce
# Version-3.03:  2008/03/10 changed the way beam is aligned for 300um collimator:
#                           now a 10um cllimator is installed instead of closing
#                           the guard slits
# Version-4.00:  2008/09/30 on-the-fly scan option with Hytec DAC added
# Version-4.01:  2009/09/22 changed to work with collimatorCtrl script
# Version-4.02:  2009/11/13 added voltage corrections for on-the-fly alignment
# Version-4.03:  2009/11/30 added optional checking for sample (warnSample)
# Version-4.04:  2010/04/24 added check that endstation shutter really opened
#                           on open command.
# Version-4.05:  2010/06/15 added flag to disable restoring collimator/pinDiode
#                           (to use with measureMiniBeam.pl).
# Version-5.00:  2010/06/15 changed to work with "signal" (e.g Pin,Beamstop,I0)
#                           as a key to find Struck input and Keithley channel
#                           from mySQL; also switched to using intensity darks
#                           from mySQL (although they are not currently used!).
#                           This required replacing the i1 and (never unused!) i0
#                           cmdline arguments with signal=[Pin|Beamstop|I0].
# Version-5.01:  2010/07/07 added displaTime and optimizeGain
# Version-5.02:  2010/07/13 added optional saving of files
# Version-6.00:  2011/01/07 added global flag for opening popups via BluIce,
#                           added multiple runs when Keithley overloads.
# Version-6.01:  2011/01/18 added restoring intensity feedback if script fails.
# Version-6.02:  2011/11/09 added final gnuplot plot for fast scanning.
# Version-6.03:  2012/03/19 added saving/restoring guard slits if collimatorCtrl
#                           script was called: it could change guard slits when
#                           moving collimator-IN.
#                           ATTENTION: since aligning shifts the beam, restored
#                           guard slits may not be centered.
# Version-7.00:  2012/05/04 added posting scan arrays to JBluIce. That required
#                           also changing &stepScanDAC in Common/dac_subs.pl and
#                           &SRstepScanDAC and &SRflyScanDAC in Common/scanrec_subs.pl
# Version-7.01:  2012/02/01 added HalfRangeVfactor=xx, HalfRangeHfactor=xx and stepScan options
# Version-7.02:  2013/04/24 changed to work with sampleEnvSrv
# Version-7.03:  2013/05/25 changed to work with $SRSCAN_FLAGS
# Version-7.04:  2013/09/01 added ADC readback "encoder" to DAC scans
# Version-7.05:  2014/03/28 changed functions calls to arrays pointers (preparation to Perl-5.16)
# Version-7.06:  2014/06/13 added popup when sample can be placed back
# Version-7.07:  2014/08/20 fixed a bug in re-checking intensity (needed to open shutter),
#                           re-enabled fly scanning, reduced default steps for fly scanning 2x,
#                           tested with Oleg's new "soft" fly scanning DAC driver
# Version-7.08:  2014/11/12 added option to skip pausing intensity feedback.
#                           added re-measurement of intensities after the feedback turned on.
# Version-7.09:  2014/11/17 added option to move sample (moveSample=...) instead of warning
# Version-7.10:  2015/08/14 added wait and error report for restoring collimator on exit
# Version-8.00:  2016/07/12 added using hardware feedback for vertical collimation (instead of scan);
#                           added saving keithley gain and intensity on successful centering
# Version-8.01:  2016/07/14 added last used energy and attenuation to the last alignment data
# Version-8.02:  2017/10/18 converted fly-scan corrections into command line parameters
# Version-8.03:  2017/10/19 converted pin sensor readout into robot_subs functions
# Version-8.04:  2017/11/01 added slope for compensating for hysteresis
# Version-8.05:  2018/11/08 added options to center on max intensity instead of FWHM
# Version-9.00:  2018/11/14 added options to center on gaussian and lorentzian peaks (via PDL functions)
# Version-9.01   2019/02/11 added beam centering with CRLs small XY positioner instead of mirrors
# Version-9.02   2019/12/13 added saving gnuplot as image
# Version-9.03   2020/10/01 added beam centering with Transfocator vertical & lateral positioners instead of mirrors
# Version-9.04   2020/10/26 changed default sample removal to up and allowed any direction
# Version-9.05   2020/10/27 changed sample auto-removal from the point&click server request to direct moving
# Version-9.06   2020/11/17 added saving transfocator positions into a PV after aligning with transfocator (see &saveTFstate())
# Version-9.07   2021/10/21 added post-aligning transfocator after scanning; added measuring elapsed time.
# Version-9.08   2022/02/22 added handling initial collimator as "None"
# Version-9.09   2022/12/12 added KBM piezo encoder, if present
#--------------------------------------------------------------
# Subroutines in this file:
#             sub scan_FPGA_V() {
#             sub scan_KBM_V() {
#             sub scan_KBM_H()
#             sub scan_CRL_V() {
#             sub scan_CRL_H()
#             sub scan_TF_V() {
#             sub scan_TF_H() {
#             sub run_guard_slits ($$;$) {
#             sub save_last_alignment ($) {
#             sub read_last_alignment ($) {
#             sub saveData($$$$$$$$$) {
#             sub tellBluIceBeforeExit($;$) {
#             sub optimizePos($$$$$$;$) {
#             sub signal_handler ($;$) {
#--------------------------------------------------------------
# use List::Util qw(min max);
# use Cwd;
  use DBI;						# mySQL DB interface (for signal darks)
  use Pezca;
  use PDL;						# Perl data language (added for 9.00)
  use PDL::Fit::LM;					# Levenberg-Marquardt fitting (added for 9.00)
  use Time::HiRes qw(gettimeofday tv_interval);		# needed for measuring total time
# use warnings::unused;
# use Test::Vars;
  use strict;
  use warnings;

  our ($NONINTERACTIVE, $IGNORE_ERRORS, $ATTENUATION_USE_FACTOR);
  our ($USE_ZENITY_FOR_EXIT, $ZENITY_VIA_BLUICE, $JBLUICE_SERVER);
  our ($beamline, $joergerName, $shutter, $shutter_readback);
  our ($struckName, $struck_ext, @StruckOffsets);
  our ($have_KBM, $have_TF, $have_CRL);
  our ($HAVE_XYZ, $scanrecName, $scanrecAbort);
  our ($gnuopen, @Offsets);
  our (@x1DAC, @y1DAC);
  our (@x1SR,  @y1SR);
  our (@x1HD,  @y1HD);
  our ($SRSCAN_FLAGS, $HDSCAN_FLAGS);			# reference to a hash with various flags for SRSCAN & HDSCAN
  my ($PEZCA, $script, $scr, $syntax, $narg, @beamlinelist);
  my ($bl_specified, $au_specified, $gs_specified, $i1_specified);
  my ($ct_specified, $vr_specified, $vs_specified, $hr_specified);
  my ($hs_specified, $co_specified, $rc_specified, $pd_specified);
  my ($tb_specified, $fs_specified, $md_specified, $ps_specified);
  my ($pg_specified, $sv_specified, $ro_specified, $gn_specified);
  my ($en_specified, $tf_specified, $im_specified, $ip_specified);
  my ($hc_specified, $vc_specified, $hf_specified);
  my ($ch_specified, $cv_specified, $center_H, $center_V);
  my ($i, $arg, $i_min, $i_max, $cmd, $prefix);
  my ($auto_choice, $usegs, $align300to10, $gssize, $i1_name, $i1);
  my (@signals, $cnt_time, @ct, $minute, $hour, $mday, $mon, $year);
  my ($scan_pmrangeH, $scan_factorH, $scan_stepH);
  my ($scan_pmrangeV, $scan_factorV, $scan_stepV);
  my ($move_collimator, $restore_collimator, $move_pindiode, $tell_bluice);
  my (@inputs, $availableH, $availableV, $overload_rescan);
  my ($setpointH, $setpointV, $interfaceMostab, $interfacePostab);
  my (@prg, $wrkdir, $scanprg, $cnt_factor, $zzz, $pos);
  my ($scalerNo, $fastScan, $gnuplot, $logscale, @gnuopts, $gnufile, $medm);
  my ($dac, $dacHFM_PV, $dacVFM_PV, $postabActH_PV, $postabActV_PV);
  my ($gsSizeH_PV, $gsAssyH_PV, $gsSizeV_PV, $gsAssyV_PV, $tmp);
  my ($colAssy_PV, $activeCol_PV, $ninputs, $scaler, $fileBkg, $afterScan);
  my ($ActH, $ActV, $scanBegin, $scanEnd, $LoH, $HiH, $LoV, $HiV);
  my ($status, $collimatorPreset, $gsSizeH_initial, $gsSizeV_initial);
  my ($FinalH, $FinalV, $npts, $mostab_initial, $postab_initial, $scanTime);
  my (@I1_max_now, $I1_max_final_V, $I1_max_final_H, $hysteresis, $delay, $inlimits);
  my ($I1_max_scan_V, $I1_max_scan_H, $I1_percent_V, $I1_percent_H);
  my ($pinsensor_PV, $pin_mounted, $msg, $displaytime, $dt_specified);
  my (@preset, $state10um, $state300um, $stateOUT, @mbbstates, $kgain_divider);
  my ($corr_fly_V, $corr_fly_H, $str, $preset_gain, $I_optimum, $kgain, $I1);
  my ($timestampDB, $epochDB, @darks, $klabelDB, $koffsetDB, $kchannel);
  my ($useDarks, $klabel, $koffset, $epoch, $bkg_age_hours, $offset_delta);
  my ($I_lo, $I_hi, $bipostHPV, $bipostVPV, $bialignStatePV, $OVERLOAD, $FOLLERR);
  my ($adc, $adcHFM_PV, $adcVFM_PV, $use_encoder, $tolerance_factor, $nscans);
  my ($tryVerticalHWfeedback, $verticalHWfeedbackSucceeded, $alignHistPV);
  my ($I1_FPGA_before, $I1_FPGA_after, $energyPV, $energy, $attn);
  my ($date_last, $energy_last, $attn_last, $i1_name_last, $kgain_last, $scantype_last);
  my ($center_last_V, $I1_last_scan_V, $I1_last_final_V, $slope_V);
  my ($center_last_H, $I1_last_scan_H, $I1_last_final_H, $slope_H);
  my ($use_CRL, $crl_specified, $crl_H_PV, $crl_V_PV, $crl_assy_PV);
  my ($use_TF, $tf_H_assy_PV, $tf_V_axis_PV, $tf_V_assy_PV, $tf_H_axis_PV);
  my ($cts2egu, $formula, $mtrOut, $struckPs, $struckNg, $struck, $png, $label);
  my ($use_KBM, $shutterAssy_PV, $protectSample, $moveSample, $sample_moved);
  my ($xyzAssyPV, $zAssyPV, @xyzAxesPV, @xyzIni, $nxyz, $direction);
  my ($omegaAxisPV, $omega, $degree2rad, $t0, $elapsed);
  my ($npts_max, $npts_max2, $scanrec, $scanAbort);
  my ($kbm_encoder_present, $encoder_array_name, $index_peak);
  my ($hfm_encoder_PV, $hfm_enc_scale, $encoder_scanpeak);
  my ($vfm_encoder_PV, $vfm_enc_scale, $encoder_finalpeak);

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;

  do $PEZCA.'/Common/motor_subs.pl';
  do $PEZCA.'/Common/joerger_subs_v2.pl';
  do $PEZCA.'/Common/dac_subs.pl';
  do $PEZCA.'/Common/scanrec_subs.pl';
  do $PEZCA.'/Common/keithley_subs.pl';
  do $PEZCA.'/Common/gnuplot_subs.pl';
  do $PEZCA.'/Common/shutter_subs.pl';
  do $PEZCA.'/Common/mostab_subs.pl';
  do $PEZCA.'/Common/attenuation_subs_v2.pl';
  do $PEZCA.'/Common/struck_subs_v2.pl';
  do $PEZCA.'/Common/crl_subs.pl';
  do $PEZCA.'/Common/robot_subs.pl';
  do $PEZCA.'/Common/common_PDL.pl';    # added for 9.00 (gaussian & loretzian peak fitting)
  do $PEZCA.'/Common/common_subs.pl';
  do $PEZCA.'/Common/beamline.pl';

  select STDOUT; $|=1;			# set unbuffered output
  select STDERR; $|=1;			# set unbuffered output

### Presume NONinteractive until we know
### if we are called from BluIce or not:
  $NONINTERACTIVE = 1;	 		# to exit from script with no prompting
  $USE_ZENITY_FOR_EXIT = 1;		# report fatal/exit errors via Zenity
# $IGNORE_ERRORS  = 1;			# ignore EZCA errors
  $ATTENUATION_USE_FACTOR=1;

  $script = &makeScriptName();
  $syntax = "\n\n".'Syntax: '.$script
          . ' [beamline]'                                                            ."\n"
          . ' [autoPinholeOrGuardSlits|align300to10|useGuardSlits[=xx]]'             ."\n"
          . ' [noKBM|useKBM] [noCRL|useCRL] [noTF|useTF]'                            ."\n"
          . ' [signal=Pin|I0|Beamstop] [countTime=xx]'                               ."\n"
          . ' [StepV=xx] [HalfRangeV=xx|HalfRangeVfactor=xx]'                        ."\n"
          . ' [StepH=xx] [HalfRangeH=xx|HalfRangeHfactor=xx]'                        ."\n"
          . ' [flyscanCorrectionH=xx] [flyscanCorrectionV=xx]'                       ."\n"
          . ' [centerH=FWHM|FWHMsmooth|Imax|Gauss|Lorentz|centroid]'                 ."\n"
          . ' [centerV=FWHM|FWHMsmooth|Imax|Gauss|Lorentz|centroid]'                 ."\n"
          . ' [noCollimatorMove] [noCollimatorRestore] [noPindiodeMove]'             ."\n"
          . ' [ignoreSample|warnSample|moveSample=xx[h|v]|moveSampleNoWarn=xx[h|v]]' ."\n"
          . ' [fastScan|stepScan] [tryVerticalHWfeedback]'                           ."\n"
          . ' [tellBluIce|bluice|nobluice] [medm|nomedm]'                            ."\n"
          . ' [gnuplot|gnuplot=image|nognuplot] [displayTime=nn]'                    ."\n"
          . ' [optimizeGain[=force|=auto]|skipGain]'                                 ."\n"
          . ' [useEncoder|noEncoder] [toleranceFactor=xx]'                           ."\n"
          . ' [pauseMostab|skipMostab] [pausePostab|skipPostab]'                     ."\n"
          . ' [save=prefix] [help|?]'                                                ."\n"
          .                                                                           "\n";
  $script  =~ s/\.pl$//i;

  $scr           = 'Align';
  @beamlinelist  = ('23i:', '23o:', '23b:', '23d:');
  $i_min         = 2;
  $i_max         = 32;
  $kgain_divider = 100.;							#keithley gain divider in case of overflow; must be 10,100, or 1000
  $gnufile       = 'gnu.tmp';
  $narg          = @ARGV;
  print STDOUT "\n".'Date: '.scalar(localtime)."\n";
  if ($narg > 0) {print STDOUT $script.': started with cmdline = '."@ARGV\n\n";}
  else           {print STDOUT $script.': started with no cmdline arguments.'."\n\n";}
  @ct = localtime(time);
  ($minute,$hour,$mday,$mon,$year) = ($ct[1],$ct[2],$ct[3],$ct[4]+1,$ct[5]+1900);

  foreach $arg (@ARGV) {
     $arg =~ s/^-{0,2}//;							#trim leading dashes
     if ($arg =~ /^help$/i || $arg =~ /^\?$/) {
        &press_key_exit($syntax);						#this will exit automatically if $NONINTERACTIVE=1
     }
     elsif (grep(/^$arg:?$/, @beamlinelist)) {
        if (defined $bl_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $bl_specified = 1;
        $arg =~ s/:$//;								#trim ':'
        if ($arg ne $beamline) {
           print STDOUT "\n".'!!! '.$script.': Cross-beamline operation:'
                        .' controlling '.$arg.' from '.$beamline."\n\n";
           $beamline = $arg;
           do $PEZCA.'/Common/beamline_'.$beamline.'.pl';
        }
     }
     elsif ($arg =~ /^useKBM$/i || $arg =~ /^noKBM$/i ||
            $arg =~ /^useCRL$/i || $arg =~ /^noCRL$/i ||
            $arg =~ /^useTF$/i  || $arg =~ /^noTF$/i) {
        if (defined $crl_specified) {
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $crl_specified = 1;
        if    ($arg =~ /^useKBM/i)   {$use_KBM=1; $use_TF=0;  $use_CRL=0;}
        elsif ($arg =~ /^noKBM$/i)   {$use_KBM=0;}
        elsif ($arg =~ /^useCRL/i)   {$use_CRL=1; $use_TF=0;  $use_KBM=0;}
        elsif ($arg =~ /^noCRL$/i)   {$use_CRL=0;}
        elsif ($arg =~ /^useTF/i)    {$use_TF=1;  $use_CRL=0; $use_KBM=0;}
        elsif ($arg =~ /^noTF$/i)    {$use_TF=0;}
     }
     elsif ($arg =~ /autopinholeorguardslits$/i) {
        if (defined $au_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $au_specified = 1;
        $auto_choice = 1;
     }
     elsif ($arg =~ /^useGuardSlits/i) {
        if (defined $gs_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $gs_specified = 1;
        $usegs = 1;
        $align300to10 = 0;
        if ($arg =~ /^useGuardSlits=/i) {
           $gssize = $arg;
           $gssize =~ s/^useGuardSlits=//i;
           if (! &is_float($gssize) || $gssize <= 0) {				#this will exit automatically if $NONINTERACTIVE=1
              &press_key_exit('*** '.$script.': Incorrect cmdline: Guard'
                             .' Slits size must be a positive nimber.'.$syntax);
           }
        }
     }
     elsif ($arg =~ /^align300to10/i) {
        if (defined $gs_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $gs_specified = 1;
        $usegs = 0;
        $align300to10 = 1;
     }
     elsif ($arg =~ /^signal=/i) {
        if (defined $i1_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $i1_specified = 1;
        $i1_name = $arg;
        $i1_name =~ s/^signal=//i;
     }
     elsif ($arg =~ /^countTime=/i) {
        if (defined $ct_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $ct_specified = 1;
        $cnt_time = $arg;
        $cnt_time =~ s/^countTime=//i;
        if (! &is_float($cnt_time) || $cnt_time <= 0) {				#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Incorrect cmdline: countTime'
                          .' must be a positive number (in seconds)'.$syntax);
        }
     }
     elsif ($arg =~ /^stepH=/i) {
        if (defined $hs_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $hs_specified = 1;
        $scan_stepH = $arg;
        $scan_stepH =~ s/^stepH=//i;
        if (! &is_float($scan_stepH)) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Incorrect cmdline: StepH'
                          .' must be a number.'.$syntax);
        }
     }
     elsif ($arg =~ /^stepV=/i) {
        if (defined $vs_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $vs_specified = 1;
        $scan_stepV = $arg;
        $scan_stepV =~ s/^stepV=//i;
        if (! &is_float($scan_stepV)) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Incorrect cmdline: StepV'
                          .' must be a number.'.$syntax);
        }
     }
     elsif ($arg =~ /^halfRangeH=/i || $arg =~ /^halfRangeHfactor=/i) {
        if (defined $hr_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $hr_specified = 1;
        if ($arg =~ /^halfRangeH=/i) {
           $scan_pmrangeH = $arg;
           $scan_pmrangeH =~ s/^halfRangeH=//i;
           if (! &is_float($scan_pmrangeH) || $scan_pmrangeH <= 0) {		#this will exit automatically if $NONINTERACTIVE=1
              &press_key_exit('*** '.$script.': Incorrect cmdline:'
                             .' HalfRangeH must be a positive number.'.$syntax);
           }
        }
        elsif ($arg =~ /^halfRangeHfactor=/i) {
           $scan_factorH = $arg;
           $scan_factorH =~ s/^halfRangeHfactor=//i;
           if (! &is_float($scan_factorH) || $scan_factorH <= 0) {		#this will exit automatically if $NONINTERACTIVE=1
              &press_key_exit('*** '.$script.': Incorrect cmdline:'
                             .' HalfRangeHfactor must be a positive number.'.$syntax);
           }
        }
     }
     elsif ($arg =~ /^halfRangeV=/i || $arg =~ /^halfRangeVfactor=/i) {
        if (defined $vr_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $vr_specified = 1;
        if ($arg =~ /^halfRangeV=/i) {
           $scan_pmrangeV = $arg;
           $scan_pmrangeV =~ s/^halfRangeV=//i;
           if (! &is_float($scan_pmrangeV) || $scan_pmrangeV <= 0) {		#this will exit automatically if $NONINTERACTIVE=1
              &press_key_exit('*** '.$script.': Incorrect cmdline:'
                             .' HalfRangeV must be a positive number.'.$syntax);
           }
        }
        elsif ($arg =~ /^halfRangeVfactor=/i) {
           $scan_factorV = $arg;
           $scan_factorV =~ s/^halfRangeVfactor=//i;
           if (! &is_float($scan_factorV) || $scan_factorV <= 0) {		#this will exit automatically if $NONINTERACTIVE=1
              &press_key_exit('*** '.$script.': Incorrect cmdline:'
                             .' HalfRangeVfactor must be a positive number.'.$syntax);
           }
        }
     }
     elsif ($arg =~ /^flyscanCorrectionH=/i) {
        if (defined $hc_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $hc_specified = 1;
        $corr_fly_H = $arg;
        $corr_fly_H =~ s/^flyscanCorrectionH=//i;
        if (! &is_float($corr_fly_H)) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Incorrect cmdline:'
                          .' flyscanCorrectionH must be a number.'.$syntax);
        }
     }
     elsif ($arg =~ /^flyscanCorrectionV=/i) {
        if (defined $vc_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $vc_specified = 1;
        $corr_fly_V = $arg;
        $corr_fly_V =~ s/^flyscanCorrectionV=//i;
        if (! &is_float($corr_fly_V)) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Incorrect cmdline:'
                          .' flyscanCorrectionV must be a number.'.$syntax);
        }
     }
     elsif ($arg =~ /^centerH=/i) {
        if (defined $ch_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $ch_specified = 1;
        $center_H = $arg;
        $center_H =~ s/^centerH=//i;
        if ($center_H !~ /^FWHM$/i && 
            $center_H !~ /^FWHMsmooth$/i && 
            $center_H !~ /^Imax$/i &&
            $center_H !~ /^centroid$/i &&
            $center_H !~ /^Gauss$/i &&
            $center_H !~ /^Lorentz$/i) {			                #this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Incorrect cmdline:'
                          .' centerH must be FWHM, FWHMsmooth, Imax, Gauss, or Lorentz.'.$syntax);
        }
        elsif ($center_H =~ /^FWHM$/i)       {$center_H = 'FWHM';}
        elsif ($center_H =~ /^FWHMsmooth$/i) {$center_H = 'FWHMsmooth';}
        elsif ($center_H =~ /^Imax$/i)       {$center_H = 'Imax';}
        elsif ($center_H =~ /^centroid$/i)   {$center_H = 'centroid';}
        elsif ($center_H =~ /^Gauss$/i)      {$center_H = 'Gauss';}
        elsif ($center_H =~ /^Lorentz$/i)    {$center_H = 'Lorentz';}
     }
     elsif ($arg =~ /^centerV=/i) {
        if (defined $cv_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $cv_specified = 1;
        $center_V = $arg;
        $center_V =~ s/^centerV=//i;
        if ($center_V !~ /^FWHM$/i && 
            $center_V !~ /^FWHMsmooth$/i && 
            $center_V !~ /^Imax$/i &&
            $center_V !~ /^centroid$/i &&
            $center_V !~ /^Gauss$/i &&
            $center_V !~ /^Lorentz$/i) {			                #this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Incorrect cmdline:'
                          .' centerV must be FWHM, FWHMsmooth, Imax, Gauss, or Lorentz.'.$syntax);
        }
        elsif ($center_V =~ /^FWHM$/i)       {$center_V = 'FWHM';}
        elsif ($center_V =~ /^FWHMsmooth$/i) {$center_V = 'FWHMsmooth';}
        elsif ($center_V =~ /^Imax$/i)       {$center_V = 'Imax';}
        elsif ($center_V =~ /^centroid$/i)   {$center_V = 'centroid';}
        elsif ($center_V =~ /^Gauss$/i)      {$center_V = 'Gauss';}
        elsif ($center_V =~ /^Lorentz$/i)    {$center_V = 'Lorentz';}
     }
     elsif ($arg =~ /^displayTime=/i) {
        if (defined $dt_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $dt_specified = 1;
        $displaytime = $arg;
        $displaytime =~ s/^displayTime=//i;
        if (! &is_int($displaytime) || $displaytime < 0) {			#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Incorrect cmdline: displayTimeout'
                           .' must be positive integer in seconds.'.$syntax);
        }
     }
     elsif ($arg =~ /^noCollimatorMove$/i) {
        if (defined $co_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $co_specified = 1;
        $move_collimator = 0;
     }
     elsif ($arg =~ /^noCollimatorRestore$/i) {
        if (defined $rc_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $rc_specified = 1;
        $restore_collimator = 0;
     }
     elsif ($arg =~ /^noPindiodeMove$/i) {
        if (defined $pd_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $pd_specified = 1;
        $move_pindiode = 0;
     }
     elsif ($arg =~ /^tellBluice$/i || 						#older version (compatibility)
            $arg =~ /^bluice$/i     ||
            $arg =~ /^nobluice$/i) {
        if (defined $tb_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $tb_specified = 1;
        if    ($arg =~ /^tellBluice$/i) {$tell_bluice = 1;}
        elsif ($arg =~ /^bluice$/i)     {$tell_bluice = 1;}
        elsif ($arg =~ /^nobluice$/i)   {$tell_bluice = 0;}
     }
     elsif ($arg =~ /^fastscan$/i || $arg =~ /^stepscan$/i) {
        if (defined $fs_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $fs_specified = 1;
        if    ($arg =~ /^fastscan$/i) {$fastScan = 1;}
        elsif ($arg =~ /^stepscan$/i) {$fastScan = 0;}
     }
     elsif ($arg =~ /^ignoreSample$/i || 
            $arg =~ /^warnSample$/i   || 
            $arg =~ /^moveSample=/i   ||
            $arg =~ /^moveSampleNoWarn=/i) {
        if (defined $ps_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $ps_specified = 1;
        if ($arg =~ /^ignoreSample$/i) {
           $protectSample = 0;						#ignore sample
           $moveSample    = 0;
        } 
        elsif ($arg =~ /^warnSample$/i) {
           $protectSample = 1;						#warn that sample needs to be moved
           $moveSample    = 0;
        } 
        elsif ($arg =~ /^moveSample=/i) {
           $protectSample = 2;						#move sample automatically
           $moveSample = lc($arg);
           $moveSample =~ s/^moveSample=//i;
        }
        elsif ($arg =~ /^moveSampleNoWarn=/i) {
           $protectSample = 3;						#move sample automatically without asking
           $moveSample = lc($arg);
           $moveSample =~ s/^moveSampleNoWarn=//i;
        }
        else {
           print STDOUT '!!! '.$script.': unexpected sample protection option! Ignoring.'."\n\n";
           $protectSample = 1;						#warn that sample needs to be moved
           $moveSample    = 0;
        }
        if ($protectSample > 1) {
           if    ($moveSample =~ /h$/i) {$direction = 'OUTBOARD'; chop $moveSample;}
           elsif ($moveSample =~ /v$/i) {$direction = 'UP';       chop $moveSample;}
           else                         {$direction = 'UP';}
           if (! &is_float($moveSample) || $moveSample == 0 || abs($moveSample) > 50) {	#this will exit automatically if $NONINTERACTIVE=1
              &press_key_exit('*** '.$script.': Incorrect cmdline: moveSample'
                                    .' must specify a positive number in mm.'
                                    .' It cannot exceed 50mm.'.$syntax);
           }
           if    ($direction eq 'UP'       && $moveSample < 0) {$direction = 'DOWN';}
           elsif ($direction eq 'OUTBOARD' && $moveSample < 0) {$direction = 'INBOARD';}
        }
     }
     elsif ($arg =~ /^medm$/i || $arg =~ /^nomedm$/i) {
        if (defined $md_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $md_specified = 1;
        if ($arg =~ /^medm$/i) {$medm = 1;}
        else                   {$medm = 0;}
     }
     elsif ($arg =~ /^gnuplot$/i || $arg =~ /^gnuplot=image$/i || $arg =~ /^nognuplot$/i) {
        if (defined $gn_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $gn_specified = 1;
        if    ($arg =~ /^gnuplot=image$/i) {$gnuplot = 2;}
        elsif ($arg =~ /^gnuplot$/i)       {$gnuplot = 1;}
        else                               {$gnuplot = 0;}
     }
     elsif ($arg =~ /^optimizeGain$/i ||
            $arg =~ /^optimizeGain=yes$/i ||
            $arg =~ /^optimizeGain=auto$/i ||
            $arg =~ /^optimizeGain=force$/i ||
            $arg =~ /^skipGain$/i) {
        if (defined $pg_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $pg_specified = 1;
        if    ($arg =~ /^optimizeGain=force$/i) {$preset_gain = 2;}
        elsif ($arg =~ /^optimizeGain/i)        {$preset_gain = 1;}		#this covers yes/auto/nokeyword
        else                                    {$preset_gain = 0;}
     }
     elsif ($arg =~ /^rescanWhenOverload$/i || $arg =~ /^ignoreOverload$/i) {
        if (defined $ro_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $ro_specified = 1;
        if ($arg =~ /^rescanWhenOverload$/i) {$overload_rescan = 1;}
        else                                 {$overload_rescan = 0;}
     }
     elsif ($arg =~ /^save=/i) {
        if (defined $sv_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $sv_specified = 1;
        $prefix = $arg;
        $prefix =~ s/^save=//i;
     }
     elsif ($arg =~ /^useEncoder$/i || $arg =~ /^noEncoder$/i) {
        if (defined $en_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                           .$arg.'"'.$syntax);
        }
        $en_specified = 1;
        if ($arg =~ /^useEncoder$/i) {$use_encoder = 1;}
        else                         {$use_encoder = 0;}
     }
     elsif ($arg =~ /^toleranceFactor=/i) {
        if (defined $tf_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                           .$arg.'"'.$syntax);
        }
        $tf_specified = 1;
        $tolerance_factor = $arg;
        $tolerance_factor =~ s/^toleranceFactor=//i;
        if (! &is_float($tolerance_factor) || $tolerance_factor <= 0) {		#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Incorrect cmdline: toleranceFactor'
                          .' must be a positive number.'.$syntax);
        }
     }
     elsif ($arg =~ /^pauseMostab$/i ||
            $arg =~ /^skipMostab$/i) {
        if (defined $im_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $im_specified = 1;
        if ($arg =~ /^pauseMostab$/i) {$interfaceMostab = 1;}
        else                          {$interfaceMostab = 0;}
     }
     elsif ($arg =~ /^pausePostab$/i ||
            $arg =~ /^skipPostab$/i) {
        if (defined $ip_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $ip_specified = 1;
        if ($arg =~ /^pausePostab$/i) {$interfacePostab = 1;}
        else                          {$interfacePostab = 0;}
     }
     elsif ($arg =~ /^tryVerticalHWfeedback$/i) {
        if (defined $hf_specified) {						#this will exit automatically if $NONINTERACTIVE=1
           &press_key_exit('*** '.$script.': Duplicate cmdline parameter "'
                          .$arg.'"'.$syntax);
        }
        $hf_specified = 1;
        $tryVerticalHWfeedback = 1;
     }
     else {									#this will exit automatically if $NONINTERACTIVE=1
        &press_key_exit('*** '.$script.': Unrecognized cmdline parameter "'
                       .$arg.'"'.$syntax);
     }
  }										#end of loop over cmdline arguments

  if (!defined $prefix) {
     $prefix = '/mnt/software/gmca_shared/logfiles/';
     if (! -e $prefix) {$prefix = './';}
  }
  if ($prefix =~ /\/$/) {							#if prefix is a directory
### Auto-construct the name:
     $prefix .= sprintf $beamline.'_%04d-%02d-%02d_%02d-%02d_align',$year,$mon,$mday,$hour,$minute;
  }

### =============================================================
### Configuration parameter -- change if beamline hardware changes:
  if (!defined $have_KBM) {$have_KBM = 1;}					#see beamline.pl
  if (!defined $have_CRL) {$have_CRL = 0;}					#see beamline.pl
  if (!defined $have_TF)  {$have_TF  = 0;}					#see beamline.pl

### Defaults (when no cmdline parameter is specified):
  if (!defined $use_KBM) {
     if ($have_KBM) {$use_KBM=1;} else {$use_KBM=0;}                            #scan with mirror piezos
  }
  if (!defined $use_TF) {
     if (!$use_KBM && $have_TF)  {$use_TF=1; $use_CRL=0;} else {$use_TF=0;}     #scan with TF
  }
  if (!defined $use_CRL) {
     if (!$use_KBM && $have_CRL) {$use_CRL=1; $use_TF=0;} else {$use_CRL=0;}    #scan with CRL
  }
  if (!defined $fastScan)             {$fastScan              = 1;}		#dont use on-the-fly scan
  if ($fastScan) {
#    print STDOUT '!!! '.$script.': switching to temporary hardcoded step scans option - questions to Sergey!'."\n\n";
#    $fastScan = 0;
  }
  if (!defined $auto_choice)          {$auto_choice           = 0;}		#option to autoselect small collimator vs Guard Slits
  if (!defined $usegs)                {$usegs                 = 0;} 		#option to work with Guard Slits
  if (!defined $align300to10)         {$align300to10          = 0;}
  if (!defined $gssize)               {$gssize                = 0.025;}		#Guard Slits size to use with the script
  if (!defined $center_H)             {$center_H              = 'FWHM';}        #center on "noisy peak" (search peak from tails)
  if (!defined $center_V)             {$center_V              = 'FWHM';}        #center on "noisy peak" (search peak from tails)
  if (!defined $cnt_time)             {$cnt_time              = 0.1;}		#scaler count time (s)
  if (!defined $move_collimator)      {$move_collimator       = 1;}		#bring collimator IN
  if (!defined $restore_collimator)   {$restore_collimator    = 1;}		#restore collimator [and pinDiode] at the end
  if (!defined $move_pindiode)        {$move_pindiode         = 1;}		#bring pindiode IN
  if (!defined $tell_bluice)          {$tell_bluice           = 0;}		#dont inform BluIce about progress
  if (!defined $protectSample)        {$protectSample         = 0;}		#dont check/warn for sample on gonio
  if (!defined $moveSample)           {$moveSample            = 0;}		#horizontal move of sample on the gonio
  if (!defined $direction)            {$direction             = 'UP';}		#UP, DOWN, OUTBOARD, INBOARD
  if (!defined $medm)                 {$medm                  = 1;}		#show scan MEDM screens (on-the-fly scan only)
  if (!defined $gnuplot)              {$gnuplot               = 1;}		#show plots with Gnuplot
# if (!defined $gnuplot)              {$gnuplot               = 2;}		#show plots with Gnuplot and save the plot as image
  if (!defined $preset_gain)          {$preset_gain           = 1;}		#by default pre-optimize Keithley gain on pin diode
  if (!defined $overload_rescan)      {$overload_rescan       = 1;}		#by default reduce gain and rescan on overload
  if (!defined $interfaceMostab)      {$interfaceMostab       = 0;}		#added 2014/11/12
  if (!defined $interfacePostab)      {$interfacePostab       = 0;}		#added 2014/11/12
  if (!defined $tryVerticalHWfeedback){$tryVerticalHWfeedback = 0;}		#added 2016/07/12
  if (!defined $displaytime)          {$displaytime           = 60;}		#how long to display the plot
  if ($tell_bluice)                   {$ZENITY_VIA_BLUICE     = 1;}
  else                                {$NONINTERACTIVE        = 0;}
  if (!defined $i1_name) {
      ### QUESTION: should we use beamstop instead of I0
      ### when we do not place the PinDiode IN ???
      if ($move_pindiode)             {$i1_name               = 'Pin';}		#scaler input for the pindiode
      else                            {$i1_name               = 'I0';}		#scaler input for the intensity after GS (actually I0)
  }
  if ($use_KBM && (! $use_CRL) && (! $use_TF)) { #---these-are-defaults-for-mirror-dac-scanning-only----------
     if (!defined $use_encoder)       {$use_encoder           = 1;}		#verify DAC output with ADC readback
     if (!defined $tolerance_factor)  {$tolerance_factor      = 0.5;}		#DAC readback tolerance in scan steps
     $dac        = $beamline.':DAC2_';
     $adc        = $beamline.':pmac21:59e:ai';
     $dacVFM_PV  = $dac.'2';   $adcVFM_PV = $adc.'5';  				#to use fast scans, must be DAC2
     $dacHFM_PV  = $dac.'3';   $adcHFM_PV = $adc.'6';  				#to use fast scans, one must use HYTEC, which is DAC2
#
     if (!$use_encoder) {undef $adcVFM_PV; undef $adcHFM_PV;}
     if (!defined $corr_fly_V)        {$corr_fly_V            = 0;}		#correction to position found by vertical on-fly scan [2013.11.01]
     if (!defined $corr_fly_H) {
        if    ($beamline eq '23i')    {$corr_fly_H            = 0.;}		#correction to position found by horizontal on-fly scan [2017.10.18] (inverted piezo on NSLS mirror)
        elsif ($beamline eq '23o')    {$corr_fly_H            =+0.01;}		#correction to position found by horizontal on-fly scan [2013.11.01]
        else                          {$corr_fly_H            = 0.;}
     }
     if (!defined $scan_stepH)        {$scan_stepH            = 0.08;           #DAC step in Volts
        if ($fastScan)                {$scan_stepH           /= 2;}             #DAC step in Volts (2x less in fast scans)
     }                                                               
     if (!defined $scan_stepV)        {
        if    ($beamline eq '23i')    {$scan_stepV            = 0.04;}		#DAC step in Volts
        elsif ($beamline eq '23o')    {$scan_stepV            = 0.02;}
        else                          {$scan_stepV            = 0.02;}
        if ($fastScan)                {$scan_stepV           /= 2;}		#DAC step in Volts (2x less in fast scans)
     }
     if (!defined $scan_pmrangeH)     {$scan_pmrangeH         = 2.0;  		#+-deviation from current position
        if (defined $scan_factorH)    {$scan_pmrangeH        *= $scan_factorH;}
     }
     if (!defined $scan_pmrangeV)     {
        if    ($beamline eq '23i')    {$scan_pmrangeV         = 1.5;}  		#+-deviation from current position
        elsif ($beamline eq '23o')    {$scan_pmrangeV         = 0.7;}		#correction to position found by horizontal on-fly scan [2013.11.01]
        else                          {$scan_pmrangeV         = 1.;}
        if (defined $scan_factorV)    {$scan_pmrangeV        *= $scan_factorV;}
     }
  } 
  elsif ($use_CRL) { #---these-are-defaults-for-CRL-XY-motor-scanning-only--------------------
     $crl_H_PV    = $beamline.':CRL:XS';
     $crl_V_PV    = $beamline.':CRL:YS';
     $crl_assy_PV = $beamline.':CRL:StS';    
     ($LoH, $HiH) = &getAxisDRVLH ($crl_H_PV,\$status);
     if ($status) {&press_key_exit('*** '.$script.': error='.$status.' getting soft limits for '.$crl_H_PV);}
     elsif ($LoH == 0 && $HiH == 0) {&press_key_exit('*** '.$script.': soft limits for '.$crl_H_PV.' are not set');}
     ($LoV, $HiV) = &getAxisDRVLH ($crl_V_PV,\$status);
     if ($status) {&press_key_exit('*** '.$script.': error='.$status.' getting soft limits for '.$crl_V_PV);}
     elsif ($LoV == 0 && $HiV == 0) {&press_key_exit('*** '.$script.': soft limits for '.$crl_V_PV.' are not set');}
#
     if ($fastScan == 0) {
           print STDOUT '!!! '.$script.': Only fly scans are implemented for CRL. Ignoring step scans request.'."\n";
           if ($tell_bluice) {&tellBluIce($scr.': Ignoring step scans request for CRL!','warn');}
        $fastScan = 1;
     }
     if (!defined $corr_fly_V)        {$corr_fly_V            = 0;}			#correction to position found by vertical on-fly scan (in mm)
     if (!defined $corr_fly_H)        {$corr_fly_H            = 0.;}			#correction to position found by horizontal on-fly scan (in mm)
     if (!defined $scan_stepH)        {$scan_stepH            = 0.001;}			#H-step in mm
     if (!defined $scan_stepV)        {$scan_stepV            = 0.001;}			#V-step in mm
#    if (!defined $scan_pmrangeH)     {$scan_pmrangeH         = ($HiH-$LoH)/4.;		#+-deviation from current position (plan for 50% of travel range)
     if (!defined $scan_pmrangeH)     {$scan_pmrangeH         = 200*$scan_stepH;	#+-deviation from current position (plan for 400 steps)
        if (defined $scan_factorH)    {$scan_pmrangeH        *= $scan_factorH;}
     }
#    if (!defined $scan_pmrangeV)     {$scan_pmrangeV         = ($HiV-$LoV)/4.;		#+-deviation from current position (plan for 50% of trave; range)
     if (!defined $scan_pmrangeV)     {$scan_pmrangeV         = 200*$scan_stepV;	#+-deviation from current position (plan for 400 steps)
        if (defined $scan_factorV)    {$scan_pmrangeV        *= $scan_factorV;}
     }
  }
  elsif ($use_TF) { #---these-are-defaults-for-Transfocator-XY-motor-scanning-only--------------------
     $tf_H_axis_PV = $beamline.':CRL:Y';
     $tf_H_assy_PV = $beamline.':CRL:Ct';    
     $tf_V_axis_PV = $beamline.':CRL:Z';
     $tf_V_assy_PV = $beamline.':CRL:Cv';    
     ($LoH, $HiH) = &getAxisDRVLH ($tf_H_axis_PV,\$status);
     if ($status) {&press_key_exit('*** '.$script.': error='.$status.' getting soft limits for '.$tf_H_axis_PV);}
     elsif ($LoH == 0 && $HiH == 0) {&press_key_exit('*** '.$script.': soft limits for '.$tf_H_axis_PV.' are not set');}
     ($LoV, $HiV) = &getAxisDRVLH ($tf_V_axis_PV,\$status);
     if ($status) {&press_key_exit('*** '.$script.': error='.$status.' getting soft limits for '.$tf_V_axis_PV);}
     elsif ($LoV == 0 && $HiV == 0) {&press_key_exit('*** '.$script.': soft limits for '.$tf_V_axis_PV.' are not set');}
#
     if ($fastScan == 0) {
           print STDOUT '!!! '.$script.': Only fly scans are implemented for Transfocator. Ignoring step scans request.'."\n";
           if ($tell_bluice) {&tellBluIce($scr.': Ignoring step scans request for Transfocator!','warn');}
        $fastScan = 1;
     }
     if (!defined $corr_fly_V)        {$corr_fly_V            = 0;}			#correction to position found by vertical on-fly scan (in mm)
     if (!defined $corr_fly_H)        {$corr_fly_H            = 0.;}			#correction to position found by horizontal on-fly scan (in mm)
     if (!defined $scan_stepH)        {$scan_stepH            = 0.001;}			#H-step in mm
     if (!defined $scan_stepV)        {$scan_stepV            = 0.001;}			#V-step in mm
#    if (!defined $scan_pmrangeH)     {$scan_pmrangeH         = ($HiH-$LoH)/4.;		#+-deviation from current position (plan for 50% of travel range)
#    if (!defined $scan_pmrangeH)     {$scan_pmrangeH         = 200*$scan_stepH;	#+-deviation from current position (plan for 400 steps)
     if (!defined $scan_pmrangeH)     {$scan_pmrangeH         = 0.15;			#+-deviation from current position (plan for +-150um)
        if (defined $scan_factorH)    {$scan_pmrangeH        *= $scan_factorH;}
     }
#    if (!defined $scan_pmrangeV)     {$scan_pmrangeV         = ($HiV-$LoV)/4.;		#+-deviation from current position (plan for 50% of trave; range)
#    if (!defined $scan_pmrangeV)     {$scan_pmrangeV         = 200*$scan_stepV;	#+-deviation from current position (plan for 400 steps)
     if (!defined $scan_pmrangeV)     {$scan_pmrangeV         = 0.15;			#+-deviation from current position (plan for +-150um)
        if (defined $scan_factorV)    {$scan_pmrangeV        *= $scan_factorV;}
     }
  }
  elsif (!$use_KBM) {
     &press_key_exit('*** '.$script.': no alignment device is specified or the specified device is unavalable.');
  } #---end-of-defaults-------------------------------------------------------------

  if ($have_KBM && $use_KBM) {
     if ($beamline eq '23i') {$kbm_encoder_present=1;}
     if ($kbm_encoder_present) {
        $hfm_encoder_PV = $beamline.':pmac11:re1:ActPos';
        $vfm_encoder_PV = $beamline.':pmac11:re2:ActPos';
        $hfm_enc_scale  = 1./1000.;					#nm to um
        $vfm_enc_scale  = 1./1000.;					#nm to um
        ### Oleg: the arm between encoder scale and mirror rotation axis is 451mm
        ### (this may be needed for recalculating encoder reading into mrad):
        #### angle_mrad = 1um_enc / 451mm
        ### Oleg: the arm between piezo and mirror rotation axis is 500mm
        #### angle_mrad = 1um_piezo / 500mm
        ### shift_piezo = 1um_enc * (500mm/451mm)
        ### Oleg: 0 to 10V DAC change corresponds to 40 micron piezo movement (1V => 4um)
#       $hfm_enc_scale /= 4.;					#1um = 1/4V
#       $vfm_enc_scale /= 4.;					#1um = 1/4V
        $hfm_enc_scale *= ((500./451.)/4.);			#1um_enc_to_V
        $vfm_enc_scale *= ((500./451.)/4.);;			#1um_enc_to_V
         
     }
  }
  if (!defined $kbm_encoder_present) {$kbm_encoder_present=0;}

  $gnuopen       = 0;
  $scalerNo      = 1;									#scaler1=Struck, scaler3=Joerger32
  $useDarks      = 0;									#re-measure offsets if needed (we do not use offsets in scans!)
  $logscale      = 0;									#1=yes, 0=no (for Gnuplot)
  @gnuopts       = ('-persist');							#set to '' in not needed
  $postabActV_PV = $beamline.':SoftPID1_incalc.F';
  $postabActH_PV = $beamline.':SoftPID1_incalc.N';
  $gsSizeV_PV    = $beamline.':GS:ZS';
  $gsAssyV_PV    = $beamline.':GS:Av';
  $gsSizeH_PV    = $beamline.':GS:YS';
  $gsAssyH_PV    = $beamline.':GS:Ah';
  $colAssy_PV    = $beamline.':COL:St';
  $activeCol_PV  = $beamline.':COL:St:ActivePreset';
  $shutterAssy_PV= $beamline.':SH:Ps';
  $energyPV      = $beamline.':MO:E:ActPos';
  $pinsensor_PV  = &getPinSensorPV($beamline);						#comment OUT if no pin sensor present
  if ($tell_bluice) {
     $bipostHPV  = $beamline.':bi:alignDataH';
     $bipostVPV  = $beamline.':bi:alignDataV';
     $bialignStatePV = $beamline.':bi:alignState';
  }
  $xyzAssyPV = $beamline.':GO:Cr';
  if (defined $HAVE_XYZ) {
     $zAssyPV   = $xyzAssyPV;
     @xyzAxesPV = ($beamline.':GO:CX',$beamline.':GO:CY',$beamline.':GO:CZ');
  } else {
     $zAssyPV = $beamline.':GO:St';
     if ($beamline eq '23b') {$zAssyPV .= 'B';}
     @xyzAxesPV = ($beamline.':GO:CX',$beamline.':GO:CY',$beamline.':GO:SH');
  }
  $nxyz =  @xyzAxesPV; 
  $omegaAxisPV = $beamline.':GO:OM';
  $degree2rad = atan2(1.,1.) / 45.;							#atan2(1)=pi/4 = 45.

  $availableH   = 1;									#H-mirror available
  $availableV   = 1;									#V-mirror available
# if ($beamline =~ /^23i/) {$availableV=0;}						#ID-in mirror temporary removed
  $afterScan    = 4;		     							#0=stay, 1=startpos, 2=priorpos 3=maxpos 4=peakpos 5=minpos 6=valleypos
  $OVERLOAD     = 100000;								#scan failure flag when overload detected
  $FOLLERR      = 200000;								#scan failure flag when overload detected
  $fileBkg      = '';									#not used
  $scaler       = $joergerName.$scalerNo;
  $struck       = $struckName;
  $scanAbort    = $scanrecAbort;
  $scanrec      = $scanrecName;
  if ($scanAbort        !~ /^${beamline}/) {$scanAbort        = $beamline.':'.$scanAbort;}
  if ($scanrec          !~ /^${beamline}/) {$scanrec          = $beamline.':'.$scanrec;}
  if ($scaler           !~ /^${beamline}/) {$scaler           = $beamline.':'.$scaler;}
  if ($struck           !~ /^${beamline}/) {$struck           = $beamline.':'.$struck;}
  if ($shutter          !~ /^${beamline}/) {$shutter          = $beamline.':'.$shutter;}
  if ($shutter_readback !~ /^${beamline}/) {$shutter_readback = $beamline.':'.$shutter_readback;}
  &setJoerger ($scaler, $cnt_time, $struck);
  @Offsets  = &getJoergerOffsets ($fileBkg,$scaler);				#we make it GLOBAL to access from main
  $npts_max = &SRmaxPts($scanrec);
  if ($tell_bluice) {
     ### Restrict number of points if it does not fit the JBluIce post array:
     ($status,$npts_max2) = Pezca::GetDouble($bipostHPV.'.NELM');
     if ($status) {
        &report_only($script, $status, $bipostHPV.'.NELM'); 
     } else {
        $npts_max2 = int(($npts_max2-4)/2);
        if ($npts_max2 < $npts_max) {$npts_max = $npts_max2;}
     }
  }
# print STDOUT $script.': Maximum number of points in the scans will be limited to '.$npts_max."\n";

  $alignHistPV  = $beamline.':bi:AlignBeamHistory';

#----------------------------------------------------------------
### Protection from multiple instances:
  &check_Nrunning_and_host_allowed();
###----------------------------------------------------------------
### Warn that sample may be affected
  if ($protectSample) {
     if (defined $pinsensor_PV) {
        print STDOUT $script.': Checking Pin Sensor ['.$pinsensor_PV.']'
                            .' to see if any sample (pin) is mounted'."\n";
        if ($tell_bluice) {
           &tellBluIce($scr.': checking pin sensor for mounted sample');
        }
        ($status,$pin_mounted) = &pinMounted($pinsensor_PV);
        if ($status) {
           print STDOUT '!!! '.$script.': Presuming pin is mounted.'."\n\n";
           if ($tell_bluice) {&tellBluIce($scr.': Failed to get pin state;'
                                         .' presuming it is mounted!','warn');}
           $pin_mounted = 1;			#presuming pin is mounted
        }
     } else {
        $pin_mounted = 1;			#presuming pin is mounted
     }
     if ($pin_mounted) {
        $zzz = $ZENITY_VIA_BLUICE;
        if (! defined $JBLUICE_SERVER || ! $tell_bluice) {$ZENITY_VIA_BLUICE = 0;}
        if ($protectSample == 1) {
           $msg = 'ATTENTION: there is a pin mounted on the goniometer!'."\n"
                . 'Please make sure there is no sample in the beam!'."\n"
                . 'Press "OK" to start alignment or "Cancel" to abort.';
           if (&zenity_question($msg,'OK','Cancel') == 0) {exit 2;}		#this will return "512" to calling perl script
        }
        elsif ($protectSample > 1) {
           if ($protectSample == 2) {						#auto-move sample (first ask)
              $msg = sprintf
                     'ATTENTION: there is a pin mounted on the goniometer!'."\n"
                   . 'To protect it from radiation, I am going to retract'."\n"
                   . 'the pin %s by %.3fmm (%.1fum).'."\n"
                   . 'Press "OK" to retract the pin or "Cancel" to proceed.'."\n"
                   . 'with aligning without retracting the pin.', $direction, abs($moveSample), abs(1000*$moveSample);
              if (&zenity_question($msg,'OK','Cancel') == 0) {
                 $moveSample = 0;
                 $msg = 'Are you sure you want to proceed with the sample in the beam?'."\n"
                      . 'Press "Yes" to proceed or "No" to abort beam aligning.';
                 if (&zenity_question($msg,'Yes','No') == 0) {exit 2;}		#this will return "512" to calling perl script
              }
           }
           if ($moveSample != 0) {
              for ($i=0; $i<$nxyz; $i++) {$xyzIni[$i] = &getAxisAct($xyzAxesPV[$i]);}
              $omega = &getAxisAct($omegaAxisPV);
              ### Move the sample out of the beam:
              if ($tell_bluice) {&tellBluIce($scr.': moving sample '.$direction,'warn');}
              if ($direction =~ /^U/ || $direction =~ /^D/) {
                 $status = &runNMotors ($xyzAssyPV, 2, [$xyzAxesPV[0],$xyzAxesPV[1]],
                                        [$xyzIni[0]+$moveSample*sin($omega*$degree2rad),
                                         $xyzIni[1]+$moveSample*cos($omega*$degree2rad)]);
              } else {
                 $status = &runNMotors ($zAssyPV, 1, [$xyzAxesPV[2]],
                                        [$xyzIni[2]+$moveSample]);
              }
              if ($status) {
                 $msg = 'Error='.$status.' moving sample out of the beam.';
                 print STDOUT '*** '.$script.': '.$msg."\n";
                 if ($tell_bluice) {&tellBluIce($scr.': '.$msg,'error');}
                 &zenity_error_exit('Failed moving sample '.$direction.'.'."\n"
                                   .'Energy not changed (script aborted).',1);
              } else {
                 $sample_moved = 1;
              }
#          } else {
#             ### Commented out because ($moveSample=0) is possible if we allowed not moving the pin
#             &zenity_error_exit('Moving sample '.$direction.' requred, but specified distance is zero.'."\n"
#                               .'Energy not changed (script aborted).',1);
           }
        }
        $ZENITY_VIA_BLUICE = $zzz;
     } else {
        print STDOUT $script.': Pin Sensor ['.$pinsensor_PV.'] reports that there is no pin mounted'."\n";
        if ($tell_bluice) {&tellBluIce($scr.': no pin mounted according to pin sensor');}
     }
  }


  $t0 = [gettimeofday];										#start measuring time

#----------------------------------------------------------------
### Query list of available signals from mySQL:
  $status = &getDarks4KeithleyList(\@signals);
  if ($status) {
     &zenity_error_exit('*** '.$script.': failed to get list of signals from mySQL');
  }
  if (! grep(/^${i1_name}$/, @signals)) {
     &zenity_error_exit('*** '.$script.': incorrect signal='.$i1_name
                                      .' -- must be one of '."(@{signals})");
  }

### Read data from previous alignment (if present):
  ($date_last, $energy_last, $attn_last, $i1_name_last, $kgain_last,
   $center_last_V, $I1_last_scan_V, $I1_last_final_V,
   $center_last_H, $I1_last_scan_H, $I1_last_final_H, $scantype_last) = &read_last_alignment($alignHistPV);

### Find current energy and attenuation:
  ($status, $energy) = Pezca::GetDouble($energyPV);
  if ($status) {&report_only($script, $status, $energyPV); $energy=12;}
  ($status,$attn) = &getAttenuation2($energy);
  if ($status) {print STDOUT '*** '.$script.': failed to get current attenuation'."\n"; $attn=1;}
  $energy = 1000.*$energy;						#convert to eV

### Read scaler/keithley data for I1 from mySQL and respective darks:
### (We ignore darks in this program, but use I1 data)
  $status = &getDarks4KeithleyGains($i1_name,\$timestampDB,\$epochDB,
                                    \$kchannel,\$klabelDB,\$koffsetDB,
                                    \$i1,\@darks);
  if ($status) {
     if ($tell_bluice) {&tellBluIce($scr.': reading darks from mySQL failed','error');}
     &zenity_error_exit('*** '.$script.': failed to get data for '.$i1_name.' from mySQL');
  }
  $klabel  = &keithley_read_label($kchannel);
  $koffset = &keithley_read_offset($kchannel);
### Verify if I1 darks need to be re-measured and re-do the measurements if needed:
  if ($useDarks) {
     $epoch = time();
     $bkg_age_hours = ($epoch - $epochDB)/3600.;
     $offset_delta  =  abs($koffset-$koffsetDB) / &findMaxRef([abs($koffset),abs($koffsetDB),1]);
     if ($bkg_age_hours > 24 || $offset_delta > 0.01 || $klabel ne $klabelDB) {
        print STDOUT $script.': Acquiring '.$i1_name.' darks...'."\n";
        if ($tell_bluice) {&tellBluIce($scr.': acquiring '.$i1_name.' darks');}
        @prg = ('perl', '-W', $PEZCA.'/measureDark1.pl', $beamline, $i1_name);
        $status = &execute_external_script(\@prg);
        if ($status) {
           if ($tell_bluice) {&tellBluIce($scr.': acquiring '.$i1_name.' darks failed','error');}
           &zenity_error_exit ('*** '.$script.': Failed to acquire '.$i1_name.' darks');
        }
        print STDOUT $script.': darks are acquired.'."\n";
        if ($tell_bluice) {&tellBluIce($scr.': '.$i1_name.' darks acquired');}
### Re-read updated I1 background:
        $status = &getDarks4KeithleyGains($i1_name,\$timestampDB,\$epochDB,
                                          \$kchannel,\$klabelDB,\$koffsetDB,
                                          \$i1,\@darks);
        if ($status) {
           if ($tell_bluice) {&tellBluIce($scr.': re-reading darks from mySQL failed','error');}
           &zenity_error_exit('*** '.$script.': failed to get data for '.$i1_name.' from mySQL');
        }
     }
  }
  $kgain = &keithley_read_gain($kchannel);
  @inputs  = ($i1);								# $i1 is the input number
  $ninputs = @inputs;
###----------------------------------------------------------------
### Query what can be possible names of collimator presets:
  @mbbstates = ('ZRST', 'ONST', 'TWST', 'THST', 'FRST', 'FVST', 'SXST', 'SVST');
  for ($i=0; $i<8; $i++) {
     ($status,$preset[$i]) = Pezca::GetString($activeCol_PV.'.'.$mbbstates[$i]);
     if ($status) {
        &report_only($script, $status, $activeCol_PV.'.'.$mbbstates[$i]);
        $preset[$i] = 'unknown';
     }
     if    ($preset[$i] =~ /10[ ]*um /)  {$state10um  = $i;}
     elsif ($preset[$i] =~ /300[ ]*um /) {$state300um = $i;}
     elsif ($preset[$i] =~ /^None/i)     {$stateOUT   = $i;}
  }
  if (!defined $state10um || !defined $state300um || !defined $stateOUT) {
     print STDOUT '!!! '.$script.': Cannot identify collimator states'
                                .' -- proceeding with guard slits'."\n\n";
     $collimatorPreset = 0;							#0: collimator=none (out)
     $align300to10 = 0;
     $usegs        = 1;
  } else {
     ### Query what is current requested collimator preset (0=none 1=300um 2=20um 3=10um 4=5um ...)?
     ($status,$collimatorPreset) = Pezca::GetDouble($activeCol_PV);		#what collimator installed?
     if ($status) {
        &report_only($script, $status, $activeCol_PV);
        print STDOUT '!!! '.$script.': Presuming no collimator'
                           .' and proceeding with guard slits'."\n\n";
        $collimatorPreset = 0;							#0: collimator=none (out)
        $align300to10 = 0;
        $usegs        = 1;
     } 
  }

  if ($auto_choice) {
    ### We are here if flag "autoPinholeOrGuardSlits":
    if ($collimatorPreset >= $state10um) {					#0=none 1=300um 2=20um 3=10um 4=5um ...
      $usegs        = 0;
      $align300to10 = 0;
    } else {
      if    ($usegs == 0 && $align300to10 == 0) {$usegs = 0; $align300to10 = 1;}
      elsif ($usegs == 0 && $align300to10 == 1) {$usegs = 0; $align300to10 = 1;}
      elsif ($usegs == 1 && $align300to10 == 0) {$usegs = 1; $align300to10 = 0;}
      elsif ($usegs == 1 && $align300to10 == 1) {$usegs = 1; $align300to10 = 0;}
    }
  }

  if ($usegs == 0 || $align300to10 == 1) {
     if ($beamline eq '23b' || $beamline eq '23d') {
        print STDOUT "\n".'!!! '.$script.': Beamline '.$beamline
                         .' does not have pinhole. Using guard slits.'."\n";
        $usegs = 1;
        $align300to10 = 0;
     }
  }
  if ($usegs == 1 && $align300to10 == 1) {
### This will exit automatically if $NONINTERACTIVE=1:
        &press_key_exit("\n".'*** '.$script.': both useGuardSlits and align300to10 cannot be set.'."\n".$syntax);
  }
  if ($collimatorPreset == 0 && $align300to10 == 1) {
     print STDOUT '!!! '.$script.': Collimator is currently set to "None".'."\n"
                        .'The script will attempt to install 10um collimator.'."\n"
                        .'If instead of that you want to align with the guard slits,'."\n"
                        .'then add the "useGuardSlits" flag and remove "align300to10".'."\n\n";
  }


### Check id DAC supports fast scanning:
  if ($use_KBM && $fastScan && ($dacHFM_PV !~ /:DAC2_/ || $dacVFM_PV  !~ /:DAC2_/)) {
     print STDOUT "\n".'!!! '.$script.': DAC does not support on-the-fly scanning. Switching to step mode!'."\n";
     $fastScan = 0;
  }
### Option to use vertical position feedback:
  if (($use_CRL || $use_TF) && $tryVerticalHWfeedback) {
     print STDOUT "\n".'!!! '.$script.': CRL and Transfocator scannings are incompatible with the "tryVerticalHWfeedback" option. Ignoring.'."\n";
     $tryVerticalHWfeedback = 0;
  }
  if ($tryVerticalHWfeedback) {
     if ($i1_name !~ /^Pin/i) {
        print STDOUT "\n".'!!! '.$script.': Ignoring option to use FPGA vertical position feedback'
                                        .' because the signal='.$i1_name.' is not PinDiode'."\n";
        $tryVerticalHWfeedback = 0;
     }
     elsif (!defined $i1_name_last || $i1_name_last ne $i1_name       ||
            !defined $energy_last  || abs($energy-$energy_last) > 100 ||
            !defined $attn_last    || abs(1-$attn_last/$attn) > 0.01  ||
            !defined $I1_last_final_V) {
        print STDOUT "\n".'!!! '.$script.': Ignoring option to use FPGA vertical position feedback'
                                        .' because the previous alignment data anavailable'
                                        .' (no intensity data for the success/fail criteria)'."\n";
        $tryVerticalHWfeedback = 0;
     }
  }
### Option to optimize gain and previous data:
  if (defined $i1_name_last && $i1_name_last eq $i1_name       &&
      defined $energy_last  && abs($energy-$energy_last) < 100 &&
      defined $attn_last    && abs(1-$attn_last/$attn) < 0.01  &&
      defined $kgain_last) {
     if ($kgain != $kgain_last) {
        $status = &keithley_set_gain($kchannel,$kgain_last,'nowait');
        if (! $status) {$kgain  = $kgain_last;}
     }
     if ($preset_gain == 1) {
        print STDOUT "\n".'!!! '.$script.': Ignoring option to optimize Keithley gain and'
                                        .' using last beam alignment gain instead.'
                                        .' Use "optimizeGain=force" to override.'."\n";
        $preset_gain = 0;
     }
  }

#----------------------------------------------------------------
### Print Defaults:
# if ($narg > 0)              {print  STDOUT $script.': starting with cmdline "'."@ARGV".'"'."\n";}
# else                        {print  STDOUT $script.': starting with all defaults (none cmdline arguments)'."\n";}
                               print  STDOUT "\n";
                               print  STDOUT 'Beam alignment is running with the following parameters:'."\n";
                               print  STDOUT '^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^'."\n";
                               print  STDOUT '  Beamline = '.$beamline."\n";
                               print  STDOUT '  Signal   = '.$i1_name."\n";
  if ($auto_choice)           {print  STDOUT '  Automatically choose guard slits or collimator = Yes'."\n";}
  else                        {print  STDOUT '  Automatically choose guard slits or collimator = No'."\n";}
  if ($usegs)                 {print  STDOUT '  Use guards_slits for alignment = Yes, close them to size = '.$gssize.'mm'."\n";}
  else                        {print  STDOUT '  Use guards_slits for alignment = No (will be using collimator)'."\n";}
  if ($align300to10)          {print  STDOUT '  Replace 300um collimator by 10um collimator for alignment = Yes'."\n";
     if (!$move_collimator)   {print  STDOUT '*** '.$script.': Contradictory requirements: beam cannot be aligned'."\n";
                               print  STDOUT '*** '.$script.': to 10um collimator when collimator motion disabled.'."\n";}
  } else                      {print  STDOUT '  Replace 300um collimator by 10um collimator for alignment = No'."\n";}
  if ($move_collimator)       {print  STDOUT '  Bring collimator IN before alighnment = Yes'."\n";}
  else                        {print  STDOUT '  Bring collimator IN before alighnment = No'."\n";}
  if ($restore_collimator)    {print  STDOUT '  Restore collimator IN/OUT state after alighnment = Yes'."\n";}
  else                        {print  STDOUT '  Restore collimator IN/OUT state after alighnment = No'."\n";}
  if ($move_pindiode)         {print  STDOUT '  Bring pindiode IN before alighnment = Yes'."\n";}
  else                        {print  STDOUT '  Bring pindiode IN before alighnment = No'."\n";}
  if ($fastScan)              {print  STDOUT '  Use fastscan for alighnment = Yes'."\n";}
  else                        {print  STDOUT '  Use fastscan for alighnment = No'."\n";}
  if    ($protectSample == 0) {print  STDOUT '  Ignore possible sample on goniometer'."\n";}
  elsif ($protectSample == 1) {print  STDOUT '  Warn about possible sample on goniometer = Yes'."\n";}
  elsif ($protectSample == 2) {print  STDOUT '  Retract sample on goniometer from the beam: Ask, by '.abs($moveSample).'mm '.$direction."\n";}
  elsif ($protectSample == 3) {print  STDOUT '  Retract sample on goniometer from the beam: Yes, by '.abs($moveSample).'mm '.$direction."\n";}
  else                        {print  STDOUT '  Warn about possible sample on goniometer = No'."\n";}
  if($tryVerticalHWfeedback)  {print  STDOUT '  Try FPGA vertical position feedback instead of scan = Yes'."\n";}
  else                        {print  STDOUT '  Try FPGA vertical position feedback instead of scan = No'."\n";}
  if ($preset_gain)           {print  STDOUT '  Pre-optimize '.$i1_name.' gain = Yes'."\n";}
  else                        {print  STDOUT '  Pre-optimize '.$i1_name.' gain = No'."\n";}
  if ($overload_rescan)       {print  STDOUT '  Auto-reduce '.$i1_name.' gain and rescan on overload = Yes'."\n";}
  else                        {print  STDOUT '  Auto-reduce '.$i1_name.' gain and rescan on overload = No'."\n";}
  if    ($scalerNo == 1)      {print  STDOUT '  Scaler = Struck ('.$scaler.')'."\n";}
  elsif ($scalerNo == 3)      {print  STDOUT '  Scaler = Joerger ('.$scaler.')'."\n";}
  else                        {print  STDOUT '  Scaler = ***UNKNOWN***'."\n";}
                               print  STDOUT '  Scaler: count_time = '.$cnt_time.'s,  I1_input = '.$i1.' ('.$i1_name.')'."\n";
  if ($use_CRL || $use_TF) {
                               printf STDOUT '  CRL/TF horizntl scan range = +/- %.3fmm,  step = %.3fmm'."\n",$scan_pmrangeH,$scan_stepH;
                               printf STDOUT '  CRL/TF vertical scan range = +/- %.3fmm,  step = %.3fmm'."\n",$scan_pmrangeV,$scan_stepV;
  } else {
     if ($availableH)         {printf STDOUT '  HFM: piezo_range = +/- %.3fV,  piezo_step = %.3fV'."\n",$scan_pmrangeH,$scan_stepH;}
     if ($availableV)         {printf STDOUT '  VFM: piezo_range = +/- %.3fV,  piezo_step = %.3fV'."\n",$scan_pmrangeV,$scan_stepV;}
  }
                               print  STDOUT '  Horizntl alignment choice: '.$center_H."\n";
                               print  STDOUT '  Vertical alignment choice: '.$center_V."\n";
                               print  STDOUT '^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^'."\n";
  if ($scalerNo != 1 && $scalerNo != 3) {
     &press_key_exit($script.': Incorrect scaler');				#this will exit automatically if $NONINTERACTIVE=1
  }
#  &press_key_exit();
#----------------------------------------------------------------
### Change working directory:
  if ($^O !~ /Win/i) {
     $wrkdir = $ENV{HOME}.'/bluice_temp';
     if (!(-e $wrkdir)) {
        if (mkdir($wrkdir) != 1) {
           print STDOUT '!!! '.$script.': cannot make directory '.$wrkdir."\n\n";
           $wrkdir = $ENV{HOME};
        }
     }
  } else {
     $wrkdir = $ENV{TEMP};
  }
  if (chdir($wrkdir) != 1) {
     print STDOUT '!!! '.$script.': cannot cwd to '.$wrkdir."\n\n";
  }

  $SIG{HUP}  = 'signal_handler';	#(1) hangup (see signum.h)
  $SIG{INT}  = 'signal_handler';	#(2) interrupt (see signum.h)
  $SIG{QUIT} = 'signal_handler';	#(3) quit (see signum.h)
# $SIG{KILL} = 'signal_handler';	#(9) this is what is sent by kill -9 -- unblockable
  $SIG{TERM} = 'signal_handler';	#(15) terminate -- sent by kill without paramaters
# $SIG{STOP} = 'signal_handler';	#(19) stop -- unblockable

  &closeShutter ($shutter,$shutter_readback,$shutterAssy_PV);

  if (&getCshutter($beamline) != 1) {
     if (&zenity_question($script.': Please open C-shutter!'."\n"
                         .'Press "OK" when ready or "Cancel" to abort alignment.',
                          'OK','Cancel') != 1) {&signal_handler('quit');}
  }
  if (&getFrontEndShutter($beamline) != 1) {
     if (&zenity_question($script.': Please open FE-shutter!'."\n"
                         .'Press "OK" when ready or "Cancel" to abort alignment.',
                          'OK','Cancel') != 1) {&signal_handler('quit');}
  }
  if ($tell_bluice) {
     $status = Pezca::PutDouble($bialignStatePV,1);				#tell BluIce that the script has started,
     &report_only($script, $status, $bialignStatePV);				#so BluIce should disable some buttons
     if ($status) {undef $bialignStatePV;}
  }

###=====================================================================
### -1. Record coords: size may change when $usegs#0 or when putting collimator IN
  if ($usegs || $move_collimator || $move_pindiode) {
     $gsSizeH_initial = &getAxisAct ($gsSizeH_PV);
     $gsSizeV_initial = &getAxisAct ($gsSizeV_PV);
  }


###=====================================================================
### 0. Close Guard Slits
  if ($usegs) {
     print STDOUT $script.': Closing guard slits to '.$gssize.' x '.$gssize."\n";
     if ($tell_bluice) {&tellBluIce($scr.': closing guard slits');}
     $status = &run_guard_slits($gssize,$gssize);
     if ($status) {
        &tellBluIceBeforeExit($scr.': closing GS failed');
        &zenity_error_exit('*** '.$script.': failed to close guard slits');
     }
  }


###=====================================================================
### 1. Move collimator and pin diode to "IN" position (external script)
  if ($move_collimator || $move_pindiode) {
     if ($move_collimator && $align300to10 == 1 && $collimatorPreset <= 1) {	#Preset: 0=none 1=300um
        if    ($collimatorPreset == 0) {					#0=none 1=300um 2=20um 3=10um, 4=5um ...
           print STDOUT "\n".$script.': Temporary bringing'
                            .' in 10um collimator'."\n";
           if ($tell_bluice) {&tellBluIce($scr.': collimator none->10 (temp)');}
        }
        elsif ($collimatorPreset == 1) {					#0=none 1=300um 2=20um 3=10um, 4=5um ...
           print STDOUT "\n".$script.': Temporary changing active'
                            .' collimator from 300um to 10um'."\n";
           if ($tell_bluice) {&tellBluIce($scr.': collimator 300->10 (temp)');}
        }
        $status = Pezca::PutDouble($activeCol_PV,$state10um);
        &report_only($script, $status, $activeCol_PV);
        ### Track the change of collimator:
        if (&waitMotor($colAssy_PV,5,0.3) == 0) {
           &tellBluIceBeforeExit($scr.': changing collimator failed');
           &zenity_error_exit('*** '.$script.': timeout on changing collimator');
        }

     }
     @prg = ('IN', $beamline, 'backlightOUT');
     if ($move_collimator) {
        print STDOUT "\n".$script.': Moving collimator IN'."\n";
        if ($tell_bluice) {&tellBluIce($scr.': moving collimator IN');}
     } else {
        push @prg, 'nocollimator';
     }
     if ($move_pindiode) {
        print STDOUT "\n".$script.': Moving pindiode IN and beamstop OUT'."\n";
        if ($tell_bluice) {&tellBluIce($scr.': moving pindiode IN, beamstop OUT');}
        push @prg, ('beamstopOUT', 'pindiodeIN');
     } else {
        push @prg, 'nobeamstop';
     }
     $status = &setSampleEnv(\@prg);
     if ($status) {
        print STDOUT "\n".'*** '.$script.': failure='.$status
                                .' from setSampleEnv'."\n\n";
        &tellBluIceBeforeExit($scr.': setSampleEnv failed');
        &zenity_error_exit('*** '.$script.': failed on setSampleEnv.');
     }
  }


###=====================================================================
### 2a. Pause positional feedback (#0=run, 3=pause)
  if ($interfacePostab) {
     $postab_initial = &getPostabState ($beamline);
     if ($postab_initial == -1) {						#failure
        &tellBluIceBeforeExit($scr.': Postab stat failed');
        &zenity_error_exit('*** '.$script.': failed to get Postab status');
     }
     if ($postab_initial != 3) {						#if not paused
        print STDOUT "\n".$script.': Pausing positional feedback'."\n";
        if ($tell_bluice) {&tellBluIce($scr.': pausing pos.feedback');}
        $status = &pausePostab ($beamline);
        if ($status) {								#failure
           &tellBluIceBeforeExit($scr.': Postab pause failed');
           &zenity_error_exit('*** '.$script.': failed to pause Postab');
        }
     }
  } else {
     $postab_initial = 3							#a.k.a. paused
  }

###=====================================================================
### 2b. Pause intensity feedback (#0=run, 1=pause)
  if ($interfaceMostab) {
     $mostab_initial = &getMostabState ($beamline);
     if ($mostab_initial == -1) {						#failure
        &tellBluIceBeforeExit($scr.': Mostab stat failed');
        &zenity_error_exit('*** '.$script.': failed to get Mostab status');
     }
     if ($mostab_initial != 1) {						#if not paused
        print STDOUT "\n".$script.': Pausing intensity feedback'."\n";
        if ($tell_bluice) {&tellBluIce($scr.': pausing intensity feedback');}
        $status = &pauseMostab ($beamline);
        if ($status) {					#faiulre
           &tellBluIceBeforeExit($scr.': Mostab pause failed');
           &zenity_error_exit('*** '.$script.': failed to pause Mostab');
        }
     }
  } else {
     $mostab_initial = 1							#a.k.a. paused
  }


###=====================================================================
###  3. Optimize keithley gain
  if ($preset_gain) {
     print STDOUT "\n".$script.': Pre-optimizing Keithley gain'."\n";
     if ($tell_bluice) {&tellBluIce($scr.': pre-optimizing Keithley gain');}
     $I_optimum = 1.0 * 1.E5;
     ($status, $kgain, $I1, $msg) = &keithleyOptimizeGain($i1_name,$kchannel,$scaler,$i1,
                                                          $I_optimum,undef,
                                                          undef,undef,undef,
                                                          $tell_bluice,$scr);
     if ($status) {
        if ($tell_bluice) {&tellBluIce($scr.': Keithley optimize failed','error');}
        &zenity_error('*** '.$script.': Keithley optimization failed'."\n".$msg);
        &signal_handler('quit',1);
     }
  }

###=====================================================================
###  3a. Clear BluIce post arrays  (optional)
  if ($tell_bluice) {
     $status = Pezca::PutList($bipostVPV, 'DOUBLE', (1,0,0,10));
     if ($status) {
        &report_only($script, $status, $bipostVPV);
        print STDOUT '!!! '.$script.': Further posting into '.$bipostVPV.' will be disabled.'."\n\n";
        undef $bipostVPV;
     }
     $status = Pezca::PutList($bipostHPV, 'DOUBLE', (1,0,0,10));
     if ($status) {
        &report_only($script, $status, $bipostHPV);
        print STDOUT '!!! '.$script.': Further posting into '.$bipostHPV.' will be disabled.'."\n\n";
        undef $bipostHPV;
     }
  }

###=====================================================================
###  4. Open shutters
  print STDOUT "\n".$script.': Opening shutters'."\n";
  if ($tell_bluice) {&tellBluIce($scr.': opening shutters');}
  $status = &getEndStationShutter($beamline);
  if ($status != 1) {					#if closed
     $status = &openEndStationShutter($beamline);
     if ($status) {
        if ($tell_bluice) {&tellBluIce($scr.': Station shutter failed','error');}
        &zenity_error('*** '.$script.': failed to open endstation shutter.');
        &signal_handler('quit');
     }
  }
  ### In principle, this is opened below by scan routine, but here it provides checking:
  $status = &openShutter ($shutter,$shutter_readback,$shutterAssy_PV);
  if ($status) {
     if ($tell_bluice) {&tellBluIce($scr.': Fast shutter failed','error');}
     &zenity_error('*** '.$script.': failed to open fast shutter.');
     &signal_handler('quit');
  }

###=====================================================================
###  5x.  Optionally attempt vertical optimization with FPGA
  $verticalHWfeedbackSucceeded = 0;
  if ($tryVerticalHWfeedback) {&scan_FPGA_V();}

###=====================================================================
###  5a. Scan VFM Pz (DAC) and then move to maximum intensity (+-2V)
  if ($use_KBM && $availableV && $verticalHWfeedbackSucceeded == 0) {&scan_KBM_V();}

###=====================================================================
###  5b. Scan CRL vertical of the small CRL positioner and then move to maximum intensity
  if ($use_CRL) {&scan_CRL_V();}

###=====================================================================
###  5c. Scan Transfocator horizontal and then move to maximum intensity
  if ($use_TF) {&scan_TF_H();}			#for TF the H/V order is reverted per David Kissick request

###=====================================================================
###  6a. Scan HFM Pz (DAC) and then move to maximum intensity (+-2V)
  if ($use_KBM && $availableH) {&scan_KBM_H();}

###=====================================================================
###  6b. Scan CRL horizontal of the small CRL positioner and then move to maximum intensity
  if ($use_CRL) {&scan_CRL_H();}

###=====================================================================
###  6c. Scan Transfocator vertical and then move to maximum intensity
  if ($use_TF) {&scan_TF_V();}							#for TF the H/V order is reverted per David Kissick request

###=====================================================================
### 7. Start intensity feedback (0=run, 1=pause) + wait for locking
  if ($interfaceMostab && $mostab_initial != 1) {				#if not paused initially
     print STDOUT $script.': Restoring intensity feedback'."\n";
     if ($tell_bluice) {&tellBluIce($scr.': restoring Intensity feedback');}
     $status = &restoreMostab ($beamline);
#    if ($status) {&press_key_exit();}
     if ($status == 0) {
        if ($tell_bluice) {&tellBluIce($scr.': waiting Intensity lock');}
        $status = &waitMostabLock ($beamline);
        if ($status) {
           print STDOUT "\n".'*** '.$script.': Failed to get Intensity lock'."\n";
           &tellBluIce('Intensity lock failed','warn');
        } else {
           &tellBluIce('Intensity locked');
        }
     }
  }


###=====================================================================
###  8. Verify HFM and VFM maxima after the feedback was turned on:
  if ($interfaceMostab && $mostab_initial != 1) {				#if not paused initially
     if ($tell_bluice) {&tellBluIce($scr.': verifying intensity after restoring intensity feedback');}
     &setJoerger ($scaler, $cnt_time, $struck);
     &openShutter ($shutter,$shutter_readback);
     @I1_max_now  = &runJoerger ($scaler, $cnt_time, $ninputs, @inputs);
     &closeShutter ($shutter,$shutter_readback);
     &subtractJoergerOffsets(\@I1_max_now,\@inputs,\@Offsets);
     $I1_max_final_V = $I1_max_final_H = $I1_max_now[0];
     $I1_percent_V = 100. * ($I1_max_final_V / $I1_max_scan_V);
     $I1_percent_H = 100. * ($I1_max_final_H / $I1_max_scan_H);
     printf STDOUT "\n".$script.': Intensity at final V-pos. is %1.0f%s'
                       .' of scan maximum (scanmax=%.3g now=%.3g)'."\n\n",
                       $I1_percent_V, '%', $I1_max_scan_V, $I1_max_final_V;
     printf STDOUT "\n".$script.': Intensity at final H-pos. is %1.0f%s'
                       .' of scan maximum (scanmax=%.3g now=%.3g)'."\n\n",
                       $I1_percent_H, '%', $I1_max_scan_H, $I1_max_final_H;
  }
###---------------------------------------------------------###
### Tweak final pos. if the final intensity is < 95% or so. ###
###---------------------------------------------------------###
  if ($use_KBM) {
     if ($availableV && $I1_percent_V < 95) {
        print STDOUT $script.': tweaking vertical DAC'."\n";
        ($status,$dac,$I1) = &optimizeDAC($dacVFM_PV, $scan_stepV, $scaler,
                                          $cnt_time, $i1, $fileBkg, 0.2);
        if ($status == 0) {
           $FinalV = $dac;
           $I1_percent_V = 100. * ($I1 / $I1_max_scan_V);

        }
     }
     if ($availableH && $I1_percent_H < 95) {
        print STDOUT $script.': tweaking horizontal DAC'."\n";
        ($status,$dac,$I1) = &optimizeDAC($dacHFM_PV, $scan_stepH, $scaler,
                                       $cnt_time, $i1, $fileBkg, 0.2);
        if ($status == 0) {
           $FinalH = $dac;
           $I1_percent_H = 100. * ($I1 / $I1_max_scan_H);

        }
     }

     print STDOUT "\n".$script.': Final DAC(s):'."\n";
     if ($availableV) {printf STDOUT '   VFM=%-5.4f (%.0f%s)'."\n",$FinalV,$I1_percent_V,'%';}
     if ($availableH) {printf STDOUT '   HFM=%-5.4f (%.0f%s)'."\n",$FinalH,$I1_percent_H,'%';}
     print STDOUT "\n";
  }
  elsif ($use_TF) {
     if ($I1_percent_V < 95) {
        printf STDOUT $script.': tweaking vertical TF position because final V-peak is %.0f%s of the scan peak'."\n",$I1_percent_V,'%';
        ($status,$pos,$I1) = &optimizePos($tf_V_assy_PV, $tf_V_axis_PV, 
                                          $scan_stepV, $scaler,
                                          $cnt_time, $i1, $fileBkg, 0.2);
        if ($status == 0) {
           $FinalV = $pos;
           $I1_percent_V = 100. * ($I1 / $I1_max_scan_V);

        }
     }
     if ($I1_percent_H < 95) {
        printf STDOUT $script.': tweaking horizontal TF position because final H-peak is %.0f%s of the scan peak.'."\n",$I1_percent_H,'%';
        ($status,$pos,$I1) = &optimizePos($tf_H_assy_PV,$tf_H_axis_PV,
                                          $scan_stepH, $scaler,
                                          $cnt_time, $i1, $fileBkg, 0.2);
        if ($status == 0) {
           $FinalH = $pos;
           $I1_percent_H = 100. * ($I1 / $I1_max_scan_H);

        }
     }

     print STDOUT "\n".$script.': Final TF positions:'."\n";
     printf STDOUT '   V=%-5.4f (%.0f%s)'."\n",$FinalV,$I1_percent_V,'%';
     printf STDOUT '   H=%-5.4f (%.0f%s)'."\n",$FinalH,$I1_percent_H,'%';
     print STDOUT "\n";
  }


###=====================================================================
###  9. Store new set points in the positional feedback
  print STDOUT "\n".$script.': Loading new set points for positional feedback'."\n";
  if ($tell_bluice) {&tellBluIce($scr.': loading pos.setpoints');}
#
  ($status,$setpointV) = Pezca::GetDouble($postabActV_PV);			#read current V-mirror DAC
  if ($status) {
     &report_only($script,$status,$postabActV_PV);
#    &signal_handler('quit');
#    &press_key_exit();								#this will exit automatically if $NONINTERACTIVE=1
     $postab_initial =-3;							#indicate failure & do not restart postab
  }
  ($status,$setpointH) = Pezca::GetDouble($postabActH_PV);			#read current H-mirror DAC
  if ($status) {
     &report_only($script,$status,$postabActH_PV);
#    &signal_handler('quit');
#    &press_key_exit();								#this will exit automatically if $NONINTERACTIVE=1
     $postab_initial =-3;							#indicate failure & do not restart postab
  }
  if ($interfacePostab && $postab_initial != -3) {				#no failure
     printf STDOUT $script.':  Vertical  setpoint = %7.3f'."\n",$setpointV;
     printf STDOUT $script.': Horizontal setpoint = %7.3f'."\n",$setpointH;
     $status = &setPointsPostab ($beamline,$setpointH,$setpointV);
#    if ($status) {&signal_handler('quit');}
     if ($status) {$postab_initial = 3;}					#do not start postab
  } else {									#failure
     $postab_initial = 3;							#do not start postab
  }

###=====================================================================
### 10. Start positional feedback (0=run, 3=pause)
  if ($postab_initial != 3) {							#if not paused initially
     print STDOUT $script.': Restoring positional feedback'."\n";
     if ($tell_bluice) {&tellBluIce($scr.': restoring pos.feedback');}
     $status = &restorePostab ($beamline);
#    if ($status) {&press_key_exit();}
  }
# &press_key_continue();


###=====================================================================
### 11. Close fast shutter
# if ($tell_bluice) {&tellBluIce($scr.': closing shutter');}
  &closeShutter ($shutter,$shutter_readback,$shutterAssy_PV);


###=====================================================================
### 12. Restore collimator, pindiode and beamstop:
  if (($move_collimator || $move_pindiode) && $restore_collimator) {
     if ($move_collimator && $align300to10 == 1 && $collimatorPreset <= 1) {	#Preset: 0=none 1=300um
        if    ($collimatorPreset == 0) {					#Preset: 0=none 1=300um
           print STDOUT "\n".$script.': Keeping 10um collimator'."\n";
           if ($tell_bluice) {&tellBluIce($scr.': collimator 10->10 (restore)');}
           $status = Pezca::PutDouble($activeCol_PV,$state10um);
        }
        elsif ($collimatorPreset == 1) {					#Preset: 0=none 1=300um
           print STDOUT "\n".$script.': Restoring active collimator from 10um to 300um'."\n";
           if ($tell_bluice) {&tellBluIce($scr.': collimator 10->300 (restore)');}
           $status = Pezca::PutDouble($activeCol_PV,$state300um);
        }
        &report_only($script, $status, $activeCol_PV);
     }
     @prg = ('IN', $beamline);
     if ($move_collimator) {
        print STDOUT "\n".$script.': Moving collimator IN'."\n";
        if ($tell_bluice) {&tellBluIce($scr.': moving collimator IN');}
     } else {
        push @prg, 'nocollimator';
     }
     if ($move_pindiode) {
        print STDOUT "\n".$script.': Moving beamstop IN and pin diode OUT'."\n";
        if ($tell_bluice) {&tellBluIce($scr.': raising beamstop');}
        push @prg, 'pindiodeOUT';
     } else {
        push @prg, 'nobeamstop';
     }
     $status = &setSampleEnv(\@prg);
     if ($status) {
        print STDOUT "\n".'*** '.$script.': failure='.$status.' from setSampleEnv'."\n\n";
        if ($tell_bluice) {&tellBluIce($scr.': setSampleEnv failed','error');}
        &zenity_error('*** '.$script.': failed on setSampleEnv.');
     }
  }


###=====================================================================
### 13. restore Guard Slits
  if ($usegs || $move_collimator || $move_pindiode) {
     print STDOUT $script.': Restoring guard slits'."\n";
     if ($tell_bluice) {&tellBluIce($scr.': restoring guard slits');}
     $status = &run_guard_slits($gsSizeH_initial,$gsSizeV_initial,0);		#0=track, 1=startonly
     if ($status) {
        if ($tell_bluice) {&tellBluIce($scr.': restoring GS failed','error');}
        &zenity_error('*** '.$script.': failed to restore guard slits.');
     }
  }


###=====================================================================
### 14a. Tell finals to BluIce:
  if ($tell_bluice) {
     if    ($use_CRL) {$str = '---> Final optimized CRLs positions:';         $label='CRL';}
     elsif ($use_TF)  {$str = '---> Final optimized Transfocator positions:'; $label='TF';}
     else             {$str = '---> Final optimized DAC(s):';}
     print STDOUT "\n".'------------------'."\n".$str."\n";
     select(undef,undef,undef,.4);
     &tellBluIce($str,'warn');
     if ($use_CRL || $use_TF) {
         $str = sprintf '  %s_vert=%-.4f',$label,$FinalV;
         print STDOUT $str."\n";
         select(undef,undef,undef,.4);
         &tellBluIce($str,'warn');
         $str = sprintf 'Signal('.$i1_name.') @ final vert.pos. %1.0f%s',$I1_percent_V,'%';
         print STDOUT $str."\n";
         select(undef,undef,undef,.4);
         &tellBluIce($str,'warn');
#
         $str = sprintf '  %s_horiz=%-.4f',$label,$FinalH;
         print STDOUT $str."\n";
         select(undef,undef,undef,.4);
         &tellBluIce($str,'warn');
         $str = sprintf 'Signal('.$i1_name.') @ final horiz.pos. %1.0f%s',$I1_percent_H,'%';
         print STDOUT $str."\n";
         select(undef,undef,undef,.4);
         &tellBluIce($str,'warn');
     } else {
        if ($availableV) {
            $str = sprintf '  VFM=%-.4f',$FinalV;
            print STDOUT $str."\n";
            select(undef,undef,undef,.4);
            &tellBluIce($str,'warn');
            $str = sprintf 'Signal('.$i1_name.') @ final VFM DAC %1.0f%s',$I1_percent_V,'%';
            print STDOUT $str."\n";
            select(undef,undef,undef,.4);
            &tellBluIce($str,'warn');
        }
        if ($availableH) {
            $str = sprintf '  HFM=%-.4f',$FinalH;
            print STDOUT $str."\n";
            select(undef,undef,undef,.4);
            &tellBluIce($str,'warn');
            $str = sprintf 'Signal('.$i1_name.') @ final HFM DAC: %1.0f%s',$I1_percent_H,'%';
            print STDOUT $str."\n";
            select(undef,undef,undef,.4);
            &tellBluIce($str,'warn');
        }
     }
     print STDOUT '------------------'."\n";
  }


###=====================================================================
### 14b. Save to history PV:
  $status = &save_last_alignment($alignHistPV);
  if ($use_TF)  {
     $str = 'Saving aligned transfocator positions for future reference';
     print STDOUT $str."\n";
     select(undef,undef,undef,.4);
     &tellBluIce($str,'warn');
     $status = &saveTFstate();
     if ($status) {
        $str = 'Saving aligned transfocator positions failed with status='.$status.' in "saveTFstate"!';
        print STDOUT '!!! '.$script.': '.$str."\n";
        select(undef,undef,undef,.4);
        &tellBluIce($str,'warn');
     }
  }

###=====================================================================
### 15. Done!

  $elapsed = tv_interval ($t0);									#measure time (in seconds) - exclude sample returning time

  $zzz = $ZENITY_VIA_BLUICE;
  if (! defined $JBLUICE_SERVER || ! $tell_bluice) {$ZENITY_VIA_BLUICE = 0;}
  if ($protectSample == 1 && $pin_mounted) {
     $msg = 'Beam optimization is complete'."\n"
          . 'Now you can put your crystal back into the beam.'."\n";
#    &zenity_info($msg, 'nowait');
     &zenity_info($msg);
  }
  elsif ($protectSample > 1 && defined $sample_moved && $moveSample != 0) {
     if    ($direction eq 'UP')       {$direction = 'DOWN';}
     elsif ($direction eq 'DOWN')     {$direction = 'UP';}
     elsif ($direction eq 'OUTBOARD') {$direction = 'INBOARD';}
     elsif ($direction eq 'INBOARD')  {$direction = 'OUTBOARD';}
     $msg = sprintf 'Moving sample back (%s) by %.1fum.',$direction,abs(1000*$moveSample);
     if ($protectSample == 2) {
        $msg = 'Beam optimization is complete'."\n".$msg;
        &zenity_info($msg, 'nowait');
     } else {
        print STDOUT $script.': '.$msg."\n";
        if ($tell_bluice) {&tellBluIce($scr.': '.$msg);}
     }
     if ($xyzAssyPV eq $zAssyPV) {
        $status = &startNMotors ($xyzAssyPV,$nxyz,\@xyzAxesPV,\@xyzIni);
     } else {
        $status  = &startNMotors ($xyzAssyPV,2,\@xyzAxesPV,\@xyzIni);
        $status += &startNMotors ($zAssyPV,2,[$xyzAxesPV[2]],[$xyzIni[2]]);
     }
     if ($status) {
        $msg = ': Failed to put the pin back into the beam!';
        print STDOUT '!!! '.$script.$msg."\n\n";
        if ($tell_bluice) {&tellBluIce($scr.$msg,'warn');}
     }
  }
  else {
     $msg = ': beam optimization is complete';
     print STDOUT $script.$msg."\n";
     if ($tell_bluice) {&tellBluIce($scr.$msg,'warn'); select(undef,undef,undef,0.2);}	#short sleep
     $msg = ': now you can mount crystals on the goniometer.'."\n";
     print STDOUT $script.$msg."\n";
     if ($tell_bluice) {&tellBluIce($scr.$msg,'warn'); select(undef,undef,undef,1);}	#sleep
  }
  $ZENITY_VIA_BLUICE = $zzz;


  $msg = sprintf ': Beam alignment finished at [%s]. Elapsed time=%.2fmin', scalar(localtime), $elapsed/60.;
  print STDOUT $script.$msg."\n";
  if ($tell_bluice) {&tellBluIce($scr.$msg);}

  &clearAllCaMonitors();

  $SIG{INT}  = sub {die 'Exit on kill request';};
  $SIG{QUIT} = sub {die 'Exit on kill request';};
  if ($tell_bluice) {&tellBluIceBeforeExit($scr.': Aligning finished', '');}
  else              {print STDOUT "\n".$script.': ALL DONE!'."\n\n";}
  if (!$tell_bluice && !$fastScan && $gnuplot && $displaytime > 0) {
     printf STDOUT $script.': This window will auto-close in %1.0fs.'."\n",$displaytime;
     print  STDOUT $script.': Press Ctrl+C to close now.'."\n";
     sleep($displaytime);
  }
# &press_key_exit();								#this will exit automatically if $NONINTERACTIVE=1
  exit 0;

#=================================================================
#=================================================================

sub scan_FPGA_V() {
### Optionally attempt vertical optimization with FPGA
  &setJoerger ($scaler, $cnt_time, $struck);
  @I1_max_now  = &runJoerger ($scaler, $cnt_time, $ninputs, @inputs);
  &subtractJoergerOffsets(\@I1_max_now,\@inputs,\@Offsets);
  $I1_FPGA_before = $I1_max_now[0];
  if ($I1_FPGA_before > 0.05*$I1_last_final_V) {
     printf STDOUT $script.': Trying FPGA positional feedback. Initial intensity=%.1f, previousMax=%.1f'."\n",$I1_FPGA_before, $I1_last_final_V;
     if ($tell_bluice) {&tellBluIce($scr.': Trying FPGA positional feedback');}
     $status = &setVertPosFeedbackFPGAstate($beamline,'Active','string');
     if ($status == 0) {
        sleep 2;								#should I wait for the "Center" state?
        $status = &setVertPosFeedbackFPGAstate($beamline,'Paused','string');
        @I1_max_now  = &runJoerger ($scaler, $cnt_time, $ninputs, @inputs);
        &subtractJoergerOffsets(\@I1_max_now,\@inputs,\@Offsets);
        $I1_FPGA_after = $I1_max_now[0];
        if ($I1_FPGA_after > 0.5*$I1_last_final_V && $I1_FPGA_after > $I1_FPGA_before) {
           printf STDOUT $script.': FPGA positional feedback succeeded. Optimized intensity=%.1f, previousMax=%.1f'."\n",$I1_FPGA_after, $I1_last_final_V;
           if ($tell_bluice) {&tellBluIce($scr.': FPGA positional feedback succeeded.');}
           $verticalHWfeedbackSucceeded = 1;
           &closeShutter ($shutter,$shutter_readback);
           $I1_max_scan_V = $I1_max_final_V = $I1_FPGA_after;
           $I1_percent_V = 100.;						#100%
           $ActV = &readDAC ($dacVFM_PV);
           @x1SR    = ();
           @y1SR    = ();
           @x1DAC   = ();
           @y1DAC   = ();
        } else {
           printf STDOUT '!!! '.$script.': FPGA positional feedback failed. Optimized intensity=%.1f, previousMax=%.1f'."\n",$I1_FPGA_after, $I1_last_final_V;
           if ($tell_bluice) {&tellBluIce($scr.': FPGA positional feedback failed.','error');}
        }
     } else {
        print STDOUT '!!! '.$script.': error='.$status.' while starting FPGA feedback'."\n";
        if ($tell_bluice) {&tellBluIce($scr.': Error starting FPGA positional feedback','error');}
     }
  } else {
     printf STDOUT '!!! '.$script.': Skipping FPGA positional feedback because initial intensity=%.1f is too low (previousMax=%.1f)'."\n",$I1_FPGA_before, $I1_last_final_V;
     if ($tell_bluice) {&tellBluIce($scr.': Skipping FPGA positional feedback','warn');}
  }
  return;
}

#=================================================================

sub scan_KBM_V() {
  my ($datfile, $title, $actPts, @pvlist);
### Scan VFM Pz (DAC) and then move to maximum intensity (+-2V)
  print STDOUT "\n".$script.': Scanning VFM piezo (vertical steering)..'."\n";
  if ($tell_bluice) {&tellBluIce($scr.': scanning VFM piezo');}
  $status = &setupDAC ($dacVFM_PV);
  if ($status) {
     &report_only($script,$status,$dacVFM_PV);
     if ($tell_bluice) {&tellBluIce($scr.': VFM DAC failed','error');}
     &signal_handler('quit');
  }
  ($LoV, $HiV) = &limitsDAC ($dacVFM_PV);
  $ActV = &readDAC ($dacVFM_PV);
  $scanBegin = $ActV - $scan_pmrangeV;
  $scanEnd   = $ActV + $scan_pmrangeV;
  if    ($scanBegin > $HiV) {$scanBegin = $HiV;}
  elsif ($scanBegin < $LoV) {$scanBegin = $LoV;}
  if    ($scanEnd   > $HiV) {$scanEnd   = $HiV;}
  elsif ($scanEnd   < $LoV) {$scanEnd   = $LoV;}
  $SRSCAN_FLAGS->{'has run'} = 0;

  if ($kbm_encoder_present) {
     @pvlist = ($vfm_encoder_PV);
     $encoder_array_name = sprintf 'y%d', ($ninputs+1);
     if ($fastScan) {$encoder_array_name .= 'SR';}
     else           {$encoder_array_name .= 'DAC';}
  } else {
     @pvlist = undef;
  }

  if ($fastScan) {
     if ($scan_stepV != 0) {
        if (($scanEnd - $scanBegin)*$scan_stepV < 0) {
           $tmp      = $scanEnd;
           $scanEnd   = $scanBegin;
           $scanBegin = $tmp;
#          $scan_stepV = -$scan_stepV;
        }
        $npts = int(($scanEnd - $scanBegin)/$scan_stepV + 1.5);
        if ($npts > $npts_max) {
           printf STDOUT '!!! '.$script.': VFM scan npts=%d exceeds buffer(s) capacity=%d.'."\n"
                        .'Reducing to max, but be aware that accuracty will be reduced.'."\n",
                        $npts, $npts_max;
           $npts = $npts_max;
        }
     } else {
        $npts = 1;
     }
     $scanTime = $npts * $cnt_time;
     printf STDOUT $script.': VFM scan npts='.$npts.' countTime='.$cnt_time.' scanTime='.$scanTime
                          .' range=[%.3f:%.3f]V, search for '.$center_V."\n", $scanBegin, $scanEnd;
  }
  $nscans = 0;
  $cnt_factor = 1;
  if    ($center_V eq 'Imax')       {$afterScan = 3;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_V eq 'FWHM')       {$afterScan = 4;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_V eq 'FWHMsmooth') {$afterScan = 5;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_V eq 'centroid')   {$afterScan = 9;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_V eq 'Gauss')      {$afterScan = 10;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_V eq 'Lorentz')    {$afterScan = 11;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  do {	#-------- loop over keithley changes on overload -----------
     if ($kgain < 1.01E3 && $overload_rescan) {					#gain=1E3 -- no room to decrease
        print STDOUT "\n".'!!! '.$script.': No room to decrease '
                         .'Keithley gain on '.$i1_name."\n\n";
        if ($tell_bluice) {&tellBluIce($scr.': no room to decrease '
                                      .'gain on '.$i1_name,'warn');}
        $overload_rescan = 0;
     }
     if ($kbm_encoder_present) {
        no strict 'refs';
        @$encoder_array_name = ();								#clear array for encoder values
        use strict 'refs';
     }  
     if ($fastScan) {
        ### scan returns globals: @x1SR,@y1SR,...@y{$ninputs+$npvs}SR, @Offsets
        @x1SR = ();
        @y1SR = ();
        $scanprg = 'SRflyScanDAC';
        ($status,$slope_V) = &SRflyScanDAC ($dacVFM_PV,$scanBegin,$scanEnd,	#see scanrec_subs.pl
                                 $npts,$cnt_factor*$scanTime,
                                 $afterScan,					#afterScan=4 (drive to peak; if not found then to max)
                                 $medm,$scaler,$fileBkg,
                                 \@inputs,\@pvlist,
                                 $overload_rescan,
                                 $bipostVPV,undef,$adcVFM_PV,
                                 $tolerance_factor*$scan_stepV);
     } else {
        ### scan returns global arrays @x1DAC, @y1DAC, ... @y{$ninputs+$npvs}DAC
        @x1DAC = ();
        @y1DAC = ();
        $scanprg = 'stepScanDAC';
        $status = &stepScanDAC ($dacVFM_PV,$scanBegin,$scanEnd,$scan_stepV,	#see dac_subs.pl
                                $scaler,$cnt_time,
                                \@inputs,\@pvlist,$fileBkg,
                                $gnuplot,$logscale,\@gnuopts,
                                $afterScan,					#afterScan=4 (drive to peak; if not found then to max)
                                $overload_rescan,$gnufile,
                                $bipostVPV,undef,$adcVFM_PV,
                                $tolerance_factor*$scan_stepV);
        $slope_V = 0;
        if ($status == 0 && $gnuplot == 2 && fileno(\*GNUPLOT)) {$gnuopen = 1;}
     }
     if ($status == $OVERLOAD) {						#if scan detected overload
        if ($overload_rescan) {
           ### reduce keithley gain
           $kgain = $kgain/$kgain_divider;					#changed 2013.01.31
           if (abs(1-$kgain/1.E5) < 0.01) {$kgain=1.E4} 			#avoid resonance at 1E5
           printf STDOUT $script.': '.$i1_name.' overload.'
                                .' Reducing Keithley gain to %1.0e'
                                .' and rescanning.'."\n",$kgain;
           if ($tell_bluice) {&tellBluIce($scr.': '.$i1_name.' overload.'
                                  .' Rescanning at lower gain','warn');}
           &keithley_set_gain($kchannel, $kgain);
        } else {
           if ($tell_bluice) {&tellBluIce($scr.': '.$i1_name.' overload. Ignored.','warn');}
        }
     }
     elsif ($status == $FOLLERR) {
        print STDOUT '!!! '.$script.': following error detected.'."\n";
        if ($tell_bluice) {&tellBluIce($scr.': following error detected.','warn');}
        $nscans++;
        if ($nscans <= 3) {
           if ($fastScan) {
              print STDOUT $script.': Reducing speed and rescanning...'."\n";
              if ($tell_bluice) {&tellBluIce($scr.': Reducing speed and rescanning...','warn');}
              $cnt_factor = 2*$cnt_factor;
           } else {
              print STDOUT $script.': Rescanning...'."\n";
              if ($tell_bluice) {&tellBluIce($scr.': Rescanning...','warn');}
           }
        }
     }
  } while (($status == $OVERLOAD && $overload_rescan) ||
           ($status == $FOLLERR && $nscans <= 3));  #--------------------
  if ($status) {
     print STDOUT "\n".'*** '.$script.': failure='.$status
                             .' from '.$scanprg.', dac='.$dacVFM_PV."\n\n";
     if ($tell_bluice) {&tellBluIce($scr.': VFM scan failed','error');}
     &signal_handler('quit');
#    &press_key_exit();								#this will exit automatically if $NONINTERACTIVE=1
  }
  if ($fastScan && $gnuplot) {
     ### Plot scan data:
     $gnuopen = &open_gnuplot ($scanBegin, $scanEnd, $logscale, \@gnuopts);
     if ($gnuopen) {
        $I_lo = +1.E+37;
        $I_hi = -1.E+37;
        $actPts = @x1SR;
        for ($i=0; $i<$actPts; $i++) {
           if ($y1SR[$i] < $I_lo) {$I_lo = $y1SR[$i];}
           if ($y1SR[$i] > $I_hi) {$I_hi = $y1SR[$i];}
        }
        if ($actPts > 2) {
           if ($I_hi <= $I_lo) {printf GNUPLOT 'set yrange [%.3f:%.3f]'."\n",$I_hi-1,$I_lo+1;}
           else                {print  GNUPLOT 'set autoscale y'."\n";}
           $title = 'title "VFM Scan Intensity (scaler-'.$inputs[0].')"';
           if ($gnuplot == 1) {							#image is plot later
              print GNUPLOT 'plot "-" using 1:2 '.$title."\n";
              for ($i=0; $i<$actPts; $i++) {
                 printf GNUPLOT ' %.3f  %.1f'."\n", $x1SR[$i], $y1SR[$i];
                 if ($i == 10*int($i/10)) {&gnuplot_filess_delay();}
              }
              print GNUPLOT 'e'."\n";						#end of data
           }
        } else {
           print STDOUT '!!! '.$script.'(scan_KBM_V): nothing to plot, actPts='.$actPts."\n";
           &close_gnuplot();							#close vertical plot
           $gnuopen = 0;
        }
     }
  }

  if ($tell_bluice) {&tellBluIce($scr.': steering VFM to max');}
  $FinalV = &readDAC ($dacVFM_PV);
  if ($fastScan) {
     if ($tell_bluice) {&tellBluIce($scr.': ...with V-correction='.$corr_fly_V);}
     $FinalV += $corr_fly_V;
     if    ($FinalV > $HiV) {$FinalV = $HiV;}
     elsif ($FinalV < $LoV) {$FinalV = $LoV;}
  }
  ### Compensate for hysteresis
  $delay = 0.5;									#in seconds
  $inlimits = 1;								#bring rqspos into limits
  $hysteresis = $FinalV - 10*$scan_stepV;
  if (!defined $slope_V) {$slope_V = 0;}
  &driveDAC ($dacVFM_PV,$scanBegin,$LoV,$HiV, $delay, $inlimits);
  &driveDAC ($dacVFM_PV,$FinalV,$LoV,$HiV, $delay, $inlimits);
# &driveDAC ($dacVFM_PV,$hysteresis,$LoV,$HiV, $delay, $inlimits);
# &driveDAC_slope ($dacVFM_PV,$FinalV,$LoV,$HiV,$slope_V);
# $npts = int(($FinalV - $scanBegin)/$scan_stepV + 1.5);
# for ($i=$npts-1; $i>=0; $i--) {
#    $dac = $FinalV + $scan_stepV*$i;
#    &driveDAC ($dacVFM_PV,$dac,$LoV,$HiV);
# }
  sleep(4);

### Verify, if we have arrived at maximum:
  if ($tell_bluice) {&tellBluIce($scr.': verifying intensity');}
  if ($fastScan) {$I1_max_scan_V = &findMaxRef(\@y1SR);}
  else           {$I1_max_scan_V = &findMaxRef(\@y1DAC);}
  if ($kbm_encoder_present) {
     if ($fastScan) {$index_peak = &find_closest_index($FinalV,\@x1SR);}
     else           {$index_peak = &find_closest_index($FinalV,\@x1DAC);}
     no strict 'refs';
     $encoder_scanpeak = @$encoder_array_name[$index_peak] * $vfm_enc_scale;
     use strict 'refs';
  }
  if ($I1_max_scan_V < 1.) {$I1_max_scan_V=1.;}
  &setJoerger ($scaler, $cnt_time, $struck);
  &openShutter ($shutter,$shutter_readback);
  @I1_max_now  = &runJoerger ($scaler, $cnt_time, $ninputs, @inputs);
  &closeShutter ($shutter,$shutter_readback);
  &subtractJoergerOffsets(\@I1_max_now,\@inputs,\@Offsets);
  $I1_max_final_V = $I1_max_now[0];
  $I1_percent_V   = 100. * ($I1_max_final_V / $I1_max_scan_V);
  printf STDOUT "\n".$script.': Intensity at final VFM DAC is %1.0f%s'
                    .' of scan maximum (scanmax=%.3g, final=%.3g)'."\n\n",
                    $I1_percent_V,'%', $I1_max_scan_V, $I1_max_final_V;
  if ($kbm_encoder_present) {
     $encoder_finalpeak = &readDAC($vfm_encoder_PV) * $vfm_enc_scale;
     printf STDOUT $script.': VFM Incremental Encoder ['.$vfm_encoder_PV.'] readings:'
                  .' scanpeak=%.3f, finalpeak=%.3f'."\n\n",
                    $encoder_scanpeak, $encoder_finalpeak;
  }
  ### Optionally save data:
  ($actPts,$datfile) = &saveData('V',$use_CRL,$use_TF,$dacVFM_PV,$FinalV,$ActV,$I1_max_scan_V,$I1_max_final_V,$I1_percent_V);
  if ($gnuopen) {
     if ($gnuplot == 2 && -e $datfile && $actPts > 2) {
        $png = $datfile; $png =~ s/\.dat$/\.png/;;
        print GNUPLOT 'plot "'.$datfile .'" using 1:2 '.$title."\n";
        &image_gnuplot($png,undef,undef,1,1);					#$file,$width,$height,$view,$GUIprompt}
     } else {&close_gnuplot();} 						#close vertical plot
     $gnuopen = 0;
  }
  return;
}

#=================================================================

sub scan_KBM_H() {
  my ($datfile, $title, $actPts, @pvlist);
### Scan HFM Pz (DAC) and then move to maximum intensity (+-2V)
  print STDOUT "\n".$script.': Scanning HFM piezo (horizontal steering)..'."\n";
  if ($tell_bluice) {&tellBluIce($scr.': scanning HFM piezo');}
  $status = &setupDAC ($dacHFM_PV);
  if ($status) {
     &report_only($script,$status,$dacHFM_PV);
     if ($tell_bluice) {&tellBluIce($scr.': HFM DAC failed','error');}
     &signal_handler('quit');
#    &press_key_exit();								#this will exit automatically if $NONINTERACTIVE=1
  }
  ($LoH, $HiH) = &limitsDAC ($dacHFM_PV);
  $ActH = &readDAC ($dacHFM_PV);
  $scanBegin = $ActH - $scan_pmrangeH;
  $scanEnd   = $ActH + $scan_pmrangeH;
  if    ($scanBegin > $HiH) {$scanBegin = $HiH;}
  elsif ($scanBegin < $LoH) {$scanBegin = $LoH;}
  if    ($scanEnd   > $HiH) {$scanEnd   = $HiH;}
  elsif ($scanEnd   < $LoH) {$scanEnd   = $LoH;}
  $SRSCAN_FLAGS->{'has run'} = 0;

  if ($kbm_encoder_present) {
     @pvlist = ($hfm_encoder_PV);
     $encoder_array_name = sprintf 'y%d', ($ninputs+1);
     if ($fastScan) {$encoder_array_name .= 'SR';}
     else           {$encoder_array_name .= 'DAC';}
  } else {
     @pvlist = undef;
  }

  if ($fastScan) {
     if ($scan_stepH != 0) {
        if (($scanEnd - $scanBegin)*$scan_stepH < 0) {
           $tmp      = $scanEnd;
           $scanEnd   = $scanBegin;
           $scanBegin = $tmp;
#          $scan_stepH = -$scan_stepH;
        }
        $npts = int(($scanEnd - $scanBegin)/$scan_stepH + 1.5);
        if ($npts > $npts_max) {
           printf STDOUT '!!! '.$script.': HFM scan npts=%d exceeds buffer(s) capacity=%d.'."\n"
                        .'Reducing to max, but be aware that accuracty will be reduced.'."\n",
                        $npts, $npts_max;
           $npts = $npts_max;
        }
     } else {
        $npts = 1;
     }
     $scanTime = $npts * $cnt_time;
     printf STDOUT $script.': HFM scan npts='.$npts.' countTime='.$cnt_time.' scanTime='.$scanTime
                          .' range=[%.3f:%.3f]V'."\n", $scanBegin, $scanEnd;
  }
  $nscans = 0;
  if    ($center_H eq 'Imax')       {$afterScan = 3;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_H eq 'FWHM')       {$afterScan = 4;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_H eq 'FWHMsmooth') {$afterScan = 5;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_H eq 'centroid')   {$afterScan = 9;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_H eq 'Gauss')      {$afterScan = 10;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_H eq 'Lorentz')    {$afterScan = 11;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  do {	#-------- loop over keithley changes on overload -----------
     if ($kgain < 1.01E3 && $overload_rescan) {					#gain=1E3 -- no room to decrease
        print STDOUT "\n".'!!! '.$script.': No room to decrease '
                         .'Keithley gain on '.$i1_name."\n\n";
        if ($tell_bluice) {&tellBluIce($scr.': no room to decrease '
                                      .'gain on '.$i1_name,'warn');}
        $overload_rescan = 0;
     }
     if ($kbm_encoder_present) {
        no strict 'refs';
        @$encoder_array_name = ();								#clear array for encoder values
        use strict 'refs';
     }  
     if ($fastScan) {
        ### scan returns globals: @x1SR,@y1SR,...@y{$ninputs+$npvs}SR, @Offsets
        @x1SR = ();
        @y1SR = ();
        $scanprg = 'SRflyScanDAC';
        ($status,$slope_H) = &SRflyScanDAC ($dacHFM_PV,$scanBegin,$scanEnd,	#see scanrec_subs.pl
                                 $npts,$cnt_factor*$scanTime,
                                 $afterScan,					#afterScan=4 (drive to peak; if not found then to max)
                                 $medm,$scaler,$fileBkg,
                                 \@inputs,\@pvlist,
                                 $overload_rescan,
                                 $bipostHPV,undef,$adcHFM_PV,
                                 $tolerance_factor*$scan_stepH);
#       &press_key_continue();
     } else {
        ### scan returns global arrays @x1DAC, @y1DAC, ... @y{$ninputs+$npvs}DAC
        @x1DAC = ();
        @y1DAC = ();
        $scanprg = 'stepScanDAC';
        $status = &stepScanDAC ($dacHFM_PV,$scanBegin,$scanEnd,$scan_stepH,	#see dac_subs.pl
                                $scaler,$cnt_time,
                                \@inputs,\@pvlist,$fileBkg,
                                $gnuplot,$logscale,\@gnuopts,
                                $afterScan,					#afterScan=4 (drive to peak; if not found then to max)
                                $overload_rescan,$gnufile,
                                $bipostHPV,undef,$adcHFM_PV,
                                $tolerance_factor*$scan_stepH);
        $slope_H = 0;
        if ($status == 0 && $gnuplot == 2 && fileno(\*GNUPLOT)) {$gnuopen = 1;}
     }
     if ($status == $OVERLOAD) {						#if scan detected overload
        if ($overload_rescan) {
           ### reduce keithley gain
           $kgain = $kgain/$kgain_divider;					#changed 2013.01.13
           if (abs(1-$kgain/1.E5) < 0.01) {$kgain=1.E4} 			#avoid resonance at 1E5
           printf STDOUT $script.': '.$i1_name.' overload.'
                                .' Reducing Keithley gain to %1.0e'
                                .' and rescanning.'."\n", $kgain;
           if ($tell_bluice) {&tellBluIce($scr.': '.$i1_name.' overload.'
                                     .' Rescanning at lower gain','warn');}
           &keithley_set_gain($kchannel, $kgain);
        } else {
           if ($tell_bluice) {&tellBluIce($scr.': '.$i1_name.' overload. Ignored.','warn');}
        }
     }
     elsif ($status == $FOLLERR) {
        print STDOUT '!!! '.$script.': following error detected.'."\n";
        if ($tell_bluice) {&tellBluIce($scr.': following error detected.','warn');}
        $nscans++;
        if ($nscans <= 3) {
           if ($fastScan) {
              print STDOUT $script.': Reducing speed and rescanning...'."\n";
              if ($tell_bluice) {&tellBluIce($scr.': Reducing speed and rescanning...','warn');}
              $cnt_factor = 2*$cnt_factor;
           } else {
              print STDOUT $script.': Rescanning...'."\n";
              if ($tell_bluice) {&tellBluIce($scr.': Rescanning...','warn');}
           }
        }
     }
  } while (($status == $OVERLOAD && $overload_rescan) ||
           ($status == $FOLLERR && $nscans <= 3));  #--------------------
  if ($status) {
     print STDOUT "\n".'*** '.$script.': failure='.$status
                 .' from '.$scanprg.', dac='.$dacHFM_PV."\n\n";
     if ($tell_bluice) {&tellBluIce($scr.': HFM scan failed','error');}
     &signal_handler('quit');
#    &press_key_exit();								#this will exit automatically if $NONINTERACTIVE=1
  }
  if ($fastScan && $gnuplot) {
     ### Plot scan data:
     $gnuopen = &open_gnuplot ($scanBegin, $scanEnd, $logscale, \@gnuopts);
     if ($gnuopen) {
        $I_lo = +1.E+37;
        $I_hi = -1.E+37;
        $actPts = @x1SR;
        for ($i=0; $i<$actPts; $i++) {
           if ($y1SR[$i] < $I_lo) {$I_lo = $y1SR[$i];}
           if ($y1SR[$i] > $I_hi) {$I_hi = $y1SR[$i];}
        }
        if ($actPts > 2) {
           if ($I_hi <= $I_lo) {printf GNUPLOT 'set yrange [%.3f:%.3f]'."\n",$I_hi-1,$I_lo+1;}
           else                {print  GNUPLOT 'set autoscale y'."\n";}
           $title = 'title "HFM Scan Intensity (scaler-'.$inputs[0].')"';
           if ($gnuplot == 1) {							#image is plot later
              print GNUPLOT 'plot "-" using 1:2 '.$title."\n";
              for ($i=0; $i<$actPts; $i++) {
                 printf GNUPLOT ' %.3f  %.1f'."\n", $x1SR[$i], $y1SR[$i];
                 if ($i == 10*int($i/10)) {&gnuplot_filess_delay();}
              }
              print GNUPLOT 'e'."\n";						#end of data
           }
        } else {
           print STDOUT '!!! '.$script.'(scan_KBM_H): nothing to plot, actPts='.$actPts."\n";
           &close_gnuplot();							#close vertical plot
           $gnuopen = 0;
        }
     }
  }

  if ($tell_bluice) {&tellBluIce($scr.': steering HFM to max');}
  $FinalH = &readDAC ($dacHFM_PV);
  if ($fastScan) {
     if ($tell_bluice) {&tellBluIce($scr.': ...with H-correction='.$corr_fly_H);}
     $FinalH += $corr_fly_H;
     if    ($FinalH > $HiH) {$FinalH = $HiH;}
     elsif ($FinalH < $LoH) {$FinalH = $LoH;}
  }
### Compensate for hysteresis
  $delay = 0.5;									#in seconds
  $inlimits = 1;								#bring rqspos into limits
  if (!defined $slope_H) {$slope_H = 0;}
  $hysteresis = $FinalH - 10*$scan_stepH;
  &driveDAC ($dacHFM_PV,$scanBegin,$LoH,$HiH, $delay, $inlimits);
  &driveDAC ($dacHFM_PV,$FinalH,$LoH,$HiH, $delay, $inlimits);
# &driveDAC ($dacHFM_PV,$hysteresis,$LoH,$HiH, $delay, $inlimits);
# &driveDAC_slope ($dacHFM_PV,$FinalH,$LoH,$HiH,$slope_H);
# $npts = int(($FinalH - $scanBegin)/$scan_stepH + 1.5);
# for ($i=$npts-1; $i>=0; $i--) {
#   $dac = $FinalH + $scan_stepH*$i;
#   &driveDAC ($dacHFM_PV,$dac,$LoH,$HiH);
# }
# &press_key_continue();
  sleep(4);

### Verify, if we have arrived at maximum:
  if ($tell_bluice) {&tellBluIce($scr.': verifying intensity');}
  if ($fastScan) {$I1_max_scan_H = &findMaxRef(\@y1SR);}
  else           {$I1_max_scan_H = &findMaxRef(\@y1DAC);}
  if ($kbm_encoder_present) {
     if ($fastScan) {$index_peak = &find_closest_index($FinalH,\@x1SR);}
     else           {$index_peak = &find_closest_index($FinalH,\@x1DAC);}
     no strict 'refs';
     $encoder_scanpeak = @$encoder_array_name[$index_peak] * $hfm_enc_scale;
     use strict 'refs';
  }
  if ($I1_max_scan_H < 1.) {$I1_max_scan_H=1.;}
  &setJoerger ($scaler, $cnt_time, $struck);
  &openShutter ($shutter,$shutter_readback);
  @I1_max_now  = &runJoerger ($scaler, $cnt_time, $ninputs, @inputs);
  &closeShutter ($shutter,$shutter_readback);
  &subtractJoergerOffsets(\@I1_max_now,\@inputs,\@Offsets);
  $I1_max_final_H = $I1_max_now[0];
  $I1_percent_H   = 100. * ($I1_max_final_H / $I1_max_scan_H);
  printf STDOUT "\n".$script.':Intensity at final HFM DAC is %1.0f%s'
                    .' of scan maximum (scanmax=%.3g now=%.3g)'."\n\n",
                    $I1_percent_H, '%', $I1_max_scan_H, $I1_max_final_H;
  if ($kbm_encoder_present) {
     $encoder_finalpeak = &readDAC($hfm_encoder_PV) * $hfm_enc_scale;
     printf STDOUT $script.': HFM Incremental Encoder ['.$hfm_encoder_PV.'] readings:'
                  .' scanpeak=%.3f, finalpeak=%.3f'."\n\n",
                    $encoder_scanpeak, $encoder_finalpeak;
  }
  ### Optionally save data:
  ($actPts,$datfile) = &saveData('H',$use_CRL,$use_TF,$dacHFM_PV,$FinalH,$ActH,$I1_max_scan_H,$I1_max_final_H,$I1_percent_H);
  if ($gnuopen) {
     if ($gnuplot == 2 && -e $datfile && $actPts > 2) {
        $png = $datfile; $png =~ s/\.dat$/\.png/;;
        print GNUPLOT 'plot "'.$datfile .'" using 1:2 '.$title."\n";
        &image_gnuplot($png,undef,undef,1,1);					#$file,$width,$height,$view,$GUIprompt}
     } else {&close_gnuplot();} 						#close vertical plot
     $gnuopen = 0;
  }
  return;
}

#=================================================================

sub scan_CRL_V() {
  my ($datfile, $title, $actPts);
### Scan CRL vertical of the small CRL positioner and then move to maximum intensity
  print STDOUT "\n".$script.': Scanning CRL height (vertical position)..'."\n";
  if ($tell_bluice) {&tellBluIce($scr.': scanning CRL height');}
  $status = &configAssy($crl_assy_PV);
  if ($status) {
     print STDOUT "\n".'*** '.$script.': Failed to setup CRL support assembly='.$crl_assy_PV."\n";

     if ($tell_bluice) {&tellBluIce($scr.': CRL support assembly='.$crl_assy_PV.' failed','error');}
     &signal_handler('quit');
  }
# ($LoV, $HiV) = &getAxisDRVLH ($crl_V_PV,\$status);
  $ActV = &getAxisAct($crl_V_PV,\$status);
  $scanBegin = $ActV - $scan_pmrangeV;
  $scanEnd   = $ActV + $scan_pmrangeV;
  if    ($scanBegin > $HiV) {$scanBegin = $HiV;}
  elsif ($scanBegin < $LoV) {$scanBegin = $LoV;}
  if    ($scanEnd   > $HiV) {$scanEnd   = $HiV;}
  elsif ($scanEnd   < $LoV) {$scanEnd   = $LoV;}
  $HDSCAN_FLAGS->{'has run'} = 0;
  if ($scan_stepV != 0) {
     if (($scanEnd - $scanBegin)*$scan_stepV < 0) {
        $tmp      = $scanEnd;
        $scanEnd   = $scanBegin;
        $scanBegin = $tmp;
#       $scan_stepV = -$scan_stepV;
     }
     $npts = int(($scanEnd - $scanBegin)/$scan_stepV + 1.5);
     if ($npts > $npts_max) {
        printf STDOUT '!!! '.$script.': CRL vert. scan npts=%d exceeds buffer(s) capacity=%d.'."\n"
                     .'Reducing to max, but be aware that accuracty will be reduced.'."\n",
                     $npts, $npts_max;
        $npts = $npts_max;
     }
  } else {
     $npts = 1;
  }
  $scanTime = $npts * $cnt_time;
  printf STDOUT $script.': CRL vert. scan npts='.$npts.' countTime='.$cnt_time.' scanTime='.$scanTime
                       .' range=[%.3f:%.3f]mm, search for '.$center_V."\n", $scanBegin, $scanEnd;
  if    ($center_V eq 'Imax')       {$afterScan = 3;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_V eq 'FWHM')       {$afterScan = 4;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_V eq 'FWHMsmooth') {$afterScan = 5;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_V eq 'centroid')   {$afterScan = 9;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_V eq 'Gauss')      {$afterScan = 10;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_V eq 'Lorentz')    {$afterScan = 11;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz

  ($cts2egu,$formula,$mtrOut,$struckPs,$struckNg) = &searchMotorList($crl_assy_PV,$crl_V_PV,'harsync');
  if (!defined $formula) {
     print STDOUT "\n".'*** '.$script.': Failed to find fastscan formula for '.$crl_V_PV."\n";
     if ($tell_bluice) {&tellBluIce($scr.': Failed to find fastscan formula for '.$crl_V_PV,'error');}
     &signal_handler('quit');
  }

  do {	#-------- loop over keithley changes on overload -----------
     if ($kgain < 1.01E3 && $overload_rescan) {					#gain=1E3 -- no room to decrease
        print STDOUT "\n".'!!! '.$script.': No room to decrease '
                         .'Keithley gain on '.$i1_name."\n\n";
        if ($tell_bluice) {&tellBluIce($scr.': no room to decrease '
                                      .'gain on '.$i1_name,'warn');}
        $overload_rescan = 0;
     }
     ### scan returns globals: @x1HD,@y1HD,...@y{$ninputs}HD
     @x1HD = ();
     @y1HD = ();
     $scanprg = 'HDflyScan';
     ($status,$actPts) = &HDflyScan($crl_assy_PV,$crl_V_PV,			#see struck_subs_v2.pl
                                 $scanBegin,$scanEnd,$npts,$scanTime,
                                 $mtrOut,$cts2egu,$formula,
                                 $struckPs,$struckNg,
                                 $medm,$fileBkg,$ninputs,\@inputs,
                                 $scaler,$afterScan,$overload_rescan,		#afterScan=4 (drive to peak; if not found then to max)
                                 $bipostVPV);
     if ($status == $OVERLOAD) {						#if scan detected overload
        if ($overload_rescan) {
           ### reduce keithley gain
           $kgain = $kgain/$kgain_divider;					#changed 2013.01.31
           if (abs(1-$kgain/1.E5) < 0.01) {$kgain=1.E4} 			#avoid resonance at 1E5
           printf STDOUT $script.': '.$i1_name.' overload.'
                                .' Reducing Keithley gain to %1.0e'
                                .' and rescanning.'."\n",$kgain;
           if ($tell_bluice) {&tellBluIce($scr.': '.$i1_name.' overload.'
                                  .' Rescanning at lower gain','warn');}
           &keithley_set_gain($kchannel, $kgain);
        } else {
           if ($tell_bluice) {&tellBluIce($scr.': '.$i1_name.' overload. Ignored.','warn');}
        }
     }
  } while (($status == $OVERLOAD && $overload_rescan));		#--------------------
  if ($status) {
     print STDOUT "\n".'*** '.$script.': failure='.$status
                             .' from '.$scanprg.', axis='.$crl_V_PV."\n\n";
     if ($tell_bluice) {&tellBluIce($scr.': CRL vert. scan failed','error');}
     &signal_handler('quit');
#    &press_key_exit();								#this will exit automatically if $NONINTERACTIVE=1
  }
  if ($gnuplot) {
     ### Plot scan data:
     $gnuopen = &open_gnuplot ($scanBegin, $scanEnd, $logscale, \@gnuopts);
     if ($gnuopen) {
        $I_lo = +1.E+37;
        $I_hi = -1.E+37;
        $actPts = @x1HD;
        for ($i=0; $i<$actPts; $i++) {
           if ($y1HD[$i] < $I_lo) {$I_lo = $y1HD[$i];}
           if ($y1HD[$i] > $I_hi) {$I_hi = $y1HD[$i];}
        }
        if ($actPts > 2) {
           if ($I_hi <= $I_lo) {printf GNUPLOT 'set yrange [%.3f:%.3f]'."\n",$I_hi-1,$I_lo+1;}
           else                {print  GNUPLOT 'set autoscale y'."\n";}
           $title = 'title "CRL Vert. Scan Intensity (scaler input-'.$inputs[0].')"';
           if ($gnuplot == 1) {							#image is plot later
              print GNUPLOT 'plot "-" using 1:2 '.$title."\n";
              for ($i=0; $i<$actPts; $i++) {
                 printf GNUPLOT ' %.3f  %.1f'."\n", $x1HD[$i], $y1HD[$i];
                 if ($i == 10*int($i/10)) {&gnuplot_filess_delay();}
              }
              print GNUPLOT 'e'."\n"	;					#end of data
           }
        } else {
           print STDOUT '!!! '.$script.'(scan_CRL_V): nothing to plot, actPts='.$actPts."\n";
           &close_gnuplot();							#close vertical plot
           $gnuopen = 0;
        }
     }
  }
  if ($tell_bluice) {&tellBluIce($scr.': steering CRL vert. to max');}
  $FinalV = &getAxisAct($crl_V_PV,\$status);
  if ($tell_bluice) {&tellBluIce($scr.': ...with V-correction='.$corr_fly_V);}
  $FinalV += $corr_fly_V;
  if    ($FinalV > $HiV) {$FinalV = $HiV;}
  elsif ($FinalV < $LoV) {$FinalV = $LoV;}
  $status = &runMotor ($crl_assy_PV, $crl_V_PV, $FinalV, 'checkAuto');
  sleep(1);

### Verify, if we have arrived at maximum:
  if ($tell_bluice) {&tellBluIce($scr.': verifying intensity');}
  $I1_max_scan_V = &findMaxRef(\@y1HD);
  if ($I1_max_scan_V < 1.) {$I1_max_scan_V=1.;}
  &setJoerger ($scaler, $cnt_time, $struck);
  &openShutter ($shutter,$shutter_readback);
  @I1_max_now  = &runJoerger ($scaler, $cnt_time, $ninputs, @inputs);
  &closeShutter ($shutter,$shutter_readback);
  &subtractJoergerOffsets(\@I1_max_now,\@inputs,\@Offsets);
  $I1_max_final_V = $I1_max_now[0];
  $I1_percent_V   = 100. * ($I1_max_final_V / $I1_max_scan_V);
  printf STDOUT "\n".$script.': Intensity at final CRL-vert. is %1.0f%s'
                    .' of scan maximum (scanmax=%.3g, now=%.3g)'."\n\n",
                    $I1_percent_V,'%', $I1_max_scan_V, $I1_max_final_V;
  ### Optionally save data:
  ($actPts,$datfile) = &saveData('V',$use_CRL,$use_TF,$crl_V_PV,$FinalV,$ActV,$I1_max_scan_V,$I1_max_final_V,$I1_percent_V);
  if ($gnuopen) {
     if ($gnuplot == 2 && -e $datfile && $actPts > 2) {
        $png = $datfile; $png =~ s/\.dat$/\.png/;;
        print GNUPLOT 'plot "'.$datfile .'" using 1:2 '.$title."\n";
        &image_gnuplot($png,undef,undef,1,1);					#$file,$width,$height,$view,$GUIprompt}
     } else {&close_gnuplot();} 						#close vertical plot
     $gnuopen = 0;
  }
  return;
}

#=================================================================

sub scan_CRL_H() {
  my ($datfile, $title, $actPts);
### Scan CRL horizontal of the small CRL positioner and then move to maximum intensity
  print STDOUT "\n".$script.': Scanning CRL outboard (horizontal position)..'."\n";
  if ($tell_bluice) {&tellBluIce($scr.': scanning CRL horizontal');}
  $status = &configAssy($crl_assy_PV);
  if ($status) {
     print STDOUT "\n".'*** '.$script.': Failed to setup CRL support assembly='.$crl_assy_PV."\n";

     if ($tell_bluice) {&tellBluIce($scr.': CRL support assembly='.$crl_assy_PV.' failed','error');}
     &signal_handler('quit');
  }
# ($LoH, $HiH) = &getAxisDRVLH ($crl_H_PV,\$status);
  $ActH = &getAxisAct($crl_H_PV,\$status);
  $scanBegin = $ActH - $scan_pmrangeH;
  $scanEnd   = $ActH + $scan_pmrangeH;
  if    ($scanBegin > $HiH) {$scanBegin = $HiH;}
  elsif ($scanBegin < $LoH) {$scanBegin = $LoH;}
  if    ($scanEnd   > $HiH) {$scanEnd   = $HiH;}
  elsif ($scanEnd   < $LoH) {$scanEnd   = $LoH;}
  $HDSCAN_FLAGS->{'has run'} = 0;
  if ($scan_stepH != 0) {
     if (($scanEnd - $scanBegin)*$scan_stepH < 0) {
        $tmp      = $scanEnd;
        $scanEnd   = $scanBegin;
        $scanBegin = $tmp;
#       $scan_stepH = -$scan_stepH;
     }
     $npts = int(($scanEnd - $scanBegin)/$scan_stepH + 1.5);
     if ($npts > $npts_max) {
        printf STDOUT '!!! '.$script.': CRL horiz. scan npts=%d exceeds buffer(s) capacity=%d.'."\n"
                     .'Reducing to max, but be aware that accuracty will be reduced.'."\n",
                     $npts, $npts_max;
        $npts = $npts_max;
     }
  } else {
     $npts = 1;
  }
  $scanTime = $npts * $cnt_time;
  printf STDOUT $script.': CRL horiz. scan npts='.$npts.' countTime='.$cnt_time.' scanTime='.$scanTime
                       .' range=[%.3f:%.3f]mm, search for '.$center_H."\n", $scanBegin, $scanEnd;
  if    ($center_H eq 'Imax')       {$afterScan = 3;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_H eq 'FWHM')       {$afterScan = 4;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_H eq 'FWHMsmooth') {$afterScan = 5;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_H eq 'centroid')   {$afterScan = 9;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_H eq 'Gauss')      {$afterScan = 10;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_H eq 'Lorentz')    {$afterScan = 11;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz

  ($cts2egu,$formula,$mtrOut,$struckPs,$struckNg) = &searchMotorList($crl_assy_PV,$crl_H_PV,'harsync');
  if (!defined $formula) {
     print STDOUT "\n".'*** '.$script.': Failed to find fastscan formula for '.$crl_H_PV."\n";
     if ($tell_bluice) {&tellBluIce($scr.': Failed to find fastscan formula for '.$crl_H_PV,'error');}
     &signal_handler('quit');
  }

  do {	#-------- loop over keithley changes on overload -----------
     if ($kgain < 1.01E3 && $overload_rescan) {					#gain=1E3 -- no room to decrease
        print STDOUT "\n".'!!! '.$script.': No room to decrease '
                         .'Keithley gain on '.$i1_name."\n\n";
        if ($tell_bluice) {&tellBluIce($scr.': no room to decrease '
                                      .'gain on '.$i1_name,'warn');}
        $overload_rescan = 0;
     }
     ### scan returns globals: @x1HD,@y1HD,...@y{$ninputs}HD
     @x1HD = ();
     @y1HD = ();
     $scanprg = 'HDflyScan';
     ($status,$actPts) = &HDflyScan($crl_assy_PV,$crl_H_PV,			#see struck_subs_v2.pl
                                 $scanBegin,$scanEnd,$npts,$scanTime,
                                 $mtrOut,$cts2egu,$formula,
                                 $struckPs,$struckNg,
                                 $medm,$fileBkg,$ninputs,\@inputs,
                                 $scaler,$afterScan,$overload_rescan,		#afterScan=4 (drive to peak; if not found then to max)
                                 $bipostHPV);
     if ($status == $OVERLOAD) {						#if scan detected overload
        if ($overload_rescan) {
           ### reduce keithley gain
           $kgain = $kgain/$kgain_divider;					#changed 2013.01.31
           if (abs(1-$kgain/1.E5) < 0.01) {$kgain=1.E4} 			#avoid resonance at 1E5
           printf STDOUT $script.': '.$i1_name.' overload.'
                                .' Reducing Keithley gain to %1.0e'
                                .' and rescanning.'."\n",$kgain;
           if ($tell_bluice) {&tellBluIce($scr.': '.$i1_name.' overload.'
                                  .' Rescanning at lower gain','warn');}
           &keithley_set_gain($kchannel, $kgain);
        } else {
           if ($tell_bluice) {&tellBluIce($scr.': '.$i1_name.' overload. Ignored.','warn');}
        }
     }
  } while (($status == $OVERLOAD && $overload_rescan));  #--------------------
  if ($status) {
     print STDOUT "\n".'*** '.$script.': failure='.$status
                             .' from '.$scanprg.', axis='.$crl_H_PV."\n\n";
     if ($tell_bluice) {&tellBluIce($scr.': CRL horiz. scan failed','error');}
     &signal_handler('quit');
#    &press_key_exit();								#this will exit automatically if $NONINTERACTIVE=1
  }
  if ($gnuplot) {
     ### Plot scan data:
     $gnuopen = &open_gnuplot ($scanBegin, $scanEnd, $logscale, \@gnuopts);
     if ($gnuopen) {
        $I_lo = +1.E+37;
        $I_hi = -1.E+37;
        $actPts = @x1HD;
        for ($i=0; $i<$actPts; $i++) {
           if ($y1HD[$i] < $I_lo) {$I_lo = $y1HD[$i];}
           if ($y1HD[$i] > $I_hi) {$I_hi = $y1HD[$i];}
        }
        if ($actPts > 2) {
           if ($I_hi <= $I_lo) {printf GNUPLOT 'set yrange [%.3f:%.3f]'."\n",$I_hi-1,$I_lo+1;}
           else                {print  GNUPLOT 'set autoscale y'."\n";}
           $title = 'title "CRL Horiz. Scan Intensity (scaler input-'.$inputs[0].')"';
           if ($gnuplot == 1) {							#image is plot later
              print GNUPLOT 'plot "-" using 1:2 '.$title."\n";
              for ($i=0; $i<$actPts; $i++) {
                 printf GNUPLOT ' %.3f  %.1f'."\n", $x1HD[$i], $y1HD[$i];
                 if ($i == 10*int($i/10)) {&gnuplot_filess_delay();}
              }
              print GNUPLOT 'e'."\n";						#end of data
           }
        } else {
           print STDOUT '!!! '.$script.'(scan_CRL_H): nothing to plot, actPts='.$actPts."\n";
           &close_gnuplot();							#close vertical plot
           $gnuopen = 0;
        }
     }
  }
  if ($tell_bluice) {&tellBluIce($scr.': steering CRL horiz. to max');}
  $FinalH = &getAxisAct($crl_H_PV,\$status);
  if ($tell_bluice) {&tellBluIce($scr.': ...with H-correction='.$corr_fly_H);}
  $FinalH += $corr_fly_H;
  if    ($FinalH > $HiH) {$FinalH = $HiH;}
  elsif ($FinalH < $LoH) {$FinalH = $LoH;}
  $status = &runMotor ($crl_assy_PV, $crl_H_PV, $FinalH, 'checkAuto');
  sleep(1);

### Verify, if we have arrived at maximum:
  if ($tell_bluice) {&tellBluIce($scr.': verifying intensity');}
  $I1_max_scan_H = &findMaxRef(\@y1HD);
  if ($I1_max_scan_H < 1.) {$I1_max_scan_H=1.;}
  &setJoerger ($scaler, $cnt_time, $struck);
  &openShutter ($shutter,$shutter_readback);
  @I1_max_now  = &runJoerger ($scaler, $cnt_time, $ninputs, @inputs);
  &closeShutter ($shutter,$shutter_readback);
  &subtractJoergerOffsets(\@I1_max_now,\@inputs,\@Offsets);
  $I1_max_final_H = $I1_max_now[0];
  $I1_percent_H   = 100. * ($I1_max_final_H / $I1_max_scan_H);
  printf STDOUT "\n".$script.': Intensity at final CRL horiz. is %1.0f%s'
                    .' of scan maximum (scanmax=%.3g, now=%.3g)'."\n\n",
                    $I1_percent_H,'%', $I1_max_scan_H, $I1_max_final_H;
  ### Optionally save data:
  ($actPts,$datfile) = &saveData('H',$use_CRL,$use_TF,$crl_H_PV,$FinalH,$ActH,$I1_max_scan_H,$I1_max_final_H,$I1_percent_H);
  if ($gnuopen) {
     if ($gnuplot == 2 && -e $datfile && $actPts > 2) {
        $png = $datfile; $png =~ s/\.dat$/\.png/;;
        print GNUPLOT 'plot "'.$datfile .'" using 1:2 '.$title."\n";
        &image_gnuplot($png,undef,undef,1,1);					#$file,$width,$height,$view,$GUIprompt}
     } else {&close_gnuplot();} 						#close vertical plot
     $gnuopen = 0;
  }
  return;
}

#=================================================================

sub scan_TF_V() {
  my ($datfile, $title, $actPts);
### Scan Transfocator vertical and then move to maximum intensity
  print STDOUT "\n".$script.': Scanning Transfocator height (vertical position)..'."\n";
  if ($tell_bluice) {&tellBluIce($scr.': scanning Transfocator vertical');}
  $status = &configAssy($tf_V_assy_PV);
  if ($status) {
     print STDOUT "\n".'*** '.$script.': Failed to setup Transfocator support assembly='.$tf_V_assy_PV."\n";

     if ($tell_bluice) {&tellBluIce($scr.': Transfocator support assembly='.$tf_V_assy_PV.' failed','error');}
     &signal_handler('quit');
  }
# ($LoV, $HiV) = &getAxisDRVLH ($tf_V_axis_PV,\$status);
  $ActV = &getAxisAct($tf_V_axis_PV,\$status);
  $scanBegin = $ActV - $scan_pmrangeV;
  $scanEnd   = $ActV + $scan_pmrangeV;
  if    ($scanBegin > $HiV) {$scanBegin = $HiV;}
  elsif ($scanBegin < $LoV) {$scanBegin = $LoV;}
  if    ($scanEnd   > $HiV) {$scanEnd   = $HiV;}
  elsif ($scanEnd   < $LoV) {$scanEnd   = $LoV;}
  $SRSCAN_FLAGS->{'has run'} = 0;
  if ($scan_stepV != 0) {
     if (($scanEnd - $scanBegin)*$scan_stepV < 0) {
        $tmp      = $scanEnd;
        $scanEnd   = $scanBegin;
        $scanBegin = $tmp;
#       $scan_stepV = -$scan_stepV;
     }
     $npts = int(($scanEnd - $scanBegin)/$scan_stepV + 1.5);
     if ($npts > $npts_max) {
        printf STDOUT '!!! '.$script.': Transfocator vert. scan npts=%d exceeds buffer(s) capacity=%d.'."\n"
                     .'Reducing to max, but be aware that accuracty will be reduced.'."\n",
                     $npts, $npts_max;
        $npts = $npts_max;
     }
  } else {
     $npts = 1;
  }
  $scanTime = $npts * $cnt_time;
  printf STDOUT $script.': Transfocator vert. scan npts='.$npts.' countTime='.$cnt_time.' scanTime='.$scanTime
                       .' range=[%.3f:%.3f]mm, search for '.$center_V."\n", $scanBegin, $scanEnd;
  if    ($center_V eq 'Imax')       {$afterScan = 3;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_V eq 'FWHM')       {$afterScan = 4;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_V eq 'FWHMsmooth') {$afterScan = 5;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_V eq 'centroid')   {$afterScan = 9;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_V eq 'Gauss')      {$afterScan = 10;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_V eq 'Lorentz')    {$afterScan = 11;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz

  do {	#-------- loop over keithley changes on overload -----------
     if ($kgain < 1.01E3 && $overload_rescan) {					#gain=1E3 -- no room to decrease
        print STDOUT "\n".'!!! '.$script.': No room to decrease '
                         .'Keithley gain on '.$i1_name."\n\n";
        if ($tell_bluice) {&tellBluIce($scr.': no room to decrease '
                                      .'gain on '.$i1_name,'warn');}
        $overload_rescan = 0;
     }
     ### scan returns globals: @x1SR,@y1SR,...@y{$ninputs}SR
     @x1SR = ();
     @y1SR = ();
     $scanprg = 'SRflyScan';
     ($status,$actPts) = &SRflyScan2($tf_V_assy_PV,$tf_V_axis_PV,
                                     $scanBegin,$scanEnd,$npts,$scanTime,
                                     $medm,$scaler,$fileBkg,\@inputs,undef,
                                     $afterScan,$overload_rescan,
                                     $bipostVPV);
     if ($status == $OVERLOAD) {						#if scan detected overload
        if ($overload_rescan) {
           ### reduce keithley gain
           $kgain = $kgain/$kgain_divider;					#changed 2013.01.31
           if (abs(1-$kgain/1.E5) < 0.01) {$kgain=1.E4} 			#avoid resonance at 1E5
           printf STDOUT $script.': '.$i1_name.' overload.'
                                .' Reducing Keithley gain to %1.0e'
                                .' and rescanning.'."\n",$kgain;
           if ($tell_bluice) {&tellBluIce($scr.': '.$i1_name.' overload.'
                                  .' Rescanning at lower gain','warn');}
           &keithley_set_gain($kchannel, $kgain);
        } else {
           if ($tell_bluice) {&tellBluIce($scr.': '.$i1_name.' overload. Ignored.','warn');}
        }
     }
  } while (($status == $OVERLOAD && $overload_rescan));  #--------------------
  if ($status) {
     print STDOUT "\n".'*** '.$script.': failure='.$status
                             .' from '.$scanprg.', axis='.$tf_V_axis_PV."\n\n";
     if ($tell_bluice) {&tellBluIce($scr.': Transfocator vert. scan failed','error');}
     &signal_handler('quit');
#    &press_key_exit();								#this will exit automatically if $NONINTERACTIVE=1
  }
  if ($gnuplot) {
     ### Plot scan data:
     $gnuopen = &open_gnuplot ($scanBegin, $scanEnd, $logscale, \@gnuopts);
     if ($gnuopen) {
        $I_lo = +1.E+37;
        $I_hi = -1.E+37;
        $actPts = @x1SR;
        for ($i=0; $i<$actPts; $i++) {
           if ($y1SR[$i] < $I_lo) {$I_lo = $y1SR[$i];}
           if ($y1SR[$i] > $I_hi) {$I_hi = $y1SR[$i];}
        }
        if ($actPts > 2) {
           if ($I_hi <= $I_lo) {printf GNUPLOT 'set yrange [%.3f:%.3f]'."\n",$I_hi-1,$I_lo+1;}
           else                {print  GNUPLOT 'set autoscale y'."\n";}
           $title = 'title "Transfocator Vert. Scan Intensity (scaler input-'.$inputs[0].')"';
           if ($gnuplot == 1) {							#image is plot later
              print GNUPLOT 'plot "-" using 1:2 '.$title."\n";
              for ($i=0; $i<$actPts; $i++) {
                 printf GNUPLOT ' %.3f  %.1f'."\n", $x1SR[$i], $y1SR[$i];
                 if ($i == 10*int($i/10)) {&gnuplot_filess_delay();}
              }
              print GNUPLOT 'e'."\n";						#end of data
           }
        } else {
           print STDOUT '!!! '.$script.'(scan_TF_V): nothing to plot, actPts='.$actPts."\n";
           &close_gnuplot();							#close vertical plot
           $gnuopen = 0;
        }
     }
  }
  if ($tell_bluice) {&tellBluIce($scr.': steering Transfocator vert. to max');}
  $FinalV = &getAxisAct($tf_V_axis_PV,\$status);
  if ($tell_bluice) {&tellBluIce($scr.': ...with V-correction='.$corr_fly_V);}
  $FinalV += $corr_fly_V;
  if    ($FinalV > $HiV) {$FinalV = $HiV;}
  elsif ($FinalV < $LoV) {$FinalV = $LoV;}
  $status = &runMotor ($tf_V_assy_PV, $tf_V_axis_PV, $FinalV, 'checkAuto');
  sleep(1);

### Verify, if we have arrived at maximum:
  if ($tell_bluice) {&tellBluIce($scr.': verifying intensity');}
  $I1_max_scan_V = &findMaxRef(\@y1SR);
  if ($I1_max_scan_V < 1.) {$I1_max_scan_V=1.;}
  &setJoerger ($scaler, $cnt_time, $struck);
  &openShutter ($shutter,$shutter_readback);
  @I1_max_now  = &runJoerger ($scaler, $cnt_time, $ninputs, @inputs);
  &closeShutter ($shutter,$shutter_readback);
  &subtractJoergerOffsets(\@I1_max_now,\@inputs,\@Offsets);
  $I1_max_final_V = $I1_max_now[0];
  $I1_percent_V   = 100. * ($I1_max_final_V / $I1_max_scan_V);
  printf STDOUT "\n".$script.': Intensity at final Transfocator-vert. is %1.0f%s'
                    .' of scan maximum (scanmax=%.3g, now=%.3g)'."\n\n",
                    $I1_percent_V,'%', $I1_max_scan_V, $I1_max_final_V;
  ### Optionally save data:
  ($actPts,$datfile) = &saveData('V',$use_CRL,$use_TF,$tf_V_axis_PV,$FinalV,$ActV,$I1_max_scan_V,$I1_max_final_V,$I1_percent_V);
  if ($gnuopen) {
     if ($gnuplot == 2 && -e $datfile && $actPts > 2) {
        $png = $datfile; $png =~ s/\.dat$/\.png/;;
        print GNUPLOT 'plot "'.$datfile .'" using 1:2 '.$title."\n";
        &image_gnuplot($png,undef,undef,1,1);					#$file,$width,$height,$view,$GUIprompt}
     } else {&close_gnuplot();} 						#close vertical plot
     $gnuopen = 0;
  }
  return;
}

#=================================================================

sub scan_TF_H() {
  my ($datfile, $title, $actPts);
### Scan Transfocator horizontal positioner and then move to maximum intensity
  print STDOUT "\n".$script.': Scanning Transfocator outboard (horizontal position)..'."\n";
  if ($tell_bluice) {&tellBluIce($scr.': scanning Transfocator horizontal');}
  $status = &configAssy($tf_H_assy_PV);
  if ($status) {
     print STDOUT "\n".'*** '.$script.': Failed to setup Transfocator support assembly='.$tf_H_assy_PV."\n";

     if ($tell_bluice) {&tellBluIce($scr.': Transfocator support assembly='.$tf_H_assy_PV.' failed','error');}
     &signal_handler('quit');
  }
# ($LoH, $HiH) = &getAxisDRVLH ($tf_H_axis_PV,\$status);
  $ActH = &getAxisAct($tf_H_axis_PV,\$status);
  $scanBegin = $ActH - $scan_pmrangeH;
  $scanEnd   = $ActH + $scan_pmrangeH;
  if    ($scanBegin > $HiH) {$scanBegin = $HiH;}
  elsif ($scanBegin < $LoH) {$scanBegin = $LoH;}
  if    ($scanEnd   > $HiH) {$scanEnd   = $HiH;}
  elsif ($scanEnd   < $LoH) {$scanEnd   = $LoH;}
  $SRSCAN_FLAGS->{'has run'} = 0;
  if ($scan_stepH != 0) {
     if (($scanEnd - $scanBegin)*$scan_stepH < 0) {
        $tmp      = $scanEnd;
        $scanEnd   = $scanBegin;
        $scanBegin = $tmp;
#       $scan_stepH = -$scan_stepH;
     }
     $npts = int(($scanEnd - $scanBegin)/$scan_stepH + 1.5);
     if ($npts > $npts_max) {
        printf STDOUT '!!! '.$script.': Transfocator horiz. scan npts=%d exceeds buffer(s) capacity=%d.'."\n"
                     .'Reducing to max, but be aware that accuracty will be reduced.'."\n",
                     $npts, $npts_max;
        $npts = $npts_max;
     }
  } else {
     $npts = 1;
  }
  $scanTime = $npts * $cnt_time;
  printf STDOUT $script.': Transfocator horiz. scan npts='.$npts.' countTime='.$cnt_time.' scanTime='.$scanTime
                       .' range=[%.3f:%.3f]mm, search for '.$center_H."\n", $scanBegin, $scanEnd;
  if    ($center_H eq 'Imax')       {$afterScan = 3;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_H eq 'FWHM')       {$afterScan = 4;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_H eq 'FWHMsmooth') {$afterScan = 5;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_H eq 'centroid')   {$afterScan = 9;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_H eq 'Gauss')      {$afterScan = 10;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz
  elsif ($center_H eq 'Lorentz')    {$afterScan = 11;}				#0=stay 1=start 2=prior 3=max 4=noisypeak 5=smoothpeak 6=min 7=noisyvalley 8=smoothvalley 9=centroid 10=gauss 11=lorentz

  do {	#-------- loop over keithley changes on overload -----------
     if ($kgain < 1.01E3 && $overload_rescan) {					#gain=1E3 -- no room to decrease
        print STDOUT "\n".'!!! '.$script.': No room to decrease '
                         .'Keithley gain on '.$i1_name."\n\n";
        if ($tell_bluice) {&tellBluIce($scr.': no room to decrease '
                                      .'gain on '.$i1_name,'warn');}
        $overload_rescan = 0;
     }
     ### scan returns globals: @x1SR,@y1SR,...@y{$ninputs}SR
     @x1SR = ();
     @y1SR = ();
     $scanprg = 'SRflyScan';
     ($status,$actPts) = &SRflyScan2($tf_H_assy_PV,$tf_H_axis_PV,
                                     $scanBegin,$scanEnd,$npts,$scanTime,
                                     $medm,$scaler,$fileBkg,\@inputs,undef,
                                     $afterScan,$overload_rescan,
                                     $bipostHPV);
     if ($status == $OVERLOAD) {						#if scan detected overload
        if ($overload_rescan) {
           ### reduce keithley gain
           $kgain = $kgain/$kgain_divider;					#changed 2013.01.31
           if (abs(1-$kgain/1.E5) < 0.01) {$kgain=1.E4} 			#avoid resonance at 1E5
           printf STDOUT $script.': '.$i1_name.' overload.'
                                .' Reducing Keithley gain to %1.0e'
                                .' and rescanning.'."\n",$kgain;
           if ($tell_bluice) {&tellBluIce($scr.': '.$i1_name.' overload.'
                                  .' Rescanning at lower gain','warn');}
           &keithley_set_gain($kchannel, $kgain);
        } else {
           if ($tell_bluice) {&tellBluIce($scr.': '.$i1_name.' overload. Ignored.','warn');}
        }
     }
  } while (($status == $OVERLOAD && $overload_rescan));  #--------------------
  if ($status) {
     print STDOUT "\n".'*** '.$script.': failure='.$status
                             .' from '.$scanprg.', axis='.$tf_H_axis_PV."\n\n";
     if ($tell_bluice) {&tellBluIce($scr.': Transfocator horiz. scan failed','error');}
     &signal_handler('quit');
#    &press_key_exit();								#this will exit automatically if $NONINTERACTIVE=1
  }
  if ($gnuplot) {
     ### Plot scan data:
     $gnuopen = &open_gnuplot ($scanBegin, $scanEnd, $logscale, \@gnuopts);
     if ($gnuopen) {
        $I_lo = +1.E+37;
        $I_hi = -1.E+37;
        $actPts = @x1SR;
        for ($i=0; $i<$actPts; $i++) {
           if ($y1SR[$i] < $I_lo) {$I_lo = $y1SR[$i];}
           if ($y1SR[$i] > $I_hi) {$I_hi = $y1SR[$i];}
        }
        if ($actPts > 2) {
           if ($I_hi <= $I_lo) {printf GNUPLOT 'set yrange [%.3f:%.3f]'."\n",$I_hi-1,$I_lo+1;}
           else                {print  GNUPLOT 'set autoscale y'."\n";}
           $title = 'title "Transfocator Horiz. Scan Intensity (scaler input-'.$inputs[0].')"';
           if ($gnuplot == 1) {							#image is plot later
              print GNUPLOT 'plot "-" using 1:2 '.$title."\n";
              for ($i=0; $i<$actPts; $i++) {
                 printf GNUPLOT ' %.3f  %.1f'."\n", $x1SR[$i], $y1SR[$i];
                 if ($i == 10*int($i/10)) {&gnuplot_filess_delay();}
              }
              print GNUPLOT 'e'."\n";						#end of data
           }
        } else {
           print STDOUT '!!! '.$script.'(scan_TF_H): nothing to plot, actPts='.$actPts."\n";
           &close_gnuplot();							#close vertical plot
           $gnuopen = 0;
        }
     }
  }
  if ($tell_bluice) {&tellBluIce($scr.': steering Transfocator horiz. to max');}
  $FinalH = &getAxisAct($tf_H_axis_PV,\$status);
  if ($tell_bluice) {&tellBluIce($scr.': ...with H-correction='.$corr_fly_H);}
  $FinalH += $corr_fly_H;
  if    ($FinalH > $HiH) {$FinalH = $HiH;}
  elsif ($FinalH < $LoH) {$FinalH = $LoH;}
  $status = &runMotor ($tf_H_assy_PV, $tf_H_axis_PV, $FinalH, 'checkAuto');
  sleep(1);

### Verify, if we have arrived at maximum:
  if ($tell_bluice) {&tellBluIce($scr.': verifying intensity');}
  $I1_max_scan_H = &findMaxRef(\@y1SR);
  if ($I1_max_scan_H < 1.) {$I1_max_scan_H=1.;}
  &setJoerger ($scaler, $cnt_time, $struck);
  &openShutter ($shutter,$shutter_readback);
  @I1_max_now  = &runJoerger ($scaler, $cnt_time, $ninputs, @inputs);
  &closeShutter ($shutter,$shutter_readback);
  &subtractJoergerOffsets(\@I1_max_now,\@inputs,\@Offsets);
  $I1_max_final_H = $I1_max_now[0];
  $I1_percent_H   = 100. * ($I1_max_final_H / $I1_max_scan_H);
  printf STDOUT "\n".$script.': Intensity at final Transfocator horiz. is %1.0f%s'
                    .' of scan maximum (scanmax=%.3g, now=%.3g)'."\n\n",
                    $I1_percent_H,'%', $I1_max_scan_H, $I1_max_final_H;
  ### Optionally save data:
  ($actPts,$datfile) = &saveData('H',$use_CRL,$use_TF,$tf_H_axis_PV,$FinalH,$ActH,$I1_max_scan_H,$I1_max_final_H,$I1_percent_H);
  if ($gnuopen) {
     if ($gnuplot == 2 && -e $datfile && $actPts > 2) {
        $png = $datfile; $png =~ s/\.dat$/\.png/;;
        print GNUPLOT 'plot "'.$datfile .'" using 1:2 '.$title."\n";
        &image_gnuplot($png,undef,undef,1,1);					#$file,$width,$height,$view,$GUIprompt}
     } else {&close_gnuplot();} 						#close vertical plot
     $gnuopen = 0;
  }
  return;
}

#=================================================================

sub run_guard_slits ($$;$) {		# ($sizeH, $sizeV, [$startonly])

  my ($sizeH, $sizeV, $startonly, $motionTimeH, $motionTimeV);

  $sizeH     = shift(@_);
  $sizeV     = shift(@_);
  $startonly = shift(@_);
  if (!defined $startonly) {$startonly = 0;}

# my ($gsSizeH_PV, $gsAssyH_PV, $gsSizeV_PV, $gsAssyV_PV);
# $gsSizeH_PV    = $beamline.':GS:YS';		#defined in main
# $gsAssyH_PV    = $beamline.':GS:Ah';		#defined in main
# $gsSizeV_PV    = $beamline.':GS:ZS';		#defined in main
# $gsAssyV_PV    = $beamline.':GS:Av';		#defined in main

### Setup motion:
  $status = &configAssy ($gsAssyH_PV);
  if ($status) {
     print STDOUT '*** '.$script.': configAssy '.$gsAssyH_PV.' detected error: status='.$status."\n";
     return -1;
  }
  $status = &configAssy ($gsAssyV_PV);
  if ($status) {
     print STDOUT '*** '.$script.': configAssy '.$gsAssyV_PV.' detected error: status='.$status."\n";
     return -1;
  }

  if (! $startonly) {
     $motionTimeH = &calcMotorTime($gsAssyH_PV, $gsSizeH_PV, $sizeH);
     $motionTimeV = &calcMotorTime($gsAssyV_PV, $gsSizeV_PV, $sizeV);
#    $motionTime = max($motionTimeH, $motionTimeV);
  }

### Start motion:
  $status = &startNMotors ($gsAssyH_PV, 1, [$gsSizeH_PV], [$sizeH], 'checkAuto');
  if ($status) {
     print STDOUT '*** '.$script.': startMotor '.$gsSizeH_PV.' detected error: status='.$status."\n";
     return -1;
  }
  $status = &startNMotors ($gsAssyV_PV, 1, [$gsSizeV_PV], [$sizeV], 'checkAuto');
  if ($status) {
     print STDOUT '*** '.$script.': startMotor '.$gsSizeV_PV.' detected error: status='.$status."\n";
     return -1;
  }

### Track motion:
  if (! $startonly) {
     $status = &trackMotor($gsAssyH_PV,$motionTimeH,$gsSizeH_PV);
     if ($status) {
        print STDOUT '*** '.$script.': Tracking '.$gsSizeH_PV.' detected error: status='.$status."\n";
        return -1;
     }
     $status = &trackMotor($gsAssyV_PV,$motionTimeV,$gsSizeV_PV);
     if ($status) {
        print STDOUT '*** '.$script.': Tracking '.$gsSizeV_PV.' detected error: status='.$status."\n";
        return -1;
     }
  }
  return 0;
}

#=================================================================

#  $status = &save_last_alignment($alignHistPV);
sub save_last_alignment ($) {
  my ($pv, $status, $str);
  $pv = shift(@_);
  ### The waveform PV should be 256 chars:
  ### $str = 'D=2016.06.07-13:00 E=13500 A=1E5 S=Pin Gain=1E5 V=-5.000 IscnV=1234567.1 IfinV=1234567.1 H=-5.000 IscnH=1234567.1 IfinH=1234567.1';
  $str  = sprintf 'D=%04d.%02d.%02d-%02d:%02d', $year, $mon, $mday, $hour, $minute;
  $str .= sprintf ' E=%d A=%1.0e S=%s Gain=%1.0e', int($energy+0.5), $attn, $i1_name, $kgain;
  if ($availableV) {$str .= sprintf ' V=%.3f IscnV=%.1f IfinV=%.1f', $FinalV, $I1_max_scan_V, $I1_max_final_V;}
  if ($availableH) {$str .= sprintf ' H=%.3f IscnH=%.1f IfinH=%.1f', $FinalH, $I1_max_scan_H, $I1_max_final_H;}
  if    ($use_CRL) {$str .= ' Scan=CRL';}
  elsif ($use_TF)  {$str .= ' Scan=Transfocator';}
  else             {$str .= ' Scan=DAC';}
  
  $status = &PezcaPutStringWF($alignHistPV,$str);
  if ($status) {
     print STDOUT "\n".'!!! '.$script.': error='.$status.' saving last alignment data into '.$pv."\n";
  }
  return $status;
}

#=================================================================

# ($date_last, $energy_last, $attn_last, $i1_name_last, $kgain_last,
#  $center_last_V, $I1_last_scan_V, $I1_last_final_V,
#  $center_last_H, $I1_last_scan_H, $I1_last_final_H) = &read_last_alignment($alignHistPV);
sub read_last_alignment ($) {
  my ($pv, $status, $str, %data, @kwords, $nw, @vals, $key, $i);
  $pv = shift(@_);
# $str = 'D=2016.06.07-13:00 E=13500 A=1E5 S=Pin Gain=1E5 V=-5.000 IscnV=1234567.1 IfinV=1234567.1 H=-5.000 IscnH=1234567.1 IfinH=1234567.1 Scan=DAC';
  ($status,$str) = &PezcaGetStringWF($pv);
  if ($status) {return;}
  %data = map {split /=/, $_, 2} map {split /\s+/, $_} $str;
  @kwords = qw(D E A S Gain V IscnV IfinV H IscnH IfinH Scan);
  $nw = @kwords;
  @vals = ();
  for ($i=0; $i<$nw; $i++) {
     $key = $kwords[$i];
     if (defined $data{$key}) {push(@vals,$data{$key});}
     else                     {push(@vals,undef);}
  }
  return @vals;
}

#=================================================================

#sub saveData($hv,$use_CRL,$use_TF,$driverPV,$driverFinal,$driverAct,		#$hv: H=horizontal or V=vertical
#             $I1_max_scan,$I1_max_final,$I1_percent) {
sub saveData($$$$$$$$$) {							#$hv: H=horizontal or V=vertical
### Optionally save data
  my ($hv, $use_CRL, $use_TF, $PV, $Final, $Act);
  my ($I1_max_scan, $I1_max_final, $actPts);
  my ($I1_percent, $file, $handler, $label);
  $hv           = shift(@_);  $hv = uc($hv);
  $use_CRL      = shift(@_);
  $use_TF       = shift(@_);
  $PV           = shift(@_);
  $Final        = shift(@_);
  $Act          = shift(@_);
  $I1_max_scan  = shift(@_);
  $I1_max_final = shift(@_);
  $I1_percent   = shift(@_);

  $actPts = 0;
  $file = $prefix.'_'.$hv.'.dat';
  $handler = open (DAT,' > '.$file);
  if (defined $handler) {
     print STDOUT $script.': Writing scan data into file="'.$file.'"'."\n";
     if ($tell_bluice) {&tellBluIce($scr.': writing '.$hv.'-scan data into "'.$file.'"');}
     printf DAT '# Date=%04d-%02d-%02d_%02d:%02d'."\n", $year, $mon, $mday, $hour, $minute;
     printf DAT '# Signal=%s (scaler=%s, input=%d)'."\n", $i1_name, $scaler, $i1;
     printf DAT '# Keithey=%s (channel=%d, gain=%1.0e, offset=%.3g)'."\n", $klabel, $kchannel, $kgain, $koffset;
     if    ($use_CRL) {$label='CRL';}
     elsif ($use_TF)  {$label='Transfocator';}
     else             {$label='DAC';}
     printf DAT '# %s-%s:  PV=%s'."\n",
                                        $label, $hv, $PV;
     printf DAT '# %s-%s:  Optimized=%.3fV   Initial=%.3fV'."\n",
                                        $label, $hv, $Final, $Act;
     printf DAT '# %s-%s:  I0max(scan)=%.1f   I0(final)=%.1f  (%.1f%s)'."\n",
                   $label, $hv, $I1_max_scan, $I1_max_final, $I1_percent, '%';
     if ($use_CRL) {
        $actPts = @x1HD;
        for ($i=0; $i<$actPts; $i++) {printf DAT ' %.3f  %.1f'."\n", $x1HD[$i], $y1HD[$i];}
     }
     elsif ($use_TF) {
        $actPts = @x1SR;
        for ($i=0; $i<$actPts; $i++) {printf DAT ' %.3f  %.1f'."\n", $x1SR[$i], $y1SR[$i];}
     } 
     else {
        if ($fastScan) {
           $actPts = @x1SR;
           for ($i=0; $i<$actPts; $i++) {printf DAT ' %.3f  %.1f'."\n", $x1SR[$i], $y1SR[$i];}
        } else {
           $actPts = @x1DAC;
           for ($i=0; $i<$actPts; $i++) {printf DAT ' %.3f  %.1f'."\n", $x1DAC[$i], $y1DAC[$i];}
        }
     }
     close(DAT);
  } else {
     print STDOUT '*** '.$script.': cannot open file="'.$file.'" for writing'."\n";
     if ($tell_bluice) {&tellBluIce($scr.': Failed to save '.$hv.'-scan data','error');}
  }
  return ($actPts,$file);
}

#=================================================================

sub tellBluIceBeforeExit($;$) {
  my $msg  = shift(@_);
  my $type = shift(@_);		if(!defined $type) {$type='error';}
  if ($tell_bluice) {
     &tellBluIce($msg,$type);
     if (defined $bialignStatePV) {
        $status = Pezca::PutDouble($bialignStatePV,0);				#tell BluIce that the script has finished,
        &report_only($script, $status, $bialignStatePV);
     }
  }
}

#=================================================================

#  ($status,$pos,$i0) = &optimizePos($assyPV, $axisPV, $step, $scaler, $countTime, $input, [$fileBkg, $delay]);
sub optimizePos($$$$$$;$$) {
###
### This subroutine tweaks asseembly from the current position to find the maximum $i0.
  our ($beamline, $shutter, $shutter_readback, @Offsets);
  our ($IGNORE_ERRORS);
  my ($script, $axisPV, $assyPV, $step, $scaler, $countTime, $input, $fileBkg, $delay);
  my ($status, $pos, $i0, $pos_ini, $posLo, $posHi, @Intensity, $off);
  my ($i0_ini, $i0_1, $i0_2, $direction, $n);

  $script    = 'optimizePos';
  $assyPV    = shift(@_);		#assy to scan
  $axisPV    = shift(@_);		#axis to scan
  $step      = shift(@_);		#scan step
  $scaler    = shift(@_);		#scaler PV
  $countTime = shift(@_);		#scaler count time
  $input     = shift(@_);		#pointer to the array of scaler inputs
  $fileBkg   = shift(@_);		#joerger background file ('auto' for automatic, '' for none)
  $delay     = shift(@_);		#delay after changing DAC and before running scaler

  if (!defined $fileBkg) {$fileBkg = 'auto';}
  if (!defined $delay)   {$delay   = 0;}

  if (!defined $IGNORE_ERRORS)  {$IGNORE_ERRORS = 0;}		#global flag
  if (defined $shutter) {
     if ($shutter !~ /^${beamline}:/) {$shutter = $beamline.':'.$shutter;}
  }
  if (defined $shutter_readback) {
     if ($shutter_readback !~ /^${beamline}:/) {
                              $shutter_readback = $beamline.':'.$shutter_readback;}
  }

  if (defined $shutter) {&closeShutter ($shutter,$shutter_readback);}

  print  STDOUT "\n".$script.': Tweaking position of axis='.$axisPV.', assy='.$assyPV
         .' to max intensity on scaler='.$scaler.' input='.$input."\n";
  printf STDOUT $script.': position step=%.3f'."\n",$step;
  print  STDOUT $script.': Scaler='.$scaler.', countTime='
         .$countTime.'s, input='.$input."\n";

  if (@Offsets == 0) {
     ### Read the scaler background offsets:
     @Offsets = &getJoergerOffsets ($fileBkg, $scaler);
     ### set scaler:
     &setJoerger ($scaler, $countTime);
     ### setup motor assembly (prepare to scan):
     $status = &configAssy($assyPV); if($status) {return -1;}
  }
  $off = $Offsets[$input-1];

### get current position:
  $pos_ini = &getAxisAct($axisPV);

### get axis limits:
  ($posLo,$posHi) = &getAxisDRVLH($axisPV);

  if (defined $shutter) {
     $status = &openShutter ($shutter,$shutter_readback);
     if ($status) {return $status;}
  }

### get current intensity:
  @Intensity = &runJoerger ($scaler, $countTime, 1, ($input));
  $i0_ini = $Intensity[0] - $off;
  printf STDOUT $script.': axis='.$axisPV.' I0[0]=%.2f'."\n",$i0_ini;

### = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
### Find the direction to go:

  $step = abs($step);
  $pos = $pos_ini - $step;
  if ($posHi != 0 || $posLo != 0) {
     if ($pos < $posLo) {
        printf STDOUT '!!! '.$script.': motor='.$axisPV.' is a lower limit=%.2f.'."\n",$posLo;
        undef $pos;
     }
  }
  if (defined $pos) {
     $status = &runMotor ($assyPV, $axisPV, $pos);
     if ($status) {return $status;}
     if ($delay > 0) {select(undef,undef,undef,$delay);}			#sleep in sec
     @Intensity = &runJoerger ($scaler, $countTime, 1, ($input));
     $i0_1 = $Intensity[0] - $off;
     printf STDOUT $script.': axis='.$axisPV.' I0[-1]=%.2f'."\n",$i0_1;
  } else {
     $i0_1 = -100000000;
  }

  $pos = $pos_ini + $step;
  if ($posHi != 0 || $posLo != 0) {
     if ($pos > $posHi) {
        printf STDOUT '!!! '.$script.': motor='.$axisPV.' is a upper limit=%.2f.'."\n",$posHi;
        undef $pos;
     }
  }
  if (defined $pos) {
     $status = &runMotor ($assyPV, $axisPV, $pos);
     if ($status) {return $status;}
     if ($delay > 0) {select(undef,undef,undef,$delay);}			#sleep in sec
     @Intensity = &runJoerger ($scaler, $countTime, 1, ($input));
     $i0_2 = $Intensity[0] - $off;
     printf STDOUT $script.': axis='.$axisPV.' I0[+1]=%.2f'."\n",$i0_2;
  } else {
     $i0_2 = -100000000;
  }

  if    ($i0_2 > $i0_ini && $i0_ini > $i0_1) {					#positive
     $direction = 1;
     $i0 = $i0_2;
     print STDOUT $script.': correcting axis='.$axisPV.' in positive direction.'."\n";
  }
  elsif ($i0_1 > $i0_ini && $i0_ini > $i0_2) {					#negative
     $direction =-1;
     $i0 = $i0_1;
     print STDOUT $script.': correcting axis='.$axisPV.' in negative direction.'."\n";
  }
  else {
     if (defined $shutter) {&closeShutter ($shutter,$shutter_readback);}
     print STDOUT $script.': No optimization required for axis='.$axisPV."\n";
     &runMotor ($assyPV, $axisPV, $pos_ini);
     return 0,$pos_ini,$i0_ini;
  }

### Run measurements:
  $n = 1;
  do {
     $n++;
     $pos = $pos_ini + $n*$step*$direction;
     if ($posHi != 0 || $posLo != 0) {
        if ($pos > $posHi) {undef $pos;}
        if ($pos < $posLo) {undef $pos;}
     }
     if (defined $pos) {
        &runMotor ($assyPV, $axisPV, $pos);
        if ($delay > 0) {select(undef,undef,undef,$delay);}			#sleep in sec
        @Intensity = &runJoerger ($scaler, $countTime, 1, ($input));
        $i0_1 = $i0;
        $i0 = $Intensity[0] - $off;
        printf STDOUT $script.': axis='.$axisPV.' I0[%d]=%.2f'."\n",$n*$direction,$i0;
     } else {
        if (defined $shutter) {&closeShutter ($shutter,$shutter_readback);}
        printf STDOUT '!!! '.$script.': optimization for axis='.$axisPV
                     .' brought it to the limit.'."\n";
        $pos = $pos_ini + ($n-1)*$step*$direction;
        return 0,$pos,$i0;
     }
  } while ($i0 > $i0_1);							#while intensity is growing
  if (defined $shutter) {&closeShutter ($shutter,$shutter_readback);}
  $n--;
  $pos = $pos_ini + $n*$step*$direction;
  &runMotor ($assyPV, $axisPV, $pos);
  printf STDOUT $script.': axis='.$axisPV.' position corrected from %.4f to %.4f (by %d steps)'."\n", $pos_ini,$pos,$n;
  return 0,$pos,$i0_1;
}

#=================================================================

sub signal_handler ($;$) {							#($signal,$exitcode)
  our ($ABORT_IN_PROCESS);
  my ($signame, $code);
  $signame   = shift(@_);
  $code      = shift(@_);
  if ($signame ne 'quit') {print STDOUT "\n".$script.': *** Caught a SIG'.$signame."\n\n";}
  else                    {print STDOUT "\n".$script.': !!! cleaning up before quit'."\n\n";}

  if (defined $ABORT_IN_PROCESS) {return;}
  $ABORT_IN_PROCESS = 1;
  &abort_external_script();

  print STDOUT $script.'('.$signame.'): ..aborting scan record ['.$scanrec.']'."\n";
  &abortSR($scanAbort,$scanrec);

  if (defined $shutter) {
     if ($shutter !~ /^${beamline}/) {$shutter = $beamline.':'.$shutter;}
     print STDOUT $script.'('.$signame.'): ..closing shutter = '.$shutter."\n";
     if ($tell_bluice) {&tellBluIce($scr.': closing shutter');}
     &closeShutter ($shutter);
  }
  if (defined $gnuopen) {
     if ($gnuopen) {&close_gnuplot();}
  }
### Drive DACs or CRLs/Transfocator to initial values:
  if (defined $ActV) {
     if (defined $dacVFM_PV) {
        if ($fastScan) {
           print STDOUT $script.'('.$signame.'): ..aborting VFM DAC scan ('.$dacVFM_PV.')'."\n";
           &abortDAC ($dacVFM_PV);
        }
        printf STDOUT $script.'('.$signame.'): ..driving vert. mirror piezo to initial pos. = %.3f'."\n",$ActV;
        if ($tell_bluice) {&tellBluIce($scr.': restoring V-pos.');}
        &driveDAC ($dacVFM_PV,$ActV,$LoV,$HiV);
     }
     if (defined $crl_V_PV) {
        printf STDOUT $script.'('.$signame.'): ..driving CRL vert. to initial pos. = %.3f'."\n",$ActV;
        &abortAssy ($crl_assy_PV,1,$crl_V_PV);
        &runMotor ($crl_assy_PV, $crl_V_PV, $ActV, 'checkAuto');
     }
     elsif (defined $tf_V_axis_PV) {
        printf STDOUT $script.'('.$signame.'): ..driving Transfocator vert. to initial pos. = %.3f'."\n",$ActV;
        &abortAssy ($tf_V_assy_PV,1,$tf_V_axis_PV);
        &runMotor ($tf_V_assy_PV, $tf_V_axis_PV, $ActV, 'checkAuto');
     }
  }
  if (defined $ActH) {
     if (defined $dacHFM_PV) {
        if ($fastScan) {
           print STDOUT $script.'('.$signame.'): ..aborting HFM DAC scan ('.$dacHFM_PV.')'."\n";
           &abortDAC ($dacHFM_PV);
        }
        printf STDOUT $script.'('.$signame.'): ..driving horiz.mirror piezo to initial pos. = %.3f'."\n",$ActH;
        if ($tell_bluice) {&tellBluIce($scr.': restoring H-pos.');}
        &driveDAC ($dacHFM_PV,$ActH,$LoH,$HiH);
     }
     if (defined $crl_H_PV) {
        printf STDOUT $script.'('.$signame.'): ..driving CRL horiz. to initial pos. = %.3f'."\n",$ActH;
        &abortAssy ($crl_assy_PV,1,$crl_H_PV);
        &runMotor ($crl_assy_PV, $crl_H_PV, $ActH, 'checkAuto');
     }
     elsif (defined $tf_H_axis_PV) {
        printf STDOUT $script.'('.$signame.'): ..driving Transfocator horiz. to initial pos. = %.3f'."\n",$ActH;
        &abortAssy ($tf_H_assy_PV,1,$tf_H_axis_PV);
        &runMotor ($tf_H_assy_PV, $tf_H_axis_PV, $ActH, 'checkAuto');
     }
  }
  if (defined $struck) {&abortStruck($struck);}

  if (defined $mostab_initial) {
     if ($interfaceMostab && $mostab_initial != 1) {				#if not paused initially
        print STDOUT $script.'('.$signame.'): ..restoring intensity feedback'."\n";
        if ($tell_bluice) {&tellBluIce($scr.': restoring I0 feedback');}
        &restoreMostab ($beamline);
     }
  }
  if (defined $postab_initial) {
     if ($interfacePostab && $postab_initial != 3) {				#if not paused initially
        print STDOUT $script.'('.$signame.'): ..restoring positional feedback'."\n";
        if ($tell_bluice) {&tellBluIce($scr.': restoring pos.feedback');}
        &restorePostab ($beamline);
     }
  }
  if (defined $activeCol_PV && defined $state300um) {
     if (($move_collimator || $move_pindiode) && $restore_collimator) {
        print STDOUT $script.'('.$signame.'): ..restoring collimator, beamstop, pindiode'."\n";
        if ($tell_bluice) {&tellBluIce($scr.': restoring Coll.,BS,PinDiode');}
        if ($move_collimator && $align300to10 == 1 && $collimatorPreset <= 1) {
           if    ($collimatorPreset == 0) {Pezca::PutDouble($activeCol_PV,$state10um);}
           elsif ($collimatorPreset == 1) {Pezca::PutDouble($activeCol_PV,$state300um);}
           if (&waitMotor($colAssy_PV,5,0.3) == 0) {
              if ($tell_bluice) {&tellBluIce($scr.': restoring collimator failed','warn');}
              print STDOUT '*** '.$script.'('.$signame.'): timeout on restoring collimator'."\n";
           }
        }
        @prg = ('IN', $beamline);
        if (! $move_collimator) {push @prg, 'nocollimator';}
        if ($move_pindiode)     {push @prg, 'pindiodeOUT';}
        else                    {push @prg, 'nobeamstop';}
        &setSampleEnv(\@prg,'nowait');
     }
  }
  if (defined $gsSizeH_initial && defined $gsSizeV_initial) {
     print STDOUT $script.'('.$signame.'): ..returning guard slits to initial size and position'."\n";
     if ($tell_bluice) {&tellBluIce($scr.': restoring Guard Slits');}
     &run_guard_slits($gsSizeH_initial,$gsSizeV_initial,1);			#start only
  }
  if ($protectSample > 1 && defined $sample_moved && $moveSample != 0) {
     if    ($direction eq 'UP')       {$direction = 'DOWN';}
     elsif ($direction eq 'DOWN')     {$direction = 'UP';}
     elsif ($direction eq 'OUTBOARD') {$direction = 'INBOARD';}
     elsif ($direction eq 'INBOARD')  {$direction = 'OUTBOARD';}
     printf STDOUT $script.': moving sample back (%s) by %.1fum.'."\n",$direction,abs(1000*$moveSample);
     if ($xyzAssyPV eq $zAssyPV) {
        $status = &startNMotors ($xyzAssyPV,$nxyz,\@xyzAxesPV,\@xyzIni);
     } else {
        $status = &startNMotors ($xyzAssyPV,2,\@xyzAxesPV,\@xyzIni);
        $status = &startNMotors ($zAssyPV,1,[$xyzAxesPV[2]],[$xyzIni[2]]);
     }
  }

  &clearAllCaMonitors();
  &tellBluIceBeforeExit($scr.': Aligning script finished.', '');
  print STDOUT $script.'('.$signame.'): Done!'."\n\n";
  if (defined $code) {exit $code;}						#this will return '256*n' to calling perl script (when n is small)
  else               {exit 1;}							#this will return '256' to calling perl script
}
