#!/usr/bin/perl -W
#----------------------------------------------------------------
# This script helps to setup beamline for a user. It aligns beamline
# components and then checks signals and positions agains "historic" values.
# See "user_setup_expectations.pl" where these "historic" values are defined.
#
#                          Initial Author: Derek Yoder (dwy)
# Current maintainer & GUI implementation: Sergey Stepanov (ss)
#----------------------------------------------------------------
# Version-1.01	2005/06/13 dwy (Derek Yoder)
# Version-1.02	2012/09/19 Sergey (ss)
# Version-1.03	2012/11/22 added Mime:Lite and checking for EPS monitor script.
# Version-1.04	2013/02/05 moved closing guard slits up in the timeline.
# Version-1.05	2013/04/24 changed to work with sampleEnvSrv.
# Version-1.06	2013/05/03 added switching between crystal-XYZ and gonio-SH.
# Version-1.07	2013/05/17 added checking for DCM monitor script.
# Version-1.08	2014/05/12 added determining max speed of Gonio & XYZ
#                          and noting in mySQL
# Version-1.09	2014/05/29 added recoding temporary behaviour of 5um intensity;
#                          added call of white beam slits centering with
#                          steering popups.
# Version-2.00	2014/09/30 some cleanup (renamed subs and changed their order);
#                          changed the call of alignMiniBeam.pl from
#                          'align300to10' to 'autoPinholeOrGuardSlits'
#                          (same as in JBluIce);
#                          switched to recording intensity with Joerger
#                          instead of ADC;
#                          added Keithley gain auto adjustment for the
#                          measurements of intensity fluctuations.
# Version-2.01	2014/10/03 added call for Mono energy calibration.
# Version-3.00	2014/10/07 added GUI frontend.
# Version-3.01	2015/03/04 changed printing to postscript file and
#                          printing/viewing with gv.
# Version-3.02	2015/05/04 added SelectAll & UnselectAll buttons.
# Version-3.03	2015/05/06 added beamline specification in command line;
#                          changed lists to array of hashes;
#                          added optional setup functions.
# Version-3.04	2016/11/18 moved guard slits centering to the main section.
# Version-3.05	2017/04/07 moved alingMinibeam before aligning guard slits;
#                          moved disabling mirror motors upward, right after 
#                          mirror centering.
# Version-3.06	2018/03/27 changed the way subprocesses are killed: always
#                          kill all sub-subprocesses.
# Version-3.07	2018/06/12 changed &record_voltages to read number of 
#                          segments from sy900s_subs.pl.
# Version-3.08	2019/03/05 changed individual components PVs to 2D arrays
#                          of @wbmtr, @mbmtr, @esmtr (white-beam-hutch
#                          motors, mono-beam-hutch motors, endstation motors);
#                          improved "minibeam_intensity" to print all presets
#                          and use arrays; 
#                          simplified "check_mirror_mtr_disabled" and
#                          "record_motors".
# Version-3.09	2019/07/09 changed aborting and added kill buttons.
# Version-3.10	2019/11/18 added &check_gripper and &check_dectris_eth.
# Version-3.11	2019/12/16 changed to save plots as images, removed producing
#                          pdfs from logs and replaced it by opening logs
#                          in editor; added window for the setuper comments.
# Version-3.12	2019/12/19 added buttons for various stats (pos. history).
# Version-3.13	2020/01/21 changed &check_gripper to &check_robot_homing.
# Version-3.14	2020/01/29 added notes history.
# Version-3.15	2020/02/06 added &center_WB2_slits and made &center_WB_slits
#                          optional.
# Version-3.16	2020/02/10 added &verifyConfigTcl.
# Version-3.17	2020/10/06 added &record_CRL and &record_Transfocator.
# Version-3.18	2020/11/04 added tee and logging of subscripts output to log2.
# Version-3.19	2021/01/20 corrections for new Eiger DCU OS.
# Version-3.20	2021/02/10 added date range to GnuPlot and adapted to new 
#                          motor positions saving format in log2.
# Version-3.21	2021/03/27 bug fixes for &record_motors and for &fill_mtrList.
# Version-3.22	2021/10/21 added &chec_keithley_times
# Version-3.23	2021/12/03 added setting short Keithley time to &record_I0_temp_data_5um()
# Version-3.24	2021/12/07 added checks for running monoStabilizer and transfocatorServer
# Version-3.25	2022/11/08 added caen_set_offsets
# Version-3.26	2023/02/02 adapted EPS monitoring from ACCEL EPS to new APS BLEPS
#
# To-do: HDM recording, beam profiling, KBM DAC-piezo recording.
#----------------------------------------------------------------
# Subroutines in this file:
# ^^^^^^^^^^^^^^^^^^^^^^^^
#                          &draw_main();
#                          &refresh_gui();
#                          &reconf_startBtn();
#                          &reconf_abortBtn();
#                          &reconf_startBtnUser();
#                          &switch_more_less();
#                          &parse_geometry(\$window);
#                          &open_comments_window();
#                          &save_notes();
#                          &run_selected_steps();
#                          &do_step();
#                          &system_with_fork(@);
#                          &unselectAll();
#                          &selectAll();
#                          &clearResults();
#                          &goodbye();
#                          &goodbye_real();
#                          &abort_step();
#                          &abort_real();
#                          &test_perl();
#                          &verifyConfigTcl();				==> verifyConfigTcl.pl
#                          &check_EPS_monitor();
#                          &check_DCM_monitor();
#                          &check_MBSH();
#                          &check_MonoStabilizer();
#                          &check_TransfocatorServer();
#                          &check_keithley_times();
#                          &caen_set_offsets();				==> caen_set_offsets.pl
#                          &center_WB_slitsV();
#                          &center_WB_slitsH();
#                          &center_WB_slits($slit);			==> centerSlits_WB.pl
#                          &center_WB2_slitsV();
#                          &center_WB2_slitsH();
#                          &center_WB2_slits($slit);			==> centerSlits_WB2.pl
#                          &center_MB_slitsV();
#                          &center_MB_slitsH();
#                          &center_MB_slits($slit);			==> centerSlits_MB.pl
#                          &center_GS_slitsV();
#                          &center_GS_slitsH();
#                          &center_GS_slits($slit);			==> centerSlits_GS.pl
#                          &center_HFM();
#                          &center_VFM();
#                          &center_KBM($mirror);			==> centerMirror_v3.pl
#                          &calibrate_mono();				==> calibrateMono.pl
#                          &align_minibeam();				==> alignMiniBeam.pl
#                          &minibeam_intensity();
#        ($err,$msg,$I0) = &measure_pinhole_cts($pinhole,$gain,
#                                               $expected_intensity,
#                                               $warn_level,$date);
# ($pts,$rmsVal,$rmsDev) = &record_I0_temp_data_5um();
#                          &measure_Beamsize();				==> measureMiniBeam.pl
#                          &bluice_settings();
#                          &check_mirror_mtr_disabled();
#                          &backup_motors();				==> mtrPositionsBackup.pl
#                          &record_motors();
#                          &backup_voltages();				==> mirrorsBackup.pl
#                          &record_voltages();
#                          &record_CRL();
#                          &record_Transfocator();
#                          &record_maxspeeds();				==> saveMaxSpeeds2sql.pl
#                          &check_robot_homing();			==> checkRobotHoming.pl
#                          &check_dectris_eth();			==> checkdetectorconnections.pl
#                          &mono_calibration_history();
#                          &beam_steering_history();
#                          &I0_history();
#                          &minibeam_intensities_history();
#                          &beam_sizes_history();
#                          &collimator_preset_pos_history();
#                          &collimator_preset_ang_history();
#                          &guard_slits_calib_history();
#                          &notes_history();
#                          &reconf_Hist();
#                          &fill_mtrList();
#                          &motor_ActPos_history();
#                   $err = &getPV($pv2get,$err_message,$value_type,
#                                 $value1,$value2);
#                          &part1($statement);
#                          &part2($statement);
#                          &part3($statement);
#                          &email($header,$msg,$email_list);
#                $status = &read_file($file,\$buf);
#         ($start,$last) = &calc_date_range($ndays,[$last,$separator]);
#                          &signal_handler();				#blocks Ctrl+C
#----------------------------------------------------------------

  use strict;
  use warnings;
  use Pezca;
  use DBI;
  use MIME::Lite;
  use Tk;
  use Tk::BrowseEntry;
  use Tk::Dialog;
  use Tk::ROText;
  use warnings::unused;

### Declare variables ---------------------------------------------
  our ($beamline, $shutter, $shutter_readback, $nRetry);		# these are from beamline.pl
  our ($struckName, $struck_ext, $struckChName, $joergerName);		# these are from beamline.pl
  our ($have_KBM, $have_CRL, $have_TF);					# these are from beamline.pl
  our ($HAVE_XYZ, $ATTENUATION_USE_FACTOR);				# these are from beamline.pl
  our ($NONINTERACTIVE, $USE_ZENITY_FOR_EXIT, $IGNORE_ERRORS, $report); # these are from beamline.pl

  my ($PEZCA, $script, $arg, $status);
  my ($email_list, $hostname, $username, $cross, $i);
  my ($message, $message_300, $message_20, $message_10, $message_5);
  my ($current_time, @ct, $sec, $min, $hour, $mday, $mon, $year);
  my ($filename, $logdir, $logroot, $logfile, $filehandle, $syntax);
  my ($ca_error, $pv_value, $err_log, $err, @prg, $status1, $history);
  my ($part2_message, $part3_message, $step_err, $teelog, $useTee);
  my ($v_size, $h_size);
  my ($inten, $i0_int, $i0_int_avg);
  my ($intensity_300, $intensity_20, $intensity_10, $intensity_5, $minibeam_summary);
  my ($gs_vs, $gs_hs, $gon_h_in, $gon_h_out, $aps_current, $dcm_vsum);
  my ($int_avg, $lo_ok, $hi_ok, $energy, $att_percentage, $att_actual, $att_factor_actual);
  my ($part1_text, $part2_text, $part3_text, $part3_text_header, $minibeam_messages);
  my ($bl_host, $bl_user, $part1_print, $log_exist);
  my (@scriptsPrefix, $inline, $pause_after, $struck_root, $struck_file);
  my ($makePDF, $viewerPDF, $editor, $filePS, $filePDF, @beamlinelist, $debug);

### These are global for user_setup_expectations.pl:
  our ($CAQTDM, $expectations, $simulate, $gh_assy, $gh_axis);
  our (@mostab_status, @mostab_status_b, @mostab_bpm2_ch, @mostab_use_bpm2);
  our (@beamstop_chk, @shutter_chk, @feedback_chk, @egap_chk);
  our (@wbmtr, $wbsl_distance2source, $wbsl2_distance2source);
  our (@mbmtr, @esmtr, @vals, $att_factor, $minibeam_date);
  our ($kl_pin_channel, $struck_pin_input, $adc_pin_input, $struck_dwell);
  our ($kl_i0_channel,  $struck_i0_input,  $adc_i0_input, $kl_i0_gain_hist);
  our ($adc_scanrate, $scaler, $cnt_time, $historic_aps_cur, $bmstp_pos);
  our (@minibeam_300, @minibeam_20, @minibeam_10, @minibeam_5);
  our ($hist_date, $hist_aps_current, $hist_energy, $hist_dcm_vsum, $hist_att);
  our ($hist_gs_v, $hist_gs_h, $hist_i0, $hist_bmstp);
  our ($hist_vsize, $hist_hsize, $hist_gs_v_mini, $hist_gs_h_mini);
  our ($nkeithley, $maxkeithleytime);
### These are added for the GUI
  our (@hostList, $mw, $pad, $w_moreBtn, @mtrList, $mtrSelected);
  our (@w_stepChkBtn, @w_stepresultLbl);
  our (@steps, $nSetupSteps, $currStep, $currDisplay, $includeExtras);
  our ($w_startBtn, $w_cancelBtn, $w_clearBtn, $w_abortBtn);
  our ($column, $use_displayer);
  our ($process_running, $process_aborted, $need2exit);
  our ($fontHdr, $fontLbl, $fontBtn, $fontSml, $fontTbl);
  our ($bgcolor, $fgcolor, $abgcolor, $afgcolor);
  our ($notesWin, $notesEditor, $notesBtnSav);
  our (@histList, $historySpin, $w_statsMtrSpin);

  if    (defined $ENV{PEZCA})     {$PEZCA=$ENV{PEZCA};}
  elsif (defined $ENV{FASTSCANS}) {$PEZCA=$ENV{FASTSCANS}; $PEZCA=~s/fastscans$/pezca/i;}
  else {if ($^O !~ /Win/i) {$PEZCA='/home/gmca/epics/pezca';}else{$PEZCA='c:/gmca/pezca';}}
  $PEZCA =~ s/\\/\//g;
  if ($PEZCA !~ /\/$/) {$PEZCA .='/';}

  select STDERR; $|=1;					#set unbuffered output
  select STDOUT; $|=1;					#set unbuffered output

  $SIG{HUP}  = 'goodbye';				#(1) hangup (see signum.h)
  $SIG{INT}  = 'goodbye';				#(2) interrupt (see signum.h)
  $SIG{QUIT} = 'goodbye';				#(3) quit (see signum.h)
# $SIG{KILL} = 'goodbye';				#(9) this is what is sent by kill -9 -- unblockable
  $SIG{TERM} = 'goodbye';				#(15) terminate -- sent by kill without paramaters
# $SIG{STOP} = 'goodbye';				#(19) stop -- unblockable

  do $PEZCA.'Common/attenuation_subs_v2.pl';
  do $PEZCA.'Common/joerger_subs_v2.pl';
  do $PEZCA.'Common/keithley_subs.pl';
  do $PEZCA.'Common/motor_subs.pl';
  do $PEZCA.'Common/shutter_subs.pl';
  do $PEZCA.'Common/struck_subs_v2.pl';
  do $PEZCA.'Common/sy900s_subs.pl';
  do $PEZCA.'/Common/crl_subs.pl';
  do $PEZCA.'Common/gnuplot_subs.pl';
  do $PEZCA.'Common/common_subs.pl';
  do $PEZCA.'Common/beamline.pl';

  $script = &makeScriptName();

  $nRetry  = 5;
  &setEpicsRetry();
  Pezca::AutoErrorMessageOff();

  @beamlinelist  = ('23i:', '23o:', '23b:', '23d:');
  @histList = qw(5 10 20 30 40 60 90);
  $history       = 28;                                          # days (default, can be overwritten by command line)
  $use_displayer = 1;                                           # use displayer.pl to show non-blocking history dialog boxes
  $makePDF       = 0;						# if "0", will open TXT in editor, "1": make PDF, "2": print directly
  $viewerPDF = &Which('masterpdfeditor5');			# this is preferred
  if (!defined $viewerPDF) {$viewerPDF = &Which('evince');}	# this is default file viewer of GNOME
  if (!defined $viewerPDF) {$viewerPDF = &Which('atril');}	# this is a MATE fork of evince.
  if (!defined $viewerPDF) {$viewerPDF = &Which('okular');}	# this is default file viewer of KDE
  if (!defined $viewerPDF) {$viewerPDF = &Which('acroread');}	# rarely installed
  if (!defined $viewerPDF) {$viewerPDF = &Which('gv');}		# does not allow choosing printer (just default)
  $editor = &Which('geany');                                    # this is a preferred desktop editor
  if (!defined $editor)    {$editor = &Which('gedit');}	        # this is a GNOME default editor
  if (!defined $editor)    {$editor = &Which('pluma');}	        # this is a MATE fork of gedit
  if (!defined $editor)    {$editor = &Which('kate');}          # this is a KDE desktop default editor

  $email_list = 'rfischetti@anl.gov'
              . ',sstepanov@anl.gov'
              . ',makarov@anl.gov'
              . ',mhilgart@anl.gov'
              . ',qxu@anl.gov'
              . ',ogata@anl.gov'
              . ',vnagas@anl.gov'
              . ',mbecker@anl.gov'
              . ',scorcoran@anl.gov'
              . ',fergusond@anl.gov'
              . ',dkissick@anl.gov'
              ;

  @hostList = (''
              ,'Craig Ogata'
              ,'Naga Venugopalan'
              ,'Michael Becker'
              ,'Steve Corcoran'
              ,'Dale Ferguson'
              ,'David Kissick'
              ,'Mark Hilgart'
              ,'Qingping Xu'
              ,'Robert Fischetti'
               );

  @steps = (
            {func=>'verifyConfigTcl',		txt=>'Verify Config.tcl against EPICS',	 main=>0,},     # Check if configNN.tcl corresponds to motor data in Tsub
            {func=>'check_EPS_monitor',		txt=>'Check EPS monitor',		 main=>1,},     # Check if EPS IOC & Monitor are running & EPS sections are in monitor mode
            {func=>'check_DCM_monitor',		txt=>'Check DCM monitor',		 main=>1,},     # Check if DCM IOC & Monitor are running and not paused
            {func=>'check_MBSH',		txt=>'Check & open MBSH',		 main=>1,},     # Check if monochromatic beam shutter is open
            {func=>'check_MonoStabilizer',	txt=>'Start & check intensity feedback', main=>1,},     # Check if monoStabilizer is running and prompt to start it, then confirm it is running
            {func=>'check_TransfocatorServer',	txt=>'Start & check Transfocator Server',main=>1,},     # Check if transocatorServer is running and prompt to start it
            {func=>'check_keithley_times', 	txt=>'Check Keithley integration times', main=>1,},     # Check if all keithley amplifiers have low integration times
            {func=>'caen_set_offsets',		txt=>'Determine CAEN counters offsets',	 main=>1,},     # Run caen_set_offsets.pl
#           {func=>'center_WB_slitsV',		txt=>'Check white-beam steering V(old)', main=>0,},     # Run White Beam Slits centering and show needed steering (restore the slits)
#           {func=>'center_WB_slitsH',		txt=>'Check white-beam steering H(old)', main=>0,},     # Run White Beam Slits centering and show needed steering (restore the slits)
            {func=>'center_WB2_slitsV',		txt=>'Check white-beam steering vert.',	 main=>1,},     # Run White Beam Slits centering and show needed steering (restore the slits)
            {func=>'center_WB2_slitsH',		txt=>'Check white-beam steering horiz.', main=>1,},     # Run White Beam Slits centering and show needed steering (restore the slits)
            {func=>'calibrate_mono',		txt=>'Calibrate mono energy',		 main=>1,},     # Run mono energy calibration
            {func=>'center_MB_slitsV',		txt=>'Center mono-beam slits vert.',	 main=>0,},     # Run script centering collimator slits vertically
            {func=>'center_MB_slitsH',		txt=>'Center mono-beam slits horiz.',	 main=>0,},     # Run script centering collimator slits horizontally
            {func=>'center_HFM',		txt=>'Center horiz. focusing mirror',	 main=>0,},     # Run script centering horiz. focusing mirror
            {func=>'center_VFM',		txt=>'Center vert. focusing mirror',	 main=>0,},     # Run script centering vert. focusing mirror
            {func=>'check_mirror_mtr_disabled',	txt=>'Verify mirrors motors disabled',	 main=>1,},     # Verify that mirror motors are disabled
            {func=>'align_minibeam',		txt=>'Align minibeam',			 main=>1,},     # Align minibeam through pinhole
            {func=>'center_GS_slitsV',		txt=>'Center guard slits vert.',	 main=>1,},     # Run script centering guard slits vertically
            {func=>'center_GS_slitsH',		txt=>'Center guard slits horiz.',	 main=>1,},     # Run script centering guard slits horizontally
            {func=>'minibeam_intensity',	txt=>'Record minibeam intensities',	 main=>1,},     # Record minibeam intensities
            {func=>'measure_Beamsize',		txt=>'Measure minibeam HxV sizes',	 main=>1,},     # Measure the beam size at the collimator position
#           {func=>'center_pin',		txt=>'Center pin to crosshair',		 main=>0,},     # Center goniometer axis to crosshair (mount pin, etc) -- TODO
            {func=>'bluice_settings',		txt=>'Check JBluIce runtime settings',	 main=>1,},     # Verify BluIce runtime settings
            {func=>'backup_motors',		txt=>'Backup motors positions',		 main=>1,},     # Backup motor positions script
            {func=>'record_motors',		txt=>'Check motors positions',		 main=>1,},     # Get motor position and compare to expected values
            {func=>'backup_voltages',		txt=>'Backup KBM voltages',		 main=>1,},     # Backup KBM voltages
            {func=>'record_voltages',		txt=>'Check KBM voltages',		 main=>1,},     # Verify non-zero.  (Anything else?)
            {func=>'record_maxspeeds',		txt=>'Record Omega & XYZ max speeds',	 main=>1,},     # Update mySQL with max speeds of Gonio Omega and Crystal XYZ
            {func=>'check_robot_homing',	txt=>'Check Robot homing',		 main=>0,},     # Check homing of Cartesian robot gripper and Dewar
            {func=>'check_dectris_eth',		txt=>'Check PAD network connections',	 main=>1,},     # Check network connections of Pilatus/Eiger detectors (on ws2 as gmca only)
            {func=>'record_CRL',		txt=>'Record CRL box config & position', main=>1,},     # Record CRL box config and position
            {func=>'record_Transfocator',	txt=>'Record Transfocator config & pos.',main=>1,},     # Record Transfocator config and position
#           {func=>'record_misc',		txt=>'Record miscellaneous PVs',	 main=>0,},     # Record KL gains, puck type, DAC voltages  (Anything else?)
#           {func=>'print_motors',		txt=>'Print motors positions',		 main=>0,},     # Print motor medm (on wish list)
#           {func=>'print_voltages',		txt=>'Print KBM voltages',		 main=>0,},     # Print voltage medms (on wish list)
           );

  $inline        = 1;						# default can be overwritten by command line
# $useTee        = 1;						# create second log from all routines (set later depending on $inline)
  $pause_after   = 0;						# default can be overwritten by command line
  $debug         = 0;						# default can be overwritten by command line
  $simulate      = 0;						# 1=do not run any function, just enter-sleep-return
  $includeExtras = 0;						# 1=show extra functions, 0=no
  $CAQTDM        = 0;						# What screens to open: 0=medm (preferred), 1=CaQtDM (may not work)
#-----------------------------------------------------------------------
  $SIG{HUP}  = 'signal_handler';	#(1) hangup (see signum.h)
  $SIG{INT}  = 'signal_handler';	#(2) interrupt (see signum.h)
  $SIG{QUIT} = 'signal_handler';	#(3) quit (see signum.h)
# $SIG{KILL} = 'signal_handler';	#(9) this is what is sent by kill -9 -- unblockable
  $SIG{TERM} = 'signal_handler';	#(15) terminate -- sent by kill without paramaters
# $SIG{STOP} = 'signal_handler';	#(19) stop -- unblockable
#-----------------------------------------------------------------------
  $syntax = "\n\n".'Syntax: '.$script;
  foreach $arg (@beamlinelist) {$syntax .=' ['.$arg.']';}
  $syntax .= ' [inline|inxterm] [tee|notee] [pause] [extras] [debug] [simulate] [history=N]'."\n\n";
  $cross = 0;
  foreach $arg (@ARGV) {
     if (grep(/^$arg:?$/, @beamlinelist)) {
        $arg=~s/:$//g;
        if ($beamline ne $arg) {
           $beamline = $arg;
           $cross = 1;
           print STDOUT "\n".'!!! '.$script.': Cross operation for beamline='.$beamline."\n";
           do $PEZCA.'/Common/beamline_'.$beamline.'.pl';
        }
     }
     elsif ($arg =~ /^inline$/i)   {$inline        = 1; print STDOUT $script.': inline option requested!'."\n";}
     elsif ($arg =~ /^inxterm$/i)  {$inline        = 0; print STDOUT $script.': inxterm option requested!'."\n";}
     elsif ($arg =~ /^tee$/i)      {$useTee        = 1; print STDOUT $script.': tee option requested!'."\n";}
     elsif ($arg =~ /^notee$/i)    {$useTee        = 0; print STDOUT $script.': notee option requested!'."\n";}
     elsif ($arg =~ /^pause$/i)    {$pause_after   = 1; print STDOUT $script.': pausing after each step requested!'."\n";}
     elsif ($arg =~ /^debug$/i)    {$debug         = 1; print STDOUT $script.': debug requested!'."\n";}
     elsif ($arg =~ /^simulate$/i) {$simulate      = 1; print STDOUT $script.': simulate mode requested!'."\n";}
     elsif ($arg =~ /^extras$/i)   {$includeExtras = 1; print STDOUT $script.': including extra steps requested!'."\n";}
     elsif ($arg =~ /^history=/i) {
        $arg =~ s/^history=//i;
        if (&is_int($arg)) {$history=abs($arg); print STDOUT $script.': history='.$history.' days requested!'."\n";}
        else               {&zenity_error_exit('*** '.$script.': incorrect history parameter command line.'.$syntax);}
     }
     else                         {&zenity_error_exit('*** '.$script.': unrecornized command line argument '.$arg.$syntax);}
  }
  if (!defined $useTee) {$useTee = $inline;}
#-----------------------------------------------------------------------
  
  if ($^O =~ /Win/i && ! $simulate) {die 'This script cannot run under Windows OS';}

  if (!defined $have_KBM) {$have_KBM = 1;}					#see beamline.pl
  if (!defined $have_CRL) {$have_CRL = 0;}					#see beamline.pl
  if (!defined $have_TF)  {$have_TF  = 0;}					#see beamline.pl

  $hostname = &gethost('short');
  if    (defined $ENV{USER})     {$username = $ENV{USER};}
  elsif (defined $ENV{USERNAME}) {$username = $ENV{USERNAME};}
  else                           {$username = 'unknown';}
### Transfocator is present at IDD only:
  if ($beamline ne '23i') {
     @steps = grep {$_->{func} ne 'check_TransfocatorServer'} @steps;
  }
### Exclude checking Dectris detectors eth if we are not on ws2 because we cannot ssh to detector computer without prompt
  if ($cross || $hostname !~ /^bl[1-2]ws2/ || $username ne 'gmca') {
     print STDOUT 'Excluding "check_dectris_eth" because of not authorized host='.$hostname.' or user='.$username."\n";
     @steps = grep {$_->{func} ne 'check_dectris_eth'} @steps;
  }
### No old-style White Beam slits at ID-B:
  if ($beamline ne '23i') {
     @steps = grep {$_->{func} ne 'center_WB_slitsV'} @steps;
     @steps = grep {$_->{func} ne 'center_WB_slitsH'} @steps;
  }
### If no bimorphs at the beamline:
  if ($beamline ne '23o') {
     @steps = grep {$_->{func} ne 'backup_voltages'} @steps;
     @steps = grep {$_->{func} ne 'record_voltages'} @steps;
  }
### If no mirrors in the beam:
  if (! $have_KBM) {
     @steps = grep {$_->{func} ne 'center_HFM'} @steps;
     @steps = grep {$_->{func} ne 'center_VFM'} @steps;
#    @steps = grep {$_->{func} ne 'check_mirror_mtr_disabled'} @steps;
     @steps = grep {$_->{func} ne 'backup_voltages'} @steps;
     @steps = grep {$_->{func} ne 'record_voltages'} @steps;
  }

### If no CRL at the beamline:
  if (! $have_CRL) {
     @steps = grep {$_->{func} ne 'record_CRL'} @steps;
  }
### If no Transfocator at the beamline:
  if (! $have_TF) {
     @steps = grep {$_->{func} ne 'record_Transfocator'} @steps;
  }

# foreach my $s (@steps) {print STDOUT $s->{func},"\n";}

### Add dummy test function:
  if ($debug) {
     unshift(@steps,{func=>'test_perl',txt=>'Test Perl execution',main=>1,});   # Test running and aborting external Perl script (for debugging only)
     $email_list = ',sstepanov@anl.gov';
  } else {
#################################################################################
     $expectations = $PEZCA.'user_setup_expectations.pl'; 			# This file should be modified each run!
#    $status = do $expectations;						# load the expectations file
     $status = eval `cat $expectations`;					# load the expectations file (more strict)
     if ($@)                  {die 'Could not parse '.$expectations.': '.$@;}	# '$@' is error message
     elsif (!defined $status) {die 'Could not do '   .$expectations.': '.$!;}	# '$!' is error
#    else                     {print STDOUT 'Expectations are OK'."\n";}	#
     &setup_expectations();							# load the defaults & PV names from setup_expectations.pl
#################################################################################
  }

  if ($inline) {@scriptsPrefix = ('/usr/bin/perl', '-W');}
  else         {@scriptsPrefix = ('/usr/bin/xterm', '-e', '/usr/bin/perl', '-W');}

# $logdir          = '/home/gmca/shared/logfiles/';             if (! -e $logdir) {$logdir='./';}
  $logdir          = '/mnt/software/gmca_shared/logfiles/';     if (! -e $logdir) {$logdir='./';}
  chdir $logdir;
  $logroot         = 'user_setup';
  $struck_root     = '5um_intensity_timeslice';
  $process_running = 0;
  $process_aborted = 0;
  $need2exit       = 0;

  if (! grep(/^$history$/,@histList)) {push(@histList,$history);}
  @histList = sort{$a<=>$b} @histList;

  $bl_host = $hostList[0];
  $bl_user = '';

  $part1_print = '';
  $part1_text  = '';
  $part2_text  = '';
  $part3_text  = '';
  $part2_message     = '';
  $part3_message     = '';
  $part3_text_header = '';
  $step_err    = '';

  $nSetupSteps = @steps;

  $current_time = scalar localtime();
  print STDOUT "\n".$script.': starting at '.$current_time."\n\n";

  &draw_main();


### ---------------------------------------------------------------
### Subs ----------------------------------------------------------
### ---------------------------------------------------------------

#=================================================================

sub  draw_main () {
  my ($hdr, $title, $txt, $ncol, $line, $i);
  my ($w_hdrFrame, $w_hdrLbl);
  my ($w_usrFrame, $w_userLbl, $w_userEntry);
  my ($w_hostLbl, $w_hostSpin);
  my ($w_stepsFrame, @w_stepLbl);
  my ($w_ctrlFrame, $w_unselBtn, $w_selBtn);
  my ($nlines, $w_statsFrame, @w_statsBtn, $histBtnWidth);
  my ($w_statsHdrFrm, $w_statsHdrLbl, $w_statsHdrSpin, $w_statsHdrDays); 
  
  $fgcolor = '#000000';     $afgcolor = '#770000';
  $bgcolor = '#11A8FF';     $abgcolor = '#0097EE';

  $hdr     = 'User Setup';
  if    ($beamline eq '23i') {$hdr = '23-IDD '.$hdr;}
  elsif ($beamline eq '23o') {$hdr = '23-IDB '.$hdr;}
  elsif ($beamline eq '23o') {$hdr = '23-BM '.$hdr;}
  if ($simulate) {$hdr .= ' (simulate)';}
  $title   = 'Select User Setup Steps';

  ### Create main window:
  $mw = MainWindow->new(-width=>10,-height=>4,-title=>$title);
  $mw->resizable( 0, 0 );

  $pad  = 2;
  $ncol = 2;
  $line = 0;
  $fontHdr = $mw->fontCreate(-size=>18, -weight=>'bold');
  $fontLbl = $mw->fontCreate(-size=>9,  -weight=>'bold');
  $fontBtn = $mw->fontCreate(-size=>11, -weight=>'bold');
  $fontSml = $mw->fontCreate(-size=>7,  -weight=>'bold');
  $fontTbl = $mw->fontCreate(-size=>11, -weight=>'bold', -family=>'Courier');

  $w_hdrFrame = $mw->Frame(-borderwidth=>0,
                           -relief=>'ridge',
                             )->grid(-row=>$line++, -column=>0, -columnspan=>2, -padx=>2, -pady=>2, -ipadx=>0, -ipady=>0, -sticky=>'ew');
  if (! $includeExtras) {$txt = 'More';} else {$txt = 'Less';}
  $w_moreBtn   = $w_hdrFrame->Button(-text=>$txt, -font=>$fontSml, 
                                     -background=>$bgcolor, -activebackground=>$abgcolor, 
                                     -foreground=>$fgcolor, -activeforeground=>$afgcolor,
                                     -command=>\&switch_more_less,
                             )->pack(-pady=>0, -padx=>0, -ipady=>0, -ipadx=>0, -anchor=>'ne');
  $w_hdrLbl   = $w_hdrFrame->Label(-text=>$hdr, -font=>$fontHdr,
                                   -foreground => 'blue', 
                             )->pack(-pady=>0, -padx=>0, -ipady=>0, -ipadx=>0);


  $w_usrFrame = $mw->Frame(-borderwidth=>0,
                           -relief=>'ridge',
                             )->grid(-row=>$line++, -column=>0, -columnspan=>2, -padx=>2, -pady=>2, -ipadx=>0, -ipady=>0, -sticky=>'e');
  $txt = 'User';
  $w_userLbl = $w_usrFrame->Label(-text =>$txt,
                                  -font=>$fontLbl, -justify=>'left'
                             )->grid(-row=>0, -column=>0, -padx=>$pad, -pady=>$pad, -sticky=>'e');
  $w_userEntry = $w_usrFrame->Entry(-width=>21, 
                                    -exportselection=>1,
                                    -textvariable=>\$bl_user,
                                    -validatecommand=>\&reconf_startBtnUser, 
                                    -validate=>'all',                           	#focus, focusin, focusout, key, all.
                                    -font=>$fontLbl, -justify=>'left',
                                    -background=>'white', -disabledbackground=>'#999999', -highlightbackground=>'white',
                                    -foreground=>'black', -disabledforeground=>'#999999',
                             )->grid(-row=>0, -column=>1, -padx=>$pad,-pady=>$pad, -sticky=>'w');

  $txt = 'Host';
  $w_hostLbl   = $w_usrFrame->Label(-text =>$txt,
                                    -font=>$fontLbl, -justify=>'left'
                             )->grid(-row=>1, -column=>0, -padx=>$pad, -pady=>$pad, -sticky=>'e');
  $w_hostSpin  = $w_usrFrame->BrowseEntry(-variable =>\$bl_host,
                                          -choices=>\@hostList,
                                          -state=>'readonly', 
                                          -command=>\&reconf_startBtn,
                                          -font=>$fontLbl, 
                                          -justify=>'left',
	   			          -background=>'white', -disabledbackground=>'white', 
                                          -foreground=>'black', -disabledforeground=>'black',
	   			          -highlightbackground=>'white'
                             )->grid(-row=>1, -column=>1, -padx=>$pad, -pady=>$pad, -sticky=>'w');
  $w_hostSpin->Subwidget('arrow')->configure(-background=>$bgcolor, -activebackground=>$abgcolor, 
                                             -foreground=>$fgcolor, -activeforeground=>$afgcolor,);
# $w_hostSpin->Subwidget('choices')->configure(-background=>$bgcolor, -foreground=>$fgcolor,);
# $w_hostSpin->Subwidget('slistbox')->configure(-background=>$bgcolor, -foreground=>$fgcolor,);

  $w_stepsFrame = $mw->Frame(-borderwidth=>4, 
                             -relief=>'ridge'
                         )->grid(-row=>$line++, -column=>0, -padx=>$pad,-pady=>$pad, -sticky=>'w');

  for ($i=0; $i<$nSetupSteps; $i++) {
     if (!defined $steps[$i]{main}) {$steps[$i]{main} = 0;}
     if ($steps[$i]{main} == 1 || $includeExtras) {
        if (!defined $steps[$i]{sel}) {$steps[$i]{sel} = 1;}				#initially mark to run all

        $w_stepLbl[$i] = $w_stepsFrame->Label(-text =>$steps[$i]{txt},
                                    -font=>$fontLbl, 
                                    -justify=>'left'
                         )->grid(-row=>$i, -column=>0, -padx=>$pad, -pady=>$pad, -sticky=>'e');
        if ($steps[$i]{main} == 0) {$w_stepLbl[$i]->configure(-foreground=>'navy');}

        $txt = '';
        $w_stepChkBtn[$i] = $w_stepsFrame->Checkbutton(-text=>$txt, 
                                                        -variable=>\$steps[$i]{sel}, 
                                                        -command=>\&reconf_startBtn,
                                                        -font=>$fontLbl, 
                                                        -justify=>'left'
                         )->grid(-row=>$i, -column=>1, -padx=>2, -pady=>2, -sticky=>'w');
        $w_stepresultLbl[$i] = $w_stepsFrame->Label(-text=>$txt,
                                                     -font=>$fontLbl, 
                                                     -justify=>'left',
                                                     -width=>18,
                         )->grid(-row=>$i, -column=>2, -padx=>2, -pady=>2, -sticky=>'w');
     } else {
        $steps[$i]{sel} = 0;
     }
  }

  $w_ctrlFrame = $mw->Frame(-borderwidth=>0,
                            -relief=>'ridge',
                             )->grid(-row=>$line++, -column=>0, -columnspan=>2, -padx=>10, -pady=>10);

  $column = 0;
  $txt = 'Unselect All';
  $w_unselBtn  = $w_ctrlFrame->Button(-text=>$txt, 
                                      -font=>$fontBtn, 
                                      -command=>\&unselectAll,
                                      -background=>$bgcolor, -activebackground=>$abgcolor, 
                                      -foreground=>$fgcolor, -activeforeground=>$afgcolor,
                             )->grid(-row=>0, -column=>$column++, -pady=>$pad, -padx=>7, -sticky=>'w',);

  $txt = 'Select All';
  $w_selBtn    = $w_ctrlFrame->Button(-text=>$txt, 
                                      -font=>$fontBtn, 
                                      -command=>\&selectAll,
                                      -background=>$bgcolor, -activebackground=>$abgcolor, 
                                      -foreground=>$fgcolor, -activeforeground=>$afgcolor,
                             )->grid(-row=>0, -column=>$column++, -pady=>$pad, -padx=>7, -sticky=>'w',);

  $txt = 'Clear';
  $w_clearBtn  = $w_ctrlFrame->Button(-text=>$txt, 
                                      -font=>$fontBtn, 
                                      -command=>\&clearResults,
                                      -background=>$bgcolor, -activebackground=>$abgcolor, 
                                      -foreground=>$fgcolor, -activeforeground=>$afgcolor,
                             )->grid(-row=>0, -column=>$column++, -pady=>$pad, -padx=>7, -sticky=>'w',);

  $txt = 'Start';
  $w_startBtn  = $w_ctrlFrame->Button(-text=>$txt, 
                                      -font=>$fontBtn, 
                                      -command=>\&run_selected_steps,
                                      -background=>$bgcolor, -activebackground=>$abgcolor, 
                                      -foreground=>$fgcolor, -activeforeground=>$afgcolor,
                             )->grid(-row=>0, -column=>$column++, -pady=>$pad, -padx=>7, -sticky=>'w',);

  $txt = 'Abort Step';
  $w_abortBtn  = $w_ctrlFrame->Button(-text=>$txt, 
                                      -font=>$fontBtn, 
                                      -command=>\&abort_step,
                                      -background=>$bgcolor, -activebackground=>$abgcolor, 
                                      -foreground=>$fgcolor, -activeforeground=>$afgcolor,
                             )->grid(-row=>0, -column=>$column++, -pady=>$pad, -padx=>7, -sticky=>'w',);

  $txt = 'Exit';
  $w_cancelBtn   = $w_ctrlFrame->Button(-text=>$txt, 
                                        -font=>$fontBtn, 
                                        -command=>\&goodbye,
                                        -background=>$bgcolor, -activebackground=>$abgcolor, 
                                        -foreground=>$fgcolor, -activeforeground=>$afgcolor,
                             )->grid(-row=>0, -column=>$column++, -pady=>$pad, -padx=>7, -sticky=>'e',);

  $nlines = $line-3;

  $w_statsFrame = $mw->Frame(-borderwidth=>5,
                             -relief=>'ridge',
                             )->grid(-row=>2, -column=>1, -rowspan=>$nlines, -padx=>10, -pady=>10, -sticky=>'n');

  $histBtnWidth=24;
  $line = 0;

  $w_statsHdrFrm = $w_statsFrame->Frame(-borderwidth=>0,
                             -relief=>'ridge',
                             )->grid(-row=>$line++, -column=>0, -padx=>7, -pady=>$pad, );

  $w_statsHdrLbl = $w_statsHdrFrm->Label(-text=>'HISTORY', 
                                      -font=>$fontBtn,
                                      -foreground => 'blue', 
                             )->grid(-row=>0, -column=>0, -padx=>7, -ipady=>0);

  $historySpin = $history;
  $w_statsHdrSpin  = $w_statsHdrFrm->BrowseEntry(-variable =>\$historySpin,
                                          -choices=>\@histList,
                                          -browsecmd=>\&reconf_Hist,
                                          -width=>4,
                                          -state=>'readonly', 
                                          -font=>$fontLbl, 
                                          -justify=>'right',
                                          -background=>'white', -disabledbackground=>'white', 
                                          -foreground=>'black', -disabledforeground=>'black',
                                          -highlightbackground=>'white'
                             )->grid(-row=>0, -column=>1, -padx=>7, -pady=>$pad, -sticky=>'w');
  $w_statsHdrSpin->Subwidget('arrow')->configure(-background=>$bgcolor, -activebackground=>$abgcolor, 
                                             -foreground=>$fgcolor, -activeforeground=>$afgcolor,);
### This is for state=>'normal' (default). Validattion of edits does not work well; so we use state=>'readonly'
# $w_statsHdrSpin->Subwidget('entry')->configure(-validate=>'all', -validatecommand=>\&reconf_Hist,);

  $w_statsHdrDays = $w_statsHdrFrm->Label(-text=>'days', 
                                      -font=>$fontBtn,
                             )->grid(-row=>0, -column=>2, -padx=>7, -ipady=>0);

  $w_statsBtn[$line-1]  = $w_statsFrame->Button(-text=>'Mono calibration history', 
                                      -width=>$histBtnWidth,
                                      -font=>$fontBtn, 
                                      -command=>\&mono_calibration_history,
                                      -background=>$bgcolor, -activebackground=>$abgcolor, 
                                      -foreground=>$fgcolor, -activeforeground=>$afgcolor,
                             )->grid(-row=>$line++, -column=>0, -padx=>7, -pady=>$pad, );
  
  $w_statsBtn[$line-1]  = $w_statsFrame->Button(-text=>'Beam steering history', 
                                      -width=>$histBtnWidth,
                                      -font=>$fontBtn, 
                                      -command=>\&beam_steering_history,
                                      -background=>$bgcolor, -activebackground=>$abgcolor, 
                                      -foreground=>$fgcolor, -activeforeground=>$afgcolor,
                             )->grid(-row=>$line++, -column=>0, -padx=>7, -pady=>$pad, );
  
  $w_statsBtn[$line-1]  = $w_statsFrame->Button(-text=>'DCM-VSUM and I0 history', 
                                      -width=>$histBtnWidth,
                                      -font=>$fontBtn, 
                                      -command=>\&I0_history,
                                      -background=>$bgcolor, -activebackground=>$abgcolor, 
                                      -foreground=>$fgcolor, -activeforeground=>$afgcolor,
                             )->grid(-row=>$line++, -column=>0, -padx=>7, -pady=>$pad, );
  
  $w_statsBtn[$line-1]  = $w_statsFrame->Button(-text=>'Beam through collimators hist.', 
                                      -width=>$histBtnWidth,
                                      -font=>$fontBtn, 
                                      -command=>\&minibeam_intensities_history,
                                      -background=>$bgcolor, -activebackground=>$abgcolor, 
                                      -foreground=>$fgcolor, -activeforeground=>$afgcolor,
                             )->grid(-row=>$line++, -column=>0, -padx=>7, -pady=>$pad, );
  
  $w_statsBtn[$line-1]  = $w_statsFrame->Button(-text=>'Beam sizes history', 
                                      -width=>$histBtnWidth,
                                      -font=>$fontBtn, 
                                      -command=>\&beam_sizes_history,
                                      -background=>$bgcolor, -activebackground=>$abgcolor, 
                                      -foreground=>$fgcolor, -activeforeground=>$afgcolor,
                             )->grid(-row=>$line++, -column=>0, -padx=>7, -pady=>$pad, );
  
  $w_statsBtn[$line-1]  = $w_statsFrame->Button(-text=>'Collimator preset pos.history', 
                                      -width=>$histBtnWidth,
                                      -font=>$fontBtn, 
                                      -command=>\&collimator_preset_pos_history,
                                      -background=>$bgcolor, -activebackground=>$abgcolor, 
                                      -foreground=>$fgcolor, -activeforeground=>$afgcolor,
                             )->grid(-row=>$line++, -column=>0, -padx=>7, -pady=>$pad, );
  
  $w_statsBtn[$line-1]  = $w_statsFrame->Button(-text=>'Collimator preset ang.history', 
                                      -width=>$histBtnWidth,
                                      -font=>$fontBtn, 
                                      -command=>\&collimator_preset_ang_history,
                                      -background=>$bgcolor, -activebackground=>$abgcolor, 
                                      -foreground=>$fgcolor, -activeforeground=>$afgcolor,
                             )->grid(-row=>$line++, -column=>0, -padx=>7, -pady=>$pad, );
  
  $w_statsBtn[$line-1]  = $w_statsFrame->Button(-text=>'Guard Slits calibration history', 
                                      -width=>$histBtnWidth,
                                      -font=>$fontBtn, 
                                      -command=>\&guard_slits_calib_history,
                                      -background=>$bgcolor, -activebackground=>$abgcolor, 
                                      -foreground=>$fgcolor, -activeforeground=>$afgcolor,
                             )->grid(-row=>$line++, -column=>0, -padx=>7, -pady=>$pad, );
  
  $w_statsBtn[$line-1]  = $w_statsFrame->Button(-text=>'Staff notes history', 
                                      -width=>$histBtnWidth,
                                      -font=>$fontBtn, 
                                      -command=>\&notes_history,
                                      -background=>$bgcolor, -activebackground=>$abgcolor, 
                                      -foreground=>$fgcolor, -activeforeground=>$afgcolor,
                             )->grid(-row=>$line++, -column=>0, -padx=>7, -pady=>$pad, );

  &fill_mtrList();
# foreach my $m (@mtrList) {print STDOUT $m,"\n";}
  if (@mtrList > 0) {
     $mtrSelected = $mtrList[0];
     $w_statsMtrSpin  = $w_statsFrame->BrowseEntry(-variable =>\$mtrSelected,
                                          -choices=>\@mtrList,
                                          -state=>'readonly', 
                                          -font=>$fontLbl, 
                                          -justify=>'right',
                                          -background=>'white', -disabledbackground=>'white', 
                                          -foreground=>'black', -disabledforeground=>'black',
                                          -highlightbackground=>'white'
                             )->grid(-row=>$line++, -column=>0, -padx=>7, -pady=>$pad, -sticky=>'we');
     $w_statsMtrSpin->Subwidget('arrow')->configure(-background=>$bgcolor, -activebackground=>$abgcolor, 
                                                    -foreground=>$fgcolor, -activeforeground=>$afgcolor,);
  
    $w_statsBtn[$line-2]  = $w_statsFrame->Button(-text=>'Motor ActPos history', 
                                      -width=>$histBtnWidth,
                                      -font=>$fontBtn, 
                                      -command=>\&motor_ActPos_history,
                                      -background=>$bgcolor, -activebackground=>$abgcolor, 
                                      -foreground=>$fgcolor, -activeforeground=>$afgcolor,
                             )->grid(-row=>$line++, -column=>0, -padx=>7, -pady=>$pad, );
  }

  &reconf_startBtn();
  &reconf_abortBtn();

  $mw->protocol('WM_DELETE_WINDOW', [\&goodbye]);
# $mw->bind('<Destroy>', [\&goodbye]);                  #this does not work
  $mw->bind('<KeyPress-Escape>', [\&goodbye]);
# $mw->bind('<KeyPress-Return>', [\&goodbye]);
  $mw->focus();
# $mw->repeat(50=>\&refresh_gui);                       #this does not help (does not execute when we go into a callback)
  MainLoop;
}

#--------------------------------------------------------------------------
sub refresh_gui() {
  $mw->update;
}

#--------------------------------------------------------------------------
sub reconf_startBtn () {
  my ($nselected, $i);
  $nselected = 0;
  for ($i=0; $i<$nSetupSteps; $i++) {
     if (defined $steps[$i]{sel}) {
        if ($steps[$i]{sel} == 1) {$nselected++;}
     }
  }

  if (Exists($w_startBtn)) {
     if (length($bl_host) == 0 ||
         length($bl_user) == 0 ||
         $nselected       == 0) {$w_startBtn->configure(-state=>'disabled');}
     else                       {$w_startBtn->configure(-state=>'normal');}
  }
  $mw->update;
  return 1;
}

#--------------------------------------------------------------------------
sub reconf_abortBtn () {
  if (Exists($w_abortBtn)) {
     if ($process_running && !$process_aborted) {$w_abortBtn->configure(-state=>'normal');}
     else                                       {$w_abortBtn->configure(-state=>'disabled');}
  }
  $mw->update;
  return 1;
}

#--------------------------------------------------------------------------
sub reconf_startBtnUser () {
  my $proposed_user = shift(@_);
  if (defined $proposed_user && Exists($w_startBtn)) {
     if (length($bl_host) == 0 || length($proposed_user) == 0) {$w_startBtn->configure(-state=>'disabled');}
     else                                                      {$w_startBtn->configure(-state=>'normal');}
  }
  $mw->update;
  return 1;
}

#--------------------------------------------------------------------------
sub switch_more_less() {
# $mw->packForget;
  $mw->destroy();
  $includeExtras = 1 - $includeExtras;
  &draw_main();
}

#--------------------------------------------------------------------------
sub parse_geometry(\$) {                                        #pass window
  my ($windowRef, $geom, $width, $height, $offh, $offv);
  $windowRef = shift(@_);
  $geom = $$windowRef->geometry();
  ($width, $height, $offh, $offv) = (0,0,0,0);
  if ($geom =~ /^=?((\d+)x(\d+))?([+-](\d+)[+-](\d+))?$/) {
### If the string is 464x800+0+30, parsing above gives: $1=464x800 $2=464 $3=800 $4=+0+30 $5=0 $6=30
#    ($width, $height, $offh, $offv) = ($2,$3,$5,$6);
     if (defined $2) {$width  = $2;}
     if (defined $3) {$height = $3;}
     if (defined $5) {$offh   = $5;}
     if (defined $6) {$offv   = $6;}
  }
  return ($width, $height, $offh, $offv);
}

#--------------------------------------------------------------------------
sub open_comments_window() {

# my ($notesWin, $notesEditor, $notesBtnSav);
  my ($notesFnt, $notesBtnClr);
  my ($columns, $rows, $i);
  my ($mw_width, $mw_height, $mw_offh, $mw_offv, $offh, $offv, $geom);

  ($mw_width, $mw_height, $mw_offh, $mw_offv) = &parse_geometry(\$mw);
# print STDOUT "GEOMETRY: $mw_width, $mw_height, $mw_offh, $mw_offv\n";
  $offh = $mw_offh + $mw_width+10;
  $offv = $mw_offv;
  $geom = '600x400+'.$offh.'+'.$offv;
  print STDOUT 'GEOMETRY: '.$geom."\n";
# $notesWin = $mw->Toplevel(-title=>'Notes and Comments', -width=>'600', -height=>'200');
  $notesWin = $mw->Toplevel(-title=>'Notes and Comments');
  $notesWin->geometry($geom);
  $notesFnt = $mw->fontCreate(
          -size => 14,
          -weight => 'normal',
#         -weight => 'bold',
          -family => 'Courier',);

  $notesEditor = $notesWin->Scrolled('Text',
                            -scrollbars=>'se',
#                           -width=>'80',
#                           -height=>'30',
                            -font=>$notesFnt,
                            -foreground=>'#000000',
                            -background=>'#FFFFFF',    )->grid(-row=>0, -column=>0, -columnspan=>2, -padx=>5, -pady=>5, -sticky=>'nsew');

  $notesBtnSav = $notesWin->Button(-text=>'Add to log',
                            -background=>$bgcolor, -activebackground=>$abgcolor,
                            -foreground=>$fgcolor, -activeforeground=>$afgcolor,
#                           -width=>12,
#                           -state=>'disabled',
                            -command=>\&save_notes,     )->grid(-row=>1, -column=>0, -padx=>5, -pady=>5);

  $notesBtnClr = $notesWin->Button(-text=>'Clear',
                            -background=>$bgcolor, -activebackground=>$abgcolor,
                            -foreground=>$fgcolor, -activeforeground=>$afgcolor,
#                           -width=>12,
#                           -state=>'disabled',
                            -command=>sub{$notesEditor->Contents('')},)->grid(-row=>1, -column=>1, -padx=>5, -pady=>5);

  ($columns, $rows) = $mw->gridSize(); 
  for ($i = 0; $i<$columns; $i++) {
     $notesWin->gridColumnconfigure($i, -weight=>1);
  } 
  $notesWin->gridRowconfigure(0,-weight=>1);                            #allow resizing entry window
  
  $notesBtnSav->configure(-state=>'disabled');                          #disable saving for now
}

#--------------------------------------------------------------------------
sub save_notes() {
  my ($notes);
  $notes = $notesEditor->Contents();
  if (length($notes) > 0) {
     if (! fileno(\*DAT)) {
        $filehandle = open (DAT,'>> '.$logfile); 				#open DAT file for writing
        if (!defined $filehandle) {
           print STDOUT "\n".'*** '.$script.': Could not open '.$logfile."\n\n";
           return;
        }
        select DAT; $|=1; 							#set unbuffered output
     }
     print DAT "\n".'--------------------------------------------------'."\n\n";
     print DAT '===> HOST NOTES by '.$bl_host.':'."\n\n";
     print DAT $notes."\n";
     close(DAT);
  }
  $notesEditor->Contents('');
  $notesWin->destroy;
}

#--------------------------------------------------------------------------
sub run_selected_steps() {
  my ($i, $func, $confirmBox, $answer, $str, $opt);

  $process_running = 1;
  $process_aborted = 0;
  &reconf_abortBtn();

### Open file with timestamp included in filename -----------------

  $current_time = scalar localtime();
  @ct           = localtime();						# This is ($sec,$min,$hour,$mday,$mon,$year$,$wday,$yday,$isdst)
  ($sec,$min,$hour,$mday,$mon,$year) =
                  ($ct[0],$ct[1],$ct[2],$ct[3],$ct[4]+1,$ct[5]+1900);
### Attention: do not change extension from .log to anything else - it is parsed below
  $logfile      = sprintf $logdir.$logroot.'_%s_%04d-%02d-%02d_%02d-%02d-%02d.log',
                                           $beamline,$year,$mon,$mday,$hour,$min,$sec;
  $logfile      =~ s/ /0/g;
  $teelog       = $logfile.'2';
  $struck_file  = sprintf $logdir.$struck_root.'_%s_%04d-%02d-%02d_%02d-%02d-%02d.dat',
                                           $beamline,$year,$mon,$mday,$hour,$min,$sec;
  $struck_file  =~ s/ /0/g;
  if (! $simulate) {
     print STDOUT "\n".'Opening log file ['.$logfile.']...'."\n";
     $filehandle = open (DAT,'> '.$logfile); 				#open DAT file for writing
     if (!defined $filehandle) {
        &press_key_exit('*** '.$script.': Could not open '.$logfile);
     }
     select DAT; $|=1; 							#set unbuffered output

     ### Print header
     print DAT 'USER: '.$bl_user."\n";
     print DAT 'DATE: '.$current_time."\n";
     print DAT 'HOST: '.$bl_host."\n";
     print DAT 'BMLN: '.$beamline."\n\n\n";
     print DAT '--------------------------------------------------'."\n\n";
     print DAT 'EXECUTING SETUP STEPS:'."\n\n";

     ### Part 2 and 3 initialization
     $part2_text  = "\n\n";
     $part2_text .= '--------------------------------------------------'."\n\n\n";
     $part2_text .= 'DETAILS:'."\n\n";

     $part3_text  = '';
     $part3_text_header  = '--------------------------------------------------'."\n\n\n";
     $part3_text_header .= 'ERRORS:'."\n\n";
  }

  ### Clear previous progress labels
  &clearResults();
  $w_moreBtn->configure(-state=>'disabled');
  $w_cancelBtn->configure(-text=>'Abort All');
  $w_startBtn->focus();
  $w_startBtn->configure(-state=>'disabled');
  $w_clearBtn->configure(-state=>'disabled');
  $mw->update;

  ### Open second window where one can add comments to the end of log:
  &open_comments_window();
  
  ### Loop over all functions:
  $currDisplay = 0;
  for ($i=0; $i<$nSetupSteps; $i++) {
     $currStep = $i;
     if ($steps[$i]{main} == 1 || $includeExtras) {
        if ($steps[$i]{sel}) {
           $w_stepresultLbl[$i]->configure(-text=>'Working...', -foreground=>'Blue');
           $mw->update;
           $currDisplay++;
           $func = $steps[$i]{func};
           $status = &do_step(\&$func);
           if    ($process_aborted) {$w_stepresultLbl[$i]->configure(-text=>'Aborted', -foreground=>'Red');}
           elsif ($status)          {$w_stepresultLbl[$i]->configure(-text=>'Done, some problem', -foreground=>'Red');}
           else                     {$w_stepresultLbl[$i]->configure(-text=>'Done, all good',     -foreground=>'Green');}
           if ($process_aborted == 2) {$process_aborted = 0; &reconf_abortBtn();}
        } else {
           $w_stepresultLbl[$i]->configure(-text=>'Skipped', -foreground=>'Grey');
        }
        $w_stepChkBtn[$i]->configure(-state=>'disabled');
        if ($need2exit) {&goodbye_real('confirm');}
        $mw->update;
        if ($process_aborted == 1)  {last;} 
     }
  }
  ### Cleanup "progress" (disabled checkboxes)
  for ($i=0; $i<$nSetupSteps; $i++) {
     if ($steps[$i]{main} == 1 || $includeExtras) {
        $w_stepChkBtn[$i]->configure(-state=>'normal');
        $mw->update;
     }
  }
  $process_running = 0;
  $process_aborted = 0;
  &reconf_abortBtn();
    
  $w_clearBtn->configure(-state=>'normal');
  $w_moreBtn->configure(-state=>'normal');
  $w_cancelBtn->configure(-text=>'Exit');
  $w_startBtn->focus();
  $mw->update;

  if (! $simulate) {
     ### Add report footer and close the file:
     print DAT $part1_print;
     if (length($part3_text) > 0) {print DAT $part3_text_header,$part3_text;}
     print DAT $part2_text;

     print STDOUT "\n".$part1_print."\n";
     if (length($part3_text) > 0) {print STDOUT "\n",$part3_text_header,$part3_text."\n";}
     close DAT;
  
### If comments windows exists and contain text, ask if you want it to be added to the log now.
### If not, activate the "Add comment to the log" button and wait for completion:
     if (Exists($notesWin)) {
        $confirmBox = $mw->Dialog(-title=>$script,
                                  -text=>'Do you want to add notes?',
                                  -wrap=>0,
                                  -font=>$fontBtn,
                                  -buttons=>['Save Now', 'Edit & Save', 'No'],
                                  -default_button=>'Edit & Save',
                                  -bitmap=>'question',
                                  -takefocus=>1,
                                  -width=>30,
                                  );
        $confirmBox->Subwidget('B_Save Now')->   configure(-background=>$bgcolor, -activebackground=>$abgcolor, 
                                                           -foreground=>$fgcolor, -activeforeground=>$afgcolor,);
        $confirmBox->Subwidget('B_Edit & Save')->configure(-background=>$bgcolor, -activebackground=>$abgcolor, 
                                                           -foreground=>$fgcolor, -activeforeground=>$afgcolor,);
        $confirmBox->Subwidget('B_No')->         configure(-background=>$bgcolor, -activebackground=>$abgcolor, 
                                                           -foreground=>$fgcolor, -activeforeground=>$afgcolor,);
        $answer = $confirmBox->Show;
        if    ($answer eq 'Save Now') {
           &save_notes();
        }
        elsif ($answer eq 'Edit & Save') {
           $notesBtnSav->configure(-state=>'normal');                   #enable saving
           print STDOUT 'Waiting for either notes saved or notes window closed...'."\n";
           do {
              select(undef,undef,undef,0.05);
              &refresh_gui();
           } while (Exists($notesWin));
           print STDOUT '... OK!'."\n";
        } else {
           $notesWin->destroy;
        }  
     }

     if (!defined $makePDF) {$makePDF = 0;}

     if ($makePDF == 1) {
        $filePS  = $logfile;  $filePS  =~ s/\.log$/\.ps/i;   if ($filePS  eq $logfile) {$filePS  .= '.ps';}
        $filePDF = $logfile;  $filePDF =~ s/\.log$/\.pdf/i;  if ($filePDF eq $logfile) {$filePDF .= '.pdf';}
        @prg = ('/usr/bin/enscript', '-t', 'User_Setup_Log', '-p', $filePS, $logfile);
        $status1 = &system_with_fork(@prg);
        if (-f $filePS) {
           @prg = ('/usr/bin/ps2pdf', $filePS, $filePDF);
           $status1 = &system_with_fork(@prg);
           if (-f $filePDF) {
              unlink $filePS;
              if (defined $viewerPDF) {
                 @prg = ($viewerPDF, $filePDF, '2>/dev/null', '&');
                 $status1 = system(join(' ',@prg));
              } else {
                print STDOUT '!!! '.$script.': no proper viewer found for '.$filePDF."\n";
              }
           } else {
              print STDOUT '!!! '.$script.': failed to create '.$filePDF."\n";
              if (defined $viewerPDF) {
                 @prg = ($viewerPDF, $filePS, '2>/dev/null', '&');
                 $status1 = system(join(' ',@prg));
              } else {
                 print STDOUT '!!! '.$script.': no proper viewer found for '.$filePS."\n";
              }
           }
        } else {
           @prg = ('/usr/bin/enscript', $logfile, '&');
           $str = join(' ',@prg);
           print STDOUT '!!! '.$script.': failed to create '.$filePS."\n"
                       .'Using the following command to print Setup Log:'."\n"
                       .$str."\n\n";
           $status1 = system($str);
        }
     } 
     elsif ($makePDF == 2) {
        ### Request printing option:
        $mw->focus(-force);
        $confirmBox = $mw->Dialog(-title=>$script,
                                  -text=>'Print the report?',
                                  -wrap=>0,
                                  -font=>$fontBtn,
                                  -buttons=>['Yes', 'No'],
                                  -default_button=>'Yes',
                                  -bitmap=>'question',
                                  -takefocus=>1,
                                  -width=>30,
                                  );
        $confirmBox->Subwidget('B_Yes')->configure(-background=>$bgcolor, -activebackground=>$abgcolor, 
                                                  -foreground=>$fgcolor, -activeforeground=>$afgcolor,);
        $confirmBox->Subwidget('B_No')-> configure(-background=>$bgcolor, -activebackground=>$abgcolor, 
                                                  -foreground=>$fgcolor, -activeforeground=>$afgcolor,);
        $answer = $confirmBox->Show;
        if ($answer eq 'Yes') {
           @prg = ('/usr/bin/enscript', '-B', $logfile, '2>/dev/null', '&');
           $str = join(' ',@prg);
           print STDOUT 'Printing Setup Log:'."\n".$str."\n\n";
           $status1 = system($str);
        }
     }
     else {
        ### Open the log in editor where we can also print:
        if ($editor =~ /geany/) {$opt='-r -s -m';} 
        else                    {$opt=' ';}
        @prg = ($editor, $opt);
        $log_exist = 0;
        if ($useTee && -e $teelog) {push(@prg,$teelog);  $log_exist++;}
        if (-e $logfile)           {push(@prg,$logfile); $log_exist++;}
        push(@prg,'2>/dev/null','&');
        $str = join(' ',@prg);
        if ($log_exist) {
           print STDOUT 'Opening Setup Log(s) in editor:'."\n".$str."\n\n";
           $status1 = system($str);
        } else {
           print STDOUT '!!! Not opening Setup Log(s) in editor: logs not found.'."\n\n";
        }
     }
  }
  &reconf_startBtn();
}

#--------------------------------------------------------------------------
sub do_step() {
  my ($func);
  $func = $_[0];
  printf STDOUT "\n\n".'==================================================='."\n"
                      .'Step=%d:  %s'."\n\n", $currDisplay, $steps[$currStep]{txt};
  $step_err = 0;
  if (defined $func) {
     if (! $simulate) {$func->();} else {sleep 1;}
  } else {
     printf STDOUT '*** '.$script.': function [%s] is undefined.'."\n", $steps[$currStep]{func};
     $step_err = -1;
  }
  return $step_err;
}

#--------------------------------------------------------------------------
sub system_with_fork(@) {
  our ($EXTERNAL_PID, $EXTERNAL_PRG);
  my ($script, $pid, $kid, $WNOHANG, $code, $arg, @tee);

### WNOHANG is a constant defined in POSIX:
# use POSIX ":sys_wait_h";
### This constant is actually "1". So instead
### of using POSIX we define a variable:
  $WNOHANG = 1;

  $pid = fork;
  if (!defined $pid) {
#    $script = &makeScriptName();
     print STDOUT '*** '.$script.': cannot fork to execute '."@_\n";
     return 1;
  }
  elsif ($pid == 0) {
     ### Client process: execute external program
     if (! $useTee) {
        print STDOUT '    ==> Calling '."@_\n";
        exec(@_);					#this will completely replace the child with called program, but retain PID
     } else {
        ### Record all output (see https://stackoverflow.com/questions/692000/how-do-i-write-stderr-to-a-file-while-using-tee-with-a-pipe)
        @tee = @_;
#       push(@tee,        '|&', 'tee', '-a', $teelog);
        push(@tee,'2>&1', '|',  'tee', '-a', $teelog);
        if ($tee[0] =~ /term/ && $tee[1] eq '-e') {
           $tee[2] = "'".$tee[2];
           $tee[-1] .= "'";
        }
        foreach $arg (@_) {				#find the name of executed script
           if ($arg =~ /\.pl$/) {
              $arg =~ s/^.*\///g;			#eliminate path from script name          
              system('echo -e "\n=============================================== Executing '.$arg.'\n" >> '.$teelog);
              last;
           }
        }
        print STDOUT '    ==> Calling '."@tee\n";
        exec(join(' ',@tee));				#this will completely replace the child with called program, but retain PID
     }
  }
  else {
     ### Parent process: wait for child to exit
     $EXTERNAL_PID = $pid;				#this is for aborting by signal handler
     undef $EXTERNAL_PRG;
     foreach $arg (@_) {				#find the name of executed script
        if ($arg =~ /\.pl$/) {
           $EXTERNAL_PRG = $arg;
           $EXTERNAL_PRG =~ s/^.*\///g;			#eliminate path from script name          
           last;
        }
     }
     do {						#wait the script to finish
        select(undef,undef,undef,0.05);
        &refresh_gui();
#       $mw->update;
#       $kid = waitpid(-1, WNOHANG);			#wait for all childs
        $kid = waitpid($pid, $WNOHANG);			#wait for particular child
     } while ($kid == 0);				#this should be "0" while process runs
     $code = $? >> 8;
     if ($code) {
#       $script = &makeScriptName();
        print STDOUT '!!!!'.$script.': execution of ['."@_".'] finished with status='.$code."\n";
     }
     undef $EXTERNAL_PID;
     undef $EXTERNAL_PRG;
     return 0;
  }
}

#--------------------------------------------------------------------------
sub unselectAll() {
  my ($i);
  for ($i=0; $i<$nSetupSteps; $i++) {
     if ($steps[$i]{main} == 1 || $includeExtras) {
        $steps[$i]{sel} = 0;
     }
  }
  $w_startBtn->configure(-state=>'disabled');
}

#--------------------------------------------------------------------------
sub selectAll() {
  my ($i);
  for ($i=0; $i<$nSetupSteps; $i++) {
     if ($steps[$i]{main} == 1 || $includeExtras) {
        $steps[$i]{sel} = 1;
     }
  }
  &reconf_startBtn();
}

#--------------------------------------------------------------------------
sub clearResults() {
  my ($i);
  for ($i=0; $i<$nSetupSteps; $i++) {
     if ($steps[$i]{main} == 1 || $includeExtras) {
        $w_stepresultLbl[$i]->configure(-text=>' ', -foreground=>'Black');
     }
  }
}

#--------------------------------------------------------------------------
sub goodbye() {
  our ($EXTERNAL_PID, $EXTERNAL_PRG);
  if ($process_running) {
     $need2exit = 1; 
#    if (defined $EXTERNAL_PID) {&goodbye_real();}			#no confirmation
     &goodbye_real();							#no confirmation (changed 2020/11/09)
     return;
  } else {
     &goodbye_real('confirm');
  }
}

#--------------------------------------------------------------------------
sub goodbye_real(;$) {
  our ($EXTERNAL_PID, $EXTERNAL_PRG);
  my ($dialogBox, $confirm, $answer);
  $confirm = shift(@_);
  if (!defined $confirm) {$confirm = 'no';}
  $need2exit = 0;
### Confirmation dialog:
  if ($confirm =~ /^confirm$/i) {
     $mw->focus(-force);
     if ($process_running) {
        $dialogBox = $mw->Dialog(-title=>$script,
                              -text=>'Interrupt user setup?',
                              -wrap=>0,
                              -font=>$fontBtn,
                              -buttons=>['Exit', 'Abort All', 'Abort Step', 'Continue'],
                              -default_button=>'Continue',
                              -bitmap=>'question',
                              -takefocus=>1,
                              -width=>30,
                              );
        $dialogBox->Subwidget('B_Exit')      ->configure(-background=>$bgcolor, -activebackground=>$abgcolor, 
                                                       -foreground=>$fgcolor, -activeforeground=>$afgcolor,);
        $dialogBox->Subwidget('B_Abort All') ->configure(-background=>$bgcolor, -activebackground=>$abgcolor, 
                                                       -foreground=>$fgcolor, -activeforeground=>$afgcolor,);
        $dialogBox->Subwidget('B_Abort Step')->configure(-background=>$bgcolor, -activebackground=>$abgcolor, 
                                                       -foreground=>$fgcolor, -activeforeground=>$afgcolor,);
        $dialogBox->Subwidget('B_Continue')  ->configure(-background=>$bgcolor, -activebackground=>$abgcolor, 
                                                       -foreground=>$fgcolor, -activeforeground=>$afgcolor,);
     } else {
        $dialogBox = $mw->Dialog(-title=>$script,
                              -text=>'Exit user setup?',
                              -wrap=>0,
                              -font=>$fontBtn,
                              -buttons=>['Yes', 'No'],
                              -default_button=>'No',
                              -bitmap=>'question',
                              -takefocus=>1,
                              -width=>30,
                             );
        $dialogBox->Subwidget('B_Yes')->configure(-background=>$bgcolor, -activebackground=>$abgcolor, 
                                                  -foreground=>$fgcolor, -activeforeground=>$afgcolor,);
        $dialogBox->Subwidget('B_No')-> configure(-background=>$bgcolor, -activebackground=>$abgcolor, 
                                                  -foreground=>$fgcolor, -activeforeground=>$afgcolor,);
     }
     $answer = $dialogBox->Show;
  } else {
     if ($process_running) {$answer = 'Abort All';}
     else                  {$answer = 'Exit';}
  }
#----
  if ($answer eq 'Yes' || $answer eq 'Exit') {
     &clearAllCaMonitors();
     exit;
  }
  &abort_real($answer);
}

#--------------------------------------------------------------------------
sub abort_step() {
  &abort_real('Abort Step');
}

#--------------------------------------------------------------------------
sub abort_real(;$) {
  our ($EXTERNAL_PID, $EXTERNAL_PRG);
  my ($answer, $nkilled);
  my (@subpid, $pid, $pname, $pnow);
  my ($waitmax, $waitstep, $wait_, $n, $i);
  $answer = shift(@_);
  if (!defined $answer) {return;}
#
  if ($answer eq 'Abort All' || $answer eq 'Abort Step') {
     if ($answer eq 'Abort All') {$process_aborted = 1;}
     else                        {$process_aborted = 2;}
     &reconf_abortBtn();
     if (defined $EXTERNAL_PID) {
        if (!defined $EXTERNAL_PRG ) {$EXTERNAL_PRG = 'external script';}
        print STDOUT '!!! '.$script.': killing ['.$EXTERNAL_PRG.'] with PID='.$EXTERNAL_PID."\n";
        @subpid = split(/\n/, `/usr/bin/pgrep -P $EXTERNAL_PID`);		#find all subprocesses for given parent
        $n = @subpid;
        if ($n > 0) {print STDOUT '--- Including subrocesses: '."@subpid\n";}
        unshift(@subpid,$EXTERNAL_PID);
        $n++;
        $nkilled = 0;
        for ($i=$n; $i>0; $i--) {
           $pid = $subpid[$i-1];
           $pname = `/bin/ps -h -p $pid -o command`;
           $pname =~ s/[\r\n]//g;
           if (length($pname) == 0 || $pname =~ /\<defunct\>/) {next;}
           print STDOUT '  Sending SIGINT to PID='.$pid. '('.$pname.')...';
           kill('SIGINT',$pid);         					#kill all sub-subrocesses
           $waitmax  = 7;
           $waitstep = 0.5;
           $wait_    = 0;
           while ($wait_ < $waitmax) {
              select(undef,undef,undef,$waitstep);
              $wait_ += $waitstep;
              $pnow = `/bin/ps -h -p $pid -o comm=`;
              $pnow =~ s/[\s\r\n]//g;
              if    (length($pnow) == 0)     {print STDOUT ' OK!'."\n"; last;}
              elsif ($pnow =~ /\<defunct\>/) {print STDOUT ' OK (defunct)!'."\n"; last;}
           }
           if (length($pnow) > 0 && $pnow !~ /\<defunct\>/) {
              print STDOUT ' Failed to abort, killing!'."\n";
              print STDOUT '  Sending SIGKILL to PID='.$pid. '('.$pname.')'."\n";
              kill('SIGKILL',$pid);         					#kill all sub-subrocesses
           }
           $nkilled++;                                				#kill all sub-subrocesses
        }
        print STDOUT '!!! '.$script.': aborted/killed '.$nkilled.' processes.'."\n";
     }
  }
}

#=================================================================

sub test_perl () {
  ### This is only used for debugging how to run/abort external scripts:
  my ($scr);
  $scr = 'test_perl.pl';
  $step_err = 0;
  $part2_message = '';
  $part3_message = '';
  print STDOUT "\t".'Testing Perl'."\n";
  @prg = (@scriptsPrefix, $PEZCA.$scr, $beamline, 'save=./', 'medm');
  $status1 = &system_with_fork(@prg);
  if ($pause_after) {&press_key_continue('Finished '.$scr.', status='.$status1);}

  if ($status1) {
     $message = "\t".$scr.' reports unclean exit.'."\n";
     $step_err = 1;
  }
}

#=================================================================

sub verifyConfigTcl () {
  ### This is only used for debugging how to run/abort external scripts:
  my ($scr);
  $scr = 'verifyConfigTcl.pl';
  $step_err = 0;
  $part2_message = '';
  $part3_message = '';
  printf STDOUT "\t".'Verifying config%s.tcl'."\n", $beamline;
  @prg = (@scriptsPrefix, $PEZCA.$scr, $beamline);
  $status1 = &system_with_fork(@prg);
  if ($pause_after) {&press_key_continue('Finished '.$scr.', status='.$status1);}

  if ($status1) {
     $message = "\t".$scr.' reports unclean exit.'."\n";
     $step_err = 1;
  }
}

#=================================================================

sub check_EPS_monitor () {

  ### check if EPS monotor is running. If not, report to host & send emails.

  my ($subscript, $status, $msg, @monitorSectionPVs, @monitorSectionFlags);
  my ($timestamp, $pv, $flag, $fl, $flagstr, $nerr, $nflags, $desc, $i);
  my ($epoch, $epochEPS, $timestampEPS, $epochPV, $timestampPV);

  $subscript = 'check_EPS_monitor';
  $step_err  = 0;
  $epoch     = time;
  $timestamp = localtime($epoch);
  $msg       = "\n".'Current time: '.$timestamp."\n";
  $nerr      = 0;

  if ($beamline eq '23i' || $beamline eq '23o') {

     print STDOUT 'Checking EPS...'."\n";

### Check if EPS IOC and EPS monitoring script are running:
     $epochPV     = '23o:epsmon_epoch';
     $timestampPV = '23o:epsmon_timestamp';
     ($status, $epochEPS) = Pezca::GetDouble($epochPV);
     if ($status) {
        $step_err = 1;
        &report_only($subscript,$status,$epochPV);
        $epochEPS = 0;
        $nerr++;
        $msg .= 'Error accessing "'.$epochPV.'".'."\n"
              . 'It is possible that DirectNet IOC is not running.'."\n";
        &part2($msg);
        &part3($msg);
        &zenity_error ($msg,1,'nowait');
        &email('EPS',$msg,$email_list);
        return;
     }
     if (($epoch-$epochEPS) > 3600) {			#last activity is more than 1h ago
        $nerr++;
        ($status, $timestampEPS) = Pezca::GetString($timestampPV);
        if ($status) {
           &report_only($subscript,$status,$timestampPV);
           $timestampEPS = 'UNKNOWN';
        }
        if (length($timestampEPS) == 0) {$timestampEPS = 'UNKNOWN';}
        $msg .= 'Last detected activity of EPS monitoring script is: '.$timestampEPS."\n"
              . 'It is possible that EPS monitoring script is not running.'."\n";
        &part2($msg);
        &part3($msg);
        &zenity_error ($msg,1,'nowait');
        &email('EPS',$msg,$email_list);
        return;
     }

### Check if any EPS section is tripped or switched in the operator mode:
     @monitorSectionPVs = ();
     @monitorSectionFlags = ();
     push (@monitorSectionPVs,'23id:BLEPS:A_TRIP_EXISTS');        push (@monitorSectionFlags,1);
     push (@monitorSectionPVs,'23id:BLEPS:A_FAULT_EXISTS');       push (@monitorSectionFlags,1);
     push (@monitorSectionPVs,'23id:BLEPS:WARNING_EXISTS');       push (@monitorSectionFlags,1);

     for ($i=1; $i<=11; $i++) {
       push (@monitorSectionPVs,'23ID:BLEPS:GV'.$i.'_CLOSED_LS'); push (@monitorSectionFlags,1);
       push (@monitorSectionPVs,'23id:BLEPS:GV'.$i.'_FAULTED');   push (@monitorSectionFlags,1);
     }
     for ($i=1; $i<=12; $i++) {
       push (@monitorSectionPVs,'23ID:BLEPS:IP'.$i.'_STATUS');    push (@monitorSectionFlags,0);
     }
     for ($i=1; $i<=10; $i++) {
       push (@monitorSectionPVs,'23id:BLEPS:TEMP'.$i.'_TRIP');    push (@monitorSectionFlags,1);
     }
     for ($i=1; $i<=12; $i++) {
       push (@monitorSectionPVs,'23id:BLEPS:VS'.$i.'_TRIP');      push (@monitorSectionFlags,1);
     }
     for ($i=1; $i<=2; $i++) {
       push (@monitorSectionPVs,'23ID:BLEPS:Cryo'.$i.'_Ready');   push (@monitorSectionFlags,0);
     }
     for ($i=1; $i<=2; $i++) {
       push (@monitorSectionPVs,'23id:BLEPS:FLOW'.$i.'_TRIP');    push (@monitorSectionFlags,1);
     }

     $nflags = @monitorSectionPVs;
     for ($i=0; $i<$nflags; $i++) {
        &refresh_gui();
        $pv = $monitorSectionPVs[$i];
        $fl = $monitorSectionFlags[$i];
        ($status, $flag) = Pezca::GetDouble($pv);
        if ($status) {
           &report_only($subscript,$status,$pv);
           $flag = 0;
           $nerr++;
           $msg .= 'Error accessing "'.$pv.'".'."\n";
        } else {
           if ($flag eq $fl) {
              $nerr++;
              ($status,$flagstr) = Pezca::GetString($pv);
              if ($status) {$flagstr = $flag;}
              ($status,$desc) = Pezca::GetString($pv.'.DESC');
              if ($status) {$desc = '';}
              $msg .= 'BLEPS Fault/Trip: '.$pv.'='.$flagstr.' ['.$desc.']'."\n";
           }
        }
     }
  }				#end ($beamline == '23i'|'23o')
  if ($nerr > 0) {
     $step_err = 1;
     &part2($msg);
     &part3($msg);
     &zenity_error ($msg,1,'nowait');
     &email('EPS',$msg,$email_list);
  } else {
     print STDOUT '... EPS is OK'."\n";
  }
}

#=================================================================

sub check_DCM_monitor () {

  ### check if DCM monotor is running. If not, report to host & send emails.

  my ($subscript, $status, $msg);
  my ($timestamp, $epoch);
  my ($timestampDCM, $epochDCM, $pauseDCM);
  my ($timestampPV, $epochPV, $pausePV);

  $subscript = 'check_DCM_monitor';
  $step_err  = 0;
  $epoch     = time;						#seconds since (00:00:00 UTC, January 1, 1970)
  $timestamp = localtime($epoch);
  $msg       = "\n".'Current time: '.$timestamp."\n";

  if ($beamline ne '23d') {
     print STDOUT 'Checking DCM monitoring...'."\n";

     $epochPV     = $beamline.':dcmmon_epoch';
     $timestampPV = $beamline.':dcmmon_timestamp';
     $pausePV     = $beamline.':dcmmon_pause';

### Check if DCM monitoring IOC and script are running:
     ($status, $epochDCM) = Pezca::GetDouble($epochPV);		#seconds since (00:00:00 UTC, January 1, 1990)
     if ($status) {
        $step_err = 1;
        &report_only($subscript,$status,$epochPV);
        $epochDCM = 0;
        $msg .= 'Error accessing "'.$epochPV.'".'."\n"
              . 'It is possible that either DCM monitoring or Keithley IOC at '
              .$beamline.' is not running.'."\n";
        &part2($msg);
        &part3($msg);
        &zenity_error ($msg,1,'nowait');
        &email('DCM',$msg,$email_list);
        return;
     }
     $epochDCM += 631152000;					#add 20 years because epoch is counted from 1970 while epochDCM from 1990
     if (($epoch-$epochDCM) > 3600) {				#last activity is more than 1h ago
        $step_err = 1;
        $msg .= 'epoch(current)='.$epoch.'  epoch(DCM)='.$epochDCM."\n";
        ($status, $timestampDCM) = Pezca::GetString($timestampPV);
        if ($status) {
           &report_only($subscript,$status,$timestampPV);
           $msg .= 'Error='.$status.' reading PV='.$timestampPV."\n";
           $timestampDCM = 'UNKNOWN DUE TO PV ERROR';
        }
        if (length($timestampDCM) == 0) {$timestampDCM = 'ZERO LENGTH';}
        $msg .= 'Last detected activity of DCM monitoring script at '
              . $beamline.' is: '.$timestampDCM."\n"
              . 'PV = '.$timestampPV."\n"
              . 'It is possible that DCM monitoring script is not running.'."\n";
        &part2($msg);
        &part3($msg);
        &zenity_error ($msg,1,'nowait');
        &email('DCM',$msg,$email_list);
        return;
     }

     ($status, $pauseDCM) = Pezca::GetDouble($pausePV);
     if ($status) {
        $step_err = 1;
        &report_only($subscript,$status,$pausePV);
        $pauseDCM=0;
        $msg .= 'Error accessing "'.$pausePV.'".'."\n"
              . 'It is possible that either DCM monitoring or Keithley IOC at '
              . $beamline.' is not running.'."\n";
        &part2($msg);
        &part3($msg);
        &zenity_error ($msg,1,'nowait');
        &email('DCM',$msg,$email_list);
        return;
     }
     if ($pauseDCM != 0) {
        $msg .= 'Un-pausing DCM monitoring script.'."\n";
        &part2($msg);
        &part3($msg);
        $status = Pezca::PutDouble($pausePV,0);
        &report_only($subscript,$status,$pausePV);
     }
     print STDOUT "\t".'DCM is OK'."\n";
  }
}

#=================================================================

sub check_MBSH () {
  ### Check MBSH open
  $step_err = 0;
  $part2_message = '';
  $part3_message = '';
  $message = 'MBSH CHECK:';
  &part1(sprintf '%-40.40s',$message);
  print STDOUT $message."\n";
  $status1 = &getEndStationShutter($beamline);		#1=open,0=close,-1=not available
  if ($status1 == -1) {
     $pv_value = 0;
     $step_err = 1;
     $message = "\t".'Could not read MBSH status. CAGET error.'."\n";
     &part2($message);
     &part3($message);
  }
  elsif ($status1 == 0) {
     $pv_value = 1;
     $step_err = 1;
     $message = "\t".'MBSH closed!'."\n";
     ### TRy to correct:
     print STDOUT '      Attempting to open MBSH that is currently closed...'."\n";
     $status1 = &openEndStationShutter($beamline);
     if ($status1 == 0) {
        print STDOUT "\t".'Success!'."\n";
        $message .= "\t".'MBSH: attempt to open succeeded.'."\n";
        $step_err = 0;
        $pv_value = 0;
     } else {
        print STDOUT "\t".'!!! Failure!'."\n";
        $message .= "\t".'MBSH: attempt to open failed.'."\n";
     }
     &part2($message);
     &part3($message);
  }
  else {
     $pv_value = 0;
     $step_err = 0;
     $message = "\t".'OK: MBSH is open.'."\n";
     print STDOUT $message;
     &part1(sprintf '%-40.40s',$message);
  }
}

#=================================================================

sub check_MonoStabilizer () {
  ### Check if the monoStabilizer script is running and prompt to start it if not

  my ($i, $mostab_script, $mostab_name, $mostab_state);
  my ($mostab_adl, $mostab_par, $mostab_line_pv, $mostab_line);
  my ($confirmBox, $answer, $msg);

  $step_err = 0;
  $part2_message = '';
  $part3_message = '';
# $message = 'INTENSITY FEEDBACK:';
  $message = 'MonoStabilizer CHECK:';
  &part1(sprintf '%-40.40s',$message);
  print STDOUT $message."\n";

  $mostab_script = 'monoStabilizer.pl';
  $mostab_name   = 'monoStabilizer';
  $mostab_line_pv= $beamline.':mostab:statusline';
  $mostab_adl    = $PEZCA.'../gmcaApp/adl/feedback_'.$beamline.'.adl';
  $mostab_adl   =~ s/pezca\/\.\.\///;
  $mostab_par   = 'BL='.$beamline.':,SECTOR=23';
  if    ($beamline eq '23i') {$mostab_par .= ',ID=ID23ds:';}
  elsif ($beamline eq '23o') {$mostab_par .= ',ID=ID23us:';}
  &open_medm($mostab_adl,$mostab_par);
  
  if ($mostab_status[0]   !~ /^${beamline}/) {$mostab_status[0]   = $beamline.':'.$mostab_status[0];}
  if ($mostab_status_b[0] !~ /^${beamline}/) {$mostab_status_b[0] = $beamline.':'.$mostab_status_b[0];}
  if ($mostab_bpm2_ch[0]  !~ /^${beamline}/) {$mostab_bpm2_ch[0]  = $beamline.':'.$mostab_bpm2_ch[0];}
  if ($mostab_use_bpm2[0] !~ /^${beamline}/) {$mostab_use_bpm2[0] = $beamline.':'.$mostab_use_bpm2[0];}

  ### Check if mostab is running:
  print STDOUT '   Checking if '.$mostab_name.' status...'."\n";
  ($err,$mostab_state) = Pezca::GetDouble($mostab_status[0]);
  if ($err) {
     &report_only($script,$err,$mostab_status[0]);
     $step_err = 1;
     $message = "\t".'Could not probe '.$mostab_name.': CAGET error on '.$mostab_status[0]."\n";
     &part1('ERR'."\n");
     &part2($message);
     &part3($message);
     print STDOUT $part1_text;
     $part1_text = '';
     print STDOUT "\n";
     return;
  }  
  if ($mostab_state == 0) {
     $mw->focus(-force);
     $confirmBox = $mw->Dialog(-title=>$script,
                               -text=>$mostab_name.' is not running. Start?',
                               -wrap=>0,
                               -font=>$fontBtn,
                               -buttons=>['Yes', 'No'],
                               -default_button=>'Yes',
                               -bitmap=>'question',
                               -takefocus=>1,
                               -width=>50,
                               );
     $confirmBox->Subwidget('B_Yes')->configure(-background=>$bgcolor, -activebackground=>$abgcolor, 
                                               -foreground=>$fgcolor, -activeforeground=>$afgcolor,);
     $confirmBox->Subwidget('B_No')-> configure(-background=>$bgcolor, -activebackground=>$abgcolor, 
                                               -foreground=>$fgcolor, -activeforeground=>$afgcolor,);
     $answer = $confirmBox->Show;
     if ($answer eq 'Yes') {
        $msg = 'nohup /usr/bin/perl -W '.$PEZCA.$mostab_script.' >/dev/null 2>&1 &';
        print STDOUT 'Strarting:'."\n".$msg."\n\n";
        $err = system($msg);
        if ($err) {
           $step_err = 1;
           $message = "\t".'Could not start '.$mostab_script."\n";
           &part1('ERR'."\n");
           &part2($message);
           &part3($message);
           print STDOUT $part1_text;
           $part1_text = '';
           print STDOUT "\n";
           return;
        }
        for ($i=0; $i<800; $i++) {
           select(undef,undef,undef,0.01);
           &refresh_gui();
        }
     } else {
        $step_err = 1;
        $message = "\t".'Operator declined starting '.$mostab_script."\n";
        &part1('ERR'."\n");
        &part2($message);
        &part3($message);
        print STDOUT $part1_text;
        $part1_text = '';
        print STDOUT "\n";
        return;
     }
  }

  ### Probe mostab status
  print STDOUT '   Probing '.$mostab_name.' status...'."\n";
  $err = Pezca::PutDoubleOldCa($mostab_status[0],0);
  if ($err) {
     &report_only($script,$err,$mostab_status[0]);
     $step_err = 1;
     $message = "\t".'Could not probe '.$mostab_name.': CAPUT error on '.$mostab_status[0]."\n";
     &part1('ERR'."\n");
     &part2($message);
     &part3($message);
     print STDOUT $part1_text;
     $part1_text = '';
     print STDOUT "\n";
     return;
  }

# select(undef,undef,undef,5);
  for ($i=0; $i<600; $i++) {
     select(undef,undef,undef,0.01);
     &refresh_gui();
  }

  $err = &getPV(@mostab_status);   if ($err) {$step_err = 1;}		#this is checking if monoStabilizer is running. Do we need it?
  $err = &getPV(@mostab_status_b); 					#this  is checking if monoStabilizer is in "active feedback" state
  if ($err) {
     $step_err = 1;
     ($err,$mostab_line) = Pezca::GetString($mostab_line_pv);
     if ($err) {&report_only($script,$err,$mostab_line_pv); $mostab_line='Unknown / PV error';}
     $msg = '	(Current mostab state: '.$mostab_line.')'."\n";
#    $part2_message .= $msg;						#details
     $part3_message .= $msg;						#summary
  }
  $err = &getPV(@mostab_bpm2_ch);  if ($err) {$step_err = 1;}
  $err = &getPV(@mostab_use_bpm2); if ($err) {$step_err = 1;}

  if ($step_err) {
     &part1('ERR'."\n");
     &part2($part2_message."\n");
     &part3($part3_message."\n");
  }
  else {
     &part1('OK!'."\n");
     &part2("\t".$mostab_name.' is operating correctly.'."\n\n");
  }
  print STDOUT $part1_text;
  $part1_text = '';
  print STDOUT "\n";
}

#=================================================================

sub check_TransfocatorServer () {
  ### Check if the transfocatorServer script is running and prompt to start it if not
  my ($i, $ts_script, $ts_pv, $ts_name, $ts_state);
  my ($ts_adl, $ts_par, $ts_req, $msg);
  my ($confirmBox, $answer, $bl, $sd);

  $step_err = 0;
  $part2_message = '';
  $part3_message = '';
  $message = 'Transfocator Server CHECK:';
  &part1(sprintf '%-40.40s',$message);
  print STDOUT $message."\n";

  $ts_script = 'transfocatorServer.pl';
  $ts_name   = 'transfocatorServer';
  $ts_pv     = $beamline.':focusSts';
  $ts_adl    = $PEZCA.'../gmcaApp/adl/transfocatorServer.adl';
  $ts_adl    =~ s/pezca\/\.\.\///;
  $bl        = $beamline.':';
  $sd        = $PEZCA.'../stdApp/burt';
  $sd        =~ s/pezca\/\.\.\///;
  $ts_par    = 'BL='.$beamline.',P1='.$bl.'pmac20:acc65e1,P2='.$bl.'pmac20:acc65e2'
                                     .',S1=pmac20acc65e1,'     .'S2=pmac20acc65e2'
                                        .',SAVE_DIR='.$sd;
  &open_medm($ts_adl,$ts_par);
  
  ### Check if Transfocator Server  is running:
  print STDOUT '   Checking if '.$ts_name.' status...'."\n";
  ($err,$ts_state) = Pezca::GetDouble($ts_pv);
  if ($err) {
     &report_only($script,$err,$ts_pv);
     $step_err = 1;
     $message = "\t".'Could not probe '.$ts_name.': CAGET error on '.$ts_pv."\n";
     &part1('ERR'."\n");
     &part2($message);
     &part3($message);
     print STDOUT $part1_text;
     $part1_text = '';
     print STDOUT "\n";
     return;
  }  
  if ($ts_state == 0) {
     $mw->focus(-force);
     $confirmBox = $mw->Dialog(-title=>$script,
                               -text=>$ts_name.' is not running. Start?',
                               -wrap=>0,
                               -font=>$fontBtn,
                               -buttons=>['Yes', 'No'],
                               -default_button=>'Yes',
                               -bitmap=>'question',
                               -takefocus=>1,
                               -width=>50,
                               );
     $confirmBox->Subwidget('B_Yes')->configure(-background=>$bgcolor, -activebackground=>$abgcolor, 
                                               -foreground=>$fgcolor, -activeforeground=>$afgcolor,);
     $confirmBox->Subwidget('B_No')-> configure(-background=>$bgcolor, -activebackground=>$abgcolor, 
                                               -foreground=>$fgcolor, -activeforeground=>$afgcolor,);
     $answer = $confirmBox->Show;
     if ($answer eq 'Yes') {
        $msg = 'nohup /usr/bin/perl -W '.$PEZCA.$ts_script.' >/dev/null 2>&1 &';
        print STDOUT 'Strarting:'."\n".$msg."\n\n";
        $err = system($msg);
        if ($err) {
           $step_err = 1;
           $message = "\t".'Could not start '.$ts_script."\n";
           &part1('ERR'."\n");
           &part2($message);
           &part3($message);
           print STDOUT $part1_text;
           $part1_text = '';
           print STDOUT "\n";
           return;
        }
        for ($i=0; $i<800; $i++) {
           select(undef,undef,undef,0.01);
           &refresh_gui();
        }
     } else {
        $step_err = 1;
        $message = "\t".'Operator declined starting '.$ts_script."\n";
        &part1('ERR'."\n");
        &part2($message);
        &part3($message);
        print STDOUT $part1_text;
        $part1_text = '';
        print STDOUT "\n";
        return;
     }
  }

  $ts_pv = $beamline.':focusRqs';
### Probe TS requested pos.
  print STDOUT '   Probing '.$ts_name.' requested focus...'."\n";
  ($err,$ts_req) = Pezca::GetDouble($ts_pv);
  if ($err) {
     &report_only($script,$err,$ts_pv);
     $step_err = 1;
     $message = "\t".'Could not probe '.$ts_name.': CAGET error on '.$ts_pv."\n";
     &part1('ERR'."\n");
     &part2($message);
     &part3($message);
     print STDOUT $part1_text;
     $part1_text = '';
     print STDOUT "\n";
     return;
  }  
  if ($ts_req <= 0) {
     $step_err = 1;
     $message = "\t".$ts_name.' Requested focus is not set (val='.$ts_req.')!'."\n";
     &part1('ERR'."\n");
     &part2($part2_message."\n");
     &part3($part3_message."\n");
  }
  else {
     &part1('OK!'."\n");
     &part2("\t".$ts_name.' is operating correctly.'."\n\n");
  }
  print STDOUT $part1_text;
  $part1_text = '';
  print STDOUT "\n";
}

#=================================================================

sub check_keithley_times () {
  ### Check that Keithley amplifiers integration times does not exceed threshold
  my ($msg, $maxtime, $ktime, $klabel, $i);

  $step_err = 0;
  $part2_message = '';
  $part3_message = '';
  $message = 'KEITHLEY TIMES CHECK:';
  &part1(sprintf '%-40.40s',$message);
  print STDOUT $message."\n";
   
  $maxtime = $maxkeithleytime;
  $maxtime =~ s/ //g;
  if    ($maxtime =~ /us$/) {$maxtime =~ s/us//g; $maxtime = $maxtime/1000.;}	#convert to milliseconds
  elsif ($maxtime =~ /ms$/) {$maxtime =~ s/ms//g;}
  elsif ($maxtime =~ /s$/)  {$maxtime =~ s/s//g;  $maxtime = $maxtime*1000.;}	#convert to milliseconds
  else  {print STDOUT '*** '.$script.': incorrect maxtime='.$maxtime."\n"; return -1;}

  $msg = '';
  for ($i=1; $i<=$nkeithley; $i++) {
     $ktime = &keithley_read_time($i);
     if ($ktime > $maxtime) {
        $klabel = &keithley_read_label($i);
        $msg .= sprintf 'keithley-%d integration time='.$ktime.'ms exceeds max='.$maxtime.'ms -- label=[%s]'."\n", $i, $klabel;
     }
  }
  if (length($msg) > 0) {
     &zenity_error($msg,undef,'nowait');
     $step_err = 1;
     $message = "\t".'Some Keithley integration times are high:'."\n";
     $message .= $msg;
#    &part2($message);
     &part3($message);
  }
}

#=================================================================

sub caen_set_offsets () {
  my ($caen_offset_PV, $pv, @offsets_before, $offsets_after, $i);

  $caen_offset_PV = $beamline.':AH501:CurrentOffset';

  $step_err = 0;
  $part2_message = '';
  $part3_message = '';
  $message = 'CAEN OFFSETS RECALIBRATION:';
  &part1(sprintf '%-40.40s',$message);
  print STDOUT $message."\n";
  for ($i=0; $i<4; $i++) {
     $pv = sprintf '%s%d', $caen_offset_PV, ($i+1);
     ($status,$offsets_before[$i]) = Pezca::GetDouble($pv);
     if ($status) {&report_only($script, $status, $pv); $offsets_before[$i]=0;}
  } 

  @prg = (@scriptsPrefix, $PEZCA.'caen_set_offsets.pl', $beamline);
  $status1 = &system_with_fork(@prg);
  if ($pause_after) {
     &press_key_continue('Finished caen_set_offsets.pl, status='.$status1);
  }
  if ($status1) {
     $message = "\t".'caen_set_offsets.pl reports unclean exit.'."\n";
     &part1('ERR'."\n");
     &part2($message);
     &part3($message);
     $step_err = 1;
  } else {
     $message = '';
     for ($i=0; $i<4; $i++) {
        $pv = sprintf '%s%d', $caen_offset_PV, ($i+1);
        ($status,$offsets_after) = Pezca::GetDouble($pv);
        if ($status) {&report_only($script, $status, $pv); $offsets_after=0;}
        $message .= sprintf "\t".'CAEN offset-%d: %9.0f => %9.0f'."\n",
                              ($i+1), $offsets_before[$i], $offsets_after;
     }
     &part2($message);
     &part1('OK'."\n".$message);
  }
}




#=================================================================

sub center_WB_slitsV () {
  &center_WB_slits('vertical');
}

#=================================================================

sub center_WB_slitsH () {
  &center_WB_slits('horizontal');
}

#=================================================================

sub center_WB_slits ($) {
  my ($dir, $D);
  my ($WB_cntr_PV, $WB_cntr_mempos_PV, $WB_cntr_reference);
  my ($WB_cntr_lastfound, $steering);
  $dir = shift(@_);
  $dir = lc($dir);
  if    ($dir =~ /^vert/) {$D = 'V'; $WB_cntr_PV = 'WS:ZC';}
  elsif ($dir =~ /^hor/)  {$D = 'H'; $WB_cntr_PV = 'WS:YC';}
  else                    {$D = '?'; $WB_cntr_PV = 'unknown';}

  $WB_cntr_mempos_PV = $beamline.':'.$WB_cntr_PV.':MemPos';

  $step_err = 0;
  $part2_message = '';
  $part3_message = '';
  $message = 'WB-SLITS '.uc($dir).' CENTERING:';
  &part1(sprintf '%-40.40s',$message);
  print STDOUT $message."\n";
  @prg = (@scriptsPrefix, $PEZCA.'centerSlits_WB.pl', 'slit='.$D,
                                                      $beamline,
                                                      'touching=memory',
                                                      'showSteeringOnly',
                                                      'nostartasking',
                                                      'noShutter');
  $status1 = &system_with_fork(@prg);
  if ($pause_after) {
     &press_key_continue('Finished centerSlits_WB.pl slit='.$D.', status='.$status1);
  }
  if ($status1) {
     $message = "\t".'centerSlits_WB.pl slit='.$D.' reports unclean exit.'."\n";
     &part1('ERR'."\n");
     &part2($message);
     &part3($message);
     $step_err = 1;
  } else {
     ($status1,$WB_cntr_lastfound) = Pezca::GetString($WB_cntr_mempos_PV.'.DESC');
     if ($status1) {
        &report_only($script,$status1,$WB_cntr_mempos_PV.'.DESC');
        $message = "\t".'Failed to read '.$WB_cntr_mempos_PV.'.DESC.'."\n";
        &part1('ERR'."\n");
        &part2($message);
        &part3($message);
        $step_err = 1;
        $WB_cntr_lastfound = 0;
     }
     elsif ($WB_cntr_lastfound !~ /^LastFoundCenter=/) {
        $message = "\t".'No data in '.$WB_cntr_mempos_PV.'.DESC'."\n";
        &part1('ERR'."\n");
        &part2($message);
        &part3($message);
        $step_err = 1;
        $WB_cntr_lastfound = 0;
     }
     else  {
        $WB_cntr_lastfound =~ s/^LastFoundCenter=//;
        $WB_cntr_lastfound =~ s/mm//;
        $WB_cntr_lastfound =~ s/ //g;
        if (! &is_float($WB_cntr_lastfound)) {
           $message = "\t".'Non-number in '.$WB_cntr_mempos_PV.'.DESC : '.$WB_cntr_lastfound."\n";
           &part1('ERR'."\n");
           &part2($message);
           &part3($message);
           $step_err = 1;
           $WB_cntr_lastfound = 0;
        }
        ($status1,$WB_cntr_reference) = Pezca::GetDouble($WB_cntr_mempos_PV);
        if ($status1) {
           &report_only($script,$status1,$WB_cntr_mempos_PV);
           $message = "\t".'Failed to read '.$WB_cntr_mempos_PV."\n";
           &part1('ERR'."\n");
           &part2($message);
           &part3($message);
           $step_err = 1;
           $WB_cntr_reference = 0;
        }
        else {
           if ($wbsl_distance2source > 0) {
              $steering =-1000*($WB_cntr_lastfound-$WB_cntr_reference)
                        / $wbsl_distance2source;				#in urad
           } else {
              $steering = 9999;
           }
           $message = sprintf "\t".'WB-%s measured  center=%.4fmm.'."\n"
                             ."\t".'WB-%s reference center=%.4fmm.'."\n"
                             ."\t".'Required beam steering=%.1furad'."\n",
                              $dir, $WB_cntr_lastfound,
                              $dir, $WB_cntr_reference, $steering;
           if (abs($steering) >= 5) {&part2($message);}
           &part1('OK'."\n".$message);
        }
     }
  }
}

#=================================================================

sub center_WB2_slitsV () {
  &center_WB2_slits('vertical');
}

#=================================================================

sub center_WB2_slitsH () {
  &center_WB2_slits('horizontal');
}

#=================================================================

sub center_WB2_slits ($) {
  my ($dir, $D);
  my ($WB_cntr_PV, $WB_cntr_mempos_PV, $WB_cntr_reference);
  my ($WB_cntr_lastfound, $steering);
  $dir = shift(@_);
  $dir = lc($dir);
  if    ($dir =~ /^vert/) {$D = 'V'; $WB_cntr_PV = 'WB:V';}
  elsif ($dir =~ /^hor/)  {$D = 'H'; $WB_cntr_PV = 'WB:H';}
  else                    {$D = '?'; $WB_cntr_PV = 'unknown';}

  $WB_cntr_mempos_PV = $beamline.':'.$WB_cntr_PV.':MemPos';

  $step_err = 0;
  $part2_message = '';
  $part3_message = '';
  $message = 'WB-SLITS '.uc($dir).'-2020 CENTERING:';
  &part1(sprintf '%-40.40s',$message);
  print STDOUT $message."\n";
  @prg = (@scriptsPrefix, $PEZCA.'centerSlits_WB2.pl', 'slit='.$D,
                                                      $beamline,
                                                      'touching=memory',
                                                      'showSteeringOnly',
                                                      'nostartasking',
                                                      'gnuplot=image',
                                                      'directory='.$logdir,
#                                                     'searchMode=fullscan',
                                                      'noShutter');
  $status1 = &system_with_fork(@prg);
  if ($pause_after) {
     &press_key_continue('Finished centerSlits_WB2.pl slit='.$D.', status='.$status1);
  }
  if ($status1) {
     $message = "\t".'centerSlits_WB2.pl slit='.$D.' reports unclean exit.'."\n";
     &part1('ERR'."\n");
     &part2($message);
     &part3($message);
     $step_err = 1;
  } else {
     ($status1,$WB_cntr_lastfound) = Pezca::GetString($WB_cntr_mempos_PV.'.DESC');
     if ($status1) {
        &report_only($script,$status1,$WB_cntr_mempos_PV.'.DESC');
        $message = "\t".'Failed to read '.$WB_cntr_mempos_PV.'.DESC.'."\n";
        &part1('ERR'."\n");
        &part2($message);
        &part3($message);
        $step_err = 1;
        $WB_cntr_lastfound = 0;
     }
     elsif ($WB_cntr_lastfound !~ /^LastFoundCenter=/) {
        $message = "\t".'No data in '.$WB_cntr_mempos_PV.'.DESC'."\n";
        &part1('ERR'."\n");
        &part2($message);
        &part3($message);
        $step_err = 1;
        $WB_cntr_lastfound = 0;
     }
     else  {
        $WB_cntr_lastfound =~ s/^LastFoundCenter=//;
        $WB_cntr_lastfound =~ s/mm//;
        $WB_cntr_lastfound =~ s/ //g;
        if (! &is_float($WB_cntr_lastfound)) {
           $message = "\t".'Non-number in '.$WB_cntr_mempos_PV.'.DESC : '.$WB_cntr_lastfound."\n";
           &part1('ERR'."\n");
           &part2($message);
           &part3($message);
           $step_err = 1;
           $WB_cntr_lastfound = 0;
        }
        ($status1,$WB_cntr_reference) = Pezca::GetDouble($WB_cntr_mempos_PV);
        if ($status1) {
           &report_only($script,$status1,$WB_cntr_mempos_PV);
           $message = "\t".'Failed to read '.$WB_cntr_mempos_PV."\n";
           &part1('ERR'."\n");
           &part2($message);
           &part3($message);
           $step_err = 1;
           $WB_cntr_reference = 0;
        } else {
           if ($wbsl2_distance2source > 0) {
              $steering =-1000*($WB_cntr_lastfound-$WB_cntr_reference)
                        / $wbsl2_distance2source;				#in urad
           } else {
              $steering = 9999;
           }
           $message = sprintf "\t".'WB-%s measured  center=%.4fmm.'."\n"
                             ."\t".'WB-%s reference center=%.4fmm.'."\n"
                             ."\t".'Required beam steering=%.1furad'."\n",
                              $dir, $WB_cntr_lastfound,
                              $dir, $WB_cntr_reference, $steering;
           if (abs($steering) >= 5) {&part2($message);}
           &part1('OK'."\n".$message);
        }
     }
  }
}

#=================================================================

sub center_MB_slitsV () {
  &center_MB_slits('vertical');
}

#=================================================================

sub center_MB_slitsH () {
  &center_MB_slits('horizontal');
}

#=================================================================

sub center_MB_slits ($) {
  my ($dir, $D);
  my ($MB_cntr_PV, $MB_cntr_actpos_PV, $MB_cntr_mempos_PV);
  my ($MB_cntr_reference, $MB_cntr_lastfound);
  $dir = shift(@_);
  $dir = lc($dir);
  if    ($dir =~ /^vert/) {$D = 'V'; $MB_cntr_PV = 'CS:ZC';} 		#from user_setup_expectations.pl
  elsif ($dir =~ /^hor/)  {$D = 'H'; $MB_cntr_PV = 'CS:YC';} 		#from user_setup_expectations.pl
  else                    {$D = '?'; $MB_cntr_PV = 'unknown';}
  $MB_cntr_mempos_PV = $beamline.':'.$MB_cntr_PV.':MemPos';
  $MB_cntr_actpos_PV = $beamline.':'.$MB_cntr_PV.':ActPos';
  $MB_cntr_mempos_PV = $beamline.':'.$MB_cntr_PV.':MemPos';

  $step_err = 0;
  $part2_message = '';
  $part3_message = '';
  $message = 'CS-SLITS '.uc($dir).' CENTERING:';
  &part1(sprintf '%-40.40s',$message);
  print STDOUT $message."\n";
  @prg = (@scriptsPrefix, $PEZCA.'centerSlits_MB.pl', 'slit='.$D,
                                                      $beamline,
                                                      'touching=initial',
                                                      'nostartasking');
  $status1 = &system_with_fork(@prg);
  if ($pause_after) {
     &press_key_continue('Finished centerSlits_MB.pl slit='.$D.', status='.$status1);
  }
  if ($status1) {
     $message = "\t".'centerSlits_MB.pl slit='.$D.' reports unclean exit.'."\n";
     &part1('ERR'."\n");
     &part2($message);
     &part3($message);
     $step_err = 1;
  } else {
     ($status1,$MB_cntr_lastfound) = Pezca::GetDouble($MB_cntr_actpos_PV);
     if ($status1) {
        $MB_cntr_lastfound = 0;
        &report_only($script,$status1,$MB_cntr_actpos_PV);
        $message = "\t".'Failed to read '.$MB_cntr_actpos_PV."\n";
        &part1('ERR'."\n");
        &part2($message);
        &part3($message);
        $step_err = 1;
     }
     else  {
        ($status1,$MB_cntr_reference) = Pezca::GetDouble($MB_cntr_mempos_PV);
        if ($status1) {
           $MB_cntr_reference = 0;
           &report_only($script,$status1,$MB_cntr_mempos_PV);
           $message = "\t".'Failed to read '.$MB_cntr_mempos_PV."\n";
           &part1('ERR'."\n");
           &part2($message);
           &part3($message);
           $step_err = 1;
        } else {
           $message = sprintf "\t".'MB-%s measured  center=%.4fmm.'."\n"
                             ."\t".'MB-%s reference center=%.4fmm.'."\n",
                              $dir, $MB_cntr_lastfound,
                              $dir, $MB_cntr_reference;
           &part1('OK'."\n".$message);
        }
     }
  }
}

#=================================================================

sub center_GS_slitsV () {
  &center_GS_slits('vertical');
}

#=================================================================

sub center_GS_slitsH () {
  &center_GS_slits('horizontal');
}

#=================================================================

sub center_GS_slits ($) {
  my ($dir, $D, $i);
  my ($GS_cntr_PV, $GS_cntr_actpos_PV, $GS_cntr_mempos_PV);
  my ($GS_cntr_reference, $GS_cntr_lastfound);
  my (@jaw_offset_PV, @jaw_off_old, @jaw_off_new);

  $dir = shift(@_);
  $dir = lc($dir);
  if    ($dir =~ /^vert/) {	 						#from user_setup_expectations.pl
     $D = 'V';
     $GS_cntr_PV    = 'GS:ZC';
     @jaw_offset_PV = ('GS:ZT:Offset', 'GS:ZB:Offset');
  }
  elsif ($dir =~ /^hor/)  {					 		#from user_setup_expectations.pl
     $D = 'H';
     $GS_cntr_PV    = 'GS:YC';
     @jaw_offset_PV = ('GS:YO:Offset', 'GS:YI:Offset');
  }
  else {
     $D = '?';
     $GS_cntr_PV    = 'unknown';
     @jaw_offset_PV = ('unknown', 'unknown');
  }
  $GS_cntr_actpos_PV = $beamline.':'.$GS_cntr_PV.':ActPos';
  $GS_cntr_mempos_PV = $beamline.':'.$GS_cntr_PV.':MemPos';
  for ($i=0; $i<2; $i++) {
    $jaw_offset_PV[$i] = $beamline.':'.$jaw_offset_PV[$i];
  }

  $step_err = 0;
  $part2_message = '';
  $part3_message = '';
  $message = 'GUARD-SLITS '.uc($dir).' CENTERING:';
  &part1(sprintf '%-40.40s',$message);
  print STDOUT $message."\n";
  for ($i=0; $i<2; $i++) {
     ($status1,$jaw_off_old[$i]) = Pezca::GetDouble($jaw_offset_PV[$i]);
     if ($status1) {
        &report_only($script,$status1,$jaw_offset_PV[$i]);
        $message = "\t".'Failed to read '.$jaw_offset_PV[$i]."\n";
        &part1('ERR'."\n");
        &part2($message);
        &part3($message);
        $jaw_off_old[$i] = 1E10;
        $step_err = 1;
     }
  }
  @prg = (@scriptsPrefix, $PEZCA.'centerSlits_GS.pl', 'slit='.$D,
                                                      $beamline,
#                                                     'calibrateSize=ask',
                                                      'calibrateSize=warn',
                                                      'nostartasking');
  $status1 = &system_with_fork(@prg);
  if ($pause_after) {
     &press_key_continue('Finished centerSlits_GS.pl slit='.$D.', status='.$status1);
  }
  if ($status1) {
     $message = "\t".'centerSlits_GS.pl slit='.$D.' reports unclean exit.'."\n";
     &part1('ERR'."\n");
     &part2($message);
     &part3($message);
     $step_err = 1;
  } else {
     ### Record change in slit center position with respect to reference:
     ($status1,$GS_cntr_lastfound) = Pezca::GetDouble($GS_cntr_actpos_PV);
     if ($status1) {
        $GS_cntr_lastfound = 0;
        &report_only($script,$status1,$GS_cntr_actpos_PV);
        $message = "\t".'Failed to read '.$GS_cntr_actpos_PV."\n";
        &part1('ERR'."\n");
        &part2($message);
        &part3($message);
        $step_err = 1;
     }
     else  {
        ($status1,$GS_cntr_reference) = Pezca::GetDouble($GS_cntr_mempos_PV);
        if ($status1) {
           $GS_cntr_reference = 0;
           &report_only($script,$status1,$GS_cntr_mempos_PV);
           $message = "\t".'Failed to read '.$GS_cntr_mempos_PV."\n";
           &part1('ERR'."\n");
           &part2($message);
           &part3($message);
           $step_err = 1;
        } else {
           $message = sprintf "\t".'GS-%s measured  center=%.4fmm.'."\n"
                             ."\t".'GS-%s reference center=%.4fmm.'."\n",
                              $dir, $GS_cntr_lastfound,
                              $dir, $GS_cntr_reference;
           &part1('OK'."\n".$message);
        }
     }
  }
  for ($i=0; $i<2; $i++) {
     ($status1,$jaw_off_new[$i]) = Pezca::GetDouble($jaw_offset_PV[$i]);
     if ($status1) {
        &report_only($script,$status1,$jaw_offset_PV[$i]);
        $message = "\t".'Failed to read '.$jaw_offset_PV[$i]."\n";
        &part1('ERR'."\n");
        &part2($message);
        &part3($message);
        $jaw_off_new[$i] = 1E10;
        $step_err = 1;
     } else {
        if ($jaw_off_old[$i] ne 1E10 && abs($jaw_off_new[$i]-$jaw_off_old[$i]) > 0.0001) {		#difference exceeds 0.1um
           $message = sprintf "\t".'GS-%s jaw-%d new calibration offset=%.4fmm.'."\n"
                             ."\t".'GS-%s jaw-%d old calibration offset=%.4fmm.'."\n",
                              $dir, ($i+1), $jaw_off_new[$i],
                              $dir, ($i+1), $jaw_off_old[$i];
           &part1('OK'."\n".$message);
        }
     }
  }
}

#=================================================================

sub center_HFM () {
  &center_KBM('horizontal');
}

#=================================================================

sub center_VFM () {
  &center_KBM('vertical');
}

#=================================================================

sub center_KBM ($) {
  my ($dir, $D);
  my ($mirror_cntr_PV, $mirror_cntr_actpos_PV, $mirror_cntr_mempos_PV);
  my ($mirror_cntr_reference, $mirror_cntr_lastfound);
  $dir = shift(@_);
  $dir = lc($dir);
  if    ($dir =~ /^vert/) {$D = 'VFM'; $mirror_cntr_PV = 'KB:VZ';}	 		#from user_setup_expectations.pl
  elsif ($dir =~ /^hor/)  {$D = 'HFM'; $mirror_cntr_PV = 'KB:HY';}	 		#from user_setup_expectations.pl
  else                    {$D = '???'; $mirror_cntr_PV = 'unknown';}

  $mirror_cntr_actpos_PV = $beamline.':'.$mirror_cntr_PV.':ActPos';
  $mirror_cntr_mempos_PV = $beamline.':'.$mirror_cntr_PV.':MemPos';

  $step_err = 0;
  $part2_message = '';
  $part3_message = '';
  $message = 'KBM-'.$D.' CENTERING:';
  &part1(sprintf '%-40.40s',$message);
  print STDOUT $message."\n";
  @prg = (@scriptsPrefix, $PEZCA.'centerMirror_v3.pl', $D, 'beamline='.$beamline);
  $status1 = &system_with_fork(@prg);
  if ($pause_after) {
     &press_key_continue('Finished centerMirror_v3.pl '.$D.', status='.$status1);
  }
  if ($status1) {
     $message = "\t".'centerMirror_v3.pl '.$D.' reports unclean exit.'."\n";
     &part1('ERR'."\n");
     &part2($message);
     &part3($message);
     $step_err = 1;
  } else {
     ($status1,$mirror_cntr_lastfound) = Pezca::GetDouble($mirror_cntr_actpos_PV);
     if ($status1) {
        $mirror_cntr_lastfound = 0;
        &report_only($script,$status1,$mirror_cntr_actpos_PV);
        $message = "\t".'Failed to read '.$mirror_cntr_actpos_PV."\n";
        &part1('ERR'."\n");
        &part2($message);
        &part3($message);
        $step_err = 1;
     }
     else  {
        ($status1,$mirror_cntr_reference) = Pezca::GetDouble($mirror_cntr_mempos_PV);
        if ($status1) {
           $mirror_cntr_reference = 0;
           &report_only($script,$status1,$mirror_cntr_mempos_PV);
           $message = "\t".'Failed to read '.$mirror_cntr_mempos_PV."\n";
           &part1('ERR'."\n");
           &part2($message);
           &part3($message);
           $step_err = 1;
        } else {
           $message = sprintf "\t".'KBM-%s measured  center=%.4fmm.'."\n"
                             ."\t".'KBM-%s reference center=%.4fmm.'."\n",
                              $D, $mirror_cntr_lastfound,
                              $D, $mirror_cntr_reference;
           &part1('OK'."\n".$message);
        }
     }
  }
}

#=================================================================

sub calibrate_mono () {
  my ($assyEn_PV, $axisQB_PV, $offsetQB_old, $offsetQB_new);
  my ($QB_new, $QB_old, $dQB, $ESinTheta, $E_new, $E_old, $dE);

  $assyEn_PV = $beamline.':MO:En';
  $axisQB_PV = $beamline.':MO:QB';
  $step_err = 0;
  $part2_message = '';
  $part3_message = '';
  $message = 'MONO CALIBRATION:';
  &part1(sprintf '%-40.40s',$message);
  print STDOUT $message."\n";
  ($status, $offsetQB_old) = Pezca::GetDouble($axisQB_PV.':Offset');
  if ($status) {
     &report_only($script,$status,$axisQB_PV.':Offset');
     $message = "\t".'Cannot read Bragg angle offset: '.$axisQB_PV.':Offset'."\n";
     &part2($message);
     &part3($message);
     $step_err = 1;
     $offsetQB_old = 0;
  }

  @prg = (@scriptsPrefix, $PEZCA.'calibrateMono.pl',
                                                'Au-L3',
                                                'beamline='.$beamline,
#                                               'i1=9',
                                                'useCAEN',
#                                               'recoveryTime=120',     #not needed with CAEN
                                                'fluorescence',
                                                'speedFactor=0.05',
                                                'idOffset=mySQL',
                                                'cutEnds=20.0',
                                                'nomedm',
                                                'pauseFeedback',
                                                'foilIn=bpm',
                                                'gnuplot=image',
                                                'nostartasking',
                                                'GUImode');
  $status1 = &system_with_fork(@prg);
  if ($pause_after) {&press_key_continue('Finished calibrateMono.pl, status='.$status1);}
  if ($status1) {
     $message = "\t".'calibrateMono.pl reports unclean exit.'."\n";
     &part2($message);
     &part3($message);
     $step_err = 1;
  } else {
     ($status, $offsetQB_new) = Pezca::GetDouble($axisQB_PV.':Offset');
     if ($status) {
        &report_only($script,$status,$axisQB_PV.':Offset');
        $message = "\t".'Cannot read Bragg angle offset: '.$axisQB_PV.':Offset'."\n";
        &part2($message);
        &part3($message);
        $step_err = 1;
        $offsetQB_new = 0;
     }
     if ($step_err == 0) {
        $dQB = $offsetQB_new - $offsetQB_old;
        $ESinTheta = &getESinTheta($assyEn_PV,\$status);
        if ($status) {$step_err = 1; return;}
        $QB_new = &getAxisAct ($axisQB_PV,\$status);
        if ($status) {$step_err = 1; return;}
        $QB_old = $QB_new - $dQB;
        $E_new = &theta2E($assyEn_PV,$QB_new,$ESinTheta);
        $E_old = &theta2E($assyEn_PV,$QB_old,$ESinTheta);
        $dE = 1000*($E_new - $E_old);					#in eV
        $message = sprintf "\t".'Mono energy was corrected by dE=%.2f eV'."\n"
              ."\t".'Mono Bragg angle was corrected by dTheta=%.5f degr.'."\n"
              ."\t".'Mono offsets:  old=%.5f,  new=%.5f'."\n"
              ,$dE, $dQB, $offsetQB_old,  $offsetQB_new;
        if (abs($dE) > 0.1) {&part2($message);}
        &part1('OK'."\n".$message);
     }
  }
}

#=================================================================

sub align_minibeam () {
  ### Optimize beam:
  $step_err = 0;
  $part2_message = '';
  $part3_message = '';
  $message = 'MINIBEAM ALIGNMENT:';
  &part1(sprintf '%-40.40s',$message);
  print STDOUT $message."\n";
  print STDOUT "\t".'Optimizing beam to 10um pinhole'."\n";
  @prg = (@scriptsPrefix, $PEZCA.'alignMiniBeam.pl', $beamline, 'save=./');
  push (@prg,'signal=Pin');
  push (@prg,'fastScan');
  push (@prg,'align300to10');						#before 2014/09/30
  push (@prg,'autoPinholeOrGuardSlits');				#after 2014/09/30, JBluIce style.
  push (@prg,'moveSample=-15h');					#move YAG out of the beam, if mounted
  push (@prg,'gnuplot=image');						#save gnuplots as images
  push (@prg,'skipMostab');
  push (@prg,'skipPostab');
  push (@prg,'displaytime=2');
  push (@prg,'medm');
# if ($beamline eq '23i') {
#    push (@prg,'useKBM');						#overwrite the TF default at IDD
#    push (@prg,'HalfRangeV=1.0');					#overwrite the IDD default
#    push (@prg,'HalfRangeH=2.0');					#overwrite the IDD default
# }
  $status1 = &system_with_fork(@prg);
  if ($pause_after) {&press_key_continue('Finished alignMiniBeam.pl, status='.$status1);}

  if ($status1) {
     $message = "\t".'alignMiniBeam.pl reports unclean exit.'."\n";
     &part2($message);
     &part3($message);
     $step_err = 1;
  } else {
     &part1('OK'."\n");
  }
}

#=================================================================

sub minibeam_intensity () {

  our ($EXTRA_MOTION_TIME, $nonstop);
  my ($struck_npts, $rmsVal_5time, $rmsDev_5time, $yag_shift);
  my (@cols, $col, @mtrs, $mtr, @lbls, %pos, $pv, $i, $j, $txt);
  my ($dialBox, $answer, $nonstop_saved);

  $step_err = 0;
  $minibeam_messages = '';
  $part2_message = '';
  $part3_message = '';
  $message = 'MINIBEAM INTENSITY:';
  print STDOUT $message."\n";
  &part1(sprintf '%-40.40s',$message);

  $EXTRA_MOTION_TIME = 10;					# default is 5s
  if (!defined $nonstop) {$nonstop = 0;}
  $nonstop_saved = $nonstop;

  print STDOUT '   Check for pre-requisite conditions...'."\n";

  ### Check E = 12 keV
  print STDOUT '      E = 12 keV'."\n";
  $energy = &getAxisAct ($beamline.':MO:E',\$ca_error);
  if ($ca_error) {
     $energy = 0;
     $step_err = 1;
     $message = "\t".'Could not read energy. CAGET error.'."\n";
     &part2($message);
     &part3($message);
  }
  elsif (abs($energy-12) > 0.01) {
     $step_err = 1;
     $message = sprintf "\t".'Non-standard energy=%.2f KeV (expect 12.00 KeV).'."\n",$energy;
     &part2($message);
     &part3($message);
  }
  else {
     $message = "\t".'Energy is OK.'."\n";
     &part1($message);
  }

  ### Currently, this sub-routine assumes that the mirrors and
  ### downstream components are aligned.  There aren't checks
  ### to verify this, however.

  ### Put attenuators where needed (leave w/ attenuation after)
  ### Calculate percentage from factor
  print STDOUT '      Setting attenuation to '.$att_factor.'x'."\n";
  if ($ATTENUATION_USE_FACTOR == 1) {				#this must be set in expectations!
     ($status,$att_actual) = &setAttenuation1($att_factor);
  } else {
     print STDOUT '!!! Please set "$ATTENUATION_USE_FACTOR=1"!'."\n";
     $att_percentage = (1 - (1/$att_factor)) * 100;
     ($status,$att_actual) = &setAttenuation1($att_percentage);
  }
  if ($status) {
     $step_err = 1;
     $message = "\t".'Could not set attenuation.'."\n";
     &part2($message);
     &part3($message);
  } else {
     if ($ATTENUATION_USE_FACTOR == 1) {				#this must be set in expectations!
        $att_factor_actual = $att_actual;
     } else {
        $att_factor_actual = 1. / (1. - $att_actual/100);
     }
     $message = sprintf "\t".'Attenuation set at %.1fx'."\n",$att_factor_actual;
     &part1($message);
  }

  $nonstop = 1;

  ### Set GS size
  printf STDOUT '      Setting guard slit size to %.2fmm x %.2fmm'."\n", $hist_gs_v,$hist_gs_h;
  $status1 = &runMotor ($beamline.':GS:Av', $beamline.':GS:ZS', $hist_gs_v, 'checkAuto');
  if ($status1) {
     $message = "\t".'Could not move GS vert size.'."\n";
     &part2($message);
     &part3($message);
     $step_err = 1;
  }
  &refresh_gui();

  $status1 = &runMotor ($beamline.':GS:Ah', $beamline.':GS:YS', $hist_gs_h, 'checkAuto');
  if ($status1) {
     $message = "\t".'Could not move GS horz size.'."\n";
     &part2($message);
     &part3($message);
     $step_err = 1;
  }
  &refresh_gui();

  $gs_vs = &getAxisAct ($beamline.':GS:ZS',\$ca_error);
  $gs_hs = &getAxisAct ($beamline.':GS:YS',\$ca_error);
  if (((abs($gs_vs - $hist_gs_v)) > 0.005) || ((abs($gs_hs - $hist_gs_h)) > 0.005)) {
     $step_err = 1;
     $message = "\t".'GS not properly sized.'."\n";
     &part2($message);
     &part3($message);
  }

  ### Remove YAG w/ Gon Horz
  $yag_shift = 15;
  print STDOUT '      Removing potential YAG interference (moving YAG inboard by '.$yag_shift.'mm)'."\n";
  $gon_h_in = &getAxisAct ($beamline.':'.$gh_axis,\$ca_error);
  if ($ca_error) {
     $step_err = 1;
     $message = "\t".'Could not read YAG position.'."\n";
     &part2($message);
     &part3($message);
  }
  else {
     $gon_h_out = $gon_h_in - $yag_shift;
     $status1 = &runMotor ($beamline.':'.$gh_assy, $beamline.':'.$gh_axis, $gon_h_out, 'checkAuto');
     if ($status1) {
        $message = "\t".'Could not remove YAG for intensity measurements.'."\n";
        &part2($message);
        &part3($message);
        $step_err = 1;
     }
     &refresh_gui();
  }

  ### Set ADC rate
  print STDOUT '      Setting ADC rate'."\n";
  $ca_error = Pezca::PutDouble($beamline.':ADC_'.$adc_i0_input.'.SCAN', $adc_scanrate);
  &report_only($script,$ca_error,$beamline.':ADC_'.$adc_i0_input.'.SCAN');
  $ca_error += Pezca::PutDouble($beamline.':ADC_'.$adc_pin_input.'.SCAN', $adc_scanrate);
  &report_only($script,$ca_error,$beamline.':ADC_'.$adc_pin_input.'.SCAN');
  if ($ca_error) {
     $step_err = 1;
     $message = "\t".'ADC rate not properly set.'."\n";
     &part2($message);		# This was commented at some point.  Why?
     &part3($message);
  }

  ### Record ring current
  print STDOUT '   Recording ring current'."\n";
  ($ca_error, $aps_current) = Pezca::GetDouble('S:SRcurrentAI.VAL');
  if ($ca_error) {
     &report_only($script,$ca_error,'S:SRcurrentAI.VAL');
     $aps_current = 0;
     $message = "\t".'Could not read ring current.'."\n";
     &part2($message);
     &part3($message);
     $step_err = 1;
  }

  ### Record DCM Vsum
  print STDOUT '   Recording DCM Vsum'."\n";
  ($ca_error, $dcm_vsum) = Pezca::GetDouble($beamline.':ADC_1.VAL');
  if ($ca_error) {
     &report_only($script,$ca_error,$beamline.':ADC_1.VAL');
     $dcm_vsum = 0;
     $message = "\t".'Could not read DCM Vsum.'."\n";
     &part2($message);
     &part3($message);
     $step_err = 1;
  }

  ### Record beamstop position
  print STDOUT '   Recording beamstop position'."\n";
  ($ca_error, $bmstp_pos) = Pezca::GetDouble($beamline.':BS:D:ActPos');
  if ($ca_error) {
     &report_only($script,$ca_error,$beamline.':BS:D:ActPos');
     $bmstp_pos = 0;
     $message = "\t".'Could not read beamstop position.'."\n";
     &part2($message);
     &part3($message);
     $step_err = 1;
  }

  ### Set initial beamstop/collimator/pin diode position
  print STDOUT '   Setting initial beamstop/collimator/pin diode positions'."\n";
  @prg = ('IN', $beamline, 'backlightOUT', 'pindiodeOUT');
  $status1 = &setSampleEnv(\@prg);
  if ($status1) {
     $message = "\t".'Could not put collimator in.'."\n";
     &part2($message);
     &part3($message);
     $step_err = 1;
  }

  ### Record I0
  print STDOUT '   Recording I0'."\n";
  $bmstp_pos = &getAxisAct ($beamline.':BS:D',\$ca_error);

  ### adjust I0 gain
  printf STDOUT '      Adjusting Keithley gain on I0 to %.0e'."\n",$kl_i0_gain_hist;
  $status = &keithley_set_gain($kl_i0_channel,$kl_i0_gain_hist);
  if ($status) {
     $step_err = 1;
     $message = "\t".'Unable to set I0 gain.';
     &part2($message);
     &part3($message);
  }

  ### Prepare scaler:
  if (defined $cnt_time) {&setJoerger ($scaler, $cnt_time);}

  #print STDOUT '      Opening shutter'."\n";
  $status = &openShutter ($shutter,$shutter_readback);
  if ($status) {
     $message = "\t".'Unable to open shutter.';
     &closeShutter ($shutter,$shutter_readback);
     &part2($message);
     &part3($message);
  }
  #print STDOUT '      Measuring'."\n";
  if (defined $cnt_time) {
     ($i0_int_avg) = &runJoergerRef ($scaler, $cnt_time, [$struck_i0_input]);
     $i0_int_avg /= (1E5);                                     #convert from counts to current
     &refresh_gui();
  } else {
     $int_avg = 0;
     for ($i=0; $i<10; $i++) {
        ($ca_error, $i0_int) = Pezca::GetDouble($beamline.':ADC_'.$adc_i0_input.'.VAL');
        if ($ca_error) {
           &report_only($script,$ca_error,$beamline.':ADC_'.$adc_i0_input.'.VAL');
           $i0_int = 0;
           $message = "\t".'Unable to read I0.';
           &part2($message);
           &part3($message);
        }
        $i0_int_avg += $i0_int;
        select(undef,undef,undef,0.2);
     }
     $i0_int_avg /= 10;
     &refresh_gui();
  }
  &closeShutter ($shutter,$shutter_readback);

  ### Record minibeam intensities
  print STDOUT '   Recording scatter guard intensity'."\n";
  ($status,$message_300,$intensity_300) = &measure_pinhole_cts(@minibeam_300);
  if ($status == 1) {
     $step_err = 1;
     &part2("\t".'300 um: '.$message."\n");
     &part3("\t".'300 um: '.$message."\n");
  }
  ###-----------------------------------------------------------------------Moved up 2013/02/05
  ### Set GS size for minibeam:
  printf STDOUT '      Setting guard slit size to %.2fmm x %.2fmm'."\n", $hist_gs_v_mini, $hist_gs_h_mini;
  $status1 = &runMotor($beamline.':GS:Av', $beamline.':GS:ZS', $hist_gs_v_mini);
  if ($status1) {
     $message = "\t".'Could not move GS vert size.'."\n";
     &part2($message);
     &part3($message);
     $step_err = 1;
  }
  &refresh_gui();

  $status1 = &runMotor($beamline.':GS:Ah', $beamline.':GS:YS', $hist_gs_h_mini);
  if ($status1) {
     $message = "\t".'Could not move GS horz size.'."\n";
     &part2($message);
     &part3($message);
     $step_err = 1;
  }
  &refresh_gui();

  ($ca_error, $gs_vs) = Pezca::GetDouble($beamline.':GS:ZS:ActPos');
  if ($ca_error) {&report_only($script,$ca_error,$beamline.':GS:ZS:ActPos'); $gs_vs=0;}
  ($ca_error, $gs_hs) = Pezca::GetDouble($beamline.':GS:YS:ActPos');
  if ($ca_error) {&report_only($script,$ca_error,$beamline.':GS:YS:ActPos'); $gs_hs=0;}
  if ((abs($gs_vs - $hist_gs_v_mini) > 0.005) || (abs($gs_hs - $hist_gs_h_mini) > 0.005)) {
     $step_err = 1;
     $message = "\t".'GS not properly sized.'."\n";
     &part2($message);
     &part3($message);
  }
  ###-----------------------------------------------------------------------End of moved up 2013/02/05

  print STDOUT '   Recording 20 um intensity'."\n";
  ($status,$message_20, $intensity_20)  = &measure_pinhole_cts(@minibeam_20);
  if ($status == 1) {
     $step_err = 1;
     &part2("\t".'20 um:  '.$message."\n");
     &part3("\t".'20 um:  '.$message."\n");
  }
  print STDOUT '   Recording 10 um intensity'."\n";
  ($status,$message_10, $intensity_10)  = &measure_pinhole_cts(@minibeam_10);
  if ($status == 1) {
     $step_err = 1;
     &part2("\t".'10 um:  '.$message."\n");
     &part3("\t".'10 um:  '.$message."\n");
  }
  print STDOUT '   Recording 5 um intensity'."\n";
  ($status,$message_5,  $intensity_5)   = &measure_pinhole_cts(@minibeam_5);
  if ($status == 1) {
     $step_err = 1;
     &part2("\t".'5 um:   '.$message."\n");
     &part3("\t".'5 um:   '.$message."\n");
  }

  ### Grab minibeam presets:
  @cols = qw(300 20 10 5);
  @mtrs = qw(V H P Y);
  @lbls = qw(Vert Horz Pitch Yaw);
  for ($j=0; $j<@mtrs; $j++) {$lbls[$j] .= ' (mm)';}
  print STDOUT '   Recording collimator preset positions'."\n";
  for ($j=0; $j<@mtrs; $j++)    {$mtr = $mtrs[$j];
     for ($i=0; $i<@cols; $i++) {$col = $cols[$i];
        $pv  = $beamline.':COL:'.$mtr.':RqsPosMem'.($i+2);                      #300=Mem2 20=Mem3 10=Mem4 5=Mem5
        ($ca_error, $pos{$col.'-'.$mtr}) = Pezca::GetDouble($pv);
        if ($ca_error) {&report_only($script,$ca_error,$pv);  $pos{$col.'-'.$mtr} = 0;}
     }
  }

##############################################################################
  ($struck_npts, $rmsVal_5time, $rmsDev_5time) = &record_I0_temp_data_5um(); #
##############################################################################

  ### Return pin diode, collimator, beamstop to desired location
  ### change preset
  print STDOUT '   Returning collimator to scatter guard position'."\n";
  $ca_error = Pezca::PutDouble($beamline.':COL:St:ActivePreset', $minibeam_300[0]);
  &report_only($script,$ca_error,$beamline.':COL:St:ActivePreset');
  if ($ca_error) {
     $step_err = 1;
     $message = "\t".'Could not select desired minibeam.'."\n";
     &part2($message);
     &part3($message);
  }

  print STDOUT '   Removing pin diode, returning beamstop'."\n";
  @prg = ('IN', $beamline, 'backlightOUT', 'pindiodeOUT');
  $status1 = &setSampleEnv(\@prg);
  if ($status1) {
     $message = "\t".'Could not put collimator in.'."\n";
     &part2($message);
     &part3($message);
     $step_err = 1;
  }
  ### Return YAG to initial position
  # print STDOUT '   Restoring gonio horz'."\n";
  # $status1 = &runMotor($beamline.':'.$gh_assy, $beamline.':'.$gh_axis, $gon_h_in);
  # if ($status1) {
  #    $message = "\t".'Could not return gonio.'."\n";
  #    &part2($message);
  #    &part3($message);
  #    $step_err = 1;
  # }
  # &refresh_gui();

  $minibeam_summary .= "\n\n";
  $minibeam_summary .= sprintf "\t".'%-20.20s %10.10s %10.10s'."\n", ' ',               'Measured',         'Expected';
  $minibeam_summary .= "\t----------------------------------------\n";
  $minibeam_summary .= sprintf "\t".'%-20.20s %10.1f %10.1f'  ."\n", 'APS current (mA)', $aps_current,       $hist_aps_current;
  $minibeam_summary .= sprintf "\t".'%-20.20s %10.3f %10.3f'  ."\n", 'Energy (keV)',     $energy,            $hist_energy;
  $minibeam_summary .= sprintf "\t".'%-20.20s %10.2f %10.2f'  ."\n", 'DCM Vsum (V)',     $dcm_vsum,          $hist_dcm_vsum;
  $minibeam_summary .= sprintf "\t".'%-20.20s %10.1f %10.1f'  ."\n", 'Attenuation (x)',  $att_factor_actual, $hist_att;
  $minibeam_summary .= sprintf "\t".'%-20.20s %10.3f %10.3f'  ."\n", 'GS Vsize (mm)',    $gs_vs,             $hist_gs_v;
  $minibeam_summary .= sprintf "\t".'%-20.20s %10.3f %10.3f'  ."\n", 'GS Hsize (mm)',    $gs_hs,             $hist_gs_h;
  $minibeam_summary .= sprintf "\t".'%-20.20s %10.2f %10.2f'  ."\n", 'I0 intensity (V)', $i0_int_avg,        $hist_i0;
  $minibeam_summary .= sprintf "\t".'%-20.20s %10.0E %10.0E'  ."\n", 'I0 gain',          $kl_i0_gain_hist,   $kl_i0_gain_hist;
  $minibeam_summary .= sprintf "\t".'%-20.20s %10.3f %10.3f'  ."\n", 'Beamstop (mm)',    $bmstp_pos,         $hist_bmstp;
  $minibeam_summary .= "\n\n";

  $minibeam_summary .= sprintf "\t".'%-15.15s %9.9s %9.9s %9.9s %9.9s'."\n", ' ','300 um','20 um','10 um','5 um';
  $minibeam_summary .= "\t".'---------------------------------------------------'."\n";
  $minibeam_summary .= sprintf "\t".'%-15.15s %9.2f %9.2f %9.2f %9.2f'."\n", 'Intensity (V)',$intensity_300,$intensity_20,$intensity_10,$intensity_5;
  $minibeam_summary .= sprintf "\t".'%-15.15s %9.2f %9.2f %9.2f %9.2f'."\n", 'Expected  (V)',$minibeam_300[2],$minibeam_20[2],$minibeam_10[2],$minibeam_5[2];
  $minibeam_summary .= "\n";
  $minibeam_summary .= sprintf "\t".'%-15.15s %9.0E %9.0E %9.0E %9.0E'."\n", 'Keithley Gain',$minibeam_300[1],$minibeam_20[1],$minibeam_10[1],$minibeam_5[1];
  for ($j=0; $j<@mtrs; $j++)    {$mtr = $mtrs[$j]; $minibeam_summary .= sprintf "\t".'%-15.15s', $lbls[$j];
     for ($i=0; $i<@cols; $i++) {$col = $cols[$i]; $minibeam_summary .= sprintf ' %9.3f', $pos{$col.'-'.$mtr};}
     $minibeam_summary .= "\n";
  }
  $minibeam_summary .= "\n";
  $minibeam_summary .= "\t".'Expected values were measured on: '.$hist_date."\n";

  $minibeam_summary .= sprintf "\t".'%s %d times with dwell=%.2fms :'."\n", 'Time dependence of 5um-pinhole Intensity sampled', $struck_npts, 1000*$struck_dwell;
  $minibeam_summary .= sprintf "\t".'  - average count rate (rms) = %.2f cts'."\n", $rmsVal_5time;
  if ($rmsVal_5time == 0) {$rmsVal_5time = 1;}
  $minibeam_summary .= sprintf "\t".'  - average deviations (rms) = %.2f cts [%.2f%s]'."\n", $rmsDev_5time, 100*$rmsDev_5time/$rmsVal_5time, '%';

  &part2($minibeam_summary);

  if ($step_err) {
     $txt = 'ATTENTION!'."\n"
          . 'The measurement of minibeam intensities resulted in one or more errors.'."\n"
          . 'Please review the data and choose whether to proceed...'."\n"
          . '-------------------------------------------------------'."\n"
          . $minibeam_messages."\n\n"
          . 'Do you wish to continue or quit to correct errors?';
     $mw->focus(-force);
     $dialBox = $mw->Dialog(-title=>$script,
                           -text=>$txt,
                           -wrap=>0,
                           -font=>$fontBtn,
                           -buttons=>['Continue', 'Quit'],
                           -default_button=>'Continue',
                           -bitmap=>'question',
                           -takefocus=>1
                           );
     $dialBox->Subwidget('B_Continue')->configure(-background=>$bgcolor, -activebackground=>$abgcolor, 
                                                  -foreground=>$fgcolor, -activeforeground=>$afgcolor,);
     $dialBox->Subwidget('B_Quit')->    configure(-background=>$bgcolor, -activebackground=>$abgcolor, 
                                                  -foreground=>$fgcolor, -activeforeground=>$afgcolor,);
     $answer = $dialBox->Show();
     if ($answer eq 'Quit') {&clearAllCaMonitors(); exit;}
  }

  if ($step_err) {&part1('ERR'."\n")}
  else           {&part1('OK!'."\n")}
  &part2("\n\n");
  print STDOUT $part1_text;
  $part1_text = '';
  print STDOUT "\n";
  $nonstop = $nonstop_saved;
}

#=================================================================

# sub measure_pinhole_cts ($pinhole,$gain,$expected_intensity,$warn_level,$date) {
sub measure_pinhole_cts ($$$$$) {

  my $pinhole      = shift (@_);
  my $gain         = shift (@_);
  my $expected_int = shift (@_);
  my $warn_level   = shift (@_);
# my $date         = shift (@_);

  ### change preset
  $ca_error = Pezca::PutDouble($beamline.':COL:St:ActivePreset', $pinhole);
  if ($ca_error) {
     &report_only($script,$ca_error,$beamline.':COL:St:ActivePreset');
     $message = 'Unable to select collimator. CAPUT error.';
     return (1,$message,0);
  }

  @prg = ('IN', $beamline, 'beamstopOUT', 'backlightOUT', 'pindiodeIN');
  $status1 = &setSampleEnv(\@prg);
  if ($status1) {
     $message = "\t".'!!! Could not put pindiode in (ignoring for now).'."\n";
  }

  ### adjust gain
  printf STDOUT '      Adjusting Keithley gain on pindiode to %.0e'."\n",$gain;
  $status = &keithley_set_gain($kl_pin_channel,$gain);
  if ($status) {
     $message = 'Unable to set gain.';
     return (1,$message,0);
  }

  ### measure
  $status = &openShutter ($shutter,$shutter_readback);
  if ($status) {
     $message = 'Unable to open shutter.';
     $ca_error = Pezca::PutDouble($beamline.':SH:mp:bo', 0);
     &report_only($script,$ca_error,$beamline.':SH:mp:bo');
     return (1,$message,0);
  }
  if (defined $cnt_time) {
     ($int_avg) = &runJoergerRef ($scaler, $cnt_time, [$struck_pin_input]);
     $int_avg /= (1E5);                                     	#convert from counts to current
     &refresh_gui();
  } else {
     $int_avg = 0;
     for ($i=0; $i<10; $i++) {
        ($ca_error, $inten) = Pezca::GetDouble($beamline.':ADC_'.$adc_pin_input.'.VAL');
        if ($ca_error) {
           &report_only($script,$ca_error,$beamline.':ADC_'.$adc_pin_input.'.VAL');
           $inten = 0;
           $message = 'Unable to read ADC.';
           return (1,$message,0);
        }
        $int_avg += $inten;
        select(undef,undef,undef,0.2);
     }
     $int_avg /= 10;
     &refresh_gui();
  }
  &closeShutter ($shutter,$shutter_readback);

  ### compare value to historical
  if ($int_avg < ((1-$warn_level)*$expected_int)) {
     $message = sprintf 'Warning: Low intensity! (%1.0f%s lower than expected; threshold=%1.0f%s)!',
                        100*(1-$int_avg/$expected_int), '%', 100*$warn_level, '%';
     return (1,$message,$int_avg);
  }
  elsif ($int_avg > ((1+$warn_level)*$expected_int)) {
     $message = sprintf 'Intensity higher than expected by %1.0f%s (threshold=%1.0f%s).',
                       100*($int_avg/$expected_int-1), '%', 100*$warn_level, '%';
     return (1,$message,$int_avg);
  }
  else {
     $message = 'Intensity ok.';
     return (0,$message,$int_avg);
  }
}

#=================================================================

sub record_I0_temp_data_5um () {
  ### Record 5um intensities as a function of time:
  my ($struck, @struck_timer, @struck_i0, @struck_pin, $timeline);
  my ($struck_nmax, $struck_npts, $struck_intext, $struck_iserased);
  my ($struck_npts_saved, $struck_intext_saved, $struck_dwell_saved);
  my ($rmsVal_5time, $rmsDev_5time, $rmsI0_5time, $i0, $pin);
  my ($attenuation_saved, $attenuation_temp, $status, $msg);
  my ($kl_i0_gain_saved, $kl_pin_gain_saved, $kl_i0_time_saved, $kl_pin_time_saved);
  my ($kl_i0_gain_temp,  $kl_pin_gain_temp,  $kl_i0_time_temp,  $kl_pin_time_temp);
  
  $step_err = 0;
  $part2_message = '';
  $part3_message = '';
  $message = uc $steps[$currStep]{txt}.':';
  print STDOUT $message."\n";
  &part1(sprintf '%-40.40s',$message);

  print STDOUT '   Recording 5 um intensity variations over time'."\n";
  ($status,$attenuation_saved) = &getAttenuation1();
  printf STDOUT '     - Temporarily setting beam attenuation to 1x (was %.1fx)'."\n",$attenuation_saved;
  ($status,$attenuation_temp) = &setAttenuation1(1);

  $kl_i0_gain_saved  = &keithley_read_gain($kl_i0_channel);
  $kl_pin_gain_saved = &keithley_read_gain($kl_pin_channel);
  print STDOUT '     - Optimizing I0 and Pindiode Keithley gains for 5um measurements'."\n";
# $report = 1;									#temp for debug -- comment out!
  ($status,$msg) = &keithleyOptimizeGains(['I0','Pin'],$scaler);
  $report = 0;
  if ($status) {								#if failed
      $step_err = 1;
      $message = '     - Keithley optimization failed with status='.$status."\n"
               . '     - Message: '.$msg."\n";
      &part3($message);
      print STDOUT $message;
      $kl_i0_gain_temp  = $kl_i0_gain_saved/10;
      $kl_pin_gain_temp = $kl_pin_gain_saved/10;
      $status = &keithley_set_gain($kl_i0_channel,$kl_i0_gain_temp);
      $status = &keithley_set_gain($kl_pin_channel,$kl_pin_gain_temp);
      $status = 0;
  } else {
      $message = '     - Keithley optimization succeeded'."\n";
      &part2($message);
      print STDOUT $message;
  }
  &refresh_gui();
  $kl_i0_gain_temp  = &keithley_read_gain($kl_i0_channel);
  $kl_pin_gain_temp = &keithley_read_gain($kl_pin_channel);
  printf STDOUT '     - Temporarily set I0 Keithley gain to %.0e (was %.0e)'."\n",      $kl_i0_gain_temp, $kl_i0_gain_saved;
  printf STDOUT '     - Temporarily set Pindiode Keithley gain to %.0e (was %.0e)'."\n",$kl_pin_gain_temp,$kl_pin_gain_saved;

  $kl_i0_time_saved  = &keithley_read_time($kl_i0_channel,'label');
  $kl_pin_time_saved = &keithley_read_time($kl_pin_channel,'label');
  $kl_i0_time_temp  = '10 us'; 
  $kl_pin_time_temp = '10 us';
  if ($kl_i0_time_temp ne $kl_i0_time_saved) {
     printf STDOUT '     - Temporarily set I0 Keithley time to %s (was %s)'."\n",      $kl_i0_time_temp, $kl_i0_time_saved;
     $status = &keithley_set_time($kl_i0_channel,$kl_i0_time_temp,'nowait','label');
  }
  if ($kl_pin_time_temp ne $kl_pin_time_saved) {
     printf STDOUT '     - Temporarily set Pindiode Keithley time to %s (was %s)'."\n",$kl_pin_time_temp,$kl_pin_time_saved;
     $status = &keithley_set_time($kl_pin_channel,$kl_pin_time_temp,'nowait','label');
  }
  &refresh_gui();
  $struck = $struckName;
  if ($struck !~ /^${beamline}/) {$struck  = $beamline.':'.$struck;}
  $struck_npts_saved   = &getStruckNuseMax($struck);
  $struck_intext_saved = &getStruckAdvance($struck);
  $struck_dwell_saved  = &getStruckDwell ($struck);
  $struck_nmax         = &getStruckNmax($struck, $struck_ext);
  $struck_npts         = $struck_nmax;
  $struck_intext       = 0;							#internal advance
  $struck_iserased     = 0;							#not erased yet
  &setStruck ($struck, $struck_ext, $struck_intext, $struck_dwell, $struck_npts, $scaler);
  print STDOUT '   Opening data file '.$struck_file."\n";
  $filehandle = open (STRUCK,'> '.$struck_file); 				#open Struck data file for writing
  if (!defined $filehandle) {
     &press_key_exit('*** Could not open '.$struck_file);
  }
  select STRUCK; $|=1; 								#set unbuffered output
  printf STRUCK '# Date: %s'."\n", scalar(localtime);
  printf STRUCK '# Time/channel=%.0fms'."\n", 1000*$struck_dwell;
  printf STRUCK '# Struck  inputs:  Pindiode=%d    I0=%d'."\n",   $struck_pin_input, $struck_i0_input;
  printf STRUCK '# Keithley gains:  Pindiode=%.0e  I0=%.0e'."\n", $kl_pin_gain_temp, $kl_i0_gain_temp;
  print  STRUCK '#  Timeline(ms)    Pindiode           I0          Pindiode/I0'."\n";
  printf STDOUT '   Running Struck scaler '.$struck.' for %.0f seconds'."\n", $struck_dwell*$struck_nmax;
  $status = &openShutter($shutter,$shutter_readback);
  if ($status) {
     $step_err = 1;
     $message = "\t".'Could not open shutter.'."\n";
     &part2($message);
     &part3($message);
  }
  printf STDOUT '     - Running Struck: time/channel=%.0fms, Inputs=(%d,%d)'."\n",1000*$struck_dwell, $struck_pin_input, $struck_i0_input;
  &refresh_gui();
  &runStruck ($struck, $struck_iserased);
  &waitStruck($struck);
  &closeShutter($shutter,$shutter_readback);
  &refresh_gui();
  $struck_npts = &readStruckTimerRef ($struck, $struck_ext, \@struck_timer);
  &readStruckInputRef($struck, $struck_ext, $struck_npts, $struck_pin_input,  \@struck_timer, \@struck_pin);
  &readStruckInputRef($struck, $struck_ext, $struck_npts, $struck_i0_input,   \@struck_timer, \@struck_i0);
  $timeline = 0;
  ($rmsI0_5time, $rmsDev_5time) = &findRMS(\@struck_i0);
  for ($i=0; $i<$struck_npts; $i++) {
     $timeline += $struck_timer[$i];
     $i0 = $struck_i0[$i];
     if ($i0 <= 1) {$i0 = 1;}
     $pin = $struck_pin[$i] * ($rmsI0_5time/$i0);
     printf STRUCK ' %10.2f    %12.2f    %12.2f    %12.2f'."\n",
                   1000*$timeline, $struck_pin[$i], $struck_i0[$i], $pin;
     $struck_pin[$i] = $pin;							#normalize to I0 fluctuations
  }
  close(STRUCK);
  ($rmsVal_5time, $rmsDev_5time) = &findRMS(\@struck_pin);
  ### Restore Struck:
  printf STDOUT '     - Restoring Struck dwell to %.0fms'."\n",1000*$struck_dwell_saved;
  &setStruck ($struck, $struck_ext, $struck_intext_saved, $struck_dwell_saved, $struck_npts_saved, $scaler);
  printf STDOUT '     - Restoring beam attenuation to %.1fx'."\n",$attenuation_saved;
  ($status,$attenuation_temp) = &setAttenuation1($attenuation_saved);
  &refresh_gui();
  printf STDOUT '     - Restoring I0 Keithley gain to %.0e'."\n",$kl_i0_gain_saved;
  $status = &keithley_set_gain($kl_i0_channel,$kl_i0_gain_saved);
  if ($kl_i0_time_temp ne $kl_i0_time_saved) {
     printf STDOUT '     - Restoring I0 Keithley time to %s'."\n",      $kl_i0_time_saved;
     $status = &keithley_set_time($kl_i0_channel,$kl_i0_time_saved,'nowait','label');
  }
  &refresh_gui();
  printf STDOUT '     - Restoring Pindiode Keithley gain to %.0e'."\n",$kl_pin_gain_saved;
  $status = &keithley_set_gain($kl_pin_channel,$kl_pin_gain_saved);
  if ($kl_pin_time_temp ne $kl_pin_time_saved) {
     printf STDOUT '     - Restoring Pindiode Keithley time to %s'."\n",$kl_pin_time_saved;
     $status = &keithley_set_time($kl_pin_channel,$kl_pin_time_saved,'nowait','label');
  }
  &refresh_gui();
  print STDOUT '   5 um intensity variations over time are recorded.'."\n";
  &refresh_gui();
  return ($struck_npts, $rmsVal_5time, $rmsDev_5time);
}

#=================================================================

sub measure_Beamsize () {

  ### Reset messages

  my ($ca_error1, $ca_error2);

  $step_err = 0;
  $part2_message = '';
  $part3_message = '';
  $message = 'BEAM SIZE MEASUREMENTS:';
  &part1(sprintf '%-40.40s',$message);
  print STDOUT $message."\n";

  ### Measure beam profiles

  @prg = (@scriptsPrefix, $PEZCA.'measureMiniBeam.pl', 'noninteractive', 'noAlign');
  $status1 = &system_with_fork(@prg);
  if ($pause_after) {&press_key_continue('Finished measureMiniBeam.pl, status='.$status1);}
  if ($status1) {
     $message = "\t".'Backup reports unclean exit.'."\n";
     &part2($message);
     &part3($message);
     $step_err = 1;
  }
  else {
     ($ca_error1,$v_size) = Pezca::GetDouble($beamline.':bi:beamsize_v_act1.VAL');
     if ($ca_error1) {&report_only($script,$ca_error1,$beamline.':bi:beamsize_v_act1.VAL'); $v_size=0;}
     ($ca_error2,$h_size) = Pezca::GetDouble($beamline.':bi:beamsize_h_act1.VAL');
     if ($ca_error2) {&report_only($script,$ca_error2,$beamline.':bi:beamsize_h_act1.VAL'); $h_size=0;}
     if (($ca_error1) || ($ca_error2)) {
        $step_err = 1;
        $message = "\t".'Could not read beam sizes. CAGET error.'."\n";
        &part2($message."\n");
        &part3($message."\n");
     }
     else {
        $message  = sprintf "\t".'Beam size: Vert=%.4fum (expected=%.4fum)'."\n",$v_size,$hist_vsize;
        $message .= sprintf "\t".'Beam size: Horz=%.4fum (expected=%.4fum)'."\n",$h_size,$hist_hsize;
        &part1('OK?'."\n");
        &part2($message."\n");
     }
  }

  ### Return YAG to initial position
  print STDOUT '   Restoring gonio horz'."\n";
  &refresh_gui();
  $status1 = &runMotor ($beamline.':'.$gh_assy, $beamline.':'.$gh_axis, $gon_h_in, 'checkAuto');
  if ($status1) {
     $message = "\t".'Could not return gonio.'."\n";
     &part2($message);
     &part3($message);
     $step_err = 1;
  }
  &refresh_gui();

  print STDOUT $part1_text;
  $part1_text = '';
  print STDOUT "\n";
}

#=================================================================

sub bluice_settings () {

  $step_err = 0;
  $part2_message = '';
  $part3_message = '';
  $message = 'BLUICE RUNTIME SETTINGS:';
  &part1(sprintf '%-40.40s',$message);
  print STDOUT $message."\n";
  print STDOUT '   Checking JBluIce runtime settings...'."\n";

### MH: The JBluIce "beamstop check" was going to use the active beamstop
### in some way. It was never implemented.  JBluIce currently does not
### use this PV, so it can be removed from the user startup script and
### the caQtDM window.
# $err = &getPV(@beamstop_chk); if ($err) {$step_err = 1;}
  $err = &getPV(@shutter_chk);  if ($err) {$step_err = 1;}
  $err = &getPV(@feedback_chk); if ($err) {$step_err = 1;}
  $err = &getPV(@egap_chk);     if ($err) {$step_err = 1;}

  if ($step_err) {
     &part3($part3_message."\n");
  }
  else {
     &part1('OK!'."\n");
     &part2("\t".'JBluIce settings are correct.'."\n\n");
  }
  print STDOUT $part1_text;
  $part1_text = '';
  print STDOUT "\n";
}

#=================================================================

sub check_mirror_mtr_disabled () {

  my (@kbm, @hdm, $nkbm, $nhdm, $i, $value_type, $value);

### Verify mirror motors amplifiers are disabled (AmpEna)
### HDM:
  $i = 0;
  $hdm[$i] = ['HD:muy1', 'HDM UY1 should be disabled.']; $i++;
  $hdm[$i] = ['HD:muy2', 'HDM UY2 should be disabled.']; $i++;
  $hdm[$i] = ['HD:mdy1', 'HDM DY1 should be disabled.']; $i++;
  $hdm[$i] = ['HD:mdy2', 'HDM DY2 should be disabled.']; $i++;
  $hdm[$i] = ['HD:mz1',	 'HDM Z1 should be disabled.'];  $i++;
  $hdm[$i] = ['HD:mz2',	 'HDM Z2 should be disabled.'];  $i++;
  $hdm[$i] = ['HD:mz3',	 'HDM Z3 should be disabled.'];  $i++;
  $nhdm = $i;
### KBM:
  $i = 0;
  $kbm[$i] = ['KB:mhy1', 'HFM Y1 should be disabled.'];  $i++;
  $kbm[$i] = ['KB:mhy2', 'HFM Y2 should be disabled.'];  $i++;
  $kbm[$i] = ['KB:mhz1', 'HFM Z1 should be disabled.'];  $i++;
  $kbm[$i] = ['KB:mhz2', 'HFM Z2 should be disabled.'];  $i++;
  $kbm[$i] = ['KB:mhz3', 'HFM Z3 should be disabled.'];  $i++;
  $kbm[$i] = ['KB:mvy1', 'VFM Y1 should be disabled.'];  $i++;
  $kbm[$i] = ['KB:mvy2', 'VFM Y2 should be disabled.'];  $i++;
  $kbm[$i] = ['KB:mvz1', 'VFM Z1 should be disabled.'];  $i++;
  $kbm[$i] = ['KB:mvz2', 'VFM Z2 should be disabled.'];  $i++;
  $kbm[$i] = ['KB:mvz3', 'VFM Z3 should be disabled.'];  $i++;
  $nkbm = $i;

  $value_type = 1;                      #check for single integer $value
  $value = 0;

  $step_err = 0;
  $part2_message = '';
  $part3_message = '';
  $message = 'MIRROR MOTORS DISABLED:';
  &part1(sprintf '%-40.40s',$message);
  print STDOUT $message."\n";

# no strict 'refs';  #-----------------------------------------------------------------------#
# foreach $m (qw/hfm_y1 hfm_y2 hfm_z1 hfm_z2 hfm_z3 vfm_y1 vfm_y2 vfm_z1 vfm_z2 vfm_z3/) {
#    if (@$m[0] !~ /:AmpEna$/) {@$m[0] .= ':AmpEna';}
#    $err = &getPV(@$m); if ($err) {$step_err = 1;}
# }
#    foreach $m (qw/hdm_uy1 hdm_uy2 hdm_dy1 hdm_dy2 hdm_z1 hdm_z2 hdm_z3/) {
#       if (@$m[0] !~ /:AmpEna$/) {@$m[0] .= ':AmpEna';}
#       $err = &getPV(@$m); if ($err) {$step_err = 1;}
#    }
# use strict 'refs'; #-----------------------------------------------------------------------#

  print STDOUT '   Checking to see if KBM motors are disabled...'."\n";
  for ($i=0; $i<$nkbm; $i++) {
     $kbm[$i][0] = $beamline.':'.$kbm[$i][0].':AmpEna';
     $err = &getPV($kbm[$i][0],$kbm[$i][1],$value_type,$value); if ($err) {$step_err = 1;}
  }
  if ($beamline eq '23o') {
     print STDOUT '   Checking to see if HDM motors are disabled...'."\n";
     for ($i=0; $i<$nhdm; $i++) {
        $hdm[$i][0] = $beamline.':'.$hdm[$i][0].':AmpEna';
        $err = &getPV($hdm[$i][0],$hdm[$i][1],$value_type,$value); if ($err) {$step_err = 1;}
     }
  }

  if ($step_err) {
     &part1('ERR'."\n");
     &part3($part3_message."\n");
  }
  else {
     &part1('OK!'."\n");
     &part2("\t".'All mirror motors are disabled.'."\n\n");
  }
  print STDOUT $part1_text;
  $part1_text = '';
  print STDOUT "\n";
}

#=================================================================

sub backup_motors () {

  $step_err = 0;
  $part2_message = '';
  $part3_message = '';
  $message = 'BACKING UP MOTORS:';
  print STDOUT $message."\n";
  &part1(sprintf '%-40.40s',$message);


  $filename = sprintf $logdir.'%s-%04d-%02d-%02d_%02d-%02d.snap',$beamline,$year,$mon,$mday,$hour,$min;
  $filename =~ s/ /0/g;

  print STDOUT '   Backing up motors to:'."\n";
  print STDOUT '   '.$filename."\n";
  @prg = (@scriptsPrefix, $PEZCA.'mtrPositionsBackup.pl', $filename, $beamline, 'noninteractive');
  $status1 = &system_with_fork(@prg);
  if ($pause_after) {&press_key_continue('Finished mtrPositionsBackup.pl, status='.$status1);}

  if ($status1) {
     $message = "\t".'Backup reports unclean exit.'."\n";
     &part2($message);
     &part3($message);
     $step_err = 1;
  }

  if (-e $filename) {
     &part2("\t".'File: '.$filename."\n");
  }
  else {
     $message = "\t".'Backup failed!'."\n";
     &part2($message);
     &part3($message);
     $message = "\t".'No file: '.$filename."\n";
     &part2($message);
     &part3($message);
     $step_err = 1;
  }

  if ($step_err) {&part1('ERR'."\n");}
  else           {&part1('OK!'."\n");}
  print STDOUT $part1_text,"\n";
  $part1_text = '';
}

#=================================================================

sub record_motors () {

  my ($i, $n);

  $step_err = 0;
  $part2_message = '';
  $part3_message = '';
  $message = 'MOTOR POSITIONS:';
  print STDOUT $message."\n";
  &part1(sprintf '%-40.40s',$message);				#Part1 is for the screen (step header)
  &part2($message."\n");					#Part2 is for the printout (details)
  &part3($message."\n");					#Part is for the printout (error summary)

  print STDOUT '   Recording beamline motor positions (white-beam hutch)'."\n";
  $n = @wbmtr;
  for ($i=0; $i<$n; $i++) {
     if ($wbmtr[$i][0] !~ /:ActPos$/) {$wbmtr[$i][0] .= ':ActPos';}
     $err = &getPV(@{$wbmtr[$i]});       if ($err) {$step_err = 1;}
  }
  print STDOUT '   Recording beamline motor positions (mono-beam hutch)'."\n";
  $n = @mbmtr;
  for ($i=0; $i<$n; $i++) {
     if ($mbmtr[$i][0] !~ /:ActPos$/) {$mbmtr[$i][0] .= ':ActPos';}
     $err = &getPV(@{$mbmtr[$i]});    if ($err) {$step_err = 1;}
  }
  print STDOUT '   Recording end-station motor positions'."\n";
  $n = @esmtr;
  for ($i=0; $i<$n; $i++) {
     if ($esmtr[$i][0] !~ /:ActPos$/) {$esmtr[$i][0] .= ':ActPos';}
     $err = &getPV(@{$esmtr[$i]});       if ($err) {$step_err = 1;}
  }
  print STDOUT '   Recording misc.PVs (DAC, ADC,...)'."\n";
  $n = @vals;
  for ($i=0; $i<$n; $i++) {
     $err = &getPV(@{$vals[$i]});        if ($err) {$step_err = 1;}
  }

  if ($step_err) {
     &part1('ERR'."\n");
     &part2($part2_message."\n");
     &part3($part3_message."\n");
  } else {
     &part1('OK!'."\n");
     &part2($part2_message."\n");
     &part3("\n");
  }
  print STDOUT $part1_text;
  $part1_text = '';
  print STDOUT "\n";
}

#=================================================================

sub backup_voltages () {

  $step_err = 0;
  $part2_message = '';
  $part3_message = '';
  $message = 'BACKING UP MIRROR VOLTAGES:';
  print STDOUT $message."\n";
  &part1(sprintf '%-40s',$message);

  $filename = sprintf $logdir.$beamline.'-'.$year.'-%02d-%02d-%02d.%02d.sy900',$mon,$mday,$hour,$min;
  $filename =~ s/ /0/g;

  print STDOUT '   Backing up mirror voltages to:'."\n";
  print STDOUT '   '.$filename."\n";
  @prg = (@scriptsPrefix, $PEZCA.'mirrorsBackup.pl', $filename, $beamline, 'noninteractive');
# @prg = (@scriptsPrefix, $PEZCA.'mirrorsBackup.pl', $filename, $beamline);
  $status1 = &system_with_fork(@prg);
  if ($pause_after) {&press_key_continue('Finished mirrorsBackup.pl, status='.$status1);}

  if ($status1) {
     $message = "\t".'Backup reports unclean exit.'."\n";
     &part3($message);
     $step_err = 1;
  }

  if (-e $filename) {
     &part2("\t".'File: '.$filename."\n");
  }
  else {
     $message = "\t".'Backup failed!'."\n";
     &part3($message);
     $message = "\t".'No file: '.$filename."\n";
     &part3($message);
     $step_err = 1;
  }

  if ($status1) {&part1('ERR'."\n");}
  else          {&part1('OK!'."\n");}
  print STDOUT $part1_text,"\n";
}

#=================================================================

sub record_voltages () {

  my ($maxVdiff, $segmentsHFM, $lengthHFM);
  my ($segmentsVFM, $lengthVFM, $i);
  my ($pv, $msg, $value_type);
  my ($h_zero, $v_zero);
  
  $step_err = 0;
  $part2_message = '';
  $part3_message = '';
  $message = 'MIRROR VOLTAGES:';
  print STDOUT $message."\n";
  &part1(sprintf '%-40.40s',$message);

  $step_err = &getSY900layout($beamline,\$maxVdiff,
                      \$segmentsHFM,\$lengthHFM,
                      \$segmentsVFM,\$lengthVFM);
 
  if ($step_err) {
     $step_err = 0;
     $message = "\t".'Cannot find bimorph mirrors data for beamline='.$beamline."\n";
     &part3($message);
     ($segmentsVFM,$segmentsHFM) = (0,0);
  }

  $msg = undef;
  $value_type = 4;                              #mirror voltages (no expectations)
  
  # Vertical mirror
  print STDOUT '   Recording VFM voltages'."\n";
  $v_zero = 1;                                  #presume V-mirror is off (all voltages < 1v)
  for ($i=0; $i<$segmentsVFM; $i++) {
     $pv = sprintf $beamline.':VFM:GET-VOUT%02d', $i;
     ### RETURNS: 0=OK, 1=ERROR, 2=Mirror is off (voltage < 1V) in $value_type=4
     $err = &getPV($pv,$msg,$value_type);  
     if    ($err == 1) {$step_err=1;}  
     elsif ($err == 0) {$v_zero=0;}             #if at least one abs(voltage)>1ev, then V-mirror voltages are OK
  }

  ### Horizontal mirror
  print STDOUT '   Recording HFM voltages'."\n";
  $h_zero = 1;                                  #presume H-mirror is off (all voltages < 1v)
  for ($i=0; $i<$segmentsHFM; $i++) {
     $pv = sprintf $beamline.':HFM:GET-VOUT%02d', $i;
     ### RETURNS: 0=OK, 1=ERROR, 2=Mirror is off (voltage < 1V) in $value_type=4
     $err = &getPV($pv,$msg,$value_type);  
     if    ($err == 1) {$step_err=1;}  
     elsif ($err == 0) {$h_zero=0;}             #if at least one abs(voltage)>1ev, then H-mirror voltages are OK
  }

  if ($v_zero) {&part3("\t".'Vertical mirror appears to be turned off.'."\n")}
  if ($h_zero) {&part3("\t".'Horizontal mirror appears to be turned off.'."\n")}


  #print STDOUT "($step_err) || ($v_zero) || ($h_zero)\n";

  if (($step_err) || ($v_zero) || ($h_zero)) {
     &part3($part3_message."\n");
  }
  else {
     &part1('OK!'."\n");
     &part2($part2_message."\n");
  }
  print STDOUT $part1_text;
  $part1_text = '';
  print STDOUT "\n";
}

#=================================================================

sub record_CRL () {

  my ($pv, $msg, $value_type);
  my ($device, @deviceHashData, $nrecords);
  my ($nCRL, $radiusCRL, $materialCRL);
  my ($energy, $focus, @CRL_axes, $a);
  
  $step_err = 0;
  $part2_message = '';
  $part3_message = '';
  $message = 'CRL CONFIG & POSITION:';
  print STDOUT $message."\n";
  &part1(sprintf '%-40.40s',$message);

  $device = 'CRL-old';
  $nrecords = &readCRLconfigSQLpopulated($device,\@deviceHashData);
  if ($nrecords == 1) {
     $nCRL        = $deviceHashData[0]{'Number'};
     $radiusCRL   = $deviceHashData[0]{'Radius'};
     $materialCRL = $deviceHashData[0]{'Material'};
     $message = 'nCRL='.$nCRL.' radius='.$radiusCRL.' material='.$materialCRL."\n";
     &part2(sprintf '%-40.40s',$message);
     $energy = &getAxisAct ($beamline.':MO:E',\$ca_error);
     if ($ca_error) {
        $energy = 12;
        $step_err = 1;
        $message = "\t".'Could not read energy. CAGET error.'."\n";
        &part3($message);
     } else {
        $focus = &calcCRLfocus($energy,$radiusCRL,$nCRL,$materialCRL);
        $message = sprintf "\t".'CRL focus=%.2fmm at E=%.2fKeV',$focus,$energy;
        &part2($message."\n");
     }
  } else {
     $step_err = 1;
     $message ="\t".'Unexpected number of SQL records='.$nrecords.' for device='.$device;
     &part3($message."\n");
  }

  $msg = undef;
  $value_type = 0;                              #single value (no expectations)
  @CRL_axes = qw(XS YS YB ZB AP AW);
  print STDOUT '   Recording CRL box positions'."\n";
  foreach $a (@CRL_axes) {
     $pv = $beamline.':CRL:'.$a.':ActPos';
     $err = &getPV($pv,$msg,$value_type);  
     if ($err == 1) {$step_err=1;}  
  }  
  if ($step_err) {
     &part1('ERR'."\n");
     &part3($part3_message."\n");
  } else {
     &part1('OK!'."\n");
     &part2($part2_message."\n");
  }
}

#=================================================================

sub record_Transfocator () {

  my ($pv, $msg, $value_type);
  my ($device, @deviceHashData, $nrecords);
  my ($energy, $focus, @TF_axes, $a);
  my (@InOut, @states);
  
  $step_err = 0;
  $part2_message = '';
  $part3_message = '';
  $message = 'TRANSFOCATOR CONFIG & POSITION:';
  print STDOUT $message."\n";
  &part1(sprintf '%-40.40s',$message);

  $device = 'CRL-2020';
  @states = qw(Out In);
  $nrecords = &readCRLconfigSQLpopulated($device,\@deviceHashData);
  if ($nrecords > 0) {
     for ($i=0; $i<$nrecords; $i++) {
        $pv = $deviceHashData[$i]{'PV_ctrl'};
        $pv =~ s/^${beamline}:pmac20:acc65//;
        $message = sprintf "\t".'%02d. name=%-12s'
#                .' PV(ctrl)=%-20s PV(in)=%-20s PV(out)=%-20s'
                 .' PV(ctrl)=%-8s'
                 .' offset=%-3.0f radius=%-3.0f number=%-2d material=%s'
                ,($i+1)
#               ,$deviceHashData[$i]{'Device'} 		#CRL-box name
                ,$deviceHashData[$i]{'Name'} 		#Name of lenses subset driven by $PV_ctrl
                ,$pv				 	#Digital output for lenses In/Out
#               ,$deviceHashData[$i]{'PV_ctrl'} 	#Digital output for lenses In/Out
#               ,$deviceHashData[$i]{'PV_rb_in'} 	#Digital input for lenses in the beam
#               ,$deviceHashData[$i]{'PV_rb_out'} 	#Digital input for lenses out of the beam
                ,$deviceHashData[$i]{'Offset'} 		#Offset (mm) of lenses subset from the centre of the whole box (may not be used)
                ,$deviceHashData[$i]{'Radius'} 		#Lenses radius (um). If $radius=0 no lenses are installed in the subset
                ,$deviceHashData[$i]{'Number'} 		#Number of lenses in the box. In $number=0, no lenses are installed in the subset
                ,$deviceHashData[$i]{'Material'} 	#Lenses material, normally Be. If $material is blank, no lenses are installed in the subset
                ;
        &part2($message."\n");
     }
     $energy = &getAxisAct ($beamline.':MO:E',\$ca_error);
     if ($ca_error) {
        $energy = 12;
        $step_err = 1;
        $message = "\t".'Could not read energy. CAGET error.'."\n";
        &part3($message);
     } else {
        $focus = &getTransfocatorFocus4($energy,$device,\$msg,\@InOut);
        $message = sprintf "\t".'Transfocator ('.$device.') focus=%.2fmm at E=%.2fKeV',$focus,$energy;
        &part2($message."\n");
        for ($i=0; $i<$nrecords; $i++) {
           $pv = $deviceHashData[$i]{'PV_ctrl'};
           $pv =~ s/^${beamline}:pmac20:acc65//;
#          $message = sprintf "\t".'%02d. name=%-12s PV(ctrl)=%-20s state=%s'
           $message = sprintf "\t".'%02d. name=%-12s PV(ctrl)=%-8s state=%s'
                ,($i+1)
                ,$deviceHashData[$i]{'Name'} 		#Name of lenses subset driven by $PV_ctrl
                ,$pv					#Digital output for lenses In/Out
#               ,$deviceHashData[$i]{'PV_ctrl'} 	#Digital output for lenses In/Out
                ,$states[$InOut[$i]]
                ;
        &part2($message."\n");
     }
     }
  } else {
     $step_err = 1;
     $message ="\t".'Unexpected number of SQL records='.$nrecords.' for device='.$device."\n";
     &part3($message);
  }

  $msg = undef;
  $value_type = 0;                              #single value (no expectations)
  @TF_axes = qw(X Z Y PT YW RL);
  print STDOUT '   Recording Transfocator box positions'."\n";
  foreach $a (@TF_axes) {
     $pv = $beamline.':CRL:'.$a.':ActPos';
     $err = &getPV($pv,$msg,$value_type);  
     if    ($err == 1) {$step_err=1;}  
  }  
  if ($step_err) {
     &part1('ERR'."\n");
     &part3($part3_message."\n");
  } else {
     &part1('OK!'."\n");
     &part2($part2_message."\n");
  }
}

#=================================================================

sub record_maxspeeds () {	        # Update mySQL with max speeds of Gonio Omega and Crystal XYZ
  
  $step_err = 0;
  $part2_message = '';
  $part3_message = '';
  $message = uc $steps[$currStep]{txt}.':';
  print STDOUT $message."\n";
  &part1(sprintf '%-40.40s',$message);

  @prg = (@scriptsPrefix, $PEZCA.'saveMaxSpeeds2sql.pl', $beamline);
  $status1 = &system_with_fork(@prg);
  if ($pause_after) {&press_key_continue('Finished saveMaxSpeeds2sql.pl, status='.$status1);}
  if ($status1) {
     $message = "\t".'Backup reports unclean exit.'."\n";
     &part1('ERR'."\n");
     &part3($message);
     $step_err = 1;
  } else {
    &part1("\n".'Speeds saved OK'."\n");
  }
}

#=================================================================

sub check_robot_homing () {

  my ($pv, $tag, $ca_error, @words, $nwords, $scr);
  $scr = 'checkRobotHoming.pl';
  
  $step_err = 0;
  $part2_message = '';
  $part3_message = '';
  $message = uc $steps[$currStep]{txt}.':';
  print STDOUT $message."\n";
  &part1(sprintf '%-40.40s',$message);

  @prg = (@scriptsPrefix, $PEZCA.$scr, $beamline, 'nopopup');
  $status1 = &system_with_fork(@prg);
  if ($pause_after) {&press_key_continue('Finished '.$scr.', status='.$status1);}
  if ($status1) {
     $message = "\t".$scr.' reports unclean exit.'."\n";
     &part1('ERR'."\n");
     &part3($message);
     $step_err = 1;
     return;
  }
### The checkRobotHoming script reports deviations in one of 3 formats depending
### what fits the string (gripper-V, gripper-H, gripper-Rot, dewar-Rot):
### V=%.3f H=%.3f R=%.3f D=%.3f
### %.3f %.3f %.3f %.3f
### %.2f %.2f %.2f %.3f
  $pv = $beamline.':RB:Gr:Sync.DESC';
  ($ca_error,$tag) = Pezca::GetString($pv);		#we ignore this error
  if ($ca_error) {
     &report_only($script,$ca_error,$pv);
     $tag = '';
  }
  @words = split(/\s+/,$tag);
  $nwords = @words;
  if ($nwords != 4) {
     $message = "\t".$scr.' did not fill deviations properly: ['.$tag.']'."\n";
     &part1('ERR'."\n");
     &part3($message);
     $step_err = 1;
     return;
  }
  if ($tag =~ /V=/ && $tag =~ /H=/ && $tag =~ /R=/ && $tag =~ /D=/) {
     $status1 = 0;
     if ($words[0] !~ /V=/) {$status1++;} else {$words[0] =~ s/V=//;}
     if ($words[1] !~ /H=/) {$status1++;} else {$words[1] =~ s/H=//;}
     if ($words[2] !~ /R=/) {$status1++;} else {$words[2] =~ s/R=//;}
     if ($words[3] !~ /D=/) {$status1++;} else {$words[3] =~ s/D=//;}
     if ($status1) {
        $message = "\t".$scr.' did not fill homing deviations properly: ['.$tag.']'."\n";
        &part1('ERR'."\n");
        &part3($message);
        $step_err = 1;
        return;
     }
  }
  $status1 = 0;
  for ($i=0; $i<$nwords; $i++) {if (! &is_decimal($words[$i])) {$status1++;}}
  if ($status1) {
     $message = "\t".$scr.' did not fill homing deviations properly: ['.$tag.']'."\n";
     &part1('ERR'."\n");
     &part3($message);
     $step_err = 1;
     return;
  }
### Test deviations for reasonable values (ask Oleg). Should we place them in expectations?
  $status1 = 0;
  if (abs($words[0]) > 0.01)   {$status1++;}                      #V in mm (10um)
  if (abs($words[1]) > 0.01)   {$status1++;}                      #H in mm (10um)
  if (abs($words[2]) > 0.0029) {$status1++;}                      #R in degr. (10um if radius is 200mm)
  if (abs($words[3]) > 0.05)   {$status1++;}                      #D in degr.
  if ($status1) {
     $message = sprintf "\t".$scr.': LARGE homing deviation(s): GripperV=%.4f GripperH=%.4f GripperR=%.4f Dewar'."\n",$words[0],$words[1],$words[2],$words[3];
     &part1('ERR'."\n");
     &part3($message);
     $step_err = 1;
     return;
  }
  &part1("\n".'Robot homing is OK'."\n");
}

#=================================================================

sub check_dectris_eth () {
        
  my ($ssh, $cmd, $dcu, $port, $loc, $buf, @lines);

  $step_err = 0;
  $part2_message = '';
  $part3_message = '';
  $message = uc $steps[$currStep]{txt}.':';
  print STDOUT $message."\n";
  &part1(sprintf '%-40.40s',$message);

# if    ($hostname =~ /^bl1ws2/) {$dcu = 'bl1pilatus'; $port=22;    $loc='/mnt/software/gmca_epics/pezca';}
  if    ($hostname =~ /^bl1ws2/) {$dcu = 'bl1dcu';     $port=22;    $loc='/mnt/software/gmca_epics/pezca';}
  elsif ($hostname =~ /^bl2ws2/) {$dcu = 'bl2dcu';     $port=22222; $loc='/home/gmca';}
  else {
     $message = "\t".'check_dectris_eth: host='.$hostname.' is not allowed to check PAD connections.'."\n";
     &part1('ERR'."\n");
     &part3($message);
     $step_err = 1;
     return;
  }

  $ssh = '/usr/bin/ssh -p '.$port.' -aqxt -o BatchMode=yes -o StrictHostKeyChecking=no gmca@'.$dcu;
  $cmd = '/usr/bin/sudo -n '.$loc.'/checkdetectorconnections.pl';
  $ssh .= ' '.$cmd;
  $buf = `$ssh`;
  if ($buf !~ /No today\'s failed 10Gbps connections/) { 
     @lines = split(/\r?\n/,$buf);
     @lines = grep {$_ !~ /^$/} @lines;
     $buf = join("\n",@lines);
     if ($buf !~ /Today\'s failed 10Gbps connections/) {
        $message = "\t".'check_dectris_eth: problem while checking PAD connections: ['.$buf.']'."\n"
                 . "\t".'CMD: ['.$ssh.']'."\n";
     } else {
        $message = "\t".'check_dectris_eth: found failed PAD connections: ['.$buf.']'."\n";
     }
     &part1('ERR'."\n");
     &part3($message);
     $step_err = 1;
     return;
  } else {
     &part1("\n".$buf);
  }
}

#=================================================================

sub mono_calibration_history() {
  our ($monoHistWin);
  my ($script, $str, @logs, $log, @lines, $line);
  my ($date, $out, @dates, @calib, $nLog, $i);
  my (@gnuopts, $gnuopen, $gnutitle, $gnulogscale, $title, $answer, $rt);
  my ($start, $last);
  $script = 'mono_calibration_history';
  $str = 'find '.$logdir.' -mtime -'.$history.' -name "'.$logroot.'_'.$beamline.'_*.log"';
  @logs = split(/\r?\n/,`$str`);
  @logs = sort @logs;
  @dates = ();
  @calib = ();
  foreach $log (@logs) {
     if ($debug) {print STDOUT $script.'(DEBUG): logfile='.$log."\n";}
     $date = $log;
     $date =~ s/^.*user_setup_${beamline}_//;
     $date = substr($date, 0, 10);
     $date =~ s/[\/\._]/-/g;
     $str = 'grep "Mono energy was corrected by dE=" '.$log;
     if ($debug) {print STDOUT $script.'(DEBUG): str='.$str."\n";}
     @lines = split(/\r?\n/,`$str`);
     if ($debug) {foreach $line (@lines) {print STDOUT $script.'(DEBUG): line='.$line."\n";}}
     if (@lines < 1)                {next;}
     $i = $#lines;							#use last correction
     $lines[$i] =~ s/^.*Mono energy\ was\ corrected\ by\ dE=//;
     $lines[$i] =~ s/\s*eV$//;
     if (! &is_float($lines[$i]))   {next;}
     push(@dates,$date);
     push(@calib,$lines[$i]);
  }
  ($start,$last) = &calc_date_range($history,undef,'-');		#must match format of $date
  $title    = 'MONO CALIBRATION HISTORY';
  $out      = $title."\n\n";
  $gnutitle = $title.' Plot';
  @gnuopts = ('-persist');
  $gnulogscale = 0;
  $nLog = @dates;
  if ($nLog > 0) {
     $out .= sprintf '%-10s  %s'."\n", 'Date', 'Energy corr.(eV)';
     for ($i=0; $i<$nLog; $i++) {$out .=  sprintf '%s  %11s'."\n", $dates[$i],$calib[$i];}
     if ($nLog > 1) {
        system ('wmctrl -F -c "'.$gnutitle.'" &');
        $gnuopen = &open_gnuplot (undef, undef, $gnulogscale, \@gnuopts);
        if ($gnuopen) {
           print GNUPLOT 'set term x11 title "'.$gnutitle.'"'."\n";
           print GNUPLOT 'set xdata time'."\n";
           print GNUPLOT 'set timefmt "%Y-%m-%d"'."\n";			#must match format of $date
           print GNUPLOT 'set xrange ["'.$start.'":"'.$last.'"]'."\n";
           print GNUPLOT 'set format y "%6.2f"'."\n";
           print GNUPLOT 'set xlabel "Date"'."\n";
           print GNUPLOT 'set ylabel "Correction (eV)"'."\n";
           print GNUPLOT 'plot "-" using 1:2 title "Energy corr." with linespoints pt 7 ps 2'."\n";
           for ($i=0; $i<$nLog; $i++) {printf GNUPLOT '%10s  %11s'."\n", $dates[$i],$calib[$i];}
           print GNUPLOT 'e'."\n";
           &close_gnuplot();
        }
     }
  } else {
    $out .= ' No data for the last '.$history.' days'."\n";
  }
# print STDOUT $out;
  if (Exists($monoHistWin)) {$monoHistWin->destroy();}

  if ($use_displayer) {system('displayer.pl "'.$out.'" "'.$gnutitle.'" &'); return;}

  $monoHistWin = $mw->DialogBox(  -title=>$title,
                                  -buttons=>['OK'],
                                  -default_button=>'OK',
                                  -takefocus=>1,
                                  );
  $rt = $monoHistWin->ScrlROText( -width=>32,
                                  -height=>40,
#                                 -scrollbars=>'se',
                                  -scrollbars=>'e',
                                  -font=>$fontTbl,
                                  -wrap=>'none',
                                  )->pack();
# $rt->Subwidget("xscrollbar")->configure(-background=>$bgcolor,);
  $rt->Subwidget("yscrollbar")->configure(-background=>$bgcolor,);
  $rt->insert(end => $out);

# $monoHistWin = $mw->Dialog(     -title=>$title,
#                                 -text=>$out,
#                                 -wrap=>0,
#                                 -font=>$fontBtn,
#                                 -buttons=>['OK'],
#                                 -default_button=>'OK',
#                                 -takefocus=>1,
#                                 -width=>32,
#                                 );
  $monoHistWin->Subwidget('B_OK')->configure(-font=>$fontBtn,
                                             -background=>$bgcolor, -activebackground=>$abgcolor, 
                                             -foreground=>$fgcolor, -activeforeground=>$afgcolor,
                                             );
  $answer = $monoHistWin->Show;
  if ($answer eq 'OK') {
     $monoHistWin->destroy();
     system ('wmctrl -F -c "'.$gnutitle.'" &');
  }
}

#=================================================================

sub beam_steering_history() {
  our ($steerHistWin);
  my ($script, $str, @logs, $log, @lines, @hw, $type, $line);
  my ($date, %steer, $out, @dates, @steerV, @steerH, $nLog, $i, $j);
  my (@gnuopts, $gnuopen, $gnutitle, $gnulogscale, $title, $answer, $rt);
  my ($start, $last);
  $script = 'beam_steering_history';
  @hw = qw(VERTICAL HORIZONTAL);
  $str = 'find '.$logdir.' -mtime -'.$history.' -name "'.$logroot.'_'.$beamline.'_*.log"';
  @logs = split(/\r?\n/,`$str`);
  @logs = sort @logs;
  @dates  = ();
  @steerV = ();
  @steerH = ();
  foreach $log (@logs) {
     if ($debug) {print STDOUT $script.'(DEBUG): logfile='.$log."\n";}
     $date = $log;
     $date =~ s/^.*user_setup_${beamline}_//;
     $date = substr($date, 0, 10);
     $date =~ s/[\/\._]/-/g;
     foreach $type (@hw) {
        $steer{$type} = undef;
        $str = 'grep -A 5 "WB-SLITS '.$type.'.* CENTERING:" '.$log.' | grep "Required beam steering"';
        if ($debug) {print STDOUT $script.'(DEBUG): str='.$str."\n";}
        @lines = split(/\r?\n/,`$str`);
        if ($debug) {foreach $line (@lines) {print STDOUT $script.'(DEBUG): line='.$line."\n";}}
        if (@lines < 1)                                 {next;}
        $i = $#lines;
        if ($lines[$i] !~ /^.*Required beam steering=/) {next;}
        $steer{$type} = $lines[$i];
        $steer{$type} =~ s/^.*Required beam steering=//;
        $steer{$type} =~ s/urad//;
     }
     if (defined $steer{$hw[0]} || defined $steer{$hw[1]}) {
        push(@dates,$date);
        if (defined $steer{$hw[0]}) {push(@steerV,$steer{$hw[0]});} else {push(@steerV,0);}
        if (defined $steer{$hw[1]}) {push(@steerH,$steer{$hw[1]});} else {push(@steerH,0);}
     }
  }
  ($start,$last) = &calc_date_range($history,undef,'-');		#must match format of $date
  $title    = 'BEAM STEERING HISTORY'; 
  $out      = $title."\n\n";
  $gnutitle = $title.' Plot';
  @gnuopts = ('-persist');
  $gnulogscale = 0;
  $nLog = @dates;
  if ($nLog > 0) {
     $out .= sprintf '%-10s  %11s  %11s'."\n",'Date', 'Vert(urad)',  'Horiz(urad)';
     for ($i=0; $i<$nLog; $i++) {$out .=  sprintf '%10s  %11s  %11s'."\n", $dates[$i],$steerV[$i],$steerH[$i];}
     if ($nLog > 1) {
        system ('wmctrl -F -c "'.$gnutitle.'" &');
        $gnuopen = &open_gnuplot (undef, undef, $gnulogscale, \@gnuopts);
        if ($gnuopen) {
           print GNUPLOT 'set term x11 title "'.$gnutitle.'"'."\n";
           print GNUPLOT 'set xdata time'."\n";
           print GNUPLOT 'set timefmt "%Y-%m-%d"'."\n";			#must match format of $date
           print GNUPLOT 'set xrange ["'.$start.'":"'.$last.'"]'."\n";
           print GNUPLOT 'set format y "%5.1f"'."\n";
           print GNUPLOT 'set xlabel "Date"'."\n";
           print GNUPLOT 'set ylabel "Steering (urad)"'."\n";
           ### See https://stackoverflow.com/questions/4585770/gnuplot-stdin-how-to-plot-two-lines
           ### (I managed to work around this by sending the data twice, terminated by "e" on it's own line after each block)
           print GNUPLOT 'plot "-" using 1:2 title "'.$hw[0].'" with linespoints pt 7 ps 2,'
                             .' "" using 1:3 title "'.$hw[1].'" with linespoints pt 5 ps 2'."\n";
           for ($j=0; $j<2; $j++) {
              for ($i=0; $i<$nLog; $i++) {printf GNUPLOT '%10s  %11s  %11s'."\n", $dates[$i],$steerV[$i],$steerH[$i];}
              print GNUPLOT 'e'."\n";
           }
           &close_gnuplot();
        }
     }
  } else {
    $out .= ' No data for the last '.$history.' days'."\n";
  }
# print STDOUT $out;
  if (Exists($steerHistWin)) {$steerHistWin->destroy();}

  if ($use_displayer) {system('displayer.pl "'.$out.'" "'.$gnutitle.'" &'); return;}

  $steerHistWin = $mw->DialogBox( -title=>$title,
                                  -buttons=>['OK'],
                                  -default_button=>'OK',
                                  -takefocus=>1,
                                  );
  $rt = $steerHistWin->ScrlROText(-width=>38,
                                  -height=>40,
#                                 -scrollbars=>'se',
                                  -scrollbars=>'e',
                                  -font=>$fontTbl,
                                  -wrap=>'none',
                                  )->pack();
# $rt->Subwidget("xscrollbar")->configure(-background=>$bgcolor,);
  $rt->Subwidget("yscrollbar")->configure(-background=>$bgcolor,);
  $rt->insert(end => $out);

# $steerHistWin = $mw->Dialog(    -title=>$script,
#                                 -text=>$out,
#                                 -wrap=>0,
#                                 -font=>$fontBtn,
#                                 -buttons=>['OK'],
#                                 -default_button=>'OK',
#                                 -takefocus=>1,
#                                 -width=>32,
#                                 );
  $steerHistWin->Subwidget('B_OK')->configure(-font=>$fontBtn,
                                              -background=>$bgcolor, -activebackground=>$abgcolor, 
                                              -foreground=>$fgcolor, -activeforeground=>$afgcolor,
                                              );
  $answer = $steerHistWin->Show;
  if ($answer eq 'OK') {
     $steerHistWin->destroy();
     system ('wmctrl -F -c "'.$gnutitle.'" &');
  }
}

#=================================================================

sub I0_history() {
  our ($I0HistWin);
  my ($script, $str, @logs, $log, @hw, $type, @lines, @words);
  my ($date, %intensity, $out, @dates, @DCM, @I0, $nLog, $i, $j, $line);
  my (@gnuopts, $gnuopen, $gnutitle, $gnulogscale, $title, $answer, $rt);
  my ($start, $last);
  $script = 'I0_history';
  @hw = ('DCM Vsum (V)', 'I0 intensity (V)');
  $str = 'find '.$logdir.' -mtime -'.$history.' -name "'.$logroot.'_'.$beamline.'_*.log"';
  @logs = split(/\r?\n/,`$str`);
  @logs = sort @logs;
  @dates = ();
  @DCM   = ();
  @I0    = ();
LOG: 
  foreach $log (@logs) {
     if ($debug) {print STDOUT $script.'(DEBUG): logfile='.$log."\n";}
     $date = $log;
     $date =~ s/^.*user_setup_${beamline}_//;
     $date = substr($date, 0, 10);
     $date =~ s/[\/\._]/-/g;
     foreach $type (@hw) {
        $intensity{$type} = undef;
        $str = 'grep "^\s*'.$type.'" '.$log;
        if ($debug) {print STDOUT $script.'(DEBUG): str='.$str."\n";}
        @lines = split(/\r?\n/,`$str`);
        if ($debug) {foreach $line (@lines) {print STDOUT $script.'(DEBUG): line='.$line."\n";}}
        if (@lines < 1)             {next LOG;}
        $i = $#lines;
        $lines[$i] =~ s/^\s*\Q${type}\E\s*//;
        @words = split(/\s+/,$lines[$i]);
        if (@words != 2)            {next LOG;}
        if (! &is_float($words[0])) {next LOG;}
        $intensity{$type} = $words[0];
     }
     push(@dates,$date);
     push(@DCM,$intensity{$hw[0]});
     push(@I0, $intensity{$hw[1]});
  }
  ($start,$last) = &calc_date_range($history,undef,'-');
  $title    = 'DCM-VSUM and I0 HISTORY';
  $out      = $title."\n\n";
  $gnutitle = $title.' Plot';
  @gnuopts = ('-persist');
  $gnulogscale = 0;
  $nLog = @dates;
  if ($nLog > 0) {
     $out .= sprintf '%-10s  %16s  %16s'."\n", 'Date', $hw[0], $hw[1];
     for ($i=0; $i<$nLog; $i++) {$out .=  sprintf '%10s  %16s  %16s'."\n", $dates[$i],$DCM[$i],$I0[$i];}
     if ($nLog > 1) {
        system ('wmctrl -F -c "'.$gnutitle.'" &');
        $gnuopen = &open_gnuplot (undef, undef, $gnulogscale, \@gnuopts);
        if ($gnuopen) {
           print GNUPLOT 'set term x11 title "'.$gnutitle.'"'."\n";
           print GNUPLOT 'set xdata time'."\n";
           print GNUPLOT 'set timefmt "%Y-%m-%d"'."\n";			#must match format of $date
           print GNUPLOT 'set xrange ["'.$start.'":"'.$last.'"]'."\n";
           print GNUPLOT 'set format y "%6.2f"'."\n";
           print GNUPLOT 'set xlabel "Date"'."\n";
           print GNUPLOT 'set ylabel "Intensity (V)"'."\n";
           ### See https://stackoverflow.com/questions/4585770/gnuplot-stdin-how-to-plot-two-lines
           ### (I managed to work around this by sending the data twice, terminated by "e" on it's own line after each block)
           print GNUPLOT 'plot "-" using 1:2 title "'.$hw[0].'" with linespoints pt 7 ps 2,'
                             .' "" using 1:3 title "'.$hw[1].'" with linespoints pt 5 ps 2'."\n";
           for ($j=0; $j<2; $j++) {
              for ($i=0; $i<$nLog; $i++) {printf GNUPLOT '%10s  %16s   %16s'."\n", $dates[$i],$DCM[$i],$I0[$i];}
              print GNUPLOT 'e'."\n";
           }
           &close_gnuplot();
        }
     }
  } else {
    $out .= ' No data for the last '.$history.' days'."\n";
  }
# print STDOUT $out;
  if (Exists($I0HistWin)) {$I0HistWin->destroy();}

  if ($use_displayer) {system('displayer.pl "'.$out.'" "'.$gnutitle.'" &'); return;}

  $I0HistWin = $mw->DialogBox(    -title=>$title,
                                  -buttons=>['OK'],
                                  -default_button=>'OK',
                                  -takefocus=>1,
                                  );
  $rt = $I0HistWin->ScrlROText(   -width=>48,
                                  -height=>40,
#                                 -scrollbars=>'se',
                                  -scrollbars=>'e',
                                  -font=>$fontTbl,
                                  -wrap=>'none',
                                  )->pack();
# $rt->Subwidget("xscrollbar")->configure(-background=>$bgcolor,);
  $rt->Subwidget("yscrollbar")->configure(-background=>$bgcolor,);
  $rt->insert(end => $out);

# $I0HistWin = $mw->Dialog(       -title=>$title,
#                                 -text=>$out,
#                                 -wrap=>0,
#                                 -font=>$fontBtn,
#                                 -buttons=>['OK'],
#                                 -default_button=>'OK',
#                                 -takefocus=>1,
#                                 );
  $I0HistWin->Subwidget('B_OK')->configure(-font=>$fontBtn,
                                           -background=>$bgcolor, -activebackground=>$abgcolor, 
                                           -foreground=>$fgcolor, -activeforeground=>$afgcolor,
                                           );
  $answer = $I0HistWin->Show;
  if ($answer eq 'OK') {
     $I0HistWin->destroy();
     system ('wmctrl -F -c "'.$gnutitle.'" &');
  }
}

#=================================================================

sub minibeam_intensities_history() {
  our ($minibHistWin);
  our (@minibeam_300, @minibeam_20, @minibeam_10, @minibeam_5);
  my ($script, $str, @logs, $log, @hw, @lines, @words);
  my ($date, @counts, $out, @dates, $nLog, $i, $j, $line);
  my (@gnuopts, $gnuopen, $gnutitle, $gnulogscale, $title, $answer, $rt);
  my ($start, $last);
  $script = 'minibeam_intensitie_history';
  @hw = qw(300um 20um 10um 5um);
  $str = 'find '.$logdir.' -mtime -'.$history.' -name "'.$logroot.'_'.$beamline.'_*.log"';
  @logs = split(/\r?\n/,`$str`);
  @logs = sort @logs;
  $nLog = 0;
  @dates = ();
  @counts = ();
LOG: 
  foreach $log (@logs) {
     if ($debug) {print STDOUT $script.'(DEBUG): logfile='.$log."\n";}
     $date = $log;
     $date =~ s/^.*user_setup_${beamline}_//;
     $date = substr($date, 0, 10);
     $date =~ s/[\/\._]/-/g;
     $str = 'grep "^\s*Intensity (V)" '.$log;
     if ($debug) {print STDOUT $script.'(DEBUG): str='.$str."\n";}
     @lines = split(/\r?\n/,`$str`);
     if ($debug) {foreach $line (@lines) {print STDOUT $script.'(DEBUG): line='.$line."\n";}}
     if (@lines < 1)                 {next LOG;}
     $i = $#lines;
     $lines[$i] =~ s/^\s*\QIntensity (V)\E\s*//;
     @words = split(/\s+/,$lines[$i]);
     if (@words != 4)                {next LOG;}
     for ($i=0; $i<4; $i++) {
        if (! &is_float($words[$i])) {next LOG;}
     }
     $dates[$nLog] = $date;
     for ($i=0; $i<4; $i++) {
        $counts[$nLog][$i] = $words[$i];
     }
     $nLog++;
  }
  ($start,$last) = &calc_date_range($history,undef,'-');		#must match format of $date
  $title    = 'MINIBEAM INTENSITIES HISTORY';
  $out      = $title."\n\n";
  $gnutitle = $title.' Plot';
  @gnuopts = ('-persist');
  $gnulogscale = 0;
# $nLog = @dates;
  if ($nLog > 0) {
     $out .= sprintf '%-10s  %10s  %10s  %10s  %10s'."\n",         'Date',$hw[0],$hw[1],$hw[2],$hw[3];
     $out .= sprintf '%-10s  %10.0e  %10.0e  %10.0e  %10.0e'."\n", 'Gains:',$minibeam_300[1],
                                                                            $minibeam_20[1],
                                                                            $minibeam_10[1],
                                                                            $minibeam_5[1];
     for ($i=0; $i<$nLog; $i++) {
        $out .= sprintf '%10s  %10s  %10s  %10s  %10s'."\n", $dates[$i],
                                                             $counts[$i][0],
                                                             $counts[$i][1],
                                                             $counts[$i][2],
                                                             $counts[$i][3];
     }
     if ($nLog > 1) {
        system ('wmctrl -F -c "'.$gnutitle.'" &');
        $gnuopen = &open_gnuplot (undef, undef, $gnulogscale, \@gnuopts);
        if ($gnuopen) {
           print GNUPLOT 'set term x11 title "'.$gnutitle.'"'."\n";
           print GNUPLOT 'set xdata time'."\n";
           print GNUPLOT 'set timefmt "%Y-%m-%d"'."\n";			#must match format of $date
           print GNUPLOT 'set xrange ["'.$start.'":"'.$last.'"]'."\n";
           print GNUPLOT 'set format y "%6.2f"'."\n";
           print GNUPLOT 'set format y2 "%6.2f"'."\n";
           print GNUPLOT 'set y2tics'."\n";
           print GNUPLOT 'set xlabel "Date"'."\n";
           print GNUPLOT 'set ylabel "Intensity '.$hw[0].','.$hw[1].','.$hw[2].'"'."\n";
           print GNUPLOT 'set y2label "Intensity '.$hw[3].'"'."\n";

           ### See https://stackoverflow.com/questions/4585770/gnuplot-stdin-how-to-plot-two-lines
           ### (I managed to work around this by sending the data twice, terminated by "e" on it's own line after each block)
           print GNUPLOT 'plot "-" using 1:2 axes x1y1 title "'.$hw[0].'" with linespoints pt 7 ps 2,'
                             .' "" using 1:3 axes x1y1 title "'.$hw[1].'" with linespoints pt 5 ps 2,'
                             .' "" using 1:4 axes x1y1 title "'.$hw[2].'" with linespoints pt 9 ps 2,'
                             .' "" using 1:5 axes x1y2 title "'.$hw[3].'" with linespoints pt 11 ps 2'."\n";
           for ($j=0; $j<4; $j++) {
              for ($i=0; $i<$nLog; $i++) {
                 printf GNUPLOT '%10s  %10s  %10s  %10s  %10s'."\n", $dates[$i],
                                                                     $counts[$i][0],
                                                                     $counts[$i][1],
                                                                     $counts[$i][2],
                                                                     $counts[$i][3]; 
              }
              print GNUPLOT 'e'."\n";
           }
           &close_gnuplot();
        }
     }
  } else {
    $out .= ' No data for the last '.$history.' days'."\n";
  }
# print STDOUT $out;
  if (Exists($minibHistWin)) {$minibHistWin->destroy();}

  if ($use_displayer) {system('displayer.pl "'.$out.'" "'.$gnutitle.'" &'); return;}

  $minibHistWin = $mw->DialogBox( -title=>$title,
                                  -buttons=>['OK'],
                                  -default_button=>'OK',
                                  -takefocus=>1,
                                  );
  $rt = $minibHistWin->ScrlROText(-width=>60,
                                  -height=>40,
#                                 -scrollbars=>'se',
                                  -scrollbars=>'e',
                                  -font=>$fontTbl,
                                  -wrap=>'none',
                                  )->pack();
# $rt->Subwidget("xscrollbar")->configure(-background=>$bgcolor,);
  $rt->Subwidget("yscrollbar")->configure(-background=>$bgcolor,);
  $rt->insert(end => $out);

# $minibHistWin = $mw->Dialog(    -title=>$title,
#                                 -text=>$out,
#                                 -wrap=>0,
#                                 -font=>$fontBtn,
#                                 -buttons=>['OK'],
#                                 -default_button=>'OK',
#                                 -takefocus=>1,
#                                 );
  $minibHistWin->Subwidget('B_OK')->configure(-font=>$fontBtn,
                                              -background=>$bgcolor, -activebackground=>$abgcolor, 
                                              -foreground=>$fgcolor, -activeforeground=>$afgcolor,
                                              );
  $answer = $minibHistWin->Show;
  if ($answer eq 'OK') {
     $minibHistWin->destroy();
     system ('wmctrl -F -c "'.$gnutitle.'" &');
  }
}

#=================================================================

sub beam_sizes_history() {
  our ($bsHistWin);
  my ($script, $str, @logs, $log, @hw, $type, @lines);
  my ($date, %size, $out, @dates, @bsV, @bsH, $nLog, $i, $j, $line);
  my (@gnuopts, $gnuopen, $gnutitle, $gnulogscale, $title, $answer, $rt);
  my ($start, $last);
  $script = 'beam_sizes_history';
  @hw = ('Vert', 'Horz');
  $str = 'find '.$logdir.' -mtime -'.$history.' -name "'.$logroot.'_'.$beamline.'_*.log"';
  @logs = split(/\r?\n/,`$str`);
  @logs = sort @logs;
  @dates = ();
  @bsV   = ();
  @bsH   = ();
LOG: 
  foreach $log (@logs) {
     if ($debug) {print STDOUT $script.'(DEBUG): logfile='.$log."\n";}
     $date = $log;
     $date =~ s/^.*user_setup_${beamline}_//;
     $date = substr($date, 0, 10);
     $date =~ s/[\/\._]/-/g;
     foreach $type (@hw) {
        $size{$type} = undef;
        $str = 'grep "^\s*Beam size: '.$type.'=" '.$log;
        if ($debug) {print STDOUT $script.'(DEBUG): str='.$str."\n";}
        @lines = split(/\r?\n/,`$str`);
        if ($debug) {foreach $line (@lines) {print STDOUT $script.'(DEBUG): line='.$line."\n";}}
        if (@lines < 1)              {next LOG;}
        $i = $#lines;
        $lines[$i] =~ s/^\s*Beam size: ${type}=//;
        $lines[$i] =~ s/um.*$//;
        if (! &is_float($lines[$i])) {next LOG;}
        $size{$type} = $lines[$i];
     }
     push(@dates,$date);
     push(@bsV,$size{$hw[0]});
     push(@bsH,$size{$hw[1]});
  }
  ($start,$last) = &calc_date_range($history,undef,'-');		#must match format of $date
  $title    = 'BEAM SIZES HISTORY';
  $out      = $title."\n\n";
  $gnutitle = $title.' Plot';
  @gnuopts = ('-persist');
  $gnulogscale = 0;
  $nLog = @dates;
  if ($nLog > 0) {
     $out .= sprintf '%-10s  %10s   %10s'."\n", 'Date', $hw[0].'(um)', $hw[1].'(um)';
     for ($i=0; $i<$nLog; $i++) {$out .= sprintf '%10s  %10s   %10s'."\n", $dates[$i],$bsV[$i],$bsH[$i];}
     if ($nLog > 1) {
        system ('wmctrl -F -c "'.$gnutitle.'" &');
        $gnuopen = &open_gnuplot (undef, undef, $gnulogscale, \@gnuopts);
        if ($gnuopen) {
           print GNUPLOT 'set term x11 title "'.$gnutitle.'"'."\n";
           print GNUPLOT 'set xdata time'."\n";
           print GNUPLOT 'set timefmt "%Y-%m-%d"'."\n";			#must match format of $date
           print GNUPLOT 'set xrange ["'.$start.'":"'.$last.'"]'."\n";
           print GNUPLOT 'set format y "%6.2f"'."\n";
           print GNUPLOT 'set format y2 "%6.2f"'."\n";
           print GNUPLOT 'set y2tics'."\n";
           print GNUPLOT 'set xlabel "Date"'."\n";
           print GNUPLOT 'set ylabel "'.$hw[0].' size (um)"'."\n";
           print GNUPLOT 'set y2label "'.$hw[1].' size (um)"'."\n";
           ### See https://stackoverflow.com/questions/4585770/gnuplot-stdin-how-to-plot-two-lines
           ### (I managed to work around this by sending the data twice, terminated by "e" on it's own line after each block)
           print GNUPLOT 'plot "-" using 1:2 axes x1y1 title "'.$hw[0].'" with linespoints pt 7 ps 2,'
                             .' "" using 1:3 axes x1y2 title "'.$hw[1].'" with linespoints pt 5 ps 2'."\n";
           for ($j=0; $j<2; $j++) {
              for ($i=0; $i<$nLog; $i++) {printf GNUPLOT '%10s  %10s   %10s'."\n", $dates[$i],$bsV[$i],$bsH[$i];}
              print GNUPLOT 'e'."\n";
           }
           &close_gnuplot();
        }
     }
  } else {
    $out .= ' No data for the last '.$history.' days'."\n";
  }
# print STDOUT $out;
  if (Exists($bsHistWin)) {$bsHistWin->destroy();}

  if ($use_displayer) {system('displayer.pl "'.$out.'" "'.$gnutitle.'" &'); return;}

  $bsHistWin = $mw->DialogBox(    -title=>$title,
                                  -buttons=>['OK'],
                                  -default_button=>'OK',
                                  -takefocus=>1,
                                  );
  $rt = $bsHistWin->ScrlROText(   -width=>38,
                                  -height=>40,
#                                 -scrollbars=>'se',
                                  -scrollbars=>'e',
                                  -font=>$fontTbl,
                                  -wrap=>'none',
                                  )->pack();
# $rt->Subwidget("xscrollbar")->configure(-background=>$bgcolor,);
  $rt->Subwidget("yscrollbar")->configure(-background=>$bgcolor,);
  $rt->insert(end => $out);

# $bsHistWin = $mw->Dialog(       -title=>$title,
#                                 -text=>$out,
#                                 -wrap=>0,
#                                 -font=>$fontBtn,
#                                 -buttons=>['OK'],
#                                 -default_button=>'OK',
#                                 -takefocus=>1,
#                                 );
  $bsHistWin->Subwidget('B_OK')->configure(-font=>$fontBtn,
                                           -background=>$bgcolor, -activebackground=>$abgcolor, 
                                           -foreground=>$fgcolor, -activeforeground=>$afgcolor,
                                           );
  $answer = $bsHistWin->Show;
  if ($answer eq 'OK') {
     $bsHistWin->destroy();
     system ('wmctrl -F -c "'.$gnutitle.'" &');
  }
}

#=================================================================

sub collimator_preset_pos_history() {
  our ($colpsHistWin);
  my ($script, $str, @logs, $log, @hw, @lines, $line, @words);
  my ($date, @pos, @posVH, $vh, $out, @dates, $nLog, $i, $j, $l, $n, $m);
  my (@gnuopts, $gnuopen, $gnutitle, $gnulogscale, $title, $answer, $rt);
  my ($start, $last);
  $script = 'collimator_preset_pos_history';
  @hw = qw(300um 20um 10um 5um);
  $str = 'find '.$logdir.' -mtime -'.$history.' -name "'.$logroot.'_'.$beamline.'_*.log"';
  @logs = split(/\r?\n/,`$str`);
  @logs = sort @logs;
  $nLog = 0;
  @dates = ();
  @posVH = ();
LOG: 
  foreach $log (@logs) {
     if ($debug) {print STDOUT $script.'(DEBUG): logfile='.$log."\n";}
     $date = $log;
     $date =~ s/^.*user_setup_${beamline}_//;
     $date = substr($date, 0, 10);
     $date =~ s/[\/\._]/-/g;
     @pos = ();
     foreach $vh (qw(Vert Horz)) {
        $str = 'grep "^\s*'.$vh.' (mm)" '.$log;
        if ($debug) {print STDOUT $script.'(DEBUG): str='.$str."\n";}
        @lines = split(/\r?\n/,`$str`);
        if ($debug) {foreach $line (@lines) {print STDOUT $script.'(DEBUG): line='.$line."\n";}}
        if (@lines < 1)                 {next LOG;}
        $i = $#lines;
        $lines[$i] =~ s/^\s*${vh}\Q (mm)\E\s*//;
        @words = split(/\s+/,$lines[$i]);
        if (@words != 4)                {next LOG;}
        for ($i=0; $i<4; $i++) {
           if (! &is_float($words[$i])) {next LOG;}
        }
        push(@pos,@words);
     }
     $dates[$nLog] = $date;
     for ($j=0; $j<8; $j++) {
        $posVH[$nLog][$j] = $pos[$j];
     }
     $nLog++;
  }
  ($start,$last) = &calc_date_range($history,undef,'-');		#must match format of $date
  $title    = 'COLLIMATOR PRESET POSITIONS HISTORY';
  $out      = $title."\n\n";
  $gnutitle = $title.' Plot';
  @gnuopts = ('-persist');
  $gnulogscale = 0;
# $nLog = @dates;
  if ($nLog > 0) {
     $out .= sprintf '%-10s  %34s  %34s'."\n", ' ','___________Vertical (mm)__________',
                                                   '__________Horizontal (mm)_________';
     $out .= sprintf '%-10s','Date';
     for ($i=0; $i<2; $i++) {
        for ($j=0; $j<4; $j++) {$out .= sprintf '  %7s',$hw[$j];}
     }
     $out .= "\n";
     for ($i=0; $i<$nLog; $i++) {
        $out .= sprintf '%10s', $dates[$i];
        for ($j=0; $j<8; $j++) {$out .= sprintf '  %7s',$posVH[$i][$j];}
        $out .= "\n";
     }
     $nLog = 0;                                 #SS: disabling GnuPlot because it is overloaded
     if ($nLog > 1) {
        system ('wmctrl -F -c "'.$gnutitle.'" &');
        $gnuopen = &open_gnuplot (undef, undef, $gnulogscale, \@gnuopts);
        if ($gnuopen) {
           print GNUPLOT 'set term x11 title "'.$gnutitle.'"'."\n";
           print GNUPLOT 'set xdata time'."\n";
           print GNUPLOT 'set timefmt "%Y-%m-%d"'."\n";
           print GNUPLOT 'set xrange ["'.$start.'":"'.$last.'"]'."\n";
           print GNUPLOT 'set format y "%7.3f"'."\n";
           print GNUPLOT 'set format y2 "%7.3f"'."\n";
           print GNUPLOT 'set y2tics'."\n";
           print GNUPLOT 'set xlabel "Date"'."\n";
           print GNUPLOT 'set ylabel "Vertical positions (mm)"'."\n";
           print GNUPLOT 'set y2label "Horizontal positions (mm)"'."\n";

           ### See https://stackoverflow.com/questions/4585770/gnuplot-stdin-how-to-plot-two-lines
           ### (I managed to work around this by sending the data twice, terminated by "e" on it's own line after each block)
           $l = '-';
           $n = 2;
           $m = 1;
           print GNUPLOT 'plot ';
           foreach $vh (qw(V H)) {
              for ($j=0; $j<4; $j++) {
                 printf GNUPLOT '"%s" using 1:%d axes x1y%d title "%s-%s" with linespoints ps 2',$l,$n,$m,$hw[$j],$vh;
                 $l = '';
                 if ($j != 3 || $vh ne 'H') {print GNUPLOT ', ';} 
                 $n++;
              }
              $m++;   
           }
           print GNUPLOT "\n";
           for ($n=0; $n<8; $n++) {
              for ($i=0; $i<$nLog; $i++) {
                 printf GNUPLOT '%10s', $dates[$i];
                 for ($j=0; $j<8; $j++) {printf GNUPLOT '  %7s',$posVH[$i][$j];}
                 print GNUPLOT "\n";
              }
              print GNUPLOT 'e'."\n";
           }
           &close_gnuplot();
        }
     }
  } else {
    $out .= ' No data for the last '.$history.' days'."\n";
  }
# print STDOUT $out;
  if (Exists($colpsHistWin)) {$colpsHistWin->destroy();}

  if ($use_displayer) {system('displayer.pl "'.$out.'" "'.$gnutitle.'" &'); return;}

  $colpsHistWin = $mw->DialogBox( -title=>$title,
                                  -buttons=>['OK'],
                                  -default_button=>'OK',
                                  -takefocus=>1,
                                  );
  $rt = $colpsHistWin->ScrlROText(-width=>84,
                                  -height=>40,
#                                 -scrollbars=>'se',
                                  -scrollbars=>'e',
                                  -font=>$fontTbl,
                                  -wrap=>'none',
                                  )->pack();
# $rt->Subwidget("xscrollbar")->configure(-background=>$bgcolor,);
  $rt->Subwidget("yscrollbar")->configure(-background=>$bgcolor,);
  $rt->insert(end => $out);

# $colpsHistWin = $mw->Dialog(    -title=>$title,
#                                 -text=>$out,
#                                 -wrap=>0,
#                                 -font=>$fontBtn,
#                                 -buttons=>['OK'],
#                                 -default_button=>'OK',
#                                 -takefocus=>1,
#                                 );
  $colpsHistWin->Subwidget('B_OK')->configure(-font=>$fontBtn,
                                              -background=>$bgcolor, -activebackground=>$abgcolor, 
                                              -foreground=>$fgcolor, -activeforeground=>$afgcolor,
                                              );
  $answer = $colpsHistWin->Show;
  if ($answer eq 'OK') {
     $colpsHistWin->destroy();
     system ('wmctrl -F -c "'.$gnutitle.'" &');
  }
}

#=================================================================

sub collimator_preset_ang_history() {
  our ($colanHistWin);
  my ($script, $str, @logs, $log, @hw, @lines, $line, @words);
  my ($date, @ang, @angPY, $py, $out, @dates, $nLog, $i, $j, $l, $n, $m);
  my (@gnuopts, $gnuopen, $gnutitle, $gnulogscale, $title, $answer, $rt);
  my ($start, $last);
  $script = 'collimator_preset_ang_history';
  @hw = qw(300um 20um 10um 5um);
  $str = 'find '.$logdir.' -mtime -'.$history.' -name "'.$logroot.'_'.$beamline.'_*.log"';
  @logs = split(/\r?\n/,`$str`);
  @logs = sort @logs;
  $nLog = 0;
  @dates = ();
  @angPY = ();
LOG: 
  foreach $log (@logs) {
     if ($debug) {print STDOUT $script.'(DEBUG): logfile='.$log."\n";}
     $date = $log;
     $date =~ s/^.*user_setup_${beamline}_//;
     $date = substr($date, 0, 10);
     $date =~ s/[\/\._]/-/g;
     @ang = ();
     foreach $py (qw(Pitch Yaw)) {
        $str = 'grep "^\s*'.$py.' (mm)" '.$log;
        if ($debug) {print STDOUT $script.'(DEBUG): str='.$str."\n";}
        @lines = split(/\r?\n/,`$str`);
        if ($debug) {foreach $line (@lines) {print STDOUT $script.'(DEBUG): line='.$line."\n";}}
        if (@lines < 1)                 {next LOG;}
        $i = $#lines;
        $lines[$i] =~ s/^\s*${py}\Q (mm)\E\s*//;
        @words = split(/\s+/,$lines[$i]);
        if (@words != 4)                {next LOG;}
        for ($i=0; $i<4; $i++) {
           if (! &is_float($words[$i])) {next LOG;}
        }
        push(@ang,@words);
     }
     $dates[$nLog] = $date;
     for ($j=0; $j<8; $j++) {
        $angPY[$nLog][$j] = $ang[$j];
     }
     $nLog++;
  }
  ($start,$last) = &calc_date_range($history,undef,'-');		#must match format of $date
  $title    = 'COLLIMATOR PRESET ANGLES HISTORY';
  $out      = $title."\n\n";
  $gnutitle = $title.' Plot';
  @gnuopts = ('-persist');
  $gnulogscale = 0;
# $nLog = @dates;
  if ($nLog > 0) {
     $out .= sprintf '%-10s  %34s  %34s'."\n", ' ','___________Vertical (mm)__________',
                                                   '__________Horizontal (mm)_________';
     $out .= sprintf '%-10s','Date';
     for ($i=0; $i<2; $i++) {
        for ($j=0; $j<4; $j++) {$out .= sprintf '  %7s',$hw[$j];}
     }
     $out .= "\n";
     for ($i=0; $i<$nLog; $i++) {
        $out .= sprintf '%10s', $dates[$i];
        for ($j=0; $j<8; $j++) {$out .= sprintf '  %7s',$angPY[$i][$j];}
        $out .= "\n";
     }
     $nLog = 0;                                 #SS: disabling GnuPlot because it is overloaded
     if ($nLog > 1) {
        system ('wmctrl -F -c "'.$gnutitle.'" &');
        $gnuopen = &open_gnuplot (undef, undef, $gnulogscale, \@gnuopts);
        if ($gnuopen) {
           print GNUPLOT 'set term x11 title "'.$gnutitle.'"'."\n";
           print GNUPLOT 'set xdata time'."\n";
           print GNUPLOT 'set timefmt "%Y-%m-%d"'."\n";			#must match format of $date
           print GNUPLOT 'set xrange ["'.$start.'":"'.$last.'"]'."\n";
           print GNUPLOT 'set format y "%7.3f"'."\n";
           print GNUPLOT 'set format y2 "%7.3f"'."\n";
           print GNUPLOT 'set y2tics'."\n";
           print GNUPLOT 'set xlabel "Date"'."\n";
           print GNUPLOT 'set ylabel "Pitch (mm)"'."\n";
           print GNUPLOT 'set y2label "Yaw (mm)"'."\n";

           ### See https://stackoverflow.com/questions/4585770/gnuplot-stdin-how-to-plot-two-lines
           ### (I managed to work around this by sending the data twice, terminated by "e" on it's own line after each block)
           $l = '-';
           $n = 2;
           $m = 1;
           print GNUPLOT 'plot ';
           foreach $py (qw(Pt Yw)) {
              for ($j=0; $j<4; $j++) {
                 printf GNUPLOT '"%s" using 1:%d axes x1y%d title "%s-%s" with linespoints ps 2',$l,$n,$m,$hw[$j],$py;
                 $l = '';
                 if ($j != 3 || $py ne 'Yw') {print GNUPLOT ', ';} 
                 $n++;
              }
              $m++;   
           }
           print GNUPLOT "\n";
           for ($n=0; $n<8; $n++) {
              for ($i=0; $i<$nLog; $i++) {
                 printf GNUPLOT '%10s', $dates[$i];
                 for ($j=0; $j<8; $j++) {printf GNUPLOT '  %7s',$angPY[$i][$j];}
                 print GNUPLOT "\n";
              }
              print GNUPLOT 'e'."\n";
           }
           &close_gnuplot();
        }
     }
  } else {
    $out .= ' No data for the last '.$history.' days'."\n";
  }
# print STDOUT $out;
  if (Exists($colanHistWin)) {$colanHistWin->destroy();}

  if ($use_displayer) {system('displayer.pl "'.$out.'" "'.$gnutitle.'" &'); return;}

  $colanHistWin = $mw->DialogBox( -title=>$title,
                                  -buttons=>['OK'],
                                  -default_button=>'OK',
                                  -takefocus=>1,
                                  );
  $rt = $colanHistWin->ScrlROText(-width=>84,
                                  -height=>40,
#                                 -scrollbars=>'se',
                                  -scrollbars=>'e',
                                  -font=>$fontTbl,
                                  -wrap=>'none',
                                  )->pack();
# $rt->Subwidget("xscrollbar")->configure(-background=>$bgcolor,);
  $rt->Subwidget("yscrollbar")->configure(-background=>$bgcolor,);
  $rt->insert(end => $out);

# $colanHistWin = $mw->Dialog(    -title=>$title,
#                                 -text=>$out,
#                                 -wrap=>0,
#                                 -font=>$fontBtn,
#                                 -buttons=>['OK'],
#                                 -default_button=>'OK',
#                                 -takefocus=>1,
#                                 );
  $colanHistWin->Subwidget('B_OK')->configure(-font=>$fontBtn,
                                              -background=>$bgcolor, -activebackground=>$abgcolor, 
                                              -foreground=>$fgcolor, -activeforeground=>$afgcolor,
                                              );
  $answer = $colanHistWin->Show;
  if ($answer eq 'OK') {
     $colanHistWin->destroy();
     system ('wmctrl -F -c "'.$gnutitle.'" &');
  }
}

#=================================================================

sub guard_slits_calib_history() {
  our ($gscalHistWin);
  my ($script, $str, @logs, $log, @hw, @lines, $line, $autosavedir);
  my ($date, @tmp, @offset, $slit, $out, @dates, $nLog, $i, $j, $n);
  my (@gnuopts, $gnuopen, $gnutitle, $gnulogscale, $title, $answer, $rt);
  my ($start, $last);
  $script = 'guard_slits_calib_history';
  @hw = qw(GS:ZT GS:ZB GS:YO GS:YI);
  if    ($beamline eq '23i') {$autosavedir = '/mnt/software/_BACKUPS/_BACKUPS_BL1UPPER/AUTOSAVE';}
  elsif ($beamline eq '23o') {$autosavedir = '/mnt/software/_BACKUPS/_BACKUPS_BL2UPPER/AUTOSAVE';}
  elsif ($beamline eq '23b') {$autosavedir = '/mnt/software/_BACKUPS/_BACKUPS_BL3UPPER/AUTOSAVE';}
  else {return;}
  $str = 'find '.$autosavedir.' -mtime -'.$history.' -name "auto_settings_xxx2_*.sav"';
  @logs = split(/\r?\n/,`$str`);
  @logs = sort @logs;
  $nLog = 0;
  @dates = ();
  @offset = ();
LOG: 
  foreach $log (@logs) {
     if ($debug) {print STDOUT $script.'(DEBUG): logfile='.$log."\n";}
     $date = $log;
     $date =~ s/^.*auto_settings_xxx2_//;
     $date = substr($date, 0, 10);
     $date =~ s/[\/\._]/-/g;
     @tmp = ();
     foreach $slit (@hw) {
        $str = 'grep "^'.$beamline.':'.$slit.':Offset" '.$log;
        if ($debug) {print STDOUT $script.'(DEBUG): str='.$str."\n";}
        @lines = split(/\r?\n/,`$str`);
        if ($debug) {foreach $line (@lines) {print STDOUT $script.'(DEBUG): line='.$line."\n";}}
        if (@lines < 1)                {next LOG;}
        $i = $#lines;
        $lines[$i] =~ s/^${beamline}:${slit}:Offset\s*//;
        if (! &is_float($lines[$i]))   {next LOG;}
        push(@tmp,$lines[$i]);
     }
     $dates[$nLog] = $date;
     for ($j=0; $j<4; $j++) {
        $offset[$nLog][$j] = $tmp[$j];
     }
     $nLog++;
  }
  ($start,$last) = &calc_date_range($history,undef,'-');		#must match format of $date
  $title    = 'GUARD SLITS CALIBRATION HISTORY';
  $out      = $title."\n\n";
  $gnutitle = $title.' Plot';
  @gnuopts = ('-persist');
  $gnulogscale = 0;
  if ($nLog > 0) {
     $out .= sprintf '%-10s  %26s  %26s'."\n", ' ','______Vertical Slit_______',
                                                   '_____Horizontal Slit______';
     $out .= sprintf '%-10s','Date';
#    for ($j=0; $j<4; $j++) {$out .= sprintf '  %16s',$beamline.':'.$hw[$j].':Offset';}
     for ($j=0; $j<4; $j++) {$out .= sprintf '  %12s',$hw[$j].':Offset';}
     $out .= "\n";
     for ($i=0; $i<$nLog; $i++) {
        $out .= sprintf '%10s', $dates[$i];
        for ($j=0; $j<4; $j++) {$out .= sprintf '  %12s',$offset[$i][$j];}
        $out .= "\n";
     }
     if ($nLog > 1) {
        system ('wmctrl -F -c "'.$gnutitle.'" &');
        $gnuopen = &open_gnuplot (undef, undef, $gnulogscale, \@gnuopts);
        if ($gnuopen) {
           print GNUPLOT 'set term x11 title "'.$gnutitle.'"'."\n";
           print GNUPLOT 'set xdata time'."\n";
           print GNUPLOT 'set timefmt "%Y-%m-%d"'."\n";			#must match format of $date
           print GNUPLOT 'set xrange ["'.$start.'":"'.$last.'"]'."\n";
           print GNUPLOT 'set format y "%7.3f"'."\n";
           print GNUPLOT 'set format y2 "%7.3f"'."\n";
           print GNUPLOT 'set y2tics'."\n";
           print GNUPLOT 'set xlabel "Date"'."\n";
           print GNUPLOT 'set ylabel "Vertical Slit Offsets (mm)"'."\n";
           print GNUPLOT 'set y2label "Horizontal Slit Offsets (mm)"'."\n";

           ### See https://stackoverflow.com/questions/4585770/gnuplot-stdin-how-to-plot-two-lines
           ### (I managed to work around this by sending the data twice, terminated by "e" on it's own line after each block)
           print GNUPLOT 'plot ';
           printf GNUPLOT '"-" using 1:2 axes x1y1 title "'.$beamline.':'.$hw[0].':Offset" with linespoints pt 7 ps 2, ';
           printf GNUPLOT '""  using 1:3 axes x1y1 title "'.$beamline.':'.$hw[1].':Offset" with linespoints pt 5 ps 2, ';
           printf GNUPLOT '""  using 1:4 axes x1y2 title "'.$beamline.':'.$hw[2].':Offset" with linespoints pt 9 ps 2, ';
           printf GNUPLOT '""  using 1:5 axes x1y2 title "'.$beamline.':'.$hw[3].':Offset" with linespoints pt 11 ps 2'."\n";
           for ($n=0; $n<4; $n++) {
              for ($i=0; $i<$nLog; $i++) {
                 printf GNUPLOT '%10s', $dates[$i];
                 for ($j=0; $j<4; $j++) {printf GNUPLOT '  %10s',$offset[$i][$j];}
                 print GNUPLOT "\n";
              }
              print GNUPLOT 'e'."\n";
           }
           &close_gnuplot();
        }
     }
  } else {
    $out .= ' No data for the last '.$history.' days'."\n";
  }
# print STDOUT $out;
  if (Exists($gscalHistWin)) {$gscalHistWin->destroy();}

  if ($use_displayer) {system('displayer.pl "'.$out.'" "'.$gnutitle.'" &'); return;}

  $gscalHistWin = $mw->DialogBox( -title=>$title,
                                  -buttons=>['OK'],
                                  -default_button=>'OK',
                                  -takefocus=>1,
                                  );
  $rt = $gscalHistWin->ScrlROText(-width=>76,
                                  -height=>40,
#                                 -scrollbars=>'se',
                                  -scrollbars=>'e',
                                  -font=>$fontTbl,
                                  -wrap=>'none',
                                  )->pack();
# $rt->Subwidget("xscrollbar")->configure(-background=>$bgcolor,);
  $rt->Subwidget("yscrollbar")->configure(-background=>$bgcolor,);
  $rt->insert(end => $out);

# $gscalHistWin = $mw->Dialog(    -title=>$title,
#                                 -text=>$out,
#                                 -wrap=>0,
#                                 -font=>$fontBtn,
#                                 -buttons=>['OK'],
#                                 -default_button=>'OK',
#                                 -takefocus=>1,
#                                 );
  $gscalHistWin->Subwidget('B_OK')->configure(-font=>$fontBtn,
                                              -background=>$bgcolor, -activebackground=>$abgcolor, 
                                              -foreground=>$fgcolor, -activeforeground=>$afgcolor,
                                              );
  $answer = $gscalHistWin->Show;
  if ($answer eq 'OK') {
     $gscalHistWin->destroy();
     system ('wmctrl -F -c "'.$gnutitle.'" &');
  }
}

#=================================================================

sub notes_history() {
  our ($notesHistWin);
  my ($script, $str, @logs, $log);
  my ($buf, $out, $notes);
  my ($rt, $answer, $title);
# my ($date);
  $script = 'notes_history';
  $str = 'find '.$logdir.' -mtime -'.$history.' -name "'.$logroot.'_'.$beamline.'_*.log"';
  @logs = split(/\r?\n/,`$str`);
  @logs = sort @logs;
  $notes = '';
LOG: 
  foreach $log (@logs) {
     if ($debug) {print STDOUT $script.'(DEBUG): logfile='.$log."\n";}
#    $date = $log;
#    $date =~ s/^.*user_setup_${beamline}_//;
#    $date = substr($date, 0, 10);
#    $date =~ s/[\/\._]/-/g;
     $str = 'grep "^===> HOST NOTES by " '.$log;
     if ($debug) {print STDOUT $script.'(DEBUG): str='.$str."\n";}
     $str = `$str`;
     if ($debug) {print STDOUT $script.'(DEBUG): line='.$str."\n";}
     if (length($str) > 0) {
        if (&read_file($log,\$buf)) {next LOG;}
        $buf =~ s/.*===\> HOST NOTES by //s;
        if (length($buf) < 4)       {next LOG;}
        $notes .= "\n".'==================== file='.$log
                 ."\n".'==================== host='.$buf;
     }
  }
  $title    = 'HOSTS NOTES HISTORY';
  $out      = $title."\n\n";
  if (length($notes) > 0) {
    $out .= $notes;
  } else {
    $out .= ' No hosts notes for the last '.$history.' days'."\n";
  }
# print STDOUT $out;
  if (Exists($notesHistWin)) {$notesHistWin->destroy();}

  if ($use_displayer) {system('displayer.pl "'.$out.'" &'); return;}

  $notesHistWin = $mw->DialogBox( -title=>$title,
                                  -buttons=>['OK'],
                                  -default_button=>'OK',
                                  -takefocus=>1,
                                  );
  $rt = $notesHistWin->ScrlROText(-width=>80,
                                  -height=>40,
#                                 -scrollbars=>'se',
                                  -scrollbars=>'e',
                                  -font=>$fontTbl,
                                  -wrap=>'none',
                                  )->pack();
# $rt->Subwidget("xscrollbar")->configure(-background=>$bgcolor,);
  $rt->Subwidget("yscrollbar")->configure(-background=>$bgcolor,);
  $rt->insert(end => $out);

# $notesHistWin = $mw->Dialog(    -title=>$title,
#                                 -text=>$out,
#                                 -wrap=>0,
#                                 -font=>$fontBtn,
#                                 -buttons=>['OK'],
#                                 -default_button=>'OK',
#                                 -takefocus=>1,
#                                 );
  $notesHistWin->Subwidget('B_OK')->configure(-font=>$fontBtn,
                                              -background=>$bgcolor, -activebackground=>$abgcolor, 
                                              -foreground=>$fgcolor, -activeforeground=>$afgcolor,
                                             );
  $answer = $notesHistWin->Show;
  if ($answer eq 'OK') {
     $notesHistWin->destroy();
  }
}

#=================================================================

sub reconf_Hist() {
  my ($nHist);
  $nHist = @histList;
  if (! &is_int($historySpin)       || 
      $historySpin < $histList[0]   || 
      $historySpin > $histList[$nHist-1]) {$historySpin = $history;}
  else                                    {$history = $historySpin;}
# print STDOUT "reconf_Hist: history=$history, historySpin=$historySpin\n";
  return 1;
}

#=================================================================

sub fill_mtrList() {
  our (@mtrList);
  my ($script, $str, @logs, $log, @lines, $line, @words, $hist);
  $script = 'fill_mtrList';
  @mtrList = ();
  $hist = $histList[-1];							#last (max) element
### Find all beamline .log files that are newer than $hist days old:
  $str = 'find '.$logdir.' -mtime -'.$hist.' -name "'.$logroot.'_'.$beamline.'_*.log"';
  if ($debug) {print STDOUT $script.': '.$str."\n";}
  @logs = split(/\r?\n/,`$str`);
  @logs = sort  {$b cmp $a} @logs;
  foreach $log (@logs) {
     if ($debug) {print STDOUT $script.': log='.$log."\n";}
### Find records like: 
### 	23o:RB:RA:ActPos                   9.910 (Robot dewar angle)
     $str = 'grep -E "^\s*'.$beamline.':.*:ActPos\s*\-?\.?[0-9]+.*" '.$log;
     if ($debug) {print STDOUT $script.': '.$str."\n";}
     @lines = split(/\r?\n/,`$str`);
     if (@lines > 50) {
        foreach $line (@lines) {
           $line =~ s/^\s*//;
           @words = split(/\s+/,$line);
           if (@words < 2) {next;}
           $words[0] =~ s/:ActPos$//;
           splice @words,1,1;
           push(@mtrList,join(' ',@words));					#EPICS name and description
        }
        if (@mtrList > 0) {return;}						#return after first found log
     }
  }
### If we did not find anything in the main log, then search log2 which 
### contains records from mtrPositionsBackup.pl

### Find all beamline .log2 files that are newer than $hist days old:
  $str = 'find '.$logdir.' -mtime -'.$hist.' -name "'.$logroot.'_'.$beamline.'_*.log2"';
  if ($debug) {print STDOUT $script.': '.$str."\n";}
  @logs = split(/\r?\n/,`$str`);
  @logs = sort  {$b cmp $a} @logs;
  foreach $log (@logs) {
     if ($debug) {print STDOUT $script.': log='.$log."\n";}
### Find records like: 
###Getting ActPos for 23o:WB:H: ...        0.100
     $str = 'grep -E "^Getting ActPos for '.$beamline.':.*$" '.$log;
     if ($debug) {print STDOUT $script.': '.$str."\n";}
     @lines = split(/\r?\n/,`$str`);
     if (@lines > 50) {
        foreach $line (@lines) {
           $line =~ s/^Getting ActPos for //;
           @words = split(/\s+/,$line);
           if (@words < 2) {next;}
           $words[0] =~ s/:$//;
           push(@mtrList,$words[0]);						#EPICS name only
        }
        if (@mtrList > 0) {return;}						#return after first found log
     }
  }
  if (@mtrList == 0) {
     print STDOUT "\n".'!!! '.$script.': No motor positions data for the last '.$hist.' days'."\n\n";
  }
  return;
}

#=================================================================

sub motor_ActPos_history() {
  our ($mtrActHistWin, $mtrSelected);
  my ($script, $str, @logs, $log, @lines, @words, $place);
  my ($date, $mtr, $out, @dates, @actPos, $nLog, $i);
  my (@gnuopts, $gnuopen, $gnutitle, $gnulogscale);
  my ($title, $answer, $rt, $pattern1, $pattern2);
  my ($start, $last);
  $script = 'motor_ActPos_history';
  $str = 'find '.$logdir.' -mtime -'.$history.' -name "'.$logroot.'_'.$beamline.'_*.log"';
  @logs = split(/\r?\n/,`$str`);
  @logs = sort @logs;
  @dates = ();
  @actPos = ();
  $mtr = $mtrSelected;
  $mtr =~ s/ \(.*$//;                           #remove motor name comment
  $pattern1 = '"^\s*'.$mtr.':ActPos\s*\-?\.?[0-9]+.*" ';
  $pattern2 = '"Getting ActPos for '.$mtr.': \.{3}\s*\-?\.?[0-9]+.*" ';
  foreach $log (@logs) {
#    print STDOUT $script.': checking file '.$log."\n";
     $date = $log;
     $date =~ s/^.*user_setup_${beamline}_//;
     $date = substr($date, 0, 10);
     $date =~ s/[\/\._]/-/g;
     $str = 'grep -Eo '.$pattern1.$log;						#old format (.log)
     @lines = split(/\r?\n/,`$str`);
     $place = 2;
     if (@lines < 1) {								#check new format
        $log .= '2';
        if (-e $log) {								#if exists .log2
#          print STDOUT $script.': checking file '.$log."\n";
           $str = 'grep -Eo '.$pattern2.$log;					#new format (.log2)
           @lines = split(/\r?\n/,`$str`);
           $place = 6;
        }
     }
     if (@lines < 1) {next;}
     $i = $#lines;
     $lines[$i] =~ s/^\s*//;							#use last line in case of duplicates
     @words = split(/\s+/,$lines[$i]);
     if    ($log =~ /\.log$/  && @words < 2) {next;}
     elsif ($log =~ /\.log2$/ && @words < 6) {next;}
     if (! &is_float($words[$place-1]))      {next;}
     push(@dates,$date);
     push(@actPos,$words[$place-1]);
  }
  ($start,$last) = &calc_date_range($history,undef,'-');		#must match format of $date
  $title    = $mtr.' ActPos HISTORY';
  $out      = $title."\n\n";
  $gnutitle = $title.' Plot';
  @gnuopts = ('-persist');
  $gnulogscale = 0;
  $nLog = @dates;
  if ($nLog > 0) {
     $out .= sprintf '%-10s  %s'."\n", 'Date', $mtrSelected;
     for ($i=0; $i<$nLog; $i++) {$out .=  sprintf '%10s  %16s'."\n", $dates[$i],$actPos[$i];}
     if ($nLog > 1) {
        system ('wmctrl -F -c "'.$gnutitle.'" &');
        $gnuopen = &open_gnuplot (undef, undef, $gnulogscale, \@gnuopts);
        if ($gnuopen) {
           print GNUPLOT 'set term x11 title "'.$gnutitle.'"'."\n";
           print GNUPLOT 'set xdata time'."\n";
           print GNUPLOT 'set timefmt "%Y-%m-%d"'."\n";			#must match format of $date
           print GNUPLOT 'set xrange ["'.$start.'":"'.$last.'"]'."\n";
           print GNUPLOT 'set format y "%8.3f"'."\n";
           print GNUPLOT 'set xlabel "Date"'."\n";
           print GNUPLOT 'set ylabel "ActPos"'."\n";
           ### See https://stackoverflow.com/questions/4585770/gnuplot-stdin-how-to-plot-two-lines
           ### (I managed to work around this by sending the data twice, terminated by "e" on it's own line after each block)
           print GNUPLOT 'plot "-" using 1:2 title "'.$mtr.'" with linespoints pt 7 ps 2'."\n";
           for ($i=0; $i<$nLog; $i++) {printf GNUPLOT '%10s  %16s'."\n", $dates[$i],$actPos[$i];}
           print GNUPLOT 'e'."\n";
           &close_gnuplot();
        }
     }
  } else {
    $out .= ' No data for '.$mtr.' within the last '.$history.' days'."\n";
  }
# print STDOUT $out;
  if (Exists($mtrActHistWin)) {$mtrActHistWin->destroy();}

  if ($use_displayer) {system('displayer.pl "'.$out.'" "'.$gnutitle.'" &'); return;}

  $mtrActHistWin = $mw->DialogBox(-title=>$title,
                                  -buttons=>['OK'],
                                  -default_button=>'OK',
                                  -takefocus=>1,
                                  );
  $rt = $mtrActHistWin->ScrlROText(-width=>42,
                                  -height=>40,
                                  -scrollbars=>'se',
#                                 -scrollbars=>'e',
                                  -font=>$fontTbl,
                                  -wrap=>'none',
                                  )->pack();
  $rt->Subwidget("xscrollbar")->configure(-background=>$bgcolor,);
  $rt->Subwidget("yscrollbar")->configure(-background=>$bgcolor,);
  $rt->insert(end => $out);

# $mtrActHistWin = $mw->Dialog(   -title=>$title,
#                                 -text=>$out,
#                                 -wrap=>0,
#                                 -font=>$fontBtn,
#                                 -buttons=>['OK'],
#                                 -default_button=>'OK',
#                                 -takefocus=>1,
#                                 );
  $mtrActHistWin->Subwidget('B_OK')->configure(-font=>$fontBtn,
                                               -background=>$bgcolor, -activebackground=>$abgcolor, 
                                               -foreground=>$fgcolor, -activeforeground=>$afgcolor,
                                               );
  $answer = $mtrActHistWin->Show;
  if ($answer eq 'OK') {
     $mtrActHistWin->destroy();
     system ('wmctrl -F -c "'.$gnutitle.'" &');
  }
}

#=================================================================

#sub getPV($pv2get,$err_message,$value_type,[$value1,$value2]) {
sub getPV ($$$;$$) {			#returns $err

  ### Declare local variables

  my ($pv2get, $value_type, $err_message, $value1, $value2, $err, $tag);

  #print STDOUT "@_\n";

  $pv2get       = shift (@_);    if ($pv2get !~ /^${beamline}:/) {$pv2get = $beamline.':'.$pv2get;}
  $err_message  = shift (@_);                   #can be used as $tag in mode=0
  $value_type   = shift (@_);
  $value1       = shift (@_);                   #used in some modes only
  $value2       = shift (@_);                   #used in some modes only

  ### @desc_name = (pv_name, value_type, error message, values...)
  ###
  ### value_type - description (values)
  ###
  ###	 0	- no expectation (0,0) - USE ONLY WITH MOTORS!
  ###	 1	- single integer value   (x,0)
  ###	 2	- single value with tolerance (x,abs_tol) - USE ONLY WITH MOTORS!
  ###	 3	- range of values (lo,hi)
  ###    4      - mirror voltages (no expectations)
  ### Currently unused:
  ###	 5	- unwanted integer value (x,0)
  ###	 6	- multiple integer values (will try to avoid this)
  ### -----------------------------------------------------------
  ### RETURNS:
  ###    0 = OK
  ###    1 = ERROR
  ###    2 = Mirror is turned off (voltage < 1V) in mode=4
  &refresh_gui();
  $err = 0;
  if ($value_type == 0) {       #---------------------------------------0=no expectation (0,0) - USE ONLY WITH MOTORS!
     ($ca_error,$pv_value) = Pezca::GetDouble($pv2get);
     &report_only($script,$ca_error,$pv2get);
     if ($ca_error) {
        $pv_value = 999999;
        $err_log = 'CA error: could not read '.$pv2get;
        $err = 1;
     }
     else {
        $err_log = '';
        $err = 0;
     }
  }
  elsif ($value_type == 1) {    #---------------------------------------1=single integer value   (x,0)
     ($ca_error,$pv_value) = Pezca::GetDouble($pv2get);
     #print STDOUT $pv_value."\n";
     if ($ca_error) {
        &report_only($script,$ca_error,$pv2get);
        $pv_value = 999999;
        $err_log = 'CA error: could not read '.$pv2get;
        $err = 1;
     }
     elsif ($pv_value != $value1) {
        $err_log = $err_message;
        $err = 1;
     }
     else {
        $err_log = '';
        $err = 0;
     }
  }
  elsif ($value_type == 2) {    #---------------------------------------2=single value with tolerance (x,abs_tol) - USE ONLY WITH MOTORS!
     $lo_ok = $value1 - $value2;
     $hi_ok = $value1 + $value2;
     ($ca_error,$pv_value) = Pezca::GetDouble($pv2get);
     if ($ca_error) {
        &report_only($script,$ca_error,$pv2get);
        $pv_value = 999999;
        $err_log = 'CA error: could not read '.$pv2get;
        $err = 1;
     }
     elsif (($pv_value > $hi_ok) || ($pv_value < $lo_ok)) {
        $err_log = $err_message;
        $err = 1;
     }
     else {
        $err_log = '';
        $err = 0;
     }
  }
  elsif ($value_type == 3) {    #---------------------------------------3=range of values (lo,hi)
     ($ca_error,$pv_value) = Pezca::GetDouble($pv2get);
     if ($ca_error) {
        &report_only($script,$ca_error,$pv2get);
        $pv_value = 999999;
        $err_log = 'CA error: could not read '.$pv2get;
        $err = 1;
     }
     elsif (($pv_value > $value2) || ($pv_value < $value1)) {
        $err_log = $err_message;
        $err = 1;
     }
     else {
        $err_log = '';
        $err = 0;
     }
  }
  elsif ($value_type == 4) {    #---------------------------------------4=mirror voltages
     ($ca_error,$pv_value) = Pezca::GetDouble($pv2get);
     if ($ca_error) {
        &report_only($script,$ca_error,$pv2get);
        $pv_value = 999999;
        $err_log = 'CA error: could not read '.$pv2get;
        $err = 1;
     }
     elsif (abs($pv_value) < 1) {
        $err = 2;
     }
     else {
        $err_log = '';
        $err = 0;
     }
  }
  else {        #-------------------------------------------------------not implemented
     $pv_value = 999999;
     $err_log = 'Error: '.$pv2get.' is of an unknown type='.$value_type;
     $err = 1;
  }

  if (($value_type == 0) || ($value_type == 2)) {#---------------------------------------0=no expectations, 2=single value with tolerance 
     $part2_message .= "\t";
     $part2_message .= sprintf '%-30.30s',$pv2get;
     if ($value_type == 2 || !defined $err_message || length($err_message) == 0) {
        ($ca_error,$tag) = Pezca::GetString($pv2get.'.DESC');		#we ignore this error
        if ($ca_error) {$tag = '';}
     } else {
        $tag = $err_message;
     }
     if (length($tag) == 0) {$tag = 'no descr.';}
     $part2_message .= sprintf '%10.3f (%s)'."\n",$pv_value,$tag;
     #&part2($part2_message);
  }
  elsif ($value_type == 4) {                     #----------------------------------------4=mirror voltages
     $part2_message .= "\t";
     $part2_message .= sprintf '%-30.30s',$pv2get;
     $part2_message .= sprintf '%6.1f'."\n",$pv_value;
     #&part2($part2_message);
  }
  else {                                         #----------------------------------------1=single integer value, 3=range of values
     $part2_message .= "\t";
     $part2_message .= sprintf '%-30.30s',$pv2get;
     $part2_message .= $pv_value."\n";
     #&part2($part2_message);
  }

  if ($err == 1) {
     $step_err = 1;
     $part3_message .= "\t";
     $part3_message .= sprintf '%-30.30s',$pv2get;
     $part3_message .= $err_log."\n";
     #&part3($part3_message);
  }
  &refresh_gui();
  return $err;
}

#=================================================================

#sub part1 ($statement) {
sub part1 ($) {

  ### Part one is printed to the screen
  ### Format:
  ###
  ### TASK                  ERR (or OK)
  ###    - Give indented error summary (if not 'OK')

  my ($statement);

  $statement = shift (@_);
  $part1_text .= $statement."\n";
  $part1_print .= $statement."\n";

}

#=================================================================

#sub part2 ($statement) {
sub part2 ($) {

  ### Part two is for the printout
  ### It is the detailed part

  my ($statement);

  $statement = shift (@_);
  $part2_text .= $statement;
  $minibeam_messages .= $statement;

}

#=================================================================

#sub part3 ($statement) {
sub part3 ($) {

  ### Part three is for the printout
  ### It provides a quick error summary for the hosts

  my ($statement);

  $statement = shift (@_);
  $part3_text .= $statement;

}

#=================================================================

# sub email ($header,$msg,$email_list) {
sub email ($$$) {

  my ($header, $msg, $email_list, $mail);

  $header     = shift(@_);
  $msg        = shift(@_);
  $email_list = shift(@_);

  &refresh_gui();
  $mail = MIME::Lite->new(
    From     => 'sstepanov@anl.gov',
    To       => $email_list,
    Subject  => $script.' at '.$beamline.': '.$header.' monitoring ALERT!',
    Type     => 'multipart/mixed',
  );
  $mail->attach(
    Type     => 'TEXT', 			#case-sensitive TEXT/BINARY/AUTO
    Encoding => 'quoted-printable',
    Data     => $msg."\n",
  );
  $mail->send();
  print STDOUT 'Mail sent to '.$email_list.' .'."\n";
  &refresh_gui();
}

#==========================================================================

# sub read_file($file,\$buf) {
sub read_file($\$) {
# use Fcntl;					                #not used (defines O_RDONLY constant (O_RDONLY=0,O_WRONLY=1,O_RDWR=2))
  my ($script, $file, $bufref, $status);

  $script = 'read_file';
  $file   = shift(@_);
  $bufref = shift(@_);

  $$bufref = '';
  $status = sysopen(FH, $file, 0);		                #0 is used instead of O_RDONLY defined in "use Fcntl";
  if (!defined $status) {print STDOUT '*** '.$script.': Could not open '.$file."\n"; return 1;}
  sysread(FH, $$bufref, -s FH);
  close(FH);
  return 0;
}

#==========================================================================

sub calc_date_range ($;$$) {						#($ndays,[$last_data],$separator)
  my ($ndays, $start, $last, $separator);
  my (@ct, $epoch);
  $ndays     = shift(@_);
  $last      = shift(@_);
  $separator = shift(@_);						#"-", ".", "/"
  if (!defined $separator) {$separator='-';}
  if (!defined $last) {							#use today
     @ct = localtime();						# This is ($sec,$min,$hour,$mday,$mon,$year$,$wday,$yday,$isdst)
     ($mday,$mon,$year) = ($ct[3],$ct[4]+1,$ct[5]+1900);
     $last = sprintf '%04s'.$separator.'%02s'.$separator.'%02s', $year, $mon, $mday;
  }
  $epoch = &timestamp2epoch($last.' 00:00');
  $epoch = $epoch - $ndays*24*3600;
  @ct = localtime($epoch);
  ($mday,$mon,$year) = ($ct[3],$ct[4]+1,$ct[5]+1900);
  $start = sprintf '%04s'.$separator.'%02s'.$separator.'%02s', $year, $mon, $mday;
  return ($start,$last);
}

#==========================================================================

sub signal_handler ($;$) {				                        #($signal,$exitcode)
  return;
}

# ---------------------------------------------------------------

### To-Do:
###
### Begin at line 831 - make motions in 'minibeam intensity' non-interactive in external window
### Auto-print
